#include "AD7768_Mb.h"
#include <Poco/Format.h>
#include <algorithm>
#include <FloatingPtSupport_Mb.h>
#include <SystemSupport_Mb.h>

namespace Innovative {
namespace ad7768 {

//-----------------------------------------------------------------------------
//  Helpers  --
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//  AD7768_ADC::AD7768_ADC()  --
//-----------------------------------------------------------------------------

AD7768_ADC::AD7768_ADC(IRequires_AD7768_ADC_Io *io, unsigned int devId) :
    _io(io), _devId(devId)
{
    _sleepMode = SleepMode_t::ACTIVE;
    _powerMode = PowerMode_t::ECO;
    _mclkDiv   = MclkDiv_t::MCLK_DIV_32;
    _dclkDiv   = DclkDiv_t::DCLK_DIV_8;
    _convOp    = ConvOp_t::STANDARD_CONV;
    _crcSel    = CrcSel_t::NO_CRC;
    // Varibles initialized to IC reseted defaults
    fill(std::begin(_chState), std::end(_chState), ChState_t::ENABLED);
    fill(std::begin(_chMode), std::end(_chMode), ChMode_t::MODE_A);
    fill(std::begin(_filtType), std::end(_filtType), Filt_t::FILTER_SINC);
    fill(std::begin(_decRate), std::end(_decRate), DecRate_t::DEC_X1024);

    _diag = 0;
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::~AD7768_ADC()  --
//-----------------------------------------------------------------------------

AD7768_ADC::~AD7768_ADC()
{
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::SetPowerMode()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::SetPowerMode(PowerMode_t mode)
{
    _powerMode = mode;
    SpiField(REG_PWR_MODE, 4, 2, to_number(mode));
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::SetSleepMode()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::SetSleepMode(SleepMode_t mode)
{
    _sleepMode = mode;
    SpiBit(REG_PWR_MODE, 7, to_number(mode));
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::SetMClkDivMode()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::SetMClkDivMode(MclkDiv_t mode)
{
    _mclkDiv = mode;
    SpiField(REG_PWR_MODE, 0, 2, to_number(mode));
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::SetDClkDivMode()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::SetDClkDivMode(DclkDiv_t mode)
{
    _dclkDiv = mode;
    SpiField(REG_INTERFACE_CFG, 0, 2, to_number(mode));
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::SetConversionMode()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::SetConversionMode(ConvOp_t mode)
{
    _convOp = mode;
    SpiBit(REG_DATA_CTRL, 4, to_number(mode));
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::SetCrcMode()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::SetCrcMode(CrcSel_t mode)
{
    _crcSel = mode;
    SpiField(REG_INTERFACE_CFG, 2, 2, to_number(mode));
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::SetChannelStandby()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::SetChannelStandby(unsigned int ch, ChState_t mode)
{
    _chState[ch] = mode;
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::SetChannelMode()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::SetChannelMode(unsigned int ch, ChMode_t mode)
{
    _chMode[ch] = mode;
    SpiBit(REG_CH_MODE_SEL, ch, to_number(mode));
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::SetChannelMode()  -- Sets all channels to the given mode
//-----------------------------------------------------------------------------

void AD7768_ADC::SetChannelMode(ChMode_t mode)
{
    for (int ch : {0, 1, 2, 3, 4, 5, 6, 7})
         _chMode[ch] = mode;

    char value = (mode == ChMode_t::MODE_A) ? 0x0 : 0xff;
    Spi(REG_CH_MODE_SEL, value);
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::SetModeA()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::SetModeA(Filt_t fi, DecRate_t dec)
{
    _filtType[0] = fi;
    _decRate[0]  = dec;

    MemoryBitManipulator Reg(0);
    Reg.Field(0, 3, to_number(dec));
    Reg.Bit(3, to_number(fi));

    Spi(REG_CH_MODE_A, Reg.Value());
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::SetModeB()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::SetModeB(Filt_t fi, DecRate_t dec)
{
    _filtType[1] = fi;
    _decRate[1]  = dec;

    MemoryBitManipulator Reg(0);
    Reg.Field(0, 3, to_number(dec));
    Reg.Bit(3, to_number(fi));

    Spi(REG_CH_MODE_B, Reg.Value());
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::SetDiagnosticMode()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::SetDiagnosticMode(DiagMode_t da, DiagMode_t db)
{
   MemoryBitManipulator Reg(0);
   Reg.Field(0, 3, to_number(da));
   Reg.Field(4, 3, to_number(db));

   Spi(REG_DIAG_CTRL, Reg.Value());
}


//-----------------------------------------------------------------------------
//  AD7768_ADC::SetDiagnosticMode()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::SetDiagnosticMode(int diag)
{
    if (diag < 4) _diag = diag;
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::SetDiagnosticChannel()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::SetDiagnosticChannel(unsigned int ch, bool en)
{
    SpiBit(REG_DIAG_SEL, ch, en ? 1 : 0);
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::Reset()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::Reset()
{
    Spi(REG_DATA_CTRL, 0x3);
    Spi(REG_DATA_CTRL, 0x2);
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::AdcInit()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::AdcInit()
{
    Debug.Log("AD7768_ADC::AdcInit()");
    Reset();
    Sleep(10);
    // Read ID
    Debug.Log("Adc Chip ID =", Spi(REG_REV_ID), 16);
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::Spi()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::Spi(unsigned int addr, unsigned int value)
{
    _io->AdcSpi(_devId, addr, value);
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::Spi()  --
//-----------------------------------------------------------------------------

unsigned int AD7768_ADC::Spi(unsigned int addr)
{
    (void)_io->AdcSpi(_devId, addr);

    return _io->AdcSpi(_devId, addr);
}

//---------------------------------------------------------------------------
//  AD7768_ADC::SpiBit() --
//---------------------------------------------------------------------------

void AD7768_ADC::SpiBit(int addr, int bit, bool data)
{
    MemoryBitManipulator Reg(Spi(addr));
    Reg.Bit(bit, data);
    _io->AdcSpi(_devId, addr, Reg.Value());
}

//---------------------------------------------------------------------------
//  AD7768_ADC::SpiField() --
//---------------------------------------------------------------------------

void   AD7768_ADC::SpiField(int addr, int bit, int size, unsigned int data)
{
    MemoryBitManipulator Reg(Spi(addr));
    Reg.Field(bit, size, data);
    _io->AdcSpi(_devId, addr, Reg.Value());

}

//---------------------------------------------------------------------------
//  AD7768_ADC::BistEnable() --
//---------------------------------------------------------------------------

void AD7768_ADC::BistEnable()
{
    SpiBit(REG_BIST_CTRL, 0, true);
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::DisplayRegisters()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::DisplayRegisters()
{
    Debug.Log( "REG_CH_STANDBY    = ", Spi(REG_CH_STANDBY), 16);
    Debug.Log( "REG_CH_MODE_A     = ", Spi(REG_CH_MODE_A), 16);
    Debug.Log( "REG_CH_MODE_B     = ", Spi(REG_CH_MODE_B), 16);
    Debug.Log( "REG_CH_MODE_SEL   = ", Spi(REG_CH_MODE_SEL), 16);
    Debug.Log( "REG_PWR_MODE      = ", Spi(REG_PWR_MODE), 16);
    Debug.Log( "REG_GENERAL_CFG   = ", Spi(REG_GENERAL_CFG), 16);
    Debug.Log( "REG_DATA_CTRL     = ", Spi(REG_DATA_CTRL), 16);
    Debug.Log( "REG_INTERFACE_CFG = ", Spi(REG_INTERFACE_CFG), 16);
    Debug.Log( "REG_DEV_STATUS    = ", Spi(REG_DEV_STATUS), 16);
    Debug.Log( "REG_REV_ID        = ", Spi(REG_REV_ID), 16);
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::DisplayDivisors()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::DisplayDivisors()
{
    static string decstr[] {
        "DEC_X32",       "DEC_X64",
        "DEC_X128",      "DEC_X256",
        "DEC_X512",      "DEC_X1024",
        "DEC_X1024_2ND", "DEC_X1024_3RD",
    };
    static string mclkstr[] {
        "MCLK_DIV_32",   "INVALID",
        "MCLK_DIV_8",    "MCLK_DIV_4",
    };
    static string dclkstr[] {
        "DCLK_DIV_8",    "DCLK_DIV_4",
        "DCLK_DIV_2",    "DCLK_DIV_1",
    };

    Debug.Log("Decimation = ",  decstr[to_number(_decRate[0])]);
    Debug.Log("CurrentMCLK = ", mclkstr[to_number(_mclkDiv)]);
    Debug.Log("CurrentDCLK = ", dclkstr[to_number(_dclkDiv)]);
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::CalculateDivisors()  --
//-----------------------------------------------------------------------------

double AD7768_ADC::CalculateDivisors(double adc_rate)
{
    Debug.Log("AD7768_ADC::CalculateDivisors(), adc_rate (Hz) = ", adc_rate);

    double ODR{adc_rate};

    vector<std::pair<DecRate_t, unsigned int>> decs{
        {DecRate_t::DEC_X1024,1024},
        {DecRate_t::DEC_X512,  512},
        {DecRate_t::DEC_X256,  256},
        {DecRate_t::DEC_X128,  128},
        {DecRate_t::DEC_X64,    64},
        {DecRate_t::DEC_X32,    32}
    };

    vector<std::pair<MclkDiv_t, unsigned int>> mclks{
        {MclkDiv_t::MCLK_DIV_32, 32},
        {MclkDiv_t::MCLK_DIV_8,   8},
        {MclkDiv_t::MCLK_DIV_4,   4}
    };

    vector<std::pair<DclkDiv_t, unsigned int>> dclks{
        {DclkDiv_t::DCLK_DIV_8,  8},
        {DclkDiv_t::DCLK_DIV_4,  4},
        {DclkDiv_t::DCLK_DIV_2,  2},
        {DclkDiv_t::DCLK_DIV_1,  1}
    };

    //-------- Modes of operation ----------
    //  Fast Mode: Fmod between 4.194304MHz and 8.388608MHz
    //  Median Mode: Fmod between 1.048578MHz and 4.194304MHz
    //  Eco Mode: Fmod between 0.036MHz and 1.048578MHzMHz
    vector<std::pair<PowerMode_t, std::pair<double, double>>> pows{
        {PowerMode_t::FAST,    {8388608, 4194304}},
        {PowerMode_t::MEDIAN,  {4194303, 1048578}},
        {PowerMode_t::ECO,     {1048577,   36000}}
    };

    struct ValidFactorType {
        PowerMode_t validPow;
        DecRate_t   validDec;
        MclkDiv_t	validMclk;
        DclkDiv_t	validDclk;
    };

    vector<ValidFactorType> valid;

    for (const auto &p : pows) {
        // ODR >= Lowest boundary with max decimation (1024)
        if (ODR >= p.second.second / 1024) {
            double UpperLimit, LowerLimit;
            tie(UpperLimit, LowerLimit) = p.second;
            // Decimation range
            for (const auto &d : decs) {
                // Check within the power mode range
                if (WithinRange(d.second * ODR, LowerLimit, UpperLimit)) {
                    double FMOD = d.second * ODR;
                    // MCLK divisors range
                    for (const auto &m : mclks) {
                        double MClk = m.second * FMOD;
                        if (MClk <= 34000000) {
                            double RequiredDCLK = ODR * 32;
                            // DCLK divisors range
                            for (const auto &k : dclks) {
                                if (RequiredDCLK == MClk / k.second) {
                                    valid.push_back({p.first, d.first, m.first, k.first});
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    if (!valid.size()) {
        valid.push_back({PowerMode_t::FAST, DecRate_t::DEC_X64, MclkDiv_t::MCLK_DIV_4, DclkDiv_t::DCLK_DIV_8});
        Debug.Log("No Valid divisors found for ODR (Hz)  = ", ODR);
    }

    _powerMode       = valid[0].validPow;
    _mclkDiv         = valid[0].validMclk;
    _dclkDiv         = valid[0].validDclk;
    _decRate[0]      = valid[0].validDec;
    _decRate[1]      = valid[0].validDec;

    static int    dclkMap[] = {8, 4, 2, 1};
    static int    mclkMap[] = {32, 0, 8, 4};
    static int    decMap [] = {32, 64, 128, 256, 512, 1024};
    vector<string> powMap {"FAST", "INVALID", "MEDIAN", "ECO"};

    double MCLK = ODR *  decMap[to_number(_decRate[0])] * mclkMap[to_number(_mclkDiv)];

    AdcTuningEvent e{MCLK, _powerMode, _decRate[0], _mclkDiv, _dclkDiv};
    OnAdcTuning.Execute(e);

    DisplayDivisors();

    Debug.Log("Power Mode  = ", powMap[to_number(_powerMode)]);
    Debug.Log("MCLK  (Hz)  = ", MCLK);
    Debug.Log("DCLK  (Hz)  = ", MCLK / dclkMap[to_number(_dclkDiv)]);

    return MCLK;
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::OpenHardware()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::OpenHardware()
{
    if (!DeviceEnabled())
        return;
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::ConfigureHardware()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;

    AdcInit();

    MemoryBitManipulator Reg(0);
    for (size_t ch = 0; ch < 8; ch++) {
        ChState_t en = _io->AdcChannelEnabled((_devId * 8) + ch) ? ChState_t::ENABLED : ChState_t::STANDBY;
        SetChannelStandby(ch, en);
        Reg.Bit(ch, to_number(_chState[ch]));
    }

    Spi(REG_CH_STANDBY, (Reg.Value() == 0xff) ? 0xff : Reg.Value() & 0xfe);

    Spi(REG_GPIO_CTRL, 0x10);
    Spi(REG_GPIO_OUT, 0x0);

    SetPowerMode(SleepMode_t::ACTIVE, _powerMode, _mclkDiv);
    SetConversionMode(ConvOp_t::STANDARD_CONV);
    SetDClkDivMode(_dclkDiv);
    SetModeA(Filt_t::FILTER_WIDEBAND, _decRate[0]);
    SetChannelMode(ChMode_t::MODE_A);

    Spi(REG_GENERAL_CFG, 0x02);

    if (_diag) {
        ad7768::DiagMode_t dmap[] = {DiagMode_t::OFF          ,
                                     DiagMode_t::FULL_POSITIVE,
                                     DiagMode_t::FULL_NEGATIVE,
                                     DiagMode_t::FULL_ZERO};

        Spi(REG_DIAG_SEL, 0xf0);
        SetDiagnosticMode(dmap[_diag], dmap[_diag]);
    }

    DisplayRegisters();
}

//-----------------------------------------------------------------------------
//  AD7768_ADC::SetPowerMode()  --
//-----------------------------------------------------------------------------

void AD7768_ADC::SetPowerMode(SleepMode_t slp, PowerMode_t pwr, MclkDiv_t clk)
{
    MemoryBitManipulator Reg(0);
    Reg.Field(0, 2, to_number(clk));
    Reg.Bit(3, 1); // LVDS enable
    Reg.Field(4, 2, to_number(pwr));
    Reg.Bit(7, to_number(slp));

    Spi(REG_PWR_MODE, Reg.Value());
}


} } // Innovative::ad7768
