#ifndef AD7768_MB_H
#define AD7768_MB_H

#include <type_traits>

#include <X6Support_Mb.h>

namespace Innovative {
namespace ad7768 {

//===========================================================================
//  CLASS AAD7768 Enums
//===========================================================================

enum class SleepMode_t {
    ACTIVE,
    SLEEP,
};

enum class PowerMode_t {
    ECO    = 0,
    MEDIAN = 2,
    FAST   = 3,
};

enum class MclkDiv_t {
    MCLK_DIV_32 = 0,
    MCLK_DIV_8  = 2,
    MCLK_DIV_4  = 3,
};

enum class DclkDiv_t {
    DCLK_DIV_8,
    DCLK_DIV_4,
    DCLK_DIV_2,
    DCLK_DIV_1,
};

enum class ConvOp_t {
    STANDARD_CONV,
    ONE_SHOT_CONV,
};

enum class CrcSel_t {
    NO_CRC,
    CRC_4,
    CRC_16,
    CRC_16_2ND,
};

enum class Channel_t {
    CH0, CH1, CH2, CH3,
    CH4, CH5, CH6, CH7,
};

enum class ChState_t {
    ENABLED,
    STANDBY,
};

enum class ChMode_t {
    MODE_A,
    MODE_B,
};

enum class Filt_t {
    FILTER_WIDEBAND,
    FILTER_SINC,
};

enum class DecRate_t  {
    DEC_X32,
    DEC_X64,
    DEC_X128,
    DEC_X256,
    DEC_X512,
    DEC_X1024,
    DEC_X1024_2ND,
    DEC_X1024_3RD,
};

enum class DiagMode_t {
    OFF           = 0,
    FULL_POSITIVE = 3,
    FULL_NEGATIVE = 4,
    FULL_ZERO     = 5
};

const unsigned int REG_CH_STANDBY    = 0x00;
const unsigned int REG_CH_MODE_A     = 0x01;
const unsigned int REG_CH_MODE_B     = 0x02;
const unsigned int REG_CH_MODE_SEL   = 0x03;
const unsigned int REG_PWR_MODE      = 0x04;
const unsigned int REG_GENERAL_CFG   = 0x05;
const unsigned int REG_DATA_CTRL     = 0x06;
const unsigned int REG_INTERFACE_CFG = 0x07;
const unsigned int REG_BIST_CTRL     = 0x08;
const unsigned int REG_DEV_STATUS    = 0x09;
const unsigned int REG_REV_ID        = 0x0a;
const unsigned int REG_DIAG_SEL      = 0x56;
const unsigned int REG_DIAG_CTRL     = 0x57;
const unsigned int REG_GPIO_CTRL     = 0x0e;
const unsigned int REG_GPIO_OUT      = 0x0f;

// Convert it to the underlying type
template <typename T>
constexpr typename std::underlying_type<T>::type to_number(T value) noexcept
{
    return static_cast<typename std::underlying_type<T>::type>(value);
}

//===========================================================================
//  CLASS IRequires_AD7768_ADC_Io -- X6 "SubDevice" Requirement Interface
//===========================================================================

struct IRequires_AD7768_ADC_Io
{
    // Io Interface
    virtual void         AdcSpi(unsigned int device, unsigned int address, unsigned int data) = 0;
    virtual unsigned int AdcSpi(unsigned int device, unsigned int address) = 0;

    virtual bool AdcChannelEnabled(int ch) = 0;
    virtual void AdcChannelEnabled(int ch, bool state) = 0;

    // Customization Interface
    virtual ~IRequires_AD7768_ADC_Io()
        {}
};


struct AdcTuningEvent : public OpenWire::Event
{
    double          &_mclkRateHz;
    PowerMode_t     &_powerMode;
    DecRate_t       &_decRate;
    MclkDiv_t       &_mclkDiv;
    DclkDiv_t       &_dclkDiv;

    AdcTuningEvent(double &a, PowerMode_t &b, DecRate_t &c, MclkDiv_t &d, DclkDiv_t &e)
        : _mclkRateHz(a), _powerMode(b), _decRate(c), _mclkDiv(d), _dclkDiv(e)
    {}
};

//===========================================================================
//  CLASS AD7768_ADC --
//===========================================================================

class AD7768_ADC : public IX6SubDevice
{
public:
    AD7768_ADC(IRequires_AD7768_ADC_Io *io, unsigned int devId);
    virtual ~AD7768_ADC();
    // IX6SubDevice Interface
    void OpenHardware() override;
    void ConfigureHardware() override;
    //
    void AdcInit();
    void Reset();
    void SetPowerMode(PowerMode_t mode);
    void SetPowerMode(SleepMode_t slp, PowerMode_t pwr, MclkDiv_t clk);
    void SetSleepMode(SleepMode_t mode);
    void SetMClkDivMode(MclkDiv_t mode);
    void SetDClkDivMode(DclkDiv_t mode);
    void SetConversionMode(ConvOp_t mode);
    void SetCrcMode(CrcSel_t mode);
    void SetChannelStandby(unsigned int ch, ChState_t mode);
    void SetChannelMode(unsigned int ch, ChMode_t mode);
    void SetChannelMode(ChMode_t mode);
    void SetModeA(Filt_t fi, DecRate_t dec);
    void SetModeB(Filt_t fi, DecRate_t dec);
    void SetDiagnosticMode(DiagMode_t da, DiagMode_t db);
    void SetDiagnosticChannel(unsigned int ch, bool en);
    //
    void         Spi(unsigned int address, unsigned int data);
    unsigned int Spi(unsigned int addr);
    void         SpiBit(int addr, int bit, bool data);
    void         SpiField(int addr, int bit, int size, unsigned int data);

    void   BistEnable();
    void   DisplayRegisters();
    void   DisplayDivisors();
    double CalculateDivisors(double );

    void    SetDiagnosticMode(int diag);

    OpenWire::EventHandler<AdcTuningEvent> OnAdcTuning;

private:
    IRequires_AD7768_ADC_Io  *_io;
    unsigned int              _devId;

    int                       _spiDev;
    SleepMode_t               _sleepMode;
    PowerMode_t               _powerMode;
    MclkDiv_t                 _mclkDiv;
    DclkDiv_t                 _dclkDiv;
    ConvOp_t                  _convOp;
    CrcSel_t                  _crcSel;
    ChState_t                 _chState[8];
    ChMode_t                  _chMode[8];
    Filt_t                    _filtType[2];
    DecRate_t                 _decRate[2];
    // Diagnostic support
    int                       _diag;

};


} // namespace ad7768
} // namespace Innovative


#endif // AD7768_MB_H
