//---------------------------------------------------------------------------
// AD9162_Dac_SubDevice_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2016 Innovative Integration Corporation.
//  All Rights Reserved.

#include <AD9162_Dac_SubDevice_Mb.h>
#include <SystemSupport_Mb.h>
#include <cmath>
#include <FloatingPtSupport_Mb.h>
#include <Magic_Mb.h>

using namespace std;

namespace Innovative
{

#ifdef _MSC_VER
    #pragma warning(push)
    #pragma warning(disable : 4355)
#endif

//===========================================================================
//  CLASS AD9162_DAC -- JESD D/A Support "SubDevice"
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class AD9162_DAC
//---------------------------------------------------------------------------
const unsigned int DacLaneCount        = 8;
const unsigned int FramesPerMultiFrame = 32;
const bool         EnableScrambler     = true;
const unsigned int JesdSubclass        = 1;     // 0 or 1 only supportred by the part

//---------------------------------------------------------------------------
//  constructor for class AD9162_DAC
//---------------------------------------------------------------------------

AD9162_DAC::AD9162_DAC(IRequires_AD9162_Dac_Io * mdc, unsigned int jesd_device_id)
    :  LanePN_Inverted(DacLaneCount),  PhysToLogical_LaneMap(DacLaneCount),
       MDC(mdc), DeviceID(jesd_device_id)
{
    // Initialize
    FFullScaleCurrent_mA = 40.0;
}

//---------------------------------------------------------------------------
//  destructor for class AD9162_DAC
//---------------------------------------------------------------------------

AD9162_DAC::~AD9162_DAC()
{
}

//---------------------------------------------------------------------------
//  AD9162_DAC::ConstructCustomParts()
//---------------------------------------------------------------------------

void  AD9162_DAC::ConstructCustomParts()
{
    if (!DeviceEnabled())
        return;

    MDC->SetDacLaneConfiguration(DeviceID);
}

//---------------------------------------------------------------------------
//  AD9162_DAC::OpenHardware()
//---------------------------------------------------------------------------

void  AD9162_DAC::OpenHardware()
{
    if (!DeviceEnabled())
        return;

    // Enable 4-wire serial port operation
    Spi(0x0,0x18);
}

//---------------------------------------------------------------------------
//  AD9162_DAC::ConfigureHardware()
//---------------------------------------------------------------------------

void  AD9162_DAC::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;

    ScopedTraceRef ccc(Debug, "AD9162_DAC::ConfigureHardware()");

    if (ChannelEnabled(DeviceID)==false)
        {
        Debug.Log("DAC channel disabled, skipping config for DAC", DeviceID);
        return;   // no need to go on, channel is disabled
        }
    else
        {
        Debug.Log("------------------------------------------------------------");
        Debug.Log("Configure DAC", DeviceID);
        Debug.Log("------------------------------------------------------------");
        }

    EnNcoDmMode = false;

    ConfigureDacStartUp();
    ConfigureDacNco();
    ConfigureJesdStartUp();
}

//---------------------------------------------------------------------------
//  AD9162_DAC::ConfigureDacStartUp()
//---------------------------------------------------------------------------

void AD9162_DAC::ConfigureDacStartUp()
{
    ScopedTraceRef ccc(Debug, "AD9162_DAC::ConfigureDacStartUp()");

    // Clear the error occurrence indicator
    ErrorOccurred = false;

    Debug.Log("DAC Interpolation Factor = ", InterpolationFactor);
    Debug.Log("DAC Serial Line Rate(Gbps) = ", (LineRate/1.0e9));
    Debug.Log("DAC Number of serial lanes = ", NumLanes);
    Debug.Log("DAC Number of converters = ", NumConverters);

    // Enable 4-wire serial port operation and soft-reset the DAC
    Spi(0x0,0x99);
    Sleep(1);

    // Reset internal calibration registers
    Spi(0xD2,0x52);
    Spi(0xD2,0xD2);

    // Configure the NVRAM
    Spi(0x606,0x2);
    Spi(0x607,0x0);

    // Load the factory calibration factors from the NVRAM
    Spi(0x604,0x1);

    Debug.Log("Dac Chip Type =", Spi(0x3), 16);

    MemoryBitManipulator Data(0);
    Data.Field(0, 8, Spi(0x4));
    Data.Field(8, 8, Spi(0x5));
    Debug.Log("Dac Product ID =", Data.Value(), 16);

    Data.Value(Spi(0x6));
    unsigned int DevRevision = Data.Field(0, 4);
    Debug.Log("Dac Device Revision =", DevRevision, 16);
    unsigned int ProdGrade = Data.Field(4, 4);
    Debug.Log("Dac Product Grade =", ProdGrade, 16);

    // Poll the boot load status indicator
    Data.Value(Spi(0x604));
    bool BootLoadStatus = Data.Bit(1);
    for (unsigned int i = 0; (i < 10) && !BootLoadStatus; ++i)
        {
            Sleep(1);
            Data.Value(Spi(0x604));
            BootLoadStatus = Data.Bit(1);
        }
    Debug.Log("Dac Boot Load Pass =", BootLoadStatus);

    if (BootLoadStatus==false)
        {
        Debug.Log("Cancel DAC configuration due to boot loader failure");
        ErrorOccurred = true;
        return;
        }

    // Enable the band gap reference
    Spi(0x58,0x3);

    ConfigureDacClkDll();

    // Cancel further configuration if an error was detected
    if (ErrorOccurred)
        return;

    // Configure the Full Scale Current
    ConfigureFullScaleCurrent();

    // Enable the interrupts
    Spi(0x20,0x0F);

    // Enable calibration factors
    Spi(0xE8,0x20);

    // Configure the DAC decode mode
    unsigned int DecodeModeReg = (DecodeMode==dm2xNRZ)? dmNRZ :DecodeMode;
    Spi(0x152,DecodeModeReg);

    // Use TX_ENABLE to reset the NCO and mask the DAC output
    Spi(0x3F,0x0E);
}

//---------------------------------------------------------------------------
//  AD9162_DAC::ConfigureDacClkDll()
//---------------------------------------------------------------------------

void AD9162_DAC::ConfigureDacClkDll()
{
    ScopedTraceRef ccc(Debug, "AD9162_DAC::ConfigureDacClkDll()");

    unsigned int SysrefPhase;

    // Power cycle the DAC clock DLL till the captured SYSREF phase is 0
    for (unsigned int i=0; i<100; i++)
        {
        Debug.Log("DAC DLL phase calibration loop index = ", i);

        // Disable the DAC clock DLL
        // Power down the DAC clock DLL
        Spi(0x90,0x1F);
        // Power down the DAC clock receiver
        Spi(0x80,0x1);

        // Enable the DAC clock DLL
        // Power up the DAC clock DLL
        Spi(0x90,0x1E);
        // Power up the DAC clock receiver
        Spi(0x80,0x0);
        // Enable the DAC bias circuits
        Spi(0x40,0x0);
        // Configure DAC analog parameters
        Spi(0x9E,0x85);
        // Enable the DAC clock DLL
        Spi(0x91,0xE9);
        WaitForDacClkDllLock();
        // Cancel further configuration if an error was detected
        if (ErrorOccurred)
            return;

        // Save the SYSREF data for readback
        Spi(0x37, 0);
        MemoryBitManipulator Data(Spi(0x37));
        Debug.Log("DAC SYSREF_PHASE0 Reg0x37 = ", Data.Value(), 16);
        SysrefPhase = Data.Field(4,4);
        if (SysrefPhase==0)
            break;
        }

    if (SysrefPhase!=0)
        {
        Debug.Log("Cancel DAC configuration due to SYSREF phase calibration failure");
        ErrorOccurred = true;
        }
}

//---------------------------------------------------------------------------
//  AD9162_DAC::WaitForDacClkDllLock()
//---------------------------------------------------------------------------

void AD9162_DAC::WaitForDacClkDllLock()
{
    MemoryBitManipulator Data(0);
    // Poll the DLL_STATUS register for DLL_LOCKED
    Data.Value(Spi(0x92));
    bool DllLocked = Data.Bit(0);
    for (unsigned int i = 0; (i < 10) && !DllLocked; i++)
        {
            Sleep(1);
            Data.Value(Spi(0x92));
            DllLocked = Data.Bit(0);
        }
    Debug.Log("DAC DLL Locked =", DllLocked);
    Debug.Log("DAC DLL Status Reg0x92 = ", Data.Value(), 16);

    if (DllLocked==false)
        {
        Debug.Log("Cancel DAC configuration due to DLL Lock failure");
        ErrorOccurred = true;
        }
}

//---------------------------------------------------------------------------
//  AD9162_DAC::ConfigureFullScaleCurrent()
//---------------------------------------------------------------------------

void AD9162_DAC::ConfigureFullScaleCurrent()
{
    const double FullScaleCurrentMax = 40.0;
    double FullScaleCurrentCapped = std::min(FFullScaleCurrent_mA,FullScaleCurrentMax);

    if (FFullScaleCurrent_mA != FullScaleCurrentCapped)
        {
        Debug.Log("User Full Scale Current = ", FFullScaleCurrent_mA);
        Debug.Log("Capped Full Scale Current = ", FullScaleCurrentCapped);
        }

    unsigned int FSC = NearestInteger((FullScaleCurrentCapped - 8.0)*1023/32);
    Debug.Log("Full Scale Current Code = ", FSC, 16);

    MemoryBitManipulator Data(Spi(0x41));
    MemoryBitManipulator Src(FSC);
    Data.Field(0, 2, Src.Field(0, 2));
    Spi(0x41,Data.Value());
    Data.Value(Src.Field(2, 8));
    Spi(0x42,Data.Value());
}

//---------------------------------------------------------------------------
//  AD9162_DAC::ConfigureDacNco()
//---------------------------------------------------------------------------

void AD9162_DAC::ConfigureDacNco()
{
    ScopedTraceRef ccc(Debug, "AD9162_DAC::ConfigureDacNco()");

    if (ErrorOccurred)
        {
        Debug.Log("Exit due to earlier error");
        return;
        }

    if (NcoMode == nmNone)
        {
        Debug.Log("Exit since NCO is not enabled");
        return;
        }

    // Configure the DC_TEST_EN bit
    if (NcoMode == nmNcoOnly)
        Spi(0x150,0x2);
    else
        Spi(0x150,0x0);

    // Cap the DC_TEST_DATA value
    const int AmplitudeMax = (1<<15)-1;
    const int AmplitudeMin = 0;
    int NcoOnlyAmplitudeCapped;

    // Check the NcoOnlyAmplitude value if within range, saturate if not within range
    NcoOnlyAmplitudeCapped = std::min(AmplitudeMax,NcoOnlyAmplitude);
    NcoOnlyAmplitudeCapped = std::max(AmplitudeMin,NcoOnlyAmplitudeCapped);

    if (NcoOnlyAmplitude != NcoOnlyAmplitudeCapped)
        {
        Debug.Log("GUI NCO Only Amplitude = ", NcoOnlyAmplitude, 16);
        Debug.Log("Capped NCO Only Amplitude = ", NcoOnlyAmplitudeCapped, 16);
        }

    MemoryBitManipulator Data(0);
    MemoryBitManipulator Src(NcoOnlyAmplitudeCapped);

    // Write the amplitude value for tone amplitude in NCO only mode
    Data.Field(0, 8, Src.Field(8, 8));
    Spi(0x14E,Data.Value());
    Data.Field(0, 8, Src.Field(0, 8));
    Spi(0x14F,Data.Value());

    // Check the NcoPhase value if within range, saturate if not within range
    const int MaxPos16b = (1<<15)-1;
    const int MaxNeg16b = -(1<<15);

    int NcoPhaseCapped;
    NcoPhaseCapped = std::min(MaxPos16b,NcoPhase);
    NcoPhaseCapped = std::max(MaxNeg16b,NcoPhaseCapped);
    if (NcoPhase != NcoPhaseCapped)
        {
        Debug.Log("GUI NCO Phase ", NcoPhase, 16);
        Debug.Log("Capped NCO Phase = ", NcoPhaseCapped, 16);
        }

    // Calculate the NCO parameters
    CalcNcoParams();

    // Ensure the frequency tuning word write request is low
    Spi(0x113,0x0);

    // Write FTW
    for (unsigned int i=0; i<6; i++)
        {
        Data.Field(0, 8, (static_cast<unsigned int>(NcoFtw>>(i*8))));
        Spi(0x114+i, Data.Value());
        }

    // Write NCO_PHASE_OFFSET
    for (unsigned int i=0; i<2; i++)
        {
        Data.Field(0, 8, (NcoPhaseCapped>>(i*8)));
        Spi(0x11C+i, Data.Value());
        }

    // Write ACC_MODULUS
    for (unsigned int i=0; i<6; i++)
        {
        Data.Field(0, 8, (static_cast<unsigned int>(NcoModulus>>(i*8))));
        Spi(0x124+i, Data.Value());
        }

    // Write ACC_DELTA
    for (unsigned int i=0; i<6; i++)
        {
        Data.Field(0, 8, (static_cast<unsigned int>(NcoDelta>>(i*8))));
        Spi(0x12A+i, Data.Value());
        }

    // Load the FTW to the NCO
    Spi(0x113,0x1);

    // Perform this write if NCO only mode is desired
    // Set INTERP_MODE = 0 and JESD_LANES = 8
    if (NcoMode == nmNcoOnly)
        Spi(0x110,0x80);

    // Configure datapath options
    Data.Value(0);
    Data.Bit(0,(DecodeMode==dm2xNRZ)?1:0);                 // FIR85_FILT_EN
    Data.Bit(1,(NcoSideband==nsbUpper)?0:1);               // SEL_SIDEBAND
    Data.Bit(2,(NcoMode!=nmNone && EnNcoDmMode)? 1 : 0);   // MODULUS_EN
    Data.Bit(4,(HbFilterBw?1:0));                          // FILT_BW
    Data.Bit(6,(NcoMode==nmNone)? 0 : 1);                  // NCO_EN
    Data.Bit(7,(EnInvSinc?1:0));                           // INVSINC_EN
    Spi(0x111,Data.Value());
}

//---------------------------------------------------------------------------
//  AD9162_DAC::CalcNcoParams()
//---------------------------------------------------------------------------

void AD9162_DAC::CalcNcoParams()
{
    double NcoFreqCapped;

    // Check NcoFreq if within range, saturate if not within range
    if (DecodeMode == dm2xNRZ)      // 0 <= fCARRIER < fDAC When FIR85 is enabled
        {
        if (NcoFreq < 0.0)
            NcoFreqCapped = NcoFreq * -1.0;         // Inverse polarity
        else
            NcoFreqCapped = NcoFreq;
        if (NcoFreqCapped >= DacSamplingRate)
            NcoFreqCapped = DacSamplingRate-1.0;    // Saturate to Max Value
        }
    else                            // -(fDAC/2) <= fCARRIER < (fDAC/2) When FIR85 is not enabled
        {
        if (NcoFreq < (-DacSamplingRate/2))
            NcoFreqCapped = -DacSamplingRate/2;      // Saturate to Max Value
        else
            NcoFreqCapped = NcoFreq;
        if (NcoFreqCapped >= (DacSamplingRate/2))
            NcoFreqCapped = (DacSamplingRate/2)-1.0; // Saturate to Max Value
        }

    if (NcoFreq != NcoFreqCapped)
        {
            Debug.Log("NCO frequency out of range");
            Debug.Log("GUI NCO frequency = ", NcoFreq);
            Debug.Log("Capped NCO frequency = ", NcoFreqCapped);
        }

    // Guard against a case where the NCO frequency selected is 0
    if (IsAboutEqual(NcoFreqCapped, 0.0, 1.0e-15))
        {
        Debug.Log("Skip NCO parameter calculation since NCO frequency = 0");
        Debug.Log("Select Integer NCO mode");
        EnNcoDmMode = false;
        NcoFtw      = 0;
        NcoDelta    = 0;
        NcoModulus  = 0;
        }
    else
        {
        s_ii64 sign;
        if (NcoFreqCapped < 0.0)
            {
            sign = -1;
            NcoFreqCapped = -1 * NcoFreqCapped;
            }
        else
            sign = 1;

        // Check if the ratio between DacSamplingRate and NcoFreqCapped is power of 2
        // If true, then enable Integer NCO mode, else enable Dual Modulus NCO mode
        double FreqRatio     = DacSamplingRate/NcoFreqCapped;
        ii64   FreqRatio_Int = static_cast<ii64>(IntegerPart(FreqRatio));
        double FreqRatio_Rem = FractionalPart(FreqRatio);

        if (IsPowerOfTwo(FreqRatio_Int) && IsAboutEqual(FreqRatio_Rem, 0.0, 1.0e-15))    // is power of 2
            EnNcoDmMode = false;
        else                                                                    // is not power of 2
            EnNcoDmMode = true;

        if (EnNcoDmMode==false)     // Integer NCO mode
            {
            Debug.Log("Integer NCO mode");
            NcoFtw        = sign*static_cast<s_ii64>(static_cast<double>(static_cast<s_ii64>(1)<<48)/FreqRatio);
            NcoDelta      = 0;
            NcoModulus    = 0;
            }
        else                        // Dual Modulus NCO mode
            {
            Debug.Log("Dual Modulus NCO mode");

             // Calculate fCARRIER/fDAC = M/N in smallest terms
            ii64 gcd = Gcd(static_cast<ii64>(IntegerPart(NcoFreqCapped)), static_cast<ii64>(IntegerPart(DacSamplingRate)));
            Debug.Log("Gcd = ", gcd);
            double M = IntegerPart(NcoFreqCapped/gcd);
            double N = IntegerPart(DacSamplingRate/gcd);
            Debug.Log("M = ", M);
            Debug.Log("N = ", N);

            // Calculate M/N = (X+(A/B))/2^48
            double Q = M/N * (static_cast<ii64>(1)<<48);
            ii64 X = static_cast<ii64>(IntegerPart(Q));
            Debug.Log("X = ", X);
            ii64 NxFrac = static_cast<ii64>(FractionalPart(Q)* N);
            Debug.Log("NxFrac = ", NxFrac);
            ii64 A      = NxFrac;
            ii64 B      = static_cast<ii64>(N);
            gcd         = Gcd(A, B);
            Debug.Log("Gcd = ", gcd);
            A           = A/gcd;
            B           = B/gcd;

            // Assign the calculated values to the output
            NcoFtw     = sign*X;
            NcoDelta   = sign*A;
            NcoModulus = B;
            }
        }

    double M_Actual = (NcoDelta==0)? static_cast<double>(NcoFtw) :
                                     static_cast<double>(NcoFtw) + static_cast<double>(NcoDelta/NcoModulus);
    double NcoFreqActual = DacSamplingRate*(M_Actual/static_cast<double>(static_cast<s_ii64>(1)<<48));

    Debug.Log("NCO FTW = ", NcoFtw);
    Debug.Log("NCO Delta = ", NcoDelta);
    Debug.Log("NCO Modulus = ", NcoModulus);
    Debug.Log("NCO Frequency Actual = ", NcoFreqActual);
}

//---------------------------------------------------------------------------
//  AD9162_DAC::ConfigureJesdStartUp()
//---------------------------------------------------------------------------

void AD9162_DAC::ConfigureJesdStartUp()
{
    ScopedTraceRef ccc(Debug, "AD9162_DAC::ConfigureJesdStartUp()");

    if (ErrorOccurred)
        {
        Debug.Log("Exit due to earlier error");
        return;
        }

    if (NcoMode == nmNcoOnly)
        {
        Debug.Log("Exit since in NCO only mode");
        return;
        }

    // Ensure the SERDES links are disabled before configuring them
    Spi(0x300,0x0);

    // Enable JESD204B interrupts
    Spi(0x4B8,0xFF);
    Spi(0x4B9,0x01);

    // Enable SERDES error counters
    for (unsigned int i=0x480; i<=0x487; i++)
        Spi(i, 0x38);

    // Encode the interpolation rate to the desired register format
    unsigned int InterpMode = (InterpolationFactor==24)? 8 :
                              (InterpolationFactor==16)? 7 :
                              (InterpolationFactor==12)? 6 :
                              (InterpolationFactor==8 )? 5 :
                              (InterpolationFactor==6 )? 4 :
                              (InterpolationFactor==4 )? 3 :
                              (InterpolationFactor==3 )? 2 :
                              (InterpolationFactor==2 )? 1 : 0;

    MemoryBitManipulator Data(0);

    // Configure number of lanes and interpolation rate
    Data.Field(0, 4, InterpMode);
    Data.Field(4, 4, NumLanes);
    Spi(0x110,Data.Value());

    // Configure datapath options
    Data.Value(0);
    Data.Bit(0,(DecodeMode==dm2xNRZ)?1:0);                 // FIR85_FILT_EN
    Data.Bit(1,(NcoSideband==nsbUpper)?0:1);               // SEL_SIDEBAND
    Data.Bit(2,(NcoMode!=nmNone && EnNcoDmMode)? 1 : 0);   // MODULUS_EN
    Data.Bit(4,(HbFilterBw?1:0));                          // FILT_BW
    Data.Bit(6,(NcoMode==nmNone)? 0 : 1);                  // NCO_EN
    Data.Bit(7,(EnInvSinc?1:0));                           // INVSINC_EN
    Spi(0x111,Data.Value());

    ConfigureDacSerDesPll();

    // Enable JESD204B block (disable master SERDES power-down)
    Data.Value(Spi(0x200));
    Data.Bit(0, 0);          // SPI_PD_MASTER
    Spi(0x200,Data.Value());

    // Soft reset the JESD204B quad-byte deframer
    Data.Value(Spi(0x475));
    Data.Bit(3, 1);          // QBD SOFT RESET
    Spi(0x475,Data.Value());

    ConfigureJesdParams();

    // Bring the JESD204B quad-byte deframer out of reset
    Data.Value(Spi(0x475));
    Data.Bit(3, 0);          // QBD SOFT RESET
    Spi(0x475,Data.Value());

    ConfigureDacLanes();

    // Override defaults in the SERDES PLL settings
    Spi(0x29E,0x1F);

    // Reset the CDR
    Spi(0x206,0x0);

    // Enable the CDR
    Spi(0x206,0x1);

    // Enable the SERDES PLL
    Spi(0x280,0x03);

    // Read back SERDES PLL status register until the SERDES PLL is locked
    bool PllLock;
    for (unsigned int i=0; i<5; i++)
        {
        Data.Value(Spi(0x281));
        Debug.Log("DAC SerDes PLL status Reg0x281 = ", Data.Value(), 16);
        PllLock = Data.Bit(0);
        if (PllLock)
            break;
        Sleep(5);
        }
    if (PllLock)
        Debug.Log("DAC SerDes PLL locked");
    else
        Debug.Log("DAC SerDes PLL lock failed");

    if (PllLock==false)
        {
        Debug.Log("Cancel DAC configuration due to SerDes PLL Lock failure");
        ErrorOccurred = true;
        return;
        }

    // Run the sync procedure to align the LMFC to the sysref
    ConfigureSyncProcedure();

    // Enable SERDES links (begin bringing up the link)
    Data.Value(Spi(0x300));
    Data.Bit(0, 1);          // LINK_EN
    Spi(0x300,Data.Value());

    // Soft reset the JESD204B quad-byte deframer
    Data.Value(Spi(0x475));
    Data.Bit(3, 1);          // QBD SOFT RESET
    Spi(0x475,Data.Value());

    // Bring the JESD204B quad-byte deframer out of reset
    Data.Value(Spi(0x475));
    Data.Bit(3, 0);          // QBD SOFT RESET
    Spi(0x475,Data.Value());

    Sleep(1);

    // Check link status
    GetJesdLinkStatus();

    // Clear the interrupts
    Spi(0x24,0x1F);

    // Clear the SERDES interrupts
    Spi(0x4BA,0xFF);
    Spi(0x4BB,0x1);

    // More register dumps to help diagnose link issues
    //GetRcvdILASConfiguration();

    // Report the measured dynamic link latency
    Debug.Log("DAC dynamic link latency Reg0x302 = ", Spi(0x302), 16);
}

//---------------------------------------------------------------------------
//  AD9162_DAC::ConfigureDacSerDesPll()
//---------------------------------------------------------------------------

void AD9162_DAC::ConfigureDacSerDesPll()
{
    ScopedTraceRef ccc(Debug, "AD9162_DAC::ConfigureDacSerDesPll()");

    unsigned int PllRefClkRate;
    unsigned int SerdesPllDivFactor;
    bool         SpiEnHalfRate;
    unsigned int SpiDivisionRate;

    if (LineRate >= 6000.0e6)
        {
        PllRefClkRate      = 0;     // 1x
        SerdesPllDivFactor = 0;     // divide by 4
        SpiEnHalfRate      = true;  // half rate
        SpiDivisionRate    = 0;     // no divide
        }
    else if (LineRate >= 3000.0e6)
        {
        PllRefClkRate      = 0;     // 1x
        SerdesPllDivFactor = 1;     // divide by 2
        SpiEnHalfRate      = false; // full rate
        SpiDivisionRate    = 0;     // no divide
        }
    else if (LineRate >= 1500.0e6)
        {
        PllRefClkRate      = 0;     // 1x
        SerdesPllDivFactor = 2;     // divide by 1
        SpiEnHalfRate      = false; // full rate
        SpiDivisionRate    = 1;     // divide by 2
        }
    else
        {
        PllRefClkRate      = 1;     // 2x
        SerdesPllDivFactor = 2;     // divide by 1
        SpiEnHalfRate      = false; // full rate
        SpiDivisionRate    = 2;     // divide by 4
        }

    MemoryBitManipulator Data(Spi(0x230));

    // Configure the CDR block
    Data.Field(1, 2, SpiDivisionRate);
    Data.Bit(5,(SpiEnHalfRate?1:0));
    Spi(0x230,Data.Value());

    // Set up the SERDES PLL divider
    Data.Value(Spi(0x289));
    Data.Field(0, 2, SerdesPllDivFactor);
    Spi(0x289,Data.Value());

    Data.Bit(2, 0);         // AD916x_FLD_LDO_BYPASS_FILT
    Spi(0x289,Data.Value());

    // Set up the PLL reference clock rate
    Data.Value(Spi(0x084));
    Data.Field(4, 2, PllRefClkRate);
    Spi(0x084,Data.Value());
}

//---------------------------------------------------------------------------
//  AD9162_DAC::ConfigureJesdParams()
//---------------------------------------------------------------------------

void AD9162_DAC::ConfigureJesdParams()
{
    ScopedTraceRef ccc(Debug, "AD9162_DAC::ConfigureJesdParams()");

    MemoryBitManipulator Data(Spi(0x453));

    // Enable scrambling on SERDES lanes
    Data.Bit(7, (EnableScrambler?1:0));
    Spi(0x453, Data.Value());

    // Set the number of frames per multiframe
    Data.Value(Spi(0x455));
    Data.Field(0, 5, FramesPerMultiFrame-1);
    Spi(0x455, Data.Value());

    // Set the subclass type
    Data.Value(Spi(0x458));
    Data.Field(5, 3, JesdSubclass); // Subclass
    Spi(0x458, Data.Value());

    // Set the JESD204x version to JESD204B
    Data.Value(Spi(0x459));
    Data.Field(5, 3, 1);    // JESD204B
    Spi(0x459, Data.Value());

    // Program the calculated checksum value for Lane 0 from values in Register 0x450 to Register 0x45C
    Data.Value(0);
    Data.Field(0, 8, CalcChecksum());
    Spi(0x45D, Data.Value());

    // Enable ILA support
    JesdDac().Poke(0x8, 0x1);
    // Set the scrambling mode
    JesdDac().Poke(0xC, (EnableScrambler?1:0));
    // Set sysref handling mode
    JesdDac().Poke(0x10, 0x0);
    // Set ILA multiframes
    JesdDac().Poke(0x14, 0x3);
    // Set the octets per frame F
    JesdDac().Poke(0x20, (OctetsPerFrame-1));
    // Set the frames per multi frame K
    JesdDac().Poke(0x24, (FramesPerMultiFrame-1));
    // Set the lanes in use
    JesdDac().Poke(0x28, (1 << NumLanes)-1);
    // Select Subclass
    JesdDac().Poke(0x2C, JesdSubclass);

    // Set ILA Config Data 4
    Data.Value(0);
    Data.Field(0, 8, NumConverters-1);
    Data.Field(8, 5, 0xF);      // N = 15
    Data.Field(16, 5, 0xF);     // N' = 15
    Data.Field(24, 2, 0x0);     // CS = 0
    JesdDac().Poke(0x810, Data.Value());

    // Set ILA Config Data 5
    Data.Value(0);
    Data.Field(8, 5, SamplesPerFrame-1);
    Data.Bit(16, (HighDensityMode? 1:0));
    Data.Field(24, 5, 0x0);     // CF = 0
    JesdDac().Poke(0x814, Data.Value());

    // Reset the JESD204B IP
    JesdDac().Poke(0x4, 0x1);
}

//---------------------------------------------------------------------------
//  AD9162_DAC::CalcChecksum()
//---------------------------------------------------------------------------

unsigned int AD9162_DAC::CalcChecksum()
{
    // Calculate the checksum value for Lane 0 from values in Register 0x450 to Register 0x45C
    unsigned int Checksum = (EnableScrambler? 1:0)  +
                            (NumLanes-1)            +
                            (OctetsPerFrame-1)      +
                            (FramesPerMultiFrame-1) +
                            (NumConverters-1)       +
                            15                      + // N
                            JesdSubclass            + // JESD204 SubClass
                            15                      + // NP
                             1                      + // JESD204B version
                            (SamplesPerFrame-1)     +
                            (HighDensityMode? 1:0);
    return Checksum;
}

//---------------------------------------------------------------------------
//  AD9162_DAC::ConfigureDacLanes()
//---------------------------------------------------------------------------

void AD9162_DAC::ConfigureDacLanes()
{
    ScopedTraceRef ccc(Debug, "AD9162_DAC::ConfigureDacLanes()");

    // Set any bits to 1 to power down the appropriate physical lane
    MemoryBitManipulator SerDesLanePd(0);
    for (unsigned int i=0; i<PhysToLogical_LaneMap.size(); i++)
        SerDesLanePd.Bit(i, (PhysToLogical_LaneMap[i]<NumLanes) ? false : true);
    Spi(0x201, SerDesLanePd.Value());

    // Set the logical to physical lanes mapping
    vector<unsigned int> LogicalToPhys_LaneMap; // logical to physical lanes mapping
    for (unsigned int LogicalIdx=0; LogicalIdx<PhysToLogical_LaneMap.size(); LogicalIdx++)
        {
        for (unsigned int PhysIdx=0; PhysIdx<PhysToLogical_LaneMap.size(); PhysIdx++)
            {
            if (LogicalIdx==PhysToLogical_LaneMap[PhysIdx])
                {
                LogicalToPhys_LaneMap.push_back(PhysIdx);
                break;
                }
            }
        }

    MemoryBitManipulator Data(Spi(0x308));

    // Set the crossbar switch
    Data.Field( 0, 3, LogicalToPhys_LaneMap[0]);
    Data.Field( 3, 3, LogicalToPhys_LaneMap[1]);
    Spi(0x308, Data.Value());

    Data.Value(Spi(0x309));
    Data.Field( 0, 3, LogicalToPhys_LaneMap[2]);
    Data.Field( 3, 3, LogicalToPhys_LaneMap[3]);
    Spi(0x309, Data.Value());

    Data.Value(Spi(0x30A));
    Data.Field( 0, 3, LogicalToPhys_LaneMap[4]);
    Data.Field( 3, 3, LogicalToPhys_LaneMap[5]);
    Spi(0x30A, Data.Value());

    Data.Value(Spi(0x30B));
    Data.Field( 0, 3, LogicalToPhys_LaneMap[6]);
    Data.Field( 3, 3, LogicalToPhys_LaneMap[7]);
    Spi(0x30B, Data.Value());

    // Set the lane inversion
    MemoryBitManipulator PinPairInv(0);
    unsigned int LogicalIdx;
    for (unsigned int PhysIdx=0; PhysIdx<LanePN_Inverted.size(); PhysIdx++)
        {
        LogicalIdx = PhysToLogical_LaneMap[PhysIdx];
        PinPairInv.Bit(LogicalIdx, (LanePN_Inverted[PhysIdx]) ? true : false);
        }
    Spi(0x334, PinPairInv.Value());

    // Equaliser Settings
    Spi(0x2AA,0xC3);
    Spi(0x2AB,0x93);
    Spi(0x2B1,0xC3);
    Spi(0x2B2,0x93);

    // Calibrate SERDES PHY Termination Block 1 (PHY 0, PHY 1, PHY 6, PHY 7)
    Data.Value(Spi(0x2A7));
    Data.Bit(0, 1);
    Spi(0x2A7, Data.Value());

    // Calibrate SERDES PHY Termination Block 2 (PHY 2, PHY 3, PHY 4, PHY 5)
    Data.Value(Spi(0x2AE));
    Data.Bit(0, 1);
    Spi(0x2AE, Data.Value());
}

//---------------------------------------------------------------------------
//  AD9162_DAC::ConfigureSyncProcedure()
//---------------------------------------------------------------------------

void AD9162_DAC::ConfigureSyncProcedure()
{
    ScopedTraceRef ccc(Debug, "AD9162_DAC::ConfigureSyncProcedure()");

    // Set SYSREF jitter window
    Spi(0x39, 4);

    // Reset the SYSREF counter
    Spi(0x36, 0);
    // Save the SYSREF data for readback
    Spi(0x34, 0);
    // Read the SYSREF counter
    Debug.Log("DAC SYSREF counter Reg0x36 = ", Spi(0x36), 16);

    // Enable one-shot sync mode
    Spi(0x3A, 2);

    // Save the SYSREF data for readback
    Spi(0x34, 0);
    // Read the SYNC_LMFC_STATx registers
    Debug.Log("DAC SYNC_LMFC_STAT0 Reg0x34 = ", Spi(0x34), 16);
    Debug.Log("DAC SYNC_LMFC_STAT1 Reg0x35 = ", Spi(0x35), 16);

    // Read the SYSREF_PHASEx registers
    Debug.Log("DAC SYSREF_PHASE0 Reg0x37 = ", Spi(0x37), 16);
    Debug.Log("DAC SYSREF_PHASE1 Reg0x38 = ", Spi(0x38), 16);
}

//---------------------------------------------------------------------------
//  AD9162_DAC::GetJesdLinkStatus()
//---------------------------------------------------------------------------

void AD9162_DAC::GetJesdLinkStatus()
{
    ScopedTraceRef ccc(Debug, "AD9162_DAC::GetJesdLinkStatus()");

    Debug.Log("Jesd204 Link Status for DAC", DeviceID);

    // Read the DAC link status indicators
    Debug.Log("DAC lane deskew status Reg0x46C = ", Spi(0x46C), 16);
    Debug.Log("DAC bad disparity status Reg0x46D = ", Spi(0x46D), 16);
    Debug.Log("DAC not in table status Reg0x46E = ", Spi(0x46E), 16);
    Debug.Log("DAC unexpected k char status Reg0x46F = ", Spi(0x46F), 16);
    Debug.Log("DAC CGS status Reg0x470 = ", Spi(0x470), 16);
    Debug.Log("DAC frame sync status Reg0x471 = ", Spi(0x471), 16);
    Debug.Log("DAC good checksum status Reg0x472 = ", Spi(0x472), 16);
    Debug.Log("DAC initial lane sync status Reg0x473 = ", Spi(0x473), 16);

    // Status reported by the JESD IP in the FPGA
    JesdDac().DumpDebugStatus();

    // Error counter status per serial lane
    Debug.Log("DAC ECNT_STAT0 Reg0x490 = ", Spi(0x490), 16);
    Debug.Log("DAC ECNT_STAT1 Reg0x491 = ", Spi(0x491), 16);
    Debug.Log("DAC ECNT_STAT2 Reg0x492 = ", Spi(0x492), 16);
    Debug.Log("DAC ECNT_STAT3 Reg0x493 = ", Spi(0x493), 16);
    Debug.Log("DAC ECNT_STAT4 Reg0x494 = ", Spi(0x494), 16);
    Debug.Log("DAC ECNT_STAT5 Reg0x495 = ", Spi(0x495), 16);
    Debug.Log("DAC ECNT_STAT6 Reg0x496 = ", Spi(0x496), 16);
    Debug.Log("DAC ECNT_STAT7 Reg0x497 = ", Spi(0x497), 16);

    // Link status per serial lane
    Debug.Log("DAC LINK_STATUS0 Reg0x4B0 = ", Spi(0x4B0), 16);
    Debug.Log("DAC LINK_STATUS1 Reg0x4B1 = ", Spi(0x4B1), 16);
    Debug.Log("DAC LINK_STATUS2 Reg0x4B2 = ", Spi(0x4B2), 16);
    Debug.Log("DAC LINK_STATUS3 Reg0x4B3 = ", Spi(0x4B3), 16);
    Debug.Log("DAC LINK_STATUS4 Reg0x4B4 = ", Spi(0x4B4), 16);
    Debug.Log("DAC LINK_STATUS5 Reg0x4B5 = ", Spi(0x4B5), 16);
    Debug.Log("DAC LINK_STATUS6 Reg0x4B6 = ", Spi(0x4B6), 16);
    Debug.Log("DAC LINK_STATUS7 Reg0x4B7 = ", Spi(0x4B7), 16);
}

//---------------------------------------------------------------------------
//  AD9162_DAC::GetRcvdILASConfiguration()
//---------------------------------------------------------------------------

void AD9162_DAC::GetRcvdILASConfiguration()
{
    Debug.Log("Jesd204 Receuived ILAS configuration for DAC", DeviceID);

    // Read the received ILAS configuration on lane0
    Debug.Log("DAC DID_REG Reg0x400 = ", Spi(0x400), 16);
    Debug.Log("DAC BID_REG Reg0x401 = ", Spi(0x401), 16);
    Debug.Log("DAC LID0_REG Reg0x402 = ", Spi(0x402), 16);
    Debug.Log("DAC SCR_L_REG Reg0x403 = ", Spi(0x403), 16);
    Debug.Log("DAC F_REG Reg0x404 = ", Spi(0x404), 16);
    Debug.Log("DAC K_REG Reg0x405 = ", Spi(0x405), 16);
    Debug.Log("DAC M_REG Reg0x406 = ", Spi(0x406), 16);
    Debug.Log("DAC CS_N_REG Reg0x407 = ", Spi(0x407), 16);
    Debug.Log("DAC NP_REG Reg0x408 = ", Spi(0x408), 16);
    Debug.Log("DAC S_REG Reg0x409 = ", Spi(0x409), 16);
    Debug.Log("DAC HD_CF_REG Reg0x40A = ", Spi(0x40A), 16);
    Debug.Log("DAC RES1_REG Reg0x40B = ", Spi(0x40B), 16);
    Debug.Log("DAC RES2_REG Reg0x40C = ", Spi(0x40C), 16);

    Debug.Log("DAC CHECKSUM0_REG Reg0x40D = ", Spi(0x40D), 16);
    Debug.Log("DAC COMPSUM0_REG Reg0x40E = ", Spi(0x40E), 16);

    Debug.Log("DAC LID1_REG Reg0x412 = ", Spi(0x412), 16);
    Debug.Log("DAC CHECKSUM1_REG Reg0x415 = ", Spi(0x415), 16);
    Debug.Log("DAC COMPSUM1_REG Reg0x416 = ", Spi(0x416), 16);

    Debug.Log("DAC LID2_REG Reg0x41A = ", Spi(0x41A), 16);
    Debug.Log("DAC CHECKSUM2_REG Reg0x41D = ", Spi(0x41D), 16);
    Debug.Log("DAC COMPSUM2_REG Reg0x41E = ", Spi(0x41E), 16);

    Debug.Log("DAC LID3_REG Reg0x422 = ", Spi(0x422), 16);
    Debug.Log("DAC CHECKSUM3_REG Reg0x425 = ", Spi(0x425), 16);
    Debug.Log("DAC COMPSUM3_REG Reg0x426 = ", Spi(0x426), 16);

    Debug.Log("DAC LID4_REG Reg0x42A = ", Spi(0x42A), 16);
    Debug.Log("DAC CHECKSUM4_REG Reg0x42D = ", Spi(0x42D), 16);
    Debug.Log("DAC COMPSUM4_REG Reg0x42E = ", Spi(0x42E), 16);

    Debug.Log("DAC LID5_REG Reg0x432 = ", Spi(0x432), 16);
    Debug.Log("DAC CHECKSUM5_REG Reg0x435 = ", Spi(0x435), 16);
    Debug.Log("DAC COMPSUM5_REG Reg0x436 = ", Spi(0x436), 16);

    Debug.Log("DAC LID6_REG Reg0x43A = ", Spi(0x43A), 16);
    Debug.Log("DAC CHECKSUM6_REG Reg0x43D = ", Spi(0x43D), 16);
    Debug.Log("DAC COMPSUM6_REG Reg0x43E = ", Spi(0x43E), 16);

    Debug.Log("DAC LID7_REG Reg0x442 = ", Spi(0x442), 16);
    Debug.Log("DAC CHECKSUM7_REG Reg0x445 = ", Spi(0x445), 16);
    Debug.Log("DAC COMPSUM7_REG Reg0x446 = ", Spi(0x446), 16);

    Debug.Log("DAC User Calculated Checksum Reg0x45D = ", Spi(0x45D), 16);
}

//---------------------------------------------------------------------------
//  AD9162_DAC::CalcDacJesdLineRate() -- Calculate the DAC JESD line rate
//---------------------------------------------------------------------------

double AD9162_DAC::CalcDacJesdLineRate(double SamplingRate)
{
    ScopedTraceRef ccc(Debug, "AD9162_DAC::CalcDacJesdLineRate()");

    DacSamplingRate = SamplingRate;

    CalcDacParams();

    if (ChannelEnabled(DeviceID) == false || NcoMode == nmNcoOnly)
        LineRate = 0.0;
    else
        LineRate = JesdDac().CalcJesdLineRate(NumConverters, 2, SamplingRate, InterpolationFactor, NumLanes);

    return LineRate;
}

//---------------------------------------------------------------------------
//  AD9162_DAC::CalcDacParams() --
//---------------------------------------------------------------------------

void AD9162_DAC::CalcDacParams()
{
    ScopedTraceRef ccc(Debug, "AD9162_DAC::CalcDacParams()");

    // Force the interpolation factor to 1 when in Nco only mode
    InterpolationFactor = (NcoMode == nmNcoOnly)? 1 : InterpolationRate;

    // Set the number of virtual converters based on the interpolation rate and
    // Nco only mode
    if (NcoMode == nmNcoOnly || InterpolationFactor == 1)
        NumConverters = 1;
    else
        NumConverters = 2;

    // Set the number of Octets Per Frame Per Lane and JESD lanes based on the interpolation
    // factor and Nco only mode
    if (NcoMode == nmNcoOnly || InterpolationFactor == 1 || InterpolationFactor == 2 ||
        InterpolationFactor == 3 || InterpolationFactor == 4)
        {
        OctetsPerFrame = 1;
        NumLanes       = 8;
        }
    else if (InterpolationFactor == 6 || InterpolationFactor == 8)
        {
        OctetsPerFrame = 1;
        NumLanes       = 4;
        }
    else if (InterpolationFactor == 16 || InterpolationFactor == 12)
        {
        OctetsPerFrame = 2;
        NumLanes       = 2;
        }
    else    // InterpolationFactor == 24
        {
        OctetsPerFrame = 4;
        NumLanes       = 1;
        }

    SamplesPerFrame = (NumLanes*OctetsPerFrame)/(2*NumConverters);

    HighDensityMode = (OctetsPerFrame==1)? true : false;
}

//---------------------------------------------------------------------------
//  AD9162_DAC::CalcLmfcRate() --
//---------------------------------------------------------------------------

double AD9162_DAC::CalcLmfcRate()
{
    double LmfcRate = LineRate/(10.0 * FramesPerMultiFrame * OctetsPerFrame);
    return LmfcRate;
}

//---------------------------------------------------------------------------
//  AD9162_DAC::ValidSampleRate() --
//---------------------------------------------------------------------------

bool AD9162_DAC::ValidSampleRate(string & text)
{
    double MaxDacUpdateRate = 6.0e9;
    double MinDacUpdateRate = 1.5e9;
    double MaxJesdLineRate  = 12.5e9;
    double MinJesdLineRate  = 0.75e9;

    double MaxSamplingRate   = std::min(((MaxJesdLineRate*InterpolationFactor*NumLanes)/(20*NumConverters)),MaxDacUpdateRate);
    double MinSamplingRate   = std::max(((MinJesdLineRate*InterpolationFactor*NumLanes)/(20*NumConverters)),MinDacUpdateRate);
    bool   SamplingRateValid = WithinRange(DacSamplingRate, MinSamplingRate, MaxSamplingRate);

    stringstream ss;
    bool Valid;

    if (NcoMode == nmNcoOnly && ChannelEnabled(DeviceID))
        {
        if (DacSamplingRate > MaxDacUpdateRate)
            {
            ss << " Sampling Rate faster than the Max Rate of " << (MaxDacUpdateRate/1.0e9) << " GSPS";
            Valid = false;
            }
        else if (DacSamplingRate < MinDacUpdateRate)
            {
            ss << " Sampling Rate slower than the Min Rate of " << (MinDacUpdateRate/1.0e9) << " GSPS";
            Valid = false;
            }
        else
            {
            ss << "";
            Valid = true;
            }
        }
    else if (!SamplingRateValid && ChannelEnabled(DeviceID))
        {
        if (DacSamplingRate > MaxSamplingRate)
            ss << " Sampling Rate faster than the Max Rate of " << (MaxSamplingRate/1.0e9) << " GSPS for the selected interpolation mode";
        else
            ss << " Sampling Rate slower than the Min Rate of " << (MinSamplingRate/1.0e9) << " GSPS for the selected interpolation mode";
        Valid = false;
        }
    else
        {
        ss << "";
        Valid = true;
        }

    text = ss.str();
    return Valid;
}

//---------------------------------------------------------------------------
//  AD9162_DAC::SetInterpolationRate() --
//---------------------------------------------------------------------------

void  AD9162_DAC::SetInterpolationRate(unsigned int value)
{
    InterpolationRate = value;
}

//---------------------------------------------------------------------------
//  AD9162_DAC::SetDecodeMode() --
//---------------------------------------------------------------------------

void AD9162_DAC::SetDecodeMode(unsigned int value)
{
    DecodeMode = static_cast<DecodeModes>(value);
}

//---------------------------------------------------------------------------
//  AD9162_DAC::SetHbFilterBw() --
//---------------------------------------------------------------------------

void AD9162_DAC::SetHbFilterBw(unsigned int value)
{
    HbFilterBw = static_cast<HbFilterBws>(value);
}

//---------------------------------------------------------------------------
//  AD9162_DAC::SetEnInvSinc() --
//---------------------------------------------------------------------------

void  AD9162_DAC::SetEnInvSinc(bool value)
{
    EnInvSinc = value;
}

//---------------------------------------------------------------------------
//  AD9162_DAC::SetNcoMode() --
//---------------------------------------------------------------------------

void AD9162_DAC::SetNcoMode(unsigned int value)
{
    NcoMode = static_cast<NcoModes>(value);
}

//---------------------------------------------------------------------------
//  AD9162_DAC::SetNcoSideband() --
//---------------------------------------------------------------------------

void AD9162_DAC::SetNcoSideband(unsigned int value)
{
    NcoSideband = static_cast<NcoSidebands>(value);
}

//---------------------------------------------------------------------------
//  AD9162_DAC::SetNcoFreq() --
//---------------------------------------------------------------------------

void AD9162_DAC::SetNcoFreq(double value)
{
    NcoFreq = value;
}

//---------------------------------------------------------------------------
//  AD9162_DAC::SetNcoPhase() --
//---------------------------------------------------------------------------

void AD9162_DAC::SetNcoPhase(int value)
{
    NcoPhase = value;
}

//---------------------------------------------------------------------------
//  AD9162_DAC::SetNcoOnlyAmplitude() --
//---------------------------------------------------------------------------

void AD9162_DAC::SetNcoOnlyAmplitude(int value)
{
    NcoOnlyAmplitude = value;
}

//---------------------------------------------------------------------------
//  AD9162_DAC::Spi() -- write to a Spi device
//---------------------------------------------------------------------------

void  AD9162_DAC::Spi(unsigned int address, unsigned int data)
{
    MDC->DacSpi(DeviceID, address, data);
}

//---------------------------------------------------------------------------
//  AD9162_DAC::Spi() -- read from a Spi device
//---------------------------------------------------------------------------

unsigned int  AD9162_DAC::Spi(unsigned int address)
{
    return MDC->DacSpi(DeviceID, address);
}

//---------------------------------------------------------------------------
//  AD9162_DAC::ChannelEnabled() -- read cached channel enable state
//---------------------------------------------------------------------------

bool  AD9162_DAC::ChannelEnabled(int ch)
{
    return MDC->DacChannelEnabled(ch);
}

//---------------------------------------------------------------------------
//  AD9162_DAC::ChannelEnabled() -- read cached channel enable state
//---------------------------------------------------------------------------

void AD9162_DAC::ChannelEnabled(int ch, bool state)
{
    return MDC->DacChannelEnabled(ch, state);
}

}  // namespace Innovative
