//---------------------------------------------------------------------------
// AD9689_SubDevice_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2016 Innovative Integration Corporation.
//  All Rights Reserved.

#include <AD9689_SubDevice_Mb.h>
#include <SystemSupport_Mb.h>

using namespace std;

namespace Innovative
{

#ifdef _MSC_VER
    #pragma warning(push)
    #pragma warning(disable : 4355)
#endif

//===========================================================================
//  CLASS AD9689_ADC -- JESD A/D Support "SubDevice"
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class AD9689_ADC
//---------------------------------------------------------------------------

const unsigned int InputWidth_Bits     = 14;
const unsigned int InputWidth_Bits_Ddc = 16;
const unsigned int AdcLaneCount        = 4;

//---------------------------------------------------------------------------
//  constructor for class AD9689_ADC
//---------------------------------------------------------------------------

AD9689_ADC::AD9689_ADC(IRequires_AD9689_ADC_Io * mdc, unsigned int jesd_device_id)
    :  LanePN_Inverted(AdcLaneCount), PhysToLogical_LaneMap(AdcLaneCount),
       MDC(mdc), DeviceID(jesd_device_id)
{
}

//---------------------------------------------------------------------------
//  destructor for class AD9689_ADC
//---------------------------------------------------------------------------

AD9689_ADC::~AD9689_ADC()
{
}

//---------------------------------------------------------------------------
//  AD9689_ADC::ConstructCustomParts()
//---------------------------------------------------------------------------

void  AD9689_ADC::ConstructCustomParts()
{
    if (!DeviceEnabled())
        return;

    MDC->SetAdcLaneConfiguration(DeviceID);
}

//---------------------------------------------------------------------------
//  AD9689_ADC::OpenHardware()
//---------------------------------------------------------------------------

void  AD9689_ADC::OpenHardware()
{
    if (!DeviceEnabled())
        return;

}

//---------------------------------------------------------------------------
//  AD9689_ADC::ConfigureHardware()
//---------------------------------------------------------------------------

void  AD9689_ADC::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;

    AdcInit();
    ConfigureAdcInputs();
    ConfigureJesdParams();
    ConfigureDdcParams();
    ConfigureCommonRegs();
}

//---------------------------------------------------------------------------
//  AD9689_ADC::Bits() - return bit width of data
//---------------------------------------------------------------------------

unsigned int AD9689_ADC::Bits()
{
    return (IsInDdcMode()) ? InputWidth_Bits_Ddc : InputWidth_Bits;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::OverrideChannelEnabled() - change physical channel enabled
//---------------------------------------------------------------------------
//  Must preload booleans with current physical channel enabled setting

void AD9689_ADC::OverrideChannelEnabled()
{
    bool Ch0Enabled = ChannelEnabled(0);
    bool Ch1Enabled = ChannelEnabled(1);

    // Override the channel enable in DDC mode
    switch (AdcChipOperatingMode)
        {
        default:
        case comFullBandwidth:
            // keep physical settings as-is
            break;
        case comDdc0:
            Ch0Enabled = (DDC[0].Iinput==0) || (DDC[0].Qinput==0) || DDC[0].Ioutput;
            Ch1Enabled = (DDC[0].Iinput==1) || (DDC[0].Qinput==1) || DDC[0].Qoutput;
            break;
        case comDdc0Ddc1:
            Ch0Enabled = false;
            Ch1Enabled = false;
            for (unsigned int i=0; i<2; i++)
                {
                Ch0Enabled = Ch0Enabled || (DDC[i].Iinput==0) || (DDC[i].Qinput==0) || DDC[i].Ioutput;
                Ch1Enabled = Ch1Enabled || (DDC[i].Iinput==1) || (DDC[i].Qinput==1) || DDC[i].Qoutput;
                }
            break;
        case comAllFour:
            Ch0Enabled = false;
            Ch1Enabled = false;
            for (unsigned int i=0; i<4; i++)
                {
                Ch0Enabled = Ch0Enabled || (DDC[i].Iinput==0) || (DDC[i].Qinput==0) || DDC[i].Ioutput;
                Ch1Enabled = Ch1Enabled || (DDC[i].Iinput==1) || (DDC[i].Qinput==1) || DDC[i].Qoutput;
                }
            break;
        }

    ChannelEnabled(0, Ch0Enabled);
    ChannelEnabled(1, Ch1Enabled);
}

//---------------------------------------------------------------------------
//  AD9689_ADC::AdcInit()
//---------------------------------------------------------------------------

void AD9689_ADC::AdcInit()
{
    Debug.Log("AD9689_ADC::AdcInit()");
    // Soft reset the ADC
    Spi(0x0, 0x81);
    Sleep(20);

    // Datapath Soft reset
    Spi(0x1, 0x2);
    Sleep(5);

    // Power down the link
    Spi(0x571, 0x15);

    // Read ADC chip_type
    Debug.Log("Adc Chip Type =", Spi(0x3), 16);

    // Read ADC chip_id
    Debug.Log("Adc Chip ID (low byte) =", Spi(0x4), 16);
    Debug.Log("Adc Chip ID (high byte) =", Spi(0x5), 16);

    // Read ADC chip_grade
    Debug.Log("Adc Chip Grade =", Spi(0x6), 16);

    // Read clock status
    Debug.Log("Adc Clock Status =", Spi(0x11C), 16);

    Debug.Log("end AD9689_ADC::AdcInit()");
}

//---------------------------------------------------------------------------
//  AD9689_ADC::ConfigureAdcInputs()
//---------------------------------------------------------------------------

void AD9689_ADC::ConfigureAdcInputs()
{
    Debug.Log("AD9689_ADC::ConfigureAdcInputs()");

    if (ActiveChannels()==2)
        {
        // Select Channels A and B and power them up
        Spi(0x8, 0x3);
        Spi(0x2, 0x0);
        }
    else
        {
        if (ChannelEnabled(0)) // Channel A
            {
            // Select Channel B and power it down
            Spi(0x8, 0x2);
            Spi(0x2, 0x3);
            // Select Channel A and power it up
            Spi(0x8, 0x1);
            Spi(0x2, 0x0);
            }
        else                   // Channel B
            {
            // Select Channel A and power it down
            Spi(0x8, 0x1);
            Spi(0x2, 0x3);
            // Select Channel B and power it up
            Spi(0x8, 0x2);
            Spi(0x2, 0x0);
            if (!IsInDdcMode())
              // Swap output channels
              Spi(0x564, 0x1);
            }
        }

    if (ChannelEnabled(0))
        ConfigureAdcCh(0);
    if (ChannelEnabled(1))
        ConfigureAdcCh(1);
    if (ActiveChannels()==2)
        {
        // Select Channels A and B
        Spi(0x8, 0x3);
        }

    Debug.Log("end AD9689_ADC::ConfigureAdcInputs()");
}

//---------------------------------------------------------------------------
//  AD9689_ADC::ConfigureAdcCh()
//---------------------------------------------------------------------------

void AD9689_ADC::ConfigureAdcCh(unsigned int AdcChIndex)
{
    // Select the channel
    Spi(0x8, (1 << AdcChIndex));

    MemoryBitManipulator Data(0);

    if (AdcCh[AdcChIndex].OverrideChannelSetting)
        {
        // Analog input differential termination
        Data.Value(Spi(0x16));
        Data.Field(4, 4, AdcCh[AdcChIndex].InputTermination);
        Spi(0x16, Data.Value());

        // Buffer control1
        Data.Value(Spi(0x18));
        Data.Field(4, 4, AdcCh[AdcChIndex].BufferControl1);
        Spi(0x18, Data.Value());

        // Buffer control2
        Data.Value(Spi(0x19));
        Data.Field(4, 4, AdcCh[AdcChIndex].BufferControl2);
        Spi(0x19, Data.Value());

        // High frequency setting
        Data.Value(Spi(0x11A));
        Data.Bit(5, AdcCh[AdcChIndex].HighFreqSetting);
        Spi(0x11A, Data.Value());

        // Input capacitance
        Data.Value(Spi(0x934));
        Data.Field(0, 5, AdcCh[AdcChIndex].InputCapacitance);
        Spi(0x934, Data.Value());

        // Low frequency operation
        Data.Value(Spi(0x935));
        Data.Bit(2, AdcCh[AdcChIndex].LowFreqOperation);
        Spi(0x935, Data.Value());
        }

    // Set the selected test mode
    Spi(0x550, (AdcCh[AdcChIndex].AdcTestMode & 0x0F));
}

//---------------------------------------------------------------------------
//  AD9689_ADC::ConfigureJesdParams()
//---------------------------------------------------------------------------

void AD9689_ADC::ConfigureJesdParams()
{
    Debug.Log("AD9689_ADC::ConfigureJesdParams()");

    unsigned int OctetsPerFrame;
    switch (NumConverters)
        {
        default:
        case 1:
            OctetsPerFrame = 2;
            break;
        case 2:
            OctetsPerFrame = (NumLanes == 1)? 4 : 2;
            break;
        case 4:
            if (NumLanes == 4)
                OctetsPerFrame = 2;
            else if (NumLanes == 2)
                OctetsPerFrame = 4;
            else
                OctetsPerFrame = 8;
            break;
        case 8:
            if (NumLanes == 4)
                OctetsPerFrame = 4;
            else if (NumLanes == 2)
                OctetsPerFrame = 8;
            else
                OctetsPerFrame = 16;
            break;
        }

    // Set the octets per frame F
    JesdAdc().Poke(0x20, (OctetsPerFrame-1));

    MemoryBitManipulator Data (0);
    Data.Field(0,3, log2(OctetsPerFrame));
    Data.Field(3,3, log2(NumConverters));
    Data.Field(6,2, log2(NumLanes));

    // JESD204B quick configuration L=4, M=1, F=2
    Spi(0x570, Data.Value());

    Debug.Log(" end AD9689_ADC::ConfigureJesdParams()");
}

//---------------------------------------------------------------------------
//  AD9689_ADC::ConfigureDdcParams()
//---------------------------------------------------------------------------

void AD9689_ADC::ConfigureDdcParams()
{
    if (IsInDdcMode()== false)
        return;
    Debug.Log("AD9689_ADC::ConfigureDdcParams()");

    switch (AdcChipOperatingMode)
        {
        default :
        case comFullBandwidth:
            break;
        case comDdc0:
            ConfigureDdcRegisters(0);
            break;
        case comDdc0Ddc1:
            ConfigureDdcRegisters(0);
            ConfigureDdcRegisters(1);
            break;
        case comAllFour:
            ConfigureDdcRegisters(0);
            ConfigureDdcRegisters(1);
            ConfigureDdcRegisters(2);
            ConfigureDdcRegisters(3);
            break;
        }

    // Set the DDC synchronization mode and soft reset the DDC NCO
    MemoryBitManipulator Data(0);
    Data.Bit(4, 1);
    Data.Field(0, 2, DdcSyncMode);
    Spi(0x300, Data.Value());
    Data.Bit(4, 0);
    Spi(0x300, Data.Value());

    Debug.Log("end AD9689_ADC::ConfigureDdcParams()");
}

//---------------------------------------------------------------------------
//  AD9689_ADC::ConfigureDdcRegisters()
//---------------------------------------------------------------------------

void AD9689_ADC::ConfigureDdcRegisters(unsigned int DdcIndex)
{
    unsigned int BaseAddress = (DdcIndex * 0x20) + 0x310;

    MemoryBitManipulator Data(0);

    Data.Bit(7, (DDC[DdcIndex].Mixer) ? 1 : 0);
    Data.Bit(6, (DDC[DdcIndex].Gain) ? 1 : 0);
    Data.Field(4, 2, DDC[DdcIndex].IFintermediateFreq);
    Data.Bit(3, DDC[DdcIndex].ComplexToReal);
    Data.Field(0, 2, DDC[DdcIndex].DecimationRate);
    // DDC control
    Spi(BaseAddress, Data.Value());

    Data.Value(0);
    Data.Bit(2, (DDC[DdcIndex].Qinput) ? 1 : 0);
    Data.Bit(0, (DDC[DdcIndex].Iinput) ? 1 : 0);
    // DDC input selection
    Spi((BaseAddress+1), Data.Value());

    Data.Value(0);
    Data.Field(0, 8, DDC[DdcIndex].NCOfreq);
    // DDC frequency LSB
    Spi((BaseAddress+4), Data.Value());

    Data.Value(0);
    Data.Field(0, 4, (DDC[DdcIndex].NCOfreq >> 8));
    // DDC frequency MSB
    Spi((BaseAddress+5), Data.Value());

    Data.Value(0);
    Data.Field(0, 8, DDC[DdcIndex].NCOphase);
    // DDC phase LSB
    Spi((BaseAddress+0x10), Data.Value());

    Data.Value(0);
    Data.Field(0, 4, (DDC[DdcIndex].NCOphase >> 8));
    // DDC phase MSB
    Spi((BaseAddress+0x11), Data.Value());

    Data.Value(0);
    Data.Bit(2, DDC[DdcIndex].Qoutput);
    Data.Bit(0, DDC[DdcIndex].Ioutput);
    // DDC output test mode selection
    Spi((BaseAddress+0x17), Data.Value());
}

//---------------------------------------------------------------------------
//  AD9689_ADC::ConfigureCommonRegs()
//---------------------------------------------------------------------------

void AD9689_ADC::ConfigureCommonRegs()
{
    Debug.Log("AD9689_ADC::ConfigureCommonRegs()");

    // Set FastDetect pin functions (to be changed later)
    Spi(0x40, 0x11);

    if (IsInDdcMode())
        // Enable N shot sysref
        Spi(0x120, 0x4);
    else
        // Enable continuous sysref
        Spi(0x120, 0x2);
    JesdAdc().Poke(0x10, 0x0);

    // Select normal synchronization mode
    Spi(0x1FF, 0x0);

    unsigned int data = (ChipQignore << 5) | (AdcChipOperatingMode & 3);
    // Set Chip operating mode
    Spi(0x200, data);

    data = log2(ChipDecimationRate);
    // Set Chip decimation ratio select
    Spi(0x201, data);

    // Set the user pattern registers
    Spi(0x551, (AdcTestPattern1 & 0xFF));
    Spi(0x552, ((AdcTestPattern1 >> 8) & 0xFF));
    Spi(0x553, (AdcTestPattern2 & 0xFF));
    Spi(0x554, ((AdcTestPattern2 >> 8) & 0xFF));
    Spi(0x555, (AdcTestPattern3 & 0xFF));
    Spi(0x556, ((AdcTestPattern3 >> 8) & 0xFF));
    Spi(0x557, (AdcTestPattern4 & 0xFF));
    Spi(0x558, ((AdcTestPattern4 >> 8) & 0xFF));

    // Set CS bit1 to overrange (Only used when CS = 2 or 3)
    Spi(0x559, 0x10);
    // Set CS bit2 to sysref (Only used when CS = 1, 2, or 3)
    Spi(0x55A, 0x5);

    // Set the data format to twos complement
    Spi(0x561, 0x1);

    // Set the serial line rate control
    if (LineRate <= 6250.0e6)
      Spi(0x56E, 0x10);
    else
      Spi(0x56E, 0x0);

    // Enable 8b10b inversion when all ADC SerDes P/N lanes are inverted
    if (AllLanesInverted())
        Spi(0x572, 0x2);
    else
        Spi(0x572, 0x0);

    // Disable JESD204B test mode patterns
    Spi(0x573, 0x0);

    // Enable scrambling
    Spi(0x58B, 0x80);
    JesdAdc().Poke(0xC, 0x1);

    // Set number of frames per multiframe K=32
    Spi(0x58D, 0x1F);
    JesdAdc().Poke(0x24, 0x1F);

    if (IsInDdcMode())
        Spi(0x58F, 0x0F); // 16-bit resolution
    else
        Spi(0x58F, 0x8D); // 14-bit resolution, 2 control bits

    // Select Subclass1 and 16-bit samples
    Spi(0x590, 0x2F);
    JesdAdc().Poke(0x2C, 0x1);

    // Enable ILA
    JesdAdc().Poke(0x8, 0x1);

    // Reset the JESD204B IP
    JesdAdc().Poke(0x4, 0x1);

    // Turn-off the unused lanes
    JesdAdc().Poke(0x28, ((1 << NumLanes)-1));
    MemoryBitManipulator LanePowerDown(0xAA);
    for (unsigned int i=0; i<PhysToLogical_LaneMap.size(); i++)
        LanePowerDown.Bit((2*i), (PhysToLogical_LaneMap[i]<NumLanes) ? false : true);
    Spi(0x5B0, LanePowerDown.Value());

    // Set JESD lane ID and mapping
    Spi(0x583, PhysToLogical_LaneMap[0]);
    Spi(0x584, PhysToLogical_LaneMap[1]);
    Spi(0x585, PhysToLogical_LaneMap[2]);
    Spi(0x586, PhysToLogical_LaneMap[3]);
    Spi(0x5B2, PhysToLogical_LaneMap[0]);
    Spi(0x5B3, PhysToLogical_LaneMap[1]);
    Spi(0x5B5, PhysToLogical_LaneMap[2]);
    Spi(0x5B6, PhysToLogical_LaneMap[3]);

    // Adjust the JESD serializer driver voltage swing
    if (LineRate > 6250.0e6)
      Spi(0x5BF, 0xF);

    // Power up the link with ILA enabled
    Spi(0x571, 0x14);
    Sleep(20);

    // Reset and enable the link error counter
    JesdAdc().Poke(0x34, 0x0);
    JesdAdc().Poke(0x34, 0x101);

    // Read ADC status registers
    Debug.Log("Adc Sysref Setup and Hold status Reg(0x128) =", Spi(0x128), 16);
    Debug.Log("Adc PLL lock status Reg(0x56F) =", Spi(0x56F), 16);
    Debug.Log("Adc JESD204B F parameter Reg(0x58C) =", Spi(0x58C), 16);
    Debug.Log("Adc JESD204B M parameter Reg(0x58E) =", Spi(0x58E), 16);
    Debug.Log("Adc JESD204B S parameter Reg(0x591) =", Spi(0x591), 16);
    Debug.Log("Adc JESD204B HD and CF parameters Reg(0x592) =", Spi(0x592), 16);

    JesdAdc().DumpDebugStatus();

    Debug.Log("end AD9689_ADC::ConfigureCommonRegs()");
}

//---------------------------------------------------------------------------
//  AD9689_ADC::AllLanesInverted() --
//---------------------------------------------------------------------------

bool AD9689_ADC::AllLanesInverted()
{
    for (unsigned int i=0; i<LanePN_Inverted.size(); i++)
        {
        if (!LanePN_Inverted[i])
            return false;
        }
    return true;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::CalcAdcParams() --
//---------------------------------------------------------------------------
//  Calculate the Chip Decimation, Qignore,
//  and number of virtual converters

void AD9689_ADC::CalcAdcParams()
{
    const unsigned int MaxChipDecimationRate = 16;
    switch (AdcChipOperatingMode)
        {
        default:
        case comFullBandwidth:
            ChipDecimationRate = (1 << AdcFullBwDecimationRate);
            ChipQignore   = false;
            NumConverters = ((ActiveChannels()==0)? 1 : ActiveChannels());
            break;
        case comDdc0:
            ChipDecimationRate = CalcDdcDecRate(0);
            ChipQignore   = DDC[0].ComplexToReal;
            NumConverters = ChipQignore? 1 : 2;
            break;
        case comDdc0Ddc1:
            ChipDecimationRate = MaxChipDecimationRate;
            for (unsigned int i=0; i<2; i++)
                {
                if (CalcDdcDecRate(i) < ChipDecimationRate)
                    ChipDecimationRate = CalcDdcDecRate(i);
                }
            ChipQignore   = DDC[0].ComplexToReal && DDC[1].ComplexToReal;
            NumConverters = ChipQignore? 2 : 4;
            break;
        case comAllFour:
            ChipDecimationRate = MaxChipDecimationRate;
            for (unsigned int i=0; i<4; i++)
                {
                if (CalcDdcDecRate(i) < ChipDecimationRate)
                    ChipDecimationRate = CalcDdcDecRate(i);
                }
            ChipQignore   = DDC[0].ComplexToReal && DDC[1].ComplexToReal &&
                            DDC[2].ComplexToReal && DDC[3].ComplexToReal;
            NumConverters = ChipQignore? 4 : 8;
            break;
        }
}

//---------------------------------------------------------------------------
//  AD9689_ADC::CalcDdcDecRate() -- Calculate the DDC rate value
//---------------------------------------------------------------------------

unsigned int AD9689_ADC::CalcDdcDecRate(unsigned int DdcIndex)
{
    unsigned int DdcDecRate;
    switch (DDC[DdcIndex].DecimationRate)
        {
        default:
        case 3:
            DdcDecRate = (DDC[DdcIndex].ComplexToReal) ? 1 : 2;
            break;
        case 0:
            DdcDecRate = (DDC[DdcIndex].ComplexToReal) ? 2 : 4;
            break;
        case 1:
            DdcDecRate = (DDC[DdcIndex].ComplexToReal) ? 4 : 8;
            break;
        case 2:
            DdcDecRate = (DDC[DdcIndex].ComplexToReal) ? 8 : 16;
            break;
        }
    return DdcDecRate;
}


//---------------------------------------------------------------------------
//  AD9689_ADC::CalcAdcJesdLineRate() -- Calculate the ADC JESD line rate
//---------------------------------------------------------------------------

double AD9689_ADC::CalcAdcJesdLineRate(double SamplingRate)
{
    CalcAdcParams();

    const unsigned int MaxNumLanes = 4;
    const double MaxLineRate = 12500.0e6;   // from the Adc datasheet
    const double MinLineRate = 3125.0e6;    // from the Adc datasheet

    // Reduce the number of lanes if the calculated line rate is less than
    // the minimum rate supported by the Adc
    // Select the maximum number of lanes that results in a line rate within range
    // The Adc supports NumLanes values of 4, 2, 1
    for (unsigned int i=MaxNumLanes; i>=1; i/=2)
        {
        LineRate = JesdAdc().CalcJesdLineRate(NumConverters, 2, SamplingRate, ChipDecimationRate, i);
        if ((LineRate >= MinLineRate && LineRate <= MaxLineRate) || (i==1))
            {
            NumLanes = i;
            break;
            }
        }
    return LineRate;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::IsInDdcMode() -- User friendly call
//---------------------------------------------------------------------------

bool AD9689_ADC::IsInDdcMode()
{
    return (AdcChipOperatingMode==comFullBandwidth) ? false : true;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::Spi() -- write to a SPI device
//---------------------------------------------------------------------------

void  AD9689_ADC::Spi(unsigned int address, unsigned int data)
{
    MDC->AdcSpi(DeviceID, address, data);
}

//---------------------------------------------------------------------------
//  AD9689_ADC::Spi() -- read from a SPI device
//---------------------------------------------------------------------------

unsigned int  AD9689_ADC::Spi(unsigned int address)
{
    return MDC->AdcSpi(DeviceID, address);
}

//---------------------------------------------------------------------------
//  AD9689_ADC::ChannelEnabled() -- read cached channel enable state
//---------------------------------------------------------------------------

bool  AD9689_ADC::ChannelEnabled(int ch)
{
    return MDC->AdcChannelEnabled(ch);
}

//---------------------------------------------------------------------------
//  AD9689_ADC::ChannelEnabled() -- set channel enable state
//---------------------------------------------------------------------------

void AD9689_ADC::ChannelEnabled(int ch, bool state)
{
    return MDC->AdcChannelEnabled(ch, state);
}

//---------------------------------------------------------------------------
//  AD9689_ADC::ActiveChannels() -- count enabled channels
//---------------------------------------------------------------------------

unsigned int  AD9689_ADC::ActiveChannels()
{
    return MDC->AdcActiveChannels();
}

//---------------------------------------------------------------------------
//  AD9689_ADC::log2()  -- Calculate Log2
//---------------------------------------------------------------------------

unsigned int AD9689_ADC::log2(unsigned int value)
{
    unsigned int log = 0;
    while (value != 0)
        {
        value >>= 1;
        log++;
        }
    return log-1;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::EffectiveAdcChannels() -- Set the number of channels Per SID for Binview
//---------------------------------------------------------------------------

unsigned int AD9689_ADC::EffectiveAdcChannels()
{
    return NumConverters;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::EffectiveAdcDevices() -- Set the number of SIDs for Binview
//---------------------------------------------------------------------------

unsigned int AD9689_ADC::EffectiveAdcDevices()
{
    return 1;
}


//---------------------------------------------------------------------------
//  AD9689_ADC::SetAdcChipOperatingMode() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetAdcChipOperatingMode(const unsigned int Mode)
{
    AdcChipOperatingMode = Mode;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetAdcFullBwDecimationRate() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetAdcFullBwDecimationRate(const unsigned int value)
{
    AdcFullBwDecimationRate = value;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetDdcSyncMode() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetDdcSyncMode(const unsigned int Mode)
{
    DdcSyncMode = Mode;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::AdcDeviceTestPattern() - set device user test pattern
//---------------------------------------------------------------------------

void AD9689_ADC::AdcDeviceTestPattern(unsigned int pattern1, unsigned int pattern2,
                                              unsigned int pattern3, unsigned int pattern4)
{
    AdcTestPattern1 = pattern1;
    AdcTestPattern2 = pattern2;
    AdcTestPattern3 = pattern3;
    AdcTestPattern4 = pattern4;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetOverrideChannelSetting() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetOverrideChannelSetting(const bool value,const unsigned int AdcChIndex)
{
    AdcCh[AdcChIndex].OverrideChannelSetting = value;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetHighFreqSetting() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetHighFreqSetting(const bool value,const unsigned int AdcChIndex)
{
    AdcCh[AdcChIndex].HighFreqSetting = value;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetLowFreqOperation() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetLowFreqOperation(const bool value,const unsigned int AdcChIndex)
{
    AdcCh[AdcChIndex].LowFreqOperation = value;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetBufferControl1() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetBufferControl1(const unsigned int value,const unsigned int AdcChIndex)
{
    AdcCh[AdcChIndex].BufferControl1 = value;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetBufferControl2() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetBufferControl2(const unsigned int value,const unsigned int AdcChIndex)
{
    AdcCh[AdcChIndex].BufferControl2 = value;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetInputCapacitance() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetInputCapacitance(const unsigned int value,const unsigned int AdcChIndex)
{
    AdcCh[AdcChIndex].InputCapacitance = value;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetInputTermination() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetInputTermination(const unsigned int value,const unsigned int AdcChIndex)
{
    AdcCh[AdcChIndex].InputTermination = value;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetAdcTestMode() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetAdcTestMode(const unsigned int value,const unsigned int AdcChIndex)
{
    AdcCh[AdcChIndex].AdcTestMode = value;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetMixer() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetMixer(const unsigned int Mode,const unsigned int DdcIndex)
{
    DDC[DdcIndex].Mixer = Mode;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetGain() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetGain(const unsigned int Mode,const unsigned int DdcIndex)
{
    DDC[DdcIndex].Gain = Mode;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetIFintermediateFreq() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetIFintermediateFreq(const unsigned int Mode,const unsigned int DdcIndex)
{
    DDC[DdcIndex].IFintermediateFreq = Mode;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetDecimationRate() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetDecimationRate(const unsigned int Mode,const unsigned int DdcIndex)
{
    DDC[DdcIndex].DecimationRate = Mode;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetComplexToReal() --
//---------------------------------------------------------------------------

void AD9689_ADC::SetComplexToReal(const bool enabled,const unsigned int DdcIndex)
{
   DDC[DdcIndex].ComplexToReal = enabled;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetNCOfreq() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetNCOfreq(const int Mode,const unsigned int DdcIndex)
{
   DDC[DdcIndex].NCOfreq = Mode;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetNCOphase() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetNCOphase(const int Mode,const unsigned int DdcIndex)
{
    DDC[DdcIndex].NCOphase = Mode;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetQinput() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetQinput(const unsigned int Mode,const unsigned int DdcIndex)
{
    DDC[DdcIndex].Qinput = Mode;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetIinput() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetIinput(const unsigned int Mode,const unsigned int DdcIndex)
{
    DDC[DdcIndex].Iinput = Mode;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetQoutput() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetQoutput(const bool enabled,const unsigned int DdcIndex)
{
    DDC[DdcIndex].Qoutput = enabled;
}

//---------------------------------------------------------------------------
//  AD9689_ADC::SetIinput() --
//---------------------------------------------------------------------------

void  AD9689_ADC::SetIoutput(const bool enabled,const unsigned int DdcIndex)
{
    DDC[DdcIndex].Ioutput = enabled;
}

}  // namespace Innovative
