//
// ADF4355_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2016 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <ADF4355_Mb.h>
#include <SystemSupport_Mb.h>
#include <VcxoControl_Mb.h>
#include <algorithm>
#include <limits>
#include <cmath>
#include <Magic_Mb.h>
#include <Trace_Mb.h>
#include <sstream>
#include <FloatingPtSupport_Mb.h>

using namespace std;

namespace Innovative
{

//=============================================================================
//  CLASS ADF4355  --  ADF4355 Pll Implementation
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class ADF4355
//-----------------------------------------------------------------------------

const unsigned int  TotalReferences  = 1;
const unsigned int  TotalOutputs     = 2;
const unsigned int  TotalOutPairs    = TotalOutputs>>1;  // the single output is a pair....?

const int           SpiRegisterCount = 14;

const double        MinInputReferenceFrequency         = 10.0e6;
const double        MaxInputReferenceFrequency_SE      = 250.0e6;
const double        MaxInputReferenceFrequency_DF      = 600.0e6;
const double        MaxInputReferenceFrequency_Doubler = 100.0e6;

const double        MaxPhaseDetectorFrequency = 125.0e6;

const double        MinOutputFrequency           = 53.125e6;
const double        MaxOutputFrequency_adf4355   = 6800.0e6;
const double        MaxOutputFrequency_adf4355_2 = 4400.0e6;
const double        MaxOutputFrequency_adf4356   = 6800.0e6;

namespace {

struct VcoRangesType {
    double Low;
    double High;
};

VcoRangesType ADF4355_VcoRanges[] = {
    {3400.0e6, 6800.0e6},    // adf4355
    {3400.0e6, 6800.0e6},    // adf4355_2
    {3400.0e6, 6800.0e6},    // adf4356
};

const unsigned int  TotalVcoTypes = sizeof(ADF4355_VcoRanges)/sizeof(VcoRangesType) ;


}  // nameless namespace

const unsigned int  kR_Min   =  1;
const unsigned int  kR_Max   =  1023;

const unsigned int  kN_INT_Min_Ps45   =  23;      // min N_INT for prescaler 4/5
const unsigned int  kN_INT_Max_Ps45   =  32767;
const unsigned int  kN_INT_Min_Ps89   =  75;      // min N_INT for prescaler 8/9
const unsigned int  kN_INT_Max_Ps89   =  65535;

const unsigned int  kN_FRAC1_Min      =  0;
const unsigned int  kN_FRAC1_Max      =  16777215;

const unsigned int  kN_FRAC2_Min           =  0;
const unsigned int  kN_FRAC2_Max_adf4355   =  16383;
const unsigned int  kN_FRAC2_Max_adf4355_2 =  16383;
const unsigned int  kN_FRAC2_Max_adf4356   =  268435455;

const unsigned int  kN_MOD2_Min            =  2;
const unsigned int  kN_MOD2_Max_adf4355    =  16383;
const unsigned int  kN_MOD2_Max_adf4355_2  =  16383;
const unsigned int  kN_MOD2_Max_adf4356    =  268435455;

//  Output Dividers
//const unsigned int  kOD_Min   =  1;
const unsigned int  kOD_Max   =  64;

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif
//-----------------------------------------------------------------------------
//  constructor for class ADF4355
//-----------------------------------------------------------------------------

ADF4355::ADF4355(IAddressingSpace & space, ADF4355Registers & regs)
    : PllTimebaseDevice(TotalReferences, TotalOutputs),
      FChipType(adf4356),
      FReferenceSource(refPrimary),
      RegisterImage(SpiRegisterCount),
      FDeviceRouting(Dev_Both),
      LocalVcxo_Standard(*this),
      ControlReg(space, regs.Control_Addr),
      SpiReg(space, regs.SpiIntf_Addr),
      FPowered(false),
      Debug(true)
{
    //  Fill max frequency checker array
    double DefaultMaxOutputFreq;
    if (FChipType==adf4355)
        DefaultMaxOutputFreq = MaxOutputFrequency_adf4355;
    else if (FChipType==adf4355_2)
        DefaultMaxOutputFreq = MaxOutputFrequency_adf4355_2;
    else
        DefaultMaxOutputFreq = MaxOutputFrequency_adf4356;

    for (unsigned int i=0; i<TotalOutputs; i++)
        {
        OutputFrequencyLimit.push_back(static_cast<float>(DefaultMaxOutputFreq));
        }
    //
    //  Set Default Input Configuration
    SourceInUse(0, true);    SetSourceMapping(0, 0);   // primary = internal

    SetVcoRange(FChipType);

    InitDividerSupport();

    delay_us = 160;
}

#ifdef _MSC_VER
	#pragma warning(pop)
#endif

//-----------------------------------------------------------------------------
//  destructor for class ADF4355
//-----------------------------------------------------------------------------

ADF4355::~ADF4355()
{
}

//-----------------------------------------------------------------------------
//  ADF4355::Init()  --  Initialize Chip
//-----------------------------------------------------------------------------

void ADF4355::Init()
{
    ScopedTraceRef ccc(Debug, "ADF4355::Init()");

    SetDeviceRouting(FDeviceRouting);
    SetDevicePowerDown(FDeviceRouting, true);
    SetDeviceEnable(FDeviceRouting, false);

    UpdateMode = updFrequency;      // always in frequency mode...

    FPhaseDetectorFrequency = DefaultPhaseDetectorFrequency();
    InitDataRegisters();   // define initial state

    OnInit();
}

//-----------------------------------------------------------------------------
//  ADF4355::InitDataRegisters()  --  Initialize all data registers into Image vector
//-----------------------------------------------------------------------------
//  On this device we can't read the true registers, but unused fields and
//  registers must have certain codes in them. Set things up here.

void ADF4355::InitDataRegisters()
{
    RegisterImage[0]  = 0x00000000;
    RegisterImage[1]  = 0x00000001;
    RegisterImage[2]  = 0x00000002;
    RegisterImage[3]  = 0x00000003;
    RegisterImage[4]  = 0x00000004;
    RegisterImage[5]  = 0x00800025;
    RegisterImage[6]  = 0x14000006;
    RegisterImage[7]  = 0x10000007;
    RegisterImage[8]  = 0x102d0428;
    RegisterImage[9]  = 0x00000009;
    RegisterImage[10] = 0x00c0000a;
    RegisterImage[11] = 0x0061300b;
    RegisterImage[12] = 0x0000041c;
    RegisterImage[13] = 0x0000000d;

    // Register 3
    SetOutputPhase(0);
    SetOutputPhaseAdjustEnabled(false);
    SetOutputPhaseResyncEnabled(false);
    SetSigmaDeltaLoadResetDisabled(false);

    // Register 4
    SetCounterReset(false);
    SetCpThreeStateEnabled(false);
    SetPowerDown(false);
    SetPhaseDetectorPolarity(pdpPositive);
    SetMuxoutLevelSelect(ldp_1p8v);
    SetReferenceMode(refDifferential);
    SetChargePumpCurrent(cpc_0p94);
    SetReg6_DoubleBuffering(true);
    SetMuxOut(muxDigitalLock);

    // Register 6
    SetRfOutputPower(out_p5dbm);
    SetAuxOutputPower(out_n1dbm);
    SetMTLDEnabled(false);
    SetChargePumpBleedCurrent(90);      // just initialize (will be overridden later)
    SetFeedbackSelect(fsDivided);
    SetBleedCurrentEnabled(true);
    SetGatedBleedEnabled(false);
    SetBleedPolarity(false);

    // Register 7
    SetLockDetectMode(ldmFractionalN);
    SetFractionalNPrecision(ldp_12ns);
    SetLossOfLockModeEnabled(false);
    SetLossOfLockCycleCount(lol_1024);
    SetLESyncEnabled(true);

    // Register 9
    SetReg9(MaxPhaseDetectorFrequency); // just initialize (will be overridden later)

    // Register 10
    SetReg10(MaxPhaseDetectorFrequency);// just initialize (will be overridden later)

    // Register 12
    PhaseResyncClockDivider(0);
}

//-----------------------------------------------------------------------------
//  ADF4355::WriteDataRegisters()  -- Write config data
//-----------------------------------------------------------------------------

void ADF4355::WriteDataRegisters()
{
    ScopedTraceRef ccc(Debug, "ADF4355::WriteDataRegisters()");

    for (int reg=GetRegCount()-1; reg>=0; reg--)
        Port(reg, RegisterImage[reg]);

    SetCounterReset(true);
    Port(4, RegisterImage[4]);    // counter reset enabled

    SetVcoAutocal(false);
    Port(0, RegisterImage[0]);    // auto calibration disabled

    SetCounterReset(false);
    Port(4, RegisterImage[4]);    // counter reset disabled

    // Wait > 16 ADC_CLK cycles
    uSleep(delay_us);

    SetVcoAutocal(true);
    Port(0, RegisterImage[0]);    // auto calibration enabled
}

//-----------------------------------------------------------------------------
//  ADF4355::Reapply()  --  Reapply settings to hardware from saved state
//-----------------------------------------------------------------------------

void ADF4355::Reapply()
{
    ScopedTraceRef ccc(Debug, "ADF4355::Reapply()");
    OnReapply();

    SetDeviceRouting(FDeviceRouting);
    SetDevicePowerDown(FDeviceRouting, false);
    SetDeviceEnable(FDeviceRouting, true);

    Update();              // Update state of the device based on sources

    WriteDataRegisters();  // Write the data registers
}

//-----------------------------------------------------------------------------
//  ADF4355::ReferenceSource()  --  Change Clock Reference
//-----------------------------------------------------------------------------

void ADF4355::ReferenceSource(AdfReferenceSource value)
{
    FReferenceSource = value;
    SetReferenceSource(value);
}

//-----------------------------------------------------------------------------
//  ADF4355::MaxOutputDivider()  --  Return max output post-divider
//-----------------------------------------------------------------------------

int ADF4355::MaxOutputDivider() const
{
    return kOD_Max;
}

//-----------------------------------------------------------------------------
//  ADF4355::N_Mod2_Max()  --  chip dependent Max Mod2
//-----------------------------------------------------------------------------

unsigned int ADF4355::N_Mod2_Max()
{
    if (FChipType==adf4355)
        return kN_MOD2_Max_adf4355;
    else if (FChipType==adf4355_2)
        return kN_MOD2_Max_adf4355_2;
    else // adf4356
        return kN_MOD2_Max_adf4356;
}

//-----------------------------------------------------------------------------
//  ADF4355::FrequencyModeUpdate()  --  Update using Frequency as baseline
//-----------------------------------------------------------------------------

void ADF4355::FrequencyModeUpdate()
{
    ScopedTraceRef ccc(Debug, "ADF4355::FrequencyModeUpdate()");

    if (Reference() == 0 || EffectiveFrequency() == 0)
        return;

    Debug.Log("Reference() = ", Reference());
    Debug.Log("EffectiveFrequency() = ", EffectiveFrequency());

    // Calculate and maximize the PFD
    unsigned int RMul2, RDiv2, R;

    if (Reference() < MaxInputReferenceFrequency_Doubler)
        RMul2 = 2;
    else
        RMul2 = 1;

    // Initialize
    R = 0;
    double PFD = min(MaxPhaseDetectorFrequency, PhaseDetectorFrequency());

    double PFD_Actual;
    do
        {
        R++;
        PFD_Actual = (Reference()* static_cast<double>(RMul2))/static_cast<double>(R);
        } while (PFD_Actual > PFD);

    if (R%2 == 0)
        {
        RDiv2 = 2;
        R /= 2;
        }
    else
        RDiv2 = 1;
    Debug.Log("RMul2 = ", RMul2);
    Debug.Log("RDiv2 = ", RDiv2);
    Debug.Log("R = ", R);
    Debug.Log("Target PFD = ", PFD);
    Debug.Log("Actual PFD = ", PFD_Actual);

    //  Calculate VCO Frequency and OD
    unsigned int OD;
    double VcoFreq;
    FindVcoRate(VcoFreq, OD);
    Debug.Log("OD = ", OD);
    Debug.Log("Initial VcoFreq = ", VcoFreq);

    //  Calculate N using the selected feedback frequency
    double N;
    if (GetFeedbackSelect() == fsDivided) // using the output frequency for the feedback leg
        N = Frequency()/PFD_Actual;
    else                                  // using the VCO frequency for the feedback leg
        N = VcoFreq/PFD_Actual;
    Debug.Log("N = ", N);

    // N_Int
    unsigned int N_Int = static_cast<unsigned int>(IntegerPart(N));
    Debug.Log("N_Int = ", N_Int);

    unsigned int N_Frac1, N_Frac2, N_Mod2;
    const unsigned int N_Mod1 = 16777216;

    double N_Rem = FractionalPart(N);
    if (IsAboutEqual(N_Rem, 0.0, 1.0e-15))
        {
        FOperatingMode = modeIntOnly;
        Debug.Log("---Integer Only Mode---");
        N_Frac1 = 0;
        N_Frac2 = 0;
        N_Mod2  = kN_MOD2_Min;  // min allowed
        }
    else
        {
        FOperatingMode = modeFractional;
        Debug.Log("---Fractional Mode---");

        double N_RemXMod1 = N_Rem * N_Mod1;

        // N_Frac1
        N_Frac1 = static_cast<unsigned int>(IntegerPart(N_RemXMod1));
        Debug.Log("N_Frac1 = ", N_Frac1);
        double N_Frac1_Rem = FractionalPart(N_RemXMod1);    // = Frac2/Mod2

        // N_Frac2, N_Mod2
        SetFChannelSpacingFrequency(DefaultChannelSpacingValue());
        N_Mod2 = Calculate_NMod2_fromSpacing(PFD_Actual, ChannelSpacingFrequency());
        N_Frac2 = static_cast<unsigned int>(IntegerPart(N_Frac1_Rem*N_Mod2));

        unsigned int gcd = Gcd(N_Mod2, N_Frac2);
        N_Mod2 /= gcd;
        N_Frac2 /= gcd;
        }

    double N_Actual = double(N_Int) + ((double(N_Frac1)+(double(N_Frac2)/double(N_Mod2)))/double(N_Mod1));
    Debug.Log("Actual N = ", N_Actual);

    Debug.Log("PFD (R Leg) = ", PFD_Actual);
    double PFD_NLeg;
    if (GetFeedbackSelect() == fsDivided) // using the output frequency for the feedback leg
        PFD_NLeg = VcoFreq/(OD*N_Actual);
    else                                  // using the VCO frequency for the feedback leg
        PFD_NLeg = VcoFreq/N_Actual;
    Debug.Log("PFD (N Leg) = ", PFD_NLeg);

    //  Allow User to Override Settings
    Debug.Log("ADF4355::RMul2 = ", RMul2);
    Debug.Log("ADF4355::RDiv2 = ", RDiv2);
    Debug.Log("ADF4355::R = ", R);
    Debug.Log("ADF4355::N_Int = ", N_Int);
    Debug.Log("ADF4355::N_Frac1 = ", N_Frac1);
    Debug.Log("ADF4355::N_Frac2 = ", N_Frac2);
    Debug.Log("ADF4355::N_Mod2 = ", N_Mod2);
    Debug.Log("ADF4355::Output Divider = ", OD);
    Debug.Log("Mid VcoFreq = ", VcoFreq);

    Debug.Log("Pll Tuning Event");
    ADF4355PllTuningEvent e(RDiv2, RMul2, R, N_Int, N_Frac1, N_Frac2, N_Mod2, OD);
    OnPllTuning.Execute(e);

    Debug.Log("Final Values");
    Debug.Log("ADF4355::RMul2 = ", RMul2);
    Debug.Log("ADF4355::RDiv2 = ", RDiv2);
    Debug.Log("ADF4355::R = ", R);
    Debug.Log("ADF4355::N_Int = ", N_Int);
    Debug.Log("ADF4355::N_Frac1 = ", N_Frac1);
    Debug.Log("ADF4355::N_Frac2 = ", N_Frac2);
    Debug.Log("ADF4355::N_Mod2 = ", N_Mod2);
    Debug.Log("ADF4355::Output Divider = ", OD);

    PFD_Actual = (Reference()* static_cast<double>(RMul2)) /
                 (static_cast<double>(R)*static_cast<double>(RDiv2));
    Debug.Log("ADF4355::Final Actual PFD = ", PFD_Actual);

    N_Actual = double(N_Int) + ((double(N_Frac1)+(double(N_Frac2)/double(N_Mod2)))/double(N_Mod1));
    Debug.Log("ADF4355::Final Actual N = ", N_Actual);

    if (GetFeedbackSelect() == fsDivided) // using the output frequency for the feedback leg
        VcoFreq = PFD_Actual * N_Actual * static_cast<double>(OD);
    else                                  // using the VCO frequency for the feedback leg
        VcoFreq = PFD_Actual * N_Actual;
    Debug.Log("ADF4355::Final VcoFreq = ", VcoFreq);

    // Load Registers
    SetRMul2(RMul2);
    SetRDiv2(RDiv2);
    SetR(R);
    SetN_Int(N_Int);
    SetN_Frac1(N_Frac1);
    SetN_Frac2(N_Frac2);
    SetN_Mod2(N_Mod2);

    // Set up the output dividers properly
    SetOutputDividers(VcoFreq, OD);

    DualModulusPrescaler prescaler;
    if (N_Int >= kN_INT_Min_Ps89 && N_Int <= kN_INT_Max_Ps89 )
        prescaler = pre_8_9;
    else
        prescaler = pre_4_5;
    SetPrescaler(prescaler);

    // Calculate the charge pump bleed current (IBleed > ICp*4/N)
    float ICp = GetChargePumpCurrent();
    float IBleed_uA = (1000*ICp*4)/static_cast<float>(N_Actual);
    SetChargePumpBleedCurrent(IBleed_uA);

    SetReg9(PFD_Actual);
    SetReg10(PFD_Actual);
}

//-----------------------------------------------------------------------------
//  ADF4355::Calculate_NMod2_fromSpacing() -- suggested calc for NMod2  (?)
//-----------------------------------------------------------------------------

unsigned int ADF4355::Calculate_NMod2_fromSpacing(double PFD, double channel_spacing_freq)
{
    unsigned int gcd = Gcd(static_cast<unsigned int>(IntegerPart(PFD)),
                           static_cast<unsigned int>(IntegerPart(channel_spacing_freq)));
    Debug.Log("Gcd = ", gcd);

    unsigned int N_Mod2 = static_cast<unsigned int>(IntegerPart(PFD/static_cast<double>(gcd)));
    Debug.Log("Initial N_Mod2 = ", N_Mod2);
    N_Mod2 = std::max(N_Mod2, kN_MOD2_Min); // trim if out of range
    N_Mod2 = std::min(N_Mod2, N_Mod2_Max());
    Debug.Log("Capped N_Mod2 = ", N_Mod2);
    return N_Mod2;
}

//-----------------------------------------------------------------------------
//  ADF4355::TuneMatch() -- Calculate Parameters for PLL
//-----------------------------------------------------------------------------

bool  ADF4355::TuneMatch(std::vector<TuneMatchResults>& TM_results)
{
    //ScopedTraceRef ccc(Debug, "ADF4355::TuneMatch()");
    TM_results.clear();

    bool  TuneMatchFound(false);
    for (ValidDividerArray::reverse_iterator div = Output_DividerValues.rbegin();
             div !=  Output_DividerValues.rend();  ++div)
        {
        unsigned int CandidateOD  = *div;   // try ODs in reverse order...
        double       CandidateVco = EffectiveFrequency()*CandidateOD;

        if (CandidateVco<=LocalVcxo().Max() && CandidateVco>=LocalVcxo().Min())
            {
            //  We have a candidate Divider. Check if we can reset to this freq.
            //     (some Vco's have gaps in the range).
            if (LocalVcxo().CanTuneTo(CandidateOD*EffectiveFrequency()))
                {
                TuneMatchFound = true;
                TuneMatchResults TMR(CandidateVco, CandidateOD);
                TM_results.push_back(TMR);
                // keep looking
                }
            }
        }

    Debug.Log("TuneMatchFound ", TuneMatchFound);
    Debug.Log("TuneMatches ", TM_results.size());
    return TuneMatchFound;
}

//-----------------------------------------------------------------------------
//  ADF4355::FindVcoRate() -- Calculate Parameters for PLL
//-----------------------------------------------------------------------------

bool  ADF4355::FindVcoRate(double & VcoRate, unsigned int & OD)
{
    //  Standard One Pass Tune Match
    std::vector<TuneMatchResults>  TM_results;
    bool TuneMatchFound = TuneMatch(TM_results);
    //
    //  Determine Vco Frequency If the scan failed
    if (!TuneMatchFound)
        {
        //  We missed finding a VCO Match. Pick a VCO freq we can use
        VcoRate = LocalVcxo().BestTunableFrequency(EffectiveFrequency());  // assume OD=1 ?
        //  have to figure a reasonable FB value too.
        OD =  static_cast<unsigned int>(VcoRate / EffectiveFrequency());
        if (OD==0)
            OD=1;
        }
    else
        {
        //  Take Tune Match Results and put them in variables for use
        VcoRate = TM_results[0].CandidateVco;
        OD = TM_results[0].OD;

        // At least for now, lock in the PFD and let the N adjustment work out
        //   any issues.  This seems to be the intent of the hardware.
        }
    return TuneMatchFound;
}

//-----------------------------------------------------------------------------
//  ADF4355::DividerModeUpdate() -- Update state of dividers
//-----------------------------------------------------------------------------
//  In this mode, the reference is sent to the outputs, with possible division

void  ADF4355::DividerModeUpdate()
{
    Debug.Log("ADF4355::DividerModeUpdate()");
    Debug.Log("This mode is not supported by hardware");
}

//-----------------------------------------------------------------------------
//  ADF4355::SetOutputDividers()  --
//-----------------------------------------------------------------------------

void  ADF4355::SetOutputDividers(double VcoFreq, unsigned int OutDiv)
{
    Debug.Log("ADF4355::SetOutputDividers()");
    double OutputFreq = VcoFreq/OutDiv;

    for (unsigned int output=0; output<TotalOutputs; output++)
        {
        if (OutputInUse(output)==true)
            {
            float   CandidateDivider = static_cast<float>(OutDiv);
            double  CandidateFreq    = OutputFreq;          
            //
            //  Check for frequency exceeding applied Limits
            Debug.Log("FC CandidateDivider = ", CandidateDivider);
            Debug.Log("FC CandidateFreq = ", CandidateFreq);
            if (OutputFrequencyLimit[output]>=CandidateFreq)
                OutputDivider(output, static_cast<unsigned int>(CandidateDivider));
            else
                {
                Debug.Log("Out of Frequency Limit Range");

                double MinDivideRatio = VcoFreq/OutputFrequencyLimit[output];
                OutDiv = NearestInteger(MinDivideRatio);

                int above, below, divider;
                GetNearestDividers(output, static_cast<float>(OutDiv), below, above);

                // Find which divider is closest to desired frequency
                double centerFreq = VcoFreq/OutDiv;
                double hiFreq     = VcoFreq/below;
                double lowFreq    = VcoFreq/above;
                if (std::abs(centerFreq-lowFreq) > std::abs(hiFreq-centerFreq))
                    divider = below;
                else
                    divider = above;

                OutputDivider(output, divider);
                }
            //
            //  Calculate Actual Frequency
            if (OutputDividerArray[output]!=0)  // if illegal divider, skip it
                {
                OutputFreqActualArray[output] = VcoFreq / OutputDividerArray[output];
                }
            }
        else
            {
            OutputFreqActualArray[output] = 0.0;
            }
        }
}

//-----------------------------------------------------------------------------
//  ADF4355::Locked()  --  Return PLL lock status
//-----------------------------------------------------------------------------

bool  ADF4355::Locked() const
{
    MemoryBitManipulator bits(ControlReg.MuxOut.Value());

    if (FDeviceRouting == Dev_Both)
        return (bits.Bit(0) && bits.Bit(1));
    else if (FDeviceRouting == Dev1)
        return bits.Bit(1);
    else
        return bits.Bit(0);
}

//-----------------------------------------------------------------------------
//  ADF4355::Port()  --  Write to PLL data port
//-----------------------------------------------------------------------------

void ADF4355::Port(unsigned int reg, unsigned int value)
{
    // Wait for previous operation to complete
    for (unsigned int i = 0; (i < 10) && !ControlReg.Spi_Ready.Value(); ++i)
        Sleep(1);
    if (!ControlReg.Spi_Ready.Value())
        {
        Debug.Log("ADF4355 PLL SPI ready WR timeout error!");
        return;
        }

    if (reg>=GetRegCount())
        {
        Debug.Log("ADF4355 skipping out of range register write");
        return;
        }

    MemoryBitManipulator  MBM(value);
    MBM.Field(0, 4, reg);

    RegisterImage[reg] = MBM.Value();

    SpiReg.Value(MBM.Value());

    std::stringstream msg;
    msg << "ADF4355::Port[0x" << std::hex << static_cast<unsigned int>(reg)
        << "]=0x" << static_cast<unsigned int>(MBM.Value());
    Debug.Log(msg.str());
}


//-----------------------------------------------------------------------------
//  ADF4355::Port()  --  Read from PLL data port
//-----------------------------------------------------------------------------

unsigned int ADF4355::Port(unsigned int reg)
{
    // Cannot physically read this device, return the cached value
    return RegisterImage[reg];
}

//-----------------------------------------------------------------------------
//  ADF4355::ApplyImage()  --  Write register image data to physical register
//-----------------------------------------------------------------------------

void ADF4355::ApplyImage(unsigned int reg)
{
    Port(reg, RegisterImage[reg]);
}

//-----------------------------------------------------------------------------
//  ADF4355::PortImage()  --  Write to Object Image Data Cache
//-----------------------------------------------------------------------------

void  ADF4355::PortImage(unsigned int reg, unsigned int value)
{
    if (reg < static_cast<unsigned int>(RegisterImage.size()))
        RegisterImage[reg] = value;
}

//-----------------------------------------------------------------------------
//  ADF4355::PortImage()  --  Read from Object Image Data Cache
//-----------------------------------------------------------------------------

unsigned int  ADF4355::PortImage(unsigned int reg)
{
    if (reg < static_cast<unsigned int>(RegisterImage.size()))
        return RegisterImage[reg];
    else
        return 0;
}

//-----------------------------------------------------------------------------
//  ADF4355::SetVcoAutocal()  --  Set Auto-Cal bit
//-----------------------------------------------------------------------------

void ADF4355::SetVcoAutocal(bool state)
{
    PtrBitManipulator Reg0(&RegisterImage[0]);
    Reg0.Bit(21, state);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetPrescaler()  --  Set Prescaler in use
//-----------------------------------------------------------------------------

void ADF4355::SetPrescaler(DualModulusPrescaler prescaler)
{
    PtrBitManipulator Reg0(&RegisterImage[0]);
    bool state;

    switch (prescaler)
        {
        default:
        case pre_4_5:  state=false;  break;
        case pre_8_9:  state=true;   break;
        }

    Reg0.Bit(20, state);
}

//-----------------------------------------------------------------------------
//  ADF4355::GetPrescaler()  --  Read prescaler in use
//-----------------------------------------------------------------------------

ADF4355::DualModulusPrescaler ADF4355::GetPrescaler()
{
    PtrBitManipulator Reg0(&RegisterImage[0]);
    bool state = Reg0.Bit(20);
    if (state)
        return pre_8_9;
    else
        return pre_4_5;
}

//-----------------------------------------------------------------------------
//  ADF4355::SetN_Int()  --  Configure N_Int  (PLL Feedback Divider, INT part)
//-----------------------------------------------------------------------------

void ADF4355::SetN_Int(unsigned int n_int)
{
    PtrBitManipulator Reg0(&RegisterImage[0]);
    Reg0.Field(4, 16, n_int);
}

//-----------------------------------------------------------------------------
//  ADF4355::GetN_Int()  --  Read N_Int  (PLL Feedback Divider, INT part)
//-----------------------------------------------------------------------------

unsigned int ADF4355::GetN_Int()
{
    PtrBitManipulator Reg0(&RegisterImage[0]);
    return Reg0.Field(4, 16);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetN_Frac1()  --  Configure N_FRAC1  (PLL1 Feedback Divider)
//-----------------------------------------------------------------------------

void ADF4355::SetN_Frac1(unsigned int n_Frac1)
{
    PtrBitManipulator Reg1(&RegisterImage[1]);
    Reg1.Field(4, 24, n_Frac1);
}

//-----------------------------------------------------------------------------
//  ADF4355::GetN_Frac1()  --  Read N_FRAC1  (PLL1 Feedback Divider)
//-----------------------------------------------------------------------------

unsigned int ADF4355::GetN_Frac1()
{
    PtrBitManipulator Reg1(&RegisterImage[1]);
    return Reg1.Field(4, 24);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetN_Frac2()  --  Configure N_FRAC2  (PLL Feedback Divider)
//-----------------------------------------------------------------------------

void ADF4355::SetN_Frac2(unsigned int n_Frac2)
{
    if (FChipType==adf4355)
        {
        //  use only 14 bit "LSB"
        SetN_Frac2_LSB(n_Frac2);
        }
    else if (FChipType==adf4355_2)
        {
        //  use only 14 bit "LSB"
        SetN_Frac2_LSB(n_Frac2);
        }
    else // adf4356
        {
        //  use 28 bit value in 2 registers
        SetN_Frac2_LSB(n_Frac2);
        n_Frac2 >>= 14;
        SetN_Frac2_MSB(n_Frac2);
        }
}

//-----------------------------------------------------------------------------
//  ADF4355::GetN_Frac2()  --  read N_FRAC2  (PLL Feedback Divider)
//-----------------------------------------------------------------------------

unsigned int ADF4355::GetN_Frac2()
{
    if (FChipType==adf4355)
        {
        //  use only 14 bit "LSB"
        return GetN_Frac2_LSB();
        }
    else if (FChipType==adf4355_2)
        {
        //  use only 14 bit "LSB"
        return GetN_Frac2_LSB();
        }
    else // adf4356
        {
        //  use 28 bit value in 2 registers
        unsigned int rtn = GetN_Frac2_MSB();
        rtn <<= 14;
        rtn |= GetN_Frac2_LSB();
        return rtn;
        }
}

//-----------------------------------------------------------------------------
//  ADF4355::SetN_Mod2()  --  Configure N_Mod2  (PLL Feedback Divider)
//-----------------------------------------------------------------------------

void ADF4355::SetN_Mod2(unsigned int n_Mod2)
{
    if (FChipType==adf4355)
        {
        //  use only 14 bit "LSB"
        SetN_Mod2_LSB(n_Mod2);
        }
    else if (FChipType==adf4355_2)
        {
        //  use only 14 bit "LSB"
        SetN_Mod2_LSB(n_Mod2);
        }
    else // adf4356
        {
        //  use 28 bit value in 2 registers
        SetN_Mod2_LSB(n_Mod2);
        n_Mod2 >>= 14;
        SetN_Mod2_MSB(n_Mod2);
        }
}

//-----------------------------------------------------------------------------
//  ADF4355::GetN_Mod2()  --  read N_Mod2  (PLL Feedback Divider)
//-----------------------------------------------------------------------------

unsigned int ADF4355::GetN_Mod2()
{
    if (FChipType==adf4355)
        {
        //  use only 14 bit "LSB"
        return GetN_Mod2_LSB();
        }
    else if (FChipType==adf4355_2)
        {
        //  use only 14 bit "LSB"
        return GetN_Mod2_LSB();
        }
    else // adf4356
        {
        //  use 28 bit value in 2 registers
        unsigned int rtn = GetN_Mod2_MSB();
        rtn <<= 14;
        rtn |= GetN_Mod2_LSB();
        return rtn;
        }
}

//-----------------------------------------------------------------------------
//  ADF4355::SetN_Frac2_LSB()  --  Configure N_FRAC2  (PLL Feedback Divider)
//-----------------------------------------------------------------------------

void ADF4355::SetN_Frac2_LSB(unsigned int n_Frac2_l)
{
    PtrBitManipulator Reg2(&RegisterImage[2]);
    Reg2.Field(18, 14, n_Frac2_l);
}

//-----------------------------------------------------------------------------
//  ADF4355::GetN_Frac2_LSB()  --  Read N_FRAC2 (PLL Feedback Divider)
//-----------------------------------------------------------------------------

unsigned int ADF4355::GetN_Frac2_LSB()
{
    PtrBitManipulator Reg2(&RegisterImage[2]);
    return Reg2.Field(18, 14);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetN_Mod2_LSB()  --  Read N_MOD2 (PLL Feedback Divider)
//-----------------------------------------------------------------------------

void ADF4355::SetN_Mod2_LSB(unsigned int n_Mod2_l)
{
    PtrBitManipulator Reg2(&RegisterImage[2]);
    Reg2.Field(4, 14, n_Mod2_l);
}

//-----------------------------------------------------------------------------
//  ADF4355::GetN_Mod2_LSB()  --  Read N_MOD2 (PLL Feedback Divider)
//-----------------------------------------------------------------------------

unsigned int ADF4355::GetN_Mod2_LSB()
{
    PtrBitManipulator Reg2(&RegisterImage[2]);
    return Reg2.Field(4, 14);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetN_Frac2_MSB()  --  Configure N_FRAC2  (PLL Feedback Divider)
//-----------------------------------------------------------------------------

void ADF4355::SetN_Frac2_MSB(unsigned int n_Frac2_h)
{
    PtrBitManipulator Reg13(&RegisterImage[13]);
    Reg13.Field(18, 14, n_Frac2_h);
}

//-----------------------------------------------------------------------------
//  ADF4355::GetN_Frac2_MSB()  --  Read N_FRAC2 (PLL Feedback Divider)
//-----------------------------------------------------------------------------

unsigned int ADF4355::GetN_Frac2_MSB()
{
    PtrBitManipulator Reg13(&RegisterImage[13]);
    return Reg13.Field(18, 14);
}


//-----------------------------------------------------------------------------
//  ADF4355::SetN_Mod2_MSB()  --  Read N_MOD2 (PLL Feedback Divider)
//-----------------------------------------------------------------------------

void ADF4355::SetN_Mod2_MSB(unsigned int n_Mod2_h)
{
    PtrBitManipulator Reg13(&RegisterImage[13]);
    Reg13.Field(4, 14, n_Mod2_h);
}


//-----------------------------------------------------------------------------
//  ADF4355::GetN_Mod2_MSB()  --  Read N_MOD2 (PLL Feedback Divider)
//-----------------------------------------------------------------------------

unsigned int ADF4355::GetN_Mod2_MSB()
{
    PtrBitManipulator Reg13(&RegisterImage[2]);
    return Reg13.Field(4, 14);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetOutputPhase()  --  Configure Output Phase
//-----------------------------------------------------------------------------

void ADF4355::SetOutputPhase(unsigned int phase_value)
{
    PtrBitManipulator Reg3(&RegisterImage[3]);
    Reg3.Field(4, 24, phase_value);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetOutputPhaseAdjustEnabled()  --  Configure Output Phase
//-----------------------------------------------------------------------------

void ADF4355::SetOutputPhaseAdjustEnabled(bool state)
{
    PtrBitManipulator Reg3(&RegisterImage[3]);
    Reg3.Bit(28, state);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetOutputPhaseResyncEnabled()  --  Configure Output Phase
//-----------------------------------------------------------------------------

void ADF4355::SetOutputPhaseResyncEnabled(bool state)
{
    PtrBitManipulator Reg3(&RegisterImage[3]);
    Reg3.Bit(29, state);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetSigmaDeltaLoadResetDisabled()  --  Configure SigmaDelta reset
//-----------------------------------------------------------------------------

void ADF4355::SetSigmaDeltaLoadResetDisabled(bool state)
{
    PtrBitManipulator Reg3(&RegisterImage[3]);
    Reg3.Bit(30, state);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetCounterReset()  --  Reset R, N, Vco Counter
//-----------------------------------------------------------------------------

void ADF4355::SetCounterReset(bool state)
{
    PtrBitManipulator Reg4(&RegisterImage[4]);
    Reg4.Bit(4, state);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetCpThreeStateEnabled()  --  set charge pump to three-state
//-----------------------------------------------------------------------------

void ADF4355::SetCpThreeStateEnabled(bool state)
{
    PtrBitManipulator Reg4(&RegisterImage[4]);
    Reg4.Bit(5, state);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetPowerDown()  --  power down chip
//-----------------------------------------------------------------------------

void ADF4355::SetPowerDown(bool state)
{
    PtrBitManipulator Reg4(&RegisterImage[4]);
    Reg4.Bit(6, state);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetPhaseDetectorPolarity()  --  set polarity of phase detect
//-----------------------------------------------------------------------------

void ADF4355::SetPhaseDetectorPolarity(PhaseDetectPolarity polarity)
{
    PtrBitManipulator Reg4(&RegisterImage[4]);
    bool state;

    switch (polarity)
        {
        default:
        case pdpNegative:  state=false;  break;
        case pdpPositive:  state=true;   break;
        }

    Reg4.Bit(7, state);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetMuxoutLevelSelect()  --  set level of MUXOUT signal
//-----------------------------------------------------------------------------

void ADF4355::SetMuxoutLevelSelect(MuxoutLogicLevels level_setting)
{
    PtrBitManipulator Reg4(&RegisterImage[4]);
    bool state;

    switch (level_setting)
        {
        default:
        case ldp_1p8v:  state=false;  break;
        case ldp_3p3v:  state=true;   break;
        }

    Reg4.Bit(8, state);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetReferenceMode()  --  set mode of reference input
//-----------------------------------------------------------------------------

void ADF4355::SetReferenceMode(ReferenceInOptions ref_mode)
{
    PtrBitManipulator Reg4(&RegisterImage[4]);
    bool state;

    switch (ref_mode)
        {
        default:
        case refSingleEnded:   state=false;  break;
        case refDifferential:  state=true;   break;
        }

    Reg4.Bit(9, state);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetChargePumpCurrent()  --  set charge pump current
//-----------------------------------------------------------------------------

void ADF4355::SetChargePumpCurrent(ChargePumpCurrentOptions current_setting)
{
    PtrBitManipulator Reg4(&RegisterImage[4]);
    Reg4.Field(10, 4, current_setting);
}

//-----------------------------------------------------------------------------
//  ADF4355::GetChargePumpCurrent()  --  get charge pump current
//-----------------------------------------------------------------------------

float ADF4355::GetChargePumpCurrent()
{
    PtrBitManipulator Reg4(&RegisterImage[4]);
    unsigned int CpSetting = Reg4.Field(10, 4);

    if (CpSetting == cpc_0p31)
        return 0.31f;
    else if (CpSetting == cpc_0p63)
        return 0.63f;
    else if (CpSetting == cpc_0p94)
        return 0.94f;
    else if (CpSetting == cpc_1p25)
        return 1.25f;
    else if (CpSetting == cpc_1p56)
        return 1.56f;
    else if (CpSetting == cpc_1p88)
        return 1.88f;
    else if (CpSetting == cpc_2p19)
        return 2.19f;
    else if (CpSetting == cpc_2p50)
        return 2.5f;
    else if (CpSetting == cpc_2p81)
        return 2.81f;
    else if (CpSetting == cpc_3p13)
        return 3.13f;
    else if (CpSetting == cpc_3p44)
        return 3.44f;
    else if (CpSetting == cpc_3p75)
        return 3.75f;
    else if (CpSetting == cpc_4p06)
        return 4.06f;
    else if (CpSetting == cpc_4p38)
        return 4.38f;
    else if (CpSetting == cpc_4p69)
        return 4.69f;
    else // (CpSetting == cpc_5p00)
        return 5.0f;
}

//-----------------------------------------------------------------------------
//  ADF4355::SetReg6_DoubleBuffering()  --  double buffer register 6
//-----------------------------------------------------------------------------

void ADF4355::SetReg6_DoubleBuffering(bool state)
{
    PtrBitManipulator Reg4(&RegisterImage[4]);
    Reg4.Bit(14, state);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetR()  --  Configure R  (PLL Reference Divider)
//-----------------------------------------------------------------------------

void ADF4355::SetR(unsigned int R)
{
    PtrBitManipulator Reg4(&RegisterImage[4]);
    Reg4.Field(15, 10, R);
}

//-----------------------------------------------------------------------------
//  ADF4355::GetR()  --  Read R  (PLL Reference Divider)
//-----------------------------------------------------------------------------

unsigned int ADF4355::GetR()
{
    PtrBitManipulator Reg4(&RegisterImage[4]);
    return Reg4.Field(15, 10);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetRDiv2()  --  Configure RDIV2  (PLL Reference Divider)
//-----------------------------------------------------------------------------

void ADF4355::SetRDiv2(unsigned int rDiv2)
{
    PtrBitManipulator Reg4(&RegisterImage[4]);
    bool state = (rDiv2==2) ? true : false ;
    Reg4.Bit(25, state);
}

//-----------------------------------------------------------------------------
//  ADF4355::GetRDiv2()  --  Read RDIV2  (PLL Reference Divider)
//-----------------------------------------------------------------------------

unsigned int ADF4355::GetRDiv2()
{
    PtrBitManipulator Reg4(&RegisterImage[4]);
    bool state = Reg4.Bit(25);
    unsigned int value = (state) ? 2 : 1 ;
    return value;
}

//-----------------------------------------------------------------------------
//  ADF4355::SetRMul2()  --  Configure RMUL2  (PLL Reference Divider)
//-----------------------------------------------------------------------------

void ADF4355::SetRMul2(unsigned int rMul2)
{
    PtrBitManipulator Reg4(&RegisterImage[4]);
    bool state = (rMul2==2) ? true : false ;
    Reg4.Bit(26, state);
}

//-----------------------------------------------------------------------------
//  ADF4355::GetRMul2()  --  Read RMUL2  (PLL Reference Divider)
//-----------------------------------------------------------------------------

unsigned int ADF4355::GetRMul2()
{
    PtrBitManipulator Reg4(&RegisterImage[4]);
    bool state = Reg4.Bit(26);
    unsigned int value = (state) ? 2 : 1 ;
    return value;
}

//-----------------------------------------------------------------------------
//  ADF4355::SetMuxOut()  --  Set MUXOUT output
//-----------------------------------------------------------------------------

void ADF4355::SetMuxOut(MuxOutOptions option)
{
    PtrBitManipulator Reg4(&RegisterImage[4]);
    Reg4.Field(27, 3, option);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetRfOutputPower()  --  Set RF output power
//-----------------------------------------------------------------------------

void ADF4355::SetRfOutputPower(OutputPowerOptions option)
{
    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Field(4, 2, option);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetRfOutputEnabled()  --  Set RF output enable
//-----------------------------------------------------------------------------

void ADF4355::SetRfOutputEnabled(bool enabled)
{
    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Bit(6, enabled);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetAuxOutputPower()  --  Set AUX output power
//-----------------------------------------------------------------------------

void ADF4355::SetAuxOutputPower(OutputPowerOptions option)
{
    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Field(7, 2, option);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetAuxOutputEnabled()  --  Set AUX output enable
//-----------------------------------------------------------------------------

void ADF4355::SetAuxOutputEnabled(bool enabled)
{
    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Bit(9, enabled);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetMTLDEnabled()  --  Set Mute Till Lock Detect
//-----------------------------------------------------------------------------

void ADF4355::SetMTLDEnabled(bool enabled)
{
    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Bit(11, enabled);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetChargePumpBleedCurrent()  --  Set bleed current
//-----------------------------------------------------------------------------

void ADF4355::SetChargePumpBleedCurrent(float current_uA)
{
    // steps are 3.75 uA apart
    unsigned int current_bits = static_cast<unsigned int>(current_uA/3.75);

    const unsigned int MaxValue = 255;
    unsigned int current_bits_capped = min (current_bits, MaxValue);

    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Field(13, 8, current_bits_capped);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetFeedbackSelect()  --  Set feedback from VCO -> N
//-----------------------------------------------------------------------------

void ADF4355::SetFeedbackSelect(FeedbackSelectOptions option)
{
    PtrBitManipulator Reg6(&RegisterImage[6]);
    bool state;

    switch (option)
        {
        default:
        case fsDivided:      state=false;  break;
        case fsFundamental:  state=true;   break;
        }

    Reg6.Bit(24, state);
}

//-----------------------------------------------------------------------------
//  ADF4355::GetFeedbackSelect()  --  Get feedback from VCO -> N
//-----------------------------------------------------------------------------

ADF4355::FeedbackSelectOptions ADF4355::GetFeedbackSelect()
{
    PtrBitManipulator Reg6(&RegisterImage[6]);

    bool state = Reg6.Bit(24);
    if (state)
        return fsFundamental;
    else
        return fsDivided;
}

//-----------------------------------------------------------------------------
//  ADF4355::SetBleedCurrentEnabled()  --  Set Negative Bleed Current
//-----------------------------------------------------------------------------

void ADF4355::SetBleedCurrentEnabled(bool enabled)
{
    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Bit(29, enabled);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetGatedBleedEnabled()  --  Set Gated Bleed Current
//-----------------------------------------------------------------------------

void ADF4355::SetGatedBleedEnabled(bool enabled)
{
    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Bit(30, enabled);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetBleedPolarity()  --  Set Gated Bleed Polarity
//-----------------------------------------------------------------------------

void ADF4355::SetBleedPolarity(bool bleed_pos)
{
    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Bit(31, bleed_pos);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetLockDetectMode()  --  Set Lock Detect Mode
//-----------------------------------------------------------------------------

void ADF4355::SetLockDetectMode(LockDetectModeOptions option)
{
    PtrBitManipulator Reg7(&RegisterImage[7]);
    bool state;

    switch (option)
        {
        default:
        case ldmFractionalN:  state=false;  break;
        case ldmIntegerN:     state=true;   break;
        }

    Reg7.Bit(4, state);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetFractionalNPrecision()  --  Set LD Precision
//-----------------------------------------------------------------------------

void ADF4355::SetFractionalNPrecision(FractionalNPrecisionOptions option)
{
    PtrBitManipulator Reg7(&RegisterImage[7]);
    Reg7.Field(5, 2, option);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetLossOfLockModeEnabled()  --  Set Gated Bleed Current
//-----------------------------------------------------------------------------

void ADF4355::SetLossOfLockModeEnabled(bool enabled)
{
    PtrBitManipulator Reg7(&RegisterImage[7]);
    Reg7.Bit(7, enabled);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetLossOfLockCycleCount()  --  Set LD Precision
//-----------------------------------------------------------------------------

void ADF4355::SetLossOfLockCycleCount(LossOfLockCycleCountOptions option)
{
    PtrBitManipulator Reg7(&RegisterImage[7]);
    Reg7.Field(8, 2, option);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetLESyncEnabled()  --  Set LE Sync Enabled
//-----------------------------------------------------------------------------

void ADF4355::SetLESyncEnabled(bool enabled)
{
    PtrBitManipulator Reg7(&RegisterImage[7]);
    Reg7.Bit(25, enabled);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetReg9()  --  Set the contents of register 9
//-----------------------------------------------------------------------------

void ADF4355::SetReg9(double pfd)
{
    // ALC Wait = 2.5 x Synthesizer Lock Timeout
    unsigned int SynthesizerLockTimeout = 12;
    unsigned int ALC_wait = 30;

    SetSynthesizerLockTimeout(SynthesizerLockTimeout);
    SetAutomaticLevelCalTimeout(ALC_wait);

    unsigned int Timeout = static_cast<unsigned int>(ceil(pfd * 50.0e-6/ALC_wait));
    SetTimeout(Timeout);

    unsigned int VcoBandDivision = static_cast<unsigned int>(ceil(pfd/2400000));
    SetVcoBandDivision(VcoBandDivision);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetSynthesizerLockTimeout()  --  Set Synth Lock Timeout
//-----------------------------------------------------------------------------

void ADF4355::SetSynthesizerLockTimeout(unsigned int value)
{
    PtrBitManipulator Reg9(&RegisterImage[9]);
    Reg9.Field(4, 5, value);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetAutomaticLevelCalTimeout()  --  Set Auto Level Calibration Timeout
//-----------------------------------------------------------------------------

void ADF4355::SetAutomaticLevelCalTimeout(unsigned int value)
{
    PtrBitManipulator Reg9(&RegisterImage[9]);
    Reg9.Field(9, 5, value);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetTimeout()  --  Set Timeout
//-----------------------------------------------------------------------------

void ADF4355::SetTimeout(unsigned int value)
{
    PtrBitManipulator Reg9(&RegisterImage[9]);
    Reg9.Field(14, 10, value);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetVcoBandDivision()  --  Set Vco Band Division Clock
//-----------------------------------------------------------------------------

void ADF4355::SetVcoBandDivision(unsigned int value)
{
    PtrBitManipulator Reg9(&RegisterImage[9]);
    Reg9.Field(24, 8, value);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetReg10()  --  Set the contents of register 10
//-----------------------------------------------------------------------------

void ADF4355::SetReg10(double pfd)
{
    VTuneAdcEnabled(true);
    VTuneAdcConversionEnabled(true);

    unsigned int AdcClkDiv = static_cast<unsigned int>(ceil(((pfd/100000)-2)/4));
    const unsigned int MaxAdcClkDiv = 255;
    unsigned int AdcClkDivCapped = min(AdcClkDiv,MaxAdcClkDiv);

    VTuneAdcClockDivider(AdcClkDivCapped);

    // Delay > 16 ADC_CLK cycles
    delay_us = static_cast<unsigned int>(ceil((16 * 1000000 * (4*AdcClkDivCapped+2))/pfd));
}

//-----------------------------------------------------------------------------
//  ADF4355::VTuneAdcEnabled()  --  Enable VTUNE Adc
//-----------------------------------------------------------------------------

void ADF4355::VTuneAdcEnabled(bool enabled)
{
    PtrBitManipulator Reg10(&RegisterImage[10]);
    Reg10.Bit(4, enabled);
}

//-----------------------------------------------------------------------------
//  ADF4355::VTuneAdcConversionEnabled()  --  Enable VTUNE Adc Conversion
//-----------------------------------------------------------------------------

void ADF4355::VTuneAdcConversionEnabled(bool enabled)
{
    PtrBitManipulator Reg10(&RegisterImage[10]);
    Reg10.Bit(5, enabled);
}

//-----------------------------------------------------------------------------
//  ADF4355::VTuneAdcClockDivider()  --  Set VTUNE Adc Clock Divider
//-----------------------------------------------------------------------------

void ADF4355::VTuneAdcClockDivider(unsigned int divider)
{
    PtrBitManipulator Reg10(&RegisterImage[10]);
    Reg10.Field(6, 8, divider);
}

//-----------------------------------------------------------------------------
//  ADF4355::PhaseResyncClockDivider()  --  Divider for Phase Resync
//-----------------------------------------------------------------------------

void ADF4355::PhaseResyncClockDivider(unsigned int divider)
{
    PtrBitManipulator Reg12(&RegisterImage[12]);
    Reg12.Field(16, 16, divider);
}

//-----------------------------------------------------------------------------
//  ADF4355::OnSetPower()  --  Perform Pwr Setting operation
//-----------------------------------------------------------------------------

void  ADF4355::OnSetPower(bool & new_state)
{
    //  POWER DOWN
    SetPowerDown(new_state);
    //  pause to allow chip to activate - needed?
    Sleep(10);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetOutEnabledHdw()  --  Update the Output Enabled hardware
//-----------------------------------------------------------------------------

void ADF4355::SetOutEnabledHdw(unsigned int output, bool enabled)
{
    if (output==0)
        SetRfOutputEnabled(enabled);
    else
        SetAuxOutputEnabled(enabled);
}

//-----------------------------------------------------------------------------
//  ADF4355::SetDividerHdw()  --  Update the Divider setting hardware
//-----------------------------------------------------------------------------

void ADF4355::SetDividerHdw(unsigned int output, unsigned int divide_by)
{
    ScopedTraceRef ccc(Debug, "ADF4355::SetDividerHdw()");

    Debug.Log("output = ", output);
    Debug.Log("divide_by = ", divide_by);

    // Both outputs share the same divider...
    if (!IsValidDivider(divide_by))
        return;

    PtrBitManipulator Reg6(&RegisterImage[6]);
    unsigned int div_field = ToDividerInput(divide_by);
    Reg6.Field(21, 3, div_field);
}

//-----------------------------------------------------------------------------
//  ADF4355::GetDivider()  --  read back the Divider setting
//-----------------------------------------------------------------------------

unsigned int  ADF4355::GetDivider(unsigned int /*output*/)
{
    PtrBitManipulator Reg6(&RegisterImage[6]);
    unsigned int div_field = Reg6.Field(21, 3);
    unsigned int divider = FromDividerInput(div_field);
    return divider;
}

//-----------------------------------------------------------------------------
//  ADF4355::SetReferenceSource()  --  Set the Reference Source
//-----------------------------------------------------------------------------

void ADF4355::SetReferenceSource(AdfReferenceSource source)
{
    Debug.Log("ADF4355::SetReferenceSource");
    Debug.Log("...source=", source);

    unsigned int source_no;
    switch (source)
        {
        default:
        case refPrimary:    source_no = 0;  break;
        }

    Debug.Log("...source_no=", source_no);

    //  There's only one source so no register change

    SetSourceSelection(source_no);   // inform base class too

}

//-----------------------------------------------------------------------------
//  ADF4355::SetDeviceRouting()  --  Set which devices SPI calls affect
//-----------------------------------------------------------------------------

void ADF4355::SetDeviceRouting(ADF4355::AdfDevice device)
{
    ControlReg.Device.Value(static_cast<unsigned int>(device));
}

//-----------------------------------------------------------------------------
//  ADF4355::SetDevicePowerDown()  --  Set which devices power dowm
//-----------------------------------------------------------------------------

void ADF4355::SetDevicePowerDown(ADF4355::AdfDevice device, bool state)
{
    unsigned int bits = (state) ? 0 : device;
    ControlReg.RF_PowerDown.Value(static_cast<unsigned int>(bits));
}

//-----------------------------------------------------------------------------
//  ADF4355::SetDeviceEnable()  --  Set which devices enable
//-----------------------------------------------------------------------------

void ADF4355::SetDeviceEnable(ADF4355::AdfDevice device, bool state)
{
    unsigned int bits = (state) ? device : 0;
    ControlReg.EnablePLL.Value(static_cast<unsigned int>(bits));
}

//-----------------------------------------------------------------------------
//  ADF4355::GetMuxOut()  --  Get Muxout bits
//-----------------------------------------------------------------------------

unsigned int ADF4355::GetMuxOut()
{
    return ControlReg.MuxOut.Value();
}

//-----------------------------------------------------------------------------
//  ADF4355::SetVcoRange()  --  Load VCO Range based off of Chip Type
//-----------------------------------------------------------------------------

void  ADF4355::SetVcoRange(AdfType chip_type)
{
    unsigned int ct = chip_type;
    if (ct >= TotalVcoTypes) ct = 0;

    // Load internal VCO mods
    LocalVcxo_Standard.Load(ADF4355_VcoRanges[ct].Low, ADF4355_VcoRanges[ct].High);
}

//-----------------------------------------------------------------------------
//  ADF4355::GetRegCount()  --  Get the number of SPI registers for the part
//-----------------------------------------------------------------------------

unsigned int ADF4355::GetRegCount()
{
    if (FChipType==adf4355 || FChipType==adf4355_2)
        return 13;
    else // adf4356
        return 14;
}

//-----------------------------------------------------------------------------
//  ADF4355::LocalVcxo()  --  Load VCO Range based off of Chip Type
//-----------------------------------------------------------------------------

IVcxoControl &  ADF4355::LocalVcxo()
{
    return LocalVcxo_Standard;
}

//-----------------------------------------------------------------------------
//  ADF4355::InitDividerSupport()  --  Initialize Special Divider Maps
//-----------------------------------------------------------------------------

void  ADF4355::InitDividerSupport()
{
    //  Output Divider Values
    Output_DividerValues.push_back(1);     Output_DividerValues.push_back(2);
    Output_DividerValues.push_back(4);     Output_DividerValues.push_back(8);
    Output_DividerValues.push_back(16);    Output_DividerValues.push_back(32);
    Output_DividerValues.push_back(64);

    //  Valid dividers is keyed by divide_by, value is Input Value
    Output_ValidDividers[1]    = 0x00;       Output_ValidDividers[2]    = 0x01;
    Output_ValidDividers[4]    = 0x02;       Output_ValidDividers[8]    = 0x03;
    Output_ValidDividers[16]   = 0x04;       Output_ValidDividers[32]   = 0x05;
    Output_ValidDividers[64]   = 0x06;

    //     Valid dividers is keyed by Input Value, value is divide_by
    Output_InverseDividers[0x00]  =    1;      Output_InverseDividers[0x01]  =    2;
    Output_InverseDividers[0x02]  =    4;      Output_InverseDividers[0x03]  =    8;
    Output_InverseDividers[0x04]  =   16;      Output_InverseDividers[0x05]  =   32;
    Output_InverseDividers[0x06]  =   64;

}

//-----------------------------------------------------------------------------
//  ADF4355::IsValidDivider() -- tests if divide_by is a valid divider
//-----------------------------------------------------------------------------
//  Output Group Output Dividers

bool  ADF4355::IsValidDivider(unsigned int divide_by)
{
    ValidDividerMap::iterator iter = Output_ValidDividers.find(divide_by);
    return iter != Output_ValidDividers.end();
}

//-----------------------------------------------------------------------------
//  ADF4355::ToDividerInput() -- Returns input value for divide_by
//-----------------------------------------------------------------------------
//  this maps divider ratios to the field value in the chip hardware

unsigned int ADF4355::ToDividerInput(unsigned int divide_by)
{
    if (IsValidDivider(divide_by))
        return Output_ValidDividers[divide_by];
    else
        return Output_ValidDividers[1];   // if invalid divider, divide by 1.
}

//-----------------------------------------------------------------------------
//  ADF4355::FromDividerInput() -- Returns divide_by for input value
//-----------------------------------------------------------------------------
//  this maps the field value in the chip hardware to divider ratios

unsigned int  ADF4355::FromDividerInput(unsigned int divider_bits)
{
    InverseDividerMap::iterator iter = Output_InverseDividers.find(divider_bits);

    if (iter != Output_InverseDividers.end())
        return Output_InverseDividers[divider_bits];
    else
        return 1;   // if invalid divider_bits, claim it is divide by 1
}

//-----------------------------------------------------------------------------
//  ADF4355::CheckDivider() -- Check for valid divider value
//-----------------------------------------------------------------------------

bool  ADF4355::CheckDivider(unsigned int /*output*/, unsigned int divide_by)
{
    return IsValidDivider(divide_by);
}

//-----------------------------------------------------------------------------
//  ADF4355::GetNearestDividers() -- Find nearest divider to a candidate
//-----------------------------------------------------------------------------

void  ADF4355::GetNearestDividers(unsigned int /*output*/, float candidate, int & below, int & above)
{
    if (IsAboutEqual(candidate, NearestIntegerPart(candidate)))
        {
        //  We are essentially an integer, so treat as such.
        unsigned int int_candidate(static_cast<unsigned int>(NearestInteger(candidate)));
        if (IsValidDivider(int_candidate))
            {
            above = below = int_candidate;
            return;
            }
        }
    above = BestDividerGreaterThan(candidate);
    below = BestDividerLessThan(candidate);
}

//-----------------------------------------------------------------------------
//  ADF4355::BestDividerGreaterThan() -- returns next largest valid Divider
//-----------------------------------------------------------------------------

unsigned int  ADF4355::BestDividerGreaterThan(float min_divide_by)
{
    unsigned int CandidateD = 0;
    for (ValidDividerArray::iterator div = Output_DividerValues.begin();
             div !=  Output_DividerValues.end();  ++div)
        {
        CandidateD = *div;
        if (static_cast<float>(CandidateD) > min_divide_by)
            break;
        }
    return CandidateD;
}

//-----------------------------------------------------------------------------
//  ADF4355::BestDividerLessThan() -- returns next smallest valid Divider
//-----------------------------------------------------------------------------

unsigned int  ADF4355::BestDividerLessThan(float max_divide_by)
{
    unsigned int CandidateD = 0;
    for (ValidDividerArray::reverse_iterator div = Output_DividerValues.rbegin();
             div !=  Output_DividerValues.rend();  ++div)
        {
        CandidateD = *div;
        if (static_cast<float>(CandidateD) < max_divide_by)
            break;
        }
    return CandidateD;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  CLASS ADF4355::InternalVcxo  --  In-chip VCXO interface
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//-----------------------------------------------------------------------------
//  ADF4355::InternalVcxo::CanTuneTo() -- return true if VCO can tune here
//-----------------------------------------------------------------------------

bool    ADF4355::InternalVcxo::CanTuneTo(double freq)
{
    bool retValue;

    if (freq >= Min() && freq <= Max())
        retValue = true;
    else
        retValue = false;

    return retValue;
}

//-----------------------------------------------------------------------------
//  ADF4355::InternalVcxo::BestTunableFrequency() -- find closest allowed freq
//-----------------------------------------------------------------------------

double ADF4355::InternalVcxo::BestTunableFrequency(double freq)
{
    double retValue;

    if (freq < Min())
        retValue = Min();
    else if (freq >= Min() && freq <= Max())
        retValue = freq;
    else
        retValue = Max();

    return retValue;
}

//-----------------------------------------------------------------------------
//  ADF4355::InternalVcxo::SetVcoFrequency()
//-----------------------------------------------------------------------------

void  ADF4355::InternalVcxo::SetVcoFrequency(double freq)
{
    ScopedTraceRef ccc(Pll.Debug, "ADF4355::InternalVcxo::SetVcoFrequency()");

    Pll.FVcoFrequency = freq;

    Pll.Debug.Log("VCO Values");
    Pll.Debug.Log("ADF4355::Output Vco Freq (MHz) =", Pll.FVcoFrequency/1.0e6);
}

//-----------------------------------------------------------------------------
//  ADF4355::InternalVcxo::GetVcoFrequency()  --  read Vco Frequency
//-----------------------------------------------------------------------------

double  ADF4355::InternalVcxo::GetVcoFrequency()
{
    return Pll.FVcoFrequency;
}


} // namespace Innovative
