// AccessDatagrams_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2007 Innovative Integration Corporation.
//  All Rights Reserved.
//
//---------------------------------------------------------------------------

#ifndef AccessDatagrams_MbH
#define AccessDatagrams_MbH

#include <iterator>

#include <cstddef>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

void ThrowDGRangeException( size_t index );

//==============================================================================
//  CLASS IIteratable -- Interface required to support iteration over data
//==============================================================================

class IIteratable
{
public:
    IIteratable()   {  }
    virtual ~IIteratable()   {  }

    virtual char *  Base() = 0;
	virtual size_t  SizeInBytes() const = 0;
};

//==============================================================================
//  CLASS DatagramIterator  -- Access Template Class Iterator
//==============================================================================

template <class T>
class DatagramIterator : public std::iterator<std::random_access_iterator_tag, T>
{
public:
    typedef T             element_type;    // the type of the iterated element
    typedef T*            element_ptr;     // type of pointer to the iterated element
    typedef ptrdiff_t     difference_type;
    typedef size_t        size_type;

public:
    //
    //  constructors and destructors
    DatagramIterator( const IIteratable *buffer, size_t position )    // default to 0 pos?
        :  Data( reinterpret_cast<element_ptr>(const_cast<IIteratable *>(buffer)->Base()) ),     // Save Iteratable * ?
           Size(const_cast<IIteratable *>(buffer)->SizeInBytes()/sizeof(element_type) ),
           Pos( position )
        {
        }
    DatagramIterator( const DatagramIterator &other )
        :  Data( other.Data ),
           Size( other.Size ),
           Pos( other.Pos )
        {
        }
    ~DatagramIterator()
        {  }
    //
    //  Assignment operator
    DatagramIterator & operator = ( const DatagramIterator & other )
        {
        Data = other.Data;
        Size = other.Size;
        Pos  = other.Pos;
        return *this;
        }

public:
    //
    //  Accessors and Methods
    //  ...current method
    element_type & operator *() const
        {
        if ( Pos >= Size )
            ThrowDGRangeException( Pos );

        return Data[Pos];
        }
    element_type * operator ->() const
        {
        if( Pos >= Size )
            ThrowDGRangeException( Pos );

        return Data + Pos;
        }
    element_type & operator [] ( size_type index ) const
        {
        return Data[ Pos + index ];
        }
    element_type & at( size_type index ) const
        {
        if( Pos + index >= Size )
            ThrowDGRangeException( Pos + index );

        return Data[ Pos + index ];
        }

    DatagramIterator & operator +=( difference_type value )
        {
        Pos += value;
        if( Pos > Size + 1 )
            Pos = Size + 1;

        return *this;
        }
    DatagramIterator & operator -=( difference_type value )
        {
        element_ptr pos_new = Data + Pos;
        pos_new -= value;
        if (pos_new < Data)
            Pos = 0;
        else
            Pos -= value;

        return *this;
        }

    DatagramIterator operator +( difference_type value )
        {
        DatagramIterator iter = *this;
        iter += value;
        return iter;
        }
    DatagramIterator operator -( difference_type value )
        {
        DatagramIterator iter = *this;
        iter -= value;
        return iter;
        }
    // Span between 2 iterators
    difference_type operator -( DatagramIterator value )
        {
        return Pos - value.Pos;
        }

    DatagramIterator & operator ++()
        {
        ++Pos;
        if ( Pos > Size + 1 )
            Pos = Size + 1;

        return *this;
        }
    DatagramIterator operator ++(int)
        {
        DatagramIterator tmp = *this;
        Pos++;
        if ( Pos > Size + 1 )
            Pos = Size + 1;

        return tmp;
        }

    DatagramIterator & operator --()
        {
        if ( Pos )
          --Pos;

        return *this;
        }

    DatagramIterator operator --(int)
        {
        DatagramIterator tmp = *this;
        if( Pos )
          --Pos;

        return tmp;
        }

    bool operator == ( const DatagramIterator &other ) const
        {
        return ( Data == other.Data && Pos == other.Pos );
        }
    bool operator != ( const DatagramIterator &other )  const
        {
        return ( Data == other.Data && Pos != other.Pos );
        }
    bool operator > ( const DatagramIterator &other )  const
        {
        return ( Data == other.Data && Pos > other.Pos );
        }
    bool operator < ( const DatagramIterator &other )  const
        {
        return ( Data == other.Data && Pos < other.Pos );
        }
    bool operator >= ( const DatagramIterator &other )  const
        {
        return ( Data == other.Data && Pos >= other.Pos );
        }
    bool operator <= ( const DatagramIterator &other )  const
        {
        return ( Data == other.Data && Pos <= other.Pos );
        }

private:
    element_ptr         Data;
    size_type           Size;
    mutable size_type   Pos;
};

//------------------------------------------------------------------------------
//  operator + (n, DatagramIterator<T>) - Support wrong-way adding: (n + iter)
//------------------------------------------------------------------------------

template <class T>
DatagramIterator<T> operator + ( typename DatagramIterator<T>::difference_type value,
                                   DatagramIterator<T> & a_iter )
{
    DatagramIterator<T> iter = a_iter;
    iter += value;
    return iter;
}

//==============================================================================
//  CLASS IDatagrammable -- Interface required to support datagrams
//==============================================================================

class IDatagrammable
{
public:
    virtual ~IDatagrammable() {}

    virtual size_t          DatagramSize() = 0;
    virtual int *           DatagramBasePtr() = 0;
    virtual bool            MakeWritable() = 0;       // returns 'true' if buffer renewed
    virtual void            Resize(size_t size_in_ints) = 0;
	virtual size_t          PadBytes()			      //  bytes at the end are NOT full of data, we skip 'em		
								{  return 0;  }
};

//==============================================================================
//  CLASS AccessDatagram  -- General Access Datagram Interface Template
//==============================================================================

template <class T>
class AccessDatagram : public IIteratable
{
public:
    typedef T             element_type;    // the type of the access element
    typedef T*            element_ptr;     // type of pointer to the element
    typedef ptrdiff_t     difference_type;
    typedef size_t        size_type;

public:
    typedef DatagramIterator<element_type> iterator;
    typedef DatagramIterator<const element_type> const_iterator;

    typedef std::reverse_iterator<iterator> reverse_iterator;
    typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

    // STL compliance methods.
    size_t size() const
        {
        return Size;
        }
    size_t max_size() const             // never heard of this one
        {
        return size();
        }
    bool empty() const
        {
        return ( size() == 0 );
        }

    // Forward iterators.
    iterator begin()
        {
        write_split_n_update();
        return iterator( this, 0 );
        }
    const const_iterator begin() const
        {
        return const_iterator( this, 0 );
        }

    iterator end()
        {
        write_split_n_update();
        return iterator( this, size() );
        }
    const const_iterator end() const
        {
        return const_iterator( this, size() );
        }

    // Reverse iterators.
    reverse_iterator rbegin()
        {
        return reverse_iterator( end() );
        }
    const const_reverse_iterator rbegin() const
        {
        return const_reverse_iterator( end() );
        }

    reverse_iterator rend()
        {
        return reverse_iterator( begin() );
        }
    const const_reverse_iterator rend() const
        {
        return const_reverse_iterator( begin() );
        }
    //
    //  Constructors and Destructors
    AccessDatagram(const IDatagrammable & intf)
        : Intf(const_cast<IDatagrammable *>(&intf))
        {
        init();
        }
    AccessDatagram(IDatagrammable & intf)
        : Intf(&intf)
        {
        init();
        }
    AccessDatagram(IDatagrammable * intf)
        : Intf(intf)
        {
        init();
        }
    AccessDatagram(const IDatagrammable * intf)
        : Intf(const_cast<IDatagrammable *>(intf))
        {
        init();
        }
    virtual ~AccessDatagram()
    	{}

    //
    //  Random access interface
    // ....not range checked
    element_type operator [] ( size_type index ) const
        {
        return Data[ index ];
        }
    element_type & operator [] ( size_type index )
        {
        write_split_n_update();
        return Data[ index ];
        }
    // ....range checked
    element_type at( size_type index ) const
        {
        if( index >= Size )
            ThrowDGRangeException( index );
        return Data[ index ];
        }
    element_type & at( size_type index )
        {
        write_split_n_update();
        if( index >= Size )
            ThrowDGRangeException( index );
        return Data[ index ];
        }
	element_ptr Ptr() const
        {  return Data;  }
	element_ptr PtrAt(size_type index) const
        {  return &Data[ index ];  }

    void  Resize(size_t size_in_elements)
        {
        Intf->Resize(size_in_elements*sizeof(element_type)/sizeof(int));
        Renew();
        }
    void  Renew()                   // Renew info from interface
        {
        init();
        }
    size_t          SizeInElements() const
                         {  return Size;  }
    size_t          SizeInInts() const
                         {  return Size*sizeof(element_type)/sizeof(int); }
    size_t          ElementSizeBytes() const
                         {  return sizeof(element_type); }
    //
    //  IIteratable Support Functions
    virtual char *  Base() {  return reinterpret_cast<char *>(Data); }
    virtual size_t  SizeInBytes() const {  return Size*sizeof(element_type); }

protected:
    //
    //  Member Data
    element_ptr               Data;
    size_type                 Size;       // in elements
    IDatagrammable *          Intf;
    //
    //  Swappable AD support
    AccessDatagram()
        : Intf(0)
        {
        }
    void load_intf(IDatagrammable * intf)
        {
        Intf = intf;
        init();
        }

    void init()
        {
        Data = reinterpret_cast<element_ptr>(Intf->DatagramBasePtr());
		size_t  size_in_bytes = Intf->DatagramSize()*sizeof(int);
		if (Intf->PadBytes())
			{
			if (size_in_bytes < Intf->PadBytes())
				size_in_bytes = 0;
			else
				size_in_bytes -= Intf->PadBytes();
			}
        Size = size_in_bytes/sizeof(element_type);
        }
    void write_split_n_update()
        {
        if (Intf->MakeWritable())
            Renew();
        }
};

//==============================================================================
//  TEMPLATE CLASS SafeMemoryRangeAccess  -- "Safe" access to memory
//==============================================================================
//  This template makes access to a range safe by redirecting accesses that 
//    are out of range, or if the base is invalid, to a safe location

template <class T>
class SafeMemoryRangeAccess
{
public:
    typedef T             element_type;    // the type of the access element
    typedef T*            element_ptr;     // type of pointer to the element
    typedef ptrdiff_t     difference_type;
    typedef size_t        size_type;

    SafeMemoryRangeAccess(element_ptr base, size_type size, element_type oor = 0)
        :   Base(base), Size(size), OutOfRange(oor), Dummy(oor)
        {}

    void write_split_n_update()
        {
        //  memory is assumed to be unshared, so do nothing here
        }

    size_t size() const
        {
        return Size;
        }
    size_t max_size() const             // never heard of this one
        {
        return size();
        }
     element_type operator [] ( size_type index ) const
        {
        if (Base==0 || index>=Size)
            return OutOfRange;
        else
            return Base[ index ];
        }
    element_type & operator [] ( size_type index )
        {
        write_split_n_update();
        return Base[ index ];
        }

    element_ptr Ptr()
        {  return Base;  }

    void Detach()
        {  Base = 0; }
        
    void OOR(element_type oor)
        {  OutOfRange = oor;  Dummy = oor;  }

protected:
    //
    //  Member Data
    element_ptr               Base;
    size_type                 Size;       // in elements
    element_type              OutOfRange;
    element_type              Dummy;

};

//==============================================================================
//  SafeMemoryRange  -- "Safe" access to memory as unsigned integers
//==============================================================================

class SafeMemoryRange : public SafeMemoryRangeAccess<unsigned int>
{
public:
    SafeMemoryRange()
        :   SafeMemoryRangeAccess<unsigned int>(0, 0)
        {}
    SafeMemoryRange(element_ptr base, size_type size, element_type oor = 0)
        :   SafeMemoryRangeAccess<unsigned int>(base, size, oor)
        {}
};

//==============================================================================
//  CLASS DatagrammableSlice -- Datagrammable subset of IDatagramable
//==============================================================================

class DatagrammableSlice : public IDatagrammable
{
public:
    DatagrammableSlice(const IDatagrammable & intf, size_t init_ofs, size_t init_size)
         : Intf(const_cast<IDatagrammable *>(&intf)), Offset(init_ofs), Size(init_size)
         {
         init();
         }
    DatagrammableSlice(IDatagrammable & intf,  size_t init_ofs, size_t /*init_size*/)
         : Intf(&intf), Offset(init_ofs)
         {
         init();
         }
    DatagrammableSlice(IDatagrammable * intf,  size_t init_ofs, size_t /*init_size*/)
         : Intf(intf), Offset(init_ofs)
         {
         init();
         }
    DatagrammableSlice(const IDatagrammable * intf, size_t init_ofs, size_t /*init_size*/)
         : Intf(const_cast<IDatagrammable *>(intf)), Offset(init_ofs)
         {
         init();
         }
    virtual ~DatagrammableSlice() {}

    void Reslice(size_t new_ofs, size_t new_size)
        {  Offset=new_ofs;  Size=new_size;  init();  }

    virtual size_t  DatagramSize()
        {  return Size;  }
    virtual int *   DatagramBasePtr()
        {  return Intf->DatagramBasePtr() + Offset;  }
    virtual bool    MakeWritable()
        {  return Intf->MakeWritable();  }
    virtual void    Resize(size_t size_in_ints)
        {  Intf->Resize(size_in_ints);  init();  }
    virtual size_t  PadBytes()			      //  bytes at the end are NOT full of data, we skip 'em
        {  return Intf->PadBytes();  }
protected:
    //
    //  Member Data
    IDatagrammable *          Intf;
    size_t                    Offset;
    size_t                    Size;

    void  init()
        {
        // sanity check slice boundaries
        if (Offset>Intf->DatagramSize())
            Offset=Intf->DatagramSize();
        if (Offset+Size>Intf->DatagramSize())
            Size = Intf->DatagramSize()- Offset;
        }
};


#ifdef __CLR_VER
#pragma managed(pop)
#endif
}  // namespace Innovative

#endif
