//
// Ad9508_Tb_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2015 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <Ad9508_Tb_Mb.h>
#include <SystemSupport_Mb.h>
#include <cmath>
#include <FloatingPtSupport_Mb.h>

using namespace std;

#ifdef __GNUC__
    #pragma GCC diagnostic ignored "-Wunused-but-set-variable"
#endif

namespace Innovative
{

//=============================================================================
//  CLASS Ad9508_Tb  --  Ad9508_Tb Pll Implementation
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class Ad9508_Tb
//-----------------------------------------------------------------------------

const unsigned int  TotalReferences  = 1;
const unsigned int  TotalOutputs     = 4;
const int           SpiRegisterCount = 0x2D;
const float         DefaultMaxOutFreq = 1650.0e6;

const unsigned int  kMaxOutputDivider = 1024;


#ifdef LINUX
//    #pragma GCC diagnostic push
    #pragma GCC diagnostic ignored "-Wunused-but-set-variable"
#endif

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif
//-----------------------------------------------------------------------------
//  constructor for class Ad9508_Tb
//-----------------------------------------------------------------------------

Ad9508_Tb::Ad9508_Tb(IAddressingSpace & space, MapRegisters& regs, unsigned int instance)
    : PllTimebaseDevice(TotalReferences, TotalOutputs),
      RegisterImage(SpiRegisterCount),
      PllStatusReg(space, regs.PllControl_Addr),
      SpiControlReg(space, regs.PllSpi_Addr),
      SpiStatusReg(space, regs.PllSpi_Addr),
      Instance(instance),
      Debug(true)
{
    //  Fill max frequency checker array
    for (unsigned int i=0; i<TotalOutputs; i++)
        {
        OutputFrequencyLimit.push_back(DefaultMaxOutFreq);
        }

}

#ifdef _MSC_VER
	#pragma warning(pop)
#endif

//-----------------------------------------------------------------------------
//  destructor for class Ad9508_Tb
//-----------------------------------------------------------------------------

Ad9508_Tb::~Ad9508_Tb()
{
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::Init()  --  Initialize Chip
//-----------------------------------------------------------------------------

void Ad9508_Tb::Init()
{
    ReadDataRegisters();   // read initial state from chip
    SettingsInit();        // common INIT settings
    OnInit();
    Update();    // Update state of the device based on sources
    WriteDataRegisters();
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::SettingsInit()  --  Initial Settings Loading
//-----------------------------------------------------------------------------
//   Note board specific implementations can override these in OnInit

void Ad9508_Tb::SettingsInit()
{
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::Reapply()  --  Reapply settings to hardware from saved state
//-----------------------------------------------------------------------------

void Ad9508_Tb::Reapply()
{
    OutputSyncPin(false);
    OnReapply();

    Update();    // Update state of the device based on sources

    WriteDataRegisters();

    OutputSyncPin(true);
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::OutputSyncPin()  --  Toggle Sync
//-----------------------------------------------------------------------------

void Ad9508_Tb::OutputSyncPin(bool state)
{
    PllStatusReg.Sync.Value(state);
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::MaxOutputDivider()  --  Return max output post-divider
//-----------------------------------------------------------------------------

int Ad9508_Tb::MaxOutputDivider() const
{
    return kMaxOutputDivider;
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::FrequencyModeUpdate()  --  Update using Frequency as baseline
//-----------------------------------------------------------------------------

void Ad9508_Tb::FrequencyModeUpdate()
{
    DividerModeUpdate();
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::DividerModeUpdate() -- Update state of dividers
//-----------------------------------------------------------------------------
//  In this mode, the reference is sent to the outputs, with possible division

void  Ad9508_Tb::DividerModeUpdate()
{
    if (Reference() == 0 || Frequency() == 0)
        return;

    double divider_ratio = Reference()/Frequency();
    unsigned int OutDiv = NearestInteger(divider_ratio);
    SetOutputDividers(Reference(), OutDiv);
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::SetOutputDividers()  --
//-----------------------------------------------------------------------------

void  Ad9508_Tb::SetOutputDividers(double VcoFreq, unsigned int OutDiv)
{
    unsigned int BestDivider = OutDiv;
    double       OutputFreq  = VcoFreq/BestDivider;

    for (unsigned int output=0; output<TotalOutputs; output++)
        {
        if (OutputInUse(output)==true)
            {
            float   CandidateDivider = static_cast<float>(BestDivider);
            double  CandidateFreq    = OutputFreq;
            //
            //  If the Output is the Master, use the best values.
            //     if not the master, and Frequency is unspecified, then
            //     the make the output a copy of Master.
            //     Otherwise, figure the best divider possible for each output
            if ((output!=FMasterOutput) && (OutputFrequency(output)!=0.0))
                {
                CandidateFreq = OutputFrequency(output);
                CandidateDivider = static_cast<float>(VcoFreq / CandidateFreq);

                int above, below, divider;
                GetNearestDividers(output, CandidateDivider, below, above);

                // Find which divider is closest to desired frequency
                double centerFreq = VcoFreq/CandidateDivider;
                double hiFreq     = VcoFreq/below;
                double lowFreq    = VcoFreq/above;
                if (std::abs(centerFreq-lowFreq) > std::abs(hiFreq-centerFreq))
                    divider = below;
                else
                    divider = above;

                CandidateDivider = static_cast<float>(divider);
                }

            //
            //  Check for frequency exceeding applied Limits
            if (OutputFrequencyLimit[output]>=CandidateFreq)
                OutputDivider(output, static_cast<unsigned int>(CandidateDivider));
            else
                {
                double MinDivideRatio = VcoFreq/OutputFrequencyLimit[output];
                OutputDivider(output, BestDividerGreaterThan(static_cast<float>(MinDivideRatio)));
                }
            //
            //  Calculate Acutal Frequency
            if (OutputDividerArray[output]!=0)  // if illegal divider, skip it
                {
                OutputFreqActualArray[output] = VcoFreq / OutputDividerArray[output];
                }
            }
        else
            {
            OutputFreqActualArray[output] = 0.0;
            }
        }
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::Port()  --  Write to device data port
//-----------------------------------------------------------------------------

void Ad9508_Tb::Port(unsigned short reg, unsigned int value)
{
    // Wait for previous operation to complete
    for (unsigned int i = 0; (i < 10) && !PllStatusReg.SpiReady.Value(); ++i)
        Sleep(1);
    if (!PllStatusReg.SpiReady.Value())
        Debug.Log("PLL SPI WR ready timeout error!");

    bool device_bit = (Instance) ? true : false ;  // Dev 0 -> bit is 0

    SpiControlReg.Address.Value(reg);
    SpiControlReg.Data.Value(value);
    SpiControlReg.Rw.Value(0);
    SpiControlReg.PllSelect.Value(device_bit);
    SpiControlReg.Apply();

    std::stringstream msg;
    msg << "Ad9508_Tb::Port[0x" << std::hex << static_cast<int>(reg)
        << "]=0x" << static_cast<int>(value);
    Debug.Log(msg.str());

}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::Port()  --  Read from device data port
//-----------------------------------------------------------------------------

unsigned int Ad9508_Tb::Port(unsigned short reg)
{
    unsigned char result;

    // Wait for previous operation to complete
    for (unsigned int i = 0; (i < 10) && !PllStatusReg.SpiReady.Value(); ++i)
        Sleep(1);
    if (!PllStatusReg.SpiReady.Value())
        Debug.Log("PLL SPI RD ready timeout error!");

    bool device_bit = (Instance) ? true : false ;  // Dev 0 -> bit is 0

    SpiControlReg.Address.Value(reg);
    SpiControlReg.Rw.Value(1);
    SpiControlReg.PllSelect.Value(device_bit);
    SpiControlReg.Apply();

    // Wait for previous operation to complete
    for (unsigned int i = 0; (i < 10) && !PllStatusReg.SpiDataValid.Value(); ++i)
        Sleep(1);
    if (!PllStatusReg.SpiDataValid.Value())
        Debug.Log("PLL SPI RD data valid timeout error!");

    result = SpiStatusReg.Data.Value();

    return result;
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::SetDividerHdw()  --  Update the Divider setting hardware
//-----------------------------------------------------------------------------

void Ad9508_Tb::SetDividerHdw(unsigned int output, unsigned int divide_by)
{
    if (!IsValidDivider(divide_by))
        return;

    if (output < TotalOutputs)
        {
        int divider_value = divide_by-1;
        PtrBitManipulator Reg0(&RegisterImage[OutputOffset(output, 0)]);
        Reg0.Value(divider_value);     // lower 8 bits
        PtrBitManipulator Reg1(&RegisterImage[OutputOffset(output, 1)]);
        Reg1.Field(0, 2, divider_value>>8); // upper 2 bits
        }
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::SetPhaseAdjust()  --  Update the Phase Adjust setting
//-----------------------------------------------------------------------------

void Ad9508_Tb::SetPhaseAdjust(unsigned int output, unsigned int divider_setting,
                                                    unsigned int phase_setting)
{
    if (!IsValidDivider(divider_setting))
        return;

    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg2(&RegisterImage[OutputOffset(output, 2)]);
        Reg2.Value(phase_setting);     // lower 8 bits
        PtrBitManipulator Reg3(&RegisterImage[OutputOffset(output, 3)]);
        Reg3.Field(0, 3, phase_setting>>8); // upper 3 bits
        }
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::SetOutEnabledHdw()  --  Update the Output Enabled hardware
//-----------------------------------------------------------------------------

void Ad9508_Tb::SetOutEnabledHdw(unsigned int output, bool enabled)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg4(&RegisterImage[OutputOffset(output, 4)]);
        Reg4.Bit(7, enabled==false);  // actually divider power down
        }
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::SetOutputSync()  --  Enable sync of output divider
//-----------------------------------------------------------------------------

void Ad9508_Tb::SetOutputSync(unsigned int output, bool is_synced)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg4(&RegisterImage[OutputOffset(output, 4)]);
        Reg4.Bit(6, is_synced==false);
        }
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::SetOutputDriverPhaseMode()  --  Enable sync of output divider
//-----------------------------------------------------------------------------

void Ad9508_Tb::SetOutputDriverPhaseMode(unsigned int output, unsigned int setting)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg4(&RegisterImage[OutputOffset(output, 4)]);
        Reg4.Field(5, 2, setting);
        }
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::SetOutputConfig()  --  Update the Output Config for a channel
//-----------------------------------------------------------------------------

void Ad9508_Tb::SetOutputConfig(unsigned int output, unsigned int cfg_setting,
                                bool /*should_cache_setting*/)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg4(&RegisterImage[OutputOffset(output, 4)]);
        Reg4.Field(1, 3, cfg_setting);
        }
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::SetCmosP_Enabled()  --  enable the CMOS P side
//-----------------------------------------------------------------------------

void Ad9508_Tb::SetCmosP_Enabled(unsigned int output, bool enabled)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg5(&RegisterImage[OutputOffset(output, 5)]);
        Reg5.Bit(7, enabled);
        }
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::SetCmosP_DriverPhaseMode()  --  Set mode of CMOS P side
//-----------------------------------------------------------------------------

void Ad9508_Tb::SetCmosP_DriverPhaseMode(unsigned int output, unsigned int setting)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg5(&RegisterImage[OutputOffset(output, 5)]);
        Reg5.Field(5, 2, setting);
        }
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::SetCmosN_Enabled()  --  enable the CMOS P side
//-----------------------------------------------------------------------------

void Ad9508_Tb::SetCmosN_Enabled(unsigned int output, bool enabled)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg5(&RegisterImage[OutputOffset(output, 5)]);
        Reg5.Bit(4, enabled);
        }
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::SetCmosN_DriverPhaseMode()  --  Set mode of CMOS P side
//-----------------------------------------------------------------------------

void Ad9508_Tb::SetCmosN_DriverPhaseMode(unsigned int output, unsigned int setting)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg5(&RegisterImage[OutputOffset(output, 5)]);
        Reg5.Field(3, 2, setting);
        }
}


//-----------------------------------------------------------------------------
//  Ad9508_Tb::GetDivider()  --  read back the Divider setting
//-----------------------------------------------------------------------------

unsigned int  Ad9508_Tb::GetDivider(unsigned int output)
{
    unsigned int divider(1);
    if (output < TotalOutputs)
        {
        int divider_value = 0;
        PtrBitManipulator Reg0(&RegisterImage[OutputOffset(output, 0)]);
        divider_value = Reg0.Value();         // lower 8 bits
        PtrBitManipulator Reg1(&RegisterImage[OutputOffset(output, 1)]);
        divider_value += Reg1.Field(0, 2)<<8; // upper 2 bits
        divider = divider_value+1;
        }
    return divider;
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::GetDividerEnable()  --  Read the Divider Enable setting
//-----------------------------------------------------------------------------

bool Ad9508_Tb::GetDividerEnable(unsigned int output)
{
    bool bit = false;
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg4(&RegisterImage[OutputOffset(output, 4)]);
        bit = Reg4.Bit(7);
        }
    return bit;
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::OutputOffset()  --  Update the Divider setting
//-----------------------------------------------------------------------------

int Ad9508_Tb::OutputOffset(int bank, int offset)
{
    int base;
    switch(bank)
        {
        default:
        case 0:   base=0x15;  break;
        case 1:   base=0x1b;  break;
        case 2:   base=0x21;  break;
        case 3:   base=0x27;  break;
        }

    return base+offset;
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::ReadDataRegisters()  --  Read all data registers into Image vector
//-----------------------------------------------------------------------------

void Ad9508_Tb::ReadDataRegisters()
{
    for (unsigned int i=0; i<RegisterImage.size(); i++)
        {
        RegisterImage[i] = Port(i);
        }
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::WriteDataRegisters()  --  Write all data registers from Image vector
//-----------------------------------------------------------------------------

void Ad9508_Tb::WriteDataRegisters()
{
    ScopedTraceRef ccc(Debug, "Ad9508_Tb::WriteDataRegisters()");

    for (unsigned int i=0x15; i<RegisterImage.size(); i++)
        {
        Port(i, RegisterImage[i]);
        }
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::IsValidDivider() -- tests if divide_by is a valid divider
//-----------------------------------------------------------------------------

bool  Ad9508_Tb::IsValidDivider(unsigned int divide_by)
{
    if (divide_by <= kMaxOutputDivider)
        return true;
    else
        return false;
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::RoundToNearestInt() -- round fraction to int
//-----------------------------------------------------------------------------

//double  Ad9508_Tb::RoundToNearestInt(double number)
//{
//    double div = 1.0;
//    double fraction = modf(number, &div);
//    div += (fraction > 0.5) ? 1.0 : 0.0;
//    return div;
//}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::BestDividerGreaterThan() -- returns next largest valid Divider
//-----------------------------------------------------------------------------

unsigned int  Ad9508_Tb::BestDividerGreaterThan(float min_divide_by)
{
	if (min_divide_by > static_cast<float>(kMaxOutputDivider-1))
        return kMaxOutputDivider;
    else if (min_divide_by < 1)
        return 1;
    else
        return static_cast<unsigned int>(IntegerPart(min_divide_by)) + 1;
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::BestDividerLessThan() -- returns next smallest valid Divider
//-----------------------------------------------------------------------------

unsigned int  Ad9508_Tb::BestDividerLessThan(float max_divide_by)
{
    if (max_divide_by > static_cast<float>(kMaxOutputDivider))
        return kMaxOutputDivider;
    else if (max_divide_by < 2)
        return 1;
    else
        return static_cast<unsigned int>(IntegerPart(max_divide_by)) - 1;
}

//-----------------------------------------------------------------------------
//  Ad9508_Tb::Ad9508GetNearestDividers()  -- find divider set around a candidate
//-----------------------------------------------------------------------------

void   Ad9508_Tb::Ad9508GetNearestDividers(float candidate, int & below, int & above)
{
    if (IsAboutEqual(candidate, NearestIntegerPart(candidate)))
        {
        //  We are essentially an integer, so treat as such.
        unsigned int int_candidate(static_cast<unsigned int>(NearestInteger(candidate)));
        if (IsValidDivider(int_candidate))
            {
            above = below = int_candidate;
            return;
            }
        }
    above = BestDividerGreaterThan(candidate);
    below = BestDividerLessThan(candidate);

}


#ifdef LINUX
//    #pragma GCC diagnostic pop
#endif

} // namespace Innovative
