// Ad9511Support_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2010 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <VcxoControl_Mb.h>
#include <Ad9511Support_Mb.h>
#include <limits>
//#include <cmath>
#include <cstring>
#include <algorithm>
#include <FloatingPtSupport_Mb.h>

using namespace std;

namespace Innovative
{
//=============================================================================
//  CLASS Ad951xAutoTune
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class Ad951xAutoTune
//-----------------------------------------------------------------------------
static const int A_bits = 6;
static const int MaxA = (1 << A_bits) - 1;
static const int B_bits = 13;
static const int MaxB = (1 << B_bits) - 1;
\
//---------------------------------------------------------------------------
//  constructor for class Ad951xAutoTune
//---------------------------------------------------------------------------

Ad951xAutoTune::Ad951xAutoTune(IVcxoControl *Vco)
	: FSampleRate(0.0f), FMultiplier(1), FLowestRate(0.0f), FHighestRate(0.0f),
      FReferenceRate(100.0e6), FPfdRate(100.0e3), FPfdError(2.0f),
      FPostDivider(false),  FPostDividerStyle(pdsAd9511),
      FAllowOddDivisors(false), FAllowDecimation(false),
      FEnforceOutputMax(false), FPllOutputMaxRate(0.0),
      FVco(Vco)
{
    memset(&Best, 0, sizeof(Best));

	ConfigureValidDividers(Dividers);
	PostDivider(FPostDivider);
	ConfigureValidPrescalers();
}

//---------------------------------------------------------------------------
//  Ad951xAutoTune::Calculate()
//---------------------------------------------------------------------------

bool Ad951xAutoTune::Calculate()
{
    bool result = true;

    // Insure that
    Best.Decimation = Best.R = Best.D[0] = Best.D[1] = 1;

    double delta = (FPfdError/100)*std::fabs(FPfdRate);
    double PfdLo = FPfdRate - delta;
    double PfdHi = FPfdRate + delta;

    int R_Lo = static_cast<int>(FReferenceRate/PfdHi);
    int R_Hi = static_cast<int>(FReferenceRate/PfdLo);

    IntArray::reverse_iterator j = PostDividers.rbegin();
    IntArray::reverse_iterator i = Dividers.rbegin();
    FLowestRate = VcoLow() /  (FMultiplier*(*j)*(*i));
    FHighestRate = VcoHigh() / FMultiplier;

    double samplerate = FSampleRate;
 
    if (!WithinRange(samplerate, FLowestRate, FHighestRate))
        {
        if (FAllowDecimation && (samplerate < FLowestRate))
            {
            Best.Decimation = NearestInteger(FLowestRate / FSampleRate);
            samplerate *= Best.Decimation;
            }

        result = false;
        }

    samplerate = std::min<double>(samplerate, FHighestRate);
    samplerate = std::max<double>(samplerate, FLowestRate);

    bool found = false;
    double error = std::numeric_limits<double>::max();
    for (IntArray::const_iterator j = PostDividers.begin(); j != PostDividers.end(); ++j)
        {
        int post = *j;

        for (IntArray::const_iterator i = Dividers.begin(); i != Dividers.end(); ++i)
            {
            int pre = *i;
            double dividers = FMultiplier*pre*post;
            double setpoint = samplerate*dividers;

            // Clamp setpoint if out of range
            if (setpoint <  VcoLow())
                setpoint = VcoLow();
            else if (setpoint >  VcoHigh())
                setpoint = VcoHigh();

            for (int R = R_Lo; R <= R_Hi; ++R)
                {
                double PfdRate = FReferenceRate/R;
                double CandidateVco = setpoint;
                double N = setpoint/PfdRate;                                

                int P;
                int P_bits;
                int B;
                int A;

                AdvancedPBACalculation(static_cast<int>(N), P, P_bits, B, A, CandidateVco);

                double out    = (PfdRate * (P*B + A))/dividers;
                double VcoOut = (PfdRate * (P*B + A));
                double err    = std::abs(samplerate-out);

                // we enforce device output rate, not post-divider out rate
                if (FEnforceOutputMax && (out*post)>FPllOutputMaxRate)
                    continue;   // cases where PLL output above max can never be good

                if ((err <= error) && FVco->CanTuneTo(VcoOut))
                    {
                    found = true;
                    error = err;
                    Best.P = P;
                    Best.P_bits = P_bits;
                    Best.R = R;
                    Best.B = B;
                    Best.A = A;
                    Best.D[0] = pre;
                    Best.D[1] = post;
                    }
                }
            }
        }

    if (!found)
        result = false;

    FActualPfdRate = FReferenceRate/Best.R;
    FActualVcoRate = FReferenceRate * (Prescaler()*Best.B + Best.A) / Best.R;
    FActualVcoRate = std::min<double>(FActualVcoRate, VcoHigh());
    FActualVcoRate = std::max<double>(FActualVcoRate, VcoLow());

    FActualSampleRateBeforePostDivider = FActualVcoRate / (Best.D[0]*FMultiplier);
    FActualSampleRateBeforeDecimation  = FActualVcoRate / (Best.D[0]*Best.D[1]*FMultiplier);
    FActualSampleRate                  = FActualVcoRate / (Best.D[0]*Best.D[1]*FMultiplier*Best.Decimation);

    return result;
}

//---------------------------------------------------------------------------
//  Ad951xAutoTune::Calculate_External()
//---------------------------------------------------------------------------
//  Calculate dividers, post_dividers, and decimation for the 'no pll' case

bool Ad951xAutoTune::Calculate_External()
{
    bool result = true;

    // Insure that
    Best.Decimation = Best.R = Best.D[0] = Best.D[1] = 1;

    //  Since we are dividing down from the Reference, it is the max rate.
    IntArray::reverse_iterator j = PostDividers.rbegin();
    IntArray::reverse_iterator i = Dividers.rbegin();
    FLowestRate = FReferenceRate /  (FMultiplier*(*j)*(*i));
    FHighestRate = FReferenceRate / FMultiplier;

    double samplerate = FSampleRate;

    if (!WithinRange(samplerate, FLowestRate, FHighestRate))
        {
        if (FAllowDecimation && (samplerate < FLowestRate))
            {
            Best.Decimation = NearestInteger(FLowestRate / FSampleRate);
            samplerate *= Best.Decimation;
            }

        result = false;
        }

    samplerate = std::min<double>(samplerate, FHighestRate);
    samplerate = std::max<double>(samplerate, FLowestRate);

    bool found = false;
    double error = std::numeric_limits<double>::max();
    for (IntArray::const_iterator j = PostDividers.begin(); j != PostDividers.end(); ++j)
        {
        int post = *j;

        for (IntArray::const_iterator i = Dividers.begin(); i != Dividers.end(); ++i)
            {
            int pre = *i;
            double dividers = FMultiplier*pre*post;

            double out    = FReferenceRate/dividers;
            double err    = std::abs(samplerate-out);

            if (FEnforceOutputMax && out>FPllOutputMaxRate)
                continue;   // cases where PLL output above max can never be good

            if (err <= error)
                {
                found = true;
                error = err;
                Best.D[0] = pre;
                Best.D[1] = post;
                }
            }
        }

    if (!found)
        result = false;

    FActualPfdRate = 0;
    FActualVcoRate = 0;

    FActualSampleRateBeforePostDivider = FReferenceRate / (Best.D[0]*FMultiplier);
    FActualSampleRateBeforeDecimation  = FReferenceRate / (Best.D[0]*Best.D[1]*FMultiplier);
    FActualSampleRate                  = FReferenceRate / (Best.D[0]*Best.D[1]*FMultiplier*Best.Decimation);

    return result;
}

//-----------------------------------------------------------------------------
//  Ad951xAutoTune::AdvancedPBACalculation()  --  Advanced Calculation
//-----------------------------------------------------------------------------

void  Ad951xAutoTune::AdvancedPBACalculation(int N, int & P, int & P_bits, int & B, int & A, double VcoRate)
{
    // calc P, B, A terms such that N = P*B+A

    //  We need to pick a P that reduces B and A input to 300 MHz, and does
    //    not violate P input rate restrictions for Dual Modulus Modes
    bool DM_Mode(false);
    if (VcoRate<=300.0e6)
        {
        P = 1;  P_bits = 0x0;  DM_Mode = false;
        }
    else if (VcoRate<=600.0e6)
        {
        P = 2;  P_bits = 0x1;  DM_Mode = false;
        }
    else if (VcoRate<=900.0e6)
        {
        P = 3;  P_bits = 0x7;  DM_Mode = false;
        }
    else if (VcoRate<=1000.0e6)
        {
        P = 4;  P_bits = 0x3;  DM_Mode = true;
        }
    else  // if (VcoRate<=1600.0e6)
        {
        P = 8;  P_bits = 0x4;  DM_Mode = true;
        }
    //
    //  Calculate B from P and N
    static const int MaxB = 1 << 13;
    B = std::max(1, std::min(N/P, MaxB-1));
    if (B == 2)
        {
        if (DM_Mode)
            {
            B = 1;  // A will cover the gap
            }
        else
            {
            B = 1;
            P = 2;   P_bits = 1;     // single mod P=2
            }
        }
    //
    //  Calculate A
    if (DM_Mode)
        A = std::max(std::min(N - (P * B), 63), 0);
    else
        A = 0;
}

//-----------------------------------------------------------------------------
// Ad951xAutoTune::ConfigureValidDividers() -  Initialize table of dividers
//-----------------------------------------------------------------------------

void Ad951xAutoTune::ConfigureValidDividers(IntArray & dividers)
{
	dividers.clear();

	dividers.push_back(1);
	for (int i = 2; i <= 32; ++i)
		{
		bool odd = i&1;
		if (!odd || (odd && FAllowOddDivisors))
			dividers.push_back(i);
		}
}

//-----------------------------------------------------------------------------
// Ad951xAutoTune::ConfigureValidPrescalers()- Initialize table of valid prescalers
//-----------------------------------------------------------------------------

void Ad951xAutoTune::ConfigureValidPrescalers()
{
	Prescalers.clear();

	Prescalers.push_back(1);
	Prescalers.push_back(2);
	Prescalers.push_back(2);
	Prescalers.push_back(4);
	Prescalers.push_back(8);
	Prescalers.push_back(16);
	Prescalers.push_back(32);
    Prescalers.push_back(3);
    }

//--------------------------------------------------------------------------------
// Ad951xAutoTune::ConfigurePostDividers()- configure post_dividers based by type
//-----------------------------------------------------------------------------

void Ad951xAutoTune::ConfigurePostDividers()
{
    switch (FPostDividerStyle)
        {
        default:
        case pdsAd9511:
            ConfigureValidDividers(PostDividers);
            break;
        case pdsAd9508:
            //  1024 divides, all 50 pct duty cycle.
            PostDividers.clear();
            for (int i = 1; i <= 1024; ++i)
                PostDividers.push_back(i);
            break;
        case pdsAd9508_SDF:
            //  FMC_SDF requires restricted divider set to match ADC
            PostDividers.clear();

            //PostDividers.push_back(16);  // no good on AD7763?
            PostDividers.push_back(32);
            PostDividers.push_back(64);
            PostDividers.push_back(128);
            PostDividers.push_back(256);
            break;

        }
}

//-----------------------------------------------------------------------------
// Ad951xAutoTune::AllowOddDivisors() -
//-----------------------------------------------------------------------------

void Ad951xAutoTune::AllowOddDivisors(bool state)
{
	FAllowOddDivisors = state;
    ConfigurePostDividers();
	PostDivider(FPostDivider);
}

//-----------------------------------------------------------------------------
// Ad951xAutoTune::PostDivider() -
//-----------------------------------------------------------------------------

void Ad951xAutoTune::PostDivider(bool state)
{
    FPostDivider = state;
	if (state)
		{
        ConfigurePostDividers();
        }
	else
		{
		PostDividers.clear();
		PostDividers.push_back(1);
		}
}

} // namespace Innovative
