//
// Ad9511_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2003 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <Ad9511_Mb.h>
#include <Ad9511Support_Mb.h>
#include <SystemSupport_Mb.h>
#include <algorithm>
#include <limits>
#include <cmath>
#include <Magic_Mb.h>
#include <Trace_Mb.h>
#include <sstream>

//---------------------------------------------------------------------------
static const double _VcoLow = 200.0e6;
static const double _VcoHigh = 270.0e6;

namespace Innovative
{

//=============================================================================
//  CLASS Ad9511  --  Ad9511 Pll Implementation
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class Ad9511
//-----------------------------------------------------------------------------

Ad9511::Ad9511(int index, AddressingSpace & space, int ctl)
	: ClockBase(index),
      Vco(new RangedVcxoControl(_VcoLow, _VcoHigh)),
      Tuner(Vco),
      ControlReg(space, ctl), StatusReg(space, ctl),
      FOutputClock(oVco), FReferenceClock(rInternal),
	  FAllowAsymmetricDutyCycle(false), FPFDPolarity(true),
	  FChargePumpCurrent(c600), FUseAdvancedPBACalculation(false),
      Debug(true)
{
    Tuner.ReferenceRate(Reference());
    Tuner.Multiplier(1);
    Tuner.AllowDecimation(true);
}

//-----------------------------------------------------------------------------
//  constructor for class Ad9511
//-----------------------------------------------------------------------------

Ad9511::Ad9511(double freq, double ref, bool enabled, int index,
				 AddressingSpace & space, int ctl)
	: ClockBase(freq, ref, enabled, index),
	  Vco(new RangedVcxoControl(_VcoLow, _VcoHigh)),
      Tuner(Vco),
	  ControlReg(space, ctl),
	  StatusReg(space, ctl),
	  FOutputClock(oVco), FReferenceClock(rInternal),
	  FAllowAsymmetricDutyCycle(false), FPFDPolarity(true),
	  FChargePumpCurrent(c600), FUseAdvancedPBACalculation(false),
      Debug(true)
{
	DoSetFrequency(freq);
	DoSetEnabled(enabled);

    Tuner.ReferenceRate(Reference());
    Tuner.Multiplier(1);
    Tuner.AllowDecimation(true);
}

//-----------------------------------------------------------------------------
//  Ad9511::LoadVco()  --  Replace Vco with a new Vco object
//-----------------------------------------------------------------------------

void Ad9511::LoadVco(IVcxoControl * aVco)
{
    if (aVco)   // we need a valid Vco object to work
        {
        IVcxoControl * tmp = Vco;
        Vco = aVco;
        Tuner.LoadVco(Vco);
        delete tmp;
        }
}

//-----------------------------------------------------------------------------
//  Ad9511::Max()  --  Set maximum output rate
//-----------------------------------------------------------------------------

void Ad9511::Max(double value)
{
    Vco->SetRange(Vco->Min(), value);
}

//-----------------------------------------------------------------------------
//  Ad9511::Decimation()  --  Return decimation factor
//-----------------------------------------------------------------------------

int	Ad9511::Decimation() const
{
    if (FOutputClock==oVco)
        return Tuner.Decimation();
    else
        return 1;
}

//-----------------------------------------------------------------------------
//  Ad9511::Max()  --  Return maximum output rate
//-----------------------------------------------------------------------------

double Ad9511::Max() const
{
	return Vco->Max();
}

//-----------------------------------------------------------------------------
//  Ad9511::Min()  --  Set minimum output rate
//-----------------------------------------------------------------------------

void Ad9511::Min(double value)
{
    Vco->SetRange(value * MaxOutputDivider(), Vco->Max());
}

//-----------------------------------------------------------------------------
//  Ad9511::Min()  --  Return minimum output rate
//-----------------------------------------------------------------------------

double Ad9511::Min() const
{
	return Vco->Min() / MaxOutputDivider();
}

//-----------------------------------------------------------------------------
//  Ad9511::MaxOutputDivider()  --  Return max output post-divider
//-----------------------------------------------------------------------------

int Ad9511::MaxOutputDivider() const
{
	return 32;
}

//---------------------------------------------------------------------------
//  Ad9511::OutputClock() -  Specify output
//---------------------------------------------------------------------------

void Ad9511::OutputClock(IIOutputClock value)
{
	FOutputClock = value;
}

//---------------------------------------------------------------------------
//  Ad9511::Mux() -  Query output
//---------------------------------------------------------------------------

Ad9511::IIOutputClock Ad9511::OutputClock() const
{
	return FOutputClock;
}

//---------------------------------------------------------------------------
//  Ad9511::ReferenceClock() -  Select reference clock source
//---------------------------------------------------------------------------

void Ad9511::ReferenceClock(IIReferenceClock value)
{
	FReferenceClock = value;
    ControlReg.ReferenceSource.Value((value == rExternal) ? 1 : 0);
}

//---------------------------------------------------------------------------
//  Ad9511::Mux() -  Query reference clock source
//---------------------------------------------------------------------------

Ad9511::IIReferenceClock Ad9511::ReferenceClock() const
{
	return FReferenceClock;
}

//-----------------------------------------------------------------------------
//  Ad9511::GetFrequencyActual()  --  Calculate actual clock rate
//-----------------------------------------------------------------------------

double Ad9511::GetFrequencyActual() const
{
	return FFrequencyActual/FClockFactor;
}

//-----------------------------------------------------------------------------
//  Ad9511::SetFrequency()  --  Calculate settings for a frequency
//-----------------------------------------------------------------------------

void Ad9511::SetFrequency(double freq)
{
	DoSetFrequency(freq);
}

//-----------------------------------------------------------------------------
//  Ad9511::Round()  --  Round to nearest integer
//-----------------------------------------------------------------------------

double Ad9511::Round(double original)
{
	if ((original-std::floor(original))<0.50)
		return std::floor(original);
	else
		return std::ceil(original);
}

//-----------------------------------------------------------------------------
//  Ad9511::DoSetFrequency()  --  Calculate settings for a frequency
//-----------------------------------------------------------------------------

void Ad9511::DoSetFrequency(double freq)
{
	Debug.Log("Ad9511 DoSetFrequency() Begins...: ");

	if (FReference == 0.0)
		return;

	//  Cache Frequency
    freq *= FClockFactor;

	// initialize device serial port
	Port(0x00, 0xB0);
	Port(0x00, 0x90);

	// Search for optimum PLL and VCO settings.
	// Ensure 50/50 duty cycle output)
	if (FOutputClock==oVco)
		{
        Tuner.ReferenceRate(Reference());
        Tuner.SampleRate(freq);

        Tuner.Calculate();

        int R  = Tuner.Best.R;
        int P  = Tuner.Best.P;
        int P_bits  = Tuner.Best.P_bits;
        int B  = Tuner.Best.B;
        int A  = Tuner.Best.A;
        outdiv = Tuner.Best.D[0];

        Debug.Log("Ad9511::R = ", R);
        Debug.Log("Ad9511::P = ", P);
        Debug.Log("Ad9511::P_bits = ", P_bits);
        Debug.Log("Ad9511::B = ", B);
        Debug.Log("Ad9511::A = ", A);
        Debug.Log("Ad9511::OutDiv = ", outdiv);

        Debug.Log("Pll Tuning Event");
        PllTuningEvent e(R, P_bits, B, A, outdiv);
        OnPllTuning.Execute(e);

        P_bits &= 7;
        static short PMap[] = {1, 2, 2, 4, 8, 16, 32, 3};
        P = PMap[P_bits];

        Debug.Log("Ad9511::R = ", R);
        Debug.Log("Ad9511::P = ", P);
        Debug.Log("Ad9511::P_bits = ", P_bits);
        Debug.Log("Ad9511::B = ", B);
        Debug.Log("Ad9511::A = ", A);
        Debug.Log("Ad9511::OutDiv = ", outdiv);
        Debug.Log("Ad9511::Decimation = ", Decimation());

        // Cache actual frequency
        FFrequencyActual = (Reference() * (P * B + A))/(outdiv * R );

        // correct if too low
        if(FFrequencyActual < (Vco->Min()/outdiv))
            FFrequencyActual = Vco->Min()/outdiv;

        // correct if too high
        if(FFrequencyActual > (Vco->Max()/outdiv))
            FFrequencyActual = Vco->Max()/outdiv;

		Debug.Log("Ad9511::FrequencyActual = ", FFrequencyActual);
		Debug.Log("Ad9511::Vco Frequency = ", FFrequencyActual*outdiv);
        
		// Attempt to set the Vco to the frequency desired
        Vco->SetVcoFrequency(FFrequencyActual*outdiv);

		Port(0x04, A);				// A
		Port(0x05, (B >> 8));		// B MSBs
		Port(0x06, (B & 0xff));		// B LSBs
        unsigned char Cpm = 0x07 | (FPFDPolarity ? 0x40 : 0x00);
		Port(0x08, Cpm);			// CP mode = normal, PFD polarity (as above),
									// status pin = PLL lock
		Port(0x0A, ((B == 1) ? 0x40 : 0x0) | (P_bits << 2));	
		Port(0x0B, (R >> 8));		// R MSBs
		Port(0x0C, (R & 0xff));		// R LSBs
		Port(0x0D, 0x00);			// default anti-backlash
		}
	else
		{
	    Debug.Log("...Using External Clock...: ");
        if (freq)
            outdiv = std::max(std::min(static_cast<int>(Round(Reference()/freq)), 32), 1);
		else
			outdiv = 32;

		if (!FAllowAsymmetricDutyCycle && (outdiv>1))
			outdiv &= ~1;

		FFrequencyActual = Reference()/outdiv;
		}

	// Clock distribution setup
	Debug.Log("Ad9511 output divider: ", outdiv);
	// encode outdiv for register use
	int outdiv_bits = ToDividerValue(outdiv);
	Debug.Log("Ad9511 outdiv bits: ", outdiv_bits);

	// disable output dividers if outdiv = 1
	int outdiv_bypass = ToBypassValue(outdiv);

	// all clock inputs on, PLL output is as selected
	unsigned char ClkOut = (FOutputClock==oVco)? 0 : 1;
	Port(0x45, ClkOut);

	ConfigureOutputs(outdiv_bits, outdiv_bypass);
	Debug.Log("Pll Divider Event");

	// Adjust CP drive current
	Port(0x09, FChargePumpCurrent << 4);

	// update registers
	Port(0x58, 0x04);     // set soft SYNC
	Port(0x5A, 0x01);
	Port(0x58, 0x00);     // Release soft SYNC
	Port(0x5A, 0x01);

	FEnabled = true;
}


//-----------------------------------------------------------------------------
//  Ad9511::ConfigureOutputs()  --  Set clock output modes and signalling types
//
// This version is tailored for the ClockGen module
//-----------------------------------------------------------------------------

void Ad9511::ConfigureOutputs(int outdiv_bits, int outdiv_bypass)
{
    PllConfigureOutputEvent e(*this, outdiv_bits, outdiv_bypass);
    OnPllConfigureOutput.Execute(e);

	Port(0x41, 0x01);
	Port(0x4a, outdiv_bits);	// output 0 divider
	Port(0x4b, outdiv_bypass);	// set divider 0 bypass
	Port(0x3D, 0x08);
	Port(0x3E, 0x03);
	Port(0x3F, 0x03);
	Port(0x40, 0x03);
}

//-----------------------------------------------------------------------------
//  Ad9511::Locked()  --  Return PLL lock status
//-----------------------------------------------------------------------------

bool  Ad9511::Locked() const
{
	return StatusReg.Locked() ? true : false;
}

//-----------------------------------------------------------------------------
//  Ad9511::SetEnabled()  --  Activate / Deactivate output
//-----------------------------------------------------------------------------

void  Ad9511::SetEnabled(bool state)
{
	DoSetEnabled(state);
}

//-----------------------------------------------------------------------------
//  Ad9511::DoSetEnabled()  --  Activate / Deactivate output
//-----------------------------------------------------------------------------

void  Ad9511::DoSetEnabled(bool /*state*/)
{
}

//-----------------------------------------------------------------------------
//  Ad9511::OutputDivider()  --  Set best-fit output divider
//-----------------------------------------------------------------------------

void Ad9511::OutputDivider(int value)
{
	outdiv = value;
}

//-----------------------------------------------------------------------------
//  Ad9511::OutputDivider()  --  Return best-fit output divider
//-----------------------------------------------------------------------------

int Ad9511::OutputDivider() const
{
	return outdiv;
}

//-----------------------------------------------------------------------------
//  Ad9511::PFDPolarity()  --  Set Phase-Frequency Detector (PFD) Polarity
//                             0 = Negative (Default), 1 = Positive.
//-----------------------------------------------------------------------------

void Ad9511::PFDPolarity(bool state)
{
    FPFDPolarity = state;
}

//-----------------------------------------------------------------------------
//  Ad9511::PFDPolarity()  --  Return Phase-Frequency Detector (PFD) Polarity
//-----------------------------------------------------------------------------

bool Ad9511::PFDPolarity() const
{
    return FPFDPolarity;
}

//-----------------------------------------------------------------------------
//  Ad9511::Reapply()  --  Reapply settings to hardware from saved state
//-----------------------------------------------------------------------------

void  Ad9511::Reapply()
{
    //  Set control register first to get reference right first
	ControlReg.Apply();
	Debug.Log("Ad9511 Load Control Register: ", (int)ControlReg.Value());

    Frequency(FFrequency);
}

//-----------------------------------------------------------------------------
//  Ad9511::AllowAsymmetricDutyCycle()  --
//-----------------------------------------------------------------------------

void Ad9511::AllowAsymmetricDutyCycle(bool state)
{
	FAllowAsymmetricDutyCycle = state;
}

//-----------------------------------------------------------------------------
//  Ad9511::AllowAsymmetricDutyCycle()  --
//-----------------------------------------------------------------------------

bool Ad9511::AllowAsymmetricDutyCycle() const
{
	return FAllowAsymmetricDutyCycle;
}

//-----------------------------------------------------------------------------
//  Ad9511::Port()  --  Set charge-pump current mode
//-----------------------------------------------------------------------------

void Ad9511::ChargePumpCurrent(IIChargePumpCurrent mode)
{
	FChargePumpCurrent = mode;
}

//-----------------------------------------------------------------------------
//  Ad9511::Port()  --  Query charge-pump current mode
//-----------------------------------------------------------------------------

Ad9511::IIChargePumpCurrent	Ad9511::ChargePumpCurrent() const
{
	return FChargePumpCurrent;
}

//-----------------------------------------------------------------------------
//  Ad9511::Port()  --  Write to PLL data port
//-----------------------------------------------------------------------------

void Ad9511::Port(unsigned char reg, unsigned char value)
{
	std::stringstream msg;
	msg << "Ad9511::Write 0x" << std::hex << static_cast<int>(value)
		<< " to reg " << static_cast<int>(reg);
	Debug.Log(msg.str());

	unsigned short regval = reg;
	int data = (regval<<8) | value;
	ControlReg.Value(data);
	Sleep(5);
}

//-----------------------------------------------------------------------------
//  Ad9511::Port()  --  Read from PLL data port
//-----------------------------------------------------------------------------

unsigned char Ad9511::Port(unsigned char reg)
{
	unsigned short regval = reg;
	int data = (1 << 23) | (regval<<8);
	ControlReg.Value(data);
	Sleep(10);

	return static_cast<unsigned char>(StatusReg.Value() & 0xff);
}

//---------------------------------------------------------------------------
//  Ad9511::ToDividerValue()  -- "Divider" register value from a divider
//---------------------------------------------------------------------------
//  Calculate bits written given a divider

int  Ad9511::ToDividerValue(int divide_by)
{
	const int odd = (divide_by < 32) ? (divide_by & 1) : 0;
	const int lo = (divide_by >> 1) - 1;
	const int hi = lo + odd;
	int divider_bits = ( (hi << 4) | lo );
    return divider_bits;
}

//---------------------------------------------------------------------------
//  Ad9511::ToBypassValue()  -- "Bypass" register value from a divider
//---------------------------------------------------------------------------

int  Ad9511::ToBypassValue(int divide_by)
{
	if (divide_by==1)
		return 0x80;	
	else
		return 0x00;
}

//=============================================================================
//  CLASS A9511Pfd  --  Ad9511Pfd Pll Implementation
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class Ad9511Pfd
//-----------------------------------------------------------------------------

Ad9511Pfd::Ad9511Pfd(int index, AddressingSpace & space, int ctl)
	: Ad9511(index, space, ctl)
{
    UseAdvancedPBACalculation(true);	
}

//-----------------------------------------------------------------------------
//  constructor for class Ad9511Pfd
//-----------------------------------------------------------------------------

Ad9511Pfd::Ad9511Pfd(double freq, double ref, bool enabled, int index,
				 AddressingSpace & space, int ctl)
	:  Ad9511(freq, ref, enabled, index, space, ctl)

{
    DoSetFrequency(freq);
	DoSetEnabled(enabled);
    UseAdvancedPBACalculation(true);
}




} // namespace Innovative
