//
// Ad9516_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2003 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <Ad9516_Mb.h>
#include <SystemSupport_Mb.h>
#include <algorithm>
#include <limits>
#include <cmath>
#include <Magic_Mb.h>
#include <Trace_Mb.h>
#include <sstream>

//---------------------------------------------------------------------------
static const double _VcoLow = 2550.0e6;
static const double _VcoHigh = 2995.0e6;

namespace Innovative
{

//=============================================================================
//  CLASS Ad9516  --  Ad9516 Pll Implementation
//=============================================================================

//-----------------------------------------------------------------------------
//  constructor for class Ad9516
//-----------------------------------------------------------------------------

Ad9516::Ad9516(int index, AddressingSpace & space, int ctl, int spi)
	: ClockBase(index),
	  VcoLow(_VcoLow), VcoHigh(_VcoHigh),
      SpiControlReg(space, spi), SpiStatusReg(space, spi),
      PllControlReg(space, ctl), PllStatusReg(space, ctl),
      FOutputClock(oVco), FReferenceClock(rInternal),
	  FPFDPolarity(true), Debug(true)
{
}

//-----------------------------------------------------------------------------
//  constructor for class Ad9516
//-----------------------------------------------------------------------------

Ad9516::Ad9516(double freq, double ref, bool enabled, int index,
				 AddressingSpace & space, int ctl, int spi)
	: ClockBase(freq, ref, enabled, index),
      VcoLow(_VcoLow), VcoHigh(_VcoHigh),
      SpiControlReg(space, spi), SpiStatusReg(space, spi),
      PllControlReg(space, ctl), PllStatusReg(space, ctl),
      FOutputClock(oVco), FReferenceClock(rInternal),
	  FPFDPolarity(true), Debug(true)
{
	DoSetFrequency(freq);
	DoSetEnabled(enabled);
}

//-----------------------------------------------------------------------------
//  Ad9516::Max()  --  Set maximum output rate
//-----------------------------------------------------------------------------

void Ad9516::Max(double value)
{
	VcoHigh = value;
}

//-----------------------------------------------------------------------------
//  Ad9516::Max()  --  Return maximum output rate
//-----------------------------------------------------------------------------

double Ad9516::Max() const
{
	return VcoHigh;
}

//-----------------------------------------------------------------------------
//  Ad9516::Min()  --  Set minimum output rate
//-----------------------------------------------------------------------------

void Ad9516::Min(double value)
{
	VcoLow = value * MaxOutputDivider();
}

//-----------------------------------------------------------------------------
//  Ad9516::Min()  --  Return minimum output rate
//-----------------------------------------------------------------------------

double Ad9516::Min() const
{
	return VcoLow / MaxOutputDivider();
}

//-----------------------------------------------------------------------------
//  Ad9516::MaxOutputDivider()  --  Return max output post-divider
//-----------------------------------------------------------------------------

int Ad9516::MaxOutputDivider() const
{
	return 32;
}

//---------------------------------------------------------------------------
//  Ad9516::OutputClock() -  Specify output
//---------------------------------------------------------------------------

void Ad9516::OutputClock(IIOutputClock value)
{
	FOutputClock = value;
    PllControlReg.ClockSource.Value((value == oExternal) ? 1 : 0);
}

//---------------------------------------------------------------------------
//  Ad9516::Mux() -  Query output
//---------------------------------------------------------------------------

Ad9516::IIOutputClock Ad9516::OutputClock() const
{
	return FOutputClock;
}

//---------------------------------------------------------------------------
//  Ad9516::ReferenceClock() -  Select reference clock source
//---------------------------------------------------------------------------

void Ad9516::ReferenceClock(IIReferenceClock value)
{
	FReferenceClock = value;
    PllControlReg.ReferenceSource.Value((value == rExternal) ? 1 : 0);
}

//---------------------------------------------------------------------------
//  Ad9516::Mux() -  Query reference clock source
//---------------------------------------------------------------------------

Ad9516::IIReferenceClock Ad9516::ReferenceClock() const
{
	return FReferenceClock;
}

//-----------------------------------------------------------------------------
//  Ad9516::GetFrequencyActual()  --  Calculate actual clock rate
//-----------------------------------------------------------------------------

double Ad9516::GetFrequencyActual() const
{
	return FFrequencyActual/FClockFactor;
}

//-----------------------------------------------------------------------------
//  Ad9516::SetFrequency()  --  Calculate settings for a frequency
//-----------------------------------------------------------------------------

void Ad9516::SetFrequency(double freq)
{
	DoSetFrequency(freq);
}

//-----------------------------------------------------------------------------
//  Ad9516::Round()  --  Round to nearest integer
//-----------------------------------------------------------------------------

double Ad9516::Round(double original)
{
	if ((original-std::floor(original))<0.50)
		return std::floor(original);
	else
		return std::ceil(original);
}

//-----------------------------------------------------------------------------
//  Ad9516::DoSetFrequency()  --  Calculate settings for a frequency
//-----------------------------------------------------------------------------

void Ad9516::DoSetFrequency(double freq)
{
	if ((freq == 0.0) || (FReference == 0.0))
		return;

	//  Cache Frequency
    freq *= FClockFactor;

	// initialize device serial port
	Port(0x00, 0x18);

	if (FOutputClock==oVco)
		{
        // Search for vcodiv and outdiv values that mitigate freq error
        CalculateDivisors(freq);

		// calc N/R ratio which gives actual_freq = (Reference * N)/(outdiv * R)
		// Restrict to 14-bit approximation
		float NR = static_cast<float>(freq*totaldiv/Reference());
		Debug.Log("N/R: ", NR);
		RationalApproximation Approx(NR, 14);
		int R = Approx.Denominator();
		int N = Approx.Numerator();

        // Maximize value of N to maximize likelyhood of DM mode operation
        const int full = (1 << 14)-1;
        const int scale = std::min(full/N, full/R);
        N *= scale;
        R *= scale;

		Debug.Log("Rational: ", static_cast<float>(N)/R);

		// calc P, B, A terms such that N = P*B+A
        // But, P must not exceed 300 MHz
        // and, B must be either 1 or within the range 3..8191
        double VcoOut = Reference() * N / R;
		Debug.Log("Ad9516::Vco freq = ", VcoOut);
        int P = std::min(NextLargestPowerOfTwo(static_cast<unsigned int>(VcoOut/300.e6)), 32u);
        int B = N / P;
        int A = N - (P*B);
        // A may be non-zero only in FD mode
        if (B < 3)
            A = 0;

		Debug.Log("Ad9516::N = ", N);
		Debug.Log("Ad9516::R = ", R);
		Debug.Log("Ad9516::P = ", P);
		Debug.Log("Ad9516::B = ", B);
		Debug.Log("Ad9516::A = ", A);

		// Cache actual frequency
		FFrequencyActual = (Reference() * (P * B + A))/(totaldiv * R);

		// correct if too low
		if(FFrequencyActual < (VcoLow/totaldiv))
			FFrequencyActual = VcoLow/totaldiv;
		// correct if too high
		if(FFrequencyActual > (VcoHigh/totaldiv))
			FFrequencyActual = VcoHigh/totaldiv;

      	// Load A register
		Port(0x13, A);

        // Load both portions of B register
		Port(0x14, (B & 0xff));
		Port(0x15, (B >> 8));

        // CP mode = normal
        unsigned char Cpm = 0x7c;
		Port(0x10, Cpm);

		Port(0x11, (R & 0xff));		// R LSBs
		Port(0x12, (R >> 8));		// R MSBs

        // B engaged, P set (DM mode)
		// encode P value for register use
		int P_bits;
		switch(P)
			{
			default:
			case  1:	P_bits = 0;		break;
			case  2:	P_bits = 1;		break;
			case  3:	P_bits = 7;		break;
			case  4:	P_bits = 3;		break;
			case  8:	P_bits = 4;		break;
			case 16:    P_bits = 5;		break;
			case 32:	P_bits = 6;		break;
			}
		Port(0x16, ((B == 1) ? 0x08 : 0x0) | P_bits);

        // default anti-backlash, status pin DC
		Port(0x17, 0x00);

        // Differential reference mode
		Port(0x1c, 0x01);

        // VCO as input to VCO divider
        Port(0x1E1, 0x02);

        // Set appropriate VCO divider ratio
        Port(0x1E0, std::max(vcodiv-2, 0));
		}
	else
		{
        // Bypass the VCO divider
        Port(0x1E1, 0x01);

        vcodiv = 1;
        outdiv = static_cast<int>(Reference()/freq);
        totaldiv = std::max(1, outdiv);

		// Cache actual frequency
        FFrequencyActual = Reference()/totaldiv;

        // Power-down PLL
		Port(0x10, 0x01);
		}

    Debug.Log("vcodiv: ", vcodiv);
    Debug.Log("outdiv: ", outdiv);
    Debug.Log("totaldiv: ", totaldiv);

    // Divider 0 duty cycle, encode duty cycle into register
    const int odd = outdiv & 1;
    const int lo = (outdiv >> 1) - 1;
    const int hi = lo + odd;
    int outdiv_bits = ( (lo << 4) | hi );
    Debug.Log("Ad9516 outdiv bits: ", outdiv_bits);

	int outdiv_bypass;
	if(outdiv == 1)
		outdiv_bypass = 0x80;
	else
		outdiv_bypass = 0;

    ConfigureOutputs(outdiv_bits, outdiv_bypass);

    // update registers
    Port(0x232, 0x01);

    // Cal VCO
    Port(0x18, 00);
    Port(0x232, 0x01);
    Port(0x18, 01);
    Port(0x232, 0x01);

    Sleep(1);

	FEnabled = true;
}

//-----------------------------------------------------------------------------
//  Ad9516::ConfigureOutputs()  --  Set clock output modes and signalling types
// This version is tailored for the GSPS module, but may be overriden
//-----------------------------------------------------------------------------

void Ad9516::ConfigureOutputs(int outdiv_bits, int outdiv_bypass)
{
    // LVPECL outputs
    Port(0xf0, 0x0c);       // OUT0 normal operation 960mV
	Port(0xf1, 0x0c);       // OUT1 normal operation 960mV
	Port(0xf2, 0x0c);       // OUT2 normal operation 960mV

	Port(0xf3, 0x03);       // OUT3 power-down
    Port(0xf4, 0x03);       // OUT4 power-down
    Port(0xf5, 0x03);       // OUT5 power-down
    //LVDS/LVCMOS outputs
    Port(0x140, 0x01);       // OUT6 power-down
    Port(0x141, 0x01);       // OUT7 power-down
    Port(0x142, 0x01);       // OUT8 power-down
    Port(0x143, 0x01);       // OUT9 power-down

    // Configure duty-cycle and bypass for channels 0, 1 & 2.  Others disabled
    for (int i = 0; i < 3; ++i)
        {
        int channel_offset = i*3;
        // Set duty cycle
        Port(0x190+channel_offset, outdiv_bits);
        // Enable/bypass divider
        if (i == 1)
            // Start high on ch 1
            Port(0x191+channel_offset, outdiv_bypass | 0x10);
        else
            Port(0x191+channel_offset, outdiv_bypass);

        // Enable DCCONF for duty cycle correction
        Port(0x192+channel_offset, 0);
        }
}

//-----------------------------------------------------------------------------
//  Ad9516::CalculateDivisors()  --  Derive best-fit divisors
//-----------------------------------------------------------------------------

void Ad9516::CalculateDivisors(double & freq)
{
    // VCO and output dividers which mitigate output frequency error
    double best_error = std::numeric_limits<double>::max();
    int best_outdiv = -1;
    int best_vcodiv = -1;

    for (vcodiv = 2; vcodiv <= 6; ++vcodiv)
        {
        for (outdiv = 1; outdiv <= 32; outdiv += (outdiv < 2) ? 1 : 2)
            {
            totaldiv = outdiv*vcodiv;

            double vco_freq = std::min(std::max(VcoLow, freq*totaldiv), VcoHigh);
            double Error = freq-(vco_freq/totaldiv);
            if (std::fabs(Error) < std::fabs(best_error))
                {
                best_error = Error;
                best_outdiv = outdiv;
                best_vcodiv = vcodiv;
                }
            }
        }

    vcodiv = best_vcodiv;
    outdiv = best_outdiv;
    totaldiv = vcodiv*outdiv;
    Debug.Log("best error: ", best_error);

    freq -= best_error;
}

//-----------------------------------------------------------------------------
//  Ad9516::Locked()  --  Return PLL lock status
//-----------------------------------------------------------------------------

bool  Ad9516::Locked() const
{
	return PllStatusReg.Locked.Value() ? true : false;
}

//-----------------------------------------------------------------------------
//  Ad9516::SetEnabled()  --  Activate / Deactivate output
//-----------------------------------------------------------------------------

void  Ad9516::SetEnabled(bool state)
{
	DoSetEnabled(state);
}

//-----------------------------------------------------------------------------
//  Ad9516::DoSetEnabled()  --  Activate / Deactivate output
//-----------------------------------------------------------------------------

void  Ad9516::DoSetEnabled(bool /*state*/)
{
}

//-----------------------------------------------------------------------------
//  Ad9516::OutputDivider()  --  Set best-fit output divider
//-----------------------------------------------------------------------------

void Ad9516::OutputDivider(int value)
{
	outdiv = value;
}

//-----------------------------------------------------------------------------
//  Ad9516::OutputDivider()  --  Return best-fit output divider
//-----------------------------------------------------------------------------

int Ad9516::OutputDivider() const
{
	return outdiv;
}

//-----------------------------------------------------------------------------
//  Ad9516::PFDPolarity()  --  Set Phase-Frequency Detector (PFD) Polarity
//                             0 = Negative (Default), 1 = Positive.
//-----------------------------------------------------------------------------

void Ad9516::PFDPolarity(bool state)
{
    FPFDPolarity = state;
}

//-----------------------------------------------------------------------------
//  Ad9516::PFDPolarity()  --  Return Phase-Frequency Detector (PFD) Polarity
//-----------------------------------------------------------------------------

bool Ad9516::PFDPolarity() const
{
    return FPFDPolarity;
}

//-----------------------------------------------------------------------------
//  Ad9516::Reset()  --  Reset state of all internal registers
//-----------------------------------------------------------------------------

void  Ad9516::Reset()
{
    PllControlReg.Reset.Set();
    Sleep(1);
    PllControlReg.Reset.Clear();
}

//-----------------------------------------------------------------------------
//  Ad9516::Reapply()  --  Reapply settings to hardware from saved state
//-----------------------------------------------------------------------------

void  Ad9516::Reapply()
{
    Reset();

    Frequency(FFrequency);
}

//-----------------------------------------------------------------------------
//  Ad9516::Port()  --  Write to PLL data port
//-----------------------------------------------------------------------------

void Ad9516::Port(unsigned short reg, unsigned char value)
{
    // Wait for previous operation to complete
    for (unsigned int i = 0; (i < 10) && !SpiStatusReg.Ready.Value(); ++i)
        Sleep(1);
    if (!SpiStatusReg.Ready.Value())
        Debug.Log("PLL SPI ready timeout error!");

    SpiControlReg.Address.Value(reg);
    SpiControlReg.Data.Value(value);
    SpiControlReg.Rw.Value(0);
    SpiControlReg.Apply();
/*
    std::stringstream msg;
    msg << "Ad9516::Port[0x" << std::hex << static_cast<int>(reg)
        << "]=0x" << static_cast<int>(reg);
    Debug.Log(msg.str());
*/
}

//-----------------------------------------------------------------------------
//  Ad9516::Port()  --  Read from PLL data port
//-----------------------------------------------------------------------------

unsigned char Ad9516::Port(unsigned short reg)
{
    unsigned char result;

    // Wait for previous operation to complete
    for (unsigned int i = 0; (i < 10) && !SpiStatusReg.Ready.Value(); ++i)
        Sleep(1);
    if (!SpiStatusReg.Ready.Value())
        Debug.Log("PLL SPI ready timeout error!");

    SpiControlReg.Address.Value(reg);
    SpiControlReg.Rw.Value(1);
    SpiControlReg.Apply();

    // Wait for previous operation to complete
    while (!SpiStatusReg.DataValid.Value())
        Sleep(1);

    result = SpiStatusReg.Data.Value();
/*
    std::stringstream msg;
    msg << std::hex << "0x" << result << "=Ad9516::Port[0x"
    << static_cast<int>(reg) << "]";
    Debug.Log(msg.str());
*/
    return result;
}

} // namespace Innovative
