//
// Ad9952_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2003 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <Ad9952_Mb.h>
#include <SystemSupport_Mb.h>
#include <algorithm>
#include <sstream>
#include <Exception_Mb.h>

//---------------------------------------------------------------------------

namespace Innovative
{

//=============================================================================
//  CLASS Ad9952  --  Ad9952 DDS Implementation
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class Ad9952
//-----------------------------------------------------------------------------

const double Ref_Freq_Mhz = 40.0;

//-----------------------------------------------------------------------------
//  constructor for class Ad9952
//-----------------------------------------------------------------------------

Ad9952::Ad9952(int index, AddressingSpace &space, int addr, int data, int ctl)
	: ClockBase(index), MaxIntAsFloat(4294967296.0), 
      SysClk(20.0e6),
	  Debug(false),
      AddrReg(space, addr), 
      DataReg(space, data),
	  CtlReg(space, ctl)
{
	Reference(Ref_Freq_Mhz*1.e6);    // 14.4 Mhz internal clock
}

//-----------------------------------------------------------------------------
//  constructor for class Ad9952
//-----------------------------------------------------------------------------

Ad9952::Ad9952(double freq, double ref, bool enabled, int index,
			   AddressingSpace &space, int addr, int data, int ctl)
	: ClockBase(freq, ref, enabled, index), 
      MaxIntAsFloat(4294967296.0),
      SysClk(20.0e6), 
	  Debug(false),
      AddrReg(space, addr),
	  DataReg(space, data), 
      CtlReg(space, ctl)
{
	Reference(ref*1.e6);

	DoSetFrequency(freq);
	DoSetEnabled(enabled);
}

//-----------------------------------------------------------------------------
//  Ad9952::GetFrequencyActual()  --  Calculate actual clock rate
//-----------------------------------------------------------------------------

double Ad9952::GetFrequencyActual() const
{
	return FDivisor*SysClk/MaxIntAsFloat;
}

//-----------------------------------------------------------------------------
//  Ad9952::Max()  --  Return maximum output frequency
//-----------------------------------------------------------------------------

double Ad9952::Max() const
{
	return 30.0e6;
}

//-----------------------------------------------------------------------------
//  Ad9952::Min()  --  Return minimum output frequency
//-----------------------------------------------------------------------------

double Ad9952::Min() const
{
	return 0.0;
}

//-----------------------------------------------------------------------------
//  Ad9952::SetFrequency()  --  Calculate settings for a frequency
//-----------------------------------------------------------------------------

void Ad9952::SetFrequency(double freq)
{
	DoSetFrequency(freq);
}

//-----------------------------------------------------------------------------
//  Ad9952::DoSetFrequency()  --  Output specified frequency
//-----------------------------------------------------------------------------

void Ad9952::DoSetFrequency(double freq)
{
	if (FReference == 0.0)
		return;

	FFrequency = std::max(std::min(freq, Max()), Min());

	// reset
	Disable();

	// phase and control
	Data(0, 0);

//	const int multiplier_max = 20;
	const int multiplier_min = 4;
	const double ref_max = 300.0e6;	// Backed down from 400, which fails to lock
	const double ref_min = 100.0e6; // Increased from 80, which may not work

	if ((Reference() < (ref_min/multiplier_min)) || (Reference() > (ref_max/multiplier_min)))
		throw MalibuException("Reference out of bounds");

	double target_reference = freq * 50.0f;
	int multiplier = static_cast<int>(target_reference/Reference());
	if ((Reference()*multiplier) < ref_min)
		multiplier = static_cast<int>(ref_min/Reference());
	if ((Reference()*multiplier) > ref_max)
		multiplier = static_cast<int>(ref_max/Reference());
    multiplier = std::min(std::max(multiplier, 4), 20);

	Debug.Log("Ad9952 multiplier: ", multiplier);
	SysClk = multiplier * Reference();
	Debug.Log("Ad9952 SysClk: ", static_cast<float>(SysClk));

	Data(1, multiplier << 3);
    int high_speed_sync = (freq > 250.0e6) ? 0x04 : 0;
	Data(2, 0 | high_speed_sync);
	Data(3, 0);

	FDivisor = static_cast<unsigned int>(FFrequency * MaxIntAsFloat / SysClk);
	Data(4, FDivisor);
	Data(5, 0);

	// update output frequency
	Control(1, 1);
	Control(1, 0);

	FEnabled = true;
}

//-----------------------------------------------------------------------------
//  Ad9952::Disable()  --  Device reset
//-----------------------------------------------------------------------------

void  Ad9952::Disable()
{
	// reset
	Control(0, 1);
	Control(0, 0);
}

//-----------------------------------------------------------------------------
//  Ad9952::SetEnabled()  --  Activate / Deactivate output
//-----------------------------------------------------------------------------

void  Ad9952::SetEnabled(bool state)
{
	DoSetEnabled(state);
}

//-----------------------------------------------------------------------------
//  Ad9952::DoSetEnabled()  --  Activate / Deactivate output
//-----------------------------------------------------------------------------

void  Ad9952::DoSetEnabled(bool /*state*/)
{
}

//-----------------------------------------------------------------------------
//  Ad9952::Reapply()  --  Reapply settings to hardware from saved state
//-----------------------------------------------------------------------------

void  Ad9952::Reapply()
{
	DoSetFrequency(FFrequency);
	DoSetEnabled(FEnabled);
}

//-----------------------------------------------------------------------------
//  Ad9952::Data()  --  Write to PLL data port
//-----------------------------------------------------------------------------

void Ad9952::Data(unsigned char reg, unsigned int value)
{
	std::stringstream msg;
	msg << "Ad9952::Write 0x" << std::hex << value << " to reg " << static_cast<int>(reg);
	Debug.Log(msg.str());

	Address(reg);

	DataReg.Value(value);
	Innovative::Sleep(100);
}

//-----------------------------------------------------------------------------
//  Ad9952::Data()  --  Read from PLL data port
//-----------------------------------------------------------------------------

unsigned int Ad9952::Data(unsigned char reg)
{
	Address(reg | 0x80);

	return DataReg.Value();
}

//-----------------------------------------------------------------------------
//  Ad9952::Control()  --  Write bit in PLL control port
//-----------------------------------------------------------------------------

void Ad9952::Control(unsigned int bit, bool value)
{
	CtlReg.Bit(bit, value);
	Sleep(10);
}

//-----------------------------------------------------------------------------
//  Ad9952::Address()  --  Set address of register to be written
//-----------------------------------------------------------------------------

void Ad9952::Address(unsigned char value)
{
	AddrReg.Value(value);
	Sleep(10);
}


} // namespace Innovative
