// AdDac9122_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2013 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <AdDac9122_Mb.h>
#include <SystemSupport_Mb.h>
#include <HardwareRegister_Mb.h>

//---------------------------------------------------------------------------
namespace Innovative
{
//---------------------------------------------------------------------------
namespace AdDac9122
{
//===========================================================================
//  CLASS DacDevice  -- Per-DAC controls for AD9122
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class DacDevice
//---------------------------------------------------------------------------

const int Max_P2_Reps = 5;
const int Max_P3_Reps = 5;

//---------------------------------------------------------------------------
//  constructor for class DacDevice
//---------------------------------------------------------------------------

DacDevice::DacDevice(IDacToHardware * hdw)
    :  Hdw(hdw), FEnterSEDMode(false)
{
    // default pattern
    FCompare_I[0]=0x5555;  FCompare_I[1]=0xaaaa;
    FCompare_Q[0]=0xaaaa;  FCompare_Q[1]=0x5555;
}

//---------------------------------------------------------------------------
//  destructor for class DacDevice
//---------------------------------------------------------------------------

DacDevice::~DacDevice()
{
    delete Hdw;   Hdw = 0;
}

//---------------------------------------------------------------------------
//  DacDevice::PreconfigHardware() --
//---------------------------------------------------------------------------

void  DacDevice::PreconfigHardware()
{
    ScopedTraceRef ccc(Debug, "DacDevice::PreconfigHardware()");

}

//---------------------------------------------------------------------------
//  DacDevice::ConfigureHardware() --
//---------------------------------------------------------------------------

void  DacDevice::ConfigureHardware()
{
    ScopedTraceRef STR(Debug, "AdDac9122::DacDevice::ConfigureHardware");

    // If DACs are off, power everything off.
    if (FChannelEnabled[0]==false && FChannelEnabled[1]==false)
        {
        SpiValue(0x01, 0xF0);   // power down all
        Debug.Log("DAC Device powered down...");
        return;
        }

    Setup_Phase1();

    unsigned int fifo_offset=0;
    for (int i=0; i<Max_P2_Reps; i++)
        {
        Setup_Phase2(fifo_offset);

        //  Setup_Phase3
        Debug.Log("Setup_Phase3");
        //  Get Sync Lost/Lock
        unsigned int sync_lock;
        for (int i=0; i<Max_P3_Reps; i++)
            {
            sync_lock = SpiField(0x6, 4, 2);
            if (sync_lock==0)
                continue;
            else
                break;
            }
        if (sync_lock!=1)
            {
            //  Reset, go back to step 2
            Setup_Phase9();
            continue;
            }

        //  Sync/Lock == 1
        //  Setup_Phase4
        Debug.Log("Setup_Phase4");
        //   enable FIFO warning interrupt
        SpiField(0x4, 0, 2, 0x3);

        //  Setup_Phase5 - thermometer status
        Debug.Log("Setup_Phase5");
        unsigned int thermometer = 1;
        bool stable = Setup_Phase5(thermometer);

        if (!IsThermometerValueLegal(thermometer))
            {
            Setup_Phase1();
            continue;
            }
        if (!stable)
            {
            Setup_Phase9();
            continue;
            }

        //  Setup_Phase6
        Debug.Log("Setup_Phase6");
        unsigned int flags = SpiField(0x6, 0, 2);
        if (flags)
            {
            Setup_Phase10();  // disable fifo warnings
            Setup_Phase9();
            continue;
            }

        Debug.Log("Setup_Phase7");
        // calculate fifo offset
        unsigned int thermo_value = ThermometerValueDecode(thermometer);
        Debug.Log("Decoded Thermo Value", thermo_value);
        if (thermo_value <= 4)
            fifo_offset = 4 - thermo_value;
        else
            fifo_offset = 12 - thermo_value;

        //  Repeat Phase 2 to apply the settings
        Setup_Phase2(fifo_offset);

        //  Setup_Phase8 - wake up DACs and run
        Debug.Log("Setup_Phase8");
        SleepDacs(false);

        //  break out of loop - we are set up
        break;
        }
    //  Setup_Phase11 - do test mode
    Setup_Phase11();

    Write_I_Offset(Offset_I);
    Write_Q_Offset(Offset_Q);
    Write_I_Gain(Gain_I);
    Write_Q_Gain(Gain_Q);

}

//---------------------------------------------------------------------------
//  DacDevice::Reset() -- Reset DAC device
//---------------------------------------------------------------------------

void  DacDevice::Reset()
{
    ScopedTraceRef STR(Debug, "AdDac9122::DacDevice::Reset");
    //  Dac Reset
    Hdw->DacReset(false);
    Hdw->DacReset(true);
    Hdw->DacReset(false);

    SpiBit(0x00, 5, false);  // soft reset
    SpiBit(0x00, 5, true);
    SpiBit(0x00, 5, false);

}

//---------------------------------------------------------------------------
//  DacDevice::SleepDacs() --
//---------------------------------------------------------------------------

void  DacDevice::SleepDacs(bool state)
{
    // Sleep DACs
    SpiBit(0x41, 7, state);
    SpiBit(0x45, 7, state);
}

//---------------------------------------------------------------------------
//  DacDevice::Setup_Phase1() --
//---------------------------------------------------------------------------

void  DacDevice::Setup_Phase1()
{
    ScopedTraceRef STR(Debug, "AdDac9122::DacDevice::Setup_Phase1");

    Reset();

    // Sleep DACs
    SleepDacs(true);

    // Set Channel Enable
    SpiBit(0x01, 7, FChannelEnabled[0]==false);   // power off if channel is off
    SpiBit(0x01, 6, FChannelEnabled[1]==false);   // power off if channel is off

    // Set Interpolation Mode
    SpiValue(0x1c, 0x1);   // bypass
    SpiValue(0x1d, 0x1);
    SpiValue(0x1e, 0x1);

    // Run all clocks

    // Set SYNC mode
    SpiBit(0x10, 6, false);

}

//---------------------------------------------------------------------------
//  DacDevice::Setup_Phase2() -- FIFO / SYNC setup
//---------------------------------------------------------------------------

void  DacDevice::Setup_Phase2(unsigned int fifo_offset)
{
    ScopedTraceRef STR(Debug, "AdDac9122::DacDevice::Setup_Phase2");
    Debug.Log("Fifo Offset", fifo_offset);

    //  Set Fifo Offset
    SpiValue(0x17, fifo_offset);

    //  Enable Sync Engine
    SpiValue(0x10, 0x80);  // 0x80 for falling edge sync, 0x88 for rising edge sync

    //  Enable sync lost interrupt)
    SpiBit(0x04, 5, true);

}

//---------------------------------------------------------------------------
//  DacDevice::Setup_Phase5() -- Thermometer Status
//---------------------------------------------------------------------------

bool  DacDevice::Setup_Phase5(unsigned int & thermo_status)
{
    ScopedTraceRef STR(Debug, "AdDac9122::DacDevice::Setup_Phase5");
    const unsigned int rep_count = 5;
    //  disable sync/interrupt/flag
    for (unsigned int reps=0; reps<rep_count; reps++)
        {
        unsigned int current_thermo = SpiValue(0x19);
        if (reps==0)
            {
            thermo_status=current_thermo;
            continue;
            }
        if (thermo_status!=current_thermo)
            {
            Debug.Log("Return False - not stable");
            return false;   // not stable
            }
        }
    //  we got all matches
    Debug.Log("Thermo Status", thermo_status);
    return true;
}

//---------------------------------------------------------------------------
//  DacDevice::Setup_Phase9() -- Clean up phase to repeat setup
//---------------------------------------------------------------------------

void  DacDevice::Setup_Phase9()
{
    ScopedTraceRef STR(Debug, "AdDac9122::DacDevice::Setup_Phase9");

    //  disable sync/interrupt/flag
    SpiBit(0x10, 7, false);    // disable sync
    SpiBit(0x04, 5, false);    // disable sync lost interrupt
    SpiBit(0x06, 5, true);     // clear sync lost flag
}

//---------------------------------------------------------------------------
//  DacDevice::Setup_Phase10() -- disable FIFO warnings and clear flags
//---------------------------------------------------------------------------

void  DacDevice::Setup_Phase10()
{
    ScopedTraceRef STR(Debug, "AdDac9122::DacDevice::Setup_Phase10");
    SpiField(0x04, 0, 2, 0x00);
    SpiField(0x06, 0, 2, 0x03);
}

//---------------------------------------------------------------------------
//  DacDevice::Setup_Phase11() -- SED (Sample Error Detection) mode
//---------------------------------------------------------------------------

void  DacDevice::Setup_Phase11()
{
    ScopedTraceRef STR(Debug, "AdDac9122::DacDevice::Setup_Phase11");

    //  Load comparison values
    Load_Compare_I(0);
    Load_Compare_Q(0);
    Load_Compare_I(1);
    Load_Compare_Q(1);

    if (FEnterSEDMode==false)
        {
        Debug.Log("Not in SED mode - test skipped");
        return;
        }

    //  Enable SED error detect flag
    SpiBit(0x05, 2, true);

    //  Enable SED pattern test mode)
    //  210000 106 fmcd!

    //  Enable SED
    SpiValue(0x67, 0x88);
    Sleep(1);

    SpiValue(0x67, 0x88);  // clear spurious error flags
    SpiValue(0x07, 0x00);

}

//---------------------------------------------------------------------------
//  DacDevice::GetSED_Status() -- Read and save SED status
//---------------------------------------------------------------------------

void  DacDevice::GetSED_Status()
{
    ScopedTraceRef STR(Debug, "AdDac9122::DacDevice::GetSED_Status");

    SedStatus.R7 = SpiValue(0x7);
    SedStatus.R67 = SpiValue(0x67);
    SedStatus.R70 = SpiValue(0x70);
    SedStatus.R71 = SpiValue(0x71);
    SedStatus.R72 = SpiValue(0x72);
    SedStatus.R73 = SpiValue(0x73);

    Debug.Log("R7",   SedStatus.R7, 16);
    Debug.Log("R67",  SedStatus.R67, 16);
    Debug.Log("R70",  SedStatus.R70, 16);
    Debug.Log("R71",  SedStatus.R71, 16);
    Debug.Log("R72",  SedStatus.R72, 16);
    Debug.Log("R73",  SedStatus.R73, 16);

}

//---------------------------------------------------------------------------
//  DacDevice::Write_I_Offset() -- Read and save SED status
//---------------------------------------------------------------------------

void  DacDevice::Write_I_Offset(unsigned int value)
{
    ScopedTraceRef STR(Debug, "AdDac9122::DacDevice::Write_I_Offset");
    Debug.Log("value",  value, 16);

    Offset_I = value;

    unsigned int R3C = value&0xFF;
    unsigned int R3D = (value>>8)&0xFF;

    SpiValue(0x3C, R3C);
    SpiValue(0x3D, R3D);

    Debug.Log("R3C",  R3C, 16);
    Debug.Log("R3D",  R3D, 16);

}

//---------------------------------------------------------------------------
//  DacDevice::Write_Q_Offset() -- Read and save SED status
//---------------------------------------------------------------------------

void  DacDevice::Write_Q_Offset(unsigned int value)
{
    ScopedTraceRef STR(Debug, "AdDac9122::DacDevice::Write_Q_Offset");
    Debug.Log("value",  value, 16);

    Offset_Q = value;

    unsigned int R3E = value&0xFF;
    unsigned int R3F = (value>>8)&0xFF;

    SpiValue(0x3E, R3E);
    SpiValue(0x3F, R3F);

    Debug.Log("R3E",  R3E, 16);
    Debug.Log("R3F",  R3F, 16);

}

//---------------------------------------------------------------------------
//  DacDevice::Write_I_Gain() -- Read and save SED status
//---------------------------------------------------------------------------

void  DacDevice::Write_I_Gain(unsigned int value)
{
    ScopedTraceRef STR(Debug, "AdDac9122::DacDevice::Write_I_Gain");
    Debug.Log("value",  value, 16);

    Gain_I = value;

    unsigned int R40 = value&0xFF;
    unsigned int R41 = (value>>8)&0xFF;

    SpiValue(0x40, R40);
    SpiField(0x41, 0, 2, R41);

    Debug.Log("R40",  R40, 16);
    Debug.Log("R41",  R41, 16);

}

//---------------------------------------------------------------------------
//  DacDevice::Write_Q_Gain() -- Read and save SED status
//---------------------------------------------------------------------------

void  DacDevice::Write_Q_Gain(unsigned int value)
{
    ScopedTraceRef STR(Debug, "AdDac9122::DacDevice::Write_Q_Gain");
    Debug.Log("value",  value, 16);

    Gain_Q = value;

    unsigned int R44 = value&0xFF;
    unsigned int R45 = (value>>8)&0xFF;

    SpiValue(0x44, R44);
    SpiField(0x45, 0, 2, R45);

    Debug.Log("R44",  R44, 16);
    Debug.Log("R45",  R45, 16);

}

//---------------------------------------------------------------------------
//  DacDevice::IsThermometerValueLegal() -- check thermo value
//---------------------------------------------------------------------------

bool  DacDevice::IsThermometerValueLegal(unsigned int thermo_value)
{
    switch (thermo_value)
        {
        default:  return false;   // not legal
        case 0x01:
        case 0x03:
        case 0x07:
        case 0x0F:
        case 0x1F:
        case 0x3F:
        case 0x7F:
        case 0xFF:  return true;   // legal
        }

}

//---------------------------------------------------------------------------
//  DacDevice::ThermometerValueDecode() -- convert thermo value
//---------------------------------------------------------------------------

unsigned int  DacDevice::ThermometerValueDecode(unsigned int thermo_value)
{
    switch (thermo_value)
        {
        default:
        case 0x01:  return 0;
        case 0x03:  return 1;
        case 0x07:  return 2;
        case 0x0F:  return 3;
        case 0x1F:  return 4;
        case 0x3F:  return 5;
        case 0x7F:  return 6;
        case 0xFF:  return 7;
        }

}

//---------------------------------------------------------------------------
//  DacDevice::Load_Compare_I() -- load compare value
//---------------------------------------------------------------------------

void  DacDevice::Load_Compare_I(unsigned int index)
{
    unsigned int msb_addr;
    unsigned int lsb_addr;
    unsigned int data_idx;

    switch (index)
        {
        default:
        case 0:  data_idx=0;  lsb_addr=0x68 ;  msb_addr=0x69;  break;
        case 1:  data_idx=1;  lsb_addr=0x6c ;  msb_addr=0x6d;  break;
        }

    SpiValue(lsb_addr, FCompare_I[data_idx]&0xFF);
    SpiValue(msb_addr, FCompare_I[data_idx]>>8);
}

//---------------------------------------------------------------------------
//  DacDevice::Load_Compare_Q() -- load compare value
//---------------------------------------------------------------------------

void  DacDevice::Load_Compare_Q(unsigned int index)
{
    unsigned int msb_addr;
    unsigned int lsb_addr;
    unsigned int data_idx;

    switch (index)
        {
        default:
        case 0:  data_idx=0;  lsb_addr=0x6a ;  msb_addr=0x6b;  break;
        case 1:  data_idx=1;  lsb_addr=0x6e ;  msb_addr=0x6f;  break;
        }

    SpiValue(lsb_addr, FCompare_Q[data_idx]&0xFF);
    SpiValue(msb_addr, FCompare_Q[data_idx]>>8);
}

//---------------------------------------------------------------------------
//  DacDevice::SpiBit() --  Read Modify Write SPI
//---------------------------------------------------------------------------

void   DacDevice::SpiBit(int addr, int bit, bool state)
{
    // Read - Modify - Write SPI
    MemoryBitManipulator Reg(Hdw->Spi(addr));

    Reg.Bit(bit, state);

    Hdw->Spi(addr, Reg.Value());

}

//---------------------------------------------------------------------------
//  DacDevice::SpiField() --  Read Modify Write SPI
//---------------------------------------------------------------------------

void   DacDevice::SpiField(int addr, int start_bit, int size, unsigned int data)
{
    // Read - Modify - Write SPI
    MemoryBitManipulator Reg(Hdw->Spi(addr));

    Reg.Field(start_bit, size, data);

    Hdw->Spi(addr, Reg.Value());

}

//---------------------------------------------------------------------------
//  DacDevice::SpiField() --  Read SPI Field
//---------------------------------------------------------------------------

unsigned int   DacDevice::SpiField(int addr, int start_bit, int size)
{
    // Read SPI
    MemoryBitManipulator Reg(Hdw->Spi(addr));
    Debug.Log("AdDac9122::DacDevice::SpiField ADDR = ", addr, 16);
    Debug.Log("AdDac9122::DacDevice::SpiField DATA = ", Reg.Value(), 16);

    return Reg.Field(start_bit, size);
}

//---------------------------------------------------------------------------
//  DacDevice::SpiValue() --  Write SPI
//---------------------------------------------------------------------------

void   DacDevice::SpiValue(int addr, int data)
{
    Hdw->Spi(addr, data);
}

//---------------------------------------------------------------------------
//  DacDevice::SpiValue() --  read SPI
//---------------------------------------------------------------------------

unsigned int   DacDevice::SpiValue(int addr)
{
    unsigned int value = Hdw->Spi(addr);
    Debug.Log("AdDac9122::DacDevice::SpiValue ADDR = ", addr, 16);
    Debug.Log("AdDac9122::DacDevice::SpiValue DATA = ", value, 16);

    return value;
    //return Hdw->Spi(addr);
}

/*
//===========================================================================
//  CLASS DacCalibrator  -- DAC calibration procedure for TiDac6682
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class DacCalibrator
//---------------------------------------------------------------------------

DacCalibrator::DacCalibrator(IDacCalibrationRegister * reg, IDacCalibratorToHdw * hdw)
    :  Reg(reg), Hdw(hdw)
{

}

//---------------------------------------------------------------------------
//  destructor for class DacCalibrator
//---------------------------------------------------------------------------

DacCalibrator::~DacCalibrator()
{
    delete Hdw;   Hdw = 0;
    delete Reg;   Reg = 0;
}

//---------------------------------------------------------------------------
//  DacCalibrator::Calibrate() -- Perform complete Cal for Dacs
//---------------------------------------------------------------------------

void  DacCalibrator::Calibrate()
{
    ScopedTraceRef ccc(Debug, "DacCalibrator::Calibrate()");

    if (ShouldCalibrate()==false)
        {
        Debug.Log("...No DACs enabled, no Latency Cal");
        return;
        }

    const unsigned int kMaxTestReps = 10;
    //
    //  Apply Gain/Offset Coefficients - need this so output is on for CAL
    Hdw->ApplyDacGainCoefficients();
    //
    //  Save old dac delays, change to delay of 0.
    unsigned int Dac0Delay_cache(0);
    unsigned int Dac1Delay_cache(0);

    ReadDacDelay(Dac0Delay_cache, Dac1Delay_cache);
    LoadDacDelay(0, 0);

    //
    //  Reset DAC/FPGA
    Hdw->DacReset();
    Hdw->PreconfigureHardware();

    unsigned int should_be( GetShouldBeValue());

    FCalibrationOk = false;
    for (unsigned int i=0; i<kMaxTestReps; i++)
        {
        Debug.Log("...Latency Calibration");

        DoInternalCalibration();

        unsigned int result = GetInternalCalibrationResults();
        Debug.Log("Cal Result ", (int)result);
        if (result==should_be && ShiftCountCheck())
            {
            Debug.Log("Latency Calibration Successful");
            FCalibrationOk = true;
            break;
            }
        Debug.Log("Latency Calibration Failed");
        }

    double output_rate = Hdw->ClockFrequencyActualMhz();
    if ((output_rate>=220.0 && output_rate<=225.0)  ||
        (output_rate>=630.0 && output_rate<=650.0)      )
        {
        Debug.Log("Latency Calibration Failed - false positive");
        }
    //
    //  Restore original dac delays and apply them
    LoadDacDelay(Dac0Delay_cache, Dac1Delay_cache);
}

//------------------------------------------------------------------------------
//  DacCalibrator::ShiftCountCheck() -- require that the shift counts be small
//------------------------------------------------------------------------------

bool  DacCalibrator::ShiftCountCheck()
{
    if ((Reg->GetShiftCount(0) < 0xB) && (Reg->GetShiftCount(1) < 0xB))
        return true;
    else
        return false;
}

//------------------------------------------------------------------------------
//  DacCalibrator::DoInternalCalibration() --
//------------------------------------------------------------------------------

void DacCalibrator::DoInternalCalibration()
{
    // trigger in software, unframed mode
    Hdw->SetSoftwareUnframedTrigger();

    SetDacTestMode();

    Hdw->SetTestEnable(true);
    SetDacRunControl(true);
    Reg->CalEnable(true);

    Sleep(5);    // wait till done

    SetDacRunControl(false);
    Reg->CalEnable(false);
    Hdw->SetTestEnable(false);

}

//------------------------------------------------------------------------------
//  DacCalibrator::GetInternalCalibrationResults() --
//------------------------------------------------------------------------------

unsigned int  DacCalibrator::GetInternalCalibrationResults()
{
    return Reg->GetCalResults();
}

//------------------------------------------------------------------------------
//  DacCalibrator::GetShouldBeValue() --
//------------------------------------------------------------------------------

unsigned int  DacCalibrator::GetShouldBeValue()
{
    unsigned int should_be(0x3);
    return should_be;
}

//------------------------------------------------------------------------------
//  DacCalibrator::ShouldCalibrate() --
//------------------------------------------------------------------------------

bool  DacCalibrator::ShouldCalibrate()
{
    return Hdw->Dac(0).Enabled();
}

//---------------------------------------------------------------------------
//  DacCalibrator::SetDacTestMode()
//---------------------------------------------------------------------------

void  DacCalibrator::SetDacTestMode()
{
    // Single Device Version
    if (Hdw->Dac(0).Enabled())
        {
        if (Hdw->Dac(0).SingleActiveChannel())
            Hdw->SetTestMode_Alternating();
        else
            Hdw->SetTestMode_DoubleAlternating();
        }
}

//---------------------------------------------------------------------------
//  DacCalibrator::SetDacRunControl()
//---------------------------------------------------------------------------

void  DacCalibrator::SetDacRunControl(bool state)
{
    IDacCalibratorToHdw::EngDeviceState which = IDacCalibratorToHdw::engAll;
    if (Hdw->Dac(0).Enabled()==false)
        which = IDacCalibratorToHdw::engDevice1;

    Hdw->DeviceRun(state, which);    // start/stop Cal
}

//---------------------------------------------------------------------------
//  DacCalibrator::ReadDacDelay()
//---------------------------------------------------------------------------

void  DacCalibrator::ReadDacDelay(unsigned int & delay_0, unsigned int & delay_1)
{
    //
    //  Read dac delays
    delay_0 = delay_1 = Hdw->Dac(0).DacDelay();
}

//---------------------------------------------------------------------------
//  DacCalibrator::LoadDacDelay()
//---------------------------------------------------------------------------

void  DacCalibrator::LoadDacDelay(unsigned int delay, unsigned int)
{
    //
    //  Restore original dac delays and apply them
    Hdw->Dac(0).LoadDacDelay(delay);
}

//===========================================================================
//  CLASS DualDacCalibrator  -- DAC calibration procedure for TiDac6682
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class DualDacCalibrator
//---------------------------------------------------------------------------

DualDacCalibrator::DualDacCalibrator(IDacCalibrationRegister * reg, IDacCalibratorToHdw * hdw)
    :  DacCalibrator(reg, hdw)
{
}

//---------------------------------------------------------------------------
//  destructor for class DualDacCalibrator
//---------------------------------------------------------------------------

DualDacCalibrator::~DualDacCalibrator()
{
}

//------------------------------------------------------------------------------
//  DualDacCalibrator::GetShouldBeValue() --
//------------------------------------------------------------------------------

unsigned int  DualDacCalibrator::GetShouldBeValue()
{
    unsigned int should_be(0xF);
    if (Hdw->Dac(0).Enabled()==false)
        should_be = 0xC;
    else if (Hdw->Dac(1).Enabled()==false)
        should_be = 0x3;

    return should_be;
}

//------------------------------------------------------------------------------
//  DualDacCalibrator::ShouldCalibrate() --
//------------------------------------------------------------------------------

bool  DualDacCalibrator::ShouldCalibrate()
{
    return Hdw->Dac(0).Enabled() || Hdw->Dac(1).Enabled() ;
}

//---------------------------------------------------------------------------
//  DualDacCalibrator::SetDacTestMode()
//---------------------------------------------------------------------------

void  DualDacCalibrator::SetDacTestMode()
{
    if (Hdw->Dac(0).Enabled())
        {
        if (Hdw->Dac(0).SingleActiveChannel())
            Hdw->SetTestMode_Alternating();
        else
            Hdw->SetTestMode_DoubleAlternating();
        }
    else  // Use Dac 1
        {
        if (Hdw->Dac(1).SingleActiveChannel())
            Hdw->SetTestMode_Alternating();
        else
            Hdw->SetTestMode_DoubleAlternating();
        }
}

//---------------------------------------------------------------------------
//  DualDacCalibrator::SetDacRunControl()
//---------------------------------------------------------------------------

void  DualDacCalibrator::SetDacRunControl(bool state)
{
    IDacCalibratorToHdw::EngDeviceState which = IDacCalibratorToHdw::engAll;
    if (Hdw->Dac(0).Enabled()==false)
        which = IDacCalibratorToHdw::engDevice1;
    if (Hdw->Dac(1).Enabled()==false)
        which = IDacCalibratorToHdw::engDevice0;

    Hdw->DeviceRun(state, which);    // start/stop Cal

}

//---------------------------------------------------------------------------
//  DualDacCalibrator::ReadDacDelay()
//---------------------------------------------------------------------------

void  DualDacCalibrator::ReadDacDelay(unsigned int & delay_0, unsigned int & delay_1)
{
    //
    //  Read dac delays
    delay_0 = Hdw->Dac(0).DacDelay();
    delay_1 = Hdw->Dac(1).DacDelay();
}

//---------------------------------------------------------------------------
//  DualDacCalibrator::LoadDacDelay()
//---------------------------------------------------------------------------

void  DualDacCalibrator::LoadDacDelay(unsigned int delay_0, unsigned int delay_1)
{
    //
    //  Restore original dac delays and apply them
    Hdw->Dac(0).LoadDacDelay(delay_0);
    Hdw->Dac(1).LoadDacDelay(delay_1);
}
*/
} // namespace TiDac6682

} // namespace Innovative

