// AdcStats_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <AdcStats_Mb.h>
#include <cmath>
#include <Stats_Mb.h>
#include <algorithm>
#include <Magic_Mb.h>
#include <Exception_Mb.h>
#include <MovingAverage_Mb.h>
#include <math.h>

using namespace Innovative;
using namespace std;

namespace   Innovative
{

static double pi = 4.0 * atan(1.0);

//---------------------------------------------------------------------------
//  Constructor
//---------------------------------------------------------------------------

AdcStats::AdcStats()
	: FBits(16), FRmsNoise(0), FRmsSignal(0), FHarmonic(0), FEnob(0),
      FHd(0), FThd(0), FdBThd(0),
      FSinad(0), FSfdr(0), FSnr(0), 
	  FExtent(4), FPowerSpec(0), FdB(0), FReach(0),
	  FHunt(5), FHarmonicWidth(5), FKeepDc(false), FKeepFs2(false),
      FOrder(12), FWindow(wtBlackman), FHarmonics(6),
      FScaleFactor(1.0), FOffset(0.0), FIntermodulation(false),
      FSpectrumMode(smAmplitude), ZerodB(0), MindB(0)
{
    SetOrder(FOrder);
}

//---------------------------------------------------------------------------
//  Destructor
//---------------------------------------------------------------------------

AdcStats::~AdcStats()
{
}

//---------------------------------------------------------------------------
// AdcStats::SetOrder() - Set analysis buffer size
//---------------------------------------------------------------------------

void AdcStats::SetOrder(int order)
{
    FOrder = order;
    unsigned int One = 1;
	FSize = One<<FOrder;
}

//---------------------------------------------------------------------------
// AdcStats::SetSize() - Set analysis buffer size
//---------------------------------------------------------------------------

void AdcStats::SetSize(size_t samples)
{
	samples = IsPowerOfTwo(samples) ? samples : NextLargestPowerOfTwo(samples);
	SetOrder(TrailingZeroCount(static_cast<int>(samples)));
	FReach = (FSize/2);

    Reset();
}

//---------------------------------------------------------------------------
// AdcStats::SetReach() - Force size of result buffer
//---------------------------------------------------------------------------

void AdcStats::SetReach(size_t samples)
{
    FReach = samples;

    Reset();
}

//---------------------------------------------------------------------------
// AdcStats::Reset() - Reset ADC statistics
//---------------------------------------------------------------------------

void AdcStats::Reset()
{
	Initialize();

	for (unsigned int i = 0; i < FExtent.size(); ++i)
		FExtent[i] = 0;

	FStride = 0;
	FPoints = 0;
	PSStats.Init();
	DataStats.Init();
	Noise.Init();
	Distortion.Init();
    Signal.Init();
}

//---------------------------------------------------------------------------
// AdcStats::Initialize() -
//---------------------------------------------------------------------------

void AdcStats::Initialize()
{
	// Number of Points
	FMargin = std::max<unsigned int>(FReach/1000, 1);

	FHarmonic.resize(FHarmonics);

	// Init FFT buffers
	FftData.Order(FOrder);
	FftData.Window(FWindow);
	FftData.SpectrumEnabled(true);
	FftData.SpectrumMode(FSpectrumMode==smAmplitude ? Fourier::smAmplitude : Fourier::smPower);
	FftData.IgnoreDC(!FKeepDc);
	FftData.IgnoreFs2(!FKeepFs2);

	// Calculate reference spectrum
	ReferenceSpectrum();

	FPowerSpec.resize(FSize);
	FdB.resize(FSize);
	FAmplitude.resize(FSize);
	FPhase.resize(FSize);
}

//---------------------------------------------------------------------------
// AdcStats::SetBits() - Set number of bits/sample
//---------------------------------------------------------------------------

void AdcStats::SetBits(int bits)
{
	FBits = bits;
}

//---------------------------------------------------------------------------
// AdcStats::SetHarmonics() - Set number of harmonics to analyze
//---------------------------------------------------------------------------

void AdcStats::SetHarmonics(unsigned int value)
{
	FHarmonics = value;
	Initialize();
}

//---------------------------------------------------------------------------
// AdcStats::SetWindow() - Set analysis buffer size
//---------------------------------------------------------------------------

void AdcStats::SetWindow(Innovative::WindowTypes window)
{
	FWindow = window;
	Initialize();
}

//---------------------------------------------------------------------------
// AdcStats::GetExtent() - Return extent associated within index
//---------------------------------------------------------------------------

unsigned int AdcStats::GetExtent(unsigned int index)
{
	if (index < FExtent.size())
		return FExtent[index];

	return 0;
}

//---------------------------------------------------------------------------
// AdcStats::SetExtent() - Set extent associated within index
//---------------------------------------------------------------------------

void AdcStats::SetExtent(unsigned int index, unsigned int value)
{
	if (index < FExtent.size())
		FExtent[index] = value;
}

//---------------------------------------------------------------------------
// AdcStats::GetMaxSnr() - Return theoretical SNR
//---------------------------------------------------------------------------

double AdcStats::GetMaxSnr() const
{
    return DataStats.SNR();
}

//---------------------------------------------------------------------------
// AdcStats::GetHarmonic() - Return extent associated within index
//---------------------------------------------------------------------------

unsigned int AdcStats::GetHarmonic(unsigned int index)
{
	if (index < FHarmonics)
		return FHarmonic[index];

	return 0;
}

//---------------------------------------------------------------------------
// AdcStats::GetPS() - Return power spectrum element
//---------------------------------------------------------------------------

double AdcStats::GetPowerSpec(unsigned int index)
{
	if (index < FReach)
		return FPowerSpec[index];

	return 0.0;
}

//---------------------------------------------------------------------------
// AdcStats::SetdB() - Set dB array element
//---------------------------------------------------------------------------

void AdcStats::SetdB(unsigned int index, double value)
{
	if (index < FReach)
		FdB[index] = value;
}

//---------------------------------------------------------------------------
// AdcStats::GetdB() - Return dB array element
//---------------------------------------------------------------------------

double AdcStats::GetdB(unsigned int index)
{
	if (index < FReach)
		return FdB[index];

    return 0.0;
}

//---------------------------------------------------------------------------
// AdcStats::Datasample() - Add a data sample
//---------------------------------------------------------------------------

void AdcStats::Datasample(double value)
{
	Sample v(value);
	Datasample(v);
}

//---------------------------------------------------------------------------
// AdcStats::Datasample() - Add a data sample
//---------------------------------------------------------------------------

void AdcStats::Datasample(Sample value)
{
	if (FPoints >= Size())
		return;

	Complex c(value.Real.Float(), value.Imag.Float());
	if (FRescaled)
		{
		c += static_cast<float>(FOffset);
		c *= static_cast<float>(FScaleFactor);
		}

	DataStats.Sample(std::real(c));
	FftData.Time()[FPoints++] = c;
    FReach = FPoints >> 1;
}

//---------------------------------------------------------------------------
// AdcStats::CalculateExtents() - Calculate extents for PS mode
//---------------------------------------------------------------------------

void AdcStats::CalculateExtents(unsigned int peak, unsigned int *extent)
{
	// Use manually-selected extents, if available
	if (extent[0] || extent[1])
		return;

	// Walk down the PS mode pulse to the floor, right-left
	const double Threshold = .01;    // in dB
	size_t sz = std::max<size_t>(std::min<size_t>(FPowerSpec.size()/1000, 10), 1);
	MovingAverage Avg(sz);
	for (unsigned int i = peak; i > Avg.Length(); --i)
		{
		double delta = 20*log10(FPowerSpec[i]/FPowerSpec[i-1]);
		if ((Avg.Process(delta) < Threshold) && Avg.Settled())
			{
			extent[0] = i+Avg.Length()/2;
			break;
			}
		}

	// Walk down the PS mode pulse to the floor, left-right
	Avg.Reset();
	for (unsigned int i=peak; i < FReach-Avg.Length(); ++i)
		{
		double delta = 0.0f;
		if (FPowerSpec[i] && FPowerSpec[i+1])
			delta = 20*log10(FPowerSpec[i]/FPowerSpec[i+1]);
		if ((Avg.Process(delta) < Threshold) && Avg.Settled())
			{
			extent[1] = i-Avg.Length()/2;
			break;
            }
        }
}

//---------------------------------------------------------------------------
// AdcStats::CalculatePeakIndex2() -
//---------------------------------------------------------------------------

unsigned int AdcStats::CalculatePeakIndex2(unsigned int peak)
{
    unsigned int result = 0;

	// Start search 20dB below peak
    double threshold = PSStats.Maximum()/10;

    // Look left
    unsigned int i = peak;
    while (i && (FPowerSpec[i] > threshold))
        --i;

    double Min = std::numeric_limits<double>::min();
    for (unsigned int j = 0; j < i; ++j)
        {
        if (FPowerSpec[j] > Min)
            {
            Min = FPowerSpec[j];
            result = j;
            }
        }

    // Look right
    i = peak;
    while ((i<FReach-10) && (FPowerSpec[i] > threshold))
        ++i;

    for (unsigned int j = i; j < FReach-10; ++j)
        {
        if (FPowerSpec[j] > Min)
            {
            Min = FPowerSpec[j];
            result = j;
            }
        }

    return result;
}

//---------------------------------------------------------------------------
// AdcStats::analyze() - Analyze samples added so far
//---------------------------------------------------------------------------

void AdcStats::analyze()
{
    // Bits/sample
    PSStats.Bits(FBits);
	Noise.Bits(FBits);
	Distortion.Bits(FBits);
	Signal.Bits(FBits);
	DataStats.Range(pow(2.0, static_cast<double>(FBits))*FScaleFactor);

    FSinad = 0;
    FEnob = 0;
    FSfdr = 0;
	FHd = 0;
	FThd = 0;
	FdBThd = 0;
	FSnr = 0;
	FPsd = 0.0;
	FNoiseFloor = 0.0;
	FRmsNoisePlusDistortion = 0;
	FRmsNoise = 0;
	FRmsSignal = 0;
    for (unsigned int i = 0; i < FHarmonics; ++i)
        FHarmonic[i] = 0;

	// Adjust peak hunt size according to FFT resolution
	if (FReach)
		FHunt = static_cast<int>(Innovative::LogN(2, static_cast<double>(FReach)));
	else
		FHunt = 1;

    if (FIntermodulation)
        analyzeIM();
    else
        analyzeTHD();

    // Sin + Distortion
    if (FRmsNoisePlusDistortion && FRmsSignal)
        FSinad = (20.0 * log10(FRmsSignal/FRmsNoisePlusDistortion));

    // Effective number of bits
    FEnob = (FSinad-1.76)/6.02;

    // Spurious free dynamic range
    if (FRmsSignal)
        {
		double peak_amplitude = FPowerSpec[PSStats.MaxIndex()];
        double spur_amplitude = FSpurPeakAmplitude;
        FSfdr = 20.0*log10(peak_amplitude/spur_amplitude);

		FThd = (FHd/FRmsSignal) * 100;
        if (FThd)
            FdBThd = -20.0 * log10(FThd / 100);
        }

	// Signal to noise
	if (FRmsNoise && FRmsSignal)
		FSnr = 20.0 * log10(FRmsSignal/FRmsNoise);

	FPsd = FRmsNoise/sqrt(static_cast<double>(FReach)*NoisePowerBandwidth());

	double sf = std::pow(2.0, static_cast<double>(FBits-1)) * FScaleFactor;
	double psd = FPsd/(sf);
	FNoiseFloor = (psd > 0.0) ? 20.0 * log10(psd) : 0;

    ProcessCompletionEvent Evt(0);
    OnAnalyzed.Execute(Evt);
}

//---------------------------------------------------------------------------
// AdcStats::analyzeTHD() - Analyze for total harmonic distortion
//---------------------------------------------------------------------------

void AdcStats::analyzeTHD()
{
    // Determine peak of power spectrum
    FPeakIndex[0] = PSStats.MaxIndex();

	CalculateExtents(FPeakIndex[0], &FExtent[0]);
    //int l = FExtent[0];
    //int r = FExtent[1];

    // Cache for the hunt region
    std::vector<double> HuntBuffer(FHunt*2+1);
    std::vector<int> Add(FHarmonics+1);

    for (unsigned int i = 0; i < FHarmonics; ++i)
        {
		// Index of current harmonic
        unsigned int Fa;
		if (FStride)
			Fa = (i+1)*FStride + FPeakIndex[0];
		else
			Fa = (i+2)*FPeakIndex[0];

		// Wrap if beyond Nyquist
		unsigned int harmonic = (Fa > FReach) ? Fa % FReach : Fa;

		harmonic = std::max<unsigned int>(harmonic, FHunt);
		harmonic = std::min(harmonic, FReach-FHunt-1);

        // Cache the hunt region, locate local maxima
        double max_value = std::numeric_limits<double>::min();
        int HuntBufferIndex = 0;
        for (int k = -FHunt; k < FHunt+1; ++k)
            {
			int PowerSpecIndex = k+harmonic;
			if (IsExcluded(PowerSpecIndex))
                continue;

            HuntBuffer[HuntBufferIndex] = FPowerSpec[PowerSpecIndex];
            if (HuntBuffer[HuntBufferIndex] > max_value)
                {
                max_value = HuntBuffer[HuntBufferIndex];
				Add[i] = PowerSpecIndex;
				}

            ++HuntBufferIndex;
            }
		}

    int idx = 0;
	for (unsigned int i = 0; i < FHarmonics; ++i)
		if (!Within(Add[i], FExtent[2], FExtent[3]))
			FHarmonic[idx++] = Add[i];

	// Analyze from DC to Left extent and from Right extent to Nyquist
	//
	// DC is lower-most .1% of spectrum, Fs/2 is upper-most .1% of spectrum
	unsigned int left = FKeepDc ? 0 : FMargin;
	unsigned int right = FKeepFs2 ? FReach : FReach - FMargin;
	FSpurPeakAmplitude = 0.0;

	for (unsigned int i = left; i < right; ++i)
		{
		if (IsExcluded(i))
			continue;

		if (Within(i, FExtent[0], FExtent[1]))
			{
			Signal.Sample(FPowerSpec[i]);
			continue;
			}

		RetainLargestSpur(i, FPowerSpec[i]);

		if (IsHarmonic(i))
			{
			Distortion.Sample(FPowerSpec[i]);
			continue;
			}

		Noise.Sample(FPowerSpec[i]);
		}

	FRmsSignal = sqrt(Signal.Integral2());
	FRmsNoise = sqrt(Noise.Integral2());
	FHd = sqrt(Distortion.Integral2());
	FRmsNoisePlusDistortion = FRmsNoise + FHd;
}

//---------------------------------------------------------------------------
// AdcStats::analyzeIM() - Analyze for inter-modulation distortion
//---------------------------------------------------------------------------

void AdcStats::analyzeIM()
{
	// Determine peak of power spectrum
	FPeakIndex[0] = PSStats.MaxIndex();
	FPeakIndex[1] = FIntermodulation ? CalculatePeakIndex2(FPeakIndex[0]) : FPeakIndex[0];

	// Insure that first extent group is lower in frequency
	if (FExtent[2] && (FExtent[2] < FExtent[0]))
		{
		std::swap(FExtent[0], FExtent[2]);
		std::swap(FExtent[1], FExtent[3]);
        }

    CalculateExtents(FPeakIndex[0], &FExtent[0]);
    CalculateExtents(FPeakIndex[1], &FExtent[2]);

    // Cache for the hunt region
    std::vector<double> HuntBuffer(FHunt*2+1);
    std::vector<int> Add;

    const int cases = 7;
    for (unsigned int j = 0; j < cases*4; ++j)
        {
        int idx = j/cases;
        int i = idx + 1;

        // Index of current harmonic
        int harmonic = 0;
        switch (j%cases)
			{
            case 0:
                harmonic = i*FPeakIndex[1] - i*FPeakIndex[0];
				break;
			case 1:
                harmonic = i*FPeakIndex[1] + i*FPeakIndex[0];
                break;
            case 2:
                harmonic = i*FPeakIndex[1] + (i+1)*FPeakIndex[0];
                break;
            case 3:
				harmonic = i*FPeakIndex[1] - (i+1)*FPeakIndex[0];
                break;
			case 4:
                harmonic = (i+1)*FPeakIndex[1] + i*FPeakIndex[0];
                break;
            case 5:
                harmonic = (i+1)*FPeakIndex[1] - i*FPeakIndex[0];
                break;
            case 6:
				harmonic = (i+1)*FPeakIndex[0] - i*FPeakIndex[1];
                break;
            }

        if (harmonic < FHunt)
            continue;

        if (Within(harmonic, FExtent[0], FExtent[1])
            || Within(harmonic, FExtent[1], FExtent[2]))
            continue;

        // Exit if no more room to search for harmonics
        if (static_cast<unsigned int>(harmonic) >= (FReach-10))
            break;

        // Cache the hunt region, locate local maxima
        double max_value = std::numeric_limits<double>::min();
        int index = 0;
        int HuntBufferIndex = 0;
		for (int k = -FHunt; k < FHunt+1; ++k)
            {
            int PowerSpecIndex = k+harmonic;
			if (IsExcluded(PowerSpecIndex))
				continue;

            HuntBuffer[HuntBufferIndex] = FPowerSpec[PowerSpecIndex];
            if (HuntBuffer[HuntBufferIndex] > max_value)
                {
                max_value = HuntBuffer[HuntBufferIndex];
                index = PowerSpecIndex;
				}
            ++HuntBufferIndex;
			}

        std::vector<int>::iterator iter=std::find(Add.begin(), Add.end(), index);
        if (iter == Add.end())
            if (index > 5)
                Add.push_back(index);
        }

    unsigned int limit = std::min(FHarmonics, static_cast<unsigned int>(Add.size()));
    for (unsigned int i = 0; i < limit; ++i)
        FHarmonic[i] = Add[i];

	// Analyze from DC to Left extent and from Right extent to Nyquist
    //
    // Return the RMS noise level of the power spectrum
    // RMS noise, less DC component
	// DC is lower-most .1% of spectrum
	unsigned int left = FKeepDc ? 0 : FMargin;
    unsigned int right = FKeepFs2 ? FReach : FReach-FMargin;
	FSpurPeakAmplitude = 0.0;

	for (unsigned int i = left; i < right; ++i)
		{
		double power = IsExcluded(i) ? 0.0f : FPowerSpec[i];

		if (Within(i, FExtent[0], FExtent[1]) || Within(i, FExtent[2], FExtent[3]))
			{
			Signal.Sample(power);
			continue;
			}

		RetainLargestSpur(i, FPowerSpec[i]);

		if (IsHarmonic(i))
			{
			Distortion.Sample(FPowerSpec[i]);
			continue;
			}

		Noise.Sample(FPowerSpec[i]);
		}

	FRmsSignal = sqrt(Signal.Integral2());
	FRmsNoise = sqrt(Noise.Integral2());
	FHd = sqrt(Distortion.Integral2());
	FRmsNoisePlusDistortion = FRmsNoise + FHd;
}

//---------------------------------------------------------------------------
// AdcStats::RetainLargestSpur() -
//---------------------------------------------------------------------------

void AdcStats::RetainLargestSpur(unsigned int idx, double value)
{
	if (FSpurPeakAmplitude >= value)
        return;

    FSpurPeakAmplitude = value;
    FSpurPeakIndex = idx;
}

//---------------------------------------------------------------------------
// AdcStats::IsHarmonic() -
//---------------------------------------------------------------------------

bool AdcStats::IsHarmonic(int idx)
{
    for (unsigned int i = 0; i < FHarmonics; ++i)
        {
        if (Within(idx, FHarmonic[i]-FHarmonicWidth, FHarmonic[i]+FHarmonicWidth))
            return true;
        }

    return false;
}

//---------------------------------------------------------------------------
// AdcStats::IsExcluded() -
//---------------------------------------------------------------------------

bool AdcStats::IsExcluded(int idx)
{
    for (unsigned int i = 0; i < FExcluded.size(); ++i)
        {
        if (Within(idx, FExcluded[i].Low, FExcluded[i].High))
            return true;
        }

    return false;
}

//---------------------------------------------------------------------------
// AdcStats::Analyze() - Analyze samples added so far
//---------------------------------------------------------------------------

void AdcStats::Analyze()
{
    try
        {
        // Figure power spectrum of input data
        PowerSpectrum();
        analyze();
        }
    catch(...)
        {
        }
}

//---------------------------------------------------------------------------
// AdcStats::Analyze() - Analyze a float buffer
//---------------------------------------------------------------------------

void AdcStats::Analyze(float const *buffer, size_t size)
{
	Reset();

	size_t limit = std::min(size, Size());

	for (size_t i = 0; i < limit; i++)
		Datasample(buffer[i]);

	for (size_t i = size; i < Size(); i++)
		Datasample(0.0f);

	Analyze();
}

//---------------------------------------------------------------------------
// AdcStats::Analyze() - Analyze a float buffer
//---------------------------------------------------------------------------

void AdcStats::Analyze(double const * buffer, size_t size)
{
	Reset();

	size_t limit = std::min(size, Size());

	for (size_t i = 0; i < limit; i++)
		Datasample(buffer[i]);

	for (size_t i = size; i < Size(); i++)
		Datasample(0.0f);

	Analyze();
}

//---------------------------------------------------------------------------
// AdcStats::Analyze() - Analyze a complex buffer
//---------------------------------------------------------------------------

void AdcStats::Analyze(const ComplexDG & buffer)
{
	Reset();

	for (size_t i = 0; i < buffer.size(); i++)
		{
		Sample s(buffer[i].real(), buffer[i].imag());
		Datasample(s);
		}

	Analyze();
}

//---------------------------------------------------------------------------
// AdcStats::Analyze() - Analyze a float datagram
//---------------------------------------------------------------------------

void AdcStats::Analyze(const Innovative::IppFloatDG & data)
{
	Analyze(data.Ptr(), data.size());
}

//---------------------------------------------------------------------------
// AdcStats::Analyze() - Analyze a float datagram
//---------------------------------------------------------------------------

void AdcStats::Analyze(const Innovative::FloatDG & data)
{
	Analyze(data.Ptr(), data.size());
}

//---------------------------------------------------------------------------
// AdcStats::Analyze() - Analyze a double datagram
//---------------------------------------------------------------------------

void AdcStats::Analyze(const Innovative::DoubleDG & data)
{
    Analyze(data.Ptr(), data.size());
}

//---------------------------------------------------------------------------
// AdcStats::ReferenceSpectrum() - Calculate magnitude of reference sine wave
//---------------------------------------------------------------------------

void AdcStats::ReferenceSpectrum()
{
    ZerodB = pow(2.0f, FBits-1) * FScaleFactor;// * WindowGain();
    MindB = FScaleFactor;
}

//---------------------------------------------------------------------------
// AdcStats::PowerSpectrun() - Scale FFT of input waveform
//---------------------------------------------------------------------------

void AdcStats::PowerSpectrum()
{
    if (ZerodB == 0.0f)
        return;

    // FFT of input signal
    FftData.Transform();

    Calculate_dBV();
}

//---------------------------------------------------------------------------
// AdcStats::Calculate_dBV() - Calculate spectrum scaled in dBV
//---------------------------------------------------------------------------

void AdcStats::Calculate_dBV()
{
    for (size_t i=0; i<FPowerSpec.size(); ++i)
        {
        FPowerSpec[i] = FftData.Spectrum()[i];
        FPhase[i] = FftData.PhaseAngle()[i];

        if (FPowerSpec[i] > 0)
            FAmplitude[i] = 20 * log10(FPowerSpec[i]/ZerodB);
        else
            FAmplitude[i] = 20 * log10(MindB/ZerodB);
        }

    Pivot(FAmplitude, 3);
    Pivot(FAmplitude, 4);

    DoubleArray::iterator p = std::max_element(FAmplitude.begin(), FAmplitude.end());
    FAmplitudePeakIndex = static_cast<unsigned int>(p-FAmplitude.begin());

    Merge(FPhase);
    Merge(FPowerSpec);

    for (size_t i=0; i<FPowerSpec.size(); ++i)
        {
        if (FPowerSpec[i] > 0)
            FdB[i] = 20 * log10(FPowerSpec[i]/ZerodB);
        else
            FdB[i] = 20 * log10(MindB/ZerodB);
        }

    for (size_t i=0; i<FPowerSpec.size(); ++i)
        PSStats.Sample(FPowerSpec[i]);
}

//---------------------------------------------------------------------------
// AdcStats::AnalyzePowerSpectrum() -
//---------------------------------------------------------------------------

void AdcStats::AnalyzePowerSpectrum(size_t stride)
{
	Reset();
    FStride = stride;

    // Extract Power Spectrum
    for (size_t i=0; i < FPowerSpec.size(); ++i)
        {
		FPowerSpec[i] = pow(10.0, FdB[i]/20.0) * ZerodB;
        PSStats.Sample(FPowerSpec[i]);
        }

    try
        {
        analyze();
        }
    catch(...)
        {
        }
}

//---------------------------------------------------------------------------
// AdcStats::Pivot() - Pivot complex samples about axis
// Axis - Pivot about:  1: First half, 2: Second half, 3: Both halves
//                      4: Array midpoint
//---------------------------------------------------------------------------

void AdcStats::Pivot(DoubleArray & array, int axis)
{
    size_t half = array.size()/2;
    size_t quarter = array.size()/4;

    if (axis & 1)
        {
        size_t left = 0;
        size_t right = half;
        for (size_t i = 0; i < quarter; ++i)
            std::swap(array[left++], array[--right]);
        }

    if (axis & 2)
        {
        size_t left = half;
        size_t right = array.size();
        for (size_t i = 0; i < quarter; ++i)
            std::swap(array[left++], array[--right]);
        }

    if (axis & 4)
        {
        size_t left = 0;
        size_t right = array.size();
        for (size_t i = 0; i < half; ++i)
            std::swap(array[left++], array[--right]);
        }
}

//---------------------------------------------------------------------------
// AdcStats::Merge() - Merge real and imaginary samples
//---------------------------------------------------------------------------

void AdcStats::Merge(DoubleArray & array)
{
    Pivot(array, 2);

    size_t half = array.size()/2;
    for (size_t i=0; i<half; ++i)
        array[i] += array[i+half];
}

//---------------------------------------------------------------------------
// AdcStats::Within() -
//---------------------------------------------------------------------------

bool AdcStats::Within(int value, int lower, int upper)
{
    return (value > lower) && (value < upper);
}


//---------------------------------------------------------------------------
// AdcStats::NoisePowerBandwidth() -
//---------------------------------------------------------------------------

double AdcStats::NoisePowerBandwidth() const
{
    double result = 1.0;

    switch (FWindow)
        {
        case wtRect:        result = 1.0;       break;     // Uniform
        case wtBlackmanOpt:
        case wtBlackman:    result = 1.73;      break;     // alpha = 0.16
        case wtHamming:     result = 1.37;      break;
        case wtHanning:     result = 1.5;       break;
        case wtBartlett:    result = 1.46;      break;
        case wtKaiser:      result = 1.5;       break;     // alpha=2, 1.8:a=3
        }

    return result;
}

//---------------------------------------------------------------------------
// AdcStats::WindowGain() -
//---------------------------------------------------------------------------

double AdcStats::WindowGain() const
{
    double result = 1.0;

    switch (FWindow)
        {
        case wtRect:        result = 1.0;       break;     // Uniform
        case wtBlackmanOpt:
        case wtBlackman:    result = 0.42;      break;     // alpha = 0.16
        case wtHamming:     result = 0.54;      break;
        case wtHanning:     result = 0.50;      break;
        case wtBartlett:    result = 0.50;      break;
        case wtKaiser:      result = 0.50;      break;
        }

    return result;
}

//---------------------------------------------------------------------------
// AdcStats::Jitter() -  Calculates required jitter level
//---------------------------------------------------------------------------

double AdcStats::Jitter(int bits, double freq_in) const
{
    return 1.0 / (pow(2.0, bits+1)*pi*freq_in);
}



} // namespace Innovative



