//  AlertSystem_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2006 Innovative Integration Corporation.
//  All Rights Reserved.
//


#include <AlertSystem_Mb.h>
#include <BufferHeader_Mb.h>
#include <BufferDatagrams_Mb.h>

namespace Innovative
{

//==============================================================================
// CLASS  AlertSystem  --  Alert Processing Class
//==============================================================================
//------------------------------------------------------------------------------
//   AlertSystem::ClearAlerts() -- erase any Alert Info
//------------------------------------------------------------------------------

void  AlertSystem::ClearAlerts()
{
    Alerts.clear();
}

//------------------------------------------------------------------------------
//   AlertSystem::LoadAlert() -- Add alert information to database
//------------------------------------------------------------------------------

void  AlertSystem::LoadAlert(const AlertData & alert)
{
    Alerts.push_back(alert);
}

//------------------------------------------------------------------------------
//   AlertSystem::DoAlertEnable() -- Change Enable state of alert
//------------------------------------------------------------------------------

void  AlertSystem::DoAlertEnable(int key, bool state)
{
    AlertData * alert = FindAlertByKey(key);
    if (alert)
        alert->Enabled = state;
}

//------------------------------------------------------------------------------
//   AlertSystem::DoAlertPassThru() -- Change PassThru state of alert
//------------------------------------------------------------------------------

void  AlertSystem::DoAlertPassThru(int key, bool state)
{
    AlertData * alert = FindAlertByKey(key);
    if (alert)
        alert->PassThru = state;
}

//------------------------------------------------------------------------------
//   AlertSystem::DoAlertClear() -- Clear alert logic blocking
//------------------------------------------------------------------------------

void  AlertSystem::DoAlertClear(int key)
{
    AlertData * alert = FindAlertByKey(key);
    if (alert)
        {
        unsigned int enable_bit(0);

        enable_bit = alert->EnableBit;
        ApplyAlertClearHdw( enable_bit );
        }
}

//------------------------------------------------------------------------------
//   AlertSystem::FindAlertByKey() -- Find alert by key value
//------------------------------------------------------------------------------

AlertData *  AlertSystem::FindAlertByKey(int key)
{
    for (unsigned int i=0; i<Alerts.size(); ++i)
        if (Alerts[i].Key==key)
            return &Alerts[i];
    return 0;
}

//------------------------------------------------------------------------------
//   AlertSystem::FindAlertByStatusMask() -- Find alert by key value
//------------------------------------------------------------------------------

AlertData *  AlertSystem::FindAlertByStatusMask(unsigned int mask)
{
    for (unsigned int i=0; i<Alerts.size(); ++i)
        if (Alerts[i].StatusMask==mask)
            return &Alerts[i];
    return 0;
}

//------------------------------------------------------------------------------
//   AlertSystem::CheckAndProcess() -- Check Buffer for Alerts and process them
//------------------------------------------------------------------------------
//   Return code to Pass the event through to the Application or not

bool  AlertSystem::CheckAndProcess(const Buffer & buffer)
{
    const bool PassThru = true;
    bool ShouldPassThru(false);

    //  Is this an alert event?  If not, always PassThru
    PacketBufferHeader Hdr(buffer);
    int pid = Hdr.PeripheralId();
    if (pid != AlertPid)
        return PassThru;
    //  ...So we have an alert. Check the Status Word for alert bits
    for (unsigned int bit=0; bit<32; ++bit)
        {
        IntegerDG BufferDG(buffer);
        unsigned int mask = 1<<bit;
        if (BufferDG[0] & mask)       // if the bit is on...
            {
            AlertData * alert = FindAlertByStatusMask(mask);
            if (alert && alert->Enabled)
                {
                ShouldPassThru |= alert->PassThru;

                int timestamp = ExtractAlertTimestamp(buffer);
                int argument  = ExtractAlertArgument(buffer, bit, alert->Field);

                AlertSignalEvent e(AlertSenderTag, timestamp, argument);
                alert->AlertEvent->Execute(e);
                }
            }
        }
    return ShouldPassThru;
}

//------------------------------------------------------------------------------
//   AlertSystem::ExtractAlertTimestamp() -- Return alert timestamp
//------------------------------------------------------------------------------
//   Crack the timestamp out of the raw alert

int  AlertSystem::ExtractAlertTimestamp(const Buffer & alert)
{
    IntegerDG AlertDG(alert);
    return AlertDG[1];  // timestamp is at offset 1
}

//------------------------------------------------------------------------------
//   AlertSystem::ExtractAlertArgument() -- Return alert argument
//------------------------------------------------------------------------------
//   Crack the argument out of the raw alert

int  AlertSystem::ExtractAlertArgument(const Buffer & alert, int bit, int /*field*/)
{
    IntegerDG AlertDG(alert);
    return AlertDG[bit+1];  // argument is at position bit+1
}

//------------------------------------------------------------------------------
//   AlertSystem::ApplyAlertClearHdw() -- Set Alert Clear bit in software
//------------------------------------------------------------------------------

void  AlertSystem::ApplyAlertClearHdw(unsigned int /*bit*/)
{
    // normally, isn't supported
}


} // namespace Innovative
