// AlignedBlock_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure 
//  agreement with Innovative Integration Corporation and may not be copied or 
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2007 Innovative Integration Corporation.
//  All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <AlignedBlock_Mb.h>
#include <MemoryUtils_Mb.h>
#include <memory>
#include <algorithm>

namespace   Innovative
{
//==============================================================================
//  CLASS AlignedBlock  -- Chunk of Managed Memory
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class AlignedBlock
//------------------------------------------------------------------------------

AlignedBlock::AlignedBlock(size_t size_in_integers)
    :  ReferenceCount(0),
       AllocatedSize(0),
       Size(size_in_integers)
{
    Allocate(size_in_integers);
}

//------------------------------------------------------------------------------
//  constructor for class AlignedBlock -- copy ctor
//------------------------------------------------------------------------------

AlignedBlock::AlignedBlock(const AlignedBlock & other)
    :  ReferenceCount(0),
       AllocatedSize(other.AllocatedSize),
       Size(other.Size)
{
    Allocate(AllocatedSize);
    CopyFrom(other);
}

//------------------------------------------------------------------------------
//  destructor for class AlignedBlock
//------------------------------------------------------------------------------

AlignedBlock::~AlignedBlock()
{
    Deallocate();
}

//------------------------------------------------------------------------------
//  AlignedBlock::Allocate() -- allocate a new internal buffer
//------------------------------------------------------------------------------

void  AlignedBlock::Allocate(size_t size_in_ints)
{
    AllocatedSize = CalculateAllocatedSize(size_in_ints);
    Ptr = AlignedMemoryManager::Allocate(AllocatedSize*sizeof(int));
    //Ptr = new int[AllocatedSize];
}

//------------------------------------------------------------------------------
//  AlignedBlock::Deallocate() -- deallocate the internal pointer
//------------------------------------------------------------------------------

void  AlignedBlock::Deallocate()
{
    if (!Ptr.IsNull())
        AlignedMemoryManager::Deallocate( Ptr, AllocatedSize*sizeof(int));
    //delete [] Ptr;
    AllocatedSize = 0;
}

//------------------------------------------------------------------------------
//  AlignedBlock::Resize() -- Resize block
//------------------------------------------------------------------------------

void  AlignedBlock::Resize(size_t sizeInInts )
{
    if (AllocatedSize<sizeInInts)
        {
        //  need to get a new data block
        AlignedBlock replacement(sizeInInts);   // this one will have the right sized innards
        Swap(replacement);                        // ...so steal it!
        }
    else
        {
        Size = sizeInInts;
        }
}

//------------------------------------------------------------------------------
//  AlignedBlock::CopyFrom() -- Copy data from another block
//------------------------------------------------------------------------------

void  AlignedBlock::CopyFrom(const AlignedBlock &other )
{
    CopyInt32Buffer(other.ConstBase(), Base(), static_cast<int>(std::min(Size, other.Size)));
}

//------------------------------------------------------------------------------
//  AlignedBlock::AddReference() -- Increment Ref Count
//------------------------------------------------------------------------------

void  AlignedBlock::AddReference()
{
    Atomic.Increment(ReferenceCount);
}

//------------------------------------------------------------------------------
//  AlignedBlock::RemoveReference() -- Decrement Reference Count
//------------------------------------------------------------------------------

unsigned int  AlignedBlock::RemoveReference()
{
    return Atomic.DecrementToZero(ReferenceCount);
}

//------------------------------------------------------------------------------
//  AlignedBlock::CalculateAllocatedSize() --
//------------------------------------------------------------------------------

unsigned int AlignedBlock::CalculateAllocatedSize(size_t size )
{
    // Minimum Size is 4 words, Maximium is 2^30
    for (unsigned int i = 2; i < 30; i++)
        {
        if (size <= static_cast<unsigned int>( 1 << i ))
            return ( 1 << i );
        }

    //  Throw here since buffer is smaller than requested?
    return ( 1 << 30 );
}

//------------------------------------------------------------------------------
//  AlignedBlock::Swap() -- exchange innards with another block
//------------------------------------------------------------------------------

void AlignedBlock::Swap(AlignedBlock & other)
{
    //
    //  This function is private so we can use it internally. Dangerous
    //     since it changes the buffer w/o changing reference count.
    size_t tmpASize = AllocatedSize;
    size_t tmpSize  = Size;
    AlignedMemoryPtr  tmpPtr   = Ptr;
    //  int *           tmpPtr   = Ptr;

    Size          = other.Size;
    AllocatedSize = other.AllocatedSize;
    Ptr           = other.Ptr;

    other.Size          =  tmpSize;
    other.AllocatedSize =  tmpASize;
    other.Ptr           =  tmpPtr;
}

//------------------------------------------------------------------------------
//  AlignedBlock::operator new() -- allocate a new AlignedBlock
//------------------------------------------------------------------------------

void *  AlignedBlock::operator new(std::size_t size)
{
    return AlignedMemoryManager::Allocate(size).Origin();
}

//------------------------------------------------------------------------------
//  AlignedBlock::operator delete() -- free an AlignedBlock
//------------------------------------------------------------------------------

void  AlignedBlock::operator delete(void * ptr)
{
    AlignedMemoryPtr p(reinterpret_cast<char *>(ptr));
    AlignedMemoryManager::Deallocate(p, sizeof(AlignedBlock));
}

//==============================================================================
//  CLASS AlignedBlockHandle  -- Reference Counter Handle to a AlignedBlock
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class AlignedBlockHandle  -- default, take a block
//------------------------------------------------------------------------------

AlignedBlockHandle::AlignedBlockHandle(AlignedBlock * block)
    : Blk(block)
{
    if (Blk)
        Blk->AddReference();
}

//------------------------------------------------------------------------------
//  constructor for class AlignedBlockHandle -- create block now, using size
//------------------------------------------------------------------------------

AlignedBlockHandle::AlignedBlockHandle(AutoCreate /*tag*/, size_t size_in_integers)
    : Blk(new AlignedBlock(size_in_integers))
{
    if (Blk)
        Blk->AddReference();
}

//------------------------------------------------------------------------------
//  constructor for class AlignedBlockHandle -- copy ctor
//------------------------------------------------------------------------------

AlignedBlockHandle::AlignedBlockHandle(const AlignedBlockHandle & other)
    : Blk(other.Blk)
{
    if (Blk)
        Blk->AddReference();
}

//------------------------------------------------------------------------------
//  destructor for class AlignedBlockHandle
//------------------------------------------------------------------------------

AlignedBlockHandle::~AlignedBlockHandle()
{
    if (Blk && Blk->RemoveReference()==0)
        {
        delete Blk;
        }
}

//------------------------------------------------------------------------------
//  AlignedBlockHandle::Swap() - Exchange Body Contents
//------------------------------------------------------------------------------

void  AlignedBlockHandle::Swap(AlignedBlockHandle & other)
{
    std::swap(Blk, other.Blk);
}

//------------------------------------------------------------------------------
//  AlignedBlockHandle::MakeUnique() - Make this handle point to a unique block
//------------------------------------------------------------------------------

void  AlignedBlockHandle::MakeUnique()
{
    if (IsUnique() || Blk==0)
        return;

    AlignedBlockHandle handle(new AlignedBlock(Blk->SizeInInts()));
    handle.Blk->CopyFrom(Blk->Self());
    Swap(handle);
}

//------------------------------------------------------------------------------
//  AlignedBlockHandle::DeepCopy() - Make self an unshared clone of a handle
//------------------------------------------------------------------------------

void  AlignedBlockHandle::DeepCopy(AlignedBlockHandle & other)
{
    // shallow copy
    AlignedBlockHandle handle(other);
    // now unshared
    handle.MakeUnique();
    // and we take his block for ours
    Swap(handle);
}

//------------------------------------------------------------------------------
//  AlignedBlockHandle::Resize() - Resize my block
//------------------------------------------------------------------------------

void  AlignedBlockHandle::Resize(size_t size_in_integers)
{
    if (IsNull())
        {
        AlignedBlockHandle replacement(createBlock, size_in_integers);
        Swap(replacement);
        }
    else if (IsUnique() && size_in_integers<=Blk->AllocatedInInts())
        {
        //  We can fast resize our current block without problems
        Blk->Resize(size_in_integers);
        }
    else  // need to reallocate the block
        {
        // To "resize" we just have to create a new block, then grab it
        AlignedBlockHandle replacement(createBlock, size_in_integers);
        Swap(replacement);
        }
}


}  // namespace Innovative



