// AlignedMemoryPool_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <AlignedMemoryPool_Mb.h>
#include <list>
#include <map>
#include <sstream>

namespace   Innovative
{
//==============================================================================
//  CLASS AlignedMemoryPtr -- Pointer to chunk of aligned memory
//==============================================================================
//------------------------------------------------------------------------------
//  housekeeping for class AlignedMemoryPtr
//------------------------------------------------------------------------------

const int DefaultAlignment = 32;     // to align our ptr on an N byte boundary
                                     //   we need to allocate more to slide to
                                     //   the right spot.

//------------------------------------------------------------------------------
//  constructor for class AlignedMemoryPtr -- takes a size in bytes
//------------------------------------------------------------------------------

AlignedMemoryPtr::AlignedMemoryPtr(size_t size_in_bytes)
    : FOrigin(0), FAddress(0)
{
    New(size_in_bytes);
}

//------------------------------------------------------------------------------
//  constructor for class AlignedMemoryPtr -- wrap an existing buffer
//------------------------------------------------------------------------------

AlignedMemoryPtr::AlignedMemoryPtr(char * origin)
    : FOrigin(origin)
{
    FAddress = AddressFromOrigin(FOrigin);
}

//---------------------------------------------------------------------------
//  constructor for class AlignedMemoryPtr --  default (uninitialized)
//---------------------------------------------------------------------------

AlignedMemoryPtr::AlignedMemoryPtr()
    : FOrigin(0), FAddress(0)
{
}

//---------------------------------------------------------------------------
//  destructor for class AlignedMemoryPtr --
//---------------------------------------------------------------------------

AlignedMemoryPtr::~AlignedMemoryPtr()
{
}

//---------------------------------------------------------------------------
//  AlignedMemoryPtr::New() --  Allocate block
//---------------------------------------------------------------------------

AlignedMemoryPtr AlignedMemoryPtr::New(size_t bytes)
{
    FOrigin = new char[bytes+DefaultAlignment];
    FAddress = AddressFromOrigin(FOrigin);

    return *this;
}

//---------------------------------------------------------------------------
//  AlignedMemoryPtr::Delete() --  Deallocate block
//---------------------------------------------------------------------------

void AlignedMemoryPtr::Delete()
{
    delete [] FOrigin;
    FOrigin = 0;
    FAddress = 0;
}

//---------------------------------------------------------------------------
//  AlignedMemoryPtr::IsNull() --  Check pointer state
//---------------------------------------------------------------------------

bool AlignedMemoryPtr::IsNull() const
{
    return FOrigin == 0;
}

//---------------------------------------------------------------------------
//  AlignedMemoryPtr::Origin() --  Unaligned address
//---------------------------------------------------------------------------

char * AlignedMemoryPtr::Origin() const
{
    return FOrigin;
}

//---------------------------------------------------------------------------
//  AlignedMemoryPtr::Address() --  Aligned address
//---------------------------------------------------------------------------

int * AlignedMemoryPtr::Address() const
{
    return FAddress;
}

//------------------------------------------------------------------------------
//  AlignedMemoryPtr::AddressFromOrigin() --  Aligned from unaligned address
//------------------------------------------------------------------------------

int *  AlignedMemoryPtr::AddressFromOrigin(char * origin)
{
    size_t original_loc = reinterpret_cast<size_t>(origin);
    size_t offset = DefaultAlignment - (original_loc%DefaultAlignment);

    char * addr_ptr = origin + offset;
    return reinterpret_cast<int *>(addr_ptr);

}

//==============================================================================
//  CLASS AlignedMemoryList  -- Container for Aligned Memory Ptrs
//==============================================================================
//------------------------------------------------------------------------------
//  declaration for class AlignedMemoryList
//------------------------------------------------------------------------------

class AlignedMemoryList
{
public:
    //
    //  Member Functions
    //  ...constructors and destructors
    AlignedMemoryList(size_t aligned_mem_size_bytes)
        : AM_Size(aligned_mem_size_bytes)
        {
        }

    ~AlignedMemoryList()
        {
        for (std::list<AlignedMemoryPtr>::iterator Iter = AlignedMemList.begin();
                Iter != AlignedMemList.end(); Iter++)
            {
            AlignedMemoryPtr p = *Iter;
            p.Delete();
            }
        }

    size_t Size()
        {
        return AM_Size;
        }
    size_t BlockCount()
        {
        return AlignedMemList.size();
        }

    void Preallocate(size_t count=1)
        {
        //  allocate buffers and insert on list for later use
        for (size_t i=0; i<count; i++)
            AlignedMemList.push_back(AlignedMemoryPtr(AM_Size));
        }

    AlignedMemoryPtr Allocate()
        {
        //  if we have no memory on the list, make one. Otherwise, use one
        //     from the list
        if (AlignedMemList.empty())
            {
            AlignedMemoryPtr p(AM_Size);
            return p;
            }
        else
            {
            AlignedMemoryPtr result(AlignedMemList.back());
            AlignedMemList.pop_back();
            return result;
            }
        }

    void Deallocate(const AlignedMemoryPtr & data_ptr)
        {
        if (!data_ptr.IsNull())
            AlignedMemList.push_back(data_ptr);
        }
        
protected:
    size_t                          AM_Size;
    std::list<AlignedMemoryPtr>     AlignedMemList;


};

//==============================================================================
//  CLASS AlignedMemoryPool  --  Collection of Aligned Memory Data
//==============================================================================
//------------------------------------------------------------------------------
//  declaration for class AlignedMemoryPool - holds AM Lists of all sizes
//------------------------------------------------------------------------------

class AlignedMemoryPool
    : public std::map<size_t, AlignedMemoryList *>
{
};

//==============================================================================
//  CLASS AlignedMemoryManager  --  AlignedMemory Manager
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class AlignedMemoryManager
//------------------------------------------------------------------------------

AlignedMemoryManager::AlignedMemoryManager()
{
    Pool = new AlignedMemoryPool;
}

//------------------------------------------------------------------------------
//  housekeeping for class AlignedMemoryManager
//------------------------------------------------------------------------------

unsigned int            AlignedMemoryManager::AllocationCount = 0;
AlignedMemoryManager *  AlignedMemoryManager::Instance        = 0;

//------------------------------------------------------------------------------
//  destructor for class AlignedMemoryManager
//------------------------------------------------------------------------------

AlignedMemoryManager::~AlignedMemoryManager()
{
    ReleaseAll_Impl();

    delete Pool;
    Pool = 0;

    Instance = 0;
}

//------------------------------------------------------------------------------
//  AlignedMemoryManager::Allocate() -- Singleton Allocate a block
//------------------------------------------------------------------------------

AlignedMemoryPtr AlignedMemoryManager::Allocate(size_t size_in_bytes)
{
    if (size_in_bytes == 0)
        return AlignedMemoryPtr();  // return a Null pointer
      
    if (!Instance)
        Instance = new AlignedMemoryManager;

    Instance->Atomic.Increment(AllocationCount);
    return Instance->Allocate_Impl(size_in_bytes);
}

//------------------------------------------------------------------------------
//  AlignedMemoryManager::Deallocate() --
//------------------------------------------------------------------------------

void AlignedMemoryManager::Deallocate(AlignedMemoryPtr data, size_t size_in_bytes)
{
    if (!Instance || (size_in_bytes == 0))
      return;

    Instance->Deallocate_Impl(data, size_in_bytes);

    Instance->Atomic.Decrement(AllocationCount);
}

//---------------------------------------------------------------------------
//  AlignedMemoryManager::ReleaseAll() --
//---------------------------------------------------------------------------

void AlignedMemoryManager::ReleaseAll()
{
    if (Instance)
        Instance->ReleaseAll_Impl();
}

//---------------------------------------------------------------------------
//  AlignedMemoryManager::Preallocate() --
//---------------------------------------------------------------------------

void AlignedMemoryManager::Preallocate(size_t size_in_bytes, size_t count)
{
    if (!Instance)
        Instance = new AlignedMemoryManager;

    Instance->Preallocate_Impl(size_in_bytes, count);
}
//------------------------------------------------------------------------------
//  AlignedMemoryManager::GetAM_forSize() --  Fetch AlignedMemoryList for a size
//------------------------------------------------------------------------------

AlignedMemoryList *   AlignedMemoryManager::GetAMList_forSize(size_t size_in_bytes)
{
    //
    //  Fetch the Pool node for this size of buffer. If none present this
    //    will return a null pointer
    AlignedMemoryList * anAM_List = (*Pool)[size_in_bytes];
    if (!anAM_List)
        {
        //  Make a new list if none present, insert in the Pool
        anAM_List = new AlignedMemoryList(size_in_bytes);
        (*Pool)[size_in_bytes] = anAM_List;
        }
    return anAM_List;
}

//------------------------------------------------------------------------------
//  AlignedMemoryManager::Allocate_Impl() --  Implementation of Allocate()
//------------------------------------------------------------------------------

AlignedMemoryPtr AlignedMemoryManager::Allocate_Impl(size_t size_in_bytes)
{
    MutexedLock Lock(Atomic.LockObject());

    AlignedMemoryList * anAM_List = GetAMList_forSize(size_in_bytes);
    //
    //  Allocate the block from the list
    return anAM_List->Allocate();
}

//---------------------------------------------------------------------------
//  AlignedMemoryManager::Deallocate_Impl() --
//---------------------------------------------------------------------------

void AlignedMemoryManager::Deallocate_Impl(AlignedMemoryPtr data, size_t size_in_bytes)
{
    MutexedLock Lock(Atomic.LockObject());

    AlignedMemoryList * anAM_List = GetAMList_forSize(size_in_bytes);

    anAM_List->Deallocate(data);
}


//---------------------------------------------------------------------------
//  AlignedMemoryManager::ReleaseAll_Impl() --
//---------------------------------------------------------------------------

void AlignedMemoryManager::ReleaseAll_Impl()
{
    MutexedLock Lock(Atomic.LockObject());

    //  Delete the lists on each node, then the nodes themselves
    for (AlignedMemoryPool::iterator Iter = Pool->begin() ;  Iter!=Pool->end(); Iter++)
        delete Iter->second;

    Pool->clear();
}

//---------------------------------------------------------------------------
//  AlignedMemoryManager::Preallocate_Impl() --
//---------------------------------------------------------------------------

void AlignedMemoryManager::Preallocate_Impl(size_t size_in_bytes, size_t count)
{
    MutexedLock Lock(Atomic.LockObject());

    AlignedMemoryList * anAM_List = GetAMList_forSize(size_in_bytes);

    anAM_List->Preallocate(count);
}

//---------------------------------------------------------------------------
//  AlignedMemoryManager::Info() -- dump text about the pool
//---------------------------------------------------------------------------

void AlignedMemoryManager::Info(std::string & result)
{
    if (Instance)
        Instance->Info_Impl(result);
}

//---------------------------------------------------------------------------
//  AlignedMemoryManager::Info_Impl() -- dump text about the pool
//---------------------------------------------------------------------------

void AlignedMemoryManager::Info_Impl(std::string & result)
{
    MutexedLock Lock(Atomic.LockObject());

    std::stringstream msg;
    msg << "Total Still Allocated: " << AllocationCount << '\r' << std::endl;
    for (AlignedMemoryPool::iterator Iter = Pool->begin() ;  Iter!=Pool->end(); Iter++)
        {
        AlignedMemoryList * AML = Iter->second;
        if (AML==0)
            continue;
        msg << " Chars: " << AML->Size() << " Blks: " << AML->BlockCount() << '\r' << std::endl;
        }

    result = msg.str();
}


} // namespace Innovative


