//---------------------------------------------------------------------------

#ifndef BlockChecker_AppH
#define BlockChecker_AppH


#include <BufferDatagrams_Mb.h>
#include <BufferHeader_Mb.h>
#include <vector>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//==============================================================================
//  CLASS IBlockCheckerTester  --  Interface to create a generated pattern
//==============================================================================

class IBlockCheckerTester
{
public:
    IBlockCheckerTester()
        :  FPid(0), FBlockSize(0x10000)
        {}
    virtual ~IBlockCheckerTester()
        {}

    //  Properties
    void  BlockPid(unsigned int pid)
             {  FPid = pid;  }
    unsigned int  BlockPid()
             {  return FPid;  }

    void  BlockSize(unsigned int data_size_ints)
                    {  FBlockSize = data_size_ints;  }
    unsigned int  BlockSize()
                    {  return FBlockSize;  }

    virtual void  TestInit() = 0;
    virtual void  RecvPatternGenerate(Innovative::Buffer & out_buffer) = 0;
    virtual void  XmitPatternGenerate(Innovative::Buffer & out_buffer) = 0;
    virtual bool  Compare(Innovative::Buffer & test, Innovative::Buffer & should_be ) = 0;

protected:
    bool FullCompare(Innovative::Buffer & test, Innovative::Buffer & should_be);
    bool QuickCompare(Innovative::Buffer & test, Innovative::Buffer & should_be);

    unsigned int FPid;
    unsigned int FBlockSize;
};

//==============================================================================
//  CLASS PatternBlockTester  --  Use a single pattern block as a template
//==============================================================================

class PatternBlockTester : public IBlockCheckerTester
{
public:
    PatternBlockTester()
        {}
    virtual ~PatternBlockTester()
        {}
    // Properties
    // Methods
    virtual void  TestInit();
    virtual void  RecvPatternGenerate(Innovative::Buffer & out_buffer);
    virtual void  XmitPatternGenerate(Innovative::Buffer & out_buffer);

protected:
    Innovative::Buffer  Pattern;

    virtual void  FillPatternBuffer() = 0;
};

//==============================================================================
//  CLASS RampBlockTester  --  Standard Ramp Block Tester
//==============================================================================

class RampBlockTester : public PatternBlockTester
{
public:
    RampBlockTester();
    virtual ~RampBlockTester()
        {}
    virtual bool  Compare(Innovative::Buffer & test, Innovative::Buffer & should_be );

protected:
    virtual void  FillPatternBuffer();
};

//==============================================================================
//  CLASS RampBlockTester16  --  continuous, 16 bit Ramp Block Tester
//==============================================================================

class RampBlockTester16 : public PatternBlockTester
{
public:
    RampBlockTester16();
    virtual ~RampBlockTester16()
        {}
    virtual void  TestInit()
        {
        PatternBlockTester::TestInit();
        Ramp = 0;
        }

    virtual bool  Compare(Innovative::Buffer & test, Innovative::Buffer & should_be );
    virtual void  RecvPatternGenerate(Innovative::Buffer & out_buffer);

protected:
    virtual void  FillPatternBuffer();

    unsigned short Ramp;
};

//==============================================================================
//  CLASS QuickRampBlockTester  --  Quick Ramp Block Tester (continuity only)
//==============================================================================

class QuickRampBlockTester : public RampBlockTester
{
public:
    QuickRampBlockTester();
    virtual ~QuickRampBlockTester()
        {}
    virtual bool  Compare(Innovative::Buffer & test, Innovative::Buffer & should_be );
};

//==============================================================================
//  CLASS AlternatingBitsBlockTester  --  A5-5A Ramp Block Tester
//==============================================================================

class AlternatingBitsBlockTester : public PatternBlockTester
{
public:
    AlternatingBitsBlockTester();
    virtual ~AlternatingBitsBlockTester()
        {}
    virtual bool  Compare(Innovative::Buffer & test, Innovative::Buffer & should_be );

protected:
    virtual void  FillPatternBuffer();
};
//==============================================================================
//  CLASS WalkingBitBlockTester  --  A5-5A Ramp Block Tester
//==============================================================================

class WalkingBitBlockTester : public PatternBlockTester
{
public:
    WalkingBitBlockTester();
    virtual ~WalkingBitBlockTester()
        {}
    virtual bool  Compare(Innovative::Buffer & test, Innovative::Buffer & should_be );

protected:
    virtual void  FillPatternBuffer();
};

//==============================================================================
//  CLASS RandomBitsBlockTester  --  Random bits Block Tester
//==============================================================================

class RandomBitsBlockTester : public PatternBlockTester
{
public:
    RandomBitsBlockTester();
    virtual ~RandomBitsBlockTester()
        {}
    virtual bool  Compare(Innovative::Buffer & test, Innovative::Buffer & should_be );

protected:
    virtual void  FillPatternBuffer();
};

//==============================================================================
//  CLASS BlockChecker  --  Check block data against a generated pattern
//==============================================================================

class BlockChecker
{
public:
    enum TestModes {tmWalkingBitsTest, tmRampTest, tmQuickRampTest, /*tmContinuousRampTest,*/
                     tmAlternatingBitsTest, tmRandomBitsTest, tm16BitRamp, tmCustomTest };

    BlockChecker(TestModes mode = tmRampTest);

    TestModes     Mode()  { return mMode; }
    void          Mode(TestModes mode) {  mMode = mode;  SwitchMode(); };
    void          Mode(int mode) {  Mode(static_cast<TestModes>(mode)); };

    void  Init();
    void  Fill(Innovative::Buffer & out);
    void  Check(Innovative::Buffer & in);
    void  LoadCustomTest(IBlockCheckerTester * custom)
            {  CustomTest = custom;  }

    void  LogToDisk(const std::string & file_name);

    unsigned int  BlockPid();
    void          BlockPid(unsigned int data_size_ints);
    unsigned int  BlockSize();
    void          BlockSize(unsigned int data_size_ints);
    unsigned int  MaxBlocks()  { return mMaxBlocks; }
    void          MaxBlocks(unsigned int max) {  mMaxBlocks = max;  };
    bool          KeepFirstBlocks()  { return mKeepFirst; }
    void          KeepFirstBlocks(bool state) {  mKeepFirst = state;  };

    int       BlocksWithErrors() { return mBlocksWithErrors; }
    int       BlocksSaved() {  return static_cast<int>(ErrorBlocks.size());  }
    Innovative::Buffer &      ErrorBlock(int idx)
            {  return ErrorBlocks[idx];  }
    Innovative::Buffer &      ShouldBeBlock(int idx)
            {  return ShouldBeBlocks[idx];  }
    //
    //  Standard Tests
    RampBlockTester             RampTest;
    QuickRampBlockTester        QuickRampTest;
    AlternatingBitsBlockTester  AltBitsBlockTest;
    RandomBitsBlockTester       RandomTest;
    WalkingBitBlockTester       WalkingBitsTest;
    RampBlockTester16           RampTest16;

private:
    IBlockCheckerTester * CurrentTest;
    TestModes     mMode;
    unsigned int  mMaxBlocks;
    int           mBlocksWithErrors;
    bool          mKeepFirst;

    IBlockCheckerTester * CustomTest;

    std::vector<Innovative::Buffer>      ErrorBlocks;
    std::vector<Innovative::Buffer>      ShouldBeBlocks;

    void  SwitchMode();

};

#ifdef __CLR_VER
#pragma managed(pop)
#endif
}; // namespace Innovative

#endif
