//---------------------------------------------------------------------------

#include <Application/IniSaver_App.h>
#include <SystemSupport_Mb.h>

#include "Poco/JSON/Object.h"
#include "Poco/JSON/Parser.h"
#include "Poco/JSON/Query.h"
#include "Poco/JSON/JSONException.h"
#include "Poco/Util/JSONConfiguration.h"

namespace Innovative
{

std::string DefaultIniFilenameString()
{
    return std::string("Settings.ini");
}

//==============================================================================
//  CLASS IniSettingsStore
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class IniSettingsStore
//------------------------------------------------------------------------------

IniSettingsStore::IniSettingsStore()
{

}

//------------------------------------------------------------------------------
//  destructor for class IniSettingsStore
//------------------------------------------------------------------------------

IniSettingsStore::~IniSettingsStore()
{
}

//------------------------------------------------------------------------------
//  IniSettingsStore::Clear()  --
//------------------------------------------------------------------------------

void  IniSettingsStore::Clear()
{
    settings.clear();
}

//------------------------------------------------------------------------------
//  IniSettingsStore::Install(const IIniSaveable & )  -- register a setting
//------------------------------------------------------------------------------

void  IniSettingsStore::Install(const IIniSaveable & setting)
{
    settings.push_back( CountedPtr<IIniSaveable>( setting.Clone() ));
}

//------------------------------------------------------------------------------
//  IniSettingsStore::Install(IIniSaveable * )  -- register a setting
//------------------------------------------------------------------------------

void  IniSettingsStore::Install(IIniSaveable * setting)
{
    settings.push_back( CountedPtr<IIniSaveable>( setting->Clone() ));
}

//------------------------------------------------------------------------------
//  IniSettingsStore::Load()  -- Load registered settings
//------------------------------------------------------------------------------

void  IniSettingsStore::Load(IniProtocol * Ini)
{
    for (SettingsStoreType::iterator iter=settings.begin() ;
                 iter != settings.end(); ++iter)
        {
        (*iter)->Load(Ini);
        }
}

//------------------------------------------------------------------------------
//  IniSettingsStore::Save()  -- Save registered settings
//------------------------------------------------------------------------------

void  IniSettingsStore::Save(IniProtocol * Ini)
{
    for (SettingsStoreType::iterator iter=settings.begin() ;
                 iter != settings.end(); ++iter)
        {
        (*iter)->Save(Ini);
        }
}

//==============================================================================
//  CLASS IniSaver
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class IniSaver
//------------------------------------------------------------------------------

IniSaver::IniSaver(const std::string & file, bool local_path, bool auto_save)
    :  file_name(file)
{
	Application App;
	root = App.Path();

	// Locate INI file on root of C if remote debugging
    std::string ini_file;
    if (local_path)
        {
        ini_file = root + file_name;
        }
    else
        {
        ini_file = file_name;
        }
    Ini = new IniFile(ini_file, false, 100000, auto_save);

}

//------------------------------------------------------------------------------
//  destructor for class IniSaver
//------------------------------------------------------------------------------

IniSaver::~IniSaver()
{
	delete Ini;
	Ini = 0;
}

//------------------------------------------------------------------------------
//  IniSaver::Load() -- Load settings variables from INI
//------------------------------------------------------------------------------

void  IniSaver::Load()
{
    IniSettingsStore::Load(Ini);
}

//------------------------------------------------------------------------------
//  IniSaver::Save() -- Save settings variables to INI
//------------------------------------------------------------------------------

void  IniSaver::Save()
{
    IniSettingsStore::Save(Ini);
}


//==============================================================================
//  CLASS JsonSaver
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class JsonSaver
//------------------------------------------------------------------------------

JsonSaver::JsonSaver(const std::string  file, bool local_path, bool auto_save)
    :  file_name(file)
{
    Application App;
    root = App.Path();


    std::string ini_file;
    if (local_path)
        {
        ini_file = root + file_name;
        }
    else
        {
        ini_file = file_name;
        }
    JFile = new JsonFile(ini_file, false, 100000, auto_save);

}

//------------------------------------------------------------------------------
//  destructor for class JsonSaver
//------------------------------------------------------------------------------

JsonSaver::~JsonSaver()
{
    delete JFile;
    JFile = 0;
}

//------------------------------------------------------------------------------
//  JsonSaver::Load() -- Load settings variables from INI
//------------------------------------------------------------------------------

void  JsonSaver::Load()
{
    IniSettingsStore::Load(JFile);
}

//------------------------------------------------------------------------------
//  JsonSaver::Save() -- Save settings variables to INI
//------------------------------------------------------------------------------

void  JsonSaver::Save()
{
    IniSettingsStore::Save(JFile);
}

//------------------------------------------------------------------------------
//  JsonSaver::Query() --
//------------------------------------------------------------------------------

std::string JsonSaver::Query(const std::string &s)
{
    if (s == ".") // query root
    {
        std::ostringstream out;
        JFile->getImpl()->save(out);
        return out.str();
    }

    std::string last;
    Poco::JSON::Object::Ptr obj = JFile->getImpl()->findStart(s, last);

    Poco::Dynamic::Var v = obj->get(last);
    return v.isEmpty() ? "" : v.toString();
}

//------------------------------------------------------------------------------
//  JsonSaver::Update() --
//------------------------------------------------------------------------------

void JsonSaver::Update(const std::string &s)
{
    Poco::JSON::Parser parser;
    Poco::Dynamic::Var result;

    try
    {
        result = parser.parse(s);
    }
    catch (Poco::Exception& )
    {
        return;
    }
    // get name
    Poco::JSON::Query qname(result);
    std::string name = qname.findValue("name", "");
    // get value
    Poco::JSON::Query qval(result);
    std::string value = qval.findValue("value", "");
    // if nothing to do
    if (name.empty() || value.empty()) return;

    if (name == ".")
    {
        std::stringstream in(value);
        JFile->getImpl()->load(in);
        return;
    }
    // find object
    std::string last;
    Poco::JSON::Object::Ptr obj = JFile->getImpl()->findStart(name, last);
    obj->set(last, value);
}

//==============================================================================
//  CLASS IniParser
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class IniParser
//------------------------------------------------------------------------------

IniParser::IniParser()
    :  Mem(new IniMemory)
{
}

//------------------------------------------------------------------------------
//  destructor for class IniParser
//------------------------------------------------------------------------------

IniParser::~IniParser()
{
	delete Mem;
	Mem = 0;
}

//------------------------------------------------------------------------------
//  IniParser::Load() -- Load settings variables from Mem
//------------------------------------------------------------------------------

void  IniParser::Load()
{
    IniSettingsStore::Load(Mem);
}

//------------------------------------------------------------------------------
//  IniParser::Save() -- Save settings variables to Mem
//------------------------------------------------------------------------------

void  IniParser::Save()
{
    IniSettingsStore::Save(Mem);
}

//------------------------------------------------------------------------------
//  IniParser::Text() -- Output settings as INI-style text
//------------------------------------------------------------------------------

std::string  IniParser::Text()
{
    return Mem->Text();
}

//------------------------------------------------------------------------------
//  IniParser::Text() -- Initialize Mem from INI-style text string
//------------------------------------------------------------------------------

void  IniParser::Text(const std::string & text)
{
    Mem->Text(text);
}

}  // end namespace Innovative

