//---------------------------------------------------------------------------

#ifndef IniSaver_AppH
#define IniSaver_AppH
//---------------------------------------------------------------------------
#include <string>
#include <IniFile_Mb.h>
#include <CountedPtr_Mb.h>
#include <vector>
namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

#ifdef __BORLANDC__
#define TYPENAME typename
#elif defined(__GNUC__)
#define TYPENAME typename
#else
#define TYPENAME
#endif

// Helper Macro - String Var Pair
#define _SVP_(x) std::string(#x), x

#define NEW_STYLE_ACTIVE 1

//==============================================================================
//  CLASS IIniSaveable - Interface to all INI savable objects
//==============================================================================

class IIniSaveable
{
public:
    virtual ~IIniSaveable()  {}
    virtual void  Save(IniProtocol *   Ini) const = 0;
    virtual void  Load(IniProtocol *   Ini) const = 0;

    virtual IIniSaveable * Clone() const = 0;
};

//==============================================================================
//  CLASS IniSettingsStore
//==============================================================================

class IniSettingsStore
{
public:
    //
    //  Member Functions
    IniSettingsStore();
    virtual ~IniSettingsStore();

    void  Clear();

    void  Install(const IIniSaveable & setting);
    void  Install(IIniSaveable * setting);

protected:
    void  Load(IniProtocol * Ini);
    void  Save(IniProtocol * Ini);

private:
    //
    //  Member Data
    typedef std::vector< CountedPtr<IIniSaveable> > SettingsStoreType;
    SettingsStoreType  settings;

};

std::string DefaultIniFilenameString();

//==============================================================================
//  CLASS IniSaver
//==============================================================================

class IniSaver : public IniSettingsStore
{
public:
    //
    //  Member Functions
    IniSaver(const std::string & file = DefaultIniFilenameString(),
                                                        bool local_path=true,
                                                        bool auto_save=true);
    virtual ~IniSaver();

    void  Load();
    void  Save();

    std::string     Root()
        {  return root;  }

protected:
    IniFile *   Ini;

private:
    //
    //  Member Data
    std::string             root;
    std::string             file_name;

    IniSaver(const IniSaver &);   // unimplemented, non copyable
};

//==============================================================================
//  CLASS JsonSaver
//==============================================================================

class JsonSaver : public IniSettingsStore
{
public:
    //
    //  Member Functions
    JsonSaver(const std::string file = "Settings.json",
                                                        bool local_path=true,
                                                        bool auto_save=true);
    virtual ~JsonSaver();

    void  Load();
    void  Save();

    std::string Query(const std::string &s);
    void Update(const std::string &s);

    std::string     Root()
        {  return root;  }

protected:
    JsonFile *   JFile;

private:
    //
    //  Member Data
    std::string             root;
    std::string             file_name;

    JsonSaver(const JsonSaver &);   // unimplemented, non copyable
};

//==============================================================================
//  CLASS IniReader -- Reads data from INI
//==============================================================================

class IniReader : public IniSaver
{
public:
    //
    //  Member Functions
    IniReader(const std::string & file)
        :   IniSaver(file, false, false)
        { /* absolute file, turn off auto_save on exit */ }
};

//==============================================================================
//  CLASS IniParser -- use IniMemory to parse INI file to/from strings
//==============================================================================

class IniParser : public IniSettingsStore
{
public:
    //
    //  Member Functions
    IniParser();
    virtual ~IniParser();

    void  Load();
    void  Save();

    //  output text as INI string
    std::string  Text();
    void         Text(const std::string & text);

protected:
    IniMemory *   Mem;

private:

    IniParser(const IniParser &);   // unimplemented, non copyable
};

#if !NEW_STYLE_ACTIVE

//==============================================================================
//  TEMPLATE CLASS IniSetting - Implements INI Saving for a type
//==============================================================================
//  Original, restrictive setting template

template <class T>
class IniSetting : public IIniSaveable
{
public:
    IniSetting(const std::string & tag, T & var, T default_value)
        :  Section(""), Tag(tag), Var(var), Default(default_value)
        {
        }
    IniSetting(const std::string & section, const std::string & tag, T & var, T default_value)
        :  Section(section), Tag(tag), Var(var), Default(default_value)
        {
        }

    virtual void Save(IniProtocol * Ini) const
        {
        Ini->Section = Section;
        Ini->Write(Tag, Var);
        Ini->Section = std::string("");
        }
    virtual void Load(IniProtocol *  Ini) const
        {
        Ini->Section = Section;
        Var = Ini->Read(Tag, Default);
        Ini->Section = std::string("");
        }
    virtual IIniSaveable * Clone() const
        {
        return new IniSetting<T>(*this);
        }

private:
    std::string  Section;
    std::string  Tag;
    T &          Var;
    T            Default;
};


//==============================================================================
//  TEMPLATE CLASS IniSetting - Vector specialization for IniSettings
//==============================================================================
template <class T>
class IniSetting<std::vector<T> > : public IIniSaveable
{
public:
    IniSetting(const std::string & tag, std::vector<T> & var, std::vector<T> default_value)
        :  Section(""), Tag(tag), Var(var), Default(default_value)
        {
        }
    IniSetting(const std::string & section, const std::string & tag, std::vector<T> & var, std::vector<T> default_value)
        :  Section(section), Tag(tag), Var(var), Default(default_value)
        {
        }

    virtual void Save(IniProtocol * Ini) const
        {
        Ini->Section = Section;
        for (size_t i = 0; i < Var.size(); i++)
            Ini->Write(Tag + std::string("Field") + IntToString(i), Var[i]);
        Ini->Section = std::string("");
        }
    virtual void Load(IniProtocol *  Ini) const
        {
        Ini->Section = Section;
        for (size_t i = 0; i < Var.size(); i++)
            Var[i] = Ini->Read(Tag + std::string("Field") + IntToString(i), Default[i]);
        Ini->Section = std::string("");
        }
    virtual IIniSaveable * Clone() const
        {
        return new IniSetting<std::vector<T> >(*this);
        }

private:
    std::string			Section;
    std::string			Tag;
    std::vector<T> &    Var;
    std::vector<T>      Default;
};


//==============================================================================
//  TEMPLATE CLASS IniList<T> - Implements INI Saving for a vector
//==============================================================================

template <class T>
class IniList : public IIniSaveable
{
public:
    IniList(const std::string & section, std::vector<T> & var, T default_value, bool autosize=true)
        :  Section(section), TagBase("Idx_"), Var(var), Default(default_value), AutoSize(autosize)
        {
        }
    IniList(const std::string & section, const std::string & tag_base, std::vector<T> & var, T default_value, bool autosize=true)
        :  Section(section), TagBase(tag_base), Var(var), Default(default_value), AutoSize(autosize)
        {
        }

#ifndef GCC
    virtual void Save(IniProtocol * Ini) const
        {
        //  Read Data from section given by Tag
        Ini->Section = Section;

        size_t idx = 0;
        for (TYPENAME std::vector<T>::iterator i = Var.begin(); i != Var.end(); ++i)
            {
            std::string ArrayTag(TagBase + Innovative::IntToString(static_cast<int>(idx++)));
            Ini->Write(ArrayTag, *i);
            }
        //  Write ArraySize Tag if auto-sizing
        if (AutoSize)
            {
            Ini->Write("ArraySize", Var.size());
            }
        //  Restore section
        Ini->Section = std::string("");
        }

    virtual void Load(IniProtocol *  Ini) const
        {
        //  Read Data from section given by Tag
        Ini->Section = Section;
        //  Check for override size tag parameter, resize if present
        size_t OverrideSize=Ini->Read("ArraySize", 0);
        if (OverrideSize!=0)
            Var.resize(OverrideSize);

        size_t idx = 0;
        for (typename std::vector<T>::iterator i = Var.begin(); i != Var.end(); ++i)
            {
            std::string ArrayTag(TagBase + Innovative::IntToString(static_cast<int>(idx++)));
            *i = Ini->Read(ArrayTag, Default);
            }
        //  Restore section
        Ini->Section = std::string("");
        }
#else
    // GCC has a bug in defining std::vector<T>::iterator so work around it
    virtual void Save(IniProtocol * Ini) const
        {
        //  Read Data from section given by Tag
        Ini->Section = Section;

        size_t idx = 0;
        for (size_t i = 0 ; i<Var.size(); i++)
            {
            std::string ArrayTag(TagBase + Innovative::IntToString(idx++));
            Ini->Write(ArrayTag, Var[i]);
            }
        //  Write ArraySize Tag if auto-sizing
        if (AutoSize)
            {
            Ini->Write("ArraySize", Var.size());
            }
        //  Restore section
        Ini->Section = std::string("");
        }

    virtual void Load(IniProtocol *  Ini) const
        {
        //  Read Data from section given by Tag
        Ini->Section = Section;
        //  Check for override size tag parameter, resize if present
        size_t OverrideSize=Ini->Read("ArraySize", 0);
        if (OverrideSize!=0)
            Var.resize(OverrideSize);

        size_t idx = 0;
        for (size_t i = 0; i<Var.size(); i++)
            {
            std::string ArrayTag(TagBase + Innovative::IntToString(idx++));
            Var[i] = Ini->Read(ArrayTag, Default);
            }
        //  Restore section
        Ini->Section = std::string("");
        }
#endif

    virtual IIniSaveable * Clone() const
        {
        return new IniList<T>(*this);
        }

private:
    std::string        Section;
    std::string        TagBase;
    std::vector<T> &   Var;
    T                  Default;
    bool               AutoSize;
};


//==============================================================================
//  TEMPLATE Function ToIni() - Shortcut to make a IniSetting
//==============================================================================

template <class T>
IniSetting<T>  ToIni(const std::string & tag, T & var, T default_value)
{
    return IniSetting<T>(tag, var, default_value);
}

//==============================================================================
//  TEMPLATE Function ToIni() - Shortcut to make a IniSetting (using section)
//==============================================================================

template <class T>
IniSetting<T>  ToIni(const std::string & section, const std::string & tag, T & var, T default_value)
{
    return IniSetting<T>(section, tag, var, default_value);
}

//==============================================================================
//  TEMPLATE Function ToIni() - Shortcut to make a IniSetting
//==============================================================================

template <class T>
IniList<T>  ToIni(const std::string & sect, std::vector<T> & var, T default_value)
{
    return IniList<T>(sect, var, default_value);
}

//==============================================================================
//  TEMPLATE Function ToIni() - Shortcut to make a IniSetting
//==============================================================================

template <class T>
IniList<T>  ToIni(const std::string & sect, const std::string & tag_base,
                  std::vector<T> & var, T default_value)
{
    return IniList<T>(sect, tag_base, var, default_value);
}
#else
//==============================================================================
//  TEMPLATE Function WriteVarToIni() - INI writing
//==============================================================================
//  This template should allow writing into INI without extending IniProtocol class
//   by specializing

template <class T>
void  WriteVarToIni(IniProtocol * Ini, const std::string & tag, T & var)
{
    Ini->Write(tag, var);
}

//==============================================================================
//  TEMPLATE Function ReadVarFromIni() - INI Reading
//==============================================================================
//  This template should allow reading from INI

template <class T>
void  ReadVarFromIni(IniProtocol * Ini, const std::string & tag, T & var, T default_value)
{
    var = Ini->Read(tag, default_value);
}

//==============================================================================
//  TEMPLATE CLASS IniSetting - Implements INI Saving for a type
//==============================================================================

template <class T, class T_init>
class IniSetting : public IIniSaveable
{
public:
    IniSetting(const std::string & tag, T & var, T_init default_init)
        :  Section(""), Tag(tag), Var(var), Default(default_init)
        {
        }
    IniSetting(const std::string & section, const std::string & tag, T & var, T_init default_init)
        :  Section(section), Tag(tag), Var(var), Default(default_init)
        {
        }

    virtual void Save(IniProtocol * Ini) const
        {
        Ini->Section = Section;
        WriteVarToIni(Ini, Tag, Var);
        Ini->Section = std::string("");
        }
    virtual void Load(IniProtocol *  Ini) const
        {
        Ini->Section = Section;
        ReadVarFromIni(Ini, Tag, Var, Default);
        Ini->Section = std::string("");
        }
    virtual IIniSaveable * Clone() const
        {
        return new IniSetting<T, T_init>(*this);
        }

private:
    std::string  Section;
    std::string  Tag;
    T &          Var;
    T            Default;
};

//==============================================================================
//  TEMPLATE CLASS IniSetting - Vector specialization for IniSettings
//==============================================================================

template <class T, class T_init>
class IniSetting<std::vector<T>, T_init > : public IIniSaveable
{
public:
    IniSetting(const std::string & tag, std::vector<T> & var, std::vector<T> default_value)
        :  Section(""), Tag(tag), Var(var), Default(default_value)
        {
        }
    IniSetting(const std::string & section, const std::string & tag, std::vector<T> & var, std::vector<T> default_value)
        :  Section(section), Tag(tag), Var(var), Default(default_value)
        {
        }

    virtual void Save(IniProtocol * Ini) const
        {
        Ini->Section = Section;
		for (size_t i = 0; i < Var.size(); i++)
            WriteVarToIni(Ini, Tag + std::string("Field") + IntToString(i), Var[i]);
        Ini->Section = std::string("");
        }
    virtual void Load(IniProtocol *  Ini) const
        {
        Ini->Section = Section;
		for (size_t i = 0; i < Var.size(); i++)
            ReadVarFromIni(Ini, Tag + std::string("Field") + IntToString(i), Var[i], Default[i]);
        Ini->Section = std::string("");
        }
    virtual IIniSaveable * Clone() const
        {
        return new IniSetting<std::vector<T>, T_init >(*this);
        }

private:
    std::string			Section;
    std::string			Tag;
    std::vector<T> &    Var;
    std::vector<T>      Default;
};

//==============================================================================
//  TEMPLATE CLASS IniListV<T> - Implements INI Saving for a vector
//==============================================================================

template <class T, class T_init>
class IniListV : public IIniSaveable
{
public:
    IniListV(const std::string & section, std::vector<T> & var, T_init default_value, bool autosize=true)
        :  Section(section), TagBase("Idx_"), Var(var), Default(default_value), AutoSize(autosize)
        {
        }
    IniListV(const std::string & section, const std::string & tag_base, std::vector<T> & var, T_init default_value, bool autosize=true)
        :  Section(section), TagBase(tag_base), Var(var), Default(default_value), AutoSize(autosize)
        {
        }

    virtual void Save(IniProtocol * Ini) const
        {
        //  Read Data from section given by Tag
        Ini->Section = Section;

        size_t idx = 0;
        for (size_t i = 0 ; i<Var.size(); i++)
            {
            std::string ArrayTag(TagBase + Innovative::IntToString(idx++));
            WriteVarToIni(Ini, ArrayTag, Var[i]);
            //Ini->Write(ArrayTag, Var[i]);
            }
        //  Write ArraySize Tag if auto-sizing
        if (AutoSize)
            {
            //Ini->Write("ArraySize", Var.size());
            int size = Var.size();
            WriteVarToIni(Ini, "ArraySize", size);
            }
        //  Restore section
        Ini->Section = std::string("");
        }

    virtual void Load(IniProtocol *  Ini) const
        {
        //  Read Data from section given by Tag
        Ini->Section = Section;
        //  Check for override size tag parameter, resize if present
        //size_t OverrideSize=Ini->Read("ArraySize", 0);
        int OverrideSize;  // it makes little sense to have 4 billion lines in an INI file
        ReadVarFromIni(Ini, "ArraySize", OverrideSize, 0);
        if (OverrideSize!=0)
            Var.resize(OverrideSize);

        size_t idx = 0;
        for (size_t i = 0; i<Var.size(); i++)
            {
            std::string ArrayTag(TagBase + Innovative::IntToString(idx++));
            //Var[i] = Ini->Read(ArrayTag, Default);
            ReadVarFromIni(Ini, ArrayTag, Var[i], Default);
            }
        //  Restore section
        Ini->Section = std::string("");
        }

    virtual IIniSaveable * Clone() const
        {
        return new IniListV<T, T_init>(*this);
        }

private:
    std::string        Section;
    std::string        TagBase;
    std::vector<T> &   Var;
    T                  Default;
    bool               AutoSize;
};




//==============================================================================
//  TEMPLATE Function ToIni() - Shortcut to make a IniSetting
//==============================================================================

template <class T, class T_init>
IniListV<T, T_init>  ToIni(const std::string & sect, const std::string & tag_base,
                  std::vector<T> & var, T_init default_value)
{
    return IniListV<T, T_init>(sect, tag_base, var, default_value);
}

//==============================================================================
//  TEMPLATE Function ToIni() - Shortcut to make a IniSetting
//==============================================================================

template <class T, class T_init>
IniListV<T, T_init>  ToIni(const std::string & sect, std::vector<T> & var, T_init default_value)
{
    return IniListV<T, T_init>(sect, var, default_value);
}

//==============================================================================
//  TEMPLATE Function ToIni() - Shortcut to make a IniSetting
//==============================================================================

template <class T, class T_init>
IniSetting<T, T_init>  ToIni(const std::string & tag, T & var, T_init default_value)
{
    return IniSetting<T, T_init>(tag, var, default_value);
}

//==============================================================================
//  TEMPLATE Function ToIni() - Shortcut to make a IniSetting (using section)
//==============================================================================

template <class T, class T_init>
IniSetting<T, T_init>  ToIni(const std::string & section, const std::string & tag, T & var, T_init default_value)
{
    return IniSetting<T, T_init>(section, tag, var, default_value);
}

//==============================================================================
//  TEMPLATE Function ToIni() - Shortcut to make a IniSetting (using section)
//==============================================================================

template <class T>
IniSetting<std::vector<T>, std::vector<T> >  ToIni(const std::string & section, const std::string & tag, std::vector<T> & var, std::vector<T> default_vals)
{
    return IniSetting<std::vector<T>, std::vector<T> >(section, tag, var, default_vals);
}
#endif

//==============================================================================
//  JSON
//==============================================================================
//==============================================================================
//  TEMPLATE CLASS IniSetting - Vector specialization for IniSettings
//==============================================================================
template <class T>
class JiniSetting : public IIniSaveable
{
public:
    JiniSetting(const std::string & tag, T & var, T default_value)
        :  Section(""), Tag(tag), Var(var), Default(default_value)
        {
        }
    JiniSetting(const std::string & section, const std::string & tag, T & var, T default_value)
        :  Section(section), Tag(tag), Var(var), Default(default_value)
        {
            Tag = !Section.empty() ? Section + "." + Tag : Tag;
        }

    virtual void Save(IniProtocol * JFile) const
        {
        JFile->Write(Tag, Var);
        }
    virtual void Load(IniProtocol *  JFile) const
        {
        Var = JFile->Read(Tag, Default);
        }
    virtual IIniSaveable * Clone() const
        {
        return new JiniSetting<T>(*this);
        }

private:
    std::string  Section;
    std::string  Tag;
    T &          Var;
    T            Default;
};

template <class T>
class JiniSetting<std::vector<T> > : public IIniSaveable
{
public:
    JiniSetting(const std::string & tag, std::vector<T> & var, std::vector<T> default_value)
        :  Section(""), Tag(tag), Var(var), Default(default_value)
        {
        }
    JiniSetting(const std::string & section, const std::string & tag, std::vector<T> & var, std::vector<T> default_value)
        :  Section(section), Tag(tag), Var(var), Default(default_value)
        {
            Tag = !Section.empty() ? Section + "." + Tag : Tag;
        }

    virtual void Save(IniProtocol * JFile) const
        {
        JFile->Write(Section , Var);
        }

    virtual void Load(IniProtocol *  JFile) const
        {
        JFile->Read(Section, Var);
        }

    virtual IIniSaveable * Clone() const
        {
        return new JiniSetting<std::vector<T> >(*this);
        }

private:
    std::string			Section;
    std::string			Tag;
    std::vector<T> &    Var;
    std::vector<T>      Default;
};

template <class T>
class JiniList : public IIniSaveable
{
public:
    JiniList(const std::string & section, std::vector<T> & var, T default_value, bool autosize=true)
        :  Section(section), TagBase("Idx_"), Var(var), Default(default_value), AutoSize(autosize)
        {
        }
    JiniList(const std::string & section, const std::string & tag_base, std::vector<T> & var, T default_value, bool autosize=true)
        :  Section(section), TagBase(tag_base), Var(var), Default(default_value), AutoSize(autosize)
        {
        }

    virtual void Save(IniProtocol * JFile) const
        {
        JFile->Write(Section , Var);
        }

    virtual void Load(IniProtocol *  JFile) const
        {
        JFile->Read(Section, Var);
        }

    virtual IIniSaveable * Clone() const
        {
        return new JiniList<T>(*this);
        }

private:
    std::string        Section;
    std::string        TagBase;
    std::vector<T> &   Var;
    T                  Default;
    bool               AutoSize;
};


//==============================================================================
//  TEMPLATE Function ToIni() - Shortcut to make a IniSetting
//==============================================================================

template <class T>
JiniSetting<T>  ToJini(const std::string & tag, T & var, T default_value)
{
    return JiniSetting<T>(tag, var, default_value);
}

//==============================================================================
//  TEMPLATE Function ToIni() - Shortcut to make a IniSetting (using section)
//==============================================================================

template <class T>
JiniSetting<T>  ToJini(const std::string & section, const std::string & tag, T & var, T default_value)
{
    return JiniSetting<T>(section, tag, var, default_value);
}

//==============================================================================
//  TEMPLATE Function ToIni() - Shortcut to make a IniSetting
//==============================================================================

template <class T>
JiniList<T>  ToJini(const std::string & sect, std::vector<T> & var, T default_value)
{
    return JiniList<T>(sect, var, default_value);
}

//==============================================================================
//  TEMPLATE Function ToIni() - Shortcut to make a IniSetting
//==============================================================================

template <class T>
JiniList<T>  ToJini(const std::string & sect, const std::string & tag_base, std::vector<T> & var, T default_value)
{
    return JiniList<T>(sect, tag_base, var, default_value);
}


#ifdef __CLR_VER
#pragma managed(pop)
#endif
} // namespace Innovative

#endif
