//---------------------------------------------------------------------------

#include "McImageChecker_App.h"
#include <DataLogger_Mb.h>
#include <BinView_Mb.h>

//---------------------------------------------------------------------------

namespace Innovative
{

#ifdef _MSC_VER
    #pragma warning(push)
    #pragma warning(disable : 4355)
#endif

//==============================================================================
//  CLASS IImageCheckerTester  --  Interface to create a generated pattern
//==============================================================================
//------------------------------------------------------------------------------
//  IImageCheckerTester::FullCompare() -- compare entire buffer
//------------------------------------------------------------------------------

bool  IImageCheckerTester::FullCompare(McImage& test, McBuffer& should_be)
{
    if (test.SizeInInts() != should_be.SizeInInts())
        return false;

    AccessDatagram<unsigned int> TestDG(test);
	UIntegerDG ShouldBeDG(should_be);

    for (unsigned int i=0; i<TestDG.size(); ++i)
        if (TestDG[i] != ShouldBeDG[i])
            return false;

    return true;
}

//------------------------------------------------------------------------------
//  IImageCheckerTester::QuickCompare() -- compare ends of a buffer
//------------------------------------------------------------------------------
//  Test continuity (missing data) rather than full accuracy

bool  IImageCheckerTester::QuickCompare(McImage& test, McBuffer& should_be)
{
    if (test.SizeInInts() != should_be.SizeInInts())
        return false;

    AccessDatagram<unsigned int> TestDG(test);
	UIntegerDG ShouldBeDG(should_be);

    if (TestDG[0] != ShouldBeDG[0])
        return false;
    else if (TestDG[TestDG.size()-1] != ShouldBeDG[ShouldBeDG.size()-1])
        return false;
    else
        return true;
}

//==============================================================================
//  CLASS LogicRampsImageTester  --  Keep
//==============================================================================
//------------------------------------------------------------------------------
//  LogicRampsImageTester::TestInit()
//------------------------------------------------------------------------------

void  LogicRampsImageTester::TestInit()
{
    // reset test ramp indexes
    for (size_t i=0; i<Ramps.size(); i++)
        Ramps[i]=0;
}

//------------------------------------------------------------------------------
//  LogicRampsImageTester::RecvPatternGenerate() - return the pattern buffer
//------------------------------------------------------------------------------

void  LogicRampsImageTester::RecvPatternGenerate(unsigned int channel, McBuffer& out_buffer)
{
    unsigned int BadRamp(0xbaddeed);
    unsigned int & RampIdx = (channel>=8) ? BadRamp : Ramps[channel];

    // fill this buffer with the incrementing channel ramp
    UIntegerDG OutDG(out_buffer);

    McHeaderDatagram OutHeader(out_buffer);
    OutHeader.ChannelId(channel);
    OutHeader.DataSize(OutDG.SizeInInts());

    for (unsigned int i=0; i<OutDG.size(); ++i)
        OutDG[i] = (((RampIdx++)&0x0fffffff) | channel<<28);
}


//==============================================================================
//  CLASS McImageChecker  --  Check Image data against a generated pattern
//==============================================================================
//------------------------------------------------------------------------------
//  constructor of class McImageChecker()
//------------------------------------------------------------------------------

McImageChecker::McImageChecker(TestModes mode)
    : mMode(mode), mMaxImages(20), mImagesWithErrors(0), CustomTest(0)
{
    SwitchMode();
}

//------------------------------------------------------------------------------
//  McImageChecker::Init()
//------------------------------------------------------------------------------

void  McImageChecker::Init()
{
    mImagesWithErrors = 0;
    ErrorImages.clear();
    ShouldBeImages.clear();

    CurrentTest->TestInit();
}

//------------------------------------------------------------------------------
//  McImageChecker::Check() -- Check Image against 'test pattern'
//------------------------------------------------------------------------------

void  McImageChecker::Check(Innovative::McImage & in)
{
    McBuffer should_be( in.SizeInInts() );
    //
    //  Generate pattern
    McHeaderDatagram HeaderDG(in);
    CurrentTest->RecvPatternGenerate(HeaderDG.ChannelId(), should_be);

    bool error = ! CurrentTest->Compare(in, should_be);
    //
    //  Error Image processing
    if (error || mKeepFirst)
        {
        // count this as an error...
        mImagesWithErrors++;

        if (ErrorImages.size() >= mMaxImages)
            return;
        //
        //  Save Images in store
        McBuffer in_buf(in.SizeInInts());
        in.CopyTo(in_buf);
        ErrorImages.push_back(in_buf);
        ShouldBeImages.push_back(should_be);
        }
}

//------------------------------------------------------------------------------
//  McImageChecker::SwitchMode() -- Switch current operating mode
//------------------------------------------------------------------------------

void  McImageChecker::SwitchMode()
{
    switch (mMode)
        {
        default:
        case tmLogicRampsTest:
            CurrentTest = &LogicRampsTest;
            break;
        case tmQuickLogicRampsTest:
            CurrentTest = &QuickLogicRampsTest;
            break;
        case tmCustomTest:
            CurrentTest = CustomTest;
            break;
        }
}

//------------------------------------------------------------------------------
//  McImageChecker::LogToDisk() -- Log Image data to disk
//------------------------------------------------------------------------------

void  McImageChecker::LogToDisk(const std::string & file_name)
{
    DataLogger   Logger;
    BinView      Bv;

    Logger.FileName(file_name);
    Bv.BinFile(Logger.FileName());

    Logger.Start();
    // Log Error Packets
    for (unsigned int i=0; i<ErrorImages.size(); i++)
        {
        //  Log Error Packets
        McBuffer & EbOrig(ErrorImages[i]);
        Buffer Eb;
	    PacketBufferHeader EbHeader(Eb);
        EbHeader.PeripheralId(0x0);                // Error Images header = 0
        EbHeader.DataSize(EbOrig.SizeInInts());
        Eb.SwapData(EbOrig);
        Logger.LogWithHeader(Eb);
        Eb.SwapData(EbOrig);

        //  Log "Should Be" Packets
        McBuffer & SbOrig(ShouldBeImages[i]);
        Buffer Sb;
        PacketBufferHeader SbHeader(Sb);
        SbHeader.PeripheralId(0x1);
        SbHeader.DataSize(SbOrig.SizeInInts());
        Sb.SwapData(SbOrig);
        Logger.LogWithHeader(Sb);    // Should be  header = 1
        Sb.SwapData(SbOrig);
        }

    Logger.Stop();

    if (ErrorImages.empty())
        return;

    // Init Bdd File
	Bv.Time().LowerLimit(0x80000001);
	Bv.Time().UpperLimit(0x7fffffff);
	Bv.Time().Break(1);

    McBuffer & Eb_0(ErrorImages[0]);
//    PacketBufferHeader Eb0_Header(Eb_0);
    unsigned int PacketDataSize = static_cast<unsigned int>(Eb_0.SizeInInts());
    const unsigned int PidCount = 2;
	Bv.Time().SeamSize(PidCount*PacketDataSize*sizeof(int));
    Bv.Time().AnnotateSeams(true);

    Bv.Fft().LowerLimit(-150);
    Bv.Fft().ScaleYType(BinViewOptions::FftOptions::sLog);
    Bv.Fft().ScaleXType(BinViewOptions::FftOptions::sLog);
    Bv.Fft().CriteriaThreshold(75);
    Bv.Fft().Window(BinViewOptions::FftOptions::wBlackmann);
    Bv.Fft().Average(1);

    Bv.Text().DataFormat("%11.3x");
    Bv.System().UpdateRate(BinViewOptions::SystemOptions::ms1000);
    Bv.Leap(PacketDataSize);
    Bv.SignificantBits(32);
    Bv.Polarity(BinView::pSigned);
    Bv.DataSpan(300);
    Bv.DataType(BinView::tInt);
    Bv.Units("mV");
    Bv.Samples(1000);
    Bv.Channels(2);
    Bv.Devices(2);
    Bv.System().Source(BinViewOptions::SystemOptions::sPacketFile);
    Bv.NullDataValue(.12345e-19f);
    float rate = static_cast<float>(100e6/1000.);
    Bv.SampleRate(rate);
    Bv.InputSpan(2000);
    Bv.ScalingEnabled(false);

    Bv.Save();

}

//==============================================================================
//  CLASS IMcBufferCheckerTester  --  Interface to create a generated pattern
//==============================================================================
//------------------------------------------------------------------------------
//  IMcBufferCheckerTester::FullCompare() -- compare entire buffer
//------------------------------------------------------------------------------

bool  IMcBufferCheckerTester::FullCompare(McBuffer& test, McBuffer& should_be)
{
    if (test.SizeInInts() != should_be.SizeInInts())
        return false;

    AccessDatagram<unsigned int> TestDG(test);
    UIntegerDG ShouldBeDG(should_be);

    for (unsigned int i=0; i<TestDG.size(); ++i)
        if (TestDG[i] != ShouldBeDG[i])
            return false;

    return true;
}

//------------------------------------------------------------------------------
//  IMcBufferCheckerTester::QuickCompare() -- compare ends of a buffer
//------------------------------------------------------------------------------
//  Test continuity (missing data) rather than full accuracy

bool  IMcBufferCheckerTester::QuickCompare(McBuffer& test, McBuffer& should_be)
{
    if (test.SizeInInts() != should_be.SizeInInts())
        return false;

    AccessDatagram<unsigned int> TestDG(test);
    UIntegerDG ShouldBeDG(should_be);

    if (TestDG[0] != ShouldBeDG[0])
        return false;
    else if (TestDG[TestDG.size()-1] != ShouldBeDG[ShouldBeDG.size()-1])
        return false;
    else
        return true;
}

//==============================================================================
//  CLASS LogicRampsMcBufferTester  --  Keep
//==============================================================================
//------------------------------------------------------------------------------
//  LogicRampsMcBufferTester::TestInit()
//------------------------------------------------------------------------------

void  LogicRampsMcBufferTester::TestInit()
{
    // reset test ramp indexes
    for (size_t i=0; i<Ramps.size(); i++)
        Ramps[i]=0;
    for (size_t i=0; i<TxRamps.size(); i++)
        TxRamps[i]=0;
}

//------------------------------------------------------------------------------
//  LogicRampsMcBufferTester::RecvPatternGenerate() - return the pattern buffer
//------------------------------------------------------------------------------

void  LogicRampsMcBufferTester::RecvPatternGenerate(unsigned int channel, McBuffer& out_buffer)
{
    unsigned int BadRamp(0xbaddeed);
    unsigned int & RampIdx = (channel>=8) ? BadRamp : Ramps[channel];

    // fill this buffer with the incrementing channel ramp
    UIntegerDG OutDG(out_buffer);

    McHeaderDatagram OutHeader(out_buffer);
    OutHeader.ChannelId(channel);
    OutHeader.DataSize(OutDG.SizeInInts());

    for (unsigned int i=0; i<OutDG.size(); ++i)
        OutDG[i] = (((RampIdx++)&0x0fffffff) | channel<<28);
}

//------------------------------------------------------------------------------
//  LogicRampsMcBufferTester::XmitPatternGenerate() - return the pattern buffer
//------------------------------------------------------------------------------

void  LogicRampsMcBufferTester::XmitPatternGenerate(unsigned int channel, McBuffer& out_buffer)
{
    unsigned int BadRamp(0xbaddeed);
    unsigned int & RampIdx = (channel>=8) ? BadRamp : TxRamps[channel];

    // fill this buffer with the incrementing channel ramp
    UIntegerDG OutDG(out_buffer);

    McHeaderDatagram OutHeader(out_buffer);
    OutHeader.ChannelId(channel);
    OutHeader.DataSize(OutDG.SizeInInts());

    for (unsigned int i=0; i<OutDG.size(); ++i)
        OutDG[i] = (((RampIdx++)&0x0fffffff) | channel<<28);
}

//==============================================================================
//  CLASS McBufferChecker  --  Check McBuffer data against a generated pattern
//==============================================================================
//------------------------------------------------------------------------------
//  constructor of class McBufferChecker()
//------------------------------------------------------------------------------

McBufferChecker::McBufferChecker(TestModes mode)
    : mMode(mode), mMaxImages(20), mImagesWithErrors(0), CustomTest(0)
{
    SwitchMode();
}

//------------------------------------------------------------------------------
//  McBufferChecker::Init()
//------------------------------------------------------------------------------

void  McBufferChecker::Init()
{
    mImagesWithErrors = 0;
    ErrorImages.clear();
    ShouldBeImages.clear();

    CurrentTest->TestInit();
}

//------------------------------------------------------------------------------
//  McBufferChecker::Check() -- Check Image against 'test pattern'
//------------------------------------------------------------------------------

void  McBufferChecker::Check(Innovative::McBuffer & in)
{
    McBuffer should_be( in.SizeInInts() );
    //
    //  Generate pattern
    McHeaderDatagram HeaderDG(in);
    CurrentTest->RecvPatternGenerate(HeaderDG.ChannelId(), should_be);

    bool error = ! CurrentTest->Compare(in, should_be);
    //
    //  Error Image processing
    if (error || mKeepFirst)
        {
        // count this as an error...
        mImagesWithErrors++;

        if (ErrorImages.size() >= mMaxImages)
            return;
        //
        //  Save Images in store
        ErrorImages.push_back(in);
        ShouldBeImages.push_back(should_be);
        }
}

//------------------------------------------------------------------------------
//  McBufferChecker::SwitchMode() -- Switch current operating mode
//------------------------------------------------------------------------------

void  McBufferChecker::SwitchMode()
{
    switch (mMode)
        {
        default:
        case tmLogicRampsTest:
            CurrentTest = &LogicRampsTest;
            break;
        case tmQuickLogicRampsTest:
            CurrentTest = &QuickLogicRampsTest;
            break;
        case tmCustomTest:
            CurrentTest = CustomTest;
            break;
        }
}

//------------------------------------------------------------------------------
//  McBufferChecker::Fill() -- Fill block from enerator
//------------------------------------------------------------------------------

void  McBufferChecker::Fill(unsigned int channel, Innovative::McBuffer & out)
{
    //
    //  Generate pattern
    CurrentTest->XmitPatternGenerate(channel, out);
}

//------------------------------------------------------------------------------
//  McBufferChecker::LogToDisk() -- Log Image data to disk
//------------------------------------------------------------------------------

void  McBufferChecker::LogToDisk(const std::string & file_name)
{
    DataLogger   Logger;
    BinView      Bv;

    Logger.FileName(file_name);
    Bv.BinFile(Logger.FileName());

    Logger.Start();
    // Log Error Packets
    for (unsigned int i=0; i<ErrorImages.size(); i++)
        {
        //  Log Error Packets
        McBuffer & EbOrig(ErrorImages[i]);
        Buffer Eb;
        PacketBufferHeader EbHeader(Eb);
        EbHeader.PeripheralId(0x0);                // Error Images header = 0
        EbHeader.DataSize(EbOrig.SizeInInts());
        Eb.SwapData(EbOrig);
        Logger.LogWithHeader(Eb);
        Eb.SwapData(EbOrig);

        //  Log "Should Be" Packets
        McBuffer & SbOrig(ShouldBeImages[i]);
        Buffer Sb;
        PacketBufferHeader SbHeader(Sb);
        SbHeader.PeripheralId(0x1);
        SbHeader.DataSize(SbOrig.SizeInInts());
        Sb.SwapData(SbOrig);
        Logger.LogWithHeader(Sb);    // Should be  header = 1
        Sb.SwapData(SbOrig);
        }

    Logger.Stop();

    if (ErrorImages.empty())
        return;

    // Init Bdd File
    Bv.Time().LowerLimit(0x80000001);
    Bv.Time().UpperLimit(0x7fffffff);
    Bv.Time().Break(1);

    McBuffer & Eb_0(ErrorImages[0]);
//    PacketBufferHeader Eb0_Header(Eb_0);
    unsigned int PacketDataSize = static_cast<unsigned int>(Eb_0.SizeInInts());
    const unsigned int PidCount = 2;
    Bv.Time().SeamSize(PidCount*PacketDataSize*sizeof(int));
    Bv.Time().AnnotateSeams(true);

    Bv.Fft().LowerLimit(-150);
    Bv.Fft().ScaleYType(BinViewOptions::FftOptions::sLog);
    Bv.Fft().ScaleXType(BinViewOptions::FftOptions::sLog);
    Bv.Fft().CriteriaThreshold(75);
    Bv.Fft().Window(BinViewOptions::FftOptions::wBlackmann);
    Bv.Fft().Average(1);

    Bv.Text().DataFormat("%11.3x");
    Bv.System().UpdateRate(BinViewOptions::SystemOptions::ms1000);
    Bv.Leap(PacketDataSize);
    Bv.SignificantBits(32);
    Bv.Polarity(BinView::pSigned);
    Bv.DataSpan(300);
    Bv.DataType(BinView::tInt);
    Bv.Units("mV");
    Bv.Samples(1000);
    Bv.Channels(2);
    Bv.Devices(2);
    Bv.System().Source(BinViewOptions::SystemOptions::sPacketFile);
    Bv.NullDataValue(.12345e-19f);
    float rate = static_cast<float>(100e6/1000.);
    Bv.SampleRate(rate);
    Bv.InputSpan(2000);
    Bv.ScalingEnabled(false);

    Bv.Save();

}

#ifdef _MSC_VER
    #pragma warning(pop)
#endif


}  // namespace Innovative

