//---------------------------------------------------------------------------
// Oscilloscope_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.

#ifndef Oscilloscope_AppH
#define Oscilloscope_AppH
//---------------------------------------------------------------------------

#include <NonCopyable_Mb.h>
#include <ChannelInfo_Mb.h>
#include <Buffer_Mb.h>
#include <string>
#include <Thread_Mb.h>
#include <Event_Mb.h>

namespace Innovative
{

class OscilloscopeDllFunctions;
//===========================================================================
//  CLASS Oscilloscope  --  Oscilloscope DLL access
//===========================================================================

class Oscilloscope : public Noncopyable
{
    friend class OscilloscopeDllFunctions;

    class Control
    {
    public:
        Control(Oscilloscope & scope)
            : Owner(scope)  {  }

        Oscilloscope &  Owner;
        int             Handle()
            {  return Owner.Handle;  }
        int             Handle() const
            {  return Owner.Handle;  }
        OscilloscopeDllFunctions &  Dll() const
            {  return *Owner.Dll;  }
        OscilloscopeDllFunctions &  Dll()
            {  return *Owner.Dll;  }
    };

    // Grouped trigger functions
    class TriggerControl : private Control
    {
        friend class Oscilloscope;
        TriggerControl(Oscilloscope & scope)
            : Control(scope)  {  }

    public:
        enum IISource { tsBeam0, tsBeam1, tsBeam2, tsExternal };
        enum IIEdge { teNone, teRising, teFalling};

        double          Level(int beam) const;
        bool            Level(int beam, double value);

        IISource        Source() const;
        bool            Source(IISource src);

        IIEdge          Edge() const;
        bool            Edge(IIEdge type);

        int             Holdoff() const;
        bool            Holdoff(int samples);
    };

    // Grouped amplifier functions
    class AmpControl : private Control
    {
        friend class Oscilloscope;
        AmpControl(Oscilloscope & scope)
            : Control(scope)
            {  FGain[0] = FGain[1] = FGain[2] = 1.0;  }

    public:
        double          Gain(int beam) const;
        void            Gain(int beam, double value);

        double          Scale(int beam) const;
        bool            Scale(int beam, double value);

        double          Offset(int beam) const;
        bool            Offset(int beam, double value);

        void            AutoScale();

    private:
        double          FGain[3];
    };

    // Grouped form display functions
    class FormControl : private Control
    {
        friend class Oscilloscope;
        FormControl(Oscilloscope & scope)
            : Control(scope)  {  }

    public:
        // Methods
        bool            Show();
        bool            Hide();
        int             Left() const;
        int             Top() const;
        int             Width() const;
        int             Height() const;
        int             ScreenWidth() const;
        int             ScreenHeight() const;

        bool            Position(int left, int top);
        bool            Size(int width, int height);
        bool            Caption(const std::string & value);

        // Properties
        bool            PanelVisible() const;
        bool            PanelVisible(bool state);
    };

    // Grouped pixel functions
    class CellControl : private Control
    {
        friend class Oscilloscope;
    public:
        double          SampleSize() const;
        bool            SampleSize(double value);

        int             PixelSize() const;
        bool            PixelSize(int size);

    private:
        CellControl(Oscilloscope & scope)
            : Control(scope)  {  }
    };

public:
    friend class Control;
    friend class TriggerControl;
    friend class AmpControl;
    friend class FormControl;

	// Ctors
	Oscilloscope();
	Oscilloscope(const std::string & file);
	Oscilloscope(const std::string & file, const std::string & section);
    // Dtor
    virtual ~Oscilloscope();

    // Tier 1 Methods
    void            Format(int channels, int bits, bool unipolar=false);

    bool            Plot(double * data, bool external=false);
    void            Plot(Buffer & buffer);
    void            Plot(Buffer & buffer, int samples);

    bool            Update();

    // Tier 2 Methods
    bool            CleanBuffers();
    int             MaxInstances() const;
    int             Instance() const;
    int             MaxBeams() const;

    // Properties
    bool            Enable(int ch) const;
    void            Enable(int ch, bool state);

    // Aggregate members
    TriggerControl  Trigger;
    AmpControl      Amp;
    FormControl     Form;
    CellControl     Cell;

protected:
    // Data
    static class OscilloscopeDllFunctions * Dll;
    static unsigned int     Instances;
    int                     Handle;
    UnpairedChannelInfo *   Info;
    int                     FMaxBeams;

    // Methods
    void    Create(const std::string & file, const std::string & section);

    // Plot samples for all active beams.
    template<typename T>
    void plot(Buffer & buffer, int samples)
        {
        int channels = Info->Channels();

        int bits = sizeof(T)*8;
        int shift = bits-Info->Bits();
        T* wave = reinterpret_cast<T*>(buffer.Data());
        for (int n = 0; n < samples; ++n)
            {
            int beam = 0;
            double data[] = {0., 0., 0.};
            for (int ch = 0; ch < Info->Channels(); ++ch)
                {
                if (!Info->Enabled(ch))
                    continue;

                unsigned int idx = n*channels+ch;
                T sample = static_cast<T>(((wave[idx] << shift) >> shift) * Amp.Gain(beam));

                data[beam++] = sample;

                if (beam > 2)
                    break;
                }
            Plot(data);
            }
        }
};


}   // namespace

#endif
