// StaticWaveform_App.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2007 Innovative Integration Corporation.
//  All Rights Reserved.
//

#ifndef StaticWaveform_AppH
#define StaticWaveform_AppH

#include <ProcessEvents_Mb.h>
#include <Buffer_Mb.h>
#include <BufferDatagrams_Mb.h>
#include <BinView_Mb.h>
#include <VitaPacketStream_Mb.h>
#include <DataPlayer_Mb.h>

// Forward declaration
namespace Poco {
    class Random;
}

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//==============================================================================
//  CLASS  WaveformFile
//==============================================================================

class WaveformFile
{
public:
    WaveformFile();


    void        SampleRate(double value)
        {  FSampleRate=value;  }
    double      SampleRate() const
        {  return FSampleRate;  }

    void        Format(int channels, int bits, int samples, bool unipolar=false)
                    {
                    FChannels = channels; FBits=bits;
                    FSamples=samples;  FUnipolar=unipolar;
                    }

    void        Load(const std::string & filename, Buffer & buffer);
    void        Save(const std::string & filename, const Buffer & buffer);
    void        Show();
    void        Hide();
    void        Load_noResize(const std::string & filename, Buffer & buffer);

private:
    unsigned int    FSamples;
    double          FSampleRate;
    unsigned int    FBits;
    unsigned int    FChannels;
    bool            FUnipolar;

    BinView         Bv;
};

//==============================================================================
//  CLASS  VitaWaveformFile
//==============================================================================

class VitaWaveformFile
{
public:
    VitaWaveformFile();


    void        SampleRate(double value)
        {  FSampleRate=value;  }
    double      SampleRate() const
        {  return FSampleRate;  }

    void        Format(int devices, int channels, int bits, int samples, bool unipolar=false)
                    {
                    FDevices = devices;
                    FChannels = channels; FBits=bits;
                    FSamples=samples;  FUnipolar=unipolar;
                    }

    void        Load(const std::string & filename, VeloBuffer & buffer);
    void        Save(const std::string & filename, const VeloBuffer & buffer);
    void        Save(const std::string & filename, std::vector<VeloBuffer> & buffer_list);
    void        Show();
    void        Hide();

private:
    unsigned int    FDevices;
    unsigned int    FSamples;
    double          FSampleRate;
	unsigned int    FBits;
    unsigned int    FChannels;
    bool            FUnipolar;

    BinView         Bv;

    void  MakeBddFile(const std::string & filename);
};

//==============================================================================
//  CLASS  StaticWaveform
//==============================================================================

class StaticWaveform
{
public:
	enum IIType {wtTriangle, wtSine, wtSquare, wtRamp, wtFile, wtDc, wtRandom };

	// Ctor
	StaticWaveform();
	~StaticWaveform();

	// Methods
	void        Format(int channels, int bits, int samples);
	void        Resize(Buffer & data);

	void        Zero(int ch);
	void        Zero();

	bool        Accumulate(int ch);
	bool        Accumulate();

	void        Apply(Buffer & data);

	// Properties
	void        Type(IIType type)
		{  FType = type;  }
	IIType      Type() const
		{  return FType;  }

	void        SampleRate(double value)
		{  FSampleRate=value;  }
	double      SampleRate() const
		{  return FSampleRate;  }

	void        Frequency(double value)
		{  FFrequency=value;  }
	double      Frequency() const
		{  return FFrequency;  }
	double      FrequencyActual() const
		{  return FFrequencyActual;  }

	void        Amplitude(double percent)
		{  FAmplitude=percent;  }
	double      Amplitude() const
		{  return FAmplitude;  }

	void        Offset(double percent)
		{  FOffset=percent;  }
	double      Offset() const
		{  return FOffset;  }

	void        Phase(double value)
		{  FPhase=value;  }
	double      Phase() const
		{  return FPhase;  }

    void          Mask(unsigned int value)
        {  FMask=value;  }
    unsigned int  Mask() const
        {  return FMask;  }

	OpenWire::ThunkedEventHandler<Innovative::ProcessStatusEvent> OnMessage;

protected:
	// Fields
	unsigned int    FSamples;
	double          FSampleRate;
	double          FFrequency;
	double          FFrequencyActual;
	double          FAmplitude;
	double          FOffset;
	double          FPhase;
    unsigned int 	FMask;
	unsigned int    FBits;
	bool            FUnipolar;
	unsigned int    FChannels;
	IIType          FType;

	// Data
    Poco::Random   *Generator;
	Buffer          Data;

    bool Normalize(int ch);
    void Copy(Buffer & data, int ch);
private:
    // No copy or assignment
    StaticWaveform(const StaticWaveform &);
    StaticWaveform &operator=(const StaticWaveform &);
};

//==============================================================================
//  CLASS DualWaveGenerator
//==============================================================================

class DualWaveGenerator
{
public:
    StaticWaveform   Gen;

    DualWaveGenerator()
        :  FAmplitude(.95)
        { }

    bool        SingleWave(Buffer & data, int ch);
    bool        SingleWave(Buffer & data);

    bool        DualWave(Buffer & data, int ch);
    bool        DualWave(Buffer & data);

    void        SecondaryFrequency(double value)
        {  FSecondaryFrequency = value;  }
    double      SecondaryFrequency() const
        {  return FSecondaryFrequency;  }
    void        Amplitude(double percent)
        {  FAmplitude=percent;  }
    double       Amplitude() const
        {  return FAmplitude;  }

private:
    double                       FAmplitude;
    double                       FSecondaryFrequency;
};

//==============================================================================
//  CLASS WaveBuilder -- Class to manage Wave Building Setup
//==============================================================================

class WaveBuilder
{
public:
    WaveBuilder() : ConversionFactor(1000000)
    {}

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS WaveBuilderSettings
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    struct WaveBuilderSettings
    {
        std::string     WaveFile;
        int             WaveType;
        float           WaveformFrequency;
        float           WaveformAmplitude;
        bool            TwoToneMode;
        float           TwoToneFrequency;
        bool            SingleChannelMode;
        int             SingleChannelChannel;
    };

    WaveBuilderSettings  Settings;
    //
    //  Properties
    void        SampleRate(double value)
        {  FSampleRate=value;  }
    double      SampleRate() const
        {  return FSampleRate;  }

    void        Format(int channels, int bits, int samples)
                    {  FChannels = channels; FBits=bits; FSamples=samples;  }
    void        UseKHz()
                { ConversionFactor = 1000; }

    //
    //  Methods
    void  Resize(Buffer& data);  // resize buffer to fit samples
    void  BuildWave(Buffer & Buffer, std::string tag = "");
    void  ZeroWave(Buffer & Buffer);

private:
    //
    //  Member Data
    unsigned int    FSamples;
    double          FSampleRate;
    unsigned int    FBits;
    unsigned int    FChannels;
    unsigned int    ConversionFactor;

    DualWaveGenerator   WaveGen;

};

//==============================================================================
//  CLASS VitaWaveBuilder -- Class to manage Wave Building for X6 Boards
//==============================================================================

class VitaWaveBuilder
{
public:
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS VWB_Settings
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    struct VWB_Settings
    {
        std::string     WaveFile;
        int             WaveType;
        float           WaveformFrequency;
        float           WaveformAmplitude;
        bool            TwoToneMode;
        float           TwoToneFrequency;
        bool            SingleChannelMode;
        int             SingleChannelChannel;
    };

    VWB_Settings  Settings;

    VitaWaveBuilder()
        :  FSampleRate(1e6), FMaxVitaSize(0xF000),
          ConversionFactor(1000000)
        {}
    //
    //  Properties
    void        SampleRate(double value)
        {  FSampleRate=value;  }
    double      SampleRate() const
        {  return FSampleRate;  }

    void    MaxVitaSize(size_t value)
                        {  FMaxVitaSize=value;  }
    size_t  MaxVitaSize() const
                        {  return FMaxVitaSize; }

    void        Format(const std::vector<int> & device_sids, int channels, int bits, int samples)
                    {
                    FDeviceSids = device_sids;  FChannels = channels;
                    FBits=bits; FSamples=samples;
                    }

    void        UseKHz()
                { ConversionFactor = 1000; }
    //
    //  Methods
    void  BuildWave(VeloBuffer & Buffer);

private:
    //
    //  Member Data
    unsigned int            FSamples;
    double                  FSampleRate;
    unsigned int            FBits;
    unsigned int            FChannels;
    std::vector<int>        FDeviceSids;
    size_t                  FMaxVitaSize;
    unsigned int            ConversionFactor;

    DualWaveGenerator    WaveGen;
    std::vector<Buffer>  Scratch;
    VeloBuffer           WaveformPacket;

    void    CreateScratchBuffers();
    void    LoadWaveFromFileList();
    void    GenerateWave();
    void    FillOutputWaveBuffer(VeloBuffer & OutputWaveform);

    size_t  CalculateScratchBufferSize();

    void    HandlePackedDataAvailable(VitaPacketPackerDataAvailable & event);
};

//==============================================================================
//  CLASS VitaWaveBuilderMP -- Class to manage Wave Building for X6 Boards
//==============================================================================
//  Creates Multiple Output Packets...

class VitaWaveBuilderMP
{
public:
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS VWB_Settings
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    struct VWB_Settings
    {
        std::string     WaveFile;
        int             WaveType;
        float           WaveformFrequency;
        float           WaveformAmplitude;
        bool            TwoToneMode;
        float           TwoToneFrequency;
        bool            SingleChannelMode;
        int             SingleChannelChannel;
    };

    VWB_Settings  Settings;

    VitaWaveBuilderMP()
        :  FSampleRate(1e6), FMaxVitaSize(0xF000),
          FMaxVeloSize(0x80000), ConversionFactor(1000000)
        {}
    //
    //  Properties
    void        SampleRate(double value)
        {  FSampleRate=value;  }
    double      SampleRate() const
        {  return FSampleRate;  }

    void    MaxVeloSize(size_t value)
                        {  FMaxVeloSize=value;  }
    size_t  MaxVeloSize() const
                        {  return FMaxVeloSize; }
    void    MaxVitaSize(size_t value)
                        {  FMaxVitaSize=value;  }
    size_t  MaxVitaSize() const
                        {  return FMaxVitaSize; }

    void        Format(const std::vector<int> & device_sids, int channels, int bits, int samples)
                    {
                    FDeviceSids = device_sids;  FChannels = channels;
                    FBits=bits; FSamples=samples;
                    }

    void        UseKHz()
                { ConversionFactor = 1000; }
    //
    //  Methods
    void  ClearList()  {  WF_PacketList.clear();  }
    void  BuildWave();
    std::vector<VeloBuffer> & List() {  return WF_PacketList;  }

private:
    //
    //  Member Data
    unsigned int            FSamples;
    double                  FSampleRate;
    unsigned int            FBits;
    unsigned int            FChannels;
    std::vector<int>        FDeviceSids;
    size_t                  FMaxVitaSize;
    size_t                  FMaxVeloSize;
    unsigned int            ConversionFactor;

    DualWaveGenerator        WaveGen;
    std::vector<Buffer>      Scratch;
    std::vector<VeloBuffer>  WF_PacketList;

    void    CreateScratchBuffers();
    void    LoadWaveFromFileList();
    void    GenerateWave();
    void    FillOutputWaveBuffer();

    size_t  CalculateScratchBufferSize();

    void    HandlePackedDataAvailable(VitaPacketPackerDataAvailable & event);
};

//==============================================================================
//  CLASS VitaWavePlayer -- Class to manage Wave Playing for X6 Boards
//==============================================================================

class VitaWavePlayer
{
public:

    enum PlayerMode { pmRepeat, pmZero };

    VitaWavePlayer()
        :  FMaxVitaSize(0xF000), FMaxVeloSize(0x10000), FCacheBufferSize(0x10000),
           FPlayerMode(pmZero), FPosition(0x0)
        {}
    ~VitaWavePlayer()
        {
            DeletePlayers();
        }
    //
    //  Properties
    void         WaveFile(const std::string & file)
                        {  FWaveFile = file;  }
    std::string  WaveFile() const
                        {  return FWaveFile;  }
    void    Mode(PlayerMode mode)
                        {  FPlayerMode=mode;  }
    PlayerMode  Mode() const
                        {  return FPlayerMode; }

    void    MaxVitaSize(size_t value)
                        {  FMaxVitaSize=value;  }
    size_t  MaxVitaSize() const
                        {  return FMaxVitaSize; }

    void    CacheBufferSize(size_t value)
                        {  FCacheBufferSize=value;  }
    size_t  CacheBufferSize() const
                        {  return FCacheBufferSize; }

    void    MaxVeloSize(size_t value)
                        {  FMaxVeloSize=value;  }
    size_t  MaxVeloSize() const
                        {  return FMaxVeloSize; }

    void    Sids(const std::vector<int> & sids)
        {  FDeviceSids = sids;  }

    std::vector<VeloBuffer> & Output()
               {  return OutputPackets;  }
    void  ClearOutput()
               {  OutputPackets.clear();  }

    ii64  Position()
        {   return FPosition;  }

    //
    //  Methods
    void  Start();
    void  Process();
    void  Stop();

private:
    //
    //  Member Data
    std::vector<int>        FDeviceSids;
    size_t                  FMaxVitaSize;
    size_t                  FMaxVeloSize;
    size_t                  FCacheBufferSize;
    PlayerMode              FPlayerMode;
    std::string             FWaveFile;
    ii64                    FPosition;

    std::vector<Buffer>        Scratch;
    std::vector<DataPlayer *>  DPlayer;
    std::vector<VeloBuffer>    OutputPackets;

    void    CreateScratchBuffers();
    void    FillScratchBuffers();
    void    FillOutputPacketList();

    void    MakePlayers(unsigned int count);
    void    OpenPlayers();
    void    ClosePlayers();
    void    DeletePlayers();

    void    HandlePackedDataAvailable(VitaPacketPackerDataAvailable & event);
};

#ifdef __CLR_VER
#pragma managed(pop)
#endif
} // namespace Innovative

#endif

