//  XU_OutputManager_App.cpp
//
//  Copyright 2016 Innovative Integration
//---------------------------------------------------------------------------

#include "XU_OutputManager_App.h"

namespace Innovative
{

//===========================================================================
//  CLASS XU_OutputManager -- Coordinate Output for XU output streams
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class ApplicationIo
//---------------------------------------------------------------------------

XU_OutputManager::XU_OutputManager(unsigned int channels)
    :  Mode(pmContinuousPlay)
{
    for (unsigned int i=0; i<channels; i++)
        {
        PerChannelArgs args;
        args.CacheSizeInts = 0x10000;
        PerArg.push_back(args);

        PerChannelInfo info;
        info.ChannelID = i;   // assume we use first N steam channels
        info.Player.Mode(DataPlayer::pmRepeat);
        PerCh.push_back(info);
        }
}

//---------------------------------------------------------------------------
//  XU_OutputManager::Start() - start all loggers
//---------------------------------------------------------------------------

void  XU_OutputManager::Start()
{
    if (Mode == pmContinuousPlay)
        {
        for (unsigned int i=0; i<PerCh.size(); i++)  // "play from file mode"
            {
            PerChannelInfo & info(PerCh[i]);

            info.Player.FileName(Settings(i).SourceFilename);

            info.PlayerOpened = info.Player.Start();

            }
        // Cache Buffer not used
        ReleaseCaches();
        return;
        }
    if (Mode == pmOneBlock)
        {
        for (unsigned int i=0; i<PerCh.size(); i++)
            {
            PerChannelInfo & info(PerCh[i]);

            info.Player.FileName(Settings(i).SourceFilename);

            info.PlayerOpened = info.Player.Start();
            }
        // Cache Buffer used
        AllocateCaches();
        // Fill Cache Buffers
        for (unsigned int i=0; i<PerCh.size(); i++)
            {
            PerChannelInfo & info(PerCh[i]);

            info.Player.Start();

            info.Player.Play(info.Cache);

            info.Player.Stop();
            }
        return;
        }
}

//---------------------------------------------------------------------------
//  XU_OutputManager::Stop() - stop all loggers
//---------------------------------------------------------------------------

void  XU_OutputManager::Stop()
{
    if (Mode == pmContinuousPlay)
        {
        for (unsigned int i=0; i<PerCh.size(); i++) // "play from file mode"
            {
            PerChannelInfo & info(PerCh[i]);

            info.Player.Stop();
            info.PlayerOpened = false;
            }
        return;
        }
    if (Mode == pmOneBlock)
        {
        // no need to do anything
        return;
        }
}

//---------------------------------------------------------------------------
//  XU_OutputManager::Play() - play into image
//---------------------------------------------------------------------------

void  XU_OutputManager::Play(unsigned int ch, McImage& Image)
{
    if (ch >= PerCh.size())
        return;

    PerChannelInfo & info(PerCh[ch]);
    if (Mode == pmContinuousPlay)
        {
        info.Player.Play(Image);
        }
    if (Mode == pmOneBlock)
        {
        size_t cache_remaining = info.Cache.SizeInInts() - info.CacheOffset;
        //
        //  Case 1:  Image larger than Cache Remaining
        if (Image.SizeInInts() > cache_remaining)
            TruncateImage(Image, cache_remaining);  // make image smaller
        //
        //  Case 2:  Image smaller (or equal) to Cache
        Image.CopyFrom(info.Cache, info.CacheOffset);
        //
        //  ...adjust offset
        info.CacheOffset += Image.SizeInInts();
        if (info.CacheOffset >= info.Cache.SizeInInts())
            info.CacheOffset = 0;

        return;
        }
}

//---------------------------------------------------------------------------
//  XU_OutputManager::Play() - play into image
//---------------------------------------------------------------------------

void  XU_OutputManager::Play(unsigned int ch, McBuffer& Buffer)
{
    if (ch >= PerCh.size())
        return;

    if (Mode == pmContinuousPlay)
        {
        PerChannelInfo & info(PerCh[ch]);

        info.Player.Play(Buffer);   // play from file into buffer
        }
    if (Mode == pmOneBlock)
        {
        PerChannelInfo & info(PerCh[ch]);
        // insta-copy cache into buffer - note header copied!
        Buffer = info.Cache;
        return;
        }
}

//---------------------------------------------------------------------------
//  XU_OutputManager::PlayerOpened() - check open success status
//---------------------------------------------------------------------------

bool  XU_OutputManager::PlayerOpened(unsigned int ch)
{
    if (ch >= PerCh.size())
        return false;

    PerChannelInfo & info(PerCh[ch]);
    return info.PlayerOpened;
}

//---------------------------------------------------------------------------
//  XU_OutputManager::AllPlayersOpened() - return true if all players ready
//---------------------------------------------------------------------------

bool  XU_OutputManager::AllPlayersOpened()
{
    bool result = true;
    for (unsigned int i=0; i<PerCh.size(); i++)
        {
        result = result && PlayerOpened(i);
        }
    return result;
}

//---------------------------------------------------------------------------
//  XU_OutputManager::AllocateCaches() - acquire
//---------------------------------------------------------------------------

void XU_OutputManager::AllocateCaches()
{
    for (unsigned int ch=0; ch<PerCh.size(); ch++)
        {
        PerChannelInfo & info(PerCh[ch]);
        info.CacheOffset = 0;

        size_t size = Settings(ch).CacheSizeInts;

        info.Cache.Resize(size);
        //  Correct header
        McHeaderDatagram Hdr(info.Cache);
        Hdr.ChannelId(ch);
        Hdr.DataSize(size);
        }
}

//---------------------------------------------------------------------------
//  XU_OutputManager::ReleaseCaches() - Free Cache memory
//---------------------------------------------------------------------------

void XU_OutputManager::ReleaseCaches()
{
    for (unsigned int i=0; i<PerCh.size(); i++)
        {
        PerChannelInfo & info(PerCh[i]);
        McBuffer  empty;

        info.Cache.SwapData(empty);
        }
}

//---------------------------------------------------------------------------
//  XU_OutputManager::TruncateImage() - Make image smaller to fit cache
//---------------------------------------------------------------------------

void XU_OutputManager::TruncateImage(McImage& Image, size_t size_ints)
{
    Image.Truncate(size_ints);
    //  Correct header size to match
    McHeaderDatagram Hdr(Image);
    Hdr.DataSize(size_ints);
}

}  // namespace Innovative
