// BaseEvents_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure 
//  agreement with Innovative Integration Corporation and may not be copied or 
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2006 Innovative Integration Corporation.
//  All Rights Reserved.
//

#ifndef BaseEvents_MbH
#define BaseEvents_MbH

#undef NULL
#define NULL 0

namespace OpenWire
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

class EventSyncPair;
class BaseEventHandler;

//=============================================================================
//  CLASS Event  --  
//=============================================================================

class Event
{
public:
    virtual ~Event() { }
};

//=============================================================================
//  CLASS NotifyEvent  --  
//=============================================================================

class NotifyEvent : public Event
{
public:
    NotifyEvent() { }
};

//=============================================================================
//  CLASS BaseEventPointer  --  
//=============================================================================

class BaseEventPointer
{
public:
    virtual ~BaseEventPointer()
        {  }
};

//=============================================================================
//  CLASS CommonEventPointer  --
//=============================================================================

template <class T_EVENT>
class CommonEventPointer : public BaseEventPointer
{
public:
    virtual void    Execute(T_EVENT & Event) = 0;

public:
	virtual CommonEventPointer * Clone( BaseEventHandler * owner ) = 0;
};

//=============================================================================
//  CLASS EventPointer  --  
//=============================================================================

template <class T_CLASS, class T_EVENT>
class  EventPointer : public CommonEventPointer<T_EVENT>
{
protected:
	BaseEventHandler * const     mOwner;
	T_CLASS * const	             Object;
    void (T_CLASS::*CallBackFunction)(T_EVENT & Event);

public:
    virtual void Execute(T_EVENT & Event)
	{
		if (Object)
			(Object->*CallBackFunction) (Event);
	}

public:
	virtual CommonEventPointer<T_EVENT> * Clone( BaseEventHandler * owner )
	{
		return new EventPointer<T_CLASS, T_EVENT>( owner, Object, CallBackFunction );
	}

public:
	EventPointer(BaseEventHandler * owner,
				 T_CLASS * Object,
				 void (T_CLASS:: * CallBackFunction)
				 (T_EVENT & Event))
		: mOwner(owner), Object(Object), CallBackFunction(CallBackFunction)
	{
	}

	~EventPointer()
	{
	}
};

//=============================================================================
//  CLASS GlobalEventPointer  --  
//=============================================================================

template <class T_EVENT>
class GlobalEventPointer : public CommonEventPointer<T_EVENT>
{
protected:
    void (*mCallBackFunction)(T_EVENT & Event);

public:
    virtual void Execute(T_EVENT & Event)
    {
        mCallBackFunction(Event);
    }

public:
    virtual CommonEventPointer<T_EVENT> * Clone( BaseEventHandler * owner )
    {
        return new GlobalEventPointer<T_EVENT>( mCallBackFunction );    
    }

public:
    GlobalEventPointer(void (*CallBackFunction)
                           (T_EVENT & Event)) 
        : mCallBackFunction(CallBackFunction)
    {
    }
};

//=============================================================================
//  CLASS BaseEventHandler  --
//=============================================================================

class BaseEventHandler
{
	friend class EventSyncPair;

public:
	virtual ~BaseEventHandler() { }

	virtual void    ClearEvent() = 0;
	virtual bool    Assigned() = 0;

protected:
	virtual void DirectExecute(Event *Event) = 0;
};

//=============================================================================
//  CLASS EventSynchronizerIntf  --
//=============================================================================

class EventSynchronizerIntf
{
public:
	virtual ~EventSynchronizerIntf() { }
	virtual void Execute( EventSyncPair * Event ) = 0;
};

//=============================================================================
//  CLASS EventSyncPair  --
//=============================================================================

class EventSyncPair
{
private:
	BaseEventHandler * const 	EventHandler;
	Event * const 				EventItem;

public:
	EventSyncPair( BaseEventHandler * event_handler, Event * event_item ) :
		EventHandler( event_handler ),
		EventItem( event_item )
	{
	}

	~EventSyncPair()
	{
		delete EventItem;
	}

public:
	void Execute()
	{
		EventHandler->DirectExecute( EventItem );
	}

};

//=============================================================================
//  CLASS StdEventHandler  --
//=============================================================================

template <class T_EVENT>
class StdEventHandler : public BaseEventHandler
{
protected:
	EventSynchronizerIntf *        mSynchronizer;
	CommonEventPointer<T_EVENT> *  mOnDispatch;

	void DirectExecute( Event *Event)
		{
		if(mOnDispatch)
			mOnDispatch->Execute( *(T_EVENT *)Event);
		}

public:
	StdEventHandler() : mSynchronizer( NULL ), mOnDispatch( NULL )
		{
		}

	StdEventHandler( const StdEventHandler<T_EVENT> & other ) : BaseEventHandler(other)
		{
		mSynchronizer = other.mSynchronizer;

		if( other.mOnDispatch )
			mOnDispatch = other.mOnDispatch->Clone( this );
		else
			mOnDispatch = NULL;
		}

	~StdEventHandler()
		{
		ClearEvent();
		}

	void Execute(T_EVENT & event)
		{
		if (mOnDispatch)
			{
			if ( mSynchronizer )
				mSynchronizer->Execute( new EventSyncPair( this, new T_EVENT( event )));
			else
				mOnDispatch->Execute( event );
			}
		}

	virtual void ClearEvent()
		{
            if(mOnDispatch)
            {
		        delete mOnDispatch;
		        mOnDispatch = NULL;
            }
		}

	virtual bool Assigned()
		{
		return (mOnDispatch != NULL);
		}

	template <class T_CLASS>
	void SetEvent(T_CLASS * Instance,
				  void (T_CLASS:: * CallBackFunction)
				  (T_EVENT & Event))
		{
		ClearEvent();

        
		mOnDispatch = new EventPointer<T_CLASS, T_EVENT>(this, Instance, CallBackFunction);

		#ifndef MALIBU_LIB
		if( ! mOnDispatch )
			Execute( *((T_EVENT *) 0 ) );  // Fixes a BCB 2006 linker bug!

		#endif

		}

	void SetEvent(void (*CallBackFunction) (T_EVENT & Event))
		{
		ClearEvent();

		mOnDispatch = new GlobalEventPointer<T_EVENT>(CallBackFunction);

		#ifndef MALIBU_LIB
		if( ! mOnDispatch )
			Execute( *((T_EVENT *) 0 ) );  // Fixes a BCB 2006 linker bug!

		#endif

		}

public:
	StdEventHandler<T_EVENT> & operator = ( const StdEventHandler<T_EVENT> & other )
	{
		mSynchronizer = other.mSynchronizer;

		ClearEvent();

		if( other.mOnDispatch )
			mOnDispatch = other.mOnDispatch->Clone( this );
		else
			mOnDispatch = NULL;

		return *this;
	}
};

#ifdef __CLR_VER
#pragma managed(pop)
#endif
} // OpenWire

#endif
