//  BaseboardCpuInterfaces_Mb.cpp
//
//  Contains standard implementation of Baseboard Cpus
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <BaseboardCpuInterfaces_Mb.h>

#include <vector>
#include <Exception_Mb.h>
#include <SystemSupport_Mb.h>
#include <Trace_Mb.h>

namespace Innovative
{

namespace
{
//
//  CPU needs a tiny boot application to be downloaded on Boot. This is the
//     code for this application...
unsigned int    StdMinimalBootApp[] =
{
    0x01800228,  0x0294a0fa,  0x0200002a,  0x0200ca6a,
    0x029002f6,  0x00000028,  0x0000ca68,  0x01800274,
    0x0ffffd10,  0x00008000,  0x00008000,  0x00008000,
    0x00008000,  0x00008000,  0x00008000,  0x00008000
};

const int  StdMinimalBootAppSize = sizeof(StdMinimalBootApp)/sizeof(unsigned int);

}  // end nameless namespace

//===========================================================================
//  CLASS IUsesStandardHpiCpu -- Standard Implementation of HPI Cpu
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class IUsesStandardHpiCpu
//---------------------------------------------------------------------------

IUsesStandardHpiCpu::IUsesStandardHpiCpu(IUsesHpiControl & hpic,
										 HpiEngineAddresses & HpiAddr,
                                         HpiCeSpaceValues & HpiCe )
    :  mHpiLoadVerify(true), mHpi(HpiAddr), HpiControl(hpic)
{
    //
    //  Load new HPI memory space values
    mHpi.LoadCeSpaceValues(HpiCe);

    Coff.Image.OnMemWrite.SetEvent(this, &IUsesStandardHpiCpu::HandleMemWrite);
    //  ...Forward Coff Image progress events
    Coff.Image.OnProgress.SetEvent(this, &IUsesStandardHpiCpu::HandleOnProgress);
    Coff.Image.OnComplete.SetEvent(this, &IUsesStandardHpiCpu::HandleOnComplete);
    Coff.Image.OnLoadMsg.SetEvent(this, &IUsesStandardHpiCpu::HandleOnLoadMsg);

}

//---------------------------------------------------------------------------
//  destructor for class IUsesStandardHpiCpu
//---------------------------------------------------------------------------

IUsesStandardHpiCpu::~IUsesStandardHpiCpu()
{
    Coff.Shutdown();
}

//---------------------------------------------------------------------------
//  IUsesStandardHpiCpu::Open() -- Open-time behavior...
//---------------------------------------------------------------------------

void  IUsesStandardHpiCpu::Open(ii32 * port_memory_addr)
{
    mHpi.Link( port_memory_addr );
}

//---------------------------------------------------------------------------
//  IUsesStandardHpiCpu::Cold()
//---------------------------------------------------------------------------

void  IUsesStandardHpiCpu::Cold()
{
    //  Toggle reset on DSP
    HpiControl.ResetBaseBoard();
	Sleep(100);
	Hpi().Reset();
    //
    //  Load EMIF settings for SDRAM
    Hpi().ConfigureTargetSdram();
}

//---------------------------------------------------------------------------
//  IUsesStandardHpiCpu::Boot() -- Boot the CPU
//---------------------------------------------------------------------------

void  IUsesStandardHpiCpu::Boot()
{
    Cold();

    //
    //  Poke test application into memory
    Hpi().WriteBlock((int*)StdMinimalBootApp, 0, StdMinimalBootAppSize);
    //
    //  Trigger HPI interrupt to DSP to start execution
    Hpi().Interrupt();
}

//---------------------------------------------------------------------------
//  IUsesStandardHpiCpu::DownloadCoff() -- Perform Coff Download
//---------------------------------------------------------------------------

void IUsesStandardHpiCpu::DownloadCoff(const std::string & Filename)
{
    ProcessCompletionEvent CanDownload(0);
    OnBeforeCoffLoad.Execute(CanDownload);
    if (CanDownload.Result)
        return;

    Cold();

    Coff.Load(Filename);
}

//---------------------------------------------------------------------------
//  IUsesStandardHpiCpu::CoffImage() -- Return Coff Object
//---------------------------------------------------------------------------

CoffFile &  IUsesStandardHpiCpu::CoffImage()
{
    return Coff.Image;
}

//---------------------------------------------------------------------------
//  IUsesStandardHpiCpu::DownloadCoff() -- Return Hpi Object
//---------------------------------------------------------------------------

HpiEngine &  IUsesStandardHpiCpu::Hpi()
{
    return mHpi;
}

//---------------------------------------------------------------------------
//  IUsesStandardHpiCpu::HandleMemWrite() --
//------------------------------------------------------------------------------

void IUsesStandardHpiCpu::HandleMemWrite(Innovative::MemWriteEvent & event)
{
    const int * src = reinterpret_cast<const int*>(event.Buffer);
    int size = event.Bytes/sizeof(int);

    std::vector<int> dst(size);

    const int MaxIterations = 10;
    int Iteration; 
    for (Iteration = 0; Iteration < MaxIterations; ++Iteration)
        {
        Hpi().WriteBlock(src, event.Addr, size);
        if (!mHpiLoadVerify)
            break;

        Hpi().ReadBlock(event.Addr, &dst[0], size);

        int errors = 0;
        for (int i = 0; i < size; ++i)
            errors += src[i] != dst[i];

        if (!errors)
            break;
        }

    if (!Iteration)
        return;

    if (Iteration < MaxIterations)
        Debug.Log("Recoverable HPI write failure!");
    else        
        throw MalibuException("Unrecoverable HPI write failure!");
}

//------------------------------------------------------------------------------
//  IUsesStandardHpiCpu::HandleOnProgress() --
//------------------------------------------------------------------------------

void IUsesStandardHpiCpu::HandleOnProgress(ProcessProgressEvent & event)
{
    OnCoffLoadProgress.Execute(event);
}

//------------------------------------------------------------------------------
//  IUsesStandardHpiCpu::HandleOnComplete() --
//------------------------------------------------------------------------------

void IUsesStandardHpiCpu::HandleOnComplete(ProcessCompletionEvent & event)
{
    //
    //  Trigger HPI interrupt to DSP to start execution
    Hpi().Interrupt();

    OnCoffLoadComplete.Execute(event);
}

//------------------------------------------------------------------------------
//  IUsesStandardHpiCpu::HandleOnLoadMsg() --
//------------------------------------------------------------------------------

void IUsesStandardHpiCpu::HandleOnLoadMsg(ProcessStatusEvent & event)
{
    OnCoffLoadMessage.Execute(event);
}

//===========================================================================
//  CLASS IUsesRestrictedHpiCpu -- User Interface to mix-in onboard CPU & Hpi
//  Suitable for HPI devices, such as C6711 which do not allow readback
//  during HPI load
//===========================================================================

//---------------------------------------------------------------------------
//  constructor for class IUsesRestrictedHpiCpu
//---------------------------------------------------------------------------

IUsesRestrictedHpiCpu::IUsesRestrictedHpiCpu(IUsesHpiControl & hpic,
    HpiEngineAddresses & HpiAddr, HpiCeSpaceValues & HpiCe)
        : IUsesStandardHpiCpu(hpic, HpiAddr, HpiCe)
{
    mHpiLoadVerify = false;
}


} // end namespace Innovative






