// BlockStream_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <BlockStream_Mb.h>
#include <StartStopThread_Mb.h>
#include <StopWatch_Mb.h>
#include <AccessDatagrams_Mb.h>
#include <algorithm>
#include <BufferHeader_Mb.h>
#include <SystemSupport_Mb.h>

namespace   Innovative
{

//===========================================================================
//  CLASS BlockStreamIBSQCreditor  -- Credit manager for IBSQueue
//===========================================================================

class BlockStreamIBSQCreditor : public IBSQueueCreditor
{
public:
    BlockStreamIBSQCreditor(BlockStream * owner, int stream_channel);

    virtual void  CreditStream(size_t cq_token);

private:
    //
    //  Member Data
    BlockStream * Owner;
    unsigned int  StreamChannel;

};

//---------------------------------------------------------------------------
//  constructor for class BlockStreamIBSQCreditor
//---------------------------------------------------------------------------

BlockStreamIBSQCreditor::BlockStreamIBSQCreditor(BlockStream * owner, int stream_channel)
    : Owner(owner), StreamChannel(stream_channel)
{
}

//---------------------------------------------------------------------------
//  BlockStreamIBSQCreditor::CreditStream() -- grant credit
//---------------------------------------------------------------------------

void  BlockStreamIBSQCreditor::CreditStream(size_t cq_token)
{
    //  token is packet size in words...
    size_t PacketSize = cq_token;
    Owner->BlkStreamDevice->GrantCredit(StreamChannel, PacketSize);
}

//===========================================================================
//  CLASS BlockStreamIoThread  -- I/O thread for BlockStreams
//===========================================================================
//  Threads handling the 'application' side of the buffer queues from the
//    VISR. They handle the interface between these queues and the app

class BlockStreamIoThread : public BasicSSThread
{
    friend class BlockStream;
    typedef BasicSSThread inherited;

public:
    // Ctor
    BlockStreamIoThread(BlockStream *owner, BlockStreamQueue & queue)
        : inherited("PktStreamIoThread"), Owner(owner), Queue(queue)
        {  Priority(tpHigher);  }

protected:
    class LoadBalance
    {
    public:
        // Ctor
        LoadBalance()
            : Then(0), Sleeps(0), Packets(0)
            {  }

        // Load balancing:  Reserve 10% of CPU for foreground processing
        void Balance()
        {
            ++Packets;
            if ((Time.Now() - Then) > 10.0e-3)
                {
                Sleep(1);
                ++Sleeps;
                Then = Time.Now();
                }
        }

        // Data
        double          Then;
        StopWatch       Time;
        unsigned int    Sleeps;
        unsigned int    Packets;
    };

    LoadBalance     Load;
    BlockStream *  Owner;

    //StreamManager *          BlkStreamManager;
    BlockStreamQueue &       Queue;

    enum { sQueue=inherited::sLast, sLast };
    //  override to load notify events at start
    virtual void    Init()
        {
        inherited::Init();
        AddObject(Queue.SignalSem(), sQueue);
        }
    //  do one iteration of main process
    virtual void    AtExecution(int result)
        {
        if (result == sQueue)
            ProcessPacket();
        }

    virtual void    ProcessPacket() = 0;

    void            Log(int tag, int value)
        {  Owner->BlkStreamDevice->Log(tag, value);  }
    void            Log(int tag, size_t value)
        {  Log(tag, static_cast<int>(value));  }
};

//===========================================================================
//  CLASS BlockStreamRcvrThread  -- Receiver thread for PktStreams
//===========================================================================

//---------------------------------------------------------------------------
//  declaration of  class BlockStreamRcvrThread
//---------------------------------------------------------------------------

class BlockStreamRcvrThread : public BlockStreamIoThread
{
public:
    BlockStreamRcvrThread(BlockStream *owner)
        : BlockStreamIoThread(owner, owner->RecvQ)
        {  }

    // Methods
    void  Clear()
        {  Signal.Clear();  }
    void  Recv(Buffer & packet);

protected:
    Semaphore       Signal;

    virtual void    ProcessPacket();
};

//---------------------------------------------------------------------------
//  BlockStreamRcvrThread::ProcessPacket() -- Recv data packet
//---------------------------------------------------------------------------

void  BlockStreamRcvrThread::ProcessPacket()
{
//    int depth_check = Queue.Depth();
//    Log(8, depth_check);

    //
    //  Signal packet available downstream...
    Signal.Release();

    BlockStreamDataEvent e(Owner);
    Owner->OnDataAvailable.Execute(e);

    // Load balancing
    if (Owner->RxLoadBalancing())
        Load.Balance();

//    depth_check = Queue.Depth();
//    Log(9, depth_check);
}

//---------------------------------------------------------------------------
//  BlockStreamRcvrThread::Recv() -- Return packet in data queue
//---------------------------------------------------------------------------

void  BlockStreamRcvrThread::Recv( Buffer & packet )
{
    //int depth_check = Queue.Depth();
    //Log(0x18, depth_check);
    //
    //  Wait for packet to arrive....
    Signal.Acquire();
    //
    //  Extract packet from queue
    packet = Queue.Front();
    //
    //  Pop the queue
    Queue.Pop();

    //depth_check = Queue.Depth();
    //Log(0x19, depth_check);
}

//===========================================================================
//  CLASS BlockStreamXmitThread  -- Transmit thread for BlkStreams
//===========================================================================
//  This thread is on the 'application' side of the buffer queues from the
//    VISR. It handles the interface from these queues to the app or data pins
//
//---------------------------------------------------------------------------
//  declaration of  class BlockStreamXmitThread
//---------------------------------------------------------------------------

class BlockStreamXmitThread : public BlockStreamIoThread
{
public:
    BlockStreamXmitThread(BlockStream *owner)
        : BlockStreamIoThread(owner, owner->XmitQ)
        {  }

    void  Send( Buffer & packet );

    void  Prefill(int packets)
        {
        for (int i = 0; i < packets; ++i)
            ProcessPacket();
        }

protected:
    virtual void    ProcessPacket( );
};

//---------------------------------------------------------------------------
//  BlockStreamXmitThread::ProcessPacket() -- packet needed by transmit
//---------------------------------------------------------------------------

void  BlockStreamXmitThread::ProcessPacket()
{
//    int depth_check = Queue.Depth();
//    Log(0xa, depth_check);

    //
    //  Signal upstream for data...
    BlockStreamDataEvent e(Owner);
    Owner->OnDataRequired.Execute(e);

    // Load balancing
    if (Owner->TxLoadBalancing())
        Load.Balance();

//    depth_check = Queue.Depth();
//    Log(0xb, depth_check);
}

//---------------------------------------------------------------------------
//  BlockStreamXmitThread::Send() -- Load packet for transmission
//---------------------------------------------------------------------------

void  BlockStreamXmitThread::Send( Buffer & packet )
{
    size_t depth_check = Queue.Depth();
    Log(0x1a, depth_check);

    //
    //  Push data onto outgoing queue
    Queue.Push(packet);
    Log(0x1c, depth_check);

    //depth_check = Queue.Depth();
    //Log(0x1b, depth_check);
}

//===========================================================================
//  CLASS BlockStream  -- Block I/O
//===========================================================================
//  Data Type Tags
const int PciInData  = 1;
const int PciOutData = 0;

//---------------------------------------------------------------------------
//  BlockStream::BlockStream() -   Ctor
//---------------------------------------------------------------------------
#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif

BlockStream::BlockStream()
    :  BlkStreamDevice(0), 
       RecvThread(0), 
       XmitThread(0),
       XmitQ(new BlockStreamIBSQCreditor(this, PciOutData)),
       RecvQ(new BlockStreamIBSQCreditor(this, PciInData)),
       Started(false),
       FPrefillPacketCount(8)
{
    XmitQ.QueueCapacity(2048);
    RecvQ.QueueCapacity(2048);

    RecvThread = new BlockStreamRcvrThread( this );
    RecvThread->Resume();

    XmitThread = new BlockStreamXmitThread( this );
    XmitThread->Resume();
}

//---------------------------------------------------------------------------
//  BlockStream::BlockStream() -   Dtor
//---------------------------------------------------------------------------

BlockStream::~BlockStream()
{
    if (Started)
        Stop();

    RecvThread->Shutdown();
    delete RecvThread;
    RecvThread = 0;

    XmitThread->Shutdown();
    delete XmitThread;
    XmitThread = 0;

}

//---------------------------------------------------------------------------
//  BlockStream::Start() -
//---------------------------------------------------------------------------
/*
    mBufferSize = BufferSize() / sizeof(ii32);
    mStreamDevice->SetBusmasterSize(mBufferSize);
*/

void BlockStream::Start()
{
    if (Started)
        return;
    if (!BlkStreamDevice)
        return;

    OpenWire::NotifyEvent e;
    OnBeforeStart.Execute(e);

    XmitQ.Clear();
    RecvQ.Clear();
    RecvThread->Clear();
    //
    //  Start Receiver Thread
    RecvThread->Start();
    XmitThread->Start();
    BlkStreamDevice->Reset();
    //
    //  Initial Device Configuration
    BlkStreamDevice->SetBusmasterSize(static_cast<int>(FBufferSize));
    BlkStreamDevice->ConfigureBusmastering();
    //
    //  Transmit Prefill
    Prefill();
//    StreamDevice->Log(0xC0, 0);
    //
    //  Load Xmit Busmaster Region
    BlkStreamDevice->PreloadXmitBusmasterRegion();

    // This flag must be set before starting the busmaster.
    Started = true;

    //
    //  Start busmastering
    BlkStreamDevice->EnableBusMaster(true);

    OnAfterStart.Execute(e);
}

//---------------------------------------------------------------------------
//  BlockStream::Stop() -
//---------------------------------------------------------------------------

void BlockStream::Stop()
{
    if (!Started)
        return;

    BlkStreamDevice->EnableBusMaster(false);
    RecvThread->Stop();
    XmitThread->Stop();
//    StreamDevice->Log(0xC1, 0);

    Started = false;

    OpenWire::NotifyEvent e;
    OnAfterStop.Execute(e);
}

//---------------------------------------------------------------------------
//  BlockStream::Prefill() -- Activate prefill on transmit Q
//---------------------------------------------------------------------------

void  BlockStream::Prefill()
{

    XmitQ.SignalSem().Clear();
    for (int i = 0; i < FPrefillPacketCount; ++i)
        XmitQ.SignalSem().Release();
    XmitThread->Prefill(FPrefillPacketCount);

    //StreamDevice->Log(0x0D, 0);
}

//---------------------------------------------------------------------------
//  BlockStream::RequestBuffer() - Blockstream needs a block..provide one
//---------------------------------------------------------------------------
/*
void BlockStream::RequestBuffer( Buffer &  buffer )
{
    if (!OnDataRequired.Assigned())
        {
        AccessDatagram<unsigned int> bufferDG(buffer);
        std::fill(bufferDG.begin(), bufferDG.end(), 0);
        return;    // dont block if event unassigned...
        }
        
    SendReadyEvent.Clear();
    //
    //  Signal the app to provide a buffer using Send()
    BlockStreamDataEvent BSDEvent(this);
    OnDataRequired.Execute(BSDEvent);
    //
    //  Block until buffer is delivered by send
	int result = SendObjects.WaitFor();
    if (result == sTerminated)
        return;

    //
    //  Return data delivered to system
    buffer = SendXBuffer;
    //
    //  Decrement Signal Queue
    if (mOutBuffersQueue.GetSize() != 0)
        mOutBuffersQueue.DeleteOne();
}
*/
//---------------------------------------------------------------------------
//  BlockStream::Send() -- Application provides buffer to BlockStream
//---------------------------------------------------------------------------

void BlockStream::Send( Buffer & block )
{
    //  ..and send it out
    XmitThread->Send(block);
    BlkStreamDevice->TransmitDataAvailable();
}

//---------------------------------------------------------------------------
//  BlockStream::Recv() -- Application retrieves a buffer from BlockStream
//---------------------------------------------------------------------------

void BlockStream::Recv( Buffer & block )
{
    RecvThread->Recv(block);
}

//---------------------------------------------------------------------------
//  BlockStream::Block() -- Return sized block for use in transmission
//---------------------------------------------------------------------------

Innovative::Buffer BlockStream::Block()
{
    Buffer  BlockBuffer(FBufferSize);
    return  BlockBuffer;
}

//---------------------------------------------------------------------------
//  BlockStream::SampleRate(float) --
//---------------------------------------------------------------------------
 
void BlockStream::SampleRate(float value)
{
    FSampleRate = value;
}

//---------------------------------------------------------------------------
//  BlockStream::SampleRate() --
//---------------------------------------------------------------------------
 
float BlockStream::SampleRate() const
{
    return FSampleRate;
}

//---------------------------------------------------------------------------
//  BlockStream::BufferSize(size_t) --
//---------------------------------------------------------------------------
 
void BlockStream::BufferSize(size_t value)
{
    FBufferSize = value;
}

//---------------------------------------------------------------------------
//  BlockStream::BufferSize() --
//---------------------------------------------------------------------------
 
size_t BlockStream::BufferSize() const
{
    return FBufferSize;
}

//---------------------------------------------------------------------------
//  BlockStream::AttachToPtr() --
//---------------------------------------------------------------------------
 
void BlockStream::AttachToPtr(StreamDevice *device)
{
    if (device)
        {
        if (BlkStreamDevice)
            BlkStreamDevice->RegisterStreamPlugIn(NULL);

        BlkStreamDevice = device;
        BlkStreamDevice->RegisterStreamPlugIn(this);
        }

}

//---------------------------------------------------------------------------
//  BlockStream::Detach() --
//---------------------------------------------------------------------------
 
void BlockStream::Detach()
{
	BlkStreamDevice = NULL;
    Started = false;        
}

//------------------------------------------------------------------------------
//  BlockStream::BusMasterSizeWords() --
//------------------------------------------------------------------------------

void BlockStream::BusMasterSizeWords(unsigned int words)
{
    BlkStreamDevice->BusMasterSizeWords(words);
}

//---------------------------------------------------------------------------
//  BlockStream::StreamStarted() -- returns true if the stream is started
//---------------------------------------------------------------------------

bool BlockStream::StreamStarted()
{
    return Started;
}

//---------------------------------------------------------------------------
//  BlockStream::BlockSize() -- returns true if the stream is started
//---------------------------------------------------------------------------

size_t BlockStream::BlockSize() 
{
    return FBufferSize;
}

//---------------------------------------------------------------------------
//  BlockStream::XmitQDepth() -- Depth of Xmit Q
//---------------------------------------------------------------------------

size_t  BlockStream::XmitQDepth()
{
    return XmitQ.Depth();
}

//---------------------------------------------------------------------------
//  BlockStream::RecvQDepth() -- Depth of Recv Q
//---------------------------------------------------------------------------

size_t  BlockStream::RecvQDepth()
{
    return RecvQ.Depth();
}

//---------------------------------------------------------------------------
//  BlockStream::DequeueXmitPacket() -- Remove Xmit Packet from queue
//---------------------------------------------------------------------------

void  BlockStream::DequeueXmitPacket()
{
//    BlkStreamDevice->Log(0xe0, XmitQ.QueueCapacity());
//    BlkStreamDevice->Log(0xe1, XmitQ.QueueThreshold());
//    BlkStreamDevice->Log(0xe2, XmitQ.QueueLevel());

    XmitQ.Pop();

//    BlkStreamDevice->Log(0xe4, XmitQ.QueueCapacity());
//    BlkStreamDevice->Log(0xe5, XmitQ.QueueThreshold());
//    BlkStreamDevice->Log(0xe6, XmitQ.QueueLevel());
}

//---------------------------------------------------------------------------
//  BlockStream::NextXmitPacket() -- Return a reference to next buffer
//---------------------------------------------------------------------------

Buffer  BlockStream::NextXmitPacket()
{
    size_t test_depth = XmitQ.Depth();
    //BlkStreamDevice->Log(0x60, test_depth);
    if (test_depth==0)
        {
        BlkStreamDevice->Log(0xF0, 0);
        while (1==1)
            Sleep(100);
        }

    //test_depth = XmitQ.Depth();
    //BlkStreamDevice->Log(0x61, test_depth);
    //Buffer xxx = XmitQ.Front();
    //StreamDevice->Log(0x62, 0);
    //return xxx;
    return XmitQ.Front();
}

//---------------------------------------------------------------------------
//  BlockStream::EnqueueRecvPacket() -- Add Recv Packet to queue
//---------------------------------------------------------------------------

void  BlockStream::EnqueueRecvPacket(Buffer & buffer)
{
    RecvQ.Push(buffer);
}

//===========================================================================
//  CLASS BlockStreamQueue  -- Queue to manage data from VISR to Receiver
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class BlockStreamQueue
//---------------------------------------------------------------------------

BlockStreamQueue::BlockStreamQueue(IBSQueueCreditor * creditor)
    : Capacity(0), Threshold(0), Level(0), Creditor(creditor) 
{
}

//---------------------------------------------------------------------------
//  destructor for class BlockStreamQueue
//---------------------------------------------------------------------------

BlockStreamQueue::~BlockStreamQueue()
{
    delete Creditor;
}

//---------------------------------------------------------------------------
//  BlockStreamQueue::Clear() -- clear queue
//---------------------------------------------------------------------------

void  BlockStreamQueue::Clear()
{
    //  reset state variables
    Level = 0;
    //  clear buffer queue
    BufferQ.clear();
    CreditQ.clear();
    //  reduce signal to 0
    Signal.Clear();
}

//---------------------------------------------------------------------------
//  BlockStreamQueue::Front() -- Front element in queue
//---------------------------------------------------------------------------

Buffer  BlockStreamQueue::Front()
{
    return BufferQ.front();
}

//---------------------------------------------------------------------------
//  BlockStreamQueue::Depth() -- depth of queue
//---------------------------------------------------------------------------

size_t BlockStreamQueue::Depth()
{
    return BufferQ.depth();
}

//---------------------------------------------------------------------------
//  BlockStreamQueue::QueueCapacity() -- Set queue capacity
//---------------------------------------------------------------------------

void  BlockStreamQueue::QueueCapacity(size_t cap)
{
    CriticalSectionArea guard(lock);
    Capacity = cap;
}

//---------------------------------------------------------------------------
//  BlockStreamQueue::QueueCapacity() -- Retreive capacity
//---------------------------------------------------------------------------

size_t BlockStreamQueue::QueueCapacity() const
{
    return Capacity;
}

//---------------------------------------------------------------------------
//  BlockStreamQueue::QueueThreshold() -- Set queue threshold
//---------------------------------------------------------------------------

void BlockStreamQueue::QueueThreshold(size_t value)
{
    CriticalSectionArea guard(lock);
    Threshold = value;
}

//---------------------------------------------------------------------------
//  BlockStreamQueue::QueueThreshold() -- Retreive queue threshold
//---------------------------------------------------------------------------

size_t BlockStreamQueue::QueueThreshold() const
{
    return Threshold;
}

//---------------------------------------------------------------------------
//  BlockStreamQueue::QueueLevel() -- Set queue fullness level
//---------------------------------------------------------------------------

void BlockStreamQueue::QueueLevel(size_t value)
{
    CriticalSectionArea guard(lock);
    Level = value;
}

//---------------------------------------------------------------------------
//  BlockStreamQueue::QueueLevel() -- Retreive queue fullness level
//---------------------------------------------------------------------------

size_t BlockStreamQueue::QueueLevel() const
{
    return Level;
}

//---------------------------------------------------------------------------
//  BlockStreamQueue::SignalSem() -- Return reference to semaphore
//---------------------------------------------------------------------------

Semaphore & BlockStreamQueue::SignalSem()
{
    return Signal;
}

//===========================================================================
//  CLASS XmitBSQueue  -- Queue to manage data from Receiver to VISR
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class XmitBSQueue
//---------------------------------------------------------------------------

XmitBSQueue::XmitBSQueue(IBSQueueCreditor * creditor)
    :  BlockStreamQueue(creditor), SuspendSignal(false)
{
}

//---------------------------------------------------------------------------
//  destructor for class XmitBSQueue
//---------------------------------------------------------------------------

XmitBSQueue::~XmitBSQueue()
{
}

//---------------------------------------------------------------------------
//  XmitPSQueue::Push() -- push buffer onto queue
//---------------------------------------------------------------------------

void  XmitBSQueue::Push( Buffer & buffer )
{
    CriticalSectionArea guard(lock);

    size_t PacketSize = buffer.FullSizeInInts();
    BufferQ.push(buffer);
    Level += PacketSize;

    //
    //  If we have uncredited buffers in busmaster memory, credit one
    if (!CreditQ.empty())
        {
        size_t credit_size = CreditQ.front();
        Creditor->CreditStream(credit_size);
        CreditQ.pop();
        }
}

//---------------------------------------------------------------------------
//  XmitPSQueue::Pop() -- remove block from queue
//---------------------------------------------------------------------------

void  XmitBSQueue::Pop()
{
    CriticalSectionArea guard(lock);

    //  Save size of buffer being popped
    size_t PacketSize = BufferQ.front().FullSizeInInts();
    BufferQ.pop();
    Level -= PacketSize;

    //
    //  Credit granting...we dont have the index so we dont grant credit
    //
    //  If there is room in the queue, request more input
    if (Level<Capacity && (SuspendSignal==false))
        {
        //  Signal need for additional data from host
        Signal.Release();
        }
}

//===========================================================================
//  CLASS RecvBSQueue  -- Queue to manage data from VISR to Receiver
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class RecvBSQueue
//---------------------------------------------------------------------------

RecvBSQueue::RecvBSQueue(IBSQueueCreditor * creditor)
    :  BlockStreamQueue(creditor)
{
}

//---------------------------------------------------------------------------
//  destructor for class RecvBSQueue
//---------------------------------------------------------------------------

RecvBSQueue::~RecvBSQueue()
{
}

//---------------------------------------------------------------------------
//  RecvBSQueue::Push() -- push buffer onto queue
//---------------------------------------------------------------------------

void  RecvBSQueue::Push( Buffer & buffer )
{
    CriticalSectionArea guard(lock);

    size_t PacketSize = buffer.SizeInInts();
    BufferHeaderDatagram bufferHdr(buffer);
    int  PacketIdx = bufferHdr[0];
    //  Push a buffer onto the queue. If we are above our threshold, stop
    //    granting credit to eventually shut down data flow
    BufferQ.push(buffer);
    Level += PacketSize;

    if (Level<Threshold)
        {
        //  Grant credit for packet
        Creditor->CreditStream( PacketIdx );
        }
    else
        {
        //  Push credit onto credit queue
        CreditQ.push(PacketIdx);  
        }
    //
    //  For every buffer added, signal Rcvr thread
    Signal.Release();

}

//---------------------------------------------------------------------------
//  RecvBSQueue::Pop() -- 
//---------------------------------------------------------------------------

void  RecvBSQueue::Pop()
{
    CriticalSectionArea guard(lock);

    size_t PacketSize = BufferQ.front().FullSizeInInts();

    BufferQ.pop();
    Level -= PacketSize;
    //
    //  If we have uncredited buffers in the queue, credit one
    if (!CreditQ.empty())
        {
        size_t credit_size = CreditQ.front();
        Creditor->CreditStream(credit_size);
        CreditQ.pop();
        }
}

} // Innovative

