// BlockStream_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure 
//  agreement with Innovative Integration Corporation and may not be copied or 
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation. 
//  All Rights Reserved.
//

#ifndef BlockStream_MbH
#define BlockStream_MbH

#include <StreamSupport_Mb.h>
#include <Buffer_Mb.h>
#include <Events_Mb.h>
#include <ThreadSafeQueue_Mb.h>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//===========================================================================
// CLASS BlockStreamDataEvent  --  Data I/O Event for BlockStream
//===========================================================================

class BlockStreamDataEvent : public OpenWire::Event
{
public:
    class BlockStream *  Sender;

public:
    BlockStreamDataEvent( BlockStream * sender )
        :  Sender( sender )
        {
        }
};

//===========================================================================
//  CLASS IBSQueueCreditor  -- Interface for credit management in BSQueue
//===========================================================================

class IBSQueueCreditor
{
public:
    virtual void  CreditStream(size_t cq_token) = 0;
    virtual ~IBSQueueCreditor()
    	{}
};

//===========================================================================
//  CLASS BlockStreamQueue  -- Queue to manage data from VISR to Receiver
//===========================================================================

class BlockStreamQueue
{
public:
    // Ctor
    BlockStreamQueue(IBSQueueCreditor * creditor);
    virtual ~BlockStreamQueue();

    // Methods
    Buffer          Front();
    size_t          Depth();

    void            Clear();
    virtual void    Push(Buffer & buffer) = 0;
    virtual void    Pop() = 0;

    Semaphore &     SignalSem();

    // Properties
    void         QueueCapacity(size_t cap);
    size_t       QueueCapacity() const;

    void         QueueThreshold(size_t value);
    size_t       QueueThreshold() const;

    void         QueueLevel(size_t value);
    size_t       QueueLevel() const;

protected:
    //
    //  Member Data
    size_t      Capacity;
    size_t      Threshold;
    size_t      Level;
    Semaphore   Signal;
    mutable CriticalSection  lock;

    ThreadSafeQueue<Buffer>     BufferQ;
    ThreadSafeQueue<size_t>     CreditQ;
    IBSQueueCreditor *          Creditor;
};

//===========================================================================
//  CLASS XmitBSQueue  -- Queue to manage data from Receiver to VISR
//===========================================================================

class XmitBSQueue : public BlockStreamQueue
{
public:
    XmitBSQueue(IBSQueueCreditor * creditor);
    ~XmitBSQueue();

    virtual void  Push(Buffer & buffer);
    virtual void  Pop();
    void          Suspend(bool which)  {  SuspendSignal=which;  }

private:
    //
    //  Member Data
    bool  SuspendSignal;

};

//===========================================================================
//  CLASS RecvBSQueue  -- Queue to manage data from VISR to Receiver
//===========================================================================

class RecvBSQueue : public BlockStreamQueue
{
public:
    RecvBSQueue(IBSQueueCreditor * creditor);
    ~RecvBSQueue();

    virtual void  Push(Buffer & buffer);
    virtual void  Pop();

};

//===========================================================================
//  CLASS BlockStream  -- 
//===========================================================================

class BlockStreamIoThread;
class BlockStreamIBSQCreditor;
class BlockStreamRcvrThread;
class BlockStreamXmitThread;
    
class BlockStream : public StreamManager
{
    friend class BlockStreamIoThread;
    friend class BlockStreamIBSQCreditor;
    friend class BlockStreamRcvrThread;
    friend class BlockStreamXmitThread;

public:
    BlockStream();
   ~BlockStream();

    void                Start();
    void                Stop();
    Innovative::Buffer  Block();

    void  Send( Buffer & block );
    void  Recv( Buffer & block );

    // Properties
    void                        SampleRate(float value);
    float                       SampleRate() const;

    void                        BufferSize(size_t value);
    size_t                      BufferSize() const;

    void  PrefillPacketCount(int packets)
        {  FPrefillPacketCount=packets;  }
    int   PrefillPacketCount() const
        {  return FPrefillPacketCount;  }

    void    RxLoadBalancing(bool state)
        {  FRxLoadBalancing=state;  }
    bool    RxLoadBalancing() const
        {  return FRxLoadBalancing;  }

    void    TxLoadBalancing(bool state)
        {  FTxLoadBalancing=state;  }
    bool    TxLoadBalancing() const
        {  return FTxLoadBalancing;  }

    OpenWire::EventHandler<BlockStreamDataEvent>   OnDataRequired;
    OpenWire::EventHandler<BlockStreamDataEvent>   OnDataAvailable;
    OpenWire::EventHandler<OpenWire::NotifyEvent>  OnBeforeStart;
    OpenWire::EventHandler<OpenWire::NotifyEvent>  OnAfterStart;
    OpenWire::EventHandler<OpenWire::NotifyEvent>  OnAfterStop;

    virtual void EnqueueRecvPacket(Buffer & buffer);
    virtual Buffer NextXmitPacket();
    virtual void DequeueXmitPacket();
    virtual size_t XmitQDepth();
    virtual size_t RecvQDepth();
    
    virtual bool StreamStarted();
    virtual size_t  BlockSize();

public: // Pseudo-protected
    virtual void AttachToPtr( StreamDevice *device );
    virtual void Detach();
    // Specify driver-allocated BM region, prior to open
    virtual void            BusMasterSizeWords(unsigned int value);
   // virtual bool OutBufferIndex( int buffer_index ); // Returns true if the buffer is handled.
   // virtual bool InBufferIndex( int buffer_index ); // Returns true if the buffer is handled.


protected:
    StreamDevice *                   BlkStreamDevice;
    BlockStreamRcvrThread *          RecvThread;
    BlockStreamXmitThread *          XmitThread;
    XmitBSQueue                      XmitQ;
    RecvBSQueue                      RecvQ;
    bool                             Started;
    int                              FPrefillPacketCount;
    // Load-balancing data
    bool                             FRxLoadBalancing;
    bool                             FTxLoadBalancing;

    // Fields
    float                       FSampleRate;
    size_t                      FBufferSize;

    void       Prefill();
};

#ifdef __CLR_VER
#pragma managed(pop)
#endif
} // Innovative

#endif
