// Buffer_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2010 Innovative Integration Corporation.
//  All Rights Reserved.
//
//---------------------------------------------------------------------------

#include "Buffer_Mb.h"
#include "BufferHeader_Mb.h"
#include "BufferTrailer_Mb.h"
#include <MemoryUtils_Mb.h>

#include <algorithm>

namespace   Innovative
{
//==============================================================================
//  CLASS Buffer  --  Data buffer class
//==============================================================================
//------------------------------------------------------------------------------
//  housekeeping for class Buffer
//------------------------------------------------------------------------------

const unsigned int DefaultPmcHeaderSize = 2;     // Pmc Buffers use 2 word header
const unsigned int DefaultPmcTrailerSize = 0;    // Pmc Buffers use 0 word trailer

const unsigned int DefaultVeloHeaderSize = 4;     // Velo Buffers use 2 word header
const unsigned int DefaultVeloTrailerSize = 0;    // Velo Buffers use 0 word trailer

const unsigned int DefaultVitaHeaderSize = 7;     // Vita Buffers use 7 word header
const unsigned int DefaultVitaTrailerSize = 1;    // Vita Buffers use 1 word trailer

const unsigned int DefaultMcHeaderSize = 1;     // MC Buffers use 1 word header
const unsigned int DefaultMcTrailerSize = 0;    // MC Buffers use 0 word trailer

const unsigned int DefaultHeaderSize  = DefaultPmcHeaderSize;    // Buffers use Pmc header
const unsigned int DefaultTrailerSize = DefaultPmcTrailerSize;   // Buffers use Pmc trailer

//------------------------------------------------------------------------------
//  constructor for class Buffer  --  default ctor
//------------------------------------------------------------------------------

Buffer::Buffer()
    :  HeaderH(AlignedBlockHandle::createBlock, DefaultHeaderSize),
       TrailerH(AlignedBlockHandle::createBlock, DefaultTrailerSize)
{
    // No data block is allocated yet, size 0.
}

//------------------------------------------------------------------------------
//  constructor for class Buffer  --  create data block from size
//------------------------------------------------------------------------------

Buffer::Buffer(size_t sizeInInts)
    :  DataH(AlignedBlockHandle::createBlock, sizeInInts),
       HeaderH(AlignedBlockHandle::createBlock, DefaultHeaderSize),
       TrailerH(AlignedBlockHandle::createBlock, DefaultTrailerSize)
{
}

//------------------------------------------------------------------------------
//  constructor for class Buffer  --  all sizes specifiable
//------------------------------------------------------------------------------

Buffer::Buffer(size_t dataSizeInInts, size_t headerSizeInInts, size_t trailerSizeInInts )
    :  DataH(AlignedBlockHandle::createBlock, dataSizeInInts),
       HeaderH(AlignedBlockHandle::createBlock, headerSizeInInts),
       TrailerH(AlignedBlockHandle::createBlock, trailerSizeInInts)
{
}

//------------------------------------------------------------------------------
//  constructor for class Buffer  --  copy ctor
//------------------------------------------------------------------------------

Buffer::Buffer(const Buffer & other)
    :   IDatagrammable(), DataH(other.DataH),
        HeaderH(other.HeaderH), TrailerH(other.TrailerH)
{
}

//------------------------------------------------------------------------------
//  destructor for class Buffer
//------------------------------------------------------------------------------

Buffer::~Buffer()
{
}

//------------------------------------------------------------------------------
//  Buffer::operator =() --  Assign to buffer
//------------------------------------------------------------------------------
//  resultant data will be exactly like full copy, but data is shared until split

Buffer &  Buffer::operator = (const Buffer & other)
{
    Buffer tmp(other);

    SwapData(tmp);
    SwapHeader(tmp);
    SwapTrailer(tmp);

    return *this;
}

//------------------------------------------------------------------------------
//  Buffer::operator =() --  Assign Data region (only) to Buffer
//------------------------------------------------------------------------------
//  Original data (only) shared with other buffer, header/trailer unaffected

void Buffer::AssignData(const Buffer& other)
{
    Buffer tmp(other);

    SwapData(tmp);
}

//------------------------------------------------------------------------------
//  Buffer::SizeInInts() --  Buffer Size
//------------------------------------------------------------------------------

size_t Buffer::SizeInInts() const
{
    if (DataH.IsNotNull())
        return DataH->SizeInInts();
    else
        return 0;
}

//------------------------------------------------------------------------------
//  Buffer::SizeInBytes() --  Buffer Size
//------------------------------------------------------------------------------

size_t Buffer::SizeInBytes() const
{
    return SizeInInts() * sizeof(int);
}

//------------------------------------------------------------------------------
//  Buffer::HeaderSizeInInts() --  Buffer Header Size
//------------------------------------------------------------------------------

size_t Buffer::HeaderSizeInInts() const
{
    if (HeaderH.IsNotNull())
        return HeaderH->SizeInInts();
    else
        return 0;
}

//------------------------------------------------------------------------------
//  Buffer::HeaderSizeInBytes() --  Buffer Header Size
//------------------------------------------------------------------------------

size_t Buffer::HeaderSizeInBytes() const
{
    return HeaderSizeInInts() * sizeof(int);
}

//------------------------------------------------------------------------------
//  Buffer::TrailerSizeInInts() --  Buffer Trailer Size
//------------------------------------------------------------------------------

size_t Buffer::TrailerSizeInInts() const
{
    if (TrailerH.IsNotNull())
        return TrailerH->SizeInInts();
    else
        return 0;
}

//------------------------------------------------------------------------------
//  Buffer::TrailerSizeInBytes() --  Buffer Trailer Size
//------------------------------------------------------------------------------

size_t Buffer::TrailerSizeInBytes() const
{
    return TrailerSizeInInts() * sizeof(int);
}

//------------------------------------------------------------------------------
//  Buffer::SwapData() --  Swap Data regions
//------------------------------------------------------------------------------

void  Buffer::SwapData(Buffer & other)
{
    DataH.Swap(other.DataH);
}

//------------------------------------------------------------------------------
//  Buffer::SwapHeader() --  Swap Header regions
//------------------------------------------------------------------------------

void  Buffer::SwapHeader(Buffer & other)
{
    HeaderH.Swap(other.HeaderH);
}

//------------------------------------------------------------------------------
//  Buffer::SwapTrailer() --  Swap Trailer regions
//------------------------------------------------------------------------------

void  Buffer::SwapTrailer(Buffer & other)
{
    TrailerH.Swap(other.TrailerH);
}

//------------------------------------------------------------------------------
//  Buffer::SwapAll() --  Swap everything regions
//------------------------------------------------------------------------------

void  Buffer::SwapAll(Buffer & other)
{
    SwapHeader(other);
    SwapData(other);
    SwapTrailer(other);
}

//------------------------------------------------------------------------------
//  Buffer::MakeWritable() --
//------------------------------------------------------------------------------

bool  Buffer::MakeWritable()
{
    bool buffer_has_changed = (IsUnique()==false);
    MakeUnique();
    return buffer_has_changed;
}

//------------------------------------------------------------------------------
//  Buffer::ReadHeaderField_PacketSize()() -- read my packet's size
//------------------------------------------------------------------------------
//  May be replaced by derived buffers

size_t  Buffer::ReadHeaderField_PacketSize()
{
    if (HeaderSizeInInts()==0)
        return 0;

    PmcBufferManager  PMC;

    size_t packet_size = PMC.PacketSizeFromMemoryInts(reinterpret_cast<unsigned int *>(Header()));
    return packet_size;
}

//===========================================================================
//  CLASS PmcBuffer  -- "Classic" X3/X5 Buffer Type
//===========================================================================
//------------------------------------------------------------------------------
//  constructor for class PmcBuffer  --  default ctor
//------------------------------------------------------------------------------

PmcBuffer::PmcBuffer()
    :  Buffer(0, DefaultPmcHeaderSize, DefaultPmcTrailerSize)
{
    // No data block is allocated yet, size 0.
}

//------------------------------------------------------------------------------
//  constructor for class PmcBuffer  --  create data block from size
//------------------------------------------------------------------------------

PmcBuffer::PmcBuffer(size_t sizeInInts)
    :  Buffer(sizeInInts, DefaultPmcHeaderSize, DefaultPmcTrailerSize)
{
}

//------------------------------------------------------------------------------
//  constructor for class PmcBuffer  --  copy ctor
//------------------------------------------------------------------------------

PmcBuffer::PmcBuffer(const PmcBuffer & other)
    :   Buffer(other)
{
}

//------------------------------------------------------------------------------
//  destructor for class PmcBuffer
//------------------------------------------------------------------------------

PmcBuffer::~PmcBuffer()
{
}

//------------------------------------------------------------------------------
//  PmcBuffer::operator =() --  Assign to buffer
//------------------------------------------------------------------------------

PmcBuffer &  PmcBuffer::operator = (const PmcBuffer & other)
{
    PmcBuffer tmp(other);

    SwapData(tmp);
    SwapHeader(tmp);
    SwapTrailer(tmp);

    return *this;
}

//===========================================================================
//  CLASS VeloBuffer  -- X6 "Velocia" Buffer Type
//===========================================================================
//------------------------------------------------------------------------------
//  constructor for class VeloBuffer  --  default ctor
//------------------------------------------------------------------------------

VeloBuffer::VeloBuffer()
    :  Buffer(0, DefaultVeloHeaderSize, DefaultVeloTrailerSize)
{
    // No data block is allocated yet, size 0.
}

//------------------------------------------------------------------------------
//  constructor for class VeloBuffer  --  create data block from size
//------------------------------------------------------------------------------

VeloBuffer::VeloBuffer(size_t sizeInInts)
    :  Buffer(sizeInInts, DefaultVeloHeaderSize, DefaultVeloTrailerSize)
{
}

//------------------------------------------------------------------------------
//  constructor for class VeloBuffer  --  copy ctor
//------------------------------------------------------------------------------

VeloBuffer::VeloBuffer(const VeloBuffer & other)
    :   Buffer(other)
{
}

//------------------------------------------------------------------------------
//  destructor for class VeloBuffer
//------------------------------------------------------------------------------

VeloBuffer::~VeloBuffer()
{
}

//------------------------------------------------------------------------------
//  VeloBuffer::operator =() --  Assign to buffer
//------------------------------------------------------------------------------

VeloBuffer &  VeloBuffer::operator = (const VeloBuffer & other)
{
    VeloBuffer tmp(other);

    SwapData(tmp);
    SwapHeader(tmp);
    SwapTrailer(tmp);

    return *this;
}

//===========================================================================
//  CLASS VitaBuffer  -- X6 "VITA-49" Buffer Type
//===========================================================================
//------------------------------------------------------------------------------
//  constructor for class VitaBuffer  --  default ctor
//------------------------------------------------------------------------------

VitaBuffer::VitaBuffer()
    :  Buffer(0, DefaultVitaHeaderSize, DefaultVitaTrailerSize)
{
    // No data block is allocated yet, size 0.
}

//------------------------------------------------------------------------------
//  constructor for class VitaBuffer  --  create data block from size
//------------------------------------------------------------------------------

VitaBuffer::VitaBuffer(size_t sizeInInts)
    :  Buffer(sizeInInts, DefaultVitaHeaderSize, DefaultVitaTrailerSize)
{
}

//------------------------------------------------------------------------------
//  constructor for class VitaBuffer  --  copy ctor
//------------------------------------------------------------------------------

VitaBuffer::VitaBuffer(const VitaBuffer & other)
    :   Buffer(other)
{
}

//------------------------------------------------------------------------------
//  destructor for class VitaBuffer
//------------------------------------------------------------------------------

VitaBuffer::~VitaBuffer()
{
}

//------------------------------------------------------------------------------
//  VitaBuffer::operator =() --  Assign to buffer
//------------------------------------------------------------------------------

VitaBuffer &  VitaBuffer::operator = (const VitaBuffer & other)
{
    VitaBuffer tmp(other);

    SwapData(tmp);
    SwapHeader(tmp);
    SwapTrailer(tmp);

    return *this;
}

//------------------------------------------------------------------------------
//  VitaBuffer::PadBytes() --  return size of ignored 'Padding' bytes
//------------------------------------------------------------------------------

size_t  VitaBuffer::PadBytes()
{
    VitaTrailerDatagram  trailer(*this);
    return trailer.Padding();
}

//------------------------------------------------------------------------------
//  VitaBuffer::ReadHeaderField_PacketSize()() -- read my packet's size
//------------------------------------------------------------------------------
//  May be replaced by derived buffers

size_t  VitaBuffer::ReadHeaderField_PacketSize()
{
    if (HeaderSizeInInts()==0)
        return 0;

    VitaBufferManager  VITA;

    size_t packet_size = VITA.PacketSizeFromMemoryInts(reinterpret_cast<unsigned int *>(Header()));
    return packet_size;
}

//===========================================================================
//  Buffer Utility Functions --
//===========================================================================
//------------------------------------------------------------------------------
//  ClearHeader(Buffer& ) -- fill header with 0s
//------------------------------------------------------------------------------

void  ClearHeader(Buffer & buffer)
{
    BufferHeaderDatagram  header(buffer);
    for (unsigned int i=0; i<header.size(); ++i)
        header[i] = 0;
}

//------------------------------------------------------------------------------
//  ClearTrailer(Buffer& ) -- fill trailer with 0s
//------------------------------------------------------------------------------

void  ClearTrailer(Buffer & buffer)
{
    BufferTrailerDatagram  trailer(buffer);
    for (unsigned int i=0; i<trailer.size(); ++i)
        trailer[i] = 0;
}

//------------------------------------------------------------------------------
//  InitPmcHeader(Buffer& ) -- make header 'proper' for PMC Buffer
//------------------------------------------------------------------------------

void  InitPmcHeader(Buffer & buffer)
{
    PacketHeaderDatagram  header(buffer);
    header.PacketSize(buffer.FullSizeInInts());
}

//------------------------------------------------------------------------------
//  InitHeader(PmcBuffer& ) -- make header 'proper' for PMC Buffer
//------------------------------------------------------------------------------

void  InitHeader(PmcBuffer & buffer)
{
    InitPmcHeader(buffer);
}

//------------------------------------------------------------------------------
//  InitHeader(VeloBuffer& ) -- make header 'proper' for Velo Buffer
//------------------------------------------------------------------------------

void  InitHeader(VeloBuffer & buffer)
{
    VeloHeaderDatagram  header(buffer);
    header.PacketSize(buffer.FullSizeInInts());
}

//------------------------------------------------------------------------------
//  InitHeader(VitaBuffer& ) -- make header 'proper' for Vita Buffer
//------------------------------------------------------------------------------

void  InitHeader(VitaBuffer & buffer)
{
    VitaHeaderDatagram  header(buffer);
    header.PacketSize(buffer.FullSizeInInts());
    header.Ctrr_Init();
    header.TSI(VitaHeaderDatagram::tsiOther);  // these two must be non zero
    header.TSF(VitaHeaderDatagram::tsfFree);   //  to get header size correct
    header.PacketType(0x01);
    header.DestinationMask(0x01);
    header.StartFrame(true);
    header.EndFrame(true);
}

//------------------------------------------------------------------------------
//  InitHeader(McBuffer& ) -- make header 'proper' for MultiChan Buffer
//------------------------------------------------------------------------------

void  InitHeader(McBuffer& buffer)
{
    McHeaderDatagram  header(buffer);
    header.DataSize(buffer.SizeInInts());
    header.ChannelId(0);
    header.Padding(0);
}

//------------------------------------------------------------------------------
//  InitTrailer(VitaBuffer& ) -- make trailer 'proper' for Vita Buffer
//------------------------------------------------------------------------------

void  InitTrailer(VitaBuffer & buffer)
{
    VitaTrailerDatagram  trailer(buffer);
    trailer.Padding(0);
    trailer.Ones_Init();
}

//------------------------------------------------------------------------------
//  ConvertHeader(VeloBuffer& ) -- adapt 'bad' Velo header into good one
//------------------------------------------------------------------------------
//  Because generic buffer generators can affect the header (as part of Buffer assignment)
//    we can get a Velo with a PMC sized header. 
//  Also calls InitHeader() for all cases

void  ConvertHeader(VeloBuffer & buffer)
{
	if (buffer.HeaderSizeInInts() != DefaultVeloHeaderSize)
		{
		VeloBuffer  NewH(0);   // Create new, well formed header
		ClearHeader(NewH);
		
		VeloHeaderDatagram  old_header(buffer); 
		VeloHeaderDatagram  new_header(NewH);
		size_t words = std::min<size_t>(buffer.HeaderSizeInInts(), static_cast<size_t>(DefaultVeloHeaderSize));
        for (unsigned int i=0; i<words; i++)
			new_header[i] = old_header[i];
			
		buffer.SwapHeader(NewH);   // grab the new header for our own
		}
	InitHeader(buffer);
}

//===========================================================================
//  CLASS MessageDatagram  -- Datagram access to PMC buffer data
//===========================================================================
//---------------------------------------------------------------------------
//  MessageDatagram::DatagramSize() -- return size of buffer in INTs
//---------------------------------------------------------------------------

size_t  MessageDatagram::DatagramSize() const
{
    return FBuffer.SizeInInts();
}

//---------------------------------------------------------------------------
//  MessageDatagram::DatagramBasePtr() -- pointer to base of data
//---------------------------------------------------------------------------

int *  MessageDatagram::DatagramBasePtr() const
{
    return FBuffer.Data();
}

//===========================================================================
//  CLASS PmcBufferManager  -- Management of PMC Velocia Buffers
//===========================================================================
//---------------------------------------------------------------------------
//  PmcBufferManager::NativeBuffer() -- Return properly typed Velocia Buffer
//---------------------------------------------------------------------------

Buffer  PmcBufferManager::NativeBuffer()
{
    // Pmc is the right type
    PmcBuffer Temp;
    return Temp;
}

//---------------------------------------------------------------------------
//  PmcBufferManager::HeaderSizeInts() -- Return size of native header
//---------------------------------------------------------------------------

unsigned int  PmcBufferManager::HeaderSizeInts()
{
    // Pmc is the right type
    return DefaultPmcHeaderSize;
}

//---------------------------------------------------------------------------
//  PmcBufferManager::PacketSizeFromHeaderInts() -- Return size from Header
//---------------------------------------------------------------------------

size_t        PmcBufferManager::PacketSizeFromHeaderInts(const Buffer & buffer)
{
    PmcBuffer native = Convert<PmcBuffer>(buffer);
    PmcHeaderDatagram ph_dg(native);
    return ph_dg.PacketSize();
}

//---------------------------------------------------------------------------
//  PmcBufferManager::PacketSizeFromMemoryInts() -- Return size from Memory
//---------------------------------------------------------------------------
//  This assumes that the pointer has enough of the header to get the size.
//    for PmcBuffers, this is always true.

size_t        PmcBufferManager::PacketSizeFromMemoryInts(unsigned int * start_of_header)
{
    PmcHeaderDatagram ph_dg(start_of_header);
    return ph_dg.PacketSize();
}

//---------------------------------------------------------------------------
//  PmcBufferManager::PeripheralIdFromMemory() -- Return PID from Memory
//---------------------------------------------------------------------------
//  This assumes that the pointer has enough of the header to get the PID.
//    for PmcBuffers, this is always true.

unsigned int  PmcBufferManager::PeripheralIdFromMemory(unsigned int * start_of_header)
{
    PmcHeaderDatagram ph_dg(start_of_header);
    return ph_dg.PeripheralId();
}

//---------------------------------------------------------------------------
//  PmcBufferManager::SetPeripheralIdIntoMemory() -- load PID into Memory
//---------------------------------------------------------------------------
//  This assumes that the pointer has enough of the header to get the PID.
//    for PmcBuffers, this is always true.

void  PmcBufferManager::SetPeripheralIdIntoMemory(unsigned int * start_of_header, int pid)
{
    PmcHeaderDatagram ph_dg(start_of_header);
    ph_dg.PeripheralId(pid);
}

//===========================================================================
//  CLASS VeloBufferManager  -- Management of Velo Velocia Buffers
//===========================================================================
//---------------------------------------------------------------------------
//  VeloBufferManager::NativeBuffer() -- Return properly typed Velocia Buffer
//---------------------------------------------------------------------------

Buffer  VeloBufferManager::NativeBuffer()
{
    // Velo is the right type
    VeloBuffer Temp;
    return Temp;
}

//---------------------------------------------------------------------------
//  VeloBufferManager::HeaderSizeInts() -- Return size of native header
//---------------------------------------------------------------------------

unsigned int  VeloBufferManager::HeaderSizeInts()
{
    // Velo is the right type
    return DefaultVeloHeaderSize;
}

//---------------------------------------------------------------------------
//  VeloBufferManager::PacketSizeFromHeaderInts() -- Return size from Header
//---------------------------------------------------------------------------

size_t        VeloBufferManager::PacketSizeFromHeaderInts(const Buffer & buffer)
{
    VeloBuffer native = Convert<VeloBuffer>(buffer);
    VeloHeaderDatagram vh_dg(native);
    return vh_dg.PacketSize();
}

//---------------------------------------------------------------------------
//  VeloBufferManager::PacketSizeFromMemoryInts() -- Return size from Memory
//---------------------------------------------------------------------------
//  This assumes that the pointer has enough of the header to get the size.
//    for VeloBuffers, this is always true.

size_t        VeloBufferManager::PacketSizeFromMemoryInts(unsigned int * start_of_header)
{
    VeloHeaderDatagram vh_dg(start_of_header);
    return vh_dg.PacketSize();
}

//---------------------------------------------------------------------------
//  VeloBufferManager::PeripheralIdFromMemory() -- Return PID from Memory
//---------------------------------------------------------------------------
//  This assumes that the pointer has enough of the header to get the PID.
//    for VeloBuffers, this is always true.

unsigned int  VeloBufferManager::PeripheralIdFromMemory(unsigned int * start_of_header)
{
    VeloHeaderDatagram vh_dg(start_of_header);
    return vh_dg.PeripheralId();
}

//---------------------------------------------------------------------------
//  VeloBufferManager::SetPeripheralIdIntoMemory() -- load PID into Memory
//---------------------------------------------------------------------------
//  This assumes that the pointer has enough of the header to get the PID.
//    for VeloBuffers, this is always true.

void  VeloBufferManager::SetPeripheralIdIntoMemory(unsigned int * start_of_header, int pid)
{
    VeloHeaderDatagram vh_dg(start_of_header);
    vh_dg.PeripheralId(pid);
}

//===========================================================================
//  CLASS VitaBufferManager  -- Management of Vita Velocia Buffers
//===========================================================================
//---------------------------------------------------------------------------
//  VitaBufferManager::NativeBuffer() -- Return properly typed Vita Buffer
//---------------------------------------------------------------------------

Buffer  VitaBufferManager::NativeBuffer()
{
    // Vita is the right type
    VitaBuffer Temp;
    return Temp;
}

//---------------------------------------------------------------------------
//  VitaBufferManager::HeaderSizeInts() -- Return size of native header
//---------------------------------------------------------------------------

unsigned int  VitaBufferManager::HeaderSizeInts()
{
    // Vita is the right type
    return DefaultVitaHeaderSize;
}

//---------------------------------------------------------------------------
//  VitaBufferManager::TrailerSizeInts() -- Return size of native header
//---------------------------------------------------------------------------

unsigned int  VitaBufferManager::TrailerSizeInts()
{
    // Vita is the right type
    return DefaultVitaTrailerSize;
}

//---------------------------------------------------------------------------
//  VitaBufferManager::PacketSizeFromHeaderInts() -- Return size from Header
//---------------------------------------------------------------------------

size_t        VitaBufferManager::PacketSizeFromHeaderInts(const Buffer & buffer)
{
    VitaBuffer native = Convert<VitaBuffer>(buffer);
    VitaHeaderDatagram vh_dg(native);
    return vh_dg.PacketSize();
}

//---------------------------------------------------------------------------
//  VitaBufferManager::PacketSizeFromMemoryInts() -- Return size from Memory
//---------------------------------------------------------------------------
//  This assumes that the pointer has enough of the header to get the size.
//    for VitaBuffers, this is always true.

size_t        VitaBufferManager::PacketSizeFromMemoryInts(unsigned int * start_of_header)
{
    VitaHeaderDatagram vh_dg(start_of_header);
    return vh_dg.PacketSize();
}

//===========================================================================
//  CLASS McBuffer  -- XU "MultiChan" Buffer Type
//===========================================================================
//------------------------------------------------------------------------------
//  constructor for class McBuffer  --  default ctor
//------------------------------------------------------------------------------

McBuffer::McBuffer()
    :  Buffer(0, DefaultMcHeaderSize, DefaultMcTrailerSize)
{
    // No data block is allocated yet, size 0.
}

//------------------------------------------------------------------------------
//  constructor for class McBuffer  --  create data block from size
//------------------------------------------------------------------------------

McBuffer::McBuffer(size_t sizeInInts)
    :  Buffer(sizeInInts, DefaultMcHeaderSize, DefaultMcTrailerSize)
{
}

//------------------------------------------------------------------------------
//  constructor for class McBuffer  --  copy ctor
//------------------------------------------------------------------------------

McBuffer::McBuffer(const McBuffer & other)
    :   Buffer(other)
{
}

//------------------------------------------------------------------------------
//  destructor for class McBuffer
//------------------------------------------------------------------------------

McBuffer::~McBuffer()
{
}

//------------------------------------------------------------------------------
//  McBuffer::operator =() --  Assign to buffer
//------------------------------------------------------------------------------

McBuffer &  McBuffer::operator = (const McBuffer & other)
{
    McBuffer tmp(other);

    SwapData(tmp);
    SwapHeader(tmp);
    SwapTrailer(tmp);

    return *this;
}

//------------------------------------------------------------------------------
//  McBuffer::PadBytes() --  return size of ignored 'Padding' bytes
//------------------------------------------------------------------------------

size_t  McBuffer::PadBytes()
{
    McHeaderDatagram  header(*this);
    return header.Padding();
}

//------------------------------------------------------------------------------
//  McBuffer::ReadHeaderField_PacketSize()() -- read my packet's size
//------------------------------------------------------------------------------

size_t  McBuffer::ReadHeaderField_PacketSize()
{
    if (HeaderSizeInInts()==0)
        return 0;

    McBufferManager  MC;

    size_t packet_size = MC.PacketSizeFromMemoryInts(reinterpret_cast<unsigned int *>(Header()));
    return packet_size;
}

//===========================================================================
//  CLASS McImage  -- XU "MultiChan" Image Type
//===========================================================================
//------------------------------------------------------------------------------
//  constructor for class McImage  --  default ctor
//------------------------------------------------------------------------------

McImage::McImage()
    :  DataPtr(0), DataSizeWds(0), HeaderData(0), IHeader(*this)
{
    // No data block is allocated yet, ptr is 0.
}

//------------------------------------------------------------------------------
//  constructor for class McImage  --  broadside load
//------------------------------------------------------------------------------

McImage::McImage(unsigned int * Base, size_t sizeInInts)
    :  DataPtr(Base), DataSizeWds(sizeInInts), HeaderData(0), IHeader(*this)
{
    // Broadside Load
}

//------------------------------------------------------------------------------
//  constructor for class McImage  --  copy ctor
//------------------------------------------------------------------------------

McImage::McImage(const McImage & other)
    :  DataPtr(other.DataPtr), DataSizeWds(other.DataSizeWds),
       HeaderData(other.HeaderData), IHeader(*this)
{
}

//------------------------------------------------------------------------------
//  destructor for class McImage
//------------------------------------------------------------------------------

McImage::~McImage()
{
}

//------------------------------------------------------------------------------
//  McImage::operator =() --  Assign to buffer
//------------------------------------------------------------------------------

McImage &  McImage::operator = (const McImage & other)
{
    McImage tmp(other);

    DataPtr = other.DataPtr;
    DataSizeWds = other.DataSizeWds;
    HeaderData = other.HeaderData;
    return *this;
}

//------------------------------------------------------------------------------
//  McImage::SetImage() --  Load Image position
//------------------------------------------------------------------------------

void  McImage::SetImage(unsigned int * Base, size_t sizeInInts)
{
    DataPtr = Base;
    DataSizeWds = sizeInInts;
}

//------------------------------------------------------------------------------
//  McImage::SizeInInts() --  Image Size
//------------------------------------------------------------------------------

size_t McImage::SizeInInts() const
{
    return DataSizeWds;
}

//------------------------------------------------------------------------------
//  McImage::SizeInBytes() --  Image Size
//------------------------------------------------------------------------------

size_t McImage::SizeInBytes() const
{
    return SizeInInts() * sizeof(int);
}

//---------------------------------------------------------------------------
//  McImage::CopyTo() -- Copy to data buffer from image
//---------------------------------------------------------------------------

void  McImage::CopyTo(McBuffer & buffer) const
{
    // Header
    McHeaderDatagram BufH(buffer);
    BufH[0] = HeaderData;

    // Body
    ii32 * TargetDataPtr = reinterpret_cast<ii32*>(buffer.Data());

    CopyInt64Buffer(reinterpret_cast<const ii64*>(DataPtr),
        reinterpret_cast<ii64*>(TargetDataPtr),
        static_cast<int>(DataSizeWds * sizeof(ii32) / sizeof(ii64)));
}

//---------------------------------------------------------------------------
//  McImage::CopyFrom() -- assumes enough data in buffer to fill image
//---------------------------------------------------------------------------

void  McImage::CopyFrom(const McBuffer & buffer, size_t offset)
{
    // Header
    McHeaderDatagram BufH(buffer);
    HeaderData = BufH[0];

    // Body
    const ii32 * SourceDataPtr = reinterpret_cast<const ii32*>(buffer.ConstData());
    SourceDataPtr += offset;

    CopyInt64Buffer(reinterpret_cast<const ii64*>(SourceDataPtr),
        reinterpret_cast<ii64*>(DataPtr),
        static_cast<unsigned int>(DataSizeWds * sizeof(ii32) / sizeof(ii64)));
}

//===========================================================================
//  CLASS McBufferManager  -- Management of McBuffers
//===========================================================================
//---------------------------------------------------------------------------
//  McBufferManager::NativeBuffer() -- Return properly typed Vita Buffer
//---------------------------------------------------------------------------

Buffer  McBufferManager::NativeBuffer()
{
    // McBuffer is the right type
    McBuffer Temp;
    return Temp;
}

//---------------------------------------------------------------------------
//  McBufferManager::HeaderSizeInts() -- Return size of native header
//---------------------------------------------------------------------------

unsigned int  McBufferManager::HeaderSizeInts()
{
    // Vita is the right type
    return DefaultMcHeaderSize;
}

//---------------------------------------------------------------------------
//  McBufferManager::TrailerSizeInts() -- Return size of native header
//---------------------------------------------------------------------------

unsigned int  McBufferManager::TrailerSizeInts()
{
    // Vita is the right type
    return DefaultMcTrailerSize;
}

//---------------------------------------------------------------------------
//  McBufferManager::PacketSizeFromHeaderInts() -- Return size from Header
//---------------------------------------------------------------------------

size_t        McBufferManager::PacketSizeFromHeaderInts(const Buffer & buffer)
{
    McBuffer native = Convert<McBuffer>(buffer);
    McHeaderDatagram mh_dg(native);
    return mh_dg.PacketSize();
}

//---------------------------------------------------------------------------
//  McBufferManager::PacketSizeFromMemoryInts() -- Return size from Memory
//---------------------------------------------------------------------------
//  This assumes that the pointer has enough of the header to get the size.
//    for VitaBuffers, this is always true.

size_t        McBufferManager::PacketSizeFromMemoryInts(unsigned int * start_of_header)
{
    McHeaderDatagram mh_dg(start_of_header);
    return mh_dg.PacketSize();
}



//---------------------------------------------------------------------------
}  // namespace Innovative

