//  Buffer_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2010 Innovative Integration Corporation.
//  All Rights Reserved.
//
//---------------------------------------------------------------------------

#ifndef Buffer_MbH
#define Buffer_MbH

#include <AlignedBlock_Mb.h>
#include <AccessDatagrams_Mb.h>
#include <Datagram_Mb.h>
#include <stdexcept>
#include <DataTypes_Mb.h>
#include <HardwareRegister_Mb.h>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//===========================================================================
//  Holding<T>()  --  Template function to calculate size of N Elements of a type
//===========================================================================
//  Use to construct a Buffer holding N elements of a type. 
//  For example, to make a buffer to hold 1000 Complex data, use:
//      Buffer ComplexStore( Holding<Complex>(1000) );
template <class T>
size_t Holding(size_t elements)
{
    size_t count = elements * sizeof(T) / sizeof(int);
    if (count<2) count = 2;  // minimum size of Buffer to be sent out
    return count;
}

//===========================================================================
//  CouldHold<T>()  --  Template function to calculate # of Elements in a block
//===========================================================================
//  Use to calculate number of elements that could fit in a buffer:
//      Buffer CharStore(1000);
//     int bytes = CouldHold<char>(CharStore.SizeInInts()) );
template <class T>
size_t CouldHold(size_t size_in_ints)
{
    size_t elements = size_in_ints * sizeof(int) / sizeof(T);
    return elements;
}

//===========================================================================
//  Convert<T>()  --  Template function to convert buffer types
//===========================================================================
//  Use to convert one flavor of Buffer into another copying ALL sections

template <class OutBufT, class InBufT>
OutBufT Convert(const InBufT & buffer)
{
    InBufT alias(buffer);  // need a scratch buffer
    OutBufT out;
    out.SwapAll(alias);   // out is now shared with original
    return out;
}

//===========================================================================
//  ConvertData<T>()  --  Template function to convert buffer types
//===========================================================================
//  Use to convert one flavor of Buffer into another copying ONLY data

template <class OutBufT, class InBufT>
OutBufT ConvertData(const InBufT & buffer)
{
    InBufT alias(buffer);  // need a scratch buffer
    OutBufT out;
    out.SwapData(alias);  // out's data is now shared with original
    return out;
}

//===========================================================================
//  CLASS Buffer  --  Data buffer class
//===========================================================================

class Buffer : public IDatagrammable
{
public:
    //
    //  Member Functions
    //  ...constructors and destructors
    Buffer();                         // empty buffer
    Buffer(size_t sizeInInts);
    Buffer(size_t dataSizeInInts, size_t headerSizeInInts, size_t trailerSizeInInts );
    Buffer(const Buffer & other);
    virtual ~Buffer();

    Buffer & operator = (const Buffer & other);
    void  AssignData(const Buffer & other);
    //
    //  Accessors and Sizes
    //  .....Data Region
    int *         Data()
                    {
                    MakeUnique();
                    return (DataH.IsNotNull()) ? DataH->Base() : 0 ;
                    }
    const int *   ConstData() const
                    {  return (DataH.IsNotNull()) ? DataH->ConstBase() : 0 ;  }
    size_t        SizeInInts() const;
    size_t        SizeInBytes() const;

    //  .....Header Region
    int *         Header()
                    {
                    MakeUniqueHeader();
                    return (HeaderH.IsNotNull()) ? HeaderH->Base() : 0 ;
                    }
    const int *   ConstHeader() const
                    {  return (HeaderH.IsNotNull()) ? HeaderH->ConstBase() : 0 ;  }
    size_t        HeaderSizeInInts() const;
    size_t        HeaderSizeInBytes() const;

    //  .....Trailer Region
    int *         Trailer()
                    {
                    MakeUniqueTrailer();
                    return (TrailerH.IsNotNull()) ? TrailerH->Base() : 0 ;
                    }
    const int *   ConstTrailer() const
                    {  return (TrailerH.IsNotNull()) ? TrailerH->ConstBase() : 0 ;  }
    size_t        TrailerSizeInInts() const;
    size_t        TrailerSizeInBytes() const;

    size_t        FullSizeInInts() const
                    {  return SizeInInts() + HeaderSizeInInts() + TrailerSizeInInts(); }   
    //
    //  ...operations
    void  Resize(size_t sizeInInts)
            {  DataH.Resize(sizeInInts);  }
    void  ResizeHeader(size_t sizeInInts)
            {  HeaderH.Resize(sizeInInts);  }
    void  ResizeTrailer(size_t sizeInInts)
            {  TrailerH.Resize(sizeInInts);  }

    bool  IsUnique() const
            {  return DataH.IsUnique();  }
    bool  IsUniqueHeader() const
            {  return HeaderH.IsUnique();  }
    bool  IsUniqueTrailer() const
            {  return TrailerH.IsUnique();  }

    //  Unique-ifyers
    void  MakeUnique()
            {  DataH.MakeUnique();  }
    void  MakeUniqueHeader()
            {  HeaderH.MakeUnique();  }
    void  MakeUniqueTrailer()
            {  TrailerH.MakeUnique();  }
    void  MakeAllUnique()
            {  MakeUnique();  MakeUniqueHeader();  MakeUniqueTrailer();  }

    //  Swaps
    void  Swap(Buffer & other)
            {  SwapData(other);  }
    void  SwapHeader(Buffer & other);
    void  SwapData(Buffer & other);
    void  SwapTrailer(Buffer & other);
    void  SwapAll(Buffer & other);
    //
    //  ...IDatagrammable Interface implementation
    virtual size_t          DatagramSize() 
                                {  return SizeInInts();  }
    virtual int *           DatagramBasePtr()
                                {  return const_cast<int *>(ConstData());  }
    virtual bool            MakeWritable();
    //
    //  ..."universal" header cracker methods
    virtual size_t          ReadHeaderField_PacketSize();
    virtual size_t          ReadHeaderField_DataSize()
        {  return ReadHeaderField_PacketSize() - HeaderSizeInInts() - TrailerSizeInInts();  }

protected:
    //
    //  Member Data
    AlignedBlockHandle      DataH;
    AlignedBlockHandle      HeaderH;
    AlignedBlockHandle      TrailerH;
};

//===========================================================================
//  CLASS PmcBuffer  -- "Classic" X3/X5 Buffer Type
//===========================================================================

class PmcBuffer : public Buffer
{
public:
    PmcBuffer();                         // empty buffer
    PmcBuffer(size_t sizeInInts);
    PmcBuffer(const PmcBuffer & other);
    virtual ~PmcBuffer();

    PmcBuffer & operator = (const PmcBuffer & other);
};

//===========================================================================
//  CLASS VeloBuffer  -- X6 "Velocia" Buffer Type
//===========================================================================

class VeloBuffer : public Buffer
{
public:
    VeloBuffer();                         // empty buffer
    VeloBuffer(size_t sizeInInts);
    VeloBuffer(const VeloBuffer & other);
    virtual ~VeloBuffer();

    VeloBuffer & operator = (const VeloBuffer & other);

};

//===========================================================================
//  CLASS VitaBuffer  -- X6 "VITA-49" Buffer Type
//===========================================================================

class VitaBuffer : public Buffer
{
public:
    VitaBuffer();                         // empty buffer
    VitaBuffer(size_t sizeInInts);
    VitaBuffer(const VitaBuffer & other);
    virtual ~VitaBuffer();

    VitaBuffer & operator = (const VitaBuffer & other);

  	virtual size_t    PadBytes();		 // Support for Data Padding
    //
    //  ..."universal" header cracker methods
    virtual size_t    ReadHeaderField_PacketSize();

};

//===========================================================================
//  CLASS MessageDatagram  -- Access buffer data as a message datagram
//===========================================================================

class MessageDatagram : public IDatagram
{
public:
    //
    //  Member Functions
    MessageDatagram(Buffer & buffer) :  FBuffer(buffer)
        {}

    Buffer &  Packet()  {  return FBuffer;  };

protected:
    virtual size_t          DatagramSize() const;
    virtual int *           DatagramBasePtr() const;

private:
    //
    //  Member Data
    Buffer &  FBuffer;

};

//===========================================================================
//  CLASS IVelociaBufferManager  -- "Universal" mgmt of Velocia-type Buffers
//===========================================================================

class IVelociaBufferManager
{
public:
    virtual ~IVelociaBufferManager()  {}

    virtual Buffer        NativeBuffer() = 0;
    virtual unsigned int  HeaderSizeInts() = 0;
    virtual size_t        PacketSizeFromHeaderInts(const Buffer & buffer) = 0;
    virtual size_t        PacketSizeFromMemoryInts(unsigned int * start_of_header) = 0;
    virtual unsigned int  PeripheralIdFromMemory(unsigned int * start_of_header) = 0;
    virtual void          SetPeripheralIdIntoMemory(unsigned int * start_of_header, int pid) = 0;
};

//===========================================================================
//  CLASS PmcBufferManager  -- Management of PMC Velocia Buffers
//===========================================================================

class PmcBufferManager : public IVelociaBufferManager
{
public:
    PmcBufferManager() {};
    virtual ~PmcBufferManager()  {}

    virtual Buffer        NativeBuffer();
    virtual unsigned int  HeaderSizeInts();
    virtual size_t        PacketSizeFromHeaderInts(const Buffer & buffer);
    virtual size_t        PacketSizeFromMemoryInts(unsigned int * start_of_header);
    virtual unsigned int  PeripheralIdFromMemory(unsigned int * start_of_header);
    virtual void          SetPeripheralIdIntoMemory(unsigned int * start_of_header, int pid);
};

//===========================================================================
//  CLASS VeloBufferManager  -- Management of PMC Velocia Buffers
//===========================================================================

class VeloBufferManager : public IVelociaBufferManager
{
public:
    VeloBufferManager() {};
    virtual ~VeloBufferManager()  {}

    virtual Buffer        NativeBuffer();
    virtual unsigned int  HeaderSizeInts();
    virtual size_t        PacketSizeFromHeaderInts(const Buffer & buffer);
    virtual size_t        PacketSizeFromMemoryInts(unsigned int * start_of_header);
    virtual unsigned int  PeripheralIdFromMemory(unsigned int * start_of_header);
    virtual void          SetPeripheralIdIntoMemory(unsigned int * start_of_header, int pid);
};

//===========================================================================
//  CLASS IVitaBufferManager  -- "Universal" mgmt of Vita- type Buffers
//===========================================================================

class IVitaBufferManager
{
public:
    virtual ~IVitaBufferManager()  {}

    virtual Buffer        NativeBuffer() = 0;
    virtual unsigned int  HeaderSizeInts() = 0;
    virtual unsigned int  TrailerSizeInts() = 0;
	virtual unsigned int  FullSizeDelta() {  return HeaderSizeInts() + TrailerSizeInts();  }
    virtual size_t        PacketSizeFromHeaderInts(const Buffer & buffer) = 0;
    virtual size_t        PacketSizeFromMemoryInts(unsigned int * start_of_header) = 0;

};

//===========================================================================
//  CLASS VitaBufferManager  -- Management of PMC Vita Buffers
//===========================================================================

class VitaBufferManager : public IVitaBufferManager
{
public:
    VitaBufferManager() {};
    virtual ~VitaBufferManager()  {}

    virtual Buffer        NativeBuffer();
    virtual unsigned int  HeaderSizeInts();
    virtual unsigned int  TrailerSizeInts();
    virtual size_t        PacketSizeFromHeaderInts(const Buffer & buffer);
    virtual size_t        PacketSizeFromMemoryInts(unsigned int * start_of_header);

};

//===========================================================================
//  CLASS VitaLayout  -- Fixed Size Packet Layout class definition
//===========================================================================

template <std::size_t N>
class VitaLayout
{
	struct VHdr
		{
		unsigned short PktSize;
		unsigned short Msw0;
		unsigned short Sid;
		unsigned short Msw1;
		ii32           Hdr2;
		ii32           Hdr3;
		ii32           Hdr4;
		ii32           Hdr5;
		ii32           Hdr6;
		};

public:
	//  Data Layout
	VHdr  header;
	ii32  data[N];    // fixed-size array of elements of type T
	ii32  trailer;

public:
	// type definitions
	typedef ii32              value_type;
	typedef ii32*             iterator;
	typedef const ii32*       const_iterator;
	typedef ii32&             reference;
	typedef const ii32&       const_reference;
	typedef std::size_t       size_type;
	typedef std::ptrdiff_t    difference_type;

	// iterator support
	iterator        begin()       { return data; }
	const_iterator  begin() const { return data; }
	const_iterator cbegin() const { return data; }

	iterator        end()       { return data+N; }
	const_iterator  end() const { return data+N; }
	const_iterator cend() const { return data+N; }

	// reverse iterator support
	typedef std::reverse_iterator<iterator> reverse_iterator;
	typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
	// workaround for broken reverse_iterator implementations
	// typedef std::reverse_iterator<iterator,T> reverse_iterator;
	// typedef std::reverse_iterator<const_iterator,T> const_reverse_iterator;

	reverse_iterator rbegin() { return reverse_iterator(end()); }
	const_reverse_iterator rbegin() const {
		return const_reverse_iterator(end());
	}
	const_reverse_iterator crbegin() const {
		return const_reverse_iterator(end());
	}

	reverse_iterator rend() { return reverse_iterator(begin()); }
	const_reverse_iterator rend() const {
		return const_reverse_iterator(begin());
	}
	const_reverse_iterator crend() const {
		return const_reverse_iterator(begin());
	}

	// operator[]
	reference operator[](size_type i)
	{
		return data[i];
	}

	const_reference operator[](size_type i) const
	{
		return data[i];
	}

	// at() with range check
	reference at(size_type i) { rangecheck(i); return data[i]; }
	const_reference at(size_type i) const { rangecheck(i); return data[i]; }

	// front() and back()
	reference front()
	{
		return data[0];
	}

	const_reference front() const
	{
		return data[0];
	}

	reference back()
	{
		return data[N-1];
	}

	const_reference back() const
	{
		return data[N-1];
	}

	// size is constant
	static size_type size() { return N; }
	static bool empty() { return false; }
	static size_type max_size() { return N; }
	enum { static_size = N };

	// direct access to data (read-only)
	const ii32* Data() const { return data; }
	ii32* Data() { return data; }

	// use array as C array (direct read/write access to data)
	ii32* c_array() { return data; }

	 // assign one value to all elements
	void assign (const ii32& value) { fill ( value ); }    // A synonym for fill
	void fill   (const ii32& value)
	{
		std::fill_n(begin(),size(),value);
	}

	// check range (may be private because it is static)
	static void rangecheck (size_type i) {
		if (i >= size()) {
			std::out_of_range e("array<>: index out of range");
			throw e;
		}
	}

};

//===========================================================================
//  CLASS VitaCursor  -- Walk through Vita Packets
//===========================================================================

class VitaCursor
{
public:
    typedef ii32            element_type;    // the type of the access element
    typedef ii32 *          element_ptr;     // type of pointer to the element
    //typedef ptrdiff_t     difference_type;
    typedef size_t        size_type;

public:
    //
    //  Constructors and Destructors
    VitaCursor(const IDatagrammable & intf, size_t init_ofs=0)
         : Intf(const_cast<IDatagrammable *>(&intf)),
           PacketSlice(Intf, 0, 7),  HeaderSlice(Intf, 0, 7),
           DataSlice(Intf, 0, 7),    TrailerSlice(Intf, 0, 1),
           Offset(init_ofs)
         {
         init();
         }
    VitaCursor(IDatagrammable & intf, size_t init_ofs=0)
         : Intf(&intf),
           PacketSlice(Intf, 0, 7),  HeaderSlice(Intf, 0, 7),
           DataSlice(Intf, 0, 7),    TrailerSlice(Intf, 0, 1),
           Offset(init_ofs)
         {
         init();
         }
    VitaCursor(IDatagrammable * intf, size_t init_ofs=0)
         : Intf(intf),
           PacketSlice(Intf, 0, 7),  HeaderSlice(Intf, 0, 7),
           DataSlice(Intf, 0, 7),    TrailerSlice(Intf, 0, 1),
           Offset(init_ofs)
         {
         init();
         }
    VitaCursor(const IDatagrammable * intf, size_t init_ofs=0)
         : Intf(const_cast<IDatagrammable *>(intf)),
           PacketSlice(Intf, 0, 7),  HeaderSlice(Intf, 0, 7),
           DataSlice(Intf, 0, 7),    TrailerSlice(Intf, 0, 1),
           Offset(init_ofs)
         {
         init();
         }
    virtual ~VitaCursor()
         {}

    void  advance()
        {
        Offset += VitaSize;
        rebase();  resize();  reslice();
        }

    unsigned int PacketSize()
        {  return static_cast<unsigned int>(VitaSize);  }
    unsigned short  Sid()
        {  return VitaBase[1] & 0xFFFF;  }

    bool StartFrame()
        {  return (VitaBase[3] & 0x00010000) != 0;  }

    bool EndFrame()
        {  return (VitaBase[3] & 0x00020000) != 0;  }

    bool isValid()
        {  return VitaSize!=0;  }

    IDatagrammable &  Packet()   // entire Vita
        {  return PacketSlice;  }
    IDatagrammable &  Header()   // Vita Header
        {  return HeaderSlice;  }
    IDatagrammable &  Data()     // Vita Data
        {  return DataSlice;  }
    IDatagrammable &  Trailer()  // Vita Trailer
        {  return TrailerSlice;  }

protected:
    //
    //  Member Data
    IDatagrammable *          Intf;
    DatagrammableSlice        PacketSlice;
    DatagrammableSlice        HeaderSlice;
    DatagrammableSlice        DataSlice;
    DatagrammableSlice        TrailerSlice;

    size_type                 Offset;
    size_type                 TotalSize;
    element_ptr               VitaBase;
    size_type                 VitaSize;

    void init()
        {
        rebase();  // set vita base ptr

        size_t  size_in_bytes = Intf->DatagramSize()*sizeof(int);
        if (Intf->PadBytes())
            {
            if (size_in_bytes < Intf->PadBytes())
                size_in_bytes = 0;
            else
                size_in_bytes -= Intf->PadBytes();
            }
        TotalSize = size_in_bytes/sizeof(element_type);

        resize();  // set vita size
        reslice();
        }

    void rebase()
        {
        VitaBase =  reinterpret_cast<element_type *>(Intf->DatagramBasePtr());
        VitaBase += Offset;
        }

    void resize()
        {
        if (Offset < TotalSize)
            VitaSize = VitaBase[0] & 0xFFFF;   // extract the size field
        else
            VitaSize = 0;
        }

    void  reslice()
        {
        PacketSlice.Reslice(Offset, VitaSize);
        HeaderSlice.Reslice(Offset, 7);
        DataSlice.Reslice(Offset+7, VitaSize-8);
        TrailerSlice.Reslice(Offset+VitaSize-1, 1);
        }


};


class BmPacket;

//===========================================================================
//  CLASS PacketPtr  -- containter class for BmPacket
//===========================================================================

class PacketPtr
{
    friend class BmPacket;

public:
    PacketPtr() :
        Pointer(NULL),
        SizeInInts (0)
    {}

    //keep public for fast access
    ii32 * Pointer;
    size_t SizeInInts;    //in 32-bit words
};

//===========================================================================
//  CLASS BmPacket  -- hold pointers to packets in busmaster memory
//===========================================================================

class BmPacket
{
public:
    size_t SizeInInts() {  return TopHalf.SizeInInts + BottomHalf.SizeInInts; }

    ii32 PeripheralID()
    {
        VeloBufferManager velo;
        return  velo.PeripheralIdFromMemory(TopHalf.Pointer);
    }

    //keep public for fast access
    PacketPtr TopHalf;
    PacketPtr BottomHalf;
};

//===========================================================================
//  CLASS McBuffer  -- XU "MultiChan" Buffer Type
//===========================================================================

class McBuffer : public Buffer
{
public:
    McBuffer();                         // empty buffer
    McBuffer(size_t sizeInInts);
    McBuffer(const McBuffer & other);
    virtual ~McBuffer();

    McBuffer & operator = (const McBuffer & other);

    virtual size_t    PadBytes();		 // Support for Data Padding
    //
    //  ..."universal" header cracker methods
    virtual size_t    ReadHeaderField_PacketSize();
    virtual size_t    ReadHeaderField_DataSize()    // DataSize==PacketSize
        {  return ReadHeaderField_PacketSize();  }

};

//===========================================================================
//  CLASS McImage  -- XU "MultiChan" Buffer Type
//===========================================================================

class McImage : public IDatagrammable
{
public:
    McImage();                         // empty image
    McImage(unsigned int * Base, size_t sizeInInts);
    McImage(const McImage & other);
    virtual ~McImage();

    McImage & operator = (const McImage & other);
    void      SetImage(unsigned int * Base, size_t sizeInInts);
    void      SetHeader(unsigned int headerData)
                { HeaderData = headerData;  }

    void      Truncate(size_t newSize)
                {
                if (newSize<DataSizeWds)
                    DataSizeWds = newSize;
                }

    void CopyTo(McBuffer& buffer) const;
    void CopyFrom(const McBuffer& buffer, size_t offset=0);

    size_t        SizeInInts() const;
    size_t        SizeInBytes() const;
    //
    //  ..."universal" header cracker methods
    //virtual size_t    ReadHeaderField_PacketSize();

    IDatagrammable & Header()  {  return IHeader; }
    const IDatagrammable & ConstHeader() const  {  return IHeader; }

	IDatagrammable * HeaderPtr()  {  return &IHeader; }
	const IDatagrammable * ConstHeaderPtr() const  {  return &IHeader; }

    unsigned int * RawH() {  return &HeaderData;  }
    unsigned int * ConstRawH() const  {  return const_cast<unsigned int *>(&HeaderData); }
    unsigned int   HeaderSize() const  {  return 1;  }

    //=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
    //  CLASS HeaderDatagrammable  -- Datagrammable Interface for Headers
    //=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
    class HeaderDatagrammable : public IDatagrammable
        {

        public:
            HeaderDatagrammable(McImage & image)
                :   Owner(image)
                {}

            virtual size_t        DatagramSize()
                                {  return Owner.HeaderSize();  }
            virtual int *         DatagramBasePtr()
                                {  return reinterpret_cast<int *>(&Owner.HeaderData);  }
            virtual bool          MakeWritable() {  return false;  }  // data never changes address
            virtual void          Resize(size_t /*size_in_ints*/)
                                {  /* Can't resize */  }
            McImage &  Owner;
        };

    //
    //  'Body' Datagrammable
    virtual size_t      DatagramSize()
                        {  return DataSizeWds;  }
    virtual int *         DatagramBasePtr()
                        {  return reinterpret_cast<int *>(DataPtr);  }
    virtual bool          MakeWritable()  // data never changes address
                        {  return false;  }
    virtual void          Resize(size_t /*size_in_ints*/)
                        {    }  // can't resize the body

protected:
    //
    //  Member Data
    unsigned int *          DataPtr;
    size_t                  DataSizeWds;
    unsigned int            HeaderData;
    HeaderDatagrammable     IHeader;
};


//===========================================================================
//  CLASS McAlert  -- XU "MultiChan" Alert Type
//===========================================================================

class McAlert : public IDatagrammable
{
public:
    static const unsigned int kMaxMultiChanAlertSize = 34;   // 2 "info" words + 32 Payload

    McAlert(unsigned int size = 4)
        : Size(size)
        {
        for (size_t i=0; i<kMaxMultiChanAlertSize; i++)
            Data[i] = 0;
        SetAlertSize(Size);
        }
    McAlert(const McAlert & other)
        {
        Size = other.Size;
        for (size_t i=0; i<kMaxMultiChanAlertSize; i++)
            Data[i] = other.Data[i];
        }

    //
    //  'Body' Datagrammable
    virtual size_t  DatagramSize()
                        {  return Size;  }
    virtual int *   DatagramBasePtr()
                        {  return reinterpret_cast<int *>(Data);  }
    virtual bool    MakeWritable()
                        {  return false;  }
    virtual void    Resize(size_t size_in_ints)
                        {
                        if (size_in_ints<=kMaxMultiChanAlertSize)
                            Size = static_cast<unsigned int>(size_in_ints);
                        else
                            Size = kMaxMultiChanAlertSize;
                        SetAlertSize(Size);
                        }

protected:
    unsigned int            Size;
    unsigned int            Data[kMaxMultiChanAlertSize];

    void    SetAlertSize(unsigned int size)
        {
        // poke
        Data[0] = size;
        }

};

//===========================================================================
//  CLASS McExternalInt  -- XU "MultiChan" External Interrupt Type
//===========================================================================

class McExternalInt : public IDatagrammable
{
public:
    static const int RMaskB =  0;   static const int RMaskW = 12;   // rising mask start bit, size
    static const int FMaskB = 12;   static const int FMaskW = 12;   // falling mask start bit, size

    McExternalInt(unsigned int mask=0)
        :  Data(mask), ExtBits(&Data)
        {
        }
    McExternalInt(const McExternalInt & other)
        : Data(other.Data), ExtBits(&Data)
        {  }
    //~McExternalInt()  {}

    bool  Flag(int idx)
        {        
        return RisingEdgeFlag(idx) || FallingEdgeFlag(idx);
        }

    bool  RisingEdgeFlag(int idx)
        {
        return ExtBits.Bit(RMaskB + idx);
        }
    void  RisingEdgeFlag(int idx, bool state)
        {
        ExtBits.Bit(RMaskB + idx, state);
        }
    bool  FallingEdgeFlag(int idx)
        {
        return ExtBits.Bit(FMaskB + idx);
        }
    void  FallingEdgeFlag(int idx, bool state)
        {
        ExtBits.Bit(FMaskB + idx, state);
        }
    unsigned int  RisingMask()
        {
        return ExtBits.Field(RMaskB, RMaskW);
        }
    void          RisingMask(unsigned int data)
        {
        ExtBits.Field(RMaskB, RMaskW, data);
        }
    unsigned int  FallingMask()
        {
        return ExtBits.Field(FMaskB, FMaskW);
        }
    void          FallingMask(unsigned int data)
        {
        ExtBits.Field(FMaskB, FMaskW, data);
        }
    unsigned int  Value()
        {
        return Data;
        }

    //
    //  'Body' Datagrammable
    virtual size_t        DatagramSize()
                        {  return 1;  }
    virtual int *         DatagramBasePtr()
                        {  return reinterpret_cast<int *>(&Data);  }
    virtual bool          MakeWritable()
                        {  return false;  }
    virtual void          Resize(size_t /*size_in_ints*/)
                        {    }  // can't resize the body

protected:
    unsigned int            Data;
    PtrBitManipulator       ExtBits;
};


//===========================================================================
//  CLASS IMcBufferManager  -- "Universal" mgmt of Mc-type Buffers
//===========================================================================

class IMcBufferManager
{
public:
    virtual ~IMcBufferManager()  {}

    virtual Buffer        NativeBuffer() = 0;
    virtual unsigned int  HeaderSizeInts() = 0;
    virtual unsigned int  TrailerSizeInts() = 0;
    virtual unsigned int  FullSizeDelta() {  return HeaderSizeInts() + TrailerSizeInts();  }
    virtual size_t        PacketSizeFromHeaderInts(const Buffer & buffer) = 0;
    virtual size_t        PacketSizeFromMemoryInts(unsigned int * start_of_header) = 0;

};

//===========================================================================
//  CLASS McBufferManager  -- Management of McBuffers
//===========================================================================

class McBufferManager : public IMcBufferManager
{
public:
    McBufferManager() {}
    virtual ~McBufferManager()  {}

    virtual Buffer        NativeBuffer();
    virtual unsigned int  HeaderSizeInts();
    virtual unsigned int  TrailerSizeInts();
    virtual size_t        PacketSizeFromHeaderInts(const Buffer & buffer);
    virtual size_t        PacketSizeFromMemoryInts(unsigned int * start_of_header);

};

//===========================================================================
//  Buffer Utility Functions --
//===========================================================================

void  ClearHeader(Buffer & buffer);       // fill with 0s
void  ClearTrailer(Buffer & buffer);

void  InitPmcHeader(Buffer & buffer);     // "correct" Header
void  InitHeader(PmcBuffer & buffer);     //  (make size correct and
void  InitHeader(VeloBuffer & buffer);    //     other required fields
void  InitHeader(VitaBuffer & buffer);
void  InitHeader(McBuffer & buffer);

void  InitTrailer(VitaBuffer & buffer);   // "correct" Trailer

void  ConvertHeader(VeloBuffer & buffer);  // if Velo Buffer has wrong type of header, fix it



//===========================================================================
//  CLASS DmaBuffer  -- hold pointers to dma data region
//===========================================================================

class DmaBuffer : public IDatagrammable
{
public:
    DmaBuffer(int *ptr, size_t sz) : _ptr(ptr), _size(sz)
    {}

    ~DmaBuffer()
    {}

    size_t          DatagramSize()
    { return _size; }
    int *           DatagramBasePtr()
    { return _ptr;  }
    bool            MakeWritable()
    { return false; }
    void            Resize(size_t /*size_in_ints*/)
    { }

private:
    int 		 *_ptr;
    size_t        _size;
};

#ifdef __CLR_VER
#pragma managed(pop)
#endif
}  // namespace Innovative

#endif

