//
// Cdce18005_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2003 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <CDCE18005_Mb.h>
#include <SystemSupport_Mb.h>

//---------------------------------------------------------------------------

namespace Innovative
{

//=============================================================================
//  CLASS Cdce18005  --  Programmable MUX object
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class Cdce18005
//-----------------------------------------------------------------------------

const unsigned int  TotalInputs = 3;
const unsigned int  TotalOutputs = 5;
const int           SpiRegisterCount = 7;   // Reg 8 and 9 not cached

//-----------------------------------------------------------------------------
//  constructor for class Cdce18005
//-----------------------------------------------------------------------------

Cdce18005::Cdce18005(IAddressingSpace & space, Cdce18005Registers & regs)
	:  MuxTimebaseDevice(TotalInputs, TotalOutputs),
       SpiReg(space, regs.SpiIntf_Addr),
       FPowered(false), RegisterImage(SpiRegisterCount),
       Debug(false)
{
    const unsigned int  DefaultOutputCfgBits = ocDisabled;
    //  Fill max frequency checker array
    for (unsigned int i=0; i<TotalOutputs; i++)
        {
        OutputConfigField.push_back(DefaultOutputCfgBits);
        }
}

//-----------------------------------------------------------------------------
//  Cdce18005::Init()  --  Initialize Chip
//-----------------------------------------------------------------------------

void Cdce18005::Init()
{
    ReadDataRegisters();   // read initial state from chip
    OnInit();
    Update();    // Update state of the device based on sources
    WriteDataRegisters();
}

//-----------------------------------------------------------------------------
//  Cdce18005::Reapply()  --  Load settings to Chip
//-----------------------------------------------------------------------------

void Cdce18005::Reapply()
{
    OnReapply();
    Update();    // Update state of the device based on sources
    WriteDataRegisters();
}

//-----------------------------------------------------------------------------
//  Cdce18005::Powered()  --  Change Powered state of chip
//-----------------------------------------------------------------------------

void Cdce18005::Powered(bool value)
{
    FPowered = value;
    OnSetPower(FPowered);
    if (FPowered)
        {
        Init();       // Restore complete chip state if powering up
        Reapply();
        }
}

//-----------------------------------------------------------------------------
//  Cdce18005::Port()  --  Write to PLL data port
//-----------------------------------------------------------------------------

void Cdce18005::Port(unsigned char reg, unsigned int value)
{
    std::stringstream msg;
    msg << "Cdce18005::Write 0x" << std::hex << static_cast<int>(value)
        << " to reg " <<  static_cast<int>(reg);
    Debug.Log(msg.str());

    unsigned int SpiValue(0);

    PtrBitManipulator SpiMap(&SpiValue);

    //   Overwrite register area with Reg
    SpiMap.Value(value);       // Spi Register Value (Bits 0-31)
    SpiMap.Field(0,  4, reg);  // Spi Register Addr  (Bits 0-4)

    SpiReg.Value(SpiValue);
    Sleep(1);
}

//-----------------------------------------------------------------------------
//  Cdce18005::Port()  --  Read from PLL data port
//-----------------------------------------------------------------------------

unsigned int Cdce18005::Port(unsigned char reg)
{
    unsigned int SpiValue(0);

    PtrBitManipulator SpiMap(&SpiValue);

    SpiMap.Field(4,  4, reg);     // Spi Register Addr  (Bits 4-7)
    //
    //  Send Read Command
    Port(0xE, SpiValue);
    Sleep(1);
    //
    //  Read data
    SpiValue = SpiReg.Value();

    std::stringstream msg;
    msg << "Cdce18005::Read 0x" << std::hex << SpiValue
        << " from reg " << static_cast<int>(reg);
    Debug.Log(msg.str());

    return SpiValue;
}

//-----------------------------------------------------------------------------
//  Cdce18005::SetMux()  --  Update the Mux setting for a channel
//-----------------------------------------------------------------------------

void Cdce18005::SetMux(unsigned int output, unsigned int mux_setting)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Field(8, 2, mux_setting);     // Output Cfg bits 8-9
        }
}

//-----------------------------------------------------------------------------
//  Cdce18005::SetPhaseAdjust()  --  Update the Phase Adjust setting
//-----------------------------------------------------------------------------

void Cdce18005::SetPhaseAdjust(unsigned int output, unsigned int divider_setting,
                                                    unsigned int phase_setting)
{
    if (!IsValidDivider(divider_setting))
        return;
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Field(10, 7, ToPhaseAdjustInput(divider_setting, phase_setting));     // Output Cfg bits 10-16
        }
}

//-----------------------------------------------------------------------------
//  Cdce18005::SetDividerHdw()  --  Update the Divider setting
//-----------------------------------------------------------------------------

void Cdce18005::SetDividerHdw(unsigned int output, unsigned int divider_setting)
{
    if (!IsValidDivider(divider_setting))
        return;
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Field(17, 7, ToDividerInput(divider_setting));     // Output Cfg bits 17-23
        }
}

//-----------------------------------------------------------------------------
//  Cdce18005::SetOutEnabledHdw()  --  Update the Output Enabled hardware
//-----------------------------------------------------------------------------

void Cdce18005::SetOutEnabledHdw(unsigned int output, bool enabled)
{
    if (output < TotalOutputs)
        {
        if (enabled)
            {
            // restore from cached setting...
            SetOutputConfig(output, OutputConfigField[output] );
            //  the above also disables divider if off
            }
        else
            {
            bool should_cache_setting = false;
            SetOutputConfig(output, ocDisabled, should_cache_setting );
            //  the above also disables divider if off
            }
        }
}

//-----------------------------------------------------------------------------
//  Cdce18005::SetDividerEnable()  --  Update the Divider Enable setting
//-----------------------------------------------------------------------------
//  If divider is disabled, it is Powered OFF and NO Output results

void Cdce18005::SetDividerEnable(unsigned int output, bool state)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Bit(24, state);     // Output Cfg bits 24
        }
}

//-----------------------------------------------------------------------------
//  Cdce18005::SetInputConfig()  --  Update the Input Config for a channel
//-----------------------------------------------------------------------------

void Cdce18005::SetInputConfig(unsigned int cfg_setting)
{
    PtrBitManipulator Reg(&RegisterImage[5]);
    Reg.Field(4, 2, cfg_setting);     // Output Cfg bits 4-5
}

//-----------------------------------------------------------------------------
//  Cdce18005::SetInputBufferConfig()  --  Update the Buffer Config
//-----------------------------------------------------------------------------

void Cdce18005::SetInputBufferConfig(bool dc_terminated,
                               bool primary_term_disabled, bool sec_term_disabled,
                               bool prim_bias, bool sec_bias, bool failsafe  )
{
    PtrBitManipulator Reg5(&RegisterImage[5]);
    Reg5.Bit(10, dc_terminated);          // DC/AC termination
    Reg5.Bit(11, true);                   // Hysteresis - forced true
    Reg5.Bit(12, primary_term_disabled);  // Primary Termination disabled
    Reg5.Bit(13, prim_bias);              // Primary Neg Pin Bias
    Reg5.Bit(14, sec_bias);               // Secondary Neg Pin Bias
    Reg5.Bit(15, failsafe);               // Failsafe

    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Bit(16, sec_term_disabled);      // Secondary Termination disabled
}

//-----------------------------------------------------------------------------
//  Cdce18005::SetOutputDividerSync()  --  Update the Output Divider Sync
//-----------------------------------------------------------------------------

void Cdce18005::SetOutputDividerSync(unsigned int setting)
{
    PtrBitManipulator Reg(&RegisterImage[5]);
    Reg.Field(6, 2, setting);     // Output Cfg bits 6-7
}

//-----------------------------------------------------------------------------
//  Cdce18005::SetHighSwingLvpecl()  --  Update the boosted LvPecl state
//-----------------------------------------------------------------------------

void Cdce18005::SetHighSwingLvpecl(unsigned int output, bool setting)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Bit(25, setting);     // Output Cfg bit 25
        }
}

//-----------------------------------------------------------------------------
//  Cdce18005::SetOutputConfig()  --  Update the Output Config for a channel
//-----------------------------------------------------------------------------

void Cdce18005::SetOutputConfig(unsigned int output, unsigned int cfg_setting,
                                bool should_cache_setting)
{
    if (output < TotalOutputs)
        {
        //  Cache restore-to setting
        if (should_cache_setting)
            OutputConfigField[output] = cfg_setting;

        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Field(26, 6, cfg_setting);     // Output Cfg bits 26-31

        //  By default, enable channel divider if output is enabled
        if (cfg_setting==ocDisabled)
            SetDividerEnable(output, false);
        else
            SetDividerEnable(output, true);
        }
}

//-----------------------------------------------------------------------------
//  Cdce18005::SetAuxOutputConfig()  --  Configure Aux Output
//-----------------------------------------------------------------------------

void Cdce18005::SetAuxOutputConfig(bool enable, unsigned int aux_setting)
{
    PtrBitManipulator Reg(&RegisterImage[6]);
    Reg.Bit(28, enable);
    Reg.Bit(29, aux_setting ? true : false);
}

//-----------------------------------------------------------------------------
//  Cdce18005::ReadDataRegisters()  --  Read all data registers into Image vector
//-----------------------------------------------------------------------------

void Cdce18005::ReadDataRegisters()
{
    for (unsigned int i=0; i<RegisterImage.size(); i++)
        {
        RegisterImage[i] = Port(i);
        }
}

//-----------------------------------------------------------------------------
//  Cdce18005::WriteDataRegisters()  --  Write all data registers from Image vector
//-----------------------------------------------------------------------------

void Cdce18005::WriteDataRegisters()
{
    WriteMustBeValues();
    for (unsigned int i=0; i<RegisterImage.size(); i++)
        {
        Port(i, RegisterImage[i]);
        }
}

//-----------------------------------------------------------------------------
//  Cdce18005::WriteMustBeValues()  --  Assure that "must be set" values are good
//-----------------------------------------------------------------------------

void Cdce18005::WriteMustBeValues()
{
    for (unsigned int output=0; output<TotalOutputs; output++)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Field(4, 4, 0);     // Bits 4-7 must be 0
        }
    //  Register 5 "Must Be"s
    PtrBitManipulator Reg5(&RegisterImage[5]);
    Reg5.Field(8, 2, 0x1);     // Bits 8-9 must be 0x1
    Reg5.Field(16, 16, 0);     // Bits 16-31 must be 0
    //  Register 5 "Must Be"s
    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Field(0, 12, 0);     // Bits 0-11 must be 0
    Reg6.Field(17, 11, 0);     // Bits 17-27 must be 0
    Reg6.Field(30, 2, 0);     // Bits 30-31 must be 0
}


} // namespace Innovative
