//
// Cdce62005_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2012 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <CDCE62005_Mb.h>
#include <SystemSupport_Mb.h>
#include <VcxoControl_Mb.h>
#include <algorithm>
#include <limits>
#include <cmath>
#include <Magic_Mb.h>
#include <Trace_Mb.h>
#include <sstream>

using namespace std;

#ifdef __GNUC__
    #pragma GCC diagnostic ignored "-Wunused-but-set-variable"
#endif

namespace Innovative
{

//=============================================================================
//  CLASS Cdce62005  --  Cdce62005 Pll Implementation
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class Cdce62005
//-----------------------------------------------------------------------------

const unsigned int  TotalReferences  = 3;    // four?
const unsigned int  TotalOutputs     = 5;
const int           SpiRegisterCount = 8;    // Skip Register 8 R/W
const float         DefaultMaxOutFreq = 1500.0e6;       // do these still apply?
const double        MaxDividerAbove1Freq = 1175.0e6;    // do these still apply?

const double        kVco_InternalRange_Min       =  1750.0e6;
const double        kVco_InternalRange_Crossover =  2043.0e6;  // below is low scale, above high scale
const double        kVco_InternalRange_Max       =  2356.0e6;

const unsigned int  kVco_Prescaler_Min   =  2;
const unsigned int  kVco_Prescaler_Max   =  5;


#ifdef LINUX
//    #pragma GCC diagnostic push
    #pragma GCC diagnostic ignored "-Wunused-but-set-variable"
#endif

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif
//-----------------------------------------------------------------------------
//  constructor for class Cdce62005
//-----------------------------------------------------------------------------

Cdce62005::Cdce62005(IAddressingSpace & space, Cdce62005Registers & regs)
    : PllTimebaseDevice(TotalReferences, TotalOutputs),
      FReferenceSource(refPrimary),
      FOutputSource(fromPll),
      RegisterImage(SpiRegisterCount),
      LocalVcxo(*this),
      ControlReg(space, regs.Control_Addr),
      StatusReg(space, regs.Control_Addr),
      SpiReg(space, regs.SpiIntf_Addr),
      FPowered(false), FVcoFrequency( kVco_InternalRange_Min/kVco_Prescaler_Max ),
      Debug(true)
{
    const unsigned int  DefaultOutputCfgBits = ocDisabled;
    //  Fill max frequency checker array
    for (unsigned int i=0; i<TotalOutputs; i++)
        {
        OutputFrequencyLimit.push_back(DefaultMaxOutFreq);
        OutputConfigField.push_back(DefaultOutputCfgBits);
        }

    InitDividerSupport();
}

#ifdef _MSC_VER
	#pragma warning(pop)
#endif

//-----------------------------------------------------------------------------
//  destructor for class Cdce62005
//-----------------------------------------------------------------------------

Cdce62005::~Cdce62005()
{
}

//-----------------------------------------------------------------------------
//  Cdce62005::Init()  --  Initialize Chip
//-----------------------------------------------------------------------------

void Cdce62005::Init()
{
    ControlReg.UsePrimaryReference.Value(true);
    ControlReg.SpiEnable.Value(true);
    ControlReg.ResetOff.Value(true);

    ReadDataRegisters();   // read initial state from chip
    SettingsInit();        // common INIT settings
    OnInit();
    Update();    // Update state of the device based on sources
    WriteDataRegisters();
}

//-----------------------------------------------------------------------------
//  Cdce62005::SettingsInit()  --  Initial Settings Loading
//-----------------------------------------------------------------------------
//   Note board specific implementations can override these in OnInit

void Cdce62005::SettingsInit()
{
    //  Since Shant Broke his test board readings, hack in correct defaults
    RegisterImage[0] = 0x81840320;
    RegisterImage[1] = 0x81840300;
    RegisterImage[2] = 0x81860300;
    RegisterImage[3] = 0xEB860300;
    RegisterImage[4] = 0x01860310;
    RegisterImage[5] = 0x101C0BE0;
    RegisterImage[6] = 0x04BE19A0;
    RegisterImage[7] = 0xBD0037F0;
    Debug.Log("Shant Busted Values Fixed Up");
}

//-----------------------------------------------------------------------------
//  Cdce62005::Reapply()  --  Reapply settings to hardware from saved state
//-----------------------------------------------------------------------------

void Cdce62005::Reapply()
{
    OnReapply();

    Update();    // Update state of the device based on sources

    ControlReg.Apply();

    WriteDataRegisters();

    Vco_Calibrate();
    SignalSync();
}

//-----------------------------------------------------------------------------
//  Cdce62005::Powered()  --  Change Powered state of chip
//-----------------------------------------------------------------------------

void Cdce62005::Powered(bool value)
{
    if (value==FPowered)
        return;

    FPowered = value;
    OnSetPower(FPowered);
    if (FPowered)
        {
        Init();       // Restore complete chip state
        Reapply();
        }
}

//-----------------------------------------------------------------------------
//  Cdce62005::SignalSync()  --  Use Sync to apply register updates
//-----------------------------------------------------------------------------

void Cdce62005::SignalSync()
{
    ControlReg.PllSync.Set();
    ControlReg.PllSync.Clear();
    ControlReg.PllSync.Set();
}

//-----------------------------------------------------------------------------
//  Cdce62005::MaxOutputDivider()  --  Return max output post-divider
//-----------------------------------------------------------------------------

int Cdce62005::MaxOutputDivider() const
{
    return 80;
}

//-----------------------------------------------------------------------------
//  Cdce62005::FrequencyModeUpdate()  --  Update using Frequency as baseline
//-----------------------------------------------------------------------------

void Cdce62005::FrequencyModeUpdate()
{
    if (Reference() == 0 || Frequency() == 0)
        return;
    //
    //   Phase Detector Freq = Reference/R*M = Vco/FB*X*N

    unsigned int R_P, M, I, FB_A, FB_B, OD;
    //
    //  1) Calculate Dividers to reduce Reference Frequency to Phase Detector
    //        Frequency
    CalculateRPMI(R_P, M, I);
    double Pfd_Actual = Reference()/R_P/M/I;
    //
    //  2) Calculate Best Output Divider value (OD) & VcoFreq
    //        and Calculate FB_A and FB_B Values
    double VcoFreq;
    CalculateOD_FB(OD, FB_A, FB_B, VcoFreq, Pfd_Actual);

    //
    //   Recalculate Vco Freq
    VcoFreq = FB_A * FB_B * Reference() / (R_P * M * I);

    Debug.Log("Cdce62005::R_P = ", R_P);
    Debug.Log("Cdce62005::M = ", M);
    Debug.Log("Cdce62005::I = ", I);
    Debug.Log("Cdce62005::FB_A = ", FB_A);
    Debug.Log("Cdce62005::FB_B = ", FB_B);
    Debug.Log("Cdce62005::Output Divider = ", OD);

    Debug.Log("VcoFreq = ", VcoFreq);

    Debug.Log("Pll Tuning Event");
    Cdce62005PllTuningEvent e(R_P, M, I, FB_A, FB_B, OD);
    OnPllTuning.Execute(e);

    Debug.Log("Final Values");
    Debug.Log("Cdce62005::R_P = ", R_P);
    Debug.Log("Cdce62005::M = ", M);
    Debug.Log("Cdce62005::I = ", I);
    Debug.Log("Cdce62005::FB_A = ", FB_A);
    Debug.Log("Cdce62005::FB_B = ", FB_B);
    Debug.Log("Cdce62005::Output Divider = ", OD);
    //
    //   Recalculate Vco Freq (in case changes were made)
    VcoFreq = FB_A * FB_B * Reference() / (R_P * M * I);
    //
    //  Update registers & allied hardware
    Vcxo().SetVcoFrequency(VcoFreq);    // reprogram Vcxo

    SetR_P(R_P);
    SetM(M);
    SetI(I);
    SetFB_A(FB_A);
    SetFB_B(FB_B);
    //
    //  Set up the output dividers properly
    SetOutputDividers(VcoFreq, OD);
}

//-----------------------------------------------------------------------------
//  Cdce62005::CalculateRPMI() -- Calculate Input
//-----------------------------------------------------------------------------

void  Cdce62005::CalculateRPMI(unsigned int & R_P, unsigned int & M, unsigned int & I)
{
    double CandidateRPMI = Reference()/PhaseDetectorFrequency();
    double CandidatePMI;
    bool   UsingAuxReference = ReferenceSource()==refAux;

    // if the rate at M is too high, or the total is > than M and I can divide (M=8, I=256)
    //    R must be 2.  Otherwise make it 1.
    if (UsingAuxReference)
        {
        R_P = 1;  //  R_P and M unusable on AUX reference path
        }
    else if (Reference()>MaxMInputFrequency() || CandidateRPMI > 256*8)
        {
        R_P = 2;
        CandidatePMI = CandidateRPMI/2;
        }
    else
        {
        R_P = 1;
        }
    //
    //  Distribute residual between I and M such that M is minimized, I maximized
    //    Grade it such that it should produce the closest fit to the PFD frequency
    //
    unsigned int  Max_M = 8;
    if (UsingAuxReference)
        {
        Max_M = 1;  //  R_P and M unusable on AUX reference path
        }
    const unsigned int  Max_I = 256;

    bool exact_match = false;
    double BestError = std::numeric_limits<double>::max();
    int BestM = -1;
    int BestI = -1;
    double ClosestPfd = Reference();

    //  Scan
    for (unsigned int m=1; m<=Max_M; m++)
        {
        for (unsigned int i=1; i<=Max_I; i++)
            {
            //
            //  Divider is product of I and M
            int divider = m * i;

            double candidatePfd = Reference()/R_P/divider;
            double PfdError = std::fabs(candidatePfd-PhaseDetectorFrequency());

            if (PfdError == 0.0)
                {
                exact_match = true;
                BestM = m;
                BestI = i;
                BestError = PfdError;
                ClosestPfd = candidatePfd;
                break;
                }
            else
                {
                if (PfdError < BestError)
                    {
                    BestM = m;
                    BestI = i;
                    BestError = PfdError;
                    ClosestPfd = candidatePfd;
                    }
                }
            }
        if (exact_match)    // if we've found it, leave
            break;
        }
        //  Best match dividers have been found...'return' 'em in arguments
        M = BestM;
        I = BestI;
}

//-----------------------------------------------------------------------------
//  Cdce62005::CalculateOD_FB() -- Calculate Output Divider
//-----------------------------------------------------------------------------

void  Cdce62005::CalculateOD_FB(unsigned int & OD, unsigned int & FB_A,
                            unsigned int & FB_B, double & VcoFreq, double Pfd )
{
    //
    //  Find Vco 'cap' frequency
    //      ...Dividers on this part cannot be above 1
    //         at frequencies above MaxDividerAbove1Freq due to hardware limitations
    //         so we clip the Max Vco Frequency
    double VcoCapFrequency;
    if (Frequency() < MaxDividerAbove1Freq)
        VcoCapFrequency = std::min( Vcxo().Max(), MaxDividerAbove1Freq );
    else
        VcoCapFrequency = Vcxo().Max();  // Not capped, divider = 1
    //
    //  Find divider to reduce Vco to Output Frequency (largest first)
    int   CandidateOD = 0;
    double BestError = std::numeric_limits<double>::max();
    int BestOD   = -1;
    int BestFB_A = -1;
    int BestFB_B = -1;
    double ClosestVcoFreq = VcoFreq;
    bool  TuneMatchFound(false);

    for (ValidDividerArray::reverse_iterator div = CdceDividerValues.rbegin();
             div !=  CdceDividerValues.rend();  ++div)
        {
        CandidateOD = *div;
        double candidateVcoFreq = CandidateOD*Frequency();
        if (candidateVcoFreq < VcoCapFrequency)
            {
            //  We have a candidate Divider. Check if we can reset to this freq.
            //     (some Vco's have gaps in the range).
            if (Vcxo().CanTuneTo(candidateVcoFreq))
                {
                //
                //  Find the best fit FB dividers for this frequency
                TuneMatchFound = true;

                unsigned int testFB_A;
                unsigned int testFB_B;
                CalculateFB(testFB_A, testFB_B, candidateVcoFreq, Pfd);

                double foundVco = Pfd*testFB_A*testFB_B;
                double VcoError = std::fabs(foundVco-candidateVcoFreq);

                if (VcoError == 0.0)
                    {
                    BestFB_A  = testFB_A;
                    BestFB_B  = testFB_B;
                    BestOD    = CandidateOD;
                    BestError = VcoError;
                    ClosestVcoFreq = foundVco;
                    break;
                    }
                else
                    {
                    if (VcoError < BestError)
                        {
                        BestFB_A  = testFB_A;
                        BestFB_B  = testFB_B;
                        BestOD    = CandidateOD;
                        BestError = VcoError;
                        ClosestVcoFreq = foundVco;
                        }
                    }
                }
            }
        }
    //
    //  Determine Vco Frequency - if the scan failed
    if (TuneMatchFound)
        {
        //
        //  We have a candidate divider FB;
        VcoFreq = BestOD*Frequency();
        }
    else
        {
        //  We missed finding a VCO Match. Pick a VCO freq we can use
        VcoFreq = Vcxo().BestTunableFrequency(Frequency());
        //  have to figure a reasonable FB value too.
        CandidateOD =  static_cast<int>(VcoFreq / Frequency());
        if (CandidateOD==0)
            CandidateOD=1;
        }

    OD = BestOD;
    FB_A = BestFB_A;
    FB_B = BestFB_B;
    // VcoFreq updated
}

//-----------------------------------------------------------------------------
//  Cdce62005::CalculateFB() -- Calculate FB dividers
//-----------------------------------------------------------------------------

void  Cdce62005::CalculateFB(unsigned int & FB_A, unsigned int & FB_B, double VcoFreq, double Pfd )
{
    //
    //  Distribute residual between FB_A and FB_B such that FB_B is minimized, FB_A maximized
    //    Grade it such that it should produce the closest fit to the PFD frequency
    //
    bool exact_match = false;
    double BestError = std::numeric_limits<double>::max();
    int BestFB_A = -1;
    int BestFB_B = -1;
    double ClosestPfd = VcoFreq;

    //  Scan
    for (unsigned int fb_b=0; fb_b<FB_B_DividerValues.size(); fb_b++)
        {
        for (unsigned int fb_a=0; fb_a<FB_A_DividerValues.size(); fb_a++)
            {
            //
            //  FB Divider is product of FB_A and FB_B
            int divider = FB_B_DividerValues[fb_b] * FB_A_DividerValues[fb_a];

            double candidatePfd = VcoFreq/divider;
            double PfdError = std::fabs(candidatePfd-Pfd);

            if (PfdError == 0.0)
                {
                exact_match = true;
                BestFB_A = FB_A_DividerValues[fb_a];
                BestFB_B = FB_B_DividerValues[fb_b];
                BestError = PfdError;
                ClosestPfd = candidatePfd;
                break;
                }
            else
                {
                if (PfdError < BestError)
                    {
                    BestFB_A = FB_A_DividerValues[fb_a];
                    BestFB_B = FB_B_DividerValues[fb_b];
                    BestError = PfdError;
                    ClosestPfd = candidatePfd;
                    }
                }
            if (candidatePfd < Pfd)    // if the divider is already too large, skip the rest
                break;
            }
        if (exact_match)    // if we've found it, leave
            break;
        }
    //  Best match dividers have been found...'return' 'em in arguments
    FB_A = BestFB_A;
    FB_B = BestFB_B;
}

//-----------------------------------------------------------------------------
//  Cdce62005::DividerModeUpdate() -- Update state of dividers
//-----------------------------------------------------------------------------
//  In this mode, the reference is sent to the outputs, with possible division

void  Cdce62005::DividerModeUpdate()
{
    if (Reference() == 0 || Frequency() == 0)
        return;

    double div = 1.0;
    double fraction = modf(Reference()/Frequency(), &div);
    div += (fraction > 0.5) ? 1.0 : 0.0;
    unsigned int OutDiv = static_cast<unsigned int>(div);
    OutDiv = std::max<unsigned int>(OutDiv, 1);

    SetOutputDividers(Reference(), OutDiv);
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetOutputDividers()  --
//-----------------------------------------------------------------------------

void  Cdce62005::SetOutputDividers(double VcoFreq, unsigned int OutDiv)
{
    unsigned int BestDivider = OutDiv;
    double       OutputFreq  = VcoFreq/BestDivider;

    for (unsigned int output=0; output<TotalOutputs; output++)
        {
        if (OutputInUse(output)==true)
            {
            float   CandidateDivider = static_cast<float>(BestDivider);
            double  CandidateFreq    = OutputFreq;
            //
            //  If the Output is the Master, use the best values.
            //     if not the master, and Frequency is unspecified, then
            //     the make the output a copy of Master.
            //     Otherwise, figure the best divider possible for each output
            if ((output!=FMasterOutput) && (OutputFrequency(output)!=0.0))
                {
                CandidateFreq = OutputFrequency(output);
                CandidateDivider = static_cast<float>(VcoFreq / CandidateFreq);

                int above, below, divider;
                GetNearestDividers(output, CandidateDivider, below, above);

                // Find which divider is closest to desired frequency
                double centerFreq = VcoFreq/CandidateDivider;
                double hiFreq     = VcoFreq/below;
                double lowFreq    = VcoFreq/above;
                if (std::abs(centerFreq-lowFreq) > std::abs(hiFreq-centerFreq))
                    divider = below;
                else
                    divider = above;

                CandidateDivider = static_cast<float>(divider);
                }

            //
            //  Check for frequency exceeding applied Limits
            if (OutputFrequencyLimit[output]>=CandidateFreq)
                OutputDivider(output, static_cast<unsigned int>(CandidateDivider));
            else
                {
                double MinDivideRatio = VcoFreq/OutputFrequencyLimit[output];
                OutputDivider(output, BestDividerGreaterThan(static_cast<float>(MinDivideRatio)));
                }
            //
            //  Calculate Acutal Frequency
            if (OutputDividerArray[output]!=0)  // if illegal divider, skip it
                {
                OutputFreqActualArray[output] = VcoFreq / OutputDividerArray[output];
                }
            //
            //  If Divider 'needed' is not allowed because of rate limitations,
            //     disable the output, otherwise enable it
            if (VcoFreq > MaxDividerAbove1Freq && OutputDivider(output)!=1)
                {
                SetDividerEnable(output, false);
                OutputFreqActualArray[output] = 0.0;  // no output allowed
                }
            else
                SetDividerEnable(output, true);
            }
        else
            {
            OutputFreqActualArray[output] = 0.0;
            }
        }
}

//-----------------------------------------------------------------------------
//  Cdce62005::Locked()  --  Return PLL lock status
//-----------------------------------------------------------------------------

bool  Cdce62005::Locked() const
{
    return StatusReg.Locked() ? true : false;
}

//-----------------------------------------------------------------------------
//  Cdce62005::Port()  --  Write to PLL data port
//-----------------------------------------------------------------------------

void Cdce62005::Port(unsigned char reg, unsigned int value)
{
    std::stringstream msg;
    msg << "Cdce62005::Write 0x" << std::hex << static_cast<int>(value)
        << " to reg " <<  static_cast<int>(reg);
    Debug.Log(msg.str());

    unsigned int SpiValue(0);

    PtrBitManipulator SpiMap(&SpiValue);

    //   Overwrite register area with Reg
    SpiMap.Value(value);       // Spi Register Value (Bits 0-31)
    SpiMap.Field(0,  4, reg);  // Spi Register Addr  (Bits 0-4)

    SpiReg.Value(SpiValue);
    Sleep(1);
}

//-----------------------------------------------------------------------------
//  Cdce62005::Port()  --  Read from PLL data port
//-----------------------------------------------------------------------------

unsigned int Cdce62005::Port(unsigned char reg)
{
    unsigned int SpiValue(0);

    PtrBitManipulator SpiMap(&SpiValue);

    SpiMap.Field(4,  4, reg);     // Spi Register Addr  (Bits 4-7)
    //
    //  Send Read Command
    Port(0xE, SpiValue);
    Sleep(1);
    //
    //  Read data
    SpiValue = SpiReg.Value();

    std::stringstream msg;
    msg << "Cdce62005::Read 0x" << std::hex << SpiValue
        << " from reg " << static_cast<int>(reg);
    Debug.Log(msg.str());

    return SpiValue;
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetPhaseAdjust()  --  Update the Phase Adjust setting
//-----------------------------------------------------------------------------

void Cdce62005::SetPhaseAdjust(unsigned int output, unsigned int divider_setting,
                                                    unsigned int phase_setting)
{
    if (!IsValidDivider(divider_setting))
        return;

    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Field(10, 7, ToPhaseAdjustInput(divider_setting, phase_setting));     // Output Cfg bits 10-16
        }
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetDividerHdw()  --  Update the Divider setting hardware
//-----------------------------------------------------------------------------

void Cdce62005::SetDividerHdw(unsigned int output, unsigned int divide_by)
{
    if (!IsValidDivider(divide_by))
        return;

    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Field(17, 7, ToDividerInput(divide_by));     // Output Cfg bits 17-23
        }
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetOutEnabledHdw()  --  Update the Output Enabled hardware
//-----------------------------------------------------------------------------

void Cdce62005::SetOutEnabledHdw(unsigned int output, bool enabled)
{
    if (output < TotalOutputs)
        {
        if (enabled)
            {
            // restore from cached setting...
            SetOutputConfig(output, OutputConfigField[output] );
            //  the above also disables divider if off
            }
        else
            {
            bool should_cache_setting = false;
            SetOutputConfig(output, ocDisabled, should_cache_setting );
            //  the above also disables divider if off
            }
        }
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetDividerEnable()  --  Update the Divider Enable setting
//-----------------------------------------------------------------------------

void Cdce62005::SetDividerEnable(unsigned int output, bool state)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Bit(24, state);     // Output Cfg bits 24
        }
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetPllEnable()  --  Enable PLL (Power Down if disabled)
//-----------------------------------------------------------------------------
//
//void Cdce62005::SetPllEnable(bool enabled)
//{
//    PtrBitManipulator Reg11(&RegisterImage[11]);
//    Reg11.Bit(23, enabled==false);     // 0 = enabled, 1=Powered Down
//}
//

//-----------------------------------------------------------------------------
//  Cdce62005::SetInputConfig()  --  Update the Input Config the references
//-----------------------------------------------------------------------------

void Cdce62005::SetInputConfig(unsigned int cfg_setting)
{
    PtrBitManipulator Reg(&RegisterImage[5]);
    Reg.Field(4, 2, cfg_setting);     // Input Cfg bits 4-5
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetInputBufferConfig()  --  Update the Buffer Config
//-----------------------------------------------------------------------------

void Cdce62005::SetInputBufferConfig(bool dc_terminated,
                               bool primary_input_term_disabled, bool secondary_input_term_disabled,
                               bool prim_bias, bool sec_bias, bool failsafe  )
{
    PtrBitManipulator Reg5(&RegisterImage[5]);
    Reg5.Bit(10, dc_terminated);                   // DC/AC termination
    Reg5.Bit(11, true);                            // Hysteresis - forced true
    Reg5.Bit(12, primary_input_term_disabled);     // Input Termination disabled
    Reg5.Bit(13, prim_bias);                       // Primary Neg Pin Bias
    Reg5.Bit(14, sec_bias);                        // Secondary Neg Pin Bias
    Reg5.Bit(15, failsafe);                        // Failsafe

    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Bit(16, secondary_input_term_disabled);   // Input Termination disabled

}

//-----------------------------------------------------------------------------
//  Cdce62005::SetReferenceSource()  --  Set the Reference Source
//-----------------------------------------------------------------------------

void Cdce62005::SetReferenceSource(CdReferenceSource source)
{
    // Conform to one of three possible physical references
    unsigned int source_no;
    switch (source)
        {
        default:
        case refNone:
            source_no = 0;
            break;
        case refPrimary:
            source_no = 0;
            break;
        case refSecondary:
            source_no = 1;
            break;
        case refAux:
            source_no = 2;
            break;
        case refAutoSelectRef:
            source_no = 0;
            break;
        case refAutoSelectAll:
            source_no = 0;
            break;
        }
    PtrBitManipulator Reg(&RegisterImage[5]);
    Reg.Field(6, 3, source);     // Output Cfg bits 6-8

    SetSourceSelection(source_no);   // inform base class too
    SetOutputSource(FOutputSource);  // make sure Output Mux is set correctly
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetOutputSource()  --  Sert the Output Mode (and Mux outputs)
//-----------------------------------------------------------------------------

void Cdce62005::SetOutputSource(CdOutputSource source)
{
    //
    //  decide on which internal clock soure to direct
    unsigned int channel_mux;
    if (source==fromPll)
        {
        channel_mux = osVco;    //  Output stage driven by VCO frequency
        }
    else
        {
        if (FReferenceSource == refAux)
            channel_mux = osSmartMux;
        else if (FReferenceSource == refSecondary)
            channel_mux = osSecondaryRef;
        else // (FReferenceSource == refPrimary) ..and switching modes too
            channel_mux = osPrimaryRef;
        }

    for (unsigned int output=0; output < TotalOutputs; output++)
        {
        SetOutputMuxSource(output, channel_mux);
        }
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetHighSwingLvpecl()  --  Update the boosted LvPecl state
//-----------------------------------------------------------------------------

void Cdce62005::SetHighSwingLvpecl(unsigned int output, bool setting)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Bit(25, setting);     // Output Cfg bit 25
        }
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetOutputConfig()  --  Update the Output Config for a channel
//-----------------------------------------------------------------------------

void Cdce62005::SetOutputConfig(unsigned int output, unsigned int cfg_setting,
                                bool should_cache_setting)
{
    if (output < TotalOutputs)
        {
        //  Cache restore-to setting
        if (should_cache_setting)
            OutputConfigField[output] = cfg_setting;

        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Field(26, 6, cfg_setting);     // Output Cfg bits 26-31

        //  By default, enable channel divider if output is enabled
        if (cfg_setting==ocDisabled)
            SetDividerEnable(output, false);
        else
            SetDividerEnable(output, true);
        }
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetLockConfig()  --  Lock options
//-----------------------------------------------------------------------------

void Cdce62005::SetLockConfig(bool analog_lock)
{
    PtrBitManipulator Reg(&RegisterImage[5]);
    Reg.Bit(31, analog_lock);
}

//------------------------------------------------------------------------------
//  Cdce62005::SetExternalReferenceControl()  --  Define who controls Ref select
//------------------------------------------------------------------------------

void Cdce62005::SetExternalReferenceControl(bool useInternalSelect)
{
    PtrBitManipulator Reg(&RegisterImage[5]);
    Reg.Bit(9, useInternalSelect);
}

//------------------------------------------------------------------------------
//  Cdce62005::SetExternalReferenceSelect()  --  Define who controls Ref select
//------------------------------------------------------------------------------

void Cdce62005::SetExternalReferenceSelect(bool usePrimaryReference)
{
    ControlReg.UsePrimaryReference.Value(usePrimaryReference);
}

//------------------------------------------------------------------------------
//  Cdce62005::SetOutputMuxSource()  --  Set source clock for an output
//------------------------------------------------------------------------------

void Cdce62005::SetOutputMuxSource(unsigned int output, unsigned int source)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Field(8, 2, source);     // Output Cfg bit 25
        }
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetAuxOutputConfig()  --  Configure Aux Output
//-----------------------------------------------------------------------------

void Cdce62005::SetAuxOutputConfig(bool enable, unsigned int aux_setting)
{
    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Bit(28, enable);
    Reg6.Bit(29, aux_setting ? true : false);
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetVcoRange()  --  Configure Vco Range
//-----------------------------------------------------------------------------

void Cdce62005::SetVcoRange(bool isHighRange)
{
    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Bit(4, isHighRange);
}

//-----------------------------------------------------------------------------
//  Cdce62005::GetDivider()  --  read back the Divider setting
//-----------------------------------------------------------------------------

unsigned int  Cdce62005::GetDivider(unsigned int output)
{
    unsigned int divider(1);
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        unsigned int div_field = Reg.Field(17, 7);
        divider = FromDividerInput(div_field);
        }
    return divider;
}

//-----------------------------------------------------------------------------
//  Cdce62005::GetDividerEnable()  --  Read the Divider Enable setting
//-----------------------------------------------------------------------------

bool Cdce62005::GetDividerEnable(unsigned int output)
{
    bool bit = false;
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        bit = Reg.Bit(24);     // Output Cfg bits 24
        }
    return bit;
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetM()  --  Configure M  (Reference Divider)
//-----------------------------------------------------------------------------
#ifdef _MSC_VER
    #pragma warning(push)
    #pragma warning(disable : 4800)
#endif

void Cdce62005::SetM(unsigned int M)
{
    PtrBitManipulator Reg2(&RegisterImage[2]);
    PtrBitManipulator Reg3(&RegisterImage[3]);

    unsigned int regval = M-1;  // offset by 1

    Reg2.Field(4, 2, regval);
    Reg3.Bit(4, static_cast<bool>(regval&0x4));      // Bit 2
}

#ifdef _MSC_VER
    #pragma warning(pop)
#endif
//-----------------------------------------------------------------------------
//  Cdce62005::SetI()  --  Configure I  (Input Divider)
//-----------------------------------------------------------------------------

void Cdce62005::SetI(unsigned int I)
{
    PtrBitManipulator Reg5(&RegisterImage[5]);
    Reg5.Field(18, 8, I-1);
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetR_P()  --  Configure R_P  (Reference Prescaler)
//-----------------------------------------------------------------------------

void Cdce62005::SetR_P(unsigned int R)
{
    PtrBitManipulator PrimaryReg(&RegisterImage[0]);
    PtrBitManipulator SecondaryReg(&RegisterImage[1]);

    if (R==2)
        {
        PrimaryReg.Field(4, 2, 1); SecondaryReg.Field(4, 2, 1); // Both PRI/SEC set same
        }
    else
        {
        PrimaryReg.Field(4, 2, 2); SecondaryReg.Field(4, 2, 2); // Both PRI/SEC set same
        }

}

//-----------------------------------------------------------------------------
//  Cdce62005::SetFB_A()  --  Update the Divider setting
//-----------------------------------------------------------------------------

void Cdce62005::SetFB_A(unsigned int FB_A)
{
    if (!IsValid_FB_A_Divider(FB_A))
        return;

    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Field(7, 8, To_FB_A_DividerInput(FB_A));
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetFB_B()  --  Update the Divider setting
//-----------------------------------------------------------------------------

void Cdce62005::SetFB_B(unsigned int FB_B)
{
    if (!IsValid_FB_B_Divider(FB_B))
        return;

    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Field(17, 3, To_FB_B_DividerInput(FB_B));
}

//-----------------------------------------------------------------------------
//  Cdce62005::SetVco_Ps()  --  Update the Vco Prescaler setting
//-----------------------------------------------------------------------------

void Cdce62005::SetVco_Ps(unsigned int Vco_Ps)
{
    if (!IsValid_Vco_Ps_Divider(Vco_Ps))
        return;

    PtrBitManipulator Reg6(&RegisterImage[6]);
    Reg6.Field(5, 2, To_Vco_Ps_DividerInput(Vco_Ps));
}

//-----------------------------------------------------------------------------
//  Cdce62005::GetM()  --  Read M  (Reference Divider)
//-----------------------------------------------------------------------------

unsigned int Cdce62005::GetM()
{
    PtrBitManipulator Reg2(&RegisterImage[2]);
    PtrBitManipulator Reg3(&RegisterImage[3]);
    unsigned int M = Reg2.Field(4, 2);
    if (Reg3.Bit(4))     // apply the most significant bit
        M+=4;
    return M+1;       // offset by 1
}

//-----------------------------------------------------------------------------
//  Cdce62005::GetI()  --  Read I  (Input Divider)
//-----------------------------------------------------------------------------

unsigned int Cdce62005::GetI()
{
    PtrBitManipulator Reg5(&RegisterImage[5]);
    unsigned int I = Reg5.Field(18, 8);
    return I+1;     // offset by 1
}

//-----------------------------------------------------------------------------
//  Cdce62005::GetR_P()  --  Read R_P  (Reference Prescaler)
//-----------------------------------------------------------------------------

unsigned int Cdce62005::GetR_P()
{
    PtrBitManipulator Reg(&RegisterImage[0]);
    unsigned int R_field = Reg.Field(4, 2);   // assume Primary and Secodary are set the same
    if (R_field==1)
        {
        return 2;
        }
    else // (R_field==2)
        {
        return 1;
        }

}

//-----------------------------------------------------------------------------
//  Cdce62005::GetFB_A()  --  Read the Divider setting
//-----------------------------------------------------------------------------

unsigned int Cdce62005::GetFB_A()
{
    PtrBitManipulator Reg6(&RegisterImage[6]);
    unsigned int FB_field = Reg6.Field(7, 8);
    unsigned int FB = From_FB_A_DividerInput(FB_field);
    return FB;
}

//-----------------------------------------------------------------------------
//  Cdce62005::GetFB_B()  --  Read the Divider setting
//-----------------------------------------------------------------------------

unsigned int Cdce62005::GetFB_B()
{
    PtrBitManipulator Reg6(&RegisterImage[6]);
    unsigned int FB_B_field = Reg6.Field(17, 3);
    unsigned int FB_B = From_FB_B_DividerInput(FB_B_field);
    return FB_B;
}

//-----------------------------------------------------------------------------
//  Cdce62005::GetVco_Ps()  --  Read the Divider setting
//-----------------------------------------------------------------------------

unsigned int Cdce62005::GetVco_Ps()
{
    PtrBitManipulator Reg6(&RegisterImage[6]);
    unsigned int Vco_Ps_field = Reg6.Field(5, 2);
    unsigned int Vco_Ps = From_Vco_Ps_DividerInput(Vco_Ps_field);
    return Vco_Ps;
}

//-----------------------------------------------------------------------------
//  Cdce62005::ReadDataRegisters()  --  Read all data registers into Image vector
//-----------------------------------------------------------------------------

void Cdce62005::ReadDataRegisters()
{
    for (unsigned int i=0; i<RegisterImage.size(); i++)
        {
        RegisterImage[i] = Port(i);
        }
}

//-----------------------------------------------------------------------------
//  Cdce62005::WriteDataRegisters()  --  Write all data registers from Image vector
//-----------------------------------------------------------------------------

void Cdce62005::WriteDataRegisters()
{
    ScopedTraceRef ccc(Debug, "Cdce62005::WriteDataRegisters()");

    for (unsigned int i=0; i<RegisterImage.size(); i++)
        {
        Port(i, RegisterImage[i]);
        }
}

//-----------------------------------------------------------------------------
//  Cdce62005::OnSetPower()  --  Perform Pwr Setting operation
//-----------------------------------------------------------------------------

void  Cdce62005::OnSetPower(bool & new_state)
{
    ControlReg.PowerOn.Value(new_state);
    //  pause to allow chip to activate
    Sleep(10);
}

//-----------------------------------------------------------------------------
//  Cdce62005::Vco_Min()  --  Internal Vcxo Device implementation
//-----------------------------------------------------------------------------

double  Cdce62005::Vco_Min()
{
    return kVco_InternalRange_Min / kVco_Prescaler_Max;   // lowest output value
}

//-----------------------------------------------------------------------------
//  Cdce62005::Vco_Max()  --  Internal Vcxo Device implementation
//-----------------------------------------------------------------------------

double  Cdce62005::Vco_Max()
{
    return kVco_InternalRange_Max / kVco_Prescaler_Min;   // highest output value
}

//-----------------------------------------------------------------------------
//  Cdce62005::Vco_Calibrate()
//-----------------------------------------------------------------------------

void Cdce62005::Vco_Calibrate()
{
    ScopedTraceRef ccc(Debug, "Cdce62005::Vco_Calibrate()");

    if (OutputSource()==fromExtClk)
        {
        Debug.Log("External Clock - no calibration");
        return;
        }

    //  Toggle SLEEP to trigger VCO calibration
    Port(0x8, 0x1000);   // SLEEP, notSYNC
    Port(0x8, 0x1800);   // notSLEEP, notSYNC

    // Wait up to 1mS for pending SPI to complete
    for (unsigned int i = 0; i < 5; ++i)
        {
        if (Locked())
            {
            Debug.Log("PLL Locked...");
            return;
            }
        else
            Innovative::uSleep(10000);
        }
    // Gave up
    Debug.Log("PLL Lock Wait Failure (>50ms)");
}

//-----------------------------------------------------------------------------
//  Cdce62005::Vco_CanTuneTo()
//-----------------------------------------------------------------------------

bool Cdce62005::Vco_CanTuneTo(double freq)
{
    bool retValue;

    if (freq >= Vco_Min() && freq <= Vco_Gap0Min())
        retValue = true;
    else if (freq >= Vco_Gap0Max() && freq <= Vco_Max())
        retValue = true;
    else
        retValue = false;

    return retValue;
}

//-----------------------------------------------------------------------------
//  Cdce62005::Vco_BestTunableFrequency()
//-----------------------------------------------------------------------------

double Cdce62005::Vco_BestTunableFrequency(double freq)
{
    double retValue;

    if (freq < Vco_Min())
        retValue = Vco_Min();
    else if (freq >= Vco_Min() && freq <= Vco_Gap0Min())
        retValue = freq;
    else if (freq >= Vco_Gap0Min() && freq <= (Vco_Gap0Min()+Vco_Gap0Max()/2))
        retValue = Vco_Gap0Min();
    else if (freq >= (Vco_Gap0Min()+Vco_Gap0Max()/2) && freq <= Vco_Gap0Max())
        retValue = Vco_Gap0Max();
    else if (freq >= Vco_Gap0Max() && freq <= Vco_Max())
        retValue = freq;
    else
        retValue = Vco_Max();

    return retValue;
}

//-----------------------------------------------------------------------------
//  Cdce62005::Vco_SetVcoFrequency()
//-----------------------------------------------------------------------------

void  Cdce62005::Vco_SetVcoFrequency(double freq)
{
    ScopedTraceRef ccc(Debug, "Cdce62005::Vco_SetVcoFrequency()");

    bool          in_range(false);
    double        BestError = std::numeric_limits<double>::max();
    unsigned int  BestPs = -1;
    double        ClosestFreq = freq;

    FVcoFrequency = freq;

    //  Scan Vco Ps list looking for a match, largest first
    for (unsigned int Ps=0; Ps<Vco_Ps_DividerValues.size(); Ps++)
        {
        //
        //  read dividers candidates, largest first
        int divider = Vco_Ps_DividerValues[Ps];

        double LowerLimit = kVco_InternalRange_Min/divider;
        double UpperLimit = kVco_InternalRange_Max/divider;

        if ((LowerLimit <= freq) && (UpperLimit >= freq))
            {
            in_range = true;
            BestPs = divider;
            break;
            }
        else
            {
            double LowerError = std::fabs(freq-LowerLimit);
            if (LowerError < BestError)
                {
                BestError = LowerError;
                BestPs =    divider;
                ClosestFreq = LowerLimit;
                }

            double UpperError = std::fabs(freq-UpperLimit);
            if (UpperError < BestError)
                {
                BestError = UpperError;
                BestPs = divider;
                ClosestFreq = UpperLimit;
                }
            }
        }

        if (!in_range)
            FVcoFrequency = ClosestFreq;

        //  Check which half of VCO to use
        bool usesHighHalf = (FVcoFrequency*BestPs >= kVco_InternalRange_Crossover);

        //  Program Range Control and Prescaler
        SetVcoRange( usesHighHalf );
        SetVco_Ps(BestPs);

    Debug.Log("VCO Values");
    Debug.Log("Cdce62005::VcoPs = ", BestPs);
    Debug.Log("Cdce62005::Vco Uses High Range = ", (usesHighHalf));
    Debug.Log("Cdce62005::Internal Vco Freq (MHz) = ", FVcoFrequency*BestPs/1.0e6);
    Debug.Log("Cdce62005::Output Vco Freq (MHz) =", FVcoFrequency/1.0e6);
}

//-----------------------------------------------------------------------------
//  Cdce62005::Vco_GetVcoFrequency()  --  read Vco Frequenc7y
//-----------------------------------------------------------------------------

double  Cdce62005::Vco_GetVcoFrequency()
{
    return FVcoFrequency;
}

//-----------------------------------------------------------------------------
//  Cdce62005::InitDividerSupport()  --  Initialize Special Divider Maps
//-----------------------------------------------------------------------------

void  Cdce62005::InitDividerSupport()
{
    //  FB_A is the first feedback divider
    FB_A_DividerValues.push_back(8);     FB_A_DividerValues.push_back(12);
    FB_A_DividerValues.push_back(16);    FB_A_DividerValues.push_back(20);
    FB_A_DividerValues.push_back(24);    FB_A_DividerValues.push_back(32);
    FB_A_DividerValues.push_back(36);    FB_A_DividerValues.push_back(40);
    FB_A_DividerValues.push_back(48);    FB_A_DividerValues.push_back(56);
    FB_A_DividerValues.push_back(60);    FB_A_DividerValues.push_back(64);
    FB_A_DividerValues.push_back(72);    FB_A_DividerValues.push_back(80);
    FB_A_DividerValues.push_back(84);    FB_A_DividerValues.push_back(96);
    FB_A_DividerValues.push_back(100);   FB_A_DividerValues.push_back(108);
    FB_A_DividerValues.push_back(112);   FB_A_DividerValues.push_back(120);
    FB_A_DividerValues.push_back(128);   FB_A_DividerValues.push_back(140);
    FB_A_DividerValues.push_back(144);   FB_A_DividerValues.push_back(160);
    FB_A_DividerValues.push_back(168);   FB_A_DividerValues.push_back(180);
    FB_A_DividerValues.push_back(192);   FB_A_DividerValues.push_back(200);
    FB_A_DividerValues.push_back(216);   FB_A_DividerValues.push_back(224);
    FB_A_DividerValues.push_back(240);   FB_A_DividerValues.push_back(252);
    FB_A_DividerValues.push_back(256);   FB_A_DividerValues.push_back(280);
    FB_A_DividerValues.push_back(288);   FB_A_DividerValues.push_back(300);
    FB_A_DividerValues.push_back(320);   FB_A_DividerValues.push_back(336);
    FB_A_DividerValues.push_back(360);   FB_A_DividerValues.push_back(384);
    FB_A_DividerValues.push_back(392);   FB_A_DividerValues.push_back(400);
    FB_A_DividerValues.push_back(420);   FB_A_DividerValues.push_back(432);
    FB_A_DividerValues.push_back(448);   FB_A_DividerValues.push_back(480);
    FB_A_DividerValues.push_back(500);   FB_A_DividerValues.push_back(504);
    FB_A_DividerValues.push_back(512);   FB_A_DividerValues.push_back(560);
    FB_A_DividerValues.push_back(576);   FB_A_DividerValues.push_back(588);
    FB_A_DividerValues.push_back(600);   FB_A_DividerValues.push_back(640);
    FB_A_DividerValues.push_back(672);   FB_A_DividerValues.push_back(700);
    FB_A_DividerValues.push_back(720);   FB_A_DividerValues.push_back(768);
    FB_A_DividerValues.push_back(784);   FB_A_DividerValues.push_back(800);
    FB_A_DividerValues.push_back(840);   FB_A_DividerValues.push_back(896);
    FB_A_DividerValues.push_back(960);   FB_A_DividerValues.push_back(980);
    FB_A_DividerValues.push_back(1024);  FB_A_DividerValues.push_back(1120);
    FB_A_DividerValues.push_back(1280);

    //  Fill the conversion information structures
    //     Valid dividers is keyed by divide_by, value is Input Value
    FB_A_ValidDividers[8]    = 0x00;       FB_A_ValidDividers[12]   = 0x01;
    FB_A_ValidDividers[16]   = 0x02;       FB_A_ValidDividers[20]   = 0x03;
    FB_A_ValidDividers[24]   = 0x05;       FB_A_ValidDividers[32]   = 0x06;
    FB_A_ValidDividers[36]   = 0x09;       FB_A_ValidDividers[40]   = 0x07;
    FB_A_ValidDividers[48]   = 0x0a;       FB_A_ValidDividers[56]   = 0x18;
    FB_A_ValidDividers[60]   = 0x0b;       FB_A_ValidDividers[64]   = 0x0e;
    FB_A_ValidDividers[72]   = 0x15;       FB_A_ValidDividers[80]   = 0x0f;
    FB_A_ValidDividers[84]   = 0x19;       FB_A_ValidDividers[96]   = 0x16;
    FB_A_ValidDividers[100]  = 0x13;       FB_A_ValidDividers[108]  = 0x49;
    FB_A_ValidDividers[112]  = 0x1a;       FB_A_ValidDividers[120]  = 0x17;
    FB_A_ValidDividers[128]  = 0x1e;       FB_A_ValidDividers[140]  = 0x1b;
    FB_A_ValidDividers[144]  = 0x35;       FB_A_ValidDividers[160]  = 0x1f;
    FB_A_ValidDividers[168]  = 0x39;       FB_A_ValidDividers[180]  = 0x4b;
    FB_A_ValidDividers[192]  = 0x36;       FB_A_ValidDividers[200]  = 0x33;
    FB_A_ValidDividers[216]  = 0x55;       FB_A_ValidDividers[224]  = 0x3a;
    FB_A_ValidDividers[240]  = 0x37;       FB_A_ValidDividers[252]  = 0x59;
    FB_A_ValidDividers[256]  = 0x3e;       FB_A_ValidDividers[280]  = 0x3b;
    FB_A_ValidDividers[288]  = 0x56;       FB_A_ValidDividers[300]  = 0x53;
    FB_A_ValidDividers[320]  = 0x3f;       FB_A_ValidDividers[336]  = 0x5a;
    FB_A_ValidDividers[360]  = 0x57;       FB_A_ValidDividers[384]  = 0x5e;
    FB_A_ValidDividers[392]  = 0xd8;       FB_A_ValidDividers[400]  = 0x73;
    FB_A_ValidDividers[420]  = 0x5b;       FB_A_ValidDividers[432]  = 0xb5;
    FB_A_ValidDividers[448]  = 0x7a;       FB_A_ValidDividers[480]  = 0x5f;
    FB_A_ValidDividers[500]  = 0x93;       FB_A_ValidDividers[504]  = 0xb9;
    FB_A_ValidDividers[512]  = 0x7e;       FB_A_ValidDividers[560]  = 0x7b;
    FB_A_ValidDividers[576]  = 0xb6;       FB_A_ValidDividers[588]  = 0xd9;
    FB_A_ValidDividers[600]  = 0x97;       FB_A_ValidDividers[640]  = 0x7f;
    FB_A_ValidDividers[672]  = 0xba;       FB_A_ValidDividers[700]  = 0x9b;
    FB_A_ValidDividers[720]  = 0xb7;       FB_A_ValidDividers[768]  = 0xbe;
    FB_A_ValidDividers[784]  = 0xda;       FB_A_ValidDividers[800]  = 0x9f;
    FB_A_ValidDividers[840]  = 0xbb;       FB_A_ValidDividers[896]  = 0xde;
    FB_A_ValidDividers[960]  = 0xbf;       FB_A_ValidDividers[980]  = 0xdb;
    FB_A_ValidDividers[1024] = 0xfe;       FB_A_ValidDividers[1120] = 0xdf;
    FB_A_ValidDividers[1280] = 0xff;

    //  Fill the conversion information structures
    //     Valid dividers is keyed by Input Value, value is divide_by
    FB_A_InverseDividers[0x00]  =    8;      FB_A_InverseDividers[0x01]  =   12;
    FB_A_InverseDividers[0x02]  =   16;      FB_A_InverseDividers[0x03]  =   20;
    FB_A_InverseDividers[0x05]  =   24;      FB_A_InverseDividers[0x06]  =   32;
    FB_A_InverseDividers[0x09]  =   36;      FB_A_InverseDividers[0x07]  =   40;
    FB_A_InverseDividers[0x0a]  =   48;      FB_A_InverseDividers[0x18]  =   56;
    FB_A_InverseDividers[0x0b]  =   60;      FB_A_InverseDividers[0x0e]  =   64;
    FB_A_InverseDividers[0x15]  =   72;      FB_A_InverseDividers[0x0f]  =   80;
    FB_A_InverseDividers[0x19]  =   84;      FB_A_InverseDividers[0x16]  =   96;
    FB_A_InverseDividers[0x13]  =  100;      FB_A_InverseDividers[0x49]  =  108;
    FB_A_InverseDividers[0x1a]  =  112;      FB_A_InverseDividers[0x17]  =  120;
    FB_A_InverseDividers[0x1e]  =  128;      FB_A_InverseDividers[0x1b]  =  140;
    FB_A_InverseDividers[0x35]  =  144;      FB_A_InverseDividers[0x1f]  =  160;
    FB_A_InverseDividers[0x39]  =  168;      FB_A_InverseDividers[0x4b]  =  180;
    FB_A_InverseDividers[0x36]  =  192;      FB_A_InverseDividers[0x33]  =  200;
    FB_A_InverseDividers[0x55]  =  216;      FB_A_InverseDividers[0x3a]  =  224;
    FB_A_InverseDividers[0x37]  =  240;      FB_A_InverseDividers[0x59]  =  252;
    FB_A_InverseDividers[0x3e]  =  256;      FB_A_InverseDividers[0x3b]  =  280;
    FB_A_InverseDividers[0x56]  =  288;      FB_A_InverseDividers[0x53]  =  300;
    FB_A_InverseDividers[0x3f]  =  320;      FB_A_InverseDividers[0x5a]  =  336;
    FB_A_InverseDividers[0x57]  =  360;      FB_A_InverseDividers[0x5e]  =  384;
    FB_A_InverseDividers[0xd8]  =  392;      FB_A_InverseDividers[0x73]  =  400;
    FB_A_InverseDividers[0x5b]  =  420;      FB_A_InverseDividers[0xb5]  =  432;
    FB_A_InverseDividers[0x7a]  =  448;      FB_A_InverseDividers[0x5f]  =  480;
    FB_A_InverseDividers[0x93]  =  500;      FB_A_InverseDividers[0xb9]  =  504;
    FB_A_InverseDividers[0x7e]  =  512;      FB_A_InverseDividers[0x7b]  =  560;
    FB_A_InverseDividers[0xb6]  =  576;      FB_A_InverseDividers[0xd9]  =  588;
    FB_A_InverseDividers[0x97]  =  600;      FB_A_InverseDividers[0x7f]  =  640;
    FB_A_InverseDividers[0xba]  =  672;      FB_A_InverseDividers[0x9b]  =  700;
    FB_A_InverseDividers[0xb7]  =  720;      FB_A_InverseDividers[0xbe]  =  768;
    FB_A_InverseDividers[0xda]  =  784;      FB_A_InverseDividers[0x9f]  =  800;
    FB_A_InverseDividers[0xbb]  =  840;      FB_A_InverseDividers[0xde]  =  896;
    FB_A_InverseDividers[0xbf]  =  960;      FB_A_InverseDividers[0xdb]  =  980;
    FB_A_InverseDividers[0xfe]  = 1024;      FB_A_InverseDividers[0xdf]  = 1120;
    FB_A_InverseDividers[0xff]  = 1280;

    //  FB_B is the feedback bypass divider
    FB_B_DividerValues.push_back(1);     FB_B_DividerValues.push_back(2);
    FB_B_DividerValues.push_back(5);     FB_B_DividerValues.push_back(8);
    FB_B_DividerValues.push_back(10);    FB_B_DividerValues.push_back(16);
    FB_B_DividerValues.push_back(20);

    //  Valid dividers is keyed by divide_by, value is Input Value
    FB_B_ValidDividers[1]    = 0x07;       FB_B_ValidDividers[2]    = 0x00;
    FB_B_ValidDividers[5]    = 0x01;       FB_B_ValidDividers[8]    = 0x02;
    FB_B_ValidDividers[10]   = 0x03;       FB_B_ValidDividers[16]   = 0x04;
    FB_B_ValidDividers[20]   = 0x05;

    //     Valid dividers is keyed by Input Value, value is divide_by
    FB_B_InverseDividers[0x07]  =    1;      FB_B_InverseDividers[0x00]  =    2;
    FB_B_InverseDividers[0x01]  =    5;      FB_B_InverseDividers[0x02]  =    8;
    FB_B_InverseDividers[0x03]  =   10;      FB_B_InverseDividers[0x04]  =   16;
    FB_B_InverseDividers[0x05]  =   20;

    //  Internal Vco Output Prescaler (largest first)
    Vco_Ps_DividerValues.push_back(5);
    Vco_Ps_DividerValues.push_back(4);
    Vco_Ps_DividerValues.push_back(3);
    Vco_Ps_DividerValues.push_back(2);

    //  Valid dividers is keyed by divide_by, value is Input Value
    Vco_Ps_ValidDividers[2] = 0x3;    Vco_Ps_ValidDividers[3] = 0x1;
    Vco_Ps_ValidDividers[4] = 0x2;    Vco_Ps_ValidDividers[5] = 0x0;

    //     Valid dividers is keyed by Input Value, value is divide_by
    Vco_Ps_InverseDividers[0x03] = 2;   Vco_Ps_InverseDividers[0x01] = 3;
    Vco_Ps_InverseDividers[0x02] = 4;   Vco_Ps_InverseDividers[0x00] = 5;

}

//-----------------------------------------------------------------------------
//  Cdce62005::To_FB_A_DividerInput() -- Returns input value for divide_by
//-----------------------------------------------------------------------------
//  this maps divider ratios to the field value in the chip hardware

unsigned int  Cdce62005::To_FB_A_DividerInput(unsigned int divide_by)
{
    if (IsValid_FB_A_Divider(divide_by))
        return FB_A_ValidDividers[divide_by];
    else
        return FB_A_ValidDividers[1];   // if invalid divider, divide by 1.
}

//-----------------------------------------------------------------------------
//  Cdce62005::From_FB_A_DividerInput() -- Returns divide_by for input value
//-----------------------------------------------------------------------------
//  this maps divider bits in the field value in the chip hardware to divide_by result

unsigned int  Cdce62005::From_FB_A_DividerInput(unsigned int divider_bits)
{
    InverseDividerMap::iterator iter = FB_A_InverseDividers.find(divider_bits);

    if (iter != FB_A_InverseDividers.end())
        return FB_A_InverseDividers[divider_bits];
    else
        return 1;   // if invalid divider_bits, claim it is divide by 1
}

//-----------------------------------------------------------------------------
//  Cdce62005::IsValid_FB_A_Divider() -- tests if divide_by is a valid divider
//-----------------------------------------------------------------------------

bool  Cdce62005::IsValid_FB_A_Divider(unsigned int divide_by)
{
    ValidDividerMap::iterator iter = FB_A_ValidDividers.find(divide_by);
    return iter != FB_A_ValidDividers.end();
}

//-----------------------------------------------------------------------------
//  Cdce62005::To_FB_B_DividerInput() -- Returns input value for divide_by
//-----------------------------------------------------------------------------
//  this maps divider ratios to the field value in the chip hardware

unsigned int  Cdce62005::To_FB_B_DividerInput(unsigned int divide_by)
{
    if (IsValid_FB_B_Divider(divide_by))
        return FB_B_ValidDividers[divide_by];
    else
        return FB_B_ValidDividers[1];   // if invalid divider, divide by 1.
}

//-----------------------------------------------------------------------------
//  Cdce62005::From_FB_B_DividerInput() -- Returns divide_by for input value
//-----------------------------------------------------------------------------
//  this maps divider bits in the field value in the chip hardware to divide_by result

unsigned int  Cdce62005::From_FB_B_DividerInput(unsigned int divider_bits)
{
    InverseDividerMap::iterator iter = FB_B_InverseDividers.find(divider_bits);

    if (iter != FB_B_InverseDividers.end())
        return FB_B_InverseDividers[divider_bits];
    else
        return 1;   // if invalid divider_bits, claim it is divide by 1
}

//-----------------------------------------------------------------------------
//  Cdce62005::IsValid_FB_B_Divider() -- tests if divide_by is a valid divider
//-----------------------------------------------------------------------------

bool  Cdce62005::IsValid_FB_B_Divider(unsigned int divide_by)
{
    ValidDividerMap::iterator iter = FB_B_ValidDividers.find(divide_by);
    return iter != FB_B_ValidDividers.end();
}

//-----------------------------------------------------------------------------
//  Cdce62005::To_Vco_Ps_DividerInput() -- Returns input value for divide_by
//-----------------------------------------------------------------------------
//  this maps divider ratios to the field value in the chip hardware

unsigned int  Cdce62005::To_Vco_Ps_DividerInput(unsigned int divide_by)
{
    if (IsValid_Vco_Ps_Divider(divide_by))
        return Vco_Ps_ValidDividers[divide_by];
    else
        return Vco_Ps_ValidDividers[1];   // if invalid divider, divide by 1.
}

//-----------------------------------------------------------------------------
//  Cdce62005::From_Vco_Ps_DividerInput() -- Returns divide_by for input value
//-----------------------------------------------------------------------------
//  this maps divider bits in the field value in the chip hardware to divide_by result

unsigned int  Cdce62005::From_Vco_Ps_DividerInput(unsigned int divider_bits)
{
    InverseDividerMap::iterator iter = Vco_Ps_InverseDividers.find(divider_bits);

    if (iter != Vco_Ps_InverseDividers.end())
        return Vco_Ps_InverseDividers[divider_bits];
    else
        return 1;   // if invalid divider_bits, claim it is divide by 1
}

//-----------------------------------------------------------------------------
//  Cdce62005::IsValid_Vco_Ps_Divider() -- tests if divide_by is a valid divider
//-----------------------------------------------------------------------------

bool  Cdce62005::IsValid_Vco_Ps_Divider(unsigned int divide_by)
{
    ValidDividerMap::iterator iter = Vco_Ps_ValidDividers.find(divide_by);
    return iter != Vco_Ps_ValidDividers.end();
}

#ifdef LINUX
//    #pragma GCC diagnostic pop
#endif

} // namespace Innovative
