//
// Cdce72010_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2003 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <CDCE72010_Mb.h>
#include <SystemSupport_Mb.h>
#include <VcxoControl_Mb.h>
#include <algorithm>
#include <limits>
#include <cmath>
#include <Magic_Mb.h>
#include <Trace_Mb.h>
#include <sstream>

using namespace std;

namespace Innovative
{

//=============================================================================
//  CLASS Cdce72010  --  Cdce72010 Pll Implementation
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class Cdce72010
//-----------------------------------------------------------------------------

const unsigned int  TotalReferences  = 2;
const unsigned int  TotalOutputs     = 10;
const int           SpiRegisterCount = 12;    // Skip Register 12 R/W
const float         DefaultMaxOutFreq = 1500.0e6;
const double        MaxDividerAbove1Freq = 1175.0e6;

//-----------------------------------------------------------------------------
//  constructor for class Cdce72010
//-----------------------------------------------------------------------------

Cdce72010::Cdce72010(IAddressingSpace & space, Cdce72010Registers & regs)
	: PllTimebaseDevice(TotalReferences, TotalOutputs),
      FReferenceSource(refAutoSelect),
      FOutputSource(outPll),
      RegisterImage(SpiRegisterCount),
      ControlReg(space, regs.Control_Addr),
      StatusReg(space, regs.Control_Addr),
      SpiReg(space, regs.SpiIntf_Addr),
      FPowered(false),
      Debug(true)
{
    const unsigned int  DefaultOutputCfgBits = ocDisabled;
    //  Fill max frequency checker array
    for (unsigned int i=0; i<TotalOutputs; i++)
        {
        OutputFrequencyLimit.push_back(DefaultMaxOutFreq);
        OutputConfigField.push_back(DefaultOutputCfgBits);
        }
}

//-----------------------------------------------------------------------------
//  destructor for class Cdce72010
//-----------------------------------------------------------------------------

Cdce72010::~Cdce72010()
{
}

//-----------------------------------------------------------------------------
//  Cdce72010::Init()  --  Initialize Chip
//-----------------------------------------------------------------------------

void Cdce72010::Init()
{
    ControlReg.UsePrimaryReference.Value(true);
    ControlReg.SpiEnable.Value(true);
    ControlReg.ResetOff.Value(true);

    ReadDataRegisters();   // read initial state from chip
	SettingsInit();        // common INIT settings
    OnInit();
    Update();    // Update state of the device based on sources
    WriteDataRegisters();
}

//-----------------------------------------------------------------------------
//  Cdce72010::SettingsInit()  --  Initial Settings Loading
//-----------------------------------------------------------------------------
//   Note board specific implementations can override these in OnInit

void Cdce72010::SettingsInit()
{
	SetFBClockType(false);   // set FB clock to be always CMOS
}

//-----------------------------------------------------------------------------
//  Cdce72010::Reapply()  --  Reapply settings to hardware from saved state
//-----------------------------------------------------------------------------

void Cdce72010::Reapply()
{
    OnReapply();

    Update();    // Update state of the device based on sources

	ControlReg.Apply();

    WriteDataRegisters();
}

//-----------------------------------------------------------------------------
//  Cdce72010::Powered()  --  Change Powered state of chip
//-----------------------------------------------------------------------------

void Cdce72010::Powered(bool value)
{
    if (value==FPowered)
        return;

    FPowered = value;
    OnSetPower(FPowered);
    if (FPowered)
        {
        Init();       // Restore complete chip state
        Reapply();
        }
}

//-----------------------------------------------------------------------------
//  Cdce72010::OutputSource()  --  Set Output Source
//-----------------------------------------------------------------------------

void Cdce72010::OutputSource(Cdce72010::CdOutputSource value)
{
     FOutputSource = value;
     if (FOutputSource==outPll)
         {
         // Turn ON Pll power
         SetPllEnable(true);
         // Turn ON VCO output
         Vcxo().OutputEnabled(true);
         UpdateMode = updFrequency;
         }
     else
         {
         // Turn OFF Pll power
         SetPllEnable(false);
         // Turn OFF VCO output
         Vcxo().OutputEnabled(false);
         UpdateMode = updDivider;
         }
     SetOutputSource(value);
}

//-----------------------------------------------------------------------------
//  Cdce72010::MaxOutputDivider()  --  Return max output post-divider
//-----------------------------------------------------------------------------

int Cdce72010::MaxOutputDivider() const
{
	return 80;
}

//-----------------------------------------------------------------------------
//  Cdce72010::FrequencyModeUpdate()  --  Update using Frequency as baseline
//-----------------------------------------------------------------------------

void Cdce72010::FrequencyModeUpdate()
{
    if (Reference() == 0 || Frequency() == 0)
        return;
    //
    //   Phase Detector Freq = Reference/R*M = Vco/FB*X*N
    int R, M, FB, X, N, OutDiv;
    //
    //  1) Calculate Dividers to reduce Reference Frequency to Phase Detector
    //        Frequency
    double CandidateRM = Reference()/PhaseDetectorFrequency();
    if (Reference()>MaxMInputFrequency() || CandidateRM > 16384)
        {
        R = 2;
        M = static_cast<int>(CandidateRM/2);     // round to nearest?
        }
    else
        {
        R = 1;
        M = static_cast<int>(CandidateRM);       // round to nearest?
        }
    //
    //  Find Vco 'cap' frequency
    //      ...Dividers on this part cannot be above 1
    //         at frequencies above MaxDividerAbove1Freq due to hardware limitations
    //         so we clip the Max Vco Frequency
    double VcoCapFrequency;
    if (Frequency() < MaxDividerAbove1Freq)
        VcoCapFrequency = std::min( Vcxo().Max(), MaxDividerAbove1Freq );
    else
        VcoCapFrequency = Vcxo().Max();  // Not capped, divider = 1
    //
    //  Find divider to reduce Vco to Output Frequency (largest first)
    int   CandidateFB = 0;
    bool  TuneMatchFound(false);

    for (ValidDividerArray::reverse_iterator div = CdceDividerValues.rbegin();
             div !=  CdceDividerValues.rend();  ++div)
        {
        CandidateFB = *div;
        if (CandidateFB*Frequency() < VcoCapFrequency)
            {
            //  We have a candidate Divider. Check if we can reset to this freq.
            //     (some Vco's have gaps in the range).
            if (Vcxo().CanTuneTo(CandidateFB*Frequency()))
                {
                TuneMatchFound = true;
                break;
                }
            }
        }
    //
    //  Determine Vco Frequency - if the scan failed
    double VcoFreq;
    if (TuneMatchFound)
        {
        //
        //  We have a candidate divider FB;
        VcoFreq = CandidateFB*Frequency();
        }
    else
        {
        //  We missed finding a VCO Match. Pick a VCO freq we can use
        VcoFreq = Vcxo().BestTunableFrequency(Frequency());
        //  have to figure a reasonable FB value too.
        CandidateFB =  static_cast<int>(VcoFreq / Frequency());
        if (CandidateFB==0)
            CandidateFB=1;
        }
    //
    //  Calculate Candidate N Value
    double CandidateN = VcoFreq/CandidateFB/PhaseDetectorFrequency();
    //
    //  Adjust FB and N to assure that Input N is in range
    N  = RoundToBestInt(CandidateN);
    FB = CandidateFB;
    Debug.Log("Initial N=", N);
    Debug.Log("Initial FB=", FB);
    unsigned int XFactor;

    while (N > 16384)
        {
        // Find XFactor to shift between N and FB
        XFactor = ExtractDividerFactor(FB, N);
        Debug.Log("N too big, Shift X Factor=", XFactor);
        FB *= XFactor;
        N  /= XFactor;
        if (FB>40)
            break;  // leave the loop before we overflow the FB divider
        }
    //  Adjust FB and N to assure that Input N frequency is good
    while (VcoFreq/FB > MaxNInputFrequency())
        {
        // Find XFactor to shift between N and FB
        XFactor = ExtractDividerFactor(FB, N);
        Debug.Log("N infreq too fast, Shift X Factor=", XFactor);
        FB *= XFactor;
        N  /= XFactor;
        }
    //  Adjust FB and N to assure that FB is proper
	//    When CMOS FB Clock type is used, FB can't be less than or equal to 5.
    while (FB <= 5)
        {
        // Find XFactor to shift between N and FB
        XFactor = ExtractDividerFactor(FB, N);
        Debug.Log("FB too low, Shift X Factor=", XFactor);
        FB *= XFactor;
        N  /= XFactor;
        }
    //
    //  Finish real values
    X  = 1;
    OutDiv = CandidateFB;
    //
    //   Vco Freq = FB*X*N*Reference/R*M
    VcoFreq = FB*N*X*Reference()/(R*M);

    Debug.Log("Cdce72010::M = ", M);
    Debug.Log("Cdce72010::R = ", R);
    Debug.Log("Cdce72010::N = ", N);
    Debug.Log("Cdce72010::FB = ", FB);
    Debug.Log("Cdce72010::X = ", X);
    Debug.Log("VcoFreq = ", VcoFreq);

    Debug.Log("Pll Tuning Event");
    CdcePllTuningEvent e(M, R, N, FB, X);
    OnPllTuning.Execute(e);

    Debug.Log("Final Values");
    Debug.Log("Cdce72010::M = ", M);
    Debug.Log("Cdce72010::R = ", R);
    Debug.Log("Cdce72010::N = ", N);
    Debug.Log("Cdce72010::FB = ", FB);
    Debug.Log("Cdce72010::X = ", X);
    //
    // Calcluate actual frequency
    VcoFreq = FB*N*X*Reference()/(R*M);
    //
    //  Update registers & allied hardware
    Vcxo().SetVcoFrequency(VcoFreq);    // reprogram Vcxo
    SetM(M);
    SetR(R);
    SetN(N);
    SetFB(FB);
    SetX(X);

    //
    //  Set up the output dividers properly
    SetOutputDividers(VcoFreq, OutDiv);
}

//-----------------------------------------------------------------------------
//  Cdce72010::ExtractDividerFactor()  -- Return best factor to adjust FB
//-----------------------------------------------------------------------------
//  Sometimes we have to adjust N by changing FB. For accuracy we need to factor
//    N to get X, as long as the resulting new FB, FB*X, is still a valid divider.

unsigned int  Cdce72010::ExtractDividerFactor(unsigned int FB, unsigned int N)
{
    unsigned int CandidateX;

    // for each candidate factor
    for (ValidDividerArray::iterator XFactor = CdceDividerFactors.begin();
             XFactor !=  CdceDividerFactors.end();  ++XFactor)
        {
        CandidateX = *XFactor;
        //  Test if Candidate X is a factor of N
        if (N%CandidateX!=0)
            continue;   // remainder must be 0 for exact factor

        // Now test if FB*CandidateX would still be a valid divider
        if (IsValidDivider(FB*CandidateX))
            return CandidateX;
        }
    return 2;   // we can't find any exact matches, so lose precision

}

//-----------------------------------------------------------------------------
//  Cdce72010::DividerModeUpdate() -- Update state of dividers
//-----------------------------------------------------------------------------
//  In this mode, the reference is sent to the outputs, with possible division

void  Cdce72010::DividerModeUpdate()
{
    if (Reference() == 0 || Frequency() == 0)
        return;

    double div = 1.0;
    double fraction = modf(Reference()/Frequency(), &div);
    div += (fraction > 0.5) ? 1.0 : 0.0;
    unsigned int OutDiv = static_cast<unsigned int>(div);
    OutDiv = std::max<unsigned int>(OutDiv, 1);

    SetOutputDividers(Reference(), OutDiv);
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetOutputDividers()  --
//-----------------------------------------------------------------------------

void  Cdce72010::SetOutputDividers(double VcoFreq, unsigned int OutDiv)
{
    unsigned int BestDivider = OutDiv;
    double       OutputFreq = VcoFreq/BestDivider;

    for (unsigned int output=0; output<TotalOutputs; output++)
        {
        if (OutputInUse(output)==true)
            {
            float   CandidateDivider = static_cast<float>(BestDivider);
            double  CandidateFreq    = OutputFreq;
            //
            //  If the Output is the Master, use the best values.
            //     if not the master, and Frequency is unspecified, then
            //     the make the output a copy of Master.
            //     Otherwise, figure the best divider possible for each output
            if ((output!=FMasterOutput) && (OutputFrequency(output)!=0.0))
                {
                CandidateFreq = OutputFrequency(output);
                CandidateDivider = static_cast<float>(VcoFreq / CandidateFreq);

                int above, below, divider;
                GetNearestDividers(output, CandidateDivider, below, above);

                // Find which divider is closest to desired frequency
                double centerFreq = VcoFreq/CandidateDivider;
                double hiFreq     = VcoFreq/below;
                double lowFreq    = VcoFreq/above;
                if (std::abs(centerFreq-lowFreq) > std::abs(hiFreq-centerFreq))
                    divider = below;
                else
                    divider = above;

                CandidateDivider = static_cast<float>(divider);
                CandidateFreq = VcoFreq/CandidateDivider;
                }

            //
            //  Check for frequency exceeding applied Limits
            if (OutputFrequencyLimit[output]>=CandidateFreq)
                OutputDivider(output, static_cast<unsigned int>(CandidateDivider));
            else
                {
                double MinDivideRatio = VcoFreq/OutputFrequencyLimit[output];
                OutputDivider(output, BestDividerGreaterThan(static_cast<float>(MinDivideRatio)));
                }
            //
            //  Calculate Acutal Frequency
            if (OutputDividerArray[output]!=0)  // if illegal divider, skip it
                {
                OutputFreqActualArray[output] = VcoFreq / OutputDividerArray[output];
                }
            //
            //  If Divider 'needed' is not allowed because of rate limitations,
            //     disable the output, otherwise enable it
            if (VcoFreq > MaxDividerAbove1Freq && OutputDivider(output)!=1)
                {
                SetDividerEnable(output, false);
                OutputFreqActualArray[output] = 0.0;  // no output allowed
                }
            else
                SetDividerEnable(output, true);
            }
        else
            {
            OutputFreqActualArray[output] = 0.0;
            }
        }
}

//-----------------------------------------------------------------------------
//  Cdce72010::LoadFreq()  --  Poke in register values
//-----------------------------------------------------------------------------

void  Cdce72010::LoadFreq(int m, int n, int fb, int divider)
{
    SetM(m);
    SetR(1);
    SetN(n);
    SetFB(fb);
    SetX(1);
    double VcoFrequency = fb*n*Reference()/(m);
    SetOutputDividers(VcoFrequency, divider);
    WriteDataRegisters();   // Update Hardware
}

//-----------------------------------------------------------------------------
//  Cdce72010::ReadInfo()  --  Read current state of chip, and other info
//-----------------------------------------------------------------------------

void  Cdce72010::ReadInfo(InfoBlock & info)
{
    ReadDataRegisters();   // Update from Hardware

    info.M  = GetM();
    info.R  = GetR();
    info.N  = GetN();
    info.Fb = GetFB();
    info.X  = GetX();
    info.RefFrequency = Reference();
    info.VcoFrequency = info.Fb*info.N*info.X*Reference()/(info.R*info.M);
    info.FrequencyActual = info.VcoFrequency/OutputDivider(FMasterOutput);

}

//-----------------------------------------------------------------------------
//  Cdce72010::Locked()  --  Return PLL lock status
//-----------------------------------------------------------------------------

bool  Cdce72010::Locked() const
{
	return StatusReg.Locked() ? true : false;
}

//-----------------------------------------------------------------------------
//  Cdce72010::Port()  --  Write to PLL data port
//-----------------------------------------------------------------------------

void Cdce72010::Port(unsigned char reg, unsigned int value)
{
	std::stringstream msg;
	msg << "Cdce72010::Write 0x" << std::hex << static_cast<int>(value)
		<< " to reg " <<  static_cast<int>(reg);
	Debug.Log(msg.str());

    unsigned int SpiValue(0);

    PtrBitManipulator SpiMap(&SpiValue);

    //   Overwrite register area with Reg
    SpiMap.Value(value);       // Spi Register Value (Bits 0-31)
    SpiMap.Field(0,  4, reg);  // Spi Register Addr  (Bits 0-4)

	SpiReg.Value(SpiValue);
	Sleep(1);
}

//-----------------------------------------------------------------------------
//  Cdce72010::Port()  --  Read from PLL data port
//-----------------------------------------------------------------------------

unsigned int Cdce72010::Port(unsigned char reg)
{
    unsigned int SpiValue(0);

    PtrBitManipulator SpiMap(&SpiValue);

    SpiMap.Field(4,  4, reg);     // Spi Register Addr  (Bits 4-7)
	//
    //  Send Read Command
    Port(0xE, SpiValue);
	Sleep(1);
	//
    //  Read data
    SpiValue = SpiReg.Value();

	std::stringstream msg;
	msg << "Cdce72010::Read 0x" << std::hex << SpiValue
		<< " from reg " << static_cast<int>(reg);
	Debug.Log(msg.str());

	return SpiValue;
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetPhaseAdjust()  --  Update the Phase Adjust setting
//-----------------------------------------------------------------------------

void Cdce72010::SetPhaseAdjust(unsigned int output, unsigned int divider_setting,
                                                    unsigned int phase_setting)
{
    if (!IsValidDivider(divider_setting))
        return;
    if (output==0 || output==9)
        return;                   // 0 & 9 are slaved to 1 & 8
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Field(10, 7, ToPhaseAdjustInput(divider_setting, phase_setting));     // Output Cfg bits 10-16
        }
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetDividerHdw()  --  Update the Divider setting hardware
//-----------------------------------------------------------------------------

void Cdce72010::SetDividerHdw(unsigned int output, unsigned int divide_by)
{
    if (!IsValidDivider(divide_by))
        return;
    if (output==0 || output==9)
        return;                   // 0 & 9 are slaved to 1 & 8
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Field(17, 7, ToDividerInput(divide_by));     // Output Cfg bits 17-23
        }
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetOutEnabledHdw()  --  Update the Output Enabled hardware
//-----------------------------------------------------------------------------

void Cdce72010::SetOutEnabledHdw(unsigned int output, bool enabled)
{
    if (output < TotalOutputs)
        {
        if (enabled)
            {
            // restore from cached setting...
            SetOutputConfig(output, OutputConfigField[output] );
            //  the above also disables divider if off
            }
        else
            {
            bool should_cache_setting = false;
            SetOutputConfig(output, ocDisabled, should_cache_setting );
            //  the above also disables divider if off
            }
        }
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetDividerEnable()  --  Update the Divider Enable setting
//-----------------------------------------------------------------------------

void Cdce72010::SetDividerEnable(unsigned int output, bool state)
{
    if (output==0 || output==9)
        return;                   // 0 & 9 are slaved to 1 & 8
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Bit(24, state);     // Output Cfg bits 24
        }
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetPllEnable()  --  Enable PLL (Power Down if disabled)
//-----------------------------------------------------------------------------

void Cdce72010::SetPllEnable(bool enabled)
{
    PtrBitManipulator Reg11(&RegisterImage[11]);
    Reg11.Bit(23, enabled==false);     // 0 = enabled, 1=Powered Down
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetFBDividerEnable()  --  Enable PLL Divider
//-----------------------------------------------------------------------------
//  (Powered Down if disabled)

void Cdce72010::SetFBDividerEnable(bool enabled)
{
    PtrBitManipulator Reg11(&RegisterImage[11]);
    Reg11.Bit(6, enabled==false);     // 0 = enabled, 1=Powered Down
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetInputConfig()  --  Update the Input Config for a channel
//-----------------------------------------------------------------------------

void Cdce72010::SetInputConfig(unsigned int cfg_setting)
{
    PtrBitManipulator Reg(&RegisterImage[0]);
    Reg.Field(4, 2, cfg_setting);     // Input Cfg bits 4-5
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetInputBufferConfig()  --  Update the Buffer Config
//-----------------------------------------------------------------------------

void Cdce72010::SetInputBufferConfig(bool dc_terminated,
                               bool input_term_disabled,
                               bool prim_bias, bool sec_bias, bool failsafe  )
{
    PtrBitManipulator Reg1(&RegisterImage[1]);
    Reg1.Bit(4, dc_terminated);          // DC/AC termination
    Reg1.Bit(5, true);                   // Hysteresis - forced true
    Reg1.Bit(6, input_term_disabled);    // Input Termination disabled
    Reg1.Bit(7, prim_bias);              // Primary Neg Pin Bias
    Reg1.Bit(8, sec_bias);               // Secondary Neg Pin Bias
    Reg1.Bit(9, failsafe);               // Failsafe

}

//-----------------------------------------------------------------------------
//  Cdce72010::SetVcoConfig()  --  Update the Vco Input Config for a channel
//-----------------------------------------------------------------------------

void Cdce72010::SetVcoConfig(unsigned int cfg_setting)
{
    PtrBitManipulator Reg(&RegisterImage[8]);
    Reg.Field(4, 2, cfg_setting);     // Input Cfg bits 4-5
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetVcoBufferConfig()  --  Update the Buffer Config
//-----------------------------------------------------------------------------

void Cdce72010::SetVcoBufferConfig(bool dc_terminated,
                                   bool hysteresis_enabled,
                                   bool input_term_disabled,
                                   bool vco_bias  )
{
    PtrBitManipulator Reg8(&RegisterImage[8]);
    Reg8.Bit(6, dc_terminated);          // DC/AC termination
    Reg8.Bit(7, hysteresis_enabled);     // Hysteresis 
    Reg8.Bit(8, input_term_disabled);    // Input Termination disabled
    Reg8.Bit(9, vco_bias);               // Vco Input Neg Pin Bias
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetReferenceSource()  --  Set the Reference Source
//-----------------------------------------------------------------------------

void Cdce72010::SetReferenceSource(CdReferenceSource source)
{
	unsigned int cfg_setting;
	// Conform to one of two possible physical references
	unsigned int source_no;
	switch (source)
		{
		default:
		case refNone:
			cfg_setting=0x0;
			source_no = 0;
			break;
		case refPrimary:
			cfg_setting=0x1;
			source_no = 0;
			break;
		case refSecondary:
			cfg_setting=0x2;
			source_no = 1;
			break;
		case refAutoSelect:
			cfg_setting=0x3;
			source_no = 0;
			break;
		}
	PtrBitManipulator Reg(&RegisterImage[0]);
	Reg.Field(6, 2, cfg_setting);     // Output Cfg bits 6-7

	SetSourceSelection(source_no);  // inform base class too
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetOutputSource()  --  Set the Reference Source
//-----------------------------------------------------------------------------

void Cdce72010::SetOutputSource(CdOutputSource source)
{
    unsigned int cfg_setting;
    switch (source)
        {
        default:
        case outPll:     cfg_setting=0x0; break;
        case outAux:     cfg_setting=0x3; break;
        }
    PtrBitManipulator Reg(&RegisterImage[11]);

    Reg.Field(24, 2, cfg_setting);     
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetFBClockType()  --  Update the Divider setting
//-----------------------------------------------------------------------------

void Cdce72010::SetFBClockType(bool clock_is_cml)
{
    PtrBitManipulator Reg(&RegisterImage[11]);
    Reg.Bit(7, clock_is_cml);
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetHighSwingLvpecl()  --  Update the boosted LvPecl state
//-----------------------------------------------------------------------------

void Cdce72010::SetHighSwingLvpecl(unsigned int output, bool setting)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Bit(25, setting);     // Output Cfg bit 25
        }
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetOutputConfig()  --  Update the Output Config for a channel
//-----------------------------------------------------------------------------

void Cdce72010::SetOutputConfig(unsigned int output, unsigned int cfg_setting,
                                bool should_cache_setting)
{
    if (output < TotalOutputs)
        {
        //  Cache restore-to setting
        if (should_cache_setting)
            OutputConfigField[output] = cfg_setting;

        PtrBitManipulator Reg(&RegisterImage[output]);
        Reg.Field(26, 6, cfg_setting);     // Output Cfg bits 26-31

        //  By default, enable channel divider if output is enabled
        if (cfg_setting==ocDisabled)
            SetDividerEnable(output, false);
        else
            SetDividerEnable(output, true);
        }
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetAuxInputConfig()  --  Configure Aux Input
//-----------------------------------------------------------------------------

void Cdce72010::SetAuxInputConfig(bool enable, unsigned int aux_setting)
{
    PtrBitManipulator Reg(&RegisterImage[9]);
    Reg.Bit(24, enable);
	Reg.Bit(23, aux_setting ? true : false);
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetLockConfig()  --  Lock options
//-----------------------------------------------------------------------------

void Cdce72010::SetLockConfig(bool analog_lock)
{
    PtrBitManipulator Reg(&RegisterImage[7]);
    Reg.Bit(9, analog_lock);
}

//------------------------------------------------------------------------------
//  Cdce72010::SetExternalReferenceControl()  --  Define who controls Ref select
//------------------------------------------------------------------------------

void Cdce72010::SetExternalReferenceControl(bool useInternalSelect)
{
    PtrBitManipulator Reg(&RegisterImage[0]);
    Reg.Bit(9, useInternalSelect);
}

//------------------------------------------------------------------------------
//  Cdce72010::SetExternalReferenceSelect()  --  Define who controls Ref select
//------------------------------------------------------------------------------

void Cdce72010::SetExternalReferenceSelect(bool usePrimaryReference)
{
    ControlReg.UsePrimaryReference.Value(usePrimaryReference);
}

//-----------------------------------------------------------------------------
//  Cdce72010::GetDivider()  --  read back the Divider setting
//-----------------------------------------------------------------------------

unsigned int  Cdce72010::GetDivider(unsigned int output)
{
    if (output==0)
        return GetDivider(1);   // 0 & 9 are slaved to 1 & 8
    if (output==9)
        return GetDivider(8);   // 0 & 9 are slaved to 1 & 8

    unsigned int divider(1);
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        unsigned int div_field = Reg.Field(17, 7);
        divider = FromDividerInput(div_field);
        }
    return divider;
}

//-----------------------------------------------------------------------------
//  Cdce72010::GetDividerEnable()  --  Read the Divider Enable setting
//-----------------------------------------------------------------------------

bool Cdce72010::GetDividerEnable(unsigned int output)
{
    if (output==0)
        return GetDividerEnable(1);   // 0 & 9 are slaved to 1 & 8
    if (output==9)
        return GetDividerEnable(8);   // 0 & 9 are slaved to 1 & 8

    bool bit = false;
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[output]);
        bit = Reg.Bit(24);     // Output Cfg bits 24
        }
    return bit;
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetM()  --  Configure M  (Reference Divider)
//-----------------------------------------------------------------------------

void Cdce72010::SetM(unsigned int M)
{
    PtrBitManipulator Reg(&RegisterImage[10]);
    Reg.Field(4, 14, M-1);   // offset by 1
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetN()  --  Configure N  (Vco Divider)
//-----------------------------------------------------------------------------

void Cdce72010::SetN(unsigned int N)
{
    PtrBitManipulator Reg(&RegisterImage[10]);
    Reg.Field(18, 14, N-1);    // offset by 1
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetR()  --  Configure R  (Reference Prescaler)
//-----------------------------------------------------------------------------

void Cdce72010::SetR(unsigned int R)
{
    PtrBitManipulator Reg(&RegisterImage[11]);
    if (R==2)
        {
        Reg.Bit(4, true);   Reg.Bit(5, true);     // Both PRI/SEC set same
        }
    else
        {
        Reg.Bit(4, false);  Reg.Bit(5, false);    // Both PRI/SEC set same
        }

}

//-----------------------------------------------------------------------------
//  Cdce72010::SetFB()  --  Update the Divider setting
//-----------------------------------------------------------------------------

void Cdce72010::SetFB(unsigned int FB)
{
    if (!IsValidDivider(FB))
        return;

    PtrBitManipulator Reg(&RegisterImage[11]);
    Reg.Field(9, 7, ToDividerInput(FB));
}

//-----------------------------------------------------------------------------
//  Cdce72010::SetX()  --  Configure X  (Vcxo Prescaler)  ??????
//-----------------------------------------------------------------------------

void Cdce72010::SetX(unsigned int X)
{
    PtrBitManipulator Reg(&RegisterImage[6]);
    if (X==2)
        {
        Reg.Bit(6, true);
        }
    else
        {
        Reg.Bit(6, false);
        }
}

//-----------------------------------------------------------------------------
//  Cdce72010::GetM()  --  Read M  (Reference Divider)
//-----------------------------------------------------------------------------

unsigned int Cdce72010::GetM()
{
    PtrBitManipulator Reg(&RegisterImage[10]);
    unsigned int M = Reg.Field(4, 14);
    return M+1;       // offset by 1
}

//-----------------------------------------------------------------------------
//  Cdce72010::GetN()  --  Read N  (Vco Divider)
//-----------------------------------------------------------------------------

unsigned int Cdce72010::GetN()
{
    PtrBitManipulator Reg(&RegisterImage[10]);
    unsigned int N = Reg.Field(18, 14);
    return N+1;     // offset by 1
}

//-----------------------------------------------------------------------------
//  Cdce72010::GetR()  --  Read R  (Reference Prescaler)
//-----------------------------------------------------------------------------

unsigned int Cdce72010::GetR()
{
    PtrBitManipulator Reg(&RegisterImage[11]);
    bool R_Bit = Reg.Bit(4);   // assume 4 and 5 are set the same
    if (R_Bit)
        {
        return 2;
        }
    else
        {
        return 1;
        }

}

//-----------------------------------------------------------------------------
//  Cdce72010::GetFB()  --  Update the Divider setting
//-----------------------------------------------------------------------------

unsigned int Cdce72010::GetFB()
{
    PtrBitManipulator Reg(&RegisterImage[11]);
    unsigned int FB_field = Reg.Field(9, 7);
    unsigned int FB = FromDividerInput(FB_field);
    return FB;
}

//-----------------------------------------------------------------------------
//  Cdce72010::GetX()  --  Read X  (Vcxo Prescaler)  ??????
//-----------------------------------------------------------------------------

unsigned int Cdce72010::GetX()
{
    PtrBitManipulator Reg(&RegisterImage[6]);
    bool X_Bit = Reg.Bit(6);
    if (X_Bit)
        {
        return 2;
        }
    else
        {
        return 1;
        }
}

//-----------------------------------------------------------------------------
//  Cdce72010::ReadDataRegisters()  --  Read all data registers into Image vector
//-----------------------------------------------------------------------------

void Cdce72010::ReadDataRegisters()
{
    for (unsigned int i=0; i<RegisterImage.size(); i++)
        {
        RegisterImage[i] = Port(i);
        }
}

//-----------------------------------------------------------------------------
//  Cdce72010::WriteDataRegisters()  --  Write all data registers from Image vector
//-----------------------------------------------------------------------------

void Cdce72010::WriteDataRegisters()
{
    ScopedTraceRef ccc(Debug, "Cdce72010::WriteDataRegisters()");

    for (unsigned int i=0; i<RegisterImage.size(); i++)
        {
        Port(i, RegisterImage[i]);
        }
}

//-----------------------------------------------------------------------------
//  Cdce72010::OnSetPower()  --  Perform Pwr Setting operation
//-----------------------------------------------------------------------------

void  Cdce72010::OnSetPower(bool & new_state)
{
    ControlReg.PowerOn.Value(new_state);
    //  pause to allow chip to activate
    Sleep(10);
}

//-----------------------------------------------------------------------------
//  Cdce72010::RoundToBestInt() -- convert to Int, rounding fraction
//-----------------------------------------------------------------------------

unsigned int  Cdce72010::RoundToBestInt(double number)
{
    double div = 1.0;
    double fraction = modf(number, &div);  // round to best int
    div += (fraction > 0.5) ? 1.0 : 0.0;
    return static_cast<unsigned int>( div );
}

} // namespace Innovative
