//
// CalMemBase_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure 
//  agreement with Innovative Integration Corporation and may not be copied or 
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2003 Innovative Integration Corporation. 
//  All Rights Reserved.
//
// PmcIdrom storage format:
//   [Header]
//   Offset ch 0..N
//   Gain ch 0..N
//   bool Calibrated

#ifdef __BORLANDC__
#include <mem.h>
#endif
#include <CalMemBase_Mb.h>


namespace Innovative
{

//=============================================================================
//  CLASS IUsesCalibration  --  Calibration memory access class for analog I/O
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class IUsesCalibration
//-----------------------------------------------------------------------------

IUsesCalibration::IUsesCalibration(AddressingSpace & space, unsigned int addr, int size,
								   int channels, int one)
	: IUsesPmcEeprom(space, addr, size),
	  Channels(channels), FOne(one)
{
}

//-----------------------------------------------------------------------------
//  IUsesCalibration::ApplyDefaultCoefficients() -- load default set of coeffs into memory
//-----------------------------------------------------------------------------

void  IUsesCalibration::ApplyDefaultCoefficients()
{
	for (int ch = 0; ch < Channels; ++ch)
		{
		Offset(ch, 0);
		Gain(ch, 1.0);
		}
}

//-----------------------------------------------------------------------------
//  IUsesCalibration::ApplyRomCoefficients() --
//-----------------------------------------------------------------------------

void  IUsesCalibration::ApplyRomCoefficients()
{
    if (!Calibrated())
        {
        ApplyDefaultCoefficients();
        return;
        }

	for (int ch = 0; ch < Channels; ++ch)
		{
		Offset(ch, Offset(ch));
		Gain(ch, Gain(ch));
		}
}

//-----------------------------------------------------------------------------
//  IUsesCalibration::ToFixed() - Converts a float gain coeff to internal fixed-point format
//-----------------------------------------------------------------------------

unsigned int IUsesCalibration::ToFixed(float value)
{
    return static_cast<int>(value*One());
}

//------------------------------------------------------------------------
//  IUsesCalibration::FromFixed() - Converts an internal fixed-point format to float
//------------------------------------------------------------------------

float IUsesCalibration::FromFixed(unsigned int value)
{
    return static_cast<float>(value)/One();
}

//------------------------------------------------------------------------
//  IUsesCalibration::Calibrated() - Set calibration status
//------------------------------------------------------------------------

IUsesCalibration & IUsesCalibration::Calibrated(bool state)
{
	int idx = (Channels*2);
	Device->WorkAreaAsInt(PmcIdrom::waModule, idx, state ? 1 : 0);

	return *this;
}

//------------------------------------------------------------------------
//  IUsesCalibration::Calibrated() - Retreive calibration status
//------------------------------------------------------------------------

bool IUsesCalibration::Calibrated() const
{
	int idx = (Channels*2);
	return Device->WorkAreaAsInt(PmcIdrom::waModule, idx) == 1;
}

//-----------------------------------------------------------------------------
//  IUsesCalibration::Gain()  --  Set gain coefficient
//-----------------------------------------------------------------------------

IUsesCalibration & IUsesCalibration::Gain(int ch, float gain)
{
	int idx = (Channels + ch);

	Device->WorkAreaAsFloat(PmcIdrom::waModule, idx, gain);
	SetCalibrationGain(ch, ToFixed(gain));

	return *this;
}

//-----------------------------------------------------------------------------
//  IUsesCalibration::Gain()  --  Get gain coefficient
//-----------------------------------------------------------------------------

float IUsesCalibration::Gain(int ch) const
{
	int idx = (Channels + ch);
	return Device->WorkAreaAsFloat(PmcIdrom::waModule, idx);
}

//-----------------------------------------------------------------------------
//  IUsesCalibration::Offset()  --  Set offset coefficient
//-----------------------------------------------------------------------------

IUsesCalibration & IUsesCalibration::Offset(int ch, float offset)
{
	int idx = ch;

	Device->WorkAreaAsFloat(PmcIdrom::waModule, idx, offset);
	SetCalibrationOffset(ch, (int)offset);

	return *this;
}

//-----------------------------------------------------------------------------
//  IUsesCalibration::Offset()  --  Get offset coefficient
//-----------------------------------------------------------------------------

float IUsesCalibration::Offset(int ch) const
{
	int idx = ch;
	return Device->WorkAreaAsFloat(PmcIdrom::waModule, idx);
}

//-----------------------------------------------------------------------------
//  IUsesCalibration::One()  --  Get Q-notation unity-value
//-----------------------------------------------------------------------------

int IUsesCalibration::One() const
{
	return FOne;
}

//=============================================================================
//  CLASS IUsesSectionCalibration  --  Calib memory class for analog I/O
//=============================================================================
//  Uses a section to access the IDROM for calibration (allowing sharing)

//-----------------------------------------------------------------------------
//  constructor for class IUsesSectionCalibration
//-----------------------------------------------------------------------------

IUsesSectionCalibration::IUsesSectionCalibration(PmcIdromSection section, int channels, int one)
	: Section(section), Channels(channels), FOne(one)
{
}

//-----------------------------------------------------------------------------
//  IUsesSectionCalibration::ApplyDefaultCoefficients() -- load default set of coeffs into memory
//-----------------------------------------------------------------------------

void  IUsesSectionCalibration::ApplyDefaultCoefficients()
{
	for (int ch = 0; ch < Channels; ++ch)
		{
		Offset(ch, 0);
		Gain(ch, 1.0);
		}
}

//-----------------------------------------------------------------------------
//  IUsesSectionCalibration::ApplyRomCoefficients() --
//-----------------------------------------------------------------------------

void  IUsesSectionCalibration::ApplyRomCoefficients()
{
    if (!Calibrated())
        {
        ApplyDefaultCoefficients();
        return;
        }

	for (int ch = 0; ch < Channels; ++ch)
		{
		Offset(ch, Offset(ch));
		Gain(ch, Gain(ch));
		}
}

//-----------------------------------------------------------------------------
//  IUsesSectionCalibration::ToFixed() - Converts a float gain coeff to internal fixed-point format
//-----------------------------------------------------------------------------

unsigned int IUsesSectionCalibration::ToFixed(float value)
{
    return static_cast<int>(value*One());
}

//------------------------------------------------------------------------
//  IUsesSectionCalibration::FromFixed() - Converts an internal fixed-point format to float
//------------------------------------------------------------------------

float IUsesSectionCalibration::FromFixed(unsigned int value)
{
    return static_cast<float>(value)/One();
}

//------------------------------------------------------------------------
//  IUsesSectionCalibration::Calibrated() - Set calibration status
//------------------------------------------------------------------------

IUsesSectionCalibration & IUsesSectionCalibration::Calibrated(bool state)
{
	int idx = (Channels*2);
	Section.AsInt(idx, state ? 1 : 0);

	return *this;
}

//------------------------------------------------------------------------
//  IUsesSectionCalibration::Calibrated() - Retreive calibration status
//------------------------------------------------------------------------

bool IUsesSectionCalibration::Calibrated() const
{
	int idx = (Channels*2);
	return Section.AsInt(idx) == 1;
}

//-----------------------------------------------------------------------------
//  IUsesSectionCalibration::Gain()  --  Set gain coefficient
//-----------------------------------------------------------------------------

IUsesSectionCalibration & IUsesSectionCalibration::Gain(int ch, float gain)
{
	int idx = (Channels + ch);

	Section.AsFloat(idx, gain);
	SetCalibrationGain(ch, ToFixed(gain));

	return *this;
}

//-----------------------------------------------------------------------------
//  IUsesSectionCalibration::Gain()  --  Get gain coefficient
//-----------------------------------------------------------------------------

float IUsesSectionCalibration::Gain(int ch) const
{
	int idx = (Channels + ch);
	return Section.AsFloat(idx);
}

//-----------------------------------------------------------------------------
//  IUsesSectionCalibration::Offset()  --  Set offset coefficient
//-----------------------------------------------------------------------------

IUsesSectionCalibration & IUsesSectionCalibration::Offset(int ch, float offset)
{
	int idx = ch;

	Section.AsFloat(idx, offset);
	SetCalibrationOffset(ch, (int)offset);

	return *this;
}

//-----------------------------------------------------------------------------
//  IUsesSectionCalibration::Offset()  --  Get offset coefficient
//-----------------------------------------------------------------------------

float IUsesSectionCalibration::Offset(int ch) const
{
	int idx = ch;
	return Section.AsFloat(idx);
}

//-----------------------------------------------------------------------------
//  IUsesSectionCalibration::One()  --  Get Q-notation unity-value
//-----------------------------------------------------------------------------

int IUsesSectionCalibration::One() const
{
	return FOne;
}

//=============================================================================
//  CLASS IUsesRangedSectionCalibration  --  Multiple gain range analog I/O support
//=============================================================================
//  Uses a section to access the IDROM for calibration (allowing sharing)
//-----------------------------------------------------------------------------
//  constructor for class IUsesRangedSectionCalibration
//-----------------------------------------------------------------------------
//  On initializing the section array: Fill in the PmcIdromSections at a later time
//     (such as the most derived ctor of the input or output object)

IUsesRangedSectionCalibration::IUsesRangedSectionCalibration(int channels, int one)
	: Channels(channels), FOne(one)
{
}

//-----------------------------------------------------------------------------
//  IUsesRangedSectionCalibration::Ranges() -- Return number of gain ranges
//-----------------------------------------------------------------------------

size_t IUsesRangedSectionCalibration::Ranges() const
{
    return SectionArray.size();
}

//-----------------------------------------------------------------------------
//  IUsesRangedSectionCalibration::ClearRanges() --  Clear range array
//-----------------------------------------------------------------------------

void IUsesRangedSectionCalibration::ClearRanges()
{
    SectionArray.clear();
}

//-----------------------------------------------------------------------------
//  IUsesRangedSectionCalibration::AppendRange() -- Append to range array
//-----------------------------------------------------------------------------

void IUsesRangedSectionCalibration::AppendRange(const PmcIdromSection & section)
{
    SectionArray.push_back(section);
}

//-----------------------------------------------------------------------------
//  IUsesRangedSectionCalibration::ApplyDefaultCoefficients() -- load default set of coeffs into memory
//-----------------------------------------------------------------------------

void  IUsesRangedSectionCalibration::ApplyDefaultCoefficients(size_t gain_range)
{
	for (int ch = 0; ch < Channels; ++ch)
		{
		Offset(gain_range, ch, 0);
		Gain(gain_range, ch, 1.0);
		}
}

//-----------------------------------------------------------------------------
//  IUsesRangedSectionCalibration::ApplyRomCoefficients() --
//-----------------------------------------------------------------------------

void  IUsesRangedSectionCalibration::ApplyRomCoefficients(size_t gain_range)
{
    if (!Calibrated(gain_range))
        {
        ApplyDefaultCoefficients(gain_range);
        return;
        }

	for (int ch = 0; ch < Channels; ++ch)
		{
		Offset(gain_range, ch, Offset(gain_range, ch));
		Gain(gain_range, ch, Gain(gain_range, ch));
		}
}

//-----------------------------------------------------------------------------
//  IUsesRangedSectionCalibration::ToFixed() -
//-----------------------------------------------------------------------------
//  Converts a float gain coeff to internal fixed-point format

unsigned int IUsesRangedSectionCalibration::ToFixed(float value)
{
    return static_cast<int>(value*One());
}

//------------------------------------------------------------------------
//  IUsesRangedSectionCalibration::FromFixed() -
//------------------------------------------------------------------------
//  Converts an internal fixed-point format to float

float IUsesRangedSectionCalibration::FromFixed(unsigned int value)
{
    return static_cast<float>(value)/One();
}

//------------------------------------------------------------------------
//  IUsesRangedSectionCalibration::Calibrated() - Set calibration status
//------------------------------------------------------------------------

IUsesRangedSectionCalibration & IUsesRangedSectionCalibration::Calibrated(size_t gain_range, bool state)
{
    int idx = (Channels*2);

	if (gain_range<Ranges())
	    SectionArray[gain_range].AsInt(idx, state ? 1 : 0);

	return *this;
}

//------------------------------------------------------------------------
//  IUsesRangedSectionCalibration::Calibrated() - Retreive calibration status
//------------------------------------------------------------------------

bool IUsesRangedSectionCalibration::Calibrated(size_t gain_range) const
{
	int idx = (Channels*2);
	if (gain_range<Ranges())
	    return SectionArray[gain_range].AsInt(idx) == 1;
    else
        return false;
}

//-----------------------------------------------------------------------------
//  IUsesRangedSectionCalibration::Gain()  --  Set gain coefficient
//-----------------------------------------------------------------------------

IUsesRangedSectionCalibration & IUsesRangedSectionCalibration::Gain(size_t gain_range, int ch, float gain)
{
	int idx = (Channels + ch);

	if (gain_range<Ranges())
	    {
        SectionArray[gain_range].AsFloat(idx, gain);
	    SetCalibrationGain(gain_range, ch, ToFixed(gain));
        }

	return *this;
}

//-----------------------------------------------------------------------------
//  IUsesRangedSectionCalibration::Gain()  --  Get gain coefficient
//-----------------------------------------------------------------------------

float IUsesRangedSectionCalibration::Gain(size_t gain_range, int ch) const
{
	int idx = (Channels + ch);
	if (gain_range<Ranges())
	    return SectionArray[gain_range].AsFloat(idx);
    else
        return 0.0;
}

//-----------------------------------------------------------------------------
//  IUsesRangedSectionCalibration::Offset()  --  Set offset coefficient
//-----------------------------------------------------------------------------

IUsesRangedSectionCalibration & IUsesRangedSectionCalibration::Offset(size_t gain_range, int ch, float offset)
{
	int idx = ch;

	if (gain_range<Ranges())
	    {
        SectionArray[gain_range].AsFloat(idx, offset);
	    SetCalibrationOffset(gain_range, ch, static_cast<int>(offset));
        }

	return *this;
}

//-----------------------------------------------------------------------------
//  IUsesRangedSectionCalibration::Offset()  --  Get offset coefficient
//-----------------------------------------------------------------------------

float IUsesRangedSectionCalibration::Offset(size_t gain_range, int ch) const
{
	int idx = ch;
	if (gain_range<Ranges())
	    return SectionArray[gain_range].AsFloat(idx);
    else
        return 0.0;
}

//-----------------------------------------------------------------------------
//  IUsesRangedSectionCalibration::One()  --  Get Q-notation unity-value
//-----------------------------------------------------------------------------

int IUsesRangedSectionCalibration::One() const
{
	return FOne;
}



} // namespace Innovative



