//
// BlkChannelInfo.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure 
//  agreement with Innovative Integration Corporation and may not be copied or 
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2003 Innovative Integration Corporation. 
//  All Rights Reserved.
//

#include <ChannelInfo_Mb.h>
           
namespace Innovative
{

//=============================================================================
//  CLASS BlkChannelInfo  --  Channel enable/disable and other information
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class BlkChannelInfo
//-----------------------------------------------------------------------------

BlkChannelInfo::BlkChannelInfo(int devices, int ch_per_dev, int bits, int muxChannels, bool unipolar)  
    : FBits(bits), FUnipolar(unipolar), FChannelsPerDevice(ch_per_dev), FMuxChannels(muxChannels)
{
	if(FMuxChannels == 0)
	    FEnabled.resize(devices);
	else
		FEnabled.resize(devices * FMuxChannels);
}

//-----------------------------------------------------------------------------
//  destructor for class BlkChannelInfo
//-----------------------------------------------------------------------------

BlkChannelInfo::~BlkChannelInfo()  
{
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::Devices()  --  Retreive device count
//-----------------------------------------------------------------------------

int BlkChannelInfo::Devices() const            
{
    if(FMuxChannels)
	    return static_cast<int>(FEnabled.size()/FMuxChannels);
	else
		return static_cast<int>(FEnabled.size());

}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::Channels()  --  Retreive channel count
//-----------------------------------------------------------------------------

int BlkChannelInfo::Channels() const            
{
    return Devices()*FChannelsPerDevice;
}

//-----------------------------------------------------------------------------
// BlkChannelInfo::MuxChannels() -- Retreive mux channel count.
//-----------------------------------------------------------------------------

int  BlkChannelInfo::MuxChannels() const
{
	return FMuxChannels;
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::Bits()  --  Set bits per sample
//-----------------------------------------------------------------------------

BlkChannelInfo & BlkChannelInfo::Bits(int bits)     
{
    FBits = bits;
    Update();

    return *this;
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::Bits()  --  Retreive bits per sample
//-----------------------------------------------------------------------------

int BlkChannelInfo::Bits() const            
{
    return FBits; 
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::Update()  --  Apply changes to class variables
//-----------------------------------------------------------------------------

void BlkChannelInfo::Update()
{
    FRange = 1 << (FBits-1);
    FOffset = FUnipolar ? 0 : FRange;
    FBytesPerSample = (FBits<17) ? (FBits + 7)/8 : 4;

    FActiveChannels = 0;

    for (int i = 0; i < Devices(); ++i)
        FActiveChannels += FEnabled[i]*FChannelsPerDevice;

	if( FMuxChannels == 0)
	    FBytesPerEvent = FBytesPerSample*FActiveChannels;
	else
		FBytesPerEvent = FBytesPerSample*FActiveMuxChannels;

	BuildMaps();

    //
    //  Update the hardware registers with new info, if any
    OpenWire::NotifyEvent e;
    OnHardwareUpdate.Execute(e);
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::Unipolar()  --  Set sample polarity
//-----------------------------------------------------------------------------

BlkChannelInfo & BlkChannelInfo::Unipolar(bool state)       
{
    FUnipolar = state;
    Update();

    return *this;
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::Unipolar()  --  Retrieve sample polarity
//-----------------------------------------------------------------------------

bool BlkChannelInfo::Unipolar() const           
{
    return FUnipolar; 
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::Range()  --  
//-----------------------------------------------------------------------------

int BlkChannelInfo::Range() const           
{  
    return FRange;  
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::Offset()  --  
//-----------------------------------------------------------------------------

int BlkChannelInfo::Offset() const          
{  
    return FOffset;  
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::BytesPerSample()  --  
//-----------------------------------------------------------------------------

int BlkChannelInfo::BytesPerSample() const          
{  
    return FBytesPerSample;  
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::BytesPerEvent()  --  
//-----------------------------------------------------------------------------

int BlkChannelInfo::BytesPerEvent() const           
{  
    return FBytesPerEvent;  
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::IntsPerSample()  --
//-----------------------------------------------------------------------------

int BlkChannelInfo::IntsPerSample() const
{  
    return FBytesPerSample/sizeof(int);  
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::IntsPerEvent()  --
//-----------------------------------------------------------------------------

int BlkChannelInfo::IntsPerEvent() const
{  
    return FBytesPerEvent/sizeof(int);
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::ActiveChannels()  --  
//-----------------------------------------------------------------------------

int BlkChannelInfo::ActiveChannels() const          
{  
  	if(FMuxChannels == 0)
	    return FActiveChannels;  
	else
		return FActiveMuxChannels;
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::ActiveDevices()  --  
//-----------------------------------------------------------------------------

int BlkChannelInfo::ActiveDevices() const          
{  
    return FActiveChannels/FChannelsPerDevice;  
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::EnableDevices()  -- Enable N devices from 0
//-----------------------------------------------------------------------------

BlkChannelInfo & BlkChannelInfo::EnableDevices( int dev_count )
{
    for (int i = 0; i < Devices(); ++i)
        if (i  < dev_count)
            FEnabled[i] = true;
        else
            FEnabled[i] = false;
            
    Update();        
    return *this;       
}
               
//-----------------------------------------------------------------------------
//  BlkChannelInfo::EnableChannels()  -- Enable N channels from 0
//-----------------------------------------------------------------------------

BlkChannelInfo & BlkChannelInfo::EnableChannels( int ch_count )
{
    int devices;
	if(FMuxChannels > 0)
		{
		FActiveMuxChannels = ch_count;
		ch_count = 1; 		// Special case for A16D2, 1ch, 16 mux ch.
		}

    if (ch_count%FChannelsPerDevice == 0)
        devices = ch_count/FChannelsPerDevice;
    else
        devices = 1+ch_count/FChannelsPerDevice;  // 'part-on' devices are on!
        
    for (int i = 0; i < Devices(); ++i)
        if (i<devices)
            FEnabled[i] = true;
        else
            FEnabled[i] = false;
            
    Update();        
    return *this;       
}
               
//-----------------------------------------------------------------------------
//  BlkChannelInfo::DisableAll()  --  Disable all channels
//-----------------------------------------------------------------------------

BlkChannelInfo & BlkChannelInfo::DisableAll()
{
    EnableDevices(-1);   

    Update();        
    return *this;       
}               

//-----------------------------------------------------------------------------
//  BlkChannelInfo::EnableAll()  --  Enable all channels
//-----------------------------------------------------------------------------

BlkChannelInfo & BlkChannelInfo::EnableAll()
{
    EnableDevices(Devices());

	if(FMuxChannels > 0)
		FActiveMuxChannels = FMuxChannels;

    Update();
    return *this;       
}               

//-----------------------------------------------------------------------------
//  BlkChannelInfo::DeviceOf()  -- Get device number for channel
//-----------------------------------------------------------------------------

int  BlkChannelInfo::DeviceOf(int channel) const
{
    return channel / FChannelsPerDevice;
}               

//-----------------------------------------------------------------------------
//  BlkChannelInfo::DevEnabled()  --  Retrieve device enable status
//-----------------------------------------------------------------------------

bool BlkChannelInfo::DevEnabled(int ch) const          
{
	return FEnabled[ch] ? true : false;
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::Enabled()  --  Retreive channel enable status
//-----------------------------------------------------------------------------

bool BlkChannelInfo::Enabled(int ch) const          
{
	return FEnabled[DeviceOf(ch)] ? true : false;
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::BuildMaps()  --  
//-----------------------------------------------------------------------------

void BlkChannelInfo::BuildMaps()
{
	FPhysicalToActiveChannel.clear();
	FActiveToPhysicalChannel.clear();
    int found = 0;
    for (int i=0; i<Channels(); ++i)
    	{
		FPhysicalToActiveChannel.push_back(Enabled(i) ? found++ : -1);
		FActiveToPhysicalChannel.push_back(Enabled(i) ? i : -1);
	    }

	FPhysicalToActiveDevice.clear();
	FActiveToPhysicalDevice.clear();
    found = 0;
    for (int i=0; i<Devices(); ++i)
    	{
	    FPhysicalToActiveDevice.push_back(DevEnabled(i) ? found++ : -1);
	    FActiveToPhysicalDevice.push_back(DevEnabled(i) ? i : -1);
	    }
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::ActiveChannelIndex()  --  Reindex active channels 
//-----------------------------------------------------------------------------
//  Returns the physical channel index given an active channel index
//   ...active channel idx is 0 based like channel indexes are.
//  (Returns -1 if no such active channel exists)

int BlkChannelInfo::ActiveChannelIndex(int N) const          
{
	return FActiveToPhysicalChannel[N];
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::ActiveDeviceIndex()  --  Reindex active devices 
//-----------------------------------------------------------------------------
//  Returns the physical device index given an active device index
//   ...active device idx is 0 based like device indexes are.
//  (Returns -1 if no such active device exists)

int BlkChannelInfo::ActiveDeviceIndex(int N) const          
{
	return FActiveToPhysicalDevice[N];
} 
      
//-----------------------------------------------------------------------------
//  BlkChannelInfo::PhysicalChannelIndex()  --  Reindex active channels 
//-----------------------------------------------------------------------------
//  Returns the active channel index given a physical channel index
//   ...physical channel idx is 0 based like active channel indexes are.
//  (Returns -1 if physical channel is disabled)

int BlkChannelInfo::PhysicalChannelIndex(int N) const          
{
	return FPhysicalToActiveChannel[N];
}

//-----------------------------------------------------------------------------
//  BlkChannelInfo::PhysicalDeviceIndex()  --  Reindex active devices 
//-----------------------------------------------------------------------------
//  Returns the active device index given a physical device index
//   ...physical device idx is 0 based like active channel indexes are.
//  (Returns -1 if physical device is disabled)

int BlkChannelInfo::PhysicalDeviceIndex(int N) const          
{
	return FPhysicalToActiveDevice[N];
} 
      
//=============================================================================
//  CLASS ChannelInfo  --  Channel Info allowing individual device enables
//=============================================================================

//-----------------------------------------------------------------------------
//  ChannelInfo::ChannelInfo()  --  Ctor
//-----------------------------------------------------------------------------

ChannelInfo::ChannelInfo(int devices, int ch_per_dev, int bits, int muxChannelCount, bool unipolar)
    : BlkChannelInfo(devices, ch_per_dev, bits, muxChannelCount, unipolar)
{
}
    
//-----------------------------------------------------------------------------
//  ChannelInfo::DevEnabled()  --  Set device enabled status
//-----------------------------------------------------------------------------

ChannelInfo & ChannelInfo::DevEnabled(int dev, bool state)
{
    FEnabled[dev]=state;      
    
    Update();  

    return *this;
}

//-----------------------------------------------------------------------------
//  ChannelInfo::Enabled()  --  Set channel enable status
//-----------------------------------------------------------------------------

ChannelInfo & ChannelInfo::Enabled(int ch, bool state)
{
    FEnabled[DeviceOf(ch)]=state;      
    Update();  

    return *this;
}

//-----------------------------------------------------------------------------
//  ChannelInfoPtr()  --  downcast BlkChannelInfo to ChannelInfo *
//-----------------------------------------------------------------------------

ChannelInfo * ChannelInfoPtr(BlkChannelInfo *blkCI)
{
    return dynamic_cast<ChannelInfo *>(blkCI);
}

//=============================================================================
//  CLASS PairedBlkChannelInfo  --  Channel Info for 16 bit or less paired hdw
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class PairedBlkChannelInfo 
//-----------------------------------------------------------------------------

PairedBlkChannelInfo::PairedBlkChannelInfo(int devices, int bits, int muxChannelCount, bool unipolar)
    :  BlkChannelInfo(devices, 2, bits, muxChannelCount, unipolar)
{

}

//=============================================================================
//  CLASS UnpairedBlkChannelInfo  --  Channel Info for 16 bit + unpaired hdw
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class UnpairedBlkChannelInfo 
//-----------------------------------------------------------------------------

UnpairedBlkChannelInfo::UnpairedBlkChannelInfo(int devices, int bits, int muxChannelCount, bool unipolar)
    :  BlkChannelInfo(devices, 1, bits, muxChannelCount, unipolar)
{
}

//=============================================================================
//  CLASS PairedChannelInfo  --  Channel Info for 16 bit or less paired hdw
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class PairedChannelInfo 
//-----------------------------------------------------------------------------

PairedChannelInfo::PairedChannelInfo(int devices, int bits, int muxChannelCount, bool unipolar)
    :  ChannelInfo(devices, 2, bits, muxChannelCount, unipolar)
{

}

//=============================================================================
//  CLASS UnpairedChannelInfo  --  Channel Info for 17 bit or more unpaired hdw
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class UnpairedChannelInfo 
//-----------------------------------------------------------------------------

UnpairedChannelInfo::UnpairedChannelInfo(int devices, int bits, int muxChannelCount, bool unipolar)
    :  ChannelInfo(devices, 1, bits, muxChannelCount, unipolar)
{

}

} // namespace Innovative

