// CoffFile_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure 
//  agreement with Innovative Integration Corporation and may not be copied or 
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation. 
//  All Rights Reserved.
//

#ifndef CoffFile_MbH
#define CoffFile_MbH

#include <string>
#include <map>
#include <vector>
#include <Events_Mb.h>
#include <stdio.h>
#include <ProcessEvents_Mb.h>

#include <Version_Mb.h>
#include <Params_Mb.h>

struct filehdr;
struct aouthdr;
struct scnhdr;
struct RELOC;
struct syment;
union auxent;
struct lineno;

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//===========================================================================
// MemWriteEvent CLASS:  Target section write callback event
//===========================================================================

class MemWriteEvent : public OpenWire::Event
{
public:
  const unsigned char *Buffer;
  const unsigned int Bytes;
  const unsigned int Addr;
  const unsigned char Page;

public:
  MemWriteEvent( unsigned char *buffer, unsigned int bytes, unsigned int addr, unsigned char page ) :
    Buffer( buffer ),
    Bytes( bytes ),
    Addr( addr ),
    Page( page )
      {
      }
};

//---------------------------------------------------------------------------
typedef std::map<std::string, std::string> StringMap;
typedef std::vector<std::string> Strings;
//---------------------------------------------------------------------------

//===========================================================================
// CoffFile CLASS:  C6x-compatible COFF4 loader class
//===========================================================================

class CoffFile
{
    friend class CoffThread;
    
public:
    enum IIEndian { DontCare, Little, Big };

    CoffFile();
    ~CoffFile();

    // Methods
    void            ResetProperties();
    int             Load(std::string Filename);
    int             Parse(std::string Filename);

    // Properties
    std::string TimeStamp() { return FTimeStamp; }
    bool  RelocationInfo() { return FRelocationInfo; }
    bool  ExternalReferences() { return FExternalReferences; }
    int   Magic() { return FMagic; }
    float VersionStamp() { return FVersionStamp; }
    int   Sections() { return FSections; }
    int   ExecutableCodeSize() { return FExecutableCodeSize; }
    int   ErrorCode();
    unsigned int InitializedDataSize() { return FInitializedDataSize; }
    unsigned int UninitializedDataSize() { return FUninitializedDataSize; }
    unsigned int EntryPoint() { return FEntryPoint; }
    unsigned int InitializedDataStart() { return FInitializedDataStart; }
    unsigned int TextStart() { return FTextStart; }
    Innovative::Strings     SectionInfo() { return FSectionInfo; }
    Innovative::StringMap   SymbolList() { return FSymbolList; }

public:
    OpenWire::EventHandler<ProcessProgressEvent>          OnProgress;
    OpenWire::EventHandler<ProcessCompletionEvent>        OnComplete;
    OpenWire::EventHandler<ProcessStatusEvent>            OnLoadMsg;
    OpenWire::EventHandler<Innovative::MemWriteEvent>     OnMemWrite;
    void  Verbose(bool state);

private:
    void Endian( Innovative::CoffFile::IIEndian endian);
    void ClearBss(bool state);
    void Symbols(bool state);
    void Data(bool state);

private:
    // Data
    int                             FFileMagic;
    std::string                     FTimeStamp;
    bool                            FRelocationInfo;
    bool                            FExternalReferences;
    int                             FMagic;
    float                           FVersionStamp;
    int                             FSections;
    unsigned int                    FExecutableCodeSize;
    unsigned int                    FInitializedDataSize;
    unsigned int                    FUninitializedDataSize;
    unsigned int                    FEntryPoint;
    unsigned int                    FInitializedDataStart;
    unsigned int                    FTextStart;
    unsigned int                    FDownloadSize;
    unsigned int                    FDownloadProgress;
    Innovative::Strings             FSectionInfo;
    Innovative::StringMap           FSymbolList;
    Innovative::CoffFile::IIEndian  FEndian;
    bool                            FClearBss;
    bool                            FSymbols;
    bool                            FData;

    // Methods
    bool BigEndian();
    bool ByteSwapped();

    /*----------------------------------------------------------------------------*/
    /* THIS STRUCTURE IS USED TO STORE THE RELOCATION AMOUNTS FOR SYMBOLS.        */
    /* EACH RELOCATABLE SYMBOL HAS A CORRESPONDING ENTRY IN THIS TABLE.           */
    /* THE TABLE IS SORTED BY SYMBOL INDEX; LOOKUP USES A BINARY SEARCH.          */
    /*----------------------------------------------------------------------------*/
    typedef struct
    {
       long rt_index;                       /* INDEX OF SYMBOL IN SYMBOL TABLE    */
       long rt_disp;                        /* AMOUNT OF RELOCATION               */
    } RELOC_TAB;

    /*----------------------------------------------------------------------------*/
    /* DEFINE A STRUCTURE FOR STRING TABLE BUFFERS.  THESE BUFFERS ARE STORED     */
    /* AS A LINKED LIST OF MEMORY PACKETS, EACH CONTAINING UP TO 64K OF THE       */
    /* STRING TABLE.                                                              */
    /*----------------------------------------------------------------------------*/
    typedef struct strtab
    {
       unsigned long  size;                 /* SIZE OF THIS PACKET                */
       unsigned long  offset;               /* STARTING OFFSET OF THIS PACKET     */
       unsigned long  index;                /* AMOUNT CURRENTLY FILLED            */
       struct strtab *next;                 /* NEXT BUFFER                        */
       char           buf[1];               /* STRING DATA (EXPAND AS ALLOCATED)  */
    } STRTAB;

    /*----------------------------------------------------------------------------*/
    /* APPLICATION VARIABLES                                                      */
    /*----------------------------------------------------------------------------*/
    FILE *   fin;                           /* INPUT FILE                         */
    int     verbose;                        /* PRINT PROGRESS INFO                */
    int     need_data;                      /* APPLICATION NEEDS RAW DATA         */
    int     need_symbols;                   /* APPLICATION NEEDS SYMBOL TABLE     */
    int     clear_bss;                      /* CLEAR BSS SECTION                  */

    /*----------------------------------------------------------------------------*/
    /* LOADER VARIABLES                                                           */
    /*----------------------------------------------------------------------------*/
    struct filehdr *file_hdr;                       /* FILE HEADER STRUCTURE              */
    int     coff_version;                   /* COFF VERSION USED IN FILE          */
    struct aouthdr *o_filehdr;                      /* OPTIONAL (A.OUT) FILE HEADER       */
    unsigned long  entry_point;                    /* ENTRY POINT OF MODULE              */
    unsigned long  *reloc_amount;                   /* AMOUNT OF RELOCATION PER SECTION   */
    char   *sect_hdrs;                      /* ARRAY OF SECTION HEADERS           */
    char   *o_sect_hdrs;                    /* ARRAY OF OLD COFF SECTION HEADERS  */
    int     n_sections;                     /* NUMBER OF SECTIONS IN THE FILE     */
    int     big_e_target;                   /* TARGET DATA IN BIG-ENDIAN FORMAT   */
    int     byte_swapped;                   /* BYTE ORDERING OPPOSITE OF HOST     */
    int     curr_sect;                      /* SECTION NUMBER CURRENTLY LOADING   */
    int     load_err;                       /* ERROR CODE RETURNED IF LOADER FAILS*/
    struct strtab *str_head;                /* LIST OF STRING TABLE BUFFERS       */

    unsigned short init_size;                      /* CURRENT SIZE OF C INITIALIZATION   */
    int     need_reloc;                     /* RELOCATION REQUIRED                */

    int big_e_config;                       /* ENDIANNESS CONFIGURATION           */

    RELOC_TAB * reloc_tab;                  /* RELOCATION SYMBOL TABLE            */

    int         reloc_tab_size;             /* CURRENT ALLOCATED AMOUNT           */
    int         reloc_sym_index;            /* CURRENT SIZE OF TABLE              */

    /*--------------------------------------------------------------------------*/
    /* LOADER DEFINITIONS                                                       */
    /*--------------------------------------------------------------------------*/

    unsigned long reloc;                    /* RELOCATION AMOUNT               */
    int           quiet;                    /* NO BANNER                       */
    FILE*         cofflogfile;

    //
    //  Local Statics turned into Class Members
    T_ADDR   cload_cinit_memberftn_init_addr;     // in cload_cinit() member ftn
    char     sym_name_memberftn_temp[9];          // in sym_name() member ftn

    /*--------------------------------------------------------------------------*/
    /*			  CLOAD.C PROTOTYPES				    */
    /*--------------------------------------------------------------------------*/
    int	    cload ();
    int	    cload_headers ();
    int	    cload_data ();
    int	    cload_sect_data (struct scnhdr *);
    int	    cload_cinit (unsigned char *, int *, int *);
    int	    cload_symbols ();
    int     cload_strings ();
    void    str_free (struct strtab *);
    long	sym_read (long, struct syment *, union auxent *);
    char *  sym_name (struct syment *);
    char *  sym_add_name (struct syment *);
    int	    reloc_add (long, struct syment *);
    int 	relocate ( struct RELOC *, unsigned char *, int);
    int	    reloc_read (struct RELOC *);
    int	    reloc_size (int);
    int	    reloc_offset (int);
    int     reloc_stop (int);
    long	sym_reloc_amount (struct RELOC *);
    unsigned long unpack (unsigned char *, int, int, int);
    void	repack (unsigned long, unsigned char *, int,int,int);
    int	    cload_lineno (long, int, struct lineno *, int);
    void	swap4byte (void *);
    void	swap2byte (void *);

    /*--------------------------------------------------------------------------*/
    /*			  MAIN.C PROTOTYPES				    */
    /*--------------------------------------------------------------------------*/
    void load_msg(const char* a, size_t b, size_t c, size_t d, size_t e, size_t f, size_t g);

    /*--------------------------------------------------------------------------*/
    /*  LOADER REQUIRED FUNCTIONS			                            */
    /*--------------------------------------------------------------------------*/
    int     mem_write ( unsigned char *buffer, unsigned int nbytes, unsigned int addr, unsigned char page );
    int     set_reloc_amount (void);
    void    lookup_sym       (long indx, syment *sym, auxent *aux);
    int     load_syms        (int need_reloc);
    void *  myalloc          (int size);
    void *  mralloc          (char *p, int size);

    int     iicoffld( const char *filename);

    void Error( std::string message );     
};

#ifdef __CLR_VER
#pragma managed(pop)
#endif
}; // namespace Innovative
//---------------------------------------------------------------------------
#endif
