// CommonFilter_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure 
//  agreement with Innovative Integration Corporation and may not be copied or 
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation. 
//  All Rights Reserved.
//
 
#include <CommonFilter_Mb.h>
#include <ipps.h>   // ipps.h is not redistributable!
#include <IppFloatDG_Mb.h>

namespace   Innovative
{

//---------------------------------------------------------------------------
//  ApplyWindow() --
//---------------------------------------------------------------------------
 
void ApplyWindow(float*  coef, size_t coef_count, WindowTypes window_type,
                 float alpha)
{
    int count = static_cast<int>(coef_count);

    switch (window_type)
        {
        case wtBartlett:
            ippsWinBartlett_32f_I(coef, count);
            break;
        case wtBlackman:
            ippsWinBlackman_32f_I(coef, count, alpha);
            break;
        case wtBlackmanOpt:
            ippsWinBlackmanOpt_32f_I(coef, count);
            break;
        case wtHamming:
            ippsWinHamming_32f_I(coef, count);
            break;
        case wtHanning:
            ippsWinHann_32f_I(coef, count);
            break;
        case wtKaiser:
            ippsWinKaiser_32f_I(coef, count, alpha);
            break;
        case wtRect:
            break;
        }
}

//===========================================================================
//  CLASS CommonFilter  --  
//===========================================================================

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif

//---------------------------------------------------------------------------
// CommonFilter::CommonFilter() --
//---------------------------------------------------------------------------

CommonFilter::CommonFilter()
	: FAlpha(1), FWindowType(wtRect), FCoefficient(91), FSampleRate(100000.0)
{
}

//---------------------------------------------------------------------------
// CommonFilter::CommonFilter() --
//---------------------------------------------------------------------------

CommonFilter::CommonFilter(const CommonFilter & other)
	: FAlpha(other.FAlpha), FWindowType(other.FWindowType),	
	  FCoefficient(other.FCoefficient), FSampleRate(other.FSampleRate),
	  FWork(other.FWork), FSpec(other.FSpec)
{
}

#ifdef _MSC_VER
	#pragma warning(pop)
#endif

//---------------------------------------------------------------------------
// CommonFilter::~CommonFilter() --
//---------------------------------------------------------------------------

CommonFilter::~CommonFilter()
{
	Stop();
}

//---------------------------------------------------------------------------
//  CommonFilter::operator = --
//---------------------------------------------------------------------------

CommonFilter & CommonFilter::operator = (const CommonFilter & other)
{
	FAlpha = other.FAlpha;
	FWindowType = other.FWindowType;
	FCoefficient = other.FCoefficient;
	FSampleRate = other.FSampleRate;
	FWork = other.FWork;
	FSpec = other.FSpec;

	return *this;
}

//---------------------------------------------------------------------------
//  CommonFilter::Stop() --
//---------------------------------------------------------------------------

void CommonFilter::Stop()
{
}

//---------------------------------------------------------------------------
//  CommonFilter::Start() --
//---------------------------------------------------------------------------

bool CommonFilter::Start()
{
	bool result = true;

	Stop();

	// Allocate Tap and Delay line buffers
	// Design the filter taps
	CalculateCoefficients();
	size_t num_taps = FCoefficient.size();

	//get sizes of the spec structure and the work buffer
	int specSize, bufSize;
	IppStatus status = ippsFIRSRGetSize(static_cast<int>(num_taps), ipp32f, &specSize, &bufSize);
	result &= (status == ippStsNoErr);

	FDelay.resize(num_taps-1);
	FWork.resize(bufSize);
	FSpec.resize(specSize);

	//initialize the spec structure
	status = ippsFIRSRInit_32f(reinterpret_cast<Ipp32f*>(&(FCoefficient.front())),
		static_cast<int>(num_taps), ippAlgDirect, reinterpret_cast<IppsFIRSpec_32f*>(&(FSpec.front())) );
	result &= (status == ippStsNoErr);

	return result;
}

//---------------------------------------------------------------------------
//  CommonFilter::Filter() --
//---------------------------------------------------------------------------

bool CommonFilter::Filter(const IppFloatDG & in, IppFloatDG & out)
{
	bool result = true;

	if (!FSpec.size())
		{
		Start();
		}
	else
		{
		IppStatus status = ippsFIRSR_32f(in.Ptr(),
										out.Ptr(),
										static_cast<int>(out.size()),
										reinterpret_cast<IppsFIRSpec_32f*>(&(FSpec.front())),
										NULL,
										reinterpret_cast<Ipp32f*>(&(FDelay.front())),
										reinterpret_cast<Ipp8u*>(&(FWork.front())));

		result &= (status == ippStsNoErr);
		}

	return result;
}

//---------------------------------------------------------------------------
// CommonFilter::SetOrder() --  Specify number of filter taps
//---------------------------------------------------------------------------

void CommonFilter::SetOrder(size_t count)
{
	if (FSpec.size())
		Stop();

	FCoefficient.resize(count+1);

    Start();
}

//---------------------------------------------------------------------------
// CommonFilter::Order() --  Specify number of filter taps
//---------------------------------------------------------------------------

void CommonFilter::Order(size_t count)
{
    SetOrder(count);
}

//---------------------------------------------------------------------------
// CommonFilter::Order() -- Query order of filter
//---------------------------------------------------------------------------

size_t CommonFilter::Order() const
{
	return FCoefficient.size()-1;
}

//---------------------------------------------------------------------------
// CommonFilter::Taps() -- Query number of filter taps
//---------------------------------------------------------------------------

size_t CommonFilter::Taps() const
{
	return FCoefficient.size();
}

//---------------------------------------------------------------------------
//  CommonFilter::Alpha() --  Setter
//---------------------------------------------------------------------------

void CommonFilter::Alpha(float value)
{
    FAlpha = value;
    Start();
}

//---------------------------------------------------------------------------
//  CommonFilter::Alpha() --  Getter
//---------------------------------------------------------------------------
 
float CommonFilter::Alpha() const
{
    return FAlpha;
}

//---------------------------------------------------------------------------
//  CommonFilter::SampleRate() --  Setter
//---------------------------------------------------------------------------

void CommonFilter::SampleRate(float value)
{
    FSampleRate = value;
	Start();
}

//---------------------------------------------------------------------------
//  CommonFilter::SampleRate() --  Getter
//---------------------------------------------------------------------------
 
float CommonFilter::SampleRate() const
{
    return FSampleRate;
}

//---------------------------------------------------------------------------
//  CommonFilter::WindowType() --
//---------------------------------------------------------------------------

void CommonFilter::WindowType(Innovative::WindowTypes value)
{
    FWindowType = value;
    Start();
}

//---------------------------------------------------------------------------
//  CommonFilter::WindowType() --
//---------------------------------------------------------------------------

WindowTypes CommonFilter::WindowType() const
{
    return FWindowType;
}

} // namespace Innovative

