//---------------------------------------------------------------------------
// DAC3xJ82_SubDevice_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2016 Innovative Integration Corporation.
//  All Rights Reserved.

#include <DAC3xJ82_SubDevice_Mb.h>
#include <SystemSupport_Mb.h>
#include <cmath>

using namespace std;

namespace Innovative
{

#ifdef _MSC_VER
    #pragma warning(push)
    #pragma warning(disable : 4355)
#endif

//===========================================================================
//  CLASS DAC3xJ82_DAC -- JESD A/D Support "SubDevice"
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class DAC3xJ82_DAC
//---------------------------------------------------------------------------

const unsigned int DacNumConverters = 2;  // constant for DAC38J82
const unsigned int DacLaneCount     = 8;

//---------------------------------------------------------------------------
//  constructor for class DAC3xJ82_DAC
//---------------------------------------------------------------------------

DAC3xJ82_DAC::DAC3xJ82_DAC(IRequires_DAC3xJ82_Dac_Io * mdc, unsigned int jesd_device_id)
    :  LanePN_Inverted(DacLaneCount),  PhysToLogical_LaneMap(DacLaneCount),
       MDC(mdc), DeviceID(jesd_device_id)
{
}

//---------------------------------------------------------------------------
//  destructor for class DAC3xJ82_DAC
//---------------------------------------------------------------------------

DAC3xJ82_DAC::~DAC3xJ82_DAC()
{
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::ConstructCustomParts()
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::ConstructCustomParts()
{
    if (!DeviceEnabled())
        return;

    MDC->SetDacLaneConfiguration(DeviceID);
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::OpenHardware()
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::OpenHardware()
{
    if (!DeviceEnabled())
        return;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::ConfigureHardware()
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;

    DacInit();
    ConfigureDacPll();
    ConfigureDacSerDesPll();
    ConfigureDacSerDes();
    ConfigureJesdParams();
    ConfigureDacDigBlocks();
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::LoadDacClockRates() - load needed clock rates
//---------------------------------------------------------------------------

void DAC3xJ82_DAC::LoadDacClockRates(double dacSamplingRate, double dacEffectiveRate)
{
    Debug.Log("DAC3xJ82_DAC::LoadDacClockRates()");

    DacSamplingRate = dacSamplingRate;
    DacClkRate      = dacEffectiveRate;
    Debug.Log("end DAC3xJ82_DAC::LoadDacClockRates()");
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::DacInit()
//---------------------------------------------------------------------------

void DAC3xJ82_DAC::DacInit()
{
    Debug.Log("DAC3xJ82_DAC::DacInit()");

    Debug.Log("DAC Interpolation Rate = ", InterpolationRate);
    Debug.Log("DAC Serial Line Rate(Gbps) = ", (LineRate/1.0e9));
    Debug.Log("DAC Number of serial lanes = ", NumLanes);
    Debug.Log("DAC Number of converters = ", DacNumConverters);

    // Soft reset the DAC
    Spi(0x2,0x2003);
    Sleep(10);

    // Read the Vendor and Version ID register
    Debug.Log("Dac Vendor and Version ID config127 = ", Spi(0x7f), 16);

    // Read the die ID
    MemoryBitManipulator Data(0);

    // Enable the die ID read
    Data.Value(Spi(0x2F));
    Data.Bit(14, 1);
    Spi(0x2F, Data.Value());

    // Read the 128-bit die ID registers
    Debug.Log("Dac die ID config100 =", Spi(0x64), 16);
    Debug.Log("Dac die ID config101 =", Spi(0x65), 16);
    Debug.Log("Dac die ID config102 =", Spi(0x66), 16);
    Debug.Log("Dac die ID config103 =", Spi(0x67), 16);
    Debug.Log("Dac die ID config104 =", Spi(0x68), 16);
    Debug.Log("Dac die ID config105 =", Spi(0x69), 16);
    Debug.Log("Dac die ID config106 =", Spi(0x6A), 16);
    Debug.Log("Dac die ID config107 =", Spi(0x6B), 16);

    // Disable the die ID read
    Data.Value(Spi(0x2F));
    Data.Bit(14, 0);
    Spi(0x2F, Data.Value());

    Debug.Log("end DAC3xJ82_DAC::DacInit()");
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::ConfigureDacPll()
//---------------------------------------------------------------------------

void DAC3xJ82_DAC::ConfigureDacPll()
{
    Debug.Log("DAC3xJ82_DAC::ConfigureDacPll()");

    DacPllEnabled = (DacSamplingRate == DacClkRate)? false : true;

    Debug.Log("DacSamplingRate(MHz) =", (DacSamplingRate/1.0e6));
    Debug.Log("DacClkRate(MHz) =", (DacClkRate/1.0e6));

    CalcDacPllParams();

    MemoryBitManipulator Data(0);

    // Set config26 register
    Data.Value(Spi(0x1A));
    Data.Bit(0, (((DacCh[1].CouplingMode == ocmAc)? true : false) || !ChannelEnabled(1)));  // set dacd_sleep
    Data.Bit(1, (((DacCh[1].CouplingMode == ocmDc)? true : false) || !ChannelEnabled(1)));  // set dacc_sleep
    Data.Bit(2, (((DacCh[0].CouplingMode == ocmAc)? true : false) || !ChannelEnabled(0)));  // set dacb_sleep
    Data.Bit(3, (((DacCh[0].CouplingMode == ocmDc)? true : false) || !ChannelEnabled(0)));  // set daca_sleep
    Data.Bit(5, !DacPllEnabled);  // set pll_sleep
    Spi(0x1A, Data.Value());

    // Set config49 register
    Data.Value(Spi(0x31));
    Data.Field(3, 5, Pll_N);
    Data.Bit(10, DacPllEnabled);   // set pll_ena
    Data.Bit(12, !DacPllEnabled);  // set pll_reset
    Spi(0x31, Data.Value());

    if (DacPllEnabled)
        {
        // Set config50 register
        Data.Value(Spi(0x32));
        Data.Field(4, 4, Pll_P);
        Data.Field(8, 8, Pll_M);
        Spi(0x32, Data.Value());

        // Set config51 register
        Data.Value(Spi(0x33));
        Data.Field(2, 5, Pll_Cp_Adj);
        Data.Field(7, 2, Pll_VcoITune);
        Data.Field(9, 6, Pll_Vco);
        Data.Bit(15, Pll_VcoSel);
        Spi(0x33, Data.Value());

        // Check the loop filter voltage
        Sleep(1);
        Data.Value(Spi(0x31));
        unsigned int Pll_Lfvolt = Data.Value() & 0x7;
        Debug.Log("DAC PLL loop filter voltage =", Pll_Lfvolt);

        // Check the PLL lock
        Spi(0x6C, 0);            // Write 0 to clear previous alarms
        Data.Value(Spi(0x6C));   // Read the new alarms
        bool DacPllLock = (Data.Value() & 0x1) ? false : true;
        Debug.Log("DAC PLL locked: ", DacPllLock);
        }

    Debug.Log("end DAC3xJ82_DAC::ConfigureDacPll()");
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::CalcDacPllParams()
//---------------------------------------------------------------------------

void DAC3xJ82_DAC::CalcDacPllParams()
{
    Debug.Log("DAC3xJ82_DAC::CalcDacPllParams()");

    if (!DacPllEnabled)
        {
        Pll_N = Pll_M = Pll_P = 0;
        Pll_VcoSel = false;
        Pll_Vco = 0;
        Pll_VcoITune = 0;
        Pll_Cp_Adj = 0;
        Debug.Log("Dac PLL disabled");
        return;
        }

    // Calculate the Prescaler value
    const double VcoLmin = 3700.0e6;
    const double VcoLmax = 4660.0e6;
    const double VcoHmin = 4440.0e6;
    const double VcoHmax = 5600.0e6;
    bool         InVcoLRange;
    bool         InVcoHRange;
    double VcoFreq = 5000.0e6;  // Initialize to a high value

    const unsigned int Pll_PAry[] = {2,3,4,5,6,7,8,9,4,6,8,10,12,14,16,18};
    const unsigned int Pll_PArySize = sizeof(Pll_PAry)/sizeof(unsigned int);
    double freq;

    // Initialize
    Pll_P = 0;

    // Loop over the prescaler values to find the max value
    for (unsigned int i=0; i<Pll_PArySize; i++)
        {
        // Calculate the candidate Vco frequency
        freq = Pll_PAry[i] * DacSamplingRate;

        InVcoLRange = false;
        InVcoHRange = false;
        if (freq >= VcoLmin && freq <= VcoLmax)
            InVcoLRange = true;
        if (freq >= VcoHmin && freq <= VcoHmax)
            InVcoHRange = true;

        // Latch the Vco frequency and prescaler index
        if ((InVcoLRange || InVcoHRange) && (Pll_PAry[i]>=Pll_PAry[Pll_P]))
            {
            VcoFreq = freq;
            Pll_P = i;
            }
        }

    // The 2 VCOs overlap over the range of VcoHmin-VcoLmax, calculate the center
    // frequency of this range to use it to select one of the VCOs
    double VcoCenter = (VcoHmin + VcoLmax)/2;

    // Select the 4GHz or 5GHz Vco (true for 4GHz Vco)
    Pll_VcoSel = (VcoFreq < VcoCenter)? true : false;

    Debug.Log("DAC PLL VcoFreq(MHz) =", VcoFreq/1.0e6);
    Debug.Log("DAC PLL P =", Pll_PAry[Pll_P]);

    double VcoError;
    double MinVcoError = 5000.0;

    // Loop over the Pll_Vco values and find the closest matching value
    for (unsigned int VcoRangeAdj=0; VcoRangeAdj<64; VcoRangeAdj++)
        {
        if (Pll_VcoSel)     // 4GHz Vco
            VcoError = (0.089703*VcoRangeAdj*VcoRangeAdj)+(8.8312*VcoRangeAdj)+3752.5-(VcoFreq/1.0e6);
        else                // 5GHz Vco
            VcoError = (0.10998*VcoRangeAdj*VcoRangeAdj)+(10.574*VcoRangeAdj)+4446.3-(VcoFreq/1.0e6);

        if (fabs(VcoError) < MinVcoError)
            {
            MinVcoError = fabs(VcoError);
            Pll_Vco = VcoRangeAdj;
            }
        }

    Debug.Log("DAC PLL MinVcoError =", MinVcoError);
    Debug.Log("DAC PLL Pll_Vco =", Pll_Vco);

    // Set the core bias current depending on the selected Vco
    Pll_VcoITune = (Pll_VcoSel)? 2 : 3;

    double PFD;
    const double MaxPFD = 650.0e6;
    bool   PllMFound;

    // Initialize (to be used when no match is found)
    Pll_N = 0;
    Pll_M = 0;

    // Loop over the Pll_N and Pll_M values to find the proper settings
    for (unsigned int N=1; N<=32; N++)
        {
        PllMFound = false;
        PFD = DacClkRate/N;
        if (PFD > MaxPFD)  // skip if PFD is too fast
            continue;
        for (unsigned int M=1; M<=256; M++)
            {
            if (PFD * M == DacSamplingRate)
                {
                Pll_N = N-1;
                Pll_M = M-1;
                PllMFound = true;
                Debug.Log("DAC PLL PFD(MHz) =", PFD/1.0e6);
                Debug.Log("DAC PLL N =", N);
                Debug.Log("DAC PLL M =", M);
                break;
                }
            }
        if (PllMFound)
            break;
        }

    // Hardcoded value set in the datasheet for PFD=625MHz
    Pll_Cp_Adj = 12;

    Debug.Log("end DAC3xJ82_DAC::CalcDacPllParams()");
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::ConfigureDacSerDesPll()
//---------------------------------------------------------------------------

void DAC3xJ82_DAC::ConfigureDacSerDesPll()
{
    Debug.Log("DAC3xJ82_DAC::ConfigureDacSerDesPll()");

    CalcDacSerDesPllParams();

    MemoryBitManipulator Data(0);

    // Set config59 register
    Data.Value(Spi(0x3B));
    Data.Field(11, 4, SerDesPllRefClkDiv);
    Data.Bit(15, DacPllEnabled);  // set serdes_clk_sel
    Spi(0x3B, Data.Value());

    // Set config60 register
    Data.Value(Spi(0x3C));
    Data.Field(1, 8, SerDesPllMpy);
    Data.Bit(9, SerDesPllVRange);
    Spi(0x3C, Data.Value());

    Debug.Log("end DAC3xJ82_DAC::ConfigureDacSerDesPll()");
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::CalcDacSerDesPllParams()
//---------------------------------------------------------------------------

void DAC3xJ82_DAC::CalcDacSerDesPllParams()
{
    Debug.Log("DAC3xJ82_DAC::CalcDacSerDesPllParams()");

    double SerDesPllOutFreq;

    if (LineRate > 6250.0e6)
        {
        DacSerDesLaneRate = lrFull;
        SerDesPllOutFreq = 0.25 * LineRate;
        }
    else if (LineRate > 3125.0e6)
        {
        DacSerDesLaneRate = lrHalf;
        SerDesPllOutFreq = 0.5 * LineRate;
        }
    else if (LineRate > 1562.5e6)
        {
        DacSerDesLaneRate = lrQuarter;
        SerDesPllOutFreq = LineRate;
        }
    else
        {
        DacSerDesLaneRate = lrEighth;
        SerDesPllOutFreq = 2 * LineRate;
        }

    Debug.Log("DAC SerDes PLL output frequency(MHz) =", SerDesPllOutFreq/1.0e6);

    SerDesPllVRange = (SerDesPllOutFreq < 2170.0e6)? true : false;

    const unsigned int SerDesPllMpyX4Ary[] = {16,20,24,32,33,40,48,50,60,64,66,80,88,100};
    const unsigned int SerDesPllMpyX4ArySize = sizeof(SerDesPllMpyX4Ary)/sizeof(double);

    double PFD;
    const double MaxPFD = 800.0e6;
    bool   SerDesPllMPyFound;

    // Initialize (to be used when no match is found)
    SerDesPllRefClkDiv = 0;
    SerDesPllMpy = 16;

    // Loop over the SerDes PLL RefClkDiv and MPY values to find the proper settings
    for (unsigned int N=1; N<=16; N++)
        {
        SerDesPllMPyFound = false;
        PFD = DacSamplingRate/N;
        if (PFD > MaxPFD)  // skip if PFD is too fast
            continue;
        for (unsigned int i=0; i<SerDesPllMpyX4ArySize; i++)
            {
            if (PFD * SerDesPllMpyX4Ary[i]/4 == SerDesPllOutFreq)
                {
                SerDesPllRefClkDiv = N-1;
                SerDesPllMpy = SerDesPllMpyX4Ary[i];
                SerDesPllMPyFound = true;
                Debug.Log("DAC SerDes PLL PFD(MHz) =", PFD/1.0e6);
                Debug.Log("DAC SerDes PLL RefClk divider =", N);
                Debug.Log("DAC SerDes PLL Mpy =", (SerDesPllMpyX4Ary[i]/4));
                break;
                }
            }
        if (SerDesPllMPyFound)
            break;
        }

    Debug.Log("end DAC3xJ82_DAC::CalcDacSerDesPllParams()");
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::ConfigureDacSerDes()
//---------------------------------------------------------------------------

void DAC3xJ82_DAC::ConfigureDacSerDes()
{
    Debug.Log("DAC3xJ82_DAC::ConfigureDacSerDes()");

    // Set config61 register
    Spi(0x3D, 0x88);

    MemoryBitManipulator Data(0);

    // Set config62 register
    Data.Value(Spi(0x3E));
    Data.Field(2, 3, 2);          // set buswidth to 2 (20bit)
    Data.Field(5, 2, DacSerDesLaneRate);
    Data.Field(8, 3, 1);          // set term to 1
    Spi(0x3E, Data.Value());

    // Set config63 register
    MemoryBitManipulator PinPairInv(0);
    for (unsigned int i=0; i<LanePN_Inverted.size(); i++)
       PinPairInv.Bit(i, (LanePN_Inverted[i]) ? true : false);
    Spi(0x3F, PinPairInv.Value());

    // Set JESD lane IDs
    // Set config70 - config72 registers
    MemoryBitManipulator config70(0);
    config70.Field(11, 5, PhysToLogical_LaneMap[0]);
    config70.Field( 6, 5, PhysToLogical_LaneMap[1]);
    config70.Field( 1, 5, PhysToLogical_LaneMap[2]);
    Spi(0x46, config70.Value());

    MemoryBitManipulator config71(0);
    config71.Field(11, 5, PhysToLogical_LaneMap[3]);
    config71.Field( 6, 5, PhysToLogical_LaneMap[4]);
    config71.Field( 1, 5, PhysToLogical_LaneMap[5]);
    Spi(0x47, config71.Value());

    MemoryBitManipulator config72(0x3);
    config72.Field(11, 5, PhysToLogical_LaneMap[6]);
    config72.Field( 6, 5, PhysToLogical_LaneMap[7]);
    Spi(0x48, config72.Value());

    // Set config73 register
    Spi(0x49, 0x0);      // assign all JESD lanes to link0

    // Set config74 register
    //  ..calculate SerDes Lane Enables
    MemoryBitManipulator SerDesLaneEn(0);
    for (unsigned int i=0; i<PhysToLogical_LaneMap.size(); i++)
        SerDesLaneEn.Bit(i, (PhysToLogical_LaneMap[i]<NumLanes) ? true : false);

    Data.Value(0);
    Data.Field(0, 8, 0x1E);
    Data.Bit(5, !ChannelEnabled(1));     // (specific to FMC1000 since Channel1 Dac is connected to either DACC or DACD)
    Data.Field(8, 8, SerDesLaneEn.Value());
    Spi(0x4A, Data.Value());

    // Set SerDes lane to JESD lane mapping bits
    // Set config95 - config96 registers
    MemoryBitManipulator config95(0);
    config95.Field(12, 3, PhysToLogical_LaneMap[0]);
    config95.Field( 8, 3, PhysToLogical_LaneMap[1]);
    config95.Field( 4, 3, PhysToLogical_LaneMap[2]);
    config95.Field( 0, 3, PhysToLogical_LaneMap[3]);
    Spi(0x5F, config95.Value());

    MemoryBitManipulator config96(0);
    config96.Field(12, 3, PhysToLogical_LaneMap[4]);
    config96.Field( 8, 3, PhysToLogical_LaneMap[5]);
    config96.Field( 4, 3, PhysToLogical_LaneMap[6]);
    config96.Field( 0, 3, PhysToLogical_LaneMap[7]);
    Spi(0x60, config96.Value());


    Debug.Log("end DAC3xJ82_DAC::ConfigureDacSerDes()");
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::ConfigureJesdParams()
//---------------------------------------------------------------------------

void DAC3xJ82_DAC::ConfigureJesdParams()
{
    Debug.Log("DAC3xJ82_DAC::ConfigureJesdParams()");

    MemoryBitManipulator Data(0);

    // Set config0
    Data.Value(Spi(0x0));
    Data.Bit(1, DacEnInvSinc);
    Data.Field(8, 4, DacInterpolationRate);
    Data.Bit(13, DacEnQmc);
    Data.Bit(15, DacEnQmc);
    Spi(0x0, Data.Value());

    // Set config36
    Data.Value(Spi(0x24));
    Data.Field(4, 3, 2);    // set cdrvser_sysref_mode to use only the next sysref pulse
    Spi(0x24, Data.Value());

    // Calculate the Jesd clock divider
    unsigned int JesdClkDiv = (InterpolationRate * NumLanes)/DacNumConverters;
    Debug.Log("DAC JESD clock divider =", JesdClkDiv);

    // Set config37
    Data.Value(Spi(0x25));
    Data.Field(13, 3, log2(JesdClkDiv));
    Spi(0x25, Data.Value());

    // Set the jesd parameters
    unsigned int OctetsPerFrame;
    unsigned int SamplesPerFrame;
    bool         HighDensityMode;
    const unsigned int FramesPerMultiFrame = 32;
    const bool   EnableScrambler = false;

    switch (NumLanes)
        {
        default:
        case 1:
            OctetsPerFrame = 4;
            SamplesPerFrame = 1;
            HighDensityMode = false;
            break;
        case 2:
            OctetsPerFrame = 2;
            SamplesPerFrame = 1;
            HighDensityMode = false;
            break;
        case 4:
            OctetsPerFrame = 1;
            SamplesPerFrame = 1;
            HighDensityMode = true;
            break;
        case 8:
            OctetsPerFrame = 1;
            SamplesPerFrame = 2;
            HighDensityMode = true;
            break;
        }

    // Set config75
    Data.Value(Spi(0x4B));
    Data.Field(0, 8, OctetsPerFrame-1);
    Data.Field(8, 5, FramesPerMultiFrame-2);
    Spi(0x4B, Data.Value());

    // Set config76
    Data.Value(Spi(0x4C));
    Data.Field(0, 5, NumLanes-1);
    Data.Field(8, 5, FramesPerMultiFrame-1);
    Spi(0x4C, Data.Value());

    // Set config77
    Data.Value(Spi(0x4D));
    Data.Field(0, 5, SamplesPerFrame-1);
    Data.Field(8, 8, DacNumConverters-1);
    Spi(0x4D, Data.Value());

    // Set config78
    Data.Value(Spi(0x4E));
    Data.Field(0, 5, 0xF);      // n_m1 = 15
    Data.Bit(5, EnableScrambler);
    Data.Bit(6, HighDensityMode);
    Data.Field(8, 5, 0xF);      // nprime_m1 = 15
    Spi(0x4E, Data.Value());

    // Set config92
    Spi(0x5C, 0x1133);

    // Set config97
    Spi(0x61, 0x0211);   // Set SYNCB LVDS generation from link0

    // Enable ILA support
    JesdDac().Poke(0x8, 0x1);
    // Set the scrambling mode
    JesdDac().Poke(0xC, EnableScrambler);
    // Set sysref handling mode
    JesdDac().Poke(0x10, 0x0);
    // Set ILA multiframes
    JesdDac().Poke(0x14, 0x3);
    // Set the octets per frame F
    JesdDac().Poke(0x20, (OctetsPerFrame-1));
    // Set the frames per multi frame K
    JesdDac().Poke(0x24, (FramesPerMultiFrame-1));
    // Set the lanes in use
    JesdDac().Poke(0x28, (1 << NumLanes)-1);
    // Select Subclass1
    JesdDac().Poke(0x2C, 0x1);

    // Set ILA Config Data 4
    Data.Value(0);
    Data.Field(0, 8, DacNumConverters-1);
    Data.Field(8, 5, 0xF);      // N = 15
    Data.Field(16, 5, 0xF);     // N' = 15
    Data.Field(24, 2, 0x0);     // CS = 0
    JesdDac().Poke(0x810, Data.Value());

    // Set ILA Config Data 5
    Data.Value(0);
    Data.Field(8, 5, SamplesPerFrame-1);
    Data.Bit(16, HighDensityMode);
    Data.Field(24, 5, 0x0);     // CF = 0
    JesdDac().Poke(0x814, Data.Value());

    // Reset the JESD204B IP
    JesdDac().Poke(0x4, 0x1);

    Debug.Log("end DAC3xJ82_DAC::ConfigureJesdParams()");
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::ConfigureDacDigBlocks()
//---------------------------------------------------------------------------

void DAC3xJ82_DAC::ConfigureDacDigBlocks()
{
    Debug.Log("DAC3xJ82_DAC::ConfigureDacDigBlocks()");

    MemoryBitManipulator Data(0);

    // Set config1 register
    Data.Value(Spi(0x1));
    Data.Bit(4, DacCh[1].Compliment);
    Data.Bit(5, DacCh[1].Compliment);
    Data.Bit(6, DacCh[0].Compliment);
    Data.Bit(7, DacCh[0].Compliment);
    Spi(0x1, Data.Value());

    // Set config2 register
    Data.Value(0);
    Data.Bit(1, 1);     // Set twos to enable 2s complement data format
    Data.Bit(4, DacEnFNco);
    Data.Bit(5, DacEnMixerGain);
    Data.Bit(6, DacEnMixer);
    Spi(0x2, Data.Value());

    // Set config3 register
    Data.Value(Spi(0x3));
    Data.Field(12, 4, DacCoarseGain);
    Spi(0x3, Data.Value());

    const int QmcOffsetMaxPos = (1<<12)-1; //4095
    const int QmcOffsetMaxNeg = -(1<<12);  //-4096
    // Check the QmcOffset values if within range, saturate if not within range
    DacCh[0].QmcOffset = std::min(QmcOffsetMaxPos,DacCh[0].QmcOffset);
    DacCh[0].QmcOffset = std::max(QmcOffsetMaxNeg,DacCh[0].QmcOffset);
    DacCh[1].QmcOffset = std::min(QmcOffsetMaxPos,DacCh[1].QmcOffset);
    DacCh[1].QmcOffset = std::max(QmcOffsetMaxNeg,DacCh[1].QmcOffset);

    // Set config8 register
    Data.Value(Spi(0x8));
    Data.Field(0, 13, DacCh[0].QmcOffset);
    Spi(0x8, Data.Value());

    // Set config9 register
    Data.Value(Spi(0x9));
    Data.Field(0, 13, DacCh[1].QmcOffset);
    Spi(0x9, Data.Value());

    const unsigned int QmcGainMax = (1<<11)-1; //2047
    // Check the QmcGain values if within range, saturate if not within range
    DacCh[0].QmcGain = std::min(QmcGainMax,DacCh[0].QmcGain);
    DacCh[1].QmcGain = std::min(QmcGainMax,DacCh[1].QmcGain);

    // Set config12 register
    Data.Value(Spi(0xC));
    Data.Field(0, 11, DacCh[0].QmcGain);
    Spi(0xC, Data.Value());

    // Set config13 register
    Data.Value(Spi(0xD));
    Data.Field(0, 11, DacCh[1].QmcGain);
    Data.Field(12, 4, DacCMixSelect);
    Spi(0xD, Data.Value());

    const int QmcPhaseMaxPos = (1<<11)-1; //2047
    const int QmcPhaseMaxNeg = -(1<<11);  //-2048
    // Check the DacQmcPhase value if within range, saturate if not within range
    DacQmcPhase = std::min(QmcPhaseMaxPos,DacQmcPhase);
    DacQmcPhase = std::max(QmcPhaseMaxNeg,DacQmcPhase);

    // Set config16 register
    Data.Value(Spi(0x10));
    Data.Field(0, 12, DacQmcPhase);
    Spi(0x10, Data.Value());

    const unsigned int DacFNcoPhaseOffsetMax = (1<<16)-1; //65535;
    // Check the DacFNcoPhaseOffset value if within range, saturate if not within range
    DacFNcoPhaseOffset = std::min(DacFNcoPhaseOffsetMax,DacFNcoPhaseOffset);

    // Set config18 register
    Spi(0x12, DacFNcoPhaseOffset);

    const s_ii64 DacFNcoPhaseIncMaxPos = ((static_cast<s_ii64>(1)) <<47)-1;
    const s_ii64 DacFNcoPhaseIncMaxNeg = -((static_cast<s_ii64>(1)) <<47);
    DacFNcoPhaseInc = std::min(DacFNcoPhaseIncMaxPos,DacFNcoPhaseInc);
    DacFNcoPhaseInc = std::max(DacFNcoPhaseIncMaxNeg,DacFNcoPhaseInc);

    // Set config20,21,22 registers
    Spi(0x14, (DacFNcoPhaseInc & 0xFFFF));
    Spi(0x15, ((DacFNcoPhaseInc>>16) & 0xFFFF));
    Spi(0x16, ((DacFNcoPhaseInc>>32) & 0xFFFF));

    // Set config34 register
    // (output assignment specific to FMC1000)
    // Outputs A&C are AC coupled, B&D are DC coupled
    Data.Value(0);
    Data.Field(0,  2, ((DacCh[1].CouplingMode == ocmDc)? 1:3));  // Set pathd_out_sel
    Data.Field(2,  2, ((DacCh[1].CouplingMode == ocmAc)? 1:3));  // Set pathc_out_sel
    Data.Field(4,  2, ((DacCh[0].CouplingMode == ocmDc)? 0:2));  // Set pathb_out_sel
    Data.Field(6,  2, ((DacCh[0].CouplingMode == ocmAc)? 0:2));  // Set patha_out_sel
    Data.Field(8,  2, 3);                     // Set pathd_in_sel
    Data.Field(10, 2, 2);                     // Set pathc_in_sel
    Data.Field(12, 2, 1);                     // Set pathb_in_sel
    Data.Field(14, 2, 0);                     // Set patha_in_sel
    Spi(0x22, Data.Value());

    // Set config38 register
    Data.Value(Spi(0x26));
    Data.Bit(0, DacEnZeroDither);
    Data.Field(4, 4, DacDitherAmount);
    Data.Bit(8, DacCh[0].EnDitherMixer);
    Data.Bit(9, DacCh[1].EnDitherMixer);
    Data.Bit(12, DacCh[0].EnDither);
    Data.Bit(13, DacCh[1].EnDither);
    Spi(0x26, Data.Value());

    // Set config30 register
    Spi(0x1E, 0x9999);

    // Set config31 register
    Spi(0x1F, 0x9940);

    // Set config32 register
    Spi(0x20, 0x8000);

    // Toggle the sif_sync
    Spi(0x1F, 0x9942);
    Spi(0x1F, 0x9940);

    // Check the SerDes PLL lock
    Spi(0x6C, 0);             // Write 0 to clear previous alarms
    Data.Value(Spi(0x6C));    // Read the new alarms
    bool Pll0Lock = (Data.Value() & 0x8) ? false : true;
    bool Pll1Lock = (Data.Value() & 0x4) ? false : true;
    Debug.Log("DAC SerDes PLL0 locked: ", Pll0Lock);
    Debug.Log("DAC SerDes PLL1 locked: ", Pll1Lock);

    // Start Jesd link initialization
    // Set config74 register
    Data.Value(Spi(0x4A));
    Data.Field(1, 4, 0x0);
    Spi(0x4A, Data.Value());
    Data.Bit(0, 1);
    Spi(0x4A, Data.Value());
    Sleep(10);

    // Clear alarms
    for (unsigned int i = 0x64; i<=0x6D; i++)
        Spi(i, 0x0);

    // Check alarms
    Debug.Log("DAC lane0 alarm error and fifo flag =", Spi(0x64), 16);
    Debug.Log("DAC lane1 alarm error and fifo flag =", Spi(0x65), 16);
    Debug.Log("DAC lane2 alarm error and fifo flag =", Spi(0x66), 16);
    Debug.Log("DAC lane3 alarm error and fifo flag =", Spi(0x67), 16);
    Debug.Log("DAC lane4 alarm error and fifo flag =", Spi(0x68), 16);
    Debug.Log("DAC lane5 alarm error and fifo flag =", Spi(0x69), 16);
    Debug.Log("DAC lane6 alarm error and fifo flag =", Spi(0x6A), 16);
    Debug.Log("DAC lane7 alarm error and fifo flag =", Spi(0x6B), 16);
    Debug.Log("DAC alarm =", Spi(0x6C), 16);
    Debug.Log("DAC short test and LOS alarm =", Spi(0x6D), 16);

    JesdDac().DumpDebugStatus();

    Debug.Log("end DAC3xJ82_DAC::ConfigureDacDigBlocks()");
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::CalcDacJesdLineRate() -- Calculate the DAC JESD line rate
//---------------------------------------------------------------------------

double DAC3xJ82_DAC::CalcDacJesdLineRate(double SamplingRate)
{
    InterpolationRate = (DacInterpolationRate == 0)? 1 : 2 * DacInterpolationRate;
    const unsigned int MaxNumLanes = 8;

    // The Dac can run at up to 12.5Gbps but the max rate is set to the max CPLL rate
    const double MaxLineRate = 6600.0e6;    // from the FPGA datasheet
    const double MinLineRate = 781.25e6;    // from  the Dac datasheet

    // Select the minimum number of lanes that results in a line rate within range
    // The Dac supports NumLanes values of 8, 4, 2, 1
    for (unsigned int i=1; i<=MaxNumLanes; i*=2)
        {
        LineRate = JesdDac().CalcJesdLineRate(DacNumConverters, 2, SamplingRate, InterpolationRate, i);
        if ((LineRate >= MinLineRate && LineRate <= MaxLineRate) || (i==MaxNumLanes))
            {
            if (i==1 && InterpolationRate == 1)  // skip if single lane in x1 interpolation mode
                continue;
            NumLanes = i;
            break;
            }
        }
    return LineRate;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::CalcDacClkFrequency() -- Calculate the DAC clock frequency
//---------------------------------------------------------------------------

double DAC3xJ82_DAC::CalcDacClkFrequency(double SamplingRate)
{
    double MaxfDac = 1230.0e6;
    double DacClkFrequency = (SamplingRate/InterpolationRate);

    while (DacClkFrequency > MaxfDac)
        DacClkFrequency /=2;

    return DacClkFrequency;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacInterpolationRate() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacInterpolationRate(const unsigned int value)
{
    DacInterpolationRate = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacEnInvSinc() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacEnInvSinc(const bool value)
{
    DacEnInvSinc = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacEnFNco() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacEnFNco(const bool value)
{
    DacEnFNco = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacEnMixerGain() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacEnMixerGain(const bool value)
{
    DacEnMixerGain = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacEnMixer() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacEnMixer(const bool value)
{
    DacEnMixer = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacCoarseGain() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacCoarseGain(const unsigned int value)
{
    DacCoarseGain = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacCMixSelect() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacCMixSelect(const unsigned int value)
{
    DacCMixSelect = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacEnQmc() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacEnQmc(const bool value)
{
    DacEnQmc = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacQmcPhase() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacQmcPhase(const int value)
{
    DacQmcPhase = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacFNcoPhaseOffset() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacFNcoPhaseOffset(const unsigned int value)
{
    DacFNcoPhaseOffset = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacFNcoPhaseInc() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacFNcoPhaseInc(const s_ii64 value)
{
    DacFNcoPhaseInc = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacEnZeroDither() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacEnZeroDither(const bool value)
{
    DacEnZeroDither = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacDitherAmount() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacDitherAmount(const unsigned int value)
{
    DacDitherAmount = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacOutputCouplingMode() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacOutputCouplingMode(const unsigned int value,const unsigned int DacChIndex)
{
    DacCh[DacChIndex].CouplingMode = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacCompliment() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacCompliment(const bool value,const unsigned int DacChIndex)
{
    DacCh[DacChIndex].Compliment = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacQmcOffset() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacQmcOffset(const int value,const unsigned int DacChIndex)
{
    DacCh[DacChIndex].QmcOffset = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacQmcGain() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacQmcGain(const unsigned int value,const unsigned int DacChIndex)
{
    DacCh[DacChIndex].QmcGain = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacEnDither() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacEnDither(const bool value,const unsigned int DacChIndex)
{
    DacCh[DacChIndex].EnDither = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::SetDacEnDitherMixer() --
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::SetDacEnDitherMixer(const bool value,const unsigned int DacChIndex)
{
    DacCh[DacChIndex].EnDitherMixer = value;
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::Spi() -- write to a SPI device
//---------------------------------------------------------------------------

void  DAC3xJ82_DAC::Spi(unsigned int address, unsigned int data)
{
    MDC->DacSpi(DeviceID, address, data);
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::Spi() -- read from a SPI device
//---------------------------------------------------------------------------

unsigned int  DAC3xJ82_DAC::Spi(unsigned int address)
{
    return MDC->DacSpi(DeviceID, address);
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::ChannelEnabled() -- read cached channel enable state
//---------------------------------------------------------------------------

bool  DAC3xJ82_DAC::ChannelEnabled(int ch)
{
    return MDC->DacChannelEnabled(ch);
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::ChannelEnabled() -- set channel enable state
//---------------------------------------------------------------------------

void DAC3xJ82_DAC::ChannelEnabled(int ch, bool state)
{
    return MDC->DacChannelEnabled(ch, state);
}

//---------------------------------------------------------------------------
//  DAC3xJ82_DAC::log2()  -- Calculate Log2
//---------------------------------------------------------------------------

unsigned int DAC3xJ82_DAC::log2(unsigned int value)
{
    unsigned int log = 0;
    while (value != 0)
        {
        value >>= 1;
        log++;
        }
    return log-1;
}

}  // namespace Innovative



