// DataLogger_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <DataLogger_Mb.h>
#include <MemoryUtils_Mb.h>

using namespace std;

namespace   Innovative
{

//=============================================================================
//  CLASS DataLogger  --
//=============================================================================
//---------------------------------------------------------------------------
// DataLogger::DataLogger() --
//---------------------------------------------------------------------------

DataLogger::DataLogger()
	 : FLogged(0), FCeiling(0), FEnabled(true), FSoftCeiling(false),
       FCoalesceWrites(true), Residual(0)
{
	Cache.Resize(0x1000000);
}

//---------------------------------------------------------------------------
// DataLogger::DataLogger() --
//---------------------------------------------------------------------------

DataLogger::DataLogger(const DataLogger & other)
	 : FLogged(other.FLogged), FCeiling(other.FCeiling),
	   FEnabled(other.FEnabled), FSoftCeiling(false),
	   FCoalesceWrites(other.FCoalesceWrites),
       Residual(other.Residual), Cache(other.Cache), mFile(other.mFile)
{
	OnProgress = other.OnProgress;
}

//---------------------------------------------------------------------------
// DataLogger::~DataLogger() --
//---------------------------------------------------------------------------

DataLogger::~DataLogger()
{
    Stop();
}

//---------------------------------------------------------------------------
//  DataLogger::operator = --
//---------------------------------------------------------------------------

DataLogger & DataLogger::operator = (const DataLogger & other)
{
	// protect against invalid self-assignment
	if (this != &other)
		{
		FCeiling = other.FCeiling;
		FEnabled = other.FEnabled;
		FLogged = other.FLogged;
		FSoftCeiling = other.FSoftCeiling;
		FCoalesceWrites = other.FCoalesceWrites;
		Residual = other.Residual;

		mFile = other.mFile;

		Cache = other.Cache;

		OnProgress = other.OnProgress;
        }

	return *this;
}

//---------------------------------------------------------------------------
//  DataLogger::Log() --
//---------------------------------------------------------------------------

bool DataLogger::Log(const Buffer & buffer)
{
	return Log(buffer.ConstData(), static_cast<unsigned int>(buffer.SizeInInts()));
}

//---------------------------------------------------------------------------
//  DataLogger::LogWithHeader() --
//---------------------------------------------------------------------------

bool DataLogger::LogWithHeader(const Buffer & buffer)
{
	if (FCoalesceWrites)
		return LogCoalesced(buffer);
	else
		return LogUncoalesced(buffer);
}

//---------------------------------------------------------------------------
//  DataLogger::LogCoalesced() --
//---------------------------------------------------------------------------

bool DataLogger::LogCoalesced(const Buffer & buffer)
{
	size_t size = buffer.HeaderSizeInInts() + buffer.SizeInInts()
		+ buffer.TrailerSizeInInts() + Residual;
	if (Cache.SizeInInts() < size)
		Cache.Resize(size);

	int * dst = Cache.Data()+Residual;

	CopyInt32Buffer(buffer.ConstHeader(), dst, buffer.HeaderSizeInInts());
	dst += buffer.HeaderSizeInInts();

	CopyInt32Buffer(buffer.ConstData(), dst, buffer.SizeInInts());
	dst += buffer.SizeInInts();

	CopyInt32Buffer(buffer.ConstTrailer(), dst, buffer.TrailerSizeInInts());
	dst += buffer.TrailerSizeInInts();

	// If logging headers, soft ceiling should always be enabled
	FSoftCeiling=true;

	//  Log aggregated buffer (header, body, trailer)
	static const size_t segment = 0x10000;
	Residual = size % segment;
	size_t chunk = size - Residual;

	bool result = true;
	if (chunk)
		result = Log(Cache.ConstData(), static_cast<unsigned int>(chunk));

	CopyInt32Buffer(Cache.ConstData()+chunk, Cache.Data(), Residual);

	return result;
}

//---------------------------------------------------------------------------
//  DataLogger::LogUncoalesced() --
//---------------------------------------------------------------------------

bool DataLogger::LogUncoalesced(const Buffer & buffer)
{
	bool result = true;

	// If logging headers, soft ceiling should always be enabled
	FSoftCeiling=true;

	//  Log Header, then Data
	result &= Log(buffer.ConstHeader(), static_cast<unsigned int>(buffer.HeaderSizeInInts()));
	result &= Log(buffer);

	//  If Trailer is present, log it too
	unsigned int TrailerSize = static_cast<unsigned int>(buffer.TrailerSizeInInts());
	if (TrailerSize)
		result &= Log(buffer.ConstTrailer(), TrailerSize);

	return result;
}

//---------------------------------------------------------------------------
//  DataLogger::Log() --
//---------------------------------------------------------------------------

bool DataLogger::Log(const int * data, unsigned int size_in_ints)
{
    unsigned int bytes = size_in_ints * sizeof(int);
	return DoLogging(data, bytes);
}

//---------------------------------------------------------------------------
//  DataLogger::DoLogging() --
//---------------------------------------------------------------------------

bool DataLogger::DoLogging(const void * data, unsigned int bytes)
{
	if (!mFile.Opened() || !FEnabled || !bytes)
		return false;

	if (FCeiling && !FSoftCeiling && ((FLogged + bytes) > FCeiling))
		bytes = static_cast<int>(FCeiling - FLogged);

	unsigned int written = mFile.Write(reinterpret_cast<const char*>(data), bytes);

	if (written != bytes)
		return false;

	FLogged += bytes;

	if (FCeiling && (FLogged >= FCeiling))
		{
		ProgressEvent e(100);
		OnProgress.Execute(e);

		return Stop(); // limit exeeded
		}

	if (FCeiling)
		{
		double percent = FLogged * 100.0 / FCeiling;
		int Percent = static_cast<int>(percent);

		ProgressEvent e(Percent);
		OnProgress.Execute(e);
		}

	return true;
}

//---------------------------------------------------------------------------
//  DataLogger::Start() --
//---------------------------------------------------------------------------

bool DataLogger::Start()
{
	if (mFile.Opened() || !mFile.FileName().length())
		return false;

	StartBufferEvent sb_event(mFile.FileName().length());
	OnStart.Execute(sb_event);

	if (!mFile.Create())
		return false;

	// Pre-allocate disk space
	if (FCeiling)
		mFile.Size(FCeiling);

	mFile.Position(0);
	FLogged = 0;
	Residual = 0;

	ProgressEvent evt(0);
	OnProgress.Execute(evt);

	return true;
}

//---------------------------------------------------------------------------
//  DataLogger::Stop() --
//---------------------------------------------------------------------------

bool DataLogger::Stop()
{
	if (!mFile.Opened())
		return false;

	if (Residual)
		{
		// Avoid reentrancy
		size_t residual = Residual;
		Residual = 0;

		Log(Cache.ConstData(), static_cast<unsigned int>(residual));
		}

	mFile.Close();

	OpenWire::NotifyEvent event;
	OnStop.Execute(event);

    return true;
}

//---------------------------------------------------------------------------
//  DataLogger::Logged() --
//---------------------------------------------------------------------------

ii64 DataLogger::Logged() const
{
	return FLogged;
}

//---------------------------------------------------------------------------
//  DataLogger::Ceiling() --
//---------------------------------------------------------------------------

void DataLogger::Ceiling(const ii64 & value)
{
	FCeiling = value;
}

//---------------------------------------------------------------------------
//  DataLogger::Ceiling() --
//---------------------------------------------------------------------------

ii64 DataLogger::Ceiling() const
{
	return FCeiling;
}

//---------------------------------------------------------------------------
//  DataLogger::FileName() --
//---------------------------------------------------------------------------

void DataLogger::FileName(const string & value)
{
	mFile.FileName(value);
}

//---------------------------------------------------------------------------
//  DataLogger::FileName() --
//---------------------------------------------------------------------------

string DataLogger::FileName() const
{
	return mFile.FileName();
}

//---------------------------------------------------------------------------
//  DataLogger::Enabled() --
//---------------------------------------------------------------------------

void DataLogger::Enabled(bool state)
{
	FEnabled = state;
}

//---------------------------------------------------------------------------
//  DataLogger::Enabled() --
//---------------------------------------------------------------------------

bool DataLogger::Enabled() const
{
	return FEnabled;
}

//---------------------------------------------------------------------------
//  DataLogger::SoftCeiling() --
//---------------------------------------------------------------------------

void DataLogger::SoftCeiling(bool state)
{
	FSoftCeiling = state;
}

//---------------------------------------------------------------------------
//  DataLogger::SoftCeiling() --
//---------------------------------------------------------------------------

bool DataLogger::SoftCeiling() const
{
	return FSoftCeiling;
}

//---------------------------------------------------------------------------
//  DataLogger::WriteThrough() --
//---------------------------------------------------------------------------

void DataLogger::WriteThrough(bool state)
{
	mFile.WriteThrough(state);
}

//---------------------------------------------------------------------------
//  DataLogger::WriteThrough() --
//---------------------------------------------------------------------------

bool DataLogger::WriteThrough() const
{
	return mFile.WriteThrough();
}

//---------------------------------------------------------------------------
//  DataLogger::CoalesceWrites() --
//---------------------------------------------------------------------------

void DataLogger::CoalesceWrites(bool state)
{
	FCoalesceWrites = state;
}

//---------------------------------------------------------------------------
//  DataLogger::CoalesceWrites() --
//---------------------------------------------------------------------------

bool DataLogger::CoalesceWrites() const
{
	return FCoalesceWrites;
}

}   // Innovative

