// DataSetBase.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <DataSetBase_Mb.h>
#include <cfloat>
#include <IniFile_Mb.h>
#include <cstring>
#include <algorithm>

namespace Innovative
{

//===========================================================================
// Class DatasetIncident - class for storing Graph Annotations and Alerts
//===========================================================================

//---------------------------------------------------------------------------
// DatasetIncident::operator ==() --
//---------------------------------------------------------------------------

bool DatasetIncident::operator ==(const DatasetIncident& rhs) const
{
	bool result = (Origin == rhs.Origin) && (Type == rhs.Type) && (Id == rhs.Id);

	return result;
}

//---------------------------------------------------------------------------
// DatasetIncident::operator !=() --
//---------------------------------------------------------------------------

bool DatasetIncident::operator !=(const DatasetIncident& rhs) const
{
	return !(*this == rhs);
}

//---------------------------------------------------------------------------
// DatasetIncident::operator <() --
//---------------------------------------------------------------------------

bool DatasetIncident::operator <(const DatasetIncident& rhs) const
{

	bool coincident = (Origin == rhs.Origin);
	if (coincident)
		{
		if (Type < rhs.Type)
			return true;

		return (Index < rhs.Index);
		}

	return (Origin < rhs.Origin);
}

//---------------------------------------------------------------------------
// DatasetIncident::Format() --  Return packet format
//---------------------------------------------------------------------------

std::string DatasetIncident::Format() const
{
	string result[] = { "Alert", "Velocia", "Velo", "Vita" };
	return result[Type];
}

//---------------------------------------------------------------------------
// DatasetIncident::Description() --  Return incident description
//---------------------------------------------------------------------------

std::string DatasetIncident::Description() const
{
	stringstream msg;
	msg << Format() << " " << Index;

	return msg.str();
}

//---------------------------------------------------------------------------
// DatasetIncident::Label() --  Return incident label
//---------------------------------------------------------------------------

std::string DatasetIncident::Label() const
{
	stringstream msg;
	msg << Format() << "\n" << Index;

	return msg.str();
}

//===========================================================================
// Class DataSet  --  Base class for sampled data sets
//===========================================================================

//---------------------------------------------------------------------------
// DataSet Constructor
//---------------------------------------------------------------------------

DataSet::DataSet()
	: FEnabled(false), FScalingEnabled(true), 
      FHeader(0), FTrailer(0), FEventHeader(0),
	  FEventTrailer(0), FChannels(1), FDevices(1), FChannel(0), 
      FType(Int), FBits(16), FOffset(0), FScaleFactor(1), FPolarity(Signed),	  
	  FBigEndian(false), FNullDataValue(0.0), 
	  FReadOnly(true), FSampleRate(100000), FFileName(""), FComplex(false),
	  FAbort(false)
{
}

//---------------------------------------------------------------------------
// DataSet Destructor
//---------------------------------------------------------------------------

DataSet::~DataSet()
{
}

//---------------------------------------------------------------------------
// DataSet::Init() -- Recalculate internal working variables
//---------------------------------------------------------------------------

void DataSet::Init()
{
	switch (FType)
		{
		case 0:     // Int
			FSubsampleSize = (FBits<17) ? (FBits + 7)/8 : 4;
			break;
		case 1:     // IEEE Float
		case 2:     // TI Float
			FSubsampleSize = 4;
			break;
		case 3:     // ALaw
		case 4:     // MuLaw
		default:
			FSubsampleSize = 1;
			break;
		}

    FSampleSize = FSubsampleSize * Subsamples();

	Shift  = 32 - FBits;
	FMask = 0xffffffff >> Shift;

	FEventSize = FEventHeader + (FSampleSize*FChannels) + FEventTrailer;

//    FAbort = false;
    
	//
	//  Perform any Derived Class Init Now...
	ImageInit();

    OpenWire::NotifyEvent event;
	Dependent.Execute(event);
}

//---------------------------------------------------------------------------
// DataSet::Samples() -- Return number of samples in data set
//---------------------------------------------------------------------------

ii64 DataSet::Samples() const
{
	return (Size()-FHeader-FTrailer)/FEventSize;
}

//---------------------------------------------------------------------------
// DataSet::ScaleFactor() -- Specify I/O scalefactor
//---------------------------------------------------------------------------

void  DataSet::ScaleFactor(float sf)
{
	FScaleFactor = sf;
	Init();
}

//---------------------------------------------------------------------------
// DataSet::Offset() -- Specify offset of data set
//---------------------------------------------------------------------------

void  DataSet::Offset(float offset)
{
	FOffset = offset;
	Init();
}

//---------------------------------------------------------------------------
// DataSet::Type() -- Specify type of data in data set
//---------------------------------------------------------------------------

void  DataSet::Type(IIDataType type)
{
	FType = type;
	Init();
}

//---------------------------------------------------------------------------
// DataSet::Polarity() -- Specify polarity of data in data set
//---------------------------------------------------------------------------

void  DataSet::Polarity(IIPolarity polarity)
{
	FPolarity = polarity;
	Init();
}

//---------------------------------------------------------------------------
// DataSet::Complex() -- Specify complexity of of data in data set
//---------------------------------------------------------------------------

void DataSet::Complex(bool state)
{
	FComplex = state;
	Init();
}

//---------------------------------------------------------------------------
// DataSet::Size() -- Specify memory image size
//---------------------------------------------------------------------------

void  DataSet::Size(ii64 size)
{
	ImageSize(size);
	Init();
}

//---------------------------------------------------------------------------
// DataSet::Size() -- Return memory image size
//---------------------------------------------------------------------------

ii64 DataSet::Size() const
{
	return ImageSize();
}

//---------------------------------------------------------------------------
// DataSet::Enabled() -- Enabled/Disable data set access
//---------------------------------------------------------------------------

void  DataSet::Enabled(bool state)
{
    if (state == FEnabled)
        return;

    FEnabled = state;
    if (state)
        FEnabled = ImageOpen();
    else
        ImageClose();
}

//---------------------------------------------------------------------------
// DataSet::Enabled() -- Return enabled status
//---------------------------------------------------------------------------

bool  DataSet::Enabled() const
{
	return FEnabled;
}

//---------------------------------------------------------------------------
// DataSet::Bits() -- Set number of significant bits/sample
//---------------------------------------------------------------------------

void  DataSet::Bits(int bits)
{
	FBits = std::max(0, std::min(32, bits));
	Init();
}

//---------------------------------------------------------------------------
// DataSet::Header() -- Specify image header size, in bytes
//---------------------------------------------------------------------------

void  DataSet::Header(int bytes)
{
	if (bytes < 0)
		return;

	FHeader = bytes;
	Init();
}

//---------------------------------------------------------------------------
// DataSet::Trailer() -- Specify image trailer size, in bytes
//---------------------------------------------------------------------------

void  DataSet::Trailer(int bytes)
{
	if (bytes < 0)
		return;

	FTrailer = bytes;
	Init();
}

//---------------------------------------------------------------------------
// DataSet::EventHeader() -- Specify event header size, in bytes
//---------------------------------------------------------------------------

void  DataSet::EventHeader(int bytes)
{
	if (bytes < 0)
		return;

	FEventHeader = bytes;
	Init();
}

//---------------------------------------------------------------------------
// DataSet::EventTrailer() -- Specify event trailer size, in bytes
//---------------------------------------------------------------------------

void  DataSet::EventTrailer(int bytes)
{
	if (bytes < 0)
		return;

	FEventTrailer = bytes;
	Init();
}

//---------------------------------------------------------------------------
// DataSet::Channels() -- Set number of channels in each event
//---------------------------------------------------------------------------

void  DataSet::Channels(int channels)
{
	if (channels > 0)
		FChannels = channels;

	if (FChannel >= FChannels)
		FChannel = FChannels-1;

	FChannels = std::max(FChannels, 1);

	Init();
}

//---------------------------------------------------------------------------
// DataSet::Channel() -- Set active data channel
//---------------------------------------------------------------------------

void  DataSet::Channel(int channel)
{
	if ( (channel < 0) || (channel >= FChannels) )
		return;

	FChannel = channel;
    Init();
}

//---------------------------------------------------------------------------
// DataSet::Channels() -- Set number of devices in each event
//---------------------------------------------------------------------------

void  DataSet::Devices(int devices)
{
	if (devices > 0)
		FDevices = devices;

	Init();
}

//---------------------------------------------------------------------------
// DataSet::CurrentSample() -- Extract next sample from an event
//---------------------------------------------------------------------------

Sample DataSet::CurrentSample()
{
	Sample result;
	if (!FEnabled)
		return result;

    PreAccess();

    for (unsigned int i = 0; i < Subsamples(); ++i)
        {
        ImageSampleFetch(result[i]);

        if (result.Type() == Sample::sUnknown)
            continue;

        switch (FType)
            {
            case ALaw:
			case MuLaw:
            case Int:
                // Make all bits significant
                if (FPolarity)
                    result[i].Int((result[i].Int() << Shift) >> Shift);
                else
                    result[i].Int(result[i].Int() & FMask);
                break;
            case TiFloat:
            case Float:
                break;
            default:
                break;
            }

        bool unipolar = FPolarity == Unsigned;
        // Return as float if scaling enabled
        if (FScalingEnabled && ((FScaleFactor != 1.0) || (FOffset != 0.0)))
            {
            double value = result[i].Extended(unipolar)*FScaleFactor+FOffset;
            result[i].Type(Subsample::sFloat);
            result[i].Double(value);
            }
        }
    PostAccess();

	return result;
}

//---------------------------------------------------------------------------
// DataSet::CurrentSample() -- Insert float sample into an event
//---------------------------------------------------------------------------

void  DataSet::CurrentSample(const Sample & value)
{
	if (!FEnabled)
		return;

    Sample result(value);

    PreAccess();

    for (unsigned int i = 0; i < Subsamples(); ++i)
    	ImageSampleStore(result[i]);

   PostAccess();
}

//---------------------------------------------------------------------------
// DataSet::Element() -- Read specified data element from set
//---------------------------------------------------------------------------

Sample DataSet::Element(ii64 index)
{
	Seek(index);
	return CurrentSample();
}

//---------------------------------------------------------------------------
// DataSet::Element() -- Write specified data element into set
//---------------------------------------------------------------------------

void  DataSet::Element(ii64 index, const Sample & value)
{
	Seek(index);
	CurrentSample(value);
}

//---------------------------------------------------------------------------
// DataSet::SetFileName() -- Set file name
//---------------------------------------------------------------------------

void  DataSet::SetFileName(const std::string & name)
{
	FFileName=name;
	Init();
}

//---------------------------------------------------------------------------
// DataSet::Seek() -- Seek to specified sample in file
//---------------------------------------------------------------------------

void DataSet::Seek(ii64 value)
{
	if (!FEnabled)
		return;

	ImageSeek(FHeader + (value*FEventSize));
}

//---------------------------------------------------------------------------
// DataSet::Flush() -- Flush dirty cache contents to dataset
//---------------------------------------------------------------------------

void DataSet::Flush()
{
	if (!FEnabled)
		return;

	ImageFlush();
}

//---------------------------------------------------------------------------
// DataSet::CorrectEndian() -- Reverse endian-ness of sample
//---------------------------------------------------------------------------

void DataSet::CorrectEndian(char *input)
{
	if (!FBigEndian)
		return;

    char output[4];
    for (int i = 0; i < FSubsampleSize; i++)
        output[FSubsampleSize-1-i] = input[i];

    memcpy(input, output, FSubsampleSize);
}

//---------------------------------------------------------------------------
// DataSet::Subsamples() -- Return number of subsamples in sample
//---------------------------------------------------------------------------

unsigned int DataSet::Subsamples() const
{
    return FComplex ? 2 : 1;
}

//---------------------------------------------------------------------------
// DataSet::Subchannels() -- Return number of subchannels in set
//---------------------------------------------------------------------------

unsigned int DataSet::Subchannels() const
{
    return FChannels * Subsamples();
}

//---------------------------------------------------------------------------
// DataSet::Progress() --
//---------------------------------------------------------------------------

bool DataSet::Progress(ii64 sample_idx)
{
    const int percent = Samples() ? static_cast<int>((sample_idx * 100)/Samples()) : 0;

	static int previous = -1;
	if (percent != previous)
		{
		ParseProgressEvent event(percent);
		OnProgress.Execute(event);
		previous = percent;

		FAbort = event.Abort;
		return FAbort;
		}

	return false;
}

//---------------------------------------------------------------------------
// DataSet::SortIncidentList() --
//---------------------------------------------------------------------------

void DataSet::SortIncidentList()
{
	sort(Incidents.begin(), Incidents.end());
}

//---------------------------------------------------------------------------
// DataSet::ReadIniFile() -- Read set characteristics from INI file
//---------------------------------------------------------------------------

void DataSet::ReadIniFile(const std::string & value)
{
	IniFile Ini(value);

	// Disable during param read
	bool state = FEnabled;
	FEnabled = false;

	// Data Parameters
	Ini.Section = "Data";
	FBigEndian = Ini.Read("Big Endian", false);
	FBits = Ini.Read("Significant Bits", 16);
	FPolarity = static_cast<DataSet::IIPolarity>(Ini.Read("Data Polarity", 1));
	FType = (DataSet::IIDataType)Ini.Read("Data Type", 0);
	FHeader = Ini.Read("Header Bytes", 0);
	FTrailer = Ini.Read("Trailer Bytes", 0);
	FEventHeader = Ini.Read("Event Header", 0);
	FEventTrailer = Ini.Read("Event Trailer", 0);
	FChannels = std::max(Ini.Read("Number of Channels", 1), 1);
	FDevices = Ini.Read("Number of Devices", 1);
	FChannel = Ini.Read("Current Channel", 0);
	FOffset = Ini.Read("Offset", 0.0f);
	FScaleFactor = Ini.Read("Scale Factor", 1.0f);
	FSampleRate = Ini.Read("Sample Rate", 100000.0f);
	FFileName = Ini.Read("File Name", std::string(""));
	FComplex = Ini.Read("Complex", false);

	ImageReadIniFile(Ini);

	FEnabled = state;

	Init();
}

//---------------------------------------------------------------------------
// DataSet::WriteIniFile() -- Write set characteristics to INI file
//---------------------------------------------------------------------------

void DataSet::WriteIniFile(const std::string & file)
{
	IniFile Ini(file);

	// Data Parameters
	Ini.Section = "Data";
	Ini.Write("Big Endian", FBigEndian);
	Ini.Write("Significant Bits", FBits);
	Ini.Write("Data Polarity", FPolarity);
	Ini.Write("Data Type", FType);
	Ini.Write("Header Bytes", FHeader);
	Ini.Write("Trailer Bytes", FTrailer);
	Ini.Write("Event Header", FEventHeader);
	Ini.Write("Event Trailer", FEventTrailer);
	Ini.Write("Number of Channels", FChannels);
	Ini.Write("Number of Devices", FDevices);
	Ini.Write("Current Channel", FChannel);
	Ini.Write("Offset", FOffset);
	Ini.Write("Scale Factor", FScaleFactor);
	Ini.Write("Sample Rate", FSampleRate);
	Ini.Write("File Name", FFileName);
	Ini.Write("Complex", FComplex);

	ImageWriteIniFile(Ini);
	Ini.Update();
}

//===========================================================================
// Class EventStreamDataSet  --  Base class for sampled data sets
//===========================================================================
//---------------------------------------------------------------------------
//  Constructor for class EventStreamDataSet
//---------------------------------------------------------------------------

 EventStreamDataSet::EventStreamDataSet()
	: HeaderSkip(0), TrailerSkip(0)
{
}

//---------------------------------------------------------------------------
//  destructor for class EventStreamDataSet
//---------------------------------------------------------------------------

 EventStreamDataSet::~EventStreamDataSet()
{
}

//---------------------------------------------------------------------------
//  EventStreamDataSet::ImageInit() -- On Image Initialization
//---------------------------------------------------------------------------

void  EventStreamDataSet::ImageInit()
{
    HeaderSkip = FEventHeader + FChannel*FSampleSize;
    TrailerSkip = (((FChannels-1)-FChannel)*FSampleSize) + FEventTrailer;
}

//---------------------------------------------------------------------------
//  EventStreamDataSet::PreAccess() -- Init prior to sample access
//---------------------------------------------------------------------------

void EventStreamDataSet::PreAccess()
{
	// Skip event header & fallow leading channels
	Skip(HeaderSkip);
}

//---------------------------------------------------------------------------
//  EventStreamDataSet::PostAccess() -- Finalize after sample access
//---------------------------------------------------------------------------

void EventStreamDataSet::PostAccess()
{
	// Skip fallow trailing channels & event trailer
	Skip(TrailerSkip);
}

//---------------------------------------------------------------------------
//  EventStreamDataSet::ImageSampleFetch() -- Fetch Sample N on channel R
//---------------------------------------------------------------------------

void  EventStreamDataSet::ImageSampleFetch(Subsample & value)
{
	// Fetch the data sample
  	ImageFetch(value);
}

//---------------------------------------------------------------------------
//  EventStreamDataSet::ImageSampleStore() -- Fetch Sample N on channel R
//---------------------------------------------------------------------------

void  EventStreamDataSet::ImageSampleStore(Subsample & value)
{
    ImageStore(value);
}

//---------------------------------------------------------------------------
// EventStreamDataSet::Skip() -- Skip N bytes in set
//---------------------------------------------------------------------------

void EventStreamDataSet::Skip(int bytes)
{
	if (!FEnabled)
        return;

	ImageSkip(bytes);
}

} // namespace Innovative




