//   DriveIo_Mb.cpp
//   Fast large disk file access
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//---------------------------------------------------------------------------

//  If compiling for linux define macro _FILE_OFFSET_BITS to 64, this will 
//  enable Large File System(LFS). If this bit is undefined or defined as 32 LFS
//  will be turned off. This macro has no effect when compiling in 64 bit 
//  system. Run "info libc 'Feature Test Macros'" for more information.
//  _FILE_OFFSET_BITS defenition must be found at the top of the file.

#if defined(LINUX)
#undef _FILE_OFFSET_BITS
#define _FILE_OFFSET_BITS	64
#endif

#include <DriveIo_Mb.h>
#include <BufferDatagrams_Mb.h>
#include <Exception_Mb.h>
#include <stdio.h>

namespace Innovative
{

//=============================================================================
//  CLASS DriveIoImpl  --  Windows-specific file I/O support class
//=============================================================================

class DriveIoImpl
{
    friend class DriveIo;

public:
    DriveIoImpl();
    DriveIoImpl(const DriveIoImpl & other);
    ~DriveIoImpl();

    // Assignment
    DriveIoImpl & operator = (const DriveIoImpl & other);

    bool            Open();
    bool            Opened() const
        {  return Handle != NULL;  }
    bool            Append();
    bool            Create();
    bool            Close();

    unsigned int    Write(const char * buffer, unsigned int size);
    unsigned int    Read(char * buffer, unsigned int size);

    void            Size(ii64 size);
    ii64            Size() const;

    void            FileName(const std::string & name);
    std::string     FileName() const;

    bool            Position(ii64 value);
    ii64            Position() const;

protected:

    // Data
	FILE *			Handle;
    std::string     FFileName;
    ii64            FPosition;
    ii64            FSize;

    void            GetSize();

};

//---------------------------------------------------------------------------
//  DriveIo::DriveIo() --  Ctor
//---------------------------------------------------------------------------

DriveIoImpl::DriveIoImpl()
    : Handle(NULL),
      FFileName(std::string("")), FPosition(0), FSize(0)
{
}

//---------------------------------------------------------------------------
// DriveIoImpl::DriveIoImpl() --
//---------------------------------------------------------------------------

DriveIoImpl::DriveIoImpl(const DriveIoImpl & other)
     : Handle(NULL),
	   FFileName(other.FFileName),
       FPosition(other.FPosition),
       FSize(other.FSize)
{
}

//---------------------------------------------------------------------------
//  DriveIoImpl::~DriveIoImpl() --  Set file buffer mode state
//---------------------------------------------------------------------------

DriveIoImpl::~DriveIoImpl()
{
    Close();
}

//---------------------------------------------------------------------------
//  DriveIoImpl::operator = --  Assignment operator
//---------------------------------------------------------------------------

DriveIoImpl & DriveIoImpl::operator = (const DriveIoImpl & other)
{
    // protect against invalid self-assignment
    if (this != &other)
        {
        Handle = NULL;
        FFileName = other.FFileName;
        FPosition = other.FPosition;
        FSize = other.FSize;
        }

    // by convention, always return *this
	return *this;
}

//---------------------------------------------------------------------------
//  DriveIoImpl::GetSize() --  Get current file size
//---------------------------------------------------------------------------

void DriveIoImpl::GetSize()
{
	off_t offset = 0;
	fseek(Handle, offset, SEEK_END);
	clearerr(Handle);
	FSize = static_cast<ii64>(ftell(Handle));
}

//---------------------------------------------------------------------------
//  DriveIoImpl::Open() --  Open file for reading
//---------------------------------------------------------------------------

bool DriveIoImpl::Open()
{
    if (Opened())
        return false;

    Handle = fopen(FFileName.c_str(), "rb");

    if (!Opened())
        throw MalibuException("Unable to open file");

    // Cache file size
    GetSize();
    Position(0);

    return true;
}

//---------------------------------------------------------------------------
//  DriveIoImpl::Create() --  Create file for writing
//---------------------------------------------------------------------------

bool DriveIoImpl::Create()
{
    if (Opened())
        return false;

    remove(FileName().c_str());

    Handle = fopen(FFileName.c_str(), "w+b");

    if (!Opened())
        throw MalibuException("Unable to create file");

    Position(0);

    return true;
}

//---------------------------------------------------------------------------
//  DriveIoImpl::Append() --  Open file for append
//---------------------------------------------------------------------------

bool DriveIoImpl::Append()
{
    if (Opened())
        return false;

    Handle = fopen(FFileName.c_str(), "r+b");

    if (!Opened())
        throw MalibuException("Unable to append file");

    // Cache file size
    GetSize();
    Position(FSize);

    return true;
}

//---------------------------------------------------------------------------
//  DriveIoImpl::Close() --
//---------------------------------------------------------------------------

bool DriveIoImpl::Close()
{
    if (!Opened())
        return true;

	bool Result = fclose(Handle);
    Handle = NULL;

    FSize = 0;
    FPosition = 0;

    return Result;
}

//---------------------------------------------------------------------------
//  DriveIoImpl::Read() --
//---------------------------------------------------------------------------

unsigned int DriveIoImpl::Read(char * buffer, unsigned int size)
{
    if (!Opened())
        return 0;

    size_t result;
    result = fread(buffer, 1, static_cast<size_t>(size), Handle);
    FPosition += result;

    return result;
}

//---------------------------------------------------------------------------
//  DriveIoImpl::Write() --
//---------------------------------------------------------------------------

unsigned int DriveIoImpl::Write(const char * buffer, unsigned int size)
{
    if (!Opened())
        return 0;

    size_t result;
    result = fwrite(buffer, 1, size, Handle);
    FPosition += result;
    FSize = std::max<ii64>(FPosition, FSize);

    return result;
}

//---------------------------------------------------------------------------
//  DriveIoImpl::FileName() --  Set file name
//---------------------------------------------------------------------------

void DriveIoImpl::FileName(const std::string & name)
{
    FFileName = name;
}

//---------------------------------------------------------------------------
//  DriveIoImpl::FileName() --  Get file name
//---------------------------------------------------------------------------

std::string DriveIoImpl::FileName() const
{
    return FFileName;
}

//---------------------------------------------------------------------------
//  DriveIoImpl::Position() --
//---------------------------------------------------------------------------

bool DriveIoImpl::Position(ii64 offset)
{
    if (!Opened())
        return false;

	off_t position = static_cast<off_t>(offset);
	int result = fseek(Handle, position, SEEK_SET);

	FPosition = position;

    return (result == 0);
}

//---------------------------------------------------------------------------
//  DriveIoImpl::Position() --
//---------------------------------------------------------------------------

ii64 DriveIoImpl::Position() const
{
    return FPosition;
}

//---------------------------------------------------------------------------
//  DriveIoImpl::Size() --  Set file size
//---------------------------------------------------------------------------

void DriveIoImpl::Size(ii64 size)
{
    Position(size);
	Write("\0", 1);

    FSize = size;
}

//---------------------------------------------------------------------------
//  DriveIoImpl::Size() --  Get file size
//---------------------------------------------------------------------------

ii64 DriveIoImpl::Size() const
{
	return FSize;
}

//===========================================================================
//  CLASS  DriveIo   ---  Large disk file access support class
//===========================================================================

//---------------------------------------------------------------------------
//  DriveIo::DriveIo() --  Ctor
//---------------------------------------------------------------------------

DriveIo::DriveIo()
    : Impl(new DriveIoImpl)
{
}

//---------------------------------------------------------------------------
//  DriveIo::DriveIo() --  Ctor
//---------------------------------------------------------------------------

DriveIo::DriveIo(const std::string & filename)
    : Impl(new DriveIoImpl)
{
    FileName(filename);
}

//---------------------------------------------------------------------------
// DriveIo::DriveIo() --  Copy ctor
//---------------------------------------------------------------------------

DriveIo::DriveIo(const DriveIo & other)
    : Impl(new DriveIoImpl)
{
    *Impl = *other.Impl;
}

//---------------------------------------------------------------------------
//  DriveIo::DriveIo() --  Dtor
//---------------------------------------------------------------------------

DriveIo::~DriveIo()
{
    delete Impl;
    Impl = 0;
}

//---------------------------------------------------------------------------
//  DriveIo::operator = --
//---------------------------------------------------------------------------

DriveIo & DriveIo::operator = (const DriveIo & other)
{
    // protect against invalid self-assignment
    if (this != &other)
        *Impl = *other.Impl;

    // by convention, always return *this
	return *this;
}

//---------------------------------------------------------------------------
//  DriveIo::Open() --  Open file for reading
//---------------------------------------------------------------------------

bool DriveIo::Open()
{
    return Impl->Open();
}

//---------------------------------------------------------------------------
//  DriveIo::Create() --  Create file for writing
//---------------------------------------------------------------------------

bool DriveIo::Create()
{
    return Impl->Create();
}

//---------------------------------------------------------------------------
//  DriveIo::Append() --  Open file for append
//---------------------------------------------------------------------------

bool DriveIo::Append()
{
    return Impl->Append();
}

//---------------------------------------------------------------------------
//  DriveIo::Close() --
//---------------------------------------------------------------------------

bool DriveIo::Close()
{
    return Impl->Close();
}

//---------------------------------------------------------------------------
//  DriveIo::Read() --
//---------------------------------------------------------------------------

unsigned int DriveIo::Read(char * buffer, unsigned int size)
{
    return Impl->Read(buffer, size);
}

//---------------------------------------------------------------------------
//  DriveIo::Write() --
//---------------------------------------------------------------------------

unsigned int DriveIo::Write(const char * buffer, unsigned int size)
{
    return Impl->Write(buffer, size);
}

//---------------------------------------------------------------------------
//  DriveIo::Opened() --
//---------------------------------------------------------------------------

bool DriveIo::Opened() const
{
    return Impl->Opened();
}

//---------------------------------------------------------------------------
//  DriveIo::Position() --
//---------------------------------------------------------------------------

bool DriveIo::Position(ii64 offset)
{
    return Impl->Position(offset);
}

//---------------------------------------------------------------------------
//  DriveIo::Position() --
//---------------------------------------------------------------------------

ii64 DriveIo::Position() const
{
    return Impl->Position();
}

//---------------------------------------------------------------------------
//  DriveIo::FileName() --  Set file name
//---------------------------------------------------------------------------

void DriveIo::FileName(const std::string & name)
{
    Impl->FileName(name);
}

//---------------------------------------------------------------------------
//  DriveIo::FileName() --  Get file name
//---------------------------------------------------------------------------

std::string DriveIo::FileName() const
{
    return Impl->FileName();
}

//---------------------------------------------------------------------------
//  DriveIo::Size() --  Set file size
//---------------------------------------------------------------------------

void DriveIo::Size(ii64 size)
{
    Impl->Size(size);
}

//---------------------------------------------------------------------------
//  DriveIo::Size() --  Get file size
//---------------------------------------------------------------------------

ii64 DriveIo::Size() const
{
    return Impl->Size();
}

//---------------------------------------------------------------------------
//  DriveIo::WriteThrough() --
//---------------------------------------------------------------------------

void DriveIo::WriteThrough(bool)
{
}

//---------------------------------------------------------------------------
//  DriveIo::WriteThrough() --
//---------------------------------------------------------------------------

bool DriveIo::WriteThrough() const
{
    return false;
}


}  // namespace Innovative

// Undefining _FILE_OFFSET_BITS macro, so that other classes are not affected.
#if defined(LINUX)
#undef _FILE_OFFSET_BITS
#endif
