
#include "DriverController_Linux.h"

#include <unistd.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <sstream>
#include <stdexcept>
#include <iostream>
#include <Event_Mb.h>
#include <Osal_Mb.h>
#include <Trace_Mb.h>
#include <iomanip>
#include <Poco/Format.h>
#include <Exception_Mb.h>



namespace   Innovative
{

using InnovativeKernel::Osal::OsDevice;
using Poco::format;

//------------------------------------------------------------------------------
//  DriverController::DriverController() --
//------------------------------------------------------------------------------

DriverController::DriverController(unsigned int id, OsDevice &owner)
   : _page_sz(sysconf(_SC_PAGESIZE)), _id(id), _target(-1), _fd(-1),
	 _owner(owner), _activity(this, &DriverController::runActivity), _event(0)
{
}

//------------------------------------------------------------------------------
//  DriverController::~DriverController() --
//------------------------------------------------------------------------------

DriverController::~DriverController()
{

}

//------------------------------------------------------------------------------
//  DriverController::open() --
//------------------------------------------------------------------------------

bool DriverController::open(unsigned int target)
{
	std::string txt = format("/dev/iipci_board_dev%u", target);

	_target = target;

	_fd = ::open(txt.c_str(), O_RDWR);

	if (_fd < 0) return false;

	return true;
}

//------------------------------------------------------------------------------
//  DriverController::close() --
//------------------------------------------------------------------------------

void DriverController::close()
{
	::close(_fd);
	_target = _fd = -1;
}

//------------------------------------------------------------------------------
//  DriverController::deviceCount() --
//------------------------------------------------------------------------------

int DriverController::deviceCount()
{
	std::string txt("/dev/iipci_board_dev0");

	int fd = ::open(txt.c_str(), O_RDWR);

	if (fd < 0) return 0;

	int count = 0;

	int ret = ioctl(fd, IOCTL_GET_MODULE_COUNT, &count);

	if  (ret < 0) throw MalibuException("deviceCount failed!");

	::close(fd);

	debug.Log(format("Board count : %d", count));

	return count;
}

//------------------------------------------------------------------------------
//  DriverController::deviceIds() --
//------------------------------------------------------------------------------

vector<device_id_t> DriverController::deviceIds()
{
	std::string txt("/dev/iipci_board_dev0");

	int fd = ::open(txt.c_str(), O_RDWR);

	if (fd < 0) return vector<device_id_t>();

	device_ids_t ids;

	int ret = ioctl(fd, IOCTL_GET_IDS, &ids);

	if  (ret < 0) throw MalibuException("deviceIds failed!");

	::close(fd);

	vector<device_id_t> vids;
	copy(&ids.ids[0], &ids.ids[ids.count], back_inserter(vids));

    debug.Log(format("Found %z devices.", vids.size()));

    for (size_t i = 0; i < vids.size(); i++)
        debug.Log(format("subsysid = 0x%08hX pci_devid = 0x%08hX", vids[i].pci_subsysid, vids[i].pci_devid));

	return vids;
}

//------------------------------------------------------------------------------
//  DriverController::barRegionSize() --
//------------------------------------------------------------------------------

int DriverController::barRegionSize(int bar)
{
	int size = bar;

	int ret = ioctl(_fd, IOCTL_GET_BAR_SIZE, &size);

	if  (ret < 0) throw MalibuException("barRegionSize failed!");

	debug.Log(format("BAR %d size %d bytes.", bar, size));

	return size;
}

//------------------------------------------------------------------------------
//  DriverController::mapBarRegion() --
//------------------------------------------------------------------------------

MemoryBlock DriverController::mapBarRegion(unsigned int bar)
{
	MemoryBlock mem;

	int size = barRegionSize(bar);

	if (size <= 0) throw MalibuException("BAR size cannot be retrieved!");

	char *addr = (char *) mmap(0, size, PROT_READ | PROT_WRITE, MAP_SHARED, _fd, bar * _page_sz);

	if (addr == MAP_FAILED) MalibuException("BAR map failed!");

	debug.Log(format("BAR %u mapped.", bar));

	mem.Addr      = (unsigned int *)addr;
	mem.Bar       = bar;
	mem.WasMapped = true;
	mem.Size      = size;

	return mem;
}

//------------------------------------------------------------------------------
//  DriverController::unmapBarRegion() --
//------------------------------------------------------------------------------

void DriverController::unmapBarRegion(const MemoryBlock &mem)
{
	munmap(mem.Addr, mem.Size);

	debug.Log(format("BAR %u unmapped.", mem.Bar));
}

//------------------------------------------------------------------------------
//  DriverController::mapDmaMemory() --
//------------------------------------------------------------------------------

MemoryBlock DriverController::mapDmaMemory(size_t size)
{
	MemoryBlock         mem;
	dma_memory_handle_t dh;

	mem.Size = dh.size = size;

	int ret = ioctl(_fd, IOCTL_ALLOC_DMA_MEMORY, &dh);

	if  (ret < 0) throw MalibuException("DMA memory allocation failed!");

	int index = DMA_MAPPING | dh.index;

	unsigned int *addr = (unsigned int *) mmap(0, size, PROT_READ | PROT_WRITE, MAP_SHARED, _fd, index * _page_sz);

	if (addr == MAP_FAILED) throw MalibuException("DMA memory map failed!");

	mem.Addr       = addr;
	mem.Bar        = dh.index;
	mem.PhysAddr   = reinterpret_cast<unsigned int *>(dh.paddr);
	mem.DriverAddr = dh.kaddr;

	debug.Log(format("DMA memory [%u] of %u bytes alloced.", mem.Bar, mem.Size));

	return mem;
}

//------------------------------------------------------------------------------
//  DriverController::unmapDmaMemory() --
//------------------------------------------------------------------------------

void DriverController::unmapDmaMemory(const MemoryBlock &mem)
{
	dma_memory_handle_t dh;

	munmap(mem.Addr, mem.Size);

	dh.kaddr = mem.DriverAddr;

	int ret = ioctl(_fd, IOCTL_FREE_DMA_MEMORY, &dh);

	if  (ret < 0) throw MalibuException("DMA memory could not be freed!");

	debug.Log(format("DMA memory [%u] freed.", mem.Bar));
}

//------------------------------------------------------------------------------
//  DriverController::clearInterruptQueue() --
//------------------------------------------------------------------------------

void DriverController::clearInterruptQueue()
{
   int ret = ioctl(_fd, IOCTL_INTERRUPT_CLEARQ);

   if  (ret < 0) throw MalibuException("clearInterruptQueue failed!");
}

//------------------------------------------------------------------------------
//  DriverController::getInterruptQueueDepth() --
//------------------------------------------------------------------------------

size_t DriverController::getInterruptQueueDepth()
{
   size_t size;
   int ret = ioctl(_fd, IOCTL_INTERRUPT_QDEPTH, &size);

   if  (ret < 0) throw MalibuException("getInterruptQueueDepth failed!");

   return size;
}

//------------------------------------------------------------------------------
//  DriverController::setInterruptInfo() --
//------------------------------------------------------------------------------

void DriverController::setInterruptInfo(unsigned int itype, unsigned int ioffset, unsigned int imask)
{
   InterruptInfo info;
   info.type = itype; info.offset = ioffset * sizeof(int); info.mask= imask;

   int ret = ioctl(_fd, IOCTL_SET_INTERRUPT_INFO, &info);

   if  (ret < 0) throw MalibuException("setInterruptInfo failed!");
}

//------------------------------------------------------------------------------
//  DriverController::enableInterrupt() --
//------------------------------------------------------------------------------

void DriverController::enableInterrupt()
{
	_activity.start();
}

//------------------------------------------------------------------------------
//  DriverController::disableInterrupt() --
//------------------------------------------------------------------------------

void DriverController::disableInterrupt()
{
	_activity.stop();
	 ioctl(_fd, IOCTL_INTERRUPT_KICK);
	_activity.wait();
}

//------------------------------------------------------------------------------
//  DriverController::isInterruptEnabled() --
//------------------------------------------------------------------------------

bool DriverController::isInterruptEnabled()
{
	return _activity.isRunning();
}

//------------------------------------------------------------------------------
//  DriverController::isInterruptEnabled() --
//------------------------------------------------------------------------------

void DriverController::installInterruptEventHandler(Event *isrEvent)
{
	_event = isrEvent;
}

//------------------------------------------------------------------------------
//  DriverController::runActivity() --
//------------------------------------------------------------------------------

void DriverController::runActivity()
{
	queue_snapshot_t q;
	while (!_activity.isStopped())
	{
		int ret = ioctl(_fd, IOCTL_WAIT_INTERRUPT);

		if (ret == -1) continue;

		ioctl(_fd, IOCTL_GET_INTERRUPT_BLOCK, &q);

		for (size_t i = 0; i < q.Depth; i++) {
			_owner.EnqueueIsrStatus(q.Queue[i]);
			_event->Set();
		}
	}
}

//------------------------------------------------------------------------------
//  DriverController::mapIdsToNames() --
//------------------------------------------------------------------------------

vector<string> DriverController::mapIdsToNames(const vector<device_id_t> &ids)
{
	vector<string> temp;
	for (size_t i = 0; i < ids.size(); i ++)
		temp.push_back(deviceName(ids[i]));

	return temp;
}

//------------------------------------------------------------------------------
//  DriverController::GetDeviceNames() --
//------------------------------------------------------------------------------

vector<string>  DriverController::getDeviceNames()
{
	if (deviceNames.empty())
		deviceNames = mapIdsToNames(deviceIds());

	return deviceNames;
}


//------------------------------------------------------------------------------
//  DriverController::deviceName() --
//------------------------------------------------------------------------------

string DriverController::deviceName(const Innovative::device_id_t& id)
{
	static struct {
		const char *name;
		unsigned int sid;
		unsigned int did;
	} device_map[] = {
        {"Cardsharp",                   0x0300, 0x0070},
        {"Cardsharp FMC-250",           0x0302, 0x0070},
        {"Cardsharp FMC-310" ,          0x0305, 0x0070},
        {"Cardsharp FMC-500" ,          0x0303, 0x0070},
        {"Cardsharp FMC-1000",          0x0304, 0x0070},
        {"Cardsharp FMC-9371",          0x030A, 0x0070},
        {"Cardsharp FMC-QSFP+",         0x0306, 0x0070},
        {"Cardsharp FMC-SDF",           0x0309, 0x0070},
        {"Cardsharp FMC-Servo",         0x0308, 0x0070},
        {"Cardsharp FMC-SFP+",          0x0307, 0x0070},
        {"Copious",                     0x0400, 0x0070},
        {"Copious FMC-250",             0x0402, 0x0070},
        {"Copious FMC-310",             0x0405, 0x0070},
        {"Copious FMC-500",             0x0403, 0x0070},
        {"Copious FMC-1000",            0x0404, 0x0070},
        {"Copious FMC-9371",            0x040A, 0x0070},
        {"Copious FMC-QSFP+",           0x0406, 0x0070},
        {"Copious FMC-SDF",             0x0409, 0x0070},
        {"Copious FMC-Servo",           0x0408, 0x0070},
        {"Copious FMC-SFP+",            0x0407, 0x0070},
        {"II Xilinx 7 Series Device",   0x0000, 0x0070},
        {"K7",                          0x0100, 0x0070},
        {"K7 FMC-250" ,                 0x0102, 0x0070},
        {"K7 FMC-310" ,                 0x0105, 0x0070},
        {"K7 FMC-500" ,                 0x0103, 0x0070},
        {"K7 FMC-1000",                 0x0104, 0x0070},
        {"K7 FMC-9371",                 0x010A, 0x0070},
        {"K7 FMC-QSFP+",                0x0106, 0x0070},
        {"K7 FMC-SDF",                  0x0109, 0x0070},
        {"K7 FMC-Servo",                0x0108, 0x0070},
        {"K7 FMC-SFP+",                 0x0107, 0x0070},
        {"Mini K7",                     0x0200, 0x0070},
        {"Mini K7 FMC-250",             0x0202, 0x0070},
        {"Mini K7 FMC-310",             0x0205, 0x0070},
        {"Mini K7 FMC-500",             0x0203, 0x0070},
        {"Mini K7 FMC-1000",            0x0204, 0x0070},
        {"Mini K7 FMC-9371",            0x020A, 0x0070},
        {"Mini K7 FMC-QSFP+",           0x0206, 0x0070},
        {"Mini K7 FMC-SDF",             0x0209, 0x0070},
        {"Mini K7 FMC-Servo",           0x0208, 0x0070},
        {"Mini K7 FMC-SFP+",            0x0207, 0x0070},
        {"PEX7-COP",                    0x0800, 0x0070},
        {"PEX7-COP FMC-250",            0x0802, 0x0070},
        {"PEX7-COP FMC-310",            0x0805, 0x0070},
        {"PEX7-COP FMC-500",            0x0803, 0x0070},
        {"PEX7-COP FMC-1000",           0x0804, 0x0070},
        {"PEX7-COP FMC-9371",           0x0804, 0x0070},
        {"PEX7-COP FMC-QSFP+",          0x0806, 0x0070},
        {"PEX7-COP FMC-SDF",            0x0809, 0x0070},
        {"PEX7-COP FMC-Servo",          0x0808, 0x0070},
        {"PEX7-COP FMC-SFP+",           0x0807, 0x0070},
        {"XA-160M",                     0x0501, 0x0070},
        {"XA-RX",                       0x0502, 0x0070},
        {"XA-TX",                       0x0503, 0x0070},
        {"XA-SD",                       0x0504, 0x0070},
        {"XA-DS",                       0x0505, 0x0070},
        {"XA-RT",                       0x0506, 0x0070},
        {"XA-DIO",                      0x0507, 0x0070},
        {"XA-A4D4",                     0x0508, 0x0070},
        {"XU-AQ",                       0x0604, 0x0070},
        {"XU-RT",                       0x0602, 0x0070},
        {"XU-RX",                       0x0603, 0x0070},
        {"XU-TX",                       0x0601, 0x0070},
        {"Innovative Spartan 6",        0x0,    0x0071},
        {"Sbc-ComEx Carrier",           0x0,    0x0040},
        {"Vpx-ComEx Carrier",           0x0,    0x0060},
        {"X3 XMCe",                     0x0,    0x0030},
        {"X5 XMCe",                     0x0,    0x0020},
        {"X6 XMCe",                     0x0,    0x0050},
        {"C64x DSP",                    0x0,    0x0008},
        {"ChicoPlus",                   0x0,    0x0004},
        {"Clock Generator PMC",         0x0,    0x0001},
        {"Digital Receiver PMC",        0x0,    0x0010},
        {"Fiber Channel SIO PMC",       0x0,    0x0001},
        {"M6713",                       0x0,    0x0000},
        {"Matador Family DSP",          0x0,    0x0006},
        {"P25M",                        0x0,    0x000C},
        {"Quadia",                      0x0,    0x0009},
        {"Quixote",                     0x0,    0x0007},
        {"TX PMC",                      0x0,    0x0011},
        {"UWB PMC",                     0x0,    0x001D}
	};

	string temp("Unknown");
	// Find the string for the matching sid-did pair
	for (size_t i = 0; i < sizeof(device_map)/sizeof(device_map[0]); i++)
		if (device_map[i].sid == id.pci_subsysid && device_map[i].did == id.pci_devid)
			temp = device_map[i].name;

	return temp;
}


} // namespace Malibu
