//
// Ds4026_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2009 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <Ds4026_Mb.h>
#include <sstream>
#include <HardwareRegister_Mb.h>

namespace Innovative
{

//=============================================================================
//  CLASS Ds4026 --  Programmable XO/VCXO Chip Support
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class Ds4026
//-----------------------------------------------------------------------------

const double TemperatureResolution = 0.0625;  // in degrees C

//-----------------------------------------------------------------------------
//  constructor for class Ds4026
//-----------------------------------------------------------------------------

Ds4026::Ds4026(I2CDevice & device)
    :  FTempCompEnabled(true),
       FTuningPPB(0), FLastTemperature(0.0), Device(device), 
       Debug(false)
{
}


//-----------------------------------------------------------------------------
//  Ds4026::Init()  --  Init setup
//-----------------------------------------------------------------------------

void  Ds4026::Init()
{
    SetTuningRegister();
    ReadTemperature();
}

//-----------------------------------------------------------------------------
//  Ds4026::Reapply()  --  Reapply settings to hardware from saved state
//-----------------------------------------------------------------------------

void  Ds4026::Reapply()
{
    SetTuningRegister();
}

//-----------------------------------------------------------------------------
//  Ds4026::Port()  --  Write to chip data port
//-----------------------------------------------------------------------------

void Ds4026::Port(unsigned char reg, unsigned char value)
{
	std::stringstream msg;
	msg << "Ds4026::Write 0x" << std::hex << static_cast<int>(value)
		<< " to reg " << static_cast<int>(reg);
	Debug.Log(msg.str());
    //
    //  Write using I2C interface
    Device.Reg(reg, value);
}

//-----------------------------------------------------------------------------
//  Ds4026::Port()  --  Read from chip data port
//-----------------------------------------------------------------------------

unsigned char Ds4026::Port(unsigned char reg)
{
    //
    //  Read using I2C interface
    int value = Device.Reg(reg);

	std::stringstream msg;
	msg << "Ds4026::Read 0x" << std::hex << value
		<< " from reg " << static_cast<int>(reg);
	Debug.Log(msg.str());

	return static_cast<unsigned char>(value);
}

//-----------------------------------------------------------------------------
//  Ds4026::SetTuningRegister()  --  Set Tuning Register Values
//-----------------------------------------------------------------------------

void  Ds4026::SetTuningRegister()
{
    MemoryBitManipulator Reg0(0);
    MemoryBitManipulator Reg1(0);

    Reg0.Bit(7, FTempCompEnabled==false);  // 1 = disable compensation

    Reg0.Field(0, 7, FTuningPPB>>8);     // Upper 6 bits
    Reg1.Field(0, 8, FTuningPPB & 0xFF);   // Lower 8 bits

    //  Write to hardware
    Port(0, Reg0.Value());
    Port(1, Reg1.Value());
}

//-----------------------------------------------------------------------------
//  Ds4026::ReadTemperature()  --  Read Temperature
//-----------------------------------------------------------------------------

void  Ds4026::ReadTemperature()
{
    int TempCache[4];    // only use first 2
    //
    //  Read Temp as single access to avoid mixed results when changing
    Device.MultiRegR(2, TempCache, 2);

    PtrBitManipulator Reg2(&TempCache[0]);
    PtrBitManipulator Reg3(&TempCache[1]);

    int temp(0);
    temp |= Reg2.Value();      // High part of temperature
    temp <<= 4;
    temp |= Reg3.Field(4, 4);  // Low part of temperature

    //  Convert to degrees C
    FLastTemperature = static_cast<float>(temp*TemperatureResolution);

}

} // namespace Innovative

