// FiclSystem_App.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2007 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include "FiclSystem_Mb.h"
#include <SystemSupport_Mb.h>
#include <StringSupport_Mb.h>
#include <cstring>                 // needed for memset in GCC

using namespace Innovative;
using namespace std;

namespace Innovative
{

//==============================================================================
//  CLASS  FiclSystem
//==============================================================================
//------------------------------------------------------------------------------
//  housekeeping for class FiclSystem --
//------------------------------------------------------------------------------

const int FiclTagValue = 0xdeadd00d;

//------------------------------------------------------------------------------
//  constructor for class FiclSystem --
//------------------------------------------------------------------------------

FiclSystem::FiclSystem()
    : vm(0), system(0), dictionary(0), isStarted(false), Target(0)
{
}

//------------------------------------------------------------------------------
//  destructor for class FiclSystem
//------------------------------------------------------------------------------

FiclSystem::~FiclSystem()
{
    if (vm)
        {
        ficlExtension * p = ToFiclExtensionPtr(vm);
        delete p;
        p = 0;
        }

    if (system)
        ficlTermSystem(system);
}

//-----------------------------------------------------------------------------
//  FiclSystem::ConnectTarget()  --  Connect target to engine
//-----------------------------------------------------------------------------

void  FiclSystem::ConnectTarget(IFiclTarget * target)
{
    if (target==0)
        return;
    Target = target;
    if (isStarted)
        Target->InitTarget(*this);
}

//-----------------------------------------------------------------------------
//  FiclSystem::Start()  --  Start command interpreter
//-----------------------------------------------------------------------------

int  FiclSystem::Start()
{
    CreateSys();

    CreateVm();

    dictionary = vmGetDict(vm);

    OnSystemStart();

    int status = Evaluate(".ver .( " __DATE__ " ) cr quit");

    //  If connected, load target functions
    if (Target)
        Target->InitTarget(*this);
    
    ProcessStatusEvent s(FICL_PROMPT);
    OnDisplayText.Execute(s);

    isStarted = true;
    return status;
}

//-----------------------------------------------------------------------------
//  FiclSystem::CreateSys()  -- Create FICL Sys
//-----------------------------------------------------------------------------

void  FiclSystem::CreateSys()
{
#if 0
    struct ficl_system_info
    {
        int size;           /* structure size tag for versioning */
        int nDictCells;     /* Size of system's Dictionary */
        OUTFUNC textOut;    /* default textOut function */
        void *pExtend;      /* Initializes VM's pExtend pointer - for application use */
        int nEnvCells;      /* Size of Environment dictionary */
    };
#endif

    ficl_system_info info;
    ficlInitInfo(&info);
    info.nDictCells = FICL_DEFAULT_DICT;
    info.nEnvCells = FICL_DEFAULT_ENV;
    info.textOut = FiclSpit;
    info.pExtend = new ficlExtension;

    system = ficlInitSystemEx(&info);

    ficlCompileFileLoad(system);  // add file loading to base system
}

//-----------------------------------------------------------------------------
//  FiclSystem::CreateVm()  -- Create FICL Vm
//-----------------------------------------------------------------------------

void  FiclSystem::CreateVm()
{
    vm = ficlNewVM(system);
    //
    //  Insert 'back pointer' to System into Vm

    ficlExtension * p = ToFiclExtensionPtr(vm);
    p->ficlSystemPtr = this;
    p->tag = FiclTagValue;
}

//-----------------------------------------------------------------------------
//  FiclSystem::OnSystemStart()  --  Add engine common functions
//-----------------------------------------------------------------------------
//  virtual to allow additions later

void  FiclSystem::OnSystemStart()
{
    Define("ms", Ms);

    //  Load default Stream Callback words
    Evaluate(": OnStreamPreconfig ;");
    Evaluate(": OnStreamSingleTimeAtStartConfig ;");
    Evaluate(": OnStreamConfigure ;");
    Evaluate(": OnBeforeStreamStart ;");
    Evaluate(": OnAfterStreamStart ;");
    Evaluate(": OnBeforeStreamStop ;");
    Evaluate(": OnAfterStreamStop ;");


    Evaluate(": >bit ( bit -- mask ) 1 swap lshift ;");
    Evaluate(": >mask ( n -- 2^n-1 ) >bit 1 - ;");
    Evaluate(": @bit ( data bit# -- data flag ) >bit over and 0= 0= ;");
    Evaluate(": @field ( data bit# width -- data field ) >mask >r over swap rshift r> and ;");
    Evaluate(": ?system "
                    "    cr .\" ---- System Definitions ----  \" "
                    "    cr .\"   @bit    ( data bit# -- data flag )        : extract bit flag from data  \" "
                    "    cr .\"   @field  ( data bit# size  -- data field ) : extract field from data     \" "
                    "    cr .\"   ?system  : display this message \" "
                    "    cr  ; ");

    Display("System wordset loaded \n" );
}

//-----------------------------------------------------------------------------
//  FiclSystem::Evaluate()  --  Execute a Forth command string (no prompt)
//-----------------------------------------------------------------------------

int FiclSystem::Evaluate(const std::string & cmd)
{
    return ficlEvaluate(vm, const_cast<char*>(cmd.c_str()));
}

//-----------------------------------------------------------------------------
//  FiclSystem::Execute()  --  Execute a Forth command string (with prompt)
//-----------------------------------------------------------------------------

int FiclSystem::Execute(const std::string & cmd)
{
    return ficlExec(vm, const_cast<char*>(cmd.c_str()));
}

//---------------------------------------------------------------------------
//  FiclSystem::Define() --  Add word to dictionary
//---------------------------------------------------------------------------

void  FiclSystem::Define(const std::string & word, FICL_CODE code,
    UNS8 flags)
{
    ficlBuild(system, const_cast<char*>(word.c_str()), code, flags);
}

//---------------------------------------------------------------------------
//  FiclSystem::Display() --  Display or cache string
//---------------------------------------------------------------------------

void  FiclSystem::Display(const std::string & txt)
{
    Text << txt;
    const string & s(Text.str());
    if (s.find("\n") != std::string::npos)
        {
        ProcessStatusEvent e(Trim(s));
        OnDisplayText.Execute(e);

        Text.str("");
        }
}

//---------------------------------------------------------------------------
//  FiclSystem::FiclSpit() --  Override display function for vm and sys
//---------------------------------------------------------------------------

void  FiclSystem::FiclSpit(FICL_VM *vm, char *text, int fNewline)
{
    FiclSystem * Sys = ToFiclSystemPtr(vm);
    if (!Sys)
        return;

    std::string out_text(text);
    if (fNewline)
        out_text += "\n";

    Sys->Display(out_text);

}

//---------------------------------------------------------------------------
//  FiclSystem::Do_Ms() --  Dwell in milliseconds
//---------------------------------------------------------------------------

void  FiclSystem::Do_Ms(int mS)
{
    double Then = Rtc.Now() + static_cast<double>(mS)/1000.0;
    while (Rtc.Now() < Then)
        Innovative::Sleep(1);
}

//-----------------------------------------------------------------------------
//  FiclSystem::Ms() --  Ficl MS wrapper
//-----------------------------------------------------------------------------

void FiclSystem::Ms(FICL_VM *vm)
{
    FiclSystem * Sys = ToFiclSystemPtr(vm);
    if (!Sys)
        return;

    int ms = stackPopINT(vm->pStack);
    Sys->Do_Ms(ms);
}

//-----------------------------------------------------------------------------
//  FiclSystem::OnStreamPreconfig() --
//-----------------------------------------------------------------------------

void FiclSystem::OnStreamPreconfig()
{
    Evaluate("OnStreamPreconfig");
}

//-----------------------------------------------------------------------------
//  FiclSystem::OnStreamSingleTimeAtStartConfig() --
//-----------------------------------------------------------------------------

void FiclSystem::OnStreamSingleTimeAtStartConfig()
{
    Evaluate("OnStreamSingleTimeAtStartConfig");
}

//-----------------------------------------------------------------------------
//  FiclSystem::OnStreamConfigure() --
//-----------------------------------------------------------------------------

void FiclSystem::OnStreamConfigure()
{
    Evaluate("OnStreamConfigure");
}

//-----------------------------------------------------------------------------
//  FiclSystem::OnBeforeStreamStart() --
//-----------------------------------------------------------------------------

void FiclSystem::OnBeforeStreamStart()
{
    Evaluate("OnBeforeStreamStart");
}

//-----------------------------------------------------------------------------
//  FiclSystem::OnAfterStreamStart() --
//-----------------------------------------------------------------------------

void FiclSystem::OnAfterStreamStart()
{
    Evaluate("OnAfterStreamStart");
}

//-----------------------------------------------------------------------------
//  FiclSystem::OnBeforeStreamStop() --
//-----------------------------------------------------------------------------

void FiclSystem::OnBeforeStreamStop()
{
    Evaluate("OnBeforeStreamStop");
}

//-----------------------------------------------------------------------------
//  FiclSystem::OnAfterStreamStop() --
//-----------------------------------------------------------------------------

void FiclSystem::OnAfterStreamStop()
{
    Evaluate("OnAfterStreamStop");
}

//==============================================================================
//  Stand-alone Support Functions
//==============================================================================
//---------------------------------------------------------------------------
//  ToFiclSystemPtr() -- Extract FiclSystem 'back pointer' from Vm
//---------------------------------------------------------------------------

FiclSystem *   ToFiclSystemPtr(FICL_VM * vm)
{
    ficlExtension * p = ToFiclExtensionPtr(vm);

    // Check validity of back pointer to System
    if (!vm || (p->tag != FiclTagValue))
        return 0;

    return static_cast<FiclSystem *>(p->ficlSystemPtr);
}

//---------------------------------------------------------------------------
//  ToFiclTargetPtr() -- Extract FiclTarget 'back pointer' from Vm
//---------------------------------------------------------------------------

IFiclTarget *   ToFiclTargetPtr(FICL_VM * vm)
{
    FiclSystem * Sys = ToFiclSystemPtr(vm);
    if (!Sys)
        return 0;

    return Sys->Target;
}

//---------------------------------------------------------------------------
//  ToFiclExtensionPtr() -- Extract FiclExtension pointer from Vm
//---------------------------------------------------------------------------

ficlExtension * ToFiclExtensionPtr(FICL_VM * vm)
{
    return static_cast<ficlExtension*>(vm->pExtend);
}

//==============================================================================
//  CLASS  IFiclTarget -  Interface to FICL target-side support
//==============================================================================

//---------------------------------------------------------------------------
//  Destructor
//---------------------------------------------------------------------------

IFiclTarget::~IFiclTarget()
{
}

//---------------------------------------------------------------------------
//  IFiclTarget::InitTarget() --
//---------------------------------------------------------------------------

void  IFiclTarget::InitTarget(FiclSystem & engine)
{
    engine.Define("l@", IFiclTarget::LFetch);
    engine.Define("l!", IFiclTarget::LStore);
    engine.Define("p@", IFiclTarget::PFetch);
    engine.Define("p!", IFiclTarget::PStore);
    engine.Define("d@", IFiclTarget::DFetch);
    engine.Define("d!", IFiclTarget::DStore);

    engine.Define("dFreq",       IFiclTarget::DFreq);
    engine.Define("dFreqActual", IFiclTarget::DFreqActual);

    AfterFiclStart(engine);
}

//---------------------------------------------------------------------------
//  IFiclTarget::LFetch() --  Ficl l@ wrapper
//---------------------------------------------------------------------------
//  l@  ( addr -- value )

void  IFiclTarget::LFetch(FICL_VM *vm)
{
    IFiclTarget * Target = ToFiclTargetPtr(vm);
    if (!Target)
        return;

    int addr = stackPopINT(vm->pStack);
    int value = Target->Do_LFetch(addr);

    stackPushINT(vm->pStack, value);
}

//---------------------------------------------------------------------------
//  IFiclTarget::LStore() --  Ficl l! wrapper
//---------------------------------------------------------------------------
//  l!  ( value addr -- )

void  IFiclTarget::LStore(FICL_VM *vm)
{
    IFiclTarget * Target = ToFiclTargetPtr(vm);
    if (!Target)
        return;

    int addr = stackPopINT(vm->pStack);
    int value = stackPopINT(vm->pStack);
    Target->Do_LStore(addr, value);
}

//---------------------------------------------------------------------------
//  IFiclTarget::PFetch() --  Ficl p@ wrapper
//---------------------------------------------------------------------------
//  p@  ( addr -- value )

void  IFiclTarget::PFetch(FICL_VM *vm)
{
    IFiclTarget * Target = ToFiclTargetPtr(vm);
    if (!Target)
        return;

    int addr = stackPopINT(vm->pStack);
    int value = Target->Do_PFetch(addr);

    stackPushINT(vm->pStack, value);
}

//---------------------------------------------------------------------------
//  IFiclTarget::PStore() --  Ficl p! wrapper
//---------------------------------------------------------------------------
//  p!  ( value addr -- )

void  IFiclTarget::PStore(FICL_VM *vm)
{
    IFiclTarget * Target = ToFiclTargetPtr(vm);
    if (!Target)
        return;

    int addr = stackPopINT(vm->pStack);
    int value = stackPopINT(vm->pStack);
    Target->Do_PStore(addr, value);
}

//---------------------------------------------------------------------------
//  IFiclTarget::DFetch() --  Ficl d@ wrapper
//---------------------------------------------------------------------------
//  d@  ( addr dev -- value )

void  IFiclTarget::DFetch(FICL_VM *vm)
{
    IFiclTarget * Target = ToFiclTargetPtr(vm);
    if (!Target)
        return;

    int dev = stackPopINT(vm->pStack);
    int addr = stackPopINT(vm->pStack);
    int value = Target->Do_DFetch(dev, addr);

    stackPushINT(vm->pStack, value);
}

//---------------------------------------------------------------------------
//  IFiclTarget::DStore() --  Ficl d! wrapper
//---------------------------------------------------------------------------
//  d!  ( value addr dev -- )

void  IFiclTarget::DStore(FICL_VM *vm)
{
    IFiclTarget * Target = ToFiclTargetPtr(vm);
    if (!Target)
        return;

    int dev = stackPopINT(vm->pStack);
    int addr = stackPopINT(vm->pStack);
    int value = stackPopINT(vm->pStack);
    Target->Do_DStore(dev, addr, value);
}

//---------------------------------------------------------------------------
//  IFiclTarget::DFreq() --  Ficl Set Frequency wrapper
//---------------------------------------------------------------------------
//  dFreq  ( f:frequency dev --  )

void  IFiclTarget::DFreq(FICL_VM *vm)
{
    IFiclTarget * Target = ToFiclTargetPtr(vm);
    if (!Target)
        return;

    int dev = stackPopINT(vm->pStack);
    float freq = stackPopFloat(vm->fStack);
    Target->Do_DFreq(dev, freq);
}

//---------------------------------------------------------------------------
//  IFiclTarget::DFreqActual() --  Ficl Read frequency wrapper
//---------------------------------------------------------------------------
//  dFreqActual  ( dev -- f:frequency_actual )

void  IFiclTarget::DFreqActual(FICL_VM *vm)
{
    IFiclTarget * Target = ToFiclTargetPtr(vm);
    if (!Target)
        return;

    int dev = stackPopINT(vm->pStack);
	float freq = Target->Do_DFreqActual(dev);
    stackPushFloat(vm->fStack, freq);
}

//---------------------------------------------------------------------------
//  IFiclTarget::Do_LFetch() --  default implementation
//---------------------------------------------------------------------------

int  IFiclTarget::Do_LFetch(int /*addr*/)
{
    return 0;
}

//---------------------------------------------------------------------------
//  IFiclTarget::Do_LStore() --  default implementation
//---------------------------------------------------------------------------

void  IFiclTarget::Do_LStore(int /*addr*/, int /*value*/)
{
}

//---------------------------------------------------------------------------
//  IFiclTarget::Do_PFetch() --  default implementation
//---------------------------------------------------------------------------

int  IFiclTarget::Do_PFetch(int /*addr*/)
{
    return 0;
}

//---------------------------------------------------------------------------
//  IFiclTarget::Do_PStore() --  default implementation
//---------------------------------------------------------------------------

void  IFiclTarget::Do_PStore(int /*addr*/, int /*value*/)
{
}

//---------------------------------------------------------------------------
//  IFiclTarget::Do_DFetch() --  default implementation
//---------------------------------------------------------------------------

int  IFiclTarget::Do_DFetch(int /*dev*/, int /*addr*/)
{
    return 0;
}

//---------------------------------------------------------------------------
//  IFiclTarget::Do_DStore() --  default implementation
//---------------------------------------------------------------------------

void  IFiclTarget::Do_DStore(int /*dev*/, int /*addr*/, int /*value*/)
{
}

//---------------------------------------------------------------------------
//  IFiclTarget::Do_DFetch() --  default implementation
//---------------------------------------------------------------------------

void  IFiclTarget::Do_DFreq(int /*dev*/, float /*freq*/)
{
}

//---------------------------------------------------------------------------
//  IFiclTarget::Do_DFreqActual() --  default implementation
//---------------------------------------------------------------------------

float  IFiclTarget::Do_DFreqActual(int /*dev*/)
{
    return 0.0;
}

} // namespace

