// FiclSystem_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2007 Innovative Integration Corporation.
//  All Rights Reserved.
//

#ifndef FiclSystem_MbH
#define FiclSystem_MbH

#include <sstream>              // Avoids compiler error with _algobase.h
#include <Script_Mb.h>
#include <DataTypes_Mb.h>
#include <StringList_Mb.h>
#include <StreamDeviceEvents_Mb.h>
#include <StopWatch_Mb.h>

#include "Ficl/ficl.h"

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//==============================================================================
//  CLASS  IFiclTarget -  Interface to FICL target-side support
//==============================================================================

class IFiclTarget
{
public:
    void InitTarget(class FiclSystem & engine);
	virtual ~IFiclTarget();

protected:
    //  Functions to be implemented by derived classes
    virtual void    AfterFiclStart(FiclSystem & engine) = 0;

    //  FICL callback implementation functions
    virtual int     Do_LFetch(int addr);
    virtual void    Do_LStore(int addr, int value);
    virtual int     Do_PFetch(int addr);
    virtual void    Do_PStore(int addr, int value);
    virtual int     Do_DFetch(int dev, int addr);
    virtual void    Do_DStore(int dev, int addr, int value);
    virtual void    Do_DFreq(int dev, float freq);
    virtual float   Do_DFreqActual(int dev);

    //  FICL callback hooks
    static void LFetch(FICL_VM *vm);
    static void LStore(FICL_VM *vm);
    static void PFetch(FICL_VM *vm);
    static void PStore(FICL_VM *vm);
    static void DFetch(FICL_VM *vm);
    static void DStore(FICL_VM *vm);
    static void DFreq(FICL_VM *vm);
    static void DFreqActual(FICL_VM *vm);
};

typedef struct ficlExtension
{
    void *  ficlSystemPtr;
    int     tag;
} ficlExtension;

//==============================================================================
//  CLASS  FiclSystem
//==============================================================================

class FiclSystem
{
public:
    friend class IFiclTarget;
    friend class FiclStreamDeviceEventCallbacks;
    friend IFiclTarget * ToFiclTargetPtr(FICL_VM * vm);

    // Ctor
    FiclSystem();
    virtual ~FiclSystem();

    // Methods
    void  ConnectTarget(IFiclTarget * target);
    void  DisconnectTarget()
                {  Target = 0;  }

    int   Start();
    int   Evaluate(const std::string & cmd);  // no prompt
    int   Execute(const std::string & cmd);   // prompt
    
    //int   Dispatch(const std::string & cmd);
    void  Define(const std::string & word, FICL_CODE code,
                 UNS8 flags=FW_DEFAULT);

    OpenWire::ThunkedEventHandler<Innovative::ProcessStatusEvent> OnDisplayText;

    // Pseudo protected
    void  Display(const std::string & txt);

protected:
    // Data
    FICL_VM *               vm;
    FICL_SYSTEM *           system;
    FICL_DICT *             dictionary;
    Innovative::StopWatch   Rtc;
    bool                    isStarted;
    IFiclTarget *           Target;
    std::stringstream       Text;
    ficlExtension           Extension;

    virtual void            OnSystemStart();

    //  FICL Streaming Hooks
    void  OnStreamPreconfig();
    void  OnStreamSingleTimeAtStartConfig();
    void  OnStreamConfigure();
    void  OnBeforeStreamStart();
    void  OnAfterStreamStart();
    void  OnBeforeStreamStop();
    void  OnAfterStreamStop();

    //  FICL system hooks
    static void FiclSpit(FICL_VM *vm, char *text, int fNewline);

    //  FICL callback implementation functions
    void          Do_Ms(int mS);

    //  FICL callback hooks
    static void   Ms(FICL_VM *vm);

    //  Other
    void  CreateSys();
    void  CreateVm();
};

//  Retrieve saved pointers from VM
FiclSystem *   ToFiclSystemPtr(FICL_VM * vm);
IFiclTarget *  ToFiclTargetPtr(FICL_VM * vm);
ficlExtension *   ToFiclExtensionPtr(FICL_VM * vm);

//===========================================================================
//  CLASS FiclStreamDeviceEventCallbacks  --  Stream Notification Callbacks
//===========================================================================

class FiclStreamDeviceEventCallbacks : public Innovative::StreamDeviceEventCallbacks
{
public:
    FiclStreamDeviceEventCallbacks(Innovative::IStreamDeviceEvents & owner, FiclSystem & system)
        :  Innovative::StreamDeviceEventCallbacks(owner), FiclSys(system)
        {}
    virtual ~FiclStreamDeviceEventCallbacks()
        {}
    // Callbacks
    virtual void  CallOnStreamPreconfig(OpenWire::NotifyEvent & Event)
        {
        Innovative::StreamDeviceEventCallbacks::CallOnStreamPreconfig(Event);
        FiclSys.OnStreamPreconfig();
        }
    virtual void  CallOnStreamSingleTimeAtStartConfig(OpenWire::NotifyEvent & Event)
        {
        Innovative::StreamDeviceEventCallbacks::CallOnStreamSingleTimeAtStartConfig(Event);
        FiclSys.OnStreamSingleTimeAtStartConfig();
        }
    virtual void  CallOnStreamConfigure(OpenWire::NotifyEvent & Event)
        {
        Innovative::StreamDeviceEventCallbacks::CallOnStreamConfigure(Event);
        FiclSys.OnStreamConfigure();
        }
    virtual void  CallOnBeforeStreamStart(OpenWire::NotifyEvent & Event)
        {
        Innovative::StreamDeviceEventCallbacks::CallOnBeforeStreamStart(Event);
        FiclSys.OnBeforeStreamStart();
        }
    virtual void  CallOnAfterStreamStart(OpenWire::NotifyEvent & Event)
        {
        Innovative::StreamDeviceEventCallbacks::CallOnAfterStreamStart(Event);
        FiclSys.OnAfterStreamStart();
        }
    virtual void  CallOnBeforeStreamStop(OpenWire::NotifyEvent & Event)
        {
        Innovative::StreamDeviceEventCallbacks::CallOnBeforeStreamStop(Event);
        FiclSys.OnBeforeStreamStop();
        }
    virtual void  CallOnAfterStreamStop(OpenWire::NotifyEvent & Event)
        {
        Innovative::StreamDeviceEventCallbacks::CallOnAfterStreamStop(Event);
        FiclSys.OnAfterStreamStop();
        }

private:
    FiclSystem &          FiclSys;
};


#ifdef __BORLANDC__
	#include <Ficl/FiclFixup_Mb.h>
#endif

#ifdef __CLR_VER
#pragma managed(pop)
#endif
} // namespace Innovative

#endif

