// Fourier_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <Fourier_Mb.h>
#include <ipp.h>
#include <math.h>
#include <Magic_Mb.h>
#include <algorithm>

using namespace std;

namespace   Innovative
{

const double Pi = 4.0 * atan(1.0);

//===========================================================================
//  CLASS Fourier  --
//===========================================================================

//---------------------------------------------------------------------------
// Fourier::Fourier() --
//---------------------------------------------------------------------------

Fourier::Fourier()
	: FAlpha(.01f), FWindow(wtBlackman), FIgnoreDC(false), FIgnoreFs2(false),
	  FSpectrumEnabled(false), FSpectrumMode(smAmplitude), FAverages(0),
	  FAveraging(false)
{
}

//---------------------------------------------------------------------------
// Fourier::Fourier() --
//---------------------------------------------------------------------------

Fourier::Fourier(const Fourier & other)
    : CommonFourier(other), FAlpha(other.FAlpha), FWindow(other.FWindow), FIgnoreDC(other.FIgnoreDC),
	  FIgnoreFs2(other.FIgnoreFs2), FSpectrumEnabled(other.FSpectrumEnabled),
	  FSpectrumMode(other.FSpectrumMode), FAverages(other.FAverages),
	  FAveraging(other.FAveraging)
{
}

//---------------------------------------------------------------------------
//  Fourier::operator = --
//---------------------------------------------------------------------------

Fourier & Fourier::operator = (const Fourier & other)
{
	FAlpha = other.FAlpha;
	FWindow = other.FWindow;
	FIgnoreDC = other.FIgnoreDC;
	FIgnoreFs2 = other.FIgnoreFs2;
	FSpectrumEnabled = other.FSpectrumEnabled;
	FSpectrumMode = other.FSpectrumMode;
	FAverages = other.FAverages;
	FAveraging = other.FAveraging;

	return *this;
}

//---------------------------------------------------------------------------
//  Fourier::Start() --
//---------------------------------------------------------------------------

bool Fourier::Start()
{
	Stop();

	//  Resize to contain required number of elements
    WindowedTime().Resize(FSamples);
    Time().Resize(FSamples);
    Spectrum().Resize(FSamples);
    SpectrumSum().Resize(FSamples);
    PhaseAngle().Resize(FSamples);
    Frequency().Resize(FSamples);
    FrequencySum().Resize(FSamples);
	Scratch().Resize(FSamples);

	return inherited::Start();
}

//---------------------------------------------------------------------------
//  Fourier::Transform() -- Manual forward FFT calculation
//---------------------------------------------------------------------------

bool Fourier::Transform(const IppComplexDG & in, IppComplexDG & out)
{
	if (!mSpec.size())
		Start();

	bool result = transform(in, out);

	// Optional power spectrum calculation
	if (FSpectrumEnabled)
		CalculateSpectrum(out);

	if (FAveraging)
		accumulate(out);
	else if (FAverages)
		average(out);

	return result;
}

//---------------------------------------------------------------------------
//  Fourier::Transform() -- use Internal buffers to calculate FFT
//---------------------------------------------------------------------------

bool Fourier::Transform()
{
	IppComplexDG t(mTime);
	IppComplexDG f(mFrequency);

	return Transform(t, f);
}

//---------------------------------------------------------------------------
//  Fourier::accumulate() --  Accumulate spectra
//---------------------------------------------------------------------------

void Fourier::accumulate(IppComplexDG & buffer)
{
    FrequencySum().Add(buffer);
    if (FSpectrumEnabled)
		SpectrumSum().Add(Spectrum());

    ++FAverages;
}

//---------------------------------------------------------------------------
//  Fourier::average() --  Spectral averaging
//---------------------------------------------------------------------------
#ifdef _MSC_VER		
	#pragma warning(push)	
	#pragma warning(disable : 4244)		
#endif 

void Fourier::average(IppComplexDG & buffer)
{
    buffer = FrequencySum();
    buffer.Divide(Complex(FAverages));

    Spectrum() = SpectrumSum();
    Spectrum().Divide(static_cast<float>(FAverages));

    FAverages = 0;
}

#ifdef _MSC_VER
	#pragma warning(pop)
#endif

//---------------------------------------------------------------------------
//  Fourier::transform() --  Time frequency domain transformation
//---------------------------------------------------------------------------

bool Fourier::transform(const IppComplexDG & in, IppComplexDG & out)
{
	bool result = true;

	IppComplexDG window(in);
	WindowedTime().Window(window, FWindow, FAlpha);

	/// perform forward FFT
	FFTSpec_C_32fc ** pSpec = reinterpret_cast<FFTSpec_C_32fc**>(&mState);

	IppStatus status = ippsFFTFwd_CToC_32fc(reinterpret_cast<Ipp32fc*>(WindowedTime().Ptr()),
						  reinterpret_cast<Ipp32fc*>(out.Ptr()),
						  *pSpec,
						  reinterpret_cast<Ipp8u*>(&(mWorking.front())));
	result &= (status == ippStsNoErr);

    return result;
}

//---------------------------------------------------------------------------
// Fourier::CalculateSpectrum() - Calculate power or magnitude spectrum
//---------------------------------------------------------------------------

void Fourier::CalculateSpectrum(IppComplexDG & src)
{
	//
    //  Calculate Spectrum
    if (FSpectrumMode == smPower)
        Spectrum().PowerSpectrum(src);
    else
        Spectrum().Magnitude(src);

    // Convert to single-sided form
    Spectrum() *= 2.0f;

    float MindB = Spectrum().Min();
    unsigned int margin = Margin();
    if (IgnoreDC())
        {
        // DC is lower .1% of spectrum
        for (unsigned int i = 0; i < margin; ++i)
            Spectrum()[i] = MindB;
        }
    if (IgnoreFs2())
		{
        // Fs/2 is upper .1% of spectrum
        for (unsigned int i = 0; i < margin; ++i)
            Spectrum()[FReach-1-i] = MindB;
        }

    PhaseAngle().Phase(src);
    PhaseAngle() *= static_cast<float>(180.0/Pi);
}

//---------------------------------------------------------------------------
//  Fourier::Accumulating() --  Set spectrum accumulator state
//---------------------------------------------------------------------------

void Fourier::Accumulating(bool state)
{
    FAveraging = state;

    if (!state)
        return;

    FrequencySum().Zero();
    SpectrumSum().Zero();
    FAverages = 0;
}

//---------------------------------------------------------------------------
//  Fourier::Accumulating() --  Return spectrum accumulator status
//---------------------------------------------------------------------------

bool Fourier::Accumulating() const
{
    return FAveraging;
}




} // Innovative

