// FpaLoadThread_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//
#include <FpaLoadThread_Mb.h>
#include <sstream>
#include <HardwareRegister_Mb.h>
#include <SystemSupport_Mb.h>
#include <Magic_Mb.h>
#include <algorithm>

namespace   Innovative
{

//==============================================================================
//  CLASS IFpaLoadBasePersonality -- Interface to FPA specific operations
//==============================================================================
//------------------------------------------------------------------------
//  IFpaLoadBasePersonality::FirmwareRevision()  --  Fetch the firmware revision
//------------------------------------------------------------------------

int IFpaLoadBasePersonality::FirmwareRevision() const
{
    return Fpa.FirmwareRevision();
}

//------------------------------------------------------------------------
//  IFpaLoadBasePersonality::PcbRevision()  --  Fetch the PCB revision
//------------------------------------------------------------------------
//
//int IFpaLoadBasePersonality::PcbRevision() const
//{
//    return Fpa.PcbRevision();
//}
//
//------------------------------------------------------------------------
//  IFpaLoadBasePersonality::Id()  --  Fetch the VsProm ID
//------------------------------------------------------------------------

int IFpaLoadBasePersonality::Id() const
{
    return Fpa.Id();
}

//==============================================================================
//  CLASS FpaPersonalityBase -- FPA Flash Personality Interface common impl
//==============================================================================
//------------------------------------------------------------------------------
//  FpaPersonalityBase::Clock() - Sets clock bit of control register
//------------------------------------------------------------------------------

void FpaPersonalityBase::Clock(int state)
{
    FClock = state;
    Fpa.Clock((state!=0)? true : false );
}

//------------------------------------------------------------------------------
//  FpaPersonalityBase::Clock() - Sets clock bit of control register
//------------------------------------------------------------------------------

int FpaPersonalityBase::Clock() const
{
    return FClock;
}

//------------------------------------------------------------------------------
//  FpaPersonalityBase::Dio() - Sets data bit of control register
//------------------------------------------------------------------------------

void FpaPersonalityBase::Dio(int state)
{
    FDio = state;
    Fpa.Write((state!=0)? true : false );
}

//------------------------------------------------------------------------------
//  FpaPersonalityBase::Dio() - Returns data bit of control register
//------------------------------------------------------------------------------

int FpaPersonalityBase::Dio() const
{
    return Fpa.Read() & 1;
}

//------------------------------------------------------------------------------
// FpaPersonalityBase::Strobe() - Pulses the clock bit of control register
//------------------------------------------------------------------------------

void FpaPersonalityBase::Strobe()
{
    Clock(1);
    Clock(0);
}

//------------------------------------------------------------------------------
// FpaPersonalityBase::Address() - Presents specified address to Flash
//------------------------------------------------------------------------------

void FpaPersonalityBase::Address(int addr)
{
    if (FAddress == addr)
        return;

    if (addr == (FAddress+1))
        {
        Mode(Incrementing);
        Strobe();
        FAddress = addr;
        return;
        }

    FAddress = addr;

    Mode(Addressing);

    int mask = 1 << (FAddressBits-1);
    for (int i = 0; i < FAddressBits; ++i)
        {
        Dio((addr & mask) ? 1 : 0);
        Strobe();
        addr <<= 1;
        }
}

//------------------------------------------------------------------------------
// FpaPersonalityBase::Address() - Returns last address
//------------------------------------------------------------------------------

int FpaPersonalityBase::Address() const
{
    return FAddress;
}

//------------------------------------------------------------------------------
// FpaPersonalityBase::Data() - Clocks out serial data value
//------------------------------------------------------------------------------

void FpaPersonalityBase::Data(int value)
{
    FData = value;
    Mode(Quantizing);

    for (int i = 0; i < FDataBits; ++i)
        {
        Dio((value&(1<<(FDataBits-1))) ? 1 : 0);
        Strobe();
        value <<= 1;
        }
}

//------------------------------------------------------------------------------
// FpaPersonalityBase::Data() - Returns last written data
//------------------------------------------------------------------------------

int FpaPersonalityBase::Data() const
{
    return FData;
}

//------------------------------------------------------------------------------
// FpaPersonalityBase::Mode() - Sets mode bits of control register
//------------------------------------------------------------------------------

void FpaPersonalityBase::Mode(FpaMode mode)
{
    FMode = mode;
    Fpa.Mode(mode);
}

//------------------------------------------------------------------------------
// FpaPersonalityBase::Mode() - Returns current mode setting
//------------------------------------------------------------------------------

FpaPersonalityBase::FpaMode FpaPersonalityBase::Mode() const
{
    return FMode;
}

//------------------------------------------------------------------------------
// FpaPersonalityBase::Prom() - Reads byte from specified Flash address
//------------------------------------------------------------------------------

int FpaPersonalityBase::Prom(int addr)
{
    Address(addr);

    Mode(Fetching);
    return ShiftIn();
}

//------------------------------------------------------------------------------
// FpaPersonalityBase::ShiftIn() - Shift in data from prom
//------------------------------------------------------------------------------

int FpaPersonalityBase::ShiftIn()
{
    Strobe();

    Mode(Quantizing);

    int result = 0;
    for (int i = 0; i < FDataBits; i++)
        {
        result <<= 1;
        result |= Dio();
        Strobe();
        }

    FData = result;

    return result;
}

//------------------------------------------------------------------------------
// FpaPersonalityBase::Prom() - Writes byte to specified address
//------------------------------------------------------------------------------

void FpaPersonalityBase::Prom(int addr, int data)
{
    Address(addr);
    Data(data);

    Mode(Storing);
    Strobe();
}

//-------------------------------------------------------------
// FpaPersonalityBase::Flash() - Reads byte from specified address
//-------------------------------------------------------------

int FpaPersonalityBase::Flash(int addr)
{
    return Prom(addr);
}

//-------------------------------------------------------------
// FpaPersonalityBase::Flash() - Programs byte at specified address
//-------------------------------------------------------------

bool FpaPersonalityBase::Flash(int addr, int data)
{
    //
    //  Start the flash operation
    OnFlashWrite_Start(addr, data);

    const double Timeout = 0.500;   // in sec
    double Then = Rtc.Now() + Timeout;
	while (Prom(addr) != data)
        {
        if (Rtc.Now() < Then)
            continue;

        std::stringstream msg;
        msg << "Prom wrote 0x" << std::hex << Prom(addr)
            << " instead of 0x" << data
            << " at address 0x" << addr;
        ProcessStatusEvent e(msg.str());
        OnLoadError.Execute(e);
        return false;
        }

    return true;
}

//-------------------------------------------------------------
// FpaPersonalityBase::Authenticate() - Verify board integrity
//-------------------------------------------------------------

bool FpaPersonalityBase::Authenticate()
{
    FSeed = static_cast<int>(Rtc.Now()*256.0)&0xff;
    Data(FSeed);

    Mode(Authenticating);

    // Generate the random #
    for (int i = 0; i < 8; i++)
        Strobe();

    // Read the result
    int result = 0;
    for (int i = 0; i < 8; i++)
        {
        result <<= 1;
        result |= Dio();
        Strobe();

        Randomize();
        }

    return FSeed == result;
}

//-------------------------------------------------------------
// FpaPersonalityBase::Randomize() - Generate next random number
//-------------------------------------------------------------

inline void FpaPersonalityBase::Randomize()
{
    int bit7 = (FSeed>>7) & 1;
    int bit1 = (FSeed>>1) & 1;

    FSeed <<= 1;
    FSeed |= (bit7^bit1);
    FSeed &= 0xff;
}

//-------------------------------------------------------------
// FpaPersonalityBase::Test() -
//-------------------------------------------------------------

int FpaPersonalityBase::Test(int value)
{
    int Result = 0;

    FData = value;
    Mode(Quantizing);

    const int Width = 32;

    for (int i = 0; i < Width; ++i)
        {
        Dio((value&(1<<(Width-1))) ? 1 : 0);
        Strobe();
        value <<= 1;

        Result <<= 1;
        Result |= (Dio() & 1);
        }

    return Result;
}
//==============================================================================
//  CLASS StandardFpaPersonality -- Default FPA Flash Personality Interface
//==============================================================================
//------------------------------------------------------------------------------
//  StandardFpaPersonality::Initialize() - Initializes flash for programming
//------------------------------------------------------------------------------

bool  StandardFpaPersonality::Initialize()
{
//    volatile int result = Test(0);
//    volatile bool status = Authenticate();

    // Attempt to autoselect, assuming progressively "larger" devices
    for (int i = 19; i < 24; ++i)
        {
        FAddress = -1;
        FAddressBits = i;

        Prom(0x5555, 0xaa);
        Prom(0x2aaa, 0x55);
        Prom(0x5555, 0x90);            // Autoselect
        FDeviceId = (Prom(1) & 0xff);

        // If autoselect succeeds, proceed
        if (FDeviceId && (FDeviceId != 0xff))
            break;

        // If autoselect fails, retry as "large" device
        FAddress = -1;
        }

    // Error if sector-protected
    FProtected = Prom(2) & 0x01;

    Prom(0x5555, 0xaa);
    Prom(0x2aaa, 0x55);
    Prom(0x5555, 0xf0);            // Reset

    switch(FDeviceId)
        {
        case 0x20: // 'F010
            FSectorSize = 0x04000;
            FSectors = 8;
            FAddressBits = 19;
            break;
        default:
        case 0xa4: // 'F040
            FSectorSize = 0x10000;
            FSectors = 8;
            FAddressBits = 19;
            break;
        case 0x4f: // 'LV040
            FSectorSize = 0x10000;
            FSectors = 8;
            FAddressBits = 19;
            break;
        case 0x38: // 'LV081
            FSectorSize = 0x10000;
            FSectors = 16;
            FAddressBits = 20;
            break;
        case 0xa3: // 'LV033
            FSectorSize = 0x10000;
            FSectors = 64;
            FAddressBits = 22;
            break;
        case 0xc7: // 'LV116
            FSectorSize = 0x10000;
            FSectors = 35;
            FAddressBits = 21;
            break;
        }

    // Store device size, in bytes
    FDeviceSize = 1 << FAddressBits;

    // Just a fallow read cycle
    Fallow = Prom(0);
    return true;
}

//------------------------------------------------------------------------------
//  StandardFpaPersonality::EraseSector() - Erase flash sector contents
//------------------------------------------------------------------------------

bool StandardFpaPersonality::EraseSector(int sector)
{
	Prom(0x5555, 0xaa);
	Prom(0x2aaa, 0x55);
	Prom(0x5555, 0x80);

	Prom(0x5555, 0xaa);
	Prom(0x2aaa, 0x55);
    int Addr = sector * FSectorSize;
	Prom(Addr, 0x30);

    const double Timeout = 10.0;   // in sec
    double Then = Rtc.Now() + Timeout;
	while ((Prom(Addr) & 0x80) != 0x80)
        {
        if (Rtc.Now() > Then)
            {
            ProcessStatusEvent e("Timeout erasing sector");
            OnLoadError.Execute(e);
            return false;
            }
        }

    return true;
}

//------------------------------------------------------------------------------
//  StandardFpaPersonality::Program() -- Write buffer to flash
//------------------------------------------------------------------------------

bool  StandardFpaPersonality::Program(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    // Program device
    for (unsigned int i = 0; i < Image.Size(); ++i)
        {
        if (!Flash(i, Image.Buffer[i]))
            {
            std::ostringstream msg;
            msg << "Unable to program byte " << i;

            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }

        Percent[1] = static_cast<int>((i*100)/Image.Size());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent e(100);
    OnLoadProgress.Execute(e);

    return true;
}

//------------------------------------------------------------------------------
//  StandardFpaPersonality::Verify() -- Verify buffer against flash contents
//------------------------------------------------------------------------------

bool StandardFpaPersonality::Verify(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    // Program device
    for (unsigned int i = 0; i < Image.Size(); ++i)
        {
        int dst = Flash(i);
        int src = Image.Buffer[i];
        if (dst != src)
            {
            std::ostringstream msg;

            msg << "Prom read 0x" << std::hex << dst
                << " instead of 0x" << src
                << " at address 0x" << i;
            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }

        Percent[1] = static_cast<int>((i*100)/Image.Size());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent p(100);
    OnLoadProgress.Execute(p);
    return true;
}

//---------------------------------------------------------------------------
//  StandardFpaPersonality::Read() -- Fill buffer from flash
//---------------------------------------------------------------------------

bool StandardFpaPersonality::Read(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    // Read device
    for (unsigned int i = 0; i < Image.Size(); ++i)
        {
        Image.Buffer[i] = Flash(i);

        Percent[1] = static_cast<int>((i*100)/Image.Size());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent p(100);
    OnLoadProgress.Execute(p);
    return true;
}

//------------------------------------------------------------------------------
//  StandardFpaPersonality::OnFlashWrite_Start() -- Start Flash write
//------------------------------------------------------------------------------

void  StandardFpaPersonality::OnFlashWrite_Start(int addr, int data)
{
	Prom(0x5555, 0xaa);
	Prom(0x2aaa, 0x55);
	Prom(0x5555, 0xa0);

	Prom(addr, data);
}

//==============================================================================
//  CLASS CommonNJ_FpaChipImpl -- X5 FPA Flash Sub-Personality Interface
//==============================================================================
//------------------------------------------------------------------------------
//  declaration of class CommonNJ_FpaChipImpl
//------------------------------------------------------------------------------

class CommonNJ_FpaChipImpl : public IX5FpaChip
{
public:
    CommonNJ_FpaChipImpl(class X5FpaPersonality * owner)
        :  IX5FpaChip(owner), FUnlockBypass(false),
           FUnlockBypassCapable(false)
        {}
    virtual ~CommonNJ_FpaChipImpl()
        {}

    void  CommonInit_QTest();
    void  CommonInit_DevId();
    void  CommonInit_Protected();

    virtual unsigned int ByteOrder(unsigned int data);

    virtual void        OnFlashWrite_Start(int addr, int data);
    virtual void        OnSingleCellPgm_Start();
    virtual void        OnSingleCellPgm_End();
    virtual void        OnEraseSector_Start(unsigned int sector);
    virtual void        OnWriteBufferProgram(FpgaImageDatagram & ImageDg, unsigned int Addr, unsigned int wc);

protected:
    bool  FUnlockBypass;
    bool  FUnlockBypassCapable;

};

//------------------------------------------------------------------------------
//  CommonNJ_FpaChipImpl::CommonInit_QTest() - Search for Q to test for valid chip
//------------------------------------------------------------------------------

void  CommonNJ_FpaChipImpl::CommonInit_QTest()
{
    //
    //  --- Check if ROM is present by looking for Q ---
    EnterCFIQueryMode();      // Start CFI Query Mode

    int query;
    query = Prom(0x10) & 0xff;  // should read "Q" (0x51)

        {
        std::stringstream msg;
        msg << "Prom Query: 0x" << std::hex << query;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }

    LeaveCFIQueryMode();      // end CFI Query Mode
}

//------------------------------------------------------------------------------
//  CommonNJ_FpaChipImpl::CommonInit_DevId() - Read device ID
//------------------------------------------------------------------------------

void  CommonNJ_FpaChipImpl::CommonInit_DevId()
{
    Prom(0x555, 0xaa);
    Prom(0x2aa, 0x55);
    Prom(0x555, 0x90);            // Autoselect
    unsigned int hi = (Prom(0x01) & 0xffff);
    unsigned int lo = (Prom(0x0E) & 0xffff);      // need 2nd word to get part ID
    FDeviceId = (hi<<16) | lo;
        {
        std::stringstream msg;
        msg << "Prom DevId: 0x" << std::hex << FDeviceId;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }

}
//------------------------------------------------------------------------------
//  CommonNJ_FpaChipImpl::CommonInit_Protected() - Read proctected bit
//------------------------------------------------------------------------------

void  CommonNJ_FpaChipImpl::CommonInit_Protected()
{
    // Error if sector-protected
    Prom(0x555, 0xaa);
    Prom(0x2aa, 0x55);
    Prom(0x555, 0x90);            // Autoselect
    FProtected = Prom(2) & 0x01;
        {
        std::stringstream msg;
        msg << "Protected Bit: 0x" << std::hex << FProtected;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }
}

//------------------------------------------------------------------------------
//  CommonNJ_FpaChipImpl::ByteOrder() -- Reorder Bytes between image and chip
//------------------------------------------------------------------------------

unsigned int  CommonNJ_FpaChipImpl::ByteOrder(unsigned int data)
{
    return ByteSwap(data);    // Swap Bytes
}

//------------------------------------------------------------------------------
//  CommonNJ_FpaChipImpl::OnFlashWrite_Start() -- Start Flash write
//------------------------------------------------------------------------------

void  CommonNJ_FpaChipImpl::OnFlashWrite_Start(int addr, int data)
{
    if (!FUnlockBypass) // if disabled, we need this sequence
        {
        Prom(0x555, 0xaa);
        Prom(0x2aa, 0x55);
        }

    Prom(0x555, 0xa0);  // address is don't care
	Prom(addr, data);
}

//------------------------------------------------------------------------------
//  CommonNJ_FpaChipImpl::OnSingleCellPgm_Start() -- Start Cell Program
//------------------------------------------------------------------------------

void  CommonNJ_FpaChipImpl::OnSingleCellPgm_Start()
{
    // enable Unlock Bypass to save 2 Prom transactions per word written
    if (FUnlockBypassCapable)
        {
        Prom(0x555, 0xaa);
        Prom(0x2aa, 0x55);
        Prom(0x555, 0x20);
        FUnlockBypass = true;
        }
}

//------------------------------------------------------------------------------
//  CommonNJ_FpaChipImpl::OnSingleCellPgm_End() -- End Cell Program
//------------------------------------------------------------------------------

void  CommonNJ_FpaChipImpl::OnSingleCellPgm_End()
{
    if (FUnlockBypassCapable)
        {
        Prom(0x555, 0x90);      // disable Unlock Bypass mode
        Prom(0x2aa, 0x00);      // addresses are don't care
        FUnlockBypass = false;
        }
}

//------------------------------------------------------------------------------
//  CommonNJ_FpaChipImpl::OnEraseSector_Start() -- Erase sector
//------------------------------------------------------------------------------

void  CommonNJ_FpaChipImpl::OnEraseSector_Start(unsigned int sector)
{
	Prom(0x555, 0xaa);
	Prom(0x2aa, 0x55);
	Prom(0x555, 0x80);

	Prom(0x555, 0xaa);
	Prom(0x2aa, 0x55);

    int Addr = SectorBaseAddress(sector);
	Prom(Addr, 0x30);

        {
        std::stringstream msg;
        msg << "Sector Base Address: 0x" << std::hex << Addr;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }
}

//------------------------------------------------------------------------------
//  CommonNJ_FpaChipImpl::OnWriteBufferProgram() -- Issue write buffer command
//------------------------------------------------------------------------------

void CommonNJ_FpaChipImpl::OnWriteBufferProgram(FpgaImageDatagram & ImageDG, unsigned int Addr,
                                                unsigned int wc)
{
    Prom(0x555, 0xAA);      // Unlock
    Prom(0x2AA, 0x55);      // Unlock
    Prom(Addr,  0x25);      // Write Buffer Load
    Prom(Addr,  wc-1);      // Word Count
    for (unsigned int count=0; count<wc; count++)
        {
        unsigned short data_word = static_cast<unsigned short>(ImageDG.AsShort(Addr+count));
        Prom(Addr+count, ByteSwap(data_word));
        }
    Prom(Addr,  0x29);      // Write Buffer to Flash

}

//==============================================================================
//  CLASS TypeJ_FpaChipImpl -- X5 FPA Flash Sub-Personality Implementation
//==============================================================================
//------------------------------------------------------------------------------
//  declaration of class TypeJ_FpaChipImpl
//------------------------------------------------------------------------------

class TypeJ_FpaChipImpl : public CommonNJ_FpaChipImpl
{
public:
    TypeJ_FpaChipImpl(class X5FpaPersonality * owner)
        :  CommonNJ_FpaChipImpl(owner)
        {}
    virtual ~TypeJ_FpaChipImpl()
        {}

    virtual bool  Initialize();
    virtual void  EnterCFIQueryMode();
    virtual void  LeaveCFIQueryMode();

    virtual unsigned int SectorBaseAddress(unsigned int sector);

    virtual  unsigned int   StartSector() const;
    virtual  unsigned int   Sectors() const;
    virtual  unsigned int   GoldenImageStartSector() const;
    virtual  unsigned int   GoldenImageSectors() const;

protected:
    unsigned int Fallow;
};

//------------------------------------------------------------------------------
//  TypeJ_FpaChipImpl::Initialize() - Initializes flash for programming
//------------------------------------------------------------------------------

bool  TypeJ_FpaChipImpl::Initialize()
{
    //  Display CPLD Type, Rev in Log
        {
        std::stringstream msg;
        msg << "J Type ROM Detected";

        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }

    //
    //  --- Check if ROM is present by looking for Q ---
    CommonInit_QTest();

    CommonInit_DevId();

    CommonInit_Protected();

    //
    //  Determine Unlock Bypass Mode
    //  ....On J Part, we are Unlock Bypass Capable
    FUnlockBypassCapable = true;
        {
        std::stringstream msg;
        msg << "Unlock Bypass Capable Bit: 0x" << std::hex << FUnlockBypassCapable;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }

    Prom(0x555, 0xf0);       // Reset (exit autoselect)
    Fallow = Prom(0);        // Just a fallow read cycle

    FUnlockBypass = false;
    return true;
}

//------------------------------------------------------------------------------
//  TypeJ_FpaChipImpl::EnterCFIQueryMode() -- Enter CFI Mode
//------------------------------------------------------------------------------

void  TypeJ_FpaChipImpl::EnterCFIQueryMode()
{
    Prom(0x55, 0x98);          // enter CFI query mode ("J Part")
}

//------------------------------------------------------------------------------
//  TypeJ_FpaChipImpl::LeaveCFIQueryMode() -- Enter CFI Mode
//------------------------------------------------------------------------------

void  TypeJ_FpaChipImpl::LeaveCFIQueryMode()
{
    Prom(0x0, 0xf0);            // Reset (exit CFI query mode)
}

//------------------------------------------------------------------------------
// TypeJ_FpaChipImpl::SectorBaseAddress() -
//------------------------------------------------------------------------------

unsigned int  TypeJ_FpaChipImpl::SectorBaseAddress(unsigned int sector)
{
    //
    //  S29WS128J Sector Address Calculation
    if (sector<8)
        {
        return sector*0x1000;
        }
    else if (sector<262)
        {
        return (sector-8)*0x8000 + 0x8000;
        }
    else
        {
        return 0x1000*(sector-262)+0x7F8000;
        }
}

//------------------------------------------------------------------------------
// TypeJ_FpaChipImpl::StartSector() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeJ_FpaChipImpl::StartSector() const
{
    return 0;
}

//------------------------------------------------------------------------------
// TypeJ_FpaChipImpl::Sectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeJ_FpaChipImpl::Sectors() const
{
    //  On J part, want first half, but 1st 0x20000 is 8 sections
    return 128 - 1 + 8;
}

//------------------------------------------------------------------------------
// TypeJ_FpaChipImpl::GoldenImageStartSector() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeJ_FpaChipImpl::GoldenImageStartSector() const
{
    return StartSector() + Sectors();
}

//------------------------------------------------------------------------------
// TypeJ_FpaChipImpl::GoldenImageSectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeJ_FpaChipImpl::GoldenImageSectors() const
{
    return Sectors();   // Same number of sectors as other half
}

//==============================================================================
//  CLASS TypeN_FpaChipImpl -- X5 FPA Flash Sub-Personality Implementation
//==============================================================================
//------------------------------------------------------------------------------
//  declaration of class TypeN_FpaChipImpl
//------------------------------------------------------------------------------

class TypeN_FpaChipImpl : public CommonNJ_FpaChipImpl
{
public:
    TypeN_FpaChipImpl(class X5FpaPersonality * owner)
        :  CommonNJ_FpaChipImpl(owner)
        {}
    virtual ~TypeN_FpaChipImpl()
        {}

    virtual bool  Initialize();
    virtual void  EnterCFIQueryMode();
    virtual void  LeaveCFIQueryMode();

    virtual unsigned int SectorBaseAddress(unsigned int sector);

    virtual  unsigned int       StartSector() const ;
    virtual  unsigned int       Sectors() const ;
    virtual  unsigned int       GoldenImageStartSector() const;
    virtual  unsigned int       GoldenImageSectors() const;

protected:
    unsigned int Fallow;

};

//------------------------------------------------------------------------------
//  TypeN_FpaChipImpl::Initialize() - Initializes flash for programming
//------------------------------------------------------------------------------

bool  TypeN_FpaChipImpl::Initialize()
{
    //  Display CPLD Type, Rev in Log
    {
    std::stringstream msg;
    msg << "N Type ROM Detected";

    ProcessStatusEvent e(msg.str());
    Owner->OnLoadError.Execute(e);
    }
    //
    //  --- Check if ROM is present by looking for Q ---
    CommonInit_QTest();
    //
    //  --- Read configuration register ---
    Prom(0x555, 0xaa);
    Prom(0x2aa, 0x55);
    Prom(0x555, 0xC6);            // configuration register
    int conf_reg = Prom(0x0);

    {
    std::stringstream msg;
    msg << "Prom Config: 0x" << std::hex << conf_reg;
    ProcessStatusEvent e(msg.str());
    Owner->OnLoadError.Execute(e);
    }
    Prom(0x0, 0xf0);
    //
    //  Read Device ID
    CommonInit_DevId();

    //
    //  Read Protect Bit
    CommonInit_Protected();
    //
    //  Determine Unlock Bypass Mode
    //  On N Part, we can check if we are Unlock Bypass Capable
    EnterCFIQueryMode();      // Start CFI Query Mode
    FUnlockBypassCapable = Prom(0x51) & 0x01;   // set if capable
        {
        std::stringstream msg;
        msg << "Unlock Bypass Capable Bit: 0x" << std::hex << FUnlockBypassCapable;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }


    Prom(0x555, 0xf0);       // Reset (exit autoselect)
    Fallow = Prom(0);        // Just a fallow read cycle

    FUnlockBypass = false;
    return true;
}

//------------------------------------------------------------------------------
//  TypeN_FpaChipImpl::EnterCFIQueryMode() -- Enter CFI Mode (for all parts)
//------------------------------------------------------------------------------

void  TypeN_FpaChipImpl::EnterCFIQueryMode()
{
    Prom(0x555, 0x98);         // enter CFI query mode  ("N Part")
}

//------------------------------------------------------------------------------
//  TypeN_FpaChipImpl::LeaveCFIQueryMode() -- Enter CFI Mode
//------------------------------------------------------------------------------

void  TypeN_FpaChipImpl::LeaveCFIQueryMode()
{
    Prom(0x0, 0xf0);            // Reset (exit CFI query mode)
}

//------------------------------------------------------------------------------
// TypeN_FpaChipImpl::SectorBaseAddress() -
//------------------------------------------------------------------------------

unsigned int  TypeN_FpaChipImpl::SectorBaseAddress(unsigned int sector)
{
    if (sector<4)
        {
        return sector*0x4000;
        }
    else if (sector<130)
        {
        return 0x10000*(sector-4)+0x10000;
        }
    else
        {
        return 0x4000*(sector-130)+0x7F0000;
        }
}

//------------------------------------------------------------------------------
// TypeN_FpaChipImpl::StartSector() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeN_FpaChipImpl::StartSector() const
{
    return 0;
}

//------------------------------------------------------------------------------
// TypeN_FpaChipImpl::Sectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeN_FpaChipImpl::Sectors() const
{
    //  On N part, want first half, but 1st 0x10000 is 4 sections
    return  64 - 1 + 4;
}

//------------------------------------------------------------------------------
// TypeN_FpaChipImpl::GoldenImageStartSector() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeN_FpaChipImpl::GoldenImageStartSector() const
{
    return StartSector() + Sectors();
}

//------------------------------------------------------------------------------
// TypeN_FpaChipImpl::GoldenImageSectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeN_FpaChipImpl::GoldenImageSectors() const
{
    return Sectors();   // Same number of sectors as other half
}

//==============================================================================
//  CLASS TypeX_FpaChipImpl -- X5 FPA Flash Sub-Personality Implementation
//==============================================================================
//  Xilinx Roms
//------------------------------------------------------------------------------
//  declaration of class TypeX_FpaChipImpl
//------------------------------------------------------------------------------

class TypeX_FpaChipImpl : public IX5FpaChip
{
public:
    TypeX_FpaChipImpl(class X5FpaPersonality * owner)
        :  IX5FpaChip(owner)
        {}
    virtual ~TypeX_FpaChipImpl()
        {}

    virtual bool Initialize();
    virtual void  EnterCFIQueryMode();
    virtual void  LeaveCFIQueryMode();

    virtual unsigned int SectorBaseAddress(unsigned int sector);

    virtual  unsigned int       ByteOrder(unsigned int data);
    virtual  unsigned int       StartSector() const ;
    virtual  unsigned int       Sectors() const ;
    virtual  unsigned int       GoldenImageStartSector() const;
    virtual  unsigned int       GoldenImageSectors() const;

    virtual void        OnFlashWrite_Start(int addr, int data);
    virtual void        OnSingleCellPgm_Start();
    virtual void        OnSingleCellPgm_End();
    virtual void        OnEraseSector_Start(unsigned int sector);
    virtual void        OnWriteBufferProgram(FpgaImageDatagram & ImageDg, unsigned int Addr, unsigned int wc);

protected:
    unsigned int Fallow;
    //
    //  Unique Operations
    unsigned int   TotalSectors() const;
    bool           UnlockSector(unsigned int sector);
    void           Unlock();
    void           ResetToNormal(unsigned int addr = 0x00);
    int            ReadUnlockStatus(unsigned int addr);
};

//------------------------------------------------------------------------------
//  TypeX_FpaChipImpl::Initialize() - Initializes flash for programming
//------------------------------------------------------------------------------

bool  TypeX_FpaChipImpl::Initialize()
{
    //  Display CPLD Type, Rev in Log
    {
    std::stringstream msg;
    msg << "X Type ROM Detected";

    ProcessStatusEvent e(msg.str());
    Owner->OnLoadError.Execute(e);
    }
    //
    //  Initialize Configuration Register
    //    (Before any reads)
    Prom(0xBDDF, 0x60);
    Prom(0xBDDF, 0x03);
    //   DEBUG
    //Prom(0x03, 0x05); // test lower address bits
    //
    //  Unlock all sectors
    Unlock();
    //
    //  --- Check if ROM is present by looking for Q ---
    EnterCFIQueryMode();      // Start CFI Query Mode

    int query;
    query = Prom(0x10) & 0xff;  // should read "Q" (0x51)

    {
    std::stringstream msg;
    msg << "Prom Query: 0x" << std::hex << query;
    ProcessStatusEvent e(msg.str());
    Owner->OnLoadError.Execute(e);
    }
    //
    //  --- Read Device ID ---
    unsigned int hi = (Prom(0x00) & 0xffff);    // 0x0049
    unsigned int lo = (Prom(0x01) & 0xffff);    // 0x506B
    FDeviceId = (hi<<16) | lo;
        {
        std::stringstream msg;
        msg << "Prom DevId: 0x" << std::hex << FDeviceId;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }
    LeaveCFIQueryMode();
    Prom(0x00, 0x90);          // enter ES query mode ("X Part")
   //
    //  --- Read Electronic Signature Codes ---
    int ESC[20];
    ESC[0] = Prom(2);
        {
        std::stringstream msg;
        msg << "Block Protect(0x2): 0x" << std::hex << ESC[0];
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }
    ESC[1] = Prom(5);
        {
        std::stringstream msg;
        msg << "Cfg Reg(0x5): 0x" << std::hex << ESC[1];
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }
    ESC[2] = Prom(0x80);
        {
        std::stringstream msg;
        msg << "Pr0 Lock(0x80): 0x" << std::hex << ESC[2];
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }
    ResetToNormal();

//    Owner->Flash(0x80000, ByteOrder(0xDEAD));
//    int test1 = Prom(0x80000);
//    Owner->Flash(0x80001, ByteOrder(0xD00D));
//    int test2 = Prom(0x80001);
//        {
//        std::stringstream msg;
//        msg << "Flash 0x80K: 0x" << std::hex << test1;
//        msg << "Flash 0x80K+1: 0x" << std::hex << test2;
//        ProcessStatusEvent e(msg.str());
//        Owner->OnLoadError.Execute(e);
//        }

    return true;
}


//------------------------------------------------------------------------------
//  TypeX_FpaChipImpl::EnterCFIQueryMode() -- Enter CFI Mode
//------------------------------------------------------------------------------

void  TypeX_FpaChipImpl::EnterCFIQueryMode()
{
    Prom(0x00, 0x98);          // enter CFI query mode ("X Part")
}

//------------------------------------------------------------------------------
//  TypeX_FpaChipImpl::LeaveCFIQueryMode() -- Leave CFI Mode
//------------------------------------------------------------------------------

void  TypeX_FpaChipImpl::LeaveCFIQueryMode()
{
    ResetToNormal();
    Fallow = Prom(0);
}

//------------------------------------------------------------------------------
//  TypeX_FpaChipImpl::ResetToNormal() -- Reset to normal mode
//------------------------------------------------------------------------------

void  TypeX_FpaChipImpl::ResetToNormal(unsigned int addr)
{
    Prom(addr, 0xff);     // Jason thinks this read array should do it
}

//------------------------------------------------------------------------------
// TypeX_FpaChipImpl::SectorBaseAddress() -
//------------------------------------------------------------------------------

unsigned int  TypeX_FpaChipImpl::SectorBaseAddress(unsigned int sector)
{
    if (sector<127)
        {
        return sector*0x10000;
        }
    else      //  only small ones on the top of range
        {
        return 0x4000*(sector-127)+0x7F0000;
        }
}

//------------------------------------------------------------------------------
//  TypeX_FpaChipImpl::ByteOrder() -- Reorder Bytes between image and chip
//------------------------------------------------------------------------------

unsigned int  TypeX_FpaChipImpl::ByteOrder(unsigned int data)
{
    return ByteSwap(data);    // Swap Bytes
}

//------------------------------------------------------------------------------
// TypeX_FpaChipImpl::StartSector() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeX_FpaChipImpl::StartSector() const
{
    return 0;
}

//------------------------------------------------------------------------------
// TypeX_FpaChipImpl::Sectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeX_FpaChipImpl::Sectors() const
{
    //  On X part, want first half
    return 64;
}

//------------------------------------------------------------------------------
// TypeX_FpaChipImpl::GoldenImageStartSector() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeX_FpaChipImpl::GoldenImageStartSector() const
{
    return StartSector() + Sectors();
}

//------------------------------------------------------------------------------
// TypeX_FpaChipImpl::GoldenImageSectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeX_FpaChipImpl::GoldenImageSectors() const
{
    return  64 - 1 + 4;    // Not symmetric
}

//------------------------------------------------------------------------------
// TypeX_FpaChipImpl::TotalSectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeX_FpaChipImpl::TotalSectors() const
{
    return  Sectors() + GoldenImageSectors();    
}

//------------------------------------------------------------------------------
//  TypeX_FpaChipImpl::OnFlashWrite_Start() -- Start Flash write
//------------------------------------------------------------------------------

void  TypeX_FpaChipImpl::OnFlashWrite_Start(int addr, int data)
{
    Prom(addr, 0x40);
	Prom(addr, data);
    //  Leave current mode
    ResetToNormal(addr);
}

//------------------------------------------------------------------------------
//  TypeX_FpaChipImpl::OnSingleCellPgm_Start() -- Begin sequence of flash
//------------------------------------------------------------------------------

void  TypeX_FpaChipImpl::OnSingleCellPgm_Start()
{
    //  don't think we need anything here
}

//------------------------------------------------------------------------------
//  TypeX_FpaChipImpl::OnSingleCellPgm_End() --  End sequence of flash
//------------------------------------------------------------------------------

void  TypeX_FpaChipImpl::OnSingleCellPgm_End()
{
    //  don't think we need anything here
}

//------------------------------------------------------------------------------
//  TypeX_FpaChipImpl::OnEraseSector_Start() -- Erase sector
//------------------------------------------------------------------------------

void  TypeX_FpaChipImpl::OnEraseSector_Start(unsigned int sector)
{
    int Addr = SectorBaseAddress(sector);
	Prom(Addr, 0x20);
	Prom(Addr, 0xD0);
        {
        std::stringstream msg;
        msg << "Sector Base Address: 0x" << std::hex << Addr;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }
    //  Leave current mode
    ResetToNormal(Addr);

    // DEBUG - display unlock status
//    int status = ReadUnlockStatus(Addr); // read unlock status from Signature Info
//	ResetToNormal();
//        {
//        std::stringstream msg;
//        msg << "Unlock Code: 0x" << std::hex << status;
//        ProcessStatusEvent e(msg.str());
//        Owner->OnLoadError.Execute(e);
//        }

}

//------------------------------------------------------------------------------
//  TypeX_FpaChipImpl::OnWriteBufferProgram() -- Issue write buffer command
//------------------------------------------------------------------------------

void TypeX_FpaChipImpl::OnWriteBufferProgram(FpgaImageDatagram & ImageDG, unsigned int Addr,
                                                unsigned int wc)
{
	while (1==1)
        {
        Prom(Addr, 0xE8);      // Load Address
        if ((Prom(Addr) & 0x80) == 0x80)
            break;
        }
    Prom(Addr,  wc-1);      // Word Count

    for (unsigned int count=0; count<wc; count++)
        {
        unsigned short data_word = static_cast<unsigned short>(ImageDG.AsShort(Addr+count));
        Prom(Addr+count, ByteSwap(data_word));
        Innovative::uSleep(100);
        }
    Prom(Addr,  0xD0);      // Write Buffer to Flash

    Innovative::uSleep(100);
    //  Leave current mode
    ResetToNormal(Addr);

}

//---------------------------------------------------------------------------
//  TypeX_FpaChipImpl::Unlock() -- unlock entire chip
//---------------------------------------------------------------------------

void TypeX_FpaChipImpl::Unlock()
{
    bool OriginalGoldenImageState = Owner->GoldenImageActive();
    //
    //  Unlock non golden area
    Owner->GoldenImageActive(false);
            {
            std::ostringstream msg;
            msg << "GoldenImageActive=false reads " << Owner->GoldenImageActive() ;

            ProcessStatusEvent e(msg.str());
            Owner->OnLoadError.Execute(e);
            }
    for (unsigned int i = StartSector(); i < Sectors(); ++i)
        {
        if (!UnlockSector(i))
            {
            std::ostringstream msg;
            msg << "Unable to unlock sector " << i ;

            ProcessStatusEvent e(msg.str());
            Owner->OnLoadError.Execute(e);

            goto leave_unlock;
            }
        }
    //
    //  Unlock golden area
    Owner->GoldenImageActive(true);
            {
            std::ostringstream msg;
            msg << "GoldenImageActive=true reads " << Owner->GoldenImageActive() ;

            ProcessStatusEvent e(msg.str());
            Owner->OnLoadError.Execute(e);
            }
    for (unsigned int i = GoldenImageStartSector(); i < TotalSectors(); ++i)
        {
        if (!UnlockSector(i))
            {
            std::ostringstream msg;
            msg << "Unable to unlock sector " << i ;

            ProcessStatusEvent e(msg.str());
            Owner->OnLoadError.Execute(e);

            goto leave_unlock;
            }
        }

leave_unlock:
    //
    //  Restore original Golden State
    Owner->GoldenImageActive(OriginalGoldenImageState);
}

//------------------------------------------------------------------------------
//  TypeX_FpaChipImpl::UnlockSector() - Unlock a sector
//------------------------------------------------------------------------------

bool TypeX_FpaChipImpl::UnlockSector(unsigned int sector)
{
    int Addr = SectorBaseAddress(sector);
	Prom(Addr, 0x60);
	Prom(Addr, 0xD0);    // unlock
        {
        std::stringstream msg;
        msg << "Unlock Sector Base Address: 0x" << std::hex << Addr;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }

    bool rtn = true;
    int chk = ReadUnlockStatus(Addr); // read unlock status from Signature Info
	if ((chk) != 0x00)
        {
        rtn = false;
            {
            std::stringstream msg;
            msg << "   Check Readback: 0x" << std::hex << chk;
            ProcessStatusEvent e(msg.str());
            Owner->OnLoadError.Execute(e);
            }
        }

    //  Leave current mode
    ResetToNormal(Addr);
    return rtn;
}

//------------------------------------------------------------------------------
//  TypeX_FpaChipImpl::ReadUnlockStatus() - Unlock a sector
//------------------------------------------------------------------------------

int TypeX_FpaChipImpl::ReadUnlockStatus(unsigned int Addr)
{
	Prom(Addr, 0x90);
    return Prom(Addr+2); // read unlock status from Signature Info
}

//==============================================================================
//  CLASS X5FpaPersonality -- X5 FPA Flash Personality Interface
//==============================================================================
//------------------------------------------------------------------------------
//  destructor for class X5FpaPersonality
//------------------------------------------------------------------------------

X5FpaPersonality::~X5FpaPersonality()
{
    if (Chip)
        {
        delete Chip;
        Chip = 0;
        }
}

//------------------------------------------------------------------------------
//  X5FpaPersonality::Initialize() - Initializes flash for programming
//------------------------------------------------------------------------------

bool  X5FpaPersonality::Initialize()
{
    //
    //  Detect CPLD Type - set our operation mode
    ReadCpldStatus();
    switch (FCpldType)
        {
        case 0x0F:    // X Type
            RomMode = romXtype;
            break;
        case 0x7F:    // J Type
            RomMode = romJtype;
            break;
        case 0x00:    // N Type
        default:
            RomMode = romNtype;
            break;
        }
    //
    //  Construct Chip Impl for the CPLD
    CreateChipImpl();

    //  Display CPLD Type, Rev in Log
    {
    std::stringstream msg;
    msg << "Cpld Type: 0x" << std::hex << FCpldType
        << " Cpld Rev: 0x"  << FCpldRevision << " Mode:" << RomMode;
    ProcessStatusEvent e(msg.str());
    OnLoadError.Execute(e);
    }

    FAddress = -1;
    FAddressBits = 22;
    // Store device size, in bytes
    FDeviceSize = 1 << FAddressBits;

    bool rtn = Chip->Initialize();
    //
    //  Copy Info from Chip Init
    FDeviceId = Chip->DeviceId();
    FProtected = Chip->IsProtected();
    
    FSectorSize = 0x10000;

    return rtn;
}

//------------------------------------------------------------------------------
//  X5FpaPersonality::OnFlashWrite_Start() -- Start Flash write
//------------------------------------------------------------------------------

void  X5FpaPersonality::OnFlashWrite_Start(int addr, int data)
{
    Chip->OnFlashWrite_Start(addr, data);
}

//------------------------------------------------------------------------------
//  X5FpaPersonality::EraseSector() - Erase specified flash sector contents
//------------------------------------------------------------------------------

bool X5FpaPersonality::EraseSector(int sector)
{
    Chip->OnEraseSector_Start(sector);

    int Addr = SectorBaseAddress(sector);

    Rtc.uS(5000);  // 5 ms delay before polling

    const double Timeout = 10.0;   // in seconds
    double Then = Rtc.Now() + Timeout;

	while ((Prom(Addr) & 0x80) != 0x80)
        {
        if (Rtc.Now() > Then)
            {
            ProcessStatusEvent e("Timeout erasing sector");
            OnLoadError.Execute(e);
            return false;
            }
        }

    return true;
}

//------------------------------------------------------------------------------
//  X5FpaPersonality::Program() -- Write buffer to flash
//------------------------------------------------------------------------------

bool  X5FpaPersonality::Program(FpgaImage & Image)
{
    switch (RomMode)
        {
//        case romXtype:    // X Type
        case romJtype:    // J Type
            return SingleCellProgram(Image);

        case romXtype:    // X Type
        case romNtype:    // N Type
        default:
            return BlockProgram(Image);

        }
}

//------------------------------------------------------------------------------
//  X5FpaPersonality::SingleCellProgram() -- Write buffer to flash 1 cell @ time
//------------------------------------------------------------------------------

bool  X5FpaPersonality::SingleCellProgram(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    Chip->OnSingleCellPgm_Start();

    FpgaImageDatagram Image_DG(Image);

    // Program device
    for (unsigned int i = 0; i < Image_DG.SizeInShorts(); ++i)
        {
        unsigned short data_word = static_cast<unsigned short>(Image_DG.AsShort(i));
        if (!Flash(i, Chip->ByteOrder(data_word)))
            {
            std::ostringstream msg;
            msg << "Unable to program byte " << i;

            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInShorts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    Chip->OnSingleCellPgm_End();

    ProcessProgressEvent e(100);
    OnLoadProgress.Execute(e);

    return true;
}

//------------------------------------------------------------------------------
//  X5FpaPersonality::BlockProgram() -- Write buffer to flash by blocks
//------------------------------------------------------------------------------

bool  X5FpaPersonality::BlockProgram(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;
    const int BlockWriteSize = 32;

    FpgaImageDatagram Image_DG(Image);

    // Program device
    for (unsigned int i = 0; i < Image_DG.SizeInShorts(); i+=BlockWriteSize)
        {
        Chip->OnWriteBufferProgram(Image_DG, i, BlockWriteSize);

        const double Timeout = 2.000;   // in sec
        double Then = Rtc.Now() + Timeout;

        //  Poll the final write for completion of all program writes
        unsigned short check_word = static_cast<unsigned short>(Image_DG.AsShort(i+BlockWriteSize-1));
        check_word = Chip->ByteOrder(check_word);

	    while (Prom(i+BlockWriteSize-1) != check_word)
            {
            if (Rtc.Now() < Then)
                continue;

            std::stringstream msg;
            msg << "Prom wrote 0x" << std::hex << Prom(i+BlockWriteSize-1)
            << " instead of 0x" << check_word
            << " at address 0x" << i+BlockWriteSize-1;

            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInShorts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent e(100);
    OnLoadProgress.Execute(e);

    return true;
}

//------------------------------------------------------------------------------
//  X5FpaPersonality::Verify() -- Verify buffer against flash contents
//------------------------------------------------------------------------------

bool X5FpaPersonality::Verify(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    FpgaImageDatagram Image_DG(Image);
    // Program device
    for (unsigned int i = 0; i < Image_DG.SizeInShorts(); ++i)
        {
        unsigned int dst = Flash(i);
        unsigned int src = static_cast<unsigned short>(Image_DG.AsShort(i));
        if (Chip->ByteOrder(dst) != src)
            {
            std::ostringstream msg;

            msg << "Prom read 0x" << std::hex << dst
                << " instead of 0x" << src
                << " at address 0x" << i;
            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInShorts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent p(100);
    OnLoadProgress.Execute(p);

    return true;
}

//---------------------------------------------------------------------------
//  X5FpaPersonality::Read() -- Fill buffer from flash
//---------------------------------------------------------------------------

bool X5FpaPersonality::Read(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    FpgaImageDatagram Image_DG(Image);
    // Read device
    for (unsigned int i = 0; i < Image_DG.SizeInShorts(); ++i)
        {
        Image_DG.AsShort(i, Chip->ByteOrder(Flash(i)));

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInShorts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent p(100);
    OnLoadProgress.Execute(p);
    return true;
}


//------------------------------------------------------------------------------
// X5FpaPersonality::SectorBaseAddress() -
//------------------------------------------------------------------------------

unsigned int  X5FpaPersonality::SectorBaseAddress(unsigned int sector)
{
    return Chip->SectorBaseAddress(sector);
}

//------------------------------------------------------------------------------
// X5FpaPersonality::GoldenImageActive() - set which ROM image we are using
//------------------------------------------------------------------------------

bool  X5FpaPersonality::GoldenImageActive()
{
    //ReadCpldStatus();
    return FGoldenImageActive;
}

//------------------------------------------------------------------------------
// X5FpaPersonality::GoldenImageActive() - set which ROM image we are using
//------------------------------------------------------------------------------

void  X5FpaPersonality::GoldenImageActive(bool state)
{
    // If we are already in the state, do nothing
    if (FGoldenImageActive==state)
        return;

    FGoldenImageActive = state;
    //  Toggle the Golden/normal image mode
    Mode(SwapImage);
    Strobe();

    // Update status
    ReadCpldStatus();
}

//------------------------------------------------------------------------------
// X5FpaPersonality::GoldenImageBooted() - Read state of boot jumper
//------------------------------------------------------------------------------

bool  X5FpaPersonality::GoldenImageBooted()
{
    //ReadCpldStatus();
    return FGoldenImageBooted;
}

//------------------------------------------------------------------------------
// X5FpaPersonality::CpldRevision() - Read CPLD logic version
//------------------------------------------------------------------------------

int  X5FpaPersonality::CpldRevision()
{
    //ReadCpldStatus();
    return FCpldRevision;
}

//------------------------------------------------------------------------------
// X5FpaPersonality::CpldType() - Read CPLD logic type
//------------------------------------------------------------------------------

int  X5FpaPersonality::CpldType()
{
    //ReadCpldStatus();
    return FCpldType;
}

//------------------------------------------------------------------------------
// X5FpaPersonality::ReadCpldStatus() - Read CPLD Status Word
//------------------------------------------------------------------------------

void  X5FpaPersonality::ReadCpldStatus()
{
    //  Toggle the Golden/normal image mode
    Mode(ReadStatusReg);
    Strobe();
    //
    //  Read in the data & Parse it
    int status = ShiftIn();
    MemoryBitManipulator StatusWord(status);

    FGoldenImageBooted = StatusWord.Bit(1);
    FGoldenImageActive = StatusWord.Bit(0);
    FCpldRevision      = StatusWord.Field(4, 4);
    FCpldType          = StatusWord.Field(8, 7);
        {
        std::stringstream msg;
        msg << "Raw Status:" << std::hex << status;
        ProcessStatusEvent e(msg.str());
        OnLoadError.Execute(e);
        }
}

//------------------------------------------------------------------------------
// X5FpaPersonality::CreateChipImpl() - Read CPLD Status Word
//------------------------------------------------------------------------------

void  X5FpaPersonality::CreateChipImpl()
{
    switch (RomMode)
        {
        case romXtype:    // X Type
            Chip = new TypeX_FpaChipImpl(this);
            break;
        case romJtype:    // J Type
            Chip = new TypeJ_FpaChipImpl(this);
            break;
        case romNtype:    // N Type
        default:
            Chip = new TypeN_FpaChipImpl(this);
            break;
        }
}

//------------------------------------------------------------------------------
// X5FpaPersonality::StartSector() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  X5FpaPersonality::StartSector() const
{
    return Chip->StartSector();
}

//------------------------------------------------------------------------------
// X5FpaPersonality::Sectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  X5FpaPersonality::Sectors() const
{
    return Chip->Sectors();
}

//------------------------------------------------------------------------------
// X5FpaPersonality::GoldenImageStartSector() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  X5FpaPersonality::GoldenImageStartSector() const
{
    return Chip->GoldenImageStartSector();
}

//------------------------------------------------------------------------------
// X5FpaPersonality::GoldenImageSectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  X5FpaPersonality::GoldenImageSectors() const
{
    return Chip->GoldenImageSectors();
}

//==============================================================================
//  CLASS TypeM26EW_ChipImpl -- X6 FPA Flash Sub-Personality Implementation
//==============================================================================
//  Xilinx Roms
//------------------------------------------------------------------------------
//  declaration of class TypeM26EW_ChipImpl
//------------------------------------------------------------------------------

class TypeM26EW_ChipImpl : public IX6FpaChip
{
public:
    TypeM26EW_ChipImpl(IX6FpaCpldPersonality * owner)
        :  IX6FpaChip(owner)
        {}
    virtual ~TypeM26EW_ChipImpl()
        {}

    virtual bool Initialize();
    virtual void  EnterCFIQueryMode();
    virtual void  LeaveCFIQueryMode();
    void          EnterAutoSelectMode();
    void          LeaveAutoSelectMode();

    virtual unsigned int SectorBaseAddress(unsigned int sector);

    virtual  unsigned int       ByteOrder(unsigned int data);
    virtual  unsigned int       StartSector() const ;
    virtual  unsigned int       Sectors() const ;
    virtual  unsigned int       GoldenImageStartSector() const;
    virtual  unsigned int       GoldenImageSectors() const;

    virtual void        OnFlashWrite_Start(int addr, int data);
    virtual void        OnSingleCellPgm_Start();
    virtual void        OnSingleCellPgm_End();
    virtual void        OnEraseSector_Start(unsigned int sector);
    virtual void        OnWriteBufferProgram(FpgaImageDatagram & ImageDg, unsigned int Addr, unsigned int wc);

protected:
    bool  FUnlockBypass;
    bool  FUnlockBypassCapable;

    void  CommonInit_QTest();
    void  CommonInit_DevId();
    void  CommonInit_Protected();
};

//------------------------------------------------------------------------------
//  TypeM26EW_ChipImpl::Initialize() - Initializes flash for programming
//------------------------------------------------------------------------------

bool  TypeM26EW_ChipImpl::Initialize()
{
    //  Display CPLD Type, Rev in Log
    {
    ProcessStatusEvent e("M26EW Type ROM Detected");
    Owner->OnLoadError.Execute(e);
    }

    CommonInit_QTest();

    CommonInit_DevId();
    //  Get ROM size from DevID
    switch (DeviceId()&0xFF)
        {
        case 0x22:  FTotalSectors = 256;  break;
        default:
        case 0x23:  FTotalSectors = 512;  break;
        case 0x28:  FTotalSectors = 1024;  break;
        case 0x48:  FTotalSectors = 2048;  break;
        }

    CommonInit_Protected();

    //  ....On this Part, we are Unlock Bypass Capable
    FUnlockBypassCapable = true;
        {
        std::stringstream msg;
        msg << "Unlock Bypass Capable Bit: 0x" << std::hex << FUnlockBypassCapable;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }

    FUnlockBypass = false;
    return true;
}

//------------------------------------------------------------------------------
//  TypeM26EW_ChipImpl::EnterCFIQueryMode() -- Enter CFI Mode
//------------------------------------------------------------------------------

void  TypeM26EW_ChipImpl::EnterCFIQueryMode()
{
    Prom(0x55, 0x98);          // enter CFI query mode
}

//------------------------------------------------------------------------------
//  TypeM26EW_ChipImpl::LeaveCFIQueryMode() -- Leave CFI Mode
//------------------------------------------------------------------------------

void  TypeM26EW_ChipImpl::LeaveCFIQueryMode()
{
    Prom(0x0, 0xf0);           // Reset (exit CFI query mode)
}

//------------------------------------------------------------------------------
//  TypeM26EW_ChipImpl::EnterAutoSelectMode() -- Enter AutoSelect Mode
//------------------------------------------------------------------------------

void  TypeM26EW_ChipImpl::EnterAutoSelectMode()
{
    Prom(0x555, 0xaa);
    Prom(0x2aa, 0x55);
    Prom(0x555, 0x90);            // Autoselect
}

//------------------------------------------------------------------------------
//  TypeM26EW_ChipImpl::LeaveAutoSelectMode() -- Leave CFI Mode
//------------------------------------------------------------------------------

void  TypeM26EW_ChipImpl::LeaveAutoSelectMode()
{
    Prom(0x0, 0xf0);           // Reset (leave AutoSelect mode)
}

//------------------------------------------------------------------------------
//  TypeM26EW_ChipImpl::CommonInit_QTest() - Search for Q to test for valid chip
//------------------------------------------------------------------------------

void  TypeM26EW_ChipImpl::CommonInit_QTest()
{
    //
    //  --- Check if ROM is present by looking for Q ---
    EnterCFIQueryMode();      // Start CFI Query Mode

    int query;
    query = Prom(0x10);  // should read "Q" (0x51) in Hi and Lo

        {
        std::stringstream msg;
        msg << "Prom Query: 0x" << std::hex << query << " = [0x00510051]";
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }

    LeaveCFIQueryMode();      // end CFI Query Mode
}

//------------------------------------------------------------------------------
//  TypeM26EW_ChipImpl::CommonInit_DevId() - Read device ID
//------------------------------------------------------------------------------

void  TypeM26EW_ChipImpl::CommonInit_DevId()
{
    EnterAutoSelectMode();

    unsigned int hi = (Prom(0x01) & 0xffff);
    unsigned int lo = (Prom(0x0E) & 0xffff);   // need 2nd word to get part ID
    FDeviceId = (hi<<16) | lo;
        {
        std::stringstream msg;
        msg << "Prom DevId: 0x" << std::hex << FDeviceId;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }

    LeaveAutoSelectMode();
}
//------------------------------------------------------------------------------
//  TypeM26EW_ChipImpl::CommonInit_Protected() - Read proctected bit
//------------------------------------------------------------------------------

void  TypeM26EW_ChipImpl::CommonInit_Protected()
{
    EnterAutoSelectMode();

    FProtected = Prom(2) & 0x01;
        {
        std::stringstream msg;
        msg << "Protected Bit: 0x" << std::hex << FProtected;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }

    LeaveAutoSelectMode();
}

//------------------------------------------------------------------------------
//  TypeM26EW_ChipImpl::OnFlashWrite_Start() -- Start Flash write
//------------------------------------------------------------------------------

void  TypeM26EW_ChipImpl::OnFlashWrite_Start(int addr, int data)
{
    if (!FUnlockBypass) // if disabled, we need this sequence
        {
        Prom(0x555, 0xaa);
        Prom(0x2aa, 0x55);
        }

    Prom(0x555, 0xa0);  // address is don't care
	Prom32(addr, data);     // do NOT twin program data
}

//------------------------------------------------------------------------------
//  TypeM26EW_ChipImpl::OnSingleCellPgm_Start() -- Start Cell Program
//------------------------------------------------------------------------------

void  TypeM26EW_ChipImpl::OnSingleCellPgm_Start()
{
    // enable Unlock Bypass to save 2 Prom transactions per word written
    if (FUnlockBypassCapable)
        {
        Prom(0x555, 0xaa);
        Prom(0x2aa, 0x55);
        Prom(0x555, 0x20);
        FUnlockBypass = true;
        }
}

//------------------------------------------------------------------------------
//  TypeM26EW_ChipImpl::OnSingleCellPgm_End() -- End Cell Program
//------------------------------------------------------------------------------

void  TypeM26EW_ChipImpl::OnSingleCellPgm_End()
{
    if (FUnlockBypassCapable)
        {
        Prom(0x555, 0x90);      // disable Unlock Bypass mode
        Prom(0x2aa, 0x00);      // addresses are don't care
        FUnlockBypass = false;
        }
}

//------------------------------------------------------------------------------
//  TypeM26EW_ChipImpl::OnEraseSector_Start() -- Erase sector
//------------------------------------------------------------------------------

void  TypeM26EW_ChipImpl::OnEraseSector_Start(unsigned int sector)
{
	Prom(0x555, 0xaa);
	Prom(0x2aa, 0x55);
	Prom(0x555, 0x80);

	Prom(0x555, 0xaa);
	Prom(0x2aa, 0x55);

    int Addr = SectorBaseAddress(sector);
	Prom(Addr, 0x30);

        {
        std::stringstream msg;
        msg << "Sector Base Address: 0x" << std::hex << Addr;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }
}

//------------------------------------------------------------------------------
//  TypeM26EW_ChipImpl::OnWriteBufferProgram() -- Issue write buffer command
//------------------------------------------------------------------------------

void TypeM26EW_ChipImpl::OnWriteBufferProgram(FpgaImageDatagram & ImageDG, unsigned int Addr,
                                                unsigned int wc)
{
    Prom(0x555, 0xAA);      // Unlock
    Prom(0x2AA, 0x55);      // Unlock
    Prom(Addr,  0x25);      // Write Buffer Load
    Prom(Addr,  wc-1);      // Word Count
    for (unsigned int count=0; count<wc; count++)
        {
        unsigned short data_word = static_cast<unsigned short>(ImageDG.AsShort(Addr+count));
        Prom(Addr+count, ByteOrder(data_word));
        }
    Prom(Addr,  0x29);      // Write Buffer to Flash

}

//------------------------------------------------------------------------------
// TypeM26EW_ChipImpl::SectorBaseAddress() -
//------------------------------------------------------------------------------

unsigned int  TypeM26EW_ChipImpl::SectorBaseAddress(unsigned int sector)
{
    //
    //  M26EW Sector Address Calculation
    return sector*0x10000;
}

//------------------------------------------------------------------------------
//  TypeM26EW_ChipImpl::ByteOrder() -- Reorder Bytes between image and chip
//------------------------------------------------------------------------------

unsigned int  TypeM26EW_ChipImpl::ByteOrder(unsigned int data)
{
    // this is a 32 bit image. Assume Hi and Lo short are correct, swap
    //   the rest
    unsigned short * asShort = reinterpret_cast<unsigned short *>(&data);
    ByteSwap(asShort[0]);
    ByteSwap(asShort[1]);
    return data;
}

//------------------------------------------------------------------------------
// TypeM26EW_ChipImpl::StartSector() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeM26EW_ChipImpl::StartSector() const
{
    return 0;
}

//------------------------------------------------------------------------------
// TypeM26EW_ChipImpl::Sectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeM26EW_ChipImpl::Sectors() const
{
    //  Half for image, half for golden image
    return FTotalSectors>>1;
}

//------------------------------------------------------------------------------
// TypeM26EW_ChipImpl::GoldenImageStartSector() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeM26EW_ChipImpl::GoldenImageStartSector() const
{
    return StartSector() + Sectors();
}

//------------------------------------------------------------------------------
// TypeM26EW_ChipImpl::GoldenImageSectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeM26EW_ChipImpl::GoldenImageSectors() const
{
    return Sectors();   // Same number of sectors as other half
}

//==============================================================================
//  CLASS IX6FpaCpldPersonality -- Interface to FPA specific operations
//==============================================================================
//------------------------------------------------------------------------------
//  destructor for class IX6FpaCpldPersonality
//------------------------------------------------------------------------------

IX6FpaCpldPersonality::~IX6FpaCpldPersonality()
{
    UncreateChipImpl();
}

//---------------------------------------------------------------------------
//  IX6FpaCpldPersonality::Initialize() -- Initialize Personality
//---------------------------------------------------------------------------

bool  IX6FpaCpldPersonality::Initialize()
{
    //  Make sure CPLD status is correct
    Cpld.CPLD_UpdateStatus();
    //
    //  Detect CPLD Type - set our operation mode
    switch (Id())
        {
        case 0x00:    // 2 x M29EW NOR type flash from Numonyx
        default:
            RomMode = romM29EWtype;
            break;
        }
    //
    //  Construct Chip Impl for the CPLD
    CreateChipImpl();

    //  Display CPLD Type, Rev in Log
    {
    std::stringstream msg;
    msg << "Cpld Type: 0x" << std::hex << CpldType()
        << " Cpld Rev: 0x"  << CpldRevision() << " Mode:" << RomMode;
    ProcessStatusEvent e(msg.str());
    OnLoadError.Execute(e);
    }

    bool rtn = Chip->Initialize();
    //
    //  Copy Info from Chip Init
    FDeviceId = Chip->DeviceId();
    FProtected = Chip->IsProtected();

    FSectorSize = 0x10000;
    FDeviceSize = Chip->Sectors()*FSectorSize;

    return rtn;
}

//------------------------------------------------------------------------------
//  IX6FpaCpldPersonality::OnFlashWrite_Start() -- Start Flash write
//------------------------------------------------------------------------------

void  IX6FpaCpldPersonality::OnFlashWrite_Start(int addr, int data)
{
    Chip->OnFlashWrite_Start(addr, data);
}

//------------------------------------------------------------------------------
//  IX6FpaCpldPersonality::EraseSector() - Erase specified flash sector contents
//------------------------------------------------------------------------------

bool IX6FpaCpldPersonality::EraseSector(int sector)
{
    Chip->OnEraseSector_Start(sector);

    int Addr = SectorBaseAddress(sector);

    Rtc.uS(5000);  // 5 ms delay before polling

    const double Timeout = 10.0;   // in seconds
    double Then = Rtc.Now() + Timeout;

	while (Prom(Addr) != ((unsigned int)(0xFFFFFFFF)))     // this waits for both ROMs to erase and read data
        {
        if (Rtc.Now() > Then)
            {
            ProcessStatusEvent e("Timeout erasing sector");
            OnLoadError.Execute(e);
            return false;
            }
        }

    return true;
}

//------------------------------------------------------------------------------
//  IX6FpaCpldPersonality::Program() -- Write buffer to flash
//------------------------------------------------------------------------------

bool  IX6FpaCpldPersonality::Program(FpgaImage & Image)
{
    switch (RomMode)
        {
        case romM29EWtype:    // M29EW type
            return SingleCellProgram(Image);    // test single cell case

        default:
            return BlockProgram(Image);

        }
}

//------------------------------------------------------------------------------
//  IX6FpaCpldPersonality::SingleCellProgram() -- Write buffer to flash 1 cell @ time
//------------------------------------------------------------------------------

bool  IX6FpaCpldPersonality::SingleCellProgram(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    Chip->OnSingleCellPgm_Start();

    FpgaImageDatagram Image_DG(Image);

    // Program device
    for (unsigned int i = 0; i < Image_DG.SizeInInts(); ++i)
        {
        unsigned int data_word = static_cast<unsigned int>(Image_DG.AsInt(i));
        if (!Flash(i, Chip->ByteOrder(data_word)))
            {
            std::ostringstream msg;
            msg << "Unable to program word 0x" << std::hex << i
                << "[" << data_word << "]";

            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInInts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    Chip->OnSingleCellPgm_End();

    ProcessProgressEvent e(100);
    OnLoadProgress.Execute(e);

    return true;
}

//------------------------------------------------------------------------------
//  IX6FpaCpldPersonality::BlockProgram() -- Write buffer to flash by blocks
//------------------------------------------------------------------------------

bool  IX6FpaCpldPersonality::BlockProgram(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;
    const int BlockWriteSize = 512;

    FpgaImageDatagram Image_DG(Image);

    // Program device
    for (unsigned int i = 0; i < Image_DG.SizeInInts(); i+=BlockWriteSize)
        {
        Chip->OnWriteBufferProgram(Image_DG, i, BlockWriteSize);

        const double Timeout = 2.000;   // in sec
        double Then = Rtc.Now() + Timeout;

        //  Poll the final write for completion of all program writes
        unsigned int check_word = static_cast<unsigned int>(Image_DG.AsInt(i+BlockWriteSize-1));
        check_word = Chip->ByteOrder(check_word);

	    while (Prom32(i+BlockWriteSize-1) != check_word)
            {
            if (Rtc.Now() < Then)
                continue;

            std::stringstream msg;
            msg << "Prom wrote 0x" << std::hex << Prom32(i+BlockWriteSize-1)
            << " instead of 0x" << check_word
            << " at address 0x" << i+BlockWriteSize-1;

            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInInts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent e(100);
    OnLoadProgress.Execute(e);

    return true;
}

//------------------------------------------------------------------------------
//  IX6FpaCpldPersonality::Verify() -- Verify buffer against flash contents
//------------------------------------------------------------------------------

bool IX6FpaCpldPersonality::Verify(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    FpgaImageDatagram Image_DG(Image);
    // Program device
    for (unsigned int i = 0; i < Image_DG.SizeInInts(); ++i)
        {
        unsigned int dst = Flash(i);
        unsigned int src = static_cast<unsigned int>(Image_DG.AsInt(i));
        if (Chip->ByteOrder(dst) != src)
            {
            std::ostringstream msg;

            msg << "Prom read 0x" << std::hex << dst
                << " instead of 0x" << src
                << " at address 0x" << i;
            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInInts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent p(100);
    OnLoadProgress.Execute(p);

    return true;
}

//---------------------------------------------------------------------------
//  IX6FpaCpldPersonality::Read() -- Fill buffer from flash
//---------------------------------------------------------------------------

bool IX6FpaCpldPersonality::Read(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    FpgaImageDatagram Image_DG(Image);
    // Read device
    for (unsigned int i = 0; i < Image_DG.SizeInInts(); ++i)
        {
        Image_DG.AsInt(i, Chip->ByteOrder(Flash(i)));

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInInts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent p(100);
    OnLoadProgress.Execute(p);
    return true;
}


//------------------------------------------------------------------------------
// IX6FpaCpldPersonality::SectorBaseAddress() -
//------------------------------------------------------------------------------

unsigned int  IX6FpaCpldPersonality::SectorBaseAddress(unsigned int sector)
{
    return Chip->SectorBaseAddress(sector);
}

//------------------------------------------------------------------------------
// IX6FpaCpldPersonality::CreateChipImpl() - Read CPLD Status Word
//------------------------------------------------------------------------------

void  IX6FpaCpldPersonality::CreateChipImpl()
{
    UncreateChipImpl();
    switch (RomMode)
        {
        case romM29EWtype:    // M29EW Type
        default:
            Chip = new TypeM26EW_ChipImpl(this);
            break;
        }
}

//------------------------------------------------------------------------------
//  IX6FpaCpldPersonality::UncreateChipImpl
//------------------------------------------------------------------------------

void  IX6FpaCpldPersonality::UncreateChipImpl()
{
    if (Chip)
        {
        delete Chip;
        Chip = 0;
        }
}

//---------------------------------------------------------------------------
//  IX6FpaCpldPersonality::Prom(a,d) -- write to prom device
//---------------------------------------------------------------------------

void  IX6FpaCpldPersonality::Prom(unsigned int addr, unsigned int value)
{
    // 'twin' the data to go out to the two devices on bus
    unsigned int twinned = (value<<16) | (value&0xFFFF);
    Cpld.CPLD_Write(addr, twinned);
}

//---------------------------------------------------------------------------
//  IX6FpaCpldPersonality::Prom(a) -- read from prom device
//---------------------------------------------------------------------------

unsigned int  IX6FpaCpldPersonality::Prom(unsigned int addr)
{
    return Cpld.CPLD_Read(addr);
}

//---------------------------------------------------------------------------
//  IX6FpaCpldPersonality::Prom32(a,d) -- write to prom device
//---------------------------------------------------------------------------

void  IX6FpaCpldPersonality::Prom32(unsigned int addr, unsigned int value)
{
    Cpld.CPLD_Write(addr, value);
}

//---------------------------------------------------------------------------
//  IX6FpaCpldPersonality::Prom32(a) -- read from prom device
//---------------------------------------------------------------------------

unsigned int  IX6FpaCpldPersonality::Prom32(unsigned int addr)
{
    return Cpld.CPLD_Read(addr);
}

//------------------------------------------------------------------------------
// IX6FpaCpldPersonality::StartSector() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  IX6FpaCpldPersonality::StartSector() const
{
    return Chip->StartSector();
}

//------------------------------------------------------------------------------
// IX6FpaCpldPersonality::Sectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  IX6FpaCpldPersonality::Sectors() const
{
    return Chip->Sectors();
}

//------------------------------------------------------------------------------
// IX6FpaCpldPersonality::GoldenImageStartSector() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  IX6FpaCpldPersonality::GoldenImageStartSector() const
{
    return Chip->GoldenImageStartSector();
}

//------------------------------------------------------------------------------
// IX6FpaCpldPersonality::GoldenImageSectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  IX6FpaCpldPersonality::GoldenImageSectors() const
{
    return Chip->GoldenImageSectors();
}

//-------------------------------------------------------------
// IX6FpaCpldPersonality::Flash() - Reads byte from specified address
//-------------------------------------------------------------

int IX6FpaCpldPersonality::Flash(int addr)
{
    if (GoldenImageActive())
        {
        addr += Chip->SectorBaseAddress(GoldenImageStartSector());
        }
    else
        {
        addr += Chip->SectorBaseAddress(StartSector());
        }

    return Prom32(addr);
}

//-------------------------------------------------------------
// IX6FpaCpldPersonality::Flash() - Programs byte at specified address
//-------------------------------------------------------------

bool IX6FpaCpldPersonality::Flash(int addr, int data)
{
    if (GoldenImageActive())
        {
        addr += Chip->SectorBaseAddress(GoldenImageStartSector());
        }
    else
        {
        addr += Chip->SectorBaseAddress(StartSector());
        }
    //
    //  Start the flash operation
    OnFlashWrite_Start(addr, data);

    const double Timeout = 0.500;   // in sec
    double Then = Rtc.Now() + Timeout;
	while (Prom32(addr) != static_cast<unsigned int>(data))
        {
        if (Rtc.Now() < Then)
            continue;

        std::ostringstream msg;
        msg << "Prom wrote 0x" << std::hex << Prom32(addr)
            << " instead of 0x" << data
            << " at address 0x" << addr;
        ProcessStatusEvent e(msg.str());
        OnLoadError.Execute(e);
        return false;
        }

    return true;
}

//-------------------------------------------------------------
// IX6FpaCpldPersonality::Authenticate() - Verify board integrity
//-------------------------------------------------------------

bool IX6FpaCpldPersonality::Authenticate()
{
    return true;
}

//==============================================================================
//  CLASS IVpxFpaChip -- X6 Flash Sub-Personality Interface
//==============================================================================
//------------------------------------------------------------------------------
//  IVpxFpaChip::ByteOrder() -- Reorder Bytes between image and chip
//------------------------------------------------------------------------------

unsigned int  IVpxFpaChip::ByteOrder(unsigned int data)
{
    // this is a 32 bit image. Assume Hi and Lo short are correct, swap
    //   the rest
    unsigned short * asShort = reinterpret_cast<unsigned short *>(&data);
    ByteSwap(asShort[0]);
    ByteSwap(asShort[1]);
    //  then bit reverse what is left
    return BitReverse(data);
}

//==============================================================================
//  CLASS VpxFpaPersonality -- Interface to FPA specific operations
//==============================================================================
//---------------------------------------------------------------------------
//  VpxFpaPersonality::Initialize() -- Initialize Personality
//---------------------------------------------------------------------------

bool  VpxFpaPersonality::Initialize()
{
    return Chip->Initialize();
}

/*
//------------------------------------------------------------------------------
//  VpxFpaPersonality::OnFlashWrite_Start() -- Start Flash write
//------------------------------------------------------------------------------

void  VpxFpaPersonality::OnFlashWrite_Start(int addr, int data)
{
    Chip->OnFlashWrite_Start(addr, data);
}
*/
//------------------------------------------------------------------------------
//  VpxFpaPersonality::EraseSector() - Erase specified flash sector contents
//------------------------------------------------------------------------------

bool VpxFpaPersonality::EraseSector(int sector)
{
    Chip->EraseSector(sector);

    int Addr = Chip->SectorBaseAddress(sector);

    Rtc.uS(5000);  // 5 ms delay before polling

    const double Timeout = 10.0;   // in seconds
    double Then = Rtc.Now() + Timeout;

    while (Chip->Read32(Addr) != ((unsigned int)(0xFFFFFFFF)))     // this waits for both ROMs to erase and read data
        {
        if (Rtc.Now() > Then)
            {
            ProcessStatusEvent e("Timeout erasing sector");
            OnLoadError.Execute(e);
            return false;
            }
        }

    return true;
}

//------------------------------------------------------------------------------
//  VpxFpaPersonality::Program() -- Write buffer to flash
//------------------------------------------------------------------------------

bool  VpxFpaPersonality::Program(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    FpgaImageDatagram Image_DG(Image);

    // Program device
    for (unsigned int i = 0; i < Image_DG.SizeInInts(); ++i)
        {
        unsigned int data_word = static_cast<unsigned int>(Image_DG.AsInt(i));
        unsigned int addr = i<<2;
        if (GoldenImageActive())
            {
            addr += Chip->SectorBaseAddress(GoldenImageStartSector());
            }
        else
            {
            addr += Chip->SectorBaseAddress(StartSector());
            }
        //  Program Word
        unsigned int reorder_word = Chip->ByteOrder(data_word);
        Chip->Write32(addr, reorder_word);
        //  Check Word
        unsigned int rb;
        if ((rb=Chip->Read32(addr)) != reorder_word)
            {
            std::ostringstream msg;
            msg << "Unable to program word 0x" << std::hex << i
                << "[" << reorder_word << "] read back [" << rb << "]";

            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInInts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent e(100);
    OnLoadProgress.Execute(e);

    return true;
}

//------------------------------------------------------------------------------
//  VpxFpaPersonality::Verify() -- Verify buffer against flash contents
//------------------------------------------------------------------------------

bool VpxFpaPersonality::Verify(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    FpgaImageDatagram Image_DG(Image);
    // Program device
    for (unsigned int i = 0; i < Image_DG.SizeInInts(); ++i)
        {
        unsigned int addr = i<<2;
        if (GoldenImageActive())
            {
            addr += Chip->SectorBaseAddress(GoldenImageStartSector());
            }
        else
            {
            addr += Chip->SectorBaseAddress(StartSector());
            }

        unsigned int dst = Chip->Read32(addr);
        unsigned int src = static_cast<unsigned int>(Image_DG.AsInt(i));
        if (Chip->ByteOrder(dst) != src)
            {
            std::ostringstream msg;

            msg << "Prom read 0x" << std::hex << dst
                << " instead of 0x" << src
                << " at address 0x" << i;
            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInInts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent p(100);
    OnLoadProgress.Execute(p);

    return true;
}

//---------------------------------------------------------------------------
//  VpxFpaPersonality::Read() -- Fill buffer from flash
//---------------------------------------------------------------------------

bool VpxFpaPersonality::Read(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    FpgaImageDatagram Image_DG(Image);
    // Read device
    for (unsigned int i = 0; i < Image_DG.SizeInInts(); ++i)
        {
        unsigned int addr = i<<2;
        if (GoldenImageActive())
            {
            addr += Chip->SectorBaseAddress(GoldenImageStartSector());
            }
        else
            {
            addr += Chip->SectorBaseAddress(StartSector());
            }
        Image_DG.AsInt(i, Chip->ByteOrder(Chip->Read32(addr)));

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInInts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent p(100);
    OnLoadProgress.Execute(p);
    return true;
}

/*
//-------------------------------------------------------------
// VpxFpaPersonality::Flash() - Reads byte from specified address
//-------------------------------------------------------------

int VpxFpaPersonality::Flash(int addr)
{
    if (GoldenImageActive())
        {
        addr += Chip->SectorBaseAddress(GoldenImageStartSector());
        }
    else
        {
        addr += Chip->SectorBaseAddress(StartSector());
        }

    return Chip->Prom32(addr);
}

//-------------------------------------------------------------
// VpxFpaPersonality::Flash() - Programs byte at specified address
//-------------------------------------------------------------

bool VpxFpaPersonality::Flash(int addr, int data)
{
    if (GoldenImageActive())
        {
        addr += Chip->SectorBaseAddress(GoldenImageStartSector());
        }
    else
        {
        addr += Chip->SectorBaseAddress(StartSector());
        }
    //
    //  Start the flash operation
    OnFlashWrite_Start(addr, data);

    const double Timeout = 0.500;   // in sec
    double Then = Rtc.Now() + Timeout;
    while (Prom32(addr) != static_cast<unsigned int>(data))
        {
        if (Rtc.Now() < Then)
            continue;

        std::ostringstream msg;
        msg << "Prom wrote 0x" << std::hex << Prom32(addr)
            << " instead of 0x" << data
            << " at address 0x" << addr;
        ProcessStatusEvent e(msg.str());
        OnLoadError.Execute(e);
        return false;
        }

    return true;
}
*/

//-------------------------------------------------------------
// VpxFpaPersonality::Authenticate() - Verify board integrity
//-------------------------------------------------------------

bool VpxFpaPersonality::Authenticate()
{
    return true;
}

//==============================================================================
//  CLASS TypeM29EW_8bit_ChipImpl -- X6 FPA Flash Sub-Personality Implementation
//==============================================================================
//  Xilinx Roms
//------------------------------------------------------------------------------
//  declaration of class TypeM29EW_8bit_ChipImpl
//------------------------------------------------------------------------------

class TypeM29EW_8bit_ChipImpl : public IVpxCopFpaChip
{
public:
    TypeM29EW_8bit_ChipImpl(IVpxCopFpaCpldPersonality * owner)
        :  IVpxCopFpaChip(owner)
        {}
    virtual ~TypeM29EW_8bit_ChipImpl()
        {}

    virtual bool Initialize();
    virtual void  EnterCFIQueryMode();
    virtual void  LeaveCFIQueryMode();
    void          EnterAutoSelectMode();
    void          LeaveAutoSelectMode();

    virtual unsigned int SectorBaseAddress(unsigned int sector);

    virtual  unsigned int       ByteOrder(unsigned int data);
    virtual  unsigned int       StartSector() const ;
    virtual  unsigned int       Sectors() const ;
    virtual  unsigned int       GoldenImageStartSector() const;
    virtual  unsigned int       GoldenImageSectors() const;

    virtual void        OnFlashWrite_Start(int addr, int data);
    virtual void        OnSingleCellPgm_Start();
    virtual void        OnSingleCellPgm_End();
    virtual void        OnEraseSector_Start(unsigned int sector);
    virtual void        OnWriteBufferProgram(FpgaImageDatagram & ImageDg, unsigned int Addr, unsigned int wc);
    virtual void        OnBufferProgramAbortAndReset();

protected:
    bool  FUnlockBypass;
    bool  FUnlockBypassCapable;

    void  CommonInit_QTest();
    void  CommonInit_DevId();
    void  CommonInit_Protected();
};

//------------------------------------------------------------------------------
//  TypeM29EW_8bit_ChipImpl::Initialize() - Initializes flash for programming
//------------------------------------------------------------------------------

bool  TypeM29EW_8bit_ChipImpl::Initialize()
{
    //  Display CPLD Type, Rev in Log
    {
    ProcessStatusEvent e("M29EW (8 bit) Type ROM Detected");
    Owner->OnLoadError.Execute(e);
    }

    CommonInit_QTest();

    CommonInit_DevId();
    //  Get ROM size from DevID
    switch (DeviceId()&0xFF)
        {
        case 0x22:  FTotalSectors = 256;  break;
        default:
        case 0x23:  FTotalSectors = 512;  break;
        case 0x28:  FTotalSectors = 1024;  break;
        case 0x48:  FTotalSectors = 2048;  break;
        }

    CommonInit_Protected();

    //  ....On this Part, we are Unlock Bypass Capable
    FUnlockBypassCapable = true;
        {
        std::stringstream msg;
        msg << "Unlock Bypass Capable Bit: 0x" << std::hex << FUnlockBypassCapable;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }

    FUnlockBypass = false;
    return true;
}

//------------------------------------------------------------------------------
//  TypeM29EW_8bit_ChipImpl::EnterCFIQueryMode() -- Enter CFI Mode
//------------------------------------------------------------------------------

void  TypeM29EW_8bit_ChipImpl::EnterCFIQueryMode()
{
    Prom(0xaa, 0x98);          // enter CFI query mode
}

//------------------------------------------------------------------------------
//  TypeM29EW_8bit_ChipImpl::LeaveCFIQueryMode() -- Leave CFI Mode
//------------------------------------------------------------------------------

void  TypeM29EW_8bit_ChipImpl::LeaveCFIQueryMode()
{
    Prom(0x0, 0xf0);           // Reset (exit CFI query mode)
}

//------------------------------------------------------------------------------
//  TypeM29EW_8bit_ChipImpl::EnterAutoSelectMode() -- Enter AutoSelect Mode
//------------------------------------------------------------------------------

void  TypeM29EW_8bit_ChipImpl::EnterAutoSelectMode()
{
    Prom(0xaaa, 0xaa);
    Prom(0x555, 0x55);
    Prom(0xaaa, 0x90);            // Autoselect
}

//------------------------------------------------------------------------------
//  TypeM29EW_8bit_ChipImpl::LeaveAutoSelectMode() -- Leave CFI Mode
//------------------------------------------------------------------------------

void  TypeM29EW_8bit_ChipImpl::LeaveAutoSelectMode()
{
    Prom(0x0, 0xf0);           // Reset (leave AutoSelect mode)
}

//------------------------------------------------------------------------------
//  TypeM29EW_8bit_ChipImpl::CommonInit_QTest() - Search for Q to test for valid chip
//------------------------------------------------------------------------------

void  TypeM29EW_8bit_ChipImpl::CommonInit_QTest()
{
    //
    //  --- Check if ROM is present by looking for Q ---
    EnterCFIQueryMode();      // Start CFI Query Mode

    int query;
    query = Prom(0x20);  // should read "Q" (0x51) in Hi and Lo

        {
        std::stringstream msg;
        msg << "Prom Query: 0x" << std::hex << query << " = [0x51]";
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }

    LeaveCFIQueryMode();      // end CFI Query Mode
}

//------------------------------------------------------------------------------
//  TypeM29EW_8bit_ChipImpl::CommonInit_DevId() - Read device ID
//------------------------------------------------------------------------------

void  TypeM29EW_8bit_ChipImpl::CommonInit_DevId()
{
    EnterAutoSelectMode();

    unsigned int hi = (Prom(0x02) & 0xff);
    unsigned int lo = (Prom(0x1c) & 0xff);   // need 2nd word to get part ID
    FDeviceId = (hi<<8) | lo;
        {
        std::stringstream msg;
        msg << "Prom DevId: 0x" << std::hex << FDeviceId;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }

    LeaveAutoSelectMode();
}
//------------------------------------------------------------------------------
//  TypeM29EW_8bit_ChipImpl::CommonInit_Protected() - Read proctected bit
//------------------------------------------------------------------------------

void  TypeM29EW_8bit_ChipImpl::CommonInit_Protected()
{
    EnterAutoSelectMode();

    FProtected = Prom(0x4) & 0x01;       // these not doubled?
        {
        std::stringstream msg;
        msg << "Protected Bit: 0x" << std::hex << FProtected;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }

    LeaveAutoSelectMode();
}

//------------------------------------------------------------------------------
//  TypeM29EW_8bit_ChipImpl::OnFlashWrite_Start() -- Start Flash write
//------------------------------------------------------------------------------

void  TypeM29EW_8bit_ChipImpl::OnFlashWrite_Start(int addr, int data)
{
    if (!FUnlockBypass) // if disabled, we need this sequence
        {
        Prom(0xaaa, 0xaa);
        Prom(0x555, 0x55);
        }

    Prom(0xaaa, 0xa0);  // address is don't care
    Prom(addr, data);
}

//------------------------------------------------------------------------------
//  TypeM29EW_8bit_ChipImpl::OnSingleCellPgm_Start() -- Start Cell Program
//------------------------------------------------------------------------------

void  TypeM29EW_8bit_ChipImpl::OnSingleCellPgm_Start()
{
    // enable Unlock Bypass to save 2 Prom transactions per word written
    if (FUnlockBypassCapable)
        {
        Prom(0xaaa, 0xaa);
        Prom(0x555, 0x55);
        Prom(0xaaa, 0x20);
        FUnlockBypass = true;
        }
}

//------------------------------------------------------------------------------
//  TypeM29EW_8bit_ChipImpl::OnSingleCellPgm_End() -- End Cell Program
//------------------------------------------------------------------------------

void  TypeM29EW_8bit_ChipImpl::OnSingleCellPgm_End()
{
    if (FUnlockBypassCapable)
        {
        Prom(0xaaa, 0x90);      // disable Unlock Bypass mode
        Prom(0x555, 0x00);      // addresses are don't care
        FUnlockBypass = false;
        }
}

//------------------------------------------------------------------------------
//  TypeM29EW_8bit_ChipImpl::OnEraseSector_Start() -- Erase sector
//------------------------------------------------------------------------------

void  TypeM29EW_8bit_ChipImpl::OnEraseSector_Start(unsigned int sector)
{
    Prom(0xaaa, 0xaa);
    Prom(0x555, 0x55);
    Prom(0xaaa, 0x80);

    Prom(0xaaa, 0xaa);
    Prom(0x555, 0x55);

    int Addr = SectorBaseAddress(sector);
    Prom(Addr, 0x30);

        {
        std::stringstream msg;
        msg << "Sector Base Address: 0x" << std::hex << Addr;
        ProcessStatusEvent e(msg.str());
        Owner->OnLoadError.Execute(e);
        }
}

//------------------------------------------------------------------------------
//  TypeM29EW_8bit_ChipImpl::OnWriteBufferProgram() -- Issue write buffer command
//------------------------------------------------------------------------------

void TypeM29EW_8bit_ChipImpl::OnWriteBufferProgram(FpgaImageDatagram & ImageDG, unsigned int Addr,
                                                unsigned int wc)
{
    Prom(0xaaa, 0xAA);      // Unlock
    Prom(0x555, 0x55);      // Unlock
    Prom(Addr,  0x25);      // Write Buffer Load
    Prom(Addr,  wc-1);      // Word Count
    for (unsigned int count=0; count<wc; count++)
        {
        unsigned int data_word = static_cast<unsigned short>(ImageDG.AsChar(Addr+count));
        Prom(Addr+count, ByteOrder(data_word));
        }
    Prom(Addr,  0x29);      // Write Buffer to Flash

}

//------------------------------------------------------------------------------
//  TypeM29EW_8bit_ChipImpl::OnBufferProgramAbortAndReset() -- Reset Prog Mode
//------------------------------------------------------------------------------

void TypeM29EW_8bit_ChipImpl::OnBufferProgramAbortAndReset()
{
    Prom(0xaaa, 0xAA);      // Unlock
    Prom(0x555, 0x55);      // Unlock
    Prom(0xaaa, 0xF0);      // Reset
}

//------------------------------------------------------------------------------
// TypeM29EW_8bit_ChipImpl::SectorBaseAddress() -
//------------------------------------------------------------------------------

unsigned int  TypeM29EW_8bit_ChipImpl::SectorBaseAddress(unsigned int sector)
{
    //
    //  M26EW Sector Address Calculation
    return sector*0x20000;
}

//------------------------------------------------------------------------------
//  TypeM29EW_8bit_ChipImpl::ByteOrder() -- Reorder Bytes between image and chip
//------------------------------------------------------------------------------

unsigned int  TypeM29EW_8bit_ChipImpl::ByteOrder(unsigned int data)
{
    return data;  //ByteSwap(data);
}

//------------------------------------------------------------------------------
// TypeM29EW_8bit_ChipImpl::StartSector() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeM29EW_8bit_ChipImpl::StartSector() const
{
    return 0;
}

//------------------------------------------------------------------------------
// TypeM29EW_8bit_ChipImpl::Sectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeM29EW_8bit_ChipImpl::Sectors() const
{
    //  Half for image, half for golden image
    return FTotalSectors>>1;
}

//------------------------------------------------------------------------------
// TypeM29EW_8bit_ChipImpl::GoldenImageStartSector() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeM29EW_8bit_ChipImpl::GoldenImageStartSector() const
{
    return StartSector() + Sectors();
}

//------------------------------------------------------------------------------
// TypeM29EW_8bit_ChipImpl::GoldenImageSectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  TypeM29EW_8bit_ChipImpl::GoldenImageSectors() const
{
    return Sectors();   // Same number of sectors as other half
}


//==============================================================================
//  CLASS IVpxCopFpaCpldPersonality -- Interface to FPA specific operations
//==============================================================================
//------------------------------------------------------------------------------
//  destructor for class IVpxCopFpaCpldPersonality
//------------------------------------------------------------------------------

IVpxCopFpaCpldPersonality::~IVpxCopFpaCpldPersonality()
{
    UncreateChipImpl();
}

//---------------------------------------------------------------------------
//  IVpxCopFpaCpldPersonality::Initialize() -- Initialize Personality
//---------------------------------------------------------------------------

bool  IVpxCopFpaCpldPersonality::Initialize()
{
    //  Make sure CPLD status is correct
    Cpld.CPLD_UpdateStatus();
    //
    //  Detect CPLD Type - set our operation mode
    switch (Id())
        {
        case 0x00:    // 2 x M29EW NOR type flash from Numonyx
        default:
            RomMode = romM29EWtype;
            break;
        }
    //
    //  Construct Chip Impl for the CPLD
    CreateChipImpl();

    //  Display CPLD Type, Rev in Log
    {
    std::stringstream msg;
    msg << "Cpld Type: 0x" << std::hex << CpldType()
        << " Cpld Rev: 0x"  << CpldRevision() << " Mode:" << RomMode;
    ProcessStatusEvent e(msg.str());
    OnLoadError.Execute(e);
    }

    bool rtn = Chip->Initialize();
    //
    //  Copy Info from Chip Init
    FDeviceId = Chip->DeviceId();
    FProtected = Chip->IsProtected();

    FSectorSize = 0x20000;
    FDeviceSize = Chip->Sectors()*FSectorSize;

    return rtn;
}

//------------------------------------------------------------------------------
//  IVpxCopFpaCpldPersonality::OnFlashWrite_Start() -- Start Flash write
//------------------------------------------------------------------------------

void  IVpxCopFpaCpldPersonality::OnFlashWrite_Start(int addr, int data)
{
    Chip->OnFlashWrite_Start(addr, data);
}

//------------------------------------------------------------------------------
//  IVpxCopFpaCpldPersonality::EraseSector() - Erase specified flash sector contents
//------------------------------------------------------------------------------

bool IVpxCopFpaCpldPersonality::EraseSector(int sector)
{
    Chip->OnEraseSector_Start(sector);

    int Addr = SectorBaseAddress(sector);

    Rtc.uS(5000);  // 5 ms delay before polling

    const double Timeout = 10.0;   // in seconds
    double Then = Rtc.Now() + Timeout;

    while (Prom(Addr) != ((unsigned int)(0xFF)))
        {
        if (Rtc.Now() > Then)
            {
            ProcessStatusEvent e("Timeout erasing sector");
            OnLoadError.Execute(e);
            return false;
            }
        }

    return true;
}

//------------------------------------------------------------------------------
//  IVpxCopFpaCpldPersonality::Program() -- Write buffer to flash
//------------------------------------------------------------------------------

bool  IVpxCopFpaCpldPersonality::Program(FpgaImage & Image)
{
//    switch (RomMode)
//        {
//        case romM29EWtype:    // M29EW 8 bit type
//            return SingleCellProgram(Image);    // test single cell case

//        default:
            return BlockProgram(Image);

//        }
}

//------------------------------------------------------------------------------
//  IVpxCopFpaCpldPersonality::SingleCellProgram() -- Write buffer to flash 1 cell @ time
//------------------------------------------------------------------------------

bool  IVpxCopFpaCpldPersonality::SingleCellProgram(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    Chip->OnSingleCellPgm_Start();

    FpgaImageDatagram Image_DG(Image);
        {
        std::ostringstream msg;
        msg << "Dump Bytes 0x10 to 0x40";

        ProcessStatusEvent e(msg.str());
        OnLoadError.Execute(e);
        }
        {
        std::ostringstream msg;
        for (unsigned int i=0x20; i<0x50; i++)
            {
            msg << std::hex << "[" << (((int)Image_DG.AsChar(i))&0xFF) << "]";
            }

        ProcessStatusEvent e(msg.str());
        OnLoadError.Execute(e);
        }
    // Program device
    for (unsigned int i=0; i < Image_DG.SizeInInts(); ++i)
        {
        unsigned int data_word = static_cast<unsigned int>(Image_DG.AsInt(i));
        unsigned int rev_word  = Chip->ByteOrder(data_word);

        for (unsigned int b=0; b<4; b++)
            {
            unsigned char * asChar = reinterpret_cast<unsigned char *>(&rev_word);
            unsigned int byte_idx  = 4*i+b;
            unsigned int data_byte = DataMask() & static_cast<unsigned int>(asChar[b]);

            if (!Flash(byte_idx, data_byte))
                {
                std::ostringstream msg;
                msg << "Unable to program byte 0x" << std::hex << byte_idx
                    << "[" << data_byte << "]";

                ProcessStatusEvent e(msg.str());
                OnLoadError.Execute(e);
                return false;
                }
            }

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInInts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    Chip->OnSingleCellPgm_End();

    ProcessProgressEvent e(100);
    OnLoadProgress.Execute(e);

    return true;
}

//------------------------------------------------------------------------------
//  IVpxCopFpaCpldPersonality::BlockProgram() -- Write buffer to flash by blocks
//------------------------------------------------------------------------------

bool  IVpxCopFpaCpldPersonality::BlockProgram(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;
    const int BlockWriteSize = 256;   // max for 8 bit mode

    FpgaImageDatagram Image_DG(Image);

    // Program device
    for (unsigned int i = 0; i < Image_DG.SizeInChars(); i+=BlockWriteSize)
        {
        Chip->OnWriteBufferProgram(Image_DG, i, BlockWriteSize);

        const double Timeout = 2.000;   // in sec
        double Then = Rtc.Now() + Timeout;

        //  Poll the final write for completion of all program writes
        unsigned int check_word = static_cast<unsigned int>(Image_DG.AsChar(i+BlockWriteSize-1));
        check_word &= DataMask();
        check_word = Chip->ByteOrder(check_word);

        while (Prom(i+BlockWriteSize-1) != check_word)
            {
            if (Rtc.Now() < Then)
                continue;

            Chip->OnBufferProgramAbortAndReset();

            std::stringstream msg;
            msg << "Prom wrote 0x" << std::hex << Prom(i+BlockWriteSize-1)
            << " instead of 0x" << check_word
            << " at address 0x" << i+BlockWriteSize-1;

            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInChars());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent e(100);
    OnLoadProgress.Execute(e);

    return true;
}

//------------------------------------------------------------------------------
//  IVpxCopFpaCpldPersonality::Verify() -- Verify buffer against flash contents
//------------------------------------------------------------------------------

bool IVpxCopFpaCpldPersonality::Verify(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    FpgaImageDatagram Image_DG(Image);
    // Program device
    for (unsigned int i = 0; i < Image_DG.SizeInChars(); ++i)
        {
        unsigned int dst = Flash(i);
        unsigned int src = static_cast<unsigned int>(Image_DG.AsChar(i));
        src &= DataMask();
        if (Chip->ByteOrder(dst) != src)
            {
            std::ostringstream msg;

            msg << "Prom read 0x" << std::hex << dst
                << " instead of 0x" << src
                << " at address 0x" << i;
            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInChars());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent p(100);
    OnLoadProgress.Execute(p);

    return true;
}

//---------------------------------------------------------------------------
//  IVpxCopFpaCpldPersonality::Read() -- Fill buffer from flash
//---------------------------------------------------------------------------

bool IVpxCopFpaCpldPersonality::Read(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    FpgaImageDatagram Image_DG(Image);
    // Read device
    for (unsigned int i = 0; i < Image_DG.SizeInChars(); ++i)
        {
        Image_DG.AsChar(i, Chip->ByteOrder(Flash(i)));

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInChars());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent p(100);
    OnLoadProgress.Execute(p);
    return true;
}


//------------------------------------------------------------------------------
// IVpxCopFpaCpldPersonality::SectorBaseAddress() -
//------------------------------------------------------------------------------

unsigned int  IVpxCopFpaCpldPersonality::SectorBaseAddress(unsigned int sector)
{
    return Chip->SectorBaseAddress(sector);
}

//------------------------------------------------------------------------------
// IVpxCopFpaCpldPersonality::CreateChipImpl() - Read CPLD Status Word
//------------------------------------------------------------------------------

void  IVpxCopFpaCpldPersonality::CreateChipImpl()
{
    UncreateChipImpl();
    switch (RomMode)
        {
        case romM29EWtype:    // M29EW Type
        default:
            Chip = new TypeM29EW_8bit_ChipImpl(this);
            break;
        }
}

//------------------------------------------------------------------------------
//  IVpxCopFpaCpldPersonality::UncreateChipImpl
//------------------------------------------------------------------------------

void  IVpxCopFpaCpldPersonality::UncreateChipImpl()
{
    if (Chip)
        {
        delete Chip;
        Chip = 0;
        }
}

//---------------------------------------------------------------------------
//  IVpxCopFpaCpldPersonality::Prom(a,d) -- write to prom device
//---------------------------------------------------------------------------

void  IVpxCopFpaCpldPersonality::Prom(unsigned int addr, unsigned int value)
{
    Cpld.CPLD_Write(addr, value);
}

//---------------------------------------------------------------------------
//  IVpxCopFpaCpldPersonality::Prom(a) -- read from prom device
//---------------------------------------------------------------------------

unsigned int  IVpxCopFpaCpldPersonality::Prom(unsigned int addr)
{
    return Cpld.CPLD_Read(addr);
}

//------------------------------------------------------------------------------
// IVpxCopFpaCpldPersonality::StartSector() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  IVpxCopFpaCpldPersonality::StartSector() const
{
    return Chip->StartSector();
}

//------------------------------------------------------------------------------
// IVpxCopFpaCpldPersonality::Sectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  IVpxCopFpaCpldPersonality::Sectors() const
{
    return Chip->Sectors();
}

//------------------------------------------------------------------------------
// IVpxCopFpaCpldPersonality::GoldenImageStartSector() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  IVpxCopFpaCpldPersonality::GoldenImageStartSector() const
{
    return Chip->GoldenImageStartSector();
}

//------------------------------------------------------------------------------
// IVpxCopFpaCpldPersonality::GoldenImageSectors() - Sector info impl
//------------------------------------------------------------------------------

unsigned int  IVpxCopFpaCpldPersonality::GoldenImageSectors() const
{
    return Chip->GoldenImageSectors();
}

//-------------------------------------------------------------
// IVpxCopFpaCpldPersonality::Flash() - Reads byte from specified address
//-------------------------------------------------------------

int IVpxCopFpaCpldPersonality::Flash(int addr)
{
    if (GoldenImageActive())
        {
        addr += Chip->SectorBaseAddress(GoldenImageStartSector());
        }
    else
        {
        addr += Chip->SectorBaseAddress(StartSector());
        }

    return Prom(addr);
}

//-------------------------------------------------------------
// IVpxCopFpaCpldPersonality::Flash() - Programs byte at specified address
//-------------------------------------------------------------

bool IVpxCopFpaCpldPersonality::Flash(int addr, int data)
{
    if (GoldenImageActive())
        {
        addr += Chip->SectorBaseAddress(GoldenImageStartSector());
        }
    else
        {
        addr += Chip->SectorBaseAddress(StartSector());
        }
    //
    //  Start the flash operation
    OnFlashWrite_Start(addr, data);

    const double Timeout = 0.500;   // in sec
    double Then = Rtc.Now() + Timeout;
    while (Prom(addr) != static_cast<unsigned int>(data))
        {
        if (Rtc.Now() < Then)
            continue;

        std::ostringstream msg;
        msg << "Prom wrote 0x" << std::hex << Prom(addr)
            << " instead of 0x" << data
            << " at address 0x" << addr;
        ProcessStatusEvent e(msg.str());
        OnLoadError.Execute(e);
        return false;
        }

    return true;
}

//-------------------------------------------------------------
// IVpxCopFpaCpldPersonality::Authenticate() - Verify board integrity
//-------------------------------------------------------------

bool IVpxCopFpaCpldPersonality::Authenticate()
{
    return true;
}


//==============================================================================
//  CLASS IK7FpaChip -- K7 Flash Sub-Personality Interface
//==============================================================================
//------------------------------------------------------------------------------
//  IK7FpaChip::ByteOrder() -- Reorder Bytes between image and chip
//------------------------------------------------------------------------------

unsigned int  IK7FpaChip::ByteOrder(unsigned int data)
{
    // this is a 32 bit image. Assume Hi and Lo short are correct, swap
    //   the rest
    unsigned short * asShort = reinterpret_cast<unsigned short *>(&data);
    ByteSwap(asShort[0]);
    ByteSwap(asShort[1]);
    //  then bit reverse what is left
    return BitReverse(data);
}

//==============================================================================
//  CLASS K7FpaPersonality -- Interface to FPA specific operations
//==============================================================================
//---------------------------------------------------------------------------
//  K7FpaPersonality::Initialize() -- Initialize Personality
//---------------------------------------------------------------------------

bool  K7FpaPersonality::Initialize()
{
    return Chip->Initialize();
}

//------------------------------------------------------------------------------
//  K7FpaPersonality::EraseSector() - Erase specified flash sector contents
//------------------------------------------------------------------------------

bool K7FpaPersonality::EraseSector(int sector)
{
    if (Chip->IsHWProtected())
        {
        Chip->ReleaseProtection();
        if (Chip->IsHWProtected()==true)
            {
            std::stringstream msg;
            msg << "ROM is write protected @ sector " << sector;
            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }
        }

    Chip->EraseSector(sector);

    Chip->SetProtection();

    int Addr = Chip->SectorBaseAddress(sector);

    Rtc.uS(5000);  // 5 ms delay before polling

    const double Timeout = 10.0;   // in seconds
    double Then = Rtc.Now() + Timeout;

    while (Chip->Read32(Addr) != ((unsigned int)(0xFFFFFFFF)))     // this waits for both ROMs to erase and read data
        {
        if (Rtc.Now() > Then)
            {
            ProcessStatusEvent e("Timeout erasing sector");
            OnLoadError.Execute(e);
            return false;
            }
        }

    return true;
}

//------------------------------------------------------------------------------
//  K7FpaPersonality::Program() -- Write buffer to flash
//------------------------------------------------------------------------------

bool  K7FpaPersonality::Program(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;
    const unsigned int BlockWriteSize = 256/sizeof(int);

    FpgaImageDatagram Image_DG(Image);

    if (Chip->IsHWProtected())
        {
        Chip->ReleaseProtection();
        if (Chip->IsHWProtected())
            {
            std::stringstream msg;
            msg << "ROM is write protected";
            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }
        }

    Chip->SetProgramPageMode(true);

    // Program device
    for (unsigned int i = 0; i < Image_DG.SizeInInts(); i+=BlockWriteSize)
        {
        //  Set Page Address
        unsigned int addr = i<<2;
        Chip->SetPageAddress(addr);

        //  Enqueue the Page Data
        for (unsigned int ofs=0; ofs<BlockWriteSize; ofs++)
            {
            unsigned int data_word = static_cast<unsigned int>(Image_DG.AsInt(i+ofs));
            unsigned int reorder_word = Chip->ByteOrder(data_word);
            Chip->Enqueue32(reorder_word);
            }
        //
        //  Initiate Page Write
        Chip->PageProgram();
        //
        //  Check Page Write Progress (for previous page)
        Chip->WaitForPageComplete();
        //
        //  Progress Update
        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInInts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    Chip->SetProgramPageMode(false);

    ProcessProgressEvent e(100);
    OnLoadProgress.Execute(e);

    Chip->SetProtection();

    return true;
}

//------------------------------------------------------------------------------
//  K7FpaPersonality::Verify() -- Verify buffer against flash contents
//------------------------------------------------------------------------------

bool K7FpaPersonality::Verify(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    FpgaImageDatagram Image_DG(Image);
    // Program device
    for (unsigned int i = 0; i < Image_DG.SizeInInts(); ++i)
        {
        unsigned int addr = i<<2;
        if (GoldenImageActive())
            {
            addr += Chip->SectorBaseAddress(GoldenImageStartSector());
            }
        else
            {
            addr += Chip->SectorBaseAddress(StartSector());
            }

        unsigned int dst = Chip->Read32(addr);
        unsigned int src = static_cast<unsigned int>(Image_DG.AsInt(i));
        if (Chip->ByteOrder(dst) != src)
            {
            std::ostringstream msg;

            msg << "Prom read 0x" << std::hex << dst
                << " instead of 0x" << src
                << " at address 0x" << i;
            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInInts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent p(100);
    OnLoadProgress.Execute(p);

    return true;
}

//---------------------------------------------------------------------------
//  K7FpaPersonality::Read() -- Fill buffer from flash
//---------------------------------------------------------------------------

bool K7FpaPersonality::Read(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    FpgaImageDatagram Image_DG(Image);
    // Read device
    for (unsigned int i = 0; i < Image_DG.SizeInInts(); ++i)
        {
        unsigned int addr = i<<2;
        if (GoldenImageActive())
            {
            addr += Chip->SectorBaseAddress(GoldenImageStartSector());
            }
        else
            {
            addr += Chip->SectorBaseAddress(StartSector());
            }
        Image_DG.AsInt(i, Chip->ByteOrder(Chip->Read32(addr)));

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInInts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent p(100);
    OnLoadProgress.Execute(p);
    return true;
}

//-------------------------------------------------------------
// K7FpaPersonality::Authenticate() - Verify board integrity
//-------------------------------------------------------------

bool K7FpaPersonality::Authenticate()
{
    return true;
}

//==============================================================================
//  CLASS IXUFpaChip -- XU Flash Sub-Personality Interface
//==============================================================================
//------------------------------------------------------------------------------
//  IXUFpaChip::ByteOrder() -- Reorder Bytes between image and chip
//------------------------------------------------------------------------------

unsigned int  IXUFpaChip::ByteOrder(unsigned int data)
{
    // this is a 32 bit image. Assume Hi and Lo short are correct, swap
    //   the rest
    unsigned short * asShort = reinterpret_cast<unsigned short *>(&data);
    ByteSwap(asShort[0]);
    ByteSwap(asShort[1]);
    //  then bit reverse what is left
    return BitReverse(data);
}

//==============================================================================
//  CLASS XUFpaPersonality -- Interface to FPA specific operations
//==============================================================================
//---------------------------------------------------------------------------
//  XUFpaPersonality::Initialize() -- Initialize Personality
//---------------------------------------------------------------------------

bool  XUFpaPersonality::Initialize()
{
    return Chip->Initialize();
}

//------------------------------------------------------------------------------
//  XUFpaPersonality::EraseSector() - Erase specified flash sector contents
//------------------------------------------------------------------------------

bool XUFpaPersonality::EraseSector(int sector)
{
    if (Chip->IsHWProtected())
        {
        Chip->ReleaseProtection();
        if (Chip->IsHWProtected()==true)
            {
            std::stringstream msg;
            msg << "ROM is write protected @ sector " << sector;
            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }
        }

    Chip->EraseSector(sector);

    Chip->SetProtection();

    int Addr = Chip->SectorBaseAddress(sector);

    Rtc.uS(5000);  // 5 ms delay before polling

    const double Timeout = 10.0;   // in seconds
    double Then = Rtc.Now() + Timeout;

    while (Chip->Read32(Addr) != ((unsigned int)(0xFFFFFFFF)))     // this waits for both ROMs to erase and read data
        {
        if (Rtc.Now() > Then)
            {
            ProcessStatusEvent e("Timeout erasing sector");
            OnLoadError.Execute(e);
            return false;
            }
        }

    return true;
}

//------------------------------------------------------------------------------
//  XUFpaPersonality::Program() -- Write buffer to flash
//------------------------------------------------------------------------------

bool  XUFpaPersonality::Program(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;
    const unsigned int BlockWriteSize = 256/sizeof(int);

    FpgaImageDatagram Image_DG(Image);

    if (Chip->IsHWProtected())
        {
        Chip->ReleaseProtection();
        if (Chip->IsHWProtected())
            {
            std::stringstream msg;
            msg << "ROM is write protected";
            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }
        }

    Chip->SetProgramPageMode(true);

    // Program device
    for (unsigned int i = 0; i < Image_DG.SizeInInts(); i+=BlockWriteSize)
        {
        //  Set Page Address
        unsigned int addr = i<<2;
        Chip->SetPageAddress(addr);

        //  Enqueue the Page Data
        for (unsigned int ofs=0; ofs<BlockWriteSize; ofs++)
            {
            unsigned int data_word = static_cast<unsigned int>(Image_DG.AsInt(i+ofs));
            unsigned int reorder_word = Chip->ByteOrder(data_word);
            Chip->Enqueue32(reorder_word);
            }
        //
        //  Initiate Page Write
        Chip->PageProgram();
        //
        //  Check Page Write Progress (for previous page)
        Chip->WaitForPageComplete();
        //
        //  Progress Update
        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInInts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    Chip->SetProgramPageMode(false);

    ProcessProgressEvent e(100);
    OnLoadProgress.Execute(e);

    Chip->SetProtection();

    return true;
}

//------------------------------------------------------------------------------
//  XUFpaPersonality::Verify() -- Verify buffer against flash contents
//------------------------------------------------------------------------------

bool XUFpaPersonality::Verify(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    FpgaImageDatagram Image_DG(Image);
    // Program device
    for (unsigned int i = 0; i < Image_DG.SizeInInts(); ++i)
        {
        unsigned int addr = i<<2;
        if (GoldenImageActive())
            {
            addr += Chip->SectorBaseAddress(GoldenImageStartSector());
            }
        else
            {
            addr += Chip->SectorBaseAddress(StartSector());
            }

        unsigned int dst = Chip->Read32(addr);
        unsigned int src = static_cast<unsigned int>(Image_DG.AsInt(i));
        if (Chip->ByteOrder(dst) != src)
            {
            std::ostringstream msg;

            msg << "Prom read 0x" << std::hex << dst
                << " instead of 0x" << src
                << " at address 0x" << i;
            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);
            return false;
            }

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInInts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent p(100);
    OnLoadProgress.Execute(p);

    return true;
}

//---------------------------------------------------------------------------
//  XUFpaPersonality::Read() -- Fill buffer from flash
//---------------------------------------------------------------------------

bool XUFpaPersonality::Read(FpgaImage & Image)
{
    Percent[0] = Percent[1] = 0;

    FpgaImageDatagram Image_DG(Image);
    // Read device
    for (unsigned int i = 0; i < Image_DG.SizeInInts(); ++i)
        {
        unsigned int addr = i<<2;
        if (GoldenImageActive())
            {
            addr += Chip->SectorBaseAddress(GoldenImageStartSector());
            }
        else
            {
            addr += Chip->SectorBaseAddress(StartSector());
            }
        Image_DG.AsInt(i, Chip->ByteOrder(Chip->Read32(addr)));

        Percent[1] = static_cast<int>((i*100)/Image_DG.SizeInInts());
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent p(100);
    OnLoadProgress.Execute(p);
    return true;
}

//-------------------------------------------------------------
// XUFpaPersonality::Authenticate() - Verify board integrity
//-------------------------------------------------------------

bool XUFpaPersonality::Authenticate()
{
    return true;
}

//==============================================================================
//  CLASS FpaLoadThread -- Flash load via FPA interface
//==============================================================================
//---------------------------------------------------------------------------
//  FpaLoadThread::FpaLoadThread() -- Ctor
//---------------------------------------------------------------------------

FpaLoadThread::FpaLoadThread(FpaLoadBase & fpa)
    : inherited("FpaLoadThread"), Image(1<<22),
      FpaPersonality(new StandardFpaPersonality(fpa)),
      Erasing(Event::ManualReset),
      Parsing(Event::ManualReset),
      Programming(Event::ManualReset),
      Verifying(Event::ManualReset),
      Reading(Event::ManualReset),
      FullEraseMode(true)
{
    Construct();
}

//---------------------------------------------------------------------------
//  FpaLoadThread::FpaLoadThread() -- Apply a non default personality
//---------------------------------------------------------------------------

FpaLoadThread::FpaLoadThread(IFpaPersonality * fpaType)
    : inherited("FpaLoadThread"), Image(1<<22),
      FpaPersonality(fpaType),
      Erasing(Event::ManualReset),
      Parsing(Event::ManualReset),
      Programming(Event::ManualReset),
      Verifying(Event::ManualReset),
      Reading(Event::ManualReset),
      FullEraseMode(true)
{
    Construct();
}

//------------------------------------------------------------------------
//  destructor for class FpaLoadThread
//------------------------------------------------------------------------

FpaLoadThread::~FpaLoadThread()
{
    delete FpaPersonality;    FpaPersonality = 0;
}

//---------------------------------------------------------------------------
//  FpaLoadThread::Construct() -- Shared constructor body operations
//---------------------------------------------------------------------------

void  FpaLoadThread::Construct()
{
    Image.OnProgress.SetEvent(this, &FpaLoadThread::HandleParseProgress);
    Image.OnComplete.SetEvent(this, &FpaLoadThread::HandleParseComplete);
    Image.OnError.SetEvent(this, &FpaLoadThread::HandleParseError);

    FpaPersonality->OnLoadProgress.SetEvent(this, &FpaLoadThread::HandlePersonalityLoadProgress);
    FpaPersonality->OnLoadError.SetEvent(this, &FpaLoadThread::HandlePersonalityLoadError);

    Once = false;
    Resume();
}

//---------------------------------------------------------------------------
//  FpaLoadThread::OnOpen() -- Board Open Initialization
//---------------------------------------------------------------------------

void  FpaLoadThread::OnOpen()
{
    Once = false;
}

//---------------------------------------------------------------------------
//  FpaLoadThread::Init() -- Shared constructor body operations
//---------------------------------------------------------------------------

void  FpaLoadThread::Init()
{
    inherited::Init();
    //  Initialize the FpaEvents object with the objects we'll be waiting for:
	AddObject(Erasing, sErasing);
    AddObject(Parsing, sParsing);
    AddObject(Programming, sProgramming);
    AddObject(Verifying, sVerifying);
    AddObject(Reading, sReading);
}

//---------------------------------------------------------------------------
//  FpaLoadThread::Erase() --
//---------------------------------------------------------------------------

void FpaLoadThread::Erase()
{
    if (FpaPersonality->IsProtected())
        {
        ProcessStatusEvent e("Can't erase, device is protected!");
        OnLoadError.Execute(e);
        return;
        }

    Erasing.Set();
}

//---------------------------------------------------------------------------
//  FpaLoadThread::Parse() --
//---------------------------------------------------------------------------

void FpaLoadThread::Parse(std::string exo_file)
{
    Image.FileName(exo_file);
    Parsing.Set();
}

//---------------------------------------------------------------------------
//  FpaLoadThread::Program() --
//---------------------------------------------------------------------------

void FpaLoadThread::Program()
{
    if (FpaPersonality->IsProtected())
        {
        ProcessStatusEvent e("Can't reprogram - Device is protected!");
        OnLoadError.Execute(e);
        return;
        }

    Programming.Set();
}

//---------------------------------------------------------------------------
//  FpaLoadThread::Verify() --
//---------------------------------------------------------------------------

void FpaLoadThread::Verify()
{
    Verifying.Set();
}

//---------------------------------------------------------------------------
//  FpaLoadThread::Read() --
//---------------------------------------------------------------------------

void FpaLoadThread::Read()
{
    Reading.Set();
}

//---------------------------------------------------------------------------
//  FpaLoadThread::BitFileFpgaName() --
//---------------------------------------------------------------------------

std::string FpaLoadThread::BitFileFpgaName(std::string PartStr)
{
    //
    //  Step 1: Check if the Part string is empty
    //
    if(PartStr.length() == 0)
        return "No Fpga string defined inside the bit file.";

    std::string FpgaName;

    //
    //  Step 2: get the fpga name from the bit file part string
    //          which each board has it unique string positioning.
    //
    // XU family
    if (PartStr.substr(0,2) == "ku")
        FpgaName = PartStr.substr(2,3);
    // XA family
    else if (PartStr.substr(0,2) == "7a")
        FpgaName = PartStr.substr(2,4);
    // K7
    else if (PartStr.substr(0,2) == "7k")
        FpgaName = PartStr.substr(2,4);
    // X6
    else if (PartStr.substr(0,2) == "6v")
        FpgaName = PartStr.substr(2,6);
    // X5
    else if(PartStr.substr(0,2) == "5v")
        FpgaName = PartStr.substr (2,5);
    // X3
    else if(PartStr.substr(0,1) == "3")
        FpgaName = PartStr.substr(1,7);
    else
        FpgaName = "Unknown";

    //
    //  Step 3: in the bit file the fpga name letters are small
    //          need to be capitalize.
    //
    std::transform(FpgaName.begin(), FpgaName.end()
                   ,FpgaName.begin(), ::toupper);

    return FpgaName;
}

//---------------------------------------------------------------------------
//  FpaLoadThread::CheckFpga() --
//---------------------------------------------------------------------------

bool FpaLoadThread::CheckFpga(std::string BitFileFpgaStr, std::string BoardFpgaStr)
{
    //Parse the part string to get the pure fpga name
    BitFileFpgaStr = BitFileFpgaName(BitFileFpgaStr);

    if(BitFileFpgaStr == BoardFpgaStr)
        return true;
    else
        return false;
}

//------------------------------------------------------------------------------
//  FpaLoadThread::HandleParseProgress() --
//------------------------------------------------------------------------------

void FpaLoadThread::HandleParseProgress(ProcessProgressEvent & event)
{
    OnParseProgress.Execute(event);
}

//------------------------------------------------------------------------------
//  FpaLoadThread::HandleParseComplete() --
//------------------------------------------------------------------------------

void FpaLoadThread::HandleParseComplete(ProcessCompletionEvent & event)
{
    OnParseComplete.Execute(event);
}

//------------------------------------------------------------------------------
//  FpaLoadThread::HandleParseError() --
//------------------------------------------------------------------------------

void FpaLoadThread::HandleParseError(ProcessStatusEvent & event)
{
    OnParseError.Execute(event);
}

//------------------------------------------------------------------------------
//  FpaLoadThread::HandlePersonalityLoadError() -- forward events up
//------------------------------------------------------------------------------

void FpaLoadThread::HandlePersonalityLoadError(ProcessStatusEvent & event)
{
    OnLoadError.Execute(event);
}

//------------------------------------------------------------------------------
//  FpaLoadThread::HandlePersonalityLoadProgress() -- forward events up
//------------------------------------------------------------------------------

void FpaLoadThread::HandlePersonalityLoadProgress(ProcessProgressEvent & event)
{
    OnLoadProgress.Execute(event);
}

//---------------------------------------------------------------------------
//  FpaLoadThread::Execute() --
//---------------------------------------------------------------------------

void FpaLoadThread::Execute()
{
    while (!Terminated())
        {
        // WaitResult is either a) Suicide event, b) Prom event
        int WaitResult = WaitFor();

        if (Terminated())
            break;
        if (!Once)
            {
            Initialize();
            Once = true;
            }

        //  Terminate:  Exit thread
        if (WaitResult == sParsing)
            {
            DoParse();
            Parsing.Clear();
            }
        else if (WaitResult == sProgramming)
            {
            DoProgram();
            Programming.Clear();
            }
        else if (WaitResult == sErasing)
            {
            DoErase();
            Erasing.Clear();
            }
        else if (WaitResult == sVerifying)
            {
            DoVerify();
            Verifying.Clear();
            }
        else if (WaitResult == sReading)
            {
            DoRead();
            Reading.Clear();
            }
        }
}

//------------------------------------------------------------------------------
//  FpaLoadThread::Initialize() - Initializes flash for subsequent programming
//------------------------------------------------------------------------------

void FpaLoadThread::Initialize()
{
    if (FpaPersonality->Initialize()==false)
        {
        ProcessStatusEvent e("FPA Initialize Failed.");
        OnLoadError.Execute(e);
        return;
        }

    // initialize the Image
    Image.Size(FpaPersonality->DeviceSize());
}

//------------------------------------------------------------------------
//  FpaLoadThread::Id()  --  Fetch the VsProm ID
//------------------------------------------------------------------------

int FpaLoadThread::Id() const
{
    return FpaPersonality->Id();
}

//------------------------------------------------------------------------
//  FpaLoadThread::FirmwareRevision()  --  Fetch the firmware revision
//------------------------------------------------------------------------

int FpaLoadThread::FirmwareRevision() const
{
    return FpaPersonality->FirmwareRevision();
}

//------------------------------------------------------------------------
//  FpaLoadThread::PcbRevision()  --  Fetch the PCB revision
//------------------------------------------------------------------------
//
//int FpaLoadThread::PcbRevision() const
//{
//    return FpaPersonality->PcbRevision();
//}
//
//---------------------------------------------------------------------------
//  FpaLoadThread::DoErase() -- Erase flash contents
//---------------------------------------------------------------------------

void FpaLoadThread::DoErase()
{
    unsigned int startSector(StartSector());
    unsigned int lastSector(Sectors());
    unsigned int totalSectors(Sectors());

    if (FullEraseMode==false)
        {
        // We want to speed up erasing by only clearing what the image requires.
        FpgaImageDatagram Image_DG(Image);
        size_t image_size = Image_DG.SizeInChars();
        if (image_size)
            {
            double size_sectors = static_cast<double>(image_size)/SectorSize();
            if (size_sectors)
                lastSector = totalSectors = static_cast<unsigned int>(size_sectors+1.0);
            }
        }
    //
    //  If we are using the Golden Image, we need to access the second half of the sector map
    if (HasGoldenImage() && GoldenImageActive())
        {
        startSector = GoldenImageStartSector();
        lastSector = GoldenImageStartSector() + GoldenImageSectors();
        }

    int Percent[2];
    Percent[0] = Percent[1] = 0;

    for (unsigned int i = startSector; i < lastSector; ++i)
        {
        if (!FpaPersonality->EraseSector(i))
            {
            std::ostringstream msg;
            msg << "Unable to erase sector " << i ;

            ProcessStatusEvent e(msg.str());
            OnLoadError.Execute(e);

            ProcessCompletionEvent c(false);
            OnLoadComplete.Execute(c);
            return;
            }

        //
        //  Progress Update
        Percent[1] = static_cast<int>(((i-startSector)*100)/totalSectors);
        if (Percent[1] != Percent[0])
            {
            ProcessProgressEvent e(Percent[1]);
            OnLoadProgress.Execute(e);
            Percent[0] = Percent[1];
            }
        }

    ProcessProgressEvent p(100);
    OnLoadProgress.Execute(p);

    ProcessCompletionEvent e(true);
    OnLoadComplete.Execute(e);
}

//---------------------------------------------------------------------------
//  FpaLoadThread::DoParse() -- Parse EXO image
//---------------------------------------------------------------------------

void FpaLoadThread::DoParse()
{
    Image.Parse();
}

//---------------------------------------------------------------------------
//  FpaLoadThread::DoProgram() -- Write buffer to flash
//---------------------------------------------------------------------------

void FpaLoadThread::DoProgram()
{
    bool status = FpaPersonality->Program(Image);

    ProcessCompletionEvent c(status);
    OnLoadComplete.Execute(c);
}

//---------------------------------------------------------------------------
//  FpaLoadThread::DoVerify() -- Verify buffer against flash contents
//---------------------------------------------------------------------------

void FpaLoadThread::DoVerify()
{
    bool status = FpaPersonality->Verify(Image);

    ProcessCompletionEvent e(status);
    OnLoadComplete.Execute(e);
}

//---------------------------------------------------------------------------
//  FpaLoadThread::DoRead() -- Fill buffer from flash
//---------------------------------------------------------------------------

void FpaLoadThread::DoRead()
{
    bool status = FpaPersonality->Read(Image);

    ProcessCompletionEvent e(status);
    OnLoadComplete.Execute(e);
}


}   // namespace Innovative

