//---------------------------------------------------------------------------

#ifndef GuiLoader_MbH
#define GuiLoader_MbH

#if _MSC_VER > 0x0800
//---------------------------------------------------------------------------
//   "Msvc" GuiLoader Files ...Need different definitions for .Net Gui
#error "For .NET include <Framework/Net/GuiLoader_Net.h> in your form file"
//#include <Framework/Net/GuiLoader_Net.h>

#else
//---------------------------------------------------------------------------
//   "Standard" GuiLoader definitions ...(Borland and Linux compatible)

#include <sstream>      // Avoids compiler error with _algobase.h
#include <vector>
#include <CountedPtr_Mb.h>


namespace Innovative
{

//==============================================================================
//  CLASS IGuiLoadable - Interface to all GUI Loader objects
//==============================================================================

class IGuiLoadable
{
public:
    virtual ~IGuiLoadable()  {  }
    virtual void  Save() const = 0;
    virtual void  Load() const = 0;

    virtual void Invalidate()
    {}
    virtual unsigned int *Id()
    { return 0; }

    virtual IGuiLoadable * Clone() const = 0;
};

//==============================================================================
//  CLASS GuiLoader - Automate Loading of GUI Controls from settings elements
//==============================================================================

class GuiLoader
{
public:
    //
    //  Member Functions
    GuiLoader();
    ~GuiLoader();

    void  Install(const IGuiLoadable & loader);
    void  Install(IGuiLoadable * loader);

    void  Load();
    void  Save();

    void  Invalidate();
    void  Invalidate(unsigned int *id);

    //  "settings-centric" aliases
    void  SetSettings(bool deepwrite = true)
        {  if (deepwrite) Invalidate();
           Load();
        }
    void  GetSettings(bool deepread = true)
        {   if (deepread) Invalidate();
            Save();
        }


    typedef std::vector< CountedPtr<IGuiLoadable> > LoaderStoreType;
    LoaderStoreType  loaders;

    GuiLoader(const GuiLoader &);   // unimplemented, non copyable
};

//==============================================================================
//  TEMPLATE CLASS Cached - Cache utility class.
//==============================================================================
//------------------------------------------------------------------------------
//  Template GuiCacheEquivalent() - std version, use value of variable as "same"
//------------------------------------------------------------------------------
//  Cached<T> needs to know if the situation has changed so it must reload the
//    variable and thus the GUI on a settings movement.  In some cases (like
//    Frequency_Units), value equivalence isn't right, so we allow this to be
//    overridden for specific types.

template <class T>
inline bool GuiCacheEquivalent(T left, T right)
{
    return left == right;
}


template<class T>
// requires EqualityComparable<T>()
struct Cached
{
    Cached() : _value(T()), _valid(false) {}

    bool changed(T val)
    { return !_valid || !GuiCacheEquivalent(_value, val); }

    bool valid() { return _valid; }

    T&   value() { return _value; }

    void update(T val)
    { _value = val, _valid = true;}

    void invalidate()
    { _valid = false; }

    T    _value;
    bool _valid;
};

//==============================================================================
//  TEMPLATE CLASS ControlLoader - Implements Gui load/save for a control
//==============================================================================

template <class ConT, class VarT>
class ControlLoader : public IGuiLoadable
{
public:
    ControlLoader(ConT * control, VarT & var)
        :  Control(control), Var(var)
        {
        }

    virtual void Save() const
        {
        if(!Cache.valid())
            {
            FromControl(Control, Var);
            Cache.update(Var);
            }
        else
            Var = Cache.value();
        }
    virtual void Load() const
        {
        if(Cache.changed(Var))
            {
            ToControl(Control, Var);
            Cache.update(Var);
            }
        }
    virtual ControlLoader * Clone() const
        {
        return new ControlLoader<ConT, VarT>(*this);
        }

    virtual void Invalidate()
        { Cache.invalidate(); }

    virtual unsigned int *Id()
        { return reinterpret_cast<unsigned int*>(Control); }

private:
    ConT *       Control;
    VarT &       Var;
    mutable Cached<VarT> Cache;
};


//==============================================================================
//  TEMPLATE CLASS ControlLoader_LdArg - Control loader with Arg for ToControl
//==============================================================================

template <class ConT, class VarT, class T>
class ControlLoader_LdArg : public IGuiLoadable
{

public:
    ControlLoader_LdArg(ConT * control, VarT & var, T arg)
        :  Control(control), Var(var), Arg(arg)
        {
        }

    virtual void Save() const
        {
        if(!Cache.valid())
            {
            FromControl(Control, Var);
            Cache.update(Var);
            }
        else
            Var = Cache.value();
        }
    virtual void Load() const
        {
        if(Cache.changed(Var))
            {
            ToControl(Control, Var, Arg);
            Cache.update(Var);
            }
        }
    virtual ControlLoader_LdArg * Clone() const
        {
        return new ControlLoader_LdArg<ConT, VarT, T>(*this);
        }

    virtual void Invalidate()
        { Cache.invalidate(); }

    virtual unsigned int *Id()
        { return reinterpret_cast<unsigned int*>(Control); }

private:
    ConT *       Control;
    VarT &       Var;
    T            Arg;
    mutable Cached<VarT> Cache;
};

//==============================================================================
//  TEMPLATE CLASS ControlLoader_LdArg2 - Control loader with Arg for both
//==============================================================================

template <class ConT, class VarT, class T_To, class T_From>
class ControlLoader_LdArg2 : public IGuiLoadable
{
public:
    ControlLoader_LdArg2(ConT * control, VarT & var, T_To to_arg, T_From from_arg)
        :  Control(control), Var(var), Arg_To(to_arg), Arg_From(from_arg)
        {
        }

    virtual void Save() const
        {
        if(!Cache.valid())
            {
            FromControl(Control, Var, Arg_From);
            Cache.update(Var);
            }
        else
            Var = Cache.value();
        }
    virtual void Load() const
        {
        if(Cache.changed(Var))
            {
            ToControl(Control, Var, Arg_To);
            Cache.update(Var);
            }
        }
    virtual ControlLoader_LdArg2 * Clone() const
        {
        return new ControlLoader_LdArg2<ConT, VarT, T_To, T_From>(*this);
        }

    virtual void Invalidate()
        { Cache.invalidate(); }

    virtual unsigned int *Id()
        { return reinterpret_cast<unsigned int*>(Control); }
private:
    ConT *       Control;
    VarT &       Var;
    T_To         Arg_To;
    T_From       Arg_From;
    mutable Cached<VarT> Cache;
};

//==============================================================================
//  TEMPLATE CLASS ControlLoader_LdArg3 - Control loader with Arg for both
//==============================================================================

template <class ConT, class VarT, class T_To1, class T_To2, class T_From>
class ControlLoader_LdArg3 : public IGuiLoadable
{
public:
    ControlLoader_LdArg3(ConT * control, VarT & var, T_To1 to_arg1, T_To2 to_arg2, T_From from_arg)
        :  Control(control), Var(var), Arg_To1(to_arg1), Arg_To2(to_arg2), Arg_From(from_arg)
        {
        }

    virtual void Save() const
        {
        if(!Cache.valid())
            {
            FromControl(Control, Var, Arg_From);
            Cache.update(Var);
            }
        else
            Var = Cache.value();
        }
    virtual void Load() const
        {
        if(Cache.changed(Var))
            {
            ToControl(Control, Var, Arg_To1, Arg_To2);
            Cache.update(Var);
            }
        }

    virtual ControlLoader_LdArg3 * Clone() const
        {
        return new ControlLoader_LdArg3<ConT, VarT, T_To1, T_To2, T_From>(*this);
        }

    virtual void Invalidate()
        { Cache.invalidate(); }

    virtual unsigned int *Id()
        { return reinterpret_cast<unsigned int*>(Control); }

private:
    ConT *       Control;
    VarT &       Var;
    T_To1        Arg_To1;
    T_To2        Arg_To2;
    T_From       Arg_From;
    mutable Cached<VarT> Cache;
};

//==============================================================================
//  TEMPLATE CLASS ControlLoader_LdArg4 - Control loader with Arg for ToControl
//==============================================================================

template <class ConT,class ColumnT, class RowT, class VarT, class T>
class ControlLoader_LdArg4 : public IGuiLoadable
{

public:
    ControlLoader_LdArg4(ConT * control,ColumnT col, RowT row,  VarT & var, T arg)
        :  Control(control), Var(var), Col(col), Row(row), Arg(arg)
        {
        }

    virtual void Save() const
        {
        if(!Cache.valid())
            {
            FromControl(Control, Col, Row, Var);
            Cache.update(Var);
            }
        else
            Var = Cache.value();
        }
    virtual void Load() const
        {
        if(Cache.changed(Var))
            {
            ToControl(Control, Col, Row, Var, Arg);
            Cache.update(Var);
            }
        }
    virtual ControlLoader_LdArg4 * Clone() const
        {
        return new ControlLoader_LdArg4<ConT, ColumnT, RowT,  VarT, T>(*this);
        }

    virtual void Invalidate()
        { Cache.invalidate(); }

    virtual unsigned int *Id()
        { return reinterpret_cast<unsigned int*>(Control); }

private:
    ConT *       Control;
    VarT &       Var;
    ColumnT   Col;
    RowT        Row;
    T               Arg;
    mutable Cached<VarT> Cache;
};

//==============================================================================
//  TEMPLATE CLASS ControlLoader_LdArg5 -- Load "vector columns" to controls
//==============================================================================
//  Version with ToControl argument
//  We have to/from controls to load columns to vectors.  Use them here

template <class ConT, class ColT, class VarT, class T>
class ControlLoader_LdArg5 : public IGuiLoadable
{

public:
    ControlLoader_LdArg5(ConT * control, ColT col, VarT & var, T arg)
        :  Control(control), Var(var), Col(col), Arg(arg)
        {
        }

    virtual void Save() const
        {
        if(!Cache.valid())
            {
            FromControl(Control, Col.ColumnNumber(), Var);
            Cache.update(Var);
            }
        else
            Var = Cache.value();
        }
    virtual void Load() const
        {
        if(Cache.changed(Var))
            {
            ToControl(Control, Col.ColumnNumber(), Var, Arg);
            Cache.update(Var);
            }
        }
    virtual ControlLoader_LdArg5 * Clone() const
        {
        return new ControlLoader_LdArg5<ConT, ColT, VarT, T>(*this);
        }

    virtual void Invalidate()
        { Cache.invalidate(); }

    virtual unsigned int *Id()
        { return reinterpret_cast<unsigned int*>(Control); }

private:
    ConT *       Control;
    VarT &       Var;
    ColT         Col;
    T            Arg;
    mutable Cached<VarT> Cache;
};

//==============================================================================
//  TEMPLATE CLASS ControlLoader_LdArg6 -- Load "vector columns" to controls
//==============================================================================
//  Version with no ToControl argument
//  We have to/from controls to load columns to vectors.  Use them here

template <class ConT, class ColT, class VarT>
class ControlLoader_LdArg6 : public IGuiLoadable
{

public:
    ControlLoader_LdArg6(ConT * control, ColT col, VarT & var)
        :  Control(control), Var(var), Col(col)
        {
        }

    virtual void Save() const
        {
        if(!Cache.valid())
            {
            FromControl(Control, Col.ColumnNumber(), Var);
            Cache.update(Var);
            }
        else
            Var = Cache.value();
        }
    virtual void Load() const
        {
        if(Cache.changed(Var))
            {
            ToControl(Control, Col.ColumnNumber(), Var);
            Cache.update(Var);
            }
        }
    virtual ControlLoader_LdArg6 * Clone() const
        {
        return new ControlLoader_LdArg6<ConT, ColT, VarT>(*this);
        }

    virtual void Invalidate()
        { Cache.invalidate(); }

    virtual unsigned int *Id()
        { return reinterpret_cast<unsigned int*>(Control); }

private:
    ConT *       Control;
    VarT &       Var;
    ColT         Col;
    mutable Cached<VarT> Cache;
};

//==============================================================================
//  TEMPLATE Function ToGui() - Shortcut to make a ControlLoader
//==============================================================================

template <class ConT, class VarT>
ControlLoader<ConT, VarT>  ToGui(ConT * control, VarT & var)
{
    return ControlLoader<ConT, VarT>(control, var);
}

//==============================================================================
//  TEMPLATE Function ToGui() - Shortcut to make a ControlLoader_LdArg
//==============================================================================

template <class ConT, class VarT, class T>
ControlLoader_LdArg<ConT, VarT, T>  ToGui(ConT * control, VarT & var, T arg)
{
    return ControlLoader_LdArg<ConT, VarT, T>(control, var, arg);
}

//==============================================================================
//  TEMPLATE Function ToGui() - Shortcut to make a ControlLoader_LdArg2
//==============================================================================

template <class ConT, class VarT, class T_To, class T_From>
ControlLoader_LdArg2<ConT, VarT, T_To, T_From>  ToGui(ConT * control, VarT & var, T_To arg_to, T_From arg_from)
{
    return ControlLoader_LdArg2<ConT, VarT, T_To, T_From>(control, var, arg_to, arg_from);
}

//==============================================================================
//  TEMPLATE Function ToGui() - Shortcut to make a ControlLoader_LdArg3
//==============================================================================

template <class ConT, class VarT, class T_To1, class T_To2, class T_From>
ControlLoader_LdArg3<ConT, VarT, T_To1, T_To2, T_From>  ToGui(ConT * control, VarT & var, T_To1 arg_to1, T_To2 arg_to2, T_From arg_from)
{
    return ControlLoader_LdArg3<ConT, VarT, T_To1, T_To2, T_From>(control, var, arg_to1, arg_to2, arg_from);
}

//==============================================================================
//  TEMPLATE Function ToGui() - Shortcut to make a ControlLoader_LdArg4
//==============================================================================

template <class ConT, class ColumnT, class RowT, class VarT, class T>
ControlLoader_LdArg4<ConT, ColumnT, RowT, VarT, T>  ToGui(ConT * control,ColumnT col, RowT row,  VarT & var, T arg)
{
    return ControlLoader_LdArg4<ConT, ColumnT, RowT, VarT, T>(control, col, row, var, arg);
}

//==============================================================================
//  TEMPLATE Function ToGui() - Shortcut to make a ControlLoader_LdArg5 (arg)
//==============================================================================

template <class ConT, class ColumnT, class VarT, class T>
ControlLoader_LdArg5<ConT, ColumnT, VarT, T>  ToGui(ConT * control, ColumnT col, VarT & var, T arg)
{
    return ControlLoader_LdArg5<ConT, ColumnT, VarT, T>(control, col, var, arg);
}

//==============================================================================
//  TEMPLATE Function ToGui() - Shortcut to make a ControlLoader_LdArg6 (no arg)
//==============================================================================

template <class ConT, class ColumnT, class VarT>
ControlLoader_LdArg6<ConT, ColumnT, VarT>  ToGui(ConT * control, ColumnT col, VarT & var)
{
    return ControlLoader_LdArg6<ConT, ColumnT, VarT>(control, col, var);
}


} // namespace Innovative
#endif   // end Msvc/Linux alternatives

#endif
