//  GuiLoader_Net.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2009 Innovative Integration Corporation.
//  All Rights Reserved.
//

#ifndef GuiLoader_NetH
#define GuiLoader_NetH

//---------------------------------------------------------------------------
//   "Msvc" GuiLoader Files ...Need different definitions for .Net Gui

#include <vcclr.h>
#include <vector>

namespace Innovative
{
//==============================================================================
//  CLASS IGuiLoadable - Interface to all GUI Loader objects
//==============================================================================

ref class IGuiLoadable abstract
{
public:
    virtual void  Save() = 0;
    virtual void  Load() = 0;
	virtual void Invalidate()
	{}

    virtual Object ^Id()
	{
		return nullptr;
	}
};

//==============================================================================
//  CLASS GuiLoader - Automate Loading of GUI Controls from settings elements
//==============================================================================

class GuiLoader
{
public:
    //
    //  Member Functions
    GuiLoader();
    ~GuiLoader();

    void  Install(IGuiLoadable ^ loader);

	void  Invalidate();
	void  Invalidate(System::Object ^ id);

    void  Load();
    void  Save();
    //  "settings-centric" aliases
    void  SetSettings()
        {  Load();  }
	void  GetSettings(bool deepread = true)
	{
		if (deepread) Invalidate();
		Save();
	}


    typedef std::vector< gcroot<IGuiLoadable ^> > LoaderStoreType;
    LoaderStoreType  loaders;

    GuiLoader(const GuiLoader &);   // unimplemented, non copyable
};


//==============================================================================
//  TEMPLATE CLASS Cached - Cache utility class.
//==============================================================================
//------------------------------------------------------------------------------
//  Template GuiCacheEquivalent() - std version, use value of variable as "same"
//------------------------------------------------------------------------------
//  Cached<T> needs to know if the situation has changed so it must reload the
//    variable and thus the GUI on a settings movement.  In some cases (like
//    Frequency_Units), value equivalence isn't right, so we allow this to be
//    overridden for specific types.

template <class T>
inline bool GuiCacheEquivalent(T & left, T & right)
{
    return left == right;
}

template<class T>
// requires EqualityComparable<T>() &&
// requires CopyAssignable<T>() &&
// requires CopyConstructible<T>() &&
// requires DefaultConstructible<T>()
ref struct Cached
{
    Cached() : _value(new T()), _valid(false) {}
	~Cached() { delete _value;}

    bool changed(T val)
    { return !_valid || !GuiCacheEquivalent(*_value, val); }

    bool valid() { return _valid; }

    T    value() { return *_value; }

    void update(const T &val)
    { *_value = val, _valid = true;}

    void invalidate()
    { _valid = false; }

    T   *_value;
    bool _valid;
};

//==============================================================================
//  CLASS GuiLoader
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class GuiLoader
//------------------------------------------------------------------------------

GuiLoader::GuiLoader()
{
}

//------------------------------------------------------------------------------
//  destructor for class GuiLoader
//------------------------------------------------------------------------------

GuiLoader::~GuiLoader()
{
    // must destroy all vector objects here...
    //   not needed because they are gcnew-ed objects?
}

//------------------------------------------------------------------------------
//  GuiLoader::Install(IGuiLoadable ^ )  -- register a setting
//------------------------------------------------------------------------------

void  GuiLoader::Install(IGuiLoadable ^ loader)
{
    loaders.push_back( loader );
}

//------------------------------------------------------------------------------
//  GuiLoader::Load()  -- Load registered settings to GUI
//------------------------------------------------------------------------------

void  GuiLoader::Load()
{
    for (LoaderStoreType::iterator iter=loaders.begin() ;
                 iter != loaders.end(); ++iter)
        {
        (*iter)->Load();
        }
}

//------------------------------------------------------------------------------
//  GuiLoader::Save()  -- Save registered settings to settings
//------------------------------------------------------------------------------

void  GuiLoader::Save()
{
    for (LoaderStoreType::iterator iter=loaders.begin() ;
                 iter != loaders.end(); ++iter)
        {
        (*iter)->Save();
        }
}

//------------------------------------------------------------------------------
//  GuiLoader::Invalidate()  -- Invalidate cache for all
//------------------------------------------------------------------------------

void  GuiLoader::Invalidate()
{
	for (LoaderStoreType::iterator iter = loaders.begin();
		iter != loaders.end(); ++iter)
	{
		(*iter)->Invalidate();
	}
}

//------------------------------------------------------------------------------
//  GuiLoader::Invalidate()  -- Invalidate cache for a given element
//------------------------------------------------------------------------------
void  GuiLoader::Invalidate(System::Object ^id)
{
	if (id == nullptr) return; // sanity check

	for (LoaderStoreType::iterator iter = loaders.begin();
		iter != loaders.end(); ++iter)
	{
		if (id->Equals((*iter)->Id()))
			(*iter)->Invalidate();
	}
}

//==============================================================================
//  TEMPLATE CLASS ControlLoader - Implements Gui load/save for a control
//==============================================================================

template <class ConT, class VarT>
ref class ControlLoader : public IGuiLoadable
{
public:
    ControlLoader(ConT ^ control, VarT & var)
        :  Control(control), Var(var), Cache(gcnew Cached<VarT>())
        {
        }

    virtual void Save() override 
        {
        if(!Cache->valid())
            {
            FromControl(Control, Var);
            Cache->update(Var);
            }
        else
            Var = Cache->value();
        }
    virtual void Load() override 
        {
        if(Cache->changed(Var))
            {
            ToControl(Control, Var);
            Cache->update(Var);
            }
        }
    virtual void Invalidate() override
        { Cache->invalidate(); }

    virtual Object^ Id() override
        { return Control; }

private:
    ConT ^        Control;
    VarT &        Var;
    Cached<VarT> ^Cache;
};

//==============================================================================
//  TEMPLATE CLASS ControlLoader_LdArg - Control loader with Arg for ToControl
//==============================================================================

template <class ConT, class VarT, class T>
ref class ControlLoader_LdArg : public IGuiLoadable
{
public:
    ControlLoader_LdArg(ConT ^ control, VarT & var, T arg)
        :  Control(control), Var(var), Arg(arg), Cache(gcnew Cached<VarT>())
        {
        }

    virtual void Save() override
        {
        if(!Cache->valid())
            {
            FromControl(Control, Var);
            Cache->update(Var);
            }
        else
            Var = Cache->value();
        }
    virtual void Load() override
        {
        if(Cache->changed(Var))
            {
            ToControl(Control, Var, Arg);
            Cache->update(Var);
            }
        }
    virtual void Invalidate() override
        { Cache->invalidate(); }

    virtual Object^ Id() override
        { return Control; }
private:
    ConT ^        Control;
    VarT &        Var;
    T             Arg;
    Cached<VarT> ^Cache;
};

//==============================================================================
//  TEMPLATE CLASS ControlLoader_LdArg2 - Control loader with Arg for both
//==============================================================================

template <class ConT, class VarT, class T_To, class T_From>
ref class ControlLoader_LdArg2 : public IGuiLoadable
{
public:
    ControlLoader_LdArg2(ConT ^ control, VarT & var, T_To to_arg, T_From from_arg)
        :  Control(control), Var(var), Arg_To(to_arg), Arg_From(from_arg), Cache(gcnew Cached<VarT>())
        {
        }

    virtual void Save() override
        {
        if(!Cache->valid())
            {
            FromControl(Control, Var, Arg_From);
            Cache->update(Var);
            }
        else
            Var = Cache->value();
        }
    virtual void Load() override
        {
        if(Cache->changed(Var))
            {
            ToControl(Control, Var, Arg_To);
            Cache->update(Var);
            }
        }
    virtual void Invalidate() override
        { Cache->invalidate(); }

    virtual Object^ Id() override
        { return Control; }
private:
    ConT ^        Control;
    VarT &        Var;
    T_To          Arg_To;
    T_From        Arg_From;
    Cached<VarT> ^Cache;
};

//==============================================================================
//  TEMPLATE CLASS ControlLoader_LdArg3 - Control loader with Arg for both
//==============================================================================

template <class ConT, class VarT, class T_To1, class T_To2, class T_From>
ref class ControlLoader_LdArg3 : public IGuiLoadable
{
public:
    ControlLoader_LdArg3(ConT ^ control, VarT & var, T_To1 to_arg1, T_To2 to_arg2, T_From from_arg)
        :  Control(control), Var(var), Arg_To1(to_arg1), 
		    Arg_To2(to_arg2), Arg_From(from_arg), Cache(gcnew Cached<VarT>())
        {
        }

    virtual void Save() override
        {
        if(!Cache->valid())
            {
            FromControl(Control, Var, Arg_From);
            Cache->update(Var);
            }
        else
            Var = Cache->value();
        }
    virtual void Load() override
        {
        if(Cache->changed(Var))
            {
            ToControl(Control, Var, Arg_To1, Arg_To2);
            Cache->update(Var);
            }
        }
    virtual void Invalidate() override
        { Cache->invalidate(); }

    virtual Object^ Id() override
        { return Control; }

private:
    ConT ^        Control;
    VarT &        Var;
    T_To1         Arg_To1;
    T_To2         Arg_To2;
    T_From        Arg_From;
    Cached<VarT> ^Cache;
};


//==============================================================================
//  TEMPLATE Function ToGui() - Shortcut to make a ControlLoader
//==============================================================================

template <class ConT, class VarT>
ControlLoader<ConT, VarT> ^  ToGui(ConT ^ control, VarT & var)
{
    return gcnew ControlLoader<ConT, VarT>(control, var);
}

//==============================================================================
//  TEMPLATE Function ToGui() - Shortcut to make a ControlLoader_LdArg
//==============================================================================

template <class ConT, class VarT, class T>
ControlLoader_LdArg<ConT, VarT, T> ^ ToGui(ConT ^ control, VarT & var, T arg)
{
    return gcnew ControlLoader_LdArg<ConT, VarT, T>(control, var, arg);
}

//==============================================================================
//  TEMPLATE Function ToGui() - Shortcut to make a ControlLoader_LdArg2
//==============================================================================

template <class ConT, class VarT, class T_To, class T_From>
ControlLoader_LdArg2<ConT, VarT, T_To, T_From> ^  ToGui(ConT ^ control, VarT & var, T_To arg_to, T_From arg_from)
{
    return gcnew ControlLoader_LdArg2<ConT, VarT, T_To, T_From>(control, var, arg_to, arg_from);
}

//==============================================================================
//  TEMPLATE Function ToGui() - Shortcut to make a ControlLoader_LdArg3
//==============================================================================

template <class ConT, class VarT, class T_To1, class T_To2, class T_From>
ControlLoader_LdArg3<ConT, VarT, T_To1, T_To2, T_From>  ^ ToGui(ConT ^ control, VarT & var, T_To1 arg_to1, T_To2 arg_to2, T_From arg_from)
{
    return gcnew ControlLoader_LdArg3<ConT, VarT, T_To1, T_To2, T_From>(control, var, arg_to1, arg_to2, arg_from);
}


} // namespace Innovative

#endif
