//
// File:       Event_Win.h
// Class:      Event
// Descrptn:   Classes to "Simplify" windows structures
// Author:     Christopher Smith (AIC)
//


#ifndef WinEvent_WinH
#define WinEvent_WinH

#include <windows.h>

namespace Innovative
{

namespace Windows
{

#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//---------------------------------------------------------------------------
//  Class constants
//---------------------------------------------------------------------------
//      Constants for Set() function
//
const BOOL EventSet   = true;
const BOOL EventReset = false;

//===========================================================================
//   CLASS Event
//===========================================================================
//
//  The Event class is designed to ease the use of a Windows Event object.
//  Putting it into a class also makes releasing resources less of a 
//  concern for the user.

class Event
{
public:
//
//  Member Functions
//
    Event(LPSECURITY_ATTRIBUTES lpsa = NULL, BOOL ManRst = TRUE, 
                     BOOL InitState = FALSE, LPCTSTR Name = NULL)
        : hEvt(NULL),
        Owned( true )
        {
            //
            //  Create a windows Event object to "go with" the object
            //     Accepting Defaults gives Manual Reset Event
            hEvt = CreateEventA(lpsa, ManRst, InitState, Name);
            //  If this failed, the hEvt is NULL. 
            //     This is our "Bad Handle" indicator.
            //
            //  NOTE: The default event is NOT inheritable, but
            //     you may be able to make one that is. The copy
            //     constructor will force the copy to be uninheritable.
            //     A true copy of a Inheritable event can't be made.
         };
    // Gets ownership of an existing event.
    Event( HANDLE _hEvt, bool _Owned=false ) : hEvt( _hEvt ), Owned( _Owned )  {}

    //
    //  Copy constructor.
    Event( const Event &originalEvent) :
        Owned( true )
        {  Dup(originalEvent);  }

    ~Event()
        {
        if (hEvt!=NULL)
            if( Owned )
                ::CloseHandle(hEvt);
                
        hEvt = NULL;
        };

    Event & operator = (const Event & anEvent)
        {
        HANDLE temp = hEvt;      // Save original Handle
        Dup(anEvent);            // Clone its Handle
        if (temp!=NULL)
            ::CloseHandle(temp); // Close the old handle
        return *this;
        };

    BOOL Set(BOOL toState = EventSet)
        {
        //
        //  Set or reset the event manually
        if (toState)
            return ::SetEvent(hEvt);
        else
            return ::ResetEvent(hEvt);
        }

    bool Set( bool toState )
        {
		return Set((BOOL)toState) ? true : false;
        }

    bool Clear ()
        {
		return Set( EventReset ) ? true : false;
        }

    bool Get()
        {
        //
        //  Gets state of object by waiting no time at all.
        //    event is signalled only if WAIT_OBJECT_0 is returned
        DWORD state = WaitFor(0);
        return state == WAIT_OBJECT_0;
        }

    DWORD WaitFor(DWORD Timeout = INFINITE )
        {  return ::WaitForSingleObject( hEvt, Timeout); };  // Timeout in ms

    operator HANDLE() const
        { return hEvt; };
        
    HANDLE GetHandle() const
        { return hEvt; }
    HANDLE Handle() const
        { return hEvt; }

private:
//
//  Data Members
//
    HANDLE  hEvt;
    bool    Owned;
    //
    //  Duplicate an event Handle.
    void Dup( const Event &originalEvent)
        {
        HANDLE thisProcess = GetCurrentProcess();
        if (!::DuplicateHandle(thisProcess,originalEvent, // original event
                             thisProcess,&hEvt,         // new handle
                             STANDARD_RIGHTS_REQUIRED,  // Ignored, since cloning
                             FALSE,                     // NOT Inheritable!
                             DUPLICATE_SAME_ACCESS))    // clone access
            hEvt = NULL;
        //
        //  If call failed, make sure our handle is NULLed.
        };
};

#ifdef __CLR_VER
#pragma managed(pop)
#endif

} // Windows

} // Innovative

#endif
