// FileFindThread_Win.cpp
//
//  Class to extract version information from DLLs and EXEs
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2003 Innovative Integration Corporation.
//  All Rights Reserved.
//
//   Background file search
//
//---------------------------------------------------------------------------

#include <string>
#include "FileFindThread_Win.h"
#include "FileFind_Win.h"
#include <FileSupport_Mb.h>
#include <StringSupport_Mb.h>
#include <Exception_Mb.h>

using namespace std;
namespace Innovative
{

//===========================================================================
//  CLASS  FileFindThread   ---  File finder support thread
//===========================================================================

//---------------------------------------------------------------------------
//  constructor for class FileFindThread
//---------------------------------------------------------------------------
// Thread is initially suspended.  Caller performs deletion.

FileFindThread::FileFindThread(FileFind *owner)
    : Owner(owner)
{
	// Priorities
    Priority(Thread::tpLower);

	Resume();
}

//---------------------------------------------------------------------------
//  destructor for class FileFindThread
//---------------------------------------------------------------------------

FileFindThread::~FileFindThread()
{
}

//------------------------------------------------------------------------------
//  FileFindThread::Init() -- Init multi-objects
//------------------------------------------------------------------------------

void FileFindThread::Init()
{
    //
    //  Initialize Actions with the objects we'll be waiting for:
    inherited::Init();
    AddObject(Scanning, sScan);
}

//---------------------------------------------------------------------------
//  FileFindThread::Start() -- Start scanning
//---------------------------------------------------------------------------

void FileFindThread::Start(const StringList & directories,
                           const StringList & filespecs, bool recursive)
{
    Directories = directories;
    FileSpecs = filespecs;
    Recursive = recursive;

    Abort = false;
    Scanning.Set();
}

//---------------------------------------------------------------------------
//  FileFindThread::Stop() -- Stop scanning
//---------------------------------------------------------------------------

void FileFindThread::Stop()
{
    Abort = true;
}

//---------------------------------------------------------------------------
// FileFindThread::ScanDirectories() -- Scan for subdirectories in specified directory
//---------------------------------------------------------------------------

void FileFindThread::ScanDirectories(string directory)
{
    if (Abort)
        return;

    // Spec is all files in specified dir
    directory += (directory[directory.size()-1] != '\\') ? "\\*.*" : "*.*";
    PathSpec spec(directory);
    string Path = spec.Drive() + spec.Dir();

    // Open the directory
    WIN32_FIND_DATA FileData;
    HANDLE Handle = FindFirstFile(directory.c_str(), &FileData);

    // Exit if bad directory name
    if (Handle == INVALID_HANDLE_VALUE)
        return;

    // Scan dir names from this directory into stringlist
    while (!Abort && FindNextFile(Handle, &FileData))
        {
        string Filename = FileData.cFileName;
        if (FileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
            {
            //
            //  Found directory
            if (Filename[0] != '.')
                {
                string Name = Path+Filename;

                Owner->HandleDirectoryFound(Name);

                if (Recursive)
                    ScanDirectories(Name);
                }
            }
        else
            {
            //
            //  Found a file, try and match versus all specs
            //
            //  Match with regular expressions
            for (StringList::iterator i = FileSpecs.begin(); i != FileSpecs.end(); ++i)
                {
                if (WildcardCompare(LowerCase(*i).c_str(), LowerCase(Filename).c_str()))
                    OnMatch(Path, Filename);
                }
            }
        }

    // Finished with this directory, close handle
    FindClose(Handle);
}

//---------------------------------------------------------------------------
// FileFindThread::OnMatch() -- Process a match
//---------------------------------------------------------------------------

void FileFindThread::OnMatch(const string & Path, const string & Filename)
{
    string Name = Path+Filename;

    if (!Owner->AllowDuplicates())
        if (Owner->Files().IndexOf(Name) >= 0)
            return;

    Owner->HandleFileFound(Name);
}

//---------------------------------------------------------------------------
//  FileFindThread::Scan() -- Perform scan
//---------------------------------------------------------------------------

void FileFindThread::Scan()
{
    for (StringList::iterator i = Directories.begin(); !Abort && (i != Directories.end()); ++i)
        ScanDirectories(*i);
}

//---------------------------------------------------------------------------
//  FileFindThread::Execute()  -- Idle behavior of file finder support thread
//---------------------------------------------------------------------------

void FileFindThread::Execute(void)
{
    while (!Terminated())
        {
        // WaitResult is either a) Suicide event, b) Data written event
        int WaitResult = WaitFor();

        switch (WaitResult)
            {
            //  Terminate:  Exit thread
            case sTerminate:
                Shutdown(false);
                Scanning.Clear();
                break;

            //  Log data:  Commit enqueued data to disk file
            case sScan:
                try
                    {
                    Scan();

                    Scanning.Clear();
                    Owner->HandleScanFinished(Abort);
                    }
                catch (...)
                    {
                    }
                break;
            }
        }
}

//---------------------------------------------------------------------------
//  FileFindThread::WildcardCompare()  -- Wildcard string match
//---------------------------------------------------------------------------

int FileFindThread::WildcardCompare(const char *wild, const char *string)
{
    const char *cp = NULL, *mp = NULL;

    while ((*string) && (*wild != '*'))
        {
        if ((*wild != *string) && (*wild != '?'))
            return 0;

        ++wild;
        ++string;
        }

    while (*string)
        {
        if (*wild == '*')
            {
            if (!*++wild)
                return 1;

            mp = wild;
            cp = string+1;
            }
        else if ((*wild == *string) || (*wild == '?'))
            {
            wild++;
            string++;
            }
        else
            {
            wild = mp;
            string = cp++;
            }
        }

    while (*wild == '*')
        ++wild;

    return !*wild;
}

} // namespace Innovative


