// MailSlot_Win.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2003 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <Framework/Windows/Mailslot_Win.h>
#include <Framework/Windows/Mutex_Win.h>
#include <DataTypes_Mb.h>
#include <Thread_Mb.h>
#include <sstream>
#include <iostream>
#include <algorithm>

#if defined(WIN32) || defined(WIN64)
#include <windows.h>
#endif



namespace Innovative
{

//===========================================================================
//  CLASS  SlotThread   ---  Mail slot background thread
//===========================================================================

class SlotThread : public Thread
{
    friend class MailSlot;

public:
    // Ctor
    SlotThread(MailSlot * owner)
        : Owner(owner)
        {
            // Priorities
            Priority(tpNormal);
            Resume();
        }

protected:
    // Methods
    virtual void    Execute();

private:
    // Data
    MailSlot *      Owner;
};

//---------------------------------------------------------------------------
//  SlotThread::Execute() --
//---------------------------------------------------------------------------

void SlotThread::Execute()
{
    while(!Terminated())
        {
        MailSlot::CharArray Buffer;
        if (!Owner->Read(Buffer))
            continue;
        if (!Buffer.size())
            continue;

        MailSlotMessage Msg(Buffer);
        if (!Msg.Body.size())
            continue;

        // Notify the recipient via event            
        Owner->OnMessage.Execute(Msg);
        }
}


static const int MAX_MAILSLOT_MSG_SIZE = 65535;
static const int MAILSLOT_TIMEOUT = 300;

//===========================================================================
//  CLASS  MailSlotMessage   ---
//===========================================================================

MailSlotMessage::MailSlotMessage(CharArray & msg)
{
    // Parse computer name from message
    CharArray::iterator lPos = std::find(msg.begin(), msg.end(), ';');
    if(lPos == msg.end())
        return;
    size_t len = lPos - msg.begin();
    Computer.append(&msg[0], len);
    msg.erase(msg.begin(), ++lPos);

    // Parse user name from message
    lPos = std::find(msg.begin(), msg.end(), ';');
    if(lPos == msg.end())
        return;
    len = lPos - msg.begin();
    User.append(&msg[0], len);
    msg.erase(msg.begin(), ++lPos);

    // Parse Id from message
    lPos = std::find(msg.begin(), msg.end(), ';');
    if(lPos == msg.end())
        return;
    len = lPos - msg.begin();
    Id.append(&msg[0], len);
    msg.erase(msg.begin(), ++lPos);

    Body = msg;
}

//===========================================================================
//  CLASS  MailSlot   ---  Mail slot component
//===========================================================================

//---------------------------------------------------------------------------
//  MailSlot::MailSlot() --
//---------------------------------------------------------------------------

 MailSlot::MailSlot()
    : FSlotName(""), FServerMailSlot(0), FClientMailSlot(0), Mailer(0),
      Scratch(MAX_MAILSLOT_MSG_SIZE)
{
    char name[MAX_COMPUTERNAME_LENGTH + 1];
    DWORD size = sizeof(name);

    if(GetComputerNameA(name, &size))
        FComputerName = name;

    size = sizeof(name);
    if(GetUserNameA(name, &size))
        FUserName = name;
}
 
//---------------------------------------------------------------------------
//  MailSlot::~MailSlot() --
//---------------------------------------------------------------------------

 MailSlot::~MailSlot()
{
    CloseClient();
    CloseServer();
}

//---------------------------------------------------------------------------
//  MailSlot::OpenServer() --
//---------------------------------------------------------------------------

bool MailSlot::OpenServer()
{
    if(FServerMailSlot)
        return false;

    std::string MailSlotName = "\\\\.\\mailslot\\" + FSlotName;
    FServerMailSlot = ::CreateMailslot(MailSlotName.c_str(), 0,
        MAILSLOT_TIMEOUT, NULL);
    if(FServerMailSlot == INVALID_HANDLE_VALUE)
        {
        FServerMailSlot = 0;
        return false;
        }

    delete Mailer;
    Mailer = new SlotThread(this);

    return true;
}

//---------------------------------------------------------------------------
//  MailSlot::OpenClient() --
//---------------------------------------------------------------------------

bool MailSlot::OpenClient(std::string Target, std::string Slot)
{
    // Open the server
    std::string MailSlotName = "\\\\" + Target + "\\mailslot\\" + Slot;
    FClientMailSlot = CreateFileA(MailSlotName.c_str(), GENERIC_WRITE,
        FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

    if(FClientMailSlot == INVALID_HANDLE_VALUE)
        {
        FClientMailSlot = 0;
        return false;
        }

    return true;
}

//---------------------------------------------------------------------------
//  MailSlot::OpenClient() --
//---------------------------------------------------------------------------

bool MailSlot::OpenClient(std::string Slot)
{
    // Check if server available
    Innovative::Windows::Mutex ServerLock(NULL, false, Slot.c_str());
    if (ServerLock.Acquire(0) != WAIT_TIMEOUT)
        {
        ServerLock.Release();
        return false;
        }

    return OpenClient(".", Slot);
}

//---------------------------------------------------------------------------
//  MailSlot::CloseServer() --
//---------------------------------------------------------------------------

bool MailSlot::CloseServer(void)
{
    if(FServerMailSlot)
        {
        if(Mailer)
            {
            Mailer->Shutdown();
            delete Mailer;
            Mailer = 0;
            }
        CloseHandle(FServerMailSlot);
        FServerMailSlot = 0;
        }

    return true;
}

//---------------------------------------------------------------------------
//  MailSlot::CloseClient() --
//---------------------------------------------------------------------------

bool MailSlot::CloseClient(void)
{
    if(FClientMailSlot)
        {
        CloseHandle(FClientMailSlot);
        FClientMailSlot = 0;
        }

    return true;
}

//---------------------------------------------------------------------------
//  MailSlot::Write() --
//---------------------------------------------------------------------------

bool MailSlot::Write(CharArray & msg)
{
    if(!FClientMailSlot)
        return false;

    std::stringstream hdr;
    hdr << FComputerName << ";" << FUserName << ";"
        << hdr.precision(9) << Rtc.Now() << ";";
    hdr.write(&msg[0], msg.size());

    DWORD BytesWritten;
    DWORD Size = static_cast<DWORD>(hdr.str().size());
	WriteFile(FClientMailSlot, hdr.str().c_str(), Size, &BytesWritten, NULL);

    return BytesWritten == static_cast<DWORD>(hdr.str().size());
}

//---------------------------------------------------------------------------
//  MailSlot::Write() --
//---------------------------------------------------------------------------

bool MailSlot::Write(std::string msg)
{
    if(!FClientMailSlot)
        return false;

    std::stringstream hdr;
    hdr << FComputerName << ";" << FUserName << ";"
        << hdr.precision(9) << Rtc.Now() << ";" << msg;

    DWORD BytesWritten;
    DWORD Size = static_cast<DWORD>(hdr.str().size());
	WriteFile(FClientMailSlot, hdr.str().c_str(), Size, &BytesWritten, NULL);

    return BytesWritten == static_cast<DWORD>(hdr.str().size());
}

//---------------------------------------------------------------------------
//  MailSlot::Read() --
//---------------------------------------------------------------------------

bool MailSlot::Read(CharArray & buffer)
{
    if (FServerMailSlot)
        {
        DWORD BytesRead;
        if (ReadFile(FServerMailSlot, &Scratch[0], static_cast<DWORD>(Scratch.size()), &BytesRead, NULL))
            {
            buffer.resize(BytesRead);
            memcpy(&buffer[0], &Scratch[0], BytesRead);
            }

        return true;
        }

    return false;
}

} // namespace Innovative




