//  MultiObject_Win.cpp   --  'MultipleObjects'... a Multiple Object Wait class
//
//  This concrete class encapsulates a WaitForMultipleObjects. It contains
//  member functions to specify a case statement (DWORD) for every event to wait
//  upon.
//
//  Currently, the IsValid() function only assures that a handle was
//    loaded for every declared handle in the object.
//  This class does not test for redundant handles!
//

#include <Framework/Windows/MultiObject_Win.h>

#ifdef __GNUC__
    #pragma GCC diagnostic ignored "-Wtype-limits"
#endif

namespace Innovative
{

namespace Windows
{

//===========================================================================
//  CLASS  MultipleObjects  --  Wait on multiple events & objects
//===========================================================================

//---------------------------------------------------------------------------
//  constructor for class MultipleObjects
//---------------------------------------------------------------------------

MultipleObjects::MultipleObjects(DWORD NumHandles)
    : isValid(false), Position(0), lastError(ERROR_SUCCESS),
      HandleCount(NumHandles), UseMsg(false)
{
    Init(NumHandles);
}

//---------------------------------------------------------------------------
//  constructor for class MultipleObjects
//---------------------------------------------------------------------------

MultipleObjects::MultipleObjects(DWORD NumHandles, int Case, DWORD WaitMask)
    : isValid(false), Position(0), lastError(ERROR_SUCCESS),
      HandleCount(NumHandles), Mask(WaitMask), UseMsg(true), MsgCase(Case)
{
    Init(NumHandles);
}

//---------------------------------------------------------------------------
//  destructor for class MultipleObjects
//---------------------------------------------------------------------------

MultipleObjects::~MultipleObjects()
{
    delete [] Handles;
    delete [] Cases;
    delete [] ValidHandles;
}

//---------------------------------------------------------------------------
//  MultipleObjects::Init()  --  Common object initialization
//---------------------------------------------------------------------------

void MultipleObjects::Init(DWORD /*NumHandles*/)
{
    Handles      = new HANDLE[HandleCount];
    Cases        = new int[HandleCount];
    ValidHandles = new bool[HandleCount];
    if ( ValidHandles )
        {
        for (int count=0; count<HandleCount; count++)
            ValidHandles[count]=false;
        }
}

//---------------------------------------------------------------------------
//  MultipleObjects::Load()  --  Load an object handle
//---------------------------------------------------------------------------

bool  MultipleObjects::Load(HANDLE Handle, int Position, int Case)
{
    ValidHandles[Position] = false;
    //
    //  Verify load
    if (Handle==NULL || Position>=HandleCount)
         return false;
    //
    //  It's OK, Load it
    Handles[Position] = Handle;
    ValidHandles[Position] = true;
    Cases[Position] = Case;
    //
    //  Test Validity
    isValid = true;
    for (int count=0; count< HandleCount; count++)
        {
        if (ValidHandles[count]==false)
            {
            isValid=false;
            break;
            }
         }

    return true;
}

//---------------------------------------------------------------------------
//  MultipleObjects::WaitFor()  --  wait on the loaded objects
//---------------------------------------------------------------------------
//  This function returns the pre-defined case constant loaded when the
//     object was loaded.
//

DWORD  MultipleObjects::WaitFor(DWORD Timeout, bool WaitAll)
{
    DWORD result;
    //
    //  Must have loaded all objects before waiting
    if (isValid==false)
        {
        lastError=ERROR_INVALID_HANDLE;
        return WAIT_FAILED;
        }
    //
    //  Wait on multiple objects, either all firing or any one
    if (UseMsg)
        {
        result = MsgWaitForMultipleObjects(HandleCount, Handles, WaitAll,
                                           Timeout, Mask);
        }
    else
        {
        result = WaitForMultipleObjects(HandleCount, Handles, WaitAll, Timeout);
        }

    //
    //  If successful wait, return Case for object that fired
    if ((result>=WAIT_OBJECT_0) && (result<=(WAIT_OBJECT_0+(DWORD)HandleCount)))
        {
        Position = result-WAIT_OBJECT_0;

        // Special handling for Windows message signal
        if (Position==HandleCount)
            return MsgCase;
        // Return user-supplied case associated with signalled condition
        return( Cases[Position] );
        }
    //
    //  If Timed out, return that fact
    if (result==WAIT_TIMEOUT)
        return WAIT_TIMEOUT;
    //
    //  If wait failed, grab error code save it.
    if (result==WAIT_FAILED)
        {
        lastError=::GetLastError();
        return WAIT_FAILED;
        }
    //
    //  If wait abandoned, save position and return abandon code
    if ( (result>=WAIT_ABANDONED_0) && (result<(WAIT_ABANDONED_0+(DWORD)HandleCount)) )
        {
        Position = result-WAIT_ABANDONED_0;
        return WAIT_ABANDONED;
        }
   //
   //  Don't know what happened, but it was bad
   return 0;
}

//------------------------------------------------------------------------------
//  MultipleObjects::Get() -- Test object state
//------------------------------------------------------------------------------

bool MultipleObjects::Get()
{
    //
    //  Gets state of object by waiting no time at all.
    DWORD state = WaitFor(0);
    return (( state != WAIT_FAILED ) && ( state != WAIT_TIMEOUT	 ));
}

//------------------------------------------------------------------------------
//  MultipleObjects::LastPosition() -- Returns position of last signaled object
//------------------------------------------------------------------------------

int MultipleObjects::LastPosition()
{
    return Position;
}

//------------------------------------------------------------------------------
//  MultipleObjects::LastError() -- Returns ID of last signaled object
//------------------------------------------------------------------------------

DWORD MultipleObjects::LastError()
{
    return lastError;
}

//------------------------------------------------------------------------------
//  MultipleObjects::LastCase() -- Returns ID of last signaled object
//------------------------------------------------------------------------------

DWORD MultipleObjects::LastCase()
{
    return (Position==HandleCount) ? MsgCase : Cases[Position];
}

//------------------------------------------------------------------------------
//  MultipleObjects::IsValid() -- Returns true if all handles set
//------------------------------------------------------------------------------

bool MultipleObjects::IsValid()
{
    return isValid;
}


} // Windows

} // Innovative




