// PipeThread_Win.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

// Note:  This implementation is Windows-Specific

#include <Framework/Windows/PipeThread_Win.h>
#include <Thread_Mb.h>
#include <string>
#include <BufferDatagrams_Mb.h>
#include <BufferHeader_Mb.h>
#include <Framework/Windows/Pipe_Win.h>
#include <Framework/Windows/Event_Win.h>
#include <Framework/Windows/MultiObject_Win.h>
#include <Framework/Windows/Overlapped_Win.h>
#include <Mutex_Mb.h>
#include <algorithm>

namespace   Innovative
{

//===========================================================================
//  CLASS  PipeImpl   ---  Named pipe implementation thread
//===========================================================================

class PipeImpl : public Thread
{
    typedef Thread inherited;
    
public:
    // Ctor
    PipeImpl(PipeBase * owner);
    ~PipeImpl();

    // Methods
    bool                Open();
    int                 Write(const Buffer & buffer);
    int                 Write(const std::string & msg);

    bool                Connected() const;
    bool                Opened() const
        {  return (Pipe != INVALID_HANDLE_VALUE);  }

protected:
    void                Init();
    void                Execute();

private:
    enum IIState {sIdle, sHead, sBody};

    // Data
    PipeBase *          Owner;
    unsigned long       BytesRead;
    bool                FConnected;
    std::string         PipeSpec;
    Buffer              Data;
    IIState             ReadState;


    enum { sSuicide=inherited::sLast, sConnecting, sReading, sLast };
    HANDLE                                  Pipe;
    Innovative::Windows::MultipleObjects    Action;
    Innovative::Windows::Event              Suicide;
    Innovative::Windows::Event              Reading;
    Innovative::Windows::Event              Writing;
    Innovative::Windows::Event              Connecting;

    Innovative::Windows::Overlapped         Connector;
    Innovative::Windows::Overlapped         Writer;
    Innovative::Windows::Overlapped         Reader;

private:
	Mutex									WriteLock;

    // Methods
    void                PostRead();
    void                DoRead();
    void                DoConnect();

    void                OpenServer();
    void                OpenClient();
};

//---------------------------------------------------------------------------
//  PipeImpl::PipeImpl() --  Ctor
//---------------------------------------------------------------------------

PipeImpl::PipeImpl(PipeBase * owner)
    : inherited("PipeImpl"), Owner(owner), BytesRead(0),
      FConnected(false), ReadState(sIdle),
      Pipe(INVALID_HANDLE_VALUE), Action(3)
{
    // Initialize
    Reader.SetEvent(Reading);
    Writer.SetEvent(Writing);
    Connector.SetEvent(Connecting);

    if (Open())
        Resume();
}

//---------------------------------------------------------------------------
//  PipeImpl::PipeImpl() --  Dtor
//---------------------------------------------------------------------------

PipeImpl::~PipeImpl()
{
    if (Pipe != INVALID_HANDLE_VALUE)
        {
        Shutdown(false);

        Suicide.Set();
        while (Suicide.Get())
            Sleep(1);

        CloseHandle(Pipe);
        Pipe = INVALID_HANDLE_VALUE;
        }

    Shutdown(true);
}

//---------------------------------------------------------------------------
//  PipeImpl::Init() --
//---------------------------------------------------------------------------

void PipeImpl::Init()
{
    inherited::Init();

    //  Initialize the Action object with the events we'll be waiting for:
    Action.Load(Suicide.Handle(), 0, 0);
    Action.Load(Connecting.Handle(), 1, 1);
    Action.Load(Reading.Handle(), 2, 2);
}

//---------------------------------------------------------------------------
//  PipeImpl::Write() --
//---------------------------------------------------------------------------

int PipeImpl::Write(const Buffer & buffer)
{
    if (Pipe == INVALID_HANDLE_VALUE)
        return 0;

    WriteLock.Acquire();   // one client at a time please

    // Head
    PacketBufferHeader Head(buffer);
    CharDG Body(buffer);
    Head.DataSize(Body.SizeInInts());
    DWORD BytesWritten;
	WriteFile(Pipe, &Head[0], static_cast<DWORD>(Head.SizeInBytes()), &BytesWritten, &Writer.Handle());
    GetOverlappedResult(Pipe, &Writer.Handle(), &BytesWritten, true);

    // Body
	WriteFile(Pipe, &Body[0], static_cast<DWORD>(Body.SizeInBytes()), &BytesWritten, &Writer.Handle());
    GetOverlappedResult(Pipe, &Writer.Handle(), &BytesWritten, true);

    WriteLock.Release();
    return BytesWritten;
}

//---------------------------------------------------------------------------
//  PipeImpl::Write() --
//---------------------------------------------------------------------------

int PipeImpl::Write(const std::string & msg)
{
    Buffer txt;

    CharDG Msg(txt);
    size_t size = msg.size();
    Msg.Resize(size);

//    std::copy(msg.begin(), msg.end(), Msg.begin());

    for (size_t i = 0; i < size; ++i)
        Msg[i] = msg[i];

    PacketBufferHeader Head(txt);
    Head.PeripheralId(0xff);

    return Write(txt);
}

//---------------------------------------------------------------------------
//  PipeImpl::DoConnect() --
//---------------------------------------------------------------------------

void PipeImpl::DoConnect()
{
    Connecting.Clear();

    FConnected = false;

    // Post another connection
    ConnectNamedPipe(Pipe, &Connector.Handle());
    volatile int Error = GetLastError();
    switch(Error)
        {
        default:
        case ERROR_PIPE_LISTENING:
            break;
        case ERROR_NO_DATA:
            FlushFileBuffers(Pipe);
            DisconnectNamedPipe(Pipe);
            ConnectNamedPipe(Pipe, &Connector.Handle());
            break;
        case ERROR_PIPE_CONNECTED:
            FConnected = true;
            // Post a read
            ReadState=sIdle;
            PostRead();
            break;
        }

    ProcessCompletionEvent event(FConnected);
    Owner->OnConnect.Execute(event);
}

//---------------------------------------------------------------------------
//  PipeImpl::PostRead() --
//---------------------------------------------------------------------------

void PipeImpl::PostRead()
{
    PacketBufferHeader Head(Data);

    switch (ReadState)
        {
        case sIdle:
            {
            // Read header
            ReadState = sHead;
            if (!ReadFile(Pipe, &Head[0], static_cast<DWORD>(Head.SizeInBytes()), &BytesRead, &Reader.Handle()))
				{
				ProcessStatusEvent e("Pipe head read error");
				Owner->OnError.Execute(e);
				}
            }
            break;
        case sHead:
            {
            // Body
            ReadState = sBody;
            CharDG Body(Data);
            Body.Resize( CouldHold<char>(Head.DataSize()) );
            if (!ReadFile(Pipe, &Body[0], static_cast<DWORD>(Body.SizeInBytes()), &BytesRead, &Reader.Handle()))
				{
				ProcessStatusEvent e("Pipe body read error");
				Owner->OnError.Execute(e);
				}
            }
            break;
        case sBody:
            ReadState = sIdle;

            if (Head.PeripheralId() == 0xff)
                {
                CharDG Msg(Data);
                std::string msg(&Msg[0], Msg.size());

                ProcessStatusEvent event(msg);
                Owner->OnMessage.Execute(event);
                }
            else
                {
                Buffer temp(Data);
                temp.MakeUnique();
                temp.MakeUniqueHeader();

                PipeRcvEvent event(temp);
                Owner->OnRead.Execute(event);
                }

            PostRead();
            break;

        }
}

//---------------------------------------------------------------------------
//  PipeImpl::DoRead() --
//---------------------------------------------------------------------------
//  This named pipe class posts a large read that will usually not be
//     filled by the sender. We assume that the notification we get fulfills
//     our posted read, and output that data.
//  So if the sender does 3 WRITES, we should output 3 reads in response

void PipeImpl::DoRead()
{
    Reading.Clear();

	bool ok = GetOverlappedResult(Pipe, &Reader.Handle(), &BytesRead, false) ? true : false;
    //
    //  Determine if our Read was a success
    if (!ok)
        {
        volatile int Error = GetLastError();
        switch(Error)
            {
            default:
                {
                ProcessStatusEvent e("Pipe read error");
                Owner->OnError.Execute(e);
                }
                break;
            case ERROR_IO_PENDING:
                {
                ProcessStatusEvent e("Unexpected Pipe I/O");
                Owner->OnError.Execute(e);
                }
                break;
            case ERROR_BROKEN_PIPE:
                CancelIo(Pipe);
                DoConnect();
                return;
            case ERROR_MORE_DATA:
                ProcessStatusEvent e("ERROR_MORE_DATA");
                Owner->OnError.Execute(e);
                break;
            }
        }

    // Initiate the next overlapped read
    PostRead();
}

//---------------------------------------------------------------------------
//  PipeImpl::Execute() --
//---------------------------------------------------------------------------

void PipeImpl::Execute()
{
    // Overlapped connection
    if (Owner->FServer)
        ConnectNamedPipe(Pipe, &Connector.Handle());
    else
        PostRead();

    while(!Terminated())
        {
        // WaitResult is either connection, read or write
        unsigned int WaitResult = Action.WaitFor();

        switch (WaitResult)
            {
            //  Termination
            default:
            case 0:                         break;
            //  Client Connection:
            case 1:     DoConnect();        break;
            //  Data read complete:
            case 2:     DoRead();           break;
            }
        }

    FlushFileBuffers(Pipe);
    DisconnectNamedPipe(Pipe);

    Suicide.Clear();
}

//---------------------------------------------------------------------------
//  PipeImpl::Connected() --
//---------------------------------------------------------------------------

bool PipeImpl::Connected() const
{
    return FConnected;
}

//---------------------------------------------------------------------------
//  PipeImpl::Open() --
//---------------------------------------------------------------------------

bool PipeImpl::Open()
{
    PipeSpec = "\\\\" +  Owner->FRemoteMachineName + "\\pipe\\" + Owner->FPipeName;

    if (Owner->FServer)
        OpenServer();
    else
        OpenClient();

    if (Opened())
        return true;

    LPVOID lpMsgBuf;

    FormatMessageA(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        GetLastError(),
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
        (LPTSTR) &lpMsgBuf,
        0,
        NULL
    );

    std::string error((char*)lpMsgBuf);

    // Free the buffer.
    LocalFree( lpMsgBuf );

    Pipe = INVALID_HANDLE_VALUE;

    // Display the std::string.
    ProcessStatusEvent event(error);
    Owner->OnError.Execute(event);

    return false;
}

//---------------------------------------------------------------------------
//  PipeImpl::OpenServer() --
//---------------------------------------------------------------------------

void PipeImpl::OpenServer()
{
    // The buffer size below is advise to the OS, but may not be respected.
    const DWORD BufferSize = 0x100000;

    // If client and server run under different user contexts,
    // access denied will happen. This is because of the NULL passed as
    // sec attributes in CreateNamedPipe. A NULL means that only the same
    // user or administrators can connect to it.
    //
    // Instead of NULL as security attributes, pass a security attributes
    // with a DACL that allows permission for the user running the
    // client program.
    // The following code sets up the pipe to allow access for everyone.
    SECURITY_ATTRIBUTES m_pSecAttrib;
    SECURITY_DESCRIPTOR* m_pSecDesc;

    m_pSecDesc = reinterpret_cast<SECURITY_DESCRIPTOR*>(LocalAlloc(LPTR, SECURITY_DESCRIPTOR_MIN_LENGTH));

    InitializeSecurityDescriptor(m_pSecDesc, SECURITY_DESCRIPTOR_REVISION);

    SetSecurityDescriptorDacl(m_pSecDesc, true, (PACL)NULL, false);

    m_pSecAttrib.nLength = sizeof(SECURITY_ATTRIBUTES);
    m_pSecAttrib.bInheritHandle = true;
    m_pSecAttrib.lpSecurityDescriptor = m_pSecDesc;

    Pipe = CreateNamedPipe(
        PipeSpec.c_str(),               // pointer to pipe name
        PIPE_ACCESS_DUPLEX
        | FILE_FLAG_OVERLAPPED,         // pipe open mode
        PIPE_TYPE_MESSAGE |             // message type pipe
        PIPE_READMODE_MESSAGE |         // message-read mode
        PIPE_WAIT,                      // blocking mode
        PIPE_UNLIMITED_INSTANCES,	    // maximum number of instances
        BufferSize,                     // output buffer size, in bytes
        BufferSize,	                    // input buffer size, in bytes
        NMPWAIT_WAIT_FOREVER,           // time-out time, in milliseconds
        &m_pSecAttrib);                 // security DECL access for Everyone

    LocalFree(m_pSecDesc);
}

//---------------------------------------------------------------------------
//  PipeImpl::OpenClient() --
//---------------------------------------------------------------------------

void PipeImpl::OpenClient()
{
    Pipe = CreateFileA(PipeSpec.c_str(), GENERIC_READ | GENERIC_WRITE,
        FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
        FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED, NULL);
}

//===========================================================================
//  CLASS  PipeSupport   ---  Named pipe support class
//===========================================================================

//---------------------------------------------------------------------------
//  PipeSupport::PipeSupport() --
//---------------------------------------------------------------------------

PipeSupport::PipeSupport(PipeBase * owner)
    : Impl(new PipeImpl(owner))
{
}

//---------------------------------------------------------------------------
//  PipeSupport::PipeSupport() --
//---------------------------------------------------------------------------

PipeSupport::~PipeSupport()
{
    delete Impl;
    Impl = 0;
}

//---------------------------------------------------------------------------
//  PipeSupport::Write() --
//---------------------------------------------------------------------------

int PipeSupport::Write(const Buffer & buffer)
{
    return Impl->Write(buffer);
}

//---------------------------------------------------------------------------
//  PipeSupport::Write() --
//---------------------------------------------------------------------------

int PipeSupport::Write(const std::string & msg)
{
    return Impl->Write(msg);
}

//---------------------------------------------------------------------------
//  PipeSupport::Connected() --
//---------------------------------------------------------------------------

bool PipeSupport::Connected() const
{
    return Impl->Connected();
}

//---------------------------------------------------------------------------
//  PipeSupport::Opened() --
//---------------------------------------------------------------------------

bool PipeSupport::Opened() const
{
    return Impl->Opened();
}

}   // Innovative



