// Registry_Win.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2003 Innovative Integration Corporation.
//  All Rights Reserved.
//
#include <Framework/Windows/Registry_Win.h>
#include <sstream>
#include <Exception_Mb.h>
#include <algorithm>
#include <memory>
#include <StringSupport_Mb.h>
#include <algorithm>

namespace   Innovative
{

//using namespace std;

//---------------------------------------------------------------------------
//  RegistryKey::RegistryKey() -- Ctor
//---------------------------------------------------------------------------

RegistryKey::RegistryKey(const HKEY & root, const std::string & key, bool create)
{
	FKey = key;
	bool predefined = key.find("HKEY_") == 0;

	//
	//  Move down into subkey, creating if need be
	long ret = ::RegOpenKeyEx(root, (predefined ? "" : key.c_str()), 0,
		create ? KEY_ALL_ACCESS : KEY_READ, &FHandle);

	if (create && (ret != ERROR_SUCCESS))
		{
		// Create key
		char classType[1];
		classType[0] = 0;

		DWORD   disposition = 0;
		ret = ::RegCreateKeyEx(root, key.c_str(), 0, classType,
			REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &FHandle, &disposition);

		if (ret != ERROR_SUCCESS)
			{
			FHandle = 0;
			return;
			}
		}
}

//---------------------------------------------------------------------------
//  RegistryKey::RegistryKey() -- Copy ctor
//---------------------------------------------------------------------------

RegistryKey::RegistryKey(const RegistryKey & other)
	: FKey(other.FKey)
{
	bool ok = DuplicateHandle(GetCurrentProcess(),
		other.FHandle, GetCurrentProcess(), reinterpret_cast<HANDLE*>(&FHandle),
		0, false, DUPLICATE_SAME_ACCESS) != 0;

	// Special case: Root keys are non-duplicatable
	if (!ok)
		FHandle = other.FHandle;
}

//---------------------------------------------------------------------------
//  RegistryKey::~RegistryKey() -- Dtor
//---------------------------------------------------------------------------

RegistryKey::~RegistryKey()
{
	::RegCloseKey(FHandle);
}

//---------------------------------------------------------------------------
//  RegistryKey::Handle() -- Return handle to registry key
//---------------------------------------------------------------------------

HKEY RegistryKey::Handle() const
{
	return FHandle;
}

//---------------------------------------------------------------------------
//  RegistryKey::Key() --  Return key name
//---------------------------------------------------------------------------

std::string RegistryKey::Key() const
{
	return FKey;
}

//=============================================================================
//  CLASS Registry  --
//=============================================================================

//---------------------------------------------------------------------------
// Registry::Registry() --  Ctor
//---------------------------------------------------------------------------

Registry::Registry(const HKEY & root, const std::string & key,
    const std::string & app, const std::string & org)
	: FApp(app), FOrganization(org)
{
    Fifo.push_back(RegistryKey(root, key, true));
	Home();
}

//---------------------------------------------------------------------------
// Registry::Registry() --  Dtor
//---------------------------------------------------------------------------

Registry::~Registry()
{
	Flush();
}

//---------------------------------------------------------------------------
//  Registry::Root() -- Return to root key HKEY_
//---------------------------------------------------------------------------

bool Registry::Root()
{
    bool result = true;
	while (Fifo.size() > 1)
        {
        result &= Ascend();
        }

	return result;
}

//---------------------------------------------------------------------------
//  Registry::Home() --  Return to home key HKEY_/Software/Application
//---------------------------------------------------------------------------

bool Registry::Home()
{
    Root();

	bool result = true;
	result &= Descend("Software", true);

	if (FOrganization.size())
		result &= Descend(FOrganization, true);

    if (FApp.size())
		result &= Descend(FApp, true);

	return result;
}

//---------------------------------------------------------------------------
//  Registry::Flush() --  Flush current key
//---------------------------------------------------------------------------

bool Registry::Flush()
{
	if (!Fifo.size())
		return true;
	else
		return RegFlushKey(CurrentKey()) == 0;
}

//---------------------------------------------------------------------------
//  Registry::CurrentKey() --  Return current key
//---------------------------------------------------------------------------

HKEY Registry::CurrentKey() const
{
	return Fifo.front().Handle();
}

//---------------------------------------------------------------------------
//  Registry::DefaultOrganization() --  Return default company name
//---------------------------------------------------------------------------

std::string Registry::DefaultOrganization() const
{
	return "Innovative Integration";
}

//---------------------------------------------------------------------------
//  Registry::Keys() --  Return list of keys
//---------------------------------------------------------------------------

StringList Registry::Keys()
{
	StringList result;

	DWORD sz;
	long ret = RegQueryInfoKey(
	  CurrentKey(), NULL, NULL, NULL, NULL, &sz,
	  NULL, NULL, NULL, NULL, NULL, NULL);

	if (ret)
		return result;

    std::string s;
	s.resize(sz);

	for (int i = 0; ; ++i)
		{
		DWORD size = static_cast<DWORD>(s.size());
		long ret = RegEnumKeyEx(CurrentKey(),
		  i, &s[0], &size, NULL, NULL, NULL, NULL);

        std::string s2(s, 0, size);
		result.Add(s2);

		if (ret != 0)
			break;
		}

	return result;
}

//---------------------------------------------------------------------------
//  Registry::Tags() --  Return list of tags
//---------------------------------------------------------------------------

StringList Registry::Tags()
{
	StringList result;

    std::string s;
	s.resize(32767);

	for (int i = 0; ; ++i)
		{
		DWORD size = static_cast<DWORD>(s.size());
		long ret = RegEnumValue(CurrentKey(),
		  i, &s[0], &size, NULL, NULL, NULL, NULL);

		if (ret == 0)
			{
            std::string s2(s, 0, size);
			result.Add(s2);
			}
		else
			break;
		}

	return result;
}

//---------------------------------------------------------------------------
//  Registry::DeleteKeyTree() -- Delete a key & all subkeys
//---------------------------------------------------------------------------
//  In WindowsNT the standard delete ftn does not kill subkeys. So descend
//     through the tree and delete all subkeys 'manually'
//
void  Registry::DeleteKeyTree(const std::string & key)
{
    //
    //  Move down into subkey, creating if need be
	Descend(key, true);
    //
	//  The subkey is now the current key
	//
	//  Get a list of subkeys and delete them all
    StringList list = Keys();
	for (StringList::iterator i = list.begin(); i != list.end(); ++i)
		DeleteKeyTree(*i);
	//
    //  Pop back out and delete the current key
	Ascend();
    DeleteKey(key);
}

//---------------------------------------------------------------------------
//  Registry::Write() --  Write string value to specified tag
//---------------------------------------------------------------------------

bool Registry::Write(const string & tag, const std::string & value)
{
	DWORD reserved = 0;
	long ret = ::RegSetValueEx(CurrentKey(), tag.c_str(), reserved, REG_SZ,
		reinterpret_cast<const BYTE *>(value.c_str()), static_cast<DWORD>(value.size()));

	return (ret == ERROR_SUCCESS);
}

//---------------------------------------------------------------------------
//  Registry::Write() --  Write null-terminated string to specified tag
//---------------------------------------------------------------------------

bool Registry::Write(const std::string & tag, const char * value)
{
	return Write(tag, std::string(value));
}

//---------------------------------------------------------------------------
//  Registry::Write() --  Write integer to specified tag
//---------------------------------------------------------------------------

bool Registry::Write(const std::string & tag, const int & value)
{
	DWORD reserved = 0;
	long ret = ::RegSetValueEx(CurrentKey(), tag.c_str(), reserved, REG_DWORD,
		reinterpret_cast < const BYTE * > (&value), sizeof(int));

	return (ret == ERROR_SUCCESS);
}

//---------------------------------------------------------------------------
//  Registry::Write() --  Write boolean to specified tag
//---------------------------------------------------------------------------

bool Registry::Write(const std::string & tag, const bool & value)
{
	int v = value ? 1 : 0;
	return Write(tag, v);
}

//---------------------------------------------------------------------------
//  Registry::Write() --  Write double to specified tag
//---------------------------------------------------------------------------

bool Registry::Write(const std::string & tag, const double & value)
{
	DWORD reserved = 0;
	long ret = ::RegSetValueEx(CurrentKey(), tag.c_str(), reserved, REG_BINARY,
		reinterpret_cast < const BYTE * > (&value), sizeof(double));

	return (ret == ERROR_SUCCESS);
}

//---------------------------------------------------------------------------
//  Registry::Write() --  Write 64-bit int to specified tag
//---------------------------------------------------------------------------

bool Registry::Write(const std::string & tag, const ii64 & value)
{
	DWORD reserved = 0;
	long ret = ::RegSetValueEx(CurrentKey(), tag.c_str(), reserved, REG_QWORD,
		reinterpret_cast < const BYTE * > (&value), sizeof(ii64));

	return (ret == ERROR_SUCCESS);
}

//---------------------------------------------------------------------------
//  Registry::Write() --  Write time value to specified tag
//---------------------------------------------------------------------------

bool Registry::Write(const std::string & tag, const time_t & value)
{
	DWORD reserved = 0;
	long ret = ::RegSetValueEx(CurrentKey(), tag.c_str(), reserved, REG_BINARY,
		reinterpret_cast < const BYTE * > (&value), sizeof(time_t));

	return (ret == ERROR_SUCCESS);
}

//---------------------------------------------------------------------------
//  Registry::Write() --  Write char array to specified tag
//---------------------------------------------------------------------------

bool Registry::Write(const std::string & tag, const CharArray & array)
{
	DWORD reserved = 0;
	long ret = ::RegSetValueEx(CurrentKey(), tag.c_str(), reserved, REG_BINARY,
		reinterpret_cast < const BYTE * > (&array[0]), static_cast<DWORD>(array.size()*sizeof(char)));

	return (ret == ERROR_SUCCESS);
}

//---------------------------------------------------------------------------
//  Registry::Write() --  Write int array to specified tag
//---------------------------------------------------------------------------

bool Registry::Write(const std::string & tag, const IntArray & array)
{
	DWORD reserved = 0;
	long ret = ::RegSetValueEx(CurrentKey(), tag.c_str(), reserved, REG_BINARY,
		reinterpret_cast < const BYTE * > (&array[0]), static_cast<DWORD>(array.size()*sizeof(int)));

	return (ret == ERROR_SUCCESS);
}

//---------------------------------------------------------------------------
//  Registry::Write() --  Write float array to specified tag
//---------------------------------------------------------------------------

bool Registry::Write(const std::string & tag, const FloatArray & array)
{
	DWORD reserved = 0;
	long ret = ::RegSetValueEx(CurrentKey(), tag.c_str(), reserved, REG_BINARY,
		reinterpret_cast < const BYTE * > (&array[0]), static_cast<DWORD>(array.size()*sizeof(float)));

	return (ret == ERROR_SUCCESS);
}

//---------------------------------------------------------------------------
//  Registry::WriteExpanded() --  Write string value to specified tag
//---------------------------------------------------------------------------

bool Registry::WriteExpanded(const std::string & tag, const std::string & value)
{
    DWORD reserved = 0;
	long ret = ::RegSetValueEx(CurrentKey(), tag.c_str(), reserved, REG_EXPAND_SZ,
		reinterpret_cast<const BYTE *>(value.c_str()), static_cast<DWORD>(value.size()));

	return (ret == ERROR_SUCCESS);
}

//---------------------------------------------------------------------------
//  Registry::Write() --  Write string value to specified tag
//---------------------------------------------------------------------------

bool Registry::Write(const std::string & tag, const vector<std::string> &vs)
{
    DWORD reserved = 0;
    std::vector<char> vc;

    for (size_t i = 0; i < vs.size(); i++)
        {
        std::copy(vs[i].begin(), vs[i].end(), back_inserter(vc));
        vc.push_back(0);
        }
   vc.push_back(0);

   std::string temp(vc.begin(), vc.end());

   long ret = ::RegSetValueEx(CurrentKey(), tag.c_str(), reserved, REG_MULTI_SZ,
        reinterpret_cast<const BYTE *>(temp.c_str()), static_cast<DWORD>(temp.size()));

    return (ret == ERROR_SUCCESS);
}


//---------------------------------------------------------------------------
//  Registry::IsKey() --  Does the key exist?
//---------------------------------------------------------------------------

bool Registry::IsKey(const std::string & key)
{
	RegistryKey hKey(CurrentKey(), key);
	return hKey.Handle() ? true : false;
}

//---------------------------------------------------------------------------
//  Registry::DeleteKey() -- Delete specified key
//---------------------------------------------------------------------------

bool Registry::DeleteKey(const std::string & key)
{
	return ::RegDeleteKey(CurrentKey(), key.c_str()) == 0;
}

//---------------------------------------------------------------------------
//  Registry::IsTag() -- Does the tag exist
//---------------------------------------------------------------------------

bool Registry::IsTag(const std::string & tag)
{
	long ret = ::RegQueryValueEx(CurrentKey(), tag.c_str(), 0, 0, 0, 0);
	return (ret == ERROR_FILE_NOT_FOUND) ? false : true;
}

//---------------------------------------------------------------------------
//  Registry::DeleteTag() -- Delete specified tag
//---------------------------------------------------------------------------

bool Registry::DeleteTag(const std::string & tag)
{
	return RegDeleteValue(CurrentKey(), tag.c_str()) == 0;
}

//---------------------------------------------------------------------------
//  Registry::Read() --  Read string from specified tag
//---------------------------------------------------------------------------

std::string Registry::Read(const std::string & tag, const std::string & defValue)
{
	DWORD type = REG_SZ;

	// Determine required buffer size
	DWORD sz;
	long ret = ::RegQueryValueEx(CurrentKey(), tag.c_str(), NULL, &type, 0, &sz);

	std::string value(sz, '\0');

	ret = ::RegQueryValueEx(CurrentKey(), tag.c_str(), NULL, &type,
		reinterpret_cast<LPBYTE>(&value[0]), &sz);

	if (ret == ERROR_SUCCESS)
		{
		// Eliminate trailing null, if present
		return RemoveTrailingNulls(value);
		}
	else
		return defValue;
}

//---------------------------------------------------------------------------
//  Registry::Read() --  Read string from specified tag
//---------------------------------------------------------------------------

std::string Registry::Read(const std::string & tag, const char * defValue)
{
    return Read(tag, std::string(defValue));
}

//---------------------------------------------------------------------------
//  Registry::Read() --  Read integer from specified tag
//---------------------------------------------------------------------------

int Registry::Read(const std::string & tag, int defValue)
{
	DWORD   type = REG_DWORD;
	DWORD   sz = sizeof(int);
	int value;
	long ret = ::RegQueryValueEx(CurrentKey(), tag.c_str(), NULL, &type,
		reinterpret_cast < LPBYTE > (&value), &sz);

	if (ret == ERROR_SUCCESS)
		return value;
	else
		return defValue;
}

//---------------------------------------------------------------------------
//  Registry::Read() --  Read boolean from specified tag
//---------------------------------------------------------------------------

bool Registry::Read(const std::string & tag, bool defValue)
{
	return Read(tag, defValue ? 1 : 0) ? true : false;
}

//---------------------------------------------------------------------------
//  Registry::Read() --  Read double value from specified tag
//---------------------------------------------------------------------------

double Registry::Read(const std::string & tag, double defValue)
{
	DWORD   type = REG_BINARY;
	DWORD   sz = sizeof(double);
	double value;
	long ret = ::RegQueryValueEx(CurrentKey(), tag.c_str(), NULL, &type,
		reinterpret_cast < LPBYTE > (&value), &sz);

	if (ret == ERROR_SUCCESS)
		return value;
	else
		return defValue;
}

//---------------------------------------------------------------------------
//  Registry::Read() --  Read 64-bit integer value from specified tag
//---------------------------------------------------------------------------

ii64 Registry::Read(const std::string & tag, ii64 defValue)
{
	DWORD   type = REG_QWORD;
	DWORD   sz = sizeof(ii64);
	ii64 value;
	long ret = ::RegQueryValueEx(CurrentKey(), tag.c_str(), NULL, &type,
		reinterpret_cast < LPBYTE > (&value), &sz);

	if (ret == ERROR_SUCCESS)
		return value;
	else
		return defValue;
}

//---------------------------------------------------------------------------
//  Registry::Read() --  Read time value from specified tag
//---------------------------------------------------------------------------

time_t Registry::Read(const std::string & tag, time_t defValue)
{
	DWORD   type = REG_BINARY;
	DWORD   sz = sizeof(time_t);
	ii64 value;
	long ret = ::RegQueryValueEx(CurrentKey(), tag.c_str(), NULL, &type,
		reinterpret_cast < LPBYTE > (&value), &sz);

	if (ret == ERROR_SUCCESS)
		return value;
	else
		return defValue;
}

//---------------------------------------------------------------------------
//  Registry::Read() --  Read array of float values from specified tag
//---------------------------------------------------------------------------

bool Registry::Read(const std::string & tag, FloatArray & result)
{
	size_t bytes = SizeInBytes(tag);
	result.resize(bytes/sizeof(float));

	DWORD   type = REG_BINARY;
	DWORD   sz = static_cast<DWORD>(bytes);
	long ret = ::RegQueryValueEx(CurrentKey(), tag.c_str(), NULL, &type,
		reinterpret_cast < LPBYTE > (&result[0]), &sz);

    return (ret == ERROR_SUCCESS);
}

//---------------------------------------------------------------------------
//  Registry::Read() --  Read array of string values from specified tag
//---------------------------------------------------------------------------
bool Registry::Read(const string &tag, vector<string> &result)
{
    size_t bytes = SizeInBytes(tag);
    string data(bytes, 0);

    DWORD type = REG_MULTI_SZ;
    DWORD sz  = static_cast<DWORD>(bytes);
    long  ret = RegQueryValueEx(CurrentKey(), tag.c_str(), NULL, &type,
        reinterpret_cast <LPBYTE> (&data[0]), &sz);

    if (ret == ERROR_SUCCESS) {
        int i = 0;
        for (;;) {
            string s = string(data.data() + i);
            i += s.length() + 1;

            if (s.empty()) break;

            result.push_back(s);
        }
    }

    return (ret == ERROR_SUCCESS);
}

//---------------------------------------------------------------------------
//  Registry::Read() --  Read array of integer values from specified tag
//---------------------------------------------------------------------------

bool Registry::Read(const std::string & tag, IntArray & result)
{
	DWORD bytes = SizeInBytes(tag);
	result.resize(bytes/sizeof(int));

	DWORD   type = REG_BINARY;
	DWORD   sz = bytes;
	long ret = ::RegQueryValueEx(CurrentKey(), tag.c_str(), NULL, &type,
		reinterpret_cast < LPBYTE > (&result[0]), &sz);

	return (ret == ERROR_SUCCESS);
}

//---------------------------------------------------------------------------
//  Registry::Read() --  Read array of char values from specified tag
//---------------------------------------------------------------------------

bool Registry::Read(const std::string & tag, CharArray & result)
{
	DWORD bytes = SizeInBytes(tag);
	result.resize(bytes/sizeof(char));

	DWORD   type = REG_BINARY;
	DWORD   sz = bytes;
	long ret = ::RegQueryValueEx(CurrentKey(), tag.c_str(), NULL, &type,
		reinterpret_cast < LPBYTE > (&result[0]), &sz);

	return (ret == ERROR_SUCCESS);
}

//---------------------------------------------------------------------------
//  Registry::SizeInBytes()  Return size in bytes of specified tag value
//---------------------------------------------------------------------------

DWORD Registry::SizeInBytes(const std::string & tag) const
{
	DWORD result;
    RegQueryValueEx(CurrentKey(), tag.c_str(), NULL, NULL, NULL, &result);

	return result;
}

//---------------------------------------------------------------------------
//  Registry::Descend() -- Enter (and optionally create) a subkey
//---------------------------------------------------------------------------

bool Registry::Descend(const std::string & key, bool create)
{
	StringList list(Decompose(key));

	while (list.Count())
		{
        std::string tag = list.At(0);
		RegistryKey hKey(CurrentKey(), tag, create);
		if (!hKey.Handle())
			return false;

		Fifo.push_front(hKey);
		list.Delete(0);
		}

	return true;
}

//---------------------------------------------------------------------------
//  Registry::Ascend() -- Retreat to parent of subkey
//---------------------------------------------------------------------------

bool Registry::Ascend()
{
	if (!Fifo.size())
		return false;

	Fifo.pop_front();
	return true;
}

//---------------------------------------------------------------------------
//  Registry::Decompose() -- Decompose key to constituent subkeys
//---------------------------------------------------------------------------

StringList Registry::Decompose(const string & key)
{
    std::string k=key;
	StringList list;

	size_t pos = k.find("\\");
	while (pos == 0)
		{
		k.erase(0, 1);
		pos = k.find("\\");
		}

    while (pos != std::string::npos)
		{
        std::string sub = k.substr(0, pos);
		if (sub.size())
			{
			list.Add(sub);
			k.erase(0, pos+1);
			}

		pos = k.find("\\");
		}

	if (k.size())
		list.Add(k);

	return list;
}

//---------------------------------------------------------------------------
//  Registry::Path() -- Return pathspec to current key
//---------------------------------------------------------------------------

std::string Registry::Path() const
{
    std::stringstream msg;
	for (KeyDeque::const_reverse_iterator i = Fifo.rbegin(); i != Fifo.rend(); ++i)
		msg << i->Key() << "\\";

	return msg.str();
}

//=============================================================================
//  CLASS SystemRegistry  --
//=============================================================================

//---------------------------------------------------------------------------
// SystemRegistry::SystemRegistry() --  Ctor
//---------------------------------------------------------------------------

SystemRegistry::SystemRegistry(const std::string & app, const std::string & org)
	: Registry(HKEY_CLASSES_ROOT, "HKEY_CLASSES_ROOT", app, org)
{
}


//=============================================================================
//  CLASS UserRegistry  --
//=============================================================================

//---------------------------------------------------------------------------
// UserRegistry::UserRegistry() --  Ctor
//---------------------------------------------------------------------------

UserRegistry::UserRegistry(const std::string & app, const std::string & org)
	: Registry(HKEY_CURRENT_USER, "HKEY_CURRENT_USER", app, org)
{
}

//=============================================================================
//  CLASS MachineRegistry  --
//=============================================================================

//---------------------------------------------------------------------------
// MachineRegistry::MachineRegistry() --  Ctor
//---------------------------------------------------------------------------

MachineRegistry::MachineRegistry(const std::string & app, const std::string & org)
	: Registry(HKEY_LOCAL_MACHINE, "HKEY_LOCAL_MACHINE", app, org)
{
}

}   // namespace Innovative
