//
// File:       Semaphore.h
// Class:      Semaphore
// Descrptn:   Classes to simplify windows structures
//
// Author:     Christopher Smith (AIC)

#ifndef WinSemaphore_WinH
#define WinSemaphore_WinH

#include <windows.h>

namespace Innovative
{

namespace Windows
{

#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//===========================================================================
//   CLASS Semaphore
//===========================================================================
//
//  The Semaphore class is designed to ease the use of a Windows 
//  Semaphore object. Putting it into a class also makes releasing  
//  resources less of a concern for the user.

class Semaphore
{
public:
//
//  Member Functions
//
    Semaphore(LPSECURITY_ATTRIBUTES lpsa = NULL, LONG InitCount = 0, 
                     LONG MaxCount = 1000, LPCTSTR Name = NULL)
        : hSem(NULL)
        {
            //
            //  Create a windows Semaphore object to "go with" the object
            //     Accepting Defaults a standard unnamed Semaphore
            hSem = CreateSemaphore(lpsa, InitCount, MaxCount, Name);
            //  If this failed, the hSem is NULL. 
            //     This is our "Bad Handle" indicator.
            //
            //  NOTE: The default Semaphore is NOT inheritable, but
            //     you may be able to make one that is. The copy
            //     constructor will force the copy to be uninheritable.
            //     A true copy of a Inheritable Semaphore can't be made.
         };
    //
    //  Alternate Constructor. No default MaxCount. If you need more
    //     flexibility than this, use the above constructor.
    Semaphore(LONG MaxCount, LONG InitCount = 0 )
        : hSem(NULL)
        {
            hSem = CreateSemaphore(NULL, InitCount, MaxCount, NULL);
        };
    //
    //  Copy constructor.
    Semaphore( const Semaphore &originalSemaphore)
        {  Dup( originalSemaphore);  };

    ~Semaphore()
        {
        if (hSem!=NULL)
            ::CloseHandle(hSem);
        hSem = NULL;
        };

    Semaphore & operator = (const Semaphore & aSemaphore)
        {
        HANDLE temp = hSem;      // Save original Handle
        Dup(aSemaphore);         // Clone its Handle
        if (temp!=NULL)
            ::CloseHandle(temp); // Close the old handle
        return *this;
        };

    BOOL Release(LONG howMany = 1)
        {  return ::ReleaseSemaphore(hSem, howMany, NULL); };
    //
    //  NOTE: Acquire may BLOCK! Use with caution...
    //
    DWORD Acquire(DWORD Timeout = INFINITE )
        {  return ::WaitForSingleObject( hSem, Timeout); };  // Timeout in ms

    DWORD WaitFor(DWORD Timeout = INFINITE )
        {  return Acquire();  };  // Timeout in ms

    operator HANDLE() const
        { return hSem; };

    HANDLE GetHandle() const
        { return hSem; }
    HANDLE Handle() const
        { return hSem; }

private:
//
//  Data Members
//
    HANDLE hSem;

    void Dup( const Semaphore &originalSemaphore)
        {
        HANDLE thisProcess = GetCurrentProcess();
        if (!::DuplicateHandle(thisProcess,originalSemaphore, // original Semaphore
                             thisProcess,&hSem,           // new handle
                             STANDARD_RIGHTS_REQUIRED,    // Ignored, since cloning
                             FALSE,                       // NOT Inheritable!
                             DUPLICATE_SAME_ACCESS))      // clone access
            hSem = NULL;
        //
        //  If call failed, make sure our handle is NULLed.
        };
};

#ifdef __CLR_VER
#pragma managed(pop)
#endif

} // Windows

} // Innovative

#endif
