// GcScript.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

//  This code implements a small script interpreter suitable for use
//  in the initialization of the GrayChip 5016.  The TI tools create
//  a complex set of register initialization settings which must be
//  applied to the GC in order to configure the device.  Due to the
//  complexity of the design process, it's actually more convienient
//  to simply parse the register init file and poke these values in
//  at runtime, than it is to write a generic API that supports all
//  reasonable GC initialization possibilities.
//  Applications hook callback events in order to perform the actual
//  port accesses to the baseboard hosting the GC device.

#include "GcScript_Mb.h"

#include <string>
#include <cstdlib>
#include <memory>
#include <iostream>
#include <sstream>
#include <cctype>
#include <errno.h>
#include <sstream>
#include <iomanip>
#include <limits>
#include <cstdio>

using namespace Innovative;

namespace   Innovative
{

//===========================================================================
//  CLASS GcScripter  --  Script command dispatcher
//===========================================================================

//---------------------------------------------------------------------------
//  GcScripter::GcScripter() --  Ctor
//---------------------------------------------------------------------------

GcScripter::GcScripter()
{
}

//---------------------------------------------------------------------------
//  GcScripter::Execute() --  Execute specified script
//---------------------------------------------------------------------------

void GcScripter::Execute(std::string file, int channel)
{
    if (!File.LoadFromFile(file))
        {
        ProcessStatusEvent e("Can't open GC script file");
        OnMessage.Execute(e);
        return;
        }

    for (size_t i = 0; i < File.Count(); ++i)
        {
        std::string line = File.StringAt(i);

        try
            {
            ParseLine(line, channel);
            }
        catch (...)
            {
            std::stringstream msg;
            msg << "Script parse error at line " << i;
            ProcessStatusEvent e(msg.str());
            OnMessage.Execute(e);
            return;
            }
        }
}

//---------------------------------------------------------------------------
//  GcScripter::ParseLine() --  Parse single line in script file
//---------------------------------------------------------------------------

void GcScripter::ParseLine(std::string line, int channel)
{
    std::string ln(line);

    // Skip blank lines
    if (!line.length())
        return;

    // If present, delete tag prefix.  Else, skip this line
    size_t TagIdx = line.find("dwr16");
    if (TagIdx == std::string::npos)
        return;
    line.erase(0, 6);

    // If present, extract hex value up to space delimeter
    size_t SpaceIdx = line.find(" ");
    unsigned int Offset = FromHex(line.substr(0, SpaceIdx));
    line.erase(0, SpaceIdx+1);

    // Trim trailing ctl chars
    std::string trim;
    for (std::string::iterator i=line.begin(); i != line.end(); ++i)
        {
        if (!std::iscntrl(*i))
            trim += *i;
        }
    unsigned int Value = FromHex(trim);

    GcScriptCommandEvent e(channel, Offset, Value);
    OnCommand.Execute(e);
}

//---------------------------------------------------------------------------
//  GcScripter::ToHex() --
//---------------------------------------------------------------------------

std::string GcScripter::ToHex(unsigned int value, int width)
{
    std::stringstream result;
    result << std::hex << std::setfill('0') << std::setw(width) << value;

    return result.str();
}

//---------------------------------------------------------------------------
//  GcScripter::FromHex() --
//---------------------------------------------------------------------------

#ifdef _MSC_VER
	#pragma warning(push)	
	#pragma warning(disable : 4996)		
#endif

unsigned int GcScripter::FromHex(std::string txt)
{
    unsigned int result = std::numeric_limits<unsigned int>::max();

    std::string format = std::string("%x");
    if (std::sscanf(txt.c_str(), format.c_str(), &result) == EOF)
        {
        ProcessStatusEvent e("Range error");
        OnMessage.Execute(e);
        }

    return result;
}

#ifdef _MSC_VER
	#pragma warning(pop)	
#endif

} // namespace Innovative



