// GpsMessage_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2009 Innovative Integration Corporation.
//  All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <GpsMessage_Mb.h>
#include <StringSupport_Mb.h>

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif

namespace Innovative
{
//===========================================================================
// CLASS GpsMessage  --  GPS "Partially Parsed" message
//===========================================================================
//---------------------------------------------------------------------------
//  GpsMessage::AsText() - combine all args as text string (delim separated)
//---------------------------------------------------------------------------

std::string GpsMessage::AsText(const std::string & delim) const
{
    std::string text;
    for (ArgList::const_iterator i=Args.begin(); i!=Args.end(); ++i)
        {
        if (!text.empty())
            text += delim;
        text += *i;
        }
    return text;
}

//---------------------------------------------------------------------------
//  GpsMessage::Arg() - Bounds-checked Arg value
//---------------------------------------------------------------------------

std::string GpsMessage::Arg(int Idx) const
{
    //  "Bounds Checked"
    if (Idx >= (int)Args.size())
        return "";
    else
        return Args[Idx];
}

//---------------------------------------------------------------------------
//  GpsMessage::ArgAsInt() - convert Arg to unsigned int
//---------------------------------------------------------------------------

unsigned int GpsMessage::ArgAsInt(int Idx) const
{
    // DON'T use String to int, because leading 0s are octal!
    return StringToDec(Arg(Idx));
}

//---------------------------------------------------------------------------
//  GpsMessage::ArgAsFloat() - convert Arg to double precision float
//---------------------------------------------------------------------------

double GpsMessage::ArgAsFloat(int Idx) const
{
    return StringToDouble(Arg(Idx));
}

//===========================================================================
// CLASS GpsMessageList  --  Collection of GPS "Partially Parsed" messages
//===========================================================================
//---------------------------------------------------------------------------
//  GpsMessageList::AsText() - combine all args as text string (delim separated)
//---------------------------------------------------------------------------

std::string GpsMessageList::AsText(const std::string & delim,
                                   const std::string & msg_delim ) const
{
    std::string text;
    for (MessageList::const_iterator i=Messages.begin(); i!=Messages.end(); ++i)
        {
        text += i->AsText(msg_delim);
        text += delim;
        }
    return text;
}

//---------------------------------------------------------------------------
//  GpsMessageList::FindMessage() - Find message by message name
//---------------------------------------------------------------------------

int GpsMessageList::FindMessage(const std::string & msg_name)
{
    int result = -1;

    for (MessageList::iterator i=Messages.begin(); i!=Messages.end(); ++i)
        {
        ++result;
        if (i->Arg(0) == msg_name)
            return result;
        }
    return result;
}

//==============================================================================
// CLASS MessageCracker  --  Base Class for Message Crackers
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class MessageCracker - extract data from message
//------------------------------------------------------------------------------

MessageCracker::MessageCracker(const GpsMessage & message)
    :  IsValid(false), Message(message)
{
    //
    //  Load Quality Check Information
    all  = Message.AsText();       // debug
    argc = Message.TotalArgs();
}

//------------------------------------------------------------------------------
//  MessageCracker::IsCommand() -- Test if our command is proper
//------------------------------------------------------------------------------

bool  MessageCracker::IsCommand(const std::string & shouldBe)
{
    //
    //  Check that message name matches desired string
    if (Message.Arg(0).compare(shouldBe) != 0)
        return false;     // remain invalid
    else
        return true;
}

//------------------------------------------------------------------------------
//  MessageCracker::HasArgs() -- Test if command has proper arguments
//------------------------------------------------------------------------------
//  A command string can be garbled, which usually will change the argument count

bool  MessageCracker::HasArgs(unsigned int shouldHave)
{
    if (argc != shouldHave)
        return false;     // remain invalid
    else
        return true;
}

//------------------------------------------------------------------------------
//  MessageCracker::SubString() -- Safe Substr
//------------------------------------------------------------------------------
//  A safe substring command that bounds-checks to avoid tossing exception

std::string  MessageCracker::SubString(const std::string & Str,
                                       size_t start_idx, size_t len)
{
    if (start_idx >= Str.size())
        return "";
    else
        return Str.substr(start_idx, len);
}

//==============================================================================
// CLASS GpZdaCracker  --  Conversions for $GPZDA message
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class GpZdaCracker - extract data from message
//------------------------------------------------------------------------------

GpZdaCracker::GpZdaCracker(const GpsMessage & Message)
    :  MessageCracker(Message)
{
    //
    //  Check that message is a GPZDA message
    if (!IsCommand("$GPZDA"))
        return;     // remain invalid
    //
    //  Check for message quality
    //  ...does it have proper number of arguments
    if (!HasArgs(7))
        return;    //  message probably garbled
    //
    //  We have a Zda message so crack it
    Utc   = Message.Arg(1);
    Day   = Message.ArgAsInt(2);
    Month = Message.ArgAsInt(3);
    Year  = Message.ArgAsInt(4);

    struct std::tm  unix_time;
    // Hour
    Hour = StringToDec(SubString(Utc, 0, 2));    // Utc string is HHMMSS.sss
    unix_time.tm_hour = Hour;
    // Min
    Min = StringToDec(SubString(Utc, 2, 2));
    unix_time.tm_min = Min;
    // Sec (integer only
    Sec = StringToDec(SubString(Utc, 4, 2));
    unix_time.tm_sec = Sec;
    // Day
    unix_time.tm_mday = Day;
    // Month
    unix_time.tm_mon  = Month - 1;
    // Year
    unix_time.tm_year = Year-1900;

    unix_time.tm_isdst = 0;  // not daylight savings time

    //  Convert time structure to time_t
    Time = mktime(&unix_time);

    //  Signal correct cracking performed
    IsValid = true;
    if (Time==-1)
        IsValid=false;
}

//==============================================================================
// CLASS GpGgaCracker  --  Conversions for $GPGGA message
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class GpGgaCracker - extract data from message
//------------------------------------------------------------------------------

GpGgaCracker::GpGgaCracker(const GpsMessage & message)
    :  MessageCracker(message)
{
    //
    //  Check that message is a GPGGA message
    if (!IsCommand("$GPGGA"))
        return;     // remain invalid
    //
    //  Check for message quality
    //  ...does it have proper number of arguments
    if (!HasArgs(15))
        return;    //  message probably garbled

    //
    //  We have a Gga message so crack it
    LatDegrees = StringToDec(SubString(Message.Arg(2), 0, 2));  // Lat string is DDMM.MMMM
    LatMinutes = StringToFloat(SubString(Message.Arg(2), 2 ));
    LatUnits   = Message.Arg(3);

    LongDegrees = StringToDec(SubString(Message.Arg(4), 0, 3));  // Lat string is DDDMM.MMMM
    LongMinutes = StringToFloat(SubString(Message.Arg(4), 3));
    LongUnits   = Message.Arg(5);

    Fix = static_cast<FixStatus>(Message.ArgAsInt(6));
    Satellites = Message.ArgAsInt(7);

    //  Signal correct cracking performed
    IsValid = true;
}

//==============================================================================
// CLASS GpRmcCracker  --  Conversions for $GPRMC message
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class GpRmcCracker - extract data from message
//------------------------------------------------------------------------------

GpRmcCracker::GpRmcCracker(const GpsMessage & Message)
    :  MessageCracker(Message)
{
    //
    //  Check that message is a GPRMC message
    if (!IsCommand("$GPRMC"))
        return;     // remain invalid
    //
    //  Check for message quality
    //  ...does it have proper number of arguments
    if (!HasArgs(13))
        return;    //  message probably garbled
    //
    //  We have a Rmc message so crack it
    Utc   = Message.Arg(1);

    Day   = StringToDec(SubString(Message.Arg(9), 0, 2));  // Arg 9 string is DDMMYY
    Month = StringToDec(SubString(Message.Arg(9), 2, 2));  // Arg 9 string is DDMMYY
    Year  = StringToDec(SubString(Message.Arg(9), 4, 2));  // Arg 9 string is DDMMYY

    struct std::tm  unix_time;
    // Hour
    Hour = StringToDec(SubString(Utc, 0, 2));    // Utc string is HHMMSS.sss
    unix_time.tm_hour = Hour;
    // Min
    Min = StringToDec(SubString(Utc, 2, 2));
    unix_time.tm_min = Min;
    // Sec (integer only)
    Sec = StringToDec(SubString(Utc, 4, 2));
    unix_time.tm_sec = Sec;
    // Day
    unix_time.tm_mday = Day;
    // Month
    unix_time.tm_mon  = Month - 1;
    // Year
    unix_time.tm_year = Year;

    unix_time.tm_isdst = 0;  // not daylight savings time

    //  Convert time structure to time_t
    Time = mktime(&unix_time);

    LatDegrees = StringToDec(SubString(Message.Arg(3), 0, 2));  // Lat string is DDMM.MMMM
    LatMinutes = StringToFloat(SubString(Message.Arg(3), 2 ));
    LatUnits   = Message.Arg(4);

    LongDegrees = StringToDec(SubString(Message.Arg(5), 0, 3));  // Lat string is DDDMM.MMMM
    LongMinutes = StringToFloat(SubString(Message.Arg(5), 3));
    LongUnits   = Message.Arg(6);

    //  Signal correct cracking performed
    IsValid = true;
    if (Time==-1)
        IsValid=false;
}

}  // end namespace Innovative

#ifdef _MSC_VER
	#pragma warning(pop)
#endif


