// Gps_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2009 Innovative Integration Corporation.
//  All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <Gps_Mb.h>
#include <SystemSupport_Mb.h>
#include <cmath>
#include <sstream>

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif

namespace Innovative
{

//===========================================================================
//  CLASS GpsCoordinate  -- GPS position object
//===========================================================================

//---------------------------------------------------------------------------
//  constructor for class GpsCoordinate
//---------------------------------------------------------------------------

GpsCoordinate::GpsCoordinate()
	: Degrees(0), Minutes(0.0f), Seconds(0.0f), Units("degrees")
{  
}

//---------------------------------------------------------------------------
//  constructor for class GpsCoordinate
//---------------------------------------------------------------------------

GpsCoordinate::GpsCoordinate(int degrees, double minutes, const std::string & units)
	: Degrees(degrees), Minutes(minutes), Seconds(0.0f), Units(units)
{
}

//---------------------------------------------------------------------------
//  constructor for class GpsCoordinate
//---------------------------------------------------------------------------

GpsCoordinate::GpsCoordinate(double radians, Orientation o)
{
	double Pi = 4.0*std::atan(1.0);
	double degrees = radians * 180.0/Pi;

	double integer;
	double fraction = std::modf(degrees, &integer);
	Degrees = static_cast<int>(std::fabs(integer));

	fraction = std::modf(std::fabs(fraction*60.0f), &integer);
	Minutes = integer;

	fraction = std::modf(std::fabs(fraction*60.0f), &integer);
	Seconds = integer;

	if (o == oLatitude)
		Units = (degrees < 0.0) ? "S" : "N";
	else
		Units = (degrees < 0.0) ? "W" : "E";
}

//---------------------------------------------------------------------------
//  GpsCoordinate::Print()  --  Display position
//---------------------------------------------------------------------------

std::string GpsCoordinate::Print() const
{
	std::stringstream msg;
	unsigned char degree = 0xb0;

	msg << Degrees << degree << " "
		<< Minutes << "' "
		<< Seconds << "\" "
		<< Units;

	return msg.str();
}

//===========================================================================
//  CLASS IGps  -- Interface to GPS Hardware
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class IGps
//---------------------------------------------------------------------------

IGps::IGps(IGpsMessageParser * iParser)
    :  Parser(iParser), FTime(std::time(0)),
       FGpsTriggerTime(0), FFix(fixNoFix), FSatellites(0)
{
    FLatitude.Degrees = 1,  FLatitude.Minutes = 0.0;   FLatitude.Units = "N";
    FLongitude.Degrees = 1, FLongitude.Minutes = 0.0;  FLongitude.Units = "E";
}

//---------------------------------------------------------------------------
//  destructor for class IGps
//---------------------------------------------------------------------------

IGps::~IGps()
{
    delete Parser;  Parser = 0;
}

//---------------------------------------------------------------------------
//  IGps::Init() --
//---------------------------------------------------------------------------

void IGps::Init()
{
    //  Local Init

    //  Derived Class Init
    OnInit();
}

//---------------------------------------------------------------------------
//  IGps::Reapply() --
//---------------------------------------------------------------------------

void IGps::Reapply()
{
    //  Local Reapply

    //  Derived Class Reapply
    OnReapply();
}

//---------------------------------------------------------------------------
//  IGps::Baud() --  Set baud rate of UART
//---------------------------------------------------------------------------

void IGps::Baud(BaudRates rate)
{
    GpsBaud(rate);
    Innovative::Sleep(1000);
    UartBaud(rate);
}

//---------------------------------------------------------------------------
//  IGps::Baud() --  Fetch baud rate of UART
//---------------------------------------------------------------------------

IGps::BaudRates IGps::Baud()
{
    return UartBaud();
}

//---------------------------------------------------------------------------
//  IGps::Enabled() --  Enable/disable GPS subsystem
//---------------------------------------------------------------------------

void IGps::Enabled(bool state)
{
    Queue.erase();
    HdwEnabled(state);
}

//---------------------------------------------------------------------------
//  IGps::Enabled() --  query enable state of GPS subsystem
//---------------------------------------------------------------------------

bool IGps::Enabled()
{
    return HdwEnabled();
}

//---------------------------------------------------------------------------
//  IGps::GpsTriggerTime() --  Trigger in the future
//---------------------------------------------------------------------------
//  The time passed in is when you want the trigger to go off. We save a second
//     prior, to give us time to prime the hardware on the previous interrupt

void IGps::GpsTriggerTime(const std::time_t & then)
{
    std::time_t interrupt_time = then-1;   // one second earlier
    FGpsTriggerTime = interrupt_time;
}

//---------------------------------------------------------------------------
//  IGps::CustomParser() --  replace Parser with a new one
//---------------------------------------------------------------------------

void IGps::CustomParser(IGpsMessageParser * newParser)
{
    delete Parser;
    Parser = newParser;
}

//---------------------------------------------------------------------------
//  IGps::Command() --  Send Command string out serial port
//---------------------------------------------------------------------------

void IGps::Command(const std::string & txt)
{
    for (std::string::const_iterator i = txt.begin(); i != txt.end(); ++i)
        UartData(*i);
    //
    //  Finish with the ending token
    std::string CmdEndToken = Parser->EndToken();
    for (std::string::const_iterator i = CmdEndToken.begin();
                                     i != CmdEndToken.end(); ++i)
        {
        UartData(*i);
        }
}

//---------------------------------------------------------------------------
//  IGps::DispatchGpsInterrupt() -- Process a GPS Interrupt
//---------------------------------------------------------------------------

void IGps::DispatchGpsInterrupt()
{
    GpsMessageList  Messages;

    // Eat entire GPS response string
    size_t count = UartFifoDepth();
    if (count)
        {
        for (size_t i = 0; i < count; ++i)
            Queue.push_back(UartData());
        //
        //  Parse the messages from queue
        Parser->Parse(Queue, Messages);
        //
        //  Send Message List to derived class for processing
        OnProcessMessages(Messages);
        //
        // Trigger at specified time
        //  (FTime has been updated by OnProcessMessages())
        std::time_t now = FTime;
        if (now == FGpsTriggerTime)
            {
            // Arm to fire on next epoch
            OnGpsTrigger();
            OpenWire::NotifyEvent e;
            OnGpsArm.Execute(e);
            }
        }

    // Notify application code
    GpsEpochEvent e(Messages);
    OnEpoch.Execute(e);
}

//===========================================================================
// CLASS StandardGpsMessageParser  --  Normal Message String Parser
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class StandardGpsMessageParser
//---------------------------------------------------------------------------

StandardGpsMessageParser::StandardGpsMessageParser()
    :  CmdStartToken("$"), CmdEndToken("\r\n"), CmdArgDelimiterToken(","),
       CmdChecksumDelimiterToken("*")
{
}

//---------------------------------------------------------------------------
//  constructor for class StandardGpsMessageParser
//---------------------------------------------------------------------------

StandardGpsMessageParser::StandardGpsMessageParser(
                             const std::string & start_token,
                             const std::string & end_token,
                             const std::string & delimit_token,
                             const std::string & cksum_token)
    :  CmdStartToken(start_token), CmdEndToken(end_token), CmdArgDelimiterToken(delimit_token),
       CmdChecksumDelimiterToken(cksum_token)
{
}

//---------------------------------------------------------------------------
//  StandardGpsMessageParser::Parse() -- extract text from text, make messages
//---------------------------------------------------------------------------

void  StandardGpsMessageParser::Parse(std::string & Queue, GpsMessageList & Messages)
{
    // Eliminate any characters preceeding the first Start Token ('$')
    size_t pos = Queue.find(CmdStartToken);
    if (pos != std::string::npos)
        Queue.erase(0, pos);

    //  Check if no end tokens are found, end at once
    size_t eol = Queue.find(CmdEndToken);
    if (eol == std::string::npos)
        return;

    //  Each string found is parsed to make a Message
    while (eol != std::string::npos)
        {
        std::string message_text = Queue.substr(0, eol);
        ParseMessage(message_text, Messages );
        //  purge the found message and check for more
        Queue.erase(0, eol+CmdEndToken.size());
        eol = Queue.find(CmdEndToken);
        }
    // end when last message is found
}

//---------------------------------------------------------------------------
//  StandardGpsMessageParser::ParseMessage() -- from text, make message
//---------------------------------------------------------------------------

void  StandardGpsMessageParser::ParseMessage(std::string text, GpsMessageList & Messages)
{
    GpsMessage Message;

    //  Strip the checksum from the end of the command
    size_t pos = text.find(CmdChecksumDelimiterToken);
    if (pos != std::string::npos)
        text.erase(pos);      // delete till end of string

	// Add a CmdArgDelimiterToken at end of message
	//  so that last argument gets counted.
	text += CmdArgDelimiterToken;


	//  Each string found is parsed to make a Message
    //    extract every argument one by one
    size_t eol = text.find(CmdArgDelimiterToken);
    while (eol != std::string::npos)
        {
        std::string arg = text.substr(0, eol);
        Message.AddArgument(arg);
        //  purge the found message and check for more
        text.erase(0, eol+CmdArgDelimiterToken.size());
        eol = text.find(CmdArgDelimiterToken);
        }

    //  Add on any leftovers as final argument
    if (!text.empty())
        Message.AddArgument(text);

    //  Add message to message list
    Messages.AddMessage(Message);
    
}

}  // end namespace Innovative
