// Gps_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2009 Innovative Integration Corporation.
//  All Rights Reserved.
//
//---------------------------------------------------------------------------

#ifndef Gps_MbH
#define Gps_MbH

#include <Events_Mb.h>
#include <GpsMessage_Mb.h>
#include <Trace_Mb.h>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//===========================================================================
// CLASS IGpsMessageParser  --  Interface to string parser
//===========================================================================

class IGpsMessageParser
{
public:
    virtual ~IGpsMessageParser()  {  }
    virtual std::string EndToken() = 0;

    virtual void  Parse(std::string & queue, GpsMessageList & messages) = 0;
};

//===========================================================================
// CLASS GpsEpochEvent  --
//===========================================================================

class GpsEpochEvent : public OpenWire::Event
{
public:
    GpsMessageList List;

    GpsEpochEvent(GpsMessageList & list)
        :  List(list)
        {
        }
};

//===========================================================================
//  CLASS GpsCoordinate  -- GPS position object
//===========================================================================

class GpsCoordinate
{
public:
    enum Orientation { oLatitude, oLongitude };

    // Ctors
    GpsCoordinate();
    GpsCoordinate(int degrees, double minutes, const std::string & units="degrees");
    GpsCoordinate(double radians, Orientation o);

    // Methods
    std::string Print() const;

    // Data
    int         Degrees;
    double      Minutes;
    double      Seconds;
    std::string Units;
};

//===========================================================================
//  CLASS IGps  -- Interface to GPS Hardware
//===========================================================================

class IGps
{
public:
    enum BaudRates { b115200, b57600, b38400, b19200, b9600, b4800 };
    enum FixStatus { fixNoFix, fixGps, fixDiffGps };
    enum IISentence { sGSA, sGSV, sRMC, sVTG, sZDA };

    IGps(IGpsMessageParser * iParser);
    virtual ~IGps();

    void            Init();
    void            Reapply();
    //
    //  Properties
    virtual void      Baud(BaudRates rate);
    virtual BaudRates Baud();
    void            Enabled(bool state);
    bool            Enabled();

    void            GpsTriggerTime(const std::time_t & then);
    std::time_t     Now()
        {  return Time();   }
    std::time_t     SecondsFromNow(int offset)
        {  return Time() + offset;  }
    void			CurrentPosition(const GpsCoordinate & latitude,
                                    const GpsCoordinate & longitude)
        {  FLatitude = latitude;  FLongitude = longitude;  }
    virtual bool    IsAntenna() const
        {  return true;  }
    virtual bool    IsLocked() const
        {  return false;  }

    //
    //  Current State Reporting Ftns
    std::time_t  	Time()
        {  return FTime;  }
    GpsCoordinate  	Latitude()
        {  return FLatitude;  }
    GpsCoordinate  	Longitude()
        {  return FLongitude;  }
    FixStatus  		Fix()
        {  return FFix;  }
    unsigned int  	Satellites()
        {  return FSatellites;  }

    void            Command(const std::string & txt);
    void            CustomParser(IGpsMessageParser * newParser);

    //   ...gps control
    virtual void          HdwEnabled(bool state) = 0;
    virtual bool          HdwEnabled() = 0;

    //  Events
    OpenWire::ThunkedEventHandler<Innovative::GpsEpochEvent> OnEpoch;
    OpenWire::ThunkedEventHandler<OpenWire::NotifyEvent>     OnGpsArm;

protected:
    //
    // Data
    std::string                 Queue;
    IGpsMessageParser *         Parser;
    //  ...cached current values
    std::time_t   	FTime;
    std::time_t   	FGpsTriggerTime;

    GpsCoordinate  	FLatitude;
    GpsCoordinate  	FLongitude;
    FixStatus      	FFix;
    unsigned int   	FSatellites;

    virtual void  OnInit()    = 0;
    virtual void  OnReapply() = 0;
    virtual void  OnProcessMessages(GpsMessageList & Messages) = 0;
    virtual void  OnGpsTrigger() = 0;
    //
    //  Baseboard Config
    //   ...uart
    virtual unsigned int  UartFifoDepth() = 0;
    virtual void          UartData(char c) = 0;
    virtual char          UartData() = 0;
    virtual void          UartBaud(BaudRates baud) = 0;
    virtual BaudRates     UartBaud() = 0;

    //
    //  Derived Gps Config
    virtual void          GpsBaud(BaudRates baud) = 0;

    Trace Debug;

public:   // Internals
    void   DispatchGpsInterrupt();
};

//===========================================================================
//  CLASS MultiGps  -- Switchable GPS
//===========================================================================

class MultiGps
{
public:


};

//===========================================================================
//  CLASS IGps_Plugin  -- Interface to provide 'plugin' interface to HDW
//===========================================================================

class IGps_Plugin
{
public:
    virtual void  PGPS_OnInit()    = 0;
    virtual void  PGPS_OnReapply() = 0;
    virtual void  PGPS_OnProcessMessages(GpsMessageList & Messages) = 0;
    virtual void  PGPS_OnGpsTrigger() = 0;
    //
    //  Baseboard Config
    //   ...status
    virtual bool    PGPS_IsAntenna(bool gps_value) = 0;
    virtual bool    PGPS_IsLocked(bool gps_value) = 0;
    //  (note: the value of the GPS objects "IsX" call is passed in so
    //         that it can be used at discretion

    //   ...uart
    virtual unsigned int     PGPS_UartFifoDepth() = 0;
    virtual void             PGPS_UartData(char c) = 0;
    virtual char             PGPS_UartData() = 0;
    virtual void             PGPS_UartBaud(IGps::BaudRates baud) = 0;
    virtual IGps::BaudRates  PGPS_UartBaud() = 0;
    //   ...gps control
    virtual void          PGPS_HdwEnabled(bool state) = 0;
    virtual bool          PGPS_HdwEnabled() = 0;
    //
    //  Derived Gps Config
    virtual void          PGPS_GpsBaud(IGps::BaudRates baud) = 0;

    virtual ~IGps_Plugin()
        {}

};

//===========================================================================
//  CLASS GpsPluginSocket<GpsT>  -- Interface to bind plugin to IGps descendant
//===========================================================================

template <class GpsT>
class GpsPluginSocket : public GpsT
{
public:
    GpsPluginSocket( IGps_Plugin * plug )
        :  PI(plug)
        {}
    virtual ~GpsPluginSocket()
        {
        delete PI;  PI=0;
        }

    //  Status
    virtual bool    IsAntenna() const
        {   return PI->PGPS_IsAntenna(GpsT::IsAntenna());  }
    virtual bool    IsLocked() const
        {   return PI->PGPS_IsLocked(GpsT::IsLocked());  }

protected:
    virtual void  OnInit() {  PI->PGPS_OnInit();  }
    virtual void  OnReapply() {  PI->PGPS_OnReapply();  }
    virtual void  OnProcessMessages(GpsMessageList & Messages)
                    {
                    GpsT::OnProcessMessages(Messages);
                    PI->PGPS_OnProcessMessages( Messages );
                    }
    virtual void  OnGpsTrigger() {  PI->PGPS_OnGpsTrigger();  }
    //
    //  Baseboard Config
    //   ...uart
    virtual unsigned int  UartFifoDepth() {  return PI->PGPS_UartFifoDepth();  }
    virtual void          UartData(char c) {  PI->PGPS_UartData(c);  }
    virtual char          UartData() {  return PI->PGPS_UartData();  }
    virtual void              UartBaud(typename GpsT::BaudRates baud)
                                {  return PI->PGPS_UartBaud(baud);  }
    virtual typename GpsT::BaudRates   UartBaud() {  return PI->PGPS_UartBaud();  }
    //   ...gps control
    virtual void          HdwEnabled(bool state) {  PI->PGPS_HdwEnabled(state);  }
    virtual bool          HdwEnabled() {  return PI->PGPS_HdwEnabled();  }
    //
    //  Derived Gps Config
    virtual void          GpsBaud(typename GpsT::BaudRates baud)
                            {  return PI->PGPS_GpsBaud(baud);  }

private:
    IGps_Plugin  *  PI;
};

//===========================================================================
// CLASS StandardGpsMessageParser  --  Normal Message String Parser
//===========================================================================

class StandardGpsMessageParser : public IGpsMessageParser
{
public:
    StandardGpsMessageParser();
    StandardGpsMessageParser(const std::string & start_token,
                             const std::string & end_token,
                             const std::string & delimit_token,
                             const std::string & cksum_token);
    virtual ~StandardGpsMessageParser()  {  }

    virtual std::string EndToken()  { return CmdEndToken;  }
    virtual void  Parse(std::string & queue, GpsMessageList & messages);

private:
    std::string  CmdStartToken;
    std::string  CmdEndToken;
    std::string  CmdArgDelimiterToken;
    std::string  CmdChecksumDelimiterToken;

    void  ParseMessage(std::string text, GpsMessageList & messages);
    Trace Debug;
};

#ifdef __CLR_VER
#pragma managed(pop)
#endif
}  // end namespace Innovative
//---------------------------------------------------------------------------
#endif
