// HardwareRegister_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#ifndef HardwareRegister_MbH
#define HardwareRegister_MbH

#include <DataTypes_Mb.h>
#include <StopWatch_Mb.h>
#include <sstream>
#include <vector>
#include <map>

#undef NULL
#define NULL 0

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//==============================================================================
//  CLASS IBitManipulator  -- Interface applying Bit Manipulations
//==============================================================================

class IBitManipulator
{
public:
    virtual ~IBitManipulator();

    bool            Bit(unsigned int bit) const;
    void            Bit(unsigned int bit, bool value);

    unsigned int    Field(unsigned int bit, unsigned int size) const;
    void            Field(unsigned int bit, unsigned int size, unsigned int value);

    void            Mask(unsigned int mask);

    unsigned int    Value() const;
    void            Value(unsigned int value);

protected:
    virtual unsigned int  ReadV() const = 0;
    virtual void          WriteV(unsigned int value) = 0;
};

//==============================================================================
//  CLASS MemoryBitManipulator --  applying Bit Manips to an in memory value
//==============================================================================

class MemoryBitManipulator : public IBitManipulator
{
public:
    MemoryBitManipulator(unsigned int result);
    virtual ~MemoryBitManipulator();

protected:
    virtual unsigned int  ReadV() const;
    virtual void          WriteV(unsigned int value);

    unsigned int FValue;
};

//==============================================================================
//  CLASS PtrBitManipulator --  applying Bit Manips to a pointer to memory
//==============================================================================

class PtrBitManipulator : public IBitManipulator
{
public:
    PtrBitManipulator(unsigned int * ptr);
    PtrBitManipulator(int * ptr);
    virtual ~PtrBitManipulator();

protected:
    virtual unsigned int  ReadV() const;
    virtual void          WriteV(unsigned int value);

    unsigned int *   FValuePtr;
};

//==============================================================================
//  CLASS IBitManipulator64  -- Interface applying Bit Manipulations (64 bit)
//==============================================================================

class IBitManipulator64
{
public:
    virtual ~IBitManipulator64();

    bool            Bit(unsigned int bit) const;
    void            Bit(unsigned int bit, bool value);

    ii64            Field(unsigned int bit, unsigned int size) const;
    void            Field(unsigned int bit, unsigned int size, ii64 value);

    void            Mask(ii64 mask);

    ii64            Value() const;
    void            Value(ii64 value);

protected:
    virtual ii64    ReadV() const = 0;
    virtual void    WriteV(ii64 value) = 0;
};

//==============================================================================
//  CLASS MemoryBitManipulator64 --  applying Bit Manips to an in memory value
//==============================================================================

class MemoryBitManipulator64 : public IBitManipulator64
{
public:
    MemoryBitManipulator64(ii64 result);
    virtual ~MemoryBitManipulator64();

protected:
    virtual ii64  ReadV() const;
    virtual void  WriteV(ii64 value);

    ii64 FValue;
};

//==============================================================================
//  CLASS PtrBitManipulator64 --  applying Bit Manips to a pointer to memory
//==============================================================================

class PtrBitManipulator64 : public IBitManipulator64
{
public:
    PtrBitManipulator64(ii64 * ptr);
    PtrBitManipulator64(s_ii64 * ptr);
    virtual ~PtrBitManipulator64();

protected:
    virtual ii64  ReadV() const;
    virtual void  WriteV(ii64 value);

    ii64 *   FValuePtr;
};

class Register;


//=============================================================================
//  CLASS IAddressingSpace  -- Interface to "AddressingSpace"-like objects
//=============================================================================
class Trace;

class IAddressingSpace
{
public:
    IAddressingSpace();
    virtual ~IAddressingSpace();

    ii32    Read(ii32 offset);
    void    Write(ii32 offset, ii32 data);

    ii32    At(ii32 offset);
    void    At(ii32 offset, ii32 data);

    virtual bool    IsNull() = 0;
    virtual bool    IsNotNull() = 0;

    virtual ii32 * Base() const = 0;
protected:
    Trace   *Debug;
};

//=============================================================================
//  CLASS AddressingSpace  -- Basic Mapped Memory Region
//=============================================================================

class AddressingSpace : public IAddressingSpace
{
public:
    AddressingSpace(ii32 * baseaddr = NULL);
    virtual ~AddressingSpace();

    void    Link(ii32 * address);
        void    Link(const AddressingSpace & space);

    virtual bool  IsNull()
                {  return mBaseAddress == 0;  }
    virtual bool  IsNotNull()
                {  return mBaseAddress != 0;  }

    virtual ii32 * Base() const;

private:
    ii32 *         mBaseAddress;
};

//=============================================================================
//  CLASS WishboneBusSpace  -- WishboneBus "Section" of an addressing space
//=============================================================================

class WishboneBusSpace : public IAddressingSpace
{
public:
    WishboneBusSpace(IAddressingSpace & space, int wishbone_offset);
    virtual ~WishboneBusSpace();

    bool    IsNull()
                {  return Parent.IsNull();  }
    bool    IsNotNull()
                {  return Parent.IsNotNull();  }

    virtual ii32 * Base() const;

private:
    IAddressingSpace &  Parent;
    int                 WishboneBaseOffset;
};

typedef WishboneBusSpace  AxiMasterSpace;    // Alias for AXI bus boards

//=============================================================================
//  CLASS BaseboardExtension<T>  -- Template class for register extension classes
//=============================================================================
//  When defining a set of additonal registers for a board, use this template
//      as a base class. It links the spaces with those of the board thru references
//  REQUIRES:
//         functions PortMemorySpace and LogicMemorySpace for each board to expose the AddressSpace objects
//
//  EXAMPLE:
//          class NewRegisters :  public BaseboardExtension<X5_210M>
//               {
//               NewRegister(X5_210M & board)
//                          :  inherited(board),
//                             AdcSelectReg(LogicMemory, mmDemodAdcSelect)
//                                  {}
//
//               Register AdcSelectReg;
//               };

template <class T>
class BaseboardExtension
{
public:
    typedef BaseboardExtension<T>  inherited;
    typedef T   baseboard_type;
    typedef T*  baseboard_ptr;
    typedef T&  baseboard_ref;

    BaseboardExtension(T & board)
        :  Board(board), LogicMemory(LogicMemorySpace(board)),
           PortMemory(PortMemorySpace(board))
        { }

protected:
    T & Board;
    IAddressingSpace &  LogicMemory;
    IAddressingSpace &  PortMemory;
};

//=============================================================================
//  CLASS MemoryCell  --
//=============================================================================

class MemoryCell
{
public:
    MemoryCell(IAddressingSpace & space, int offset);
    virtual ~MemoryCell();

    int Offset()
        {  return mOffset;  }

protected:
    IAddressingSpace &   mAddressingSpace;
    int                  mOffset;
};

//=============================================================================
//  CLASS ReadOnlyRegister  --
//=============================================================================

class ReadOnlyRegister : public MemoryCell
{
public:
    ReadOnlyRegister(IAddressingSpace & space, int offset);
    ReadOnlyRegister(const std::string & label, IAddressingSpace & space, int offset);

    virtual ii32    Value() const;
    bool            Bit(int bit) const;
    operator        ii32() const;

//    ii32 &          Reference() const;

protected:
    std::string             Label;
};

//=============================================================================
//  CLASS Register  --
//=============================================================================

class Register : public ReadOnlyRegister
{
public:
    Register(IAddressingSpace & space, int offset);
    Register(const std::string & label, IAddressingSpace & space, int offset);

    virtual void    Value(ii32 value);
    virtual ii32    Value() const;

    bool            Bit(int bit) const;
    void            Bit(int bit, bool value);

    Register & operator =(ii32 value);
};

//==============================================================================
//  CLASS RegisterBitManipulator  -- Bit Manipulations to a Register
//==============================================================================

class RegisterBitManipulator : public IBitManipulator
{
public:
    RegisterBitManipulator(Register & reg)
        : Reg(reg)
        {}
    virtual ~RegisterBitManipulator()
        {};

protected:
    Register & Reg;

    virtual unsigned int  ReadV() const
       {  return Reg.Value();  }
    virtual void          WriteV(unsigned int value)
       {  Reg.Value(value);  }
};

//=============================================================================
//  CLASS ShadowRegisterBase  --
//=============================================================================

class ShadowRegisterBase : public Register
{
public:
    ShadowRegisterBase(IAddressingSpace & space, int offset, ii32 value);
    ShadowRegisterBase(const std::string & label, IAddressingSpace & space, int offset, ii32 value);

    virtual void    Value(ii32 value);
    virtual ii32    Value() const;

    void            Apply();

    ShadowRegisterBase & operator =(ii32 value);

    virtual void    ShadowValue(ii32 value);

protected:
    ii32    mShadowValue;

private:
    virtual void    ShadowRegisterChanged() = 0;
};

//=============================================================================
//  CLASS ShadowRegister  --
//=============================================================================

class ShadowRegister : public ShadowRegisterBase
{
public:
    ShadowRegister(IAddressingSpace & space, int offset, ii32 value=0);
    ShadowRegister(const std::string & label, IAddressingSpace & space, int offset, ii32 value=0);

private:
    virtual void ShadowRegisterChanged();
};

//=============================================================================
//  CLASS RefreshableShadowRegister  --
//=============================================================================

class RefreshableShadowRegister : public ShadowRegisterBase
{
public:
    RefreshableShadowRegister(IAddressingSpace & space, int offset, ii32 value=0);
    RefreshableShadowRegister(const std::string & label, IAddressingSpace & space, int offset, ii32 value=0);

    virtual void Refresh();
private:
    virtual void ShadowRegisterChanged();
};

//=============================================================================
//  CLASS CachedShadowRegister  --
//=============================================================================

class CachedShadowRegister : public ShadowRegisterBase
{
public:
    CachedShadowRegister(IAddressingSpace & space, int offset, ii32 value=0);
    CachedShadowRegister(const std::string & label,
        IAddressingSpace & space, int offset, ii32 value=0);

private:
    virtual void ShadowRegisterChanged();
};

//=============================================================================
//  CLASS MultiplexedShadowRegister  --
//=============================================================================

class MultiplexedShadowRegister : public ShadowRegisterBase
{
public:
    MultiplexedShadowRegister(IAddressingSpace & space, int offset,
        ii32 value=0, int mux=0);
    MultiplexedShadowRegister(const std::string & label,
        IAddressingSpace & space, int offset, ii32 value=0, int mux=0);

    void Mux(int mux);

protected:
    // Data
    int         mMux;

    virtual void ShadowRegisterChanged();
};

//=============================================================================
//  CLASS RegisterBit  --
//=============================================================================

class RegisterBit
{
public:
    RegisterBit(Register & owner_register, int bit_no);

    // Methods
    void Set();
    void Clear();
    void Flip();
    void Toggle();

    void Value(bool value);
    bool Value() const;

    operator bool() const;
    RegisterBit & operator =(bool value);

protected:
    Register &  mOwner;
    const int   mBit;
};

//=============================================================================
//  CLASS RegisterBitGroup  --
//=============================================================================

class RegisterBitGroup
{
public:
    RegisterBitGroup(Register & owner_register, int bit_no, int bit_count);
        virtual ~RegisterBitGroup();

    virtual void    Value(ii32 value);
    virtual ii32    Value() const;

    bool            Bit(unsigned int bit) const;
    void            Bit(unsigned int bit, bool value);

    RegisterBitGroup & operator =(ii32 value);
    operator int() const;

protected:
    Register &  mOwner;
    const int   mFirstBit;
    const int   mBitCount;
    ii32        mMask;
    ii32        mOwnerMask;
};

//=============================================================================
//  CLASS ReadOnlyRegisterBit  --
//=============================================================================

class ReadOnlyRegisterBit
{
public:
    ReadOnlyRegisterBit(ReadOnlyRegister & owner_register, int bit_no);

    // Methods
    bool        Value() const;
    operator    bool() const;

protected:
    ReadOnlyRegister &  mOwner;
    const int           mBit;

private:
    ReadOnlyRegisterBit & operator =(bool value);
};

//=============================================================================
//  CLASS ReadOnlyRegisterBitGroup  --
//=============================================================================

class ReadOnlyRegisterBitGroup
{
public:
    ReadOnlyRegisterBitGroup(ReadOnlyRegister & owner_register,
                             int bit_no,
                             int bit_count);
        virtual ~ReadOnlyRegisterBitGroup();

    virtual ii32    Value() const;
    bool            Bit(unsigned int bit) const;
    operator        int() const;

protected:
    ReadOnlyRegister &  mOwner;
    const int           mFirstBit;
    const int           mBitCount;
    ii32                mMask;
    ii32                mOwnerMask;
};

//=============================================================================
//  CLASS RegisterArray  --
//=============================================================================

template <class T>
class RegisterArray
{
public:
    unsigned int Size()
        {  return mRegisters.size();  }

    void Apply()
    {
        for (int i = 0; i < Size(); ++i)
            mRegisters[i] = mRegisters[i];
    }

    RegisterArray(IAddressingSpace & space, int offset, int size)
    {
        for (int i = 0; i < size; i ++)
            mRegisters.push_back(new T(space, offset + i));
    }

    //  This one for compatibility with old registers
    RegisterArray(AddressingSpace & space, int offset, int size)
    {
        for (int i = 0; i < size; i ++)
            mRegisters.push_back(new T(space, offset + i));
    }

    virtual ~RegisterArray()
    {
        for (unsigned int i = 0; i < mRegisters.size(); i ++)
            delete mRegisters[i];
    }

    T & operator[](int index)
    {
        return *(mRegisters[index]);
    }

    const T & operator[](int index) const
    {
        return *(mRegisters[index]);
    }

protected:
    std::vector<T *>    mRegisters;
};

//===========================================================================
//  CLASS SplitRegister
//===========================================================================
//  This class of register is mapped as halves on two 32 bit registers.

class SplitRegister
{
public:
  SplitRegister( IAddressingSpace & space, int offset_low, int offset_high );
  virtual ~SplitRegister();

  virtual ii32 Value();

  virtual void Value( ii32 value );

  operator ii32();
  SplitRegister & operator =(ii32 value);

protected:
    Register Low;
    Register High;
};

//===========================================================================
//  CLASS HpiRegister  -- Special Split register for HPI accesses
//===========================================================================
//  Adds a check of the Ready register between accesses of the HPI registers

class HpiRegister : public SplitRegister
{
public:
    HpiRegister(IAddressingSpace &space,
        int offset_low, int offset_high, int offset_hpi_ready );

  void          IsReady();

  virtual ii32  Value();
  virtual void  Value(ii32 value);

  operator ii32();
  SplitRegister & operator =(ii32 value);

protected:
    Register          HpiReady;
    StopWatch         Timer;
};

#ifdef __CLR_VER
#pragma managed(pop)
#endif
} // namespace Innovative

#endif
