// HpiEngine_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//
 
#include <HpiEngine_Mb.h>
#include <SystemSupport_Mb.h>

namespace Innovative
{

//===========================================================================
//  CLASS HpiEngine  -- HPI hardware control
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class HpiEngine
//------------------------------------------------------------------------------

HpiEngine::HpiEngine(HpiEngineAddresses & Hpi)
    :  HpiAddress( PortMemory, Hpi.AddrLsw, Hpi.AddrMsw, Hpi.ControlLsw ),
       HpiData( PortMemory, Hpi.DataLsw, Hpi.DataMsw, Hpi.ControlLsw ),
       HpiDataIncr( PortMemory, Hpi.DataIncrLsw, Hpi.DataIncrMsw, Hpi.ControlLsw ),
       HpiControl( PortMemory, Hpi.ControlLsw, Hpi.ControlMsw, Hpi.ControlLsw ),
       mSlowHpi(false)
{
    //  Default CE Space Loading values
    Values.GlobalCtl = 0x3078;   Values.Ce1      = 0x40;      Values.Ce0      = 0x10910321;
    Values.Ce2       = 0x30;     Values.Ce3      = 0x40;   
    Values.SdramTime = 0x465;    Values.SdramExt = 0x544a7;   Values.SdramCtl = 0x53225000;
}

//---------------------------------------------------------------------------
//  HpiEngine::Reset() -- clear HPI system
//------------------------------------------------------------------------------

void  HpiEngine::Reset()
{
    HpiControl = 0x00010001;
}

//---------------------------------------------------------------------------
//  HpiEngine::Interrupt() -- Start program
//---------------------------------------------------------------------------
 
void HpiEngine::Interrupt()
{
    HpiControl = 0x00030003;
}

//---------------------------------------------------------------------------
//  HpiEngine::WriteBlock() --
//---------------------------------------------------------------------------
 
void  HpiEngine::WriteBlock(const int*  src, int dst, int size)
{
    if (mSlowHpi)
        {
        for (int i = 0; i < size; ++i)
            Poke(dst + i * sizeof(int), src[i]);

        return;
        }

    HpiAddress = dst;

    int Result;
    for (int i = 0; i < (size - 1); ++i)
        {
        Result = *src++;

        HpiDataIncr = Result;
        }

    // Workaround for silicon errata:  Must end incr HPI access w/fixed mode
    Result = *src;
    HpiData = Result;
}

//---------------------------------------------------------------------------
//  HpiEngine::ReadBlock() --
//---------------------------------------------------------------------------
 
void HpiEngine::ReadBlock(int src, int* dst, int size)
{
    // Must prefetch and discard some initial data, to work-around
    // apparent silicon defects
    const int Discard = 0x8;
    int source = src - (Discard * sizeof(int));

    HpiAddress.Value(source);

    for (int i = 0; i < Discard; ++i)
        {
        //  Retreive HPI data
        HpiDataIncr.Value();
        }

    for (int i = 0; i < size; ++i)
        {
        //  Retreive HPI data
        int Result = HpiDataIncr;
        *dst++ = Result;
        }

    // Workaround for silicon errata:  Must end incr HPI access w/fixed mode
    HpiData.Value();
}

//---------------------------------------------------------------------------
//  HpiEngine::Poke() --
//---------------------------------------------------------------------------
 
void HpiEngine::Poke(int address, int value)
{
    HpiAddress = address;
    HpiData = value;
}

//---------------------------------------------------------------------------
//  HpiEngine::Peek() --
//---------------------------------------------------------------------------

int  HpiEngine::Peek(int address)
{
    HpiAddress = address;

    // Workaround for silicon errata:  Avoid use of FETCH bit
    HpiData.Value();

    int Result = HpiData;
    return Result;
}

//---------------------------------------------------------------------------
//  HpiEngine::SetAddress() --
//---------------------------------------------------------------------------

void  HpiEngine::SetAddress(int address)
{
    HpiAddress = address;
}

//---------------------------------------------------------------------------
//  HpiEngine::GetAddress() --
//---------------------------------------------------------------------------

int   HpiEngine::GetAddress() const
{
    return HpiAddress;
}

//---------------------------------------------------------------------------
//  HpiEngine::ConfigureTargetSdram() --
//---------------------------------------------------------------------------

void  HpiEngine::ConfigureTargetSdram()
{
    const int mmEmifBase = 0x1800000;    
    
    const int emifGblCtl = 0 << 2;
    const int emifCe1 = 1 << 2;
    const int emifCe0 = 2 << 2;
    const int emifCe2 = 4 << 2;
    const int emifCe3 = 5 << 2;
    const int emifSdCtl = 6 << 2;
    const int emifSdTim = 7 << 2;
    const int emifSdExt = 8 << 2;

    Poke(mmEmifBase + emifGblCtl, Values.GlobalCtl);   // global control
    Poke(mmEmifBase + emifCe1,    Values.Ce1);         // CE1
    Poke(mmEmifBase + emifCe0,    Values.Ce0);         // CE0
    Poke(mmEmifBase + emifCe2,    Values.Ce2);         // CE2
    Poke(mmEmifBase + emifCe3,    Values.Ce3);         // CE3
    Poke(mmEmifBase + emifSdTim,  Values.SdramTime);   // SDRAM timing
    Poke(mmEmifBase + emifSdExt,  Values.SdramExt);    // SDRAM extension
    Poke(mmEmifBase + emifSdCtl,  Values.SdramCtl);    // SDRAM control
	Sleep(5);
}


}


