//
// I2c_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure 
//  agreement with Innovative Integration Corporation and may not be copied or 
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2003 Innovative Integration Corporation. 
//  All Rights Reserved.
//

#include <I2C_Mb.h>
#include <Endian_Mb.h>
#include <algorithm>
#include <SystemSupport_Mb.h>
                
namespace Innovative
{

#ifdef __GNUC__
    #pragma GCC diagnostic ignored "-Wunused-but-set-variable"
#endif

//=============================================================================
//  CLASS I2CDevice  --  I-squared-C device control class
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class I2CDevice
//-----------------------------------------------------------------------------

const int kDefaultPageSizeBytes = 16;

//-----------------------------------------------------------------------------
//  constructor for class I2CDevice -- SubAddress Version
//-----------------------------------------------------------------------------

I2CDevice::I2CDevice(I2CDeviceIntf & intf, SubAddrTag/*sub*/, int sub_addr,
                     bool endian, bool bit_reversed)
    : FSubAddr(sub_addr), FDevAddr(sub_addr>>4), FEndian(endian),
      FBitReversed(bit_reversed), FNackReads(false),
      FStretchableClock(false), FPageSizeBytes(kDefaultPageSizeBytes), Intf(intf)
{
}

//-----------------------------------------------------------------------------
//  constructor for class I2CDevice -- DevAddress Version
//-----------------------------------------------------------------------------

I2CDevice::I2CDevice(I2CDeviceIntf & intf, DevAddrTag/*dev*/, int dev_addr,
                     bool endian, bool bit_reversed)
    : FSubAddr(dev_addr<<4), FDevAddr(dev_addr), FEndian(endian),
      FBitReversed(bit_reversed), FNackReads(false),
      FStretchableClock(false), FPageSizeBytes(kDefaultPageSizeBytes), Intf(intf)
{
}

//-----------------------------------------------------------------------------
//  constructor for class I2CDevice -- copy ctor
//-----------------------------------------------------------------------------

I2CDevice::I2CDevice(const I2CDevice & other)
    : ISuppliesI2CDeviceRequirements(),
      FSubAddr(other.FSubAddr), FDevAddr(other.FDevAddr), FEndian(other.FEndian),
      FBitReversed(other.FBitReversed), FNackReads(other.FNackReads),
      FStretchableClock(other.FStretchableClock), FPageSizeBytes(kDefaultPageSizeBytes),
      Intf(other.Intf)
{
}

//------------------------------------------------------------------------
//  I2CDevice::~I2CDevice - Dtor
//------------------------------------------------------------------------

I2CDevice::~I2CDevice()
{
}

//-----------------------------------------------------------------------------
//  constructor for class I2CDevice -- DevAddress Version
//-----------------------------------------------------------------------------

I2CDevice & I2CDevice::operator =(const I2CDevice & other)
{
    FSubAddr = other.FSubAddr;
    FDevAddr = other.FDevAddr;
    FEndian = other.FEndian;
    FBitReversed = other.FBitReversed;
    FNackReads = other.FNackReads;
    FStretchableClock = other.FStretchableClock;
    Intf = other.Intf;

    return *this;
}

//------------------------------------------------------------------------
//  I2CDevice::tCLK()  --  Minimum clk hi/low, in nanoseconds
//------------------------------------------------------------------------

void I2CDevice::tCLK()
{
    Rtc.uS(3);
}

//------------------------------------------------------------------------
//  I2CDevice::BitReversed - Enable/disable bytewise bit reversal 
//------------------------------------------------------------------------

void I2CDevice::BitReversed(bool state)
{
    FBitReversed = state;
}

//------------------------------------------------------------------------
//  I2CDevice::BitReversed - Retrieve bytewise bit reversal state
//------------------------------------------------------------------------

bool I2CDevice::BitReversed() const
{
    return FBitReversed;
}

//------------------------------------------------------------------------
//  I2CDevice::Reg_sub() -  Register access, with sub_address
//------------------------------------------------------------------------

int I2CDevice::Reg_sub(int sub_addr, int reg, int value)
{
    // Set Sub_Address Tag
    SubAddr_Override(sub_addr);
    return Reg(reg, value);
}

//------------------------------------------------------------------------
//  I2CDevice::Reg_sub() -  Register access, with sub_address
//------------------------------------------------------------------------

int I2CDevice::Reg_sub(int sub_addr, int reg)
{
    // Set Sub_Address Tag
    SubAddr_Override(sub_addr);
    return Reg(reg);
}

//------------------------------------------------------------------------
//  I2CDevice::Reg - Writes a register on a device via an I2C interface
//------------------------------------------------------------------------
//  Arguments:
//     reg = register subaddress to access
//     value = value to write
//

int I2CDevice::Reg(int reg, int value)
{
    int return_val = 1;
    //
    // send preamble
    Start();
    // send R/W command with I2C bus address, direction
    RwCommand(FSubAddr, 0);
    if (!Acked())
        {
        return_val = 0;
        }
    // send register subaddress
    if (!SendAddress(reg))
        {
        return_val = 0;
        }

    // send data to write to register
    WriteByte(value);

    if (!Acked())
        {
        return_val = 0;
        }
    // send stop
    Stop();

    return(return_val);
}

//------------------------------------------------------------------------
//  I2CDevice::Littlereg - Writes a register on 3 line LCD device via an I2C interface
//------------------------------------------------------------------------
//  Arguments:
//     rs = data vs instruction bit (0 or 1)
//     value = value to write
//

void I2CDevice::LittleReg(int rs, int value)
{
    StartNHLCD();
    // send R/W command with I2C bus address, direction

    // send data to write to register
    WriteNHLCD(rs, value);
}



//------------------------------------------------------------------------
//  I2CDevice::Reg - reads a register on a device via an I2C interface
//------------------------------------------------------------------------
//  Arguments:
//    addr = I2C bus address of device to access
//    reg = subaddress register number of I2C device to read
//
//  Returns value read from register in LS 8 bits
//

int I2CDevice::Reg(int reg)
{
    int readback;

    // send preamble
    Start();
    // send R/W command with I2C bus address, direction = write
    RwCommand(FSubAddr, 0);
    Acked();
    // send register subaddress
    SendAddress(reg);

    Start();
    // send R/W command with I2C bus address, direction = read
    RwCommand(FSubAddr, 1);
    Acked();
    // read data from register
    readback = ReadByte();

    // Acknowledge data read
    if (NackReads())
        Nack();
    else
        Ack();

    // send stop
    Stop();

    return(readback);
}

//------------------------------------------------------------------------
//  I2CDevice::MultiRegW() - Write multiple registers on I2C
//------------------------------------------------------------------------
//  Arguments:
//     reg = register subaddress to start with
//     data = array to write
//     size = number of writes
//
//  Returns 0 if a failure was seen
//

int I2CDevice::MultiRegW(int reg, const int * data, int size)
{
    int return_val = 1;
    //
    // send preamble
    Start();
    // send R/W command with I2C bus address, direction
    RwCommand(FSubAddr, 0);
    if (!Acked())
        {
        return_val = 0;
        }
    // send register subaddress
    if (!SendAddress(reg))
        {
        return_val = 0;
        }
    // send data to write to register
    for (int i=0; i<size; i++)
        {
        WriteByte(data[i]);
        if (!Acked())
            {
            return_val = 0;
            }
        }

    // send stop
    Stop();

    return(return_val);
}


//------------------------------------------------------------------------
//  I2CDevice::MultiRegR - reads a register on a device via an I2C interface
//------------------------------------------------------------------------
//  Arguments:
//     reg = subaddress register number of I2C device to read
//     data = array to read
//     size = number of reads
//
//	 Returns 0 if a failure was seen
//

int  I2CDevice::MultiRegR(int reg, int * data, int size)
{
    int return_val = 1;

    if (size<=0)
        return 0;    // illegal attempt

    // send preamble
    Start();
    // send R/W command with I2C bus address, direction = write
    RwCommand(FSubAddr, 0);
    if (!Acked())
        {
        return_val = 0;
        }
    // send register subaddress
    if (!SendAddress(reg))
        {
        return_val = 0;
        }

    Start();
    // send R/W command with I2C bus address, direction	= read
    RwCommand(FSubAddr, 1);
    if (!Acked())
        {
        return_val = 0;
        }

    //  Read all but last word, acking each one
    for (int i=0; i<size-1; i++)
        {
        // read data from register
        data[i] = ReadByte();

        Ack();
        }

    // Nack the final word read
    data[size-1] = ReadByte();
    Nack();

    // send stop
    Stop();

    return(return_val);
}

//------------------------------------------------------------------------
//  I2CDevice::ReadBlock - Sequential read of IDROM
//------------------------------------------------------------------------

bool I2CDevice::ReadBlock(int offset, int * data, int words)
{
    if (!Ready(offset << 2))
        return false;   // throw IIException("Ready timeout accessing I2C device");

    Start();

    if (!ReadAddress(offset << 2))
        return false;
        
    //
    // Now read data from rom, assemble into words and write to buffer
    for (int word=0; word<words; ++word)   
        {
        int word_value = 0;        // Reset for next word

        for (int byte=0; byte<4; ++byte)   
            {
            int byte_value = ReadByte(FBitReversed);

            //
            //  Assemble word from bytes
            word_value |= (byte_value << (8*byte));  

            //  Acknowledge data read, except on last byte of sequence
            if( !((word==(words-1)) && (byte==3)) )
                Ack();
            }
        //
        //  Load into output buffer
        data[word] = word_value;
        }
    //
    // issue STOP 
    if (NackReads())
        Nack();
    Stop();
    
    if (FEndian)
        Endian(data, words);
            
    return true;
}

//------------------------------------------------------------------------
//  I2CDevice::WritePortion() --  Writes portion of data page efficiently
//------------------------------------------------------------------------
//  Supported by EEPROM devices

bool I2CDevice::WritePortion(int offset, const int * page_data, int words)
{
    if (!Ready(offset << 2))
        return false;
        
    //  Now write data to page
    for (int word=0; word<words; ++word)   
        {
        // Get 32 bit word value from buffer
        int word_value = page_data[word];   
        if (FEndian)
            word_value = Endian(word_value);

        for (int byte=0; byte<4; ++byte)   
            {
            // 
            //  Extract byte value from word
            int byte_value = (word_value >> (8*byte)) & 0xff ; 
            //
            //  Write to IDROM
            WriteByte(byte_value, FBitReversed);
            //
            //  Check for ACK from rom:  Return if no ack after select
            if (!Acked())
                return false;   
            }
        }
    //
    // issue STOP 
    Stop();

    return true;
}

//------------------------------------------------------------------------
//  I2CDevice::WriteBlock() --  Writes block of data across pages efficiently
//------------------------------------------------------------------------
//  Supported by EEPROM devices

bool  I2CDevice::WriteBlock(int offset, const int * data, int words)
{
    bool Result = true;
    
    while (words && Result)
        {
        int Index = offset % IntsPerPage();
        int Residual = std::min(static_cast<unsigned int>(words), IntsPerPage()-Index);
        
        Result &= WritePortion(offset, data, Residual);
        
        offset += Residual;
        data += Residual;
        words -= Residual;
        }
        
    return Result;		
}

//------------------------------------------------------------------------
//  I2CDevice::WritePage() --  Writes Page of Data efficiently
//------------------------------------------------------------------------
//  Supported by EEPROM devices

bool I2CDevice::WritePage(int offset, const int * page_data)
{
    return WriteBlock(offset, page_data, IntsPerPage());
}

//------------------------------------------------------------------------
//  I2CDevice::Byte_dev() -  Register access, with dev_address
//------------------------------------------------------------------------

void I2CDevice::Byte_dev(int dev_addr, int byte, int value)
{
    // Set dev_address Tag
    DevAddr_Override(dev_addr);
    Byte(byte, value);
}

//------------------------------------------------------------------------
//  I2CDevice::Byte_dev() -  Register access, with dev_address
//------------------------------------------------------------------------

int I2CDevice::Byte_dev(int dev_addr, int byte)
{
    // Set dev_address Tag
    DevAddr_Override(dev_addr);
    return Byte(byte);
}

//------------------------------------------------------------------------
//  I2CDevice::Byte() --  Single Byte Random access READ
//------------------------------------------------------------------------
//  Supported by EEPROM devices

int  I2CDevice::Byte(int byte_offset)
{
    int readback;

    // send preamble
    Start();

    // send register subaddress
    if (!WriteAddress(byte_offset))
        return 0;
        
    Start();
    
    if (!ReadAddress(byte_offset))
        return 0;
        
    // read data from register
    readback = ReadByte(FBitReversed);

    //
    // Note:  No ack on single-byte reads...
    //
    
    // send stop
    Stop();

    return readback;
}

//------------------------------------------------------------------------
//  I2CDevice::Byte() --  Single Byte Random access WRITE
//------------------------------------------------------------------------
//  Supported by EEPROM devices

void  I2CDevice::Byte(int byte_offset, int data)
{
    //
    //  send preamble
    Start();

    // send register subaddress
    if (!WriteAddress(byte_offset))
        return;

    // send data to write to register
    WriteByte(data, FBitReversed);
    if (!Acked())
        return;

    // send stop
    Stop();

    return;
}

//------------------------------------------------------------------------
//  I2CDevice::DeviceAddress() --  Set device address
//------------------------------------------------------------------------

bool I2CDevice::DeviceAddress(int byte_offset)
{
    // send preamble
    Start();

    // send register subaddress
    if (!WriteAddress(byte_offset))
        return false;
        
    return true;
}


//------------------------------------------------------------------------
//  I2CDevice::DeviceByte() --  Read byte at current address
//------------------------------------------------------------------------

int  I2CDevice::DeviceByte(int byte_offset)
{
    int readback;

    Start();
    
    if (!ReadAddress(byte_offset))
        return 0;
        
    // read data from register
    readback = ReadByte(FBitReversed);

    //
    // Note:  No ack on single-byte reads...
    //
    
    // send stop
    Stop();

    return readback;
}

//-----------------------------------------------------------------
//  I2CDevice::Ready() -  Write completion check
//-----------------------------------------------------------------

bool I2CDevice::Ready(int byte_offset)
{
    double Then = Rtc.Now() + 1;
    bool TimedOut = false;

    do
        {
        Start();

        TimedOut = Rtc.Now() > Then;
        }
    while (!WriteAddress(byte_offset) && !TimedOut);
    
    return !TimedOut;
}

//-----------------------------------------------------------------
//  I2CDevice::WriteAddress() - Set byte address for subsequent write
//-----------------------------------------------------------------

bool I2CDevice::WriteAddress(int byte_offset)
{
    //  Figure out where address bits go...in the address or in the
    //    RW command word
    int rw_high = 0;
    int rw_low = 0;
    int offset = 0;
    CalculateAddressParams(byte_offset, rw_high, rw_low, offset);
    //
    //  Start R/W command
    RwCommand(rw_high, rw_low, 0);
    if (!Acked())
        return false;

    // send register subaddress, if no ACK received, return false
    return SendAddress(offset) ? true : false;
}

//-----------------------------------------------------------------
//  I2CDevice::ReadAddress() - Set byte address for subsequent read
//-----------------------------------------------------------------

bool I2CDevice::ReadAddress(int byte_offset)
{
    //  Figure out where address bits go...in the address or in the
    //    RW command word
    int rw_high = 0;
    int rw_low = 0;
    int offset = 0;
    CalculateAddressParams(byte_offset, rw_high, rw_low, offset);

    // send R/W command with I2C bus address, direction	= read
    RwCommand(rw_high, rw_low, 1);

    // if no ACK received, return false
    return Acked() ? true : false;
}

//------------------------------------------------------------------------------
//  I2CDevice::CalculateAddressParams() - params for ReadAddress(), WriteAddress()
//------------------------------------------------------------------------------

void I2CDevice::CalculateAddressParams(int byte_offset, int & rw_hi, int &rw_lo, int & new_offset)
{
    //  For these roms, the upper address bits are sent in the RW Command Low bits
    rw_hi = FDevAddr;
    rw_lo = byte_offset >> 8;  // 64-word page address for 4k ROMs
    new_offset = byte_offset;
}

//-----------------------------------------------------------------
//  I2CDevice::IntsPerPage() - Return # of 32-bit ints per page
//-----------------------------------------------------------------

unsigned int I2CDevice::IntsPerPage()
{
    return FPageSizeBytes>>2;
}

//-----------------------------------------------------------------
//  I2CDevice::SendAddress() - Send address portion of I2C
//-----------------------------------------------------------------
//  This method sends the standard 1 byte address field

int  I2CDevice::SendAddress(int address)
{
    WriteByte(address);
    if (!Acked())
        return 0;
    else
        return 1;
}

//-----------------------------------------------------------------
//  I2CDevice::ReadBit() - read a bit from I2C
//-----------------------------------------------------------------

int  I2CDevice::ReadBit()
{
    tCLK();
    int bit_value = Data();

    Clock(1);
    volatile int dummy = Clock();
    tCLK();

    Clock(0);
    dummy = Clock();
    tCLK();

    return bit_value;
}

//-----------------------------------------------------------------
//  I2CDevice::WriteBit() - write a bit from 'data' to I2C 
//-----------------------------------------------------------------

void  I2CDevice::WriteBit(int data, int bit_no )
{
    Data((data & (1<<bit_no)) ? 1 : 0);
    ClockPulse(); 
}
 
//-----------------------------------------------------------------
//  I2CDevice::ReadByte() - read 8-bit byte from I2C
//-----------------------------------------------------------------

unsigned int  I2CDevice::ReadByte(bool bit_reverse)
{
    int byte_value = 0;

    for (int bit=0; bit<8; ++bit)
        {
        int shift = bit_reverse ? bit : (7-bit);
        byte_value |= ReadBit() << shift; 
        }

    return byte_value;
}

//-----------------------------------------------------------------
//  I2CDevice::WriteByte() - write 8-bit byte to I2C 
//-----------------------------------------------------------------

void  I2CDevice::WriteByte(int byte_value, bool bit_reverse)
{
    // MSB to LSB
    for (int bit=0; bit<8; ++bit)
        {
        int shift = bit_reverse ? bit : (7-bit);
        WriteBit(byte_value, shift); 
        }
}

//-----------------------------------------------------------------
//  I2CDevice::WriteNHLCD() - write 9-bit byte to I2C LCD
//-----------------------------------------------------------------

void  I2CDevice::WriteNHLCD(int rs, int value)
{
    // MSB to LSB
    WriteBit(rs, 0);
    WriteByte(value);
}

//-----------------------------------------------------------------
//  I2CDevice::RwCommand() - send address, and read/write bit 
//-----------------------------------------------------------------

void  I2CDevice::RwCommand(int addr, int rw_bit)
{
    //  Write address from MSB to LSB
    for (int i = 7; i > 0; --i)
        WriteBit(addr, i);

    //  Write R/W bit
    WriteBit(rw_bit, 0);
}

//-----------------------------------------------------------------
//  I2CDevice::RwCommand() - Send dev type, address, and r/w bit 
//-----------------------------------------------------------------

void  I2CDevice::RwCommand(int device_type, int addr, int rw_bit)
{
    //  Write device type from MSB to LSB
    WriteBit(device_type, 3);
    WriteBit(device_type, 2);
    WriteBit(device_type, 1);
    WriteBit(device_type, 0);
    //	Write address from MSB to LSB
    WriteBit(addr, 2);
    WriteBit(addr, 1);
    WriteBit(addr, 0);
    //	Write R/W bit
    WriteBit(rw_bit, 0);
}

//-----------------------------------------------------------------
//  I2CDevice::Acked() - check for ACK from I2C
//-----------------------------------------------------------------

bool I2CDevice::Acked()
{
    bool Result = false;

    // check for ACK from rom
    Data(1);
    tCLK();
    Clock(1);
    volatile int dummy = Clock();
    tCLK();

    if (Data() == 0)
        Result = true;

    // tHIGH delay
    tCLK();

    Clock(0);
    dummy = Clock();

    // tLOW delay
    tCLK();

    return Result;
}

//-----------------------------------------------------------------
//  I2CDevice::Start() - Send START to I2C 
//-----------------------------------------------------------------

void  I2CDevice::Start()
{
    // IDLE:  Data and clock to high
    Data(1);
    tCLK();
    Clock(1);
    volatile int dummy = Clock();
    tCLK();     

    // START
    Data(0);
    tCLK();      

    Clock(0);
    tCLK();
    dummy = 0;
}

//-----------------------------------------------------------------
//  I2CDevice::StartNHLCD() - Send START to I2C
//-----------------------------------------------------------------

void  I2CDevice::StartNHLCD()
{
    // IDLE:  Data and clock to high
    Data(0);
    tCLK();
    Clock(0);
    volatile int dummy = Clock();
    tCLK();

    // START
   // Data(rs);
   // tCLK();

  //  Clock(1);
  //  tCLK();
    dummy = 0;
}

//-----------------------------------------------------------------
//  I2CDevice::Stop() - Send STOP to I2C
//-----------------------------------------------------------------

void  I2CDevice::Stop()
{
    Data(0);
    tCLK();

    Clock(1);
    volatile int dummy = Clock();
    tCLK();

    Data(1);	
    tCLK();
    dummy = 0;
}

//-----------------------------------------------------------------
//  I2CDevice::Ack() - Send Ack to I2C when read complete
//-----------------------------------------------------------------

void  I2CDevice::Ack()
{
    Data(0);
    ClockPulse(); 
    Data(1);
    tCLK();        
}

//-----------------------------------------------------------------
//  I2CDevice::Nack() - Send NAck to I2C when finished read 
//-----------------------------------------------------------------

void  I2CDevice::Nack()
{
    Data(1);
    ClockPulse(); 
    Data(1);
    tCLK();        
}

//-----------------------------------------------------------------
//  I2CDevice::ClockPulse() - pulse the clock line to the I2C bus
//-----------------------------------------------------------------

void  I2CDevice::ClockPulse()
{
    tCLK();      
    Clock(1);
    volatile int dummy = Clock();
    tCLK();
    Clock(0);
    dummy = Clock();
    tCLK();
}


//-----------------------------------------------------------------
//  I2CDevice::DevAddr_Override() - Reset DevAddr
//-----------------------------------------------------------------

void  I2CDevice::DevAddr_Override(int dev_addr)
{
    FDevAddr = dev_addr;  FSubAddr = dev_addr<<4;
}

//-----------------------------------------------------------------
//  I2CDevice::SubAddr_Override() - Reset SubAddr
//-----------------------------------------------------------------

void  I2CDevice::SubAddr_Override(int sub_addr)
{
    FSubAddr = sub_addr;  FDevAddr = sub_addr>>4;
}

//-----------------------------------------------------------------
//  I2CDevice::Stretch() - Implement Clock Stretching
//-----------------------------------------------------------------
//  Some devices will hold the clock down if they are not ready.
//    We should wait for the clock to actually rise before moving on

void  I2CDevice::Stretch(bool state)
{
    if (FStretchableClock == false)
        return;
    if (state == false)
        return;               // don't stretch 0s
    //
    //  We have just written the clock..make sure it is actually up
    while (Clock()==false)
        Rtc.uS(1);
}

//-----------------------------------------------------------------
//  I2CDevice::I2C_Reg() - Write I2C devicer register
//-----------------------------------------------------------------

void  I2CDevice::I2C_Reg(int reg, int value)
{
    Reg(reg, value);
}

//-----------------------------------------------------------------
//  I2CDevice::I2C_Reg() - Read I2C devicer register
//-----------------------------------------------------------------

int  I2CDevice::I2C_Reg(int reg)
{
    return Reg(reg);
}

//=============================================================================
//  CLASS I2CDevice16  --  16 bit address I2C device
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class I2CDevice16
//-----------------------------------------------------------------------------

const unsigned int kDefaultI2CDevice16_PageSizeBytes = 32;

//-----------------------------------------------------------------------------
//  constructor for class I2CDevice16 -- SubAddrTag Version
//-----------------------------------------------------------------------------

I2CDevice16::I2CDevice16(I2CDeviceIntf & intf, SubAddrTag sub, int sub_addr,
              bool endian, bool bit_reversed)
    : I2CDeviceN(intf, sub, sub_addr, endian, bit_reversed)
{
    PageSizeBytes(kDefaultI2CDevice16_PageSizeBytes);
}

//-----------------------------------------------------------------------------
//  constructor for class I2CDevice16 -- copy ctor
//-----------------------------------------------------------------------------

I2CDevice16::I2CDevice16(const I2CDevice & other)
    : I2CDeviceN(other)
{
    PageSizeBytes(kDefaultI2CDevice16_PageSizeBytes);
}

//------------------------------------------------------------------------------
//  I2CDevice16::CalculateAddressParams() - params for ReadAddress(), WriteAddress()
//------------------------------------------------------------------------------

void I2CDevice16::CalculateAddressParams(int byte_offset, int & rw_hi, int &rw_lo, int & new_offset)
{
    //  For these roms, the full SubAddr is passed in the RW Command
    rw_hi = FSubAddr>>4;
    rw_lo = FSubAddr&0xFF;
    new_offset = byte_offset;   // offset is unchanged
}

//-----------------------------------------------------------------
//  I2CDevice16::SendAddress() - Send address portion of I2C
//-----------------------------------------------------------------
//  This method sends a 2 byte address field

int  I2CDevice16::SendAddress(int address)
{
    int addr_lo = address&0xFF;
    int addr_hi = address>>8;

    int return_val = 1;
    WriteByte(addr_hi);
    if (!Acked())
        return_val = 0;

    WriteByte(addr_lo);
    if (!Acked())
        return_val = 0;

    return return_val;
}

//=============================================================================
//  CLASS I2C_MasterCore  --  Mastercore I-squared-C device control class
//=============================================================================

const unsigned int fudge_factor = 10;

//-----------------------------------------------------------------------------
//  constructor for class I2C_MasterCore -- DevAddress Version
//-----------------------------------------------------------------------------

I2C_MasterCore::I2C_MasterCore(IAddressingSpace & space, MapRegisters & regs, int sub_addr)
    :  PrescaleLoReg(space, regs.PrescaleLo_Addr),
       PrescaleHiReg(space, regs.PrescaleHi_Addr),
       ControlReg(space, regs.Control_Addr),
       DataReg(space, regs.DataIo_Addr),
       CommandReg(space, regs.CmdStatus_Addr),
       StatusReg(space, regs.CmdStatus_Addr),
       FSubAddress(sub_addr), FMaxRetries(10), I2CEnabled(false),
       FI2CRate(100e3), FLogicRate(100.0e6), WaitTimeUs(50*fudge_factor)
{
    Debug.Log("I2C_MasterCore Wait Time Us: ", WaitTimeUs);
}

//------------------------------------------------------------------------
//  destructor for class I2C_MasterCore
//------------------------------------------------------------------------

I2C_MasterCore::~I2C_MasterCore()
{
}

//------------------------------------------------------------------------
//  I2C_MasterCore::Init() - 'one time' initialization step
//------------------------------------------------------------------------

void  I2C_MasterCore::Init()
{
    //  Set prescaler to I2C Frequency
    Enabled(false);
    SetI2CFrequency(FI2CRate);
}

//------------------------------------------------------------------------
//  I2C_MasterCore::Enabled() - Change I2C Bus Enabled state
//------------------------------------------------------------------------

void  I2C_MasterCore::Enabled(bool state)
{
    if (state && I2CEnabled)    // if turning on and is on, skip it
        return;
    I2CEnabled = state;
    CoreEnable(state);
}

//------------------------------------------------------------------------
//  I2C_MasterCore::Reg() - Writes a register on an I2C device
//------------------------------------------------------------------------
//  Arguments:
//     reg = register subaddress to access
//     value = value to write
//

void  I2C_MasterCore::Reg(int reg, int value)
{
    const bool Final(true);
    //  Make sure core is on
    Enabled(true);

    // send R/W command with I2C bus address, direction
    RwCommand(FSubAddress, i2cRW_W);
    if (!Acked())
        {
        Debug.Log("I2C_MasterCore::Reg() no ACK 1");
        return;
        }
    // send register subaddress
    WriteByte(reg);
    if (!Acked())
        {
        Debug.Log("I2C_MasterCore::Reg() no ACK 2");
        return;
        }
    // send data to write to register
    WriteByte(value, Final);
    if (!Acked())
        {
        Debug.Log("I2C_MasterCore::Reg() no ACK 3");
        }
}

//------------------------------------------------------------------------
//  I2C_MasterCore::Reg - reads a register on a device via an I2C interface
//------------------------------------------------------------------------
//  Arguments:
//     addr = I2C bus address of device to access
//     reg = subaddress register number of I2C device to read
//
//   Returns value read from register in LS 8 bits
//

int I2C_MasterCore::Reg(int reg)
{
    const bool Final(true);
    //  Make sure core is on
    Enabled(true);

    // send R/W command with I2C bus address, direction = write
    RwCommand(FSubAddress, i2cRW_W);
    if (!Acked())
        {
        Debug.Log("I2C_MasterCore::Reg() no ACK 4");
        return 0;
        }
    // send register subaddress
    WriteByte(reg);
    if (!Acked())
        {
        Debug.Log("I2C_MasterCore::Reg() no ACK 5");
        return 0;
        }
    // send R/W command with I2C bus address, direction = read
    RwCommand(FSubAddress, i2cRW_R);
    if (!Acked())
        {
        Debug.Log("I2C_MasterCore::Reg() no ACK 6");
        return 0;
        }

    // read data from register & return it
    int readback = ReadByte(Final);
    return readback;
}

//------------------------------------------------------------------------
//  I2C_MasterCore::MultiRegW() - Write block of data to I2C register addr
//------------------------------------------------------------------------
//  Arguments:
//     reg = register subaddress to start with
//     data = vector of data to write
//

bool I2C_MasterCore::MultiRegW(int reg, std::vector<int> & data)
{
    const bool Final(true);
    //  Make sure core is on
    Enabled(true);

    // send R/W command with I2C bus address, direction
    RwCommand(FSubAddress, i2cRW_W);
    if (!Acked())
        {
        Debug.Log("I2C_MasterCore::MultiRegW() no ACK 1");
        return false;
        }
    // send register subaddress
    WriteByte(reg);
    if (!Acked())
        {
        Debug.Log("I2C_MasterCore::MultiRegW() no ACK 2");
        return false;
        }
    // send all but the last data byte to register
    for (unsigned int i=0; i<data.size()-1; i++)
        {
        WriteByte(data[i]);
        if (!Acked())
            {
            Debug.Log("I2C_MasterCore::MultiRegW() no ACK 3, Byte", i);
            return false;
            }
        }
    // send the last data byte to register and issue stop
    WriteByte(data[data.size()-1], Final);
    if (!Acked())
        {
        Debug.Log("I2C_MasterCore::MultiRegW() no ACK 3, Last Byte");
        return false;
        }

    return true;
}

//------------------------------------------------------------------------
//  I2C_MasterCore::MultiRegR - reads a block of data via an I2C interface
//------------------------------------------------------------------------
//  Arguments:
//     reg = subaddress register number of I2C device to read
//     data = vector of data to read - pre-sized for bytes expected
//

bool I2C_MasterCore::MultiRegR(int reg, std::vector<int> & data)
{
    const bool Final(true);
    //  Make sure core is on
    Enabled(true);

    // send R/W command with I2C bus address, direction = write
    RwCommand(FSubAddress, i2cRW_W);
    if (!Acked())
        {
        Debug.Log("I2C_MasterCore::MultiRegR() no ACK 4");
        data.clear();
        return false;
        }
    // send register subaddress
    WriteByte(reg);
    if (!Acked())
        {
        Debug.Log("I2C_MasterCore::MultiRegR() no ACK 5");
        data.clear();
        return false;
        }
    // send R/W command with I2C bus address, direction = read
    RwCommand(FSubAddress, i2cRW_R);
    if (!Acked())
        {
        Debug.Log("I2C_MasterCore::MultiRegR() no ACK 6");
        data.clear();
        return false;
        }

    // Read all the bytes but the last
    for (unsigned int i=0; i<data.size()-1; i++)
        {
        int readback = ReadByte();
        if (!Acked())
            {
            Debug.Log("I2C_MasterCore::MultiRegR() no ACK 7, Byte", i);
            data.resize((i>0) ? i-1 : 0);
            return false;
            }
        data[i]=readback;
        }
    // read the last data byte and issue stop
    int readback = ReadByte(Final);
    if (!Acked())
        {
        Debug.Log("I2C_MasterCore::MultiRegR() no ACK 7, Last Byte");
        data.resize(data.size()-1);
        return false;
        }
    data[data.size()-1]=readback;

    return true;
}

//-----------------------------------------------------------------
//  I2C_MasterCore::RwCommand() - Start I2C Command
//-----------------------------------------------------------------

void  I2C_MasterCore::RwCommand(int addr, I2C_RW dir)
{
    //  Load device address and RW direction into Data
    Data( addr | static_cast<int>(dir));

    Command(i2c_Sta | i2c_Wr);  // issue START + WRITE commands
    WaitforTIP();
}

//-----------------------------------------------------------------
//  I2C_MasterCore::WriteByte() - Write address/data to I2C
//-----------------------------------------------------------------

void  I2C_MasterCore::WriteByte(int byte_value, bool final_byte)
{
    Data(byte_value);

    int command_data = i2c_Wr;    // issue WRITE command
    if (final_byte)
        command_data |= i2c_Sto;  // ...and also STOP command

    Command(command_data);
    WaitforTIP();
}

//-----------------------------------------------------------------
//  I2C_MasterCore::ReadByte(() - Read address/data from I2C
//-----------------------------------------------------------------

int  I2C_MasterCore::ReadByte(bool final_byte)
{
    int command_data = i2c_Rd;    // issue READ command
    if (final_byte)
        command_data |= (i2c_Sto | i2c_nAck);  // ...and also STOP + NACK command
    else
        command_data |= i2c_Ack;               // ...or also ACK command

    Command(command_data);
    WaitforTIP();

    return Data();
}

//-----------------------------------------------------------------
//  I2C_MasterCore::SetI2CFrequency() - Set frequency of I2C Clock
//-----------------------------------------------------------------

void  I2C_MasterCore::SetI2CFrequency(double rate)
{
    FI2CRate = rate;
    unsigned int psf = static_cast<unsigned int>(FLogicRate/(5.0*FI2CRate));
    psf -= 1;
    LoadPrescale(psf);

    //  Recalculate wait time as well
    double bit_time = 1.0/FI2CRate;
    WaitTimeUs = static_cast<unsigned int>(5*bit_time*1e6*fudge_factor);
    Debug.Log("I2C_MasterCore::SetI2CFrequency() Wait Time Us: ", WaitTimeUs);
}

//-----------------------------------------------------------------
//  I2C_MasterCore::Acked() - check for ACK from I2C
//-----------------------------------------------------------------

bool  I2C_MasterCore::Acked()
{
    bool Result = StatusReg.RxAck.Value();
    return (Result==false);   // 0 is 'acked'
}

//-----------------------------------------------------------------
//  I2C_MasterCore::WaitforTIP() - Wait for TIP signal to complete
//-----------------------------------------------------------------

void  I2C_MasterCore::WaitforTIP()
{
    int retries = FMaxRetries;
    while (StatusReg.TIP.Value())
        {
        //  count retries and abandon after N tries with message
        if (retries--==0)
            {
            Debug.Log("I2C_MasterCore TIP retries failed");
            return;
            }
        //Rtc.uS(WaitTimeUs);
        uSleep(WaitTimeUs);
        }
}

//-----------------------------------------------------------------
//  I2C_MasterCore::Data() - Read to data register
//-----------------------------------------------------------------

int  I2C_MasterCore::Data()
{
    return DataReg.Value() & 0xFF;
}

//-----------------------------------------------------------------
//  I2C_MasterCore::Data() - Write to data register
//-----------------------------------------------------------------

void  I2C_MasterCore::Data(int value)
{
    DataReg.Value(value);
}

//-----------------------------------------------------------------
//  I2C_MasterCore::Command() - Write to command register
//-----------------------------------------------------------------

void  I2C_MasterCore::Command(int value)
{
    CommandReg.Value(value);
}

//-----------------------------------------------------------------
//  I2C_MasterCore::CoreEnable() - Write to core enable bit
//-----------------------------------------------------------------

void  I2C_MasterCore::CoreEnable(bool state)
{
    ControlReg.CoreEnable.Value(state);
}

//-----------------------------------------------------------------
//  I2C_MasterCore::LoadPrescale() - Write value to prescale reg
//-----------------------------------------------------------------

void  I2C_MasterCore::LoadPrescale(unsigned int value)
{
    PrescaleLoReg.Ps_Value.Value(value&0xFF);
    PrescaleHiReg.Ps_Value.Value(value>>8);
}

//-----------------------------------------------------------------
//  I2C_MasterCore::I2C_Reg() - Write I2C device register
//-----------------------------------------------------------------

void  I2C_MasterCore::I2C_Reg(int reg, int value)
{
    Reg(reg, value);
}

//-----------------------------------------------------------------
//  I2C_MasterCore::I2C_Reg() - Read I2C device register
//-----------------------------------------------------------------

int  I2C_MasterCore::I2C_Reg(int reg)
{
    return Reg(reg);
}

#ifdef LINUX
//    #pragma GCC diagnostic pop
#endif


} // namespace Innovative
