//
// I2c_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2003 Innovative Integration Corporation.
//  All Rights Reserved.
//

#ifndef I2c_MbH
#define I2c_MbH

#include <StopWatch_Mb.h>
#include <HardwareRegister_Mb.h>
#include <Trace_Mb.h>

namespace Innovative
{

//
//  "I2C I/O Requirements" Interfaces - Define which pieces of the I2C protocol
//     are required by an I2C client
//
//===========================================================================
//  CLASS IRequires_I2C_RegIo -- Client Io Requirement Interface
//===========================================================================
//  This client needs Register input and output

class IRequires_I2C_RegIo
{
public:
    virtual void I2C_Reg(int reg, int value) = 0;
    virtual int  I2C_Reg(int reg) = 0;

    virtual void I2C_SetSubAddress(int subaddress) = 0;

    virtual ~IRequires_I2C_RegIo()
        {}
};

//
//  "I2C I/O  Supplier" Interfaces - Define the set of requirements supplied
//                                      by an I2C device
//===========================================================================
//  CLASS ISuppliesI2CDeviceRequirements -- What's supplied by I2CDevice
//===========================================================================

class ISuppliesI2CDeviceRequirements : public IRequires_I2C_RegIo
{
public:
    virtual ~ISuppliesI2CDeviceRequirements()
        {}

};

//===========================================================================
//  CLASS ISuppliesMastercoreRequirements -- What's supplied by I2C_MasterCore
//===========================================================================

class ISuppliesMastercoreRequirements : public IRequires_I2C_RegIo
{
public:
    virtual ~ISuppliesMastercoreRequirements()
        {}

};

//============================================================================
//  CLASS I2CDeviceIntf  --  I-squared-C device interface class
//=============================================================================

class I2CDeviceIntf
{
public:
    // Methods
    virtual void    SetClock(int state) = 0;
    virtual int     GetClock() const = 0;
    virtual void    SetData(int state) = 0;
    virtual int     GetData() const = 0;

    virtual ~I2CDeviceIntf()
        {}
    
};

//=============================================================================
//  CLASS I2CDevice  --  I-squared-C device control class
//=============================================================================

class I2CDevice : public ISuppliesI2CDeviceRequirements
{
public:
    enum SubAddrTag { SubAddress };
    enum DevAddrTag { DevAddress };
    //
    //  Member Functions
    I2CDevice(I2CDeviceIntf & intf, SubAddrTag sub, int sub_addr,
              bool endian=false, bool bit_reversed=false);
    I2CDevice(I2CDeviceIntf & intf, DevAddrTag dev, int dev_addr,
              bool endian=false, bool bit_reversed=false);
    I2CDevice(const I2CDevice & other);
    virtual ~I2CDevice();

    I2CDevice & operator =(const I2CDevice & other);

    //
    //  Properties
    int             Reg_sub(int sub_addr, int reg, int value);
    int             Reg_sub(int sub_addr, int reg);

    int             Reg(int reg, int value);
    int             Reg(int reg);
    void            LittleReg(int rs, int value);

    int             MultiRegW(int reg, const int * data, int size);
    int             MultiRegR(int reg, int * data, int size);

    bool            NackReads()
                       { return FNackReads; }
    void            NackReads(bool state)
                       { FNackReads = state; }

    void            Byte_dev(int dev_addr, int reg, int value);
    int             Byte_dev(int dev_addr, int reg);

    int             Byte(int byte_offset);
    void            Byte(int byte_offset, int data);

    void            BitReversed(bool state);
    bool            BitReversed() const;

    void            PageSizeBytes(unsigned int bytes)
                        {  FPageSizeBytes = bytes;  }
    unsigned int    PageSizeBytes() const
                        {  return FPageSizeBytes;  }

    // Methods
    bool            ReadBlock(int offset, int * data, int words);
    bool            WriteBlock(int offset, const int * data, int words);
    bool            WritePage(int offset, const int * page_data);

    // Debug
    bool            DeviceAddress(int byte_offset);
    int             DeviceByte(int byte_offset);

    //  Overrides
    unsigned int    IntsPerPage();
    void            DevAddr_Override(int dev_addr);
    void            SubAddr_Override(int sub_addr);
    int             GetSubAddr()
                        {  return FSubAddr;  }

    //
    //  Supplied Requirements
    virtual void  I2C_Reg(int reg, int value);
    virtual int   I2C_Reg(int reg);
    virtual void  I2C_SetSubAddress(int sub_addr)
                      {  SubAddr_Override(sub_addr);  }

protected:
    //
    //  Member Data
    int             FSubAddr;
    int             FDevAddr;
    StopWatch       Rtc;
    bool            FEndian;
    bool            FBitReversed;
    bool            FNackReads;
    bool            FStretchableClock;
    unsigned int    FPageSizeBytes;
    I2CDeviceIntf & Intf;

    // Methods
    bool            WritePortion(int offset, const int * page_data, int words);

    //
    //  Overrides
    bool            WriteAddress(int byte_offset);
    bool            ReadAddress(int byte_offset);
    void            tCLK();

    void            Clock(int state)    {  Intf.SetClock(state);  Stretch(state ? true : false);  }
    int             Clock() const       {  return static_cast<int>(Intf.GetClock());  }
    void            Data(int state)     {  Intf.SetData(state);     }
    int             Data() const        {  return Intf.GetData();   }

    // Override Methods
    virtual int     SendAddress(int address);   // transmit address field (here 8 bits)
    virtual void    CalculateAddressParams(int byte_offset, int & rw_hi, int &rw_lo, int & new_offset);

    int             ReadBit();
    void            WriteBit(int data, int bit_no );
    unsigned int    ReadByte(bool bit_reverse=false);
    void            WriteByte(int byte_value, bool bit_reverse=false);
    void            WriteNHLCD(int rs, int value);
    void            RwCommand(int addr, int rw_bit);
    void            RwCommand(int device_type, int addr, int rw_bit);
    bool            Acked();
    void            Start();
    void            StartNHLCD();
    void            Stop();
    void            Ack();
    void            Nack();
    void            ClockPulse();
    bool            Ready(int byte_offset);
    void            Stretch(bool clock_state);
};

//=============================================================================
//  CLASS I2CDeviceN  --  I-squared-C device control class - defaults to NAK reads
//=============================================================================

class I2CDeviceN : public I2CDevice
{
public:
    //
    //  Member Functions
    I2CDeviceN(I2CDeviceIntf & intf, SubAddrTag sub, int sub_addr,
              bool endian=false, bool bit_reversed=false)
        : I2CDevice(intf, sub, sub_addr, endian, bit_reversed)
        {  FNackReads = true; FStretchableClock=false;  }
    I2CDeviceN(I2CDeviceIntf & intf, DevAddrTag dev, int dev_addr,
              bool endian=false, bool bit_reversed=false)
         : I2CDevice(intf, dev, dev_addr, endian, bit_reversed)
        {  FNackReads = true; FStretchableClock=false; }
    I2CDeviceN(const I2CDevice & other)
        : I2CDevice(other)
        {  FNackReads = true; FStretchableClock=false; }

};

//=============================================================================
//  CLASS I2CDevice16  --  16 bit address I2C device
//=============================================================================

class I2CDevice16 : public I2CDeviceN
{
public:
    //
    //  Member Functions
    I2CDevice16(I2CDeviceIntf & intf, SubAddrTag sub, int sub_addr,
                    bool endian=false, bool bit_reversed=false);
    I2CDevice16(const I2CDevice & other);

    // Override Methods
    virtual int     SendAddress(int address);   // transmit address field (here 16 bits)
    virtual void    CalculateAddressParams(int byte_offset, int & rw_hi, int &rw_lo, int & new_offset);

};


#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif

//=============================================================================
//  CLASS I2C_MasterCore  --  Mastercore I-squared-C device control class
//=============================================================================
//  I2C device based off of MasterCore logic core

class I2C_MasterCore : public ISuppliesI2CDeviceRequirements
{
public:

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS I2C_MasterCore::MapRegisters  -- Device Register Offsets
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    class MapRegisters
    {
    public:
        unsigned int   PrescaleLo_Addr;
        unsigned int   PrescaleHi_Addr;
        unsigned int   Control_Addr;
        unsigned int   DataIo_Addr;
        unsigned int   CmdStatus_Addr;
    };
    //
    //  Member Functions
    I2C_MasterCore(IAddressingSpace & space, MapRegisters & regs, int sub_addr);
    virtual ~I2C_MasterCore();

    virtual void    Init();
    //
    //  Properties
    bool            Enabled()
                        {  return I2CEnabled;  }
    void            Enabled(bool state);

    void            Reg(int reg, int value);
    int             Reg(int reg);

    bool            MultiRegW(int reg, std::vector<int> & data);
    bool            MultiRegR(int reg, std::vector<int> & data);

    void            PresetI2CFrequency(double rate)   // use only before Init()
                        {  FI2CRate = rate;  }        //   otherwise use SetI2CFrequency
    void            SetRetryPeriod(int uS)
                        {  FMaxRetries = uS;  }
    void            SetI2CFrequency(double rate);
    void            SetLogicFrequency(double rate)
                        {  FLogicRate = rate;  }
    //
    //  Supplied Requirements
    virtual void    I2C_Reg(int reg, int value);
    virtual int     I2C_Reg(int reg);
    virtual void    I2C_SetSubAddress(int sub_addr)
                        {  FSubAddress = sub_addr;  }

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS I2C_MasterCore::PrescaleRegister  -- Prescale Register
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    class PrescaleRegister : public Register
    {
        typedef Register inherited;

    public:
        RegisterBitGroup  Ps_Value;

    public:
        PrescaleRegister( IAddressingSpace &space, int offset )
            :  inherited(space, offset),
               Ps_Value(*this, 0, 8)
            {
            }
    };

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS I2C_MasterCore::ControlRegister  -- Control Register
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    class ControlRegister : public Register
    {
        typedef Register inherited;

    public:
        RegisterBit  CoreEnable;
        RegisterBit  IntEnable;

    public:
        ControlRegister( IAddressingSpace &space, int offset )
            :  inherited(space, offset),
               CoreEnable(*this, 7),
               IntEnable(*this, 6)
            {
            }
    };
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS I2C_MasterCore::StatusRegister  -- Status Register
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    class StatusRegister : public ReadOnlyRegister
    {
        typedef ReadOnlyRegister inherited;

    public:
        ReadOnlyRegisterBit  RxAck;
        ReadOnlyRegisterBit  Busy;
        ReadOnlyRegisterBit  ArbitrationLost;
        ReadOnlyRegisterBit  TIP;     // Transfer in Progress
        ReadOnlyRegisterBit  IF;

    public:
        StatusRegister( IAddressingSpace &space, int offset )
            :  inherited(space, offset),
               RxAck(*this, 7),   Busy(*this, 6),
               ArbitrationLost(*this, 5),
               TIP(*this, 1),
               IF(*this, 0)
            {
            }
    };

private:
    //
    //  Registers
    PrescaleRegister   PrescaleLoReg;
    PrescaleRegister   PrescaleHiReg;
    ControlRegister    ControlReg;
    Register           DataReg;
    Register           CommandReg;   // Write Only!
    StatusRegister     StatusReg;

    int             FSubAddress;   // 7 bit address, left justified in LSB
    int             FMaxRetries;
    bool            I2CEnabled;
    double          FI2CRate;
    double          FLogicRate;
    unsigned int    WaitTimeUs;

    //StopWatch   Rtc;
    Trace       Debug;

    enum I2C_RW { i2cRW_W=0x0, i2cRW_R=0x1 };   // 'RW Command'
    enum I2C_Commands { i2c_Ack=0x0, i2c_IAck=0x1, i2c_nAck=0x8,
                        i2c_Wr=0x10, i2c_Rd=0x20,  i2c_Sto=0x40, i2c_Sta=0x80 };

    //
    //  Mid Level Ops
    void   RwCommand(int addr, I2C_RW dir);
    int    ReadByte(bool final_byte=false);
    void   WriteByte(int byte_value, bool final_byte=false);
    //
    //  Low Level Ops
    int    Data();
    void   Data(int value);
    void   Command(int value);
    bool   Acked();
    void   WaitforTIP();
    void   CoreEnable(bool state);
    void   LoadPrescale(unsigned int value);

};

#ifdef _MSC_VER
    #pragma warning(pop)
#endif

} // namespace Innovative

#endif
