//
// Ics843002_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2003 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <ICS843002_Mb.h>
#include <SystemSupport_Mb.h>
#include <algorithm>
#include <limits>
#include <cmath>

//---------------------------------------------------------------------------

namespace Innovative
{

//=============================================================================
//  CLASS Ics843002  --  Ics843002 Pll Implementation
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class Ics843002
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//  constructor for class Ics843002
//-----------------------------------------------------------------------------

Ics843002::Ics843002(int index, AddressingSpace & space, int offset)
	: ClockBase(index),
	  QaReg(space, offset+0x400),
	  QbReg(space, offset+0x800),
	  VclkReg(space, offset+0xc00),
	  InputReg(space, offset+0x1400), 
	  FeedbackReg(space, offset+0x1000),
	  ControlReg(space, offset),
	  StatusReg(space, offset),
      Debug(false)
{
	Source(rcTcxo);
	Reference(20.0e6);    // Pullable reference clock
}

//-----------------------------------------------------------------------------
//  constructor for class Ics843002
//-----------------------------------------------------------------------------

Ics843002::Ics843002(double freq, double ref, bool enabled, int index,
					 AddressingSpace & space, int offset)
	: ClockBase(freq, ref, enabled, index),
	  QaReg(space, offset+0x400),
	  QbReg(space, offset+0x800),
	  VclkReg(space, offset+0xc00),
	  InputReg(space, offset+0x1400), 
	  FeedbackReg(space, offset+0x1000),
	  ControlReg(space, offset),
	  StatusReg(space, offset),
      Debug(false)
{
	Reference(ref);
	Source(rcTcxo);

	DoSetFrequency(freq);
	DoSetEnabled(enabled);
}

//-----------------------------------------------------------------------------
//  Ics843002::GetFrequencyActual()  --  Calculate actual clock rate
//-----------------------------------------------------------------------------

double Ics843002::GetFrequencyActual() const
{
	double fempto = ControlReg.Multiplier ? 28.0 : 32.0;
	double result = (Reference() * fempto) / (QaReg.Divisor() * QbReg.Divisor());

	return result;
}

//-----------------------------------------------------------------------------
//  Ics843002::Max()  --  Return maximum output rate
//-----------------------------------------------------------------------------

double Ics843002::Max() const
{
	return Reference() * (32.0/1.0);
}

//-----------------------------------------------------------------------------
//  Ics843002::Min()  --  Return minimum output rate
//-----------------------------------------------------------------------------

double Ics843002::Min() const
{
	return Reference() * (28.0/128.0);
}

//-----------------------------------------------------------------------------
//  Ics843002::Source()  -- Set input clock source
//-----------------------------------------------------------------------------

Ics843002 & Ics843002::Source(IIClockSource value)
{
	switch (value)
		{
		default:
		case rcTcxo: 	ControlReg.Mux.Value(1);		break;
		case rcSyncIn: 	ControlReg.Mux.Value(2);		break;
		}
	return *this;
}

//-----------------------------------------------------------------------------
//  Ics843002::Source()  -- Fetch input clock source
//-----------------------------------------------------------------------------

Ics843002::IIClockSource Ics843002::Source() const
{
	IIClockSource result;

	switch (ControlReg.Mux.Value())
		{
		default:
		case 1:     result = rcTcxo;		break;
		case 2:		result = rcSyncIn;		break;
		}

	return result;
}

//-----------------------------------------------------------------------------
//  Ics843002::SourceFrequency()  -- Set input clock frequency
//-----------------------------------------------------------------------------

Ics843002 & Ics843002::SourceFrequency(double value)
{
	FSourceFrequency = value;
	return *this;
}

//-----------------------------------------------------------------------------
//  Ics843002::SourceFrequency()  -- Fetch input clock frequency
//-----------------------------------------------------------------------------

double Ics843002::SourceFrequency() const
{
	return FSourceFrequency;
}

//-----------------------------------------------------------------------------
//  Ics843002::Locked()  --  Return lock status
//-----------------------------------------------------------------------------

bool Ics843002::Locked() const
{
	return StatusReg.Lock.Value() ? true : false;
}

//-----------------------------------------------------------------------------
//  Ics843002::SetFrequency()  --  Calculate settings for a frequency
//-----------------------------------------------------------------------------

void Ics843002::SetFrequency(double freq)
{
	DoSetFrequency(freq);
}

//-----------------------------------------------------------------------------
//  Ics843002::DoSetFrequency()  --  Calculate settings for a frequency
//-----------------------------------------------------------------------------

void Ics843002::DoSetFrequency(double freq)
{
	if (FReference == 0.0)
		return;

	freq *= ClockFactor();

	// Adjust VCXO to match pullable XO frequency
	double ratio = Reference()/SourceFrequency();
	RationalApproximation Ratio(static_cast<float>(ratio), 14);
	InputReg.Value(Ratio.Denominator());
	FeedbackReg.Value(Ratio.Numerator());

	Debug.Log("Ics843002 feedback reg (*): ", Ratio.Numerator());
	Debug.Log("Ics843002 input reg (%): ", Ratio.Denominator());
	Debug.Log("Ics843002 ratio = ", static_cast<double>(Ratio.Numerator())/Ratio.Denominator());

	// Adjust Femptoclock and divisor to best match requested
	// output frequency
	ClosestQAMatch(freq);

	ControlReg.OutputEnable.Value(true);

	// Reset and run
	ControlReg.Reset.Value(1);
	ControlReg.Reset.Value(0);

	// Wait till lock
	for (int i = 0; !Locked() && (i < 100); ++i)
		Sleep(1);

	if (!Locked())
		Debug.Log("Ics843002 PLL lock failure");

	FEnabled = true;
}

//-----------------------------------------------------------------------------
//  Ics843002::Disable()  --  Place device in reset
//-----------------------------------------------------------------------------

void Ics843002::Disable()
{
	ControlReg.Reset.Value(1);
}

//-----------------------------------------------------------------------------
//  Ics843002::ClosestQAMatch()  --  Select optimal fempto, Qa and Qb settings
//-----------------------------------------------------------------------------

void Ics843002::ClosestQAMatch(double freq)
{
	const int Fempto[] = { 32, 28 };
	const int Qa[] = { 1, 2, 4, 8, 12, 14, 16 };
	const int Qb[] = { 1, 2, 4, 8 };

	double error = std::numeric_limits<double>::max();

	int fempto = 32;
    int qa = 0;
    int qb = 0;;

	for (int i = 0; i < 2; ++i)
		{
		for (int j = 0; j < 7; ++j)
			{
			for (int k = 0; k < 4; ++k)
				{
				double numerator = Fempto[i];
				double denominator = Qa[j]*Qb[k];
				double f = Reference()*numerator/denominator;
				double err = std::fabs(f-freq);
				if (err < error)
					{
					fempto = i;
					qa = j;
					qb = k;
					error = err;
					}
				}
			}
		}

	ControlReg.Multiplier.Value(fempto ? true : false);
	QaReg.Divisor(Qa[qa]);
	QbReg.Divisor(Qb[qb]);
}

//-----------------------------------------------------------------------------
//  Ics843002::SetEnabled()  --  Activate / Deactivate output
//-----------------------------------------------------------------------------

void  Ics843002::SetEnabled(bool state)
{
	DoSetEnabled(state);
}

//-----------------------------------------------------------------------------
//  Ics843002::DoSetEnabled()  --  Activate / Deactivate output
//-----------------------------------------------------------------------------

void  Ics843002::DoSetEnabled(bool /*state*/)
{
}

//-----------------------------------------------------------------------------
//  Ics843002::Apply()  --  Reapply settings to hardware from saved state
//-----------------------------------------------------------------------------

void  Ics843002::Apply()
{
	QaReg.Apply();
	QbReg.Apply();
	VclkReg.Apply();
	FeedbackReg.Apply();
	InputReg.Apply();
	ControlReg.Apply();
}

} // namespace Innovative
