//
// Ics8442_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2003 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <ICS8442_Mb.h>
#include <SystemSupport_Mb.h>
#include <algorithm>

//---------------------------------------------------------------------------

namespace Innovative
{

//=============================================================================
//  CLASS Ics8442  --  Ics8442 Pll Implementation
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class Ics8442
//-----------------------------------------------------------------------------

const double Ref_Freq_Mhz = 14.4;

//-----------------------------------------------------------------------------
//  constructor for class Ics8442
//-----------------------------------------------------------------------------

Ics8442::Ics8442(int index, AddressingSpace & space, int offset)
	: ClockBase(index), CtlReg(space, offset)
{
    TestMode(toDisabled);
    ReferenceClock(rcCrystal);
    MultiplierEnable(true);
    Reference(Ref_Freq_Mhz*1.e6);    // 14.4 Mhz internal clock
}

//-----------------------------------------------------------------------------
//  constructor for class Ics8442
//-----------------------------------------------------------------------------

Ics8442::Ics8442(double freq, double ref, bool enabled, int index,
				 AddressingSpace & space, int offset)
	: ClockBase(freq, ref, enabled, index),
	  CtlReg(space, offset)
{
	Reference(ref);
	TestMode(toDisabled);
	ReferenceClock(rcCrystal);
	MultiplierEnable(true);

	DoSetFrequency(freq);
	DoSetEnabled(enabled);
}

//-----------------------------------------------------------------------------
//  Ics8442::GetFrequencyActual()  --  Calculate actual clock rate
//-----------------------------------------------------------------------------

double Ics8442::GetFrequencyActual() const
{
	return (CtlReg.M.Value() * Reference()) /
		((1<<CtlReg.N.Value()) * ClockFactor());
}

//-----------------------------------------------------------------------------
//  Ics8442::ReferenceClock()  -- Set VCO reference
//-----------------------------------------------------------------------------

Ics8442 & Ics8442::ReferenceClock(IIReferenceClk value)
{
	CtlReg.Reference.Value((value==rcDivider) ? 0 : 1);
	return *this;
}

//-----------------------------------------------------------------------------
//  Ics8442::ReferenceClock()  -- Set VCO reference
//-----------------------------------------------------------------------------

Ics8442::IIReferenceClk Ics8442::ReferenceClock() const
{
	return CtlReg.Reference.Value() ? rcCrystal : rcDivider;
}

//-----------------------------------------------------------------------------
//  Ics8442::MultiplierEnable()  -- Set VCO bypass state
//-----------------------------------------------------------------------------

Ics8442 & Ics8442::MultiplierEnable(bool state)
{
    CtlReg.MultiplierEnable.Value(state ? 1 : 0);
	return *this;
}

//-----------------------------------------------------------------------------
//  Ics8442::MultiplierEnable()  -- Return VCO state
//-----------------------------------------------------------------------------

bool Ics8442::MultiplierEnable() const
{
	return CtlReg.MultiplierEnable.Value() ? true : false;
}

//-----------------------------------------------------------------------------
//  Ics8442::TestMode()  -- Output mode in test mode
//-----------------------------------------------------------------------------

Ics8442 &  Ics8442::TestMode(IITestMode output)
{
	CtlReg.TestOutput.Value(output);
	return *this;
}

//-----------------------------------------------------------------------------
//  Ics8442::TestMode()  --  Output mode in test mode
//-----------------------------------------------------------------------------

Ics8442::IITestMode Ics8442::TestMode() const
{
    IITestMode result;

    switch (CtlReg.TestOutput.Value())
        {
        case 0:     result = toDisabled;        break;
        case 1:     result = toShiftReg;        break;
        case 2:     result = toFeedbackDivider; break;
        default:
        case 3:     result = toNormal;          break;
        }

    return result;
}

//-----------------------------------------------------------------------------
//  Ics8442::SetFrequency()  --  Calculate settings for a frequency
//-----------------------------------------------------------------------------

void Ics8442::SetFrequency(double freq)
{
	DoSetFrequency(freq);
}

//-----------------------------------------------------------------------------
//  Ics8442::DoSetFrequency()  --  Calculate settings for a frequency
//-----------------------------------------------------------------------------

void Ics8442::DoSetFrequency(double freq)
{
    //  Cache Frequency
    FFrequency = freq;

	freq *= ClockFactor();
    freq /= 1.e6;
    //
    //  Determine which frequency divisor to use...
    if (freq < 87.5)       // 1st band to 87.5 Mhz (divide by 8)
        CtlReg.N.Value(3);
    else if (freq < 175.0) // 2nd band to 175 Mhz (divide by 4)
        CtlReg.N.Value(2);
    else if (freq < 350.0) // 3rd band to 350 Mhz (divide by 2)
        CtlReg.N.Value(1);
    else                   // 4th band to 700 Mhz (divide by 1)
        CtlReg.N.Value(0);
    //
    //  Calculate M field
    int M_val = static_cast<int>(freq * (1<<CtlReg.N.Value()) / (Reference()/1.e6));

	//  'M' field limits:  The "M" parameter must be chosen to provide a 'native'
	//      frequency between 250 and 700 Mhz. This can then be divided by the 'N' factor.
	//          M(min) = F(min)/F(xtal)     M(max) = F(max)/F(xtal)
	const int M_max = static_cast<int>(700.e6 / Reference()); 
	const int M_min = static_cast<int>(250.e6 / Reference()); 

    //  ...Only allow M values in our sync band...
    M_val = std::min(M_max, M_val);
    M_val = std::max(M_min, M_val);

    CtlReg.M.Value(M_val);

    CtlReg.Reset = 1;
	Sleep(1);
    CtlReg.Reset = 0;
    Sleep(1);
}

//-----------------------------------------------------------------------------
//  Ics8442::SetEnabled()  --  Activate / Deactivate output
//-----------------------------------------------------------------------------

void  Ics8442::SetEnabled(bool state)
{
	DoSetEnabled(state);
}

//-----------------------------------------------------------------------------
//  Ics8442::DoSetEnabled()  --  Activate / Deactivate output
//-----------------------------------------------------------------------------

void  Ics8442::DoSetEnabled(bool /*state*/)
{
}

//-----------------------------------------------------------------------------
//  Ics8442::Reapply()  --  Reapply settings to hardware from saved state
//-----------------------------------------------------------------------------

void  Ics8442::Reapply()
{
	CtlReg.Apply();
}

} // namespace Innovative
