// IniFile_Mb.h --  Class to access local configuration files
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2006 Innovative Integration Corporation.
//  All Rights Reserved.
//

#ifndef IniFile_MbH
#define IniFile_MbH

#include <DataTypes_Mb.h>
#include <StringList_Mb.h>
#include <vector>
#include <map>
#include <StringSupport_Mb.h>   
#include <Trace_Mb.h>





#ifdef __CLR_VER
#pragma managed(push, off)
#endif

// Forward declaration
namespace Poco { namespace Util {
    class JSONConfiguration;
} }

namespace Innovative
{

//===========================================================================
//  IniProtocol --  manage read and write Interface for INI style key/value
//===========================================================================

class IniProtocol
{
public:

	typedef std::vector<char>       CharArray;
	typedef std::vector<int>        IntArray;
	typedef std::vector<float>      FloatArray;
	typedef std::vector<double>     DoubleArray;

    //
    //  Member Functions

    // Constructor
    IniProtocol(int max_len);
    virtual ~IniProtocol();

    void        DeleteKey(const std::string & key);

    void        DeleteSection(const std::string & section);
    void        DeleteSection();

    void        ReadKeys(const std::string & section, StringList & strings);
    void        ReadKeys(StringList & strings);

    void        ReadValues(const std::string & section, StringList & list);
    void        ReadValues(StringList & list);

    bool        KeyExists(const std::string & section, const std::string & key);
    bool        KeyExists(const std::string & key);

    bool        SectionExists(const std::string & section);

    //  Read string list from tag
    virtual int Read(const std::string & key, StringList & list, bool numbered=false);
    //  Read bool from tag
    virtual bool  Read(const std::string & tag, bool defValue);
    //  Read integer from tag
    virtual int  Read(const std::string & tag, int defValue);
    virtual unsigned int  Read(const std::string & tag, unsigned int defValue);
    //  Read long integer from tag
    virtual ii64 Read(const std::string & tag, ii64 defValue);
    //  Read signed long integer from tag
    virtual s_ii64 Read(const std::string & key, s_ii64 value);
    //  Read float from tag
    virtual float  Read(const std::string & tag, float defValue);
    //  Read double from tag
    virtual double  Read(const std::string & tag, double defValue);
    virtual double  ReadDP(const std::string & tag, double defValue);
    //  Read string from tag
    virtual std::string  Read(const std::string & tag, const std::string & defValue);

    //  Read character array from tag
    virtual size_t Read(const std::string & tag, CharArray &data);
    //  Read integer array from tag
    virtual size_t Read(const std::string & tag, IntArray &data);
    //  Read float array from tag
    virtual size_t Read(const std::string & tag, FloatArray &data);

    //  Write string list from tag
    virtual void  Write(const std::string & tag, StringList & list, bool numbered=false);
    //  Write bool to tag
    virtual void  Write(const std::string & tag, bool Value);
    //  Write integer to tag
    virtual void  Write(const std::string & tag, int Value);
    virtual void  Write(const std::string & tag, unsigned int Value);
    //  Write long integer to tag
    virtual void  Write(const std::string & tag, ii64 Value);
    //  Write signed long integer to tag
    virtual void  Write(const std::string & tag, s_ii64 Value);
    //  Write float to tag
    virtual void  Write(const std::string & tag, float Value);
    //  Write double to tag
    virtual void  Write(const std::string & tag, double Value);
    virtual void  WriteDP(const std::string & tag, double Value);
    //  Write string to tag
    virtual void  Write(const std::string & tag, const std::string & Value);

    //  Write character array to tag
    virtual void Write(const std::string & tag, const CharArray &data);
    //  Write integer array to tag
    virtual void Write(const std::string & tag, const IntArray &data);
    //  Write float array to tag
    virtual void Write(const std::string & tag, const FloatArray &data);

    std::string                 Section;

protected:
    //
    //  Derived class stream Io interface
    virtual void  LoadFromStream(std::istream & stream);
    virtual void  SaveToStream(std::ostream & stream);

    virtual void  Clear();

private:
#ifdef _MSC_VER
#pragma warning (push)
#pragma warning (disable : 4996)
#endif

    struct StlIniCompareStringNoCase
    {
        // return true or false depending on whether the first string is less than the second
        inline bool operator()(const std::string& x, const std::string& y) const
            {
            return (StringCompare(x, y) < 0) ? true : false;
            }
    };

#ifdef _MSC_VER
#pragma warning (pop)
#endif

    // these typedefs just make the code a bit more readable
    typedef std::map<std::string, std::string, StlIniCompareStringNoCase > INISection;
    typedef std::map<std::string, INISection , StlIniCompareStringNoCase > INIFile;

    // Support Methods
    void PutKeys(const std::string & section, const StringList & list);
    void PutValues(const std::string & section, const StringList & list);
    void PutSetting(const std::string & section, const std::string & key,
                    const std::string & value);

    bool RemoveSetting(const std::string & section, const std::string & key);
    bool RemoveSection(const std::string & section);

    void SaveSection(std::ostream & file, INIFile::iterator & section);

    bool GetSettings(const std::string & section, StringList & list);
    std::string GetValue(const std::string & section, const std::string & key,
                         const std::string & def);

    //
    //  Member Data
    INIFile                 Ini;
    int                     MaxLength;
};


//===========================================================================
//  JsonProtocol --  manage read and write Interface for JSON style key/value
//===========================================================================

class JsonProtocol : public IniProtocol
{
public:
    //
    //  Member Functions

    // Constructor
    JsonProtocol(int max_len);
    virtual ~JsonProtocol();

    // Single element read and writes
    bool  Read(const std::string & , bool );
    int  Read(const std::string & , int ) ;
    unsigned int  Read(const std::string & , unsigned int );
    ii64 Read(const std::string & , ii64 );
    s_ii64 Read(const std::string & , s_ii64 );
    float  Read(const std::string & , float );
    double  Read(const std::string & , double );
    double  ReadDP(const std::string & , double );
    std::string  Read(const std::string & , const std::string & );

    void  Write(const std::string & tag, bool Value);
    void  Write(const std::string & tag, int Value);
    void  Write(const std::string & tag, unsigned int Value);
    void  Write(const std::string & tag, ii64 Value);
    void  Write(const std::string & tag, s_ii64 Value);
    void  Write(const std::string & tag, float Value);
    void  Write(const std::string & tag, double Value);
    void  WriteDP(const std::string & tag, double Value);
    void  Write(const std::string & tag, const std::string & Value);

    // Arrays
    int  Read(const std::string &tag , StringList & , bool numbered=false);
    void Read(const std::string &tag, vector<string> &ss);
    size_t Read(const std::string &tag, CharArray &);
    size_t Read(const std::string &tag, IntArray &);
    size_t Read(const std::string &tag, FloatArray &);
    size_t Read(const std::string &tag, DoubleArray &);
    void Write(const std::string &tag, const vector<string> &ss);
    void Write(const std::string &tag, StringList & , bool numbered=false);
    void Write(const std::string &tag, const CharArray &);
    void Write(const std::string &tag, const IntArray &);
    void Write(const std::string &tag, const FloatArray &);
    void Write(const std::string &tag, const DoubleArray &);

    // File interface
    void LoadFromStream(std::istream & stream);
    void SaveToStream(std::ostream & stream);

    Poco::Util::JSONConfiguration *getImpl();

private:
    struct Jimpl *pimpl;

};


//===========================================================================
//  IniFile  --  Read and Write file as INI format text
//===========================================================================

class IniFile :  public IniProtocol
{
public:
    IniFile(const std::string & filename, bool create=false,
            int max_len=100000, bool autosave=true);
    ~IniFile();

    //  Flush buffered output to file
    void  Update();
	void  SetAutoSave(bool state)
		{  AutoSave = state;  }

private:

    void Save();
    void Load();
    //
    //  Member Data
    std::string             Filename;
    bool                    AutoSave;
};

//===========================================================================
//  JsonFile  --  Read and Write file as JSON format text
//===========================================================================
class JsonFile :  public JsonProtocol
{
public:
    JsonFile(const std::string & filename, bool create=false,
            int max_len=100000, bool autosave=true);
    ~JsonFile();

    //  Flush buffered output to file
    void  Update();
    void  SetAutoSave(bool state)
        {  AutoSave = state;  }

private:

    void Save();
    void Load();
    //
    //  Member Data
    std::string             Filename;
    bool                    AutoSave;
};

//===========================================================================
//  IniMemory  --  Read and Write to memory region as INI format text
//===========================================================================

class IniMemory : public IniProtocol
{
public:
    IniMemory();
    IniMemory(const std::string & text, int max_len=100000);
    ~IniMemory();

    //  output text as INI string
    std::string  Text();
    void         Text(const std::string & text);

};


}  // namespace Innovative

#ifdef __CLR_VER
#pragma managed(pop)
#endif

#endif


