//---------------------------------------------------------------------------
// IppComplexDG_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2007 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include "IppComplexDG_Mb.h"
#include "IppFloatDG_Mb.h"
#include <ipps.h>   // ipps.h is not redistributable!
#include <algorithm>

namespace
{
//
//  The IPP library uses a differently named structure for its complex elements
//     than we do. They share the same in memory layout. Since C++ doesnt know
//     this, we use these functions to bludgeon the compiler into letting us do
//     what we want.
inline Ipp32fc ToIppComplex(Innovative::Complex & complex_value)
    {
    Ipp32fc v;
    v.re = complex_value.real();
    v.im = complex_value.imag();

    return *(&v);
    }
inline Ipp32fc * ToIppComplexPtr(Innovative::Complex * complex_ptr)
    {
    return reinterpret_cast<Ipp32fc *>(complex_ptr);
    }
inline const Ipp32fc * ToIppComplexPtr(const Innovative::Complex * complex_ptr)
    {
    return reinterpret_cast<const Ipp32fc *>(complex_ptr);
    }
}

namespace Innovative
{

//---------------------------------------------------------------------------
//  IppComplexDG::operator =() --
//------------------------------------------------------------------------------

IppComplexDG & IppComplexDG::operator = (const IppComplexDG & rhs)
{
    if (this != &rhs) // protect against invalid self-assignment
       {
        Complex *  DataPtr = &(*(begin()));
        const Complex*    RhsDataPtr = &(*(rhs.begin()));
        ippsCopy_32fc(ToIppComplexPtr(RhsDataPtr), ToIppComplexPtr(DataPtr), TruncatedSize());
        }

    // by convention, always return *this
    return *this;
}

//------------------------------------------------------------------------------
//  IppComplexDG::Zero() -- Set all elements to 0
//------------------------------------------------------------------------------

void IppComplexDG::Zero()
{
	Complex * DataPtr = &(*(begin()));

	ippsZero_32fc(ToIppComplexPtr(DataPtr), TruncatedSize());
}

//------------------------------------------------------------------------------
//  IppComplexDG::Set() --  Compose complex array from real part
//------------------------------------------------------------------------------

void IppComplexDG::Set(const IppFloatDG & RealPart)
{
	Buffer b(Holding<float>(RealPart.size()));
	IppFloatDG ImagPart(b);
	ImagPart.Zero();

	const float *    RealPtr = &(*(RealPart.begin()));
	const float *    ImagPtr = &(*(ImagPart.begin()));

	size_t op_size = std::min(RealPart.size(), ImagPart.size());
	Complex *       DataPtr = &(*(begin()));
    Resize(op_size);

    ippsRealToCplx_32f(RealPtr, ImagPtr, ToIppComplexPtr(DataPtr), TruncatedSize());
}

//------------------------------------------------------------------------------
//  IppComplexDG::Set() --  Compose complex array from real and imaginary parts
//------------------------------------------------------------------------------

void IppComplexDG::Set(const IppFloatDG & RealPart, const IppFloatDG & ImagPart)
{
    const float *    RealPtr = &(*(RealPart.begin()));
    const float *    ImagPtr = &(*(ImagPart.begin()));

    size_t op_size = std::min(RealPart.size(), ImagPart.size());
    Complex *       DataPtr = &(*(begin()));
    Resize(op_size);

    ippsRealToCplx_32f(RealPtr, ImagPtr, ToIppComplexPtr(DataPtr), TruncatedSize());
}

//------------------------------------------------------------------------------
//  IppComplexDG::Set() -- Set complex array to single value
//------------------------------------------------------------------------------

void IppComplexDG::Set(Complex value)
{
    Complex * DataPtr = &(*(begin()));

    ippsSet_32fc(ToIppComplex(value), ToIppComplexPtr(DataPtr), TruncatedSize());
}

//------------------------------------------------------------------------------
//  IppComplexDG::Add() -- Add a value to the buffer
//------------------------------------------------------------------------------

void IppComplexDG::Add(Complex value)
{
    if (value.real() == 0 && value.imag() == 0)
        return;

    Complex * DataPtr = &(*(begin()));

    ippsAddC_32fc_I(ToIppComplex(value), ToIppComplexPtr(DataPtr), TruncatedSize());
}

//------------------------------------------------------------------------------
//  IppComplexDG::Add() -- vector Add
//------------------------------------------------------------------------------

void IppComplexDG::Add(const IppComplexDG & rhs)
{
    Complex * DataPtr = &(*(begin()));
    const Complex * RhsDataPtr = &(*(rhs.begin()));

    int op_size = static_cast<int>(std::min(size(), rhs.size()));
    ippsAdd_32fc_I(ToIppComplexPtr(RhsDataPtr), ToIppComplexPtr(DataPtr), op_size);
}

//------------------------------------------------------------------------------
//  IppComplexDG::Multiply() --
//------------------------------------------------------------------------------

void IppComplexDG::Multiply(Complex value)
{
    if (value.real() == 1.0 && value.imag() == 1.0)
        return;

    Complex * DataPtr = &(*(begin()));

    ippsMulC_32fc_I(ToIppComplex(value), ToIppComplexPtr(DataPtr), TruncatedSize());
}

//------------------------------------------------------------------------------
//  IppComplexDG::Multiply() --
//------------------------------------------------------------------------------

void IppComplexDG::Multiply( const IppComplexDG & rhs )
{
    Complex * DataPtr = &(*(begin()));
    const Complex * RhsDataPtr = &(*(rhs.begin()));

    int op_size = static_cast<int>(std::min(size(), rhs.size()));
    ippsMul_32fc_I(ToIppComplexPtr(RhsDataPtr), ToIppComplexPtr(DataPtr), op_size);
}

//------------------------------------------------------------------------------
//  IppComplexDG::Divide() --
//------------------------------------------------------------------------------

void IppComplexDG::Divide(Complex value)
{
    if (value.real() == 1.0 && value.imag() == 1.0)
        return;

    if (value.real() && value.imag())
        {
        Complex * DataPtr = &(*(begin()));
        ippsDivC_32fc_I(ToIppComplex(value), ToIppComplexPtr(DataPtr), TruncatedSize());
        }
}

//------------------------------------------------------------------------------
//  IppComplexDG::Divide() --
//------------------------------------------------------------------------------

void IppComplexDG::Divide(const IppComplexDG & rhs)
{
    Complex * DataPtr = &(*(begin()));
    const Complex * RhsDataPtr = &(*(rhs.begin()));

    int op_size = static_cast<int>(std::min(size(), rhs.size()));
    ippsDiv_32fc_I(ToIppComplexPtr(RhsDataPtr), ToIppComplexPtr(DataPtr), op_size);
}

//------------------------------------------------------------------------------
//  IppComplexDG::Conjugate() --
//------------------------------------------------------------------------------

void IppComplexDG::Conjugate(const IppComplexDG & rhs)
{
    Complex * DataPtr = &(*(begin()));
    const Complex * RhsDataPtr = &(*(rhs.begin()));

    int op_size = static_cast<int>(std::min(size(), rhs.size()));
    ippsConj_32fc(ToIppComplexPtr(RhsDataPtr), ToIppComplexPtr(DataPtr), op_size);
}

//------------------------------------------------------------------------------
//  IppComplexDG::ConjugateFlip() --
//------------------------------------------------------------------------------

void IppComplexDG::ConjugateFlip(const IppComplexDG & rhs)
{
    Complex * DataPtr = &(*(begin()));
    const Complex * RhsDataPtr = &(*(rhs.begin()));

    int op_size = static_cast<int>(std::min(size(), rhs.size()));
    ippsConjFlip_32fc(ToIppComplexPtr(RhsDataPtr), ToIppComplexPtr(DataPtr), op_size);
}

//------------------------------------------------------------------------------
//  IppComplexDG::Flip() --
//------------------------------------------------------------------------------

void IppComplexDG::Flip(const IppComplexDG & rhs)
{
    Complex * DataPtr = &(*(begin()));
    const Complex * RhsDataPtr = &(*(rhs.begin()));

    int op_size = static_cast<int>(std::min(size(), rhs.size()));
    ippsFlip_32fc(ToIppComplexPtr(RhsDataPtr), ToIppComplexPtr(DataPtr), op_size);
}

//------------------------------------------------------------------------------
//  IppComplexDG::Sub() -- subtract constant from vector
//------------------------------------------------------------------------------

void IppComplexDG::Sub(Complex value)
{
	if (value.real() == 0 && value.imag() == 0)
        return;

    Complex * DataPtr = &(*(begin()));
    ippsSubC_32fc_I(ToIppComplex(value), ToIppComplexPtr(DataPtr), TruncatedSize());
}

//------------------------------------------------------------------------------
//  IppComplexDG::Sub() -- vector subtract
//------------------------------------------------------------------------------

void IppComplexDG::Sub(const IppComplexDG & rhs)
{
    Complex * DataPtr = &(*(begin()));
    const Complex * RhsDataPtr = &(*(rhs.begin()));

    int op_size = static_cast<int>(std::min(size(), rhs.size()));
    ippsSub_32fc_I(ToIppComplexPtr(RhsDataPtr), ToIppComplexPtr(DataPtr), op_size);
}

//------------------------------------------------------------------------------
//  IppComplexDG::Window() -- Apply window to vector
//------------------------------------------------------------------------------

bool IppComplexDG::Window(IppComplexDG & rhs, WindowTypes type, float alpha)
{
    Ipp32fc * dst = reinterpret_cast<Ipp32fc*>(&(*(begin())));
    const Ipp32fc * src = reinterpret_cast<const Ipp32fc*>(&(*(rhs.begin())));
    int sz = static_cast<int>(std::min(size(), rhs.size()));

    switch(type)
        {
        default:
        case wtRect:
            return ippsCopy_32fc(src, dst, sz) == ippStsNoErr;
        case wtBartlett:
            return ippsWinBartlett_32fc(src, dst, sz) == ippStsNoErr;
        case wtBlackman:
            return ippsWinBlackman_32fc(src, dst, sz, alpha) == ippStsNoErr;
		case wtBlackmanOpt:
			return ippsWinBlackmanOpt_32fc(src, dst, sz) == ippStsNoErr;
		case wtHamming:
            return ippsWinHamming_32fc(src, dst, sz) == ippStsNoErr;
		case wtHanning:
            return ippsWinHann_32fc(src, dst, sz) == ippStsNoErr;
        case wtKaiser:
            return ippsWinKaiser_32fc(src, dst, sz, alpha) == ippStsNoErr;
        }
}


} // end namespace Innovative
