//---------------------------------------------------------------------------
// JESDSubDevice_Stream_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2016 Innovative Integration Corporation.
//  All Rights Reserved.

#include <JESDSubDevice_Stream_Mb.h>
#include <SystemSupport_Mb.h>
#include <StringSupport_Mb.h>
#include <FloatingPtSupport_Mb.h>

using namespace std;

namespace Innovative
{

#ifdef _MSC_VER
    #pragma warning(push)
    #pragma warning(disable : 4355)
#endif


//===========================================================================
//  CLASS JESD_PHY -- JESD PHY IP interface
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class JESD_PHY
//---------------------------------------------------------------------------

JESD_PHY::JESD_PHY(IRequires_JESDPHY_Io * mdc, unsigned int device_id)
    :  MDC(mdc), DeviceID(device_id), AdcJesd(0), DacJesd(0), PllRates(0)
{
    // Initialize
    NumGtCommon = 1;
}

//---------------------------------------------------------------------------
//  destructor for class JESD_PHY
//---------------------------------------------------------------------------

JESD_PHY::~JESD_PHY()
{
    delete AdcJesd;   AdcJesd = 0;
    delete DacJesd;   DacJesd = 0;
    delete PllRates;  PllRates = 0;
}

//---------------------------------------------------------------------------
//  JESD_PHY::ConstructCustomParts()
//---------------------------------------------------------------------------

void  JESD_PHY::ConstructCustomParts()
{
    if (!DeviceEnabled())
        return;

    AdcJesd = MDC->GetAdcJesd_IoDevice(DeviceID);
    DacJesd = MDC->GetDacJesd_IoDevice(DeviceID);
    PllRates = MDC->GetJesd_PllRatesDevice(DeviceID);
}

//---------------------------------------------------------------------------
//  JESD_PHY::OpenHardware()
//---------------------------------------------------------------------------

void  JESD_PHY::OpenHardware()
{
    if (!DeviceEnabled())
        return;

    LoadFpgaInfo();
    LoadFpgaMgtPllRanges();
    LoadIpDefaultSettings();
}

//---------------------------------------------------------------------------
//  JESD_PHY::PreconfigureHardware()
//---------------------------------------------------------------------------

void  JESD_PHY::PreconfigureHardware()
{
    if (!DeviceEnabled())
        return;

    InitFpgaMgts();
}

//---------------------------------------------------------------------------
//  JESD_PHY::ConfigureHardware()
//---------------------------------------------------------------------------

void  JESD_PHY::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;
}

//---------------------------------------------------------------------------
//  JESD_PHY::InitFpgaMgts() -- Initialize and configure the FPGA MGTs
//---------------------------------------------------------------------------

void  JESD_PHY::InitFpgaMgts()
{
    LoadFpgaInfo();
    LoadFpgaMgtPllRanges();
    LoadIpDefaultSettings();

    LoadCurrentMgtPllSettings();

    Debug.Log("Initialize the FPGA MGTs and set the actual Adc and Dac parameters");

    double ActualAdcSampleRate = PllRates->AdcFrequencyActual();
    bool AdcActive = AdcJesd->IsDeviceActive();
    double AdcJesdLineRate(0);
    if (AdcActive)
        {
        Debug.Log("ActualAdcSampleRate (MHz) =", (ActualAdcSampleRate/1.0e6));
        AdcJesdLineRate = AdcJesd->CalcJesdLineRate(ActualAdcSampleRate);
        Debug.Log("Adc line rate (Gbps) =", (AdcJesdLineRate/1.0e9));
        RxPowered_All(true);
        }
    else
        RxPowered_All(false);

    double ActualDacSampleRate = PllRates->DacFrequencyActual();
    bool DacActive = DacJesd->IsDeviceActive();
    double DacJesdLineRate(0);
    if (DacActive)
        {
        Debug.Log("ActualDacSampleRate (MHz) =", (ActualDacSampleRate/1.0e6));
        DacJesdLineRate = DacJesd->CalcJesdLineRate(ActualDacSampleRate);
        Debug.Log("Dac line rate (Gbps) =", (DacJesdLineRate/1.0e9));
        TxPowered_All(true);
        }
    else
        TxPowered_All(false);

    // If One line rate is invalid, use same for both. If none are valid, skip init entirely
    if (AdcActive && !DacActive)
        DacJesdLineRate = AdcJesdLineRate;
    else if (!AdcActive && DacActive)
        AdcJesdLineRate = DacJesdLineRate;
    else if (!AdcActive && !DacActive)
        return;

    // use adc line rate if possible, otherwise dac
    double LineRate = (AdcActive && !IsAboutEqual(AdcJesdLineRate, 0.0, 10.0)) ? AdcJesdLineRate : DacJesdLineRate ;

    // Skip if LineRate is 0
    if (IsAboutEqual(LineRate, 0.0, 10.0))
        {
        Debug.Log("Skipping FPGA MGT initialization since the line rate (Gbps) =", (LineRate/1.0e9));
        return;
        }

    double ExpMgtRefClk = CalcMgtRefClk(LineRate);
    Debug.Log("Expected MGT reference clock frequency (MHz) =", (ExpMgtRefClk/1.0e6));
    double MgtRefClk = PllRates->MgtRefFrequencyActual();
    Debug.Log("Actual MGT reference clock frequency (MHz) =", (MgtRefClk/1.0e6));

    if (IsFpgaMgtPllUpdateRequired(AdcJesdLineRate, DacJesdLineRate, MgtRefClk))
        {
        CalcFpgaMgtPllSettings(AdcJesdLineRate, DacJesdLineRate, MgtRefClk);
        ConfigureFpgaMgts();
        }
    CheckFpgaMgtsPllLock();
}

//---------------------------------------------------------------------------
//  JESD_PHY::LoadFpgaInfo() -- Load the FPGA type, speed grade, package, and transceiver type
//---------------------------------------------------------------------------

void  JESD_PHY::LoadFpgaInfo()
{
    Debug.Log("------------------------------------------------------------");
    Debug.Log("Configure JESD204B PHY", DeviceID);
    Debug.Log("------------------------------------------------------------");
    IpVersion = PeekJesdPhy(0x0);
    Debug.Log("JESD204B PHY IP Version Reg(0x0) =", IpVersion, 16);

    MemoryBitManipulator Data(PeekJesdPhy(0x4));

    FpgaType = Data.Field(24,8);
    unsigned int FpgaSpeedGradeCode = Data.Field(16,8);
    unsigned int FpgaPackageCode = Data.Field(8,8);
    FpgaMgtType = Data.Field(0,8);

    std::string str;
    if (FpgaType == ft7Series)
        str = "7 Series";
    else if (FpgaType == ftUltrascale)
        str = "UltraScale";
    else if (FpgaType == ftUltrascalePlus)
        str = "UltraScale+";
    else
        str = std::string("Unknown = ") + IntToString(FpgaType, 16);
    Debug.Log("FPGA Type = ", str);

    DecodeFpgaSpeedGrade(FpgaSpeedGradeCode);
    DecodeFpgaPackage(FpgaPackageCode);

    if (FpgaMgtType == fmtGTXE2)
        str = "GTXE2";
    else if (FpgaMgtType == fmtGTHE2)
        str = "GTHE2";
    else if (FpgaMgtType == fmtGTHE3)
        str = "GTHE3";
    else if (FpgaMgtType == fmtGTYE3)
        str = "GTYE3";
    else if (FpgaMgtType == fmtGTHE4)
        str = "GTHE4";
    else
        str = std::string("Unknown = ") + IntToString(FpgaMgtType, 16);
    Debug.Log("FPGA Transceiver Type = ", str);
}

//---------------------------------------------------------------------------
//  JESD_PHY::DecodeFpgaSpeedGrade()
//---------------------------------------------------------------------------

void  JESD_PHY::DecodeFpgaSpeedGrade(unsigned int code)
{
    if (code == 1)
        FpgaSpeedGrade = fsg1;
    else if (code == 2)
        FpgaSpeedGrade = fsg2;
    else if (code == 3)
        FpgaSpeedGrade = fsg3;
    else
        FpgaSpeedGrade = code;

    std::string str;
    if (FpgaSpeedGrade == fsg1)
        str = "1";
    else if (FpgaSpeedGrade == fsg1L)
        str = "1L";
    else if (FpgaSpeedGrade == fsg1H)
        str = "1H";
    else if (FpgaSpeedGrade == fsg1HV)
        str = "1HV";
    else if (FpgaSpeedGrade == fsg1LV)
        str = "1LV";
    else if (FpgaSpeedGrade == fsg2)
        str = "2";
    else if (FpgaSpeedGrade == fsg2L)
        str = "2L";
    else if (FpgaSpeedGrade == fsg2LV)
        str = "2LV";
    else if (FpgaSpeedGrade == fsg3)
        str = "3";
    else
        str = std::string("Unknown = ") + IntToString(FpgaSpeedGrade, 16);
    Debug.Log("FPGA Speed Grade = ", str);
}

//---------------------------------------------------------------------------
//  JESD_PHY::DecodeFpgaPackage()
//---------------------------------------------------------------------------

void  JESD_PHY::DecodeFpgaPackage(unsigned int code)
{
    if (IpVersion < 0x03010100)
        {
        if (code == 1)
            FpgaPackage = fpFF;
        else if (code == 2)
            FpgaPackage = fpFB;
        else if (code == 3)
            FpgaPackage = fpFF;
        else if (code == 4)
            FpgaPackage = fpFB;
        else
            FpgaPackage = code;
        }
    else
        FpgaPackage = code;

    std::string str;
    if (FpgaPackage == fpRF)
        str = "RF";
    else if (FpgaPackage == fpFL)
        str = "FL";
    else if (FpgaPackage == fpFF)
        str = "FF";
    else if (FpgaPackage == fpFB)
        str = "FB";
    else if (FpgaPackage == fpHC)
        str = "HC";
    else if (FpgaPackage == fpFH)
        str = "FH";
    else if (FpgaPackage == fpCS)
        str = "CS";
    else if (FpgaPackage == fpCP)
        str = "CP";
    else if (FpgaPackage == fpFT)
        str = "FT";
    else if (FpgaPackage == fpFG)
        str = "FG";
    else if (FpgaPackage == fpSB)
        str = "SB";
    else if (FpgaPackage == fpRB)
        str = "RB";
    else if (FpgaPackage == fpRS)
        str = "RS";
    else if (FpgaPackage == fpCL)
        str = "CL";
    else if (FpgaPackage == fpSF)
        str = "SF";
    else if (FpgaPackage == fpBA)
        str = "BA";
    else if (FpgaPackage == fpFA)
        str = "FA";
    else
        str = std::string("Unknown = ") + IntToString(FpgaPackage, 16);
    Debug.Log("FPGA Package = ", str);
}

//---------------------------------------------------------------------------
//  JESD_PHY::LoadFpgaMgtPllRanges() -- Load the FPGA MGT PLL ranges
//---------------------------------------------------------------------------
// Loaded from JESDSubDevice_Stream_Tables_Mb.cpp file

//---------------------------------------------------------------------------
//  JESD_PHY::LoadIpDefaultSettings() -- Load the PHY IP default settings
//---------------------------------------------------------------------------

void  JESD_PHY::LoadIpDefaultSettings()
{
    // Skip if JESD204B PHY IP version is prior to v3.2
    if (IpVersion < 0x03020000)
        return;

    Debug.Log("JESD204B PHY IP RX line rate (Gbps) =", (PeekJesdPhy(0x90)/1.0e6));
    Debug.Log("JESD204B PHY IP RX reference clock frequency (MHz) =", (PeekJesdPhy(0x98)/1.0e3));
    Debug.Log("JESD204B PHY IP RX xMult ratio =", (PeekJesdPhy(0x9C)/1.0e3));

    unsigned int RxPll = PeekJesdPhy(0xA0);

    std::string str;
    if (RxPll == 0)
        str = "CPLL";
    else if (RxPll == 2)
        str = "QPLL1";
    else if (RxPll == 3 && FpgaType == ft7Series)
        str = "QPLL";
    else if (RxPll == 3)
        str = "QPLL0";
    else
        str = "Unknown";
    Debug.Log("JESD204B PHY IP RX PLL type = ", str);

    Debug.Log("JESD204B PHY IP TX line rate (Gbps) =", (PeekJesdPhy(0xB0)/1.0e6));
    Debug.Log("JESD204B PHY IP TX reference clock frequency (MHz) =", (PeekJesdPhy(0xB8)/1.0e3));
    Debug.Log("JESD204B PHY IP TX xMult ratio =", (PeekJesdPhy(0xBC)/1.0e3));

    unsigned int TxPll = PeekJesdPhy(0xC0);

    if (TxPll == 0)
        str = "CPLL";
    else if (TxPll == 2)
        str = "QPLL1";
    else if (TxPll == 3 && FpgaType == ft7Series)
        str = "QPLL";
    else if (TxPll == 3)
        str = "QPLL0";
    else
        str = "Unknown";
    Debug.Log("JESD204B PHY IP TX PLL type = ", str);

    unsigned int SwCapable = PeekJesdPhy(0xD0);
    if (SwCapable == 0)
        Debug.Log("JESD204B PHY IP is generated for static line rate");
    else if (SwCapable == 1)
        Debug.Log("JESD204B PHY IP is generated for dynamic line rate");
    else
        Debug.Log("JESD204B PHY IP is generated for unknown line rate switching type");

    Debug.Log("JESD204B PHY IP RX insertion loss (dB) =", (PeekJesdPhy(0xD4)/1.0e3));

    unsigned int Equalization = PeekJesdPhy(0xD8);
    if (Equalization == 0)
        Debug.Log("JESD204B PHY IP equalization mode is Auto (set based on insertion loss)");
    else if (Equalization == 1)
        Debug.Log("JESD204B PHY IP equalization mode is Low Loss (LPM)");
    else if (Equalization == 2)
        Debug.Log("JESD204B PHY IP equalization mode is High Loss (DFE)");
    else
        Debug.Log("JESD204B PHY IP equalization mode is unknown");

    Debug.Log("JESD204B PHY IP minimum line rate (Gbps) =", (PeekJesdPhy(0xE0)/1.0e3));
    Debug.Log("JESD204B PHY IP maximum line rate (Gbps) =", (PeekJesdPhy(0xE4)/1.0e3));
}

//---------------------------------------------------------------------------
//  JESD_PHY::LoadCurrentMgtPllSettings() -- Load the current MGT PLL settings
//---------------------------------------------------------------------------

void  JESD_PHY::LoadCurrentMgtPllSettings()
{
    Debug.Log("Load the current FPGA MGT PLL settings");

    // Read the IP registers to determine the number of GT_COMMON and GT_CHANNEL blocks
    SetNumInterfaces();

    if (NumGtCommon > 32 || NumGtChannel > 32 || NumGtChannel == 0) // Skip if illegitimate count
        return;

    GetTxRxPllClockSelect();
    GetCPLLParams();
    GetQPLLParams();
    GetOutputDividers();

    // Print the current MGT settings
    std::string str;
    str = (CurrentRxPllType == mptCPLL)? "CPLL" :
          (CurrentRxPllType == mptQPLL0)? "QPLL0" : "QPLL1";
    Debug.Log("Current MGT RX PLL type = ", str);
    str = (CurrentTxPllType == mptCPLL)? "CPLL" :
          (CurrentTxPllType == mptQPLL0)? "QPLL0" : "QPLL1";
    Debug.Log("Current MGT Tx PLL type = ", str);

    std::stringstream msg;
    msg << "Current MGT CPLL M=" << CurrentCPLLParams.M << " N1=" <<
           CurrentCPLLParams.N1 << " N2=" << CurrentCPLLParams.N2;
    Debug.Log(msg.str());
    msg.str("");

    for (unsigned int idx = 0; idx < CurrentQPLLParamsVec.size(); idx++)
        {
        str = (CurrentQPLLParamsVec[idx].QPLLIndex == 0)? "QPLL0" : "QPLL1";
        msg << "Current MGT " << str << " M=" << CurrentQPLLParamsVec[idx].M << " N=" << CurrentQPLLParamsVec[idx].N;
        Debug.Log(msg.str());
        msg.str("");
        }

    Debug.Log("Current MGT RX output divider = ", CurrentRxOutputDivider);
    Debug.Log("Current MGT TX output divider = ", CurrentTxOutputDivider);
}

//---------------------------------------------------------------------------
//  JESD_PHY::IsFpgaMgtPllUpdateRequired() -- check if the current PLL settings can support the required line rates
//---------------------------------------------------------------------------

bool  JESD_PHY::IsFpgaMgtPllUpdateRequired(double RxLineRate, double TxLineRate, double MgtRefClk)
{
    if (WithinRange(MgtRefClk, MgtRefclkRange.Low, MgtRefclkRange.High) == false)
        Debug.Log("MGT reference clock is out of range");

    double LineRate;
    unsigned int PllType;
    unsigned int OutputDivider;
    double LineRateLow;
    double LineRateHigh;
    double VcoLow;
    double VcoHigh;
    unsigned int MaxOutputDivider;
    double ClkoutFreq;
    double VcoFreq;
    double ClkoutFreqRequired;
    bool   PllUpdateIsRequired = false;

    for (unsigned int LinkDirection = 0; LinkDirection <= 1; LinkDirection++)
        {
        if (LinkDirection == ldtRx)
            {
            LineRate = RxLineRate;
            PllType = CurrentRxPllType;
            OutputDivider = CurrentRxOutputDivider;
            }
        else
            {
            LineRate = TxLineRate;
            PllType = CurrentTxPllType;
            OutputDivider = CurrentTxOutputDivider;
            }

        if (PllType == mptCPLL)
            {
            LineRateLow = CPLLRange.LineRateLow;
            LineRateHigh = CPLLRange.LineRateHigh;
            VcoLow = CPLLRange.VcoLow;
            VcoHigh = CPLLRange.VcoHigh;
            MaxOutputDivider = CPLLRange.MaxOutputDivider;
            ClkoutFreq = CalcCPLLClkout(MgtRefClk, CurrentCPLLParams);
            VcoFreq = ClkoutFreq;
            }
        else
            {
            unsigned int CurrentQPLLIndex;
            if (PllType == mptQPLL0)
                CurrentQPLLIndex = 0;
            else
                CurrentQPLLIndex = 1;
            LineRateLow = QPLLRange[CurrentQPLLIndex].LineRateLow;
            LineRateHigh = QPLLRange[CurrentQPLLIndex].LineRateHigh;
            VcoLow = QPLLRange[CurrentQPLLIndex].VcoLow;
            VcoHigh = QPLLRange[CurrentQPLLIndex].VcoHigh;
            MaxOutputDivider = QPLLRange[CurrentQPLLIndex].MaxOutputDivider;
            QPLLParamsType CurrentQPLLParams;
            if (FpgaType == ft7Series)
                CurrentQPLLParams = CurrentQPLLParamsVec[0];
            else
                CurrentQPLLParams = CurrentQPLLParamsVec[CurrentQPLLIndex];
            ClkoutFreq = CalcQPLLClkout(MgtRefClk, CurrentQPLLParams);
            VcoFreq = ClkoutFreq * 2;
            }

        if (WithinRange(LineRate, LineRateLow, LineRateHigh) == false)
            {
            PllUpdateIsRequired = true;
            Debug.Log("MGT PLL update is required, LineRate is out of range");
            break;
            }

        if (OutputDivider > MaxOutputDivider)
            {
            PllUpdateIsRequired = true;
            Debug.Log("MGT PLL update is required, OutputDivider is out of range");
            break;
            }

        if (WithinRange(VcoFreq, VcoLow, VcoHigh) == false)
            {
            PllUpdateIsRequired = true;
            Debug.Log("MGT PLL update is required, VcoFreq is out of range");
            break;
            }

        ClkoutFreqRequired = (LineRate * OutputDivider)/2.0;

        if (IsAboutEqual(ClkoutFreq, ClkoutFreqRequired, 1.0) == false)
            {
            PllUpdateIsRequired = true;
            Debug.Log("MGT PLL update is required, ClkoutFreq mismatch");
            break;
            }
        }

    if (PllUpdateIsRequired)
        Debug.Log("MGT PLL update is required");
    else
        Debug.Log("MGT PLL update is not required");

    return PllUpdateIsRequired;
}

//---------------------------------------------------------------------------
//  JESD_PHY::CalcFpgaMgtPllSettings() -- Calculate the MGT PLL settings
//---------------------------------------------------------------------------

void  JESD_PHY::CalcFpgaMgtPllSettings(double RxLineRate, double TxLineRate, double MgtRefClk)
{
    Debug.Log("Calculate FPGA MGT PLL settings");

    double LineRate;
    double LineRateLow;
    double LineRateHigh;
    unsigned int MaxOutputDivider;
    double ClkoutFreq;

    RxCPLLParamsVec.clear();
    TxCPLLParamsVec.clear();
    RxQPLLParamsVec.clear();
    TxQPLLParamsVec.clear();

    unsigned int NumMgtPlls = (IsGtCommonEnabled()? 3 : 1);

    for (unsigned int LinkDirection = 0; LinkDirection <= 1; LinkDirection++)
        {
        if (LinkDirection == ldtRx)
            LineRate = RxLineRate;
        else
            LineRate = TxLineRate;

        for (unsigned int PllIdx = 0; PllIdx < NumMgtPlls; PllIdx++)
            {
            if (PllIdx == 0)
                {
                LineRateLow = CPLLRange.LineRateLow;
                LineRateHigh = CPLLRange.LineRateHigh;
                MaxOutputDivider = CPLLRange.MaxOutputDivider;
                }
            else
                {
                LineRateLow = QPLLRange[PllIdx-1].LineRateLow;
                LineRateHigh = QPLLRange[PllIdx-1].LineRateHigh;
                MaxOutputDivider = QPLLRange[PllIdx-1].MaxOutputDivider;
                }

            if (WithinRange(LineRate, LineRateLow, LineRateHigh) == false)
                continue;

            for (unsigned int OutputDivider = MaxOutputDivider; OutputDivider >= 1; OutputDivider/=2)
                {
                ClkoutFreq = (LineRate * OutputDivider)/2.0;
                if (PllIdx == 0)
                    CalcCPLLSettings(ClkoutFreq, MgtRefClk, LinkDirection);
                else
                    CalcQPLLSettings(ClkoutFreq, MgtRefClk, LinkDirection, PllIdx-1);
                }
            }
//        // If the rates for both directions are the same, then skip the calculation of the Tx
//        if (IsAboutEqual(RxLineRate, TxLineRate, 1.0))
//            break;
        }

    // Select the MGT PLL settings from the candidates tables
    SelectFpgaMgtPllSettings();

    double CPLLClkout = 0;
    if (NewCPLLParams.IsValid())
        CPLLClkout = CalcCPLLClkout(MgtRefClk, NewCPLLParams);

    vector<double> QPLLClkout(2);
    // Initialize the QPLLClkout vector
    for (unsigned int idx = 0; idx < QPLLClkout.size(); idx++)
        QPLLClkout[idx] = 0.0;
    for (unsigned int idx = 0; idx < NewQPLLParamsVec.size(); idx++)
        QPLLClkout[NewQPLLParamsVec[idx].QPLLIndex] = CalcQPLLClkout(MgtRefClk, NewQPLLParamsVec[idx]);

    double RxPLLClkout, TxPLLClkout;
    if (NewRxPllType == mptCPLL)
        RxPLLClkout = CPLLClkout;
    else if (NewRxPllType == mptQPLL0)
        RxPLLClkout = QPLLClkout[0];
    else
        RxPLLClkout = QPLLClkout[1];

    if (NewTxPllType == mptCPLL)
        TxPLLClkout = CPLLClkout;
    else if (NewTxPllType == mptQPLL0)
        TxPLLClkout = QPLLClkout[0];
    else
        TxPLLClkout = QPLLClkout[1];

    RxOutputDivider = static_cast<unsigned int>((RxPLLClkout * 2) / RxLineRate);
    TxOutputDivider = static_cast<unsigned int>((TxPLLClkout * 2) / TxLineRate);

    Debug.Log("Calculated MGT PLL parameters are ...");

    std::string str;
    std::stringstream msg;

    str = (NewRxPllType == mptCPLL)? "CPLL" :
          (NewRxPllType == mptQPLL0)? "QPLL0" : "QPLL1";
    msg << "MGT RX PLL Type = " << str;
    Debug.Log(msg.str());
    msg.str("");

    str = (NewTxPllType == mptCPLL)? "CPLL" :
          (NewTxPllType == mptQPLL0)? "QPLL0" : "QPLL1";
    msg << "MGT TX PLL Type = " << str;
    Debug.Log(msg.str());
    msg.str("");

    if (NewRxPllType == mptCPLL || NewTxPllType == mptCPLL)
        {
        msg << "MGT CPLL M=" << NewCPLLParams.M << " N1=" << NewCPLLParams.N1 << " N2=" << NewCPLLParams.N2;
        Debug.Log(msg.str());
        msg.str("");
        Debug.Log("MGT CPLL clock output frequency (MHz) =", (CPLLClkout/1.0e6));
        }

    // non empty NewQPLLParamsVec means the QPLL is used
    for (unsigned int idx = 0; idx < NewQPLLParamsVec.size(); idx++)
        {
        unsigned int QPLLIndex = NewQPLLParamsVec[idx].QPLLIndex;
        str = (QPLLIndex == 0)? "MGT QPLL0" : "MGT QPLL1";
        msg << str << " clock output frequency (MHz) =" << (QPLLClkout[QPLLIndex]/1.0e6) <<
        " M=" << NewQPLLParamsVec[idx].M << " N=" << NewQPLLParamsVec[idx].N;
        Debug.Log(msg.str());
        msg.str("");
        }

    Debug.Log("MGT RX output divider =", RxOutputDivider);
    Debug.Log("MGT TX output divider =", TxOutputDivider);
}

//---------------------------------------------------------------------------
//  JESD_PHY::CalcCPLLSettings() -- Calculate the MGT CPLL settings
//---------------------------------------------------------------------------

void  JESD_PHY::CalcCPLLSettings(double ClkoutFreq, double RefClkFreq, unsigned int LinkDirection)
{
    // ClkoutFreq = RefClkFreq * N1*N2/M = VcoFreq
    // M  = 1,2
    // N2 = 1,2,3,4,5
    // N1 = 4,5

    double VcoFreq = ClkoutFreq;
    if (WithinRange(VcoFreq, CPLLRange.VcoLow, CPLLRange.VcoHigh) == false)
        return;

    Debug.Log("MGT CPLL clock output frequency (MHz) =", (ClkoutFreq/1.0e6));

    std::string direction;
    if (LinkDirection == ldtRx)
        direction = "RX";
    else
        direction = "TX";

    double CandidateFreq;
    CPLLParamsType CPLL;

    for (unsigned int M = 1; M <= 2; M++)
        {
        for (unsigned int N1 = 4; N1 <= 5; N1++)
            {
            for (unsigned int N2 = 1; N2<= 5; N2++)
                {
                CPLL.M = M;
                CPLL.N1 = N1;
                CPLL.N2 = N2;
                CandidateFreq = CalcCPLLClkout(RefClkFreq, CPLL);
                if (IsAboutEqual(CandidateFreq,ClkoutFreq, 1.0))
                    {
                    if (LinkDirection == ldtRx)
                        RxCPLLParamsVec.push_back(CPLL);
                    if (LinkDirection == ldtTx)
                        TxCPLLParamsVec.push_back(CPLL);
                    std::stringstream msg;
                    msg << direction << " MGT CPLL M=" << M << " N1=" << N1 << " N2=" << N2;
                    Debug.Log(msg.str());
                    }
                }
            }
        }
}

//---------------------------------------------------------------------------
//  JESD_PHY::CalcQPLLSettings() -- Calculate the MGT QPLL settings
//---------------------------------------------------------------------------

void  JESD_PHY::CalcQPLLSettings(double ClkoutFreq, double RefClkFreq, unsigned int LinkDirection, unsigned int which)
{
    // ClkoutFreq = RefClkFreq * N/(M*2) = VcoFreq/2
    // M = 1,2,3,4
    // N = 16,20,32,40,64,66,80,100                                 (UltraScale and 7Series)
    // N = 16,20,32,40,60,64,66,75,80,84,90,96,100,112,120,125,150  (UltraScale+)

    double VcoFreq = ClkoutFreq * 2;
    if (WithinRange(VcoFreq, QPLLRange[which].VcoLow, QPLLRange[which].VcoHigh) == false)
        return;

    std::string QPLLType;
    if (FpgaType == ft7Series)
        {
        if (which == 0)
            QPLLType = "MGT Upper band QPLL";
        else
            QPLLType = "MGT Lower band QPLL";
        }
    else
        {
        if (which == 0)
            QPLLType = "MGT QPLL0";
        else
            QPLLType = "MGT QPLL1";
        }

    std::stringstream msg;
    msg << QPLLType << " clock output frequency (MHz) =" <<  (ClkoutFreq/1.0e6);
    Debug.Log(msg.str());

    std::string direction;
    if (LinkDirection == ldtRx)
        direction = "RX";
    else
        direction = "TX";

    double CandidateFreq;
    QPLLParamsType QPLL;

    vector<unsigned int> N;
    if (FpgaType == ft7Series || FpgaType == ftUltrascale)
        {
        N.push_back(16);     N.push_back(20);     N.push_back(32);     N.push_back(40);
        N.push_back(64);     N.push_back(66);     N.push_back(80);     N.push_back(100);
        }
    else        // (FpgaType == ftUltrascalePlus && FpgaMgtTypes == fmtGTHE4)
        {
        N.push_back(16);     N.push_back(20);     N.push_back(32);     N.push_back(40);
        N.push_back(60);     N.push_back(64);     N.push_back(66);     N.push_back(75);
        N.push_back(80);     N.push_back(84);     N.push_back(90);     N.push_back(96);
        N.push_back(100);    N.push_back(112);    N.push_back(120);    N.push_back(125);
        N.push_back(150);
        }

    for (unsigned int M = 1; M <= 4; M++)
        {
        for (size_t NIdx = 0; NIdx < N.size(); NIdx++)
            {
            QPLL.M = M;
            QPLL.N = N[NIdx];
            QPLL.QPLLIndex = which;
            CandidateFreq = CalcQPLLClkout(RefClkFreq, QPLL);
            if (IsAboutEqual(CandidateFreq,ClkoutFreq, 1.0))
                {
                if (LinkDirection == ldtRx)
                    RxQPLLParamsVec.push_back(QPLL);
                if (LinkDirection == ldtTx)
                    TxQPLLParamsVec.push_back(QPLL);
                std::stringstream msg;
                msg << direction << " " << QPLLType << " M=" << M << " N=" << N[NIdx];
                Debug.Log(msg.str());
                }
            }
        }
}

//---------------------------------------------------------------------------
//  JESD_PHY::CalcCPLLClkout()
//---------------------------------------------------------------------------

double  JESD_PHY::CalcCPLLClkout(double RefClkFreq, CPLLParamsType CPLL)
{
    // ClkoutFreq = RefClkFreq * N1*N2/M = VcoFreq
    if (!CPLL.IsValid())
        return 0;
    else
        return (RefClkFreq * CPLL.N1 * CPLL.N2)/CPLL.M;
}

//---------------------------------------------------------------------------
//  JESD_PHY::CalcQPLLClkout()
//---------------------------------------------------------------------------

double  JESD_PHY::CalcQPLLClkout(double RefClkFreq, QPLLParamsType QPLL)
{
    // ClkoutFreq = RefClkFreq * N/(M*2) = VcoFreq/2
    if (!QPLL.IsValid())
        return 0;
    else
        return (RefClkFreq * QPLL.N)/(QPLL.M * 2);
}

//---------------------------------------------------------------------------
//  JESD_PHY::CPLLParamsType::operator ==
//---------------------------------------------------------------------------

bool JESD_PHY::CPLLParamsType::operator == (const CPLLParamsType & other)
{
    if (N1 == other.N1 && N2 == other.N2 && M  == other.M )
        return true;
    else
        return false;
}

//---------------------------------------------------------------------------
//  JESD_PHY::CheckCommonCPLLParams()
//---------------------------------------------------------------------------

bool  JESD_PHY::CheckCommonCPLLParams()
{
    bool MatchFound = false;

    for (unsigned int RxIdx=0; RxIdx<RxCPLLParamsVec.size(); RxIdx++)
        {
        for (unsigned int TxIdx=0; TxIdx<TxCPLLParamsVec.size(); TxIdx++)
            {
            if (RxCPLLParamsVec[RxIdx] == TxCPLLParamsVec[TxIdx])
                {
                MatchFound = true;
                CommonCPLLParams = RxCPLLParamsVec[RxIdx];
                break;
                }
            }
        if (MatchFound)
            break;
        }
    return MatchFound;
}

//---------------------------------------------------------------------------
//  JESD_PHY::QPLLParamsType::operator ==
//---------------------------------------------------------------------------

bool JESD_PHY::QPLLParamsType::operator == (const QPLLParamsType & other)
{
    if (N == other.N && M  == other.M && QPLLIndex == other.QPLLIndex)
        return true;
    else
        return false;
}

//---------------------------------------------------------------------------
//  JESD_PHY::CheckCommonQPLLParams()
//---------------------------------------------------------------------------

bool  JESD_PHY::CheckCommonQPLLParams(unsigned int which)
{
    bool MatchFound = false;

    for (unsigned int RxIdx=0; RxIdx<RxQPLLParamsVec.size(); RxIdx++)
        {
        for (unsigned int TxIdx=0; TxIdx<TxQPLLParamsVec.size(); TxIdx++)
            {
            if (RxQPLLParamsVec[RxIdx] == TxQPLLParamsVec[TxIdx] &&
                RxQPLLParamsVec[RxIdx].QPLLIndex == which)
                {
                MatchFound = true;
                CommonQPLLParams = RxQPLLParamsVec[RxIdx];
                break;
                }
            }
        if (MatchFound)
            break;
        }
    return MatchFound;
}

//---------------------------------------------------------------------------
//  JESD_PHY::SplitQPLLParamsVec()
//---------------------------------------------------------------------------

void  JESD_PHY::SplitQPLLParamsVec(const QPLLParamsVec &FullQPLLParamsVec, unsigned int which, QPLLParamsVec & PartialQPLLParamsVec)
{
    PartialQPLLParamsVec.clear();
    for (unsigned int Idx=0; Idx<FullQPLLParamsVec.size(); Idx++)
        {
        if (FullQPLLParamsVec[Idx].QPLLIndex == which)
            PartialQPLLParamsVec.push_back(FullQPLLParamsVec[Idx]);
        }
}

//---------------------------------------------------------------------------
//  JESD_PHY::SelectFpgaMgtPllSettings()
//---------------------------------------------------------------------------

void  JESD_PHY::SelectFpgaMgtPllSettings()
{
    QPLLParamsVec RxQPLL0ParamsVec;
    QPLLParamsVec RxQPLL1ParamsVec;
    QPLLParamsVec TxQPLL0ParamsVec;
    QPLLParamsVec TxQPLL1ParamsVec;

    SplitQPLLParamsVec(RxQPLLParamsVec, 0, RxQPLL0ParamsVec);
    SplitQPLLParamsVec(RxQPLLParamsVec, 1, RxQPLL1ParamsVec);
    SplitQPLLParamsVec(TxQPLLParamsVec, 0, TxQPLL0ParamsVec);
    SplitQPLLParamsVec(TxQPLLParamsVec, 1, TxQPLL1ParamsVec);

    size_t RxCPLLSize = RxCPLLParamsVec.size();
    size_t RxQPLLSize = RxQPLLParamsVec.size();
    size_t RxQPLL0Size = RxQPLL0ParamsVec.size();
    size_t RxQPLL1Size = RxQPLL1ParamsVec.size();
    size_t TxCPLLSize = TxCPLLParamsVec.size();
    size_t TxQPLLSize = TxQPLLParamsVec.size();
    size_t TxQPLL0Size = TxQPLL0ParamsVec.size();
    size_t TxQPLL1Size = TxQPLL1ParamsVec.size();

    Debug.Log("RX CPLL solutions =", RxCPLLSize);
    Debug.Log("RX QPLL0 solutions =", RxQPLL0Size);
    Debug.Log("RX QPLL1 solutions =", RxQPLL1Size);
    Debug.Log("TX CPLL solutions =", TxCPLLSize);
    Debug.Log("TX QPLL0 solutions =", TxQPLL0Size);
    Debug.Log("TX QPLL1 solutions =", TxQPLL1Size);

    // Initialize
    NewCPLLParams.Init();
    NewQPLLParamsVec.clear();

    // Temporary variables to hold the selector outputs
    CPLLParamsType RxCPLL;
    CPLLParamsType TxCPLL;
    QPLLParamsType RxQPLL;
    QPLLParamsType TxQPLL;

    // Prioritize selecting the same type MGT PLL as used in the previous run
    if (CurrentTxPllType == mptCPLL && CurrentRxPllType == mptCPLL && CheckCommonCPLLParams())
        {
        Debug.Log("MGT Rx/Tx Sharing a common CPLL setting");
        RxCPLL = CommonCPLLParams;
        TxCPLL = CommonCPLLParams;
        }
    else if (CurrentTxPllType == mptQPLL0 && CurrentRxPllType == mptQPLL0 && CheckCommonQPLLParams(0))
        {
        Debug.Log("MGT Rx/Tx Sharing a common QPLL0 setting");
        RxQPLL = CommonQPLLParams;
        TxQPLL = CommonQPLLParams;
        }
    else if (CurrentTxPllType == mptQPLL1 && CurrentRxPllType == mptQPLL1 && CheckCommonQPLLParams(1))
        {
        Debug.Log("MGT Rx/Tx Sharing a common QPLL1 setting");
        RxQPLL = CommonQPLLParams;
        TxQPLL = CommonQPLLParams;
        }
    else if (CurrentTxPllType == mptQPLL0 && CurrentRxPllType == mptQPLL1 && TxQPLL0Size != 0 && RxQPLL1Size != 0)
        {
        RxQPLL = RxQPLL1ParamsVec[0];
        TxQPLL = TxQPLL0ParamsVec[0];
        }
    else if (CurrentTxPllType == mptQPLL1 && CurrentRxPllType == mptQPLL0 && TxQPLL1Size != 0 && RxQPLL0Size != 0)
        {
        RxQPLL = RxQPLL0ParamsVec[0];
        TxQPLL = TxQPLL1ParamsVec[0];
        }
    else if (CurrentTxPllType == mptCPLL && CurrentRxPllType == mptQPLL0 && TxCPLLSize != 0 && RxQPLL0Size != 0)
        {
        RxQPLL = RxQPLL0ParamsVec[0];
        TxCPLL = TxCPLLParamsVec[0];
        }
    else if (CurrentTxPllType == mptCPLL && CurrentRxPllType == mptQPLL1 && TxCPLLSize != 0 && RxQPLL1Size != 0)
        {
        RxQPLL = RxQPLL1ParamsVec[0];
        TxCPLL = TxCPLLParamsVec[0];
        }
    else if (CurrentTxPllType == mptQPLL0 && CurrentRxPllType == mptCPLL && TxQPLL0Size != 0 && RxCPLLSize != 0)
        {
        RxCPLL = RxCPLLParamsVec[0];
        TxQPLL = TxQPLL0ParamsVec[0];
        }
    else if (CurrentTxPllType == mptQPLL1 && CurrentRxPllType == mptCPLL && TxQPLL1Size != 0 && RxCPLLSize != 0)
        {
        RxCPLL = RxCPLLParamsVec[0];
        TxQPLL = TxQPLL1ParamsVec[0];
        }
    // Select the same type MGT PLL to share between the Rx/Tx if possible
    else if (CheckCommonQPLLParams(0) || CheckCommonQPLLParams(1))
        {
        Debug.Log("MGT Rx/Tx Sharing a common QPLL setting");
        RxQPLL = CommonQPLLParams;
        TxQPLL = CommonQPLLParams;
        }
    else if (CheckCommonCPLLParams())
        {
        Debug.Log("MGT Rx/Tx Sharing a common CPLL setting");
        RxCPLL = CommonCPLLParams;
        TxCPLL = CommonCPLLParams;
        }
    else if (TxQPLL0Size != 0 && RxQPLL1Size != 0 && (FpgaType == ftUltrascale || FpgaType == ftUltrascalePlus))
        {
        RxQPLL = RxQPLL1ParamsVec[0];
        TxQPLL = TxQPLL0ParamsVec[0];
        }
    else if (TxQPLL1Size != 0 && RxQPLL0Size != 0 && (FpgaType == ftUltrascale || FpgaType == ftUltrascalePlus))
        {
        RxQPLL = RxQPLL0ParamsVec[0];
        TxQPLL = TxQPLL1ParamsVec[0];
        }
    else if (TxQPLLSize == 0 && TxCPLLSize == 0 && RxQPLLSize == 0 && RxCPLLSize != 0)
        {
        RxCPLL = RxCPLLParamsVec[0];
        TxCPLL = RxCPLLParamsVec[0];
        }
    else if (TxQPLLSize == 0 && TxCPLLSize == 0 && RxQPLLSize != 0 && RxCPLLSize == 0)
        {
        RxQPLL = RxQPLLParamsVec[0];
        TxQPLL = RxQPLLParamsVec[0];
        }
    else if (TxQPLLSize == 0 && TxCPLLSize != 0 && RxQPLLSize == 0 && RxCPLLSize == 0)
        {
        RxCPLL = TxCPLLParamsVec[0];
        TxCPLL = TxCPLLParamsVec[0];
        }
    else if (TxQPLLSize != 0 && TxCPLLSize == 0 && RxQPLLSize == 0 && RxCPLLSize == 0)
        {
        RxQPLL = TxQPLLParamsVec[0];
        TxQPLL = TxQPLLParamsVec[0];
        }
    else if (TxQPLLSize != 0 && TxCPLLSize != 0 && RxQPLLSize == 0 && RxCPLLSize == 0)
        {
        RxCPLL = TxCPLLParamsVec[0];
        TxCPLL = TxCPLLParamsVec[0];
        }
    else if (TxQPLLSize == 0 && TxCPLLSize == 0 && RxQPLLSize != 0 && RxCPLLSize != 0)
        {
        RxCPLL = RxCPLLParamsVec[0];
        RxCPLL = RxCPLLParamsVec[0];
        }
    else if (TxQPLLSize == 0 && TxCPLLSize != 0 && RxQPLLSize == 0 && RxCPLLSize != 0)
        {
        RxCPLL = RxCPLLParamsVec[0];
        TxCPLL = RxCPLLParamsVec[0];
        }
    else if (TxQPLLSize == 0 && TxCPLLSize != 0 && RxQPLLSize != 0 && RxCPLLSize == 0)
        {
        RxQPLL = RxQPLLParamsVec[0];
        TxCPLL = TxCPLLParamsVec[0];
        }
    else if (TxQPLLSize != 0 && TxCPLLSize == 0 && RxQPLLSize == 0 && RxCPLLSize != 0)
        {
        RxCPLL = RxCPLLParamsVec[0];
        TxQPLL = TxQPLLParamsVec[0];
        }
    else if (TxQPLLSize != 0 && TxCPLLSize == 0 && RxQPLLSize != 0 && RxCPLLSize == 0)
        {
        RxQPLL = RxQPLLParamsVec[0];
        TxQPLL = RxQPLLParamsVec[0];
        }
    else if (TxQPLLSize == 0 && TxCPLLSize != 0 && RxQPLLSize != 0 && RxCPLLSize != 0)
        {
        RxQPLL = RxQPLLParamsVec[0];
        TxCPLL = TxCPLLParamsVec[0];
        }
    else if (TxQPLLSize != 0 && TxCPLLSize == 0 && RxQPLLSize != 0 && RxCPLLSize != 0)
        {
        RxCPLL = RxCPLLParamsVec[0];
        TxQPLL = TxQPLLParamsVec[0];
        }
    else if (TxQPLLSize != 0 && TxCPLLSize != 0 && RxQPLLSize == 0 && RxCPLLSize != 0)
        {
        RxCPLL = RxCPLLParamsVec[0];
        TxQPLL = TxQPLLParamsVec[0];
        }
    else if (TxQPLLSize != 0 && TxCPLLSize != 0 && RxQPLLSize != 0 && RxCPLLSize == 0)
        {
        RxQPLL = RxQPLLParamsVec[0];
        TxCPLL = TxCPLLParamsVec[0];
        }
    else if (TxQPLLSize != 0 && TxCPLLSize != 0 && RxQPLLSize != 0 && RxCPLLSize != 0)
        {
        RxQPLL = RxQPLLParamsVec[0];
        TxCPLL = TxCPLLParamsVec[0];
        }

    // Use the selector outputs (a setting per direction), to set the final parameters
    if (RxCPLL.IsValid())
        {
        NewRxPllType  = mptCPLL;
        NewCPLLParams = RxCPLL;
        }
    else
        NewRxPllType = RxQPLL.QPLLType();

    if (TxCPLL.IsValid())
        {
        NewTxPllType  = mptCPLL;
        NewCPLLParams = TxCPLL;
        }
    else
        NewTxPllType = TxQPLL.QPLLType();

    if (RxQPLL.IsValid())
        NewQPLLParamsVec.push_back(RxQPLL);
    if (TxQPLL.IsValid() && (!RxQPLL.IsValid() || !(RxQPLL == TxQPLL)))
        NewQPLLParamsVec.push_back(TxQPLL);
}

//---------------------------------------------------------------------------
//  JESD_PHY::ConfigureFpgaMgts() -- Configure the FPGA MGTs
//---------------------------------------------------------------------------

void  JESD_PHY::ConfigureFpgaMgts()
{
    Debug.Log("Configure the FPGA MGTs");

    if (!NewCPLLParams.IsValid() && NewQPLLParamsVec.empty())       // Skip if no solution was found
        {
        Debug.Log("Skipping FPGA MGT configuration (no solution was found)");
        return;
        }

    QPLLPowered_All(0, false);
    QPLLPowered_All(1, false);
    if ((NewTxPllType != mptCPLL) || (NewRxPllType != mptCPLL))
        ConfigureMgtQPLL();

    CPLLPowered_All(false);
    if ((NewTxPllType == mptCPLL) || (NewRxPllType == mptCPLL))
        ConfigureMgtCPLL();

    for (unsigned int i=0; i<NumGtChannel; i++)
        {
        SetGTCHANSelect(i);        // Select a GT_CHANNEL

        if ((RxOutputDivider != CurrentRxOutputDivider) || (TxOutputDivider != CurrentTxOutputDivider))
            SetOutputDividers();
        // Update the Tx/Rx PLL clock selection when switching to a different MGT PLL type
        if (NewTxPllType != CurrentTxPllType)
            SetTxPllClockSelect();
        if (NewRxPllType != CurrentRxPllType)
            SetRxPllClockSelect();
        }

    for (unsigned int idx = 0; idx < NewQPLLParamsVec.size(); idx++)
        QPLLPowered_All(NewQPLLParamsVec[idx].QPLLIndex, true);

    if (NewCPLLParams.IsValid())
        CPLLPowered_All(true);
}

//---------------------------------------------------------------------------
//  JESD_PHY::ConfigureMgtCPLL() -- Configure the FPGA MGTs CPLL
//---------------------------------------------------------------------------

void  JESD_PHY::ConfigureMgtCPLL()
{
    Debug.Log("Configure the FPGA MGTs CPLL");

    if (!NewCPLLParams.IsValid())
        {
        Debug.Log("Skipping FPGA MGT CPLL configuration (no solution was found)");
        return;
        }
    else if (NewCPLLParams == CurrentCPLLParams)
        {
        Debug.Log("Skipping FPGA MGT CPLL configuration (no change in parameters)");
        return;
        }

    unsigned int CPLL_REFCLK_DIV;
    unsigned int CPLL_FB_DIV_45;
    unsigned int CPLL_FBDIV;

    if (NewCPLLParams.M == 1)
        CPLL_REFCLK_DIV = 16;
    else            // 2
        CPLL_REFCLK_DIV = 0;

    if (NewCPLLParams.N1 == 4)
        CPLL_FB_DIV_45 = 0;
    else            // 5
        CPLL_FB_DIV_45 = 1;

    if (NewCPLLParams.N2 == 1)
        CPLL_FBDIV = 16;
    else if (NewCPLLParams.N2 == 2)
        CPLL_FBDIV = 0;
    else if (NewCPLLParams.N2 == 3)
        CPLL_FBDIV = 1;
    else if (NewCPLLParams.N2 == 4)
        CPLL_FBDIV = 2;
    else                // 5
        CPLL_FBDIV = 3;

    MemoryBitManipulator Data(0);

    for (unsigned int i=0; i<NumGtChannel; i++)
        {
        SetGTCHANSelect(i);          // Select a GT_CHANNEL

        if (FpgaType == ft7Series)
            {
            Data.Value(GetDrpReg(mdtGTCHAN, 0x5E));
            Data.Field(8, 5, CPLL_REFCLK_DIV);
            Data.Bit(7, (CPLL_FB_DIV_45) ? 1 : 0);
            Data.Field(0, 7, CPLL_FBDIV);
            SetDrpReg(mdtGTCHAN, 0x5E, Data.Value());
            }

        if (FpgaType == ftUltrascale || FpgaType == ftUltrascalePlus)
            {
            if (NewCPLLParams.M != CurrentCPLLParams.M)
                {
                Data.Value(GetDrpReg(mdtGTCHAN, 0x2A));
                Data.Field(11, 5, CPLL_REFCLK_DIV);
                SetDrpReg(mdtGTCHAN, 0x2A, Data.Value());
                }

            if ((NewCPLLParams.N1 != CurrentCPLLParams.N1) || (NewCPLLParams.N2 != CurrentCPLLParams.N2))
                {
                Data.Value(GetDrpReg(mdtGTCHAN, 0x28));
                Data.Bit(7, (CPLL_FB_DIV_45) ? 1 : 0);
                Data.Field(8, 8, CPLL_FBDIV);
                SetDrpReg(mdtGTCHAN, 0x28, Data.Value());
                }
            }
        }
}

//---------------------------------------------------------------------------
//  JESD_PHY::ConfigureMgtQPLL() -- Configure the FPGA MGTs QPLL
//---------------------------------------------------------------------------

void  JESD_PHY::ConfigureMgtQPLL()
{
    Debug.Log("Configure the FPGA MGTs QPLL");

    if (CurrentQPLLParamsVec.empty())
        {
        Debug.Log("Skipping FPGA MGT QPLL configuration (no QPLLs in the logic)");
        return;
        }

    if (NewQPLLParamsVec.empty())
        {
        Debug.Log("Skipping FPGA MGT QPLL configuration (no solution was found)");
        return;
        }

     if (FpgaType == ft7Series)
         ConfigureQPLL_7S(CurrentQPLLParamsVec[0], NewQPLLParamsVec[0]);
     else               // (FpgaType == ftUltrascale || FpgaType == ftUltrascalePlus)
        {
         for (unsigned int NewIdx=0; NewIdx < NewQPLLParamsVec.size(); NewIdx++)
             ConfigureQPLL_US(CurrentQPLLParamsVec[NewQPLLParamsVec[NewIdx].QPLLIndex], NewQPLLParamsVec[NewIdx]);
        }
}

//---------------------------------------------------------------------------
//  JESD_PHY::ConfigureQPLL_7S() -- Configure the 7Series FPGA MGT QPLL
//---------------------------------------------------------------------------

void  JESD_PHY::ConfigureQPLL_7S(QPLLParamsType CurrentQPLLParams, QPLLParamsType NewQPLLParams)
{
    unsigned int CurrentQPLLIndex = CurrentQPLLParams.QPLLIndex;

    if (NewQPLLParams == CurrentQPLLParams)
        {
        if (CurrentQPLLIndex == 0)
            Debug.Log("Skipping FPGA MGT QPLL0 configuration (no change in parameters)");
        else
            Debug.Log("Skipping FPGA MGT QPLL1 configuration (no change in parameters)");
        return;
        }

    unsigned int QPLL_REFCLK_DIV;
    unsigned int QPLL_FBDIV;

    if (NewQPLLParams.M == 1)
        QPLL_REFCLK_DIV = 16;
    else if (NewQPLLParams.M == 2)
        QPLL_REFCLK_DIV = 0;
    else if (NewQPLLParams.M == 3)
        QPLL_REFCLK_DIV = 1;
    else                // 4
        QPLL_REFCLK_DIV = 2;

    MemoryBitManipulator Data(0);

    if (NewQPLLParams.N == 16)
        QPLL_FBDIV = 0x20;
    else if (NewQPLLParams.N == 20)
        QPLL_FBDIV = 0x30;
    else if (NewQPLLParams.N == 32)
        QPLL_FBDIV = 0x60;
    else if (NewQPLLParams.N == 40)
        QPLL_FBDIV = 0x80;
    else if (NewQPLLParams.N == 64)
        QPLL_FBDIV = 0xE0;
    else if (NewQPLLParams.N == 66)
        QPLL_FBDIV = 0x140;
    else if (NewQPLLParams.N == 80)
        QPLL_FBDIV = 0x120;
    else                // 100
        QPLL_FBDIV = 0x170;

    bool QPLL_FBDIV_RATIO = (NewQPLLParams.N == 66)? false : true;

    for (unsigned int i=0; i<NumGtCommon; i++)
        {
        SetGTCMNSelect(i);            // Select a GT_COMMON

        if (NewQPLLParams.QPLLIndex != CurrentQPLLParams.QPLLIndex)
            {
            Data.Value(GetDrpReg(mdtGTCMN, 0x32));
            Data.Bit(6, (NewQPLLParams.QPLLIndex) ? 1 : 0);
            SetDrpReg(mdtGTCMN, 0x32, Data.Value());
            }

        if (NewQPLLParams.M != CurrentQPLLParams.M)
            {
            Data.Value(GetDrpReg(mdtGTCMN, 0x33));
            Data.Field(11, 5, QPLL_REFCLK_DIV);
            SetDrpReg(mdtGTCMN, 0x33, Data.Value());
            }

        if (NewQPLLParams.N != CurrentQPLLParams.N)
            {
            Data.Value(GetDrpReg(mdtGTCMN, 0x36));
            Data.Field(0, 10, QPLL_FBDIV);
            SetDrpReg(mdtGTCMN, 0x36, Data.Value());

            Data.Value(GetDrpReg(mdtGTCMN, 0x37));
            Data.Bit(6, QPLL_FBDIV_RATIO);
            SetDrpReg(mdtGTCMN, 0x37, Data.Value());
            }
        }
}

//---------------------------------------------------------------------------
//  JESD_PHY::ConfigureQPLL_US() -- Configure the UltraScale, UltraScale+ FPGA MGT QPLL
//---------------------------------------------------------------------------

void  JESD_PHY::ConfigureQPLL_US(QPLLParamsType CurrentQPLLParams, QPLLParamsType NewQPLLParams)
{
    unsigned int CurrentQPLLIndex = CurrentQPLLParams.QPLLIndex;

    if (NewQPLLParams == CurrentQPLLParams)
        {
        if (CurrentQPLLIndex == 0)
            Debug.Log("Skipping FPGA MGT QPLL0 configuration (no change in parameters)");
        else
            Debug.Log("Skipping FPGA MGT QPLL1 configuration (no change in parameters)");
        return;
        }

    unsigned int QPLL_REFCLK_DIV;
    unsigned int QPLL_FBDIV;

    if (NewQPLLParams.M == 1)
        QPLL_REFCLK_DIV = 16;
    else if (NewQPLLParams.M == 2)
        QPLL_REFCLK_DIV = 0;
    else if (NewQPLLParams.M == 3)
        QPLL_REFCLK_DIV = 1;
    else                // 4
        QPLL_REFCLK_DIV = 2;

    MemoryBitManipulator Data(0);

    QPLL_FBDIV = NewQPLLParams.N - 2;

    unsigned int NRegAddr = (NewQPLLParams.QPLLIndex == 0)? 0x14 : 0x94;
    unsigned int MRegAddr = (NewQPLLParams.QPLLIndex == 0)? 0x18 : 0x98;

    for (unsigned int i=0; i<NumGtCommon; i++)
        {
        SetGTCMNSelect(i);            // Select a GT_COMMON

        if ((NewQPLLParams.N != CurrentQPLLParams.N) || (NewQPLLParams.QPLLIndex != CurrentQPLLParams.QPLLIndex))
            {
            Data.Value(GetDrpReg(mdtGTCMN, NRegAddr));
            Data.Field(0, 8, QPLL_FBDIV);
            SetDrpReg(mdtGTCMN, NRegAddr, Data.Value());
            }

        if ((NewQPLLParams.M != CurrentQPLLParams.M) || (NewQPLLParams.QPLLIndex != CurrentQPLLParams.QPLLIndex))
            {
            Data.Value(GetDrpReg(mdtGTCMN, MRegAddr));
            Data.Field(7, 5, QPLL_REFCLK_DIV);
            SetDrpReg(mdtGTCMN, MRegAddr, Data.Value());
            }
        }
}

//---------------------------------------------------------------------------
//  JESD_PHY::CheckFpgaMgtsPllLock() -- Configure the FPGA MGTs
//---------------------------------------------------------------------------

void  JESD_PHY::CheckFpgaMgtsPllLock()
{
    Debug.Log("Check the FPGA MGTs PLL lock");

    GetTxRxPllClockSelect();

    if ((CurrentTxPllType == mptCPLL) || (CurrentRxPllType == mptCPLL))
        {
        if (!CheckCPLLLock())
            {
            CPLLPowered_All(false);
            CPLLPowered_All(true);
            CheckCPLLLock();
            }
        }

    if ((CurrentTxPllType == mptQPLL0) || (CurrentRxPllType == mptQPLL0))
        {
        if (!CheckQPLLLock(0))
            {
            // Power down then up the QPLLs
            QPLLPowered_All(0, false);
            QPLLPowered_All(0, true);
            CheckQPLLLock(0);
            }
        }

    if ((CurrentTxPllType == mptQPLL1) || (CurrentRxPllType == mptQPLL1))
        {
        if (!CheckQPLLLock(1))
            {
            // Power down then up the QPLLs
            QPLLPowered_All(1, false);
            QPLLPowered_All(1, true);
            CheckQPLLLock(1);
            }
        }
}

//---------------------------------------------------------------------------
//  JESD_PHY::SetNumInterfaces()
//---------------------------------------------------------------------------

void  JESD_PHY::SetNumInterfaces()
{
    NumGtCommon  = PeekJesdPhy(0x8);    // Read number of GT_COMMON blocks
    NumGtChannel = PeekJesdPhy(0xC);    // Read number of GT_CHANNEL blocks
}

//---------------------------------------------------------------------------
//  JESD_PHY::SetGTCMNSelect()
//---------------------------------------------------------------------------

void  JESD_PHY::SetGTCMNSelect(unsigned int which)
{
    PokeJesdPhy(0x20, which);       // Set cmm_interface_sel
    SetDrpSelect(mdtGTCMN, which);
    if (IpVersion < 0x03000000)
        PokeJesdPhy(0x300, which);  // Set PLL Select
}

//---------------------------------------------------------------------------
//  JESD_PHY::SetGTCHANSelect()
//---------------------------------------------------------------------------

void  JESD_PHY::SetGTCHANSelect(unsigned int which)
{
    PokeJesdPhy(0x24, which);       // Set gt_interface_sel
    SetDrpSelect(mdtGTCHAN,which);
    if (IpVersion < 0x03000000)
        {
        PokeJesdPhy(0x400, which);  // Set Transceiver Select (bank 1)
        PokeJesdPhy(0x500, which);  // Set Transceiver Select (bank 2)
        PokeJesdPhy(0x600, which);  // Set Transceiver Select (bank 3)
        }
}

//---------------------------------------------------------------------------
//  JESD_PHY::CheckCPLLLock()
//---------------------------------------------------------------------------

bool  JESD_PHY::CheckCPLLLock()
{
    unsigned int PllStatus;
    bool Lock = false;

    // Poll the CPLL lock status
    for (unsigned int i=0; i<5; i++)
        {
        PllStatus = PeekJesdPhy(0x80);   // Read the PLL Status
        Lock = ((PllStatus & 0x4) == 0); // Check the all CPLLs locked bit (active low)
        if (Lock)
            {
            Debug.Log("JESD PHY: All CPLLs locked");
            break;
            }
        else if (i<4)
            Sleep(1);
        else
            Debug.Log("JESD PHY: CPLLs lock timeout");
        }
    Debug.Log("JESD PHY: PLL Status = ", PllStatus, 16);
    return Lock;
}

//---------------------------------------------------------------------------
//  JESD_PHY::CheckQPLLLock()
//---------------------------------------------------------------------------

bool  JESD_PHY::CheckQPLLLock(unsigned int which)
{
    unsigned int LockBitMask;

    if ((FpgaType == ftUltrascale || FpgaType == ftUltrascalePlus) && which == 1)
        LockBitMask = 0x1;
    else
        LockBitMask = 0x2;

    unsigned int PllStatus;
    bool Lock = false;

    // Poll the QPLL lock status
    for (unsigned int i=0; i<5; i++)
        {
        PllStatus = PeekJesdPhy(0x80);       // Read the PLL Status
        Lock = ((PllStatus & LockBitMask) == 0);// Check the all QPLLs locked bit (active low)
        if (Lock)
            {
            Debug.Log("JESD PHY: All QPLLs locked");
            break;
            }
        else if (i<4)
            Sleep(1);
        else
            Debug.Log("JESD PHY: QPLLs lock timeout");
        }
    Debug.Log("JESD PHY: PLL Status = ", PllStatus, 16);
    return Lock;
}

//---------------------------------------------------------------------------
//  JESD_PHY::SetDrpSelect()
//---------------------------------------------------------------------------

void  JESD_PHY::SetDrpSelect(unsigned int which, unsigned int select)
{
    unsigned int DrpAddr = (which == mdtGTCMN)? 0x100 : 0x200;
    if (IpVersion < 0x03000000)
        PokeJesdPhy(DrpAddr, select);
}

//---------------------------------------------------------------------------
//  Fmc1000::SetDrpReset()
//---------------------------------------------------------------------------

void  JESD_PHY::SetDrpReset(unsigned int which)
{
    unsigned int DrpAddr = (which == mdtGTCMN)? 0x110 : 0x210;
    PokeJesdPhy(DrpAddr, 1);
}

//---------------------------------------------------------------------------
//  JESD_PHY::SetDrpReg()
//---------------------------------------------------------------------------

void  JESD_PHY::SetDrpReg(unsigned int which, unsigned int address, unsigned int data)
{
    unsigned int DrpWrDataAddr = (which == mdtGTCMN)? 0x108 : 0x208;

    MemoryBitManipulator Data(0);
    Data.Field(0, 16, data);
    PokeJesdPhy(DrpWrDataAddr, Data.Value());

    unsigned int DrpAddr = (which == mdtGTCMN)? 0x104 : 0x204;

    Data.Value(0);
    Data.Field(0, 30, address);
    Data.Bit(31, 1);
    PokeJesdPhy(DrpAddr, Data.Value());

    std::stringstream msg;
    msg << std::hex << "Set JESD204B PHY IP DRP Dev" << DeviceID << " (0x" << address << ", 0x" << data << ")";
    Debug.Log(msg.str());
}

//---------------------------------------------------------------------------
//  JESD_PHY::GetDrpReg()
//---------------------------------------------------------------------------

unsigned int  JESD_PHY::GetDrpReg(unsigned int which, unsigned int address)
{
    unsigned int DrpAddr = (which == mdtGTCMN)? 0x104 : 0x204;

    MemoryBitManipulator Data(0);
    Data.Field(0, 30, address);
    Data.Bit(30, 1);
    PokeJesdPhy(DrpAddr, Data.Value());

    unsigned int DrpRdDataAddr = (which == mdtGTCMN)? 0x10C : 0x20C;

    Data.Value(PeekJesdPhy(DrpRdDataAddr));

    unsigned int DrpRdData = Data.Field(0,16);

    std::stringstream msg;
    msg << std::hex << "Get JESD204B PHY IP DRP Dev" << DeviceID << " (0x" << address << ") = 0x" << DrpRdData;
    Debug.Log(msg.str());

    return DrpRdData;
}

//---------------------------------------------------------------------------
//  JESD_PHY::QPLLPowered() -- GT_COMMON QPLL power control
//---------------------------------------------------------------------------

void  JESD_PHY::QPLLPowered(unsigned int QPLLIdx, bool enable)
{
    unsigned int QPLLPdAddr;
    if ((FpgaType == ftUltrascale || FpgaType == ftUltrascalePlus) && QPLLIdx == 1)
        QPLLPdAddr = 0x308;
    else
        QPLLPdAddr = 0x304;

    PokeJesdPhy(QPLLPdAddr, (enable == 0));
}

//---------------------------------------------------------------------------
//  JESD_PHY::QPLLPowered_All() -- All GT_COMMONs' QPLLs power control
//---------------------------------------------------------------------------

void  JESD_PHY::QPLLPowered_All(unsigned int QPLLIdx, bool enable)
{
    if (!IsGtCommonEnabled())
        return;
    for (unsigned int i=0; i<NumGtCommon; i++)
        {
        SetGTCMNSelect(i);            // Select a GT_COMMON
        QPLLPowered(QPLLIdx, enable);
        }
    if (enable)
        Debug.Log("Powered up FPGA MGT QPLL ", QPLLIdx);
    else
        Debug.Log("Powered down FPGA MGT QPLL ", QPLLIdx);
}

//---------------------------------------------------------------------------
//  JESD_PHY::CPLLPowered() -- GT_CHANNEL CPLL power control
//---------------------------------------------------------------------------

void JESD_PHY::CPLLPowered(bool enable)
{
    PokeJesdPhy(0x408, (enable == 0));
}

//---------------------------------------------------------------------------
//  JESD_PHY::CPLLPowered_All() -- All GT_CHANNELs' CPLL power control
//---------------------------------------------------------------------------

void JESD_PHY::CPLLPowered_All(bool enable)
{
    for (unsigned int i=0; i<NumGtChannel; i++)
        {
        SetGTCHANSelect(i);          // Select a GT_CHANNEL
        CPLLPowered(enable);
        }
    Debug.Log("Powered down the FPGA MGT CPLL");
}

//---------------------------------------------------------------------------
//  JESD_PHY::SetTxPllClockSelect()
//---------------------------------------------------------------------------

void  JESD_PHY::SetTxPllClockSelect()
{
    unsigned int PllType;
    if (NewTxPllType == mptCPLL)
        PllType = 0;
    else if ((FpgaType == ftUltrascale || FpgaType == ftUltrascalePlus) &&
             NewTxPllType == mptQPLL1)
        PllType = 2;
    else
        PllType = 3;
    PokeJesdPhy(0x40C, PllType);
}

//---------------------------------------------------------------------------
//  JESD_PHY::SetRxPllClockSelect()
//---------------------------------------------------------------------------

void  JESD_PHY::SetRxPllClockSelect()
{
    unsigned int PllType;
    if (NewRxPllType == mptCPLL)
        PllType = 0;
    else if ((FpgaType == ftUltrascale || FpgaType == ftUltrascalePlus) &&
             NewRxPllType == mptQPLL1)
        PllType = 2;
    else
        PllType = 3;
    PokeJesdPhy(0x410, PllType);
}

//---------------------------------------------------------------------------
//  JESD_PHY::GetTxRxPllClockSelect()
//---------------------------------------------------------------------------

void JESD_PHY::GetTxRxPllClockSelect()
{
    unsigned int CurrentValue = PeekJesdPhy(0x40C);
    if (CurrentValue == 0 || !IsGtCommonEnabled())
        CurrentTxPllType = mptCPLL;
    else if (CurrentValue == 2)  // UltraScale only
        CurrentTxPllType = mptQPLL1;
    else                         // UltraScale and 7Series
        CurrentTxPllType = mptQPLL0;

    CurrentValue = PeekJesdPhy(0x410);
    if (CurrentValue == 0 || !IsGtCommonEnabled())
        CurrentRxPllType = mptCPLL;
    else if (CurrentValue == 2)  // UltraScale only
        CurrentRxPllType = mptQPLL1;
    else                         // UltraScale and 7Series
        CurrentRxPllType = mptQPLL0;

    // For 7Series FPGAs, the QPLL index is read from the QPLL_CFG[6] register
    if (FpgaType == ft7Series)
        {
        unsigned int QPLLIndex = 0;
        if ((CurrentRxPllType == mptQPLL0) || (CurrentTxPllType == mptQPLL0))
            QPLLIndex = GetQPLLIndex_7S();

        if ((CurrentTxPllType == mptQPLL0) && (QPLLIndex == 1))
            CurrentTxPllType = mptQPLL1;
        if ((CurrentRxPllType == mptQPLL0) && (QPLLIndex == 1))
            CurrentRxPllType = mptQPLL1;
        }
}

//---------------------------------------------------------------------------
//  JESD_PHY::GetCPLLParams()
//---------------------------------------------------------------------------

void  JESD_PHY::GetCPLLParams()
{
    unsigned int MReg = 0;
    unsigned int N1Reg = 0;
    unsigned int N2Reg = 0;

    MemoryBitManipulator Data(0);

    SetGTCHANSelect(0);

    if (FpgaType == ft7Series)
        {
        Data.Value(GetDrpReg(mdtGTCHAN, 0x5E));
        MReg = Data.Field(8,5);
        N1Reg = Data.Field(7,1);
        N2Reg = Data.Field(0,7);
        }

    if (FpgaType == ftUltrascale || FpgaType == ftUltrascalePlus)
        {
        Data.Value(GetDrpReg(mdtGTCHAN, 0x2A));
        MReg = Data.Field(11,5);
        Data.Value(GetDrpReg(mdtGTCHAN, 0x28));
        N1Reg = Data.Field(7,1);
        N2Reg = Data.Field(8,8);
        }

    if (MReg == 16)
        CurrentCPLLParams.M = 1;
    else if (MReg == 0)
        CurrentCPLLParams.M = 2;
    else
        CurrentCPLLParams.M = 0;

    if (N1Reg == 0)
        CurrentCPLLParams.N1 = 4;
    else if (N1Reg == 1)
        CurrentCPLLParams.N1 = 5;
    else
        CurrentCPLLParams.N1 = 0;

    if (N2Reg == 16)
        CurrentCPLLParams.N2 = 1;
    else if (N2Reg == 0)
        CurrentCPLLParams.N2 = 2;
    else if (N2Reg == 1)
        CurrentCPLLParams.N2 = 3;
    else if (N2Reg == 2)
        CurrentCPLLParams.N2 = 4;
    else if (N2Reg == 3)
        CurrentCPLLParams.N2 = 5;
    else
        CurrentCPLLParams.N2 = 0;
}

//---------------------------------------------------------------------------
//  JESD_PHY::GetQPLLParams()
//---------------------------------------------------------------------------

void  JESD_PHY::GetQPLLParams()
{
    CurrentQPLLParamsVec.clear();

    if (!IsGtCommonEnabled())
        return;

    unsigned int MReg = 0;
    unsigned int NReg = 0;
    QPLLParamsType QPLLParams;

    MemoryBitManipulator Data(0);

    SetGTCMNSelect(0);

    if (FpgaType == ft7Series)
        {
        Data.Value(GetDrpReg(mdtGTCMN, 0x33));
        MReg = Data.Field(11,5);
        if (MReg == 16)
            QPLLParams.M = 1;
        else if (MReg == 0)
            QPLLParams.M = 2;
        else if (MReg == 1)
            QPLLParams.M = 3;
        else if (MReg == 2)
            QPLLParams.M = 4;
        else
            {
            QPLLParams.M = 0;
            return;         // Skip since no legit M value was found
            }

        Data.Value(GetDrpReg(mdtGTCMN, 0x36));
        NReg = Data.Field(0,10);
        if (NReg == 0x20)
            QPLLParams.N = 16;
        else if (NReg == 0x30)
            QPLLParams.N = 20;
        else if (NReg == 0x60)
            QPLLParams.N = 32;
        else if (NReg == 0x80)
            QPLLParams.N = 40;
        else if (NReg == 0xE0)
            QPLLParams.N = 64;
        else if (NReg == 0x140)
            QPLLParams.N = 66;
        else if (NReg == 0x120)
            QPLLParams.N = 80;
        else if (NReg == 0x170)
            QPLLParams.N = 100;
        else
            {
            QPLLParams.N = 0;
            return;         // Skip since no legit N value was found
            }

        // For 7Series FPGAs, the QPLL index is read from the QPLL_CFG[6] register
        QPLLParams.QPLLIndex = GetQPLLIndex_7S();

        CurrentQPLLParamsVec.push_back(QPLLParams);
        }

    if (FpgaType == ftUltrascale || FpgaType == ftUltrascalePlus)
        {
        for (unsigned int idx = 0; idx < 2; idx++)
            {
            unsigned int MRegAddr = (idx == 0)? 0x18 : 0x98;
            unsigned int NRegAddr = (idx == 0)? 0x14 : 0x94;

            Data.Value(GetDrpReg(mdtGTCMN, MRegAddr));
            MReg = Data.Field(7,5);
            if (MReg == 16)
                QPLLParams.M = 1;
            else if (MReg == 0)
                QPLLParams.M = 2;
            else if (MReg == 1)
                QPLLParams.M = 3;
            else if (MReg == 2)
                QPLLParams.M = 4;
            else
                {
                QPLLParams.M = 0;
                break;         // Skip since no legit M value was found
                }

            Data.Value(GetDrpReg(mdtGTCMN, NRegAddr));
            NReg = Data.Field(0,8);
            QPLLParams.N = NReg+2;

            QPLLParams.QPLLIndex = idx;
            CurrentQPLLParamsVec.push_back(QPLLParams);
            }
        }
}

//---------------------------------------------------------------------------
//  JESD_PHY::GetQPLLIndex_7S()
//---------------------------------------------------------------------------

unsigned int  JESD_PHY::GetQPLLIndex_7S()
{
    SetGTCMNSelect(0);

    unsigned int QPLL_CFG = GetDrpReg(mdtGTCMN, 0x32);
    if ((QPLL_CFG & 0x40) == 0)
        return 0;
    else
        return 1;
}

//---------------------------------------------------------------------------
//  JESD_PHY::SetOutputDividers()
//---------------------------------------------------------------------------

void  JESD_PHY::SetOutputDividers()
{
    if (TxOutputDivider == 0 || RxOutputDivider == 0)
        {
         Debug.Log("Skipping MGT output dividr configuration since the calculated divider value = 0");
         return;
        }

    MemoryBitManipulator Data(0);

    if (FpgaType == ft7Series)
        {
        Data.Value(GetDrpReg(mdtGTCHAN, 0x88));
        Data.Field(4, 3, log2(TxOutputDivider));
        Data.Field(0, 3, log2(RxOutputDivider));
        SetDrpReg(mdtGTCHAN, 0x88, Data.Value());
        }

    if (FpgaType == ftUltrascale || FpgaType == ftUltrascalePlus)
        {
        Data.Value(GetDrpReg(mdtGTCHAN, 0x7C));
        Data.Field(8, 3, log2(TxOutputDivider));
        SetDrpReg(mdtGTCHAN, 0x7C, Data.Value());
        Data.Value(GetDrpReg(mdtGTCHAN, 0x63));
        Data.Field(0, 3, log2(RxOutputDivider));
        SetDrpReg(mdtGTCHAN, 0x63, Data.Value());
        }
}

//---------------------------------------------------------------------------
//  JESD_PHY::GetOutputDividers()
//---------------------------------------------------------------------------

void  JESD_PHY::GetOutputDividers()
{
    unsigned int TxOutDiv = 0;
    unsigned int RxOutDiv = 0;

    MemoryBitManipulator Data(0);

    SetGTCHANSelect(0);

    if (FpgaType == ft7Series)
        {
        Data.Value(GetDrpReg(mdtGTCHAN, 0x88));
        TxOutDiv = Data.Field(4,3);
        RxOutDiv = Data.Field(0,3);
        }

    if (FpgaType == ftUltrascale || FpgaType == ftUltrascalePlus)
        {
        Data.Value(GetDrpReg(mdtGTCHAN, 0x7C));
        TxOutDiv = Data.Field(8,3);
        Data.Value(GetDrpReg(mdtGTCHAN, 0x63));
        RxOutDiv = Data.Field(0,3);
        }

    CurrentRxOutputDivider = 1 << RxOutDiv;
    CurrentTxOutputDivider = 1 << TxOutDiv;
}

//---------------------------------------------------------------------------
//  JESD_PHY::InvertTxPolarity()
//---------------------------------------------------------------------------

void  JESD_PHY::InvertTxPolarity(unsigned int which, bool invert)
{
    SetGTCHANSelect(which);          // Select a GT_CHANNEL
    PokeJesdPhy(0x510, invert?1:0);  // Set Invert TX polarity
}

//---------------------------------------------------------------------------
//  JESD_PHY::InvertRxPolarity()
//---------------------------------------------------------------------------

void  JESD_PHY::InvertRxPolarity(unsigned int which, bool invert)
{
    SetGTCHANSelect(which);          // Select a GT_CHANNEL
    PokeJesdPhy(0x604, invert?1:0);  // Set Invert RX polarity
}

//---------------------------------------------------------------------------
//  JESD_PHY::RxPowered() -- GT_CHANNEL Rx power control
//---------------------------------------------------------------------------

void JESD_PHY::RxPowered(bool enable)
{
    PokeJesdPhy(0x404, (enable?0:1));
}

//---------------------------------------------------------------------------
//  JESD_PHY::RxPowered_All() -- All GT_CHANNELs' Rx power control
//---------------------------------------------------------------------------

void JESD_PHY::RxPowered_All(bool enable)
{
    for (unsigned int i=0; i<NumGtChannel; i++)
        {
        SetGTCHANSelect(i);          // Select a GT_CHANNEL
        RxPowered(enable);
        }
}

//---------------------------------------------------------------------------
//  JESD_PHY::TxPowered() -- GT_CHANNEL Tx power control
//---------------------------------------------------------------------------

void JESD_PHY::TxPowered(bool enable)
{
    PokeJesdPhy(0x504, (enable?0:1));
}

//---------------------------------------------------------------------------
//  JESD_PHY::TxPowered_All() -- All GT_CHANNELs' Tx power control
//---------------------------------------------------------------------------

void JESD_PHY::TxPowered_All(bool enable)
{
    for (unsigned int i=0; i<NumGtChannel; i++)
        {
        SetGTCHANSelect(i);          // Select a GT_CHANNEL
        TxPowered(enable);
        }
}

//---------------------------------------------------------------------------
//  JESD_PHY::log2()  -- Calculate Log2
//---------------------------------------------------------------------------

unsigned int JESD_PHY::log2(unsigned int value)
{
    unsigned int log = 0;
    while (value != 0)
        {
        value >>= 1;
        log++;
        }
    return log-1;
}

/*
//---------------------------------------------------------------------------
//  JESD_PHY::CalcLineRate() -- Calculate JESD204B line rate
//---------------------------------------------------------------------------

double JESD_PHY::CalcJesdLineRate(unsigned int NumConverters, unsigned int BytePerSample,
                                 double SamplingRate, unsigned int Decimation, unsigned int NumLanes)
{
    return ((NumConverters * BytePerSample * SamplingRate * 10.0)/(Decimation * NumLanes));
}
*/

//---------------------------------------------------------------------------
//  JESD_PHY::CalcMgtRefClk() -- Calculate the MGT reference clock frequency
//---------------------------------------------------------------------------

double JESD_PHY::CalcMgtRefClk(double LineRate)
{
    double MgtRefClk = LineRate/20;

    if (MgtRefClk > 0.0)
        {
        // Adjust MgtRefClk rate to be within the valid range
        while (!WithinRange(MgtRefClk, MgtRefclkRange.Low, MgtRefclkRange.High))
            {
            if (MgtRefClk > MgtRefclkRange.High)
                MgtRefClk /= 2;
            else           // means < MgtRefclkRange.Low
                MgtRefClk *= 2;
            }

        // Maximize MgtRefClk rate
        while (MgtRefClk*2.0 < MgtRefclkRange.High)
            MgtRefClk = MgtRefClk * 2.0;
        }
    return MgtRefClk;
}

//===========================================================================
//  CLASS JESD_ADCDevice -- JESD FPGA ADC interface
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class JESD_ADCDevice
//---------------------------------------------------------------------------

JESD_ADCDevice::JESD_ADCDevice(IRequires_JESDADC_Io * mdc, unsigned int device_id)
    :  MDC(mdc), DeviceID(device_id)
{
}

//---------------------------------------------------------------------------
//  destructor for class JESD_PHY
//---------------------------------------------------------------------------

JESD_ADCDevice::~JESD_ADCDevice()
{
}

//---------------------------------------------------------------------------
//  JESD_ADCDevice::ConstructCustomParts()
//---------------------------------------------------------------------------

void  JESD_ADCDevice::ConstructCustomParts()
{
    if (!DeviceEnabled())
        return;
}

//---------------------------------------------------------------------------
//  JESD_ADCDevice::OpenHardware()
//---------------------------------------------------------------------------

void  JESD_ADCDevice::OpenHardware()
{
    if (!DeviceEnabled())
        return;

    IpVersion = Peek(0);
}

//---------------------------------------------------------------------------
//  JESD_ADCDevice::PreconfigureHardware()
//---------------------------------------------------------------------------

void  JESD_ADCDevice::PreconfigureHardware()
{
    if (!DeviceEnabled())
        return;
}

//---------------------------------------------------------------------------
//  JESD_ADCDevice::ConfigureHardware()
//---------------------------------------------------------------------------

void  JESD_ADCDevice::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;
}

//---------------------------------------------------------------------------
// JESD_ADCDevice::Poke() -- write to JESD logic area
//---------------------------------------------------------------------------

void  JESD_ADCDevice::Poke(unsigned int address, unsigned int data)
{
    MDC->PokeJESDADC(DeviceID, address, data);
}

//---------------------------------------------------------------------------
// JESD_ADCDevice::Peek() -- read from JESD logic area
//---------------------------------------------------------------------------

unsigned int  JESD_ADCDevice::Peek(unsigned int address)
{
    return MDC->PeekJESDADC(DeviceID, address);
}

//---------------------------------------------------------------------------
//  JESD_ADCDevice::CalcLineRate() -- Calculate JESD204B line rate
//---------------------------------------------------------------------------

double JESD_ADCDevice::CalcJesdLineRate(unsigned int NumConverters, unsigned int BytePerSample,
                                 double SamplingRate, unsigned int Decimation, unsigned int NumLanes)
{
    return ((NumConverters * BytePerSample * SamplingRate * 10.0)/(Decimation * NumLanes));
}

//---------------------------------------------------------------------------
//  JESD_ADCDevice::DumpDebugStatus() -- useful status dump code
//---------------------------------------------------------------------------

void  JESD_ADCDevice::DumpDebugStatus()
{
    Debug.Log("Adc JESD204B IP Version Reg(0x0) =", IpVersion,  16);
    Debug.Log("Adc JESD204B IP Link Error Status Reg(0x01C) =", Peek(0x1C),  16);
    Debug.Log("Adc JESD204B IP RX Buffer Delay Reg(0x030) =",   Peek(0x30),  16);
    Debug.Log("Adc JESD204B IP Sync Status Reg(0x038) =",       Peek(0x38),  16);
    if (IpVersion >= 0x07000100)
        Debug.Log("Adc JESD204B IP Debug Status Reg(0x03C) =",  Peek(0x3C),  16);
    Debug.Log("Adc JESD204B IP ILA Config Data0 Reg(0x800) =",  Peek(0x800), 16);
    Debug.Log("Adc JESD204B IP ILA Config Data1 Reg(0x804) =",  Peek(0x804), 16);
    Debug.Log("Adc JESD204B IP ILA Config Data2 Reg(0x808) =",  Peek(0x808), 16);
    Debug.Log("Adc JESD204B IP ILA Config Data3 Reg(0x80C) =",  Peek(0x80C), 16);
    Debug.Log("Adc JESD204B IP ILA Config Data4 Reg(0x810) =",  Peek(0x810), 16);
    Debug.Log("Adc JESD204B IP ILA Config Data5 Reg(0x814) =",  Peek(0x814), 16);
    Debug.Log("Adc JESD204B IP ILA Config Data6 Reg(0x818) =",  Peek(0x818), 16);
    Debug.Log("Adc JESD204B IP ILA Config Data7 Reg(0x81C) =",  Peek(0x81C), 16);
    Debug.Log("Adc JESD204B IP Link Error Count Reg(0x824) =",  Peek(0x824), 16);
    Debug.Log("Adc JESD204B IP RX Buffer Adjust Reg(0x830) =",  Peek(0x830), 16);
}

//===========================================================================
//  CLASS JESD_DACDevice -- JESD FPGA DAC interface
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class JESD_DACDevice
//---------------------------------------------------------------------------

JESD_DACDevice::JESD_DACDevice(IRequires_JESDDAC_Io * mdc, unsigned int device_id)
    :  MDC(mdc), DeviceID(device_id)
{
}

//---------------------------------------------------------------------------
//  destructor for class JESD_DACDevice
//---------------------------------------------------------------------------

JESD_DACDevice::~JESD_DACDevice()
{
}

//---------------------------------------------------------------------------
//  JESD_DACDevice::ConstructCustomParts()
//---------------------------------------------------------------------------

void  JESD_DACDevice::ConstructCustomParts()
{
    if (!DeviceEnabled())
        return;
}

//---------------------------------------------------------------------------
//  JESD_DACDevice::OpenHardware()
//---------------------------------------------------------------------------

void  JESD_DACDevice::OpenHardware()
{
    if (!DeviceEnabled())
        return;

    IpVersion = Peek(0);
}

//---------------------------------------------------------------------------
//  JESD_DACDevice::PreconfigureHardware()
//---------------------------------------------------------------------------

void  JESD_DACDevice::PreconfigureHardware()
{
    if (!DeviceEnabled())
        return;
}

//---------------------------------------------------------------------------
//  JESD_DACDevice::ConfigureHardware()
//---------------------------------------------------------------------------

void  JESD_DACDevice::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;
}

//---------------------------------------------------------------------------
// JESD_DACDevice::Poke() -- write to JESD logic area
//---------------------------------------------------------------------------

void  JESD_DACDevice::Poke(unsigned int address, unsigned int data)
{
    MDC->PokeJESDDAC(DeviceID, address, data);
}

//---------------------------------------------------------------------------
// JESD_DACDevice::Peek() -- read from JESD logic area
//---------------------------------------------------------------------------

unsigned int  JESD_DACDevice::Peek(unsigned int address)
{
    return MDC->PeekJESDDAC(DeviceID, address);
}

//---------------------------------------------------------------------------
//  JESD_DACDevice::CalcLineRate() -- Calculate JESD204B line rate
//---------------------------------------------------------------------------

double JESD_DACDevice::CalcJesdLineRate(unsigned int NumConverters, unsigned int BytePerSample,
                                 double SamplingRate, unsigned int Decimation, unsigned int NumLanes)
{
    return ((NumConverters * BytePerSample * SamplingRate * 10.0)/(Decimation * NumLanes));
}

//---------------------------------------------------------------------------
//  JESD_DACDevice::DumpDebugStatus() -- useful status dump code
//---------------------------------------------------------------------------

void  JESD_DACDevice::DumpDebugStatus()
{
    Debug.Log("Dac JESD204B IP Version Reg(0x0) =", IpVersion,  16);
    Debug.Log("Dac JESD204B IP Sync Status Reg(0x038) =", Peek(0x38), 16);
}

}  // namespace Innovative
