// KernelPci_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2007 Innovative Integration Corporation.
//  All Rights Reserved.
//

#ifndef KernelPci_MbH
#define KernelPci_MbH

#include <stddef.h>
#include <vector>

namespace InnovativeKernel
{
//==============================================================================
//  CLASS PciId -- PCI ID Code
//==============================================================================

class PciId
{
public:
    //
    //  Member Data
    unsigned int  VendorId;
    unsigned int  DeviceId;
    //
    //  Member Functions
    PciId(unsigned int vendorId = 0, unsigned int deviceId = 0);
    PciId(const PciId & other);

    bool operator<(const PciId& other) const
    {
        if (VendorId != other.VendorId)
            return this->VendorId < other.VendorId;
        else
            return this->DeviceId < other.DeviceId;
    }


};

//==============================================================================
//  CLASS PciPosition -- PCI Position on the Bus
//==============================================================================

class PciPosition
{
public:
    //
    //  Member Data
    unsigned int  Bus;
    unsigned int  Device;
    unsigned int  Function;
    //
    //  Member Functions
    PciPosition(unsigned int bus = 0, unsigned int device = 0, unsigned int function = 0);
    PciPosition(const PciPosition & other);

};

//==============================================================================
//  CLASS PciBoardLocation -- PCI Scan Return Structure
//==============================================================================

class PciBoardLocation
{
public:
    //
    //  Member Data
    PciId         ID;
    PciPosition   Location;
    //
    //  Member Functions
    PciBoardLocation()
        {}
    PciBoardLocation(const PciId & board_id, const PciPosition & pos);
    PciBoardLocation(const PciBoardLocation & other);

};

//==============================================================================
//  STRUCT PCIConfigSpace -- Mapping of Configuration Space
//==============================================================================
//31                    16 15                     0
//   -------------------------------------------------
//   |       Device ID       |       Vendor ID       | 0x00
//   -------------------------------------------------
//   |       Status          |       Command         | 0x04
//   -------------------------------------------------
//   |            Class Code           |   Rev ID.   | 0x08
//   -------------------------------------------------
//   |    BIST   | Hdr_Type  | Lat_Timer |Cache_Ln_Sz| 0x0c
//   -------------------------------------------------
//   |                                               | 0x10
//   -------------------------------------------------
//   |                                               | 0x14
//   -------------------------------------------------
//   |                                               | 0x18
//   -------------------------------------------------
//   |                                               | 0x1c
//   -------------------------------------------------
//   |                                               | 0x20
//   -------------------------------------------------
//   |                                               | 0x24
//   -------------------------------------------------
//   |              Cardbus CIS Pointer**            | 0x28
//   -------------------------------------------------
//   |     Subsystem ID**    | Subsystem Vendor ID** | 0x2c
//   -------------------------------------------------
//   |       Expansion ROM base address              | 0x30
//   -------------------------------------------------
//   |               Reserved                        | 0x34
//   -------------------------------------------------
//   |               Reserved                        | 0x38
//   -------------------------------------------------
//   |  Max_Lat  | Min_Gnt   | Intr Pin. | Intr line | 0x3c
//   -------------------------------------------------

//==============================================================================
//  CLASS DeviceEntry --
//==============================================================================
class DeviceEntry
{
public:
    // Constructor
    DeviceEntry(int bus=-1, int device=-1, int function=-1);

	unsigned int VendorId() const;
	void VendorId(unsigned vid);

	unsigned int DeviceId() const;
	void DeviceId(unsigned int did);
	
	unsigned int SubvendorId() const;
	void SubvendorId(unsigned int svid);
	
	unsigned int SubsystemId() const;
	void SubsystemId(unsigned int ssid);

    void  PositionFrom( const PciPosition & slot);
    PciPosition     Slot;

private:
	unsigned int FVendorId;
    unsigned int FDeviceId;
    unsigned int FSubvendorId;
    unsigned int FSubsystemId;
};

//==============================================================================
//  CLASS  FreeAlignedMemoryPtr -- Pointer to chunk of aligned memory
//==============================================================================

class FreeAlignedMemoryPtr
{
public:
    FreeAlignedMemoryPtr(char * origin, size_t align_to);
    FreeAlignedMemoryPtr(size_t size_in_bytes, size_t align_to);
    FreeAlignedMemoryPtr(size_t align_to = 4096);
    ~FreeAlignedMemoryPtr();

    FreeAlignedMemoryPtr    New(size_t bytes);
    void                    Delete();
    bool                    IsNull() const;

    char *                  Origin() const;
    unsigned int *          Address() const;
    size_t                  AlignTo() const;
    void                    AlignTo(size_t align_to);

private:
    char *                  FOrigin;
    unsigned int *          FAddress;
    size_t                  FAlignTo;

    unsigned int *          AddressFromOrigin(char * origin);
};


//==============================================================================
//  STRUCT MemoryBlock --
//==============================================================================

enum MemoryType { mtMemory, mtPort };

struct MemoryBlock
{
    unsigned int *  Addr;        // Mapped Address of Block Base
    unsigned int *  DriverAddr;  // Linear Address of Block Base
    unsigned int *  PhysAddr;    // Physical Address of Block Base
    unsigned int    Size;        // Block Size in bytes
    unsigned int *  RawPhysAddr; // Bus Physical Address of Block Base
    unsigned int    Bar;         // Bar number of Block Base
    MemoryType      MemType;     // Port or Memory Mapped
    bool            WasMapped;   // Was Mapped or not
};

//==============================================================================
//  CLASS ScatterBlock -- Memory Block for Scatter/Gather Memory
//==============================================================================

const size_t SG_PageSize = 4096;

class ScatterBlock
{
public:

    ScatterBlock() :  Memory(SG_PageSize), PageIdx(0)
        {}

    size_t  Pages() const
        {  return PageList.size();  }
        
    void    Reset()  {  PageIdx = 0;  }
    size_t  NextPageAddr() { return PageList[PageIdx]; }
    void    Increment()  {  PageIdx++; if (PageIdx>=Pages()) PageIdx=0;  }

    FreeAlignedMemoryPtr  Memory;
    size_t                Size;        // Block Size in bytes
    std::vector<size_t>   PageList;    // Physical Pages of Buffer
    size_t                PageSize;    // Page Size in bytes
    size_t                PageIdx;
};


}  // end namespace InnovativeKernel
//---------------------------------------------------------------------------
#endif
