//
// Lmk04800_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2012 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <LMK04800_Mb.h>
#include <SystemSupport_Mb.h>
#include <VcxoControl_Mb.h>
#include <algorithm>
#include <limits>
#include <cmath>
#include <Magic_Mb.h>
#include <Trace_Mb.h>
#include <sstream>

using namespace std;

namespace Innovative
{

//=============================================================================
//  CLASS Lmk04800  --  Lmk04800 Pll Implementation
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class Lmk04800
//-----------------------------------------------------------------------------

const unsigned int  TotalReferences  = 3;    // one, in some modes
const unsigned int  TotalOutputs     = 12;
const int           SpiRegisterCount = 31;   // last register just for doing reads

const double        MaxOutputFrequency = 1536.0e6;   // pg 55, analog delay must be 0 above this...

//  External Vco Parameters

//  Internal Vco Parameters
const unsigned int  TotalVcoTypes     = 4;

static struct VcoRangesType {
    double Low;
    double High;
} Lmk04800_VcoRanges[] = {
    {1840.0e6, 2030.0e6}, // LMK04803 default values
    {2148.0e6, 2370.0e6}, // LMK04805 default values
    {2370.0e6, 2600.0e6}, // LMK04806 default values  ** Default Vco **
    {2750.0e6, 3072.0e6}  // LMK04808 default values

};

const double        kPDF1_lower_limit = 1.0e6;    // might go as low as 500.0e3
const double        kPDF1_upper_limit = 40.0e6;   // max PFD

const double        kPDF2_upper_limit = 155.0e6;   // max PFD


const unsigned int  kR1_Min   =  1;
const unsigned int  kR1_Max   =  16383;
const unsigned int  kN1_Min   =  1;
const unsigned int  kN1_Max   =  4095;

const unsigned int  kR2_Min   =  1;
const unsigned int  kR2_Max   =  4095;
const unsigned int  kN2_Min   =  1;
const unsigned int  kN2_Max   =  262143;

const unsigned int  kOD_Max   =  1045;

//------ old constants
const double        kVco_InternalRange_Min       =  1750.0e6;
const double        kVco_InternalRange_Crossover =  2043.0e6;  // below is low scale, above high scale
const double        kVco_InternalRange_Max       =  2356.0e6;

const unsigned int  kVco_Prescaler_Min   =  2;
const unsigned int  kVco_Prescaler_Max   =  5;
const float         DefaultMaxOutFreq = 1500.0e6;       // do these still apply?
const double        MaxDividerAbove1Freq = 1175.0e6;    // do these still apply?


#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif
//-----------------------------------------------------------------------------
//  constructor for class Lmk04800
//-----------------------------------------------------------------------------

Lmk04800::Lmk04800(IAddressingSpace & space, Lmk04800Registers & regs, OperatingModes pllMode)
    : PllTimebaseDevice(TotalReferences, TotalOutputs),
      FChipType(lt4806), FReferenceSource(refPrimary),
      FClkMode(clockPll),
      FDefaultOpMode(pllMode),
      FFbMuxSource(fbm_Clkout0),
      RegisterImage(SpiRegisterCount),
      LocalVcxo(*this),
      ControlReg(space, regs.Control_Addr),
      SpiReg(space, regs.SpiIntf_Addr),
      FPowered(false),
      FVco2Range_Min(0), FVco2Range_Max(0),
      Debug(true)
{
    const unsigned int  DefaultOutputCfgBits = ocDisabled;
    //  Fill max frequency checker array
    for (unsigned int i=0; i<TotalOutputs; i++)
        {
        OutputFrequencyLimit.push_back(DefaultMaxOutFreq);
        OutputConfigField.push_back(DefaultOutputCfgBits);
        }

    SetVco2Range(FChipType);

    InitDividerSupport();
}

#ifdef _MSC_VER
	#pragma warning(pop)
#endif

//-----------------------------------------------------------------------------
//  destructor for class Lmk04800
//-----------------------------------------------------------------------------

Lmk04800::~Lmk04800()
{
}

//-----------------------------------------------------------------------------
//  Lmk04800::Init()  --  Initialize Chip
//-----------------------------------------------------------------------------

void Lmk04800::Init()
{
    FFrontEndPhaseDetectorFrequency = DefaultFrontEndPhaseDetectorFrequency();
    FFrontEndVcxoFrequency = DefaultFrontEndVcxoFrequency();
    FPhaseDetectorFrequency = DefaultPhaseDetectorFrequency();

    SetReset(true);
    SetReset(false);

    ReadDataRegisters();   // read initial state from chip
    SettingsInit();        // common INIT settings
    OnInit();
    Update();    // Update state of the device based on sources
    WriteDataRegisters();
}

//-----------------------------------------------------------------------------
//  Lmk04800::SettingsInit()  --  Initial Settings Loading
//-----------------------------------------------------------------------------
//   Note board specific implementations can override these in OnInit

void Lmk04800::SettingsInit()
{
    SetHoldoverEnable(false);    // no holdover mode
    SetHoldoverMux(hmxUWireRd);  // needed for register read.
    SetOutputGroupSource(3, ogsVco);
    SetOutputGroupSource(4, ogsVco);

    // Power on all output chains
    for (unsigned int i=0; i<TotalOutputs/2; i++)
        {
        PowerDownOutputGroup(i, false);
        SetOutputGroupDelay(i, 5);
        }

    // SYNC Setup
    SetSyncEnable(true);
    SetSyncQual(false);          SetSyncAuto(false);
    SetSyncMux(smxLogicLow);     SetSyncType(stpIn);
    SetSyncPolarityInverted(false); // 1 = SYNC

    SetNoSyncCLKout(0, false);  // Make all outputs SYNC
    SetNoSyncCLKout(1, false);
    SetNoSyncCLKout(2, false);
    SetNoSyncCLKout(3, false);
    SetNoSyncCLKout(4, false);
    SetNoSyncCLKout(5, false);

    SetPll2WindowSize();

    SetFeedbackMuxEnable(true);
    SetFeedbackMux(fbm_Clkout0);


}

//-----------------------------------------------------------------------------
//  Lmk04800::Reapply()  --  Reapply settings to hardware from saved state
//-----------------------------------------------------------------------------

void Lmk04800::Reapply()
{
    OnReapply();

    Update();    // Update state of the device based on sources

    Sync(true);
    //SetReset(true);
    WriteDataRegisters();  // first write sets reset false...
    Sync(false);
}


//-----------------------------------------------------------------------------
//  Lmk04800::Rewrite()  --  Rewrite settings without calculations
//-----------------------------------------------------------------------------
//  For use with PortImage() especially...

void Lmk04800::Rewrite()
{
    Sync(true);
    WriteDataRegisters();
    Sync(false);
}

//-----------------------------------------------------------------------------
//  Lmk04800::OutputFrequency()  --  Set Output Frequency
//-----------------------------------------------------------------------------
//  The LMK is configured by pairs only.  So set both halves to the same thing

void  Lmk04800::OutputFrequency(unsigned int output, double freq)
{
    int out_low = (output>>1)<<1;

    PllTimebaseDevice::OutputFrequency(out_low,   freq);
    PllTimebaseDevice::OutputFrequency(out_low+1, freq);
}

//-----------------------------------------------------------------------------
//  Lmk04800::Powered()  --  Change Powered state of chip
//-----------------------------------------------------------------------------

void Lmk04800::Powered(bool value)
{
    if (value==FPowered)
        return;

    FPowered = value;
    OnSetPower(FPowered);
    if (FPowered)
        {
        Init();       // Restore complete chip state
        Reapply();
        }
}

//-----------------------------------------------------------------------------
//  Lmk04800::Sync()  --  Sync Output Clocks
//-----------------------------------------------------------------------------

void Lmk04800::Sync(bool state)
{
    ControlReg.Sync.Value(state);
}

//-----------------------------------------------------------------------------
//  Lmk04800::ReferenceSource()  --  Change Clock Reference
//-----------------------------------------------------------------------------

void Lmk04800::ReferenceSource(LmkReferenceSource value)
{
    FReferenceSource = value;
    SetReferenceSource(value);
}

//-----------------------------------------------------------------------------
//  Lmk04800::MaxOutputDivider()  --  Return max output post-divider
//-----------------------------------------------------------------------------

int Lmk04800::MaxOutputDivider() const
{
    return kOD_Max;
}

//-----------------------------------------------------------------------------
//  Lmk04800::FrequencyModeUpdate()  --  Update using Frequency as baseline
//-----------------------------------------------------------------------------

void Lmk04800::FrequencyModeUpdate()
{
    Debug.Log("Lmk04800::FrequencyModeUpdate()");
    if (Reference() == 0 || Frequency() == 0)
        return;

    unsigned int R2, N2, N2_Pre;
    unsigned int N1;
    unsigned int CLKin0_PreR1, CLKin1_PreR1, R1;

    double OscIn = FrontEndVcxoFrequency();
    Debug.Log("PLL2 Reference = ", OscIn);
    //
    //  1)  Calculate PLL2 parameters
    unsigned int OD;
    double Vco2Freq;
    double PDF_Actual = PhaseDetectorFrequency();
    CalculateBackEndPLL(OscIn, Vco2Freq, OD, R2, N2, N2_Pre, PDF_Actual);
    //
    //   Phase Detector Freq = Reference/R1 = Vco/N1
    double PDF = PhaseDetectorFrequency();
    Debug.Log("Target PDF2 = ", PDF);
    Debug.Log("Actual PDF2 = ", PDF_Actual);
    //
    //   Recalculate everything (in case changes were made)
    Vco2Freq = N2_Pre * N2 * OscIn / R2;
    Debug.Log("Mid Vco2Freq = ", Vco2Freq);
    //
    //  Set up the output dividers properly
    SetOutputDividers(Vco2Freq, OD);
    //
    //  2) Calculate Parameters for Front End PLL
    //       We do this last since N frequency can depend on PLL2
    AutoselectLoopbackOutput();  // need to set LB to minimal value
    Debug.Log("N Input Freq = ", Modal_FrontEnd_N_InputFrequency());
    CalculateFrontEndPLL(CLKin0_PreR1, CLKin1_PreR1, R1, N1);

    double Pfd1_Actual_R = Reference()/R1;
    double Vco1Freq = Pfd1_Actual_R * N1;
    double Pfd1_Actual_N = Modal_FrontEnd_N_InputFrequency()/N1;

    Debug.Log("Mid Vco1Freq = ", Vco1Freq);
    Debug.Log("N Input Freq = ", Modal_FrontEnd_N_InputFrequency());
    Debug.Log("Vco1 = ", FrontEndVcxoFrequency());
    Debug.Log("Pfd1 (R Leg) = ", Pfd1_Actual_R);
    Debug.Log("Pfd1 (N Leg) = ", Pfd1_Actual_N);

    //  Allow User to Override Settings
    Debug.Log("Lmk04800::CLKin0_PreR1 = ", CLKin0_PreR1);
    Debug.Log("Lmk04800::CLKin0_PreR1 = ", CLKin1_PreR1);
    Debug.Log("Lmk04800::R1 = ", R1);
    Debug.Log("Lmk04800::N1 = ", N1);
    Debug.Log("Lmk04800::R2 = ", R2);
    Debug.Log("Lmk04800::N2 = ", N2);
    Debug.Log("Lmk04800::N2_Pre = ", N2_Pre);
    Debug.Log("Lmk04800::Output Divider = ", OD);
    Debug.Log("Vco2Freq (medium) = ", Vco2Freq);

    Debug.Log("Pll Tuning Event");
    Lmk04800PllTuningEvent e(CLKin0_PreR1, CLKin0_PreR1, R1, N1, R2, N2, N2_Pre, OD);
    OnPllTuning.Execute(e);

    Debug.Log("Final Values");
    Debug.Log("Lmk04800::CLKin0_PreR1 = ", CLKin0_PreR1);
    Debug.Log("Lmk04800::CLKin0_PreR1 = ", CLKin1_PreR1);
    Debug.Log("Lmk04800::R1 = ", R1);
    Debug.Log("Lmk04800::N1 = ", N1);
    Debug.Log("Lmk04800::R2 = ", R2);
    Debug.Log("Lmk04800::N2 = ", N2);
    Debug.Log("Lmk04800::N2_Pre = ", N2_Pre);
    Debug.Log("Lmk04800::Output Divider = ", OD);
    //
    //   Recalculate everything (in case changes were made)

    Pfd1_Actual_R = Reference()/R1;
    Vco1Freq = Pfd1_Actual_R * N1;

    Vco2Freq = N2_Pre * N2 * OscIn / R2;

    Debug.Log("Final Vco1Freq = ", Vco1Freq);
    Debug.Log("Final Vco2Freq = ", Vco2Freq);
    //
    //  Update registers & allied hardware
    LocalVcxo.SetVcoFrequency(Vco2Freq);    // reprogram Vcxo

    //  PLL2 Registers
    SetCLKin0_PreR(CLKin0_PreR1);
    SetCLKin1_PreR(CLKin1_PreR1);
    SetR1(R1);
    SetN1(N1);
    SetR2(R2);
    SetN2(N2);
    SetN2_Cal(N2);
    SetN2_Pre(N2_Pre);

    SetVcoMuxEnable(false);           // never use these

    SetPll2RefDoublerEnable(false);
    ConfigurePLL2(OscIn/1.0e6, PhaseDetectorFrequency()/1.0e6);
    //
    //  Set up the output dividers properly
    SetOutputDividers(Vco2Freq, OD);

}

//-----------------------------------------------------------------------------
//  Lmk04800::CalculateBackEndPLL() -- Calculate Parameters for PLL2
//-----------------------------------------------------------------------------

void  Lmk04800::CalculateBackEndPLL(double OscIn, double & Vco2Rate, unsigned int & OD, unsigned int & R2,
                                    unsigned int & N2, unsigned int & N2_Pre,
                                    double & PdfActual)
{
    Debug.Log("Lmk04800::CalculateBackEndPLL()");
    Debug.Log("Frequency()=", Frequency());

    bool  TuneMatchFound(false);
    for (unsigned int rep = 0; rep<kOD_Max; rep++)
        {
        unsigned int CandidateOD  = kOD_Max-rep;   // try ODs in reverse order...
        double       CandidateVco = Frequency()*CandidateOD;

        if (CandidateVco<=LocalVcxo.Max() && CandidateVco>=LocalVcxo.Min())
            {
            //  We have a candidate Divider. Check if we can reset to this freq.
            //     (some Vco's have gaps in the range).
            if (LocalVcxo.CanTuneTo(CandidateOD*Frequency()))
                {
                TuneMatchFound = true;
                Vco2Rate = CandidateVco;
                OD = CandidateOD;
                break;
                }
            }
        }
    //
    //  Determine Vco Frequency If the scan failed
    if (!TuneMatchFound)
        {
        //  We missed finding a VCO Match. Pick a VCO freq we can use
        Vco2Rate = LocalVcxo.BestTunableFrequency(Frequency());  // assume OD=1 ?
        //  have to figure a reasonable FB value too.
        OD =  static_cast<int>(Vco2Rate / Frequency());
        if (OD==0)
            OD=1;
        }
    Debug.Log("Vco2Rate=", Vco2Rate);
    //
    //  We want to adjust the PFD to match the VCO more exactly
    unsigned int TestR = static_cast<unsigned int>(OscIn/PdfActual);
    while (TestR < kR2_Max)
        {
        Debug.Log("TestR = ", TestR);

        unsigned int Div;
        Div = DivideToBestInt(Vco2Rate, PdfActual);
//        Div = DivideToBestInt(Vco2Rate/OD, PdfActual);
        Debug.Log("Div = ", Div);
        Debug.Log("PDF Actual = ", PdfActual);
        Debug.Log("Div*PdfActual = ", Div*PdfActual);

        if (Div*PdfActual == Vco2Rate)
        //if (Div*PdfActual == Vco2Rate/OD)
            {
            Debug.Log("Exact PDF Actual = ", PdfActual);
            Debug.Log("Exact Div*PdfActual = ", Div*PdfActual);
            break;
            }
        else
            {
            //  Not an exact fit, so reduce PDF
            TestR = static_cast<unsigned int>(OscIn/(PdfActual/10.0));
            if (TestR < kR2_Max)
                PdfActual /= 10.0;
            }
        }

    //
    //  Calculate Candidate N2, N2_Pre Value
    unsigned int TotalN2;
    TotalN2 = DivideToBestInt(Vco2Rate, PdfActual);

    SplitN2Factors(TotalN2, N2_Pre, N2);

//    if (FDefaultOpMode==omCascadedZeroDelay)
//        {
//        // in this mode, the actual N2 avoids PRE, is smaller by OD
//        N2 = N2*N2_Pre/OD;
//        }

    R2 = DivideToBestInt(OscIn, PdfActual);

    // check R2, N2?
}

//-----------------------------------------------------------------------------
//  Lmk04800::DivideToBestInt() -- Divide number, rounding fraction to nearest
//-----------------------------------------------------------------------------

unsigned int  Lmk04800::DivideToBestInt(double number, double divide_by)
{
    double div = 1.0;
    double fraction = modf(number/divide_by, &div);  // round to best int
    div += (fraction > 0.5) ? 1.0 : 0.0;
    return static_cast<unsigned int>( div );
}

//-----------------------------------------------------------------------------
//  Lmk04800::SplitN2Factors() -- Split Total N2 into N2, N2_Pre
//-----------------------------------------------------------------------------

void  Lmk04800::SplitN2Factors(unsigned int TotalN2, unsigned int &N2_Pre, unsigned int &N2)
{
    unsigned int CandidatePre = 2;
    for (ValidDividerArray::iterator div = N2_Pre_DividerValues.begin();
             div !=  N2_Pre_DividerValues.end();  ++div)
        {
        CandidatePre = *div;
        if (TotalN2%CandidatePre == 0)
            {
            N2_Pre = CandidatePre;
            N2 = TotalN2/N2_Pre;
            return;
            }
        }

    N2_Pre = 2;
    N2 = TotalN2/N2_Pre;
    return;
}

//-----------------------------------------------------------------------------
//  Lmk04800::CalculateFrontEndPLL() -- Calculate Parameters for FrontEnd PLL
//-----------------------------------------------------------------------------

void  Lmk04800::CalculateFrontEndPLL(unsigned int & CLKin0_PreR1,
                                     unsigned int & CLKin1_PreR1,
                                     unsigned int & R1,
                                     unsigned int & N1)
{
    Debug.Log("Lmk04800::CalculateFrontEndPLL()");

    //
    //   With the high PDFs we use, use minimal PreR
    CLKin0_PreR1 = ClkinX_PreR_DividerValues[0];  // 1
    CLKin1_PreR1 = ClkinX_PreR_DividerValues[0];  // 1
    //
    //   Phase Detector Freq = Reference/R1 = Vco/N1
    double RefRate = Reference();
    double N_Input_Rate = Modal_FrontEnd_N_InputFrequency();

    double PDF_high_limit = kPDF1_upper_limit;
    double PDF_low_limit = kPDF1_lower_limit;

    //
    //  We want to adjust the PFD to match the N1 Input
    //    But we can't go much below 1 MHz (~500K is ok) before we fail to lock
    double PDF = RefRate;
    Debug.Log("Start PDF1 = ", PDF);
    R1 = static_cast<unsigned int>( RefRate/PDF );
    for (unsigned int candidateR1 = 1; candidateR1<static_cast<unsigned int>(RefRate);  candidateR1++)
        {
        double PDF = RefRate/candidateR1;
        if (PDF >= PDF_high_limit)
            continue;
        if (PDF < PDF_low_limit)   // quit if we go too low
            {
            R1 = 1;   // failed
            break;
            }
        Debug.Log("Try R1 = ",  candidateR1);
        Debug.Log("Try PDF = ", PDF);
        //
        //  Check if this PFD evenly divides the N input rate
        unsigned int Div;
        Div = DivideToBestInt(N_Input_Rate, PDF);

        if (Div*PDF == N_Input_Rate)
            {
            R1 = candidateR1;
            Debug.Log("Exact PDF1 Actual = ", PDF);
            Debug.Log("Exact div*PDF1 = ", Div*PDF);
            break;
            }

        Debug.Log("div*PDF1 = ", Div*PDF);
        }

    // check R1
    if (R1==0)  R1=1;

    // Find effective PFD
    double Pfd1_Actual_R = Reference()/R1;

    N1 = static_cast<unsigned int>( N_Input_Rate/Pfd1_Actual_R );
    if (N1==0)  N1=1;

}

//-----------------------------------------------------------------------------
//  Lmk04800::AutoselectLoopbackOutput()  --  select output to sync PLL 1
//-----------------------------------------------------------------------------

void Lmk04800::AutoselectLoopbackOutput()
{
    //  Only do this in Dual Vco Zero Delay Mode
    if (FDefaultOpMode == DualIVcoZD)
        {
        unsigned int FeedbackIdx = 0;
        double FeedbackRate = 1e10;
        for (unsigned int output=0; output<=10;  output+=2)
            {
            // find minimum rate
            if (OutputInUse(output)==true)
                {
                double OutputRate = OutputFrequencyActual(output);
                if (OutputRate < FeedbackRate)
                    {
                    if (OutputRate==Adc().FrequencyActual() || OutputRate==Dac().FrequencyActual())
                        {
                        FeedbackIdx = output;
                        FeedbackRate = OutputRate;
                        }
                    }
                }
            }
        //
        //  Now set FB value from output with minimum rate
        switch (FeedbackIdx)
            {
            default:
            case 0:    FeedbackOutput(fbm_Clkout0) ;   break;
            case 2:    FeedbackOutput(fbm_Clkout2) ;   break;
            case 4:    FeedbackOutput(fbm_Clkout4) ;   break;
            case 6:    FeedbackOutput(fbm_Clkout6) ;   break;
            case 8:    FeedbackOutput(fbm_Clkout8) ;   break;
            case 10:   FeedbackOutput(fbm_Clkout10);   break;
            }
        }
    else
        {
        FeedbackOutput(fbm_FbClkIn);
        }

    // Apply to hardware
    SetFeedbackMux(FeedbackOutput());
}

//-----------------------------------------------------------------------------
//  Lmk04800::Modal_FrontEnd_N_InputFrequency()  --  N1 input by "operating mode"
//-----------------------------------------------------------------------------

double Lmk04800::Modal_FrontEnd_N_InputFrequency()
{
    switch (FDefaultOpMode)
        {
        default:
            Debug.Log("...unsupported operating mode");
        case DualIVco:
            return FrontEndVcxoFrequency();
//            break;

        case DualIVcoZD:
            switch (FeedbackOutput())
                {
                case fbm_Clkout0:    return OutputFrequencyActual(0);   //break;
                case fbm_Clkout2:    return OutputFrequencyActual(2);   //break;
                case fbm_Clkout4:    return OutputFrequencyActual(4);   //break;
                case fbm_Clkout6:    return OutputFrequencyActual(6);   //break;
                case fbm_Clkout8:    return OutputFrequencyActual(8);   //break;
                case fbm_Clkout10:   return OutputFrequencyActual(10);  //break;
                default:
                case fbm_FbClkIn:
                    return FrontEndVcxoFrequency();
//                    break;
                }
//            break;
        }
//    return FrontEndVcxoFrequency();   // should never get here...
}

//-----------------------------------------------------------------------------
//  Lmk04800::DividerModeUpdate() -- Update state of dividers
//-----------------------------------------------------------------------------
//  In this mode, the reference is sent to the outputs, with possible division

void  Lmk04800::DividerModeUpdate()
{
    Debug.Log("Lmk04800::DividerModeUpdate()");
    if (Reference() == 0 || Frequency() == 0)
        return;

    double div = 1.0;
    double fraction = modf(Reference()/Frequency(), &div);
    div += (fraction > 0.5) ? 1.0 : 0.0;
    unsigned int OutDiv = static_cast<unsigned int>(div);
    OutDiv = std::max<unsigned int>(OutDiv, 1);

    SetOutputDividers(Reference(), OutDiv);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetOutputDividers()  --
//-----------------------------------------------------------------------------

void  Lmk04800::SetOutputDividers(double VcoFreq, unsigned int OutDiv)
{
    Debug.Log("Lmk04800::SetOutputDividers()");
    Debug.Log("VcoFreq = ", VcoFreq);
    Debug.Log("OutDiv = ", OutDiv);

    unsigned int BestDivider = OutDiv;
    double       OutputFreq  = VcoFreq/BestDivider;


    for (unsigned int output=0; output<TotalOutputs; output++)
        {
        if (OutputInUse(output)==true)
            {
            float   CandidateDivider = static_cast<float>(BestDivider);
            double  CandidateFreq    = OutputFreq;
            //
            //  If the Output is the Master, use the best values.
            //     if not the master, and Frequency is unspecified, then
            //     the make the output a copy of Master.
            //     Otherwise, figure the best divider possible for each output
            if ((output!=FMasterOutput) && (OutputFrequency(output)!=0.0))
                {
                //Debug.Log("Calculating Optimal Frequency");
                CandidateFreq = OutputFrequency(output);
                CandidateDivider = static_cast<float>(VcoFreq / CandidateFreq);
                //Debug.Log("CandidateDivider = ", CandidateDivider);
                //Debug.Log("CandidateFreq = ", CandidateFreq);

                int above, below, divider;
                GetNearestDividers(output, CandidateDivider, below, above);

                // Find which divider is closest to desired frequency
                double centerFreq = VcoFreq/CandidateDivider;
                double hiFreq     = VcoFreq/below;
                double lowFreq    = VcoFreq/above;
                if (std::abs(centerFreq-lowFreq) > std::abs(hiFreq-centerFreq))
                    divider = below;
                else
                    divider = above;

                CandidateDivider = static_cast<float>(divider);
                }

            //
            //  Check for frequency exceeding applied Limits
            //Debug.Log("FC CandidateDivider = ", CandidateDivider);
            //Debug.Log("FC CandidateFreq = ", CandidateFreq);
            if (OutputFrequencyLimit[output]>=CandidateFreq)
                OutputDivider(output, static_cast<unsigned int>(CandidateDivider));
            else
                {
                //Debug.Log("Out of Frequency Limit Range");
                double MinDivideRatio = VcoFreq/OutputFrequencyLimit[output];

                double div = 1.0;
                double fraction = modf(MinDivideRatio, &div);
                div += (fraction > 0.0) ? 1.0 : 0.0;
                unsigned int OutDiv = static_cast<unsigned int>(div);
                OutDiv = std::max<unsigned int>(OutDiv, 1);

                OutputDivider(output, OutDiv);
                }
            //
            //  Calculate Actual Frequency
            if (OutputDividerArray[output]!=0)  // if illegal divider, skip it
                {
                OutputFreqActualArray[output] = VcoFreq / OutputDividerArray[output];
                }
            //
            //  If Divider 'needed' is not allowed because of rate limitations,
            //     disable the output, otherwise enable it
//            if (VcoFreq > MaxDividerAbove1Freq && OutputDivider(output)!=1)
//                {
//                SetDividerEnable(output, false);
//                OutputFreqActualArray[output] = 0.0;  // no output allowed
//                }
//            else
//                SetDividerEnable(output, true);
            }
        else
            {
            OutputFreqActualArray[output] = 0.0;
            }
        }

}

//-----------------------------------------------------------------------------
//  Lmk04800::Locked()  --  Return PLL lock status
//-----------------------------------------------------------------------------

bool  Lmk04800::Locked() const
{
    return ControlReg.Status_Ld.Value() ? true : false;
}

//-----------------------------------------------------------------------------
//  Lmk04800::Port()  --  Write to PLL data port
//-----------------------------------------------------------------------------

void Lmk04800::Port(unsigned char reg, unsigned int value)
{
    if (reg != 0x1f)
        {
        std::stringstream msg;
        msg << "Lmk04800::Write 0x" << std::hex << static_cast<int>(value)
            << " to reg " <<  static_cast<int>(reg);
        Debug.Log(msg.str());
        }

    unsigned int SpiValue(0);

    PtrBitManipulator SpiMap(&SpiValue);

    //   Overwrite register area with Reg
    SpiMap.Value(value);       // Spi Register Value (Bits 0-31)
    SpiMap.Field(0,  5, reg);  // Spi Register Addr  (Bits 0-5)

    Wait_UWireReady();
    SpiReg.Value(SpiValue);
    Sleep(1);
}

//-----------------------------------------------------------------------------
//  Lmk04800::Port()  --  Read from PLL data port
//-----------------------------------------------------------------------------

unsigned int Lmk04800::Port(unsigned char reg)
{
    unsigned int SpiValue(0);

    PtrBitManipulator SpiMap(&SpiValue);

    SpiMap.Field(16,  5, reg);     // Spi Register Addr  (Bits 16-20)

    //
    //  Send Read Command
    Wait_UWireReady();

    Port(0x1F, SpiValue);

    Wait_UWireDataAvailable();
    //
    //  Read data
    SpiValue = SpiReg.Value();

    std::stringstream msg;
    msg << "Lmk04800::Read 0x" << std::hex << SpiValue
        << " from reg " << static_cast<int>(reg);
    Debug.Log(msg.str());

    return SpiValue;
}

//-----------------------------------------------------------------------------
//  Lmk04800::PortImage()  --  Write to Object Image Data Cache
//-----------------------------------------------------------------------------

void  Lmk04800::PortImage(unsigned char reg, unsigned int value)
{
	if (reg < static_cast<unsigned char>(RegisterImage.size()))
        RegisterImage[reg] = value;
}

//-----------------------------------------------------------------------------
//  Lmk04800::PortImage()  --  Read from Object Image Data Cache
//-----------------------------------------------------------------------------

unsigned int  Lmk04800::PortImage(unsigned char reg)
{
	if (reg < static_cast<unsigned char>(RegisterImage.size()))
        return RegisterImage[reg];
    else
        return 0;
}

//---------------------------------------------------------------------------
//  Lmk04800::Wait_UWireReady()
//---------------------------------------------------------------------------

void Lmk04800::Wait_UWireReady()
{
	// Wait up to 5 uS for pending operation to complete
	for (unsigned int i = 0; i < 5; ++i)
		{
		if (ControlReg.UWire_Ready.Value())
			return;
		else
			Innovative::uSleep(1);
		}
	// Gave up
	std::stringstream msg;
	msg << std::hex << " Lmk04800 UWire Port Ready Failure (>5us)";
	Debug.Log(msg.str());
}

//---------------------------------------------------------------------------
//  Lmk04800::Wait_UWireDataAvailable()
//---------------------------------------------------------------------------

void Lmk04800::Wait_UWireDataAvailable()
{
	// Wait up to 5 uS for pending operation to complete
	for (unsigned int i = 0; i < 5; ++i)
		{
		if (ControlReg.UWire_DataValid.Value())
			return;
		else
			Innovative::uSleep(1);
		}
	// Gave up
	std::stringstream msg;
	msg << std::hex << " Lmk04800 UWire Port DataValid Failure (>5us)";
	Debug.Log(msg.str());
}


//-----------------------------------------------------------------------------
//  Lmk04800::SetDividerHdw()  --  Update the Divider setting hardware
//-----------------------------------------------------------------------------

void Lmk04800::SetDividerHdw(unsigned int output, unsigned int divide_by)
{
    Debug.Log("Lmk04800::SetOutputDividers()");
    Debug.Log("output = ", output);
    Debug.Log("divide_by = ", divide_by);

    if (!IsValidDivider(divide_by))
        return;

    unsigned int pair = output>>1;
    if (pair < 6)
        {
        PtrBitManipulator Reg(&RegisterImage[pair]);
        Reg.Field(5, 10, ToDividerInput(divide_by));
        }
    Debug.Log("...set register image");
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetOutEnabledHdw()  --  Update the Output Enabled hardware
//-----------------------------------------------------------------------------

void Lmk04800::SetOutEnabledHdw(unsigned int output, bool enabled)
{
    if (output < TotalOutputs)
        {
        if (enabled)
            {
            // restore from cached setting...
            SetOutputConfig(output, OutputConfigField[output] );
            //  the above also disables divider if off
            }
        else
            {
            bool should_cache_setting = false;
            SetOutputConfig(output, ocDisabled, should_cache_setting );
            //  the above also disables divider if off
            }
        }
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetReset()  --  reset the PLL - don't change the image!
//-----------------------------------------------------------------------------

void Lmk04800::SetReset(bool state)
{
    unsigned int R0_Data = Port(0x0);
    PtrBitManipulator Reg(&R0_Data);
    Reg.Bit(17, state);

    Port(0x0, R0_Data);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetOperatingMode()  --  Update the Op Mode for the PLL
//-----------------------------------------------------------------------------

void Lmk04800::SetOperatingMode(unsigned int om_setting)
{
    PtrBitManipulator Reg(&RegisterImage[11]);
    Reg.Field(27, 5, om_setting);
}

//-----------------------------------------------------------------------------
//  Lmk04800::PowerDownOutputGroup()  --  Power down/up output pair
//-----------------------------------------------------------------------------

void Lmk04800::PowerDownOutputGroup(unsigned int pair, bool state)
{
    if (pair>5)
        pair=0;
    PtrBitManipulator Reg(&RegisterImage[pair]);
    Reg.Bit(31, state);  // true = power down
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetOutputGroupSource()  --  Source of output pair
//-----------------------------------------------------------------------------

void Lmk04800::SetOutputGroupSource(unsigned int pair, unsigned int ogrp_source)
{
    if (pair!=3 && pair!=4)
        pair=3;
    PtrBitManipulator Reg(&RegisterImage[pair]);
	Reg.Bit(30, (ogrp_source)? true : false);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetOutputGroupDelayMode()  --  delay mode of pair
//-----------------------------------------------------------------------------

void Lmk04800::SetOutputGroupDelayMode(unsigned int pair, unsigned int ogrp_delaymode)
{
    if (pair>5)
        pair=0;
    PtrBitManipulator Reg(&RegisterImage[pair]);
    Reg.Field(28, 2, ogrp_delaymode);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetOutputGroupDelay()  --  delay value of pair
//-----------------------------------------------------------------------------

void Lmk04800::SetOutputGroupDelay(unsigned int pair, unsigned int delay)
{
    if (pair>5)
        pair=0;
    if (delay<5)
        delay=5;
    else if (delay>522)
        delay=522;

    PtrBitManipulator Reg(&RegisterImage[pair]);
    Reg.Field(18, 10, delay);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetOutputGroupDelay_HalfShift()  --  minus half-unit delay
//-----------------------------------------------------------------------------

void Lmk04800::SetOutputGroupDelay_HalfShift(unsigned int pair, bool hs)
{
    if (pair>5)
        pair=0;

    PtrBitManipulator Reg(&RegisterImage[pair]);
    Reg.Bit(16, hs);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetOutputConfig()  --  Update the Output Config for a channel
//-----------------------------------------------------------------------------

void Lmk04800::SetOutputConfig(unsigned int output, unsigned int cfg_setting,
                                bool should_cache_setting)
{
    if (output < TotalOutputs)
        {
        //  Cache restore-to setting
        if (should_cache_setting)
            OutputConfigField[output] = cfg_setting;
        // Figure out location of field
        unsigned int reg;
        unsigned int ofs;
        switch (output)
            {
            default:
            case  0:  reg=6;  ofs=16;  break;
            case  1:  reg=6;  ofs=20;  break;
            case  2:  reg=6;  ofs=24;  break;
            case  3:  reg=6;  ofs=28;  break;
            case  4:  reg=7;  ofs=16;  break;
            case  5:  reg=7;  ofs=20;  break;
            case  6:  reg=7;  ofs=24;  break;
            case  7:  reg=7;  ofs=28;  break;
            case  8:  reg=8;  ofs=16;  break;
            case  9:  reg=8;  ofs=20;  break;
            case 10:  reg=8;  ofs=24;  break;
            case 11:  reg=8;  ofs=28;  break;
            }

        PtrBitManipulator Reg(&RegisterImage[reg]);
        Reg.Field(ofs, 4, cfg_setting);
        }
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetOutputOsc0Config()  --  Configure Osc0 output config
//-----------------------------------------------------------------------------

void Lmk04800::SetOutputOsc0Config(unsigned int cfg_setting)
{
    PtrBitManipulator Reg(&RegisterImage[10]);
    Reg.Field(24, 4, cfg_setting);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetOutputOsc1Config()  --  Configure Osc1 Lvpecl level
//-----------------------------------------------------------------------------

void Lmk04800::SetOutputOsc1Config(unsigned int cfg_setting)
{
    PtrBitManipulator Reg(&RegisterImage[10]);
    Reg.Field(30, 2, cfg_setting);
}

//-----------------------------------------------------------------------------
//  Lmk04800::EnableOutputOsc0()  --  enable Osc0 output
//-----------------------------------------------------------------------------

void Lmk04800::EnableOutputOsc0(bool state)
{
    PtrBitManipulator Reg(&RegisterImage[10]);
    Reg.Bit(22, state);
}

//-----------------------------------------------------------------------------
//  Lmk04800::EnableOutputOsc1()  --  enable Osc1 output
//-----------------------------------------------------------------------------

void Lmk04800::EnableOutputOsc1(bool state)
{
    PtrBitManipulator Reg(&RegisterImage[10]);
    Reg.Bit(23, state);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetInputConfig()  --  Update the Input Config
//-----------------------------------------------------------------------------

void Lmk04800::SetInputConfig(unsigned int ref_idx, unsigned int cfg_setting)
{
    PtrBitManipulator Reg(&RegisterImage[14]);
    unsigned int ofs = (ref_idx==0) ? 20 : 21 ;
	Reg.Bit(ofs, (cfg_setting) ? true : false );
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetInputPolarity()  --  Update the Input Polarity CLKin_Sel_INV
//-----------------------------------------------------------------------------

void Lmk04800::SetInputPolarity(unsigned int cfg_setting)
{
    PtrBitManipulator Reg(&RegisterImage[13]);
	Reg.Bit(8, (cfg_setting) ? true : false);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetLockType()  --  Set holdover mux bit source
//-----------------------------------------------------------------------------

void Lmk04800::SetLockType(unsigned int setting)
{
    PtrBitManipulator Reg(&RegisterImage[12]);  // LD_MUX
    Reg.Field(27, 5, setting);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetHoldoverEnable() - enable holdover
//-----------------------------------------------------------------------------

void Lmk04800::SetHoldoverEnable(bool enabled)
{
    PtrBitManipulator Reg(&RegisterImage[12]);   //HOLDOVER_MODE
    unsigned int mode = (enabled) ? 0x2 : 0x1 ;
    Reg.Field(6, 2, mode);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetHoldoverMux()  --  Set holdover mux bit source
//-----------------------------------------------------------------------------

void Lmk04800::SetHoldoverMux(unsigned int setting)
{
    PtrBitManipulator Reg(&RegisterImage[13]);
    Reg.Field(27, 5, setting);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetVcoMuxEnable()  --  Enable Vco Mux
//-----------------------------------------------------------------------------

void Lmk04800::SetVcoMuxEnable(bool enabled)
{
    PtrBitManipulator Reg(&RegisterImage[10]);
    Reg.Bit(12, enabled);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetFeedbackMuxEnable()  --  Enable Feedback Mux
//-----------------------------------------------------------------------------

void Lmk04800::SetFeedbackMuxEnable(bool enabled)
{
    PtrBitManipulator Reg(&RegisterImage[10]);   // EN_FEEDBACK_MUX
    Reg.Bit(11, enabled);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetFeedbackMux()  --  Enable Vco Mux
//-----------------------------------------------------------------------------

void Lmk04800::SetFeedbackMux(unsigned int mux_state)
{
    PtrBitManipulator Reg(&RegisterImage[10]);   // FEEDBACK_MUX
    Reg.Field(5, 3, mux_state);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetPll2RefDoublerEnable()  --  Enable Reference Doubler
//-----------------------------------------------------------------------------

void Lmk04800::SetPll2RefDoublerEnable(bool enabled)
{
    PtrBitManipulator Reg(&RegisterImage[26]);
    Reg.Bit(29, enabled);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetSyncEnable()  --  Enable SYNC
//-----------------------------------------------------------------------------

void Lmk04800::SetSyncEnable(bool enabled)
{
    PtrBitManipulator Reg(&RegisterImage[11]);    // EN_SYNC
    Reg.Bit(26, enabled);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetNoSyncCLKout()  --  Disable SYNC on output pair
//-----------------------------------------------------------------------------

void Lmk04800::SetNoSyncCLKout(unsigned int pair, bool state)
{
    PtrBitManipulator Reg(&RegisterImage[11]);    // NO_SYNC_CLKoutX_Y
    Reg.Bit(20+pair, state);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetSyncMux()  --  Configure SYNC output pin
//-----------------------------------------------------------------------------

void Lmk04800::SetSyncMux(unsigned int cfg_setting)
{
    PtrBitManipulator Reg(&RegisterImage[11]);    // SYNC_MUX
    Reg.Field(18, 2, cfg_setting);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetSyncQual()  --  Enable SYNC QUAL mode
//-----------------------------------------------------------------------------

void Lmk04800::SetSyncQual(bool enabled)
{
    PtrBitManipulator Reg(&RegisterImage[11]);    // SYNC_QUAL
    Reg.Bit(17, enabled);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetSyncPolarityInverted()  --  Enable SYNC polarity inversion
//-----------------------------------------------------------------------------

void Lmk04800::SetSyncPolarityInverted(bool inverted)
{
    PtrBitManipulator Reg(&RegisterImage[11]);    // SYNC_POL_INV
    Reg.Bit(16, inverted);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetSyncAuto()  --  Enable SYNC auto mode
//-----------------------------------------------------------------------------

void Lmk04800::SetSyncAuto(bool auto_sync)
{
    PtrBitManipulator Reg(&RegisterImage[11]);    // SYNC_EN_AUTO
    Reg.Bit(15, auto_sync);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetSyncType()  --  Configure SYNC pin type
//-----------------------------------------------------------------------------

void Lmk04800::SetSyncType(unsigned int cfg_setting)
{
    PtrBitManipulator Reg(&RegisterImage[11]);    // SYNC_TYPE
    Reg.Field(12, 3, cfg_setting);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetPll2WindowSize()  --  Set Window Size
//-----------------------------------------------------------------------------

void Lmk04800::SetPll2WindowSize( )
{
    PtrBitManipulator Reg(&RegisterImage[26]);
    Reg.Field(30, 2, 0x2);                      // PLL2_WND_SIZE

}

//-----------------------------------------------------------------------------
//  Lmk04800::SetReferenceSource()  --  Set the Reference Source
//-----------------------------------------------------------------------------

void Lmk04800::SetReferenceSource(LmkReferenceSource source)
{
    Debug.Log("Lmk04800::SetReferenceSource");
    Debug.Log("...source=", source);
    // CLKin_SELECT_MODE
    unsigned int source_no;
    switch (source)
        {
        default:
        //case refCLKin0:
        case refPrimary:    source_no = 0;  break;

        //case refCLKin1:
        case refSecondary:  source_no = 1;  break;

        case refAuto:
        case refAutoNext:
        case refPinSelect:  source_no = 0;  break;
        }

    Debug.Log("...source_no=", source_no);
    PtrBitManipulator Reg(&RegisterImage[13]);
    Reg.Field(9, 3, source_no);

    SetSourceSelection(source_no);   // inform base class too
    //SetOutputSource(FOutputSource);  // make sure Output Mux is set correctly
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetOutputSource()  --  Sert the Output Mode (and Mux outputs)
//-----------------------------------------------------------------------------

void Lmk04800::SetOutputSource(LmkClockMode source)
{
    //
    //  decide on which internal clock soure to direct
    if (source==clockPll)
        {
        SetOperatingMode(FDefaultOpMode);
        }
    else
        {
        // External Mode (use reference 1 always?)
        SetOperatingMode(ClkDistro);
        }

}

//-----------------------------------------------------------------------------
//  Lmk04800::SetVco2Range()  --  Load VCO Range based off of Chip Type
//-----------------------------------------------------------------------------

void  Lmk04800::SetVco2Range(LmkType chip_type)
{
    unsigned int ct = chip_type;
    if (ct < TotalVcoTypes)
        {
        FVco2Range_Min = Lmk04800_VcoRanges[ct].Low;
        FVco2Range_Max = Lmk04800_VcoRanges[ct].High;
        }
}

//-----------------------------------------------------------------------------
//  Lmk04800::GetDivider()  --  read back the Divider setting
//-----------------------------------------------------------------------------

unsigned int  Lmk04800::GetDivider(unsigned int output)
{
    unsigned int divider(1);
    unsigned int pair(output>>1);
    if (pair < 6)
        {
        PtrBitManipulator Reg(&RegisterImage[pair]);
        unsigned int div_field = Reg.Field(5, 10);
        divider = FromDividerInput(div_field);
        }
    return divider;
}

//-----------------------------------------------------------------------------
//  Lmk04800::GetDividerEnable()  --  Read the Divider Enable setting
//-----------------------------------------------------------------------------

//bool Lmk04800::GetDividerEnable(unsigned int output)
//{
//    bool bit = false;
//    if (output < TotalOutputs)
//        {
//        PtrBitManipulator Reg(&RegisterImage[output]);
//        bit = Reg.Bit(24);     // Output Cfg bits 24
//        }
//    return bit;
//}

//-----------------------------------------------------------------------------
//  Lmk04800::ConfigurePLL2()  --  Set Rate Dependent Config of PLL2
//-----------------------------------------------------------------------------

void Lmk04800::ConfigurePLL2(double OscInRateMHz, double Pll2_PdfMHz)
{
    bool Pll2_Fast_Pdf = (Pll2_PdfMHz <= 100.0) ? false : true ;
    PtrBitManipulator Reg(&RegisterImage[29]);

    Reg.Bit(23, Pll2_Fast_Pdf);  // Set PLL2_FAST_PDF

    unsigned int OSCin_FREQ_code;
    if (OscInRateMHz <= 63.0)
        OSCin_FREQ_code = 0x0;
    else if (OscInRateMHz <= 127.0)
        OSCin_FREQ_code = 0x1;
    else if (OscInRateMHz <= 255.0)
        OSCin_FREQ_code = 0x2;
    else if (OscInRateMHz <= 400.0)
        OSCin_FREQ_code = 0x4;
    else
        OSCin_FREQ_code = 0x4;

    Reg.Field(24, 3, OSCin_FREQ_code);  // Set OSCin_FREQ
    // So what if OscIn Rate > 400.0? Is this illegal
    //  Default is frickin SEVEN?
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetCLKin0_PreR()  --  Configure CLKin0_PreR  (Input PreDivider)
//-----------------------------------------------------------------------------
//  Intent is to get 2 references balanced going into rest of system

void Lmk04800::SetCLKin0_PreR(unsigned int PreR)
{
    PtrBitManipulator Reg(&RegisterImage[27]);
    Reg.Field(20, 2, To_CLKinX_PreR_DividerInput(PreR));
}

//-----------------------------------------------------------------------------
//  Lmk04800::GetCLKin0_PreR()  --  Read CLKin0_PreR  (Input PreDivider)
//-----------------------------------------------------------------------------

unsigned int Lmk04800::GetCLKin0_PreR()
{
    PtrBitManipulator Reg(&RegisterImage[27]);
    unsigned int R_field = Reg.Field(20, 2);
    return From_CLKinX_PreR_DividerInput(R_field);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetCLKin1_PreR()  --  Configure CLKin1_PreR  (Input PreDivider)
//-----------------------------------------------------------------------------
//  Intent is to get 2 references balanced going into rest of system

void Lmk04800::SetCLKin1_PreR(unsigned int PreR)
{
    PtrBitManipulator Reg(&RegisterImage[27]);
    Reg.Field(22, 2, To_CLKinX_PreR_DividerInput(PreR));
}

//-----------------------------------------------------------------------------
//  Lmk04800::GetCLKin1_PreR()  --  Read CLKin1_PreR  (Input PreDivider)
//-----------------------------------------------------------------------------

unsigned int Lmk04800::GetCLKin1_PreR()
{
    PtrBitManipulator Reg(&RegisterImage[27]);
    unsigned int R_field = Reg.Field(22, 2);
    return From_CLKinX_PreR_DividerInput(R_field);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetR1()  --  Configure R1  (PLL1 Reference Divider)
//-----------------------------------------------------------------------------

void Lmk04800::SetR1(unsigned int R1)
{
    PtrBitManipulator Reg(&RegisterImage[27]);

    Reg.Field(6, 14, R1);
}

//-----------------------------------------------------------------------------
//  Lmk04800::GetR1()  --  Read R1  (PLL1 Reference Divider)
//-----------------------------------------------------------------------------

unsigned int Lmk04800::GetR1()
{
    PtrBitManipulator Reg(&RegisterImage[27]);

    unsigned int R1 = Reg.Field(6, 14);
    return R1;
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetN1()  --  Configure N1  (PLL1 Feedback Divider)
//-----------------------------------------------------------------------------

void Lmk04800::SetN1(unsigned int N1)
{
    PtrBitManipulator Reg(&RegisterImage[28]);

    Reg.Field(6, 14, N1);
}

//-----------------------------------------------------------------------------
//  Lmk04800::GetN1()  --  Read N1  (PLL1 Feedback Divider)
//-----------------------------------------------------------------------------

unsigned int Lmk04800::GetN1()
{
    PtrBitManipulator Reg(&RegisterImage[28]);

    unsigned int N1 = Reg.Field(6, 14);
    return N1;
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetR2()  --  Configure R2  (PLL2 Reference Divider)
//-----------------------------------------------------------------------------

void Lmk04800::SetR2(unsigned int R2)
{
    PtrBitManipulator Reg(&RegisterImage[28]);

    Reg.Field(20, 12, R2);
}

//-----------------------------------------------------------------------------
//  Lmk04800::GetR2()  --  Read R2  (PLL2 Reference Divider)
//-----------------------------------------------------------------------------

unsigned int Lmk04800::GetR2()
{
    PtrBitManipulator Reg(&RegisterImage[28]);

    unsigned int R2 = Reg.Field(20, 12);
    return R2;
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetN2()  --  Configure N2  (PLL2 Feedback Divider)
//-----------------------------------------------------------------------------

void Lmk04800::SetN2(unsigned int N2)
{
    PtrBitManipulator Reg(&RegisterImage[30]);

    Reg.Field(5, 18, N2);
}

//-----------------------------------------------------------------------------
//  Lmk04800::GetN2()  --  Read N2  (PLL2 Feedback Divider)
//-----------------------------------------------------------------------------

unsigned int Lmk04800::GetN2()
{
    PtrBitManipulator Reg(&RegisterImage[30]);

    unsigned int N2 = Reg.Field(5, 18);
    return N2;
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetN2_Cal()  --  Configure SetN2_Cal  (PLL2 Feedback Divider)
//-----------------------------------------------------------------------------

void Lmk04800::SetN2_Cal(unsigned int N2_Cal)
{
    PtrBitManipulator Reg(&RegisterImage[29]);

    Reg.Field(5, 18, N2_Cal);
}

//-----------------------------------------------------------------------------
//  Lmk04800::GetN2_Cal()  --  Read SetN2_Cal  (PLL2 Feedback Divider)
//-----------------------------------------------------------------------------

unsigned int Lmk04800::GetN2_Cal()
{
    PtrBitManipulator Reg(&RegisterImage[29]);

    unsigned int N2_Cal = Reg.Field(5, 18);
    return N2_Cal;
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetN2_Pre()  --  Configure PLL2 N Prescaler (Feedback PreDivider)
//-----------------------------------------------------------------------------

void Lmk04800::SetN2_Pre(unsigned int N2_Pre)
{
    PtrBitManipulator Reg(&RegisterImage[30]);
    Reg.Field(24, 3, To_N2_Pre_DividerInput(N2_Pre));
}

//-----------------------------------------------------------------------------
//  Lmk04800::GetN2_Pre()  --  Read PLL2 N Prescaler (Feedback PreDivider)
//-----------------------------------------------------------------------------

unsigned int Lmk04800::GetN2_Pre()
{
    PtrBitManipulator Reg(&RegisterImage[30]);
    unsigned int N2_Pre_field = Reg.Field(24, 3);
    return From_N2_Pre_DividerInput(N2_Pre_field);
}

//-----------------------------------------------------------------------------
//  Lmk04800::SetVcoDiv()  --  Configure PLL2 (Vco Postdivider)
//-----------------------------------------------------------------------------

void Lmk04800::SetVcoDiv(unsigned int VcoMux)
{
    PtrBitManipulator Reg(&RegisterImage[10]);
    Reg.Field(8, 3, To_VcoDiv_DividerInput(VcoMux));
}

//-----------------------------------------------------------------------------
//  Lmk04800::GetVcoDiv()  --  Read PLL2 (Vco Postdivider)
//-----------------------------------------------------------------------------

unsigned int Lmk04800::GetVcoDiv()
{
    PtrBitManipulator Reg(&RegisterImage[10]);
    unsigned int VcoDiv_field = Reg.Field(8, 3);
    return From_VcoDiv_DividerInput(VcoDiv_field);
}

//-----------------------------------------------------------------------------
//  Lmk04800::ReadDataRegisters()  --  Read all data registers into Image vector
//-----------------------------------------------------------------------------

void Lmk04800::ReadDataRegisters()
{
    for (unsigned int i=0; i<=16; i++)
        {
        //if (i==9)
        //    continue;
        RegisterImage[i] = Port(i);
        }
    // Skip 17-23
    for (unsigned int i=24; i<RegisterImage.size(); i++)
        {
        RegisterImage[i] = Port(i);
        }
}

//-----------------------------------------------------------------------------
//  Lmk04800::WriteDataRegisters()  --  Write all data registers from Image vector
//-----------------------------------------------------------------------------

void Lmk04800::WriteDataRegisters()
{
    ScopedTraceRef ccc(Debug, "Lmk04800::WriteDataRegisters()");

    for (unsigned int i=0; i<=16; i++)
        {
        //if (i==9)  // skip 9
        //    continue;
        Port(i, RegisterImage[i]);
        }
    // Skip 17-23
    for (unsigned int i=24; i<RegisterImage.size(); i++)
        {
        Port(i, RegisterImage[i]);
        }
}

//-----------------------------------------------------------------------------
//  Lmk04800::OnSetPower()  --  Perform Pwr Setting operation
//-----------------------------------------------------------------------------

void  Lmk04800::OnSetPower(bool & new_state)
{
    PtrBitManipulator Reg1(&RegisterImage[1]);
    Reg1.Bit(17, new_state);
    //  pause to allow chip to activate
    Sleep(10);
}


//-----------------------------------------------------------------------------
//  Lmk04800::InitDividerSupport()  --  Initialize Special Divider Maps
//-----------------------------------------------------------------------------

void  Lmk04800::InitDividerSupport()
{
    //  ClkinX_PreR are Input Prescaler dividers
    ClkinX_PreR_DividerValues.push_back(1);     ClkinX_PreR_DividerValues.push_back(2);
    ClkinX_PreR_DividerValues.push_back(4);     ClkinX_PreR_DividerValues.push_back(8);

    //  Valid dividers is keyed by divide_by, value is Input Value
    ClkinX_PreR_ValidDividers[1]    = 0x00;       ClkinX_PreR_ValidDividers[2]    = 0x01;
    ClkinX_PreR_ValidDividers[4]    = 0x02;       ClkinX_PreR_ValidDividers[8]    = 0x03;

    //     Valid dividers is keyed by Input Value, value is divide_by
    ClkinX_PreR_InverseDividers[0x00]  =    1;      ClkinX_PreR_InverseDividers[0x01]  =    2;
    ClkinX_PreR_InverseDividers[0x02]  =    4;      ClkinX_PreR_InverseDividers[0x03]  =    8;

    //  N2_Pre are PLL2 Feedback Prescaler dividers
    N2_Pre_DividerValues.push_back(2);     N2_Pre_DividerValues.push_back(3);
    N2_Pre_DividerValues.push_back(4);     N2_Pre_DividerValues.push_back(5);
    N2_Pre_DividerValues.push_back(6);     N2_Pre_DividerValues.push_back(7);
    N2_Pre_DividerValues.push_back(8);

    //  Valid dividers is keyed by divide_by, value is Input Value
    N2_Pre_ValidDividers[2]    = 0x02;       N2_Pre_ValidDividers[3]    = 0x03;
    N2_Pre_ValidDividers[4]    = 0x04;       N2_Pre_ValidDividers[5]    = 0x05;
    N2_Pre_ValidDividers[6]    = 0x06;       N2_Pre_ValidDividers[7]    = 0x07;
    N2_Pre_ValidDividers[8]    = 0x00;

    //     Valid dividers is keyed by Input Value, value is divide_by
    N2_Pre_InverseDividers[0x00]  =    8;      N2_Pre_InverseDividers[0x01]  =    2;
    N2_Pre_InverseDividers[0x02]  =    2;      N2_Pre_InverseDividers[0x03]  =    3;
    N2_Pre_InverseDividers[0x04]  =    4;      N2_Pre_InverseDividers[0x05]  =    5;
    N2_Pre_InverseDividers[0x06]  =    6;      N2_Pre_InverseDividers[0x07]  =    7;

    //  VcoDiv is PLL2 VCO post_scaler divider
    VcoDiv_DividerValues.push_back(2);     VcoDiv_DividerValues.push_back(3);
    VcoDiv_DividerValues.push_back(4);     VcoDiv_DividerValues.push_back(5);
    VcoDiv_DividerValues.push_back(6);     VcoDiv_DividerValues.push_back(7);
    VcoDiv_DividerValues.push_back(8);

    //  Valid dividers is keyed by divide_by, value is Input Value
    VcoDiv_ValidDividers[2]    = 0x02;       VcoDiv_ValidDividers[3]    = 0x03;
    VcoDiv_ValidDividers[4]    = 0x04;       VcoDiv_ValidDividers[5]    = 0x05;
    VcoDiv_ValidDividers[6]    = 0x06;       VcoDiv_ValidDividers[7]    = 0x07;
    VcoDiv_ValidDividers[8]    = 0x00;

    //     Valid dividers is keyed by Input Value, value is divide_by
    VcoDiv_InverseDividers[0x00]  =    8;      VcoDiv_InverseDividers[0x01]  =    2;
    VcoDiv_InverseDividers[0x02]  =    2;      VcoDiv_InverseDividers[0x03]  =    3;
    VcoDiv_InverseDividers[0x04]  =    4;      VcoDiv_InverseDividers[0x05]  =    5;
    VcoDiv_InverseDividers[0x06]  =    6;      VcoDiv_InverseDividers[0x07]  =    7;
}

//-----------------------------------------------------------------------------
//  Lmk04800::IsValidDivider() -- tests if divide_by is a valid divider
//-----------------------------------------------------------------------------
//  Output Group Output Dividers

bool  Lmk04800::IsValidDivider(unsigned int divide_by)
{
	if (divide_by !=0 && divide_by<=kOD_Max)
		return true;
	else
		return false;
}

//-----------------------------------------------------------------------------
//  Lmk04800::ToDividerInput() -- Returns input value for divide_by
//-----------------------------------------------------------------------------
//  this maps divider ratios to the field value in the chip hardware

unsigned int  Lmk04800::ToDividerInput(unsigned int divide_by)
{
    if (IsValidDivider(divide_by))
        return divide_by;
    else
        return 1;   // if invalid divider, divide by 1.
}

//-----------------------------------------------------------------------------
//  Lmk04800::FromDividerInput() -- Returns divide_by for input value
//-----------------------------------------------------------------------------
//  this maps the field value in the chip hardware to divider ratios

unsigned int  Lmk04800::FromDividerInput(unsigned int divide_field)
{
    return divide_field;
}

//-----------------------------------------------------------------------------
//  Lmk04800::To_CLKinX_PreR_DividerInput() -- Returns input value for divide_by
//-----------------------------------------------------------------------------
//  this maps divider ratios to the field value in the chip hardware

unsigned int  Lmk04800::To_CLKinX_PreR_DividerInput(unsigned int divide_by)
{
    if (IsValid_CLKinX_PreR_Divider(divide_by))
        return ClkinX_PreR_ValidDividers[divide_by];
    else
        return ClkinX_PreR_ValidDividers[1];   // if invalid divider, divide by 1.
}

//-----------------------------------------------------------------------------
//  Lmk04800::From_CLKinX_PreR_DividerInput() -- Returns divide_by for input value
//-----------------------------------------------------------------------------
//  this maps divider bits in the field value in the chip hardware to divide_by result

unsigned int  Lmk04800::From_CLKinX_PreR_DividerInput(unsigned int divider_bits)
{
    InverseDividerMap::iterator iter = ClkinX_PreR_InverseDividers.find(divider_bits);

    if (iter != ClkinX_PreR_InverseDividers.end())
        return ClkinX_PreR_InverseDividers[divider_bits];
    else
        return 1;   // if invalid divider_bits, claim it is divide by 1
}

//-----------------------------------------------------------------------------
//  Lmk04800::IsValid_CLKinX_PreR_Divider() -- tests if divide_by is a valid divider
//-----------------------------------------------------------------------------

bool  Lmk04800::IsValid_CLKinX_PreR_Divider(unsigned int divide_by)
{
    ValidDividerMap::iterator iter = ClkinX_PreR_ValidDividers.find(divide_by);
    return iter != ClkinX_PreR_ValidDividers.end();
}

//-----------------------------------------------------------------------------
//  Lmk04800::To_N2_Pre_DividerInput() -- Returns input value for divide_by
//-----------------------------------------------------------------------------
//  this maps divider ratios to the field value in the chip hardware

unsigned int  Lmk04800::To_N2_Pre_DividerInput(unsigned int divide_by)
{
    if (IsValid_N2_Pre_Divider(divide_by))
        return N2_Pre_ValidDividers[divide_by];
    else
        return N2_Pre_ValidDividers[1];   // if invalid divider, divide by 1.
}

//-----------------------------------------------------------------------------
//  Lmk04800::From_N2_Pre_DividerInput() -- Returns divide_by for input value
//-----------------------------------------------------------------------------
//  this maps divider bits in the field value in the chip hardware to divide_by result

unsigned int  Lmk04800::From_N2_Pre_DividerInput(unsigned int divider_bits)
{
    InverseDividerMap::iterator iter = N2_Pre_InverseDividers.find(divider_bits);

    if (iter != N2_Pre_InverseDividers.end())
        return N2_Pre_InverseDividers[divider_bits];
    else
        return 1;   // if invalid divider_bits, claim it is divide by 1
}

//-----------------------------------------------------------------------------
//  Lmk04800::IsValid_N2_Pre_Divider() -- tests if divide_by is a valid divider
//-----------------------------------------------------------------------------

bool  Lmk04800::IsValid_N2_Pre_Divider(unsigned int divide_by)
{
    ValidDividerMap::iterator iter = N2_Pre_ValidDividers.find(divide_by);
    return iter != N2_Pre_ValidDividers.end();
}

//-----------------------------------------------------------------------------
//  Lmk04800::To_VcoDiv_DividerInput() -- Returns input value for divide_by
//-----------------------------------------------------------------------------
//  this maps divider ratios to the field value in the chip hardware

unsigned int  Lmk04800::To_VcoDiv_DividerInput(unsigned int divide_by)
{
    if (IsValid_VcoDiv_Divider(divide_by))
        return VcoDiv_ValidDividers[divide_by];
    else
        return VcoDiv_ValidDividers[1];   // if invalid divider, divide by 1.
}

//-----------------------------------------------------------------------------
//  Lmk04800::From_VcoDiv_DividerInput() -- Returns divide_by for input value
//-----------------------------------------------------------------------------
//  this maps divider bits in the field value in the chip hardware to divide_by result

unsigned int  Lmk04800::From_VcoDiv_DividerInput(unsigned int divider_bits)
{
    InverseDividerMap::iterator iter = VcoDiv_InverseDividers.find(divider_bits);

    if (iter != VcoDiv_InverseDividers.end())
        return VcoDiv_InverseDividers[divider_bits];
    else
        return 1;   // if invalid divider_bits, claim it is divide by 1
}

//-----------------------------------------------------------------------------
//  Lmk04800::IsValid_VcoDiv_Divider() -- tests if divide_by is a valid divider
//-----------------------------------------------------------------------------

bool  Lmk04800::IsValid_VcoDiv_Divider(unsigned int divide_by)
{
    ValidDividerMap::iterator iter = VcoDiv_ValidDividers.find(divide_by);
    return iter != VcoDiv_ValidDividers.end();
}

//-----------------------------------------------------------------------------
//  Lmk04800::GetNearestDividers() -- Find nearest divider to a candidate
//-----------------------------------------------------------------------------

void  Lmk04800::GetNearestDividers(unsigned int /*output*/, float candidate, int & below, int & above)
{
    below = static_cast<int>(candidate);
    above = below+1;

    // range check
    if (below<1)
        below = 1;
    if (above>static_cast<int>(kOD_Max))
        above = kOD_Max;
}

//-----------------------------------------------------------------------------
//  Lmk04800::Vco_CanTuneTo()
//-----------------------------------------------------------------------------

bool Lmk04800::Vco_CanTuneTo(double freq)
{
    bool retValue;

    if (freq >= LocalVcxo.Min() && freq <= LocalVcxo.Max())
        retValue = true;
    else
        retValue = false;

    return retValue;
}

//-----------------------------------------------------------------------------
//  Lmk04800::Vco_BestTunableFrequency()
//-----------------------------------------------------------------------------

double Lmk04800::Vco_BestTunableFrequency(double freq)
{
    double retValue;

    if (freq < LocalVcxo.Min())
        retValue = LocalVcxo.Min();
    else if (freq >= LocalVcxo.Min() && freq <= LocalVcxo.Max())
        retValue = freq;
    else
        retValue = LocalVcxo.Max();

    return retValue;
}

//-----------------------------------------------------------------------------
//  Lmk04800::Vco_SetVcoFrequency()
//-----------------------------------------------------------------------------

void  Lmk04800::Vco_SetVcoFrequency(double freq)
{
    ScopedTraceRef ccc(Debug, "Lmk04800::Vco_SetVcoFrequency()");

    FVco2Frequency = freq;

    Debug.Log("VCO Values");
    Debug.Log("Lmk04800::Output Vco Freq (MHz) =", FVco2Frequency/1.0e6);
}

//-----------------------------------------------------------------------------
//  Lmk04800::Vco_GetVcoFrequency()  --  read Vco Frequenc7y
//-----------------------------------------------------------------------------

double  Lmk04800::Vco_GetVcoFrequency()
{
    return FVco2Frequency;
}


} // namespace Innovative
