//
// LMK61E2_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2016 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <LMK61E2_Mb.h>
#include <limits>
#include <sstream>
#include <cmath>
#include <HardwareRegister_Mb.h>
#include <FloatingPtSupport_Mb.h>
#include <Magic_Mb.h>
#include <SystemSupport_Mb.h>

namespace Innovative
{

//=============================================================================
//  CLASS Lmk61E2 --  Programmable XO/VCXO Chip Support
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class Lmk61E2
//-----------------------------------------------------------------------------

const double       _VcoLow  = 4600.0e6;
const double       _VcoHigh = 5600.0e6;
const unsigned int MinOutputDivider = 5;
const unsigned int MaxOutputDivider = 511;
const double       MinOutputFrequency = 10.0e6;
const double       MaxOutputFrequency = 1000.0e6;

const int I2CAddr_ADDFloating   = 0xB2;      // Bits 2 and 1 fixed by ADD pins...
const int I2CAddr_ADDPulledLow  = 0xB0;      // Bits 2 and 1 fixed by ADD pins...
const int I2CAddr_ADDPulledHigh = 0xB4;      // Bits 2 and 1 fixed by ADD pins...

const int DataRegOffset = 0;       // Base address of Data Registers
const int DataRegisterCount = 73;

//-----------------------------------------------------------------------------
//  constructor for class Lmk61E2
//-----------------------------------------------------------------------------
//  Note: we do NOT own the Device pointer, so don't delete it

Lmk61E2::Lmk61E2(IRequires_I2C_RegIo & device)
    :  Device(&device), UseClockDoubler(true),
       VcoLow(_VcoLow), VcoHigh(_VcoHigh),
       FADDMode(addFloating),
       RegisterImage(DataRegisterCount),
       Debug(true)
{
    FReference       = 50.0e6;
    FFrequency       = 156.25e6;
    FFrequencyActual = FFrequency;
    PllMode          = pmInteger;
}

//-----------------------------------------------------------------------------
//  Lmk61E2::Init()  --  Init setup
//-----------------------------------------------------------------------------

void  Lmk61E2::Init()
{
    ScopedTraceRef ccc(Debug, "Lmk61E2::Init()");

    Device->I2C_SetSubAddress(I2C_Address());
    ReadStatusRegisters();
    ReadDataRegisters();    
    OnInit();

    WriteDataRegisters();

    //  Set up for new frequency
    NewFrequency();
}

//-----------------------------------------------------------------------------
//  Lmk61E2::Reapply()  --  Reapply settings to hardware from saved state
//-----------------------------------------------------------------------------

void  Lmk61E2::Reapply()
{
    ScopedTraceRef ccc(Debug, "Lmk61E2::Reapply()");

    OnReapply();
    WriteDataRegisters();

    //  Set up for new frequency
    NewFrequency();
}

//-----------------------------------------------------------------------------
//  Lmk61E2::Frequency()  --  Set frequency
//-----------------------------------------------------------------------------

void Lmk61E2::Frequency(double freq)
{
    ScopedTraceRef ccc(Debug, "Lmk61E2::Frequency()");

    // Store the requested frequency
    FFrequency = freq;
    Debug.Log("Frequency = ", FFrequency);
    Debug.Log("Reference = ", FReference);

    if (FReference == 0)
        {
        Debug.Log("Lmk61E2: Skipping VCO configuration, reference is set to 0");
        return;
        }

    double FrequencyCapped;
    // Check range on requested frequency, cap to the closest rail when out of range
    if (CanTuneTo(freq))
        FrequencyCapped = freq;
    else
        FrequencyCapped = BestTunableFrequency(freq);

    //  Calculate PFD
    double PFD;
    if (UseClockDoubler)
        PFD = FReference*2;
    else
        PFD = FReference;
    Debug.Log("Target PFD = ", PFD);

    unsigned int OD, N_Int, N_Num, N_Den;
    CalcPllParams(FrequencyCapped, PFD, OD, N_Int, N_Num, N_Den);

    Debug.Log("OutDiv = ", OD);
    Debug.Log("N_Int = ", N_Int);
    Debug.Log("N_Num = ", N_Num);
    Debug.Log("N_Den = ", N_Den);

    // Calculate VCO Frequency
    double N_Actual;
    if (PllMode == pmInteger)
        N_Actual = double(N_Int);
    else
        N_Actual = double(N_Int) + double(N_Num)/double(N_Den);
    Debug.Log("N Actual = ", N_Actual);
    double VcoFreq = PFD * N_Actual;
    Debug.Log("VcoFreq = ", VcoFreq);

    FFrequencyActual = VcoFreq/OD;
    Debug.Log("Frequency Actual = ", FFrequencyActual);

    // Load Registers
    LoadPllParams(OD, N_Int, N_Num, N_Den);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::CalcPllParams()  --  Hunt for proper divider for a frequency
//-----------------------------------------------------------------------------

void Lmk61E2::CalcPllParams(double freq, double pfd, unsigned int &ODiv,
                            unsigned int &NInt, unsigned int &NNum, unsigned int &NDen)
{
    unsigned int MaxOD = static_cast<unsigned int>(std::floor(VcoHigh/freq));
    unsigned int MinOD = static_cast<unsigned int>(std::ceil(VcoLow/freq));

    //Debug.Log("MaxOD = ", MaxOD);
    //Debug.Log("MinOD = ", MinOD);

    std::vector<double> NCandidateVec;
    std::vector<unsigned int> ODCandidateVec;

    PllMode = pmInteger;

    // Loop over the output divider values, calculate the candidate values for N,
    // and store them in a vector
    for (unsigned int Div=MaxOD; Div>=MinOD; Div--)
        {
        double VcoFreq    = freq * Div;
        double NCandidate = VcoFreq/pfd;
        NCandidateVec.push_back(NCandidate);
        ODCandidateVec.push_back(Div);
        // Break/return upon detecting an integer N value
        if (FractionalPart(NCandidate) == 0)
            {
            ODiv = Div;
            NInt = static_cast<unsigned int>(IntegerPart(NCandidate));
            NNum = 0;
            NDen = 0;
            return;
            }
        }

    // Fractional N values are calculated below
    //Debug.Log("Number of candidate N values = ", NCandidateVec.size());

    PllMode = pmFractional;

    unsigned int OD(1), N_Int(1), N_Num(0), N_Den(1);
    double MinError = 1.0;

    // Loop over the candidate N values and find the best fit
    for (unsigned int i=0; i<NCandidateVec.size(); i++)
        {
        unsigned int integral   = static_cast<unsigned int>(IntegerPart(NCandidateVec[i]));
        double       fractional = FractionalPart(NCandidateVec[i]);
        //Debug.Log("Candidate N = ", NCandidateVec[i]);
        //Debug.Log("N Integral = ", integral);
        //Debug.Log("N Fractional = ", fractional);

        unsigned int Num, Den;
        FractionalToRational(fractional, Num, Den);
        double Error = fabs(fractional-(double(Num)/double(Den)));
        //Debug.Log("N Error = ", Error);

        // Break upon detecting an integer N value
        if (Error <= MinError)
            {
            MinError = Error;
            OD       = ODCandidateVec[i];
            N_Int    = integral;
            N_Num    = Num;
            N_Den    = Den;
            }
        }

    ODiv = OD;
    NInt = N_Int;
    NNum = N_Num;
    NDen = N_Den;
}

//-----------------------------------------------------------------------------
//  Lmk61E2::FractionalToRational()  --  Hunt for proper rational representation
//-----------------------------------------------------------------------------

void  Lmk61E2::FractionalToRational(double fractional, unsigned int & Num, unsigned int & Den)
{
    const unsigned int MaxDen = 2042040;//(1<<22)-2;

    unsigned int InitialNum = static_cast<unsigned int>(IntegerPart(fractional*MaxDen));
    unsigned int InitialDen = MaxDen;

    if (fractional == 0)
        {
        Num = 0;
        Den = 1;
        return;
        }

    unsigned int GCDFactor = Gcd(InitialNum,InitialDen);
    //Debug.Log("GCDFactor = ", GCDFactor);

    // Num and Den are 22 bit unsigned numbers
    Num = InitialNum/GCDFactor;
    Den = InitialDen/GCDFactor;
}

//-----------------------------------------------------------------------------
//  Lmk61E2::LoadPllParams()  --  Load the PLL parameters to the registers
//-----------------------------------------------------------------------------

void Lmk61E2::LoadPllParams(unsigned int OD, unsigned int N_Int, unsigned int N_Num, unsigned int N_Den)
{
    OutDiv(OD);
    NDiv(N_Int);
    PllFracNumerator(N_Num);
    PllFracDenominator(N_Den);
    PllDividerDoublerEnable(UseClockDoubler);

    if (PllMode == pmInteger)
        {
        SetMashDitherConfig(mdDisabled);
        SetChargePumpCurrent(cp6p4mA);
        SetChargePumpPhaseShift(cppsZero);

        if (UseClockDoubler)
            {
            SetLoopFilterR2(lfr2500);
            SetLoopFilterC1(0x1);
            PllEnableC3(true);
            SetLoopFilterR3(lfr31535);
            SetLoopFilterC3(0x7);
            }
        else
            {
            SetLoopFilterR2(lfr2700);
            SetLoopFilterC1(0x0);
            PllEnableC3(false);
            SetLoopFilterR3(lfr318);
            SetLoopFilterC3(0x0);
            }
        }
    else
        {
        SetMashDitherConfig(mdWeak);
        SetChargePumpCurrent(cp1p6mA);
        SetChargePumpPhaseShift(cpps1ns_at100);

        SetLoopFilterR2(lfr2500);
        SetLoopFilterC1(0x1);
        PllEnableC3(true);
        SetLoopFilterR3(lfr31535);
        SetLoopFilterC3(0x7);
        }
}

//-----------------------------------------------------------------------------
//  Lmk61E2::Port()  --  Write to chip data port
//-----------------------------------------------------------------------------

void Lmk61E2::Port(unsigned int reg, unsigned int value)
{
    std::stringstream msg;
    msg << "Lmk61E2::Write 0x" << std::hex << (value)
        << " to reg " << std::dec << (reg);
    Debug.Log(msg.str());
    //
    //  Write using I2C interface
    Device->I2C_Reg(reg, value);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::Port()  --  Read from chip data port
//-----------------------------------------------------------------------------

unsigned char Lmk61E2::Port(unsigned int reg)
{
    //
    //  Read using I2C interface
    int value = Device->I2C_Reg(reg);

    std::stringstream msg;
    msg << "Lmk61E2::Read 0x" << std::hex << value
        << " from reg " << std::dec << (reg);
    Debug.Log(msg.str());

    return static_cast<unsigned char>(value);
}

//-----------------------------------------------------------------------------
// Lmk61E2::CanTuneTo(double freq)
//-----------------------------------------------------------------------------

bool Lmk61E2::CanTuneTo(double freq)
{
    return WithinRange(freq, Min(), Max());
}

//-----------------------------------------------------------------------------
// Lmk61E2::BestTunableFrequency()
//-----------------------------------------------------------------------------

double Lmk61E2::BestTunableFrequency(double freq)
{
    if (freq <= Min())
        return Min();
    else if (freq >= Max())
        return Max();
    else
        return freq;
}

//-----------------------------------------------------------------------------
// Lmk61E2:: Min() - Tuning Range
//-----------------------------------------------------------------------------

double Lmk61E2::Min()
{
    return std::max(VcoLow/MaxOutputDivider, MinOutputFrequency);
}

//-----------------------------------------------------------------------------
// Lmk61E2::Max()
//-----------------------------------------------------------------------------

double Lmk61E2::Max()
{
    return std::min(VcoHigh/MinOutputDivider, MaxOutputFrequency);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::ReadStatusRegisters()  --  Read unchanging status registers
//-----------------------------------------------------------------------------

void Lmk61E2::ReadStatusRegisters()
{
    ScopedTraceRef ccc(Debug, "Lmk61E2::ReadStatusRegisters()");

    for (unsigned int i=0; i<4; i++)
        {
        RegisterImage[i] = Port(i);
        }
    for (unsigned int i=8; i<10; i++)
        {
        RegisterImage[i] = Port(i);
        }
    RegisterImage[66] = Port(66);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::ReadDataRegisters()  --  Read all writable data registers into Image vector
//-----------------------------------------------------------------------------

void Lmk61E2::ReadDataRegisters()
{
    ScopedTraceRef ccc(Debug, "Lmk61E2::ReadDataRegisters()");

    for (unsigned int i=0; i<RegisterImage.size(); i++)
        {
        if (IsWritableDataRegister(i))
            RegisterImage[i] = Port(i);
        }
}

//-----------------------------------------------------------------------------
//  Lmk61E2::WriteDataRegisters()  --  Write all writable data registers from Image vector
//-----------------------------------------------------------------------------

void Lmk61E2::WriteDataRegisters()
{
    ScopedTraceRef ccc(Debug, "Lmk61E2::WriteDataRegisters()");

    for (unsigned int i=0; i<RegisterImage.size(); i++)
        {
        if (IsWritableDataRegister(i))
            Port(i, RegisterImage[i]);
        }
}

//-----------------------------------------------------------------------------
//  Lmk61E2::IsWritableDataRegister()  --  Return true if reg is ok to write
//-----------------------------------------------------------------------------

bool Lmk61E2::IsWritableDataRegister(unsigned int idx)
{
    if (idx==10)
        return true;
    else if (idx==16 || idx==17)
        return true;
    else if (idx>=21 && idx<=23)
        return true;
    else if (idx>=25 && idx<=39)
        return true;
    else if (idx==42)
        return true;
    else
        return false;
}

//-----------------------------------------------------------------------------
//  Lmk61E2::NewFrequency()  --  Inform chip of new frequency setting
//-----------------------------------------------------------------------------

void Lmk61E2::NewFrequency()
{
    VcoCalibration();  // must do when configuration changes
    Sleep(10);
    Debug.Log("CalActive() = ", CalActive());
    Debug.Log("Pll_Lock() = ", Pll_Lock());
}

//-----------------------------------------------------------------------------
//  Lmk61E2::VendorID()  --  read vendor ID
//-----------------------------------------------------------------------------

unsigned int Lmk61E2::VendorID()
{
    //  R0  R1  VNDRID_BY1  VNDRID_BY0
    PtrBitManipulator Reg0(&RegisterImage[0]);
    PtrBitManipulator Reg1(&RegisterImage[1]);

    unsigned int value(0);
    value |= Reg0.Value();     // Bits 8-15
    value <<= 8;
    value |= Reg1.Value();     // LSB

    return value;
}

//-----------------------------------------------------------------------------
//  Lmk61E2::ProductID()  --  read Product ID
//-----------------------------------------------------------------------------

unsigned int Lmk61E2::ProductID()
{
    //  R2  PRODID
    PtrBitManipulator Reg2(&RegisterImage[2]);

    return Reg2.Value();
}

//-----------------------------------------------------------------------------
//  Lmk61E2::RevisionID()  --  read Revision ID
//-----------------------------------------------------------------------------

unsigned int Lmk61E2::RevisionID()
{
    //  R3  REVID
    PtrBitManipulator Reg3(&RegisterImage[3]);

    return Reg3.Value();
}

//-----------------------------------------------------------------------------
//  Lmk61E2::OutDiv()  --  Set OUTDIV registers
//-----------------------------------------------------------------------------

void  Lmk61E2::OutDiv(unsigned int out_div)
{
     //  R22  R23  OUTDIV_BY1  OUTDIV_BY0
     PtrBitManipulator Reg22(&RegisterImage[22]);
     PtrBitManipulator Reg23(&RegisterImage[23]);

     Reg22.Field(0, 1, out_div >> 8);     // MSB  (Bit 8)
     Reg23.Value(out_div & 0xFF);         // LSB
}

//-----------------------------------------------------------------------------
//  Lmk61E2::NDiv()  --  Set NDIV registers
//-----------------------------------------------------------------------------

void  Lmk61E2::NDiv(unsigned int n_div)
{
     //  R25  R26  NDIV_BY1  NDIV_BY0
     PtrBitManipulator Reg25(&RegisterImage[25]);
     PtrBitManipulator Reg26(&RegisterImage[26]);

     Reg25.Field(0, 4, n_div >> 8);      // Bits 8-11
     Reg26.Value(n_div & 0xFF);          // LSB
}

//-----------------------------------------------------------------------------
//  Lmk61E2::PllFracNumerator()  --  Write PLL FRAC Numerator to Image
//-----------------------------------------------------------------------------

void  Lmk61E2::PllFracNumerator(unsigned int value)
{
    //  R27 R28  R29  PLL_FRACNUM_BY2 PLL_FRACNUM_BY1  PLL_FRACNUM_BY0
    PtrBitManipulator Reg27(&RegisterImage[27]);
    PtrBitManipulator Reg28(&RegisterImage[28]);
    PtrBitManipulator Reg29(&RegisterImage[29]);

    Reg27.Field(0, 6, (value >> 16));   // Bits 16-21
    Reg28.Value(value >> 8);            // Bits 8-15
    Reg29.Value(value & 0xFF);          // LSB
}

//-----------------------------------------------------------------------------
//  Lmk61E2::PllFracDenominator()  --  Write PLL FRAC Denominator to Image
//-----------------------------------------------------------------------------

void  Lmk61E2::PllFracDenominator(unsigned int value)
{
    //  R30 R31  R32  PLL_FRACDEN_BY2 PLL_FRACDEN_BY1  PLL_FRACDEN_BY0
    PtrBitManipulator Reg30(&RegisterImage[30]);
    PtrBitManipulator Reg31(&RegisterImage[31]);
    PtrBitManipulator Reg32(&RegisterImage[32]);

    Reg30.Field(0, 6, (value >> 16));   // Bits 16-21
    Reg31.Value(value >> 8);            // Bits 8-15
    Reg32.Value(value & 0xFF);          // LSB
}

//-----------------------------------------------------------------------------
//  Lmk61E2::PllPowerdown()  --  power off PLL
//-----------------------------------------------------------------------------

void  Lmk61E2::PllPowerdown(bool state)
{
    //  R10  PLL_PDN
    PtrBitManipulator Reg10(&RegisterImage[10]);

    return Reg10.Bit(6, state);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::EnableFreqCalibration()  --
//-----------------------------------------------------------------------------

void  Lmk61E2::EnableFreqCalibration(bool state)
{
    //  R10  ENCAL
    PtrBitManipulator Reg10(&RegisterImage[10]);

    return Reg10.Bit(1, state);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::AutoStart()  --
//-----------------------------------------------------------------------------

void  Lmk61E2::AutoStart(bool state)
{
    //  R10  AUTOSTRT
    PtrBitManipulator Reg10(&RegisterImage[10]);

    return Reg10.Bit(0, state);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::XoCapControl()  --  XO Margining Offset
//-----------------------------------------------------------------------------

void  Lmk61E2::XoCapControl(unsigned int value)
{
    //  R16  R17  XO_CAPCTRL_BY1  XO_CAPCTRL_BY0
    PtrBitManipulator Reg16(&RegisterImage[16]);
    PtrBitManipulator Reg17(&RegisterImage[17]);

    Reg16.Field(0, 2, (value >> 8));     // Bits 8-9
    Reg17.Value(value & 0xFF);           // LSB

    // Note - doc seems confused by what bit is what here.
}

//-----------------------------------------------------------------------------
//  Lmk61E2::OutputBufferPowerdown()  --
//-----------------------------------------------------------------------------

void  Lmk61E2::OutputBufferPowerdown(bool state)
{
    //  R21  DIFF_OUT_PD
    PtrBitManipulator Reg21(&RegisterImage[21]);

    return Reg21.Bit(7, state);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::SetOutputConfig()  -- Set output driver format
//-----------------------------------------------------------------------------

void  Lmk61E2::SetOutputConfig(unsigned int setting)
{
    //  R21  OUT_SEL
    PtrBitManipulator Reg21(&RegisterImage[21]);

    return Reg21.Field(0, 2, setting);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::SetMashDitherConfig()  -- PLL Fractional Divider Settings
//-----------------------------------------------------------------------------

void Lmk61E2::SetMashDitherConfig(unsigned int setting)
{
    //  R33  PLL_DTHRMODE
    PtrBitManipulator Reg33(&RegisterImage[33]);

    return Reg33.Field(2, 2, setting);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::SetMashOrderConfig()  -- PLL Fractional Divider Settings
//-----------------------------------------------------------------------------

void Lmk61E2::SetMashOrderConfig(unsigned int setting)
{
    //  R33  PLL_ORDER
    PtrBitManipulator Reg33(&RegisterImage[33]);

    return Reg33.Field(0, 2, setting);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::PllDividerDoublerEnable()  -- Input frequency doubler
//-----------------------------------------------------------------------------

void Lmk61E2::PllDividerDoublerEnable(bool is_doubled)
{
    //  R34  PLL_D
    PtrBitManipulator Reg34(&RegisterImage[34]);

    return Reg34.Bit(5, is_doubled);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::SetChargePumpCurrent()  -- Set charge pump current
//-----------------------------------------------------------------------------

void Lmk61E2::SetChargePumpCurrent(unsigned int setting)
{
    //  R34  PLL_CP
    PtrBitManipulator Reg34(&RegisterImage[34]);

    return Reg34.Field(0, 4, setting);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::SetChargePumpPhaseShift()  --
//-----------------------------------------------------------------------------

void Lmk61E2::SetChargePumpPhaseShift(unsigned int setting)
{
    //  R35  PLL_CP_PHASE_SHIFT
    PtrBitManipulator Reg35(&RegisterImage[35]);

    return Reg35.Field(4, 3, setting);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::PllEnableC3()  --
//-----------------------------------------------------------------------------

void Lmk61E2::PllEnableC3(bool state)
{
    //  R35  PLL_ENABLE_C3
    PtrBitManipulator Reg35(&RegisterImage[35]);

    return Reg35.Bit(2, state);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::SetLoopFilterR2()  --
//-----------------------------------------------------------------------------

void Lmk61E2::SetLoopFilterR2(unsigned int setting)
{
    //  R36  PLL_LF_R2
    PtrBitManipulator Reg36(&RegisterImage[36]);

    return Reg36.Value(setting);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::SetLoopFilterC1()  --
//-----------------------------------------------------------------------------

void Lmk61E2::SetLoopFilterC1(unsigned int setting)
{
    //  R37  PLL_LF_C1
    PtrBitManipulator Reg37(&RegisterImage[37]);

    return Reg37.Field(0, 3, setting);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::SetLoopFilterR3()  --
//-----------------------------------------------------------------------------

void Lmk61E2::SetLoopFilterR3(unsigned int setting)
{
    //  R38 PLL_LF_R3
    PtrBitManipulator Reg38(&RegisterImage[38]);

    return Reg38.Field(0, 7, setting);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::SetLoopFilterC1()  --
//-----------------------------------------------------------------------------

void Lmk61E2::SetLoopFilterC3(unsigned int setting)
{
    //  R39  PLL_LF_C3
    PtrBitManipulator Reg39(&RegisterImage[39]);

    return Reg39.Field(0, 3, setting);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::ClosedLoopWaitPeriod()  --
//-----------------------------------------------------------------------------

void Lmk61E2::ClosedLoopWaitPeriod(unsigned int setting)
{
    //  R42  PLL_CLSDWAIT
    PtrBitManipulator Reg42(&RegisterImage[42]);

    return Reg42.Field(2, 2, setting);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::VcoWaitPeriod()  --
//-----------------------------------------------------------------------------

void Lmk61E2::VcoWaitPeriod(unsigned int setting)
{
    //  R42  PLL_VCOWAIT
    PtrBitManipulator Reg42(&RegisterImage[42]);

    return Reg42.Field(0, 2, setting);
}

//   R8 R9  R47-49  R51-53  R56

//-----------------------------------------------------------------------------
//  Lmk61E2::Pll_Lock()  --  read loss of lock bit
//-----------------------------------------------------------------------------

bool Lmk61E2::Pll_Lock()
{
    unsigned int value = Port(66);
    PtrBitManipulator Reg66(&value);

    return ! Reg66.Bit(1);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::CalActive()  --  read cal active bit
//-----------------------------------------------------------------------------

bool Lmk61E2::CalActive()
{
    unsigned int value = Port(66);
    PtrBitManipulator Reg66(&value);

    return Reg66.Bit(0);
}

//-----------------------------------------------------------------------------
//  Lmk61E2::VcoCalibration()  --  Start VCO Calibration (after reconfig)
//-----------------------------------------------------------------------------

void Lmk61E2::VcoCalibration()
{
    Debug.Log("Start the Vco calibration");
    Port(72, 0x02);  //
}

//-----------------------------------------------------------------------------
//  Lmk61E2::I2C_Address()  --  return proper I2C Address for ADD setting
//-----------------------------------------------------------------------------

int Lmk61E2::I2C_Address()
{
    switch (FADDMode)
        {
        default:
        case addFloating:
            return I2CAddr_ADDFloating;
        case addPulledLow:
            return I2CAddr_ADDPulledLow;
        case addPulledHigh:
            return I2CAddr_ADDPulledHigh;
        }
}

} // namespace Innovative

