// LogicLoadThread_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//
#include <algorithm>
#include <LogicLoadThread_Mb.h>
#include <FpgaLoadBase_Mb.h>
#include <FileSupport_Mb.h>
#include <sstream>
#include <SystemSupport_Mb.h>

namespace   Innovative
{

//==============================================================================
//  CLASS LogicLoadThread -- manages Logic load via select map
//==============================================================================

//---------------------------------------------------------------------------
//  LogicLoadThread::LogicLoadThread() -- Ctor
//---------------------------------------------------------------------------

LogicLoadThread::LogicLoadThread(FpgaLoadBase & fpga)
	: inherited("LogicLoadThread"), Image(0), Fpga(fpga),
      Configuring(Event::ManualReset)
{
    Resume();
}

//---------------------------------------------------------------------------
//  LogicLoadThread::LogicLoadThread() -- Ctor
//---------------------------------------------------------------------------

LogicLoadThread::~LogicLoadThread()
{
	delete Image;
	Image = 0;
}

//---------------------------------------------------------------------------
//  LogicLoadThread::Init() -- Initialize thread
//---------------------------------------------------------------------------

void LogicLoadThread::Init()
{
    inherited::Init();
	AddObject(Configuring, sConfiguring);
}

//---------------------------------------------------------------------------
//  LogicLoadThread::Configure() -- Write buffer to logic
//---------------------------------------------------------------------------

void LogicLoadThread::Configure(std::string file)
{
    PathSpec Path(file);

    delete Image;
    if (Path.Ext() == ".exo")
        Image = new SRecord;
    else
        Image = new BitFile;

    Image->FileName(file);

    Image->OnProgress.SetEvent(this, &LogicLoadThread::HandleParseProgress);
    Image->OnComplete.SetEvent(this, &LogicLoadThread::HandleParseComplete);
    Image->OnError.SetEvent(this, &LogicLoadThread::HandleParseError);

    Configuring.Set();
}

//---------------------------------------------------------------------------
//  LogicLoadThread::Load() --
//---------------------------------------------------------------------------

void LogicLoadThread::Load(char * buffer, size_t size)
{
    // Clear out FPGA image
    // INIT state should assert for a while
	bool IsInit = Fpga.IsInit();
	double Then = Rtc.Now() + 2.0;    // in secs
	Fpga.Program(true);
	while (!IsInit && (Rtc.Now() < Then))
		IsInit = Fpga.IsInit();

	if (!IsInit)
        {
        Error("INIT failed to assert following PROG");
        return;
        }

    Fpga.Program(false);
    IsInit = Fpga.IsInit();
    Then = Rtc.Now() + 2.0;    // in secs
	while (IsInit && (Rtc.Now() < Then))
        IsInit = Fpga.IsInit();

    if (IsInit)
        {
        Error("INIT failed to deassert following PROG");
        return;
        }

    // DONE should be deasserted initially
    if (Fpga.IsDone())
        {
        Error("Fpga done too early");
        return;
        }

    // Transmit the data
	const size_t BlockSize = size/100;
    size_t Residual = size;
    size_t i = 0;
    size_t chunk = 0;

    while (Residual)
        {
        // Write a chunk of EXO image to FPGA load port
        size_t ChunkSize = std::min(Residual, BlockSize);
		Fpga.Load(&buffer[i], ChunkSize);

		// Index to next chunk
		i += ChunkSize;
		Residual -= ChunkSize;
		++chunk;

		// Display progress
		int Percent = static_cast<int>((i*100)/size);
		Progress(Percent);

		// Quit if initializing during configuration
		if (Residual && Fpga.IsInit())
			{
			std::ostringstream msg;
            msg << "Fpga INIT active at chunk " << chunk;

            Error(msg.str());
            return;
            }
        }

    // Wait up to 2 seconds for Done bit to assert
    Then = Rtc.Now() + 2.0;    // in secs
	while (!Fpga.IsDone() && (Rtc.Now() < Then))
        Sleep(1);

    // Notify app after logic is loaded
    if (Fpga.IsDone())
        {
        Complete(0);
        }
    else
        {
        Error("Fpga load failure");
		return;
        }

    // Reset logic when load completes
    Sleep(1000);
    Fpga.Reset();
}

//---------------------------------------------------------------------------
//  LogicLoadThread::Load() --
//---------------------------------------------------------------------------

void LogicLoadThread::Load()
{
	Load(reinterpret_cast<char *>(&Image->Buffer[0]), Image->Size());
}

//------------------------------------------------------------------------------
//  LogicLoadThread::Progress() --
//------------------------------------------------------------------------------

void LogicLoadThread::Progress(int percent)
{
    ProcessProgressEvent e(percent);
    OnLoadProgress.Execute(e);
}

//------------------------------------------------------------------------------
//  LogicLoadThread::Complete() --
//------------------------------------------------------------------------------

void LogicLoadThread::Complete(int result)
{
    ProcessCompletionEvent e(result);
    OnLoadComplete.Execute(e);
}

//------------------------------------------------------------------------------
//  LogicLoadThread::Error() --
//------------------------------------------------------------------------------

void LogicLoadThread::Error(const std::string & msg)
{
    ProcessStatusEvent e(msg);
    OnLoadError.Execute(e);
}

//------------------------------------------------------------------------------
//  LogicLoadThread::HandleParseProgress() --
//------------------------------------------------------------------------------

void LogicLoadThread::HandleParseProgress(ProcessProgressEvent & event)
{
    OnParseProgress.Execute(event);
}

//------------------------------------------------------------------------------
//  LogicLoadThread::HandleParseComplete() --
//------------------------------------------------------------------------------

void LogicLoadThread::HandleParseComplete(ProcessCompletionEvent & event)
{
    OnParseComplete.Execute(event);
}

//------------------------------------------------------------------------------
//  LogicLoadThread::HandleParseError() --
//------------------------------------------------------------------------------

void LogicLoadThread::HandleParseError(ProcessStatusEvent & event)
{
    OnParseError.Execute(event);
}

//---------------------------------------------------------------------------
//  LogicLoadThread::Execute() --
//---------------------------------------------------------------------------

void LogicLoadThread::Execute()
{
    while (!Terminated())
        {
        // WaitResult is either a) Suicide event, b) Configure event
        int WaitResult = WaitFor();

        if (WaitResult == sConfiguring)
            {
            Image->Parse();
            Load();
            Configuring.Clear();
            }
        }
}


}   // namespace Innovative

