// Magic_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <Magic_Mb.h>
#include <limits>
#include <cfloat>
#include <cmath>
#include <algorithm>

using namespace std;

namespace Innovative
{

//---------------------------------------------------------------------------
// Population Count (Ones Count)
// The population count of a binary integer value x is the number of one bits in the value.
// Although many machines have single instructions for this, the single instructions are
// usually microcoded loops that test a bit per cycle; a log-time algorithm coded in C is
// often faster. The following code uses a variable-precision SWAR algorithm to perform 
// a tree reduction adding the bits in a 32-bit value:
//---------------------------------------------------------------------------

unsigned int Ones(unsigned int x)
{
	// 32-bit recursive reduction using SWAR...  but first step is mapping 2-bit values
	// into sum of 2 1-bit values in sneaky way
	x -= ((x >> 1) & 0x55555555);
	x = (((x >> 2) & 0x33333333) + (x & 0x33333333));
	x = (((x >> 4) + x) & 0x0f0f0f0f);
	x += (x >> 8);
	x += (x >> 16);

	return(x & 0x0000003f);
}

//---------------------------------------------------------------------------
// Given a binary integer value x, the floor of the base 2 log of that number efficiently
// can be computed by the application of two variable-precision SWAR algorithms. The
// first "folds" the upper bits into the lower bits to construct a bit vector with the 
// same most significant 1 as x, but all 1's below it. The second SWAR algorithm is 
// population count, defined elsewhere in this document.
// However, we must consider the issue of what the log2(0) should be; the log of 0 is undefined,
// so how that value should be handled is unclear. The following code for handling a 32-bit value
// gives two options: if LOG0UNDEFINED, this code returns -1 for log2(0); otherwise,
// it returns 0 for log2(0). For a 32-bit value:
//---------------------------------------------------------------------------

unsigned int FloorLog2(register unsigned int x)
{
	x |= (x >> 1);
	x |= (x >> 2);
	x |= (x >> 4);
	x |= (x >> 8);
	x |= (x >> 16);

#ifdef	LOG0UNDEFINED
	return(Ones(x) - 1);
#else
	return(Ones(x >> 1));
#endif
}

//---------------------------------------------------------------------------
// Suppose instead that you want the ceiling of the base 2 log.
// The floor and ceiling are identical if x is a power of two; otherwise,
// the result is 1 too small. This can be corrected using the power of 2
// test followed with the comparison-to-mask shift used in integer
// minimum/maximum. The result is:
//---------------------------------------------------------------------------

unsigned int CeilLog2(register unsigned int x)
{
    static const int WORDBITS = 32;
	register int y = (x & (x - 1));

    y |= -y;
    y >>= (WORDBITS - 1);
    x |= (x >> 1);
    x |= (x >> 2);
    x |= (x >> 4);
    x |= (x >> 8);
    x |= (x >> 16);

#ifdef	LOG0UNDEFINED
    return(Ones(x) - 1 - y);
#else
	return(Ones(x >> 1) - y);
#endif
}

//---------------------------------------------------------------------------
// Next Largest Power of 2
// Given a binary integer value x, the next largest power of 2 can be computed
// by a SWAR algorithm that recursively "folds" the upper bits into the lower
// bits. This process yields a bit vector with the same most significant 1 as x,
// but all 1's below it. Adding 1 to that value yields the next largest power of 2.
// For a 32-bit value:
//---------------------------------------------------------------------------

unsigned int NextLargestPowerOfTwo(register unsigned int x)
{
	x |= (x >> 1);
	x |= (x >> 2);
	x |= (x >> 4);
	x |= (x >> 8);
	x |= (x >> 16);

	return(x+1);
}

//---------------------------------------------------------------------------
// Next Smallest Power of 2 --  Return greatest power-of-two <= x

unsigned int NextSmallestPowerOfTwo(register unsigned int x)
{
    unsigned int exp = FloorLog2(x);
    return 1 << exp;
}

//---------------------------------------------------------------------------
// Bit Reversal
// Reversing the bits in an integer x is somewhat painful, but here's a SWAR
// algorithm for a 8-bit value:
//---------------------------------------------------------------------------

unsigned char BitReverse(register unsigned char x)
{
	x = (((x & 0xaaaaaaaa) >> 1) | ((x & 0x55555555) << 1));
	x = (((x & 0xcccccccc) >> 2) | ((x & 0x33333333) << 2));
	x = (((x & 0xf0f0f0f0) >> 4) | ((x & 0x0f0f0f0f) << 4));

	return x;
}

//---------------------------------------------------------------------------
// Bit Reversal algorithm for a 16-bit value:
//---------------------------------------------------------------------------

unsigned short BitReverse(register unsigned short x)
{
	x = (((x & 0xaaaaaaaa) >> 1) | ((x & 0x55555555) << 1));
	x = (((x & 0xcccccccc) >> 2) | ((x & 0x33333333) << 2));
	x = (((x & 0xf0f0f0f0) >> 4) | ((x & 0x0f0f0f0f) << 4));
	x = (((x & 0xff00ff00) >> 8) | ((x & 0x00ff00ff) << 8));

	return x;;
}

//---------------------------------------------------------------------------
// Bit Reversal algorithm for a 32-bit value:
//---------------------------------------------------------------------------

unsigned int BitReverse(register unsigned int x)
{
	x = (((x & 0xaaaaaaaa) >> 1) | ((x & 0x55555555) << 1));
	x = (((x & 0xcccccccc) >> 2) | ((x & 0x33333333) << 2));
	x = (((x & 0xf0f0f0f0) >> 4) | ((x & 0x0f0f0f0f) << 4));
	x = (((x & 0xff00ff00) >> 8) | ((x & 0x00ff00ff) << 8));

	return((x >> 16) | (x << 16));
}

//---------------------------------------------------------------------------
// Most Significant 1 Bit
// Given a binary integer value x, the most significant 1 bit (highest numbered
// element of a bit set) can be computed using a SWAR algorithm that recursively
// "folds" the upper bits into the lower bits. This process yields a bit vector
// with the same most significant 1 as x, but all 1's below it. Bitwise AND of
// the original value with the complement of the "folded" value shifted down by
// one yields the most significant bit. For a 32-bit value:
//---------------------------------------------------------------------------

unsigned int MostSignificantOneBit(register unsigned int x)
{
	x |= (x >> 1);
	x |= (x >> 2);
	x |= (x >> 4);
	x |= (x >> 8);
	x |= (x >> 16);

	return(x & ~(x >> 1));
}

//---------------------------------------------------------------------------
// Trailing Zero Count
// Given the Least Significant 1 Bit and Population Count (Ones Count) algorithms,
// it is trivial to combine them to construct a trailing zero count 
// (as pointed-out by Joe Bowbeer):
//---------------------------------------------------------------------------

unsigned int TrailingZeroCount(register int x)
{
	return(Ones((x & -x) - 1));
}

//---------------------------------------------------------------------------
// Counting bits set, Brian Kernighan's way
//---------------------------------------------------------------------------

unsigned int BitsSet(unsigned int v)
{
	unsigned int c; // c accumulates the total bits set in v
	for (c = 0; v; c++)
		v &= v - 1; // clear the least significant bit set

	return c;
}

//---------------------------------------------------------------------------
// LogN() -- Logarithm of general base, n.
//---------------------------------------------------------------------------

double LogN(double n, double x)
{
    return (x <= 0.0) ? -DBL_MAX : log(x)/log(fabs(n));
}

//-----------------------------------------------------------------------------
//  Lcm()  --  Return least common multiple
//-----------------------------------------------------------------------------

int Lcm(int a, int b)
{
	return a / Gcd(abs(a), abs(b)) * b;
}

//-----------------------------------------------------------------------------
//  RelativePrime()  --  Test for relative prime
//-----------------------------------------------------------------------------

bool RelativePrime(int a, int b)
{
	return Gcd(a, b) == 1;
}

//---------------------------------------------------------------------------
// RationalApproximation::Ctor
//
// Given a floating point number, finds the 'best' rational approximation
// (by using continued fractions).
//---------------------------------------------------------------------------

RationalApproximation::RationalApproximation(float what, int bits)
	: range(1 << (bits-1))
{
	bool invert = false;

	if (what > 1.0f)
		{
		what = 1.0f/what;
		invert = true;
		}

	try
		{
		double a,b;
		int c,d,e,f,mult;

		a=what;
		b=1.0;
		c=1;
		d=0;
		e=0;
		f=1;

		for (;;)
			{
			if (b != 0.0)
				mult=static_cast<int>(a/b);
			else
				mult=numeric_limits<int>::max();
			a-=mult*b;
			c-=mult*d;
			e-=mult*f;
			swap(num[1], num[0]);
			swap(den[1], den[0]);
			num[1]=-e;
			den[1]=c;
			ratio=num[1]/den[1];

			if ((num[1] > range) || (den[1] > range))
				break;

			if (a == 0.0)
				mult=numeric_limits<int>::max();
			else
				mult=static_cast<int>(b/a);

			b-=mult*a;
			d-=mult*c;
			f-=mult*e;
			swap(num[1], num[0]);
			swap(den[1], den[0]);
			num[1]=f;
			den[1]=-d;
			if (den[1] != 0.0)
				ratio=num[1]/den[1];
			else
				ratio=numeric_limits<double>::max();

			if ((num[1] > range) || (den[1] > range))
				break;
			}
		}
	catch(...)
		{
		}

	if (invert)
		swap(num[0], den[0]);
}

//---------------------------------------------------------------------------
// RationalApproximation::Numerator
//---------------------------------------------------------------------------

int  RationalApproximation::Numerator()
{
	return static_cast<int>(floor(num[0]+0.5f));
}

//---------------------------------------------------------------------------
// RationalApproximation::Denominator
//---------------------------------------------------------------------------

int RationalApproximation::Denominator()
{
	return static_cast<int>(floor(den[0]+0.5f));
}

//---------------------------------------------------------------------------
// RationalApproximation::Ratio
//---------------------------------------------------------------------------

double RationalApproximation::Ratio()
{
	return static_cast<double>(Numerator())/Denominator();
}

} // namespace Innovative
