//---------------------------------------------------------------------------
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2015 Innovative Integration Corporation.
//  All Rights Reserved.
//

#ifndef MultiChanStreamSupport_MbH
#define MultiChanStreamSupport_MbH

//#include <StreamDeviceEvents_Mb.h>
#include <Buffer_Mb.h>
//#include <vector>
#include <SafeLink_Mb.h>
#include <Trace_Mb.h>

#ifdef __CLR_VER
#pragma managed(push, off)
#endif


namespace InnovativeKernel
{

//=============================================================================
//  CLASS IMultiChanOwnerDM  -- Interface pointer to owning board device map
//=============================================================================

class IMultiChanOwnerDM
{
public:
    virtual ~IMultiChanOwnerDM()
        {}

    virtual void     TraceLog(int type, int data) = 0;
    virtual Innovative::Trace &  TextLog() = 0;
};

}  // end namespace InnovativeKernel


namespace Innovative
{
//==============================================================================
//  Template CLASS DataPack  --  "variable" sized array-like object with no allocation
//==============================================================================

template <class TElement, size_t nMaxCount>
class DataPack
{
public:
    //
    //  Member Functions
    DataPack()
      : Count(0), MaxCount(nMaxCount)
        {
        }

    bool  empty() const
        {
        return Count==0;
        }

    size_t size() const
        {
        return Count;
        }

    void  clear()
        {
        Count = 0;
        }

    void resize(size_t newCount)
        {
        if (newCount>MaxCount)
            Count = MaxCount;
        else
            Count = newCount;
        }

    TElement &  operator[]( size_t pos )
        {
        return Pack[pos];
        }

    const TElement & operator[]( size_t pos ) const
        {
        return Pack[pos];
        }

    void  push_back(const TElement& value)
        {
        if (Count==MaxCount)
            return;
        Pack[Count++] = value;
        }

protected:
    //
    //  Member Data
    TElement      Pack[nMaxCount];
    size_t        Count;
    const size_t  MaxCount;

};

const unsigned int kMultiChanChannelMaxX = 8;
const unsigned int kMultiChanFifoMax    = 512;
const unsigned int kMcImageHeaderWords  = 1;
const unsigned int kMcAlertSizeWords    = 4;  //? average
const unsigned int kMcExternalIntWords  = 1;

const unsigned int kMaxMcImages = kMultiChanFifoMax/kMcImageHeaderWords;
const unsigned int kMaxMcAlerts = kMultiChanFifoMax/kMcAlertSizeWords;
const unsigned int kMaxMcExternalInts = kMultiChanFifoMax/kMcExternalIntWords;


class  McImagePack : public DataPack<McImage, kMaxMcImages>
{
public:
    // use default ctor, dtor
};

class  McAlertPack : public DataPack<McAlert, kMaxMcAlerts>
{
public:
    // use default ctor, dtor
};

class  McExternalIntPack : public DataPack<McExternalInt, kMaxMcExternalInts>
{
public:
    // use default ctor, dtor
};

class  McHintPack : public DataPack<bool, kMultiChanChannelMaxX>
{
public:
    McHintPack() {  resize(MaxCount);  }

    void  Reset()
        {
        for (size_t i=0; i<size(); i++)
            Pack[i] = false;
        }
};

//
//  Some Common Info Passing Classes for MultiChan streaming
//==============================================================================
//  CLASS McImageInformation  -- Per-channel Status Information
//==============================================================================

class McImageInformation
{
public:
    size_t   Channel;
    size_t   TotalFreeSpace;
    size_t   MaxImageSize;
};

//==============================================================================
//  CLASS IMultiChanConfiguration  -- MultiChan Config Interface
//==============================================================================

class IMultiChanConfiguration
{
public:
    virtual void  SetAllBusMasterSize(ii32 words) = 0;  // set all channels
    virtual void  BusMasterSize(int chan, ii32 words) = 0;
    virtual ii32  BusMasterSize(int chan) = 0;

    virtual void  SetAllNotifySize(ii32 words) = 0;  // set all channels
    virtual void  NotifySize(int chan, ii32 words) = 0;
    virtual ii32  NotifySize(int chan) = 0;

    virtual ii32  MaxMcChannels() = 0;  // absolute system max stream channels
    virtual ii32  McChannels() = 0;     // "in use" max stream channels
    virtual bool  McChannelActive(int ch) = 0;   // true if channel is "in use" and analog ch enabled

    virtual ~IMultiChanConfiguration()
        {}
};

//==============================================================================
//  CLASS IMultiChanInitialization  -- Internal "setters" for MC system
//==============================================================================

class IMultiChanInitialization
{
public:
    virtual void  SetMcChannels(ii32 channels) = 0;        // Set total number of "In Use" McChannels
    virtual void  SetUnusedMcChannelsSize(ii32 words) = 0; // Only allocate a (small) size for unused streams
    virtual void  SetMcChannelActive(int chan, bool isActive) = 0;  // set to follow analog channel enable

    virtual ~IMultiChanInitialization()
        {}
};


//==============================================================================
//  CLASS IMultiChanStreamManagerInternals  -- MultiChan Streaming Internal Interface
//==============================================================================

class IMultiChanStreamManagerInternals
{
public:
    //  Direct Packet/Imange Transfer
    virtual void ForwardAlertPacket(McAlert & pkt) = 0;
    virtual void ForwardExternIntPacket(McExternalInt & pkt) = 0;

    //  Image Mode Handlers
    virtual void ForwardImagePacket(McImage & pkt) = 0;
    virtual void ForwardImageTransmitAck(McHintPack & OutgoingHints) = 0;

    //  Buffer "Mode" Handlers
    virtual void ForwardDataPacket(McBuffer & pkt) = 0;

    virtual bool CopyIncomingData() = 0;    // do we want actual buffers?
    virtual bool CopyOutgoingData() = 0;    // do we want actual buffers?
    virtual bool IsIncomingQueueOverThreshold() = 0;    // queue too full?

    // Buffer Mode Handlers - Transmit
    virtual size_t    ChannelXmitQ_Packets(int channel) = 0;
    virtual size_t    ChannelXmitQ_PendingNotifies(int channel) = 0;
    virtual void      ChannelXmitQ_Signal(int channel, unsigned int count) = 0;
    virtual size_t    ChannelXmitQ_NextPacketSize(int channel) = 0;
    virtual McBuffer  ChannelXmitQ_DequeueNextPacket(int channel) = 0;

    virtual void ForwardPacketTransmitAck(McHintPack & OutgoingHints) = 0;

    virtual bool HasStreamStarted() = 0;    // alias of StreamStarted()

    virtual ~IMultiChanStreamManagerInternals()
    	{}

};

//==============================================================================
//  CLASS IMultiChanStreamManager  -- MultiChan Streaming Public Interface
//==============================================================================

class IMultiChanStreamManager
{
public:
//    virtual void  Send( short PeriphId, Buffer & MultiChan ) = 0;
//    virtual void  Send( Buffer & MultiChan ) = 0;
//    virtual void  Recv( Buffer & MultiChan ) = 0;

//    virtual bool ConnectTo( IMultiChanStreamDevice *Device ) = 0;
//    virtual void Disconnect() = 0;
//    virtual bool StreamStarted() = 0;

    virtual IMultiChanStreamManagerInternals *  Internals() = 0;
    virtual ~IMultiChanStreamManager()
    	{}

};

//==============================================================================
//  CLASS IMultiChanStreamDeviceInternals  -- MultiChan Device Internal Interface
//==============================================================================

class IMultiChanStreamDeviceInternals
{
public:
    //
    //  Stream/Configuration "Time" Activations
    virtual void  AtPreconfigureTime() = 0;
    virtual void  AtStreamConfigureTime() = 0;
    virtual void  AtStreamPrefillTime() = 0;
    virtual void  AtStreamStartTime() = 0;
    virtual void  AtStreamStopTime() = 0;
    virtual void  AtDataTransmitTime(unsigned int channel) = 0;
    virtual void  AtDataReceivedTime() = 0;

    virtual bool                TxMcChannelActive(int ch) = 0;  // returns true if McChannel is used this run
    virtual McImageInformation  TxChannelInformation(size_t channel) = 0;
    virtual McImage             TxGetEmptyImage(size_t channel, size_t desired_size) = 0;
    virtual void                TransferFilledImage(McImage& image) = 0;
    virtual void                ForwardPacketTransmitAck(McHintPack & OutgoingHints)  = 0;

    virtual unsigned int  VisrIsrQDepth() = 0;

    virtual void  TraceLog(int type, int data) = 0;   // Same as MCSD->Log();

    virtual ~IMultiChanStreamDeviceInternals()
    	{}

};

//===========================================================================
//  CLASS IMultiChanStreamDevice  -- MultiChan Based Streaming device
//===========================================================================

class IMultiChanStreamDevice
{
public: // Pseudo-protected
    virtual void  RegisterStreamPlugIn( IMultiChanStreamManager *plug_in ) = 0;
//    virtual void  Reset() = 0;

    virtual IMultiChanStreamDeviceInternals *  Internals() = 0;
    virtual void  Log(int type, int data) = 0;
    virtual ~IMultiChanStreamDevice()
    	{}

};

//==========================================================================
//  CLASS SafeMultiChanStreamManagerInternalsImpl  -- "Null" Manager Internals
//==========================================================================

class SafeMultiChanStreamManagerInternalsImpl : public IMultiChanStreamManagerInternals
{
public:
    virtual void ForwardAlertPacket(McAlert & /*pkt*/) { X(); }
    virtual void ForwardExternIntPacket(McExternalInt & /*pkt*/) { X(); }
    virtual void ForwardImagePacket(McImage & /*pkt*/) { X(); }
    virtual void ForwardImageTransmitAck(McHintPack & /*OutgoingHints*/)  { X(); }

    virtual void ForwardDataPacket(McBuffer & /*pkt*/) { X(); }

    virtual bool CopyIncomingData() { X(); return false; }    // do we want actual buffers?
    virtual bool CopyOutgoingData() { X(); return false; }    // do we want actual buffers?
    virtual bool IsIncomingQueueOverThreshold() { X(); return false; }  // queue too full?

    // Buffer Mode Handlers - Transmit
    virtual size_t    ChannelXmitQ_Packets(int /*channel*/)  { X(); return 0; }
    virtual size_t    ChannelXmitQ_PendingNotifies(int /*channel*/)  { X(); return 0; }
    virtual void      ChannelXmitQ_Signal(int /*channel*/, unsigned int /*count*/)  { X(); }
    virtual size_t    ChannelXmitQ_NextPacketSize(int /*channel*/)  { X(); return 0; }
    virtual McBuffer  ChannelXmitQ_DequeueNextPacket(int /*channel*/) {  X();  McBuffer X;  return X; }

    virtual void ForwardPacketTransmitAck(McHintPack & /*OutgoingHints*/)  { X(); }

    virtual bool  HasStreamStarted() { X();  return false; }

private:
    Trace             Debug;
//    PmcBufferManager  Mgr;

    void X()
        {  Debug.Log("INVALID CALL to unconnected MultiChan StreamManagerInternals...: ");  }
};

//==============================================================================
//  CLASS SafeMultiChanStreamDeviceInternalsImpl  -- "Null" Manager
//==============================================================================
//
class SafeMultiChanStreamDeviceInternalsImpl : public IMultiChanStreamDeviceInternals
{
public:
    //
    //  Stream/Configuration "Time" Activations
    virtual void  AtPreconfigureTime() { X(); }
    virtual void  AtStreamConfigureTime() { X(); }
    virtual void  AtStreamPrefillTime() { X(); }
    virtual void  AtStreamStartTime() { X(); }
    virtual void  AtStreamStopTime() { X(); }
    virtual void  AtDataTransmitTime(unsigned int /*channel*/) { X(); }
    virtual void  AtDataReceivedTime() { X(); }

    virtual bool  TxMcChannelActive(int /*ch*/) { X(); return false; }

    virtual McImageInformation  TxChannelInformation(size_t channel)
        {
            McImageInformation x;
            x.Channel = channel;
            x.MaxImageSize = 0;
            x.TotalFreeSpace = 0;
            X();
            return x;
        }

    virtual McImage TxGetEmptyImage(size_t /*channel*/, size_t /*desired_size*/)
        {
        McImage x;
        X();
        return x;
        }

    virtual void  TransferFilledImage(McImage& /*image*/) { X(); }
    virtual void  ForwardPacketTransmitAck(McHintPack & /*OutgoingHints*/)  { X(); }

    virtual unsigned int  VisrIsrQDepth() { X();  return 0; }

    virtual void  TraceLog(int /*type*/, int /*data*/) { X(); }   // Same as MCSD->Log();

private:
    Trace                    Debug;

    void X() const
        {  Debug.Log("INVALID CALL to unconnected MultiChan StreamDeviceInternals...: ");  }

};

//==============================================================================
//  CLASS SafeMultiChanStreamManagerImpl  -- "Null" Manager
//==============================================================================

class SafeMultiChanStreamManagerImpl : public IMultiChanStreamManager
{
public:

//    virtual void  Send( short /*PeriphId*/, Buffer & /*MultiChan*/ ) { X(); };
//    virtual void  Send( Buffer & /*MultiChan*/ ) { X(); };
//    virtual void  Recv( Buffer & /*MultiChan*/ ) { X(); };

//    virtual bool ConnectTo( IMultiChanStreamDevice * /*Device*/ ) { X();  return false; };
//    virtual void Disconnect() { X(); };
//    virtual bool StreamStarted() { X();  return false;  };

   virtual IMultiChanStreamManagerInternals *  Internals() { X();  return &internals;  };

private:
    Trace                                    Debug;
    SafeMultiChanStreamManagerInternalsImpl  internals;

    void X()
        {  Debug.Log("INVALID CALL to unconnected MultiChan StreamManager...: ");  }

};

//===========================================================================
//  CLASS SafeMultiChanStreamDeviceImpl  -- "Null" Device
//===========================================================================

class SafeMultiChanStreamDeviceImpl : public IMultiChanStreamDevice
{
public: // Pseudo-protected
    virtual void  RegisterStreamPlugIn( IMultiChanStreamManager * /*plug_in*/ ) { X(); };
    //virtual void  Reset() { X(); };

    virtual IMultiChanStreamDeviceInternals *  Internals() { X(); return &internals; };
    virtual void  Log(int /*type*/, int /*data*/) { X(); };

private:
    Trace                               Debug;
    SafeMultiChanStreamDeviceInternalsImpl internals;

    void X()
        {  Debug.Log("INVALID CALL to unconnected MultiChan StreamDevice...: ");  }

};


class IIsAMultiChanStreamBoard;

//===========================================================================
// CLASS MultiChanStreamManagerSafeLink  --  Safe Linkage to a device
//===========================================================================

class MultiChanStreamManagerSafeLink : public SafeLink<IMultiChanStreamManager, SafeMultiChanStreamManagerImpl>
{
public:
    MultiChanStreamManagerSafeLink(IIsAMultiChanStreamBoard * owner)
        :  Owner(owner)
        {}

private:
    IIsAMultiChanStreamBoard *  Owner;

    virtual void  OnAfterLink();        // overrides
    virtual void  OnBeforeUnlink();
};

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif
//===========================================================================
//  CLASS IIsAMultiChanStreamBoard  -- MultiChanStream linkage interface class
//===========================================================================
class MultiChanDeviceMap;

class IIsAMultiChanStreamBoard : public IMultiChanStreamDevice
{
    friend class MultiChanStreamManagerSafeLink;
public:
    IIsAMultiChanStreamBoard(MultiChanDeviceMap & dev_map)
        :  IAP_PDevMap(dev_map), StreamManager(this)
        {};

public:  // pseudo-protected
    virtual void  RegisterStreamPlugIn( IMultiChanStreamManager *plug_in );

protected:
    MultiChanDeviceMap &            IAP_PDevMap;
    MultiChanStreamManagerSafeLink  StreamManager;
};

#ifdef _MSC_VER
	#pragma warning(pop)
#endif

#ifdef __CLR_VER
#pragma managed(pop)
#endif
}  // end namespace Innovative

#endif
