//---------------------------------------------------------------------------
// MultiChanStream_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2015 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <MultiChanStream_Mb.h>
#include <SystemSupport_Mb.h>
#include <StartStopThread_Mb.h>
#include <StopWatch_Mb.h>
#include <SystemSupport_Mb.h>
#include <BufferHeader_Mb.h>

namespace Innovative
{
//===========================================================================
// CLASS MultiChanStreamDeviceSafeLink  --  Safe Linkage to a device
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class MultiChanStream
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//  MultiChanStreamDeviceSafeLink::OnAfterLink() - handle post-link stuff
//---------------------------------------------------------------------------

void  MultiChanStreamDeviceSafeLink::OnAfterLink()
{
    Ptr()->RegisterStreamPlugIn(Owner);
}

//---------------------------------------------------------------------------
//  MultiChanStreamDeviceSafeLink::OnBeforeUnlink() - handle pre-unlink stuff
//---------------------------------------------------------------------------

void  MultiChanStreamDeviceSafeLink::OnBeforeUnlink()
{
    if (Owner->StreamStarted())
        Owner->Stop();

    Ptr()->RegisterStreamPlugIn(0);
}

//===========================================================================
//  CLASS MultiChanStreamRcvrThread  -- Receiver thread for MC Streams
//===========================================================================
//  Threads handling the 'application' side of the buffer queues from the
//    VISR. They handle the interface between these queues and the app

class MultiChanStreamRcvrThread : public BasicSSThread
{
    friend class MultiChanStream;
    typedef BasicSSThread inherited;

public:
    // Ctor
    MultiChanStreamRcvrThread(MultiChanStream * owner)
        : inherited("MultiChanStreamIoThread"), Owner(owner), Queue(owner->RecvQ)
        {  Priority(tpHigher);  }

    // Methods
    void  Clear()
        {  Signal.Clear();  }
    void  Recv(McBuffer & Packet);

protected:
    Semaphore       Signal;

    virtual void    ProcessPacket();

    class LoadBalance
    {
    public:
        // Ctor
        LoadBalance()
            : Then(0), Sleeps(0), MultiChans(0)
            {  }

        // Load balancing:  Reserve 10% of CPU for foreground processing
        void Balance()
        {
            ++MultiChans;
            if ((Time.Now() - Then) > 10.0e-3)
                {
                Sleep(1);
                ++Sleeps;
                Then = Time.Now();
                }
        }

        // Data
        double          Then;
        StopWatch       Time;
        unsigned int    Sleeps;
        unsigned int    MultiChans;
    };

    LoadBalance         Load;
    MultiChanStream *   Owner;

    MultiChanStreamQueue &       Queue;

    enum { sQueue=inherited::sLast, sLast };
    //  override to load notify events at start
    virtual void    Init()
        {
        inherited::Init();
        AddObject(Queue.SignalSem(), sQueue);
        }
    //  do one iteration of main process
    virtual void    AtExecution(int result)
        {
        if (result == sQueue)
            ProcessPacket();
        }

    void            Log(int tag, int value)
        {  Owner->DeviceInternals()->TraceLog(tag, value);  }
    void            Log(int tag, size_t value)
        {  Log(tag, static_cast<int>(value));  }
};

//---------------------------------------------------------------------------
//  MultiChanStreamRcvrThread::ProcessPacket() -- Deliver Packet to App
//---------------------------------------------------------------------------

void  MultiChanStreamRcvrThread::ProcessPacket()
{
//    int depth_check = Queue.Depth();
//    Log(8, depth_check);

    //
    //  Signal MultiChan available downstream...
    Signal.Release();

    Owner->SignalDataAvailable();

    // Load balancing
    if (Owner->RxLoadBalancing())
        Load.Balance();

//    depth_check = Queue.Depth();
//    Log(9, depth_check);
}

//---------------------------------------------------------------------------
//  MultiChanStreamRcvrThread::Recv() -- Return MultiChan in data queue
//---------------------------------------------------------------------------

void  MultiChanStreamRcvrThread::Recv(McBuffer& Packet )
{
    int depth_check = static_cast<int>(Queue.Depth());
    Log(0x18, depth_check);

    if (depth_check>0x100)
        {
        //  Spinning out of control now?
        Log(0x1f, depth_check);
        }
    //
    //  Wait for packet to arrive....
    Signal.Acquire();
    //
    //  Extract packet from queue
    Packet = Queue.Front();
    //
    //  Pop the queue
    Queue.Pop();

    depth_check = static_cast<int>(Queue.Depth());
    Log(0x19, depth_check);
}

//===========================================================================
//  CLASS MultiChanStreamXmitThread  -- Transmit thread for MC Streams
//===========================================================================
//  This thread is on the 'application' side of the buffer queues from the
//    VISR. It handles the interface from the App to the Queues bound to the
//    Busmaster areas

//---------------------------------------------------------------------------
//  declaration of  class MultiChanStreamXmitThread
//---------------------------------------------------------------------------

class MultiChanStreamXmitThread : public BasicSSThread
{
    friend class MultiChanStream;
    typedef BasicSSThread inherited;

public:
    // Ctor
    MultiChanStreamXmitThread(MultiChanStream *owner)
        : inherited("MultiChanStreamIoThread"), Owner(owner), Queues(owner->XmitQ),
          Prefilling(false)
        {  Priority(tpHigher);  }

public:
    void  Send(McBuffer& Packet );
    void  Clear()
        {  Queues.ClearSem();  }

    void  Prefill(unsigned int channel, int packets);

protected:
    // Methods
    virtual void    ProcessPacket(unsigned int channel);
    virtual void    AtStop();
    void            Signal(unsigned int channel);


    class LoadBalance
    {
    public:
        // Ctor
        LoadBalance()
            : Then(0), Sleeps(0), MultiChans(0)
            {  }

        // Load balancing:  Reserve 10% of CPU for foreground processing
        void Balance()
        {
            ++MultiChans;
            if ((Time.Now() - Then) > 10.0e-3)
                {
                Sleep(1);
                ++Sleeps;
                Then = Time.Now();
                }
        }

        // Data
        double          Then;
        StopWatch       Time;
        unsigned int    Sleeps;
        unsigned int    MultiChans;
    };

    LoadBalance         Load;
    MultiChanStream *   Owner;

    MC_XmitQPool &      Queues;
    bool            Prefilling;


    enum { sQueue_0=inherited::sLast,  sQueue_1,  sQueue_2,  sQueue_3,
           sQueue_4,   sQueue_5,   sQueue_6,   sQueue_7,  sLast };
    //  override to load notify events at start
    virtual void    Init()
        {
        inherited::Init();
        AddObject(Queues.Queue(0).SignalSem(), sQueue_0);
        AddObject(Queues.Queue(1).SignalSem(), sQueue_1);
        AddObject(Queues.Queue(2).SignalSem(), sQueue_2);
        AddObject(Queues.Queue(3).SignalSem(), sQueue_3);
        AddObject(Queues.Queue(4).SignalSem(), sQueue_4);
        AddObject(Queues.Queue(5).SignalSem(), sQueue_5);
        AddObject(Queues.Queue(6).SignalSem(), sQueue_6);
        AddObject(Queues.Queue(7).SignalSem(), sQueue_7);
        }
    //  do one iteration of main process
    virtual void    AtExecution(int result)
        {
        //  calculate ch index from result...
        unsigned int channel = result-sQueue_0;
        if (channel<Queues.QPoolSize())
            ProcessPacket(channel);
        }

    void            Log(int tag, int value)
        {  Owner->DeviceInternals()->TraceLog(tag, value);  }
    void            Log(int tag, size_t value)
        {  Log(tag, static_cast<int>(value));  }
};

//---------------------------------------------------------------------------
//  MultiChanStreamXmitThread::ProcessPacket() -- notify app to get packet
//---------------------------------------------------------------------------

void  MultiChanStreamXmitThread::ProcessPacket(unsigned int channel)
{
    if (Prefilling)
        {
        return;
        }

    Signal(channel);
}

//---------------------------------------------------------------------------
//  MultiChanStreamXmitThread::AtStop() -- Special Code when Stopped
//---------------------------------------------------------------------------

void  MultiChanStreamXmitThread::AtStop()
{
//    Log(0xec, 0x01);
    Clear();
//    Log(0xec, 0x00);
}

//---------------------------------------------------------------------------
//  MultiChanStreamXmitThread::Signal() -- packet needed by transmit
//---------------------------------------------------------------------------

void  MultiChanStreamXmitThread::Signal(unsigned int channel)
{
    //int depth_check = Queue.Depth();
    //Log(0xa1, depth_check);
    //
    //  Signal upstream for data...
    Owner->SignalDataRequired(channel);

    // Load balancing
    if (Owner->TxLoadBalancing())
        Load.Balance();

    //depth_check = Queue.Depth();
    //Log(0xa0, depth_check);
}

//---------------------------------------------------------------------------
//  MultiChanStreamXmitThread::Send() -- Load packet for transmission
//---------------------------------------------------------------------------

void  MultiChanStreamXmitThread::Send( McBuffer & Packet )
{
    McHeaderDatagram Hdr(Packet);
    unsigned int channel = Hdr.ChannelId();
    size_t depth_check = Owner->XmitQ.Queue(channel).Depth();
    Log(0x1a, depth_check);

    //
    //  Push data onto outgoing queue
    Owner->XmitQ.Queue(channel).Push(Packet);
    //Log(0x1c, depth_check);

    depth_check = Owner->XmitQ.Queue(channel).Depth();
    Log(0x1b, depth_check);
}

//---------------------------------------------------------------------------
//  MultiChanStreamXmitThread::Prefill() -- Wait for prefill to complete
//---------------------------------------------------------------------------

void  MultiChanStreamXmitThread::Prefill(unsigned int channel, int packets)
{
    Prefilling = true;
    for (int i = 0; i < packets; ++i)
        Signal(channel);
    Prefilling = false;
}

//===========================================================================
//  CLASS MultiChanStreamStoppageThread  -- Proper Termination signalling
//===========================================================================
//  We want a callback to happen when it is safe to call Stream:;Start() again
//    this thread watches the condition and calls the handler when ready

class MultiChanStreamStoppageThread : public BasicSSThread
{
//    friend class MultiChanStream;
    typedef BasicSSThread inherited;

public:
    // Ctor
    MultiChanStreamStoppageThread(MultiChanStream *owner)
        : inherited("MultiChanStreamStoppageThread"), Owner(owner)
        {  Priority(tpNormal);  }

    void  StartMonitoring()
            {  eMonitor.Set();  }
    void  EndMonitoring()
            {  eMonitor.Clear();  }

protected:

    MultiChanStream *  Owner;
    Event               eMonitor;

    enum { sMonitor=inherited::sLast, sLast };
    virtual void    Init()
        {
        inherited::Init();
        AddObject(eMonitor, sMonitor);
        }

    //  do one iteration of main process
    virtual void    AtExecution(int result);

    void            Log(int tag, int value)
        {  Owner->DeviceInternals()->TraceLog(tag, value);  }
    void            Log(int tag, size_t value)
        {  Log(tag, static_cast<int>(value));  }
};

//---------------------------------------------------------------------------
//  MultiChanStreamStoppageThread::AtExecution() -- Internal Interface
//---------------------------------------------------------------------------

void  MultiChanStreamStoppageThread::AtExecution(int /*result*/)
{
    Log(0xE3, 0x1);

    Owner->RecvThread->WaitForStopped_External();
    Log(0xE3, 0x8);
    Owner->XmitThread->WaitForStopped_External();
    Log(0xE3, 0x7);

    while (Owner->DeviceInternals()->VisrIsrQDepth() != 0)
        {
        Innovative::uSleep(100);
        Log(0xE3, 0x6);
        }

    Log(0xE3, 0x5);

    EndMonitoring();  // only call self once!

    Log(0xE3, 0x0);
    OpenWire::NotifyEvent e;
    Owner->OnAfterStop.Execute(e);
}


//===========================================================================
//  CLASS TIMultiChanInternalsImpl  -- Internal Interface Impl
//===========================================================================

class TIMultiChanInternalsImpl : public IMultiChanStreamManagerInternals
{
public:
    TIMultiChanInternalsImpl(MultiChanStream * owner)
        :  Owner(owner) {}

    virtual void ForwardAlertPacket(McAlert & pkt)  {  Owner->ForwardAlertPacket(pkt); }
    virtual void ForwardExternIntPacket(McExternalInt & pkt)  {  Owner->ForwardExternIntPacket(pkt); }
    virtual void ForwardImagePacket(McImage & pkt)  {  Owner->ForwardImagePacket(pkt); }
    virtual void ForwardImageTransmitAck(McHintPack & OutgoingHints) {  Owner->ForwardImageTransmitAck(OutgoingHints); }
    virtual void ForwardDataPacket(McBuffer & pkt) { Owner->ForwardDataPacket(pkt);  }

    virtual bool CopyIncomingData() { return Owner->CopyIncomingData(); }    // do we want actual buffers?
    virtual bool CopyOutgoingData() { return Owner->CopyOutgoingData(); }    // do we want actual buffers?
    virtual bool IsIncomingQueueOverThreshold() { return Owner->IsIncomingQueueOverThreshold();  }  // queue too full?

    virtual size_t    ChannelXmitQ_Packets(int channel) { return Owner->ChannelXmitQ_Packets(channel);  }
    virtual size_t    ChannelXmitQ_PendingNotifies(int channel)  { return Owner->ChannelXmitQ_PendingNotifies(channel); }
    virtual void      ChannelXmitQ_Signal(int channel, unsigned int count) { Owner->ChannelXmitQ_Signal(channel, count);  }
    virtual size_t    ChannelXmitQ_NextPacketSize(int channel) { return Owner->ChannelXmitQ_NextPacketSize(channel);  }
    virtual McBuffer  ChannelXmitQ_DequeueNextPacket(int channel) { return Owner->ChannelXmitQ_DequeueNextPacket(channel);  }

    virtual void ForwardPacketTransmitAck(McHintPack & OutgoingHints) { Owner->ForwardPacketTransmitAck(OutgoingHints);  }

    virtual bool  HasStreamStarted() { return Owner->StreamStarted(); }

private:
    MultiChanStream * Owner;
};


//===========================================================================
//  CLASS MultiChanStream  -- MultiChan Based Stream I/O
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class MultiChanStream
//---------------------------------------------------------------------------
#ifdef _MSC_VER
    #pragma warning(push)
    #pragma warning(disable : 4355)
#endif

MultiChanStream::MultiChanStream()
    : StreamDevice(this),
      InternalOps(new TIMultiChanInternalsImpl(this)),
      RecvThread(0),
      XmitThread(0),
      Started(false),
      FRecvPacketCache(10),
      FRxLoadBalancing(true), FTxLoadBalancing(true)//,
{
    RecvThread = new MultiChanStreamRcvrThread( this );
    RecvThread->Resume();

    XmitThread = new MultiChanStreamXmitThread( this );
    XmitThread->Resume();

    StopperThread = new MultiChanStreamStoppageThread( this );
    StopperThread->Resume();
}
#ifdef _MSC_VER
    #pragma warning(pop)
#endif

//---------------------------------------------------------------------------
//  destructor for class MultiChanStream
//---------------------------------------------------------------------------

MultiChanStream::~MultiChanStream()
{
    if (Started)
        Stop();

    RecvThread->Shutdown();
    delete RecvThread;
    RecvThread = 0;

    XmitThread->Shutdown();
    delete XmitThread;
    XmitThread = 0;

    StopperThread->Shutdown();
    delete StopperThread;
    StopperThread = 0;

    delete InternalOps;
    InternalOps = 0;
}

//---------------------------------------------------------------------------
//  MultiChanStream::Preconfigure() -- Non-repeated 'Manual' Configuration
//---------------------------------------------------------------------------

void  MultiChanStream::Preconfigure()
{
    DeviceInternals()->AtPreconfigureTime();
}

//---------------------------------------------------------------------------
//  MultiChanStream::Start() -- Start busmastering
//---------------------------------------------------------------------------

void  MultiChanStream::Start()
{
    if (Started)
        return;

    //  Call "OnBeforeStart" event
    OpenWire::NotifyEvent e;
    OnBeforeStart.Execute(e);

    RecvQ.QueueCapacity(FRecvPacketCache);

    XmitQ.Clear();
    RecvQ.Clear();
    RecvThread->Clear();
    //
    //  Start Receiver Thread
    RecvThread->Start();
    XmitThread->Start();
    StopperThread->Start();

    // Begin Busmastering
    DeviceInternals()->AtStreamConfigureTime();

    //  Prefill Outgoing Data
    Prefill();

    //
    //  Post-prefill Device Operations
    DeviceInternals()->AtStreamPrefillTime();

    // This flag must be set before starting the busmaster.
    Started = true;
    //
    //  Start busmastering
    DeviceInternals()->AtStreamStartTime();

    //  Call "OnAfterStart" event
    OpenWire::NotifyEvent e2;
    OnAfterStart.Execute(e2);

    StreamDevice->Log(0xFF, 0);
}

//---------------------------------------------------------------------------
//  MultiChanStream::Prefill() -- Activate prefill on transmit Q
//---------------------------------------------------------------------------

void  MultiChanStream::Prefill()
{
    for (unsigned int ch=0; ch<XmitQ.QPoolSize(); ch++)
        {
        XmitQ.Queue(ch).ClearSem();
        XmitThread->Prefill(ch, XmitQ.PrefillPacketCount(ch));
        XmitQ.Queue(ch).QueueCapacity(XmitQ.PrefillPacketCount(ch));
        }
}

//---------------------------------------------------------------------------
//  MultiChanStream::Stop() -- Stop busmastering
//---------------------------------------------------------------------------

void  MultiChanStream::Stop()
{
    if (!Started)
        return;

    //  Wakes up a stopper thread that will post OnAfterStop when safe
    StopperThread->StartMonitoring();

    Started = false;
    DeviceInternals()->AtStreamStopTime();
    RecvThread->Stop();
    XmitThread->Stop();
}

//---------------------------------------------------------------------------
//  MultiChanStream::SuspendOutput() -- Suspend output request for data
//---------------------------------------------------------------------------

void  MultiChanStream::SuspendOutput(unsigned int channel, bool state)
{
    XmitQ.Queue(channel).Suspend(state);
}

//---------------------------------------------------------------------------
//  MultiChanStream::Internals() -- Stream Manager Internal Interface
//---------------------------------------------------------------------------

IMultiChanStreamManagerInternals *  MultiChanStream::Internals()
{
    return InternalOps;
}

//---------------------------------------------------------------------------
//  MultiChanStream::DeviceInternals() -- Stream Device Internal Interface
//---------------------------------------------------------------------------

IMultiChanStreamDeviceInternals *  MultiChanStream::DeviceInternals()
{
    return StreamDevice->Internals();
}

//---------------------------------------------------------------------------
//  MultiChanStream::PrefillPacketCount() -- Set all prefill sizes
//---------------------------------------------------------------------------

void MultiChanStream::PrefillPacketCount(int packets)
{
    for (unsigned int ch=0; ch<XmitQ.QPoolSize(); ch++)
        {
        if (DeviceInternals()->TxMcChannelActive(ch))
            XmitQ.PrefillPacketCount(ch, packets);
        else
            XmitQ.PrefillPacketCount(ch, 0);  // if stream isn't active, don't prefill
        }
}

//---------------------------------------------------------------------------
//  MultiChanStream::PrefillPacketCount() -- Set one channels prefill sizes
//---------------------------------------------------------------------------

void MultiChanStream::PrefillPacketCount(unsigned int channel, int packets)
{
    XmitQ.PrefillPacketCount(channel, packets);
}

//---------------------------------------------------------------------------
//  MultiChanStream::PrefillPacketCount() -- read one channel's prefill level
//---------------------------------------------------------------------------

int MultiChanStream::PrefillPacketCount(unsigned int channel) const
{
    return XmitQ.PrefillPacketCount(channel);
}

//---------------------------------------------------------------------------
//  MultiChanStream::Send() -- Post MultiChan for transmission
//---------------------------------------------------------------------------

void  MultiChanStream::Send(McBuffer & Packet)
{
    //
    //  Read ChannelId
    McHeaderDatagram headerDG(Packet);
    unsigned int channel = headerDG.ChannelId();

    XmitThread->Send(Packet);

    DeviceInternals()->AtDataTransmitTime(channel);
}

//---------------------------------------------------------------------------
//  MultiChanStream::Recv() -- Return MultiChan received from host
//---------------------------------------------------------------------------

void  MultiChanStream::Recv(McBuffer& Packet)
{
    // Fill using base class method
    RecvThread->Recv(Packet);

    // Signal VISR Thread to credit a queued packet, if any
    DeviceInternals()->AtDataReceivedTime();
}

//---------------------------------------------------------------------------
//  MultiChanStream::TxChannelInformation() -- Info about TX channel image
//---------------------------------------------------------------------------

McImageInformation MultiChanStream::TxChannelInformation(size_t channel)
{
    // Get TX channel current state
    return DeviceInternals()->TxChannelInformation(channel);
}

//---------------------------------------------------------------------------
//  MultiChanStream::TxGetEmptyImage() -- Get Image for refilling
//---------------------------------------------------------------------------

McImage MultiChanStream::TxGetEmptyImage(size_t channel, size_t desired_size)
{
    return DeviceInternals()->TxGetEmptyImage(channel, desired_size);
}

//---------------------------------------------------------------------------
//  MultiChanStream::Send() -- Get Image for refilling
//---------------------------------------------------------------------------

void MultiChanStream::Send(McImage& image)
{
    DeviceInternals()->TransferFilledImage(image);
}

//---------------------------------------------------------------------------
//  MultiChanStream::SignalDataAvailable() -- Signal data available to app
//---------------------------------------------------------------------------

void  MultiChanStream::SignalDataAvailable()
{
    MultiChanStreamDataEvent e(this, 0);
    OnDataAvailable.Execute(e);
}

//---------------------------------------------------------------------------
//  MultiChanStream::SignalDataRequired() -- Signal data needed to app
//---------------------------------------------------------------------------

void  MultiChanStream::SignalDataRequired(unsigned int channel)
{
    MultiChanStreamDataEvent e(this, channel);
    OnDataRequired.Execute(e);
}

//---------------------------------------------------------------------------
//  MultiChanStream::ForwardAlertPacket() -- Signal Alert Data available
//---------------------------------------------------------------------------

void  MultiChanStream::ForwardAlertPacket(McAlert & pkt)
{
    MultiChanAlertEvent e(this, pkt);
    OnAlertAvailable.Execute(e);
}

//---------------------------------------------------------------------------
//  MultiChanStream::ForwardExternIntPacket() -- Signal Ext Int Data available
//---------------------------------------------------------------------------

void  MultiChanStream::ForwardExternIntPacket(McExternalInt & pkt)
{
    MultiChanExternalEvent e(this, pkt);
    OnExternalIntAvailable.Execute(e);
}

//---------------------------------------------------------------------------
//  MultiChanStream::ForwardImagePacket() -- Signal Image Packet available
//---------------------------------------------------------------------------

void  MultiChanStream::ForwardImagePacket(McImage & pkt)
{
    MultiChanDirectDataEvent e(this, pkt);
    OnDirectDataAvailable.Execute(e);
}

//---------------------------------------------------------------------------
//  MultiChanStream::ForwardImageTransmitAck() -- Signal Transmit Ack
//---------------------------------------------------------------------------

void  MultiChanStream::ForwardImageTransmitAck(McHintPack& OutgoingHints)
{
    MultiChanDirectTransmitAckEvent e(this, OutgoingHints);
    OnDirectDataRequired.Execute(e);
}

//---------------------------------------------------------------------------
//  MultiChanStream::ForwardPacketTransmitAck() -- Signal Transmit Ack
//---------------------------------------------------------------------------

void MultiChanStream::ForwardPacketTransmitAck(McHintPack& OutgoingHints)
{
    DeviceInternals()->ForwardPacketTransmitAck(OutgoingHints);
}

//---------------------------------------------------------------------------
//  MultiChanStream::ForwardDataPacket() -- Queue Data Packet
//---------------------------------------------------------------------------

void  MultiChanStream::ForwardDataPacket(McBuffer & pkt)
{
    RecvQ.Push(pkt);
}

//---------------------------------------------------------------------------
//  MultiChanStream::CopyIncomingData() -- do we want actual buffers?
//---------------------------------------------------------------------------
//  Use this to indicate if Int Scheme will create true copied buffers
//    for the application.  We use the fact that the ODA event is assigned a handler
//    to block that step.

bool  MultiChanStream::CopyIncomingData()
{
    return OnDataAvailable.Assigned();
}

//---------------------------------------------------------------------------
//  MultiChanStream::CopyOutgoingData() -- do we want actual buffers?
//---------------------------------------------------------------------------
//  Use this to indicate if Int Scheme will load true copied buffers
//    from the application.  We use the fact that the ODR event is assigned a handler
//    to enable that step.

bool MultiChanStream::CopyOutgoingData()
{
    return OnDataRequired.Assigned();
}

//---------------------------------------------------------------------------
//  MultiChanStream::IsIncomingQueueOverThreshold() -- do we want actual buffers?
//---------------------------------------------------------------------------
//  We need to know if this packet can be credited at once, or delayed to avoid
//    internal overflow. Delayed queuing stops up

bool  MultiChanStream::IsIncomingQueueOverThreshold()
{
    return static_cast<int>(RecvQ.Depth()) > FRecvPacketCache;
}

//---------------------------------------------------------------------------
//  MultiChanStream::ChannelXmitQ_Packets() -- return depth of Xmit Q
//---------------------------------------------------------------------------
//  Checks if packets are ready for transmission

size_t MultiChanStream::ChannelXmitQ_Packets(int channel)
{
    return XmitQ.Queue(channel).Depth();
}

//---------------------------------------------------------------------------
//  MultiChanStream::ChannelXmitQ_PendingNotifies() -- Q packets that wake
//---------------------------------------------------------------------------
//  If BM region fills, signals could be exhausted. This tells how many
//    are left.  If notifies < packets we need to take steps.

size_t MultiChanStream::ChannelXmitQ_PendingNotifies(int channel)
{
    return XmitQ.Queue(channel).SignalSem().Count();
}

//---------------------------------------------------------------------------
//  MultiChanStream::ChannelXmitQ_Signal() -- have Xmit Q signal for packets
//---------------------------------------------------------------------------
//  Notify Host that more packets are needed

void MultiChanStream::ChannelXmitQ_Signal(int channel, unsigned int count)
{
    //  If in "Async" mode, skip the signalling to avoid useless work.
    if (XmitQ.Queue(channel).QueueCapacity()!=0)
        XmitQ.Queue(channel).BumpSignal(count);
}

//---------------------------------------------------------------------------
//  MultiChanStream::ChannelXmitQ_NextPacketSize() -- return size of next pkt
//---------------------------------------------------------------------------

size_t MultiChanStream::ChannelXmitQ_NextPacketSize(int channel)
{
    if (XmitQ.Queue(channel).Depth()!=0)
        {
        return XmitQ.Queue(channel).Front().SizeInInts();
        }
    else
        return 0;
}


//---------------------------------------------------------------------------
//  MultiChanStream::ChannelXmitQ_DequeueNextPacket() -- pop & return pkt
//---------------------------------------------------------------------------

McBuffer MultiChanStream::ChannelXmitQ_DequeueNextPacket(int channel)
{
    //    size_t test_depth = XmitQ.Depth();
    //    //DeviceInternals()->TraceLog(0x60, test_depth);
    //    if (test_depth==0)
    //        {
    //        DeviceInternals()->TraceLog(0xF0, 0);
    //        while (1==1)
    //            Sleep(100);
    //        }
    //  ^Hang if no packet available...
    McBuffer rtn = XmitQ.Queue(channel).Front();
    XmitQ.Queue(channel).Pop();
    return rtn;
}

//---------------------------------------------------------------------------
//  MultiChanStream::StreamStarted() -- returns true if the stream is started
//---------------------------------------------------------------------------

bool MultiChanStream::StreamStarted()
{
    return Started;
}

//---------------------------------------------------------------------------
//  MultiChanStream::ConnectTo() -- connect stream to stream device
//---------------------------------------------------------------------------

bool  MultiChanStream::ConnectTo(IMultiChanStreamDevice * aStreamDevice)
{
    if (aStreamDevice)
        {
        StreamDevice.Link(aStreamDevice);
        return true;
        }
    else
        return false;
}

//---------------------------------------------------------------------------
//  MultiChanStream::Disconnect() -- disconnect stream from stream device
//---------------------------------------------------------------------------

void MultiChanStream::Disconnect()
{
    StreamDevice.Unlink();
}

//---------------------------------------------------------------------------
//  MultiChanStream::NativeBuffer() -- create native buffer type
//---------------------------------------------------------------------------

McBuffer  MultiChanStream::NativeBuffer(size_t data_size, bool autoinit)
{
    McBuffer  buffer(data_size);
    if (autoinit)
        {
        ClearHeader(buffer);
        InitHeader(buffer);
        }
    return buffer;
}



//===========================================================================
//  CLASS MultiChanStreamQueue  -- Queue to manage data from VISR to Receiver
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class MultiChanStreamQueue
//---------------------------------------------------------------------------

MultiChanStreamQueue::MultiChanStreamQueue()
    : CapacityInBlks(10)
{
}

//---------------------------------------------------------------------------
//  destructor for class MultiChanStreamQueue
//---------------------------------------------------------------------------

MultiChanStreamQueue::~MultiChanStreamQueue()
{
}

//---------------------------------------------------------------------------
//  MultiChanStreamQueue::Clear() -- clear queue
//---------------------------------------------------------------------------

void  MultiChanStreamQueue::Clear()
{
    //  clear buffer queue
    BufferQ.clear();
    //  reduce signal to 0
    Signal.Clear();
}

//---------------------------------------------------------------------------
//  MultiChanStreamQueue::Front() -- Front element in queue
//---------------------------------------------------------------------------

McBuffer MultiChanStreamQueue::Front()
{
    return BufferQ.front();
}

//---------------------------------------------------------------------------
//  MultiChanStreamQueue::Depth() -- depth of queue
//---------------------------------------------------------------------------

size_t MultiChanStreamQueue::Depth()
{
    return BufferQ.depth();
}

//---------------------------------------------------------------------------
//  MultiChanStreamQueue::QueueCapacity() -- Set queue capacity
//---------------------------------------------------------------------------

void  MultiChanStreamQueue::QueueCapacity(size_t cap)
{
    CriticalSectionArea guard(lock);
    CapacityInBlks = cap;
}

//---------------------------------------------------------------------------
//  MultiChanStreamQueue::QueueCapacity() -- Retreive capacity
//---------------------------------------------------------------------------

size_t MultiChanStreamQueue::QueueCapacity() const
{
    return CapacityInBlks;
}

//---------------------------------------------------------------------------
//  MultiChanStreamQueue::SignalSem() -- Return reference to semaphore
//---------------------------------------------------------------------------

Semaphore & MultiChanStreamQueue::SignalSem()
{
    return Signal;
}

//---------------------------------------------------------------------------
//  MultiChanStreamQueue::BumpSignal() -- Release Signal Sem
//---------------------------------------------------------------------------

void MultiChanStreamQueue::BumpSignal(unsigned int count)
{
    size_t total_blocks = Signal.Count() + Depth();
    if (total_blocks<CapacityInBlks)
        for (unsigned int i = 0; i < count; ++i)
            Signal.Release();
}

//===========================================================================
//  CLASS XmitMCQueue  -- Queue to manage data from Receiver to VISR
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class XmitMCQueue
//---------------------------------------------------------------------------

XmitMCQueue::XmitMCQueue()
    :  SuspendSignal(false)
{
}

//---------------------------------------------------------------------------
//  destructor for class XmitMCQueue
//---------------------------------------------------------------------------

XmitMCQueue::~XmitMCQueue()
{
}

//---------------------------------------------------------------------------
//  XmitMCQueue::Push() -- push buffer onto queue
//---------------------------------------------------------------------------

void  XmitMCQueue::Push(McBuffer& buffer )
{
    CriticalSectionArea guard(lock);

    BufferQ.push(buffer);
}

//---------------------------------------------------------------------------
//  XmitMCQueue::Pop() -- remove block from queue
//---------------------------------------------------------------------------

void  XmitMCQueue::Pop()
{
    CriticalSectionArea guard(lock);

    //  Save size of buffer being popped
    BufferQ.pop();
    //
    //  If there is room in the queue, request more input
    //
    //  Count up present blocks and pending blocks
    size_t total_blocks = Signal.Count() + Depth();
    if (total_blocks<CapacityInBlks && (SuspendSignal==false))
        {
        //  Signal need for additional data from host
        Signal.Release();
        }
}

//===========================================================================
//  CLASS RecvMCQueue  -- Queue to manage data from VISR to Receiver
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class RecvMCQueue
//---------------------------------------------------------------------------

RecvMCQueue::RecvMCQueue()
{
}

//---------------------------------------------------------------------------
//  destructor for class RecvMCQueue
//---------------------------------------------------------------------------

RecvMCQueue::~RecvMCQueue()
{
}

//---------------------------------------------------------------------------
//  RecvMCQueue::Push() -- push buffer onto queue
//---------------------------------------------------------------------------

void  RecvMCQueue::Push(McBuffer& buffer )
{
    CriticalSectionArea guard(lock);

    //  Push a buffer onto the queue.
    BufferQ.push(buffer);
    //
    //  For every buffer added, signal Rcvr thread
    Signal.Release();

}

//---------------------------------------------------------------------------
//  RecvMCQueue::Pop() --
//---------------------------------------------------------------------------

void  RecvMCQueue::Pop()
{
    CriticalSectionArea guard(lock);

    BufferQ.pop();
}


//===========================================================================
//  CLASS MC_XmitQPool  -- Manager for collection of XMIT queues for MCs.
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class MC_XmitQPool
//---------------------------------------------------------------------------

MC_XmitQPool::MC_XmitQPool()
    :  PrefillPacketCountArray(QPoolSize())
{
    std::fill(PrefillPacketCountArray.begin(), PrefillPacketCountArray.end(), 10);
}

//---------------------------------------------------------------------------
//  MC_XmitQPool::Queue() -- reference to internal XmitQ for a channel
//---------------------------------------------------------------------------

XmitMCQueue&  MC_XmitQPool::Queue(int channel)
{
    switch (channel)
        {
        default:
        case 0:    return XmitQ_0;    break;
        case 1:    return XmitQ_1;    break;
        case 2:    return XmitQ_2;    break;
        case 3:    return XmitQ_3;    break;
        case 4:    return XmitQ_4;    break;
        case 5:    return XmitQ_5;    break;
        case 6:    return XmitQ_6;    break;
        case 7:    return XmitQ_7;    break;
        }
}

//---------------------------------------------------------------------------
//  MC_XmitQPool::Clear() -- clear all data from queues
//---------------------------------------------------------------------------

void MC_XmitQPool::Clear()
{
    for (unsigned int ch=0; ch<QPoolSize(); ch++)
        Queue(ch).Clear();
}

//---------------------------------------------------------------------------
//  MC_XmitQPool::ClearSem() -- clear all transmit semaphores.
//---------------------------------------------------------------------------
//  called when stopping streaming

void MC_XmitQPool::ClearSem()
{
    for (unsigned int ch=0; ch<QPoolSize(); ch++)
        Queue(ch).ClearSem();
}

//---------------------------------------------------------------------------
//  MC_XmitQPool::PrefillPacketCount() -- load prefill depth
//---------------------------------------------------------------------------

void MC_XmitQPool::PrefillPacketCount(unsigned int channel, int packets)
{
    if (channel < PrefillPacketCountArray.size())
        PrefillPacketCountArray[channel] = packets;
}

//---------------------------------------------------------------------------
//  MC_XmitQPool::PrefillPacketCount() -- load prefill depth
//---------------------------------------------------------------------------

int MC_XmitQPool::PrefillPacketCount(unsigned int channel) const
{
    if (channel < PrefillPacketCountArray.size())
        return PrefillPacketCountArray[channel];
    else
        return 0;
}

}  // namespace Innovative

