//---------------------------------------------------------------------------
// MultiChanStream_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2015 Innovative Integration Corporation.
//  All Rights Reserved.

#ifndef MultiChanStream_MbH
#define MultiChanStream_MbH

#include <Events_Mb.h>
#include <SafeLink_Mb.h>

#include <MultiChanStreamSupport_Mb.h>
#include <ThreadSafeQueue_Mb.h>
#include <Buffer_Mb.h>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

class MultiChanStream;
class MultiChanStreamRcvrThread;
class MultiChanStreamXmitThread;
class MultiChanStreamStoppageThread;

//===========================================================================
// CLASS MultiChanStreamDataEvent  --  Data I/O Event for MultiChanStream
//===========================================================================

class MultiChanStreamDataEvent : public OpenWire::Event
{
public:
    MultiChanStream *  Sender;
    unsigned int       TxChannel;

public:
    MultiChanStreamDataEvent( MultiChanStream * sender, unsigned int channel )
        :  Sender( sender ), TxChannel(channel)
        {
        }
};

//===========================================================================
// CLASS MultiChanDirectDataEvent  --  Image Data Event for MultiChanStream
//===========================================================================

class MultiChanDirectDataEvent : public OpenWire::Event
{
public:
    MultiChanStream *  Sender;
    McImage            Data;

public:
    MultiChanDirectDataEvent( MultiChanStream * sender, const McImage & image)
        :  Sender( sender ), Data(image)
        {
        }
};


//===========================================================================
// CLASS MultiChanDirectTransmitAckEvent  --  Image Xmit Ack Notification
//===========================================================================

class MultiChanDirectTransmitAckEvent : public OpenWire::Event
{
public:
    MultiChanStream *  Sender;
    McHintPack         OutgoingHints;

public:
    MultiChanDirectTransmitAckEvent( MultiChanStream * sender, const McHintPack & hints)
        :  Sender( sender ), OutgoingHints(hints)
        {
        }
};

//===========================================================================
// CLASS MultiChanAlertEvent  --  Alert Event for MultiChanStream
//===========================================================================

class MultiChanAlertEvent : public OpenWire::Event
{
public:
    MultiChanStream *  Sender;
    McAlert            Alert;

public:
    MultiChanAlertEvent( MultiChanStream * sender, const McAlert & alert)
        :  Sender( sender ), Alert(alert)
        {
        }
};

//===========================================================================
// CLASS MultiChanExternalEvent  --  External I/O Interrupt Event
//===========================================================================

class MultiChanExternalEvent : public OpenWire::Event
{
public:
    MultiChanStream *  Sender;
    McExternalInt      ExtInt;

public:
    MultiChanExternalEvent( MultiChanStream * sender, const McExternalInt & ext_int )
        :  Sender( sender ),  ExtInt(ext_int)
        {
        }
};

//===========================================================================
// CLASS MultiChanStreamDeviceSafeLink  --  Safe Linkage to a device
//===========================================================================

class MultiChanStreamDeviceSafeLink : public SafeLink<IMultiChanStreamDevice, SafeMultiChanStreamDeviceImpl>
{
public:
    MultiChanStreamDeviceSafeLink(MultiChanStream * owner)
        :  Owner(owner)
        {}

private:
    MultiChanStream *  Owner;

    virtual void  OnAfterLink();        // overrides
    virtual void  OnBeforeUnlink();
};


//===========================================================================
//  CLASS MultiChanStreamQueue  -- Queue to manage data from VISR to Receiver
//===========================================================================

class MultiChanStreamQueue
{
public:
    // Ctor
    MultiChanStreamQueue();
    virtual ~MultiChanStreamQueue();

    // Methods
    McBuffer        Front();
    size_t          Depth();

    void            Clear();
    virtual void    Push(McBuffer & buffer) = 0;
    virtual void    Pop() = 0;

    Semaphore &     SignalSem();

    void            ClearSem()
                        {  Signal.Clear();  }
    void            BumpSignal(unsigned int count);

    // Properties
    void         QueueCapacity(size_t cap_in_blocks);
    size_t       QueueCapacity() const;

protected:
    //
    //  Member Data
    size_t      CapacityInBlks;
    Semaphore   Signal;
    mutable CriticalSection  lock;

    ThreadSafeQueue<McBuffer>  BufferQ;
};

//===========================================================================
//  CLASS XmitMCQueue  -- Queue to manage data from Receiver to VISR
//===========================================================================

class XmitMCQueue : public MultiChanStreamQueue
{
public:
    XmitMCQueue();
    ~XmitMCQueue();

    virtual void  Push(McBuffer & buffer);
    virtual void  Pop();
    void          Suspend(bool which)  {  SuspendSignal=which;  }

private:
    //
    //  Member Data
    bool  SuspendSignal;

};

//===========================================================================
//  CLASS RecvMCQueue  -- Queue to manage data from VISR to Receiver
//===========================================================================

class RecvMCQueue : public MultiChanStreamQueue
{
public:
    RecvMCQueue();
    ~RecvMCQueue();

    virtual void  Push(McBuffer & buffer);
    virtual void  Pop();

};

//===========================================================================
//  CLASS MC_XmitQPool  -- Manager for collection of XMIT queues for MCs.
//===========================================================================

class MC_XmitQPool
{
public:
    MC_XmitQPool();

    XmitMCQueue& Queue(int channel);

    void  Clear();
    void  ClearSem();

    void  PrefillPacketCount(unsigned int channel, int packets);
    int   PrefillPacketCount(unsigned int channel) const;

    size_t QPoolSize()  {   return kMultiChanChannelMaxX;  }  // hard coded system channel max

private:
    XmitMCQueue  XmitQ_0;
    XmitMCQueue  XmitQ_1;
    XmitMCQueue  XmitQ_2;
    XmitMCQueue  XmitQ_3;
    XmitMCQueue  XmitQ_4;
    XmitMCQueue  XmitQ_5;
    XmitMCQueue  XmitQ_6;
    XmitMCQueue  XmitQ_7;

    std::vector<int>  PrefillPacketCountArray;

};

//===========================================================================
//  CLASS MultiChanStream  -- MultiChan Based Stream I/O
//===========================================================================

class MultiChanStream : public IMultiChanStreamManager
{
public:
//    friend class MultiChanStreamIPSQCreditor;
    friend class MultiChanStreamRcvrThread;
    friend class MultiChanStreamXmitThread;
    friend class MultiChanStreamStoppageThread;
    friend class TIMultiChanInternalsImpl;

    MultiChanStream();
    virtual ~MultiChanStream();

    virtual void  Preconfigure();
    virtual void  Start();
    virtual void  Stop();

    void  Send(McBuffer & Packet );
    void  Recv(McBuffer & Packet );

    //  Image Mode Transmit
    McImageInformation  TxChannelInformation(size_t channel);
    McImage             TxGetEmptyImage(size_t channel, size_t desired_size);
    void                Send(McImage& image);

    void  SuspendOutput(unsigned int channel, bool state);

    virtual bool  StreamStarted();

    virtual bool  ConnectTo( IMultiChanStreamDevice *Device );
    virtual void  Disconnect( );

    McBuffer NativeBuffer(size_t data_size, bool autoinit=true);

    virtual IMultiChanStreamManagerInternals *  Internals();
    virtual IMultiChanStreamDeviceInternals *   DeviceInternals();

    OpenWire::EventHandler<OpenWire::NotifyEvent>  OnBeforeStart;
    OpenWire::EventHandler<OpenWire::NotifyEvent>  OnAfterStart;
    OpenWire::EventHandler<OpenWire::NotifyEvent>  OnAfterStop;

    OpenWire::EventHandler<MultiChanDirectTransmitAckEvent>  OnDirectDataRequired;
    OpenWire::EventHandler<MultiChanDirectDataEvent>  OnDirectDataAvailable;

    OpenWire::EventHandler<MultiChanStreamDataEvent>  OnDataRequired;
    OpenWire::EventHandler<MultiChanStreamDataEvent>  OnDataAvailable;

    OpenWire::EventHandler<MultiChanAlertEvent>       OnAlertAvailable;

    OpenWire::EventHandler<MultiChanExternalEvent>    OnExternalIntAvailable;


    //DPM

    void  PrefillPacketCount(int packets);
    void  PrefillPacketCount(unsigned int channel, int packets);
    int   PrefillPacketCount(unsigned int channel) const;

    void  RecvPacketCache(int packets)
        {  FRecvPacketCache=packets;  }
    int   RecvPacketCache() const
        {  return FRecvPacketCache;  }

    void    RxLoadBalancing(bool state)
        {  FRxLoadBalancing=state;  }
    bool    RxLoadBalancing() const
        {  return FRxLoadBalancing;  }

    void    TxLoadBalancing(bool state)
        {  FTxLoadBalancing=state;  }
    bool    TxLoadBalancing() const
        {  return FTxLoadBalancing;  }

protected:
    //
    //  Member Functions
    virtual void  SignalDataAvailable();
    virtual void  SignalDataRequired(unsigned int channel);

//    virtual IVelociaBufferManager *  ActualVelociaManager()
//                                        {  return &BufferMgr;  }

private:
    //
    //  Member Data
    MultiChanStreamDeviceSafeLink    StreamDevice;
    //PmcBufferManager                 BufferMgr;  McBufferManager
    //
    //  Member Data
    IMultiChanStreamManagerInternals *  InternalOps;
    MultiChanStreamRcvrThread *         RecvThread;
    MultiChanStreamXmitThread *         XmitThread;
    MultiChanStreamStoppageThread *     StopperThread;

    MC_XmitQPool                     XmitQ;
    RecvMCQueue                      RecvQ;

private:
    bool                             Started;
    int                              FRecvPacketCache;

private:
    // Load-balancing data
    bool                             FRxLoadBalancing;
    bool                             FTxLoadBalancing;
    //
    //  Member Functions
    void                    Prefill();


    //  StreamManagerInternals Interface Implementations
    void   ForwardAlertPacket(McAlert & pkt);
    void   ForwardExternIntPacket(McExternalInt & pkt);
    void   ForwardImagePacket(McImage & pkt);
    void   ForwardImageTransmitAck(McHintPack & OutgoingHints);
    void   ForwardPacketTransmitAck(McHintPack & OutgoingHints);
    void   ForwardDataPacket(McBuffer & pkt);
    bool   CopyIncomingData();
    bool   CopyOutgoingData();
    bool   IsIncomingQueueOverThreshold();
    size_t    ChannelXmitQ_Packets(int channel);
    size_t    ChannelXmitQ_PendingNotifies(int channel);
    void      ChannelXmitQ_Signal(int channel, unsigned int count);
    size_t    ChannelXmitQ_NextPacketSize(int channel);
    McBuffer  ChannelXmitQ_DequeueNextPacket(int channel);

    //bool  HasStreamStarted();   // alias of StreamStarted()

};

#ifdef __CLR_VER
#pragma managed(pop)
#endif
}  // namespace

#endif
