// NetImageLoadThread_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//
#define ACE_NLOGGING

#include <algorithm>
#include <ace/Log_Msg.h>
#include <ace/SOCK_Connector.h>
#include <ace/OS_NS_string.h>
#include <ace/OS_NS_stdio.h>
#include <ace/Time_Value.h>

#include <NetImageLoadThread_Mb.h>
#include <Sbc6713e_Mb.h>

namespace   Innovative
{

//==============================================================================
//  CLASS NetImageLoadThread -- manages Logic load via network
//==============================================================================
const unsigned int   MaxChunkSize = 16384;

//------------------------------------------------------------------------------
//  constructor of class NetImageLoadThread
//------------------------------------------------------------------------------

NetImageLoadThread::NetImageLoadThread(Sbc6713e&  baseboard,
                                       Sbc6713e::LoadInfoType which,
                                       std::string image_file)
    : DefaultFileName(""), Baseboard(baseboard), 
      BinImage(DefaultFileName), LoadType(which), 
      ImageLoadComplete(Event::ManualReset),
      ID(baseboard.Id()), Port(baseboard.ImageLoadPort()),
      FImageFile(image_file), client(0)       
{
    Resume();
}

//------------------------------------------------------------------------------
//  NetImageLoadThread::Execute() -- Thread routine
//------------------------------------------------------------------------------

void NetImageLoadThread::Execute()
{
    int                     Checksum(0);
    size_t                  BytesRemaining(0);
    size_t                  ImageSize(0);
    const unsigned char*    ImageDataBase(0);

    //
    //  Prepare the Image for transmission....
    switch (LoadType)
        {
        case Sbc6713e::ltLANLogic:
        case Sbc6713e::ltINTFLogic:
            SrImage.OnProgress.SetEvent(&Baseboard, &Sbc6713e::LogicParseProgress);
            SrImage.OnComplete.SetEvent(&Baseboard, &Sbc6713e::LogicParseComplete);

            //
            //  Parse the EXO file to create the image that will be sent
            SrImage.FileName(FImageFile);
            SrImage.Parse();

            //
            //  Calculate Image checksum for transmission
            for (unsigned int ck = 0; ck < SrImage.Size(); ck++)
                Checksum += SrImage.Buffer[ck];

            ImageSize = BytesRemaining = SrImage.Size();
            ImageDataBase = &SrImage.Buffer[0];
            break;

        case Sbc6713e::ltLoaderBin:
        case Sbc6713e::ltTalkerBin:
        case Sbc6713e::ltAppBin:
            BinImage.OnProgress.SetEvent(&Baseboard, &Sbc6713e::LogicParseProgress);
            BinImage.OnComplete.SetEvent(&Baseboard, &Sbc6713e::LogicParseComplete);

            //
            //  Parse the BIN file to create an image for transmission
            BinImage.FileName(FImageFile);
            BinImage.LoadFromFile();

            //
            //  Calculate Image checksum for transmission
            for (size_t ck = 0; ck < BinImage.Size(); ck++)
                Checksum += (unsigned char)BinImage.Image[ck];

            ImageSize = BytesRemaining = BinImage.Size();
            ImageDataBase = BinImage.UImageBasePtr();
            break;
        }

    ImageLoadComplete.Clear();

    //
    //  Attempt Connection to Server
    //  ...Build ourselves a Stream socket. This is a connected socket that
    //     provides reliable end-to-end communications. We will use the
    //     server object to send data to the server we connect to.
    client = new ACE_SOCK_Stream;

    //  ...And we need a connector object to establish that connection. The
    //     ACE_SOCK_Connector object provides all of the tools we need to
    //     establish a connection once we know the server's network
    //     address...
    ACE_SOCK_Connector  connector;

    //  ...Which we create with an ACE_INET_Addr object. This object is
    //     given the TCP/IP port and hostname of the server we want to
    //     connect to.
    ACE_INET_Addr       addr(Port, ID.c_str());
    ACE_Time_Value      timeout(45);

    //  ...So, we feed the Addr object and the Stream object to the
    //     connector's connect() member function. Given this information, it
    //     will establish the network connection to the server and attach
    //     that connection to the server object.
    if (connector.connect(*client, addr, &timeout) == -1)
        {
        ProcessStatusEvent evt(std::string("Connection failure"));
        Baseboard.OnLogicLoadError.Execute(evt);
        return;
        }

    //
    //  Send a "Download Request" packet
    enum PakTypes
    {
        pkLanLogicDownloadRequest   = 0xE0,
        pkIntfLogicDownloadRequest,
        pkLogicDownloadRequestAck,
        pkLogicDownloadDone,
        pkLogicDownloadDoneAck,
        pkChunkHeader,
        pkChunkHeaderAck,
        pkChunkDataAck,
        pkLoaderDownloadRequest,
        pkTalkerDownloadRequest,
        pkAppDownloadRequest
    };

    PakTypes    ReqType;
    switch (LoadType)
        {
        case Sbc6713e::ltLANLogic:
            ReqType = pkLanLogicDownloadRequest;
            break;
        case Sbc6713e::ltINTFLogic:
            ReqType = pkIntfLogicDownloadRequest;
            break;
        case Sbc6713e::ltLoaderBin:
            ReqType = pkLoaderDownloadRequest;
            break;
        case Sbc6713e::ltTalkerBin:
            ReqType = pkTalkerDownloadRequest;
            break;
        case Sbc6713e::ltAppBin:
            ReqType = pkAppDownloadRequest;
            break;
        default:
            {
            ProcessStatusEvent evt(std::string("Invalid image type"));
            Baseboard.OnLogicLoadError.Execute(evt);
            return;
            }
        }

    size_t packet[3];
    packet[0] = ReqType;
    packet[1] = ImageSize;
    packet[2] = Checksum;

    if (client->send_n(packet, sizeof(packet), &timeout) == -1)
        {
        ProcessStatusEvent evt(std::string("Target timeout sending load command request"));
        Baseboard.OnLogicLoadError.Execute(evt);
        return;
        }

    //
    //  Wait for "ack" packet
    if (client->recv_n(packet, sizeof(packet), &timeout) == -1)
        {
        ProcessStatusEvent evt(std::string("Target timeout waiting for ack to load command request"));
        Baseboard.OnLogicLoadError.Execute(evt);
        return;
        }

    //
    //  Initiate Download...
    size_t LocIdx = 0;

    // Reset progress.
    ProcessProgressEvent e(0);
    Baseboard.OnLogicLoadProgress.Execute(e);

    while (BytesRemaining)
        {
        //
        //  Send Chunk Header
        size_t Header[3];
        int Reply[3];
        size_t ChunkSize =
            (BytesRemaining > MaxChunkSize) ? MaxChunkSize : BytesRemaining;
        Header[0] = pkChunkHeader;
        Header[1] = ChunkSize;

        if (client->send_n(Header, sizeof(Header), &timeout) == -1)
            {
            ProcessStatusEvent evt(std::string("Target timeout on chunk header"));
            Baseboard.OnLogicLoadError.Execute(evt);
            break;
            }
        if (client->recv_n(Reply, sizeof(Reply), &timeout) == -1)
            {
            ProcessStatusEvent evt(std::string("Target timeout on chunk header ack"));
            Baseboard.OnLogicLoadError.Execute(evt);
            break;
            }

        //
        //  Send Chunk Data
        int                     DReply[3];
        const unsigned char*    PacketData = ImageDataBase + LocIdx;

        if (client->send_n(PacketData, ChunkSize, &timeout) == -1)
            {
            ProcessStatusEvent evt(std::string("Target timeout on chunk data"));
            Baseboard.OnLogicLoadError.Execute(evt);
            break;
            }
        if (client->recv_n(DReply, sizeof(DReply), &timeout) == -1)
            {
            ProcessStatusEvent evt(std::string("Target timeout on chunk data ack"));
            Baseboard.OnLogicLoadError.Execute(evt);
            break;
            }

        //
        //  Update pointers
        BytesRemaining -= ChunkSize;
        LocIdx += ChunkSize;

        //  Update event.
        int residual = static_cast<int>(100 - BytesRemaining*100/ImageSize);
        ProcessProgressEvent e(residual);
        Baseboard.OnLogicLoadProgress.Execute(e);
        }

    //
    //  Send final command to start program...
    packet[0] = pkLogicDownloadDone;
    packet[1] = 0;
    packet[2] = 0;

    if (client->send_n(packet, sizeof(packet), &timeout) == -1)
        {
        ProcessStatusEvent evt(std::string("Target timeout on burn command"));
        Baseboard.OnLogicLoadError.Execute(evt);
        return;
        }

    //
    //  Wait for "ack" packet
    if (client->recv_n(packet, sizeof(packet), &timeout) == -1)
        {
        ProcessStatusEvent evt(std::string("Target timeout on burn ack"));
        Baseboard.OnLogicLoadError.Execute(evt);
        return;
        }

    switch (packet[1])
        {
        case 0:     // ok
            break;
        case 1:     // Bad transmit checksum
            {
            ProcessStatusEvent evt(std::string("Bad transmit checksum"));
            Baseboard.OnLogicLoadError.Execute(evt);
            return;
            }
        case 2:     // Bad burn checksum
            {
            ProcessStatusEvent evt(std::string("Bad burn checksum"));
            Baseboard.OnLogicLoadError.Execute(evt);
            return;
            }
        }

    ProcessProgressEvent eventProgress(100);
    Baseboard.OnLogicLoadProgress.Execute(eventProgress);
    
    ProcessCompletionEvent eventComplete(static_cast<int>(packet[1]));
    Baseboard.OnLogicLoadComplete.Execute(eventComplete);
}

//------------------------------------------------------------------------------
//  NetImageLoadThread::Done() -- Clean-up
//------------------------------------------------------------------------------

void NetImageLoadThread::Done()
{
    //  Close the connection to the server.  The servers we've created so
    //  far all are based on the ACE_Reactor.  When we close(), the
    //  server's reactor will see activity for the registered event
    //  handler and invoke handle_input().  That, in turn, will try to
    //  read from the socket but get back zero bytes.  At that point, the
    //  server will know that we've closed from our side.
    if (client)
        client->close();

    ImageLoadComplete.Set();
}


}   // namespace Innovative
