//  PacketDeviceMap_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <PacketDeviceMap_Mb.h>
#include <Exception_Mb.h>
#include <Trace_Mb.h>

//---------------------------------------------------------------------------
namespace Innovative
{

namespace PacketDevMap
{

//===========================================================================
//  CLASS DeviceToScheme  -- Generic Device callbacks for Scheme
//===========================================================================
//  Data Ring Tags
const int PciInData  = 1;
const int PciOutData = 0;

class DeviceToScheme : public IPacket_DeviceToScheme
{
public:
    DeviceToScheme(PacketDeviceMap * owner)
        : Owner(owner), Debug(true) {}

    virtual Event * IsrEvent() {  return Owner->PacketDevice->IsrEvent(); }
    virtual size_t  InterruptDepth() { return Owner->PacketDevice->InterruptDepth(); }
    virtual int     InterruptAck() { return Owner->PacketDevice->InterruptAck(); }
    virtual void    Log(int type, int data)
        {
        Owner->PacketDevice->Log(type, data);
        //std::stringstream msg;
        //msg << "Type: " << std::hex << type << ", Data: " << data;
        //Debug.Log(msg.str());
        }
    virtual void    Log(int type, size_t data)
        {  Log(type, static_cast<int>(data));  }

    virtual InnovativeKernel::IDeviceDriverBmManagerUI &  BmMgr()
                {  return Owner->PacketDevice->BmManager(); }

    virtual ii32    IncomingBusMasterSizeWords() const { return Owner->IncomingBusMasterSizeWords(); }
    virtual ii32    OutgoingBusMasterSizeWords() const { return Owner->OutgoingBusMasterSizeWords(); }
    virtual void    BusMasterSizeWords(ii32 value) { Owner->BusMasterSizeWords(value); }
    virtual bool  AlertTest(unsigned int pid) { return Owner->AlertTest(pid); }
    virtual bool  AlertCheck(const Buffer & buffer) { return Owner->AlertCheck(buffer); }

    virtual void  DispatchCustomInterrupt(int CustomFlags) {  Owner->DispatchCustomInterrupt(CustomFlags);  }

    virtual ManagerInternalsSafeLink &  Link()  { return Owner->StreamManagerInternals;  }

    virtual ii32 DebugPortPeek(unsigned int port_addr)
            {
            return Owner->PacketDevice->PortMemory.Read(port_addr);
            }

private:
    PacketDeviceMap  *  Owner;
    Trace               Debug;
};

//===========================================================================
//  CLASS PacketDevicePluginImpl -- Installable Callbacks for PacketDevice
//===========================================================================

class PacketDevicePluginImpl : public IPacketDevicePlugin
{
public:
    PacketDevicePluginImpl( PacketDeviceMap * pdm )
        :  PDM(pdm)
        {}
    virtual void  AfterDeviceOpen()
        {  PDM->DoOpen();  }
    virtual void  BeforeDeviceClose()
        {  PDM->DoClose();  }

    PacketDeviceMap * PDM;
};

}  // end namespace PacketDevMap

//===========================================================================
//  CLASS PacketDeviceMap  -- Base class for Packet mode BM devices
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class PacketDeviceMap
//---------------------------------------------------------------------------

PacketDeviceMap::PacketDeviceMap( IPacketDevice * pkt_device_intf,
                    PacketDeviceMapRegisters & regs,
                    IPacket_DeviceToScheme * scheme_intf)
    : PacketDevice(pkt_device_intf),
      IntScheme(0), 
      RegAddr(regs),
      CreditQueue           ( pkt_device_intf->PortMemory,  regs.CreditQueue_Addr ),
      InterruptEnable       ( pkt_device_intf->PortMemory,  regs.InterruptEnable_Addr ),
      InitRcvCredits(0x400),
      ShouldPreloadXmit(true),
      ShouldIssueRcvCredit(true),
      SingleTimeAtStartPreconfig(true)
{
    IPacket_DeviceToScheme * DtS;

    if (scheme_intf!=0)        // if override intf passed in, use it
        DtS = scheme_intf;
    else
        DtS = new PacketDevMap::DeviceToScheme(this);
        
    IntScheme = new PacketInterruptScheme(DtS);
}

//---------------------------------------------------------------------------
//  destructor for class PacketDeviceMap
//---------------------------------------------------------------------------

PacketDeviceMap::~PacketDeviceMap()
{
    PacketDevice->Close();

	delete IntScheme;
	IntScheme = 0;
}

//---------------------------------------------------------------------------
//  PacketDeviceMap::DoOpen()
//---------------------------------------------------------------------------

void  PacketDeviceMap::DoOpen()
{
    IntScheme->Open();
    EnableCustomInterrupt(true);
}

//---------------------------------------------------------------------------
//  PacketDeviceMap::DoClose()
//---------------------------------------------------------------------------

void  PacketDeviceMap::DoClose()
{
    EnableCustomInterrupt(false);
    IntScheme->Close();
}

//---------------------------------------------------------------------------
//  PacketDeviceMap::PreconfigBusmastering()
//---------------------------------------------------------------------------

void  PacketDeviceMap::PreconfigBusmastering()
{
    //  Stream Event Callback
    OpenWire::NotifyEvent e;
    OnStreamPreconfig.Execute(e);
    // "activate" one time at-start preconfiguration
    SingleTimeAtStartPreconfig = true;
}

//---------------------------------------------------------------------------
//  PacketDeviceMap::ConfigureBusmastering()
//---------------------------------------------------------------------------

void  PacketDeviceMap::ConfigureBusmastering()
{
    IntScheme->ConfigureBusmastering();

    //  Stream Event Callback
    OpenWire::NotifyEvent e;
    OnStreamConfigure.Execute(e);

    if (SingleTimeAtStartPreconfig)
        {
        //  Stream Event Callback
        OpenWire::NotifyEvent e;
        OnStreamSingleTimeAtStartConfig.Execute(e);

        SingleTimeAtStartPreconfig = false;
        }

}

//---------------------------------------------------------------------------
//  PacketDeviceMap::PreloadXmitBusmasterRegion()
//---------------------------------------------------------------------------

void  PacketDeviceMap::PreloadXmitBusmasterRegion()
{
    if (ShouldPreloadXmit)
        {
        IntScheme->PreloadXmitBusmasterRegion();
        }
}

//---------------------------------------------------------------------------
//  PacketDeviceMap::StartBusmastering()
//---------------------------------------------------------------------------

void  PacketDeviceMap::StartBusmastering()
{
    //  Stream Event Callback
    OpenWire::NotifyEvent e;
    OnBeforeStreamStart.Execute(e);

    EnableBusmasterInterrupt(true);

    if (ShouldIssueRcvCredit)
        {
        //  ...Grant credit for part of Queue
        unsigned int RcvRegionSize = IncomingBusMasterSizeWords();
        GrantCredit(PacketDevMap::PciInData, RcvRegionSize-5);  // give a 'leetle' less than full region
        }
    //
    //  Begin full duplex busmastering
    PacketDevice->BusmasteringOn();

    //  Stream Event Callback
    OnAfterStreamStart.Execute(e);

}

//---------------------------------------------------------------------------
//  PacketDeviceMap::StopBusmastering()
//---------------------------------------------------------------------------

void  PacketDeviceMap::StopBusmastering()
{
    //  Stream Event Callback
    OpenWire::NotifyEvent e;
    OnBeforeStreamStop.Execute(e);
    //
    //  Begin shutting down the PCI transmissions
    PacketDevice->BusmasteringOff();
    //
    //  Stream Event Callback
    OnAfterStreamStop.Execute(e);

}

//---------------------------------------------------------------------------
//  PacketDeviceMap::EnableBusmasterInterrupt()
//---------------------------------------------------------------------------

void PacketDeviceMap::EnableBusmasterInterrupt( bool enable )
{
    if ( enable )
        {
        InterruptEnable.BmRecvInt.Set();
        InterruptEnable.BmTransmitInt.Set();
        }
    else
        {
        InterruptEnable.BmRecvInt.Clear();
        InterruptEnable.BmTransmitInt.Clear();
        }
}

//---------------------------------------------------------------------------
//  PacketDeviceMap::EnableCustomInterrupt()
//---------------------------------------------------------------------------

void PacketDeviceMap::EnableCustomInterrupt( bool enable )
{
    if ( enable )
        {
        InterruptEnable.CustomInt.Set();
        }
    else
        {
        InterruptEnable.CustomInt.Clear();
        }
}

//---------------------------------------------------------------------------
//  PacketDeviceMap::AlertTest() -- Alert Test Processing Hook
//---------------------------------------------------------------------------

bool  PacketDeviceMap::AlertTest(unsigned int /*pid*/)
{
    //  return FALSE to as this is not an alert
    return false;
}

//---------------------------------------------------------------------------
//  PacketDeviceMap::AlertCheck() -- Alert Event Processing Hook
//---------------------------------------------------------------------------

bool  PacketDeviceMap::AlertCheck(const Buffer & /*buffer*/)
{
    //  return TRUE if buffer is to be passed on to application.
    //  return FALSE to discard buffer now.
    return true;
}

//---------------------------------------------------------------------------
//  PacketDeviceMap::GrantCredit() - Grant credit to PMC device
//---------------------------------------------------------------------------

void  PacketDeviceMap::GrantCredit(int channel, size_t size_wds)
{
    unsigned int credit_addr = (channel==1) ? 0x17c00 : 0x17800 ;

    PacketDevice->Log(0x10, PacketDevice->PortMemory.Read(credit_addr));
    CriticalSectionArea  guard(CreditLock);
    CreditQueue.ChannelNumber.Value(channel);
    CreditQueue.Credit.Value(static_cast<ii32>(size_wds));

    CreditQueue.Apply();    // actally apply to hardware

    PacketDevice->Log(0x11, static_cast<int>(channel));
    PacketDevice->Log(0x12, static_cast<int>(size_wds));
    PacketDevice->Log(0x13, PacketDevice->PortMemory.Read(credit_addr));

    PacketDevice->Log(0x14, static_cast<int>(size_wds) & 0xffff);
    PacketDevice->Log(0x15, (static_cast<int>(size_wds) >> 16) & 0xffff );


}

//---------------------------------------------------------------------------
//  PacketDeviceMap::TransmitDataAvailable() - Signal xmit data available
//---------------------------------------------------------------------------

void PacketDeviceMap::TransmitDataAvailable()
{
    IntScheme->TransmitDataAvailable();
}

//---------------------------------------------------------------------------
//  PacketDeviceMap::DispatchCustomInterrupt() -- process custom interrupt
//---------------------------------------------------------------------------

void  PacketDeviceMap::DispatchCustomInterrupt(int /*CustomFlags*/)
{
    //  Default behavior is to do nothing...override in derived classes
}

//---------------------------------------------------------------------------
//  PacketDeviceMap::BusmasterReset() -- process custom interrupt
//---------------------------------------------------------------------------

void  PacketDeviceMap::BusmasterReset()
{
    PacketDevice->BusmasterReset();
}

//---------------------------------------------------------------------------
//  PacketDeviceMap::TraceLog() -- process custom interrupt
//---------------------------------------------------------------------------

void  PacketDeviceMap::TraceLog(int type, int data)
{
    PacketDevice->Log(type, data);
}

//---------------------------------------------------------------------------
//  PacketDeviceMap::VisrIsrQDepth() -- return depth of ISR queue
//---------------------------------------------------------------------------

unsigned int  PacketDeviceMap::VisrIsrQDepth()
{
    return static_cast<unsigned int>(IntScheme->VisrIsrQDepth());
}

//---------------------------------------------------------------------------
//  PacketDeviceMap::Error() -- throw error string
//---------------------------------------------------------------------------

void  PacketDeviceMap::Error( std::string msg )
{
    throw Innovative::MalibuException(msg);
}

//---------------------------------------------------------------------------
//  PacketDeviceMap::RegisterStreamPlugIn() -- attach to a streaming object
//---------------------------------------------------------------------------

void  PacketDeviceMap::RegisterStreamPlugIn( IPacketStreamManagerInternals * plug_in )
{
    if (plug_in)
        {
        PacketDevice->InstallStreamPlugin( new PacketDevMap::PacketDevicePluginImpl(this) );
        StreamManagerInternals.Link(plug_in);
        }
    else
        StreamManagerInternals.Unlink();
}

//---------------------------------------------------------------------------
//  PacketDeviceMap::FillBmRegion() -- fill busmaster region with pattern
//---------------------------------------------------------------------------

void  PacketDeviceMap::FillBmRegion(Dir which, unsigned int pattern, bool ramp)
{
    PacketInterruptScheme::Dir which2;
    which2 = (which==BmXmit)? PacketInterruptScheme::BmXmit : PacketInterruptScheme::BmRecv;
    IntScheme->FillBmRegion(which2, pattern, ramp);
}

//------------------------------------------------------------------------------
//  PacketDeviceMap::BusMasterSizeWords() --
//------------------------------------------------------------------------------

void PacketDeviceMap::BusMasterSizeWords(unsigned int words)
{
    PacketDevice->SetBusMasterSizeWords(words);
}

//------------------------------------------------------------------------------
//  PacketDeviceMap::IncomingBusMasterSizeWords() --
//------------------------------------------------------------------------------

void PacketDeviceMap::IncomingBusMasterSizeWords(unsigned int words)
{
    PacketDevice->SetIncomingBusMasterSizeWords(words);
}

//------------------------------------------------------------------------------
//  PacketDeviceMap::OutgoingBusMasterSizeWords() --
//------------------------------------------------------------------------------

void PacketDeviceMap::OutgoingBusMasterSizeWords(unsigned int words)
{
	PacketDevice->SetOutgoingBusMasterSizeWords(words);
}

//------------------------------------------------------------------------------
//  PacketDeviceMap::IncomingBusMasterSizeWords() --
//------------------------------------------------------------------------------

ii32 PacketDeviceMap::IncomingBusMasterSizeWords() const
{
	return PacketDevice->GetIncomingBusMasterSizeWords();
}

//------------------------------------------------------------------------------
//  PacketDeviceMap::OutgoingBusMasterSizeWords() --
//------------------------------------------------------------------------------

ii32 PacketDeviceMap::OutgoingBusMasterSizeWords() const
{
	return PacketDevice->GetOutgoingBusMasterSizeWords();
}



} // namespace Innovative


