// PacketFileDataSet_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//
//---------------------------------------------------------------------------

#ifndef PacketFileDataSet_MbH
#define PacketFileDataSet_MbH

#include <DataSetBase_Mb.h>
#include <DriveIo_Mb.h>
#include <map>
#include <vector>
#include <list>

namespace Innovative
{

#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//===========================================================================
// Class PacketHeader
//===========================================================================

class PacketHeader
{
public:
    // Ctor
    PacketHeader()
        {
            Data[1] = Data[0] = 0;
        }

    // Methods
    unsigned int        PacketSize() const
        {  return Data[0] & 0xffffff;  }
    unsigned int        PeripheralId() const
        {  return Data[0] >> 24;  }
    unsigned int *      Raw()
        {  return &Data[0];  }

    unsigned int        SizeInBytes()
        {  return sizeof(Data);  }
	unsigned int        SizeInInts()
        {  return SizeInBytes()/sizeof(int);  }

	// Data
    unsigned int        Data[2];
};

//===========================================================================
// Class PacketBlock
//===========================================================================

class PacketBlock
{
public:
	PacketBlock();
	PacketBlock(const PacketBlock & other);
    ~PacketBlock();

    void  Allocate(unsigned int size);
    void  Release();

    bool  IsAllocated()  { return Allocated; }
    unsigned int Size()  { return FSize;  }

private:
    // Data
    bool                 Allocated;
    unsigned int         FSize;           // Packet Data Size in INTs

public:
    // Data
    unsigned int *       Data;
};

//===========================================================================
// Class PacketFile
//===========================================================================

class PacketFile
{
public:
    // Ctor
    PacketFile();

    // Methods
    bool            IsOpen()
		{  return File.Opened();  }
	bool            Open(std::string filename, bool auto_create=false,
						 bool read_only=true);
	void            Close();
	void            Flush();
	ii64         	Size() const;
	bool            Read(unsigned int * buffer, unsigned int size_in_ints);
	bool            Write(unsigned int * buffer, unsigned int size_in_ints);
	bool            Skip(int size_in_ints);

	bool            Seek(ii64 offset_from_front);

	// Properties
	bool            Position(ii64 offset);
	ii64         	Position();

private:
	// Fields
	std::string             FFileName;
	bool                    FAutoCreate;
	std::streampos          FCursor;

	// Data
	DriveIo                 File;
};

//===========================================================================
// Class PacketProxy -- Information
//===========================================================================

class PacketProxy
{
public:
	PacketProxy();
    PacketProxy(ii64 sample_idx);
    ~PacketProxy();

    ii64   OffsetAfterHeader()
        {  return Offset + Header.SizeInBytes();  };
    ii64   OffsetAfterPacket()
        {  return Offset + Size*sizeof(int);  };
    void      UseFile(PacketFile * file)
        {  PFile = file;  };

    bool  operator < (const PacketProxy & rhs) const
        {  return LessThanEnd(rhs);  }
	//
	//  Search predicates
    bool PacketContainsSample(ii64 Sample) const
        {  return Sample >= BaseSampleNumber && BaseSampleNumber+Samples > Sample;  }
    bool LessThanStart(const PacketProxy & rhs) const
        {  return BaseSampleNumber < rhs.BaseSampleNumber;  }
    bool LessThanEnd(const PacketProxy & rhs) const
        {  return BaseSampleNumber+Samples-1 < rhs.BaseSampleNumber;  }

public:
	ii64             	Offset;            // Offset within image to start of packet
    unsigned int        Size;              // Number of ints in packet
    unsigned int        DataSize;          // Number of ints in data portion of packet
	ii64             	BaseSampleNumber;  // Sample # of first sample in packet
	ii64             	Samples;           // # of samples PER CHANNEL
    int                 Channels;          // # of data channels in packet
    int                 BytesPerSample;    // Width of samples
    PacketHeader        Header;
	PacketBlock         Body;
	size_t              Index;

    void                AllocateBody();
    void                ReleaseBody();

private:
    PacketFile *   PFile;
};

//===========================================================================
// Class PidPacketArray
//===========================================================================

class PidPacketArray
{
public:
    PidPacketArray();
    ~PidPacketArray();

	Subsample   CurrentSample(int channel_in_pid, ii64 index);
	void       	CurrentSample(int channel_in_pid, ii64 index, Subsample & value);

    void  Clear();

    PacketProxy *  LastPacket();

public:
	ii64             	MaxSampleIdx;
	const size_t		CacheSizeBlocks;
	size_t				CurrentProxyIdx;

	typedef std::vector<PacketProxy>  PacketArray;
	typedef std::list<size_t>  CacheList;

	PacketArray    Packets;
	CacheList      Cache;

	void  FindProxyForSample(ii64 index);
    void  AllocateProxy();

    PacketProxy &  FoundProxy();
    bool  IsFoundProxyValid();
};

//===========================================================================
// Class PacketFileDataSet
//===========================================================================

class PacketFileDataSet : public DataSet
{
public:
	PacketFileDataSet();
	virtual ~PacketFileDataSet();

	// Properties
	void				AutoCreate(bool value)
		{  FAutoCreate=value;  }
	bool				AutoCreate() const
		{  return FAutoCreate;  }

    void  FillIncidentList(ii64 StartIdx, ii64 EndIdx);

protected:
    // Overridden Methods
    virtual void        ImageInit();
	virtual void        ImageSampleFetch(Subsample &sample);
	virtual void        ImageSampleStore(Subsample &sample);
	virtual void        PreAccess();
	virtual void        PostAccess();
    // Override Methods
    virtual bool        ImageOpen();
    virtual void        ImageClose();
	virtual void        ImageSeek(ii64 offset);
	virtual ii64     	ImageSize() const;
    virtual void        ImageSize(ii64 size);
    virtual void        ImageFlush();

private:
    typedef std::map<int, PidPacketArray> PidPacketMap;

    // Data
    PacketFile              File;
    PidPacketMap            Pid;
    PidPacketArray          Alert_PPA;
	ii64                 	FSampleIdx;
    bool                    FImaginary;
	std::vector<ii64>    	FParseLimitIdx;
    bool                    FFileReadOk;
	PidPacketArray *        FLastArray;

	// Fields
	bool                    FAutoCreate;        // Auto-create files on open fail

	// Methods
	virtual void        ImageReadIniFile(IniFile &Ini);
	virtual void        ImageWriteIniFile(IniFile &Ini);

	PidPacketArray &  WorkingPid();
	int               PidFromSubchannel(int channel);
	void              ParseDataFile(bool force_scan = false);
	void              InitializeProxy(PacketProxy & Proxy, ii64 Position, ii64 SampleIdx);
	ii64           	  LastParsePosition();
    unsigned int      SubchannelsPerDevice();
    size_t            PacketCount() const;
    int               Subchannel() const;

    size_t            AlertPacketCount() const;

    ii64  LeastOffsetFromIdx(ii64 idx);
    ii64  MostOffsetFromIdx(ii64 idx);
	void  FillAlertIncident(DatasetIncident & incident, const PacketProxy & Alert, ii64 EstimatedSampleIdx);
	void  FillDataIncident(DatasetIncident & incident, const PacketProxy & Alert, ii64 EstimatedSampleIdx);
};

#ifdef __CLR_VER
#pragma managed(pop)
#endif
} // namespace Innovative

#endif
