//---------------------------------------------------------------------------
// PacketImage_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <PacketImage_Mb.h>
#include <StartStopThread_Mb.h>
#include <StopWatch_Mb.h>
#include <SystemSupport_Mb.h>
#include <BufferHeader_Mb.h>
#include <MemoryUtils_Mb.h>

namespace Innovative
{

namespace PacketImageNS
{
//==============================================================================
//  CLASS PacketImageDatagram  --  Special Copy Ops for PacketImage
//==============================================================================
//------------------------------------------------------------------------------
//  declaration for class PacketImageDatagram
//------------------------------------------------------------------------------

class PacketImageDatagram
{
public:
    PacketImageDatagram(Buffer & op, const PacketImage & image)
        : FBuffer(op), FImage(image), CurrentRange(0,0)
        {}
    void  FillBufferFromImage(size_t MaxSizeInts);

	void  CopyBodyFromImageAt( ii32 * buf, size_t my_index, size_t size);
	void  CopyHeaderFromImageAt( ii32 * buf, size_t my_index, size_t size);
    void  AdvanceExtractionByWords(size_t words);

private:
    Buffer &            FBuffer;
    const PacketImage & FImage;
    PacketImage::Range  CurrentRange;
    size_t              RangeIdx;
    size_t              ExtractionIdx;
};

//------------------------------------------------------------------------------
//  PacketImageDatagram::FillBufferFromImage() -- Fill buffer
//------------------------------------------------------------------------------
//  This fill command assumes packet data region is sized properly first
//    and that the image is larger or equal to the buffer size

void  PacketImageDatagram::FillBufferFromImage(size_t MaxSizeInts)
{
    // Resize packet to equal Max Size
    size_t buf_size = std::min(MaxSizeInts, FImage.SizeInInts());
    FBuffer.Resize(buf_size);

    //  Get the first image range
    RangeIdx = 0;
    CurrentRange = FImage.ImageRange(RangeIdx);
    ExtractionIdx = 0;

    // Header
    size_t buffer_idx = 0;
    size_t transfer_size = FBuffer.HeaderSizeInInts();
    while (transfer_size>0)
        {
        size_t words_before_end = CurrentRange.SizeInInts - ExtractionIdx;
        size_t size = std::min(words_before_end, transfer_size);

        CopyHeaderFromImageAt(CurrentRange.Addr+ExtractionIdx, buffer_idx, size);
        transfer_size -= size;
        buffer_idx += size;
        AdvanceExtractionByWords(size);
        }

    // Body
    buffer_idx = 0;
    transfer_size = FBuffer.SizeInInts();
    while (transfer_size>0)
        {
        size_t words_before_end = CurrentRange.SizeInInts - ExtractionIdx;
        size_t size = std::min(words_before_end, transfer_size);

        CopyBodyFromImageAt(CurrentRange.Addr+ExtractionIdx, buffer_idx, size);
        transfer_size -= size;
        buffer_idx += size;
        
        if (transfer_size)
            AdvanceExtractionByWords(size);
        }
}

//---------------------------------------------------------------------------
//  PacketImageDatagram::AdvanceExtractionByWords() --
//---------------------------------------------------------------------------

void  PacketImageDatagram::AdvanceExtractionByWords(size_t words)
{
    ExtractionIdx += words;
    if (ExtractionIdx==CurrentRange.SizeInInts)
        {
        // Advance to next range
        RangeIdx++;
        CurrentRange = FImage.ImageRange(RangeIdx);
        ExtractionIdx = 0;
        }
}

//---------------------------------------------------------------------------
//  PacketImageDatagram::CopyBodyFromImageAt() --
//---------------------------------------------------------------------------

void  PacketImageDatagram::CopyBodyFromImageAt( ii32 * buf, size_t my_index, size_t size)
{
    // Body
    ii32 * TargetDataPtr = reinterpret_cast<ii32*>(FBuffer.Data());

    CopyInt64Buffer(reinterpret_cast<const ii64*>(buf),
                 reinterpret_cast<ii64*>(TargetDataPtr + my_index),
                 static_cast<int>(size * sizeof(ii32) / sizeof(ii64)));
}

//---------------------------------------------------------------------------
//  PacketImageDatagram::CopyHeaderFromImageAt() --
//---------------------------------------------------------------------------

void  PacketImageDatagram::CopyHeaderFromImageAt( ii32 * buf, size_t my_index, size_t size)
{
    // Header
    ii32*    TargetDataPtr = (ii32*)FBuffer.Header();

    CopyInt64Buffer(reinterpret_cast<const ii64*>(buf),
                 reinterpret_cast<ii64*>(TargetDataPtr + my_index),
                 static_cast<int>(size * sizeof(ii32) / sizeof(ii64)));

}

}  // end namespace PacketStreamNS

//==============================================================================
//  CLASS PacketImage  -- Packet Position Information Store Object
//==============================================================================
//------------------------------------------------------------------------------
//  PacketImage::HeaderSizeInInts() -- Return packet header size
//------------------------------------------------------------------------------

size_t  PacketImage::HeaderSizeInInts() const
{
    return VBM->HeaderSizeInts();
}

//------------------------------------------------------------------------------
//  PacketImage::SizeInInts() -- Return packet data size
//------------------------------------------------------------------------------

size_t  PacketImage::SizeInInts() const
{
    return FullSizeInInts() - HeaderSizeInInts();
}

//------------------------------------------------------------------------------
//  PacketImage::FullSizeInInts() -- Return packet size
//------------------------------------------------------------------------------

size_t  PacketImage::FullSizeInInts() const
{
    return PacketSize;
}

//------------------------------------------------------------------------------
//  PacketImage::PeripheralId() -- Return packet PID
//------------------------------------------------------------------------------

unsigned int  PacketImage::PeripheralId() const
{
    if (Ranges.size())
        {
        return VBM->PeripheralIdFromMemory(Ranges[0].Addr);
        }
    else
        return 0;
}

//------------------------------------------------------------------------------
//  PacketImage::PeripheralId() -- change packet PID
//------------------------------------------------------------------------------

void  PacketImage::PeripheralId(unsigned int id)
{
    if (Ranges.size())
        {
        VBM->SetPeripheralIdIntoMemory(Ranges[0].Addr, id);
        }
}

//------------------------------------------------------------------------------
//  PacketImage::PacketFullSize() -- read packet size from memory
//------------------------------------------------------------------------------

size_t  PacketImage::PacketFullSize() const
{
    if (Ranges.size())
        {
        return VBM->PacketSizeFromMemoryInts(Ranges[0].Addr);
        }
    else
        return 0;
}

//------------------------------------------------------------------------------
//  PacketImage::CopyDataTo() -- copy data in packet image into a buffer
//------------------------------------------------------------------------------

void  PacketImage::CopyDataTo(Buffer & buffer) const
{
    // we want the entire buffer in this case
    CopyDataTo(buffer, SizeInInts());

}

//------------------------------------------------------------------------------
//  PacketImage::CopyDataTo() -- Copy buffer, with Size Limit
//------------------------------------------------------------------------------

void  PacketImage::CopyDataTo(Buffer & buffer, size_t MaxSizeInts ) const
{
    PacketImageNS::PacketImageDatagram PID( buffer, *this );
    PID.FillBufferFromImage(MaxSizeInts);
}

//==============================================================================
// CLASS BasicPacketImageParser  --  Parse packets in a flat memory image list
//==============================================================================
//------------------------------------------------------------------------------
//  BasicPacketImageParser::Clear() -- Clear state of Parser of any old data
//------------------------------------------------------------------------------

void  BasicPacketImageParser::Clear()
{
    Ranges.clear();
    TotalProcessed = 0;
}

//------------------------------------------------------------------------------
//  BasicPacketImageParser::Append() -- Append Image to Parser memory
//------------------------------------------------------------------------------

void  BasicPacketImageParser::Append(const PacketImage & Image)
{
    for (unsigned int i=0; i<Image.ImageRangeCount(); i++)
        {
        //  we wish to coalesce each new range with prev one if possible
        if (!Ranges.empty() && Ranges.back().end()==Image.ImageRange(i).begin())
            {
            Ranges.back().SizeInInts += Image.ImageRange(i).SizeInInts;
            }
        else
            Ranges.push_back(Image.ImageRange(i));
        }
}

//------------------------------------------------------------------------------
//  BasicPacketImageParser::Depth() -- Total Data in Parser in words
//------------------------------------------------------------------------------

size_t  BasicPacketImageParser::Depth() const
{
    size_t sum = 0;
    for (RangeQueue::const_iterator iter = Ranges.begin(); iter!=Ranges.end(); ++iter)
        {
        sum += iter->SizeInInts;
        }
    return sum;
}

//------------------------------------------------------------------------------
//  BasicPacketImageParser::Parse() -- Parse Images for embedded packets
//------------------------------------------------------------------------------

void  BasicPacketImageParser::Parse()
{
    if (Ranges.empty())
        return;
    while (1==1)
        {
        PacketImage image(ParsePacketImage());
        if (image.empty())
            return;
        //  remember total size processed
        TotalProcessed += image.FullSizeInInts();
        //  call handler
        PacketImageParserImageAvailable evt(this, image);
        OnImageAvailable.Execute(evt);
        }
}

//------------------------------------------------------------------------------
//  BasicPacketImageParser::ParsePacketImage() -- Extract image of next packet
//------------------------------------------------------------------------------

PacketImage  BasicPacketImageParser::ParsePacketImage()
{
    if (!Ranges.size())
        return PacketImage(VBM, 0);

    //  create image of first entry
    PacketImage image(VBM, Ranges.front() );
    // extract size of next packet from header
    size_t packet_size = image.PacketFullSize();
    size_t depth = Depth();

    if (depth >= packet_size)
        {
        // extract packet as full image
        PacketImage OutputImage(VBM, packet_size);
        size_t packet_remaining(packet_size);
        while (packet_remaining)
            {
            if (packet_remaining >= Ranges.front().SizeInInts)
                {
                // grab entire next image range
                OutputImage.AddRange(Ranges.front());
                packet_remaining -= Ranges.front().SizeInInts;
                Ranges.pop_front();  // we have processed this range now
                }
            else
                {
                // Split the front Range and send out just our part
                // ...create reduced range from Range list front
                PacketImage::Range remainder = Ranges.front();
                remainder.SizeInInts = packet_remaining;
                // ...adjust the front for the range 'extracted'
                Ranges.front().Addr += packet_remaining;
                Ranges.front().SizeInInts -= packet_remaining;
                packet_remaining = 0;
                OutputImage.AddRange(remainder);
                }
            }
        return OutputImage;
        }
    else
        {
        // return null packet image
        return PacketImage(VBM, 0);
        }
}
}  // namespace Innovative

