//---------------------------------------------------------------------------
// PacketImage_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2008 Innovative Integration Corporation.
//  All Rights Reserved.
//---------------------------------------------------------------------------

#ifndef PacketImage_MbH
#define PacketImage_MbH
//---------------------------------------------------------------------------
#include <BufferHeader_Mb.h>
#include <Events_Mb.h>
#include <deque>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

class PacketStreamBase;

//==============================================================================
//  CLASS PacketImage  -- Packet Position Information Store Object
//==============================================================================

class PacketImage
{
public:
    //-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
    //  CLASS Range - image data location
    //-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
    class Range
        {
        public:

            Range(unsigned int * addr, size_t size_in_ints)
                :  Addr(addr), SizeInInts(size_in_ints)
                {}

            unsigned int *  Addr;
            size_t          SizeInInts;
            
            unsigned int *  begin()
                {  return Addr;  }
            unsigned int *  end()
                {  return Addr+SizeInInts;  }
        };

    PacketImage(IVelociaBufferManager * vbm, size_t full_packet_size_ints)
        :  VBM(vbm), PacketSize( full_packet_size_ints )
        {}
    PacketImage(IVelociaBufferManager * vbm, const PacketImage::Range & range)
        :  VBM(vbm), PacketSize( range.SizeInInts )
        {
        AddRange(range);
        }
    PacketImage(const PacketImage & image)
        :  VBM(image.VBM), PacketSize( image.PacketSize ), Ranges(image.Ranges)
        {
        }

    //
    //  Image Information
    size_t        HeaderSizeInInts() const;
    size_t        SizeInInts() const;
    size_t        FullSizeInInts() const;
    unsigned int  PeripheralId() const ;
    void          PeripheralId(unsigned int id);
    size_t        PacketFullSize() const;


    void  ClearRanges()
        {  Ranges.clear();  }
    void  AddRange( const Range & range )
        {  Ranges.push_back( range );  }
    bool  empty() const
        {  return Ranges.size()==0;  }
    size_t ImageRangeCount() const
        {  return Ranges.size();  }
    Range ImageRange(size_t idx) const
        {  return Ranges[idx];  }

    void  CopyDataTo( Buffer & buffer ) const;
    void  CopyDataTo( Buffer & buffer, size_t MaxIntsCopied ) const;

private:
    //
    //  Member Data
    IVelociaBufferManager  *  VBM;
    size_t                    PacketSize;
    std::vector<Range>        Ranges;
};

//==============================================================================
// CLASS UnformattedDataImage  --  Unaligned Packet Data Image
//==============================================================================

class UnformattedDataImage
{
public:
    UnformattedDataImage(IVelociaBufferManager * vbm, size_t slab_size_ints)
        :  Image(vbm, slab_size_ints)
        {}

    unsigned int *  Address()
        { return (Image.empty()) ? 0 : Image.ImageRange(0).begin();  };
    size_t          SizeInInts()
        { return (Image.empty()) ? 0 : Image.ImageRange(0).SizeInInts; }

    PacketImage  Image;
};

//==============================================================================
// CLASS PacketImageParserImageAvailable  --  Parser Image of Packet event
//==============================================================================

class PacketImageParserImageAvailable : public OpenWire::Event
{
public:
    class BasicPacketImageParser *  Sender;
    PacketImage                Image;

public:
    PacketImageParserImageAvailable( BasicPacketImageParser * sender, const PacketImage & image )
        :  Sender( sender ), Image(image)
        {
        }
};

//==============================================================================
// CLASS BasicPacketImageParser  --  Parse packets in a flat memory image list
//==============================================================================

class BasicPacketImageParser
{
public:
    BasicPacketImageParser(IVelociaBufferManager * vbm)
         :  VBM(vbm), TotalProcessed(0)
        {}
    virtual ~BasicPacketImageParser()
        {}
    //
    //  Operations
    void    Clear();
    void    Append(const PacketImage & Image);
    void    Parse();
	void    LoadManager(IVelociaBufferManager * mgr)
		{  VBM = mgr;  }
	//
    //  ...data processed counter
    size_t  ProcessedInInts() const
                {  return TotalProcessed;  }
    void    ResetProcessed()
                {  TotalProcessed = 0;  }

    size_t  Depth() const;
    //
    //  Events
    OpenWire::EventHandler<PacketImageParserImageAvailable>  OnImageAvailable;

private:
    typedef std::deque<PacketImage::Range>  RangeQueue;
    //
    //  Member Data
	IVelociaBufferManager  *  VBM;

    RangeQueue     Ranges;
    size_t         TotalProcessed;
    //
    //  Member Functions
    PacketImage    ParsePacketImage();

};

//==============================================================================
// CLASS PacketImageParser  --  Parse PMC packets in a flat memory image list
//==============================================================================

class PacketImageParser : public BasicPacketImageParser
{
public:
	PacketImageParser(class PacketStream * stream = 0)
         :  BasicPacketImageParser(&P_VBM), FStream(stream)
        {}
    virtual ~PacketImageParser()
        {}

	PacketStream *   Stream() const
				{  return FStream;  }

private:
	PmcBufferManager P_VBM;
	PacketStream * FStream;
};

typedef PacketImageParser  PmcPacketImageParser;

//==============================================================================
// CLASS VeloPacketImageParser  --  Parse Velo packets in a flat memory image list
//==============================================================================

class VeloPacketImageParser : public BasicPacketImageParser
{
public:
	VeloPacketImageParser(class VitaPacketStream * stream = 0)
		 :  BasicPacketImageParser(&V_VBM), FStream(stream)
		{}
	virtual ~VeloPacketImageParser()
		{}

	class VitaPacketStream *   Stream() const
				{  return FStream;  }
private:
	VeloBufferManager V_VBM;
	class VitaPacketStream * FStream;
};

#ifdef __CLR_VER
#pragma managed(pop)
#endif
}  // end namespace Innovative

#endif
