// PacketPoll_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//


#include "PacketPoll_Mb.h"
#include <iostream>
#include "SystemSupport_Mb.h"
#include "BufferHeader_Mb.h"
#include "BufferDatagrams_Mb.h"
#include "Exception_Mb.h"

namespace Innovative
{

//===========================================================================
//  CLASS PacketPollQueue  -- Queue to store data to/from application and board
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class PacketPollQueue --
//---------------------------------------------------------------------------

PacketPollQueue::PacketPollQueue()
    :  CapacityInBlks(10)
{
}

//---------------------------------------------------------------------------
//  constructor for class PacketPollQueue --
//---------------------------------------------------------------------------

PacketPollQueue::PacketPollQueue(size_t capacity)
    :  CapacityInBlks(capacity)
{
}

//---------------------------------------------------------------------------
//  PacketPollQueue::Push() -- push buffer onto queue
//      Buffer & packet - packet to queue
//      return : true  - packet successfully inserted into queue
//               false - otherwise
//---------------------------------------------------------------------------

bool  PacketPollQueue::Push( Buffer & packet )
{
    if (Depth() < CapacityInBlks)
        {
        CriticalSectionArea guard(lock);
        BufferQ.push(packet);
        return true;
        }

    return false;
 }

//---------------------------------------------------------------------------
//  PacketPollQueue::Pop() -- remove block from queue
//      Buffer & packet - packet from queue
//      return : true  - packet successfully extracted from queue
//               false - otherwise
//---------------------------------------------------------------------------

bool PacketPollQueue::Pop(Buffer & packet)
{
    if (Depth() > 0)
        {
        CriticalSectionArea guard(lock);
        packet = BufferQ.front();
        BufferQ.pop();
        return true;
        }

    return false;
}

//---------------------------------------------------------------------------
//  PacketPollQueue::Clear() -- clear queue
//---------------------------------------------------------------------------

void  PacketPollQueue::Clear()
{
    //  clear buffer queue
    BufferQ.clear();
}

//---------------------------------------------------------------------------
//  PacketPollQueue::Depth() -- depth of queue
//---------------------------------------------------------------------------

size_t PacketPollQueue::Depth()
{
    return BufferQ.depth();
}

//---------------------------------------------------------------------------
//  PacketPollQueue::QueueCapacity() -- Set queue capacity
//---------------------------------------------------------------------------

void  PacketPollQueue::QueueCapacity(size_t cap)
{
    CriticalSectionArea guard(lock);
    CapacityInBlks = cap;
}

//---------------------------------------------------------------------------
//  PacketPollQueue::QueueCapacity() -- Retreive capacity
//---------------------------------------------------------------------------

size_t PacketPollQueue::QueueCapacity() const
{
    return CapacityInBlks;
}

//===========================================================================
//  CLASS PacketPollThread  -- Base class for Packet Threads
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class PacketPollThread
//---------------------------------------------------------------------------

PacketPollThread::PacketPollThread(const char* tName/*,
                                       PacketPollMgr *owner*/)
        : inherited(tName)//, /*Owner(owner),*/ //
{
    Priority(tpHigher);

//    CpuAffinityVec cpus;
//    cpus.push_back(1);
//    SetCpuAffinity(cpus);

    loopCount = 0;
    pktRxCount = 0;
    pktTxCount = 0;
    emptyRxFifoCount = 0;
    fullTxFifoCount = 0;
    alertCount = 0;
    once = false;
}

//---------------------------------------------------------------------------
//  destructor for class PacketPollThread
//---------------------------------------------------------------------------

PacketPollThread::~PacketPollThread()
{
    inherited::Shutdown(true);
}


//---------------------------------------------------------------------------
//  PacketPollThread::Init()
//---------------------------------------------------------------------------

void PacketPollThread::Init()
{
    inherited::Init();
}


//---------------------------------------------------------------------------
//  PacketPollThread::Start()
//---------------------------------------------------------------------------

void PacketPollThread::Start()
{
    inherited::Start();
}

//---------------------------------------------------------------------------
//  PacketPollThread::Stop()
//---------------------------------------------------------------------------

void PacketPollThread::Stop()
{
    inherited::Stop();
}

//===========================================================================
//  CLASS PacketPollDataThread  -- Polling thread to process packets received to/from board
//===========================================================================

void PacketPollDataThread::Start()
{
    packet.ResizeHeader(2);
    inherited::Start();
}

//---------------------------------------------------------------------------
//  PacketPollDataThread::GetPacketFromFifo() -- read and assemble a data packet from HW FIFO
//---------------------------------------------------------------------------
bool PacketPollDataThread::GetPacketFromFifo(Buffer & buffer)
{
    //read 1st word from fifo
    ii32 firstHeaderWord = Owner->FifoExt.ReadRecvFifoWord();

    //packet size has to be > 0
    if(firstHeaderWord != 0)
    {
        unsigned char pid = firstHeaderWord & 0xff000000;
        const unsigned char a2dPid = 0;
        const unsigned char alertPid = 0xff;

        //match with header
        if ((pid == a2dPid ) ||
            (pid == alertPid))
        {
            *buffer.Header() = firstHeaderWord;

            while(Owner->FifoExt.RecvFifoWordCount() == 0);

            //2nd word always 0
            if(Owner->FifoExt.ReadRecvFifoWord() == 0)
            {
                pktSize = firstHeaderWord & 0x00ffffff;
                buffer.Resize(pktSize-2);
                IntegerDG Packet_DG(buffer);

                for(ii32 i=0; i<pktSize-2; i++)
                {
                    while(Owner->FifoExt.RecvFifoWordCount() == 0)
                    {
                        Debug.Log("Rx Fifo empty");
                    }

                    Packet_DG[i] = Owner->FifoExt.ReadRecvFifoWord();
                }

                return true;
            }
        }
    }
    return false;
}

//---------------------------------------------------------------------------
//  PacketPollDataThread::Execute() -- per loop execution
//---------------------------------------------------------------------------
void PacketPollDataThread::SingleExecution()
{
    loopCount++;

    //read packet from board
    if (Owner->FifoExt.RecvFifoWordCount() > 0)
    {

        if (GetPacketFromFifo(packet))
        {
             if(Owner->tsLogger.State())
                Owner->tsLogger.Mark();

            if (Owner->DeviceInternals()->AlertCheck(packet) )
            {
                //data packet, add it to queue
                if (Owner->RecvQ.Push(packet))
                {
                    Owner->OnDataAvailable.Execute(event);
                    pktRxCount++;
                }
            }
            else
                alertCount++;
        }
        else
            emptyRxFifoCount++;
    }

    //request tx data
    Owner->OnDataRequired.Execute(event);

    if (Owner->SendQ.Pop(packet))
    {
        // write packet to tx fifo reg
        if((PacketPollMgr::MAX_HW_FIFO-Owner->FifoExt.SendFifoWordCount()) > pktSize)
        {
            PutPacketIntoFifo(packet);
            pktTxCount++;

             if(Owner->tsLogger.State())
                Owner->tsLogger.Mark();
        }
        else
            fullTxFifoCount++;
    }

}

//---------------------------------------------------------------------------
//  PacketPollDataThread::PutPacketIntoFifo() -- put packet on FIFO for transmission
//---------------------------------------------------------------------------
bool PacketPollDataThread::PutPacketIntoFifo(Buffer & packet)
{
    //Debug.Log("Pkt #", pktTxCount);
    for(ii32 i=0; i<packet.HeaderSizeInInts(); i++)
    {
        Owner->FifoExt.WriteSendFifoWord(*(packet.Header()+i));
        //Debug.Log("> ", *(packet.Header()+i), 16);
    }

    for(ii32 i=0; i<packet.DatagramSize(); i++)
    {
        Owner->FifoExt.WriteSendFifoWord(*(packet.Data()+i));
        //Debug.Log("> ", *(packet.Data()+i), 16);
    }

    return true;
}

//===========================================================================
//  CLASS PacketPollMgr  -- This is the main interface class between application layer
//                          and Mallibu's internal polling operations
//===========================================================================

const ii32 PacketPollMgr::MAXQSIZE = 512;
const ii32 PacketPollMgr::MAX_HW_FIFO = 512;

//---------------------------------------------------------------------------
//  PacketPollMgr() --
//---------------------------------------------------------------------------
PacketPollMgr::PacketPollMgr(IFifoExtension &fx) :
    PktPollThread(this), FifoExt(fx),
    RecvQ(MAXQSIZE), SendQ(MAXQSIZE), DeviceLink(this)
{
    State = OPEN;
    PktPollThread.Resume();
}

//---------------------------------------------------------------------------
//  ~ PacketPollMgr() --
//---------------------------------------------------------------------------
PacketPollMgr::~PacketPollMgr()
{
    if(State != STOP)
    {
        Stop();
    }

    State = CLOSE;

    tsLogger.Stop();
    tsLogger.Disable();
}

//---------------------------------------------------------------------------
//  PacketPollMgr::Start() -- start polling loop for data
//---------------------------------------------------------------------------
void PacketPollMgr::Start()
{
    if (State == OPEN || State == STOP)
    {
        //setup queues
        SendQ.Clear();
        RecvQ.Clear();

        //setup threads
        PktPollThread.Start();

        State = POLLING;
    }
}



//---------------------------------------------------------------------------
//  PacketPollMgr::Stop() -- stop polling for data
//---------------------------------------------------------------------------
void PacketPollMgr::Stop()
{
    if(State == POLLING)
        {
        PktPollThread.Stop();
        State = STOP;
        }

    tsLogger.Save();
}

//---------------------------------------------------------------------------
//  PacketPollMgr::Send() -- Send packet to device
//      throws 'bad state'
//---------------------------------------------------------------------------

void   PacketPollMgr::Send( Buffer & packet )
{
    if(State == POLLING)
        {
        // push packet on queue
        SendQ.Push(packet);
        }
    else
        throw Innovative::MalibuException("PacketPollMgr::Send(): polling stopped.");
}

//---------------------------------------------------------------------------
//  PacketPollMgr::Recv() -- Recieve packet from device
//---------------------------------------------------------------------------

void PacketPollMgr::Recv( Buffer & packet )
{
    if(State == POLLING)
        {
        if (!RecvQ.Pop(packet))
            throw Innovative::MalibuException("Receive queue is empty.");
        }
    else
        throw Innovative::MalibuException("PacketPollMgr::Recv(): polling stopped.");
}

//---------------------------------------------------------------------------
//  PacketPollMgr::SetFifoDacDelay() -- Set delay time between ADC and DAC FIFOs
//  Return: success - delay in ns
//          0       - otherwise
//---------------------------------------------------------------------------

ii32 PacketPollMgr::SetFifoDacDelay(ii32 ns)
{
    const ii32 TICK2NS = 15;

    //max=15*(2^26-1)
    const ii32 MAX=TICK2NS*static_cast<const ii32>(pow(2.0, 26)-1);
    if(ns >= 15 && ns <= MAX)
        {
        FifoExt.SetDacDelay(ns/TICK2NS);
        return (FifoExt.GetDacDelay()*TICK2NS);
        }
    else
        return 0;
}

//---------------------------------------------------------------------------
//  PacketPollMgr::ConnectTo() -- connect poll manager to a pollable device
//---------------------------------------------------------------------------

bool PacketPollMgr::ConnectTo(IPacketStreamDevice * aStreamDevice)
{
    if (aStreamDevice)
        {
        DeviceLink.Link(aStreamDevice);
        return true;
        }
    else
        return false;
}

//---------------------------------------------------------------------------
//  PacketPollMgr::Disconnect() -- disconnect poll manager to a pollable device
//---------------------------------------------------------------------------

void PacketPollMgr::Disconnect()
{
    DeviceLink.Unlink();
}

//===========================================================================
//  CLASS FmcPacketPollDataThread  -- Polling thread to process packets received to/from board
//===========================================================================

//---------------------------------------------------------------------------
//  FmcPacketPollDataThread::Init()
//---------------------------------------------------------------------------

void FmcPacketPollDataThread::Init()
{
    inherited::Init();
}

//---------------------------------------------------------------------------
//  FmcPacketPollDataThread::Start()
//---------------------------------------------------------------------------

void FmcPacketPollDataThread::Start()
{
    if (Owner->IsValid())
        {
        InputPacketSize = Owner->Adc()->AlmostFullLevel();
        }

    inherited::Start();
}

//---------------------------------------------------------------------------
//  FmcPacketPollDataThread::GetPacketFromFifo() --
//---------------------------------------------------------------------------
//  read and assemble a data packet from HW FIFO

bool FmcPacketPollDataThread::GetPacketFromFifo(Buffer & buffer)
{
    buffer.Resize(InputPacketSize);   // this size is "AF" level for input fifo

    IntegerDG BufferDG(buffer);
    for (size_t i=0; i < BufferDG.size(); i++)
         BufferDG[i] = Owner->Adc()->Data();

    return true;
}

//---------------------------------------------------------------------------
//  FmcPacketPollDataThread::PutPacketIntoFifo() -- Write data to output fifo
//---------------------------------------------------------------------------

bool FmcPacketPollDataThread::PutPacketIntoFifo(Buffer & buffer)
{
    IntegerDG BufferDG(buffer);
    for (size_t i=0; i<BufferDG.size(); i++)
         Owner->Dac()->Data(BufferDG[i]);

    return true;
}

//---------------------------------------------------------------------------
//  FmcPacketPollDataThread::SingleExecution() -- per loop execution
//---------------------------------------------------------------------------

void FmcPacketPollDataThread::SingleExecution()
{
    if (Owner->IsValid()==false)
        return;

    loopCount++;

    // Check for Incoming Data
    if (Owner->Adc()->AlmostFull())
        {
        // We have a sample event, read it
        bool rtnval = GetPacketFromFifo(packet);
        if (rtnval)
            {
            // we have a valid packet, process it
            if (Owner->tsLogger.State())
                Owner->tsLogger.Mark();

            if (Owner->RecvQ.Push(packet))
                {
                FmcPacketPollDataEvent event(Owner);
                Owner->OnDataAvailable.Execute(event);
                pktRxCount++;
                }
            }


    emptyRxFifoCount++;

    // Check for ability to write outgoing data
    if (Owner->Dac()->AlmostEmpty())
        {
        FmcPacketPollDataEvent event(Owner);
        Owner->OnDataRequired.Execute(event);

        if (Owner->SendQ.Pop(packet))
            {
            PutPacketIntoFifo(packet);
            pktTxCount++;
            }
        }
        }

    Sleep(1);

    // Check for Alert Data???
    //    alertCount++;
}

//===========================================================================
//  CLASS FmcPacketPollManager
//===========================================================================
//  Interface to FMC-Servo Packet Polling Interface

const ii32 FmcPacketPollManager::MAXQSIZE = 512;
const ii32 FmcPacketPollManager::MAX_HW_FIFO = 512;

//---------------------------------------------------------------------------
//  constructor for class FmcPacketPollManager() --
//---------------------------------------------------------------------------

FmcPacketPollManager::FmcPacketPollManager()
  :   AdcFifo(0), DacFifo(0),
      Thread(this),
      RecvQ(MAXQSIZE), SendQ(MAXQSIZE), DeviceLink(this)
{
    State = OPEN;
    Thread.Resume();
}

//---------------------------------------------------------------------------
//  destructor for class FmcPacketPollManager() --
//---------------------------------------------------------------------------

FmcPacketPollManager::~FmcPacketPollManager()
{
    if(State != STOP)
        {
        Stop();
        }

    State = CLOSE;

    tsLogger.Stop();
    tsLogger.Disable();
}

//---------------------------------------------------------------------------
//  FmcPacketPollManager::Start() -- start polling loop for data
//---------------------------------------------------------------------------

void  FmcPacketPollManager::Start()
{
    if (State == OPEN || State == STOP)
        {
        //setup queues
        SendQ.Clear();
        RecvQ.Clear();

        State = POLLING;

        //setup threads
        Thread.Start();
        }
}

//---------------------------------------------------------------------------
//  FmcPacketPollManager::Stop() -- stop polling for data
//---------------------------------------------------------------------------

void FmcPacketPollManager::Stop()
{
    if(State == POLLING)
        {
        Thread.Stop();
        State = STOP;
        }

    tsLogger.Save();
}

//---------------------------------------------------------------------------
//  FmcPacketPollManager::Send() -- Send packet to device
//      throws 'bad state'
//---------------------------------------------------------------------------

void   FmcPacketPollManager::Send(VeloBuffer& packet )
{
    if(State == POLLING)
        {
        // push packet on queue
        SendQ.Push(packet);
        }
    else
        throw Innovative::MalibuException("FmcPacketPollManager::Send(): polling stopped.");
}

//---------------------------------------------------------------------------
//  FmcPacketPollManager::Recv() -- Recieve packet from device
//---------------------------------------------------------------------------

void FmcPacketPollManager::Recv(VeloBuffer& packet )
{
    if (State == POLLING)
        {
        if (!RecvQ.Pop(packet))
            throw Innovative::MalibuException("Receive queue is empty.");
        }
    else
        throw Innovative::MalibuException("FmcPacketPollManager::Recv(): polling stopped.");
}

//---------------------------------------------------------------------------
//  FmcPacketPollManager::SetFifoDacDelay() -- Set delay time between ADC and DAC FIFOs
//---------------------------------------------------------------------------
//  Return: success - delay in ns
//          0       - otherwise

//ii32 FmcPacketPollManager::SetFifoDacDelay(ii32 ns)
//{
//    const ii32 TICK2NS = 15;

//    //max=15*(2^26-1)
//    const ii32 MAX=TICK2NS*static_cast<const ii32>(pow(2.0, 26)-1);
//    if(ns >= 15 && ns <= MAX)
//        {
//        FifoExt.SetDacDelay(ns/TICK2NS);
//        return (FifoExt.GetDacDelay()*TICK2NS);
//        }
//    else
//        return 0;
//}

//---------------------------------------------------------------------------
//  FmcPacketPollManager::ConnectTo() -- connect poll manager to a pollable device
//---------------------------------------------------------------------------

bool FmcPacketPollManager::ConnectTo(IVitaPacketStreamDevice * aStreamDevice)
{
    if (aStreamDevice)
        {
        DeviceLink.Link(aStreamDevice);
        return true;
        }
    else
        return false;
}

//---------------------------------------------------------------------------
//  FmcPacketPollManager::Disconnect() -- disconnect poll manager to a pollable device
//---------------------------------------------------------------------------

void FmcPacketPollManager::Disconnect()
{
    DeviceLink.Unlink();
}
























//---------------------------------------------------------------------------
//  QuickPacketPollMgr() --
//---------------------------------------------------------------------------
QuickPacketMgr::QuickPacketMgr(IFifoExtension &fx) :
    A2dPid(0x0), AlertPid(0xff), FifoExt(fx),
    Running(false), MAX_HW_FIFO(512), DeviceLink(this)
{
}

//---------------------------------------------------------------------------
//  ~ QuickPacketPollMgr() --
//---------------------------------------------------------------------------
QuickPacketMgr::~QuickPacketMgr()
{
}

//---------------------------------------------------------------------------
//  QuickPacketPollMgr::ConnectTo() -- connect poll manager to a pollable device
//---------------------------------------------------------------------------

bool QuickPacketMgr::ConnectTo(IPacketStreamDevice * aStreamDevice)
{
    if (aStreamDevice)
        {
        DeviceLink.Link(aStreamDevice);
        return true;
        }
    else
        return false;
}

//---------------------------------------------------------------------------
//  QuickPacketPollMgr::Disconnect() -- disconnect poll manager to a pollable device
//---------------------------------------------------------------------------

void QuickPacketMgr::Disconnect()
{
    DeviceLink.Unlink();
}

//---------------------------------------------------------------------------
//  QuickPacketMgr::GetPacketFromFifo() -- read and assemble a data packet from HW FIFO
//---------------------------------------------------------------------------
bool QuickPacketMgr::GetPacketFromFifo(Buffer & buffer)
{
    Trace debug;
    // Wait for Full Header...
    while(FifoExt.RecvFifoWordCount() < 2)
       if (StreamStopped) return false;
    PacketBufferHeader Hdr(buffer);
    // Read 1st word from FIFO(Aka. 1st Header word)
    Hdr[0] = FifoExt.ReadRecvFifoWord();
    // Read 2nd word from FIFO(Aka. 2nd Header word)
    Hdr[1] = FifoExt.ReadRecvFifoWord();

	ii32 pktSize = static_cast<ii32>(Hdr.PacketSize());
    // Nothing to do. Wait for Data...
   while(FifoExt.RecvFifoWordCount() < pktSize-2)
      if (StreamStopped) return false;
    // Stuff the data into the Packet
    buffer.Resize(pktSize-2);
    IntegerDG    Packet_DG(buffer);
    for(ii32 i = 0; i< pktSize-2; i++)
        Packet_DG[i] = FifoExt.ReadRecvFifoWord();

    return true;
}

//---------------------------------------------------------------------------
//  QuickPacketMgr::PutPacketIntoFifo() -- put packet on FIFO for transmission
//---------------------------------------------------------------------------
bool QuickPacketMgr::PutPacketIntoFifo(Buffer & packet)
{
    for(ii32 i = 0; i < packet.HeaderSizeInInts(); i++)
        {
        FifoExt.WriteSendFifoWord(*(packet.Header()+i));
        }

    for(ii32 i = 0; i < packet.DatagramSize(); i++)
        {
        FifoExt.WriteSendFifoWord(*(packet.Data()+i));
        }

    return true;
}

//---------------------------------------------------------------------------
//  QuickPacketMgr::HasTxFifoRoom() --
//---------------------------------------------------------------------------
bool QuickPacketMgr::HasTxFifoRoom(const Buffer & packet)
{
    size_t sz_needed  = packet.FullSizeInInts();
    size_t sz_free    = MAX_HW_FIFO - FifoExt.SendFifoWordCount();

    return (sz_free > sz_needed);
}
} // namespace Innovative
