//---------------------------------------------------------------------------
// PacketStreamBase_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2010 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <PacketStreamBase_Mb.h>
#include <StartStopThread_Mb.h>
#include <StopWatch_Mb.h>
#include <SystemSupport_Mb.h>

namespace Innovative
{
//===========================================================================
//  CLASS PacketStreamIPSQCreditor  -- Credit manager for IPSQueue
//===========================================================================

class PacketStreamIPSQCreditor : public IPSQueueCreditor
{
public:
    PacketStreamIPSQCreditor(PacketStreamBase * owner, int stream_channel);

    virtual void  CreditStream(size_t cq_token);

private:
    //
    //  Member Data
    PacketStreamBase * Owner;
    unsigned int  StreamChannel;
};

//---------------------------------------------------------------------------
//  constructor for class PacketStreamIPSQCreditor
//---------------------------------------------------------------------------

PacketStreamIPSQCreditor::PacketStreamIPSQCreditor(PacketStreamBase * owner, int stream_channel)
    : Owner(owner), StreamChannel(stream_channel)
{
}

//---------------------------------------------------------------------------
//  PacketStreamIPSQCreditor::CreditStream() -- grant credit
//---------------------------------------------------------------------------

void  PacketStreamIPSQCreditor::CreditStream(size_t cq_token)
{
    //  token is packet size in words...
    size_t PacketSize = cq_token;
    Owner->DeviceInternals()->GrantCredit(StreamChannel, PacketSize);
}

//===========================================================================
//  CLASS PacketStreamIoThread  -- I/O thread for PacketStreams
//===========================================================================
//  Threads handling the 'application' side of the buffer queues from the
//    VISR. They handle the interface between these queues and the app

class PacketStreamIoThread : public BasicSSThread
{
    friend class PacketStream;
    typedef BasicSSThread inherited;

public:
    // Ctor
    PacketStreamIoThread(PacketStreamBase *owner, PacketStreamQueue & queue)
        : inherited("PacketStreamIoThread"), Owner(owner), Queue(queue)
        {  Priority(tpHigher);  }

protected:
    class LoadBalance
    {
    public:
        // Ctor
        LoadBalance()
            : Then(0), Sleeps(0), Packets(0)
            {  }

        // Load balancing:  Reserve 10% of CPU for foreground processing
        void Balance()
        {
            ++Packets;
            if ((Time.Now() - Then) > 10.0e-3)
                {
                Sleep(1);
                ++Sleeps;
                Then = Time.Now();
                }
        }

        // Data
        double          Then;
        StopWatch       Time;
        unsigned int    Sleeps;
        unsigned int    Packets;
    };

    LoadBalance         Load;
    PacketStreamBase *  Owner;

    PacketStreamQueue &       Queue;

    enum { sQueue=inherited::sLast, sLast };
    //  override to load notify events at start
    virtual void    Init()
        {
        inherited::Init();
        AddObject(Queue.SignalSem(), sQueue);
        }
    //  do one iteration of main process
    virtual void    AtExecution(int result)
        {
        if (result == sQueue)
            ProcessPacket();
        }

    virtual void    ProcessPacket() = 0;

    void            Log(int tag, int value)
        {  Owner->DeviceInternals()->TraceLog(tag, value);  }
    void            Log(int tag, size_t value)
        {  Log(tag, static_cast<int>(value));  }
};

//===========================================================================
//  CLASS PacketStreamRcvrThread  -- Receiver thread for PktStreams
//===========================================================================
//---------------------------------------------------------------------------
//  declaration of  class PacketStreamRcvrThread
//---------------------------------------------------------------------------

class PacketStreamRcvrThread : public PacketStreamIoThread
{
public:
    PacketStreamRcvrThread(PacketStreamBase * owner)
        : PacketStreamIoThread(owner, owner->RecvQ)
        {  }

    // Methods
    void  Clear()
        {  Signal.Clear();  }
    void  Recv(Buffer & packet);

protected:
    Semaphore       Signal;

    virtual void    ProcessPacket();
};

//---------------------------------------------------------------------------
//  PacketStreamRcvrThread::ProcessPacket() -- Recv data packet
//---------------------------------------------------------------------------

void  PacketStreamRcvrThread::ProcessPacket()
{
//    int depth_check = Queue.Depth();
//    Log(8, depth_check);

    //
    //  Signal packet available downstream...
    Signal.Release();

    Owner->SignalDataAvailable();

//    PacketStreamDataEvent e(Owner);
//    Owner->OnDataAvailable.Execute(e);

    // Load balancing
    if (Owner->RxLoadBalancing())
        Load.Balance();

//    depth_check = Queue.Depth();
//    Log(9, depth_check);
}

//---------------------------------------------------------------------------
//  PacketStreamRcvrThread::Recv() -- Return packet in data queue
//---------------------------------------------------------------------------

void  PacketStreamRcvrThread::Recv( Buffer & packet )
{
    int depth_check = static_cast<int>(Queue.Depth());
    Log(0x18, depth_check);

    if (depth_check>0x100)
        {
        //  Spinning out of control now?
        Log(0x88, depth_check);
        }
    //
    //  Wait for packet to arrive....
    Signal.Acquire();
    //
    //  Extract packet from queue
    packet = Queue.Front();
    //
    //  Pop the queue
    Queue.Pop();

    depth_check = static_cast<int>(Queue.Depth());
    Log(0x19, depth_check);
}

//===========================================================================
//  CLASS PacketStreamXmitThread  -- Receiver thread for PktStreams
//===========================================================================
//  This thread is on the 'application' side of the buffer queues from the
//    VISR. It handles the interface from these queues to the app or data pins
//
//---------------------------------------------------------------------------
//  declaration of  class PacketStreamXmitThread
//---------------------------------------------------------------------------

class PacketStreamXmitThread : public PacketStreamIoThread
{
public:
    PacketStreamXmitThread(PacketStreamBase *owner)
        : PacketStreamIoThread(owner, owner->XmitQ), Prefilling(false), PrefilledOnce(false)
        {  }

    void  Send( Buffer & packet );
    void  Clear()
        {  Queue.ClearSem();  }

    void  Prefill(int packets);
	

protected:
    // Data
    bool            Prefilling;

    bool            PrefilledOnce;

    // Methods
    virtual void    ProcessPacket( );
	virtual void    AtStop();
    void            Signal();

    virtual void    XXX_DebugLog(int type, int code);

};

void    PacketStreamXmitThread::XXX_DebugLog(int type, int code)
    {  if (PrefilledOnce) {  Log(type, code); }  }
//---------------------------------------------------------------------------
//  PacketStreamXmitThread::ProcessPacket() -- packet needed by transmit
//---------------------------------------------------------------------------

void  PacketStreamXmitThread::ProcessPacket()
{
    //  Log(0xc, 0x00);

    if (Prefilling)
        {
        //  Log(0xc, 0xFA);
        return;
        }

    Signal();
    //  Log(0xc, 0xFF);
}

//---------------------------------------------------------------------------
//  PacketStreamXmitThread::AtStop() -- Special Code when Stopped
//---------------------------------------------------------------------------

void  PacketStreamXmitThread::AtStop()
{
    Log(0xec, 0x01);

	Clear();
	
    Log(0xec, 0x00);
}

//---------------------------------------------------------------------------
//  PacketStreamXmitThread::Signal() -- packet needed by transmit
//---------------------------------------------------------------------------

void  PacketStreamXmitThread::Signal()
{
    //int depth_check = Queue.Depth();
    //Log(0xa1, depth_check);

    //Log(0xa2, 1);
    //
    //  Signal upstream for data...
    Owner->SignalDataRequired();
    
    //Log(0xa2, 2);
    // Load balancing
    if (Owner->TxLoadBalancing())
        Load.Balance();

    //Log(0xa2, 3);
    //depth_check = Queue.Depth();
    //Log(0xa0, depth_check);
}

//---------------------------------------------------------------------------
//  PacketStreamXmitThread::Send() -- Load packet for transmission
//---------------------------------------------------------------------------

void  PacketStreamXmitThread::Send( Buffer & packet )
{
    size_t depth_check = Queue.Depth();
    Log(0x1a, depth_check);

    //
    //  Push data onto outgoing queue
    Queue.Push(packet);
    //Log(0x1c, depth_check);

    depth_check = Queue.Depth();
    Log(0x1b, depth_check);
}

//---------------------------------------------------------------------------
//  PacketStreamXmitThread::Prefill() -- Wait for prefill to complete
//---------------------------------------------------------------------------
    
void  PacketStreamXmitThread::Prefill(int packets)
{
    Prefilling = true;
    PrefilledOnce = true;

    for (int i = 0; i < packets; ++i)
        Signal();
    Prefilling = false;
}

//===========================================================================
//  CLASS PacketStreamStoppageThread  -- Proper Termination signalling
//===========================================================================
//  We want a callback to happen when it is safe to call Stream:;Start() again
//    this thread watches the condition and calls the handler when ready

class PacketStreamStoppageThread : public BasicSSThread
{
//    friend class PacketStream;
    typedef BasicSSThread inherited;

public:
    // Ctor
    PacketStreamStoppageThread(PacketStreamBase *owner)
        : inherited("PktStreamStopThd"), Owner(owner)
        {  Priority(tpNormal);  }

    void  StartMonitoring()
            {  eMonitor.Set();  }
    void  EndMonitoring()
            {  eMonitor.Clear();  }

protected:

    PacketStreamBase *  Owner;
    Event               eMonitor;

    enum { sMonitor=inherited::sLast, sLast };
    virtual void    Init()
        {
        inherited::Init();
        AddObject(eMonitor, sMonitor);
        }

    //  do one iteration of main process
    virtual void    AtExecution(int result);

    void            Log(int tag, int value)
        {  Owner->DeviceInternals()->TraceLog(tag, value);  }
    void            Log(int tag, size_t value)
        {  Log(tag, static_cast<int>(value));  }
};

//---------------------------------------------------------------------------
//  PacketStreamStoppageThread::AtExecution() -- Internal Interface
//---------------------------------------------------------------------------

void  PacketStreamStoppageThread::AtExecution(int result)
{
    if (result == sMonitor)
        {
        Log(0xE3, 0x1);

        Owner->RecvThread->WaitForStopped_External();
        Log(0xE3, 0x8);
        Owner->XmitThread->WaitForStopped_External();
        Log(0xE3, 0x7);

        while (Owner->DeviceInternals()->VisrIsrQDepth() != 0)
            {
            Innovative::uSleep(100);
            Log(0xE3, 0x6);
            }
            
        Log(0xE3, 0x5);

        EndMonitoring();  // only call self once!

        Log(0xE3, 0x0);
        OpenWire::NotifyEvent e;
        Owner->OnAfterStop.Execute(e);
        }
}

//===========================================================================
//  CLASS TIPacketInternalsImpl  -- Internal Interface Impl
//===========================================================================

class TIPacketInternalsImpl : public IPacketStreamManagerInternals
{
public:
    TIPacketInternalsImpl(PacketStreamBase * owner)
        :  Owner(owner) {}

    virtual void ForwardDirectDataPacket(BmPacket & pkt) {  Owner->ForwardDirectDataPacket(pkt); }
    virtual void ForwardDirectAlertPacket(BmPacket & pkt) {  Owner->ForwardDirectAlertPacket(pkt); }
    virtual void EnqueueRecvPacket(Buffer & buffer) {  Owner->EnqueueRecvPacket(buffer); }
    virtual void ProcessDataSlab(UnformattedDataImage & slab) { Owner->ProcessDataSlab(slab); };
    virtual void FillDataSlab(UnformattedDataImage & slab)  { Owner->FillDataSlab(slab);  };
    virtual Buffer NextXmitPacket() {  return Owner->NextXmitPacket(); }
    virtual void DequeueXmitPacket() {  Owner->DequeueXmitPacket(); }
    virtual size_t XmitQDepth() {  return Owner->XmitQDepth(); }
    virtual size_t RecvQDepth() {  return Owner->RecvQDepth(); }
    virtual PacketStreamDataModes  PacketProcessingMode()  {  return Owner->FStreamDataMode; }
    virtual bool  HasStreamStarted() { return Owner->DoStreamStarted(); };
    virtual void   XmitQSignal(size_t count) {  Owner->XmitQSignal(count); }
    virtual void   DispatchCustomInterrupt(int custom_flags) { Owner->DispatchCustomInterrupt(custom_flags); }
    virtual IVelociaBufferManager *  VelociaManager() { return Owner->VelociaBufferManager(); };

private:
    PacketStreamBase * Owner;
};

//===========================================================================
//  CLASS PacketStreamBase  -- Packet Based Stream I/O Base Class
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class PacketStreamBase
//---------------------------------------------------------------------------
#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif
PacketStreamBase::PacketStreamBase()
    : InternalOps(new TIPacketInternalsImpl(this)),
      RecvThread(0),
      XmitThread(0),
      XmitQ(new PacketStreamIPSQCreditor(this, PciOutData)),
      RecvQ(new PacketStreamIPSQCreditor(this, PciInData)),
      Started(false),
      FPrefillPacketCount(10),
      FRecvPacketCache(10),
      FStreamDataMode(ppmNormal),
      FRxLoadBalancing(true), FTxLoadBalancing(true),
      LocalCreditor(new PacketStreamIPSQCreditor(this, PciInData))
{
    RecvThread = new PacketStreamRcvrThread( this );
    RecvThread->Resume();

    XmitThread = new PacketStreamXmitThread( this );
    XmitThread->Resume();

    StopperThread = new PacketStreamStoppageThread( this );
    StopperThread->Resume();

}
#ifdef _MSC_VER
    #pragma warning(pop)
#endif

//---------------------------------------------------------------------------
//  destructor for class PacketStreamBase
//---------------------------------------------------------------------------

PacketStreamBase::~PacketStreamBase()
{
    if (Started)
        Stop();

    RecvThread->Shutdown();
    delete RecvThread;
    RecvThread = 0;

    XmitThread->Shutdown();
    delete XmitThread;
    XmitThread = 0;

    StopperThread->Shutdown();
    delete StopperThread;
    StopperThread = 0;

    delete InternalOps;
    InternalOps = 0;

    delete LocalCreditor;
    LocalCreditor = 0;
}

//---------------------------------------------------------------------------
//  PacketStreamBase::ManagerInternals() -- Internal Interface
//---------------------------------------------------------------------------

IPacketStreamManagerInternals *  PacketStreamBase::ManagerInternals()
{
    return InternalOps;
}

//---------------------------------------------------------------------------
//  PacketStreamBase::Preconfigure() -- Non-repeated 'Manual' Configuration
//---------------------------------------------------------------------------

void  PacketStreamBase::Preconfigure()
{
    DeviceInternals()->PreconfigBusmastering();
}

//---------------------------------------------------------------------------
//  PacketStreamBase::Start() -- Start busmastering
//---------------------------------------------------------------------------

void  PacketStreamBase::Start()
{
    if (Started)
        return;

    //  Call "OnBeforeStart" event
    OpenWire::NotifyEvent e;
    OnBeforeStart.Execute(e);

    RecvQ.QueueCapacity(FRecvPacketCache);

    XmitQ.Clear();
    RecvQ.Clear();
    RecvThread->Clear();
    //
    //  Start Receiver Thread
    RecvThread->Start();
    XmitThread->Start();
    StopperThread->Start();
    //
    //  Begin Busmastering Sequence
    DeviceInternals()->BusmasterReset();
    //
    //  Initial Device Configuration
    DeviceInternals()->ConfigureBusmastering();

    //  Prefill needs to be done differently in direct data mode
    if (FStreamDataMode == ppmNormal)
        {
        //
        //  Transmit Prefill
        Prefill();
        DeviceInternals()->TraceLog(0xC0, 0);
        //
        //  Load Xmit Busmaster Region
        DeviceInternals()->PreloadXmitBusmasterRegion();
        }
    else
        {
        //  We only need to do this here, it will notify on its own
        DeviceInternals()->PreloadXmitBusmasterRegion();
        }

    // This flag must be set before starting the busmaster.
    Started = true;

    //
    //  Start busmastering
    DeviceInternals()->StartBusmastering();

    //  Call "OnAfterStart" event
    OpenWire::NotifyEvent e2;
    OnAfterStart.Execute(e2);
}

//---------------------------------------------------------------------------
//  PacketStreamBase::Stop() -- Stop busmastering
//---------------------------------------------------------------------------

void  PacketStreamBase::Stop()
{
    if (!Started)
        return;

    //  Wakes up a stopper thread that will post OnAfterStop when safe
    DeviceInternals()->TraceLog(0xe4, 1);
    StopperThread->StartMonitoring();

    DeviceInternals()->TraceLog(0xe4, 10);
    Started = false;
    DeviceInternals()->TraceLog(0xe4, 9);
    DeviceInternals()->StopBusmastering();
    DeviceInternals()->TraceLog(0xe4, 8);
    RecvThread->Stop();
    DeviceInternals()->TraceLog(0xe4, 7);
    XmitThread->Stop();

//
//    OpenWire::NotifyEvent e;
//    OnAfterStop.Execute(e);
    DeviceInternals()->TraceLog(0xe4, 0);
}

//---------------------------------------------------------------------------
//  PacketStreamBase::SuspendOutput() -- Suspend output request for data
//---------------------------------------------------------------------------

void  PacketStreamBase::SuspendOutput(bool state)
{
    XmitQ.Suspend(state);
}

//---------------------------------------------------------------------------
//  PacketStreamBase::BasicSend() -- Post packet for transmission
//---------------------------------------------------------------------------

void  PacketStreamBase::BasicSend(Buffer & packet)
{
    //  ..and Send it out
    XmitThread->Send(packet);
    DeviceInternals()->TransmitDataAvailable();
}

//---------------------------------------------------------------------------
//  PacketStreamBase::BasicRecv() -- Return packet received from host
//---------------------------------------------------------------------------

void  PacketStreamBase::BasicRecv(Buffer & packet)
{
    RecvThread->Recv(packet);
}

//---------------------------------------------------------------------------
//  PacketStreamBase::ForwardDirectDataPacket() -- forward direct data packet to app
//---------------------------------------------------------------------------

void  PacketStreamBase::ForwardDirectDataPacket(BmPacket & pkt)
{
    // give credit (... when its due)
    LocalCreditor->CreditStream(pkt.SizeInInts());

    // trigger events to notify app
    PacketStreamDirectPacketDataEvent e(this, pkt);

    //  Fire signal to process data
    OnDirectPacketDataAvailable.Execute(e);
}

//---------------------------------------------------------------------------
//  PacketStreamBase::ForwardDirectAlertPacket() -- forward direct alert packet to app
//---------------------------------------------------------------------------

void  PacketStreamBase::ForwardDirectAlertPacket(BmPacket & pkt)
{
    // give credit (... when its due)
    LocalCreditor->CreditStream(pkt.SizeInInts());

    // trigger events to notify app
    PacketStreamDirectPacketAlertEvent e(this, pkt);

    //  Fire signal to process alert
    OnDirectPacketAlertAvailable.Execute(e);

}

//---------------------------------------------------------------------------
//  PacketStreamBase::EnqueueRecvPacket() -- Add Recv Packet to queue
//---------------------------------------------------------------------------

void  PacketStreamBase::EnqueueRecvPacket(Buffer & buffer)
{
    //  Alert check generates alerts on a packet - returns true
    //    if packet should be passed thru to reciever
    if (DeviceInternals()->AlertCheck(buffer))
        {
        RecvQ.Push(buffer);
        }
    else
        {
        //  If we don't enqueue, we MUST grant credit for this packet before
        //    discarding it.
        LocalCreditor->CreditStream(buffer.FullSizeInInts());
        }
}
//------------------------------------------------------------------------------
//  PacketStreamBase::ProcessDataSlab() -- Process Incoming Data Slab (empty it)
//------------------------------------------------------------------------------

void  PacketStreamBase::ProcessDataSlab(UnformattedDataImage & slab)
{
    PacketStreamDirectDataEvent e(this, slab);
    //
    //  Fire signal to process slab
    OnDirectDataAvailable.Execute(e);
}

//---------------------------------------------------------------------------
//  PacketStreamBase::FillDataSlab() -- Fill Outgoing Data Slab
//---------------------------------------------------------------------------

void  PacketStreamBase::FillDataSlab(UnformattedDataImage & slab)
{
    //
    //  Fire signal to process slab
    PacketStreamDirectDataEvent e(this, slab);
    OnDirectDataRequired.Execute(e);
}

//------------------------------------------------------------------------------
//  PacketStreamBase::DirectDataTransmitAck() -- Ack Data Transfer in DD mode
//------------------------------------------------------------------------------

void  PacketStreamBase::DirectDataTransmitAck(size_t data_size)
{
    if (FStreamDataMode == ppmDirectOutput)
        {
        PacketStreamIPSQCreditor  Creditor(this, PciOutData);
        Creditor.CreditStream(data_size);
        }
}

//------------------------------------------------------------------------------
//  PacketStreamBase::DirectDataReceiveAck() -- Ack Data Transfer in DD mode
//------------------------------------------------------------------------------

void  PacketStreamBase::DirectDataReceiveAck(size_t data_size)
{
    if (FStreamDataMode == ppmDirectOutput)
        {
        PacketStreamIPSQCreditor  Creditor(this, PciInData);
        Creditor.CreditStream(data_size);
        }
}

//---------------------------------------------------------------------------
//  PacketStreamBase::DoDataMode() -- Get Streaming Mode Flag
//---------------------------------------------------------------------------
//
//PacketStreamDataModes  PacketStreamBase::DoDataMode()
//{
//    return FStreamDataMode;
//}
//
//------------------------------------------------------------------------------
//  PacketStreamBase::DispatchCustomInterrupt() -- Dispatch User Custom Interrupt
//------------------------------------------------------------------------------

void  PacketStreamBase::DispatchCustomInterrupt(int custom_flags)
{
    PacketStreamCustomInterruptEvent e(this, custom_flags);
    //
    //  Fire signal to process custom interrupt
    OnCustomInterrupt.Execute(e);
}

//---------------------------------------------------------------------------
//  PacketStreamBase::NextXmitPacket() -- Return a reference to next buffer
//---------------------------------------------------------------------------

Buffer  PacketStreamBase::NextXmitPacket()
{
    size_t test_depth = XmitQ.Depth();
    //DeviceInternals()->TraceLog(0x60, test_depth);
    if (test_depth==0)
        {
        DeviceInternals()->TraceLog(0xF0, 0);
        while (1==1)
            Sleep(100);
        }

    //test_depth = XmitQ.Depth();
    //DeviceInternals()->TraceLog(0x61, test_depth);
    //Buffer xxx = XmitQ.Front();
    //DeviceInternals()->TraceLog(0x62, 0);
    //return xxx;
    return XmitQ.Front();
}

//---------------------------------------------------------------------------
//  PacketStreamBase::DequeueXmitPacket() -- Remove Xmit Packet from queue
//---------------------------------------------------------------------------

void  PacketStreamBase::DequeueXmitPacket()
{
    XmitQ.Pop();
}

//---------------------------------------------------------------------------
//  PacketStreamBase::XmitQDepth() -- Depth of Xmit Q
//---------------------------------------------------------------------------

size_t  PacketStreamBase::XmitQDepth()
{
    return XmitQ.Depth();
}

//---------------------------------------------------------------------------
//  PacketStreamBase::RecvQDepth() -- Depth of Recv Q
//---------------------------------------------------------------------------

size_t  PacketStreamBase::RecvQDepth()
{
    return RecvQ.Depth();
}

//---------------------------------------------------------------------------
//  PacketStreamBase::XmitQSignal() -- Bump Xmit Q signal
//---------------------------------------------------------------------------

void  PacketStreamBase::XmitQSignal(size_t bump_count)
{
    //  If in "Async" mode, skip the signalling to avoid useless work.
    if (XmitQ.QueueCapacity()!=0)
        XmitQ.BumpSignal(static_cast<unsigned int>(bump_count));
}

//---------------------------------------------------------------------------
//  PacketStreamBase::Prefill() -- Activate prefill on transmit Q
//---------------------------------------------------------------------------

void  PacketStreamBase::Prefill()
{
    XmitQ.ClearSem();
    XmitThread->Prefill(FPrefillPacketCount);
    XmitQ.QueueCapacity(FPrefillPacketCount);
    DeviceInternals()->TraceLog(0x0D, 0);
}

//---------------------------------------------------------------------------
//  PacketStreamBase::DoStreamStarted() -- returns true if the stream is started
//---------------------------------------------------------------------------

bool PacketStreamBase::DoStreamStarted()
{
    return Started;
}

//===========================================================================
//  CLASS PacketStreamQueue  -- Queue to manage data from VISR to Receiver
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class PacketStreamQueue
//---------------------------------------------------------------------------

PacketStreamQueue::PacketStreamQueue(IPSQueueCreditor * creditor)
    : CapacityInBlks(10), Creditor(creditor)
{
}

//---------------------------------------------------------------------------
//  destructor for class PacketStreamQueue
//---------------------------------------------------------------------------

PacketStreamQueue::~PacketStreamQueue()
{
    delete Creditor;
}

//---------------------------------------------------------------------------
//  PacketStreamQueue::Clear() -- clear queue
//---------------------------------------------------------------------------

void  PacketStreamQueue::Clear()
{
    //  reset state variables
    //Level = 0;
    //  clear buffer queue
    BufferQ.clear();
    CreditQ.clear();
    //  reduce signal to 0
    Signal.Clear();
}

//---------------------------------------------------------------------------
//  PacketStreamQueue::Front() -- Front element in queue
//---------------------------------------------------------------------------

Buffer  PacketStreamQueue::Front()
{
    return BufferQ.front();
}

//---------------------------------------------------------------------------
//  PacketStreamQueue::Depth() -- depth of queue
//---------------------------------------------------------------------------

size_t PacketStreamQueue::Depth()
{
    return BufferQ.depth();
}

//---------------------------------------------------------------------------
//  PacketStreamQueue::QueueCapacity() -- Set queue capacity
//---------------------------------------------------------------------------

void  PacketStreamQueue::QueueCapacity(size_t cap)
{
    CriticalSectionArea guard(lock);
    CapacityInBlks = cap;
}

//---------------------------------------------------------------------------
//  PacketStreamQueue::QueueCapacity() -- Retreive capacity
//---------------------------------------------------------------------------

size_t PacketStreamQueue::QueueCapacity() const
{
    return CapacityInBlks;
}

//---------------------------------------------------------------------------
//  PacketStreamQueue::SignalSem() -- Return reference to semaphore
//---------------------------------------------------------------------------

Semaphore & PacketStreamQueue::SignalSem()
{
    return Signal;
}

//---------------------------------------------------------------------------
//  PacketStreamQueue::BumpSignal() -- Release Signal Sem
//---------------------------------------------------------------------------

void PacketStreamQueue::BumpSignal(unsigned int count)
{
    size_t total_blocks = Signal.Count() + Depth();
    if (total_blocks<CapacityInBlks)
        for (unsigned int i = 0; i < count; ++i)
            Signal.Release();
}

//===========================================================================
//  CLASS XmitPSQueue  -- Queue to manage data from Receiver to VISR
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class XmitPSQueue
//---------------------------------------------------------------------------

XmitPSQueue::XmitPSQueue(IPSQueueCreditor * creditor)
    :  PacketStreamQueue(creditor), SuspendSignal(false)
{
}

//---------------------------------------------------------------------------
//  destructor for class XmitPSQueue
//---------------------------------------------------------------------------

XmitPSQueue::~XmitPSQueue()
{
}

//---------------------------------------------------------------------------
//  XmitPSQueue::Push() -- push buffer onto queue
//---------------------------------------------------------------------------

void  XmitPSQueue::Push( Buffer & buffer )
{
    CriticalSectionArea guard(lock);

    //size_t PacketSize = buffer.FullSizeInInts();
    BufferQ.push(buffer);
    //Level += PacketSize;

    //
    //  If we have uncredited buffers in busmaster memory, credit one
    if (!CreditQ.empty())
        {
        size_t credit_size = CreditQ.front();
        Creditor->CreditStream(credit_size);
        CreditQ.pop();
        }
}

//---------------------------------------------------------------------------
//  XmitPSQueue::Pop() -- remove block from queue
//---------------------------------------------------------------------------

void  XmitPSQueue::Pop()
{
    CriticalSectionArea guard(lock);

    //  Save size of buffer being popped
    size_t PacketSize = BufferQ.front().FullSizeInInts();
    BufferQ.pop();
    //Level -= PacketSize;

    //
    //  Credit granting...credit can either be granted now or postponed
    //    till later. Here we always grant credit. If you postpone, you
    //    need to make it so that you are assured that the last packet
    //    will be sent out in normal circumstances and NOT requiring another
    //    packet to be input to push it out.
    //  Always grant credit for packet
    Creditor->CreditStream(PacketSize);
    //
    //  If there is room in the queue, request more input
    //
    //  Count up present blocks and pending blocks
    size_t total_blocks = Signal.Count() + Depth();
    if (total_blocks<CapacityInBlks && (SuspendSignal==false))
        {
        //  Signal need for additional data from host
        Signal.Release();
        }
}

//===========================================================================
//  CLASS RecvPSQueue  -- Queue to manage data from VISR to Receiver
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class RecvPSQueue
//---------------------------------------------------------------------------

RecvPSQueue::RecvPSQueue(IPSQueueCreditor * creditor)
    :  PacketStreamQueue(creditor)
{
}

//---------------------------------------------------------------------------
//  destructor for class RecvPSQueue
//---------------------------------------------------------------------------

RecvPSQueue::~RecvPSQueue()
{
}

//---------------------------------------------------------------------------
//  RecvPSQueue::Push() -- push buffer onto queue
//---------------------------------------------------------------------------

void  RecvPSQueue::Push( Buffer & buffer )
{
    CriticalSectionArea guard(lock);

    size_t PacketSize = buffer.FullSizeInInts();
    //  Push a buffer onto the queue. If we are above our threshold, stop
    //    granting credit to eventually shut down data flow
    BufferQ.push(buffer);
    //Level += PacketSize;

    if (Depth()<CapacityInBlks)
        {
        //  Grant credit for packet
        Creditor->CreditStream( PacketSize );
        }
    else
        {
        //  Push credit onto credit queue
        CreditQ.push(PacketSize);
        }
    //
    //  For every buffer added, signal Rcvr thread
    Signal.Release();

}

//---------------------------------------------------------------------------
//  RecvPSQueue::Pop() -- 
//---------------------------------------------------------------------------

void  RecvPSQueue::Pop()
{
    CriticalSectionArea guard(lock);

    //size_t PacketSize = BufferQ.front().FullSizeInInts();

    BufferQ.pop();
    //Level -= PacketSize;
    //
    //  If we have uncredited buffers in the queue, credit one
    if (!CreditQ.empty())
        {
        size_t credit_size = CreditQ.front();
        Creditor->CreditStream(credit_size);
        CreditQ.pop();
        }
}

}  // namespace Innovative

