//---------------------------------------------------------------------------
// PacketStreamBase_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2010 Innovative Integration Corporation.
//  All Rights Reserved.

#ifndef PacketStreamBase_MbH
#define PacketStreamBase_MbH

#include <PacketStreamSupport_Mb.h>
#include <ThreadSafeQueue_Mb.h>
#include <Buffer_Mb.h>
#include <PacketImage_Mb.h>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

class PacketStreamBase;

//===========================================================================
// CLASS PacketStreamDirectPacketDataEvent  --  Data Event
//===========================================================================

class PacketStreamDirectPacketDataEvent : public OpenWire::Event
{
public:
    PacketStreamBase *    Sender;
    BmPacket Packet;

public:
    PacketStreamDirectPacketDataEvent( PacketStreamBase * sender, const BmPacket & p )
        :  Sender( sender ), Packet(p)
        {
        }
};

//===========================================================================
// CLASS PacketStreamDirectPacketAlertEvent  --  Data Event
//===========================================================================

class PacketStreamDirectPacketAlertEvent : public OpenWire::Event
{
public:
    PacketStreamBase *    Sender;
    BmPacket Packet;

public:
    PacketStreamDirectPacketAlertEvent( PacketStreamBase * sender, const BmPacket & p )
        :  Sender( sender ), Packet(p)
        {
        }
};

//===========================================================================
// CLASS PacketStreamDirectDataEvent  --  Data Logging I/O Event
//===========================================================================

class PacketStreamDirectDataEvent : public OpenWire::Event
{
public:
    PacketStreamBase *    Sender;
    UnformattedDataImage  Slab;

public:
    PacketStreamDirectDataEvent( PacketStreamBase * sender, const UnformattedDataImage & slab )
        :  Sender( sender ), Slab(slab)
        {
        }
};

//===========================================================================
// CLASS PacketStreamCustomInterruptEvent  --  Custom Interrupt Event
//===========================================================================

class PacketStreamCustomInterruptEvent : public OpenWire::Event
{
public:
    PacketStreamBase *    Sender;
    int                   CustomFlags;

public:
    PacketStreamCustomInterruptEvent( PacketStreamBase * sender, int custom_flags )
        :  Sender( sender ), CustomFlags(custom_flags)
        {
        }
};

//===========================================================================
//  CLASS IPSQueueCreditor  -- Interface for credit management in PSQueue
//===========================================================================

class IPSQueueCreditor
{
public:
    virtual void  CreditStream(size_t cq_token) = 0;
    virtual ~IPSQueueCreditor()
    	{}
};

//===========================================================================
//  CLASS PacketStreamQueue  -- Queue to manage data from VISR to Receiver
//===========================================================================

class PacketStreamQueue
{
public:
    // Ctor
    PacketStreamQueue(IPSQueueCreditor * creditor);
    virtual ~PacketStreamQueue();

    // Methods
    Buffer          Front();
    size_t          Depth();

    void            Clear();
    virtual void    Push(Buffer & buffer) = 0;
    virtual void    Pop() = 0;

    Semaphore &     SignalSem();

    void            ClearSem()
                        {  Signal.Clear();  }
    void            BumpSignal(unsigned int count);

    // Properties
    void         QueueCapacity(size_t cap_in_blocks);
    size_t       QueueCapacity() const;

protected:
    //
    //  Member Data
    size_t      CapacityInBlks;
    Semaphore   Signal;
    mutable CriticalSection  lock;

    ThreadSafeQueue<Buffer>     BufferQ;
    ThreadSafeQueue<size_t>     CreditQ;
    IPSQueueCreditor *          Creditor;
};

//===========================================================================
//  CLASS XmitPSQueue  -- Queue to manage data from Receiver to VISR
//===========================================================================

class XmitPSQueue : public PacketStreamQueue
{
public:
    XmitPSQueue(IPSQueueCreditor * creditor);
    ~XmitPSQueue();

    virtual void  Push(Buffer & buffer);
    virtual void  Pop();
    void          Suspend(bool which)  {  SuspendSignal=which;  }

private:
    //
    //  Member Data
    bool  SuspendSignal;

};

//===========================================================================
//  CLASS RecvPSQueue  -- Queue to manage data from VISR to Receiver
//===========================================================================

class RecvPSQueue : public PacketStreamQueue
{
public:
    RecvPSQueue(IPSQueueCreditor * creditor);
    ~RecvPSQueue();

    virtual void  Push(Buffer & buffer);
    virtual void  Pop();

};
//===========================================================================
//  CLASS PacketStreamBase  -- Packet Based Stream I/O Base Class
//===========================================================================
class PacketStreamRcvrThread;
class PacketStreamXmitThread;
class PacketStreamStoppageThread;

class PacketStreamBase
{
    friend class PacketStreamIPSQCreditor;
    friend class PacketStreamRcvrThread;
    friend class PacketStreamXmitThread;
    friend class PacketStreamStoppageThread;
    friend class TIPacketInternalsImpl;

public:
    enum  PciDirectionCodes { PciOutData=0,  PciInData=1 };
    
    PacketStreamBase();
    virtual ~PacketStreamBase();

    virtual void  Preconfigure();
    virtual void  Start();
    virtual void  Stop();

    void  SuspendOutput(bool state);

    void  DirectDataTransmitAck(size_t data_size);
    void  DirectDataReceiveAck(size_t data_size);

    void  PrefillPacketCount(int packets)
        {  FPrefillPacketCount=packets;  }
    int   PrefillPacketCount() const
        {  return FPrefillPacketCount;  }

    void  RecvPacketCache(int packets)
        {  FRecvPacketCache=packets;  }
    int   RecvPacketCache() const
        {  return FRecvPacketCache;  }

    void    RxLoadBalancing(bool state)
        {  FRxLoadBalancing=state;  }
    bool    RxLoadBalancing() const
        {  return FRxLoadBalancing;  }

    void    TxLoadBalancing(bool state)
        {  FTxLoadBalancing=state;  }
    bool    TxLoadBalancing() const
        {  return FTxLoadBalancing;  }

    void    DirectDataMode(bool state)
        {  FStreamDataMode = (state) ? ppmDirectOutput : ppmNormal; }
    bool    DirectDataMode() const
        {  return FStreamDataMode==ppmDirectOutput;  }

    void    NormalDataMode(bool /*state*/)
        {  FStreamDataMode = ppmNormal;  }
    bool    NormalDataMode() const
        {  return FStreamDataMode==ppmNormal;  }

    void   DirectPacketMode()
        {  FStreamDataMode = ppmDirectPacket;  }
    bool    DirectPacketMode() const
        {  return FStreamDataMode==ppmDirectPacket;  }

    PacketStreamRcvrThread *  RThread()  { return RecvThread;  }
    PacketStreamXmitThread *  XThread()  { return XmitThread;  }

    virtual IPacketStreamManagerInternals *  ManagerInternals();
    virtual IPacketStreamDeviceInternals *   DeviceInternals() = 0;

    OpenWire::EventHandler<OpenWire::NotifyEvent>  OnBeforeStart;
    OpenWire::EventHandler<OpenWire::NotifyEvent>  OnAfterStart;
    OpenWire::EventHandler<OpenWire::NotifyEvent>  OnAfterStop;

    OpenWire::EventHandler<PacketStreamDirectDataEvent>  OnDirectDataAvailable;
    OpenWire::EventHandler<PacketStreamDirectDataEvent>  OnDirectDataRequired;

    //DPM
    OpenWire::EventHandler<PacketStreamDirectPacketDataEvent>  OnDirectPacketDataAvailable;
    OpenWire::EventHandler<PacketStreamDirectPacketAlertEvent>  OnDirectPacketAlertAvailable;

    OpenWire::EventHandler<PacketStreamCustomInterruptEvent>  OnCustomInterrupt;

protected:
    //
    //  Member Functions
    //
    //  Velocia-style packet "Interface" between base and derived classes
    void  BasicSend( Buffer & packet );
    void  BasicRecv( Buffer & packet );

    virtual void  SignalDataAvailable() = 0;
    virtual void  SignalDataRequired() = 0;

    virtual IVelociaBufferManager *  ActualVelociaManager() = 0;

private:
    //
    //  Member Data
    IPacketStreamManagerInternals *  InternalOps;
    PacketStreamRcvrThread *         RecvThread;
    PacketStreamXmitThread *         XmitThread;
    PacketStreamStoppageThread *     StopperThread;
    XmitPSQueue                      XmitQ;
    RecvPSQueue                      RecvQ;
private:
    bool                             Started;
    int                              FPrefillPacketCount;
    int                              FRecvPacketCache;
    PacketStreamDataModes            FStreamDataMode;
private:
//    // Load-balancing data
    bool                             FRxLoadBalancing;
    bool                             FTxLoadBalancing;
    IPSQueueCreditor *               LocalCreditor;
    //
    //  Member Functions
    void                    Prefill();

protected:
    //
    //  Implementation of IxxxStreamManager interface
    bool                   DoStreamStarted();

    //
    //  Implementation of IPacketStreamManagerInternals interface
    void                      ForwardDirectDataPacket(BmPacket & pkt);
    void                      ForwardDirectAlertPacket(BmPacket & pkt);

    void                      EnqueueRecvPacket(Buffer & buffer);
    void                      DispatchCustomInterrupt(int custom_flags);
    IVelociaBufferManager *   VelociaBufferManager()
                                {  return ActualVelociaManager();  }

    virtual Buffer          NextXmitPacket();
    virtual void            DequeueXmitPacket();
    virtual size_t          XmitQDepth();
    virtual size_t          RecvQDepth();
    virtual void            XmitQSignal(size_t bump_count);
    virtual void            ProcessDataSlab(UnformattedDataImage & slab);
    virtual void            FillDataSlab(UnformattedDataImage & slab);

};


#ifdef __CLR_VER
#pragma managed(pop)
#endif
}  // namespace

#endif
