//---------------------------------------------------------------------------

#ifndef PacketStreamSupport_MbH
#define PacketStreamSupport_MbH

#include <StreamDeviceEvents_Mb.h>
#include <Buffer_Mb.h>
#include <vector>
#include <SafeLink_Mb.h>
#include <Trace_Mb.h>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

class IPacketStreamDevice;
class IVitaPacketStreamDevice;
class UnformattedDataImage;

enum PacketStreamDataModes { ppmNormal, ppmDirectOutput, ppmDirectPacket };

//==============================================================================
//  CLASS IPacketStreamManagerInternals  -- Packet Streaming Internal Interface
//==============================================================================

class IPacketStreamManagerInternals
{
public:
    virtual void ForwardDirectDataPacket(BmPacket & pkt) = 0;
    virtual void ForwardDirectAlertPacket(BmPacket & pkt) = 0;
    virtual void EnqueueRecvPacket(Buffer & buffer) = 0;
    virtual void ProcessDataSlab(UnformattedDataImage & slab) = 0;
    virtual void FillDataSlab(UnformattedDataImage & slab) = 0;
    virtual Buffer NextXmitPacket() = 0;
    virtual void DequeueXmitPacket() = 0;
    virtual size_t XmitQDepth() = 0;
    virtual size_t RecvQDepth() = 0;
    virtual PacketStreamDataModes  PacketProcessingMode() = 0;  // alias of DataMode()
    virtual bool HasStreamStarted() = 0;    // alias of StreamStarted()
    virtual void   XmitQSignal(size_t count) = 0;
    virtual void  DispatchCustomInterrupt(int CustomFlags) = 0;
    virtual IVelociaBufferManager *  VelociaManager() = 0;
    virtual ~IPacketStreamManagerInternals()
    	{}

};

//==============================================================================
//  CLASS IPacketStreamManager  -- Packet Streaming Public Interface
//==============================================================================

class IPacketStreamManager
{
public:
    virtual void  Send( short PeriphId, Buffer & packet ) = 0;
    virtual void  Send( Buffer & packet ) = 0;
    virtual void  Recv( Buffer & packet ) = 0;

    virtual bool ConnectTo( IPacketStreamDevice *Device ) = 0;
    virtual void Disconnect() = 0;
    virtual bool StreamStarted() = 0;

    virtual IPacketStreamManagerInternals *  Internals() = 0;
    virtual ~IPacketStreamManager()
    	{}

};

//==============================================================================
//  CLASS IVitaPacketStreamManager  -- VITA Packet Streaming Public Interface
//==============================================================================

class IVitaPacketStreamManager
{
public:
    virtual void  Send( short PeriphId, VeloBuffer & packet ) = 0;
    virtual void  Send( VeloBuffer & packet ) = 0;
    virtual void  Recv( VeloBuffer & packet ) = 0;

    virtual bool ConnectTo( IVitaPacketStreamDevice *Device ) = 0;
    virtual void Disconnect() = 0;
    virtual bool StreamStarted() = 0;

    virtual IPacketStreamManagerInternals *  Internals() = 0;
    virtual ~IVitaPacketStreamManager()
    	{}

};

//==============================================================================
//  CLASS IPacketStreamDeviceInternals  -- Packet Device Internal Interface
//==============================================================================

class IPacketStreamDeviceInternals
{
public:
	virtual void  ConfigureBusmastering() = 0;
    virtual void  PreloadXmitBusmasterRegion() = 0;
    virtual void  PreconfigBusmastering() = 0;
    virtual void  StartBusmastering() = 0;
    virtual void  StopBusmastering() = 0;
    virtual void  GrantCredit(int channel, size_t size_wds) = 0;
    virtual void  TransmitDataAvailable() = 0;
    virtual bool  AlertTest(unsigned int pid) = 0;
    virtual bool  AlertCheck(const Buffer & buffer) = 0;
    //  Alias of PSDevice functions
    virtual void  BusmasterReset() = 0;               // Same as PSD->Reset();
    virtual void  TraceLog(int type, int data) = 0;   // Same as PSD->Log();
    virtual unsigned int  VisrIsrQDepth() = 0;

	// Specify driver-allocated BM region, prior to open
	virtual void            BusMasterSizeWords(unsigned int value) = 0;
	virtual unsigned int    IncomingBusMasterSizeWords() const = 0;
	virtual unsigned int    OutgoingBusMasterSizeWords() const = 0;
    virtual ~IPacketStreamDeviceInternals()
    	{}

};

//===========================================================================
//  CLASS IPacketStreamDevice  -- Packet Based Streaming device
//===========================================================================

class IPacketStreamDevice
{
public: // Pseudo-protected
    virtual void  RegisterStreamPlugIn( IPacketStreamManager *plug_in ) = 0;
    virtual void  Reset() = 0;

	virtual IPacketStreamDeviceInternals *  Internals() = 0;
    virtual void  Log(int type, int data) = 0;
    virtual ~IPacketStreamDevice()
    	{}

};

//===========================================================================
//  CLASS IVitaPacketStreamDevice  -- VITA Packet Streaming device
//===========================================================================

class IVitaPacketStreamDevice
{
public: // Pseudo-protected
    virtual void  RegisterStreamPlugIn( IVitaPacketStreamManager *plug_in ) = 0;
    virtual void  Reset() = 0;

	virtual IPacketStreamDeviceInternals *  Internals() = 0;
    virtual void  Log(int type, int data) = 0;
    virtual ~IVitaPacketStreamDevice()
    	{}

};

//==========================================================================
//  CLASS SafePacketStreamManagerInternalsImpl  -- "Null" Manager Internals
//==========================================================================

class SafePacketStreamManagerInternalsImpl : public IPacketStreamManagerInternals
{
public:
    virtual void ForwardDirectDataPacket(BmPacket & /*pkt*/) { X(); };
    virtual void ForwardDirectAlertPacket(BmPacket & /*pkt*/) { X(); };
    virtual void EnqueueRecvPacket(Buffer & /*buffer*/) { X(); };
    virtual void ProcessDataSlab(UnformattedDataImage & /*slab*/) { X(); };
    virtual void FillDataSlab(UnformattedDataImage & /*slab*/)  { X(); };
    //virtual PacketStreamDataModes DataMode() {  X(); return ppmNormal;  };
    virtual Buffer NextXmitPacket() {  X(); Buffer X;  return X; };
    virtual void DequeueXmitPacket() { X(); };
    virtual size_t XmitQDepth() { X();  return 0; };
    virtual size_t RecvQDepth() { X();  return 0; };
    virtual PacketStreamDataModes  PacketProcessingMode() {  X(); return ppmNormal; };
    virtual bool  HasStreamStarted() { X();  return false; };
    virtual void   XmitQSignal(size_t /*count*/) { X(); };
    virtual void  DispatchCustomInterrupt(int /*CustomFlags*/) { X(); };
    virtual IVelociaBufferManager *  VelociaManager() { return &Mgr;  }   // always safe to call

private:
    Trace             Debug;
    PmcBufferManager  Mgr;

    void X()
        {  Debug.Log("INVALID CALL to unconnected StreamManagerInternals...: ");  }
};

//==============================================================================
//  CLASS SafePacketStreamDeviceInternalsImpl  -- "Null" Manager
//==============================================================================
//
class SafePacketStreamDeviceInternalsImpl : public IPacketStreamDeviceInternals
{
public:
    virtual void  ConfigureBusmastering() { X(); };
    virtual void  PreloadXmitBusmasterRegion() { X(); };
    virtual void  PreconfigBusmastering() { X(); };
    virtual void  StartBusmastering() { X(); };
    virtual void  StopBusmastering() { X(); };
    virtual void  GrantCredit(int /*channel*/, size_t /*size_wds*/) { X(); };
    virtual void  TransmitDataAvailable() { X(); };
    virtual bool  AlertTest(unsigned int /*pid*/) { X();  return false;  };
    virtual bool  AlertCheck(const Buffer & /*buffer*/) { X();  return false;};
    //  Alias of PSDevice functions
    virtual void  BusmasterReset() { X(); };               // Same as PSD->Reset();
    virtual void  TraceLog(int /*type*/, int /*data*/) { X(); };   // Same as PSD->Log();
    virtual unsigned int  VisrIsrQDepth() { X();  return 0; };   

    // Specify driver-allocated BM region, prior to open
    virtual void            BusMasterSizeWords(unsigned int /*value*/) { X(); };
    virtual unsigned int    IncomingBusMasterSizeWords() const { X();  return 0; };
    virtual unsigned int    OutgoingBusMasterSizeWords() const { X();  return 0; };

private:
    Trace                    Debug;

    void X() const
        {  Debug.Log("INVALID CALL to unconnected StreamDeviceInternals...: ");  }

};

//==============================================================================
//  CLASS SafePacketStreamManagerImpl  -- "Null" Manager
//==============================================================================

class SafePacketStreamManagerImpl : public IPacketStreamManager
{
public:

    virtual void  Send( short /*PeriphId*/, Buffer & /*packet*/ ) { X(); };
    virtual void  Send( Buffer & /*packet*/ ) { X(); };
    virtual void  Recv( Buffer & /*packet*/ ) { X(); };

    virtual bool ConnectTo( IPacketStreamDevice * /*Device*/ ) { X();  return false; };
    virtual void Disconnect() { X(); };
    virtual bool StreamStarted() { X();  return false;  };

    virtual IPacketStreamManagerInternals *  Internals() { X();  return &internals;  };

private:
    Trace                                 Debug;
    SafePacketStreamManagerInternalsImpl  internals;

    void X()
        {  Debug.Log("INVALID CALL to unconnected StreamManager...: ");  }

};

//==============================================================================
//  CLASS SafeVitaPacketStreamManagerImpl  -- "Null" Manager
//==============================================================================

class SafeVitaPacketStreamManagerImpl : public IVitaPacketStreamManager
{
public:

    virtual void  Send( short /*PeriphId*/,  VeloBuffer & /*packet*/ ) { X(); };
    virtual void  Send( VeloBuffer & /*packet*/ ) { X(); };
    virtual void  Recv( VeloBuffer & /*packet*/ ) { X(); };

    virtual bool ConnectTo( IVitaPacketStreamDevice * /*Device*/ ) { X();  return false; };
    virtual void Disconnect() { X(); };
    virtual bool StreamStarted() { X();  return false;  };

    virtual IPacketStreamManagerInternals *  Internals() { X();  return &internals;  };

private:
    Trace                                 Debug;
    SafePacketStreamManagerInternalsImpl  internals;

    void X()
        {  Debug.Log("INVALID CALL to unconnected VITA StreamManager...: ");  }

};

//===========================================================================
//  CLASS SafePacketStreamDeviceImpl  -- "Null" Device
//===========================================================================

class SafePacketStreamDeviceImpl : public IPacketStreamDevice
{
public: // Pseudo-protected
    virtual void  RegisterStreamPlugIn( IPacketStreamManager * /*plug_in*/ ) { X(); };
    virtual void  Reset() { X(); };

	virtual IPacketStreamDeviceInternals *  Internals() { X(); return &internals; };
    virtual void  Log(int /*type*/, int /*data*/) { X(); };

private:
    Trace                               Debug;
    SafePacketStreamDeviceInternalsImpl internals;

    void X()
        {  Debug.Log("INVALID CALL to unconnected StreamDevice...: ");  }

};

//===========================================================================
//  CLASS SafeVitaPacketStreamDeviceImpl  -- "Null" Device
//===========================================================================

class SafeVitaPacketStreamDeviceImpl : public IVitaPacketStreamDevice
{
public: // Pseudo-protected
    virtual void  RegisterStreamPlugIn( IVitaPacketStreamManager * /*plug_in*/ ) { X(); };
    virtual void  Reset() { X(); };

	virtual IPacketStreamDeviceInternals *  Internals() { X(); return &internals; };
    virtual void  Log(int /*type*/, int /*data*/) { X(); };

private:
    Trace                               Debug;
    SafePacketStreamDeviceInternalsImpl internals;

    void X()
        {  Debug.Log("INVALID CALL to unconnected VITA Stream Devicve...: ");  }

};

class IIsAPacketStreamBoard;

//===========================================================================
// CLASS PacketStreamManagerSafeLink  --  Safe Linkage to a device
//===========================================================================

class PacketStreamManagerSafeLink : public SafeLink<IPacketStreamManager, SafePacketStreamManagerImpl>
{
public:
    PacketStreamManagerSafeLink(IIsAPacketStreamBoard * owner)
        :  Owner(owner)
        {}

private:
    IIsAPacketStreamBoard *  Owner;

    virtual void  OnAfterLink();        // overrides
    virtual void  OnBeforeUnlink();
};

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif
//===========================================================================
//  CLASS IIsAPacketStreamBoard  -- PacketStream linkage interface class
//===========================================================================
class PacketDeviceMap;

class IIsAPacketStreamBoard : public IPacketStreamDevice
{
    friend class PacketStreamManagerSafeLink;
public:
    IIsAPacketStreamBoard(PacketDeviceMap & dev_map)
        :  IAP_PDevMap(dev_map), StreamManager(this)
        {};

public:  // pseudo-protected
    virtual void  RegisterStreamPlugIn( IPacketStreamManager *plug_in );

protected:
    PacketDeviceMap &            IAP_PDevMap;
    PacketStreamManagerSafeLink  StreamManager;
};

#ifdef _MSC_VER
	#pragma warning(pop)
#endif

class IIsAVitaPacketStreamBoard;

//===========================================================================
// CLASS VitaPacketStreamManagerSafeLink  --  Safe Linkage to a device
//===========================================================================

class VitaPacketStreamManagerSafeLink : public SafeLink<IVitaPacketStreamManager, SafeVitaPacketStreamManagerImpl>
{
public:
    VitaPacketStreamManagerSafeLink(IIsAVitaPacketStreamBoard * owner)
        :  Owner(owner)
        {}

private:
    IIsAVitaPacketStreamBoard *  Owner;

    virtual void  OnAfterLink();        // overrides
    virtual void  OnBeforeUnlink();
};

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif
//===========================================================================
//  CLASS IIsAVitaPacketStreamBoard  -- VitaPacketStream linkage interface class
//===========================================================================
class PacketDeviceMap;

class IIsAVitaPacketStreamBoard : public IVitaPacketStreamDevice
{
    friend class VitaPacketStreamManagerSafeLink;
public:
    IIsAVitaPacketStreamBoard(PacketDeviceMap & dev_map)
        :  IAVP_PDevMap(dev_map), StreamManager(this)
        {};

public:  // pseudo-protected
    virtual void  RegisterStreamPlugIn( IVitaPacketStreamManager *plug_in );

protected:
    PacketDeviceMap &                IAVP_PDevMap;
public:  // pseudo-protected
    VitaPacketStreamManagerSafeLink  StreamManager;
};
#ifdef _MSC_VER
	#pragma warning(pop)
#endif

#ifdef __CLR_VER
#pragma managed(pop)
#endif
}  // end namespace Innovative

#endif
