//---------------------------------------------------------------------------
// PmcIdrom_Mb.cpp  --  Omnibus Module Idrom Access Object
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//


#include <PmcIdrom_Mb.h>
#include <cctype>
#include <string>
#include <SystemSupport_Mb.h>

namespace Innovative
{
//=============================================================================
//  CLASS PmcIdrom  --  Omnibus Module Idrom Access Object
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class PmcIdrom
//-----------------------------------------------------------------------------

const int EEPromDeviceID = 10;

//-----------------------------------------------------------------------------
//  constructor for class PmcIdrom
//-----------------------------------------------------------------------------

PmcIdrom::PmcIdrom(I2CDeviceIntf & intf, int module_area_size, bool endian)
    :  I2CDevice(intf, I2CDevice::DevAddress, EEPromDeviceID, endian),
       ReserveSize(0), FLocationSet(false)
{
    SetAreaSizes(module_area_size);
    Clear();
}

//-----------------------------------------------------------------------------
//  destructor for class PmcIdrom
//-----------------------------------------------------------------------------

PmcIdrom::~PmcIdrom()
{
}

//-----------------------------------------------------------------------------
//  PmcIdrom::Clear() -- Blank out in-memory image
//-----------------------------------------------------------------------------

void  PmcIdrom:: Clear()
{
    for (int i=0; i<IdromPacketSize; i++)
        Packet[i] = 0;
}

//-----------------------------------------------------------------------------
//  PmcIdrom::SetAreaSizes() -- Partition work areas in ROM
//-----------------------------------------------------------------------------

void  PmcIdrom:: SetAreaSizes(int module_area_size)
{
    // Set sizes of regions on the rom...
    //  ...Reserve is set externally
    HeaderSize = 17;
    TrailerSize = 1;
    ModuleAreaSize = module_area_size;
    UserAreaSize = IdromPacketSize - ModuleAreaSize - HeaderSize - 1;       
}

//-----------------------------------------------------------------------------
//  PmcIdrom::WorkAreaBase() -- return base of work area in ROM
//-----------------------------------------------------------------------------

int *  PmcIdrom::WorkAreaBase(WorkArea area)
{
    switch (area)
        {
        case  waModule:
            return &Packet[ReserveSize+HeaderSize];
        case  waUser:
            return &Packet[ReserveSize+HeaderSize+ModuleAreaSize];
        default:
        case  waRaw:
            return &Packet[0];
        }
        
}

//-----------------------------------------------------------------------------
//  PmcIdrom::CWorkAreaBase() const -- return base of work area (const)
//-----------------------------------------------------------------------------

const int *  PmcIdrom::CWorkAreaBase(WorkArea area) const
{
    switch (area)
        {
        case  waModule:
            return &Packet[ReserveSize+HeaderSize];
        case  waUser:
            return &Packet[ReserveSize+HeaderSize+ModuleAreaSize];
        default:
        case  waRaw:
            return &Packet[0];
        }
        
}

//-----------------------------------------------------------------------------
//  PmcIdrom::CalculateChecksum() -- Calculate checksum of packet
//-----------------------------------------------------------------------------

void  PmcIdrom::CalculateChecksum()
{  
    unsigned int checksum(0);

    for ( int i = 0; i < IdromPacketSize-TrailerSize; i++)
        checksum += Packet[i];

    Packet[IdromPacketSize-TrailerSize] = checksum;
}
    
//---------------------------------------------------------------------------
//  PmcIdrom::SizeInChars() -- size of data in Chars
//---------------------------------------------------------------------------

unsigned int  PmcIdrom::SizeInChars() const
{
    return IdromPacketSize * (sizeof(int)/sizeof(char));
}

//------------------------------------------------------------------------
//   PmcIdrom::Name()  --  Read ROM name code from IDROM
//------------------------------------------------------------------------

std::string  PmcIdrom::Name() const
{
    std::string name;
    for (int i=0; i<13; i++)
        {
        char c = (char)Packet[ReserveSize+i];
        if (c=='\0')
            break;
        name += c;
        }
    return name;
}

//------------------------------------------------------------------------
//   PmcIdrom::Name()  --  Write ROM name code to Image
//------------------------------------------------------------------------

void  PmcIdrom::Name(const std::string & name)
{
    for (int i=0; i<13; i++)
        {
        Packet[ReserveSize+i] = static_cast<unsigned int>(name[i]);
        if (name[i]=='\0')
            break;
        }
    CalculateChecksum();
}
//------------------------------------------------------------------------
//   PmcIdrom::Revision()  --  Read ROM rev code from IDROM
//------------------------------------------------------------------------

std::string  PmcIdrom::Revision() const
{
    std::string name;

    name += static_cast<char>(Packet[ReserveSize+16]);
    return name;
}

//------------------------------------------------------------------------
//   PmcIdrom::Revision()  --  Write ROM rev code to Image
//------------------------------------------------------------------------

void  PmcIdrom::Revision(const std::string & revision)
{
    Packet[ReserveSize+16] = revision[0];
    CalculateChecksum();
}

//------------------------------------------------------------------------
//   PmcIdrom::SerialNumber()  --  Read SerialNumber from IDROM
//------------------------------------------------------------------------

unsigned int  PmcIdrom::SerialNumber() const
{
	const int HeaderOfs = 13;

	int MyIdx = ReserveSize + HeaderOfs;

	unsigned int SerialNum;

	SerialNum = WorkAreaAsInt(waRaw, MyIdx);
	return SerialNum;
}

//------------------------------------------------------------------------
//   PmcIdrom::SerialNumber()  --  Write SerialNumber to Image
//------------------------------------------------------------------------

void  PmcIdrom::SerialNumber(unsigned int Num)
{
	const int HeaderOfs = 13;

	int MyIdx = ReserveSize + HeaderOfs;

	WorkAreaAsInt(waRaw, MyIdx, Num);
	CalculateChecksum();
}

//------------------------------------------------------------------------
//   PmcIdrom::PartNumber()  --  Read PartNumber from IDROM
//------------------------------------------------------------------------

unsigned int  PmcIdrom::PartNumber() const
{
    const int HeaderOfs = 14;

    int MyIdx = ReserveSize + HeaderOfs;

    unsigned int part;

    part = WorkAreaAsInt(waRaw, MyIdx);
    return part;
}

//------------------------------------------------------------------------
//   PmcIdrom::PartNumber()  --  Write PartNumber to Image
//------------------------------------------------------------------------

void  PmcIdrom::PartNumber(unsigned int part)
{
    const int HeaderOfs = 14;

    int MyIdx = ReserveSize + HeaderOfs;

    WorkAreaAsInt(waRaw, MyIdx, part);
    CalculateChecksum();
}

//------------------------------------------------------------------------
//   PmcIdrom::DashNumber()  --  Read Board Dash Code from IDROM
//------------------------------------------------------------------------

unsigned int  PmcIdrom::DashNumber() const
{
    const int HeaderOfs = 15;
    const int InWordOffset = 0;

    int MyIdx = ReserveSize + HeaderOfs*4 + InWordOffset;

    unsigned int dash;

    dash = WorkAreaAsChar(waRaw, MyIdx);
    return dash;
}

//------------------------------------------------------------------------
//   PmcIdrom::DashNumber()  --  Write Board Dash Code to Image
//------------------------------------------------------------------------

void  PmcIdrom::DashNumber(unsigned int dash)
{
    const int HeaderOfs = 15;
    const int InWordOffset = 0;

    int MyIdx = ReserveSize + HeaderOfs*4 + InWordOffset;

    WorkAreaAsChar(waRaw, MyIdx, dash);
    CalculateChecksum();
}

//------------------------------------------------------------------------
//   PmcIdrom::DashVariant()  --  Read Board Dash Variant Code from IDROM
//------------------------------------------------------------------------

std::string  PmcIdrom::DashVariant() const
{
    const int HeaderOfs = 15;
    const int InWordOffset = 1;

    int MyIdx = ReserveSize + HeaderOfs*4 + InWordOffset;

    std::string dashV;

    dashV.push_back(WorkAreaAsChar(waRaw, MyIdx));
    return dashV;
}

//------------------------------------------------------------------------
//   PmcIdrom::DashVariant()  --  Write Board Dash Variant to Image
//------------------------------------------------------------------------

void  PmcIdrom::DashVariant(const std::string & dashV)
{
    const int HeaderOfs = 15;
    const int InWordOffset = 1;

    int MyIdx = ReserveSize + HeaderOfs*4 + InWordOffset;

    WorkAreaAsChar(waRaw, MyIdx, dashV[0]);
    CalculateChecksum();
}

//------------------------------------------------------------------------
//   PmcIdrom::RuggedizationNumber()  --  Read Board Dash Code from IDROM
//------------------------------------------------------------------------

unsigned int  PmcIdrom::RuggedizationNumber() const
{
    const int HeaderOfs = 15;
    const int InWordOffset = 2;

    int MyIdx = ReserveSize + HeaderOfs*4 + InWordOffset;

    unsigned int rugged;

    rugged = WorkAreaAsChar(waRaw, MyIdx);
    return rugged;
}

//------------------------------------------------------------------------
//   PmcIdrom::RuggedizationNumber()  --  Write Board Dash Code to Image
//------------------------------------------------------------------------

void  PmcIdrom::RuggedizationNumber(unsigned int rugged)
{
    const int HeaderOfs = 15;
    const int InWordOffset = 2;

    int MyIdx = ReserveSize + HeaderOfs*4 + InWordOffset;

    WorkAreaAsChar(waRaw, MyIdx, rugged);
    CalculateChecksum();
}

//------------------------------------------------------------------------
//   PmcIdrom::EcoSubRevNumber()  --  Read Board ECO Code from IDROM
//------------------------------------------------------------------------

unsigned int  PmcIdrom::EcoSubRevNumber() const
{
    const int HeaderOfs = 15;
    const int InWordOffset = 3;

    int MyIdx = ReserveSize + HeaderOfs*4 + InWordOffset;

    unsigned int eco_subrev;

    eco_subrev = WorkAreaAsChar(waRaw, MyIdx);
    return eco_subrev;
}

//------------------------------------------------------------------------
//   PmcIdrom::EcoSubRevNumber()  --  Write Board Dash Code to Image
//------------------------------------------------------------------------

void  PmcIdrom::EcoSubRevNumber(unsigned int eco_subrev)
{
    const int HeaderOfs = 15;
    const int InWordOffset = 3;

    int MyIdx = ReserveSize + HeaderOfs*4 + InWordOffset;

    WorkAreaAsChar(waRaw, MyIdx, eco_subrev);
    CalculateChecksum();
}

//------------------------------------------------------------------------
//   PmcIdrom::ChipErase()  --  Erase the Rom
//------------------------------------------------------------------------

void PmcIdrom::ChipErase(int pattern)
{
    for(int i=0;i<IdromPacketSize-1;++i)
    {
       WorkAreaAsInt(waRaw, i, pattern);
    }
}

//------------------------------------------------------------------------
//   PmcIdrom::VerifyChipErased()  --  verify the Rom has been erased
//------------------------------------------------------------------------

bool PmcIdrom::VerifyChipErased(int pattern)
{

    bool Verified = true; //will be forced if did not match the pattern
    for(int i=0;i<IdromPacketSize-1;++i)
    {
        if(WorkAreaAsInt(waRaw,i) != pattern)
            Verified = false;
    }

    return Verified;
}

//------------------------------------------------------------------------
//   PmcIdrom::PrintChipData()  --  Print the chip data
//------------------------------------------------------------------------

std::string PmcIdrom::PrintChipData()
{

    stringstream data;
    for(int i=0;i<IdromPacketSize-1;++i)
    {
        data << WorkAreaAsInt(waRaw,i) << " ";
    }

    return data.str();
}

//------------------------------------------------------------------------
//   PmcIdrom::IsEmpty()  --  Check if rom has been programmed
//------------------------------------------------------------------------

bool  PmcIdrom::IsEmpty() const
{
    char value = static_cast<char>(Packet[ReserveSize+16]);
    return std::isalnum(value) == false;
}

//-----------------------------------------------------------------------------
//  PmcIdrom::AsInt() const -- Read Data field as int
//-----------------------------------------------------------------------------

int PmcIdrom::AsInt(int idx) const
{
    return WorkAreaAsInt(waUser, idx); 
}

//-----------------------------------------------------------------------------
//  PmcIdrom::AsInt() -- Write Data field as int
//-----------------------------------------------------------------------------

void PmcIdrom::AsInt(int idx, int data)
{
    return WorkAreaAsInt(waUser, idx, data); 
}

//-----------------------------------------------------------------------------
//  PmcIdrom::AsFloat() const -- Read Data field as float
//-----------------------------------------------------------------------------

float PmcIdrom::AsFloat(int idx) const
{
    return WorkAreaAsFloat(waUser, idx); 
}

//-----------------------------------------------------------------------------
//  PmcIdrom::AsFloat() -- Write Data field as float
//-----------------------------------------------------------------------------

void PmcIdrom::AsFloat(int idx, float data)
{
    return WorkAreaAsFloat(waUser, idx, data); 
}

//-----------------------------------------------------------------------------
//  PmcIdrom::AsDouble() const -- Read Data field as double
//-----------------------------------------------------------------------------

double PmcIdrom::AsDouble(int idx) const
{
    return WorkAreaAsDouble(waUser, idx);
}

//-----------------------------------------------------------------------------
//  PmcIdrom::AsDouble() -- Write Data field as double
//-----------------------------------------------------------------------------

void PmcIdrom::AsDouble(int idx, double data)
{
    WorkAreaAsDouble(waUser, idx, data);
}

//-----------------------------------------------------------------------------
//  PmcIdrom::AsShort() const -- Read Data field as short
//-----------------------------------------------------------------------------

short PmcIdrom::AsShort(int idx) const
{
    return WorkAreaAsShort(waUser, idx); 
}

//-----------------------------------------------------------------------------
//  PmcIdrom::AsShort() -- Write Data field as short
//-----------------------------------------------------------------------------

void PmcIdrom::AsShort(int idx, short data)
{
    return WorkAreaAsShort(waUser, idx, data); 
}

//-----------------------------------------------------------------------------
//  PmcIdrom::AsChar() const -- Read Data field as char
//-----------------------------------------------------------------------------

char PmcIdrom::AsChar(int idx) const
{
    return WorkAreaAsChar(waUser, idx); 
}

//-----------------------------------------------------------------------------
//  PmcIdrom::AsChar() -- Write Data field as char
//-----------------------------------------------------------------------------

void PmcIdrom::AsChar(int idx, char data)
{
    return WorkAreaAsChar(waUser, idx, data); 
}

//-----------------------------------------------------------------------------
//  PmcIdrom::AsString() const -- Read Data field(s) as String of Char
//-----------------------------------------------------------------------------

std::string  PmcIdrom::AsString(unsigned int int_idx, unsigned int ints_to_use) const
{
	std::string result;
	//  Range check
	unsigned int char_idx = int_idx*sizeof(int);
	unsigned int chars_to_use = ints_to_use*sizeof(int);

	if (char_idx>=SizeInChars())
		return result;

	while (chars_to_use)
		{
		char next_char = AsChar(char_idx++);
		if (next_char==0)
			break;
		result.push_back(next_char);

		chars_to_use--;
		}
	return result;
}

//-----------------------------------------------------------------------------
//  PmcIdrom::AsString() -- Write Data field(s) as String of Char
//-----------------------------------------------------------------------------

void  PmcIdrom::AsString(unsigned int int_idx, unsigned int ints_to_use, const std::string & text)
{
	//  Range check
	unsigned int char_idx = int_idx*sizeof(int);
	unsigned int chars_to_use = ints_to_use*sizeof(int);

	if (char_idx>=SizeInChars())
		return;

	size_t text_idx = 0;
	while (chars_to_use)
		{
		char next_char;
		if (text_idx>=text.size())
			next_char = 0;
		else
			next_char = text[text_idx];
		text_idx++;
		AsChar(char_idx++, next_char);

		chars_to_use--;
		}
}

//-----------------------------------------------------------------------------
//  PmcIdrom::WorkAreaAsInt() const -- Read Data field as int
//-----------------------------------------------------------------------------

int PmcIdrom::WorkAreaAsInt(WorkArea area, int idx) const
{
    const int * pkt = CWorkAreaBase(area);
    //  Range check
    if (idx>=IdromPacketSize)
        return 0;
    
    return pkt[idx];
}

//-----------------------------------------------------------------------------
//  PmcIdrom::WorkAreaAsInt() -- Write Data field as int
//-----------------------------------------------------------------------------

void  PmcIdrom::WorkAreaAsInt(WorkArea area, int idx, int data)
{
    int * pkt = WorkAreaBase(area);
    //  Range check
    if (idx>=IdromPacketSize)
        return;
    
    pkt[idx] = data;
    CalculateChecksum();
}

//-----------------------------------------------------------------------------
//  PmcIdrom::WorkAreaAsFloat() const -- Read Data field as float
//-----------------------------------------------------------------------------

float  PmcIdrom::WorkAreaAsFloat(WorkArea area, int idx) const
{
    const float * pkt = reinterpret_cast<const float *>(CWorkAreaBase(area));
    //  Range check
    if (idx>=IdromPacketSize)
        return 0;
    
    return pkt[idx];
}

//-----------------------------------------------------------------------------
//  PmcIdrom::WorkAreaAsFloat() -- Write Data fieldas float
//-----------------------------------------------------------------------------

void  PmcIdrom::WorkAreaAsFloat(WorkArea area, int idx, float data)
{
    float * pkt = reinterpret_cast<float *>(WorkAreaBase(area));
    //  Range check
    if (idx>=IdromPacketSize)
        return;
    
    pkt[idx] = data;
    CalculateChecksum();
}

//-----------------------------------------------------------------------------
//  PmcIdrom::WorkAreaAsDouble() const -- Read Data field as double
//-----------------------------------------------------------------------------

double  PmcIdrom::WorkAreaAsDouble(WorkArea area, int idx) const
{
    const double * pkt = reinterpret_cast<const double *>(CWorkAreaBase(area));
    //  Range check
    if (idx>=IdromPacketSize/2)
        return 0;

    return pkt[idx];
}

//-----------------------------------------------------------------------------
//  PmcIdrom::WorkAreaAsDouble() -- Write Data field as double
//-----------------------------------------------------------------------------

void  PmcIdrom::WorkAreaAsDouble(WorkArea area, int idx, double data)
{
    double * pkt = reinterpret_cast<double *>(WorkAreaBase(area));
    //  Range check
    if (idx>=IdromPacketSize/2)
        return;

    pkt[idx] = data;
    CalculateChecksum();
}

//-----------------------------------------------------------------------------
//  PmcIdrom::WorkAreaAsShort() const -- Read Data field as short
//-----------------------------------------------------------------------------

short  PmcIdrom::WorkAreaAsShort(WorkArea area, int idx) const
{
    const short * pkt = reinterpret_cast<const short *>(CWorkAreaBase(area));
    //  Range check
    if (idx>=IdromPacketSize*2)
        return 0;
    
    return pkt[idx];
}

//-----------------------------------------------------------------------------
//  PmcIdrom::WorkAreaAsShort() -- Write Data field as short
//-----------------------------------------------------------------------------

void  PmcIdrom::WorkAreaAsShort(WorkArea area, int idx, short data)
{
    short * pkt = reinterpret_cast<short *>(WorkAreaBase(area));
    //  Range check
    if (idx>=IdromPacketSize*2)
        return;
    
    pkt[idx] = data;
    CalculateChecksum();
}

//-----------------------------------------------------------------------------
//  PmcIdrom::WorkAreaAsChar() const -- Read Data field as char
//-----------------------------------------------------------------------------

char  PmcIdrom::WorkAreaAsChar(WorkArea area, int idx) const
{
    const char * pkt = reinterpret_cast<const char *>(CWorkAreaBase(area));
    //  Range check
    if (idx>=IdromPacketSize*4)
        return 0;
    
    return pkt[idx];
}

//-----------------------------------------------------------------------------
//  PmcIdrom::WorkAreaAsChar() -- Write Data field as char
//-----------------------------------------------------------------------------

void  PmcIdrom::WorkAreaAsChar(WorkArea area, int idx, char data)
{
    char * pkt = reinterpret_cast<char *>(WorkAreaBase(area));
    //  Range check
    if (idx>=IdromPacketSize*4)
        return;
    
    pkt[idx] = data;
    CalculateChecksum();
}

//-----------------------------------------------------------------------------
//  PmcIdrom::LoadFromRom() -- Load in-memory image from ROM hardware
//-----------------------------------------------------------------------------

void  PmcIdrom::LoadFromRom()
{
    ReadBlock(0, Packet, IdromPacketSize);
}

//-----------------------------------------------------------------------------
//  PmcIdrom::StoreToRom() -- Write in-memory image to ROM hardware
//-----------------------------------------------------------------------------

void  PmcIdrom::StoreToRom()
{
    WriteBlock(0, Packet, IdromPacketSize);
//    WritePage(0, Packet);
}

//=============================================================================
//  CLASS PmcIdromSection  --  Object controlling access to a section of IDROM
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class PmcIdromSection
//-----------------------------------------------------------------------------

PmcIdromSection::PmcIdromSection(PmcIdrom & rom, PmcIdrom::WorkArea wa, int offset, int section_size_words)
    : Rom(rom), WA(wa), Offset(offset), SectionSize(section_size_words)
{
}

//-----------------------------------------------------------------------------
//  Copy constructor for class PmcIdromSection
//-----------------------------------------------------------------------------

PmcIdromSection::PmcIdromSection(const PmcIdromSection & other)
    : Rom(other.Rom), WA(other.WA), Offset(other.Offset),
      SectionSize(other.SectionSize)
{
}

//-----------------------------------------------------------------------------
//  PmcIdromSection assignment operator
//-----------------------------------------------------------------------------

PmcIdromSection & PmcIdromSection::operator =(const PmcIdromSection & other)
{
    Rom = other.Rom;
    WA = other.WA;
    Offset = other.Offset;
    SectionSize = other.SectionSize;

    return *this;
}

//-----------------------------------------------------------------------------
//  PmcIdromSection::Clear() -- Erase the section
//-----------------------------------------------------------------------------

void  PmcIdromSection::Clear()
{
    for (int i=0; i<SectionSize; i++)
        Rom.WorkAreaAsInt(WA, Offset+i, 0);
}

//-----------------------------------------------------------------------------
//  PmcIdromSection::LoadFromRom() -- Read data from ROM
//-----------------------------------------------------------------------------

void  PmcIdromSection::LoadFromRom()
{
    Rom.LoadFromRom();
}

//-----------------------------------------------------------------------------
//  PmcIdromSection::StoreToRom() -- Read data from ROM
//-----------------------------------------------------------------------------

void  PmcIdromSection::StoreToRom()
{
    Rom.StoreToRom();
}

//-----------------------------------------------------------------------------
//  PmcIdromSection::AsInt() -- Access Data as Integer
//-----------------------------------------------------------------------------

int  PmcIdromSection::AsInt(int idx) const
{
    if (idx<SectionSize)
        return Rom.WorkAreaAsInt(WA, Offset+idx);
    else
        return 0;
    
}

//-----------------------------------------------------------------------------
//  PmcIdromSection::AsInt() -- Access Data as Integer
//-----------------------------------------------------------------------------

void  PmcIdromSection::AsInt(int idx, int data)
{
    if (idx<SectionSize)
        Rom.WorkAreaAsInt(WA, Offset+idx, data);
    
}

//-----------------------------------------------------------------------------
//  PmcIdromSection::AsFloat() -- Access Data as Float
//-----------------------------------------------------------------------------

float  PmcIdromSection::AsFloat(int idx) const
{
    if (idx<SectionSize)
        return Rom.WorkAreaAsFloat(WA, Offset+idx);
    else
        return 0;

}

//-----------------------------------------------------------------------------
//  PmcIdromSection::AsFloat() -- Access Data as Float
//-----------------------------------------------------------------------------

void  PmcIdromSection::AsFloat(int idx, float data)
{
    if (idx<SectionSize)
        Rom.WorkAreaAsFloat(WA, Offset+idx, data);

}

//-----------------------------------------------------------------------------
//  PmcIdromSection::AsDouble() -- Access Data as double
//-----------------------------------------------------------------------------

double  PmcIdromSection::AsDouble(int idx) const
{
    int inv_factor = sizeof(double)/sizeof(int);

    if (idx<SectionSize/inv_factor)
        return Rom.WorkAreaAsDouble(WA, Offset/inv_factor+idx);
    else
        return 0;
}

//-----------------------------------------------------------------------------
//  PmcIdromSection::AsDouble() -- Access Data as double
//-----------------------------------------------------------------------------

void  PmcIdromSection::AsDouble(int idx, double data)
{
    int inv_factor = sizeof(double)/sizeof(int);
    
    if (idx<SectionSize/inv_factor)
        Rom.WorkAreaAsDouble(WA, Offset/inv_factor+idx, data);
}

//-----------------------------------------------------------------------------
//  PmcIdromSection::AsShort() -- Access Data as short
//-----------------------------------------------------------------------------

short  PmcIdromSection::AsShort(int idx) const
{
    int factor = sizeof(int)/sizeof(short);

    if (idx<SectionSize*factor)
        return Rom.WorkAreaAsShort(WA, Offset*factor+idx);
    else
        return 0;
    
}

//-----------------------------------------------------------------------------
//  PmcIdromSection::AsShort() -- Access Data as short
//-----------------------------------------------------------------------------

void  PmcIdromSection::AsShort(int idx, short data)
{
    int factor = sizeof(int)/sizeof(short);
    
    if (idx<SectionSize*factor)
        Rom.WorkAreaAsShort(WA, Offset*factor+idx, data);
    
}

//-----------------------------------------------------------------------------
//  PmcIdromSection::AsChar() -- Access Data as char
//-----------------------------------------------------------------------------

char  PmcIdromSection::AsChar(int idx) const
{
    int factor = sizeof(int)/sizeof(char);

    if (idx<SectionSize*factor)
        return Rom.WorkAreaAsChar(WA, Offset*factor+idx);
    else
        return 0;
    
}

//-----------------------------------------------------------------------------
//  PmcIdromSection::AsChar() -- Access Data as char
//-----------------------------------------------------------------------------

void  PmcIdromSection::AsChar(int idx, char data)
{
    int factor = sizeof(int)/sizeof(char);
    
    if (idx<SectionSize*factor)
        Rom.WorkAreaAsChar(WA, Offset*factor+idx, data);
    
}
#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif
//=============================================================================
//  CLASS AdvI2CIdrom  --  Access to a segment of the X6 ROM as an IDROM
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class AdvI2CIdrom
//-----------------------------------------------------------------------------

const int kSinglePageRomSize  =  64;   // IDROM size in words
const int kDualPageRomSize    = 128;

const unsigned int kSPA0_Code = 0x6C;
const unsigned int kSPA1_Code = 0x6E;

//-----------------------------------------------------------------------------
//  constructor for class AdvI2CIdrom
//-----------------------------------------------------------------------------

AdvI2CIdrom::AdvI2CIdrom(I2CDeviceN & rom)
    :  IdromIoDG(this->IView()), ROM(rom), IdromSizeWords(kSinglePageRomSize), DualPageMode(false)
{
    Clear();
}

#ifdef _MSC_VER
	#pragma warning(pop)
#endif
//-----------------------------------------------------------------------------
//  destructor for class AdvI2CIdrom
//-----------------------------------------------------------------------------

AdvI2CIdrom::~AdvI2CIdrom()
{
}

//-----------------------------------------------------------------------------
//  AdvI2CIdrom::DualPageRomMode() -- Switch to Dual Page ROM mode
//-----------------------------------------------------------------------------

void   AdvI2CIdrom::DualPageRomMode(bool enabled)
{
    DualPageMode = enabled;
    if (enabled)
        {
        IdromSizeWords = kDualPageRomSize;
        }
    else
        {
        IdromSizeWords = kSinglePageRomSize;
        }
    Clear();  // re-erase ROM shadow
}

//-----------------------------------------------------------------------------
//  AdvI2CIdrom::Clear() -- Blank out in-memory image
//-----------------------------------------------------------------------------

void  AdvI2CIdrom::Clear()
{
    for (int i=0; i<IdromSizeWords; i++)
        Packet[i] = 0;
}

//-----------------------------------------------------------------------------
//  AdvI2CIdrom::CalculateChecksum() -- Calculate checksum of packet
//-----------------------------------------------------------------------------

void  AdvI2CIdrom::CalculateChecksum()
{
    unsigned int checksum(0);

    for ( unsigned int i = 0; i < idrom_data_size(); i++)
        checksum += Packet[i];

    Packet[IdromSizeWords-ReservedSize] = checksum;
}

//-----------------------------------------------------------------------------
//  AdvI2CIdrom::SetRomPage() -- Swap ROM page
//-----------------------------------------------------------------------------

void  AdvI2CIdrom::SetRomPage(unsigned int page)
{
    unsigned int page_code = (page==0) ? kSPA0_Code : kSPA1_Code ;
    unsigned int subaddr = ROM.GetSubAddr();
    ROM.Reg_sub(page_code, 0, 0);
    ROM.SubAddr_Override(subaddr);  // restore original subaddr
}

//-----------------------------------------------------------------------------
//  AdvI2CIdrom::LoadFromRom() -- Use ROM to read the IDROM page into cache
//-----------------------------------------------------------------------------

void  AdvI2CIdrom::LoadFromRom()
{
    Clear();
    if (DualPageMode==false)
        {
        ROM.ReadBlock(0, Packet, kSinglePageRomSize);
        return;
        }
    //
    //  Dual Page ROM
    SetRomPage(0);
    //Sleep(100);

    ROM.ReadBlock(0, Packet, kSinglePageRomSize);

    SetRomPage(1);
   // Sleep(100);

    ROM.ReadBlock(0, Packet+kSinglePageRomSize, kSinglePageRomSize);

    SetRomPage(0);
    //Sleep(100);
}

//-----------------------------------------------------------------------------
//  AdvI2CIdrom::StoreToRom() -- Use ROM to write the IDROM page from cache
//-----------------------------------------------------------------------------

void  AdvI2CIdrom::StoreToRom()
{
    if (DualPageMode==false)
        {
        ROM.WriteBlock(0, Packet, kSinglePageRomSize);
        return;
        }
    //
    //  Dual Page ROM
    Sleep(100);
    SetRomPage(0);

    ROM.WriteBlock(0, Packet, kSinglePageRomSize);

    Sleep(100);
    SetRomPage(1);

    ROM.WriteBlock(0, Packet+kSinglePageRomSize, kSinglePageRomSize);

    Sleep(100);
    SetRomPage(0);

}

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif
//=============================================================================
//  CLASS AdvI2CRomSection  --  Object for access to a section of IDROM Page
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class AdvI2CRomSection
//-----------------------------------------------------------------------------

AdvI2CRomSection::AdvI2CRomSection(AdvI2CIdrom & rom, int offset,
                           int section_data_size, int section_reserve_size)
    :  IdromIoDG(this->IView()), Rom(rom), Offset(offset),
       SectionSize(section_data_size + section_reserve_size),
       SectionReserveSize(section_reserve_size)

{
}

#ifdef _MSC_VER
	#pragma warning(pop)
#endif
//-----------------------------------------------------------------------------
//  AdvI2CRomSection::Clear() -- Erase the section
//-----------------------------------------------------------------------------

void  AdvI2CRomSection::Clear()
{
    for (unsigned int i=0; i<SizeInInts(); i++)
        AsInt(i, 0);
}

//-----------------------------------------------------------------------------
//  AdvI2CRomSection::LoadFromRom() -- Read data from ROM
//-----------------------------------------------------------------------------

void  AdvI2CRomSection::LoadFromRom()
{
    Rom.LoadFromRom();
}

//-----------------------------------------------------------------------------
//  AdvI2CRomSection::StoreToRom() -- Write data to ROM
//-----------------------------------------------------------------------------

void  AdvI2CRomSection::StoreToRom()
{
    Rom.StoreToRom();
}

//------------------------------------------------------------------------
//  AdvI2CRomSection::Updated()  -- read 'written to' flag for a section
//------------------------------------------------------------------------

bool  AdvI2CRomSection::Updated() const
{
    if (SectionReserveSize == 0)
        return true;

    //  Use first word of reserved region as update flag
    const int * update_flag = const_idrom_base() + idrom_data_size();

    return ((*update_flag)==1) ? true : false ;
}

//------------------------------------------------------------------------
//   AdvI2CRomSection::Updated()  -- Mark data area as 'written to'
//------------------------------------------------------------------------

void  AdvI2CRomSection::Updated(bool state)
{
    if (SectionReserveSize == 0)
        return;

    //  Use first word of reserved region as update flag
    int * update_flag = idrom_base() + idrom_data_size();
    (*update_flag) = (state) ? 1 : 0 ;
}

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif
//=============================================================================
//  CLASS AdvI2C_16Idrom  --  Access to a segment of the X6 ROM as an IDROM
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class AdvI2C_16Idrom
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
//  constructor for class AdvI2C_16Idrom
//-----------------------------------------------------------------------------

AdvI2C_16Idrom::AdvI2C_16Idrom(I2CDevice16 & rom)
    :  IdromIoDG(this->IView()), ROM(rom)
{
    Clear();
}

#ifdef _MSC_VER
	#pragma warning(pop)
#endif
//-----------------------------------------------------------------------------
//  destructor for class AdvI2C_16Idrom
//-----------------------------------------------------------------------------

AdvI2C_16Idrom::~AdvI2C_16Idrom()
{
}

//-----------------------------------------------------------------------------
//  AdvI2C_16Idrom::Clear() -- Blank out in-memory image
//-----------------------------------------------------------------------------

void  AdvI2C_16Idrom:: Clear()
{
    for (int i=0; i<IdromSizeWords; i++)
        Packet[i] = 0;
}

//-----------------------------------------------------------------------------
//  AdvI2C_16Idrom::CalculateChecksum() -- Calculate checksum of packet
//-----------------------------------------------------------------------------

void  AdvI2C_16Idrom::CalculateChecksum()
{
    unsigned int checksum(0);

    for ( unsigned int i = 0; i < idrom_data_size(); i++)
        checksum += Packet[i];

    Packet[IdromSizeWords-ReservedSize] = checksum;
}

//-----------------------------------------------------------------------------
//  AdvI2C_16Idrom::LoadFromRom() -- Use ROM to read the IDROM page into cache
//-----------------------------------------------------------------------------

void  AdvI2C_16Idrom::LoadFromRom()
{
    Clear();
    ROM.ReadBlock(0, Packet, IdromSizeWords);
}

//-----------------------------------------------------------------------------
//  AdvI2C_16Idrom::StoreToRom() -- Use ROM to write the IDROM page from cache
//-----------------------------------------------------------------------------

void  AdvI2C_16Idrom::StoreToRom()
{
    ROM.WriteBlock(0, Packet, IdromSizeWords);
}

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif
//=============================================================================
//  CLASS AdvI2C_16RomSection  --  Object for access to a section of IDROM Page
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class AdvI2C_16RomSection
//-----------------------------------------------------------------------------

AdvI2C_16RomSection::AdvI2C_16RomSection(AdvI2C_16Idrom & rom, int offset,
                           int section_data_size, int section_reserve_size)
    :  IdromIoDG(this->IView()), Rom(rom), Offset(offset),
       SectionSize(section_data_size + section_reserve_size),
       SectionReserveSize(section_reserve_size)

{
}

#ifdef _MSC_VER
	#pragma warning(pop)
#endif
//-----------------------------------------------------------------------------
//  AdvI2C_16RomSection::Clear() -- Erase the section
//-----------------------------------------------------------------------------

void  AdvI2C_16RomSection::Clear()
{
    for (unsigned int i=0; i<SizeInInts(); i++)
        AsInt(i, 0);
}

//-----------------------------------------------------------------------------
//  AdvI2C_16RomSection::LoadFromRom() -- Read data from ROM
//-----------------------------------------------------------------------------

void  AdvI2C_16RomSection::LoadFromRom()
{
    Rom.LoadFromRom();
}

//-----------------------------------------------------------------------------
//  AdvI2C_16RomSection::StoreToRom() -- Write data to ROM
//-----------------------------------------------------------------------------

void  AdvI2C_16RomSection::StoreToRom()
{
    Rom.StoreToRom();
}

//------------------------------------------------------------------------
//  AdvI2C_16RomSection::Updated()  -- read 'written to' flag for a section
//------------------------------------------------------------------------

bool  AdvI2C_16RomSection::Updated() const
{
    if (SectionReserveSize == 0)
        return true;

    //  Use first word of reserved region as update flag
    const int * update_flag = const_idrom_base() + idrom_data_size();

    return ((*update_flag)==1) ? true : false ;
}

//------------------------------------------------------------------------
//   AdvI2C_16RomSection::Updated()  -- Mark data area as 'written to'
//------------------------------------------------------------------------

void  AdvI2C_16RomSection::Updated(bool state)
{
    if (SectionReserveSize == 0)
        return;

    //  Use first word of reserved region as update flag
    int * update_flag = idrom_base() + idrom_data_size();
    (*update_flag) = (state) ? 1 : 0 ;
}

//=============================================================================
//  CLASS AdvCalibrationSection  --  IDROM section for analog calibration
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class AdvCalibrationSection
//-----------------------------------------------------------------------------

AdvCalibrationSection::AdvCalibrationSection(AdvI2C_16Idrom & rom, int offset, int channels, int one,
                                           IAdvGainCalibrationDevice & gain_cal)
    :  AdvI2C_16RomSection(rom, offset, channels*2), Channels(channels), FOne(one),
       GainCal(gain_cal)
{
}

//-----------------------------------------------------------------------------
//  AdvCalibrationSection::ApplyDefaultCoefficients() -- load default set of coeffs into memory
//-----------------------------------------------------------------------------

void  AdvCalibrationSection::ApplyDefaultCoefficients()
{
    for (int ch = 0; ch < Channels; ++ch)
        {
        Offset(ch, 0);
        Gain(ch, 1.0);
        }
}

//-----------------------------------------------------------------------------
//  AdvCalibrationSection::ApplyRomCoefficients() --
//-----------------------------------------------------------------------------

void  AdvCalibrationSection::ApplyRomCoefficients()
{
    if (!Calibrated())
        {
        ApplyDefaultCoefficients();
        return;
        }

    for (int ch = 0; ch < Channels; ++ch)
        {
        Offset(ch, Offset(ch));
        Gain(ch, Gain(ch));
        }
}

//-----------------------------------------------------------------------------
//  AdvCalibrationSection::ToFixed() - Converts a float gain coeff to internal fixed-point format
//-----------------------------------------------------------------------------

unsigned int AdvCalibrationSection::ToFixed(float value)
{
    return static_cast<int>(value*One());
}

//------------------------------------------------------------------------------
//  AdvCalibrationSection::FromFixed() - Converts an internal fixed-point format to float
//------------------------------------------------------------------------------

float AdvCalibrationSection::FromFixed(unsigned int value)
{
    return static_cast<float>(value)/One();
}

//-----------------------------------------------------------------------------
//  AdvCalibrationSection::Gain()  --  Set gain coefficient
//-----------------------------------------------------------------------------

void AdvCalibrationSection::Gain(int ch, float gain)
{
    int idx = (Channels + ch);

    AsFloat(idx, gain);
    GainCal.SetCalibrationGain(ch, ToFixed(gain));
}

//-----------------------------------------------------------------------------
//  AdvCalibrationSection::Gain()  --  Get gain coefficient
//-----------------------------------------------------------------------------

float AdvCalibrationSection::Gain(int ch) const
{
    int idx = (Channels + ch);
    return AsFloat(idx);
}

//-----------------------------------------------------------------------------
//  AdvCalibrationSection::Offset()  --  Set offset coefficient
//-----------------------------------------------------------------------------

void AdvCalibrationSection::Offset(int ch, float offset)
{
    int idx = ch;

    AsFloat(idx, offset);
    GainCal.SetCalibrationOffset(ch, (int)offset);
}

//-----------------------------------------------------------------------------
//  AdvCalibrationSection::Offset()  --  Get offset coefficient
//-----------------------------------------------------------------------------

float AdvCalibrationSection::Offset(int ch) const
{
    int idx = ch;
    return AsFloat(idx);
}

//-----------------------------------------------------------------------------
//  AdvCalibrationSection::One()  --  Get Q-notation unity-value
//-----------------------------------------------------------------------------

int AdvCalibrationSection::One() const
{
    return FOne;
}

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif
//=============================================================================
//  CLASS GeneralIdromImage  --  Access to a segment of the X6 ROM as an IDROM
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class GeneralIdromImage
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
//  constructor for class GeneralIdromImage
//-----------------------------------------------------------------------------

GeneralIdromImage::GeneralIdromImage(unsigned int idrom_size_words, unsigned int reserved_size)
    :  IdromIoDG(this->IView()), Packet(idrom_size_words),
       IdromSizeWords(idrom_size_words), ReservedSize(reserved_size)
{
    Clear();
}

#ifdef _MSC_VER
    #pragma warning(pop)
#endif
//-----------------------------------------------------------------------------
//  destructor for class GeneralIdromImage
//-----------------------------------------------------------------------------

GeneralIdromImage::~GeneralIdromImage()
{
}

//-----------------------------------------------------------------------------
//  GeneralIdromImage::Reconfigure() -- Re-init Image with new parameters
//-----------------------------------------------------------------------------

void   GeneralIdromImage::Reconfigure(unsigned int idrom_size_words, unsigned int reserved_size)
{
    IdromSizeWords = idrom_size_words;
    ReservedSize = reserved_size;
    Packet.resize(idrom_size_words);
}

//-----------------------------------------------------------------------------
//  GeneralIdromImage::Clear() -- Blank out in-memory image
//-----------------------------------------------------------------------------

void  GeneralIdromImage:: Clear()
{
    for (size_t i=0; i<Packet.size(); i++)
        Packet[i] = 0;
}

//-----------------------------------------------------------------------------
//  GeneralIdromImage::CalculateChecksum() -- Calculate checksum of packet
//-----------------------------------------------------------------------------

void  GeneralIdromImage::CalculateChecksum()
{
    unsigned int checksum(0);

    for ( unsigned int i = 0; i < idrom_data_size(); i++)
        checksum += Packet[i];

    Packet[IdromSizeWords-ReservedSize-1] = checksum;
}

//-----------------------------------------------------------------------------
//  GeneralIdromImage::LoadFromRom() -- Use ROM to read the IDROM page into cache
//-----------------------------------------------------------------------------

void  GeneralIdromImage::LoadFromRom()
{
    Clear();
    ROM_ReadBlock(0, &Packet[0], IdromSizeWords);
}

//-----------------------------------------------------------------------------
//  GeneralIdromImage::StoreToRom() -- Use ROM to write the IDROM page from cache
//-----------------------------------------------------------------------------

void  GeneralIdromImage::StoreToRom()
{
    ROM_WriteBlock(0, &Packet[0], IdromSizeWords);
}

//------------------------------------------------------------------------
//  GeneralIdromImage::Reserved()  -- read reserved area
//------------------------------------------------------------------------

int  GeneralIdromImage::Reserved(int idx) const
{
    if (ReservedSize <= static_cast<unsigned int>(idx))
        return 0;

    //  Use Nth word of reserved region
    const int * reserved_ptr = const_idrom_base() + idrom_data_size();

    return *(reserved_ptr+idx) ;
}

//------------------------------------------------------------------------
// GeneralIdromImage::Reserved()  -- write to reserved area
//------------------------------------------------------------------------

void  GeneralIdromImage::Reserved(int idx, int data)
{
    if (ReservedSize <= static_cast<unsigned int>(idx))
        return;

    //  Use Nth word of reserved region
    int * reserved_ptr = idrom_base() + idrom_data_size();
    *(reserved_ptr+idx) = data ;
}


#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif
//=============================================================================
//  CLASS GeneralRomSection  --  Object for access to a section of IDROM Page
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class GeneralRomSection
//-----------------------------------------------------------------------------

GeneralRomSection::GeneralRomSection(GeneralIdromImage & rom, int offset,
                           int section_data_size, int section_reserve_size)
    :  IdromIoDG(this->IView()), Rom(rom), Offset(offset),
       SectionSize(section_data_size + section_reserve_size),
       SectionReserveSize(section_reserve_size)

{
}

#ifdef _MSC_VER
    #pragma warning(pop)
#endif
//-----------------------------------------------------------------------------
//  GeneralRomSection::Clear() -- Erase the section
//-----------------------------------------------------------------------------

void  GeneralRomSection::Clear()
{
    for (unsigned int i=0; i<SizeInInts(); i++)
        AsInt(i, 0);
}

//-----------------------------------------------------------------------------
//  GeneralRomSection::LoadFromRom() -- Read data from ROM
//-----------------------------------------------------------------------------

void  GeneralRomSection::LoadFromRom()
{
    Rom.LoadFromRom();
}

//-----------------------------------------------------------------------------
//  GeneralRomSection::StoreToRom() -- Write data to ROM
//-----------------------------------------------------------------------------

void  GeneralRomSection::StoreToRom()
{
    Rom.StoreToRom();
}

//------------------------------------------------------------------------
//  GeneralRomSection::Updated()  -- read 'written to' flag for a section
//------------------------------------------------------------------------

bool  GeneralRomSection::Updated() const
{
    if (SectionReserveSize == 0)
        return false;

    //  Use first word of reserved region as update flag
    const int * update_flag = const_idrom_base() + idrom_data_size();

    return ((*update_flag)==1) ? true : false ;
}

//------------------------------------------------------------------------
//   GeneralRomSection::Updated()  -- Mark data area as 'written to'
//------------------------------------------------------------------------

void  GeneralRomSection::Updated(bool state)
{
    if (SectionReserveSize == 0)
        return;

    //  Use first word of reserved region as update flag
    int * update_flag = idrom_base() + idrom_data_size();
    (*update_flag) = (state) ? 1 : 0 ;
}

//------------------------------------------------------------------------
//  GeneralRomSection::UpdatedM()  -- read 'written to' flag for a section
//------------------------------------------------------------------------

bool  GeneralRomSection::UpdatedM(int idx) const
{   
    if (SectionReserveSize <= idx)
        return false;

    //  Use Nth word of reserved region as update flag
    const int * update_flag = const_idrom_base() + idrom_data_size();

    return ((*(update_flag+idx))==1) ? true : false ;
}

//------------------------------------------------------------------------
//   GeneralRomSection::UpdatedM()  -- Mark data area as 'written to'
//------------------------------------------------------------------------

void  GeneralRomSection::UpdatedM(int idx, bool state)
{
    if (SectionReserveSize <= idx)
        return;

    //  Use Nth word of reserved region as update flag
    int * update_flag = idrom_base() + idrom_data_size(); // skip checksum?
    *(update_flag+idx) = (state) ? 1 : 0 ;
}

//=============================================================================
//  CLASS GeneralCalibrationSection  --  IDROM section for analog calibration
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class GeneralCalibrationSection
//-----------------------------------------------------------------------------
//  Bump Data Size by 1 to leave room for "checksum" field at the end

GeneralCalibrationSection::GeneralCalibrationSection(GeneralIdromImage & rom, int offset, int channels, int one,
                                           IAdvGainCalibrationDevice & gain_cal)
    :  GeneralRomSection(rom, offset, channels*2+1, 1), Channels(channels), FOne(one),
       GainCal(gain_cal)
{
}

//-----------------------------------------------------------------------------
//  GeneralCalibrationSection::ApplyDefaultCoefficients() -- load default set of coeffs into memory
//-----------------------------------------------------------------------------

void  GeneralCalibrationSection::ApplyDefaultCoefficients()
{
    for (int ch = 0; ch < Channels; ++ch)
        {
        Offset(ch, 0);
        Gain(ch, 1.0);
        }
}

//-----------------------------------------------------------------------------
//  GeneralCalibrationSection::ApplyRomCoefficients() --
//-----------------------------------------------------------------------------

void  GeneralCalibrationSection::ApplyRomCoefficients()
{
    if (!Calibrated())
        {
        ApplyDefaultCoefficients();
        return;
        }

    for (int ch = 0; ch < Channels; ++ch)
        {
        Offset(ch, Offset(ch));
        Gain(ch, Gain(ch));
        }
}

//-----------------------------------------------------------------------------
//  GeneralCalibrationSection::ToFixed() - Converts a float gain coeff to internal fixed-point format
//-----------------------------------------------------------------------------

unsigned int GeneralCalibrationSection::ToFixed(float value)
{
    return static_cast<int>(value*One());
}

//------------------------------------------------------------------------------
//  GeneralCalibrationSection::FromFixed() - Converts an internal fixed-point format to float
//------------------------------------------------------------------------------

float GeneralCalibrationSection::FromFixed(unsigned int value)
{
    return static_cast<float>(value)/One();
}

//-----------------------------------------------------------------------------
//  GeneralCalibrationSection::Gain()  --  Set gain coefficient
//-----------------------------------------------------------------------------

void GeneralCalibrationSection::Gain(int ch, float gain)
{
    int idx = (Channels + ch);

    AsFloat(idx, gain);
    GainCal.SetCalibrationGain(ch, ToFixed(gain));
}

//-----------------------------------------------------------------------------
//  GeneralCalibrationSection::Gain()  --  Get gain coefficient
//-----------------------------------------------------------------------------

float GeneralCalibrationSection::Gain(int ch) const
{
    int idx = (Channels + ch);
    return AsFloat(idx);
}

//-----------------------------------------------------------------------------
//  GeneralCalibrationSection::Offset()  --  Set offset coefficient
//-----------------------------------------------------------------------------

void GeneralCalibrationSection::Offset(int ch, float offset)
{
    int idx = ch;

    AsFloat(idx, offset);
    GainCal.SetCalibrationOffset(ch, (int)offset);
}

//-----------------------------------------------------------------------------
//  GeneralCalibrationSection::Offset()  --  Get offset coefficient
//-----------------------------------------------------------------------------

float GeneralCalibrationSection::Offset(int ch) const
{
    int idx = ch;
    return AsFloat(idx);
}

//-----------------------------------------------------------------------------
//  GeneralCalibrationSection::One()  --  Get Q-notation unity-value
//-----------------------------------------------------------------------------

int GeneralCalibrationSection::One() const
{
    return FOne;
}

//=============================================================================
//  CLASS GeneralRangedCalibrationSection  --  IDROM section for analog calibration
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class GeneralRangedCalibrationSection
//-----------------------------------------------------------------------------
//  Bump Data Size by 1 to leave room for "checksum" field at the end

GeneralRangedCalibrationSection::GeneralRangedCalibrationSection(
                                            GeneralIdromImage & rom, int offset,
                                            GainRanges gain_ranges, CalChannels channels,
                                            int one,
                                            IAdvGainCalibrationDevice & gain_cal)
    :  GeneralRomSection(rom, offset, gain_ranges.AsInt()*channels.AsInt()*2+1, gain_ranges.AsInt()),
       Ranges(gain_ranges), Channels(channels),
       GainSettings(channels.AsInt()), FOne(one),
       GainCal(gain_cal)
{
    std::fill(GainSettings.begin(), GainSettings.end(), 0u);
}

//-----------------------------------------------------------------------------
//  GeneralRangedCalibrationSection::ApplyDefaultCoefficients() -- load default set of coeffs into memory
//-----------------------------------------------------------------------------

void  GeneralRangedCalibrationSection::ApplyDefaultCoefficients()
{
    for (int ch = 0; ch < Channels.AsInt(); ++ch)
        {
        size_t gain_range = GainSettings[ch];
        Offset(gain_range, ch, 0);
        Gain(gain_range, ch, 1.0);
        }
}

//-----------------------------------------------------------------------------
//  GeneralRangedCalibrationSection::ApplyRomCoefficients() --
//-----------------------------------------------------------------------------

void  GeneralRangedCalibrationSection::ApplyRomCoefficients()
{

    for (int ch = 0; ch < Channels.AsInt(); ++ch)
        {
        size_t gain_range = GainSettings[ch];
        float offset;
        float gain;
        if (!Calibrated(gain_range))
            {
            offset = 0.0;
            gain = 1.0;
            }
        else
            {
            offset = Offset(gain_range, ch);
            gain = Gain(gain_range, ch);
            }
        Offset(gain_range, ch, offset);
        Gain(gain_range, ch, gain);
        }
}

//-----------------------------------------------------------------------------
//  GeneralRangedCalibrationSection::ToFixed() - Converts a float gain coeff to internal fixed-point format
//-----------------------------------------------------------------------------

unsigned int GeneralRangedCalibrationSection::ToFixed(float value)
{
    return static_cast<int>(value*One());
}

//------------------------------------------------------------------------------
//  GeneralRangedCalibrationSection::FromFixed() - Converts an internal fixed-point format to float
//------------------------------------------------------------------------------

float GeneralRangedCalibrationSection::FromFixed(unsigned int value)
{
    return static_cast<float>(value)/One();
}

//------------------------------------------------------------------------------
//  GeneralRangedCalibrationSection::GainRange() - read current gain range
//------------------------------------------------------------------------------

size_t  GeneralRangedCalibrationSection::GainRange(int ch)
{
    if (static_cast<size_t>(ch) < GainSettings.size())
        return GainSettings[ch];
    else
        return 0;
}

//------------------------------------------------------------------------------
//  GeneralRangedCalibrationSection::GainRange() - read current gain range
//------------------------------------------------------------------------------

void  GeneralRangedCalibrationSection::GainRange(int ch, size_t gain_range)
{
    if (static_cast<size_t>(ch) < GainSettings.size())
		GainSettings[ch] = static_cast<unsigned int>(gain_range);
}

//-----------------------------------------------------------------------------
//  GeneralRangedCalibrationSection::Gain()  --  Set gain coefficient
//-----------------------------------------------------------------------------

void GeneralRangedCalibrationSection::Gain(size_t gain_range, int ch, float gain)
{
    int range_size = (Channels.AsInt()*2);
	int range_ofs = static_cast<int>(gain_range)*range_size;
    int idx = range_ofs + (Channels.AsInt() + ch);

    AsFloat(idx, gain);
    GainCal.SetCalibrationGain(ch, ToFixed(gain));
}

//-----------------------------------------------------------------------------
//  GeneralRangedCalibrationSection::Gain()  --  Get gain coefficient
//-----------------------------------------------------------------------------

float GeneralRangedCalibrationSection::Gain(size_t gain_range, int ch) const
{
    int range_size = (Channels.AsInt()*2);
	int range_ofs = static_cast<int>(gain_range)*range_size;
    int idx = range_ofs + (Channels.AsInt() + ch);

    return AsFloat(idx);
}

//-----------------------------------------------------------------------------
//  GeneralRangedCalibrationSection::Offset()  --  Set offset coefficient
//-----------------------------------------------------------------------------

void GeneralRangedCalibrationSection::Offset(size_t gain_range, int ch, float offset)
{
    int range_size = (Channels.AsInt()*2);
	int range_ofs = static_cast<int>(gain_range)*range_size;
    int idx = range_ofs + ch;

    AsFloat(idx, offset);
    GainCal.SetCalibrationOffset(ch, (int)offset);
}

//-----------------------------------------------------------------------------
//  GeneralRangedCalibrationSection::Offset()  --  Get offset coefficient
//-----------------------------------------------------------------------------

float GeneralRangedCalibrationSection::Offset(size_t gain_range, int ch) const
{
    int range_size = (Channels.AsInt()*2);
	int range_ofs = static_cast<int>(gain_range)*range_size;
    int idx = range_ofs + ch;

    return AsFloat(idx);
}

//-----------------------------------------------------------------------------
//  GeneralRangedCalibrationSection::One()  --  Get Q-notation unity-value
//-----------------------------------------------------------------------------

int GeneralRangedCalibrationSection::One() const
{
    return FOne;
}

}  // end namespace Innovative
