//---------------------------------------------------------------------------
// PmcIdrom_Mb.h  --  Omnibus Module Idrom Access Object
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#ifndef PmcIdrom_MbH
#define PmcIdrom_MbH

#include <I2C_Mb.h>
#include <string>
#include <IdromDatagram_Mb.h>

namespace Innovative
{

//=============================================================================
//  CLASS PmcIdrom  --  Omnibus Module Idrom Access Object
//=============================================================================
//  The Module IDROM contains a header containing a 12 character name field
//     (1 word per character), follwed by a Revision Code character. The
//     final word is a checksum.  This leaves a Working Area of 110 words
//     that are divided up between module-specific data and fallow words
//     that can be used by an applicaton.
//
class PmcIdrom : public I2CDevice
{
    static const int IdromPacketSize = 512;   // IDROM size in words

public:
    enum  WorkArea { waModule, waUser, waRaw };
    //
    //  Member Functions
    PmcIdrom(I2CDeviceIntf & intf, int module_area_size, bool endian=false);
    ~PmcIdrom();

    void   Clear();
    void   LoadFromRom();
    void   StoreToRom();

    //
    //  Properties
    std::string     Name() const;
    void            Name(const std::string & name);
    std::string     Revision() const;
    void            Revision(const std::string & revision);
    unsigned int    SerialNumber() const;
    void            SerialNumber(unsigned int Num);
    unsigned int    PartNumber() const;
    void            PartNumber(unsigned int part);
    unsigned int    DashNumber() const;
    void            DashNumber(unsigned int dash);
    std::string     DashVariant() const;
    void            DashVariant(const std::string & dashV);
    unsigned int    RuggedizationNumber() const;
    void            RuggedizationNumber(unsigned int rugged);
    unsigned int    EcoSubRevNumber() const;
    void            EcoSubRevNumber(unsigned int eco_subrev);

    //Sanitizer
    void            ChipErase(int pattern = 0);
    bool            VerifyChipErased(int pattern = 0);
    std::string     PrintChipData();

    bool            IsEmpty() const;
    void            ModuleAreaWords(int module_area_size)
        {  SetAreaSizes(module_area_size);  Clear();  }
    int             ModuleAreaWords() const
        {  return ModuleAreaSize;  }
    int             UserAreaWords() const
        {  return UserAreaSize;  }
    int             Checksum() const
        {  return WorkAreaAsInt(waRaw, IdromPacketSize-1);  }

    void            SetReserve(int reserve_size)
        {  ReserveSize = reserve_size; ModuleAreaWords(ModuleAreaSize); }

    // ...User Area Access Functions
    int    AsInt(int idx) const;
    void   AsInt(int idx, int data);
    float  AsFloat(int idx) const;
    void   AsFloat(int idx, float data);
    double AsDouble(int idx) const;
    void   AsDouble(int idx, double data);
    short  AsShort(int idx) const;
    void   AsShort(int idx, short data);
    char   AsChar(int idx) const;
    void   AsChar(int idx, char data);
    void   AsString(unsigned int int_idx, unsigned int ints_to_use, const std::string & text);
    std::string  AsString(unsigned int int_idx, unsigned int ints_to_use) const;

    //  ...Full access I/O functions
    int    WorkAreaAsInt(WorkArea area, int idx) const;
    void   WorkAreaAsInt(WorkArea area, int idx, int data);
    float  WorkAreaAsFloat(WorkArea area, int idx) const;
    void   WorkAreaAsFloat(WorkArea area, int idx, float data);
    double WorkAreaAsDouble(WorkArea area, int idx) const;
    void   WorkAreaAsDouble(WorkArea area, int idx, double data);
    short  WorkAreaAsShort(WorkArea area, int idx) const;
    void   WorkAreaAsShort(WorkArea area, int idx, short data);
    char   WorkAreaAsChar(WorkArea area, int idx) const;
    void   WorkAreaAsChar(WorkArea area, int idx, char data);


private:
    //
    //  Member Data
    int         Packet[IdromPacketSize];
    int         ReserveSize;
    int         HeaderSize;
    int         ModuleAreaSize;
    int         UserAreaSize;
    int         TrailerSize;

    bool        FLocationSet;
    int		    FClockBit;
    int		    FDataBit;
    int		    FDataPin;

    //
    //  Support methods
    void            SetAreaSizes(int module_area_size);
    int *           WorkAreaBase(WorkArea area);
    const int *     CWorkAreaBase(WorkArea area) const;
    void            CalculateChecksum();
	unsigned int    SizeInChars() const;
};

//=============================================================================
//  CLASS PmcIdromSection  --  Object controlling access to a section of IDROM
//=============================================================================
//  When a single IDROM Work area is to be used by multiple clients, you can use a
//     predefined section object to manage the location of the data in the ROM

class PmcIdromSection
{
public:

    enum  WorkArea { waModule, waUser, waRaw };
    //
    //  Member Functions
    PmcIdromSection(PmcIdrom & rom, PmcIdrom::WorkArea wa, int offset, int section_size_words);
    virtual ~PmcIdromSection()
        {  }

    // Copy
    PmcIdromSection(const PmcIdromSection & other);
    // Copy
    PmcIdromSection & operator =(const PmcIdromSection & other);

    void   Clear();
    void   LoadFromRom();
    void   StoreToRom();

    int    AsInt(int idx) const;
    void   AsInt(int idx, int data);
    float  AsFloat(int idx) const;
    void   AsFloat(int idx, float data);
    double AsDouble(int idx) const;
    void   AsDouble(int idx, double data);
    short  AsShort(int idx) const;
    void   AsShort(int idx, short data);
    char   AsChar(int idx) const;
    void   AsChar(int idx, char data);

private:
    PmcIdrom &          Rom;
    PmcIdrom::WorkArea  WA;
    int                 Offset;
    int                 SectionSize;
};

//=============================================================================
//  CLASS AdvI2CIdrom  --  Access to an 8 bit address ROM as an IDROM
//=============================================================================
//  The last word of the IDROM page is an checksum of the rest of the page
//

class AdvI2CIdrom :  public IIdromViewable, public IdromIoDG
{
    static const int MaxIdromSizeWords = 512/sizeof(int);   // Maximum IDROM size supported
    static const int ReservedSize  = 1;  //  Checksum area

public:

    friend class AdvI2CRomSection;
    //
    //  Member Functions
    AdvI2CIdrom(I2CDeviceN & rom);
    virtual ~AdvI2CIdrom();

    // ROM Mode
    void   DualPageRomMode(bool enabled);

    //  Data Operations
    void   Clear();
    int    Checksum() const
        {  return Packet[IdromSizeWords-1];  }

    //  ROM Operations
    void   LoadFromRom();
    void   StoreToRom();

    I2CDeviceN &        Device() {  return ROM;  };

protected:
    //
    //  IIdromViewable Interface Implementation
    virtual const int *  const_idrom_base() const
                            {  return Packet;  }
    virtual int *        idrom_base()
                            {  return Packet;  }

    virtual size_t       idrom_size() const
                            {  return IdromSizeWords;  }
    virtual size_t       idrom_reserve_size() const
                            {  return ReservedSize;  }
    size_t               idrom_data_size() const
                            {  return idrom_size() - idrom_reserve_size();  }

    virtual void         after_data_write()
                            {  CalculateChecksum();  }

    IIdromViewable & IView()
        {  return *this;  }

private:
    //
    //  Member Data
    I2CDeviceN &        ROM;
    int                 Packet[MaxIdromSizeWords];
    int                 IdromSizeWords;
    bool                DualPageMode;
    //
    //  Support methods
    void        CalculateChecksum();
    void        SetRomPage(unsigned int page);
};

//=============================================================================
//  CLASS AdvI2CRomSection  --  Object for access to a section of IDROM Page
//=============================================================================
//  When a single IDROM Work area is to be used by multiple clients, you can use a
//     predefined section object to manage the location of the data in the ROM

class AdvI2CRomSection  :  public IIdromViewable, public IdromIoDG
{
public:
    //
    //  Member Functions
    AdvI2CRomSection(AdvI2CIdrom & rom, int offset,
                           int section_data_size, int section_reserve_size=1);
    virtual ~AdvI2CRomSection()
        {  }

    void   Clear();
    void   LoadFromRom();
    void   StoreToRom();

    bool   Updated() const;          // 'Section written' flag
    void   Updated(bool state);

protected:
    //
    //  IIdromViewable Interface Implementation
    virtual const int *  const_idrom_base() const
                            {  return Rom.const_idrom_base() + Offset;  }
    virtual int *        idrom_base()
                            {  return Rom.idrom_base() + Offset;;  }

    virtual size_t       idrom_size() const
                            {  return SectionSize;  }
    virtual size_t       idrom_reserve_size() const
                            {  return SectionReserveSize;  }
    size_t               idrom_data_size() const
                            {  return idrom_size() - idrom_reserve_size();  }

    virtual void         after_data_write()
                            {  Rom.after_data_write();  }

    IIdromViewable & IView()
        {  return *this;  }

private:
    AdvI2CIdrom &       Rom;
    int                 Offset;
    int                 SectionSize;
    int                 SectionReserveSize;
};

//=============================================================================
//  CLASS AdvI2C_16Idrom  --  Access to a 16 bit address ROM as an IDROM
//=============================================================================
//  The last word of the IDROM page is a checksum of the rest of the page
//

class AdvI2C_16Idrom :  public IIdromViewable, public IdromIoDG
{
    static const int IdromSizeWords = 2048;   // IDROM size in words
    static const int ReservedSize  = 1;      //  Checksum area

public:

    friend class AdvI2C_16RomSection;
    //
    //  Member Functions
    AdvI2C_16Idrom(I2CDevice16 & rom);
    virtual ~AdvI2C_16Idrom();

    //  Data Operations
    void   Clear();
    int    Checksum() const
        {  return Packet[IdromSizeWords-1];  }

    //  ROM Operations
    void   LoadFromRom();
    void   StoreToRom();

    I2CDevice16 &        Device() {  return ROM;  };

protected:
    //
    //  IIdromViewable Interface Implementation
    virtual const int *  const_idrom_base() const
                            {  return Packet;  }
    virtual int *        idrom_base()
                            {  return Packet;  }

    virtual size_t       idrom_size() const
                            {  return IdromSizeWords;  }
    virtual size_t       idrom_reserve_size() const
                            {  return ReservedSize;  }
    size_t               idrom_data_size() const
                            {  return idrom_size() - idrom_reserve_size();  }

    virtual void         after_data_write()
                            {  CalculateChecksum();  }

    IIdromViewable & IView()
        {  return *this;  }

private:
    //
    //  Member Data
    I2CDevice16 &       ROM;
    int                 Packet[IdromSizeWords];
    //
    //  Support methods
    void        CalculateChecksum();
};

//=============================================================================
//  CLASS AdvI2C_16RomSection  --  Object for access to a section of IDROM Page
//=============================================================================
//  When a single IDROM Work area is to be used by multiple clients, you can use a
//     predefined section object to manage the location of the data in the ROM

class AdvI2C_16RomSection  :  public IIdromViewable, public IdromIoDG
{
public:
    //
    //  Member Functions
    AdvI2C_16RomSection(AdvI2C_16Idrom & rom, int offset,
                           int section_data_size, int section_reserve_size=1);
    virtual ~AdvI2C_16RomSection()
        {  }

    void   Clear();
    void   LoadFromRom();
    void   StoreToRom();

    bool   Updated() const;          // 'Section written' flag
    void   Updated(bool state);

protected:
    //
    //  IIdromViewable Interface Implementation
    virtual const int *  const_idrom_base() const
                            {  return Rom.const_idrom_base() + Offset;  }
    virtual int *        idrom_base()
                            {  return Rom.idrom_base() + Offset;;  }

    virtual size_t       idrom_size() const
                            {  return SectionSize;  }
    virtual size_t       idrom_reserve_size() const
                            {  return SectionReserveSize;  }
    size_t               idrom_data_size() const
                            {  return idrom_size() - idrom_reserve_size();  }

    virtual void         after_data_write()
                            {  Rom.after_data_write();  }

    IIdromViewable & IView()
        {  return *this;  }

private:
    AdvI2C_16Idrom &    Rom;
    int                 Offset;
    int                 SectionSize;
    int                 SectionReserveSize;
};

//=============================================================================
//  CLASS IAdvGainCalibrationDevice  --  Interface to gain calibration registers
//=============================================================================

class IAdvGainCalibrationDevice
{
public:
    virtual void   SetCalibrationOffset(unsigned int ch, unsigned int value) = 0;
    virtual void   SetCalibrationGain(unsigned int ch, unsigned int value) = 0;

    virtual ~IAdvGainCalibrationDevice()
        {}

};

//=============================================================================
//  CLASS AdvCalibrationSection  --  IDROM section for analog calibration
//=============================================================================

class AdvCalibrationSection :  public AdvI2C_16RomSection
{
public:

    AdvCalibrationSection(AdvI2C_16Idrom & rom, int offset, int channels, int one,
                         IAdvGainCalibrationDevice & gain_cal);
    virtual ~AdvCalibrationSection()
        {  }

    // Properties
    bool    Calibrated() const
                {  return Updated();  }
    void    Calibrated(bool state)
                {  Updated(state);  }

    float   Offset(int ch) const;
    void    Offset(int ch, float value);

    float   Gain(int ch) const;
    void    Gain(int ch, float value);

    //  Loading functions
    void    ApplyRomCoefficients();
    void    ApplyDefaultCoefficients();

protected:
    // Data
    int                 Channels;
    int                 FOne;

    // Methods
    unsigned int    ToFixed(float value);
    float           FromFixed(unsigned int value);

    int             One() const;

    IAdvGainCalibrationDevice & GainCal;

};

//=============================================================================
//  CLASS GeneralIdromImage  --  IDROM image class
//=============================================================================
//  The last word of the IDROM page is a checksum of the rest of the page
//

class GeneralIdromImage :  public IIdromViewable, public IdromIoDG
{
public:
    friend class GeneralRomSection;
    //
    //  Member Functions
    GeneralIdromImage(unsigned int idrom_size_words, unsigned int reserved_size);
    virtual ~GeneralIdromImage();

    void   Reconfigure(unsigned int idrom_size_words, unsigned int reserved_size);

    //  Data Operations
    void   Clear();
    int    Checksum() const
        {  return Packet[IdromSizeWords-ReservedSize-1];  }

    int   Reserved(int idx) const;
    void  Reserved(int idx, int data);

    //  ROM Operations
    void   LoadFromRom();
    void   StoreToRom();

protected:
    //
    //  IIdromViewable Interface Implementation
    virtual const int *  const_idrom_base() const
                            {  return &Packet[0];  }
    virtual int *        idrom_base()
                            {  return &Packet[0];  }

    virtual size_t       idrom_size() const
                            {  return IdromSizeWords;  }
    virtual size_t       idrom_reserve_size() const
                            {  return ReservedSize;  }
    size_t               idrom_data_size() const
                            {  return idrom_size() - idrom_reserve_size();  }

    virtual void         after_data_write()
                            {  CalculateChecksum();  }

    IIdromViewable & IView()
        {  return *this;  }

    virtual void ROM_ReadBlock(int offset, int * data, int words) = 0;
    virtual void ROM_WriteBlock(int offset, const int * data, int words) = 0;

protected:
    //
    //  Member Data
    std::vector<int>  Packet;
    unsigned int IdromSizeWords;
    unsigned int ReservedSize;
    //
    //  Support methods
    void        CalculateChecksum();
};

//=============================================================================
//  CLASS GeneralRomSection  --  Object for access to a section of IDROM Page
//=============================================================================
//  When a single IDROM Work area is to be used by multiple clients, you can use a
//     predefined section object to manage the location of the data in the ROM

class GeneralRomSection :  public IIdromViewable, public IdromIoDG
{
public:
    //
    //  Member Functions
    GeneralRomSection(GeneralIdromImage & rom, int offset,
                           int section_data_size, int section_reserve_size=1);
    virtual ~GeneralRomSection()
        {  }

    void   Clear();
    void   LoadFromRom();
    void   StoreToRom();

    bool   Updated() const;          // 'Section written' flag
    void   Updated(bool state);

    bool   UpdatedM(int idx) const;          // 'Section written' flag
    void   UpdatedM(int idx, bool state);

protected:
    //
    //  IIdromViewable Interface Implementation
    virtual const int *  const_idrom_base() const
                            {  return Rom.const_idrom_base() + Offset;  }
    virtual int *        idrom_base()
                            {  return Rom.idrom_base() + Offset;;  }

    virtual size_t       idrom_size() const
                            {  return SectionSize;  }
    virtual size_t       idrom_reserve_size() const
                            {  return SectionReserveSize;  }
    size_t               idrom_data_size() const
                            {  return idrom_size() - idrom_reserve_size();  }

    virtual void         after_data_write()
                            {  Rom.after_data_write();  }

    IIdromViewable & IView()
        {  return *this;  }

private:
    GeneralIdromImage &    Rom;
    int                    Offset;
    int                    SectionSize;
    int                    SectionReserveSize;
};


//=============================================================================
//  CLASS GeneralCalibrationSection  --  IDROM section for analog calibration
//=============================================================================

class GeneralCalibrationSection :  public GeneralRomSection
{
public:

    GeneralCalibrationSection(GeneralIdromImage & rom, int offset, int channels, int one,
                         IAdvGainCalibrationDevice & gain_cal);
    virtual ~GeneralCalibrationSection()
        {  }

    // Properties
    bool    Calibrated() const
                {  return Updated();  }
    void    Calibrated(bool state)
                {  Updated(state);  }

    float   Offset(int ch) const;
    void    Offset(int ch, float value);

    float   Gain(int ch) const;
    void    Gain(int ch, float value);

    //  Loading functions
    void    ApplyRomCoefficients();
    void    ApplyDefaultCoefficients();

protected:
    // Data
    int                 Channels;
    int                 FOne;

    // Methods
    unsigned int    ToFixed(float value);
    float           FromFixed(unsigned int value);

    int             One() const;

    IAdvGainCalibrationDevice & GainCal;

};

//=============================================================================
//  CLASS GeneralRangedCalibrationSection  --  IDROM section for analog calibration
//=============================================================================

class GeneralRangedCalibrationSection :  public GeneralRomSection
{
public:
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    class CalChannels
    {
    public:
        int channels;

        explicit CalChannels(int chans) : channels(chans) {}
        int AsInt() const {  return channels;  }

    };
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    class GainRanges
    {
    public:
        int ranges;

        explicit GainRanges(int rng) : ranges(rng) {}
        int AsInt() const {  return ranges;  }

    };

    GeneralRangedCalibrationSection(GeneralIdromImage & rom, int offset,
                              GainRanges gain_ranges, CalChannels channels,
                              int one,
                              IAdvGainCalibrationDevice & gain_cal);
    virtual ~GeneralRangedCalibrationSection()
        {  }

    // Properties
    bool    Calibrated(size_t gain_range) const
                {  return UpdatedM(static_cast<int>(gain_range));  }
    void    Calibrated(size_t gain_range, bool state)
                {  UpdatedM(static_cast<int>(gain_range), state);  }

    size_t  GainRange(int ch);
    void    GainRange(int ch, size_t gain_range);

    float   Offset(size_t gain_range, int ch) const;
    void    Offset(size_t gain_range, int ch, float value);

    float   Gain(size_t gain_range, int ch) const;
    void    Gain(size_t gain_range, int ch, float value);

    //  Loading functions
    void    ApplyRomCoefficients();
    void    ApplyDefaultCoefficients();

protected:
    // Data
    GainRanges                 Ranges;
    CalChannels                Channels;
    std::vector<unsigned int>  GainSettings;
    int                        FOne;

    // Methods
    unsigned int    ToFixed(float value);
    float           FromFixed(unsigned int value);

    int             One() const;

    IAdvGainCalibrationDevice & GainCal;

};


}  // end namespace Innovative

#endif
