// SRecord_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//
#include <SRecord_Mb.h>
#include <algorithm>
#include <sstream>
#include <Exception_Mb.h>
#include <StringSupport_Mb.h>

namespace   Innovative
{

//===========================================================================
// CLASS CIISRecordException  --  Motorola S-record exception base
//===========================================================================

class CIISRecordException : public MalibuException
{
public:
    virtual const char* what() const throw() {return "class CIISRecordException";}
};

//===========================================================================
// CLASS CIISRecordFileErrorException  --  Motorola S-record exception
//===========================================================================

class CIISRecordFileErrorException : public CIISRecordException
{
public:
    virtual const char* what() const throw() {return "class CIISRecordFileErrorException";}
};

//===========================================================================
// CLASS CIISRecordFileFormatException  --  Motorola S-record exception
//===========================================================================

class CIISRecordFileFormatException : public CIISRecordException
{
public:
    virtual const char* what() const throw() {return "class CIISRecordFileFormatException";}
};

//===========================================================================
// CLASS SRecord  --  Motorola S-record file reader
//===========================================================================

//---------------------------------------------------------------------------
//  SRecord::SRecord() --  Ctor
//---------------------------------------------------------------------------

SRecord::SRecord(int size)
    : FSize(0)
{
    Size(size);
}

//---------------------------------------------------------------------------
//  SRecord::~SRecord() --  Dtor
//---------------------------------------------------------------------------

SRecord::~SRecord()
{
}

//---------------------------------------------------------------------------
//  SRecord::Parse() --  Construct the binary buffer from EXO image
//---------------------------------------------------------------------------

void SRecord::Parse()
{
    if (!Image.FileExists(FFileName))
        {
		Error("Exo file not found!");
        return;
        }

    FSize = 0;
    Image.Clear();
    std::fill(Buffer.begin(), Buffer.end(), 0xff);
    Image.LoadFromFile(FFileName);

    int OldPercent = -1;

    int tally = 0;
    const int Total = Image.Count();
    for (List::iterator i = Image.begin(); i != Image.end(); ++i)
        {
        int Percent = (tally++ * 100) / Total;
        if (Percent != OldPercent)
            {
            OldPercent = Percent;
            Progress(Percent);
            }

        Line = *i;
        if (Line.size() != 0)                   // Skip a 'blank' line
            ParseLine();
        }

    // Resize vector to hold actual number of elements
    Size(FSize);

    Progress(100);
    Complete(0);
}

//---------------------------------------------------------------------------
//  SRecord::ParseLine() --  Parse contents of specified S-record
//---------------------------------------------------------------------------

void SRecord::ParseLine()
{
    Cursor = 0;
    if (Line[Cursor++] != 'S')
        throw CIISRecordFileFormatException();  //("Line is not an S-record!");

    int type = Digit(1);

    // Residual line length in characters, excluding checksum
    int length = (Digit(2) - 1) * 2;

    switch (type)
        {
        case 0:     Comment = Line.substr(Cursor); return;
        case 1:     Address = Digit(4); ParseData(length - 4); break;
        case 2:     Address = Digit(6); ParseData(length - 6); break;
        case 3:     Address = Digit(8); ParseData(length - 8); break;
        case 7:     Address = Digit(4); break;
        case 8:     Address = Digit(6); break;
        case 9:     Address = Digit(8); break;
        default:    break;
        }
}

//---------------------------------------------------------------------------
//  SRecord::Digit() --  Extract N-character digit from S-record line
//---------------------------------------------------------------------------

int SRecord::Digit(int Width)
{
    std::istringstream  is;
    is.str(Line.substr(Cursor, Width));
    Cursor += Width;

    int rtnVal;
    is >> std::hex >> rtnVal;
    return rtnVal;
}

//---------------------------------------------------------------------------
//  SRecord::ParseData() --  Extract N-byte data field from S-record line
//---------------------------------------------------------------------------

void SRecord::ParseData(int Width)
{
    int Count = Width >> 1;

    const std::string sub = Line.substr(Cursor, Width);
    const char * src = sub.data();
    char * dst = reinterpret_cast<char*>(&Buffer[Address]);
    if (!HexToBin(src, dst, Count))
        throw CIISRecordException();  //("Parse error!");

    Address += Count;

    FSize = std::max(Address, static_cast<int>(FSize));
}

//---------------------------------------------------------------------------
//  SRecord::GetChecksum() --  Calculate 32-bit checksum
//---------------------------------------------------------------------------

int SRecord::GetChecksum() const
{
    int Result = 0;
    int Elements = FSize / sizeof(int);
    const int * p = reinterpret_cast < const int * > (&Buffer[0]);
    for (int i = 0; i < Elements; ++i)
        Result += *p++;

    return Result;
}

}  // namespace Innovative

