//---------------------------------------------------------------------------
// Sample_Mb.cpp
//  Acquired data sample container class
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <Sample_Mb.h>
#include <C3xFloat_Mb.h>
#include <Codec_Mb.h>

using namespace std;

namespace Innovative
{

static const double Pi = 4.0f * atan(1.0f);
static const double DegreesPerRadian = 180.0f/Pi;

//------------------------------------------------------------------------------
//   Subsample::operator < () -- Compare Subsamples
//------------------------------------------------------------------------------

bool Subsample::operator < (const Subsample & rhs) const
{
    bool result;

	switch (FType)
		{
		default:
		case sALaw:
		case sMuLaw:
		case sInt:
            result = Int() < rhs.Int();
            break;
		case sTiFloat:
		case sFloat:
            result = Float() < rhs.Float();
            break;
		}

    return result;
}

//------------------------------------------------------------------------------
//   Subsample::operator >= () -- Compare Subsamples
//------------------------------------------------------------------------------

bool Subsample::operator >= (const Subsample & rhs) const
{
    return !(*this < rhs);
}

//------------------------------------------------------------------------------
//   Subsample::operator > () -- Compare Subsamples
//------------------------------------------------------------------------------

bool Subsample::operator > (const Subsample & rhs) const
{
    return (*this >= rhs) && (*this != rhs);
}

//------------------------------------------------------------------------------
//   Subsample::operator - () -- Difference between Subsamples
//------------------------------------------------------------------------------

Subsample Subsample::operator- (const Subsample & rhs)
{
    Subsample temp;

	switch (FType)
		{
		default:
		case sALaw:
		case sMuLaw:
		case sInt:
            temp.Int(Int()-rhs.Int());
            break;
		case sTiFloat:
		case sFloat:
            temp.Float(Float()-rhs.Float());
            break;
		}

    return temp;
}

//------------------------------------------------------------------------------
//   Subsample::operator -- () -- Decrement Subsample
//------------------------------------------------------------------------------

Subsample Subsample::operator --()
{
	Subsample temp;

	switch (FType)
		{
		default:
		case sALaw:
		case sMuLaw:
		case sInt:
            temp.Int(Int()-1);
            break;
		case sTiFloat:
		case sFloat:
            temp.Float(Float()-1.0f);
            break;
		}

	return temp;
}

//------------------------------------------------------------------------------
//  Subsample::operator !=  Inequality test
//------------------------------------------------------------------------------

bool Subsample::operator !=(const Subsample & rhs) const
{
    return !(*this==rhs);
}

//------------------------------------------------------------------------------
//  Subsample::operator ==  Equality test
//------------------------------------------------------------------------------

bool Subsample::operator ==(const Subsample & rhs) const
{
    bool result;

	switch (FType)
		{
		default:
		case sALaw:
		case sMuLaw:
		case sInt:
            result = Int() == rhs.Int();
		case sTiFloat:
		case sFloat:
            result = Float() == rhs.Float();
		}

    return result;        
}

//---------------------------------------------------------------------------
//  Subsample::asString() --
//---------------------------------------------------------------------------

char * Subsample::asString()
{
    return FValue.Byte;
}

//---------------------------------------------------------------------------
// Subsample::Int() -- Return sample as an integer
//---------------------------------------------------------------------------

int Subsample::Int() const
{
	switch (FType)
		{
		case sInt:      return FValue.Int;
		case sFloat:    return (int)FValue.Float;
		case sTiFloat:
			{  TIFloat Value(FValue.Int);  return (int)Value.asIEEE();  }
		case sMuLaw:
			{  Codec Value((unsigned char)FValue.Int, Codec::uLaw);
			   return Value.asInt();  }
		case sALaw:
			{  Codec Value((unsigned char)FValue.Int, Codec::aLaw);
			   return Value.asInt();  }
		default:        return 0;
		}
}

//---------------------------------------------------------------------------
// Subsample::Int() -- Store as integer sample
//---------------------------------------------------------------------------

void Subsample::Int(int i)
{
	switch (FType)
		{
		case sInt:      FValue.Int = i;                             break;
		case sFloat:    FValue.Float = (float)i;                    break;
		case sTiFloat:
			{  TIFloat Value((float)i);
			   FValue.Int = Value.asTI();  }                        break;
		case sMuLaw:
			{  Codec Value(i);  FValue.Int = Value.asMuLaw();  }  break;
		case sALaw:
			{  Codec Value(i);  FValue.Int = Value.asALaw();  }   break;
		default:                                                    break;
		}
}

//---------------------------------------------------------------------------
// Subsample::Extended() -- Return value as a double
//      If FType is unsigned int, correct sign of result
//---------------------------------------------------------------------------

double Subsample::Extended(bool unipolar) const
{
	switch (FType)
		{
		case sFloat:    return FValue.Float;
		case sInt:
            {
            if (unipolar)
                return static_cast<unsigned int>(FValue.Int);
            else
                return FValue.Int;
            }
		case sTiFloat:
			{  TIFloat Value(FValue.Int);  return Value.asIEEE();  }
		case sMuLaw:
			{  Codec Value((unsigned char)FValue.Int, Codec::uLaw);
			   return Value.asInt();  }
		case sALaw:
			{  Codec Value((unsigned char)FValue.Int, Codec::aLaw);
			   return Value.asInt();  }
		default:        return 0.0;
		}
}

//---------------------------------------------------------------------------
// Subsample::Double() -- Return value as a double
//---------------------------------------------------------------------------

double Subsample::Double() const
{
    return Extended(false);
}

//---------------------------------------------------------------------------
// Subsample::Double() -- Set value as a double
//---------------------------------------------------------------------------

void Subsample::Double(double f)
{
	switch (FType)
		{
		case sFloat:    FValue.Float = (float)f;                        break;
		case sInt:      FValue.Int = (int)f;                            break;
		case sTiFloat:
			{  TIFloat Value((float)f);  FValue.Int = Value.asTI();  }  break;
		case sMuLaw:
			{  Codec Value((int)f);  FValue.Int = Value.asMuLaw();  } break;
		case sALaw:
			{  Codec Value((int)f);  FValue.Int = Value.asALaw();  }  break;
		default:                                                        break;
		}
}

//---------------------------------------------------------------------------
// Subsample::Float() -- Return value as a float
//---------------------------------------------------------------------------

float Subsample::Float() const
{
	return static_cast<float>(Double());
}

//---------------------------------------------------------------------------
// Subsample::Float() -- Set value as a float
//---------------------------------------------------------------------------

void Subsample::Float(float d)
{
	Double(d);
}

//---------------------------------------------------------------------------
// Subsample::Type() --
//---------------------------------------------------------------------------

void Subsample::Type(SampleType value)
{
	FType = value;
}

//---------------------------------------------------------------------------
// Subsample::Type() --
//---------------------------------------------------------------------------

Subsample::SampleType Subsample::Type() const
{
	return FType;
}

//---------------------------------------------------------------------------
// Subsample::SizeInBytes() --
//---------------------------------------------------------------------------

size_t Subsample::SizeInBytes(int bits) const
{
    size_t result;

	switch (FType)
		{
		case sInt:
			result = (bits<17) ? (bits + 7)/8 : 4;
			break;
		case sFloat:
		case sTiFloat:
			result = 4;
			break;
		case sALaw:
		case sMuLaw:
		default:
			result = 1;
			break;
		}

    return result;
}

//========================================================================
//  CLASS Sample -- Complex data sample container class
//========================================================================

//---------------------------------------------------------------------------
// Sample::Sample() -- Default ctor
//---------------------------------------------------------------------------

Sample::Sample()
    : Real(0), Imag(0)
{
}

//---------------------------------------------------------------------------
// Sample::Sample() -- Ctor
//---------------------------------------------------------------------------

Sample::Sample(char * p, int bytes, bool complex)
: Real(p, bytes), Imag(p+bytes, complex ? bytes : 0)
{
}

//---------------------------------------------------------------------------
// Sample::Sample() -- Ctor
//---------------------------------------------------------------------------

Sample::Sample(int r, int i)
    : Real(r), Imag(i)
{
}

//---------------------------------------------------------------------------
// Sample::Sample() -- Ctor
//---------------------------------------------------------------------------

Sample::Sample(double r, double i)
    : Real(r), Imag(i)
{
}

//---------------------------------------------------------------------------
// Sample::Sample() -- Ctor
//---------------------------------------------------------------------------

Sample::Sample(const Sample & rhs)
    : Real(rhs.Real), Imag(rhs.Imag)
{
}

//---------------------------------------------------------------------------
// Sample::operator =() -- Ctor
//---------------------------------------------------------------------------

Sample & Sample::operator =(const Sample & rhs)
{
    if (this != &rhs) // protect against invalid self-assignment
        {
        Real = rhs.Real;
        Imag = rhs.Imag;
        }

    // by convention, always return *this
    return *this;
}

//---------------------------------------------------------------------------
// Sample::operator =() -- Ctor
//---------------------------------------------------------------------------

Sample & Sample::operator =(const double & rhs)
{
    Real.Double(rhs);
    Imag.Double(0.0f);

    // by convention, always return *this
    return *this;
}

//---------------------------------------------------------------------------
// Sample::operator =() -- Ctor
//---------------------------------------------------------------------------

Sample & Sample::operator =(const int & rhs)
{
    Real.Int(rhs);
    Imag.Int(0);

    // by convention, always return *this
    return *this;
}

//------------------------------------------------------------------------------
//   Sample::operator < () -- Compare Samples
//------------------------------------------------------------------------------

bool Sample::operator < (const Sample & rhs) const
{
    return (Real < rhs.Real);// && (Imag < rhs.Imag);
}

//------------------------------------------------------------------------------
//   Sample::operator >= () -- Compare Subsamples
//------------------------------------------------------------------------------

bool Sample::operator >= (const Sample & rhs) const
{
    return (Real >= rhs.Real);// && (Imag >= rhs.Imag);
}

//------------------------------------------------------------------------------
//  Sample::operator >  Compare samples
//------------------------------------------------------------------------------

bool Sample::operator > (const Sample & rhs) const
{
    return (Real > rhs.Real);// && (Imag > rhs.Imag);
}

//------------------------------------------------------------------------------
//   Sample::operator - () -- Difference between Samples
//------------------------------------------------------------------------------

Sample Sample::operator- (const Sample & rhs)
{
    Sample temp;

    temp.Real = Real - rhs.Real;
    temp.Imag = Imag - rhs.Imag;

    return temp;
}

//------------------------------------------------------------------------------
//   Sample::operator -- () -- Decrement Samples
//------------------------------------------------------------------------------

Sample Sample::operator --()
{
	Sample temp;

	temp.Real = --Real;
	temp.Imag = --Imag;

	return temp;
}

//------------------------------------------------------------------------------
//  Sample::operator !=  Inequality test
//------------------------------------------------------------------------------

bool Sample::operator !=(const Sample & rhs) const
{
    return !(*this==rhs);
}

//------------------------------------------------------------------------------
//  Sample::operator ==  Equality test
//------------------------------------------------------------------------------

bool Sample::operator ==(const Sample & rhs) const
{
    return ((Real==rhs.Real) && (Imag==rhs.Imag));
}

//---------------------------------------------------------------------------
//  Sample::Int() --
//---------------------------------------------------------------------------

void Sample::Int(int value)
{
    Real.Int(value);
    Imag.Int(0);
}

//---------------------------------------------------------------------------
//  Sample::Int() --
//---------------------------------------------------------------------------

int Sample::Int() const
{
    return Real.Int();
}

//---------------------------------------------------------------------------
//  Sample::Float() --
//---------------------------------------------------------------------------

void Sample::Float(float value)
{
    Real.Float(value);
    Imag.Float(0.0f);
}

//---------------------------------------------------------------------------
//  Sample::Float() --
//---------------------------------------------------------------------------

float Sample::Float() const
{
    return Real.Float();
}

//---------------------------------------------------------------------------
//  Sample::Double() --
//---------------------------------------------------------------------------

void Sample::Double(double value)
{
    Real.Double(value);
    Imag.Double(0.0f);
}

//---------------------------------------------------------------------------
//  Sample::Extended() --
//---------------------------------------------------------------------------

double Sample::Extended(bool unipolar) const
{
    return Real.Extended(unipolar);
}

//---------------------------------------------------------------------------
//  Sample::Double() --
//---------------------------------------------------------------------------

double Sample::Double() const
{
    return Real.Double();
}

//---------------------------------------------------------------------------
//  Sample::Type() --
//---------------------------------------------------------------------------

void Sample::Type(SampleType value)
{
    Real.Type(static_cast<Subsample::SampleType>(value));
    Imag.Type(static_cast<Subsample::SampleType>(value));
}

//---------------------------------------------------------------------------
//  Sample::Type() --
//---------------------------------------------------------------------------

Sample::SampleType  Sample::Type() const
{
    return static_cast<SampleType>(Real.Type());
}

//---------------------------------------------------------------------------
//  Sample::Complex() --
//---------------------------------------------------------------------------

std::complex<double>  Sample::Complex() const
{
    return std::complex<double>(Real.Float(), Imag.Float());
}

//---------------------------------------------------------------------------
//  Sample::Phase() --
//---------------------------------------------------------------------------

double Sample::Phase(double err) const
{
    std::complex<double> c(Complex());

    if ( IsNan(c.real()) || IsNan(c.imag()) || (c.imag()==0.0f) )
        return err;
    else
        return arg(c);
}

//---------------------------------------------------------------------------
//  Sample::Degrees() --
//---------------------------------------------------------------------------

double Sample::Degrees(double err) const
{
    double phase = Phase(err);
    
    if (phase == err)
        return err;
    else
        return phase*DegreesPerRadian;
}

//---------------------------------------------------------------------------
//  Sample::Magnitude() --
//---------------------------------------------------------------------------

double Sample::Magnitude(double err) const
{
    std::complex<double> c(Complex());

    if ( IsNan(c.real()) || IsNan(c.imag()) )
        return err;
    else
        return abs(c);
}

//------------------------------------------------------------------------------
//  Sample::operator []() --  Read from Subsample
//------------------------------------------------------------------------------

Subsample Sample::operator [] (unsigned int idx) const
{
    return idx ? Imag : Real;
}

//------------------------------------------------------------------------------
//  Sample::operator []() --  Write to Subsample
//------------------------------------------------------------------------------

Subsample & Sample::operator [] (unsigned int idx)
{
    return idx ? Imag : Real;
}

}  // namespace



