//
// Si5338_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2010 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <Si5338_Mb.h>
#include <limits>
#include <sstream>
#include <cmath>
#include <HardwareRegister_Mb.h>
#include <SystemSupport_Mb.h>

namespace Innovative
{

//=============================================================================
//  CLASS Si5338 --  Programmable Quad Clock Generator Chip Support
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class Si5338
//-----------------------------------------------------------------------------

const unsigned int  TotalReferences  = 4;  // IN1&2=0, IN3=1, IN4=2, IN5&6=3 
const unsigned int  TotalOutputs     = 4;
const unsigned int  TotalRegisters   = 256;

const unsigned int  ReadRegAddr_Min  = 0;
const unsigned int  ReadRegAddr_Max  = 256;

const unsigned int  WriteRegAddr_Min_0  = 28;
const unsigned int  WriteRegAddr_Max_0  = 43;
const unsigned int  WriteRegAddr_Min_1  = 45;
const unsigned int  WriteRegAddr_Max_1  = 96;
const unsigned int  WriteRegAddr_Min_2  = 97;
const unsigned int  WriteRegAddr_Max_2  = 145;
const unsigned int  WriteRegAddr_Min_3  = 152;
const unsigned int  WriteRegAddr_Max_3  = 218;

//  Register 230 handled Specially

const double Si5338_VcoLow  = 2.2e9;
const double Si5338_VcoOpt  = 2.6e9;  //2.5e9;     // match the utility for now
const double Si5338_VcoHigh = 2.8e9;

//-----------------------------------------------------------------------------
//  constructor for class Si5338
//-----------------------------------------------------------------------------

Si5338::Si5338(I2CDevice & device)
	:  PllTimebaseDevice(TotalReferences, TotalOutputs),
       Device(device),
       FReferenceSource(ref3SingleEnded),
       RegisterImage(TotalRegisters),
       FPfd(0.0),  FVco(0.0),
       DenomValues(TotalOutputs+1),         // use last for N
       MustBeInteger(TotalOutputs+1),       // use last for N
       CurrentDividerInfo(TotalOutputs+1),  // use last for N
       Debug(false)  // Shant no likey
{
    InitChipInformation();

    for (unsigned int i=0; i<DenomValues.size(); i++)
        DenomValues[i] = 100;
    for (unsigned int i=0; i<MustBeInteger.size(); i++)
        MustBeInteger[i] = false;
}

//-----------------------------------------------------------------------------
//  Si5338::Init()  --  Init setup
//-----------------------------------------------------------------------------

void  Si5338::Init()
{
    ReadDataRegisters();   // Read initial state from chip
	SettingsInit();        //  common INIT settings
    OnInit();              //  Derived Class settings
    Update();              // Update state of the device based on sources
    InitiateClockUpdate();
}

//-----------------------------------------------------------------------------
//  Si5338::SettingsInit()  --  Initial Settings Loading
//-----------------------------------------------------------------------------
//   Note board specific implementations can override these in OnInit

void Si5338::SettingsInit()
{
    SetPllEnable(true);
    
    SetOutDividerSource(0, diMsDefault);  // Use M0
    SetOutDividerSource(1, diMsDefault);  // Use M1
    SetOutDividerSource(2, diMsDefault);  // Use M2
    SetOutDividerSource(3, diMsDefault);  // Use M3

    OutputDivider(0, 1);   // MUST be divide by 1 unless rate < 5 MHz
	OutputDivider(1, 1);
    OutputDivider(2, 1);
    OutputDivider(3, 1);
}

//-----------------------------------------------------------------------------
//  Si5338::Reapply()  --  Reapply settings to hardware from saved state
//-----------------------------------------------------------------------------

void  Si5338::Reapply()
{
    OnReapply();
    Update();    // Update state of the base device based on sources
    InitiateClockUpdate();
}

//-----------------------------------------------------------------------------
//  Si5338::InitiateClockUpdate()  --  Clock Settings Application Sequence
//-----------------------------------------------------------------------------

void  Si5338::InitiateClockUpdate()
{
    //  Disable All Outputs
    MasterOutputEnabled(false);
    //  Pause LOL
    DisableLOL(true);
    //  Apply Registers
    WriteDataRegisters();
    //  Validate Input Clock Status
    CheckPllClockIn();
    //  Configure PLL for locking
    FCALOverride(false);
    //  Initiate Locking of PLL
    SoftReset();
    //  Restart LOL
    DisableLOL(false);
    //  Wait 25 ms
    Sleep(25);
    //  Confirm PLL Lock Status
    CheckPllLockStatus();
    //  Copy FCAL to active Registers
    CopyFCAL();
    //  Set PLL to use FCAL values
    FCALOverride(true);
    //  Enable Outputs
    MasterOutputEnabled(true);
}

//-----------------------------------------------------------------------------
//  Si5338::VcoMin() - Tuning Range
//-----------------------------------------------------------------------------

double Si5338::VcoMin()
{
    return Si5338_VcoLow;
}

//-----------------------------------------------------------------------------
//  Si5338::VcoMax() - Tuning Range
//-----------------------------------------------------------------------------

double Si5338::VcoMax()
{
    return Si5338_VcoHigh;
}

//-----------------------------------------------------------------------------
//  Si5338::IsPllLocked() - Return Lock Status
//-----------------------------------------------------------------------------

bool  Si5338::IsPllLocked()
{
    //  Read Lock Status Register & extract PLL Lock Bit
    RegisterImage[218] = Port(218);
    PtrBitManipulator Reg218(&RegisterImage[218]);
    bool state = Reg218.Bit(4);
    return !state;    // bit on = not locked
}

//-----------------------------------------------------------------------------
//  Si5338::IsRefInPresent() - Return Ref Presence Status
//-----------------------------------------------------------------------------

bool  Si5338::IsRefInPresent()
{
    //  Read Lock Status Register & extract CLKIN Bit
    RegisterImage[218] = Port(218);
    PtrBitManipulator Reg218(&RegisterImage[218]);
    bool state = Reg218.Bit(2);
    return !state;   // bit on = not present
}

//-----------------------------------------------------------------------------
//  Si5338::IsFbClkPresent() - Return FB Clock Presence Status
//-----------------------------------------------------------------------------

bool  Si5338::IsFbClkPresent()
{
    //  Read Lock Status Register & extract FB Lock Bit
    RegisterImage[218] = Port(218);
    PtrBitManipulator Reg218(&RegisterImage[218]);
    bool state = Reg218.Bit(3);
    return !state;   // bit on = not present
}

//-----------------------------------------------------------------------------
//  Si5338::FrequencyModeUpdate()  --  Update using Frequency as baseline
//-----------------------------------------------------------------------------

void Si5338::FrequencyModeUpdate()
{
    if (Reference() == 0 || Frequency() == 0)
        return;
    Debug.Log("Si5338::FrequencyModeUpdate()");
    //
    //  Calculate true Reference Frequency
    if ( CurrentRDiv==DivFb )
        {
        unsigned int divide_by = GetFeedbackDivider();
        FPfd = Reference() / divide_by;
        }
    else if ( CurrentRDiv==DivRef )
        {
        unsigned int divide_by = GetReferenceDivider();
        FPfd = Reference() / divide_by;
        }
    else //( CurrentRDiv==Normal )
        {
        FPfd = Reference();
        }
    Debug.Log("Si5338::Reference() = ", Reference());
    Debug.Log("Si5338::FPfd = ", FPfd);
    //
    //  So we know our reference rate. Calculate N to get to Si5338_VcoOpt
    double N_trial = Si5338_VcoOpt / FPfd;
    if (ForceIntegerN()==true)
        {
        N_trial = std::floor(N_trial+0.5);  // round to nearest int
        }
    //
    //  Convert to integer representation
    DividerParams N_rep;
    N_rep.Int         = static_cast<unsigned int>(N_trial);                         // integer part
    N_rep.Numerator   = static_cast<unsigned int>((N_trial-N_rep.Int)*DenomForN()); // Numerator
    N_rep.Denominator = DenomForN();                      // Denominator
    Debug.Log("Si5338::N_trial = ", N_trial);
    Debug.Log("Si5338::N Int   = ", (int)N_rep.Int);
    Debug.Log("Si5338::N Num   = ", (int)N_rep.Numerator);
    Debug.Log("Si5338::N Denom = ", (int)N_rep.Denominator);
    //
    //  Save results
    CurrentDividerInfo[CurrentDividerInfo.size()-1] = N_rep;
    //
    //  Calculate actual Vco Rate
    FVco = FPfd*(N_rep.Int + (N_rep.Numerator/N_rep.Denominator));
    Debug.Log("Si5338::FVco = ", FVco);
    //
    //  For each active output
    for (unsigned int output=0; output<TotalOutputs; output++)
        {
        Debug.Log("Si5338::Calculating M for output ", (int)output);
        if (OutputInUse(output)==false)
            continue;
        //
        //  Calculate M
        double M_trial = FVco / OutputFrequency(output);
        if (ForceIntegerM(output)==true)
            {
            M_trial = std::floor(M_trial+0.5);  // round to nearest int
            }
        //
        //  Convert to integer representation
        DividerParams M_rep;
        M_rep.Int         = static_cast<unsigned int>(M_trial);                                // integer part
        M_rep.Numerator   = static_cast<unsigned int>((M_trial-M_rep.Int)*DenomForM(output));  // Numerator
        M_rep.Denominator = DenomForM(output);                      // Denominator

        Debug.Log("Si5338::Out Freq = ", OutputFrequency(output));
        Debug.Log("Si5338::M_trial = ", M_trial);
        Debug.Log("Si5338::M Int   = ", (int)M_rep.Int);
        Debug.Log("Si5338::M Num   = ", (int)M_rep.Numerator);
        Debug.Log("Si5338::M Denom = ", (int)M_rep.Denominator);
        //
        //  Save results
        CurrentDividerInfo[output] = M_rep;
        }
    //
    //  Convert N ratio to "output" values
    DividerParams N_out = ConvertToOutputRep(CurrentDividerInfo[CurrentDividerInfo.size()-1]);
    SetN(N_out.Int, N_out.Numerator, N_out.Denominator);
    //
    //  Convert M ratio to "output" values
    for (unsigned int output=0; output<TotalOutputs; output++)
        {
        if (OutputInUse(output)==false)
            continue;
        DividerParams M_out = ConvertToOutputRep(CurrentDividerInfo[output]);
        SetM(output, M_out.Int, M_out.Numerator, M_out.Denominator);
        }
    //
    //  Configure PLL Parameters
    //  ...KPhi
    SetKPhi( FPfd, FVco );
    //  ...Avar, RSel, BWSel
    SetAvarRSelBWSel( FPfd, FVco );
    //  ...SlewCal
    SetSlewCal( FVco );
    //  ...PiCal
    SetPiCal();

    //
    //  Set up the output dividers properly
//    SetOutputDividers(VcoFreq, OutDiv);
    Debug.Log("End of Si5338::FrequencyModeUpdate()");
}

//-----------------------------------------------------------------------------
//  Si5338::DividerModeUpdate() -- Update state of dividers
//-----------------------------------------------------------------------------
//  In this mode, the reference is sent to the outputs, with possible division

void  Si5338::DividerModeUpdate()
{
    if (Reference() == 0 || Frequency() == 0)
        return;
//
//    double div = 1.0;
//    double fraction = modf(Reference()/Frequency(), &div);
//    div += (fraction > 0.5) ? 1.0 : 0.0;
//    unsigned int OutDiv = static_cast<unsigned int>(div);
//    OutDiv = std::max<unsigned int>(OutDiv, 1);
//
//    SetOutputDividers(Reference(), OutDiv);
}

//-----------------------------------------------------------------------------
//  Si5338::ConvertToOutputRep()  --  Convert fraction to output version
//-----------------------------------------------------------------------------

Si5338::DividerParams  Si5338::ConvertToOutputRep(const DividerParams & in)
{
   DividerParams out;
   double glob = (((double)in.Int*in.Denominator + in.Numerator)*128.0);
   out.Int = static_cast<unsigned int>(std::floor(glob/in.Denominator - 512.0));
   out.Numerator =  ((unsigned int)glob) % in.Denominator;
   out.Denominator = in.Denominator;

   return out;
}

//-----------------------------------------------------------------------------
//  Si5338::SetKPhi()  --  PLL Settings Cfg
//-----------------------------------------------------------------------------

void  Si5338::SetKPhi(double Pfd_Rate, double Vco_Rate)
{
    PtrBitManipulator Reg(&RegisterImage[48]);

    double K(0.0);
    if (Pfd_Rate>=15.0e6)
        K = 925.0;
    else if (Pfd_Rate>=8.0e6)
        K = 325.0;
    else //if (Pfd_Rate>=5.0e6)
        K = 185.0;

    double Q(0.0);
    if (Vco_Rate>=2.425e9)
        Q = 3.0;
    else //if (Vco_Rate<2.425e9)
        Q = 4.0;

    double PfdMhz = Pfd_Rate/1e6;
    double VcoMhz = Vco_Rate/1e6;
    unsigned int KPhi = static_cast<unsigned int>(std::floor(0.5 +
                            (K / (Q*533.0)) * (VcoMhz/PfdMhz) * std::pow(2500.0/VcoMhz, 3)));
    Debug.Log("Si5338::Untrimmed KPhi = ", (int)KPhi);

    //  Clip KPhi
    if (KPhi<1)
        KPhi = 1;
    if (KPhi>127)
        KPhi = 127;

    Debug.Log("Si5338::KPhi = ", (int)KPhi);
    Reg.Field(0, 7, KPhi);
}

//-----------------------------------------------------------------------------
//  Si5338::SetAvarRSelBWSel()  --  PLL Settings Cfg
//-----------------------------------------------------------------------------

void  Si5338::SetAvarRSelBWSel(double Pfd_Rate, double Vco_Rate)
{
    PtrBitManipulator Reg(&RegisterImage[49]);

    unsigned int RSel(0);
    unsigned int BWSel(0);
    if (Pfd_Rate>=15.0e6)
        {
        RSel = 0x0;
        BWSel = 0x0;
        }
    else if (Pfd_Rate>=8.0e6)
        {
        RSel = 0x1;
        BWSel = 0x1;
        }
    else //if (Pfd_Rate>=5.0e6)
        {
        RSel = 0x3;
        BWSel = 0x2;
        }

    unsigned int Avar(0);
    if (Vco_Rate>=2.425e9)
        Avar = 0x0;
    else //if (Vco_Rate<2.425e9)
        Avar = 0x1;

    Debug.Log("Si5338::BWSel = ", (int)BWSel);
    Debug.Log("Si5338::RSel  = ", (int)RSel);
    Debug.Log("Si5338::Avar  = ", (int)Avar);

    Reg.Field(0, 2, BWSel);
    Reg.Field(2, 2, RSel);
    Reg.Field(4, 3, Avar);
}

//-----------------------------------------------------------------------------
//  Si5338::SetSlewCal()  -- PLL Settings Cfg
//-----------------------------------------------------------------------------

void  Si5338::SetSlewCal(double Vco_Rate)
{
    PtrBitManipulator Reg(&RegisterImage[50]);

    unsigned int SlewCal;
    double VcoMhz = Vco_Rate/1e6;

    SlewCal = static_cast<unsigned int>( -6.67*VcoMhz/1000.0 + 20.67);

    Debug.Log("Si5338::SlewCal = ", (int)SlewCal);

    Reg.Field(0, 6, SlewCal);
}

//-----------------------------------------------------------------------------
//  Si5338::SetPiCal()  -- PLL Settings Cfg
//-----------------------------------------------------------------------------

void  Si5338::SetPiCal()
{
    PtrBitManipulator Reg(&RegisterImage[51]);

    unsigned int PiCal;
    PiCal = 0x7;

    Debug.Log("Si5338::PiCal = ", (int)PiCal);

    Reg.Field(0, 3, PiCal);
}

//-----------------------------------------------------------------------------
//  Si5338::Port()  --  Write to chip data port
//-----------------------------------------------------------------------------

void Si5338::Port(unsigned char reg, unsigned char value)
{
	std::stringstream msg;
	msg << "Si5338::Write 0x" << std::hex << static_cast<int>(value)
		<< " to reg " << std::dec << static_cast<int>(reg);
	Debug.Log(msg.str());
    //
    //  Write using I2C interface
    Device.Reg(reg, value);
}

//-----------------------------------------------------------------------------
//  Si5338::Port()  --  Read from chip data port
//-----------------------------------------------------------------------------

unsigned char Si5338::Port(unsigned char reg)
{
    //
    //  Read using I2C interface
    int value = Device.Reg(reg);

	std::stringstream msg;
	msg << "Si5338::Read 0x" << std::hex << value
		<< " from reg " << std::dec << static_cast<int>(reg);
	Debug.Log(msg.str());

	return static_cast<unsigned char>(value);
}

//-----------------------------------------------------------------------------
//  Si5338::ReadDataRegisters()  --  Read all data registers into Image vector
//-----------------------------------------------------------------------------

void Si5338::ReadDataRegisters()
{
    for (unsigned int i=ReadRegAddr_Min; i<ReadRegAddr_Max; i++)
        {
        RegisterImage[i] = Port(i);
        }
    //  Apply "magic" field updates
    PtrBitManipulator Reg42(&RegisterImage[42]);
    Reg42.Bit(5, true);
    PtrBitManipulator Reg47(&RegisterImage[47]);
    Reg47.Field(2, 6, 0x5);
    PtrBitManipulator Reg106(&RegisterImage[106]);
    Reg106.Bit(7, true);
    PtrBitManipulator Reg116(&RegisterImage[116]);
    Reg116.Bit(7, true);
    PtrBitManipulator Reg241(&RegisterImage[241]);
    Reg241.Field(0, 7, 0x65);

}

//-----------------------------------------------------------------------------
//  Si5338::WriteDataRegisters()  --  Write all data registers from Image vector
//-----------------------------------------------------------------------------

void Si5338::WriteDataRegisters()
{
    for (unsigned int i=WriteRegAddr_Min_0; i<WriteRegAddr_Max_0; i++)
        {
        Port(i, RegisterImage[i]);
        }
    for (unsigned int i=WriteRegAddr_Min_1; i<WriteRegAddr_Max_1; i++)
        {
        Port(i, RegisterImage[i]);
        }
    for (unsigned int i=WriteRegAddr_Min_2; i<WriteRegAddr_Max_2; i++)
        {
        Port(i, RegisterImage[i]);
        }
    for (unsigned int i=WriteRegAddr_Min_3; i<WriteRegAddr_Max_3; i++)
        {
        Port(i, RegisterImage[i]);
        }
    //
    //  Update Register 230
    Port(230, RegisterImage[230]);
}

//-----------------------------------------------------------------------------
//  Si5338::SetInputMuxConfig()  -- Configure Select A block
//-----------------------------------------------------------------------------

void  Si5338::SetInputMuxConfig(ReferenceASettings RefA, float xtal_rate_mhz)
{
    unsigned int f_xtal;
    if (RefA != ra12Xtal)
        f_xtal = 0;
    else if (xtal_rate_mhz < 11.0)
        f_xtal = rar08t11;
    else if (xtal_rate_mhz < 19.0)
        f_xtal = rar11t19;
    else if (xtal_rate_mhz < 26.0)
        f_xtal = rar19t26;
    else if (xtal_rate_mhz < 30.0)
        f_xtal = rar26t30;
    else
        f_xtal = 0;

    PtrBitManipulator RegL(&RegisterImage[28]);
    PtrBitManipulator RegH(&RegisterImage[29]);

    RegL.Field(0, 2, f_xtal);
    RegL.Field(2, 3, RefA>>2);
    RegH.Field(3, 2, RefA);

}

//-----------------------------------------------------------------------------
//  Si5338::SetFeedbackMuxConfig()  -- Configure Select B block
//-----------------------------------------------------------------------------

void  Si5338::SetFeedbackMuxConfig(FeedbackBSettings FbB)
{

    PtrBitManipulator RegL(&RegisterImage[28]);
    PtrBitManipulator RegH(&RegisterImage[30]);

    RegL.Bit(5, (FbB>>2) ? true : false);
    RegH.Field(3, 2, FbB);

}

//-----------------------------------------------------------------------------
//  Si5338::SetPllEnable()  -- Enable PLL
//-----------------------------------------------------------------------------

void  Si5338::SetPllEnable(bool isEnabled)
{
    PtrBitManipulator Reg(&RegisterImage[50]);

    unsigned int enable_bits = (isEnabled) ? 0x3 : 0x00 ;
    Reg.Field(6, 2, enable_bits);
}

//-----------------------------------------------------------------------------
//  Si5338::SetPllReferenceInput()  -- Set Select C Input
//-----------------------------------------------------------------------------

void  Si5338::SetPllReferenceInput(PllRefInputSettings settings)
{

    PtrBitManipulator Reg(&RegisterImage[29]);

    //  Save current divider option for use in frequency calc.
    switch (settings)
        {
        default:           CurrentRDiv = Normal;  break;
        case prDivRefClk:  CurrentRDiv = DivRef;  break;
        case prDivFbClk:   CurrentRDiv = DivFb;   break;
        }
    Reg.Field(5, 3, settings);

}

//-----------------------------------------------------------------------------
//  Si5338::SetReferenceDivider()  -- Set Select C Ref Divider
//-----------------------------------------------------------------------------

void  Si5338::SetReferenceDivider(unsigned int divide_by)
{
    PtrBitManipulator Reg(&RegisterImage[29]);

    Reg.Field(0, 3, ToDividerInput(divide_by));

}

//-----------------------------------------------------------------------------
//  Si5338::SetFeedbackDivider()  -- Set Select C FB Divider
//-----------------------------------------------------------------------------

void  Si5338::SetFeedbackDivider(unsigned int divide_by)
{
    PtrBitManipulator Reg(&RegisterImage[30]);

    Reg.Field(0, 3, ToDividerInput(divide_by));

}

//-----------------------------------------------------------------------------
//  Si5338::SetFeedbackSource()  -- Set PLL Feedback Source
//-----------------------------------------------------------------------------

void  Si5338::SetFeedbackSource(bool isExternal, ExternalFeedbackSource ext)
{
    PtrBitManipulator Reg48(&RegisterImage[48]);
    PtrBitManipulator Reg30(&RegisterImage[30]);

    Reg48.Bit(7, isExternal);
    Reg30.Field(5, 3, static_cast<unsigned int>(ext));

}

//-----------------------------------------------------------------------------
//  Si5338::GetReferenceDivider()  -- Get Select C Ref Divider divide by
//-----------------------------------------------------------------------------

unsigned int  Si5338::GetReferenceDivider()
{
    PtrBitManipulator Reg(&RegisterImage[29]);

    unsigned int divider_setting = Reg.Field(0, 3);
    return FromDividerInput( divider_setting );
}

//-----------------------------------------------------------------------------
//  Si5338::GetFeedbackDivider()  -- Get Select C Ref Divider divide by
//-----------------------------------------------------------------------------

unsigned int  Si5338::GetFeedbackDivider()
{
    PtrBitManipulator Reg(&RegisterImage[30]);

    unsigned int divider_setting = Reg.Field(0, 3);
    return FromDividerInput( divider_setting );
}

//-----------------------------------------------------------------------------
//  Si5338::SetClkPowerDown()  --  Set Clk and divider power down
//-----------------------------------------------------------------------------

void  Si5338::SetClkPowerDown(unsigned int clk, bool powered_down)
{
    if (clk>=TotalOutputs)
        return;

    PtrBitManipulator RegX(&RegisterImage[31+clk]);

    RegX.Bit(0, powered_down);

}

//-----------------------------------------------------------------------------
//  Si5338::SetMultiSynthPowerDown()  --  Set multi-synth power down
//-----------------------------------------------------------------------------

void  Si5338::SetMultiSynthPowerDown(unsigned int ms_dev, bool powered_down)
{
    if (ms_dev>=TotalOutputs)
        return;

    PtrBitManipulator RegX(&RegisterImage[31+ms_dev]);

    RegX.Bit(1, powered_down);

}

//-----------------------------------------------------------------------------
//  Si5338::SetOutDividerSource()  --  Set Clk and divider power down
//-----------------------------------------------------------------------------

void  Si5338::SetOutDividerSource(unsigned int output, DividerInputSrcSettings source)
{
    if (output>=TotalOutputs)
        return;

    PtrBitManipulator RegX(&RegisterImage[31+output]);

    RegX.Field(5, 3, static_cast<unsigned int>(source));

}

//-----------------------------------------------------------------------------
//  Si5338::SetDividerHdw()  --  Update the Divider setting hardware
//-----------------------------------------------------------------------------

void Si5338::SetDividerHdw(unsigned int output, unsigned int divide_by)
{
    if (!IsValidDivider(divide_by))
        return;
    if (output < TotalOutputs)
        {
        PtrBitManipulator RegX(&RegisterImage[31+output]);

        RegX.Field(2, 3, ToDividerInput(divide_by));     
        }
}

//-----------------------------------------------------------------------------
//  Si5338::SetOutputVoltage()  --  Set Clk and divider power down
//-----------------------------------------------------------------------------

void  Si5338::SetOutputVoltage(unsigned int output, VddOutputSettings level)
{
    if (output>=TotalOutputs)
        return;

    PtrBitManipulator Reg(&RegisterImage[35]);
    unsigned int start_bit = output*2;

    Reg.Field(start_bit, 2, static_cast<unsigned int>(level));

}

//-----------------------------------------------------------------------------
//  Si5338::SetOutputConfig()  --  Update the Output Config for a channel
//-----------------------------------------------------------------------------

void Si5338::SetOutputConfig(unsigned int output, OutputCfgSettings cfg_setting,
                                               OutputInvSettings inv_setting)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator RegX(&RegisterImage[36+output]);

        RegX.Field(0, 3, static_cast<unsigned int>(cfg_setting));    // Output Cfg bits 0-2
        RegX.Field(3, 2, static_cast<unsigned int>(inv_setting));    // Output Inv bits 3-4
        }
}

//-----------------------------------------------------------------------------
//  Si5338::SetOutEnabledHdw()  --  Update the Output Enabled hardware
//-----------------------------------------------------------------------------

void Si5338::SetOutEnabledHdw(unsigned int output, bool enabled)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator Reg(&RegisterImage[230]);
		Reg.Bit(output, enabled ? false : true);
        }
}

//-----------------------------------------------------------------------------
//  Si5338::SetMasterOutEnabled()  --  Update the Output Enabled shadow
//-----------------------------------------------------------------------------

void Si5338::SetMasterOutEnabled(bool enabled)
{
    PtrBitManipulator Reg(&RegisterImage[230]);
	Reg.Bit(4, enabled ? false : true);
}

//-----------------------------------------------------------------------------
//  Si5338::SetDisableLOL()  --  Update the disable LOL bit
//-----------------------------------------------------------------------------

void Si5338::SetDisableLOL(bool dis_lol_bit)
{
    PtrBitManipulator Reg(&RegisterImage[241]);
    Reg.Bit(7, dis_lol_bit);
}

//-----------------------------------------------------------------------------
//  Si5338::SetFCALOverride()  --  Update the FCAL value usage bit
//-----------------------------------------------------------------------------

void Si5338::SetFCALOverride(bool ignore_fcal)
{
    PtrBitManipulator Reg(&RegisterImage[49]);
    Reg.Bit(7, ignore_fcal);
}

//-----------------------------------------------------------------------------
//  Si5338::SetOutputTrim()  --  Update the Output Trim setting
//-----------------------------------------------------------------------------

void Si5338::SetOutputTrim(unsigned int output, OutputTrimSettings trim_setting)
{
    switch (output)
        {
        case 0:
            {
            PtrBitManipulator Reg(&RegisterImage[40]);
            Reg.Field(0, 5, trim_setting);
            }
            break;
        case 1:
            {
            PtrBitManipulator RegL(&RegisterImage[40]);
            PtrBitManipulator RegH(&RegisterImage[41]);
            RegL.Field(5, 3, trim_setting);
            RegH.Field(0, 2, trim_setting>>3);
            }
            break;
        case 2:
            {
            PtrBitManipulator Reg(&RegisterImage[41]);
            Reg.Field(2, 5, trim_setting);
            }
            break;
        case 3:
            {
            PtrBitManipulator Reg(&RegisterImage[42]);
            Reg.Field(0, 5, trim_setting);
            }
            break;
        default:
            break;
        }
}

//-----------------------------------------------------------------------------
//  Si5338::SetOutputDisableState()  --  Update the state when disabled
//-----------------------------------------------------------------------------

void Si5338::SetOutputDisableState(unsigned int output, OutputDisableSettings dis_setting)
{
    if (output < TotalOutputs)
        {
        PtrBitManipulator RegX(&RegisterImage[110+4*output]);

        RegX.Field(6, 2, static_cast<unsigned int>(dis_setting));
        }
}

//-----------------------------------------------------------------------------
//  Si5338::SetM()  --  Write M to Image
//-----------------------------------------------------------------------------

void  Si5338::SetM(unsigned int output, unsigned int M_Int, unsigned int M_Numer,
                  unsigned int M_Denom)
{
    //  Find base registers of the three sections
    unsigned int M_IntBase   = 53 + 11*output;
    unsigned int M_NumerBase = 55 + 11*output;
    unsigned int M_DenomBase = 59 + 11*output;

    PtrBitManipulator M_IntLo(&RegisterImage[M_IntBase]);
    PtrBitManipulator M_IntMid(&RegisterImage[M_IntBase+1]);
    PtrBitManipulator M_IntHi(&RegisterImage[M_IntBase+2]);

    PtrBitManipulator M_NumerLo(&RegisterImage[M_NumerBase]);
    PtrBitManipulator M_NumerMid0(&RegisterImage[M_NumerBase+1]);
    PtrBitManipulator M_NumerMid1(&RegisterImage[M_NumerBase+2]);
    PtrBitManipulator M_NumerHi(&RegisterImage[M_NumerBase+3]);

    PtrBitManipulator M_DenomLo(&RegisterImage[M_DenomBase]);
    PtrBitManipulator M_DenomMid0(&RegisterImage[M_DenomBase+1]);
    PtrBitManipulator M_DenomMid1(&RegisterImage[M_DenomBase+2]);
    PtrBitManipulator M_DenomHi(&RegisterImage[M_DenomBase+3]);

    M_IntLo.Value(static_cast<int>(M_Int));                 // LSB
    M_IntMid.Value(static_cast<int>(M_Int >> 8));           // Bits 8-15
    M_IntHi.Field(0, 2, static_cast<int>(M_Int >> 16));     // MSB  (Bits 16-17)

    M_NumerLo.Field(2, 6, static_cast<int>(M_Numer));       // LSB
    M_NumerMid0.Value(static_cast<int>(M_Numer >> 6));      // Bits 6-13
    M_NumerMid1.Value(static_cast<int>(M_Numer >> 14));     // Bits 14-21
    M_NumerHi.Value(static_cast<int>(M_Numer >> 22));       // MSB  (Bits 22-29)

    M_DenomLo.Value(static_cast<int>(M_Denom));             // LSB
    M_DenomMid0.Value(static_cast<int>(M_Denom >> 8));      // Bits 8-15
    M_DenomMid1.Value(static_cast<int>(M_Denom >> 16));     // Bits 16-23
    M_DenomHi.Field(0, 6, static_cast<int>(M_Denom >> 24)); // MSB  (Bits 24-29)

}

//-----------------------------------------------------------------------------
//  Si5338::SetN()  --  Write N to Image
//-----------------------------------------------------------------------------

void  Si5338::SetN(unsigned int N_Int, unsigned int N_Numer, unsigned int N_Denom)
{
    //  Find base registers of the three sections
    unsigned int N_IntBase   = 97;
    unsigned int N_NumerBase = 99;
    unsigned int N_DenomBase = 103;

    PtrBitManipulator N_IntLo(&RegisterImage[N_IntBase]);
    PtrBitManipulator N_IntMid(&RegisterImage[N_IntBase+1]);
    PtrBitManipulator N_IntHi(&RegisterImage[N_IntBase+2]);

    PtrBitManipulator N_NumerLo(&RegisterImage[N_NumerBase]);
    PtrBitManipulator N_NumerMid0(&RegisterImage[N_NumerBase+1]);
    PtrBitManipulator N_NumerMid1(&RegisterImage[N_NumerBase+2]);
    PtrBitManipulator N_NumerHi(&RegisterImage[N_NumerBase+3]);

    PtrBitManipulator N_DenomLo(&RegisterImage[N_DenomBase]);
    PtrBitManipulator N_DenomMid0(&RegisterImage[N_DenomBase+1]);
    PtrBitManipulator N_DenomMid1(&RegisterImage[N_DenomBase+2]);
    PtrBitManipulator N_DenomHi(&RegisterImage[N_DenomBase+3]);

    N_IntLo.Value(static_cast<int>(N_Int));                 // LSB
    N_IntMid.Value(static_cast<int>(N_Int >> 8));           // Bits 8-15
    N_IntHi.Field(0, 2, static_cast<int>(N_Int >> 16));     // MSB  (Bits 16-17)

    N_NumerLo.Field(2, 6, static_cast<int>(N_Numer));       // LSB
    N_NumerMid0.Value(static_cast<int>(N_Numer >> 6));      // Bits 6-13
    N_NumerMid1.Value(static_cast<int>(N_Numer >> 14));     // Bits 14-21
    N_NumerHi.Value(static_cast<int>(N_Numer >> 22));       // MSB  (Bits 22-29)

    N_DenomLo.Value(static_cast<int>(N_Denom));             // LSB
    N_DenomMid0.Value(static_cast<int>(N_Denom >> 8));      // Bits 8-15
    N_DenomMid1.Value(static_cast<int>(N_Denom >> 16));     // Bits 16-23
    N_DenomHi.Field(0, 6, static_cast<int>(N_Denom >> 24)); // MSB  (Bits 24-29)

}

//-----------------------------------------------------------------------------
//  Si5338::SetReferenceSource()  --  Select reference source
//-----------------------------------------------------------------------------

void Si5338::SetReferenceSource(SiReferenceSource source)
{
    unsigned int source_no;
    switch (source)
        {
        default:
        case ref12Differential:
            source_no=0;
            SetInputMuxConfig(ra12Differential, (float)ReferenceFrequency(source_no));
            break;
        case ref3SingleEnded:
            source_no=1;
            SetInputMuxConfig(ra3SingleEnded, (float)ReferenceFrequency(source_no));
            break;
        case ref4SingleEnded:
            source_no=2;
            SetFeedbackMuxConfig(fb4SingleEnded);
            break;
        case ref56Differential:
            source_no=3;
            SetFeedbackMuxConfig(fb56Differential);
            break;
        case ref12DifferentialCrystal:
            source_no=0;
            SetInputMuxConfig(ra12Xtal, (float)ReferenceFrequency(source_no));
            break;
        }

    SetSourceSelection(source_no);  // inform base class too
}

//-----------------------------------------------------------------------------
//  Si5338::MasterOutputEnabled()  --  Disable all outputs, or enable chosen
//-----------------------------------------------------------------------------

void Si5338::MasterOutputEnabled(bool enabled)
{
    //  Apply to Shadow
    SetMasterOutEnabled(enabled);
    //  Apply to hardware
    Port(230, RegisterImage[230]);
}

//-----------------------------------------------------------------------------
//  Si5338::DisableLOL()  --  Disable LOL bit update
//-----------------------------------------------------------------------------

void Si5338::DisableLOL(bool dis_lol_bit)
{
    //  Apply to Shadow
    SetDisableLOL(dis_lol_bit);
    //  Apply to hardware
    Port(241, RegisterImage[241]);
}

//-----------------------------------------------------------------------------
//  Si5338::FCALOverride()  --  Set FCAL usage bit
//-----------------------------------------------------------------------------

void Si5338::FCALOverride(bool ignore_fcal)
{
    //  Apply to Shadow
    SetFCALOverride(ignore_fcal);
    //  Apply to hardware
    Port(49, RegisterImage[49]);
}

//-----------------------------------------------------------------------------
//  Si5338::SoftReset()  -- Apply configuration changes
//-----------------------------------------------------------------------------

void  Si5338::SoftReset()
{
    Debug.Log("Si5338::Soft Reset");
    Port(246, 0x2);
}

//-----------------------------------------------------------------------------
//  Si5338::CopyFCAL() - Copy FCAL data from chip and move to CAL registers
//-----------------------------------------------------------------------------

void  Si5338::CopyFCAL()
{
    //  Read FCAL in registers
    RegisterImage[237] = Port(237);
    RegisterImage[236] = Port(236);
    RegisterImage[235] = Port(235);

    //  Copy FCAL data
    PtrBitManipulator Reg237(&RegisterImage[237]);
    PtrBitManipulator Reg47(&RegisterImage[47]);
    Reg47.Field(0, 2, Reg237.Field(0, 2));
    Reg47.Field(2, 6, 0x5);

    RegisterImage[46] = RegisterImage[236];

    RegisterImage[45] = RegisterImage[235];

    //  Write FCAL out registers
    Port(47, RegisterImage[47]);
    Port(46, RegisterImage[46]);
    Port(45, RegisterImage[45]);
}

//-----------------------------------------------------------------------------
//  Si5338::CheckPllClockIn()  -- Test if PLL Clockin is present
//-----------------------------------------------------------------------------

void  Si5338::CheckPllClockIn()
{
    Debug.Log("Si5338::Check Pll ClockIn");
    bool ClkPresent(false);
    if (FSource==0 || FSource==1)
        ClkPresent = IsRefInPresent();
    else
        ClkPresent = IsFbClkPresent();

    Debug.Log("Si5338::Bit = ", (int)ClkPresent);
    Debug.Log("Si5338::R218 = ", RegisterImage[218]);
}

//-----------------------------------------------------------------------------
//  Si5338::CheckPllLockStatus()  -- Test if PLL lock programming worked
//-----------------------------------------------------------------------------

void  Si5338::CheckPllLockStatus()
{
    Debug.Log("Si5338::Check Pll LockStatus");
    bool lock = IsPllLocked();

    Debug.Log("Si5338::Lock = ", (int)lock);
    if (lock==false)
        Debug.Log("Si5338::Lock FAILED");
    Debug.Log("Si5338::R218 = ", RegisterImage[218]);
}

//-----------------------------------------------------------------------------
//  Si5338::InitChipInformation()  --  Set up chip information data
//-----------------------------------------------------------------------------

void Si5338::InitChipInformation()
{
    SiDividerValues.push_back(1);   SiDividerValues.push_back(2);
    SiDividerValues.push_back(4);   SiDividerValues.push_back(8);
    SiDividerValues.push_back(16);  SiDividerValues.push_back(32);

    //  Fill the conversion information structures
    //     Valid dividers is keyed by divide_by, value is Input Value
    SiValidDividers[1]  = 0x00;       SiValidDividers[2]  = 0x01;
    SiValidDividers[4]  = 0x02;       SiValidDividers[8]  = 0x03;
    SiValidDividers[16] = 0x04;       SiValidDividers[32] = 0x05;

    //  Fill the conversion information structures
    //     Inverse dividers is keyed by Input Value, value is divide_by
    SiInverseDividers[0]  = 1;       SiInverseDividers[1]  = 2;
    SiInverseDividers[2]  = 4;       SiInverseDividers[3]  = 8;
    SiInverseDividers[4]  = 16;      SiInverseDividers[5] = 32;
}

//-----------------------------------------------------------------------------
//  Si5338::IsValidDivider() -- tests if divide_by is a valid divider
//-----------------------------------------------------------------------------

bool  Si5338::IsValidDivider(unsigned int divide_by)
{
	ValidDividerMap::iterator iter = SiValidDividers.find(divide_by);
	return iter != SiValidDividers.end();
}

//-----------------------------------------------------------------------------
//  Si5338::ToDividerInput() -- Returns input value for divide_by
//-----------------------------------------------------------------------------
//  this maps divider ratios to the field value in the chip hardware

unsigned int  Si5338::ToDividerInput(unsigned int divide_by)
{
	if (IsValidDivider(divide_by))
        return SiValidDividers[divide_by];
    else
        return SiValidDividers[1];   // if invalid divider, divide by 1.
}

//-----------------------------------------------------------------------------
//  Si5338::FromDividerInput() -- Returns divide_by for input value
//-----------------------------------------------------------------------------
//  this maps divider bits in the field value in the chip hardware to divide_by result

unsigned int  Si5338::FromDividerInput(unsigned int divider_bits)
{
	InverseDividerMap::iterator iter = SiInverseDividers.find(divider_bits);

	if (iter != SiInverseDividers.end())
        return SiInverseDividers[divider_bits];
    else
        return 1;   // if invalid divider_bits, claim it is divide by 1
}

//-----------------------------------------------------------------------------
//  Si5338::BestDividerGreaterThan() -- returns next largest valid Divider
//-----------------------------------------------------------------------------

unsigned int  Si5338::BestDividerGreaterThan(unsigned int min_divide_by)
{
    unsigned int CandidateD = 0;
    for (ValidDividerArray::iterator div = SiDividerValues.begin();
             div !=  SiDividerValues.end();  ++div)
        {
        CandidateD = *div;
        if (CandidateD > min_divide_by)
            break;
        }
    return CandidateD;
}

//-----------------------------------------------------------------------------
//  Si5338::BestDividerLessThan() -- returns next smallest valid Divider
//-----------------------------------------------------------------------------

unsigned int  Si5338::BestDividerLessThan(unsigned int max_divide_by)
{
    unsigned int CandidateD = 0;
    for (ValidDividerArray::reverse_iterator div = SiDividerValues.rbegin();
             div !=  SiDividerValues.rend();  ++div)
        {
        CandidateD = *div;
        if (CandidateD < max_divide_by)
            break;
        }
    return CandidateD;
}

} // namespace Innovative

