//
// Si5338_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2010 Innovative Integration Corporation.
//  All Rights Reserved.
//

#ifndef Si5338_MbH
#define Si5338_MbH

#include <DataTypes_Mb.h>
#include <TimebaseDevice_Mb.h>
#include <I2C_Mb.h>
#include <vector>
#include <map>
#include <Trace_Mb.h>

namespace Innovative
{

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif

//=============================================================================
//  CLASS Si5338 --  Programmable Quad Clock Generator Chip Support
//=============================================================================

class Si5338 : public PllTimebaseDevice
{
public:
    enum SiReferenceSource { ref12Differential, ref3SingleEnded, ref4SingleEnded,
                             ref56Differential, ref12DifferentialCrystal};

    struct DividerParams
    {
        unsigned int Int;
        unsigned int Numerator;
        unsigned int Denominator;
    };

	Si5338(I2CDevice & device);

	virtual void        Init();
	virtual void        Reapply();

    bool  IsPllLocked();
    bool  IsRefInPresent();
    bool  IsFbClkPresent();

    //  Select the Input reference
    void  ReferenceSource(SiReferenceSource value)
             { FReferenceSource = value;  SetReferenceSource(value); }
    SiReferenceSource  ReferenceSource()
             {  return FReferenceSource;  }

    //  Conditions on frequency calculation
    void  ForceIntegerN(bool state)
        {  MustBeInteger[MustBeInteger.size()-1] = state;  }
    bool  ForceIntegerN()
		{  return MustBeInteger[MustBeInteger.size()-1] ? true : false;  }
    void  DenomForN(unsigned int denom)
        {  DenomValues[DenomValues.size()-1] = denom;  }
    unsigned int  DenomForN()
        {  return DenomValues[DenomValues.size()-1];  }

    void  ForceIntegerM(unsigned int output, bool state)
        {  MustBeInteger[output] = state;  }
    bool  ForceIntegerM(unsigned int output)
		{  return MustBeInteger[output] ? true : false;  }
    void  DenomForM(unsigned int output, unsigned int denom)
        {  DenomValues[output] = denom;  }
    unsigned int  DenomForM(unsigned int output)
        {  return DenomValues[output];  }

    //  Raw configuration accesses
	void 			Port(unsigned char reg, unsigned char value);
	unsigned char 	Port(unsigned char reg);

    void            MasterOutputEnabled(bool enabled);
    //
    //  Vcxo Parameters
    double  VcoMin();
    double  VcoMax();

    //unsigned int  ToDivider(unsigned int divider_idx);
    //unsigned int  ToDividerIdx(unsigned int divider);
    bool          IsValidDivider(unsigned int divide_by);
    unsigned int  ToDividerInput(unsigned int divide_by);
    unsigned int  FromDividerInput(unsigned int divider_bits);

    unsigned int  BestDividerGreaterThan(unsigned int min_divide_by);
    unsigned int  BestDividerLessThan(unsigned int max_divide_by);


protected:

	I2CDevice &         Device;
    SiReferenceSource   FReferenceSource;

    std::vector<int>    RegisterImage;

    virtual void    OnInit() = 0;
    virtual void    OnReapply() = 0;
    //
    //  RESERVED CONFIG FUNCTIONS (Library Only)
    //----------------------------------------------------------------------------
    enum ReferenceASettings { ra12Differential=0x0, ra12Xtal=0x16, ra3SingleEnded=0x9 };
    enum RefARateSettings  { rar26t30=0x3, rar19t26=0x2, rar11t19=0x1,
                                   rar08t11=0x0 };
    enum DividerInputSrcSettings { diFbClk=0x0, diRefClk=0x1, diP2Out=0x2,
                                   diP1Out=0x3, diXoClk=0x4, diMSZero=0x5,
                                   diMsDefault=0x6, diNoClk=0x7 };
    enum FeedbackBSettings { fb56Differential=0x0, fb4SingleEnded=0x5, fbNoClk=0x2 };
    enum VddOutputSettings { vo_3p3volts=0x0, vo_2p5volts=0x1, vo_1p8volts=0x2,
                               vo_1p5volts=0x0 };
    enum PllRefInputSettings { prRefClk=0x0, prFbClk=0x1, prDivRefClk=0x2,
                             prDivFbClk=0x3, prXoClk=0x4  };
    enum ExternalFeedbackSource { efbRefClk=0x1, efbFbClk=0x0, efbDivRefClk=0x3,
                             efbDivFbClk=0x2, efbNoClk=0x5  };

    enum OutputCfgSettings { ocLvPecl=0x04, ocLvds=0x06, ocHscl=0x7,
                             ocLvCmos_Ponly=0x01, ocLvCmos_Nonly=0x02,
                             ocLvCmos_PN=0x03   };
    enum OutputInvSettings { oiNormal=0x0, oiInvP=0x1, oiInvN=0x2, oiInvPN=0x3 };
    enum OutputTrimSettings { ot_3p3cmos=0x17, ot_2p5cmos=0x13,   ot_1p8cmos=0x15,
                              ot_1p5hstl=0x1f, ot_3p3sstl=0x04,   ot_2p5sstl=0x0d,
                              ot_1p8sstl=0x17, ot_3p3lvpecl=0x0f, ot_2p5lvpecl=0x10,
                              ot_3p3lvds=0x03, ot_2p5lvds=0x04,   ot_3p3hscl=0x07,
                              ot_2p5hscl=0x07, ot_1p8hscl=0x07  };
    enum OutputDisableSettings { odHighZ=0x0, odLow=0x1, odHi=0x2, odAlwaysOn=0x3 };

    //
    //  Internal "Setters"
    void  SetInputMuxConfig(ReferenceASettings RefA, float xtal_rate_mhz);  // rate needed if ra12Xtal used
    void  SetFeedbackMuxConfig(FeedbackBSettings FbB);
    void  SetClkPowerDown(unsigned int clk, bool powered_down);
    void  SetMultiSynthPowerDown(unsigned int ms_dev, bool powered_down);

    void  SetOutDividerSource(unsigned int output, DividerInputSrcSettings source);
    void  SetOutputVoltage(unsigned int output, VddOutputSettings level);
    void  SetOutputConfig(unsigned int output, OutputCfgSettings cfg_setting,
                                               OutputInvSettings inv_setting);
    void  SetOutputTrim(unsigned int output, OutputTrimSettings trim_setting);
    void  SetOutputDisableState(unsigned int output, OutputDisableSettings trim_setting);

    void  SetPllEnable(bool isEnabled);
    void  SetPllReferenceInput(PllRefInputSettings settings);
    void  SetReferenceDivider(unsigned int divider_setting);
    void  SetFeedbackDivider(unsigned int divider_setting);
    void  SetFeedbackSource(bool isExternal, ExternalFeedbackSource ext);

    unsigned int  GetReferenceDivider();
    unsigned int  GetFeedbackDivider();

    void  SetM(unsigned int output, unsigned int M_Int, unsigned int M_Numer,
                  unsigned int M_Denom);
    void  SetN(unsigned int N_Int, unsigned int N_Numer, unsigned int N_Denom);
    //
    //  Pll Config
    void  SetKPhi(double Pfd_Rate, double Vco_Rate);
    void  SetAvarRSelBWSel(double Pfd_Rate, double Vco_Rate);
    void  SetSlewCal(double Vco_Rate);
    void  SetPiCal();

    void  SetReferenceSource(SiReferenceSource source);
    virtual void  SetDividerHdw(unsigned int output, unsigned int divider_setting);
    virtual void  SetOutEnabledHdw(unsigned int output, bool enabled);

    void  SetMasterOutEnabled(bool enabled);
    void  SetDisableLOL(bool dis_lol_bit);
    void  SetFCALOverride(bool ignore_fcal);
    void  CheckPllClockIn();
    void  CheckPllLockStatus();
//    void  SetDividerEnable(unsigned int output, bool state);
//    //
//    //  Internal "Getters"
//    unsigned int  GetDivider(unsigned int output);
//    bool          GetDividerEnable(unsigned int output);

    virtual bool CheckDivider(unsigned int /*output*/, unsigned int divide_by)
        {  return IsValidDivider(divide_by);  }
    virtual void  GetNearestDividers(unsigned int /*output*/, float candidate, int & below, int & above)
        {
        above = BestDividerGreaterThan(static_cast<unsigned int>(candidate));
        below = BestDividerLessThan(static_cast<unsigned int>(candidate));
        }

    virtual unsigned int AdcMasterOutput()
        {  return FMasterOutput;  }
    virtual unsigned int DacMasterOutput()
        {  return FMasterOutput;  }

    virtual void  SetAdcOutputRates(double adc_rate)
        {  OutputFrequency(AdcMasterOutput(), adc_rate);  }
    virtual void  SetDacOutputRates(double /*dac_rate*/)
        {}

private:
    void            ReadDataRegisters();
    void            WriteDataRegisters();

	void            InitChipInformation();
	void            SettingsInit();
    void 			FrequencyModeUpdate();
    void            DividerModeUpdate();
    DividerParams   ConvertToOutputRep(const DividerParams & in);

	void            InitiateClockUpdate();
    void            DisableLOL(bool dis_lol_bit);
    void            SoftReset();
    void            CopyFCAL();
    void            FCALOverride(bool ignore_fcal);

    typedef std::vector<unsigned int> ValidDividerArray;
    typedef std::map<unsigned int, unsigned int> ValidDividerMap;
    typedef std::map<unsigned int, unsigned int> InverseDividerMap;

    ValidDividerMap      SiValidDividers;
    InverseDividerMap    SiInverseDividers;

    ValidDividerArray    SiDividerValues;

    //  Internal State Values
    enum  RefDividerOptions { Normal, DivRef, DivFb };

    double  FPfd;
    double  FVco;
    std::vector<unsigned int>  DenomValues;       // use last for N for all these
    std::vector<char>          MustBeInteger;
    std::vector<DividerParams> CurrentDividerInfo;

    RefDividerOptions          CurrentRDiv;

public:
    Trace			Debug;
};

#ifdef _MSC_VER
	#pragma warning(pop)
#endif

} // namespace Innovative
#endif
