//
// Si570_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2009 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <Si570_Mb.h>
#include <limits>
#include <sstream>
#include <cmath>
#include <HardwareRegister_Mb.h>

namespace Innovative
{

//=============================================================================
//  CLASS Si570 --  Programmable XO/VCXO Chip Support
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class Si570
//-----------------------------------------------------------------------------

const double _DcoLow = 4.85e9;
const double _DcoHigh = 5.67e9;

const int DataRegOffset = 7;   // Base address of Data Registers
const int ResetRegOffset = 135;   // Base address of Data Registers
const int FreezeRegOffset = 137;   // Base address of Data Registers

const int TypeC_DividerProductMinimum = 18;  // N1*HsDiv must be larger to work
const int TypeB_DividerProductMinimum = 5;   // N1*HsDiv must be larger to work

const int Standard_I2C_SubAddr = 0xAA;  // X6
const int XU_TX_I2C_SubAddr    = 0xAC;  // XU_TX

//-----------------------------------------------------------------------------
//  constructor for class Si570
//-----------------------------------------------------------------------------
//  Note: we do NOT own the Device pointer, so don't delete it

Si570::Si570(IRequires_I2C_RegIo & device, TypeCode tCode, double pwr_on_mhz)
    :  ClockBase(0), FVcxoType(tCode), Device(&device),
       DcoLow(_DcoLow), DcoHigh(_DcoHigh),
       FPowerOnFrequencyMhz(pwr_on_mhz), FCurrentFrequency(pwr_on_mhz*1e6),
       FOverrideI2C(false) , FOverrideI2CAddr(Standard_I2C_SubAddr),
       FPermaReset(false),
       RegisterImage(rcDataRegCount),
       ControlImage(rcCtrlRegCount),
       TwoTwentyEightPowerAsFloat(268435456.0),
       FSyncOnInit(true),
       Debug(true)
{
    InitValidDividers();
    //  Enable Nack on Reads
    //  device.NackReads(true);
    //  Initialize FFrequency
    FFrequency =  FPowerOnFrequencyMhz*1e6;
}

//-----------------------------------------------------------------------------
// Si570::CanTuneTo(double freq)
//-----------------------------------------------------------------------------

bool Si570::CanTuneTo(double freq)
{
    bool retValue;
    switch(FVcxoType)
        {
        case tcA:
            {
            if (freq >= Min() && freq <= Gap0Min())
                retValue = true;
            else if (freq >= Gap0Max() && freq <= Gap1Min())
                retValue = true;
            else if (freq >= Gap1Max() && freq <= Max())
                retValue = true;
            else
                retValue = false;
            }
            break;
        default:
            {
            if (freq >= Min() && freq <= Max())
                retValue = true;
            else
                retValue = false;
            }
            break;
        }
    return retValue;
}

//-----------------------------------------------------------------------------
// Si570::BestTunableFrequency()
//-----------------------------------------------------------------------------

double Si570::BestTunableFrequency(double freq)
{
    double retValue;
    switch(FVcxoType)
        {
        case tcA:
            {
            if (freq < Min())
                retValue = Min();
            else if (freq >= Min() && freq <= Gap0Min())
                retValue = freq;
            else if (freq >= Gap0Min() && freq <= (Gap0Min()+Gap0Max()/2))
                retValue = Gap0Min();
            else if (freq >= (Gap0Min()+Gap0Max()/2) && freq <= Gap0Max())
                retValue = Gap0Max();
            else if (freq >= Gap0Max() && freq <= Gap1Min())
                retValue = freq;
            else if (freq >= Gap1Min() && freq <= (Gap1Min()+Gap1Max()/2))
                retValue = Gap1Min();
            else if (freq >= (Gap1Min()+Gap1Max()/2) && freq <= Gap1Max())
                retValue = Gap1Max();
            else if (freq >= Gap1Max() && freq <= Max())
                retValue = freq;
            else
                retValue = Max();
            }
            break;
        default:
            {
            if (freq <= Min())
                retValue = Min();
            else if (freq >= Max())
                retValue = Max();
            else
                retValue = freq;
            }
            break;
        }
    return retValue;
}

//-----------------------------------------------------------------------------
// Si570:: Min() - Tuning Range
//-----------------------------------------------------------------------------

double Si570:: Min()
{ 
    return 10.0e6;   //For all Vcxos
}

//-----------------------------------------------------------------------------
// Si570::Max() 
//-----------------------------------------------------------------------------

double Si570::Max()
{ 
    double val = 280.0e6;
    switch(FVcxoType)
        {
        case tcA:
            val = 1415.0e6;   
            break;
        case tcB:
            val = 810.0e6;
            break;
        case tcC:
            val = 280.0e6;
            break;
        default:
            break;
        }
    return val;          
}

//-----------------------------------------------------------------------------
//  Si570::ReInit()  --  Reinitialize to NVRAM settings
//-----------------------------------------------------------------------------
//  Since we always need the 'current' frequency, lets reset to known state
//
//void  Si570::ReInit()
//{
//    Recall();     // Move default NVRAM settings into memory
//
//    ReadDataRegisters();
//
//    Debug.Log("Si570 ReInit -- New Frequency = ", FPowerOnFrequencyMhz*1e6);
//    Debug.Log("Si570::HSDIV = ", (int)HsDiv());
//    Debug.Log("Si570::N1 = ", (int)N1());
//    Debug.Log("Si570::RFreq = ", RFreq(), 16);
//
//    // Apply the already loaded values to chip
//
//    // Freeze the DCO
//    Freeze();
//
//    // UnFreeze the DCO
//    Unfreeze();
//    //
//    NewFrequency();
//
//    FFrequency =  FCurrentFrequency = FPowerOnFrequencyMhz*1e6;
//}
//

//-----------------------------------------------------------------------------
//  Si570::QuickUpdate()  --  Fast, no disturbance clock update
//-----------------------------------------------------------------------------

void  Si570::QuickUpdate(double new_freq)
{
    // Only when change is < 3500 ppm
    ReadDataRegisters();

    ii64 OrigRFreq = RFreq();
    ii64 NewRFreq = static_cast<ii64>(OrigRFreq * (new_freq / FCurrentFrequency));

    RFreq(NewRFreq);

    Debug.Log("Si570 Quick Update -- New Frequency = ", new_freq);
    Debug.Log("Si570::Old Frequency = ", FCurrentFrequency);
    Debug.Log("Si570::Old RFreq = ", OrigRFreq, 16);
    Debug.Log("Si570::New RFreq = ", NewRFreq, 16);
    Debug.Log("Si570::HSDIV = ", (int)HsDiv());
    Debug.Log("Si570::N1 = ", (int)N1());
	
    // Update Hardware
    WriteDataRegisters();
    FCurrentFrequency = new_freq;
}

void Si570::LoadFreq(int hsdiv, int n1, ii64 rfreq)
{
    //  Reapply updated values to image
    HsDiv(hsdiv);
    N1(n1);
    RFreq(rfreq);

    // Freeze the DCO
    Freeze();

    // Update Hardware
    WriteDataRegisters();
    // UnFreeze the DCO
    Unfreeze();
    //
    NewFrequency();

}

//-----------------------------------------------------------------------------
//  Si570::FullUpdate()  --  General clock update
//-----------------------------------------------------------------------------

void  Si570::FullUpdate(double new_freq)
{
    // Only when change is < 3500 ppm
    ReadDataRegisters();

    Debug.Log("Si570 Full Update -- New Frequency = ", new_freq);
    Debug.Log("Si570::Frequency = ", FCurrentFrequency);
    Debug.Log("Si570::HSDIV = ", (int)HsDiv());
    Debug.Log("Si570::N1 = ", (int)N1());
    Debug.Log("Si570::RFreq = ", RFreq(), 16);

    // Calculate FXtal
    double OrigRF_Ratio = (double)RFreq() / TwoTwentyEightPowerAsFloat;
    if (OrigRF_Ratio==0.0)
        OrigRF_Ratio = 1.0;   // avoid divide by 0.
    double FXtal = FCurrentFrequency * (double)HsDiv() * (double)N1() / OrigRF_Ratio;
    int NewHsDiv = HsDiv();
    int NewN1    = N1();

    if (FXtal==0)
        FXtal = 1.0;   // avoid divide by 0.

    // Calculate New Dividers
    DividerScan(new_freq, NewHsDiv,  NewN1);

    // Calculate new RFreq value
    double RF_Ratio = new_freq * (double)NewHsDiv * (double)NewN1 / FXtal;

    ii64 NewRFreq = static_cast<ii64>(RF_Ratio * TwoTwentyEightPowerAsFloat);

    Debug.Log("Si570::New Frequency = ", new_freq);
    Debug.Log("Si570::HSDIV = ", NewHsDiv);
    Debug.Log("Si570::N1 = ", NewN1);
    Debug.Log("Si570::RFreq = ", NewRFreq, 16);

    //  Reapply updated values to image
    HsDiv(NewHsDiv);
    N1(NewN1);
    RFreq(NewRFreq);
    FCurrentFrequency = new_freq;

    // Freeze the DCO
    Freeze();

    // Update Hardware
    WriteDataRegisters();
    // UnFreeze the DCO
    Unfreeze();
    //
    NewFrequency();

}

//-----------------------------------------------------------------------------
//  Si570::DividerScan()  --  Hunt for proper divider product for a frequency
//-----------------------------------------------------------------------------

void  Si570::DividerScan(double & freq, int & HsDiv, int & N1)
{
    //  Search for Best HsDiv and N1 divider values that reduces DCO range to
    //    include the requested clock frequency
    //
    bool in_range = false;
    double BestError = std::numeric_limits<double>::max();
    int BestHsDiv = -1;
    int BestN1 = -1;
    double ClosestFreq = freq;

    //  Scan
    for (unsigned int hs=0; hs<Valid_HsDiv.size(); hs++)
        {
        for (unsigned int n1=0; n1<Valid_N1.size(); n1++)
            {
            //
            //  Divider is product of HsDiv and N1
            int divider = Valid_HsDiv[hs] * Valid_N1[n1];
            //
            //  Lower speed parts are restricted in 'total divider' value
            //     supported. Skip these
            if (FVcxoType==tcC && divider<=TypeC_DividerProductMinimum)
                continue;
            if (FVcxoType==tcB && divider<=TypeB_DividerProductMinimum)
                continue;
            // Type A has no restrictions...

            double LowerLimit = DcoLow/divider;
            double UpperLimit = DcoHigh/divider;

            if ((LowerLimit <= freq) && (UpperLimit >= freq))
                {
                in_range = true;
                BestHsDiv = Valid_HsDiv[hs];
                BestN1    = Valid_N1[n1];
                break;
                }
            else
                {
                double LowerError = std::fabs(freq-LowerLimit);
                if (LowerError < BestError)
                    {
                    BestError = LowerError;
                    BestHsDiv = Valid_HsDiv[hs];
                    BestN1    = Valid_N1[n1];
                    ClosestFreq = DcoLow/divider;
                    }

                double UpperError = std::fabs(freq-UpperLimit);
                if (UpperError < BestError)
                    {
                    BestError = UpperError;
                    BestHsDiv = Valid_HsDiv[hs];
                    BestN1    = Valid_N1[n1];
                    ClosestFreq = DcoHigh/divider;
                    }
                }
            //  If we have already divided past our frequency, try another HS divider
            if (UpperLimit < freq)
                break;
            }
        if (in_range)    // if we've found it, leave
            break;
        }
        //  Best match dividers have been found...'return' 'em in arguments
        HsDiv = BestHsDiv;
        N1    = BestN1;

        if (!in_range)    
            freq = ClosestFreq;
}

//-----------------------------------------------------------------------------
//  Si570::Init()  --  Init setup
//-----------------------------------------------------------------------------

void  Si570::Init()
{
    if (FOverrideI2C)
        {
        Device->I2C_SetSubAddress(FOverrideI2CAddr);
        }
    if (FSyncOnInit)
        {
        Reset();
        // ReInit();
        }
}

//-----------------------------------------------------------------------------
//  Si570::Reapply()  --  Reapply settings to hardware from saved state
//-----------------------------------------------------------------------------

void  Si570::Reapply()
{
    Frequency(FCurrentFrequency);
}

//-----------------------------------------------------------------------------
//  Si570::Calibrate()  --  'calibrate' the Si570
//-----------------------------------------------------------------------------

void  Si570::Calibrate(double power_on_rate_MHz)
{
    FPowerOnFrequencyMhz = power_on_rate_MHz;
    Reset();
}

//-----------------------------------------------------------------------------
//  Si570::SetFrequency()  --  Change Frequency
//-----------------------------------------------------------------------------

void Si570::SetFrequency(double freq)
{
    //  Are we within 3500 PPM ?
    if (std::fabs(FCurrentFrequency - freq)/FCurrentFrequency < 3500.0e-6)
        QuickUpdate(freq);
    else
        FullUpdate(freq);
}

//-----------------------------------------------------------------------------
//  Si570::GetFrequencyActual()  --  Calculate actual clock rate
//-----------------------------------------------------------------------------

double Si570::GetFrequencyActual() const
{
    return FCurrentFrequency;
}

//-----------------------------------------------------------------------------
//  Si570::SetEnabled()  --  Set enabled state
//-----------------------------------------------------------------------------

void  Si570::SetEnabled(bool /*state*/)
{
    // do nothing
}

//-----------------------------------------------------------------------------
//  Si570::InitValidDividers()  --  Initialize Divider valid values arrays
//-----------------------------------------------------------------------------

void  Si570::InitValidDividers()
{
    //  Allowed HSDIV values ( prefer large values )
    Valid_HsDiv.push_back(11);   Valid_HsDiv.push_back(9);
    Valid_HsDiv.push_back( 7);   Valid_HsDiv.push_back(6);
    Valid_HsDiv.push_back( 5);   Valid_HsDiv.push_back(4);

    //  Allowed N1 values 1 nad even #s ( prefer small values )
    Valid_N1.push_back(1);
    for (int i=2; i<=128; i+=2)
        Valid_N1.push_back(i);

}

//-----------------------------------------------------------------------------
//  Si570::Port()  --  Write to chip data port
//-----------------------------------------------------------------------------

void Si570::Port(unsigned char reg, unsigned char value)
{
    std::stringstream msg;
    msg << "Si570::Write 0x" << std::hex << static_cast<int>(value)
        << " to reg " << static_cast<int>(reg);
    Debug.Log(msg.str());
    //
    //  Write using I2C interface
    Device->I2C_Reg(reg, value);
}

//-----------------------------------------------------------------------------
//  Si570::Port()  --  Read from chip data port
//-----------------------------------------------------------------------------

unsigned char Si570::Port(unsigned char reg)
{
    //
    //  Read using I2C interface
    int value = Device->I2C_Reg(reg);

    std::stringstream msg;
    msg << "Si570::Read 0x" << std::hex << value
        << " from reg " << static_cast<int>(reg);
    Debug.Log(msg.str());

    return static_cast<unsigned char>(value);
}

//-----------------------------------------------------------------------------
//  Si570::ReadDataRegisters()  --  Read all data registers into Image vector
//-----------------------------------------------------------------------------

void Si570::ReadDataRegisters()
{
    for (unsigned int i=0; i<RegisterImage.size(); i++)
        {
        RegisterImage[i] = Port(DataRegOffset+i);
        }
}

//-----------------------------------------------------------------------------
//  Si570::WriteDataRegisters()  --  Write all data registers from Image vector
//-----------------------------------------------------------------------------

void Si570::WriteDataRegisters()
{
    for (unsigned int i=0; i<RegisterImage.size(); i++)
        {
        Port(DataRegOffset+i, RegisterImage[i]);
        }
}

//-----------------------------------------------------------------------------
//  Si570::HsDiv()  --  Read HsDiv from Image
//-----------------------------------------------------------------------------

unsigned int  Si570::HsDiv()
{
    PtrBitManipulator Reg7(&RegisterImage[rcReg7]);

    unsigned int value;
    unsigned int code = Reg7.Field(5, 3);    // bits 5,6,7
    switch (code)
        {
        case 0:  value = 4;  break;
        case 1:  value = 5;  break;
        case 2:  value = 6;  break;
        case 3:  value = 7;  break;
        case 5:  value = 9;  break;
        default:
        case 7:  value = 11; break;
        }
    return value;
}

//-----------------------------------------------------------------------------
//  Si570::HsDiv()  --  Write HsDiv to Image
//-----------------------------------------------------------------------------

void  Si570::HsDiv(unsigned int value)
{
    PtrBitManipulator Reg7(&RegisterImage[rcReg7]);

    unsigned int code;
    switch (value)
        {
        case 4:  code = 0;  break;
        case 5:  code = 1;  break;
        case 6:  code = 2;  break;
        case 7:  code = 3;  break;
        case 9:  code = 5;  break;
        default:
        case 11: code = 7;  break;
        }

    Reg7.Field(5, 3, code);    // bits 5,6,7
}

//-----------------------------------------------------------------------------
//  Si570::N1()  --  Read N1 from Image
//-----------------------------------------------------------------------------

unsigned int  Si570::N1()
{
    PtrBitManipulator Reg7(&RegisterImage[rcReg7]);
    PtrBitManipulator Reg8(&RegisterImage[rcReg8]);

    unsigned int value_hi = Reg7.Field(0, 5);    // bits 0-4
    unsigned int value_lo = Reg8.Field(6, 2);    // bits 6,7
    return ((value_hi<<2) | value_lo) + 1;
}

//-----------------------------------------------------------------------------
//  Si570::N1()  --  Write N1 to Image
//-----------------------------------------------------------------------------

void  Si570::N1(unsigned int value)
{
    PtrBitManipulator Reg7(&RegisterImage[rcReg7]);
    PtrBitManipulator Reg8(&RegisterImage[rcReg8]);

    value -= 1;
    Reg7.Field(0, 5, value >> 2);   // bits 0-4
    Reg8.Field(6, 2, value & 0x3);  // bits 6,7
}

//-----------------------------------------------------------------------------
//  Si570::RFreq()  --  Read RFreq from Image
//-----------------------------------------------------------------------------

ii64  Si570::RFreq()
{
    PtrBitManipulator Reg8(&RegisterImage[rcReg8]);
    PtrBitManipulator Reg9(&RegisterImage[rcReg9]);
    PtrBitManipulator Reg10(&RegisterImage[rcReg10]);
    PtrBitManipulator Reg11(&RegisterImage[rcReg11]);
    PtrBitManipulator Reg12(&RegisterImage[rcReg12]);

    ii64 value(0);
    value |= Reg8.Field(0, 6);  // MSB  (Bits 32-37)
    value <<= 8;
    value |= Reg9.Value();      // Bits 24-31
    value <<= 8;
    value |= Reg10.Value();     // Bits 16-23
    value <<= 8;
    value |= Reg11.Value();     // Bits 8-15
    value <<= 8;
    value |= Reg12.Value();     // LSB

    return value;
}

//-----------------------------------------------------------------------------
//  Si570::RFreq()  --  Write RFreq to Image
//-----------------------------------------------------------------------------

void  Si570::RFreq(ii64 value)
{
    PtrBitManipulator Reg8(&RegisterImage[rcReg8]);
    PtrBitManipulator Reg9(&RegisterImage[rcReg9]);
    PtrBitManipulator Reg10(&RegisterImage[rcReg10]);
    PtrBitManipulator Reg11(&RegisterImage[rcReg11]);
    PtrBitManipulator Reg12(&RegisterImage[rcReg12]);


    Reg8.Field(0, 6, static_cast<int>(value >> 32));   // MSB  (Bits 32-37)
    Reg9.Value(static_cast<int>(value >> 24));         // Bits 24-31
    Reg10.Value(static_cast<int>(value >> 16));        // Bits 16-23
    Reg11.Value(static_cast<int>(value >> 8));         // Bits 8-15
    Reg12.Value(static_cast<int>(value & 0xFF));       // LSB
}

//-----------------------------------------------------------------------------
//  Si570::Freeze()  --  Set Freeze Bit
//-----------------------------------------------------------------------------

void Si570::Freeze()
{
    // No need to read current state
    Port(FreezeRegOffset, 0x10);
}

//-----------------------------------------------------------------------------
//  Si570::Unfreeze()  --  Clear Freeze Bit
//-----------------------------------------------------------------------------

void Si570::Unfreeze()
{
    // No need to read current state
    Port(FreezeRegOffset, 0x00);
}

//-----------------------------------------------------------------------------
//  Si570::NewFrequency()  --  Inform chip of new frequency setting
//-----------------------------------------------------------------------------

void Si570::NewFrequency()
{
    // Since all bits auto clear, no need to read
    ControlImage[rcReg135] = (FPermaReset) ? 0x80 : 0;

    PtrBitManipulator ResetReg(&ControlImage[rcReg135]);

    ResetReg.Bit(6, 1);

    // Update
    Port(ResetRegOffset, ControlImage[rcReg135]);
}

//-----------------------------------------------------------------------------
//  Si570::Recall()  --  Recall NVRAM settings
//-----------------------------------------------------------------------------

void Si570::Recall()
{
    // Read state
    // Since all bits auto clear, no need to read
    ControlImage[rcReg135] = (FPermaReset) ? 0x80 : 0;

    PtrBitManipulator ResetReg(&ControlImage[rcReg135]);

    ResetReg.Bit(0, 1);

    // Update
    Port(ResetRegOffset, ControlImage[rcReg135]);
}

//-----------------------------------------------------------------------------
//  Si570::Reset()  --  Reset Chip
//-----------------------------------------------------------------------------

void Si570::Reset()
{
    // Read state
    // Since all bits auto clear, no need to read
    ControlImage[rcReg135] = 0;
    PtrBitManipulator ResetReg(&ControlImage[rcReg135]);

    ResetReg.Bit(7, 1);
    // Update
    Port(ResetRegOffset, ControlImage[rcReg135]);

    //  Since Bit auto clears, lets clear it
    FFrequency =  FCurrentFrequency = FPowerOnFrequencyMhz*1e6;
}

} // namespace Innovative

