// SimpleInterrupt_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <SimpleInterrupt_Mb.h>
#include <StartStopThread_Mb.h>
#include <SignalQueue_Mb.h>

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif

namespace Innovative
{

namespace SimpleIsrScheme
{

//===========================================================================
//  CLASS VirtualIsrThread  -- ISR handling thread for Simple ISR scheme
//===========================================================================
//---------------------------------------------------------------------------
//  declaration of  class VirtualIsrThread
//---------------------------------------------------------------------------

class VirtualIsrThread : public StartStopThread
{
    typedef StartStopThread inherited;

public:
    VirtualIsrThread(ISimpleInterrupt & device);

protected:
    enum { sDriverEvent=inherited::sLast, sLocalIsrQ, sLast };

    Event &                     DriverEvent;
    ISimpleInterrupt &          Device;
    SignalQueue<unsigned int>   LocalIsrQ;

    size_t  MaxInterruptDepth;
    size_t  MaxLocalIsrDepth;

    virtual void Init();
    virtual void SingleExecution();
    virtual void AtStart()
        {  Device.InterruptEnable(true);  }
    virtual void AtStop()
        {  Device.InterruptEnable(false);  }
};

//---------------------------------------------------------------------------
//  housekeeping for class VirtualIsrThread
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//  constructor for class VirtualIsrThread
//---------------------------------------------------------------------------

VirtualIsrThread::VirtualIsrThread(ISimpleInterrupt & device)
    : inherited("PacketStream::VirtualIsrThread"),
      DriverEvent( *(device.InterruptEvent()) ), Device(device), 
      MaxInterruptDepth(0), MaxLocalIsrDepth(0)
{
    Priority(tpTimeCritical);
}

//---------------------------------------------------------------------------
//  VirtualIsrThread::Init() -- Initialize sync objects
//---------------------------------------------------------------------------

void VirtualIsrThread::Init()
{
    inherited::Init();

    AddObject(DriverEvent, sDriverEvent);
    AddObject(LocalIsrQ.SignalSem(), sLocalIsrQ);
}

//---------------------------------------------------------------------------
//  VirtualIsrThread::SingleExecution() -- thread function
//---------------------------------------------------------------------------

void VirtualIsrThread::SingleExecution()
{
    int result = WaitFor();

    if (Terminated() || eStopped.Check())
        return;
    //
    //  A new interrupt was signalled. Since the low level queue is of
    //     finite size, we drain that queue as fast as we can. Since we
    //     also got at least one interrupt, we process one now.
    if (result == sDriverEvent)
        {
        size_t depth = Device.InterruptDepth();
        MaxInterruptDepth = std::max(depth, MaxInterruptDepth);

        while(depth--)
            {
            int irqstatus = Device.InterruptAck();
            LocalIsrQ.push(irqstatus);
            }

        size_t Idepth = LocalIsrQ.depth();
        MaxLocalIsrDepth = std::max(Idepth, MaxLocalIsrDepth);

        //
        //  Process one element in local Q
        if (Idepth != 0)
            {
            //  since we did not wait for Sem before, must decrement sem manually
            //    when popping IRQ register value
            unsigned int IrqToProcess = LocalIsrQ.front();
            LocalIsrQ.pop(SignalQueue<unsigned int>::PopSem);

            Device.InterruptDispatch(IrqToProcess);
            }
        }
    //
    //  We were signalled because of queued local ISRs.
    if (result == sLocalIsrQ)
        {
        size_t Idepth = LocalIsrQ.depth();
        //
        //  Process one element in local Q
        if (Idepth != 0)
            {
            unsigned int IrqToProcess = LocalIsrQ.front();
            LocalIsrQ.pop();  // don't dec sem, already done

            Device.InterruptDispatch(IrqToProcess);
            }
        }
}

//===========================================================================
//  CLASS SimpleIsrThread  -- ISR handling thread for Simple ISR scheme
//===========================================================================
//---------------------------------------------------------------------------
//  declaration of  class SimpleIsrThread
//---------------------------------------------------------------------------

class SimpleIsrThread : public VirtualIsrThread
{
    typedef VirtualIsrThread inherited;

public:
    SimpleIsrThread(ISimpleInterrupt & device)
        : inherited(device)
        {  }

protected:
    virtual void AtStart()
        {  Device.InterruptEnable(false);  }
};

}  // end namespace Innovative::SimpleIsrScheme


//===========================================================================
//  CLASS SimpleInterrupt  -- ISR & Busmaster region handling
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class SimpleInterrupt
//---------------------------------------------------------------------------

SimpleInterrupt::SimpleInterrupt(ISimpleInterrupt & device)
    : Device(device), Visr(0)
{
}

//---------------------------------------------------------------------------
//  destructor for class SimpleInterrupt
//---------------------------------------------------------------------------

SimpleInterrupt::~SimpleInterrupt()
{
    Close();
}

//---------------------------------------------------------------------------
//  SimpleInterrupt::Open()
//---------------------------------------------------------------------------

void  SimpleInterrupt::Open()
{
    if (Visr)
        return;

    //
    //  Create V-ISR thread.
    Visr = new SimpleIsrScheme::SimpleIsrThread(Device);

    Visr->Resume();
    Visr->Start();
}

//---------------------------------------------------------------------------
//  SimpleInterrupt::Close()
//---------------------------------------------------------------------------

void  SimpleInterrupt::Close()
{
    //
    //  Have VISR Thread shut itself down
    if (!Visr)
        return;

    Visr->Stop();
    Visr->Shutdown();
    delete Visr;
    Visr = NULL;
}

}  // end namespace Innovative

//---------------------------------------------------------------------------


