// SpiDevice_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2009 Innovative Integration Corporation.
//  All Rights Reserved.
//

#ifndef SpiDevice_MbH
#define SpiDevice_MbH

#include <string>
#include <HardwareRegister_Mb.h>
#include <Trace_Mb.h>

namespace Innovative
{

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif

//===========================================================================
//  CLASS ISpiDevice -- Interface to SPI controlled device
//===========================================================================

class ISpiDevice
{
public:
    enum SpiDir { spiWrite, spiRead };

    ISpiDevice(std::string tag="")
        :  Tag(tag)
        {}
    virtual ~ISpiDevice()
        {}

    unsigned char  Spi(int addr);
    void           Spi(int addr, unsigned char value);

protected:
    std::string  Tag;
    Trace        Debug;

    void  SpiReady();
    void  SpiNotReady();

    virtual void           SetupSpiRead(int addr) = 0;
    virtual void           SetupSpiWrite(int addr, unsigned char value) = 0;
    virtual unsigned char  ReadSpiData() = 0;
    virtual bool           ReadSpiReadyBit() = 0;
    virtual unsigned int   DumpSpiRegister() = 0;
};

//===========================================================================
//  CLASS ISpiDevice16 -- Interface to SPI controlled device (16 bit data)
//===========================================================================

class ISpiDevice16
{
public:
    enum SpiDir { spiWrite, spiRead };

    ISpiDevice16(std::string tag="")
        :  Tag(tag)
        {}
    virtual ~ISpiDevice16()
        {}

    unsigned short  Spi(int addr);
    void            Spi(int addr, unsigned short value);

protected:
    std::string  Tag;
    Trace        Debug;

    void  SpiReady();
    void  SpiNotReady();

    virtual void           SetupSpiRead(int addr) = 0;
    virtual void           SetupSpiWrite(int addr, unsigned short value) = 0;
    virtual unsigned short ReadSpiData() = 0;
    virtual bool           ReadSpiReadyBit() = 0;
    virtual unsigned int   DumpSpiRegister() = 0;
};

//===========================================================================
//  CLASS IMuxedSpiDevice -- Interface to SPI controlled device
//===========================================================================
//  One SPI engine controls multiple devices

class IMuxedSpiDevice
{
public:
    enum SpiDir { spiWrite, spiRead };

    IMuxedSpiDevice(std::string tag="")
        :  Tag(tag)
        {}
    virtual ~IMuxedSpiDevice()
        {}

    unsigned short  Spi(int device, int addr);
    void            Spi(int device, int addr, unsigned short value);

    void            SpiWriteAll(int addr, unsigned short value);

protected:
    std::string  Tag;
    Trace        Debug;

    void  SpiReady();
    void  SpiNotReady();

    virtual void            SetupSpiDevice(int device) = 0;
    virtual void            SetupSpiAllDevice() = 0;
    virtual void            SetupSpiRead(int addr) = 0;
    virtual void            SetupSpiWrite(int addr, unsigned short value) = 0;
    virtual unsigned short  ReadSpiData() = 0;
    virtual bool            ReadSpiReadyBit() = 0;
    virtual unsigned int    DumpSpiRegister() = 0;
};



//===========================================================================
//  CLASS ILTC2494SpiDevice -- Interface to SPI controlled device
//===========================================================================

class ILTC2494SpiDevice
{
public:

    ILTC2494SpiDevice(std::string tag="")
        :  Tag(tag)
        {}
    virtual ~ILTC2494SpiDevice()
        {}

    unsigned int   Spi(unsigned int addr);

protected:
    std::string  Tag;
    Trace        Debug;

    void  SpiReady();
    void  SpiNotReady();

    virtual void           SetupSpiRead(int addr) = 0;
    virtual unsigned int   ReadSpiData() = 0;
    virtual bool           ReadSpiReadyBit() = 0;
    virtual unsigned int   DumpSpiRegister() = 0;
    virtual unsigned int   ReadReg() = 0;
    virtual unsigned int   AddrReg() = 0;
};


//===========================================================================
//  CLASS StdSpiDevice -- Implementation of 'standard' SPI interface
//===========================================================================

class StdSpiDevice : public ISpiDevice
{
public:
    //  struct Info -- holds information about SpiDeviceReg
    //       ( needed since some fields are moved in different boards )
    struct Info
    {
        int Offset;         // offset of SPI register
        int DirectionBit;
        int ReadyBit;
        int DataValidBit;
    };

    StdSpiDevice(IAddressingSpace &space, Info & info, std::string tag="")
        :  ISpiDevice(tag), SpiWriteReg(space, info),
           SpiReadReg(space, info), SpiInfo(info)
        {}
    virtual ~StdSpiDevice()
        {}

protected:
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS SpiWriteRegister  -- SPI Write register
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	class SpiWriteRegister : public CachedShadowRegister
	{
		typedef CachedShadowRegister inherited;

	public:
		RegisterBitGroup    Data;
		RegisterBitGroup    Address;
		RegisterBit         Direction;

		SpiWriteRegister(IAddressingSpace &space, Info & info)
			: inherited(space, info.Offset), Data(*this, 0, 8),
			  Address(*this, 8, 7), Direction(*this, info.DirectionBit)
		{
		}
	};

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS SpiReadRegister  -- SPI Read register
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    class SpiReadRegister : public Register
    {
        typedef Register inherited;

    public:
        RegisterBit         Status;
        RegisterBit         Ready;
        RegisterBit         DataValid;
        RegisterBitGroup    ReadData;

        SpiReadRegister(IAddressingSpace &space, Info & info)
            : inherited(space, info.Offset), Status(*this, 21),
              Ready(*this, info.ReadyBit), DataValid(*this, info.DataValidBit),
                ReadData(*this, 24, 8)
        {
        }
    };

    SpiWriteRegister            SpiWriteReg;
    SpiReadRegister             SpiReadReg;
    Info                        SpiInfo;

    virtual void           SetupSpiRead(int addr);
    virtual void           SetupSpiWrite(int addr, unsigned char value);
    virtual unsigned char  ReadSpiData();
    virtual bool           ReadSpiReadyBit();
    virtual unsigned int   DumpSpiRegister();

};

//===========================================================================
//  CLASS X6SpiDevice -- Implementation of 'X6' SPI interface
//===========================================================================

class X6SpiDevice : public ISpiDevice16
{
public:
    //  struct Info -- holds information about X6SpiDevice
    struct Info
    {
        int WriteOffset;        // offset of SPI Write register
        int ReadOffset;         // offset of SPI Read register
    };

    X6SpiDevice(IAddressingSpace &space, Info & info, std::string tag="")
        :  ISpiDevice16(tag), SpiWriteReg(space, info),
           SpiReadReg(space, info), SpiInfo(info)
        {}
    virtual ~X6SpiDevice()
        {}

protected:
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS SpiWriteRegister  -- SPI Write register
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	class SpiWriteRegister : public CachedShadowRegister
	{
		typedef CachedShadowRegister inherited;

	public:
		RegisterBitGroup    Data;
		RegisterBitGroup    Address;
		RegisterBit         Direction;

		SpiWriteRegister(IAddressingSpace &space, Info & info)
			: inherited(space, info.WriteOffset), Data(*this, 0, 16),
			  Address(*this, 16, 8), Direction(*this, 28)
		{
		}
	};

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS SpiReadRegister  -- SPI Read register
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	class SpiReadRegister : public Register
	{
		typedef Register inherited;

	public:
		RegisterBit         Ready;
		RegisterBit         DataValid;
		RegisterBitGroup    ReadData;

		SpiReadRegister(IAddressingSpace &space, Info & info)
			: inherited(space, info.ReadOffset),
              Ready(*this, 30), DataValid(*this, 31),  ReadData(*this, 0, 16)
		{
		}
	};

    SpiWriteRegister            SpiWriteReg;
    SpiReadRegister             SpiReadReg;
    Info                        SpiInfo;

    virtual void           SetupSpiRead(int addr);
    virtual void           SetupSpiWrite(int addr, unsigned short value);
    virtual unsigned short  ReadSpiData();
    virtual bool           ReadSpiReadyBit();
    virtual unsigned int   DumpSpiRegister();

};


//===========================================================================
//  CLASS FmcSpiDevice -- Implementation of 'Fmc' SPI interface
//===========================================================================

class FmcSpiDevice : public IMuxedSpiDevice
{
public:
    //=======================================================================
    //  CLASS FmcSpiDevice::MapRegisters -- Device Register Offsets
    //=======================================================================
    class MapRegisters
    {
	public:
        unsigned int  Device_Addr;       // offset of SPI Device Enable register
        unsigned int  Write_Addr;        // offset of SPI Write register
        unsigned int  Read_Addr;         // offset of SPI Read register
    };

    class RegisterFieldParams
    {
    public:
        unsigned int DirectionBit_Offset;
        unsigned int Address_Width;

        RegisterFieldParams(unsigned int db_offset, unsigned int addr_w)
            :  DirectionBit_Offset(db_offset), Address_Width(addr_w)
            { }
    };

    FmcSpiDevice(IAddressingSpace &space, MapRegisters & regs, std::string tag="",
                 RegisterFieldParams field_params = RegisterFieldParams(28, 12),
                 unsigned int all_dev_mask=0xFF)
        :  IMuxedSpiDevice(tag), 
		   SpiDeviceReg(space, regs.Device_Addr),
           SpiWriteReg(space, regs.Write_Addr, field_params),
           SpiReadReg(space,  regs.Read_Addr),
           WriteAllMask(all_dev_mask)
        {}
    virtual ~FmcSpiDevice()
        {}

protected:
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS SpiWriteRegister  -- SPI Write register
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	class SpiWriteRegister : public CachedShadowRegister
	{
		typedef CachedShadowRegister inherited;

	public:
		RegisterBitGroup    Data;
		RegisterBitGroup    Address;
		RegisterBit         Direction;

        SpiWriteRegister(IAddressingSpace &space, unsigned int addr, RegisterFieldParams & field_params)
			: inherited(space, addr), Data(*this, 0, 16),
              Address(*this, 16, field_params.Address_Width),
              Direction(*this, field_params.DirectionBit_Offset)
		{
		}
	};

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS SpiReadRegister  -- SPI Read register
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	class SpiReadRegister : public Register
	{
		typedef Register inherited;

	public:
		RegisterBit         Ready;
		RegisterBit         DataValid;
		RegisterBitGroup    ReadData;

		SpiReadRegister(IAddressingSpace &space, unsigned int addr)
			: inherited(space, addr),
              Ready(*this, 30), DataValid(*this, 31),  ReadData(*this, 0, 16)
		{
		}
	};

    Register                    SpiDeviceReg;
    SpiWriteRegister            SpiWriteReg;
    SpiReadRegister             SpiReadReg;
    unsigned int                WriteAllMask;

    virtual void            SetupSpiDevice(int device);
    virtual void            SetupSpiAllDevice();
    virtual void            SetupSpiRead(int addr);
    virtual void            SetupSpiWrite(int addr, unsigned short value);
    virtual unsigned short  ReadSpiData();
    virtual bool            ReadSpiReadyBit();
    virtual unsigned int    DumpSpiRegister();

};

//===========================================================================
//  CLASS XUSpiDevice -- currently just a slightly tweaked FMC Spi Device
//===========================================================================

class XUSpiDevice : public FmcSpiDevice
{
public:
    XUSpiDevice(IAddressingSpace &space, MapRegisters & regs, std::string tag="",
                 RegisterFieldParams field_params = RegisterFieldParams(31, 15),
                 unsigned int all_dev_mask=0xFF)
        :  FmcSpiDevice(space, regs, tag, field_params, all_dev_mask)
        {}
    virtual ~XUSpiDevice()
        {}
};

//===========================================================================
//  CLASS XASpiDevice -- currently just a slightly tweaked FMC Spi Device
//===========================================================================

class XASpiDevice : public FmcSpiDevice
{
public:
    XASpiDevice(IAddressingSpace &space, MapRegisters & regs, std::string tag="",
                 RegisterFieldParams field_params = RegisterFieldParams(28, 12),
                 unsigned int all_dev_mask=0xFF)
        :  FmcSpiDevice(space, regs, tag, field_params, all_dev_mask)
        {}
    virtual ~XASpiDevice()
        {}
};

//===========================================================================
//  CLASS FmcCpldSpiDevice -- Implementation of yet another SPI
//===========================================================================

class FmcCpldSpiDevice : public ISpiDevice
{
public:
    //  struct Info -- holds information about SpiDeviceReg
    struct Info
    {
        unsigned int  Status_Addr;
        unsigned int  Data_Addr;
    };

    FmcCpldSpiDevice(IAddressingSpace &space, Info & info, std::string tag="")
        :  ISpiDevice(tag),
           SpiWriteReg(space,  info.Data_Addr),
           SpiReadReg(space,   info.Data_Addr),
           SpiStatusReg(space, info.Status_Addr),
           SpiInfo(info)
        {}
    virtual ~FmcCpldSpiDevice()
        {}

protected:
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS SpiWriteRegister  -- SPI Write Data register
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	class SpiWriteRegister : public CachedShadowRegister
	{
		typedef CachedShadowRegister inherited;

	public:
		RegisterBitGroup    Address;
		RegisterBit         Direction;
		RegisterBitGroup    Data;

		SpiWriteRegister(IAddressingSpace &space, unsigned int offset)
			: inherited(space, offset), Address(*this, 0, 8),
			  Direction(*this, 12), Data(*this, 16, 8)
		{
		}
	};

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS SpiReadRegister  -- SPI Rwead register
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    class SpiReadRegister : public ReadOnlyRegister
    {
        typedef ReadOnlyRegister inherited;

    public:
        ReadOnlyRegisterBitGroup    ReadData;

        SpiReadRegister(IAddressingSpace &space, unsigned int offset)
            : inherited(space, offset), ReadData(*this, 24, 8)
        {
        }
    };

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS SpiStatusRegister  -- SPI Status register
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    class SpiStatusRegister : public ReadOnlyRegister
    {
        typedef ReadOnlyRegister inherited;

    public:
        ReadOnlyRegisterBit         Ready;
        ReadOnlyRegisterBit         DataValid;

        SpiStatusRegister(IAddressingSpace &space, unsigned int offset)
            : inherited(space, offset),
              Ready(*this, 30), DataValid(*this, 31)
        {
        }
    };

    SpiWriteRegister            SpiWriteReg;
    SpiReadRegister             SpiReadReg;
    SpiStatusRegister           SpiStatusReg;
    Info                        SpiInfo;

    virtual void           SetupSpiRead(int addr);
    virtual void           SetupSpiWrite(int addr, unsigned char value);
    virtual unsigned char  ReadSpiData();
    virtual bool           ReadSpiReadyBit();
    virtual unsigned int   DumpSpiRegister();

};

//===========================================================================
//  CLASS LTC2494SpiDevice -- Implementation of LTC2494 SPI interface
//===========================================================================

class LTC2494SpiDevice : public ILTC2494SpiDevice
{
public:
    //  struct Info -- holds information about X6SpiDevice
    struct Info
    {
        int WriteOffset;        // offset of SPI Write register
        int ReadOffset;         // offset of SPI Read register
    };

    LTC2494SpiDevice(IAddressingSpace &space, Info & info, std::string tag="")
        :  ILTC2494SpiDevice(tag), SpiWriteReg(space, info),
           SpiReadReg(space, info), SpiInfo(info)
        {}
    virtual ~LTC2494SpiDevice()
        {}

protected:
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS SpiWriteRegister  -- SPI Write register
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    class SpiWriteRegister : public Register
    {
        typedef Register inherited;

    public:
        RegisterBitGroup    Address;

        SpiWriteRegister(IAddressingSpace &space, Info & info)
            : inherited(space, info.WriteOffset), Address(*this, 0, 16)
        {
        }
    };

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS SpiReadRegister  -- SPI Read register
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    class SpiReadRegister : public Register
    {
        typedef Register inherited;

    public:
        RegisterBit         Ready;
        RegisterBit         DataValid;
        RegisterBitGroup    ReadData;

        SpiReadRegister(IAddressingSpace &space, Info & info)
            : inherited(space, info.ReadOffset),Ready(*this, 30),
              DataValid(*this, 31),  ReadData(*this, 0, 24)
        {
        }
    };

    SpiWriteRegister            SpiWriteReg;
    SpiReadRegister             SpiReadReg;
    Info                        SpiInfo;

    virtual void           SetupSpiRead(int addr);
    virtual unsigned int   ReadSpiData();
    virtual bool           ReadSpiReadyBit();
    virtual unsigned int   DumpSpiRegister();
    virtual unsigned int   ReadReg()
    { return SpiReadReg.Value(); }
    virtual unsigned int   AddrReg()
    { return SpiWriteReg.Value(); }

};

#ifdef _MSC_VER
	#pragma warning(pop)
#endif

} // namespace Innovative
#endif
