// Synchronizer_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <Synchronizer_Mb.h>
#include <Events_Mb.h>
#include <Exception_Mb.h>

namespace   Innovative
{

//=============================================================================
//  CLASS Synchronizer  --  class for Synchronizing with main UI thread
//=============================================================================

//---------------------------------------------------------------------------
//  constructor for class Synchronizer
//---------------------------------------------------------------------------

Synchronizer::Synchronizer()
    : mThunker(NewThunker()),
      Aborted(Event::AutoReset), SyncCallbackDone(Event::AutoReset),
      AbortCompleted(Event::AutoReset), InSyncronization(Event::ManualReset, true)
{
    WaitOnCompletion.AddObject(Aborted, sAborted);
    WaitOnCompletion.AddObject(SyncCallbackDone, sSyncCallbackDone);
}

//---------------------------------------------------------------------------
//  destructor for class Synchronizer
//---------------------------------------------------------------------------

Synchronizer::~Synchronizer()
{
    WaitOnCompletion.Clear();
    DeleteThunker(mThunker);
    mThunker = 0;
}

//---------------------------------------------------------------------------
//  Synchronizer::Notify() --
//---------------------------------------------------------------------------

bool Synchronizer::Notify()
{
    if (!mThunker->CanThunk())
        return true;

	{
	CriticalSectionArea CriticalArea( Section );
	if( Aborted.Check() )
		{
	    AbortCompleted.Set();
		return false;
		}

	//
	//  Post notification msg
    if (!mThunker->Notify())
        return false;

	InSyncronization.Clear();
	}

    //
    //  Wait for the main UI to signal completion, or Abort
    int result = WaitOnCompletion.WaitFor();

	InSyncronization.Set();
    if (result == sAborted)
        {
        //  We have aborted. Signal sender & quit
        AbortCompleted.Set();
        // "Abort called...leaving sync early
        throw MalibuException("class Innovative::Synchronizer::BeginSynchronize");
        }
    //
    //  Sync event finished, move on.
    return result == sSyncCallbackDone;
}

//---------------------------------------------------------------------------
//  Synchronizer::Abort() -- Abandon wait for sync to finish
//---------------------------------------------------------------------------

void Synchronizer::Abort(bool wait)
{
	Aborted.Set();

	{
	CriticalSectionArea CriticalArea( Section );
	if( ! InSyncronization.Check() )
		return;
	}

    //  Set abort signal, then wait for it to complete.
	if(wait)
		AbortCompleted.WaitFor();
}

//---------------------------------------------------------------------------
//  Synchronizer::Acknowledge() -- Acknowledge event completion
//---------------------------------------------------------------------------

void Synchronizer::Acknowledge()
{
    if (!mThunker->CanThunk())
        return;

    //
    //  Let 'em know we are done
    SyncCallbackDone.Set();
}

//=============================================================================
//  CLASS BasicEventSynchronizer  --
//=============================================================================

//---------------------------------------------------------------------------
//  BasicEventSynchronizer::ImplementExecute() --
//---------------------------------------------------------------------------

void BasicEventSynchronizer::ImplementExecute(OpenWire::EventSyncPair * Event,
                                     ThunkerIntf * thunker)
{
    if (!thunker->CanThunk())
        {
        Event->Execute();
        delete Event;
        return;
        }

    mEventQueue.Add(Event);
}

//---------------------------------------------------------------------------
//  BasicEventSynchronizer::OnExecuted() --
//---------------------------------------------------------------------------

void BasicEventSynchronizer::OnExecuted(OpenWire::Event & /*event*/)
{
    OpenWire::EventSyncPair * EventPair;

    while( mEventQueue.Get(EventPair))
        {
        EventPair->Execute();
        delete EventPair;
        }
}

//=============================================================================
//  CLASS EventSynchronizer  --  Block for synchronized execution in main UI thread
//=============================================================================

//---------------------------------------------------------------------------
//  EventSynchronizer::EventSynchronizer() -- Constructor
//---------------------------------------------------------------------------

EventSynchronizer::EventSynchronizer()
    : mSynchronizer(new Synchronizer)
{
    mSynchronizer->mThunker->OnNotified.SetEvent(this, &EventSynchronizer::OnExecuted);
}

//---------------------------------------------------------------------------
//  EventSynchronizer::~EventSynchronizer() -- Destructor
//---------------------------------------------------------------------------

EventSynchronizer::~EventSynchronizer()
{
    delete mSynchronizer;
    mSynchronizer = 0;
}

//---------------------------------------------------------------------------
//  EventSynchronizer::Clone() -- Create a copy of object
//---------------------------------------------------------------------------

BasicEventSynchronizer * EventSynchronizer::Clone()
{
    return new EventSynchronizer;
}

//---------------------------------------------------------------------------
//  EventSynchronizer::Execute() -- Push event while crossing thread boundary
//---------------------------------------------------------------------------

void EventSynchronizer::Execute(OpenWire::EventSyncPair * event)
{
    ImplementExecute(event, mSynchronizer->mThunker);
    mSynchronizer->Notify();
}

//---------------------------------------------------------------------------
//  EventSynchronizer::OnExecuted() -- Dispatch event within destination thread
//---------------------------------------------------------------------------

void EventSynchronizer::OnExecuted(OpenWire::Event & event)
{
    inherited::OnExecuted(event);
    mSynchronizer->Acknowledge();
}

//=============================================================================
//  CLASS EventThunker  --  "Thunk" call into main thread, unlocked with caller
//=============================================================================

//---------------------------------------------------------------------------
//  EventThunker::EventThunker() -- Constructor
//---------------------------------------------------------------------------

EventThunker::EventThunker()
    : mThunker(NewThunker())
{
    mThunker->OnNotified.SetEvent((inherited *)this, &EventThunker::OnExecuted);
}

//---------------------------------------------------------------------------
//  EventThunker::~EventThunker() -- Destructor
//---------------------------------------------------------------------------

EventThunker::~EventThunker()
{
    DeleteThunker(mThunker);
    mThunker = 0;
}

//---------------------------------------------------------------------------
//  EventThunker::Clone() -- Create a copy of object
//---------------------------------------------------------------------------

BasicEventSynchronizer * EventThunker::Clone()
{
    return new EventThunker;
}

//---------------------------------------------------------------------------
//  EventThunker::Execute() -- Push event while crossing thread boundary
//---------------------------------------------------------------------------

void EventThunker::Execute(OpenWire::EventSyncPair * event)
{
    ImplementExecute(event, mThunker);
    mThunker->Notify();
}


}  // namespace
