// SystemResponse_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <cmath>
#include <SystemResponse_Mb.h>

using namespace std;

namespace Innovative
{

//===========================================================================
//  CLASS SystemResponse  --
//===========================================================================

//---------------------------------------------------------------------------
// SystemResponse::SystemResponse() --
//---------------------------------------------------------------------------

SystemResponse::SystemResponse(StimulusResponse & io)
    : mChannels(2), mSweepMode(false),
      mIo(io), mPi(4.0*atan(1.0))
{
}

//---------------------------------------------------------------------------
//  SystemResponse::Smooth() -- Smooth response and recalculate impulse
//---------------------------------------------------------------------------

void SystemResponse::Smooth(const double octaves)
{
    if (mSweepMode)
        return;

    //generate cumulative data
    vector < double > cumulative1(mOutput);
    vector < double > cumulative2(mOutput);

    double      acc1 = 0;
    double      acc2 = 0;

    for (unsigned int i = 0; i < mOutput; ++i)
        {
        acc1 += abs(mAmplitude[i]);
        acc2 += arg(mAmplitude[i]);
        cumulative1[i] = acc1;
        cumulative2[i] = acc2;
        }

    mStats.Init();
    for (unsigned int i = 0; i < mOutput; ++i)
        {
        int low = static_cast<int>(double(i) * pow(2., -octaves));
        if (low < 1)
            low = 1;            //Don't include DC component

        if (low > static_cast<int>(Samples() / 2))
            low = static_cast<int>(Samples() / 2);

        int high = static_cast<int>(double(i) * pow(2., octaves));
        if (high < 1)
            high = 1;

        if (high > static_cast<int>(Samples() / 2))
            high = static_cast<int>(Samples() / 2);

        double  avg1 = cumulative1[high] - cumulative1[low - 1];
        avg1 /= high - low + 1;

        double  avg2 = cumulative2[high] - cumulative2[low - 1];
        avg2 /= high - low + 1;

        mAmplitude[i] = polar(avg1, avg2);
        mStats.Sample(20. * log10(abs(mAmplitude[i])));
        }

    CalculateImpulseResponse();
}

//---------------------------------------------------------------------------
//  SystemResponse::Gate() --  Gate impulse data and recalculate response
//    Doesn't destroy original impulse vector, can be repeated
//---------------------------------------------------------------------------

void SystemResponse::Gate(const unsigned int start, const unsigned int stop)
{
    if (mSweepMode)
        return;

    if (start > mImpulse.size() || stop > mImpulse.size())
        throw "Mls:Gate Wrong range";

    for (unsigned int i = 0; i < Samples(); ++i)
        mFft.Time()[i] = static_cast<float>(mImpulse[i]);

    if (stop > start)
        {
        for (unsigned int i = 0; i < start; ++i)
            mFft.Time()[i] = 0;

        for (unsigned int i = stop; i < mImpulse.size(); ++i)
            mFft.Time()[i] = 0;
        }
    else
        {
        for (unsigned int i = stop; i < start; ++i)
            mFft.Time()[i] = 0;
        }

    //Recalculate H
    mFft.Transform();

    mStats.Init();
    for (unsigned int i = 0; i < mOutput; ++i)
        {
        mAmplitude[i] = mFft.Frequency()[i];
        mStats.Sample(20. * log10(abs(mAmplitude[i])));
        }
}

//---------------------------------------------------------------------------
//  SystemResponse::GenerateStimulusBuffer() --
//---------------------------------------------------------------------------

void SystemResponse::GenerateStimulusBuffer()
{
    IppFloatDG  StimulusDG(mStimulus);
    mDebug.Log("Generating stimulus buffer...");
    for (unsigned int i = 0; i < Samples(); ++i)
        {
        StimulusDG[i] = ((mSequence)[i] == 1) ? 1.0f : -1.0f;

#ifdef SQUARE
        StimulusDG[i] = (i % 32 > 16) ? 1.0f : -1.0f;
#endif
        }
}

//---------------------------------------------------------------------------
//  SystemResponse::CalculateImpulseResponse() --
//---------------------------------------------------------------------------

void SystemResponse::CalculateImpulseResponse()
{
    Buffer I( Holding<Complex>(Samples()));
    Buffer O( Holding<Complex>(Samples()));

    IppComplexDG I_DG(I);
	IppComplexDG O_DG(O);

	for (unsigned int i = 0; i < mOutput; ++i)
		I_DG[i] = mAmplitude[i];
    mIfft.Transform(I_DG, O_DG);

//    IppFloatDG   O_DG(O);
    O_DG /= static_cast<float>(Samples());

    for (unsigned int i = 0; i < Samples(); ++i)
		mImpulse[i] = O_DG[i].real();
}

//===========================================================================
//  CLASS MaximumLengthSequence  --
//===========================================================================

//---------------------------------------------------------------------------
//  MaximumLengthSequence Ctor
//---------------------------------------------------------------------------

MaximumLengthSequence::MaximumLengthSequence(StimulusResponse & io)
    : SystemResponse(io)
{
    io.OnProgress.SetEvent(this, &MaximumLengthSequence::HandleOnStimulusProgress);
    io.OnComplete.SetEvent(this, &MaximumLengthSequence::HandleOnStimulusComplete);
}

//---------------------------------------------------------------------------
//  MaximumLengthSequence::Initialize() --
//---------------------------------------------------------------------------

void MaximumLengthSequence::Initialize(const double rate,
    const bool stereo, const unsigned int order)
{
    mSweepMode = false;
    mSampleRate = rate;
    mChannels = stereo ? 2 : 1;
    mOrder = order;
    mStereo = stereo;

    if (order > 20)
        throw "SystemResponse::Initialize: order > 20 Are you crazy?";

    mSequence.Order(order);

    mIfft.Order(order);
    mFft.Order(order);
    mFft.Window(wtBlackman);
    mIfft.Normalization(InverseFourier::Normalize);
    mOutput = (Samples() / 2) + 1;

    mStimulus.Resize(Samples());
    mResponse.Resize(Samples() * mChannels);

    GenerateStimulusBuffer();

    mIo.Initialize(rate, stereo);
}

//---------------------------------------------------------------------------
//  MaximumLengthSequence::Measure() --
//---------------------------------------------------------------------------

void MaximumLengthSequence::Measure(const double amplitude,
    const unsigned int averages, const bool reverse)
{
    IppFloatDG StimulusDG(mStimulus);
    IppFloatDG ResponseDG(mResponse);

    if (!StimulusDG.size())
        throw "SystemResponse::Measure: Run Initialize first";

    mReverse = reverse;
    mAverages = averages;
    mAverage = 0;

    mDebug.Log("SystemResponse::Measure");

    // Prepare channels
    mIo.Set(&StimulusDG[0], &ResponseDG[0], Samples(), amplitude);

    mIo.Process(2 * Samples());
}

//---------------------------------------------------------------------------
//  MaximumLengthSequence::HandleOnStimulusProgress() --
//---------------------------------------------------------------------------

void MaximumLengthSequence::HandleOnStimulusProgress(ProcessProgressEvent & e)
{
    size_t portion = (mAverage * 100) / mAverages;
    int percent = static_cast<int>((e.Percent/mAverages) + portion);

    ProcessProgressEvent progEvent(percent);
    OnProgress.Execute(progEvent);
}

//---------------------------------------------------------------------------
//  MaximumLengthSequence::HandleOnStimulusComplete() --
//---------------------------------------------------------------------------

void MaximumLengthSequence::HandleOnStimulusComplete(OpenWire::NotifyEvent & /*event*/)
{
    mDebug.Log("SystemResponse::Measure - stimulus finished");
    IppFloatDG ResponseDG(mResponse);

    std::stringstream msg;
    msg << "Signal level" << (mStereo ? "s" : "") << " = ( ";

    double mRms[2] = {0.0, 0.0};
    for (int ch = 0; ch < mChannels; ++ch)
        {
        for (unsigned int i = 0; i < Samples(); ++i)
            {
            double sample = ResponseDG[mChannels * i  + ch];
            mRms[ch] += sample * sample;
            }

        mRms[ch] /= Samples();
        mRms[ch] = sqrt(mRms[ch]);
        msg << mRms[ch] << ((!ch && mStereo) ? ", " : " ");
        }

    msg << ") rms";
    ProcessStatusEvent statusEvent(msg.str());
    OnStatus.Execute(statusEvent);

    ++mAverage;
    if ((mAverages > 1) && (mAverages == mAverage))
        Correlate(mAverage, mReverse);
    else if (mAverage)
        Correlate(1, mReverse);
    else
        Correlate(0, mReverse);

    if (mAverage < mAverages)
        {
        mIo.Process(2 * Samples());
        return;
        }

    double valueRms = std::max(mRms[0], mRms[1]);
    SystemMeasurementEvent sysEvent(valueRms);
    OnComplete.Execute(sysEvent);
}

//---------------------------------------------------------------------------
//  MaximumLengthSequence::Correlate() --
//---------------------------------------------------------------------------

void MaximumLengthSequence::Correlate(const unsigned int average,
    const bool reverse)
{
    unsigned int refOffset = reverse ? 1 : 0;
    unsigned int measOffset = reverse ? 0 : 1;
    IppFloatDG StimulusDG(mStimulus);
    IppFloatDG ResponseDG(mResponse);

    // Prepare for FFT
    mFrequency.resize(mOutput);
    mAmplitude.resize(mOutput);
    mImpulse.resize(Samples());

    // Run FFT on reference channel
    for (unsigned int i = 0; i < Samples(); ++i)
        {
        if (mChannels == 2)
            mFft.Time()[i] = ResponseDG[i * 2 + refOffset];
        else
            mFft.Time()[i] = StimulusDG[i];
        }
    mFft.Transform();

    vector < complex < double > > X(mOutput);
    for (unsigned int i = 0; i < mOutput; ++i)
        {
        mFrequency[i] = mIo.SampleRate() * i / Samples();
        X[i] = mFft.Frequency()[i];
        }

    // Run FFT on measurement channel
    for (unsigned int i = 0; i < Samples(); ++i)
        {
        if (mChannels == 2)
            mFft.Time()[i] = ResponseDG[i * 2 + measOffset];
        else
            mFft.Time()[i] = ResponseDG[i];
        }
    mFft.Transform();

    vector < complex < double > > Y(mOutput);
    for (unsigned int i = 0; i < mOutput; ++i)
        Y[i] = mFft.Frequency()[i];

    // Calculate H
    mStats.Init();
    for (unsigned int i = 0; i < mOutput; ++i)
        {
        complex<double> Hi;
        Hi = (X[i] * conj(Y[i])) / (X[i] * conj(X[i]));
        if (!average)
            mAmplitude[i] = conj(Hi);  //TODO: Why is conj needed?
        else
            mAmplitude[i] += conj(Hi); //TODO: Why is conj needed?

        mStats.Sample(20. * log10(abs(mAmplitude[i])));

        if (average > 1)
            mAmplitude[i] /= double(average);
        }

    CalculateImpulseResponse();
}

//===========================================================================
//  CLASS SineSweep  --
//===========================================================================

//---------------------------------------------------------------------------
//  SineSweep Ctor
//---------------------------------------------------------------------------

SineSweep::SineSweep(StimulusResponse & io)
    : SystemResponse(io), mMinTime(0.1)
{
    io.OnProgress.SetEvent(this, &SineSweep::HandleOnStimulusProgress);
    io.OnComplete.SetEvent(this, &SineSweep::HandleOnStimulusComplete);
}

//---------------------------------------------------------------------------
//  SineSweep::Initialize() --
//---------------------------------------------------------------------------

void SineSweep::Initialize(const double sampleRate, const bool stereo,
    const double minStepTime)
{
    mSweepMode = true;
    mMinTime = minStepTime;

    mSampleRate = sampleRate;
    mStereo = stereo;
    mChannels = stereo ? 2 : 1;

    mIo.Initialize(sampleRate, stereo);
}

//---------------------------------------------------------------------------
//  SineSweep::Measure() --
//---------------------------------------------------------------------------

void SineSweep::Measure(const double amplitude, const double start,
    const double stop, const double pointsPerDecade, const bool reverse)
{
    if (!mSweepMode)
        throw "SystemResponse::Sweep: Run Initialize first";

    mDebug.Log("SystemResponse::Measure");

    mReverse = reverse;

    double  frequency = start;
    long    now = 0;

    mFrequency.clear();
    mSweepPositions.clear();
    mAmplitude.clear();

    // Make list of frequencies
    while (frequency <= stop)
        {
        // Minimum 10 periods
        unsigned int duration = static_cast<unsigned int>(10. * mIo.SampleRate() / frequency);

        if (double(duration) < mIo.SampleRate() * mMinTime)
            duration = static_cast<unsigned int>(mIo.SampleRate() * mMinTime);
        now += duration;
        mIo.AddFrequency(duration, frequency);
        mFrequency.push_back(frequency);
        mSweepPositions.push_back(now);
        frequency *= pow(10., 1. / pointsPerDecade);
        }

    mSweepSamples = now + static_cast<unsigned int>(mIo.SampleRate()) / 2;

    if (mSweepSamples > 10e6)
        throw "SystemResponse::Measure: Are you crazy?";

    IppFloatDG ResponseDG(mResponse);
    ResponseDG.Resize(mSweepSamples * mChannels);

    mIo.Set(NULL, &ResponseDG[0], mSweepSamples, amplitude);

    mIo.Process(mSweepSamples);
}

//---------------------------------------------------------------------------
//  SineSweep::HandleOnStimulusProgress() --
//---------------------------------------------------------------------------

void SineSweep::HandleOnStimulusProgress(ProcessProgressEvent & event)
{
    OnProgress.Execute(event);
}

//---------------------------------------------------------------------------
//  SineSweep::HandleOnStimulusComplete() --
//---------------------------------------------------------------------------

void SineSweep::HandleOnStimulusComplete(OpenWire::NotifyEvent & /*event*/)
{
    mDebug.Log("SystemResponse::Measure - stimulus finished");

    double rms = SineFit(mSweepSamples, mReverse);
    SystemMeasurementEvent sysEvent(rms);
    OnComplete.Execute(sysEvent);
}

//---------------------------------------------------------------------------
//  SineSweep::GaussEliminate() --
//---------------------------------------------------------------------------

#ifdef __BORLANDC__
    #pragma option push
    #pragma warn -8004
#endif

void SineSweep::GaussEliminate(unsigned int n, vector<double> my, vector<double>&  x)
{
    unsigned int    rf = n + 1;
    unsigned int    dia = 0;
    unsigned int    row = 0;
    unsigned int    col = 0;

    for (dia = 0; dia < n - 1; ++dia)
        {
        if (my[dia + dia * rf] == 0.)
            {
            throw "Diagonal element zero";
            }
        else
            {
            for (row = dia + 1; row < n; ++row)
                {
                double  factor = my[dia + row * rf] / my[dia + dia * rf];
                for (col = dia; col < rf; ++col)
                    {
                    my[col + row * rf] = my[col + row * rf] -
                        factor * my[col + dia * rf];
                    }
                }
            }
        }

    // Check for singularity
    for (dia = 0; dia < n; ++dia)
        {
        if (my[dia + dia * rf] == 0.)
            throw "Singular matrix";
        }

    // Get X values
    x[n - 1] = my[n + (n - 1) * rf] / my[n - 1 + (n - 1) * rf];
    for (row = n - 2; static_cast<int>(row) >= 0; --row)
        {
        double  sum = 0.;
        for (col = row + 1; col < n; ++col)
            sum = sum + my[col + row * rf] * x[col];

        x[row] = (my[3 + row * rf] - sum) / my[row + row * rf];
        }
}

#ifdef __BORLANDC__
    #pragma option pop
#endif

//---------------------------------------------------------------------------
//  SineSweep::SineFit() --
//---------------------------------------------------------------------------

double SineSweep::SineFit(const unsigned int sweepSamples, const bool reverse)
{
    unsigned int    refOffset = reverse ? 1 : 0;
    unsigned int    measOffset = reverse ? 0 : 1;
    //IppFloatDG StimulusDG(mStimulus);
    IppFloatDG ResponseDG(mResponse);

    if (mChannels == 1)
        {
        refOffset = 0;
        measOffset = 0;
        }

    // Find peak input
    double      compressPeak(0.);
    for (unsigned int i = 0; i < sweepSamples; ++i)
        {
        if (compressPeak < abs(ResponseDG[i * mChannels + refOffset]))
            compressPeak = abs(ResponseDG[i * mChannels + refOffset]);
        }

    mDebug.Log("Reference peak input = ", compressPeak);
    compressPeak *= 0.05;

    // Find output to input delay
    unsigned int    window(mSweepPositions[mSweepPositions.size() - 1]);
    unsigned int    span(sweepSamples - window);
    unsigned int    delay(span / 2);
    double      pwrLo = 0.0;
    double      pwrHi = 0.0;

    while (span > 2)
        {
        unsigned int    ofLo(delay - span / 2);
        unsigned int    ofHi(delay + span / 2);
        pwrLo = 0.;
        pwrHi = 0.;
        for (unsigned int i = 0; i < window; ++i)
            {
            double  yLo = abs(ResponseDG[(i + ofLo) * mChannels + refOffset]);
            double  yHi = abs(ResponseDG[(i + ofHi) * mChannels + refOffset]);
            if (yLo > compressPeak)
                yLo = compressPeak;
            if (yHi > compressPeak)
                yHi = compressPeak;
            pwrLo += yLo * yLo;
            pwrHi += yHi * yHi;
            }

        span /= 2;
        if (pwrLo > pwrHi)
            delay -= span / 2;
        else
            delay += span / 2;
        }

    mDebug.Log("Output to input delay = ", static_cast<int>(delay));

    // Make sure we don't fit over beginning of next step
    if (delay > static_cast<unsigned int>(mIo.SampleRate() * mMinTime) / 4)
        delay -= static_cast<unsigned int>(mIo.SampleRate() * mMinTime) / 4;

    // Least squares sine wave fitting
    // Only the last half of each frequency is fitted
    vector < double > mtmmty1(12);
    vector < double > mtmmty2(12);

    unsigned int    lastStart = 0;
    unsigned int    j;
    mStats.Init();
    for (unsigned int i = 0; i < mSweepPositions.size(); ++i)
        {
        mtmmty1.assign(12, 0.);
        mtmmty2.assign(12, 0.);

        //Calculate mtm and mty
        for (j = 0; j < (mSweepPositions[i] - lastStart) / 2; ++j)
            {
            double  phi = 2. *
                mPi *
                mFrequency[i] *
                j /
                mIo.SampleRate();
            double  sinphi = sin(phi);
            double  cosphi = cos(phi);

            mtmmty1[0] += 1.;
            mtmmty1[1] += sinphi;
            mtmmty1[2] += cosphi;
            mtmmty1[4] += sinphi;
            mtmmty1[5] += sinphi * sinphi;
            mtmmty1[6] += sinphi * cosphi;
            mtmmty1[8] += cosphi;
            mtmmty1[9] += cosphi * sinphi;
            mtmmty1[10] += cosphi * cosphi;

            unsigned int    si = j + delay + lastStart +
                (mSweepPositions[i] - lastStart) / 2;
            si *= mChannels;
            mtmmty1[3] += ResponseDG[si + refOffset];
            mtmmty1[7] += sinphi * ResponseDG[si + refOffset];
            mtmmty1[11] += cosphi * ResponseDG[si + refOffset];
            if (mChannels == 2)
                {
                mtmmty2[3] += ResponseDG[si + measOffset];
                mtmmty2[7] += sinphi * ResponseDG[si + measOffset];
                mtmmty2[11] += cosphi * ResponseDG[si + measOffset];
                }
            }

        if (mChannels == 2)
            {
            mtmmty2[0] = mtmmty1[0];
            mtmmty2[1] = mtmmty1[1];
            mtmmty2[2] = mtmmty1[2];
            mtmmty2[4] = mtmmty1[4];
            mtmmty2[5] = mtmmty1[5];
            mtmmty2[6] = mtmmty1[6];
            mtmmty2[8] = mtmmty1[8];
            mtmmty2[9] = mtmmty1[9];
            mtmmty2[10] = mtmmty1[10];
            }

        vector < double > x1(3);
        vector < double > x2(3);

        try
            {
            GaussEliminate(3, mtmmty1, x1);
            if (mChannels == 2)
                {
                GaussEliminate(3, mtmmty2, x2);
                }
            }
        catch (const char*str)
            {
            mDebug.Log(str);
            continue;
            }

        complex < double > Href(x1[1], x1[2]);

        double  rmsErr(0.);
        for (j = 0; j < (mSweepPositions[i] - lastStart) / 2; ++j)
            {
            double phi = 2. * mPi * mFrequency[i] * j / mIo.SampleRate();
            unsigned int    si = j + delay + lastStart +
                (mSweepPositions[i] - lastStart) / 2;
            double err = ResponseDG[si * mChannels + refOffset] - sin(phi + arg(Href)) * abs(Href);
            rmsErr += err * err;
            }

        rmsErr /= double((mSweepPositions[i] - lastStart) / 2);
        rmsErr = sqrt(rmsErr);

        lastStart = mSweepPositions[i];

        {
        std::stringstream msg;
        msg << "F=" << mFrequency[i] << " DC=" << x1[0] <<
            " Ref=" << abs(Href) << " Err=" << rmsErr;

        ProcessStatusEvent event(msg.str());
        OnStatus.Execute(event);
        }

        complex<double> ampl;
        if (mChannels == 1)
            {
            ampl = Href;
            }
        else
            {
            complex < double > Hmeas(x2[1], x2[2]);
            ampl = Hmeas / Href;
            }

        mStats.Sample(20. * log10(abs(ampl)));
        mAmplitude.push_back(ampl);
        }

    return (pwrLo > pwrHi) ? pwrLo / double(window) : pwrHi / double(window);
}


} // namespace Innovative

