//
//  TimebaseDevice_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2009 Innovative Integration Corporation.
//  All Rights Reserved.
//


#include <TimebaseDevice_Mb.h>
#include <FloatingPtSupport_Mb.h>
#include <cmath>

namespace Innovative
{


//==============================================================================
//  MultiTimebaseDevice -- Impl to multiple input/output timebase devices
//==============================================================================
//-----------------------------------------------------------------------------
//  constructor for class MultiTimebaseDevice
//-----------------------------------------------------------------------------

MultiTimebaseDevice::MultiTimebaseDevice(unsigned int n_sources, unsigned int n_outputs)
    :  SourceFreqArray(n_sources), OutputFreqArray(n_outputs),
       OutputFreqActualArray(n_outputs), OutputDividerArray(n_outputs),
       OutputInUseArray(n_outputs), OutputEnabledArray(n_outputs),
       SourceInUseArray(n_sources), SourceMappingDirectArray(n_sources),
       SourceMappingReverseArray(n_sources)
{
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::Sources()
//-----------------------------------------------------------------------------

unsigned int  MultiTimebaseDevice::Sources()
{
    return static_cast<unsigned int>(SourceFreqArray.size());
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::Outputs()
//-----------------------------------------------------------------------------

unsigned int  MultiTimebaseDevice::Outputs()
{
    return static_cast<unsigned int>(OutputFreqArray.size());
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::SourceFrequency()
//-----------------------------------------------------------------------------

void  MultiTimebaseDevice::SourceFrequency(unsigned int source, double freq)
{
    SourceFreqArray[source] = freq;
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::SourceFrequency()
//-----------------------------------------------------------------------------

double  MultiTimebaseDevice::SourceFrequency(unsigned int source) const
{
    return SourceFreqArray[source];
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::OutputFrequency()
//-----------------------------------------------------------------------------

void  MultiTimebaseDevice::OutputFrequency(unsigned int output, double freq)
{
    OutputFreqArray[output] = freq;
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::OutputFrequency()
//-----------------------------------------------------------------------------

double  MultiTimebaseDevice::OutputFrequency(unsigned int output) const
{
    return OutputFreqArray[output];
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::OutputFrequencyActual()
//-----------------------------------------------------------------------------

double  MultiTimebaseDevice::OutputFrequencyActual(unsigned int output) const
{
    return OutputFreqActualArray[output];
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::OutputDivider()
//-----------------------------------------------------------------------------

void  MultiTimebaseDevice::OutputDivider(unsigned int output, unsigned int divide_by)
{
    if (CheckDivider(output, divide_by))
        {
        OutputDividerArray[output] = divide_by;
        SetDividerHdw(output, divide_by);
        }
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::OutputDivider()
//-----------------------------------------------------------------------------

unsigned int  MultiTimebaseDevice::OutputDivider(unsigned int output) const
{
    return OutputDividerArray[output];
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::OutputInUse()
//-----------------------------------------------------------------------------

void  MultiTimebaseDevice::OutputInUse(unsigned int output, bool in_use)
{
    OutputInUseArray[output] = in_use;
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::OutputInUse()
//-----------------------------------------------------------------------------

bool  MultiTimebaseDevice::OutputInUse(unsigned int output) const
{
	return OutputInUseArray[output] ? true : false;
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::OutputEnabled()
//-----------------------------------------------------------------------------

void  MultiTimebaseDevice::OutputEnabled(unsigned int output, bool enabled)
{
    OutputEnabledArray[output] = enabled;
    SetOutEnabledHdw(output, enabled);
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::OutputEnabled()
//-----------------------------------------------------------------------------

bool  MultiTimebaseDevice::OutputEnabled(unsigned int output) const
{
    return OutputEnabledArray[output] ? true : false;
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::CheckOutputs() - check if output and actual are "close"
//-----------------------------------------------------------------------------

bool MultiTimebaseDevice::CheckOutputs(double error_rng, unsigned int* first_error)
{
    unsigned int error_chan(-1);
    bool check_failed(false);

    for (std::size_t out=0; out < Outputs();  out++)
        {
        if (OutputInUse(static_cast<unsigned int>(out))==false)
            continue;
		if (IsAboutEqual(OutputFrequency(static_cast<unsigned int>(out)), OutputFrequencyActual(static_cast<unsigned int>(out)), error_rng))
            continue;

        // we found an error
		error_chan = static_cast<unsigned int>(out);
        check_failed = true;
        break;
        }
    if (first_error)
        *first_error = error_chan;

    return check_failed;

}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::SourceInUse()
//-----------------------------------------------------------------------------

void  MultiTimebaseDevice::SourceInUse(unsigned int source, bool in_use)
{
    SourceInUseArray[source] = in_use;
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::SourceInUse()
//-----------------------------------------------------------------------------

bool  MultiTimebaseDevice::SourceInUse(unsigned int source) const
{
    return SourceInUseArray[source] ? true : false;
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::SetSourceMapping()
//-----------------------------------------------------------------------------

void  MultiTimebaseDevice::SetSourceMapping(unsigned int source, unsigned int user_source)
{
    SourceMappingDirectArray[source] = user_source;
    SourceMappingReverseArray[user_source] = source;
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::GetSourceMapping_Reverse()
//-----------------------------------------------------------------------------

unsigned int  MultiTimebaseDevice::GetSourceMapping(unsigned int source)
{
    return SourceMappingDirectArray[source];
}

//-----------------------------------------------------------------------------
//  MultiTimebaseDevice::GetSourceMapping_Reverse()
//-----------------------------------------------------------------------------

unsigned int  MultiTimebaseDevice::GetSourceMapping_Reverse(unsigned int user_source)
{
    return SourceMappingReverseArray[user_source];
}


#ifdef _MSC_VER
    #pragma warning(push)
	#pragma warning(disable : 4355)
#endif
//==============================================================================
//  PllTimebaseDevice -- Impl to PLL multiple i/o timebase devices
//==============================================================================
//-----------------------------------------------------------------------------
//  constructor for class PllTimebaseDevice
//-----------------------------------------------------------------------------

PllTimebaseDevice::PllTimebaseDevice(unsigned int n_references,
                                     unsigned int n_outputs)
    :  MultiTimebaseDevice(n_references, n_outputs),
       FSource(0), FMasterOutput(0),
       UpdateMode(updFrequency),  ShouldAutoReapply(true),
       SysTBR(*this),     AdcTBR(*this),   DacTBR(*this)
{
}

#ifdef _MSC_VER
    #pragma warning(pop)
#endif
//-----------------------------------------------------------------------------
//  PllTimebaseDevice::UpdateState() -- perform one of several update opns
//-----------------------------------------------------------------------------

void  PllTimebaseDevice::UpdateState()
{
    switch (UpdateMode)
        {
        default:
        case updFrequency:
            FrequencyModeUpdate();
            break;
        case updDivider:
            DividerModeUpdate();
            break;
        }
}

//-----------------------------------------------------------------------------
//  PllTimebaseDevice::SelectMasterOutput() -- Choose proper Master Output
//-----------------------------------------------------------------------------

void  PllTimebaseDevice::SelectMasterOutput()
{
    if (AdcRate >= DacRate)
        FMasterOutput = AdcMasterOutput();
    else
        FMasterOutput = DacMasterOutput();
}

//-----------------------------------------------------------------------------
//  PllTimebaseDevice::SysEffectiveFactor() -- Choose proper factor for Sys out
//-----------------------------------------------------------------------------

unsigned int  PllTimebaseDevice::SysEffectiveFactor()
{
    if (AdcRate >= DacRate)
        return AdcEffectiveFactor();
    else
        return DacEffectiveFactor();
}

//-----------------------------------------------------------------------------
//  PllTimebaseDevice::SysMultiplierFactor() -- Choose proper M factor for Sys out
//-----------------------------------------------------------------------------

unsigned int  PllTimebaseDevice::SysMultiplierFactor()
{
    if (AdcRate >= DacRate)
        return AdcMultiplierFactor();
    else
        return DacMultiplierFactor();
}

//==============================================================================
//  PllTimebaseDevice::SystemTimebaseRate -- Changing ADC/DAC rates together
//==============================================================================
//-----------------------------------------------------------------------------
//  PllTimebaseDevice::SystemTimebaseRate::Frequency() - set frequency
//-----------------------------------------------------------------------------

void    PllTimebaseDevice::SystemTimebaseRate::Frequency(double freq)
{
    double multipliedAdcFreq = freq*PTB.AdcMultiplierFactor();
    double multipliedDacFreq = freq*PTB.DacMultiplierFactor();

    // Configure ADC&DAC clock outputs
    PTB.SetAdcOutputRates(multipliedAdcFreq);
    PTB.SetDacOutputRates(multipliedDacFreq);

    double effectiveAdcFreq = multipliedAdcFreq/PTB.AdcEffectiveFactor();
    double effectiveDacFreq = multipliedDacFreq/PTB.DacEffectiveFactor();

    // Set ADC frequency
    PTB.AdcRate = effectiveAdcFreq;
    // Set DAC frequency
    PTB.DacRate = effectiveDacFreq;
    // update FMaster output
    PTB.SelectMasterOutput();

    if (PTB.ShouldAutoReapply)
        PTB.Reapply();
}

//==============================================================================
//  PllTimebaseDevice::AdcTimebaseRate -- Changing ADC rates
//==============================================================================
//-----------------------------------------------------------------------------
//  PllTimebaseDevice::AdcTimebaseRate::Frequency() - set frequency
//-----------------------------------------------------------------------------

void    PllTimebaseDevice::AdcTimebaseRate::Frequency(double freq)
{
    double multipliedAdcFreq = freq*PTB.AdcMultiplierFactor();

    // Configure ADC clock outputs
    PTB.SetAdcOutputRates(multipliedAdcFreq);

    double effectiveFreq = multipliedAdcFreq/PTB.AdcEffectiveFactor();
    // Set ADC frequency
    PTB.AdcRate = effectiveFreq;
    // update FMaster output
    PTB.SelectMasterOutput();

    if (PTB.ShouldAutoReapply)
        PTB.Reapply();
}

//==============================================================================
//  PllTimebaseDevice::DacTimebaseRate -- Changing DAC rates
//==============================================================================
//-----------------------------------------------------------------------------
//  PllTimebaseDevice::DacTimebaseRate::Frequency() - set frequency
//-----------------------------------------------------------------------------

void    PllTimebaseDevice::DacTimebaseRate::Frequency(double freq)
{
    double multipliedDacFreq = freq*PTB.DacMultiplierFactor();
    // Configure DAC clock outputs
    PTB.SetDacOutputRates(multipliedDacFreq);

    double effectiveFreq = multipliedDacFreq/PTB.DacEffectiveFactor();
    // Set DAC frequency
    PTB.DacRate = effectiveFreq;
    // update FMaster output
    PTB.SelectMasterOutput();

    if (PTB.ShouldAutoReapply)
        PTB.Reapply();
}

//==============================================================================
//  MuxTimebaseDevice -- Impl to MUX multiple i/o timebase devices
//==============================================================================
//-----------------------------------------------------------------------------
//  constructor for class MuxTimebaseDevice
//-----------------------------------------------------------------------------

MuxTimebaseDevice::MuxTimebaseDevice(unsigned int n_sources,
                                     unsigned int n_outputs)
    :  MultiTimebaseDevice(n_sources, n_outputs),
       OutputMuxArray(n_outputs), UpdateMode(updDivider)
{
}

//-----------------------------------------------------------------------------
//  MuxTimebaseDevice::UpdateState() -- perform one of several update opns
//-----------------------------------------------------------------------------

void  MuxTimebaseDevice::UpdateState()
{
    switch (UpdateMode)
        {
        default:
        case updDivider:
            DividerModeUpdate();
            break;
        case updFrequency:
            FrequencyModeUpdate();
            break;
        }
}

//-----------------------------------------------------------------------------
//  MuxTimebaseDevice::DividerModeUpdate() -- Update state of Mux
//-----------------------------------------------------------------------------
//  In this mode, we assume the input divider values are correct and set
//     the output frequencies to match

void  MuxTimebaseDevice::DividerModeUpdate()
{
    for (unsigned int out=0; out<OutputFreqActualArray.size(); out++)
        {
        //  Obtain the current Source Frequency
        double In = SourceFreqArray[OutputMuxArray[out]];
        if (OutputDividerArray[out]==0)
            continue;    // illegal divider, skip it

        // Frequency Actual is Source divided by divider
        OutputFreqActualArray[out] = In / OutputDividerArray[out];
        }
}

//-----------------------------------------------------------------------------
//  MuxTimebaseDevice::FrequencyModeUpdate() -- Update state of Mux
//-----------------------------------------------------------------------------
//  In this mode, we assume the output frequency values are correct and set
//     the output dividers to match

void  MuxTimebaseDevice::FrequencyModeUpdate()
{
    for (unsigned int out=0; out<OutputFreqActualArray.size(); out++)
        {
        //  Obtain the current Source Frequency
        double In = SourceFreqArray[OutputMuxArray[out]];
        if (OutputFreqArray[out]==0)
            continue;    // frequency not set, skip it
        // Calculate a possible divider
        float candidateDiv = static_cast<float>(In / OutputFreqArray[out]);

        int above, below, divider;
        GetNearestDividers(out, candidateDiv, below, above);

        // Find which divider is closest to desired frequency
        double centerFreq = In/candidateDiv;
        double hiFreq     = In/below;
        double lowFreq    = In/above;
        if (std::abs(centerFreq-lowFreq) > std::abs(hiFreq-centerFreq))
            divider = below;
        else
            divider = above;

        OutputDivider(out, divider);
        // Frequency Actual is Source divided by divider
        OutputFreqActualArray[out] = In / OutputDividerArray[out];
        }
}

} // namespace Innovative

