//
//  TimebaseDevice_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2009 Innovative Integration Corporation.
//  All Rights Reserved.
//

#ifndef TimebaseDevice_MbH
#define TimebaseDevice_MbH

#include <vector>

//---------------------------------------------------------------------------

namespace Innovative
{

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif

//==============================================================================
//  ITimebaseRate -- Interface to set/read a timebase rate (1 or more)
//==============================================================================

class ITimebaseRate
{
public:
	virtual ~ITimebaseRate()  {}

	virtual void    Frequency(double freq) = 0;
	virtual double  Frequency() const = 0;
	virtual double  FrequencyActual() const = 0;

    virtual double  EffectiveFrequency() const = 0;
    virtual double  EffectiveFrequencyActual() const = 0;

    virtual double  MultipliedFrequency() const = 0;
    virtual double  MultipliedFrequencyActual() const = 0;

};

//==============================================================================
//  IMultiTimebaseDevice -- Interface to multiple input/output timebase devices
//==============================================================================

class IMultiTimebaseDevice
{
public:
    virtual ~IMultiTimebaseDevice()  {}

    virtual unsigned int  Sources() = 0;
    virtual unsigned int  Outputs() = 0;

    virtual void    SourceFrequency(unsigned int source, double freq) = 0;
    virtual double  SourceFrequency(unsigned int source) const = 0;

    virtual void    OutputFrequency(unsigned int output, double freq) = 0;
    virtual double  OutputFrequency(unsigned int output) const = 0;

    virtual double  OutputFrequencyActual(unsigned int output) const = 0;

    virtual void          OutputDivider(unsigned int output, unsigned int divide_by) = 0;
    virtual unsigned int  OutputDivider(unsigned int output) const = 0;

    virtual void          OutputEnabled(unsigned int output, bool enabled) = 0;
    virtual bool          OutputEnabled(unsigned int output) const = 0;

};

//==============================================================================
//  MultiTimebaseDevice -- Impl to multiple input/output timebase devices
//==============================================================================

class MultiTimebaseDevice : public IMultiTimebaseDevice
{
public:
    MultiTimebaseDevice(unsigned int n_sources, unsigned int n_outputs);
    virtual ~MultiTimebaseDevice()  {}

    virtual unsigned int  Sources();
    virtual unsigned int  Outputs();

    virtual void    SourceFrequency(unsigned int source, double freq);
    virtual double  SourceFrequency(unsigned int source) const;

    virtual void    OutputFrequency(unsigned int output, double freq);
    virtual double  OutputFrequency(unsigned int output) const;

    virtual double  OutputFrequencyActual(unsigned int output) const;

    virtual void          OutputDivider(unsigned int output, unsigned int divide_by);
    virtual unsigned int  OutputDivider(unsigned int output) const;

    virtual void          OutputEnabled(unsigned int output, bool enabled);
    virtual bool          OutputEnabled(unsigned int output) const;

    bool            CheckOutputs(double error_rng, unsigned int * first_error = 0);

protected:
    std::vector<double>  SourceFreqArray;
    std::vector<double>  OutputFreqArray;
    std::vector<double>  OutputFreqActualArray;

    std::vector<unsigned int>  OutputDividerArray;
    std::vector<int>           OutputInUseArray;
    std::vector<int>           OutputEnabledArray;

    std::vector<int>           SourceInUseArray;
    std::vector<int>           SourceMappingDirectArray;
    std::vector<int>           SourceMappingReverseArray;


public:
    virtual void       SourceInUse(unsigned int source, bool in_use);
    virtual bool       SourceInUse(unsigned int source) const;

    virtual void           SetSourceMapping(unsigned int source, unsigned int user_source);
    virtual unsigned int   GetSourceMapping(unsigned int source);
    virtual unsigned int   GetSourceMapping_Reverse(unsigned int user_source);

protected:
    virtual void       OutputInUse(unsigned int output, bool in_use);
    virtual bool       OutputInUse(unsigned int output) const;

    //  Internal Update Interface
    void  Update()
        {
        UpdateSources();
        UpdateState();
        UpdateOutputs();
        }
    virtual void  UpdateSources()
        {}
    virtual void  UpdateState()
        {}
    virtual void  UpdateOutputs()
        {}
    //
    //  Derived Class
    virtual void  SetDividerHdw(unsigned int output, unsigned int divide_by) = 0;
    virtual bool  CheckDivider(unsigned int output, unsigned int divide_by) = 0;
    virtual void  GetNearestDividers(unsigned int output, float candidate,
                                    int & below,
                                    int & above) = 0;
    virtual void  SetOutEnabledHdw(unsigned int output, bool enabled) = 0;

};

//==============================================================================
//  PllTimebaseDevice -- Impl to PLL multiple i/o timebase devices
//==============================================================================

class PllTimebaseDevice : public MultiTimebaseDevice, public ITimebaseRate
{
public:

	enum PllUpdateMode  { updDivider, updFrequency };

    PllTimebaseDevice(unsigned int n_references, unsigned int n_outputs);
    virtual ~PllTimebaseDevice()  {}

    virtual void   Init() = 0;
    virtual void   Reapply() = 0;

    void    AutoReapply(bool state)
                {  ShouldAutoReapply=state;  }
//    bool    AutoReapply()
//                {  return ShouldAutoReapply;  }

    void    ReferenceFrequency(unsigned int source, double freq)
        {  SourceFrequency(source, freq);  }
    double  ReferenceFrequency(unsigned int source) const
        {  return SourceFrequency(source);  }
    double  Reference() const
        {  return SourceFrequency(FSource);  }
    void    Reference(double freq)
        {  SourceFrequency(FSource, freq) ;  }

    void    Frequency(double freq)
        {  SysTBR.Frequency(freq);  }
    double  Frequency() const
        {  return SysTBR.Frequency();  }
    double  FrequencyActual() const
        {  return SysTBR.FrequencyActual();  }
    double  EffectiveFrequency() const
        {  return SysTBR.EffectiveFrequency();  }
    double  EffectiveFrequencyActual() const
        {  return SysTBR.EffectiveFrequencyActual();  }
    double  MultipliedFrequency() const
        {  return SysTBR.MultipliedFrequency();  }
    double  MultipliedFrequencyActual() const
        {  return SysTBR.MultipliedFrequencyActual();  }
    double  MasterOutputFrequency() const  // Freq to calculate PLL on, with all factors included.
        {  return OutputFrequency(MasterOutput());  }

    ITimebaseRate & Adc()   {  return AdcTBR;  }
    ITimebaseRate & Dac()   {  return DacTBR;  }
    ITimebaseRate & System() {  return SysTBR;  }

    void  SetMasterOutputSelection(unsigned int output)
        {  FMasterOutput = output;  }

	virtual unsigned int AdcMasterOutput() = 0;
	virtual unsigned int DacMasterOutput() = 0;


    unsigned int MasterOutput() const {  return FMasterOutput;  }
    //
    //  Clock Rate Modifiers - ratios between "Frequency" and real OutputFrequency
    virtual unsigned int AdcEffectiveFactor() {  return 1;  }   // device needs lower rate than sample rate
    virtual unsigned int DacEffectiveFactor() {  return 1;  }
    unsigned int SysEffectiveFactor();

    virtual unsigned int AdcMultiplierFactor() {  return 1;  }   // "Clock Factor" - device needs faster rate than sample rate
    virtual unsigned int DacMultiplierFactor() {  return 1;  }
    unsigned int SysMultiplierFactor();

protected:
	unsigned int   FSource;
    unsigned int   FMasterOutput;
    PllUpdateMode  UpdateMode;
    bool           ShouldAutoReapply;

    void  SetSourceSelection(unsigned int source)
        {  FSource = source;  }
    unsigned int  GetSourceSelection()
        {  return FSource;  }

    virtual void  UpdateState();
    //  Mode Update types
    virtual void  FrequencyModeUpdate() = 0;
    virtual void  DividerModeUpdate() = 0;

	virtual void  SetAdcOutputRates(double adc_rate) = 0;
    virtual void  SetDacOutputRates(double dac_rate) = 0;

    void  SelectMasterOutput();


    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//  CLASS PllTimebaseDevice::SystemTimebaseRate - combined
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	class SystemTimebaseRate : public ITimebaseRate
    {
    public:
        SystemTimebaseRate( PllTimebaseDevice &ptb )
            :  PTB(ptb)//,
            {
            }

        void    Frequency(double freq);
        double  Frequency() const
              {  return EffectiveFrequency() * PTB.SysEffectiveFactor() / PTB.SysMultiplierFactor();  }
        double  FrequencyActual() const
              {  return EffectiveFrequencyActual() * PTB.SysEffectiveFactor() / PTB.SysMultiplierFactor();  }
        double  EffectiveFrequency() const
              {  return PTB.OutputFrequency(PTB.MasterOutput());  }
        double  EffectiveFrequencyActual() const
              {  return PTB.OutputFrequencyActual(PTB.MasterOutput());  }
        double  MultipliedFrequency() const
              {  return PTB.OutputFrequency(PTB.MasterOutput()) * PTB.SysEffectiveFactor();  }
        double  MultipliedFrequencyActual() const
              {  return PTB.OutputFrequencyActual(PTB.MasterOutput()) * PTB.SysEffectiveFactor();  }
    private:
        PllTimebaseDevice &PTB;
    };
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS PllTimebaseDevice::AdcTimebaseRate - adc timebase rate
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	class AdcTimebaseRate : public ITimebaseRate
    {
    public:
        AdcTimebaseRate( PllTimebaseDevice &ptb )
            :  PTB(ptb)//,
            {
            }

        void    Frequency(double freq);
        double  Frequency() const
              {  return EffectiveFrequency() * PTB.AdcEffectiveFactor() / PTB.AdcMultiplierFactor();  }
        double  FrequencyActual() const
              {  return EffectiveFrequencyActual() * PTB.AdcEffectiveFactor() / PTB.AdcMultiplierFactor();  }
        double  EffectiveFrequency() const
              {  return PTB.OutputFrequency(PTB.AdcMasterOutput());  }
        double  EffectiveFrequencyActual() const
              {  return PTB.OutputFrequencyActual(PTB.AdcMasterOutput());  }
        double  MultipliedFrequency() const
              {  return PTB.OutputFrequency(PTB.AdcMasterOutput()) * PTB.AdcEffectiveFactor();  }
        double  MultipliedFrequencyActual() const
              {  return PTB.OutputFrequencyActual(PTB.AdcMasterOutput()) * PTB.AdcEffectiveFactor();  }
    private:
        PllTimebaseDevice &PTB;
    };

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS PllTimebaseDevice::DacTimebaseRate - dac timebase rate
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	class DacTimebaseRate : public ITimebaseRate
    {
    public:
        DacTimebaseRate( PllTimebaseDevice &ptb )
            :  PTB(ptb)//,
            {
            }
        void    Frequency(double freq);
        double  Frequency() const
              {  return EffectiveFrequency() * PTB.DacEffectiveFactor() / PTB.DacMultiplierFactor();  }
        double  FrequencyActual() const
              {  return EffectiveFrequencyActual() * PTB.DacEffectiveFactor() / PTB.DacMultiplierFactor();  }
        double  EffectiveFrequency() const
              {  return PTB.OutputFrequency(PTB.DacMasterOutput());  }
        double  EffectiveFrequencyActual() const
              {  return PTB.OutputFrequencyActual(PTB.DacMasterOutput());  }
        double  MultipliedFrequency() const
              {  return PTB.OutputFrequency(PTB.DacMasterOutput()) * PTB.DacEffectiveFactor();  }
        double  MultipliedFrequencyActual() const
              {  return PTB.OutputFrequencyActual(PTB.DacMasterOutput()) * PTB.DacEffectiveFactor();  }

    private:
        PllTimebaseDevice &PTB;
    };
	friend class SystemTimebaseRate;
	friend class AdcTimebaseRate;
	friend class DacTimebaseRate;
	//
    //  Timebase Rate Interfaces
    SystemTimebaseRate  SysTBR;
    AdcTimebaseRate     AdcTBR;
    DacTimebaseRate     DacTBR;
    double              AdcRate;
    double              DacRate;
};

//==============================================================================
//  MuxTimebaseDevice -- Impl to MUX type multiple i/o timebase devices
//==============================================================================

class MuxTimebaseDevice : public MultiTimebaseDevice
{
public:
    enum MuxUpdateMode  { updDivider, updFrequency };

    MuxTimebaseDevice(unsigned int n_sources, unsigned int n_outputs);

    virtual unsigned int  InputSource(unsigned int output) const
        {  return OutputMuxArray[output];  }
    virtual void          InputSource(unsigned int output, unsigned int source)
        {  SetMuxSelection(output, source);   }


protected:
    std::vector<unsigned int>  OutputMuxArray;
    MuxUpdateMode  UpdateMode;

    void  SetMuxSelection(unsigned int output, unsigned int source)
        {  OutputMuxArray[output] = source;  SetMuxHdw(output, source); }

    virtual void  UpdateState();
    //  Mode Update types
    virtual void  DividerModeUpdate();
    virtual void  FrequencyModeUpdate();

    virtual void  SetMuxHdw(unsigned int output, unsigned int source) = 0;
};


#ifdef _MSC_VER
	#pragma warning(pop)
#endif

}  // end namespace Innovative
#endif
