//---------------------------------------------------------------------------
// Trace_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.

#ifndef Trace_MbH
#define Trace_MbH

#include <string>
#include <DataTypes_Mb.h>
#include <ThreadSafeQueue_Mb.h>

namespace InnovativeKernel {
    namespace Osal {
        class OsSupport;
    }
}

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//==============================================================================
//  CLASS Trace -- Persistent trace log
//==============================================================================
//  Verbosity Level is matched against Trace's Priority Level to determine if a
//     message will be output. If Priority <= Verbosity, then it will be printed
//
//  Each Trace's effective Priority can be adjusted..the default cannot.
//  The system Verbosity is set by TraceVerbosity().
//
class Trace
{
public:
    enum VerbosityLevels { vSilent=0, vOnlyImportant=2, vNormal=5, vEvenTrivial=7, vAll=10 };
    enum PriorityLevels { pMaximum=1, pImportant=2, pNormal=5, pTrivial=7, pMinimum=10 };

    friend void         TraceVerbosity(unsigned int level);
    friend unsigned int TraceVerbosity();

	Trace(bool enabled=true,  enum PriorityLevels default_priority=pNormal);

    //  Properties
    void            Enabled(bool state);
    bool            Enabled() const;

    void            Prefix(const std::string & value);
    std::string     Prefix() const;

    //  Priority Management
    void            PushPriority(unsigned int priority);
    void            PopPriority();
    unsigned int    Priority() const;
    bool            WillTrace() const;

    // Methods
    void            Log(const std::string & s) const;
    void            Log(const std::string & s, int value, int radix=10) const;
    void            Log(const std::string & s, unsigned int value, int radix=10) const
                        {  Log(s, static_cast<int>(value), radix);  }
	void            Log(const std::string & s, ii64 value, int radix=10) const;
    void            Log(const std::string & s, s_ii64 value, int radix=10) const;
    void            Log(const std::string & s, std::string y) const;
    void            Log(const std::string & s, float value) const;
	void            Log(const std::string & s, double value) const;
	void            Log(const std::string & s, bool state) const;

    unsigned int    Verbosity() const;

private:
    // Fields
    bool            FEnabled;
    bool            FAutoCr;
    std::string     FPrefix;

    typedef  ThreadSafeStack<unsigned int> PriorityStack;

    const unsigned int    FDefaultPriority;
    static unsigned int   FVerbosity;
    PriorityStack         FPriority;

    // Methods
    void            Spit(const std::string & s) const;

	const InnovativeKernel::Osal::OsSupport & LogOutput() const;

};

//==============================================================================
//  CLASS TracePriority -- Scope of object reported with Trace log
//==============================================================================

class TracePriority
{
public:
    TracePriority(const Trace & trace, int priority_level)
        :  Debug(const_cast<Trace &>(trace))
        {  Debug.PushPriority(priority_level);  }
    TracePriority(Trace & trace, int priority_level)
        :  Debug(trace)
        {  Debug.PushPriority(priority_level);  }
    ~TracePriority()
        {  Debug.PopPriority();  }

    bool  WillTrace()
        {  return  Debug.WillTrace();  }
    int  Priority()
        {  return  Debug.Priority();  }


private:
    Trace &     Debug;
};

//==============================================================================
//  CLASS ScopedTrace -- Scope of object reported with Trace log
//==============================================================================

class ScopedTrace
{
public:
    ScopedTrace(const std::string &text);
    ~ScopedTrace();

private:
    std::string Text;
    Trace       Debug;
};

//==============================================================================
//  CLASS ScopedTraceRef-- Scope of object reported with Trace log
//==============================================================================

class ScopedTraceRef
{
public:
    ScopedTraceRef(Trace & debug, const std::string &text);
    ~ScopedTraceRef();

private:
    std::string Text;
    Trace &     Debug;
};

//---------------------------------------------------------------------------
//  TraceVerbosity() --  Access the global current verbosity level
//---------------------------------------------------------------------------

void         TraceVerbosity(unsigned int level);
unsigned int TraceVerbosity();


#ifdef __CLR_VER
#pragma managed(pop)
#endif
}  // namespace

#endif
