// Tuner_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2009 Innovative Integration Corporation.
//  All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <Tuner_Mb.h>
#include <SystemSupport_Mb.h>
#include <Trace_Mb.h>

#ifdef _MSC_VER
    #pragma warning(push)
    #pragma warning(disable : 4355)
#endif

namespace Innovative
{

//===========================================================================
//  CLASS ITuner  -- Interface to Tuner Hardware
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class ITuner
//---------------------------------------------------------------------------

ITuner::ITuner()
	: Uart(0), Parser(new StandardTunerMessageParser), FEnabled(false)
{
}

//---------------------------------------------------------------------------
//  constructor for class ITuner (takes uart)
//---------------------------------------------------------------------------

ITuner::ITuner(IMulticharUart * uart)
	: Uart(uart), Parser(new StandardTunerMessageParser), FEnabled(false)
{
}

//---------------------------------------------------------------------------
//  constructor for class ITuner (takes uart & parser)
//---------------------------------------------------------------------------

ITuner::ITuner(IMulticharUart * uart, ITunerMessageParser * parser)
	: Uart(uart), Parser(parser), FEnabled(false)
{
}

//---------------------------------------------------------------------------
//  destructor for class ITuner
//---------------------------------------------------------------------------

ITuner::~ITuner()
{
	Uart = 0;   // we don't own the Uart
	delete Parser;  Parser = 0;
}

//---------------------------------------------------------------------------
//  ITuner::Init() -  "Open" time specific operations
//---------------------------------------------------------------------------

void  ITuner::Init()
{
	OnInit();  // unit specific operations
}

//---------------------------------------------------------------------------
//  ITuner::Reapply() -  "StreamStart" time specific operations
//---------------------------------------------------------------------------

void  ITuner::Reapply()
{
	OnReapply();  // unit specific operations
}

//---------------------------------------------------------------------------
//  ITuner::Close() -  "Close" time specific operations
//---------------------------------------------------------------------------

void  ITuner::Close()
{
	OnClose();   // unit specific operations
}

//---------------------------------------------------------------------------
//  ITuner::Command() --  Send Command string out serial port
//---------------------------------------------------------------------------

void ITuner::Command(const std::string & txt)
{
	if (!FEnabled || Uart==0 || Parser==0)
		return;

	Uart->PutString(txt);
	//
    //  Finish with the ending token
	std::string CmdEndToken = Parser->EndToken();
	Uart->PutString(CmdEndToken);
}

//---------------------------------------------------------------------------
//  ITuner::Command() --  Send Command string out serial port
//---------------------------------------------------------------------------


void ITuner::Command(const std::string & txt, std::string &OutputStr, unsigned int delay)
{
    if (!FEnabled || Uart==0 )
        return;

    Uart->PutString(txt);
    //
    //  Finish with the ending token
    std::string CmdEndToken = Parser->EndToken();
    Uart->PutString(CmdEndToken);

    Sleep(delay);

    Uart->GetString(OutputStr);
}

//---------------------------------------------------------------------------
//  ITuner::CustomParser() --  replace Parser with a new one
//---------------------------------------------------------------------------

void ITuner::CustomParser(ITunerMessageParser * newParser)
{
    delete Parser;
    Parser = newParser;
}

//---------------------------------------------------------------------------
//  ITuner::DispatchTunerInterrupt() -- Process a GPS Interrupt
//---------------------------------------------------------------------------

void ITuner::DispatchTunerInterrupt()
{
	if (!FEnabled || Uart==0 || Parser==0)
		return;

	TunerMessageList  Messages;

	// Read string from UART
	std::string interrupt_text;
	Uart->GetString(interrupt_text);

	if (!interrupt_text.empty())
		{
		Queue += interrupt_text;
		//
        //  Parse the messages from queue
        Parser->Parse(Queue, Messages);
        //
        //  Send Message List to derived class for processing
        OnProcessMessages(Messages);
		}

    // Notify application code
	TunerOutputEvent e(Messages);
	OnOutput.Execute(e);
}

//===========================================================================
// CLASS StandardTunerMessageParser  --  Normal Message String Parser
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class StandardTunerMessageParser
//---------------------------------------------------------------------------

StandardTunerMessageParser::StandardTunerMessageParser()
	:  CmdEndToken("\r\n")
{
}

//---------------------------------------------------------------------------
//  constructor for class StandardTunerMessageParser
//---------------------------------------------------------------------------

StandardTunerMessageParser::StandardTunerMessageParser(
							 const std::string & end_token)
	:  CmdEndToken(end_token)
{
}

//---------------------------------------------------------------------------
//  StandardTunerMessageParser::Parse() -- extract text from text, make messages
//---------------------------------------------------------------------------

void  StandardTunerMessageParser::Parse(std::string & Queue, TunerMessageList & Messages)
{
    //  Check if no end tokens are found, end at once
    size_t eol = Queue.find(CmdEndToken);
	if (eol == std::string::npos)
        return;

    //  Each string found is parsed to make a Message
	while (eol != std::string::npos)
        {
        std::string message_text = Queue.substr(0, eol);
        ParseMessage(message_text, Messages );
		//  purge the found message and check for more
        Queue.erase(0, eol+CmdEndToken.size());
        eol = Queue.find(CmdEndToken);
        }
	// end when last message is found
}

//---------------------------------------------------------------------------
//  StandardTunerMessageParser::ParseMessage() -- from text, make message
//---------------------------------------------------------------------------

void  StandardTunerMessageParser::ParseMessage(std::string text, TunerMessageList & Messages)
{
	//  Right now we don't do any argument parsing, so add the message
	//  Add on any leftovers as final argument
	if (!text.empty())
		Messages.push_back(text);
}


}
