// Units_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2017 Innovative Integration Corporation.
//  All Rights Reserved.
//
#include <Units_Mb.h>
#include <algorithm>

//using namespace std;

namespace   Innovative
{
using namespace Frequency_Impl;

//===========================================================================
//  CLASS Frequency_Units  -- Frequency class with flexible units
//===========================================================================
//---------------------------------------------------------------------------
//  Frequency_Units::Text() - canonical text representation
//---------------------------------------------------------------------------

std::string Frequency_Units::Text(int precision)
{
    return DoubleToString(number, precision) + Name(unit);
}

//---------------------------------------------------------------------------
//  Frequency_Units::Value() - return "natural" value of the frequency
//---------------------------------------------------------------------------

double Frequency_Units::Value()
{
    return number * CFactor(unit);
}

//---------------------------------------------------------------------------
//  Frequency_Units::AsUnit() - return value in the given units
//---------------------------------------------------------------------------

double Frequency_Units::AsUnit(Types a_unit)
{
    return Value() / CFactor(a_unit);
}

//---------------------------------------------------------------------------
//  Frequency_Units::AskHz() - return value in kHz
//---------------------------------------------------------------------------

double Frequency_Units::AskHz()
{
    return Value() / unit_info<kHz>::CFactor();
}

//---------------------------------------------------------------------------
//  Frequency_Units::AsMHz() - return value in MHz
//---------------------------------------------------------------------------

double Frequency_Units::AsMHz()
{
    return Value() / unit_info<MHz>::CFactor();
}

//---------------------------------------------------------------------------
//  Frequency_Units::AsGHz() - return value in GHz
//---------------------------------------------------------------------------

double Frequency_Units::AsGHz()
{
    return Value() / unit_info<GHz>::CFactor();
}


//---------------------------------------------------------------------------
//  Frequency_Units::Normalize() - change unit to minimize 'number' but > 1.
//---------------------------------------------------------------------------

void  Frequency_Units::Normalize()
{
    Types newUnit = unit;
    double newNumber = number;

    for (int i=0; i<Frequency_Count; i++)
        {
        Types type(static_cast<Types>(i));
        double cur_number = AsUnit(type);
        if (cur_number < 1.0)
            continue;

        // We found a legal number, see if it the best
        if (newNumber > cur_number || newNumber < 1.0)
            {
            newNumber = cur_number;
            newUnit = type;
            }
        }
   *this = ObjectAs(newUnit);
}

//---------------------------------------------------------------------------
//  Frequency_Units::ObjectAs() - return object with same value, new unit
//---------------------------------------------------------------------------

Frequency_Units  Frequency_Units::ObjectAs(Types a_unit)
    {
    Frequency_Units  NewF(0.0, a_unit);
    NewF.Number( AsUnit(a_unit) );
    return NewF;
    }

//---------------------------------------------------------------------------
//  Frequency_Units::print() - canonical "dump" of info as text
//---------------------------------------------------------------------------

std::string Frequency_Units::print()
{
    return Text(3);
}

//---------------------------------------------------------------------------
//  Frequency_Units::CFactor() - select CFactor from valid units
//---------------------------------------------------------------------------

double Frequency_Units::CFactor(Types a_unit)
{
    switch (a_unit)
        {
        default:
        case Hz:  return unit_info<Hz>::CFactor();
        case kHz: return unit_info<kHz>::CFactor();
        case MHz: return unit_info<MHz>::CFactor();
        case GHz: return unit_info<GHz>::CFactor();
        }
}

//---------------------------------------------------------------------------
//  Frequency_Units::Name() - select CFactor from valid units
//---------------------------------------------------------------------------

std::string Frequency_Units::Name(Types a_unit)
{
    switch (a_unit)
        {
        default:
        case Hz:  return unit_info<Hz>::Name();
        case kHz: return unit_info<kHz>::Name();
        case MHz: return unit_info<MHz>::Name();
        case GHz: return unit_info<GHz>::Name();
        }
}

//---------------------------------------------------------------------------
//  Frequency_Units::FillUnitNameList() - select CFactor from valid units
//---------------------------------------------------------------------------

void Frequency_Units::FillUnitNameList(std::vector<std::string> & list)
{
    list.clear();
    for (int i=0; i<Frequency_Count; i++)
        {
        Types type(static_cast<Types>(i));
        std::string unit_name = Name(type);
        list.push_back(unit_name);
        }

}

//---------------------------------------------------------------------------
//  Frequency_Units::ParseString() - fill string data from string rep
//---------------------------------------------------------------------------

void Frequency_Units::ParseString(const std::string & str)
{
    std::string local(str);
    // check for valid units string
    std::transform(local.begin(), local.end(), local.begin(), ::tolower);

    //  order of parsing the Types. Need to assure a short string
    Types type_order[] = { GHz, MHz, kHz, Hz };

    //  find the unit
    for (int i=0; i<Frequency_Count; i++)
        {
        //  Find units string
        Types type = type_order[i];
        std::string unit_name = Name(type);

        std::string unit_text = unit_name.substr(1);  // search for unit tag w/o space
        std::transform(unit_text.begin(), unit_text.end(), unit_text.begin(), ::tolower);

        size_t pos = local.find(unit_text);
        if (pos != std::string::npos)
            {
            // found it!
            Units(type);
            break;
            }

        }

    //  find the "number"
    double number = StringToDouble(str);
    Number(number);

}

//===========================================================================
//  CLASS U64_as_Hex  --
//===========================================================================

//---------------------------------------------------------------------------
//  U64_as_Hex::ParseString() - extract data from a string representation
//---------------------------------------------------------------------------

void U64_as_Hex::ParseString(const std::string & str)
{
    Data = StringToHex64(str);

}

//---------------------------------------------------------------------------
//  U64_as_Hex::Text() - Return string rep of class
//---------------------------------------------------------------------------

std::string U64_as_Hex::Text()
{
    return Int64ToString(Data, 16);
}


//===========================================================================
//  CLASS U64_as_dec  --
//===========================================================================

//---------------------------------------------------------------------------
//  U32_as_dec::ParseString() - extract data from a string representation
//---------------------------------------------------------------------------

void U64_as_dec::ParseString(const std::string & str)
{
    Data = StringToDec64(str);
}

//---------------------------------------------------------------------------
//  U64_as_dec::Text() - Return string rep of class
//---------------------------------------------------------------------------

std::string U64_as_dec::Text()
{
    return Int64ToString(Data);
}

//===========================================================================
//  CLASS U32_as_Hex  --
//===========================================================================

//---------------------------------------------------------------------------
//  U32_as_Hex::ParseString() - extract data from a string representation
//---------------------------------------------------------------------------

void U32_as_Hex::ParseString(const std::string & str)
{
    Data = StringToHex(str);

}

//---------------------------------------------------------------------------
//  U64_as_Hex::Text() - Return string rep of class
//---------------------------------------------------------------------------

std::string U32_as_Hex::Text()
{
    return IntToString(Data, 16);
}

//===========================================================================
//  CLASS U32_as_dec  --
//===========================================================================

//---------------------------------------------------------------------------
//  U32_as_dec::ParseString() - extract data from a string representation
//---------------------------------------------------------------------------

void U32_as_dec::ParseString(const std::string & str)
{
    Data = StringToUDec(str);

}

//---------------------------------------------------------------------------
//  U64_as_Hex::Text() - Return string rep of class
//---------------------------------------------------------------------------

std::string U32_as_dec::Text()
{
    return UIntToString(Data);
}

//===========================================================================
//  CLASS S64_as_dec  --
//===========================================================================

//---------------------------------------------------------------------------
//  S64_as_dec::ParseString() - extract data from a string representation
//---------------------------------------------------------------------------

void S64_as_dec::ParseString(const std::string & str)
{
    Data = StringToSignedDec64(str);

}

//---------------------------------------------------------------------------
//  U64_as_Hex::Text() - Return string rep of class
//---------------------------------------------------------------------------

std::string S64_as_dec::Text()
{
    return Int64ToString(Data, 10);
}

} // namespace Innovative

