// Units_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure 
//  agreement with Innovative Integration Corporation and may not be copied or 
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2017 Innovative Integration Corporation. 
//  All Rights Reserved.
//

#ifndef Units_MbH
#define Units_MbH

#include <string>
#include <vector>
#include <StringSupport_Mb.h>
#include <IniFile_Mb.h>
#include <DataTypes_Mb.h>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//===========================================================================
//  CLASS Frequency_Units  -- Frequency class with flexible units
//===========================================================================
//---------------------------------------------------------------------------
//  Frequency_Impl namespace - definitions for Frequency_Units class
//---------------------------------------------------------------------------

namespace Frequency_Impl
{
    enum Types { Hz, kHz, MHz, GHz, Frequency_Count  };

} // end namespace Frequency_Impl

//---------------------------------------------------------------------------
//  declaration for class Frequency_Units
//---------------------------------------------------------------------------

class Frequency_Units
{

public:
    Frequency_Units(double num=0.0, Frequency_Impl::Types aUnit=Frequency_Impl::Hz)
        :  number(num), unit(aUnit)
        {
        if (unit>=Frequency_Impl::Frequency_Count)
            unit = Frequency_Impl::Hz;  // "clip" invalid units to standard
        }
    Frequency_Units(const std::string & rep)  // default arg for default unit?
        :  number(0.0), unit(Frequency_Impl::Hz)
        {
        ParseString(rep);
        }

    // use standard copy, assignment ops
    

    // Gui Loader specific equivalence...any rep change gets applied
    bool  GuiCacheEquivalent(Frequency_Units other)
        {  return (number==other.number) && (unit==other.unit);  }   

    // value, unit read
    double                 Number()  {  return number;  }
    void                   Number(double a_number)  {  number = a_number;  }
    Frequency_Impl::Types  Units()  {  return unit;  }
    void                   Units(Frequency_Impl::Types a_unit)  {  unit = a_unit;  }
    // convert from standard, canonical string format
    std::string            Text(int precision=1);  // field with # digits after fraction?
    //void                   Text(const std::string & string_rep );

    // base value (in "normal" units
    double              Value();
    //void                Value(double value);

    double              AsUnit(Frequency_Impl::Types a_unit);
    double              AsHz()  {  return Value();  }
    double              AskHz();
    double              AsMHz();
    double              AsGHz();

    void                Normalize();  // change unit where number is >1, but smallest

    Frequency_Units     ObjectAs(Frequency_Impl::Types a_unit);

    std::string print();

    //
    //  typeinfo "database" lookup
    static double       CFactor(Frequency_Impl::Types a_unit);   // make static member function?
    static std::string  Name(Frequency_Impl::Types x);

    static void         FillUnitNameList(std::vector<std::string>& list);

private:
    double                   number;
    Frequency_Impl::Types    unit;

    void  ParseString(const std::string & str);
};

// standard value using comparisons...

inline bool operator == (Frequency_Units left, Frequency_Units right)
{
    return left.Value() == right.Value();

}

inline bool operator < (Frequency_Units left, Frequency_Units right)
{
    return left.Value() < right.Value();
}


inline bool operator != (Frequency_Units left, Frequency_Units right)
{
    return ! (left == right);

}

inline bool operator > (Frequency_Units left, Frequency_Units right)
{
    return right < left;
}


inline bool operator <= (Frequency_Units left, Frequency_Units right)
{
    return ! (left > right);
}


inline bool operator >= (Frequency_Units left, Frequency_Units right)
{
    return ! (left < right);
}


// ... move to GuiXfer for each platform
inline bool GuiCacheEquivalent(Frequency_Units left, Frequency_Units right)
{
    return left.GuiCacheEquivalent(right);
}

//==============================================================================
//   Function WriteVarToIni() - INI writing
//==============================================================================
//  This ftn allows writing into INI without extending IniProtocol class
//   by specializing

inline void  WriteVarToIni(IniProtocol * Ini, const std::string & tag, Frequency_Units & var)
{
    Ini->Write(tag, var.Text(16));  // 16 is precision of double
}

//==============================================================================
//  TEMPLATE Function ReadVarFromIni() - INI writing
//==============================================================================
//  This ftn allows reading from INI without extending IniProtocol class
//   by specializing

inline void  ReadVarFromIni(IniProtocol * Ini, const std::string & tag, Frequency_Units & var, Frequency_Units default_value)
{
    std::string text;
    text = Ini->Read(tag, default_value.Text(16));  // 16 is precision of double
    Frequency_Units freq(text);
    var = freq;
}

// "dense constructor wrappers"  -- make default ToIni() less verbosely

inline Frequency_Units  FreqU(double num=0.0, Frequency_Impl::Types aUnit=Frequency_Impl::Hz)
{
    return Frequency_Units(num, aUnit);
}

inline Frequency_Units  FreqU(std::string rep)
{
    return Frequency_Units(rep);
}
//---------------------------------------------------------------------------
//  class Frequency_Units type info templates
//---------------------------------------------------------------------------

//  "illegal" type
template <Frequency_Impl::Types T>
struct unit_info;

template <>
struct unit_info<Frequency_Impl::Hz>
{
    static double      CFactor() {  return 1.0;  }
    static std::string Name()  { return " Hz";  }
};

template <>
struct unit_info<Frequency_Impl::kHz>
{
    static double      CFactor() {  return 1000.0;  }
    static std::string Name()  { return " kHz";  }
};

template <>
struct unit_info<Frequency_Impl::MHz>
{
    static double      CFactor() {  return 1.0e6;  }
    static std::string Name()  { return " MHz";  }
};

template <>
struct unit_info<Frequency_Impl::GHz>
{
    static double      CFactor() {  return 1.0e9;  }
    static std::string Name()  { return " GHz";  }
};

//---------------------------------------------------------------------------
//  declaration for class U64_as_Hex
//---------------------------------------------------------------------------

class U64_as_Hex
{
public:

    U64_as_Hex(ii64 event=0)
        :  Data(event)
        {
        }
    U64_as_Hex(const std::string & rep)
        :  Data(0)
        {
        ParseString(rep);
        }

    bool  GuiCacheEquivalent(U64_as_Hex other)
        {  return (number==other.number);  }

    // value, unit read
    ii64                 Number()  {  return number;  }
    void                   Number(ii64 a_number)  {  number = a_number;  }
    // convert from standard, canonical string format

    std::string            Text();


    // full broadside load/store
    ii64          Value()  {  return Data;  }
    void          Value(ii64 value) {  Data = value;  }

private:
    ii64  Data;
    ii64                   number;

    // convert from standard, canonical string format

    void  ParseString(const std::string & str);

};

// ... move to GuiXfer for each platform
inline bool GuiCacheEquivalent(U64_as_Hex left, U64_as_Hex right)
{
    return left.GuiCacheEquivalent(right);
}

//==============================================================================
//  Function WriteVarToIni() - INI writing
//==============================================================================
//  This template should allow writing into INI without extending IniProtocol class
//   by specializing

inline void  WriteVarToIni(IniProtocol * Ini, const std::string & tag, U64_as_Hex & var)
{
    Ini->Write(tag, var.Text());
}

//==============================================================================
//  Function ReadVarFromIni() - INI writing
//==============================================================================
//  This template should allow reading from INI without extending IniProtocol class
//   by specializing

inline void  ReadVarFromIni(IniProtocol * Ini, const std::string & tag, U64_as_Hex & var, U64_as_Hex default_value)
{
    std::string text;
    text = Ini->Read(tag, default_value.Text());
    U64_as_Hex number(text);
    var = number;
}

//---------------------------------------------------------------------------
//  declaration for class U64_as_dec
//---------------------------------------------------------------------------

class U64_as_dec
{
public:

    U64_as_dec(ii64 event=0)
        :  Data(event)
        {
        }
    U64_as_dec(const std::string & rep)
        :  Data(0)
        {
        ParseString(rep);
        }
    bool  GuiCacheEquivalent(U64_as_dec other)
        {  return (Data==other.Data);  }

    // value, unit read
   // ii64                 Number()  {  return number;  }
    //void                   Number(ii64 a_number)  {  number = a_number;  }


    // convert from standard, canonical string format

    std::string            Text();


    // full broadside load/store
    ii64          Value()  {  return Data;  }
    void          Value(ii64 value) {  Data = value;  }

private:
    ii64  Data;
   // ii64                   number;

    // convert from standard, canonical string format

    void  ParseString(const std::string & str);

};
// ... move to GuiXfer for each platform
inline bool GuiCacheEquivalent(U64_as_dec left, U64_as_dec right)
{
    return left.GuiCacheEquivalent(right);
}

//==============================================================================
//  Function WriteVarToIni() - INI writing
//==============================================================================
//  This template should allow writing into INI without extending IniProtocol class
//   by specializing

inline void  WriteVarToIni(IniProtocol * Ini, const std::string & tag, U64_as_dec & var)
{
    Ini->Write(tag, var.Text());
}

//==============================================================================
//  Function ReadVarFromIni() - INI writing
//==============================================================================
//  This template should allow reading from INI without extending IniProtocol class
//   by specializing

inline void  ReadVarFromIni(IniProtocol * Ini, const std::string & tag, U64_as_dec & var, U64_as_dec default_value)
{
    std::string text;
    text = Ini->Read(tag, default_value.Text());
    U64_as_dec number(text);
    var = number;
}

//---------------------------------------------------------------------------
//  declaration for class U32_as_Hex
//---------------------------------------------------------------------------

class U32_as_Hex
{
public:

    U32_as_Hex(ii32 event=0)
        :  Data(event)
        {
        }
    U32_as_Hex(const std::string & rep)
        :  Data(0)
        {
        ParseString(rep);
        }
    bool  GuiCacheEquivalent(U32_as_Hex other)
        {  return (number==other.number);  }

    // value, unit read
    ii32                 Number()  {  return number;  }
    void                   Number(ii32 a_number)  {  number = a_number;  }


    // convert from standard, canonical string format

    std::string            Text();


    // full broadside load/store
    ii32          Value()  {  return Data;  }
    void          Value(ii32 value) {  Data = value;  }

private:
    ii32  Data;
    ii32    number;

    // convert from standard, canonical string format

    void  ParseString(const std::string & str);

};

// ... move to GuiXfer for each platform
inline bool GuiCacheEquivalent(U32_as_Hex left, U32_as_Hex right)
{
    return left.GuiCacheEquivalent(right);
}

//==============================================================================
//  Function WriteVarToIni() - INI writing
//==============================================================================
//  This template should allow writing into INI without extending IniProtocol class
//   by specializing

inline void  WriteVarToIni(IniProtocol * Ini, const std::string & tag, U32_as_Hex & var)
{
    Ini->Write(tag, var.Text());
}

//==============================================================================
//  Function ReadVarFromIni() - INI writing
//==============================================================================
//  This template should allow reading from INI without extending IniProtocol class
//   by specializing

inline void  ReadVarFromIni(IniProtocol * Ini, const std::string & tag, U32_as_Hex & var, U32_as_Hex default_value)
{
    std::string text;
    text = Ini->Read(tag, default_value.Text());
    U32_as_Hex number(text);
    var = number;
}

//---------------------------------------------------------------------------
//  declaration for class U32_as_dec
//---------------------------------------------------------------------------

class U32_as_dec
{
public:

    U32_as_dec(ii32 event=0)
        :  Data(event)
        {
        }
    U32_as_dec(const std::string & rep)
        :  Data(0)
        {
        ParseString(rep);
        }
    bool  GuiCacheEquivalent(U32_as_dec other)
        {  return (number==other.number);  }

    // value, unit read
    ii32                 Number()  {  return number;  }
    void                   Number(ii32 a_number)  {  number = a_number;  }


    // convert from standard, canonical string format

    std::string            Text();


    // full broadside load/store
    ii32          Value()  {  return Data;  }
    void          Value(ii32 value) {  Data = value;  }

private:
    ii32  Data;
    ii32    number;

    // convert from standard, canonical string format

    void  ParseString(const std::string & str);

};

// ... move to GuiXfer for each platform
inline bool GuiCacheEquivalent(U32_as_dec left, U32_as_dec right)
{
    return left.GuiCacheEquivalent(right);
}

//==============================================================================
//  Function WriteVarToIni() - INI writing
//==============================================================================
//  This template should allow writing into INI without extending IniProtocol class
//   by specializing

inline void  WriteVarToIni(IniProtocol * Ini, const std::string & tag, U32_as_dec & var)
{
    Ini->Write(tag, var.Text());
}

//==============================================================================
//  Function ReadVarFromIni() - INI writing
//==============================================================================
//  This template should allow reading from INI without extending IniProtocol class
//   by specializing

inline void  ReadVarFromIni(IniProtocol * Ini, const std::string & tag, U32_as_dec & var, U32_as_dec default_value)
{
    std::string text;
    text = Ini->Read(tag, default_value.Text());
    U32_as_dec number(text);
    var = number;
}

//---------------------------------------------------------------------------
//  declaration for class S64_as_dec
//---------------------------------------------------------------------------

class S64_as_dec
{
public:

    S64_as_dec(s_ii64 event=0)
        :  Data(event)
        {
        }
    S64_as_dec(const std::string & rep)
        :  Data(0)
        {
        ParseString(rep);
        }

    bool  GuiCacheEquivalent(S64_as_dec other)
        {  return (number==other.number);  }

    // value, unit read
    s_ii64                 Number()  {  return number;  }
    void                   Number(s_ii64 a_number)  {  number = a_number;  }



    // convert from standard, canonical string format

    std::string            Text();


    // full broadside load/store
    s_ii64          Value()  {  return Data;  }
    void          Value(s_ii64 value) {  Data = value;  }

private:
    s_ii64  Data;
    s_ii64  number;

    // convert from standard, canonical string format

    void  ParseString(const std::string & str);

};

// ... move to GuiXfer for each platform
inline bool GuiCacheEquivalent(S64_as_dec left, S64_as_dec right)
{
    return left.GuiCacheEquivalent(right);
}

//==============================================================================
//  Function WriteVarToIni() - INI writing
//==============================================================================
//  This template should allow writing into INI without extending IniProtocol class
//   by specializing

inline void  WriteVarToIni(IniProtocol * Ini, const std::string & tag, S64_as_dec & var)
{
    Ini->Write(tag, var.Text());
}

//==============================================================================
//  Function ReadVarFromIni() - INI writing
//==============================================================================
//  This template should allow reading from INI without extending IniProtocol class
//   by specializing

inline void  ReadVarFromIni(IniProtocol * Ini, const std::string & tag, S64_as_dec & var, S64_as_dec default_value)
{
    std::string text;
    text = Ini->Read(tag, default_value.Text());
    S64_as_dec number(text);
    var = number;
}
#ifdef __CLR_VER
#pragma managed(pop)
#endif
}  // namespace Innovative
#endif
