//---------------------------------------------------------------------------
// VitaPacketStream_Mb.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2010 Innovative Integration Corporation.
//  All Rights Reserved.

#ifndef VitaPacketStream_MbH
#define VitaPacketStream_MbH

#include <PacketStreamBase_Mb.h>
#include <ThreadSafeQueue_Mb.h>
#include <SafeLink_Mb.h>
#include <deque>
#include <StartStopThread_Mb.h>
#include <ThreadSafeLog_Mb.h>
#include <CountedPtr_Mb.h>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

class VitaPacketStream;

//==============================================================================
//  CLASS VitaPacketImage  -- Vita Packet Position Information Store Object
//==============================================================================

class VitaPacketImage
{
public:

    friend class VitaPacketParser;  // for logging
    //-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
    //  CLASS Range - image data location
    //-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
    class Range
        {
        public:
            Range(const VeloBuffer & buffer, size_t offset=0)
                :  Velo(buffer), Position(offset)
                {  SizeInInts = buffer.SizeInInts() - offset;  }

            VeloBuffer  Velo;
            size_t      Position;
            size_t      SizeInInts;

            unsigned int *  begin()
                {  return reinterpret_cast<unsigned int *>(Velo.Data()) + Position;  }
            unsigned int *  end()
                {  return begin() + SizeInInts ;  }
            const unsigned int *   cbegin() const
                {  return reinterpret_cast<const unsigned int *>(Velo.ConstData()) + Position;  }
            const unsigned int *   cend() const
                {  return cbegin() + SizeInInts ;  }
        };

    VitaPacketImage()   // empty image
        {   }
    VitaPacketImage(const VeloBuffer & buffer, size_t offset=0)
        {  Ranges.push_back( Range(buffer, offset) );  }
    //
    //  Data Information
    size_t            DataDepth() const;
    VitaPacketImage   ExtractNextImage();
    //
    //  Image Information
    size_t         NextPacketFullSize();
    size_t         NextPacketDataSize();

    unsigned int   ImageWord(size_t image_idx);
    unsigned int   HeaderWord(size_t header_idx)
                        {  return ImageWord(header_idx);  }
    unsigned int   BodyWord(size_t body_idx);
    unsigned int   TrailerWord(size_t trailer_idx);


    void  ClearRanges()
        {  Ranges.clear();  }
    void  AddRange( const Range & range )
        {  Ranges.push_back( range );  }
    bool  empty() const
        {  return Ranges.size()==0;  }
    size_t ImageRangeCount() const
        {  return Ranges.size();  }
    Range ImageRange(size_t idx) const
        {  return Ranges[idx];  }

    void  CopyDataTo( VitaBuffer & buffer );
    void  CopyOnlyDataSectionTo( VitaBuffer & buffer );

    void  AreaDump(std::vector<unsigned int> & vec);

private:

    typedef std::deque<Range>  RangePack;
    //
    //  Member Data
    RangePack        Ranges;
};

//==============================================================================
// CLASS VitaPacketParserImageAvailable  --  Parser Image of Packet event
//==============================================================================

class VitaPacketParserImageAvailable : public OpenWire::Event
{
public:
    class VitaPacketParser *  Sender;
    VitaPacketImage           Image;

public:
    VitaPacketParserImageAvailable( VitaPacketParser * sender, const VitaPacketImage & image )
        :  Sender( sender ), Image(image)
        {
        }
};

//==============================================================================
// CLASS VitaPacketParser  --  Parse Velo packets to extract VITA packets
//==============================================================================

class VitaPacketParser
{
public:
    VitaPacketParser() :  DebugLog(10000)
        {}
    virtual ~VitaPacketParser()
        {}
    //
    //  Operations
    void    Clear();
    void    Append(const VeloBuffer & buffer);
    void    Parse();

    size_t  Depth() const;
    //
    //  Events
    OpenWire::EventHandler<VitaPacketParserImageAvailable>  OnImageAvailable;

private:
    //
    //  Member Data
    VitaPacketImage     VeloQ;
    //
    //  Member Functions
    VitaPacketImage    ParsePacketImage();

	ThreadSafeLog      DebugLog;

public:
	//  Debug
	void  Log(int code, int data) {  DebugLog.append(code, data);  };
	void  Log(int code, size_t data) {  DebugLog.append(code, static_cast<int>(data));  };
	void  DumpLog(std::vector<int> & vec);

};

//==============================================================================
// CLASS VitaPacketPackerDataAvailable  --  Data I/O Event
//==============================================================================

class VitaPacketPackerDataAvailable : public OpenWire::Event
{
public:
    class VitaPacketPacker *  Sender;
    VeloBuffer          Data;

public:
    VitaPacketPackerDataAvailable( VitaPacketPacker * sender, VeloBuffer & data )
        :  Sender( sender ), Data(data)
        {
        }
};

//==============================================================================
// CLASS VitaPacketPackerDataRequired  --  Data I/O Event
//==============================================================================

class VitaPacketPackerDataRequired : public OpenWire::Event
{
public:
    class VitaPacketPacker *  Sender;

public:
    VitaPacketPackerDataRequired( VitaPacketPacker * sender )
        :  Sender( sender )
        {
        }
};

//==============================================================================
// CLASS VitaPacketPacker  --  Pack Velo packets with VITA packets
//==============================================================================

class VitaPacketPacker
{
public:
    VitaPacketPacker(size_t out_size = 0x10000)
        :  Out(out_size), OutPos(0), OutputSize(out_size)
        {}
    virtual ~VitaPacketPacker()
        {}
    //
    //  Properties
    size_t  OutputSizeInInts()
                {  return OutputSize;  }
    void    OutputSizeInInts(size_t packet_data_size)
                {  OutputSize = packet_data_size;  }
    //
    //  Operations
    void    Clear();
    void    Pack(const VitaBuffer & buffer);
    void    Flush();

    //
    //  Events
    OpenWire::EventHandler<VitaPacketPackerDataAvailable>  OnDataAvailable;
    OpenWire::EventHandler<VitaPacketPackerDataRequired>   OnDataRequired;

private:
    //
    //  Member Data
    VeloBuffer     Out;
    size_t         OutPos;
    size_t         OutputSize;

};

//==============================================================================
// CLASS VitaPacketCombinerDataAvailable  --  Data I/O Event
//==============================================================================

class VitaPacketCombinerDataAvailable : public OpenWire::Event
{
public:
    class VitaPacketCombiner *  Sender;
    VitaBuffer          Data;

public:
    VitaPacketCombinerDataAvailable( VitaPacketCombiner * sender, VitaBuffer & data )
        :  Sender( sender ), Data(data)
        {
        }
};

//==============================================================================
// CLASS VitaPacketCombiner  --  Merge VITA packets into larger units
//==============================================================================

class VitaPacketCombiner
{
public:
    VitaPacketCombiner(size_t out_size = 0x10000)
        :  Out(out_size), OutPos(0), OutputSize(out_size)
        {}
    virtual ~VitaPacketCombiner()
        {}
    //
    //  Properties
    size_t  OutputSizeInInts()
                {  return OutputSize;  }
    void    OutputSizeInInts(size_t packet_data_size)
                {  OutputSize = packet_data_size;  }
    //
    //  Operations
    void    Clear();
    void    Append(VitaPacketImage & buffer);
    void    Flush();

    //
    //  Events
    OpenWire::EventHandler<VitaPacketCombinerDataAvailable>  OnDataAvailable;

private:
    //
    //  Member Data
    VitaBuffer     Out;
    size_t         OutPos;
    size_t         OutputSize;

};

//==============================================================================
//  CLASS AlignedVeloPacketExQ  -- Velocia Packet Vita Extraction Queue
//==============================================================================

class AlignedVeloPacketExQ
{
public:

    //-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
    //  CLASS Range - image data location
    //-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
    class Range
        {
        public:
            Range(const VeloBuffer & buffer)
                :  Velo(new VeloBuffer(buffer)), Position(0)
                {  SizeInInts = buffer.SizeInInts();  }

            CountedPtr<VeloBuffer>  Velo;
            size_t      Position;
            size_t      SizeInInts;

            unsigned int *  begin()
                {  return reinterpret_cast<unsigned int *>(Velo->Data()) + Position;  }
            unsigned int *  end()
                {  return begin() + SizeInInts ;  }
            const unsigned int *   cbegin() const
                {  return reinterpret_cast<const unsigned int *>(Velo->ConstData()) + Position;  }
            const unsigned int *   cend() const
                {  return cbegin() + SizeInInts ;  }
        };

    AlignedVeloPacketExQ() :  UnknownSidCount(0)
        {   }

    void    Init(std::vector<int> & sids);
    void    ClearRanges();

    void    UnpackVitas(const VeloBuffer & buffer);

    bool  vita_empty() const
        {  return OutVitaRanges.size()==0;  }

    size_t  vita_image_bytes() const
        {  return OutVitaRanges[0].SizeInInts*sizeof(int); }
    Range & vita_image(size_t idx)
        {  return OutVitaRanges[idx]; }

    unsigned int *  vita_trailer_ptr();
    unsigned int *  vita_header_ptr();

    void  vita_pop()
        {  OutVitaRanges.pop_front();  }

private:

    typedef std::deque<Range>  RangePack;
    std::vector<int>           Sids;
    unsigned int               UnknownSidCount;
    //
    //  Member Data
    RangePack        OutVitaRanges;

    size_t          NextPacketFullSizeNC(Range & InVelo);
    int             NextPacketSID(AlignedVeloPacketExQ::Range & InVelo);
    
    void  AddOutRange( const Range & range )
        {  OutVitaRanges.push_back( range );  }

};

//==============================================================================
// CLASS VeloMergeParserDataAvailable  --  Aligned Merge Parser Data Available
//==============================================================================

class VeloMergeParserDataAvailable : public OpenWire::Event
{
public:
    class VeloMergeParser *  Sender;
    VeloBuffer &             Data;

public:
    VeloMergeParserDataAvailable( VeloMergeParser * sender, VeloBuffer & data )
        :  Sender( sender ), Data(data)
        {
        }
};

//==============================================================================
// CLASS VeloMergeParser  --  Merge contained VITA packets into large Velos
//==============================================================================
//  Assumes aligned vita packets

class VeloMergeParser
{
public:
    VeloMergeParser()
        :  OutDataSize(0x10000)
       {}
    virtual ~VeloMergeParser()
        {}
    //
    //  Operations
    void    Init(std::vector<int> & sids);

    void    Clear();
    void    Append(const VeloBuffer & buffer);
    void    Parse();
    void    Flush();

    void    Resize(size_t out_data_size);

    //
    //  Events
    OpenWire::EventHandler<VeloMergeParserDataAvailable>  OnDataAvailable;

    struct OutInfo
    {
        unsigned int  Sid;
        size_t        Pos;
        VeloBuffer    OutData;

        OutInfo() : Sid(0), Pos(0)
            {}
    };

private:
    //
    //  Member Data
    AlignedVeloPacketExQ  VeloQ;
    std::vector<int>      Sids;
    std::vector<OutInfo>  To;
    size_t                OutDataSize;

    //
    //  Member Functions
    unsigned int   VitaHeaderBytes()  {  return 7*sizeof(int);  }
    unsigned int   VitaTrailerBytes()  {  return 1*sizeof(int);  }
    unsigned int   VitaPadding();
    unsigned int   GetCurrentSid();
    OutInfo *      GetOutInfo( unsigned int sid );

    void   PackVitaIntoVelo();
    void   CopyToVelo(AlignedVeloPacketExQ::Range & inRange, size_t InPos,
                                  VeloBuffer & outBuffer, size_t outPos, size_t bytes);

};

//===========================================================================
// CLASS VitaPacketStreamDataEvent  --  Data I/O Event for VitaPacketStream
//===========================================================================

class VitaPacketStreamDataEvent : public OpenWire::Event
{
public:
    class VitaPacketStream *  Sender;

public:
    VitaPacketStreamDataEvent( VitaPacketStream * sender )
        :  Sender( sender )
        {
        }
};

//===========================================================================
// CLASS VitaPacketStreamDeviceSafeLink  --  Safe Linkage to a device
//===========================================================================

class VitaPacketStreamDeviceSafeLink : public SafeLink<IVitaPacketStreamDevice,
                                                       SafeVitaPacketStreamDeviceImpl>
{
public:
    VitaPacketStreamDeviceSafeLink(VitaPacketStream * owner)
        :  Owner(owner)
        {}

private:
    VitaPacketStream *  Owner;

    virtual void  OnAfterLink();        // overrides
    virtual void  OnBeforeUnlink();
};

//===========================================================================
//  CLASS VitaRcvPacketQueue  -- Manage VITA packets
//===========================================================================

class VitaRcvPacketQueue
{
public:
    // Ctor
    VitaRcvPacketQueue();
    virtual ~VitaRcvPacketQueue();

    // Methods
    VitaBuffer      Front();
    size_t          Depth();

    void            Clear();
    void            Push(VitaBuffer & buffer);
    void            Pop();

    Semaphore &     SignalSem();

    void            ClearSem()
                        {  Signal.Clear();  }
    void            BumpSignal(unsigned int count);

    // Properties
    void         QueueCapacity(size_t cap_in_blocks);
    size_t       QueueCapacity() const;

protected:
    //
    //  Member Data
    size_t      CapacityInBlks;
    Semaphore   Signal;
    Event       IsNotEmpty;
    Event       IsNotFull;
    mutable CriticalSection  lock;

    ThreadSafeQueue<VitaBuffer>   VitaQ;
};

//===========================================================================
//  CLASS VitaStreamRcvrThread  -- Recieve VITA packets to app
//===========================================================================
//  This thread handles the VITA stream to application interface

class VitaStreamRcvrThread : public BasicSSThread
{
    //friend class VitaPacketStream;
    typedef BasicSSThread inherited;

public:
    // Ctor
    VitaStreamRcvrThread(VitaPacketStream * owner, VitaRcvPacketQueue & queue)
        : inherited("VitaStreamRcvrThread"), Owner(owner), Queue(queue)
        {  Priority(tpHigher);  }

    void  Clear()
        {  Signal.Clear();  }

    void  Recv(VitaBuffer & packet);

protected:
    VitaPacketStream *    Owner;
    VitaRcvPacketQueue &  Queue;
    Semaphore       Signal;

    enum { sQueue=inherited::sLast, sLast };
    //  override to load notify events at start
    virtual void    Init()
        {
        inherited::Init();
        AddObject(Queue.SignalSem(), sQueue);
        }
    //  do one iteration of main process
    virtual void    AtExecution(int result)
        {
        if (result == sQueue)
            ProcessPacket();
        }

    virtual void    ProcessPacket();

    void            Log(int tag, int value);
    void            Log(int tag, size_t value);
};

//===========================================================================
//  CLASS VitaStreamXmitThread  -- Signal VITA packets needed to app
//===========================================================================
//  This thread handles the VITA stream requests to application interface

class VitaStreamXmitThread : public BasicSSThread
{
    typedef BasicSSThread inherited;

public:
    // Ctor
    VitaStreamXmitThread(VitaPacketStream * owner)
        : inherited("VitaStreamXmitThread"), Owner(owner)
        {  Priority(tpHigher);  }

    void  Clear()
        {  Signal.Clear();  }

    void  SignalDataRequired();

protected:
    VitaPacketStream *    Owner;
    Semaphore             Signal;

    enum { sSignal=inherited::sLast, sLast };
    //  override to load notify events at start
    virtual void    Init()
        {
        inherited::Init();
        AddObject(Signal, sSignal);
        }
    //  do one iteration of main process
    virtual void    AtExecution(int result)
        {
        if (result == sSignal)
            ProcessPacket();
        }

    virtual void    ProcessPacket();

    void            Log(int tag, int value);
    void            Log(int tag, size_t value);
};

//===========================================================================
//  CLASS VitaPacketStream  -- Packet Based Stream I/O
//===========================================================================

class VitaPacketStream : public PacketStreamBase,
                         public IVitaPacketStreamManager
{
public:
    enum VitaStreamModes { vsmVitaMode, vsmVelociaMode };

    friend class VitaStreamRcvrThread;
    friend class VitaStreamXmitThread;

    VitaPacketStream();
    virtual ~VitaPacketStream();

    virtual void  Preconfigure();
    virtual void  Start();
    virtual void  Stop();

//    size_t  OutputPacketSizeInInts()
//                {  return Packer.OutputSizeInInts();  }
//    void    OutputPacketSizeInInts(size_t packet_data_size)
//                {  Packer.OutputSizeInInts(packet_data_size);  }

    //  Vita Buffer Packet I/O
    virtual void  Send( VitaBuffer & packet );
    virtual void  Recv( VitaBuffer & packet );

    void  Pokey(VitaBuffer & packet);
    void  Pokeyinit();
    unsigned int  VDepth();
    //
    //  "test" Velocia packet insertion interface
    virtual void  Send( short PeriphId, VeloBuffer & packet );
    virtual void  Send( VeloBuffer & packet );
    virtual void  Recv( VeloBuffer & packet );

    virtual bool  StreamStarted();

    virtual bool  ConnectTo( IVitaPacketStreamDevice *Device );
    virtual void  Disconnect( );

    VeloBuffer  NativeBuffer(size_t data_size, bool autoinit=true);
    VitaBuffer  NativeVitaBuffer(size_t data_size, bool autoinit=true);

    virtual IPacketStreamManagerInternals *  Internals();
    virtual IPacketStreamDeviceInternals *   DeviceInternals();

    void    VelociaStreamMode(bool state)
        {  FVitaMode = (state) ? vsmVelociaMode : vsmVitaMode;  }
    void    VitaStreamMode(bool state)
        {  FVitaMode = (state) ? vsmVitaMode : vsmVelociaMode;  }
    bool    VelociaStreamMode() const
        {  return FVitaMode==vsmVelociaMode;  }
    bool    VitaStreamMode() const
        {  return FVitaMode==vsmVitaMode;  }

    // Vita mode data handlers
    OpenWire::EventHandler<VitaPacketStreamDataEvent>  OnDataAvailable;
    OpenWire::EventHandler<VitaPacketStreamDataEvent>  OnDataRequired;

    // Velocia mode data handlers
    OpenWire::EventHandler<VitaPacketStreamDataEvent>  OnVeloDataRequired;
    OpenWire::EventHandler<VitaPacketStreamDataEvent>  OnVeloDataAvailable;

protected:
    //
    //  Member Functions
    //
    //  VITA-style packet "Interface" between base and derived classes
    void  SignalVitaDataAvailable();
    void  SignalVitaDataRequired();
    //
    //  Velocia-style packet "Interface" between base and derived classes
    virtual void  SignalDataAvailable();
    virtual void  SignalDataRequired();

    virtual IVelociaBufferManager *  ActualVelociaManager()
                                        {  return &BufferMgr;  }

private:
    //
    //  Member Data
    VitaPacketStreamDeviceSafeLink    StreamDevice;
    VeloBufferManager                 BufferMgr;

    VitaStreamModes                   FVitaMode;

    //  Vita 'outgoing' mode
    //VitaPacketPacker                  Packer;
    VitaStreamXmitThread              VitaOut;
    
    //  Vita 'inooming' mode
    VitaPacketParser                  Parser;

    VitaRcvPacketQueue                VitaInQ;
    VitaStreamRcvrThread              VitaIn;
    //
    //  Member Functions
    void  HandlePackerDataAvailable(VitaPacketPackerDataAvailable & event);
    void  HandlePackerDataRequired(VitaPacketPackerDataRequired & event);
    void  HandleParserImageAvailable(VitaPacketParserImageAvailable & event);

};

#ifdef __CLR_VER
#pragma managed(pop)
#endif
}  // namespace

#endif
