//---------------------------------------------------------------------------
// VpxSubDevice_Misc_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2012 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <VpxSubDevice_Misc_Mb.h>
#include <SystemSupport_Mb.h>

namespace Innovative
{

#ifdef _MSC_VER
    #pragma warning(push)
    #pragma warning(disable : 4355)
#endif

//===========================================================================
//  CLASS VpxFpaLoader -- X6 FPA Loader "SubDevice"
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class VpxFpaLoader
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//  constructor for class VpxFpaLoader
//---------------------------------------------------------------------------

VpxFpaLoader::VpxFpaLoader(IRequires_Nothing * mdc, RomType which_rom,
                           IAddressingSpace & space, MapRegisters & regs)
    :  MDC(mdc),
       AddressControlReg(space, regs.AddressControl_Addr),
       DataReg(space, regs.Data_Addr),
       OpcodeReg(space, regs.Opcode_Addr),
       Chip(MakeChipObject(which_rom)), RawProm(this),
       Thread( new VpxFpaPersonality( Chip ))
{
}

//---------------------------------------------------------------------------
//  destructor for class VpxFpaLoader
//---------------------------------------------------------------------------

VpxFpaLoader::~VpxFpaLoader()
{
    Thread.Shutdown();

    delete Chip;
}

//---------------------------------------------------------------------------
//  VpxFpaLoader::OpenHardware()
//---------------------------------------------------------------------------

void  VpxFpaLoader::OpenHardware()
{
    if (!DeviceEnabled())
        return;

    Thread.OnOpen();
}

//---------------------------------------------------------------------------
//  VpxFpaLoader::ConfigureHardware()
//---------------------------------------------------------------------------

void  VpxFpaLoader::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;
}

//---------------------------------------------------------------------------
//  VpxFpaLoader::ReleaseBlockProtection() -- Remove all block protection
//---------------------------------------------------------------------------

void  VpxFpaLoader::ReleaseBlockProtection()
{
//    if (!DeviceEnabled())
//        return;

    FlashOp(0x06);   // WREN - enable writing
    FlashOp(0x01, 0x00, 0x80);  // WRSR (write status register)
    // ...the above sets BP0-BP3 to 0, removing all protection
    //    SRWD is 1, using the hardware protect signal.
}

//---------------------------------------------------------------------------
//  VpxFpaLoader::EraseSector() -- erase 1 IDROM sector
//---------------------------------------------------------------------------

void  VpxFpaLoader::EraseSector(unsigned int sector)
{
//    if (!DeviceEnabled())
//        return;

    FlashOp(0x06);   // WREN - enable writing
    FlashOp(0xd8, Chip->SectorBaseAddress(sector));
    
    Flash_NotBusy();   // wait for chip to report completed Erase
}

//---------------------------------------------------------------------------
//  VpxFpaLoader::Read32() -- read 1 IDROM page
//---------------------------------------------------------------------------

unsigned int  VpxFpaLoader::Read32(unsigned int addr)
{
//    if (!DeviceEnabled())
//        return;

    return FlashOp(0x03, addr);
}

//---------------------------------------------------------------------------
//  VpxFpaLoader::Write32() -- write 1 word
//---------------------------------------------------------------------------

void  VpxFpaLoader::Write32(unsigned int addr, unsigned int value)
{
//    if (!DeviceEnabled())
//        return;

    char * value_asChar = reinterpret_cast<char *>(&value);

    ByteProgram(addr+3, *value_asChar++);
    ByteProgram(addr+2, *value_asChar++);
    ByteProgram(addr+1, *value_asChar++);
    ByteProgram(addr+0, *value_asChar++);

}

//---------------------------------------------------------------------------
//  VpxFpaLoader::ByteProgram() -- Flash Operation - Byte Program
//---------------------------------------------------------------------------

void  VpxFpaLoader::ByteProgram(unsigned int addr, unsigned int byte)
{
    FlashOp(0x06);   // WREN - enable writing
    FlashOp(0x02, addr, byte );

    Flash_QuickNotBusy();
}

//---------------------------------------------------------------------------
//  VpxFpaLoader::Rdsr() -- Flash Operation - Read status register
//---------------------------------------------------------------------------

unsigned int  VpxFpaLoader::Rdsr()
{
    return FlashOp(0x05, 0);    // address unused
}

//---------------------------------------------------------------------------
//  VpxFpaLoader::Rdid() -- Flash Instruction - RDID
//---------------------------------------------------------------------------

unsigned int  VpxFpaLoader::Rdid()
{
    return FlashOp(0x9F, 0);   // address unused
}

//---------------------------------------------------------------------------
//  VpxFpaLoader::FlashOp() -- primitive Flash Operation - "command-like" ops
//---------------------------------------------------------------------------

void  VpxFpaLoader::FlashOp(unsigned int opcode)
{
    if (!DeviceEnabled())
        return;

    Flash_Ready();

    // Write opcode to trigger the transaction
    OpcodeReg.Value(opcode);
}

//---------------------------------------------------------------------------
//  VpxFpaLoader::FlashOp() -- primitive Flash Operation - "read-like" ops
//---------------------------------------------------------------------------

unsigned int  VpxFpaLoader::FlashOp(unsigned int opcode, unsigned int addr)
{
    if (!DeviceEnabled())
        return 0;

    Flash_Ready();

    // Write address
    AddressControlReg.Addr.Value(addr);

    // Write opcode to trigger the transaction
    OpcodeReg.Value(opcode);

    Flash_ReadValid();

    unsigned int data = DataReg.Value();
    return data;
}

//---------------------------------------------------------------------------
//  VpxFpaLoader::FlashOp() -- primitive Flash Operation - "write-like" ops
//---------------------------------------------------------------------------

void  VpxFpaLoader::FlashOp(unsigned int opcode, unsigned int addr, unsigned int data)
{
    if (!DeviceEnabled())
        return;

    Flash_Ready();

    // Write data
    DataReg.Value(data);

    // Write address
    AddressControlReg.Addr.Value(addr);

    // Write opcode to trigger the transaction
    OpcodeReg.Value(opcode);

}

//---------------------------------------------------------------------------
//  VpxFpaLoader::Flash_Ready() -- wait for flash ready indicator
//---------------------------------------------------------------------------

void  VpxFpaLoader::Flash_Ready()
{
    // Wait up to 5 uS for pending flash to be ready
    for (unsigned int i = 0; i < 5; ++i)
        {
        bool ready = AddressControlReg.Ready.Value();
        if (ready)
            return;
        else
            Innovative::uSleep(40);
        }
    // Gave up
    std::stringstream msg;
    msg << std::hex << " FlashRom Ready Failure (>200us)";
    Debug.Log(msg.str());

}

//---------------------------------------------------------------------------
//  VpxFpaLoader::Flash_ReadValid() -- wait for flash readvalid indicator
//---------------------------------------------------------------------------

void  VpxFpaLoader::Flash_ReadValid()
{
    // Wait up to 5 uS for pending flash read to be valid
    for (unsigned int i = 0; i < 5; ++i)
        {
        bool readvalid = AddressControlReg.ReadDataValid.Value();
        if (readvalid)
            return;
        else
            Innovative::uSleep(40);
        }
    // Gave up
    std::stringstream msg;
    msg << std::hex << " FlashRom Read Valid Failure (>200us)";
    Debug.Log(msg.str());

}

//---------------------------------------------------------------------------
//  VpxFpaLoader::Flash_NotBusy() -- wait for flash busy=0 indicator
//---------------------------------------------------------------------------

void  VpxFpaLoader::Flash_NotBusy()
{
    // Wait up to 50 mS for pending flash read to be validRdsr
    for (unsigned int i = 0; i < 10; ++i)
        {
        unsigned int rdsr = Rdsr();
        if ((rdsr&0x1)==false)
            return;
        else
            Innovative::Sleep(500);
        }
    // Gave up
    std::stringstream msg;
    msg << std::hex << " FlashRom Not Busy Failure (>5s)";
    Debug.Log(msg.str());

}

//---------------------------------------------------------------------------
//  VpxFpaLoader::Flash_QuickNotBusy() -- wait for flash busy=0 indicator
//---------------------------------------------------------------------------

void  VpxFpaLoader::Flash_QuickNotBusy()
{
    // Wait up to 150 uS for pending flash read to be validRdsr
    for (unsigned int i = 0; i < 10; ++i)
        {
        unsigned int rdsr = Rdsr();
        if ((rdsr&0x1)==false)
            return;
        else
            Innovative::uSleep(15);
        }
    // Gave up
    std::stringstream msg;
    msg << std::hex << " FlashRom Quick Not Busy Failure (>150us)";
    Debug.Log(msg.str());

}

//---------------------------------------------------------------------------
//  VpxFpaLoader::MakeChipObject() -- create proper chip object
//---------------------------------------------------------------------------

IVpxFpaChip *   VpxFpaLoader::MakeChipObject(RomType which_rom)
{
    if (which_rom==romM25P64)
        return new VpxFpaChip(this);
    else //(which_rom==romS25FL128S)
        return new K7FpaChip(this);
}

//===========================================================================
//  CLASS VpxFpaLoader::K7FpaChip  -- Hardware Interface
//===========================================================================
//---------------------------------------------------------------------------
//  VpxFpaLoader::K7FpaChip::Initialize() -- init chip object
//---------------------------------------------------------------------------

//bool  VpxFpaLoader::K7FpaChip::Initialize()
//{
//    if (!Owner->DeviceEnabled())
//        return false;

//    FDeviceId = Owner->Rdid();

//    Owner->ReleaseBlockProtection();
//    FProtected = false;

//    return true;   // or check if RDID is right?
//}

//---------------------------------------------------------------------------
//  VpxFpaLoader::K7FpaChip::Read32() -- Read 32 bit word
//---------------------------------------------------------------------------

//unsigned int  VpxFpaLoader::K7FpaChip::Read32(unsigned int addr)
//{
//    if (!Owner->DeviceEnabled())
//        return false;

//    return Owner->Read32(addr);
//}

//---------------------------------------------------------------------------
//  VpxFpaLoader::K7FpaChip::Write32() -- Write 32 bit word
//---------------------------------------------------------------------------

//void  VpxFpaLoader::K7FpaChip::Write32(unsigned int addr, unsigned int value)
//{
//    if (!Owner->DeviceEnabled())
//        return;

//     Owner->Write32(addr, value);
//}

//---------------------------------------------------------------------------
//  VpxFpaLoader::K7FpaChip::EraseSector() -- Erase a sector
//---------------------------------------------------------------------------

//void  VpxFpaLoader::K7FpaChip::EraseSector(unsigned int sector)
//{
//    if (!Owner->DeviceEnabled())
//        return;

//    Owner->EraseSector(sector);
//}

//---------------------------------------------------------------------------
//  VpxFpaLoader::K7FpaChip::SectorBaseAddress() -- base addr of sector
//---------------------------------------------------------------------------

//unsigned int  VpxFpaLoader::K7FpaChip::SectorBaseAddress(unsigned int sector)
//{
//    unsigned int base_addr = 0x0;
//    for (unsigned int s=0; s<sector; s++)
//        base_addr += SectorSize(sector);
//    return base_addr;
//}

//===========================================================================
//  CLASS VpxFpaLoader::VpxFpaChip  -- CPLD Hardware Interface
//===========================================================================
//---------------------------------------------------------------------------
//  VpxFpaLoader::VpxFpaChip::Initialize() -- init chip object
//---------------------------------------------------------------------------

bool  VpxFpaLoader::VpxFpaChip::Initialize()
{
    if (!Owner->DeviceEnabled())
        return false;

    FDeviceId = Owner->Rdid();

    Owner->ReleaseBlockProtection();
    FProtected = false;

    return true;   // or check if RDID is right?
}

//---------------------------------------------------------------------------
//  VpxFpaLoader::VpxFpaChip::Read32() -- Read 32 bit word
//---------------------------------------------------------------------------

unsigned int  VpxFpaLoader::VpxFpaChip::Read32(unsigned int addr)
{
    if (!Owner->DeviceEnabled())
        return false;

    return Owner->Read32(addr);
}

//---------------------------------------------------------------------------
//  VpxFpaLoader::VpxFpaChip::Write32() -- Write 32 bit word
//---------------------------------------------------------------------------

void  VpxFpaLoader::VpxFpaChip::Write32(unsigned int addr, unsigned int value)
{
    if (!Owner->DeviceEnabled())
        return;

     Owner->Write32(addr, value);
}

//---------------------------------------------------------------------------
//  VpxFpaLoader::VpxFpaChip::EraseSector() -- Erase a sector
//---------------------------------------------------------------------------

void  VpxFpaLoader::VpxFpaChip::EraseSector(unsigned int sector)
{
    if (!Owner->DeviceEnabled())
        return;

    Owner->EraseSector(sector);
}

//===========================================================================
//  CLASS VpxApplication -- FMC device information "SubDevice"
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class VpxApplication
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//  constructor for class VpxApplication
//---------------------------------------------------------------------------

VpxApplication::VpxApplication(IRequires_Nothing * mdc, IAddressingSpace & space, MapRegisters & regs)
    :  MDC(mdc),
       ControlStatusReg(space, regs.VpxAppControlStatus_Addr)
{
}

//---------------------------------------------------------------------------
//  destructor for class VpxApplication
//---------------------------------------------------------------------------

VpxApplication::~VpxApplication()
{
}

//---------------------------------------------------------------------------
//  VpxApplication::OpenHardware()
//---------------------------------------------------------------------------

void  VpxApplication::OpenHardware()
{
    if (!DeviceEnabled())
        return;

    // nothing
}

//---------------------------------------------------------------------------
//  VpxApplication::ConfigureHardware()
//---------------------------------------------------------------------------

void  VpxApplication::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;

    // nothing
}

//---------------------------------------------------------------------------
//  VpxApplication::LoopbackEnabled() - Read LoopbackEnabled bit
//---------------------------------------------------------------------------

bool  VpxApplication::LoopbackEnabled()
{
	return ControlStatusReg.Loopback.Value();
}

//---------------------------------------------------------------------------
//  VpxApplication::LoopbackEnabled() - Set LoopbackEnabled bit
//---------------------------------------------------------------------------

void  VpxApplication::LoopbackEnabled(bool value)
{
	ControlStatusReg.Loopback.Value( value );
}

//---------------------------------------------------------------------------
//  VpxApplication::TestGenEnabled() - Read TestGenEnabled bit
//---------------------------------------------------------------------------

bool  VpxApplication::TestGenEnabled()
{
	return ControlStatusReg.TestGen.Value();
}

//---------------------------------------------------------------------------
//  VpxApplication::TestGenEnabled() - Set TestGenEnabled bit
//---------------------------------------------------------------------------

void  VpxApplication::TestGenEnabled(bool value)
{
	ControlStatusReg.TestGen.Value( value );
}

//===========================================================================
//  CLASS VpxDdrMemory -- Vpx DDR3 memory "SubDevice"
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class VpxDdrMemory
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//  constructor for class VpxDdrMemory
//---------------------------------------------------------------------------

VpxDdrMemory::VpxDdrMemory(IRequires_Nothing * mdc, IAddressingSpace & space, MapRegisters & regs)
    :  MDC(mdc),
	   DdrControlReg(space, regs.VpxDdrMemControlStatus_Addr),
	   DdrStatusReg(space, regs.VpxDdrMemControlStatus_Addr),
	   DdrMaskReg(space, regs.VpxDdrMemDataMask_Addr),
       DdrWrite0_Reg(space, regs.VpxDdrMemWriteData0_Addr),
       DdrWrite1_Reg(space, regs.VpxDdrMemWriteData1_Addr),
       DdrWrite2_Reg(space, regs.VpxDdrMemWriteData2_Addr),
       DdrWrite3_Reg(space, regs.VpxDdrMemWriteData3_Addr),
       DdrRead0_Reg(space, regs.VpxDdrMemReadData0_Addr),
	   DdrRead1_Reg(space, regs.VpxDdrMemReadData1_Addr),
       DdrRead2_Reg(space, regs.VpxDdrMemReadData2_Addr),
       DdrRead3_Reg(space, regs.VpxDdrMemReadData3_Addr)
{
}

//---------------------------------------------------------------------------
//  destructor for class VpxDdrMemory
//---------------------------------------------------------------------------

VpxDdrMemory::~VpxDdrMemory()
{
}

//---------------------------------------------------------------------------
//  VpxDdrMemory::OpenHardware()
//---------------------------------------------------------------------------

void  VpxDdrMemory::OpenHardware()
{
    if (!DeviceEnabled())
        return;

    // nothing
}

//---------------------------------------------------------------------------
//  VpxDdrMemory::ConfigureHardware()
//---------------------------------------------------------------------------

void  VpxDdrMemory::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;

    // nothing
}

//---------------------------------------------------------------------------
//  VpxDdrMemory::Read() - read ddr address
//---------------------------------------------------------------------------

void  VpxDdrMemory::Read(unsigned int address, DdrData & data)
{
	// Wait for ready
	ReadReady();

	//  Write address & direction
    DdrControlReg.RWSelect.Value(true);
	DdrControlReg.Address.Value(address);
    DdrControlReg.Apply();

	//  Wait for data avail bit
	//ReadDataAvailable();        -- not needed

	// Read Data
	data.Word[0] = DdrRead0_Reg.Value();
	data.Word[1] = DdrRead1_Reg.Value();
	data.Word[2] = DdrRead2_Reg.Value();
	data.Word[3] = DdrRead3_Reg.Value();
}

//---------------------------------------------------------------------------
//  VpxDdrMemory::Write() - Write ddr address
//---------------------------------------------------------------------------

void  VpxDdrMemory::Write(unsigned int address, const DdrData & data)
{
	// Wait for ready
	WriteReady();

	// Write Data
	DdrWrite0_Reg.Value(data.Word[0]);
	DdrWrite1_Reg.Value(data.Word[1]);
	DdrWrite2_Reg.Value(data.Word[2]);
	DdrWrite3_Reg.Value(data.Word[3]);

	//  Write address & direction
    DdrControlReg.RWSelect.Value(false);
	DdrControlReg.Address.Value(address);
    DdrControlReg.Apply();

}

//---------------------------------------------------------------------------
//  VpxDdrMemory::WriteReady()
//---------------------------------------------------------------------------

void VpxDdrMemory::WriteReady()
{
	// Wait up to 5 uS for pending operation to complete
	for (unsigned int i = 0; i < 5; ++i)
		{
		if (DdrStatusReg.Ready.Value())
			return;
		else
			Innovative::uSleep(1);
		}
	// Gave up
	std::stringstream msg;
	msg << std::hex << " DDR Write Ready Failure (>5us)";
	Debug.Log(msg.str());
}

//---------------------------------------------------------------------------
//  VpxDdrMemory::ReadReady()
//---------------------------------------------------------------------------

void VpxDdrMemory::ReadReady()
{
	// Wait up to 5 uS for pending operation to complete
	for (unsigned int i = 0; i < 5; ++i)
		{
		if (DdrStatusReg.Ready.Value())
			return;
		else
			Innovative::uSleep(1);
		}
	// Gave up
	std::stringstream msg;
	msg << std::hex << " DDR Read Ready Failure (>5us)";
	Debug.Log(msg.str());
}

//---------------------------------------------------------------------------
//  VpxDdrMemory::ReadDataAvailable()
//---------------------------------------------------------------------------

void VpxDdrMemory::ReadDataAvailable()
{
	// Wait up to 5 uS for pending operation to complete
	for (unsigned int i = 0; i < 5; ++i)
		{
		if (DdrStatusReg.DataValid.Value())
			return;
		else
			Innovative::uSleep(1);
		}
	// Gave up
	std::stringstream msg;
	msg << std::hex << " DDR Read DataValid Failure (>5us)";
	Debug.Log(msg.str());
}

//===========================================================================
//  CLASS VpxQdrMemory -- Vpx QDR3 memory "SubDevice"
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class VpxQdrMemory
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//  constructor for class VpxQdrMemory
//---------------------------------------------------------------------------

VpxQdrMemory::VpxQdrMemory(IRequires_Nothing * mdc, IAddressingSpace & space, MapRegisters & regs)
    :  MDC(mdc),
       QdrControlReg(space, regs.VpxQdrMemControlStatus_Addr),
       QdrMaskReg(space, regs.VpxQdrMemDataMask_Addr),
       QdrWrite0_Reg(space, regs.VpxQdrMemWriteData0_Addr),
       QdrWrite1_Reg(space, regs.VpxQdrMemWriteData1_Addr),
       QdrWrite2_Reg(space, regs.VpxQdrMemWriteData2_Addr),
       QdrReadCSR_Reg(space, regs.VpxQdrMemReadCSR_Addr),
       QdrRead0_Reg(space, regs.VpxQdrMemReadData0_Addr),
       QdrRead1_Reg(space, regs.VpxQdrMemReadData1_Addr),
       QdrRead2_Reg(space, regs.VpxQdrMemReadData2_Addr)
{
}

//---------------------------------------------------------------------------
//  destructor for class VpxQdrMemory
//---------------------------------------------------------------------------

VpxQdrMemory::~VpxQdrMemory()
{
}

//---------------------------------------------------------------------------
//  VpxQdrMemory::OpenHardware()
//---------------------------------------------------------------------------

void  VpxQdrMemory::OpenHardware()
{
    if (!DeviceEnabled())
        return;

    // nothing
}

//---------------------------------------------------------------------------
//  VpxQdrMemory::ConfigureHardware()
//---------------------------------------------------------------------------

void  VpxQdrMemory::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;

    // nothing
}

//---------------------------------------------------------------------------
//  VpxQdrMemory::Read() - read qdr address
//---------------------------------------------------------------------------

void  VpxQdrMemory::Read(unsigned int address, QdrData & data)
{
	// Wait for ready
	ReadReady();

	//  Write address
	QdrReadCSR_Reg.Address.Value(address);

	//  Wait for data avail bit
	//  ReadDataAvailable();

	// Read Data
	data.Word[0] = QdrRead0_Reg.Value();
	data.Word[1] = QdrRead1_Reg.Value();
	data.Word[2] = QdrRead2_Reg.Value();
}

//---------------------------------------------------------------------------
//  VpxQdrMemory::Write() - Write qdr address
//---------------------------------------------------------------------------

void  VpxQdrMemory::Write(unsigned int address, const QdrData & data)
{
	// Wait for ready
	WriteReady();

	// Write Data
	QdrWrite0_Reg.Value(data.Word[0]);
	QdrWrite1_Reg.Value(data.Word[1]);
	QdrWrite2_Reg.Value(data.Word[2]);

	// Write Addr
	QdrControlReg.Address.Value(address);

}

//---------------------------------------------------------------------------
//  VpxQdrMemory::WriteReady()
//---------------------------------------------------------------------------

void VpxQdrMemory::WriteReady()
{
	// Wait up to 5 uS for pending operation to complete
	for (unsigned int i = 0; i < 5; ++i)
		{
		if (QdrControlReg.Ready.Value())
			return;
		else
			Innovative::uSleep(1);
		}
	// Gave up
	std::stringstream msg;
	msg << std::hex << " QDR Write Ready Failure (>5us)";
	Debug.Log(msg.str());
}

//---------------------------------------------------------------------------
//  VpxQdrMemory::ReadReady()
//---------------------------------------------------------------------------

void VpxQdrMemory::ReadReady()
{
	// Wait up to 5 uS for pending operation to complete
	for (unsigned int i = 0; i < 5; ++i)
		{
		if (QdrReadCSR_Reg.Ready.Value())
			return;
		else
			Innovative::uSleep(1);
		}
	// Gave up
	std::stringstream msg;
	msg << std::hex << " QDR Read Ready Failure (>5us)";
	Debug.Log(msg.str());
}

//---------------------------------------------------------------------------
//  VpxQdrMemory::ReadDataAvailable()
//---------------------------------------------------------------------------

void VpxQdrMemory::ReadDataAvailable()
{
	// Wait up to 5 uS for pending operation to complete
	for (unsigned int i = 0; i < 5; ++i)
		{
		if (QdrReadCSR_Reg.DataValid.Value())
			return;
		else
			Innovative::uSleep(1);
		}
	// Gave up
	std::stringstream msg;
	msg << std::hex << " QDR Read DataValid Failure (>5us)";
	Debug.Log(msg.str());
}

//===========================================================================
//  CLASS VpxTrigger -- Vpx Trigger "SubDevice"
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class VpxTrigger
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//  constructor for class VpxTrigger
//---------------------------------------------------------------------------

VpxTrigger::VpxTrigger(IRequires_Nothing * mdc, IAddressingSpace & space, MapRegisters & regs)
    :  MDC(mdc),
       TriggerControlReg(space, regs.VpxTrigger_Control_Addr),
       ItfOtfReg(space, regs.VpxTrigger_ItfOtf_Addr),
       PlLtReg(space, regs.VpxTrigger_PlLt_Addr),
       PlMdReg(space, regs.VpxTrigger_PlMd_Addr),
       GlobalReg(space, regs.VpxTrigger_PpsGt_Addr)
{
}

//---------------------------------------------------------------------------
//  destructor for class VpxTrigger
//---------------------------------------------------------------------------

VpxTrigger::~VpxTrigger()
{
}

//---------------------------------------------------------------------------
//  VpxTrigger::OpenHardware()
//---------------------------------------------------------------------------

void  VpxTrigger::OpenHardware()
{
    if (!DeviceEnabled())
        return;

    // nothing
}

//---------------------------------------------------------------------------
//  VpxTrigger::ConfigureHardware()
//---------------------------------------------------------------------------

void  VpxTrigger::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;

    // nothing
}

//===========================================================================
//  CLASS Comex_ControlStatus -- EPC CSR "SubDevice"
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class Comex_ControlStatus
//---------------------------------------------------------------------------

Comex_ControlStatus::Comex_ControlStatus(IRequires_Nothing * mdc,
                        IAddressingSpace & bar_0space, MapRegisters & regs)
    :  MDC(mdc),
       ControlReg(bar_0space, regs.Bar0Control_Addr),
       StatusReg(bar_0space,  regs.Bar0Status_Addr)
{
}

//---------------------------------------------------------------------------
//  destructor for class Comex_ControlStatus
//---------------------------------------------------------------------------

Comex_ControlStatus::~Comex_ControlStatus()
{
}

//---------------------------------------------------------------------------
//  Comex_ControlStatus::OpenHardware()
//---------------------------------------------------------------------------

void  Comex_ControlStatus::OpenHardware()
{
    if (!DeviceEnabled())
        return;

    // nothing
}

//---------------------------------------------------------------------------
//  Comex_ControlStatus::ConfigureHardware()
//---------------------------------------------------------------------------

void  Comex_ControlStatus::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;

    // nothing
}

//---------------------------------------------------------------------------
//  Comex_ControlStatus::Led() -- Read LED state
//---------------------------------------------------------------------------

bool Comex_ControlStatus::Led() const
{
    return ControlReg.Led.Value();
}

//---------------------------------------------------------------------------
//  Comex_ControlStatus::Led() -- Write LED state
//---------------------------------------------------------------------------

void Comex_ControlStatus::Led(bool state)
{
    ControlReg.Led.Value(state);
}

//---------------------------------------------------------------------------
//  Comex_ControlStatus::Slot64() -- Read LED state
//---------------------------------------------------------------------------

bool Comex_ControlStatus::Slot64() const
{
    return StatusReg.Slot64.Value();
}

//===========================================================================
//  CLASS VpxComex_SmbOneControl  -- Master Time Config Controller
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class VpxComex_SmbOneControl
//---------------------------------------------------------------------------

const unsigned int Smb1_Subaddr = 0x7C;

//---------------------------------------------------------------------------
//  constructor for class VpxComex_SmbOneControl
//---------------------------------------------------------------------------

VpxComex_SmbOneControl::VpxComex_SmbOneControl(IRequires_Nothing * mdc,
                    IAddressingSpace & space, I2C_MasterCore::MapRegisters & regs)
    :  MDC(mdc),  Smb1(space, regs, Smb1_Subaddr), Smb1Mux(0x0)
{
}

//---------------------------------------------------------------------------
//  VpxComex_SmbOneControl::OpenHardware() -- on Open Configuration
//---------------------------------------------------------------------------

void  VpxComex_SmbOneControl::OpenHardware()
{
    Smb1.Init();
    Smb1Mux.Value(0x60);

    // configure all as outputs
    Smb1.I2C_Reg(0x3, 0x0);

    // Select Default Mux
    ExternalSourceSelect(inJ1);
    // Select Default Trigger
    ExternalTriggerSelect(inJ2);
    // Select Default PPS source
    ExternalPpsSourceSelect(inJ3);

}

//---------------------------------------------------------------------------
//  VpxComex_SmbOneControl::ConfigureHardware() -- on Start Configuration
//---------------------------------------------------------------------------

void  VpxComex_SmbOneControl::ConfigureHardware()
{
}

//---------------------------------------------------------------------------
//  VpxComex_SmbOneControl::StreamStopHardware() -- on Stop Configuration
//---------------------------------------------------------------------------

void  VpxComex_SmbOneControl::StreamStopHardware()
{
}

//---------------------------------------------------------------------------
//  VpxComex_SmbOneControl::ExternalSourceSelect() --
//---------------------------------------------------------------------------

void  VpxComex_SmbOneControl::ExternalSourceSelect( Mux0SourceSelect which )
{
    switch (which)
        {
        default:
        case inJ1:
            Smb1Mux.Bit(0, false);
            break;
        case inC_Clk_I:
            Smb1Mux.Bit(0, true);
            break;
        }
    Smb1.I2C_Reg(0x1, Smb1Mux.Value());
}

//---------------------------------------------------------------------------
//  VpxComex_SmbOneControl::ExternalTriggerSelect() --
//---------------------------------------------------------------------------

void  VpxComex_SmbOneControl::ExternalTriggerSelect( Mux7SourceSelect which )
{
    switch (which)
        {
        default:
        case inJ1:
            Smb1Mux.Bit(7, false);
            break;
        case inC_Trig_I:
            Smb1Mux.Bit(7, true);
            break;
        }
    Smb1.I2C_Reg(0x1, Smb1Mux.Value());
}

//---------------------------------------------------------------------------
//  VpxComex_SmbOneControl::ExternalPpsSourceSelect() --
//---------------------------------------------------------------------------

void  VpxComex_SmbOneControl::ExternalPpsSourceSelect( Mux3SourceSelect which )
{
    switch (which)
        {
        default:
        case inJ3:
            Smb1Mux.Bit(3, false);
            break;
        case inC_PPS_I:
            Smb1Mux.Bit(3, true);
            break;
        }
    Smb1.I2C_Reg(0x1, Smb1Mux.Value());
}

//===========================================================================
//  CLASS VpxComex_Trigger -- Vpx Epc Trigger "SubDevice"
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class VpxComex_Trigger
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//  constructor for class VpxComex_Trigger
//---------------------------------------------------------------------------

VpxComex_Trigger::VpxComex_Trigger(IRequires_ComexSignalMux * mdc, IAddressingSpace & space, MapRegisters & regs)
    :  MDC(mdc),
       TriggerModeReg(space, regs.TriggerModeCfg_Addr),
       TriggerFrameReg(space, regs.FrameSizeCfg_Addr),
       DecimationReg(space, regs.Decimation_Addr),
       SwTriggerReg(space, regs.SoftwareTrigger_Addr),
       ExtSyncSelectReg(space, regs.ExtSyncSelect_Addr)
{
}

//---------------------------------------------------------------------------
//  destructor for class VpxComex_Trigger
//---------------------------------------------------------------------------

VpxComex_Trigger::~VpxComex_Trigger()
{
    SigMux = 0;   // Not owned by us.
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::ConstructCustomParts()
//---------------------------------------------------------------------------

void  VpxComex_Trigger::ConstructCustomParts()
{
    if (!DeviceEnabled())
        return;

    SigMux = MDC->MakeComexSignalMuxPtr();
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::OpenHardware()
//---------------------------------------------------------------------------

void  VpxComex_Trigger::OpenHardware()
{
    if (!DeviceEnabled())
        return;

    AllOff();
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::ConfigureHardware()
//---------------------------------------------------------------------------

void  VpxComex_Trigger::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;

    AllOff();
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::AllOff()
//---------------------------------------------------------------------------

void  VpxComex_Trigger::AllOff()
{
    SoftwareTrigger( false );
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::SoftwareTrigger() --  Set SW Trigger state
//---------------------------------------------------------------------------

void VpxComex_Trigger::SoftwareTrigger(bool state)
{
    SwTriggerReg.SwTrigger.Value(state);
    SwTriggerReg.Apply();
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::SoftwareTrigger() --  Read SW Trigger state
//---------------------------------------------------------------------------

bool  VpxComex_Trigger::SoftwareTrigger() const
{
    return SwTriggerReg.SwTrigger.Value();
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::FramedTriggerSize() --  Set Frame Size
//---------------------------------------------------------------------------

void VpxComex_Trigger::FramedTriggerSize(unsigned int samples)
{
    TriggerFrameReg.FrameSize.Value(samples);
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::FramedTriggerSize() --  Read Frame Size
//---------------------------------------------------------------------------

unsigned int  VpxComex_Trigger::FramedTriggerSize() const
{
    return TriggerFrameReg.FrameSize.Value();
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::Decimation() --  Set Decimation Factor
//---------------------------------------------------------------------------

void VpxComex_Trigger::Decimation(unsigned int samples)
{
    DecimationReg.Decimation.Value(samples);
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::Decimation() --  Read Decimation Factor
//---------------------------------------------------------------------------

unsigned int  VpxComex_Trigger::Decimation() const
{
    return DecimationReg.Decimation.Value();
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::FramedTriggerEnabled() --  Set Framed Trigger state
//---------------------------------------------------------------------------

void VpxComex_Trigger::FramedTriggerEnabled(bool state)
{
    TriggerModeReg.FramedMode.Value(state);
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::FramedTriggerEnabled() --  Read Framed Trigger state
//---------------------------------------------------------------------------

bool  VpxComex_Trigger::FramedTriggerEnabled() const
{
    return TriggerModeReg.FramedMode.Value();
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::EdgeTrigger() --  Set Edge Trigger state
//---------------------------------------------------------------------------

void VpxComex_Trigger::EdgeTrigger(bool state)
{
    TriggerModeReg.RisingEdge.Value(state);
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::EdgeTrigger() --  Read Edge Trigger state
//---------------------------------------------------------------------------

bool  VpxComex_Trigger::EdgeTrigger() const
{
    return TriggerModeReg.RisingEdge.Value();
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::ExternalTrigger() --  Set External Trigger state
//---------------------------------------------------------------------------

void VpxComex_Trigger::ExternalTrigger(bool state)
{
    TriggerModeReg.ExternalTrigger.Value(state);
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::ExternalTrigger() --  Read External Trigger state
//---------------------------------------------------------------------------

bool  VpxComex_Trigger::ExternalTrigger() const
{
    return TriggerModeReg.ExternalTrigger.Value();
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::ExternalSyncSelection() --  Set Ext Trigger source option
//---------------------------------------------------------------------------

void VpxComex_Trigger::ExternalSyncSelection(ExtSyncOptions option)
{
    switch (option)
        {
        case essFrontPanel:
            ExtSyncSelectReg.ExtSyncOption.Value(0);
            break;
        case essPps:
            ExtSyncSelectReg.ExtSyncOption.Value(1);
            break;
        }
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::ExternalSyncSelection() --  Set Ext Trigger source option
//---------------------------------------------------------------------------

VpxComex_Trigger::ExtSyncOptions VpxComex_Trigger::ExternalSyncSelection() const
{
    bool option =  ExtSyncSelectReg.ExtSyncOption.Value();
    if (option)
        return essPps;
    else
        return essFrontPanel;
}

//---------------------------------------------------------------------------
//  VpxComex_Trigger::LTrigSource(() -- set LTrig source
//---------------------------------------------------------------------------

void  VpxComex_Trigger::LTrigSource(LTrigSourceSelect source)
{
    if (source==srcJ2)
        {
        SigMux->ExternalTriggerSelect(VpxComex_SmbOneControl::inJ2);
        }
    else   // srcC_Trig_I
        {
        SigMux->ExternalTriggerSelect(VpxComex_SmbOneControl::inC_Trig_I);
        }

}
//===========================================================================
//  CLASS VpxComex_Utility --  Vpx Comex Utility "SubDevice"
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class VpxComex_Utility
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//  constructor for class VpxComex_Utility
//---------------------------------------------------------------------------

VpxComex_Utility::VpxComex_Utility(IRequires_Nothing * mdc, IAddressingSpace & space, MapRegisters & regs)
    :  MDC(mdc),
       VoltageStatusReg(space, regs.VoltageStatus_Addr),
       GpioDirectionReg(space, regs.GpioDirection_Addr),
       GpioDataReg(space, regs.GpioData_Addr)
{
}

//---------------------------------------------------------------------------
//  destructor for class VpxComex_Utility
//---------------------------------------------------------------------------

VpxComex_Utility::~VpxComex_Utility()
{
}

//---------------------------------------------------------------------------
//  VpxComex_Utility::ConstructCustomParts()
//---------------------------------------------------------------------------

void  VpxComex_Utility::ConstructCustomParts()
{
}

//---------------------------------------------------------------------------
//  VpxComex_Utility::OpenHardware()
//---------------------------------------------------------------------------

void  VpxComex_Utility::OpenHardware()
{
    if (!DeviceEnabled())
        return;
}

//---------------------------------------------------------------------------
//  VpxComex_Utility::ConfigureHardware()
//---------------------------------------------------------------------------

void  VpxComex_Utility::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;
}

//---------------------------------------------------------------------------
//  VpxComex_Utility::V2_5Status() --  Read Voltage Status
//---------------------------------------------------------------------------

bool  VpxComex_Utility::V2_5Status() const
{
    return VoltageStatusReg.V2_5Status.Value();
}

//---------------------------------------------------------------------------
//  VpxComex_Utility::V1_0Status() --  Read Voltage Status
//---------------------------------------------------------------------------

bool  VpxComex_Utility::V1_0Status() const
{
    return VoltageStatusReg.V1_0Status.Value();
}

//---------------------------------------------------------------------------
//  VpxComex_Utility::V1_2Status() --  Read Voltage Status
//---------------------------------------------------------------------------

bool  VpxComex_Utility::V1_2Status() const
{
    return VoltageStatusReg.V1_2Status.Value();
}

//---------------------------------------------------------------------------
//  VpxComex_Utility::V1_5Status() --  Read Voltage Status
//---------------------------------------------------------------------------

bool  VpxComex_Utility::V1_5Status() const
{
    return VoltageStatusReg.V1_5Status.Value();
}

//---------------------------------------------------------------------------
//  VpxComex_Utility::Ddr3Status() --  Read Voltage Status
//---------------------------------------------------------------------------

bool  VpxComex_Utility::Ddr3Status() const
{
    return VoltageStatusReg.Ddr3Status.Value();
}

//---------------------------------------------------------------------------
//  VpxComex_Utility::GpioDirection() --  Set direction of GPIO
//---------------------------------------------------------------------------

void  VpxComex_Utility::GpioDirection(bool isOutput)
{
    return GpioDirectionReg.Bit(0, isOutput);
}

//---------------------------------------------------------------------------
//  VpxComex_Utility::Gpio() --  Read Gpio Data
//---------------------------------------------------------------------------

bool  VpxComex_Utility::Gpio() const
{
    return GpioDataReg.Bit(0);
}

//---------------------------------------------------------------------------
//  VpxComex_Utility::Gpio() --  Write Gpio Data
//---------------------------------------------------------------------------

void  VpxComex_Utility::Gpio(bool data)
{
    return GpioDataReg.Bit(0, data);
}

//===========================================================================
//  CLASS VpxComex_PexSwitch  -- Vpx Comex PEX Switch Subsystem
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class VpxComex_PexSwitch
//---------------------------------------------------------------------------

const unsigned int Pex_Subaddr = 0xE8;
const unsigned int PexRetries = 5;

//---------------------------------------------------------------------------
//  constructor for class VpxComex_PexSwitch
//---------------------------------------------------------------------------

VpxComex_PexSwitch::VpxComex_PexSwitch(IRequires_Nothing * mdc,
                    IAddressingSpace & space, I2C_MasterCore::MapRegisters & regs)
    :  MDC(mdc),  Pex(space, regs, Pex_Subaddr)
{
    Pex.PresetI2CFrequency(400e3);   // Use 400 kHz I2C rate
}

//---------------------------------------------------------------------------
//  VpxComex_PexSwitch::OpenHardware() -- on Open Configuration
//---------------------------------------------------------------------------

void  VpxComex_PexSwitch::OpenHardware()
{
    Pex.Init();
}

//---------------------------------------------------------------------------
//  VpxComex_PexSwitch::ConfigureHardware() -- on Start Configuration
//---------------------------------------------------------------------------

void  VpxComex_PexSwitch::ConfigureHardware()
{
}

//---------------------------------------------------------------------------
//  VpxComex_PexSwitch::StreamStopHardware() -- on Stop Configuration
//---------------------------------------------------------------------------

void  VpxComex_PexSwitch::StreamStopHardware()
{
}

//---------------------------------------------------------------------------
//  VpxComex_PexSwitch::CsrPort() -- Read 32 bit CSR Register
//---------------------------------------------------------------------------

unsigned int  VpxComex_PexSwitch::CsrPort(unsigned int addr)
{
    ScopedTraceRef ccc(Debug, "VpxComex_PexSwitch::CsrPort(addr)");

    std::vector<int> ReadRequest;
    ReadRequest.push_back(0x03);             // byte count of remainder
    ReadRequest.push_back((int)kDWordRead);  // Command code 4 byte read
    ReadRequest.push_back(addr & 0xFF);      // Address LSB
    ReadRequest.push_back(addr >> 8);        // Address MSB

    bool XfrRequestOk(false);
    for (unsigned int i=0; i<PexRetries; i++)
        {
        Debug.Log("Read Request Xfr #", i);
        XfrRequestOk = Pex.MultiRegW((int)kCsrBlockIo, ReadRequest);
        if (XfrRequestOk)
            break;
        else
            {
            Debug.Log("Request Failed");
            }
        }
    if (!XfrRequestOk)
        {
        return 0;
        }
    //
    //  We sent the request OK...so get the reply
    for (unsigned int i=0; i<PexRetries; i++)
        {
        Debug.Log("Reply #", i);
        std::vector<int> ReadReply(8);    // expect 8 words back
        bool ReadOk = Pex.MultiRegR((int)kCsrBlockIo, ReadReply);
        //
        //  Check response
        if (ReadOk)
            {
            MemoryBitManipulator RtnVal(0);
            RtnVal.Field(24, 8, ReadReply[7]);
            RtnVal.Field(16, 8, ReadReply[6]);
            RtnVal.Field( 8, 8, ReadReply[5]);
            RtnVal.Field( 0, 8, ReadReply[4]);
            return RtnVal.Value();
            }
        //
        //  Else was an error
        Debug.Log("Readback Not Ready/Error: size=", ReadReply.size());
        for (unsigned int x=0; x<ReadReply.size(); x++)
            Debug.Log("Word: ", ReadReply[x]);
        //  keep trying
        }

    // if we get here we gave up on read retries
    return 0;

}


//---------------------------------------------------------------------------
//  VpxComex_PexSwitch::CsrPort() -- Write 32 bit CSR Register
//---------------------------------------------------------------------------

void  VpxComex_PexSwitch::CsrPort(unsigned int addr, unsigned int data)
{
    ScopedTraceRef ccc(Debug, "VpxComex_PexSwitch::CsrPort(addr, data)");

    //
    //  Construct tranfer packet
    MemoryBitManipulator RtnVal(data);
    std::vector<int> WriteInfo;
    WriteInfo.push_back(0x07);                  // byte count of remainder
    WriteInfo.push_back((int)kDWordWrite);      // Command code 4 byte write
    WriteInfo.push_back(addr & 0xFF);           // Address LSB
    WriteInfo.push_back(addr >> 8);             // Address MSB
    WriteInfo.push_back(RtnVal.Field( 0, 8));   // Data LSB
    WriteInfo.push_back(RtnVal.Field( 8, 8));   // Data
    WriteInfo.push_back(RtnVal.Field(16, 8));   // Data
    WriteInfo.push_back(RtnVal.Field(24, 8));   // Data MSB

    bool WriteOk(false);
    for (unsigned int i=0; i<PexRetries; i++)
        {
        Debug.Log("Write Xfr #", i);
        WriteOk = Pex.MultiRegW((int)kCsrBlockIo, WriteInfo);
        if (WriteOk)
            break;
        else
            {
            Debug.Log("Try Failed");
            }
        }
    if (!WriteOk)
        {
        Debug.Log("Write Command Failed");
        }

}

//===========================================================================
//  CLASS VpxComex_Repeater  -- Vpx Comex PCIE Repeater Subsystem
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class VpxComex_Repeater
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//  constructor for class VpxComex_Repeater
//---------------------------------------------------------------------------

VpxComex_Repeater::VpxComex_Repeater(IRequires_Nothing * mdc, unsigned int SubAddr,
                    IAddressingSpace & space,
                    I2C_MasterCore::MapRegisters & regs)
    :  MDC(mdc),  Repeater(space, regs, SubAddr)
{
    Repeater.PresetI2CFrequency(400e3);   // Use 400 kHz I2C rate
}

//---------------------------------------------------------------------------
//  VpxComex_Repeater::OpenHardware() -- on Open Configuration
//---------------------------------------------------------------------------

void  VpxComex_Repeater::OpenHardware()
{
    Repeater.Init();
}

//---------------------------------------------------------------------------
//  VpxComex_Repeater::ConfigureHardware() -- on Start Configuration
//---------------------------------------------------------------------------

void  VpxComex_Repeater::ConfigureHardware()
{
}

//---------------------------------------------------------------------------
//  VpxComex_Repeater::StreamStopHardware() -- on Stop Configuration
//---------------------------------------------------------------------------

void  VpxComex_Repeater::StreamStopHardware()
{
}

//---------------------------------------------------------------------------
//  VpxComex_Repeater::Write() -- Write data packet to repeater
//---------------------------------------------------------------------------

bool  VpxComex_Repeater::Write(std::vector<int> & data)
{
    return Repeater.MultiRegW(0, data);
}

//---------------------------------------------------------------------------
//  VpxComex_Repeater::Read() -- Write data packet to repeater
//---------------------------------------------------------------------------

bool  VpxComex_Repeater::Read(std::vector<int> & data)
{
    return Repeater.MultiRegR(0, data);
}


//===========================================================================
//  CLASS Vpx_ClockGen  -- Vpx Baseboard ClockGen Subsystem
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class Vpx_ClockGen
//---------------------------------------------------------------------------

const float  defaultLClkFreq = 100.0e6;
const float  defaultFmcClk1Freq = 90.0e6;

//---------------------------------------------------------------------------
//  constructor for class Vpx_ClockGen
//---------------------------------------------------------------------------

Vpx_ClockGen::Vpx_ClockGen(IRequires_VpxClkGenParts * mdc, IAddressingSpace & /*space*/, MapRegisters & /*regs*/ )
    :  MDC(mdc), I2C_uDev(0), VcxoDev(0), PllDev(0), InMuxDev(0),
       LClk_Frequency(defaultLClkFreq), FmcClk1_Frequency(defaultFmcClk1Freq),
       ClockGenIo(this), ClockGenDevs(this)
{
}

//---------------------------------------------------------------------------
//  destructor for class Vpx_ClockGen
//---------------------------------------------------------------------------

Vpx_ClockGen::~Vpx_ClockGen()
{
    delete InMuxDev;   InMuxDev  = 0;
    delete PllDev;     PllDev  = 0;
    delete VcxoDev;    VcxoDev = 0;
    delete I2C_uDev;   I2C_uDev = 0;
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::ConstructCustomParts()
//---------------------------------------------------------------------------

void  Vpx_ClockGen::ConstructCustomParts()
{
    if (!DeviceEnabled())
        return;

    InMuxDev = MDC->MakeVCG_ClkMuxObject();
    I2C_uDev = MDC->MakeVCG_VcxoI2CLink();
    VcxoDev  = MDC->MakeVCG_VcxoObject(I2C_uDev->Device());
    PllDev   = MDC->MakeVCG_PllObject();
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::OpenHardware() -- on Open Configuration
//---------------------------------------------------------------------------

void  Vpx_ClockGen::OpenHardware()
{
    if (!DeviceEnabled())
        return;

    //  Mux
//    InMuxDev->Powered(true);
    InMuxDev->Init();
    //  Vcxo
    I2C_uDev->Init();
    EnableVcxo(true);
    VcxoDev->Init();
    //  Pll
    PllDev->Powered(true);
    PllDev->Init();
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::ConfigureHardware() -- on Start Configuration
//---------------------------------------------------------------------------

void  Vpx_ClockGen::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;

    // Reapply Mux settings
    InMuxDev->Reapply();
    // Reapply Vcxo and PLL settings
    VcxoDev->Reapply();
    PllDev->Reapply();
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::StreamStopHardware() -- on Stop Configuration
//---------------------------------------------------------------------------

void  Vpx_ClockGen::StreamStopHardware()
{
    if (!DeviceEnabled())
        return;
    //  turn off clocks, disable outputs?
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::EnableVcxo()
//---------------------------------------------------------------------------

void  Vpx_ClockGen::EnableVcxo(bool state)
{
    I2C_uDev->DeviceEnable(state);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::Vcxo()
//---------------------------------------------------------------------------

Si570 & Vpx_ClockGen::Vcxo()
{
    return *VcxoDev;
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::InMux() --
//---------------------------------------------------------------------------

Cdce18005 & Vpx_ClockGen::InMux()
{
    return *InMuxDev;
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::Pll() --
//---------------------------------------------------------------------------

Cdce72010 & Vpx_ClockGen::Pll()
{
    return *PllDev;
}

//===========================================================================
//  CLASS Vpx_ClockGen::VpxClkGenIo -- X6 AFE Clock "SubDevice" User Interface
//===========================================================================
//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::SetPllMasterOutputSelection(() --
//---------------------------------------------------------------------------

void  Vpx_ClockGen::VpxClkGenIo::SetPllMasterOutputSelection(int idx)
{
    Owner->Pll().SetMasterOutputSelection(idx);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::ToDivider() -- convert divider list idx to divider
//---------------------------------------------------------------------------

unsigned int  Vpx_ClockGen::VpxClkGenIo::ToDivider(unsigned int divider_idx)
{
    return Owner->InMux().ToDivider(divider_idx);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::FpgaClk1Source() --
//---------------------------------------------------------------------------

void  Vpx_ClockGen::VpxClkGenIo::FpgaClk1Source(unsigned int mux_source_select)
{
    Owner->InMux().InputSource(2, mux_source_select);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::FpgaClk1Divider() --
//---------------------------------------------------------------------------

void  Vpx_ClockGen::VpxClkGenIo::FpgaClk1Divider(unsigned int divide_by)
{
    Owner->InMux().OutputDivider(2, divide_by);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::FpgaClk1OutputEnabled() --
//---------------------------------------------------------------------------

void  Vpx_ClockGen::VpxClkGenIo::FpgaClk1OutputEnabled(bool enabled)
{
    Owner->InMux().OutputEnabled(2, enabled);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::FpgaClk1OutputFrequencyActual() --
//---------------------------------------------------------------------------

double  Vpx_ClockGen::VpxClkGenIo::FpgaClk1OutputFrequencyActual()
{
    return Owner->InMux().OutputFrequencyActual(2);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::FpgaClk0OutputEnabled() --
//---------------------------------------------------------------------------

void  Vpx_ClockGen::VpxClkGenIo::FpgaClk0OutputEnabled(bool enabled)
{
    Owner->PllDev->OutputEnabled(1, enabled);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::FpgaClk0OutputFrequency() --
//---------------------------------------------------------------------------

void  Vpx_ClockGen::VpxClkGenIo::FpgaClk0OutputFrequency(double frequency)
{
    Owner->PllDev->OutputFrequency(1, frequency);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::FpgaClk0OutputFrequencyActual() --
//---------------------------------------------------------------------------

double  Vpx_ClockGen::VpxClkGenIo::FpgaClk0OutputFrequencyActual()
{
    return Owner->PllDev->OutputFrequencyActual(1);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::GClkOutputEnabled() --
//---------------------------------------------------------------------------

void  Vpx_ClockGen::VpxClkGenIo::GClkOutputEnabled(unsigned int idx, bool enabled)
{
    if (idx>3)
        idx=3;
     Owner->PllDev->OutputEnabled(2+idx, enabled);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::GClkOutputFrequency() --
//---------------------------------------------------------------------------

void  Vpx_ClockGen::VpxClkGenIo::GClkOutputFrequency(unsigned int idx, double frequency)
{
    if (idx>3)
        idx=3;
    Owner->PllDev->OutputFrequency(2+idx, frequency);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::GClkOutputFrequencyActual() --
//---------------------------------------------------------------------------

double  Vpx_ClockGen::VpxClkGenIo::GClkOutputFrequencyActual(unsigned int idx)
{
    if (idx>3)
        idx=3;
    return  Owner->PllDev->OutputFrequencyActual(2+idx);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::PlClkOutputEnabled() --
//---------------------------------------------------------------------------

void  Vpx_ClockGen::VpxClkGenIo::PlClkOutputEnabled(bool enabled)
{
    Owner->PllDev->OutputEnabled(6, enabled);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::PlClkOutputFrequency() --
//---------------------------------------------------------------------------

void  Vpx_ClockGen::VpxClkGenIo::PlClkOutputFrequency(double frequency)
{
    Owner->PllDev->OutputFrequency(6, frequency);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::PlClkOutputFrequencyActual() --
//---------------------------------------------------------------------------

double  Vpx_ClockGen::VpxClkGenIo::PlClkOutputFrequencyActual()
{
    return Owner->PllDev->OutputFrequencyActual(6);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::FmcClk2OutputEnabled() --
//---------------------------------------------------------------------------

void  Vpx_ClockGen::VpxClkGenIo::FmcClk2OutputEnabled(bool enabled)
{
    Owner->PllDev->OutputEnabled(7, enabled);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::FmcClk2OutputFrequency() --
//---------------------------------------------------------------------------

void  Vpx_ClockGen::VpxClkGenIo::FmcClk2OutputFrequency(double frequency)
{
    Owner->PllDev->OutputFrequency(7, frequency);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::FmcClk2OutputFrequencyActual() --
//---------------------------------------------------------------------------

double  Vpx_ClockGen::VpxClkGenIo::FmcClk2OutputFrequencyActual()
{
    return Owner->PllDev->OutputFrequencyActual(7);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::FmcClk3OutputEnabled() --
//---------------------------------------------------------------------------

void  Vpx_ClockGen::VpxClkGenIo::FmcClk3OutputEnabled(bool enabled)
{
    Owner->PllDev->OutputEnabled(8, enabled);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::FmcClk3OutputFrequency() --
//---------------------------------------------------------------------------

void  Vpx_ClockGen::VpxClkGenIo::FmcClk3OutputFrequency(double frequency)
{
    Owner->PllDev->OutputFrequency(8, frequency);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::FmcClk3OutputFrequencyActual() --
//---------------------------------------------------------------------------

double  Vpx_ClockGen::VpxClkGenIo::FmcClk3OutputFrequencyActual()
{
    return Owner->PllDev->OutputFrequencyActual(8);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::Source() -- Retrieve clock source
//---------------------------------------------------------------------------

IVpxClkGenIo::VpxPllClockSource Vpx_ClockGen::VpxClkGenIo::Source() const
{
    if (Owner->PllDev->OutputSource() == Cdce72010::outPll)
        return csInternal;
    else
        return csExternal;
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::Source() -- Select clock source
//---------------------------------------------------------------------------

void Vpx_ClockGen::VpxClkGenIo::Source(VpxPllClockSource src)
{
    if (src == csInternal)
        Owner->PllDev->OutputSource(Cdce72010::outPll);
    else
        Owner->PllDev->OutputSource(Cdce72010::outAux);
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::ExternalClkSelect() -- Retrieve selected clock
//---------------------------------------------------------------------------

IVpxClkGenIo::VpxExtClockSource Vpx_ClockGen::VpxClkGenIo::ExternalClkSelect() const
{
    if (Owner->InMux().InputSource(1)==0)  // LClk is Primary
        return cslLClk;
    else
        return cslFmcClk1;                  // FmcClk1 is secondary
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::ExternalClkSelect() -- Select clock for external IN
//---------------------------------------------------------------------------
//  Switch OUT 1 and OUT 4 together

void Vpx_ClockGen::VpxClkGenIo::ExternalClkSelect(VpxExtClockSource src)
{
    switch (src)
        {
        case cslLClk:
            Owner->InMux().InputSource(1, 0);      // LClk is Primary
            Owner->InMux().InputSource(4, 0);
            break;
        default:
            Owner->InMux().InputSource(1, 1);      // FmcClk1 is secondary
            Owner->InMux().InputSource(4, 1);
            break;
        }
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::Reference() -- Retrieve reference source
//---------------------------------------------------------------------------

IVpxClkGenIo::VpxPllRefSource Vpx_ClockGen::VpxClkGenIo::Reference() const
{
    if (Owner->PllDev->ReferenceSource() == Cdce72010::refPrimary)
        return rsExternal;
    else
        return rsInternal;
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::Reference() -- Select reference source
//---------------------------------------------------------------------------

void Vpx_ClockGen::VpxClkGenIo::Reference(VpxPllRefSource src)
{
    if (src == rsInternal)
        Owner->PllDev->ReferenceSource(Cdce72010::refSecondary);
    else
        Owner->PllDev->ReferenceSource(Cdce72010::refPrimary);
    //
    //  Turn off Ref VCO if not using it
    //Owner->EnableRefVco( src == rsInternal );
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::ReferenceFrequency() -- Change reference frequency
//---------------------------------------------------------------------------
//  Since this system has 2 reference frequencies, we use an argument

void Vpx_ClockGen::VpxClkGenIo::ReferenceFrequency(VpxExtClockSource clk, double value)
{
    switch (clk)
        {
        case cslLClk:
            Owner->SetLClkFrequency(value);
            break;
        case cslFmcClk1:
            Owner->SetFmcClk1Frequency(value);
            break;
        default:
            break;
        }
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::ReferenceFrequency() -- Query reference frequency
//---------------------------------------------------------------------------
//  Since this system has 2 reference frequencies, we use an argument

double Vpx_ClockGen::VpxClkGenIo::ReferenceFrequency(VpxExtClockSource clk) const
{
    switch (clk)
        {
        case cslLClk:
            return Owner->LClkFrequency();
        case cslFmcClk1:
            return Owner->FmcClk1Frequency();
        default:
            return 0;
        }
}

//---------------------------------------------------------------------------
//  Vpx_ClockGen::VpxClkGenIo::Locked() -- Query PLL Locked state
//---------------------------------------------------------------------------

bool  Vpx_ClockGen::VpxClkGenIo::Locked() const
{
    return Owner->PllDev->Locked();
}

//==============================================================================
//  CLASS VpxCopFpaCpldPersonality -- VpxCop FPA Personality
//==============================================================================
//  Uses IX6FpaCpld to implement its personality

class VpxCopFpaCpldPersonality : public IVpxCopFpaCpldPersonality
{
public:
    VpxCopFpaCpldPersonality(IX6FpaCpld & cpld)
        :  IVpxCopFpaCpldPersonality(cpld)
        {}
    virtual ~VpxCopFpaCpldPersonality()
        {}
};

//===========================================================================
//  CLASS VpxCopFpaLoader -- VpxCop FPA Loader "SubDevice"
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class VpxCopFpaLoader
//---------------------------------------------------------------------------

const double cpldBusyWaitTimeout    = 0.001;  // in seconds
const double cpldReadValidTimeout   = 0.5;    // in seconds
const double cpldFlashReadyTimeout  = 0.5;    // in seconds

//---------------------------------------------------------------------------
//  constructor for class VpxCopFpaLoader
//---------------------------------------------------------------------------

VpxCopFpaLoader::VpxCopFpaLoader(IRequires_Nothing * mdc, IAddressingSpace & space, MapRegisters & regs)
    :  MDC(mdc),
       FpaAddress(space, regs.FpaAddress_Addr),
       FpaData(space, regs.FpaData_Addr),
       FpaCommand(space, regs.FpaCommand_Addr),
       FGoldenImageMode(false),
       CpldIntf(this),
       Thread( new VpxCopFpaCpldPersonality( CpldIntf ))
{
}

//---------------------------------------------------------------------------
//  destructor for class VpxCopFpaLoader
//---------------------------------------------------------------------------

VpxCopFpaLoader::~VpxCopFpaLoader()
{
    Thread.Shutdown();
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::OpenHardware()
//---------------------------------------------------------------------------

void  VpxCopFpaLoader::OpenHardware()
{
    if (!DeviceEnabled())
        return;

    Thread.OnOpen();
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::ConfigureHardware()
//---------------------------------------------------------------------------

void  VpxCopFpaLoader::ConfigureHardware()
{
    if (!DeviceEnabled())
        return;
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_Address()
//---------------------------------------------------------------------------

void  VpxCopFpaLoader::CPLD_Address(unsigned int addr)
{
    FpaAddress.Addr.Value(addr);
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_Data() -- write data
//---------------------------------------------------------------------------

void  VpxCopFpaLoader::CPLD_Data(unsigned int data)
{
    FpaData.Value(data);
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_Data()  -- read data
//---------------------------------------------------------------------------

unsigned int  VpxCopFpaLoader::CPLD_Data()
{
    return FpaData.Value();
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_Command()
//---------------------------------------------------------------------------

void  VpxCopFpaLoader::CPLD_Command(unsigned int code)
{
    FpaCommand.Command.Value(code);
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_Version()  -- read data
//---------------------------------------------------------------------------

unsigned int  VpxCopFpaLoader::CPLD_Version()
{
    return FpaCommand.CPLDVersion.Value();
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_FlashType()  -- read data
//---------------------------------------------------------------------------

unsigned int  VpxCopFpaLoader::CPLD_FlashType()
{
    return FpaCommand.FlashType.Value();
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_GoldenImageBooted()  -- read state of golden jumper
//---------------------------------------------------------------------------

bool  VpxCopFpaLoader::CPLD_GoldenImageBooted()
{
    return FpaCommand.GoldenImageBooted.Value()==false;  // 0 = Golden Boot
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_GoldenImageActive()  -- read golden image state
//---------------------------------------------------------------------------

bool  VpxCopFpaLoader::CPLD_GoldenImageActive()
{
    return FGoldenImageMode;
    //return FpaCommand.GoldenImageInUse.Value();
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_Busy()  -- read bit
//---------------------------------------------------------------------------

bool  VpxCopFpaLoader::CPLD_Busy()
{
    return FpaCommand.Busy.Value();
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_ReadValid()  -- read bit
//---------------------------------------------------------------------------

bool  VpxCopFpaLoader::CPLD_ReadValid()
{
    return FpaCommand.ReadValid.Value();
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_FlashReady()  -- read bit
//---------------------------------------------------------------------------

bool  VpxCopFpaLoader::CPLD_FlashReady()
{
    return FpaCommand.FlashReady.Value();
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_OverTemperature()  -- read bit (inverted)
//---------------------------------------------------------------------------

bool  VpxCopFpaLoader::CPLD_OverTemperature()
{
    return FpaCommand.OverTemp.Value()==0;
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_Write()  -- Write To Flash
//---------------------------------------------------------------------------

void  VpxCopFpaLoader::CPLD_Write(unsigned int addr, unsigned int data)
{
    CPLD_BusyWait();
    CPLD_Address(addr);

    CPLD_BusyWait();
    CPLD_Data(data);

    CPLD_BusyWait();
    CPLD_Command(cmdFlashWrite);

    CPLD_FlashReadyWait();
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_Read()  -- Read Flash
//---------------------------------------------------------------------------

unsigned int  VpxCopFpaLoader::CPLD_Read(unsigned int addr)
{
    CPLD_BusyWait();
    CPLD_Address(addr);

    CPLD_BusyWait();
    CPLD_Command(cmdFlashRead);

    CPLD_ReadValidWait();
    return CPLD_Data();
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_ReadPowerStatus()  -- Read Pwr Status
//---------------------------------------------------------------------------

unsigned int  VpxCopFpaLoader::CPLD_ReadPowerStatus()
{
    CPLD_BusyWait();
    CPLD_Command(cmdPowerStatus);

    CPLD_ReadValidWait();
    return CPLD_Data();
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_GoldenImageActive()  -- Make selected image current
//---------------------------------------------------------------------------

void  VpxCopFpaLoader::CPLD_GoldenImageActive(bool state)
{
    FGoldenImageMode = state;      // no longer read hardware for state

    //  Toggle the Golden/normal image mode
    // If we are already in the state, do nothing
    //if (CPLD_GoldenImageActive()==state)
    //    return;
    //CPLD_BusyWait();
    //CPLD_Command(cmdSwapImage);

    // Update status - automatic after command
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_RestartConfig()  -- Restart ROM confiugration
//---------------------------------------------------------------------------

void  VpxCopFpaLoader::CPLD_RestartConfig(bool restart_golden_image)
{
    // write a number not in the pattern to reset state machine
    CPLD_BusyWait();
    CPLD_Address(0x00000000);

    CPLD_BusyWait();
    CPLD_Address(0xAAAAAAAA);
    CPLD_BusyWait();
    CPLD_Address(0xBBBBBBBB);
    CPLD_BusyWait();
    CPLD_Address(0xCCCCCCCC);

    unsigned int image_code(0);
    if (restart_golden_image)
        image_code = 0x07000000;

    CPLD_BusyWait();
    CPLD_Address(image_code);    // tell which image to boot
    CPLD_BusyWait();
    CPLD_Command(cmdRestartCfg);
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_UpdateStatus()  -- Update status register
//---------------------------------------------------------------------------

void  VpxCopFpaLoader::CPLD_UpdateStatus()
{
    CPLD_BusyWait();
    CPLD_Command(cmdReadStatus);
}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_BusyWait()  -- Wait For BUSY flag
//---------------------------------------------------------------------------

void  VpxCopFpaLoader::CPLD_BusyWait()
{
    // Update status - automatic after command
    const double Timeout = cpldBusyWaitTimeout;   // in sec
    double Then = Rtc.Now() + Timeout;
	while (CPLD_Busy() == true)
        {
        if (Rtc.Now() > Then)
            {
            Debug.Log("Timeout waiting for CPLD Busy OFF");
            break;
            }
        }

}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_ReadValidWait()  -- Wait For Read Valid flag
//---------------------------------------------------------------------------

void  VpxCopFpaLoader::CPLD_ReadValidWait()
{
    RefreshStatus();
    // Update status - automatic after command
    const double Timeout = cpldReadValidTimeout;   // in sec
    double Then = Rtc.Now() + Timeout;
	while (CPLD_ReadValid() == false)
        {
        RefreshStatus();
        if (Rtc.Now() > Then)
            {
            Debug.Log("Timeout waiting for CPLD Read Valid ON");
            break;
            }
        }

}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::CPLD_FlashReadyWait()  -- Wait For Flash Ready flag
//---------------------------------------------------------------------------

void  VpxCopFpaLoader::CPLD_FlashReadyWait()
{
    RefreshStatus();
    // Update status - automatic after command
    const double Timeout = cpldFlashReadyTimeout;   // in sec
    double Then = Rtc.Now() + Timeout;
	while (CPLD_FlashReady() == false)
        {
        RefreshStatus();
        if (Rtc.Now() > Then)
            {
            Debug.Log("Timeout waiting for CPLD Flash Ready ON");
            break;
            }
        }

}

//---------------------------------------------------------------------------
//  VpxCopFpaLoader::RefreshStatus()  -- Update Status Register Bits
//---------------------------------------------------------------------------

void  VpxCopFpaLoader::RefreshStatus()
{
    CPLD_UpdateStatus();
}


#ifdef _MSC_VER
    #pragma warning(pop)
#endif


}  // namespace Innovative

