// X3AlertSystem.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <X3AlertSystem_Mb.h>
#include <X3DeviceMap_Mb.h>
#include <X3MemoryMap_Mb.h>
#include <BufferDatagrams_Mb.h>

namespace Innovative
{

using namespace X3_MemoryMap;

//===========================================================================
//  CLASS X3AlertSystem
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class X3AlertSystem
//---------------------------------------------------------------------------

const int X3AlertPid = 0xff;

//---------------------------------------------------------------------------
//  constructor for class X3AlertSystem
//---------------------------------------------------------------------------

X3AlertSystem::X3AlertSystem(X3DeviceMap * device)
	:  AlertSystem(X3AlertPid), Device(device),
	   AlertControlReg( device->LogicMemory, mmAlertControl ),
	   AlertEnableReg(  device->LogicMemory, mmAlertEnable ),
	   SoftwareAlertReg( device->LogicMemory, mmAlertSWTrigger ),
       FRolloverCount(0)
{
    SenderTag(reinterpret_cast<size_t>(this));

	// Load Alerts
	LoadAlert(AlertData((int)alertTimeStampRollover,       0, 1 << 0, false, &OnTimeStampRolloverAlert));
	LoadAlert(AlertData((int)alertSoftware,                1, 1 << 1, false, &OnSoftwareAlert));
	LoadAlert(AlertData((int)alertWarningTemperature,      3, 1 << 3, false, &OnWarningTemperature));
	LoadAlert(AlertData((int)alertPllLost, 			       4, 1 << 4, false, &OnPllLost));
	LoadAlert(AlertData((int)alertInputFifoOverrun,	       8, 1 << 8, false, &OnInputFifoOverrun));
	LoadAlert(AlertData((int)alertInputTrigger,	 	       9, 1 << 9, false, &OnInputTrigger));
	LoadAlert(AlertData((int)alertInputPretriggerError,   10, 1 << 10, false, &OnInputPretriggerError));
	LoadAlert(AlertData((int)alertInputStopTrigger,       11, 1 << 11, false, &OnInputStopTrigger));
	LoadAlert(AlertData((int)alertInputOverrange, 	      16, 1 << 16, false, &OnInputOverrange));
	LoadAlert(AlertData((int)alertOutputFifoUnderrun,     24, 1 << 24, false, &OnOutputFifoUnderrun));
	LoadAlert(AlertData((int)alertOutputTrigger,          25, 1 << 25, false, &OnOutputTrigger));
}

//---------------------------------------------------------------------------
//  X3AlertSystem::AlertReset() - Reset Alert Fifo
//---------------------------------------------------------------------------

void  X3AlertSystem::AlertReset()
{
    AlertControlReg.AlertFifoReset.Set();
    AlertControlReg.AlertFifoReset.Clear();

    FRolloverCount = 0;
}

//---------------------------------------------------------------------------
//  X3AlertSystem::SoftwareAlert() - Reset Alert Fifo
//---------------------------------------------------------------------------

void  X3AlertSystem::SoftwareAlert(unsigned int tag)
{
    SoftwareAlertReg.Value(tag);
}

//---------------------------------------------------------------------------
//  X3AlertSystem::Open() - Open-Time behavior
//---------------------------------------------------------------------------

void  X3AlertSystem::Open()
{
    // nothing special
}

//---------------------------------------------------------------------------
//  X3AlertSystem::Configure() - Configure-Time behavior
//---------------------------------------------------------------------------

void  X3AlertSystem::Configure()
{
	AlertReset();
	//
	//  Turn on Timestamp Clock
	AlertControlReg.TimestampOn.Set();

	// Enable all configured alerts
	for (unsigned int i=0; i<Alerts.size(); ++i)
		AlertEnableReg.Bit(Alerts[i].EnableBit, Alerts[i].Enabled);
}

//---------------------------------------------------------------------------
//  X3AlertSystem::Append() - Add a custom user alert
//---------------------------------------------------------------------------

void X3AlertSystem::Append(const AlertData & alert)
{
    LoadAlert(alert);
}

//------------------------------------------------------------------------------
//   AlertSystem::ExtractAlertArgument() -- Return alert argument
//------------------------------------------------------------------------------
//   Crack the argument out of the raw alert

int  X3AlertSystem::ExtractAlertArgument(const Buffer & alert, int bit, int field)
{
    int result;
    const IntegerDG  alertDG(alert);

    switch (bit)
        {
        case 0:     // TimeStampRollover
            result = ++FRolloverCount;
            break;
        case 1:     // Software
            result = alertDG[1+2];     // returns Tag    // bit # plus 2 offse
            break;
        case 3:     //WarningTemperature
            result = alertDG[3+2];
            break;
        case 4:     // PllLost
            result = 0;
            break;
        case 8:     // InputFifoOverrun
            result = 0;
            Device->UserFpgaControlReg.InMqReset.Toggle();
            break;
        case 9:     // InputTrigger
            result = alertDG[9+2];           // sample number at start trigger
            break;
        case 10:    // InputPrefillError
            result = alertDG[10+2];          // # points actually prefilled
            break;
        case 11:     // InputStopTrigger
            result = alertDG[11+2];          // sample number at stop trigger
            break;
        case 16:    // InputOverrange
            result = 0;
            break;
        case 24:    // OutputFifoUnderrun
            result = 0;
            Device->UserFpgaControlReg.OutMqReset.Toggle();
            break;
        case 25:    // OutputTrigger
            result = 0;
            break;
        default:    // Spurious, unknown
            result = alertDG[field+2];
            break;
        }

    return result;
}


//===========================================================================
//  CLASS X3AlertSystem::AlertControlsRegister
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class X3AlertSystem::AlertControlsRegister
//---------------------------------------------------------------------------

X3AlertSystem::AlertControlsRegister::AlertControlsRegister(AddressingSpace &space, int offset )
	: inherited( space, offset ),
	  AlertFifoReset( *this, 1 ),
	  TimestampOn( *this, 0 ),
      Pdn(*this, 24, 8)
{
    Pdn.Value(X3AlertPid);
}

//===========================================================================
//  CLASS X3AlertSystem::AlertEnableRegister
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class X3AlertSystem::AlertEnableRegister
//---------------------------------------------------------------------------

X3AlertSystem::AlertEnableRegister::AlertEnableRegister(AddressingSpace &space, int offset )
	: inherited( space, offset ),
	  TimestampRollover( *this, 0 ),
      Software( *this, 1 ),
      WarningTemperature( *this, 3 ),
      PllLost( *this, 4 ),
      InputFifoOverrun( *this, 8 ),
      InputTrigger( *this, 9 ),
      InputPretriggerError( *this, 10 ),
      InputStopTrigger( *this, 11 ),
      InputOverrange( *this, 16 ),
      OutputFifoUnderrun( *this, 24 ),
      OutputTrigger( *this, 25 )
{
}

} // namespace Innovative









