// X3_10MDeviceMap_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <X3_10M_Mb.h>
#include <X3_10MDeviceMap_Mb.h>
#include <X3MemoryMap_Mb.h>
#include <SystemSupport_Mb.h>

//---------------------------------------------------------------------------
namespace Innovative
{

using namespace X3_MemoryMap;

//---------------------------------------------------------
//  Clock and Module Constants Definition

const int    MaxDevices        = 4;
const int    InputWidth_Bits   = 16;
const int    InputMuxCount     = 0;
const bool   InputIsUnipolar   = false;
const int    InputClockFactor  = 1;
const float  InputMaxRate      = 25.0e6;
const float  InputMinRate      = 3.125e6;
const int    InputGangSize     = 2;
const float  InputReferenceClk = 100.0e6;
const float  InputVcoMinClk    = 100.0e6;
const float  InputVcoMaxClk    = 140.0e6;

//---------------------------------------------------------
//  IDROM Section Map
//  ...First section: Input Coefficient section
//       (8 channels x 2 values + 1 calibration flag + 1 spare) x 4 ranges
const int romX3_10M_InputCoeffOffset = 0;
const int romX3_10M_Channels = 8;
const int romX3_10M_Ranges = 4;
const int romX3_10M_RangeSize = romX3_10M_Channels * 2 + 1 + 1;
const int romX3_10M_RangeCount = 4;
const int romX3_10M_InputCoeffSize = romX3_10M_RangeSize * romX3_10M_RangeCount + 4;

//  ...Second  section: Clock Parameters
//       (1 param + 1 calibrated word + 2 spare)
const int romX3_10M_ClockParamOffset = romX3_10M_InputCoeffSize;
const int romX3_10M_ClockParamSize = 4;

class X3_10MInfoIntf : public IoInfoIntf
{
public:
	X3_10MInfoIntf()
		: FClockFactor(InputClockFactor),
			   //    devices,           bits,    muxChannelCount,   unipolar
		  FChannels(MaxDevices, InputWidth_Bits, InputMuxCount,  InputIsUnipolar)
        {
        FSpan.push_back(SpanInfo(2000, -2000));
        FSpan.push_back(SpanInfo(1000, -1000));
        FSpan.push_back(SpanInfo(400, -400));
        FSpan.push_back(SpanInfo(200, -200));
        }
	virtual ~X3_10MInfoIntf()  {  }

	virtual PairedChannelInfo &	    	Channels() 		        {  return FChannels;  }
	virtual float               		MaxRate() const			{  return InputMaxRate;  }
	virtual float               		MinRate() const			{  return InputMinRate;  }
	virtual const SpanArray &   		Span() const			{  return FSpan;  }
	virtual int                 		GangSize() const		{  return InputGangSize;  }
	virtual bool                		IsStackable() const		{  return false;  }
	virtual bool                		IsInverted() const		{  return true;  }
	virtual bool                		IsMuxed() const			{  return false;  }
	virtual bool                		MuxDepth() const		{  return InputMuxCount;  }
	virtual int                 		BufferSize() const		{  return 65536;  }
	virtual int                 		ClockFactor() const		{  return FClockFactor;  }
	virtual float               		ReferenceClock() const	{  return InputReferenceClk;  }
	virtual int               	        SamplesPerFrame() const {  return 1;  }

    virtual bool                    HasPretriggering() const    {  return true;  }
    virtual bool                    HasTriggerDelay() const     {  return true;  }
    
	void                                SetClockFactor(unsigned int factor)
											{  FClockFactor = factor;  }
private:
	unsigned int                FClockFactor;
	PairedChannelInfo	        FChannels;
	SpanArray					FSpan;
};

//===========================================================================
//  CLASS X3_10MInputDevice  -- Register & low level support for X3
//===========================================================================

class X3_10MInputDevice : public X3RangedIoDevice
{
	friend class X3_10MDeviceMap;

public:
    // Ctor
	X3_10MInputDevice(X3DeviceMap * owner, X3IoDeviceRegisters & reg)
        : X3RangedIoDevice(owner, reg, new X3_10MInfoIntf, 0x10000)  /* 'one' value for gain registers */
        {
        for (int i = 0; i < romX3_10M_Ranges; ++i)
            {
            AppendRange(PmcIdromSection(owner->Rom(), PmcIdrom::waModule,
                i * romX3_10M_RangeSize, romX3_10M_RangeSize));
            }
        }
	// Dtor
	~X3_10MInputDevice()
        {
        }
};

static X3IoDeviceRegisters X3_10MInputDeviceRegisters =
    { mmI2c, 0,
      mmInControl, mmInGainCoef, mmInOffsetCoef, mmInGainMode,
      mmInChannelEnable, mmInAmpEnable, mmInTriggerControl,
      mmInDecimation, mmInPacketHeader, mmInTriggerDelay
    };

//=============================================================================
//  CLASS X3_10MClock  --  PLL timebase control
//=============================================================================

class X3_10MClock : public Ad9511Pfd
{
    typedef Ad9511Pfd inherited;

public:
	enum SamplingMode { smSingle, smDual, smQuad };
	enum ClockFactors { cf128, cf256, cf512 };
	//
	//  Member Functions
	X3_10MClock(X3_10MDeviceMap * owner, int index, AddressingSpace &space, int ctl)
		: inherited(index, space, ctl), Owner(owner), FPDF(100.0e3)
		{
		Reference(InputReferenceClk);
		Min(InputVcoMinClk/MaxOutputDivider());
		Max(InputVcoMaxClk);
		PFDPolarity(true);
        Tuner.PfdRate(PhaseDetectorFrequency());
		}
	X3_10MClock(X3_10MDeviceMap * owner, double freq, double ref,
		bool enabled, int index, AddressingSpace & space, int ctl)
		: inherited(freq, ref, enabled, index, space, ctl), Owner(owner), FPDF(100.0e3)
		{
		Min(InputVcoMinClk/MaxOutputDivider());
		Max(InputVcoMaxClk);
		PFDPolarity(true);
		Tuner.PfdRate(PhaseDetectorFrequency());
        }

	//  Set clock output modes and signalling types
	void ConfigureOutputs(int outdiv_bits, int outdiv_bypass)
	{
        PllConfigureOutputEvent e(*this, outdiv_bits, outdiv_bypass);
        OnPllConfigureOutput.Execute(e);

		// Identical duty cycle and phase for all outputs
		for (int i = 0; i < 8; ++i)
			{
			Port(0x48+i*2, outdiv_bits);	    // duty cycle
			Port(0x48+(i*2+1), outdiv_bypass);	// phase
			}

		Port(0x3C, 0x08);       // Clock 0 LVPECL on @810mV
		Port(0x3D, 0x03);       // Clock 1 disabled
		Port(0x3E, 0x03);       // Clock 2 disabled
		Port(0x3F, 0x03);       // Clock 3 disabled

		Port(0x40, 0x0a);       // Clocks 4..7 LVCMOS on @3.5 mA
		Port(0x41, 0x0a);
		Port(0x42, 0x0a);
		Port(0x43, 0x0a);

        PllConfigureOutputEvent e1(*this, outdiv_bits, outdiv_bypass);
        OnAfterPllConfigureOutput.Execute(e1);
	}

	void SetReferenceCalibrationFactor(double value)
	{
	Owner->ClockParamRomSection.AsFloat(0, static_cast<float>(value));
	// set calibrated field
	Owner->ClockParamRomSection.AsInt(1, static_cast<int>('Y'));
	}
	double GetReferenceCalibrationFactor() const
	{
	//  Check Calibrated bit
	if (Owner->ClockParamRomSection.AsInt(1)=='Y')
		{
		return static_cast<double>(Owner->ClockParamRomSection.AsFloat(0));
		}
	else
		return 1.0;  // default value
	}

protected:
	// Data
	X3_10MDeviceMap *         Owner;
    double                    FPDF;

	virtual void    SetFrequency(double freq)
	{
        Tuner.PfdRate(PhaseDetectorFrequency());
		Ad9511::SetFrequency(freq);
		Owner->ResetAnalog();
	}

    virtual double          PhaseDetectorFrequency() const
                                {  return FPDF;   }
    virtual void            SetPhaseDetectorFrequency(double freq)
                               {  FPDF = freq;   }


};

//===========================================================================
//  CLASS X3_10MDeviceMap  -- Register & low level support for SD module
//===========================================================================

//---------------------------------------------------------------------------
//  constructor
//---------------------------------------------------------------------------

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif

X3_10MDeviceMap::X3_10MDeviceMap(X3_10M * owner)
	: IUsesStreamDeviceEventCallbacks(owner),
	  FInput(new X3_10MInputDevice(this, X3_10MInputDeviceRegisters)),
	  Owner(owner),
	  Clock(new X3_10MClock(this, 0, LogicMemory, mmPll)),
	  ClockParamRomSection(Rom(), PmcIdrom::waModule,
						   romX3_10M_ClockParamOffset, romX3_10M_ClockParamSize)
{
	//
	//  Hook stream events from base class
	PDMPtr()->OnStreamConfigure.SetEvent(this,   &X3_10MDeviceMap::OnStreamConfigureHandler );
	PDMPtr()->OnBeforeStreamStart.SetEvent(this, &X3_10MDeviceMap::OnBeforeStreamStartHandler );
	PDMPtr()->OnAfterStreamStart.SetEvent(this,   &X3_10MDeviceMap::OnAfterStreamStartHandler );
	PDMPtr()->OnBeforeStreamStop.SetEvent(this,   &X3_10MDeviceMap::OnBeforeStreamStopHandler );
	PDMPtr()->OnAfterStreamStop.SetEvent(this,   &X3_10MDeviceMap::OnAfterStreamStopHandler );

}

#ifdef _MSC_VER
	#pragma warning(pop)
#endif

//---------------------------------------------------------------------------
//  destructor
//---------------------------------------------------------------------------

X3_10MDeviceMap::~X3_10MDeviceMap()
{
	delete Clock;
    Clock = 0;

	delete FInput;
	FInput = 0;
}

//---------------------------------------------------------------------------
//  X3_10MDeviceMap::OpenHardware()
//---------------------------------------------------------------------------

void  X3_10MDeviceMap::OpenHardware()
{
	X3DeviceMap::OpenHardware();

	// Configure A/Ds
    for (size_t range = 0; range < FInput->Ranges(); ++range)
		{
		for (int ch = 0; ch < Input().Info().Channels().Channels(); ++ch)
			{
            Input().PacketId(0x0);
		    Input().PacketSize(0x4000);
			FInput->Gain(range, ch, 1.0);
			FInput->Offset(range, ch, 0.0);
			Input().Unframed();
			}
		}
}

//---------------------------------------------------------------------------
//  X3_10MDeviceMap::OnStreamConfigureHandler()
//---------------------------------------------------------------------------

void  X3_10MDeviceMap::OnStreamConfigureHandler(OpenWire::NotifyEvent & Event)
{
    // Family-common init
	X3DeviceMap::ConfigureHardware();
    
    // Reapply Hardware Settings for  X3 IO Devices
    FInput->ConfigureHardware();

    // Clear any FIFO overrun condition
    FInput->AnalogControlReg.FailedReset.Toggle();

	//
	//  Apply Clock Settings from saved state
	Clock->Reapply();

	//
	//  Apply Gain/Offset Coefficients
    for (size_t i = 0; i < FInput->Ranges(); ++i)
        {
        if (FInput->Calibrated(i))
            FInput->ApplyRomCoefficients(i);
        else
            FInput->ApplyDefaultCoefficients(i);
        }

	//
	//  Apply input settings
	for (int ch = 0; ch < FInput->Info().Channels().Channels(); ++ch)
		{
		FInput->ChannelEnableReg.Bit(ch, FInput->Info().Channels().Enabled(ch));
        Input().PacketId(Input().PacketId());
		Input().PacketSize(Input().PacketSize());
        FInput->Range(ch, FInput->Range(ch));
		}

	// Configure triggers
	FInput->TriggerReg.Apply();

	//  Forward notify to baseboard handler
    OwnerCB()->CallOnStreamConfigure(Event);
}

//---------------------------------------------------------------------------
//  X3_10MDeviceMap::OnBeforeStreamStartHandler()
//---------------------------------------------------------------------------

void  X3_10MDeviceMap::OnBeforeStreamStartHandler(OpenWire::NotifyEvent & Event)
{
	//  Forward notify to baseboard handler
    OwnerCB()->CallOnBeforeStreamStart(Event);
	//
	//  Start execution
	UserFpgaControlReg.InGo.Set();
	UserFpgaControlReg.Apply();
}

//---------------------------------------------------------------------------
//  X3_10MDeviceMap::OnAfterStreamStartHandler()
//---------------------------------------------------------------------------

void  X3_10MDeviceMap::OnAfterStreamStartHandler(OpenWire::NotifyEvent & Event)
{
	//  Forward notify to baseboard handler
    OwnerCB()->CallOnAfterStreamStart(Event);
}

//---------------------------------------------------------------------------
//  X3_10MDeviceMap::OnBeforeStreamStopHandler()
//---------------------------------------------------------------------------

void  X3_10MDeviceMap::OnBeforeStreamStopHandler(OpenWire::NotifyEvent & Event)
{
	//  Forward notify to baseboard handler
    OwnerCB()->CallOnBeforeStreamStop(Event);
}

//---------------------------------------------------------------------------
//  X3_10MDeviceMap::OnAfterStreamStopHandler()
//---------------------------------------------------------------------------

void  X3_10MDeviceMap::OnAfterStreamStopHandler(OpenWire::NotifyEvent & Event)
{
	//  Disable User Logic Streaming
	UserFpgaControlReg.InGo.Clear();
	UserFpgaControlReg.Apply();

	Trigger(tInput, false);

	//  Forward notify to baseboard handler
    OwnerCB()->CallOnAfterStreamStop(Event);
}

//---------------------------------------------------------------------------
//  X3_10MDeviceMap::Input() -- Reference to input object
//---------------------------------------------------------------------------

X3RangedIoDevice & X3_10MDeviceMap::Input()
{
	return *FInput;
}

//---------------------------------------------------------------------------
//  X3_10MDeviceMap::ResetAnalog() -- Reset Analog Chip
//---------------------------------------------------------------------------

void  X3_10MDeviceMap::ResetAnalog()
{
	//  Reset analog
	FInput->AnalogControlReg.Reset.Set();
	FInput->AnalogControlReg.Reset.Clear();
	Sleep(1);   // delay 1 ms

}


} // namespace Innovative


