// X5AlertSystem.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <X5AlertSystem_Mb.h>
#include <X5DeviceMap_Mb.h>
#include <X5MemoryMap_Mb.h>
#include <BufferDatagrams_Mb.h>

namespace Innovative
{

using namespace X5_MemoryMap;

//===========================================================================
//  CLASS X5AlertSystem
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class X5AlertSystem
//---------------------------------------------------------------------------

const int X5AlertPid = 0xff;

//---------------------------------------------------------------------------
//  constructor for class X5AlertSystem
//---------------------------------------------------------------------------

X5AlertSystem::X5AlertSystem(X5DeviceMap * device)
	:  AlertSystem(X5AlertPid), Device(device),
	   AlertControlReg( device->LogicMemory, mmAlertControl ),
	   AlertEnableReg(  device->LogicMemory, mmAlertEnable ),
	   SoftwareAlertReg( device->LogicMemory, mmAlertSWTrigger ),
       FRolloverCount(0)
{
    InitAlerts();
}

//---------------------------------------------------------------------------
//  constructor for class X5AlertSystem
//---------------------------------------------------------------------------

X5AlertSystem::X5AlertSystem(AddressingSpace & space, int alertControl, int alertEnable, int alertSWTrigger)
	:  AlertSystem(X5AlertPid), Device(0),
	   AlertControlReg( space, alertControl ),
	   AlertEnableReg(  space, alertEnable ),
	   SoftwareAlertReg( space, alertSWTrigger ),
       FRolloverCount(0)
{
    InitAlerts();
}

//---------------------------------------------------------------------------
//  X5AlertSystem::InitAlerts() - Initialize Alert Database
//---------------------------------------------------------------------------

void  X5AlertSystem::InitAlerts()
{
    SenderTag(reinterpret_cast<size_t>(this));

	// Load Alerts
	LoadAlert(AlertData((int)alertTimeStampRollover,    0, 1 << 0,  false, &OnTimeStampRolloverAlert));
	LoadAlert(AlertData((int)alertSoftware,             1, 1 << 1,  false, &OnSoftwareAlert));
	LoadAlert(AlertData((int)alertPatternDone,          2, 1 << 2,  false, &OnPatternDoneAlert));
    LoadAlert(AlertData((int)alertWarningTemperature,   3, 1 << 3 , false, &OnWarningTemperature));
	LoadAlert(AlertData((int)alertInputFifoOverrun,	    8, 1 << 8,  false, &OnInputFifoOverrun));
	LoadAlert(AlertData((int)alertInputTrigger,	        9, 1 << 9,  false, &OnInputTrigger));
	LoadAlert(AlertData((int)alertInputOverrange, 	   16, 1 << 16, false, &OnInputOverrange));
	LoadAlert(AlertData((int)alertOutputFifoUnderrun,  24, 1 << 24, false, &OnOutputFifoUnderrun));
	LoadAlert(AlertData((int)alertOutputTrigger ,	   25, 1 << 25, false, &OnOutputTrigger));
	LoadAlert(AlertData((int)alertNumericError ,	   26, 1 << 26, false, &OnNumericError));
    LoadAlert(AlertData((int)alertDdrDacQAlmostFull,   27, 1 << 27, false, &OnDdrDacQAlmostFull));

}

//---------------------------------------------------------------------------
//  X5AlertSystem::AlertReset() - Reset Alert Fifo
//---------------------------------------------------------------------------

void  X5AlertSystem::AlertReset()
{
    AlertControlReg.AlertFifoReset.Set();
    AlertControlReg.AlertFifoReset.Clear();

    FRolloverCount = 0;
}

//---------------------------------------------------------------------------
//  X5AlertSystem::SoftwareAlert() - Reset Alert Fifo
//---------------------------------------------------------------------------

void  X5AlertSystem::SoftwareAlert(unsigned int tag)
{
    SoftwareAlertReg.Value(tag);
}

//---------------------------------------------------------------------------
//  X5AlertSystem::Open() - Open-Time behavior
//---------------------------------------------------------------------------

void  X5AlertSystem::Open()
{
    // nothing special
}

//---------------------------------------------------------------------------
//  X5AlertSystem::Configure() - Configure-Time behavior
//---------------------------------------------------------------------------

void  X5AlertSystem::Configure()
{
	AlertReset();
	//
	//  Turn on Timestamp Clock
	AlertControlReg.TimestampOn.Set();

	// Enable all configured alerts
	for (unsigned int i=0; i<Alerts.size(); ++i)
		AlertEnableReg.Bit(Alerts[i].EnableBit, Alerts[i].Enabled);
}

//---------------------------------------------------------------------------
//  X5AlertSystem::Append() - Add a custom user alert
//---------------------------------------------------------------------------

void X5AlertSystem::Append(const AlertData & alert)
{
    LoadAlert(alert);
}

//------------------------------------------------------------------------------
//   AlertSystem::ExtractAlertArgument() -- Return alert argument
//------------------------------------------------------------------------------
//   Crack the argument out of the raw alert

int  X5AlertSystem::ExtractAlertArgument(const Buffer & alert, int bit, int field)
{
    int result;
    const IntegerDG  alertDG(alert);

    switch (bit)
        {
        case 0:     // TimeStampRollover
            result = ++FRolloverCount;
            break;
        case 1:     // Software
            result = alertDG[3];
            break;
        case 2:     // PatternDone
            result = alertDG[4];
            break;
        case 3:     //TemperatureWarning
            result = alertDG[5];
            break;
        case 8:     // InputFifoOverrun
            result = alertDG[10];
            break;
        case 9:     // InputTrigger
            result = alertDG[11];
            break;
        case 16:    // InputOverrange
            result = alertDG[18];
            break;
        case 24:    // OutputFifoUnderrun
            result = alertDG[26];
            break;
        case 25:    // OutputTrigger
            result = alertDG[27];
            break;
        case 26:    // NumericError
            result = alertDG[28];
            break;
        case 27:     // OutputQueueAlmostFull
            result = alertDG[29];
            break;
        default:    // Spurious, unknown
            result = alertDG[field+2];
            break;
        }

    return result;
}

//===========================================================================
//  CLASS X5AlertSystem::AlertControlsRegister
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class X3AlertSystem::AlertControlsRegister
//---------------------------------------------------------------------------

X5AlertSystem::AlertControlsRegister::AlertControlsRegister(AddressingSpace &space, int offset )
	: inherited( space, offset ),
	  AlertFifoReset( *this, 1 ),
	  TimestampOn( *this, 0 ),
      Pdn(*this, 24, 8)
{
    Pdn.Value(X5AlertPid);
}

//===========================================================================
//  CLASS X5AlertSystem::AlertEnableRegister
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class X5AlertSystem::AlertEnableRegister
//---------------------------------------------------------------------------

X5AlertSystem::AlertEnableRegister::AlertEnableRegister(AddressingSpace &space, int offset )
	: inherited( space, offset ),
	  TimestampRollover ( *this, 0 ),
      Software          ( *this, 1 ),
      PatternDone       ( *this, 2 ),
      WarningTemperature( *this, 3 ),
      InputFifoOverrun  ( *this, 8 ),
      InputTrigger      ( *this, 9 ),
      InputOverrange    ( *this, 16),      
      OutputFifoUnderrun( *this, 24),
      OutputTrigger     ( *this, 25),
      NumericError      ( *this, 26),
	  DdrDacQAlmostFull ( *this, 27)
{
}

} // namespace Innovative









