// X5_G12DeviceMap_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <X5_G12_Mb.h>
#include <X5_G12DeviceMap_Mb.h>
#include <X5MemoryMap_Mb.h>
#include <X5_G12MemoryMap_Mb.h>
#include <Magic_Mb.h>
#include <sstream>
#include <cmath>
#include <SystemSupport_Mb.h>

//---------------------------------------------------------------------------
namespace Innovative
{

using namespace X5_MemoryMap;
using namespace X5_G12MemoryMap;

//---------------------------------------------------------
//  Clock and Module Constants Definition
const double  kRefVcxoPowerOnRate = 100.0;    // Ref Si570 power on rate (MHz)
const double  kVcxoPowerOnRate = 10.0;        // Si570 power on rate (MHz)
const int     VcxoI2C_SubAddr  = 0xAA;        //

//---------------------------------------------------------
//  Module Constants Definition, Clock is either external or internal fixed crystal.

const int    MaxHsDevices      = 2;
const int    MaxLsDevices      = 2;
const int    MaxInDevices      = MaxHsDevices + MaxLsDevices;
const int    InputWidth_Bits   = 12;
const int    InputMuxCount     = 0;
const bool   InputIsUnipolar   = false;
const int    InputClockFactor  = 1;
const float  InputMaxRate      = 1000.0e6f;
const float  InputMinRate      = 100.0e6f;
const int    InputGangSize     = 1;
const float  Input_ReferenceClk = 100.0e6;
const int    InputMax_mV       = 1000;

const float  DesMinRate        = InputMaxRate;

//---------------------------------------------------------
//  IDROM Section Map
//  ...First section: Input and Output Coefficient section
//       2channels x 2 values + 1 calibration flag + 1 spare
const int romX5_G12_InputCoeffOffset   = 0;
const int romX5_G12_InChannels         = MaxInDevices;
const int romX5_G12_InputCoeffSize     = romX5_G12_InChannels * 2 + 1 + 1;

//  ...Second section: Phase Offset Coefficient section
//       2channels x 1 values + 1 calibration flag + 1 spare
const int romX5_G12_PhaseCoeffOffset   = romX5_G12_InputCoeffOffset + romX5_G12_InputCoeffSize;
const int romX5_G12_PhaseCoeffSize     = romX5_G12_InChannels * 1 + 1 + 1;
const int romX5_G12_Phase_CalibratedIdx  = romX5_G12_PhaseCoeffSize-1;

//const int romX5_G12_CoeffSize          = romX5_G12_InputCoeffSize + romX5_G12_PhaseCoeffSize;

//=============================================================================
//  CLASS X5_G12_Pll  --  PLL timebase control
//=============================================================================
//---------------------------------------------------------------------------
//  housekeeping for class X5_G12_Pll
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//  constructor for class X5_G12_Pll
//---------------------------------------------------------------------------

X5_G12_Pll::X5_G12_Pll(X5_G12DeviceMap * owner, AddressingSpace &space, Cdce72010Registers & reg)
    : Cdce72010(space, reg), Owner(owner)
{
    SetMasterOutputSelection(1);  // Use Output 1 as guideline output

    ReferenceFrequency(0, Input_ReferenceClk);  // Claim Front Panel Ref is 100
    ReferenceFrequency(1, Input_ReferenceClk);
}

//---------------------------------------------------------------------------
//  X5_G12_Pll::Vcxo() -- return Vcxo
//---------------------------------------------------------------------------

IVcxoControl &  X5_G12_Pll::Vcxo()
{
    return Owner->Vcxo;
}

//---------------------------------------------------------------------------
//  X5_G12_Pll::OnInit() -- inital configuration
//---------------------------------------------------------------------------

void  X5_G12_Pll::OnInit()
{
    //  Devcie Powered/Enabled Setup
    SetPllEnable(true);
    SetFBDividerEnable(true);
    //
    //  Set Input Configuration
    SetInputConfig(icLvPecl);
    SetInputBufferConfig(true,   // DC Terminated
                         false,  // Input Term NOT disabled
                         false,  // Primary Bias OFF
                         false,  // Secondary Bias OFF
                         false   // NO failsafe
                         );
    //
    //  Set Vco/Aux Configuration
    SetVcoConfig(vcLvPecl);
    SetVcoBufferConfig(true,   // DC Terminated
                       false,  // Hysteresis
                       false,  // Input Term disabled
                       false   // Vco Bias OFF
                       );
    //
    //  Set Output Configuration
    OutputInUse(0, false);  SetOutputConfig(0, ocDisabled);
    OutputInUse(1, true);   SetOutputConfig(1, ocLvPecl);       // A/D 0
    OutputInUse(2, true);   SetOutputConfig(2, ocLvPecl);       // A/D 1
    OutputInUse(3, false);  SetOutputConfig(3, ocDisabled);
    OutputInUse(4, true);   SetOutputConfig(4, ocLvPecl);       //  Trigger Clock
    OutputInUse(5, false);  SetOutputConfig(5, ocDisabled);
    OutputInUse(6, false);  SetOutputConfig(6, ocDisabled);
    OutputInUse(7, false);  SetOutputConfig(7, ocDisabled);
    OutputInUse(8, false);  SetOutputConfig(8, ocDisabled);
    OutputInUse(9, false);  SetOutputConfig(9, ocDisabled);   // used as input
    //  Configure Aux
    SetAuxInputConfig(aeEnabled, aiNormal);
    SetLockConfig(false);   // NOT analog lock
    //
    //  Set Frequency Limits for Outputs (Preset to 1500.0 MHz)
    //    (Outputs 0 and 1 match preset value, so leave alone)
    //OutputFrequencyLimit[6] = 100.0e6;
    //OutputFrequencyLimit[8] = 250.0e6;
    //
    //  Reference
    SetExternalReferenceControl(true);  // Use internal bits, not external pin
}

//---------------------------------------------------------------------------
//  X5_G12_Pll::OnReapply()
//---------------------------------------------------------------------------

void  X5_G12_Pll::OnReapply()
{
}

//===========================================================================
// class X5_G12InputInfoIntf
//===========================================================================

class X5_G12InputInfoIntf : public IoInfoIntf
{
public:
    X5_G12InputInfoIntf(X5_G12DeviceMap * owner)
        : //    devices,           bits,    muxChannelCount,   unipolar
          FChannels(MaxInDevices, InputWidth_Bits, InputMuxCount, InputIsUnipolar),
          Owner(owner)
        {  FSpan.push_back(SpanInfo(InputMax_mV, -InputMax_mV));  }
    virtual ~X5_G12InputInfoIntf()  {  }

    virtual UnpairedChannelInfo &       Channels()              {  return FChannels;}
    virtual float  MaxRate() const
                {
                    if (Owner->Input().DESModeCapable())
                        return InputMaxRate*2;
                    else
                        return InputMaxRate;
                }
    virtual float                       MinRate() const         {  return InputMinRate;  }
    virtual const SpanArray &           Span() const            {  return FSpan;  }
    virtual int                         GangSize() const        {  return InputGangSize;  }
    virtual bool                        IsStackable() const     {  return true;  }
    virtual bool                        IsInverted() const      {  return true;  }
    virtual bool                        IsMuxed() const         {  return false;  }
    virtual bool                        MuxDepth() const        {  return InputMuxCount;  }
    virtual int                         BufferSize() const      {  return 65536;  }
    virtual int                         ClockFactor() const     {  return InputClockFactor;  }
    virtual float                       ReferenceClock() const  {  return Input_ReferenceClk;  }
    virtual int               	        SamplesPerFrame() const {  return 32;  }

private:
    UnpairedChannelInfo         FChannels;
    SpanArray                   FSpan;
    X5_G12DeviceMap *           Owner;
};

//===========================================================================
//  CLASS X5_G12InputDevice  -- Register & low level support for X5
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class X5_G12InputDevice
//---------------------------------------------------------------------------

//  Fill in the Info argument for the SPI device constructors
//      -- address-  -direction bit- -ready bit-  -data valid bit-
StdSpiDevice::Info  Adc0Info = { mmAdc0Spi, 16, 22, 23 };
StdSpiDevice::Info  Adc1Info = { mmAdc1Spi, 16, 22, 23 };

const unsigned int X5_G12_Adc_One = 0x8000;

//---------------------------------------------------------------------------
//  constructor for class X5_G12InputDevice
//---------------------------------------------------------------------------

X5_G12InputDevice::X5_G12InputDevice(X5_G12DeviceMap * owner,
                                       X5IoDeviceRegisters & reg)
    : inherited(owner,
                reg,
                new X5_G12InputInfoIntf(owner),
                PmcIdromSection(owner->Rom(),
                                PmcIdrom::waModule,
                                romX5_G12_InputCoeffOffset,
                                romX5_G12_InputCoeffSize),
                X5_G12_Adc_One),	// 'one' value for gain registers
      Owner(owner),
      Adc0(owner->LogicMemory, Adc0Info, "Adc0"),
      Adc1(owner->LogicMemory, Adc1Info, "Adc1"),
      Debug(true),
      AdcTestGenReg(owner->LogicMemory, mmAdcTestGenerator),
      TestGenRateReg(owner->LogicMemory, mmTestGeneratorRate),
      AdcControlReg(owner->LogicMemory, mmAdcControl),
      LsAdcRateReg(owner->LogicMemory, mmLsRateConfig),
      TriggerReg(owner->LogicMemory, reg.TriggerControlRegister_Addr),
      //TriggerDelayReg(owner->LogicMemory, mmExtTriggerIDelay),
      SoftwareTriggerReg(owner->LogicMemory, mmAdcSoftwareTrigger),
      LowSpeedPacketHeaderReg(owner->LogicMemory, mmLowSpeedPacketHeader),
      //DclkSkewReg(owner->LogicMemory, mmAdcDclkSkew),
      FFTControlReg(owner->LogicMemory, mmFFTControl),
      FFTWindowReg(owner->LogicMemory, mmFFTWindow),
      FFTPacketHeaderReg(owner->LogicMemory, mmFFTPacketHeader),
      Adc0PhaseReg(owner->LogicMemory, mmAdc0ClockPhase),
      Adc1PhaseReg(owner->LogicMemory, mmAdc1ClockPhase),
      FFrameSize(0x1000), FFrameMode(false),
      FEdgeTriggered(true), FTestGenRate(0.0f),
      FLowSpeedSampleRate(50000.0f),  FIsBoardDESCapable(false)

{
}

//---------------------------------------------------------------------------
//  destructor for class X5_G12InputDevice
//---------------------------------------------------------------------------

X5_G12InputDevice::~X5_G12InputDevice()
{
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::Framed() --
//---------------------------------------------------------------------------

void X5_G12InputDevice::Framed(int samples)
{
    FFrameSize = samples;
    FFrameMode = true;
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::Unframed() --
//---------------------------------------------------------------------------

void X5_G12InputDevice::Unframed()
{
    FFrameMode = false;
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::ExternalTrigger() --  Set external trigger enable
//---------------------------------------------------------------------------

void X5_G12InputDevice::ExternalTrigger(bool state)
{
    TriggerReg.External.Value(state);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::ExternalTrigger() --  Query ext trigger enable
//---------------------------------------------------------------------------

bool X5_G12InputDevice::ExternalTrigger() const
{
    return TriggerReg.External.Value() ? true : false;
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::EdgeTriggered() --  Set EdgeTrigger mode
//---------------------------------------------------------------------------

void X5_G12InputDevice::EdgeTriggered(bool state)
{
    TriggerReg.EdgeTriggered.Value(state);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::EdgeTriggered() --  Query EdgeTrigger mode
//---------------------------------------------------------------------------

bool X5_G12InputDevice::EdgeTriggered() const
{
    return TriggerReg.EdgeTriggered.Value() ? true : false;
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::HsRun() --  Set HS A/D run state mode
//---------------------------------------------------------------------------

void X5_G12InputDevice::HsRun(bool state)
{
    AdcControlReg.HsRun.Value(state);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::LsRun() -- Set size of received data packets
//---------------------------------------------------------------------------

void X5_G12InputDevice::LsRun(bool state)
{
    AdcControlReg.LsRun.Value(state);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::ConfigureHardware() -- Apply hardware settings after reset
//---------------------------------------------------------------------------

void X5_G12InputDevice::ConfigureHardware()
{
//    if (Owner->Clock.Frequency() < 400.0e6)
//        DclkSkewReg.Skew.Value(10);
//    else
//        DclkSkewReg.Skew.Value(0);

    LowSpeedPacketHeaderReg.Apply();

    LsRun(false);
    HsRun(false);
    //
    //  Run A/D
    if (EnabledLsChannels())
        LsRun(true);
    HsRun(true);

    //
    //  Wait for Logic Ready Bits
    WaitForReady();

    // Apply frame settings
    // 4 samples/ADC clock * 64 samples per trigger pulse in FPGA implementation
    const int SamplesPerFrame = 4 * 64;
    TriggerReg.Count.Value(FFrameSize/SamplesPerFrame);
    TriggerReg.FramedMode.Value(FFrameMode);
    TriggerReg.EdgeTriggered.Value(FEdgeTriggered);

    // Test
    TestFrequency(FTestGenRate);
    AdcTestGenReg.Apply();

    // LS A/D
    LowSpeedSampleRate(FLowSpeedSampleRate);

    // Reapply trigger settings
    TriggerReg.Apply();
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::Spi()
//---------------------------------------------------------------------------

void X5_G12InputDevice::Spi(int adc, int addr, unsigned char data)
{
    if (adc==0)
        Adc0.Spi(addr, data);
    else
        Adc1.Spi(addr, data);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::Spi()
//---------------------------------------------------------------------------

unsigned int X5_G12InputDevice::Spi(int adc, int addr)
{
    if (adc==0)
        return Adc0.Spi(addr) & 0xff;
    else
        return Adc1.Spi(addr) & 0xff;
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::ChannelEnabled() -- Enable specified I/O channel
//---------------------------------------------------------------------------

void X5_G12InputDevice::ChannelEnabled(int ch, bool state)
{
    AdcControlReg.ChannelEnable.Bit(ch, state);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::ChannelEnabled() -- Query state of specified I/O channel
//---------------------------------------------------------------------------

bool X5_G12InputDevice::ChannelEnabled(int ch) const
{
    return AdcControlReg.ChannelEnable.Bit(ch);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::WaitForReady() -- Test PHY Ready registers
//---------------------------------------------------------------------------

void X5_G12InputDevice::WaitForReady()
{
    const int Reps = 10;
    unsigned int EnableBits = AdcControlReg.ChannelEnable.Value() & 3;
    //
    //  We try a few times, but eventually give up
    for (int i=0; i<Reps; i++)
        {
        unsigned int ReadyBits =  AdcControlReg.ChannelPHYReady.Value();
        //  If an enable bit is 0, we don't care about its state. Otherwise, wait
        //    for them all to be 1. Bit 'and' with Enable does this
        if ((ReadyBits & EnableBits)==EnableBits)
            return;

        Innovative::Sleep(1);
        }
    //
    // if we get here we timed out. Log error
    Debug.Log("Warning: PHY Ready did not occur - continuing...");
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::TestEnable()
//---------------------------------------------------------------------------

void X5_G12InputDevice::TestEnable(bool enable)
{
    AdcTestGenReg.TestEnable.Value(enable);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::TestEnable()
//---------------------------------------------------------------------------

bool X5_G12InputDevice::TestEnable() const
{
    return AdcTestGenReg.TestEnable.Value() ? true : false;
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::TestMode()
//---------------------------------------------------------------------------

void X5_G12InputDevice::TestMode(X5_G12InputDevice::TestModes mode)
{
    AdcTestGenReg.TestMode.Value(mode);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::TestMode()
//---------------------------------------------------------------------------

X5_G12InputDevice::TestModes X5_G12InputDevice::TestMode() const
{
    return static_cast<TestModes>(AdcTestGenReg.TestMode.Value());
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::TestMode()
//---------------------------------------------------------------------------

void X5_G12InputDevice::TestFrequency(float value)
{
    FTestGenRate = value;

    double Fs = Owner->Clock.Frequency();
    if (Fs == 0.0f)
        return;
    unsigned int phase = static_cast<unsigned int>(std::pow(2.0, 32.0)/Fs * value);

    TestGenRateReg.Value(phase);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::TestMode()
//---------------------------------------------------------------------------

float X5_G12InputDevice::TestFrequency() const
{
    return FTestGenRate;
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::SoftwareTrigger() -- set software trigger
//---------------------------------------------------------------------------

void X5_G12InputDevice::SoftwareTrigger(bool state)
{
    SoftwareTriggerReg.Trigger.Value(state);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::SoftwareTrigger()  -- read software trigger
//---------------------------------------------------------------------------

bool X5_G12InputDevice::SoftwareTrigger() const
{
    return SoftwareTriggerReg.Trigger.Value() ? true : false ;
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::LowSpeedSampleRate() --
//---------------------------------------------------------------------------

void X5_G12InputDevice::LowSpeedSampleRate(float rate)
{
    FLowSpeedSampleRate = rate;

    double Fs = Owner->Clock.Frequency();

    unsigned int channels = EnabledLsChannels();
    if (!channels)
        return;

    unsigned int divider = static_cast<unsigned int>(Fs/(channels*4 * rate));
    LsAdcRateReg.Divider.Value(divider);

    unsigned int period = static_cast<unsigned int>(std::ceil(3.2e-6 * Fs / 4.0f));
    LsAdcRateReg.HighPeriod.Value(period);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::LowSpeedSampleRate() --
//---------------------------------------------------------------------------

float X5_G12InputDevice::LowSpeedSampleRate() const
{
    return FLowSpeedSampleRate;
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::LowSpeedPacketId() -- Set ID tage of received data packets
//---------------------------------------------------------------------------

void  X5_G12InputDevice::LowSpeedPacketId(unsigned char id)
{
    LowSpeedPacketHeaderReg.Id.Value(id);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::LowSpeedPacketId() -- Query ID tage of received data packets
//---------------------------------------------------------------------------

unsigned char X5_G12InputDevice::LowSpeedPacketId() const
{
    return LowSpeedPacketHeaderReg.Id.Value();
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::LowSpeedPacketSize() -- Set size of received data packets
//---------------------------------------------------------------------------

void  X5_G12InputDevice::LowSpeedPacketSize(unsigned int size_words)
{
   LowSpeedPacketHeaderReg.PacketSize.Value(size_words);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::LowSpeedPacketSize() -- Set size of received data packets
//---------------------------------------------------------------------------

unsigned int  X5_G12InputDevice::LowSpeedPacketSize() const
{
    return LowSpeedPacketHeaderReg.PacketSize.Value();
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::LowSpeedPacketSize() -- Set size of received data packets
//---------------------------------------------------------------------------

unsigned int X5_G12InputDevice::EnabledLsChannels()
{
    unsigned int channels = 0;
    for (int ch = 2; ch < Info().Channels().Channels(); ++ch)
        channels += ChannelEnabled(ch) ? 1 : 0;

    return channels;
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::SetCalibrationOffset()
//---------------------------------------------------------------------------

void X5_G12InputDevice::SetCalibrationOffset(int /*ch*/, int value)
{
    // NOTE: A/D does not appear to respect writes to offset 'bit8' via
    //       register 3.  This bit appears to be always asserted.
    //       Consequently, offset range cannot be adjusted below 0 mV properly

    int offset = static_cast<int>(value);   // need signed offset
    //  Max offset that can be calibrated by the hardware (120uVsteps)
    int offset_cal_range_counts = static_cast<int>(30.72/((float)InputMax_mV/(1<<(InputWidth_Bits-1))));
    //  Trim range to fit hardware limitations
    if (offset>0 && offset>offset_cal_range_counts)
        offset = offset_cal_range_counts;
    else if (offset<0 && offset>offset_cal_range_counts)
        offset = -offset_cal_range_counts;
    //
    //  Convert 'in range' offset to acutal units of 1/4 LSB, unipolar mode
    offset +=  offset_cal_range_counts;
    offset <<= 2;

    // Virtual override of the SetCalibrationOffset method is essential:  The base class behavior disrupts the board initialization and breaks the streaming subsystem.  
    // inherited::SetCalibrationOffset(ch, offset);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::SetCalibrationGain()
//---------------------------------------------------------------------------

void X5_G12InputDevice::SetCalibrationGain(int ch, int value)
{
    const float HdwStep_mV = 0.120f;

    float gain = static_cast<float>(value)/X5_G12_Adc_One;
    unsigned int hdw_steps;
    if (gain<=1.0)
        {
        hdw_steps = 0;   // no gain shift
        }
    else
        {
        float full_range_mV =  2.0*InputMax_mV;
        float adjustment_mV =  full_range_mV*(1.0f-1.0f/gain);
        hdw_steps = static_cast<unsigned int>(adjustment_mV/HdwStep_mV);
        }

    inherited::SetCalibrationGain(ch, hdw_steps);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::DESModeCapable() - return DES Capability state
//---------------------------------------------------------------------------

bool  X5_G12InputDevice::DESModeCapable() const
{
    return FIsBoardDESCapable;
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::DESMode() - read DES Mode State
//---------------------------------------------------------------------------

bool  X5_G12InputDevice::DESMode() const
{
    return AdcControlReg.DesModeEnable.Value();
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::DESMode() - enable or disable DES mode
//---------------------------------------------------------------------------

void  X5_G12InputDevice::DESMode(bool enabled)
{
    AdcControlReg.DesModeEnable.Value(enabled);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::DESPhase() - set DES Phase
//---------------------------------------------------------------------------

void  X5_G12InputDevice::DESPhase(int ch, unsigned int phase)
{
    if (ch!=0)
        Adc1PhaseReg.Value(phase);
    else
        Adc0PhaseReg.Value(phase);

    //  Update ROM Image
    Owner->SetDESPhase(ch, phase);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::DESPhase() - set DES Phase
//---------------------------------------------------------------------------

unsigned int  X5_G12InputDevice::DESPhase(int ch) const
{
    return Owner->GetDESPhase(ch);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::DESPhaseCalibrated() - set DES Phase
//---------------------------------------------------------------------------

void  X5_G12InputDevice::DESPhaseCalibrated(bool flag)
{
    //  Update ROM Image
    Owner->SetDESPhaseCalibrated(flag);
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::DESPhaseCalibrated() - set DES Phase
//---------------------------------------------------------------------------

bool  X5_G12InputDevice::DESPhaseCalibrated() const
{
    return Owner->GetDESPhaseCalibrated();
}

//---------------------------------------------------------------------------
//  X5_G12InputDevice::DESCapabilityCheck() - determine DES ability
//---------------------------------------------------------------------------

void  X5_G12InputDevice::DESCapabilityCheck()
{
    unsigned int variant = Owner->FpgaLogicVariant();
    if (variant == 0x8)
        FIsBoardDESCapable = true;
    else
        FIsBoardDESCapable = false;
}

//==============================================================================
//  CLASS X5_G12Clock -- G12 clock control interface
//==============================================================================
//---------------------------------------------------------------------------
//  housekeeping for class X5_G12Clock
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//  constructor for class X5_G12Clock
//---------------------------------------------------------------------------

X5_G12Clock::X5_G12Clock(X5_G12DeviceMap * owner)
    : Owner(owner), FClockFactor(1.0)
{
}

//---------------------------------------------------------------------------
//  X5_G12Clock::Frequency() -- Change output frequency
//---------------------------------------------------------------------------

void X5_G12Clock::Frequency(double value)
{
    if (Owner->Input().DESModeCapable())
        {
        if (value > DesMinRate)
            {
            FClockFactor = 0.5;
            Owner->Input().DESMode(true);
            }
        else
            {
            FClockFactor = 1.0;
            Owner->Input().DESMode(false);
            }
        }
    else
        FClockFactor = 1.0;

    Owner->Pll.Frequency(value*FClockFactor);
}

//---------------------------------------------------------------------------
//  X5_G12Clock::Frequency() -- Query output frequency
//---------------------------------------------------------------------------

double X5_G12Clock::Frequency() const
{
    return Owner->Pll.Frequency() / FClockFactor;
}

//---------------------------------------------------------------------------
//  X5_G12Clock::FrequencyActual() -- Query actual output frequency
//---------------------------------------------------------------------------

double X5_G12Clock::FrequencyActual() const
{
    return Owner->Pll.FrequencyActual() / FClockFactor;
}

//---------------------------------------------------------------------------
//  X5_G12Clock::Source() -- Retrieve clock source
//---------------------------------------------------------------------------

X5_G12Clock::IIClockSource X5_G12Clock::Source() const
{
    if (Owner->Pll.OutputSource() == X5_G12_Pll::outPll)
        return csInternal;
    else
        return csExternal;
}

//---------------------------------------------------------------------------
//  X5_G12Clock::Source() -- Select clock source
//---------------------------------------------------------------------------

void X5_G12Clock::Source(IIClockSource src)
{
    if (src == csInternal)
        Owner->Pll.OutputSource(X5_G12_Pll::outPll);
    else
        Owner->Pll.OutputSource(X5_G12_Pll::outAux);
}

//---------------------------------------------------------------------------
//  X5_G12Clock::Reference() -- Retrieve reference source
//---------------------------------------------------------------------------

X5_G12Clock::IIReferenceSource X5_G12Clock::Reference() const
{
    if (Owner->Pll.ReferenceSource() == X5_G12_Pll::refPrimary)
        return rsExternal;
    else
        return rsInternal;
}

//---------------------------------------------------------------------------
//  X5_G12Clock::Reference() -- Select reference source
//---------------------------------------------------------------------------

void X5_G12Clock::Reference(IIReferenceSource src)
{
    if (src == rsInternal)
        Owner->Pll.ReferenceSource(X5_G12_Pll::refAutoSelect);
    else
        Owner->Pll.ReferenceSource(X5_G12_Pll::refPrimary);
    //
    //  Turn off Ref VCO if not using it
    Owner->EnableRefVco( src == rsInternal );
}

//---------------------------------------------------------------------------
//  X5_G12Clock::ReferenceFrequency() -- Change reference frequency
//---------------------------------------------------------------------------

void X5_G12Clock::ReferenceFrequency(double value)
{
    Owner->Pll.Reference(value);
}

//---------------------------------------------------------------------------
//  X5_G12Clock::ReferenceFrequency() -- Query reference frequency
//---------------------------------------------------------------------------

double X5_G12Clock::ReferenceFrequency() const
{
    return Owner->Pll.Reference();
}

//---------------------------------------------------------------------------
//  X5_G12Clock::Reapply() -- Reapply all clock settings
//---------------------------------------------------------------------------

void X5_G12Clock::Reapply()
{
    // Route clock to active analog devices
    Source(Source());

    // Select reference clock
    Reference(Reference());
}

//---------------------------------------------------------------------------
//  X5_G12Clock::Locked() -- Return PLL lock status
//---------------------------------------------------------------------------

bool X5_G12Clock::Locked() const
{
   return Owner->Pll.Locked();
}

//---------------------------------------------------------------------------
//  X5_G12Clock::EnableRefVco() -- enable/disable Reference Vco
//---------------------------------------------------------------------------

void  X5_G12Clock::EnableRefVco(bool enabled)
{
    Owner->EnableRefVco(enabled);
}



//===========================================================================
//  CLASS X5_G12DeviceMap  -- Register & low level support for SD module
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class X5_G12DeviceMap
//---------------------------------------------------------------------------

static Cdce72010Registers X5_G12_PllRegisters =
    { mmPllControl,  mmPllSpiIntf  };

static X5IoDeviceRegisters X5G12IoDeviceRegisters =
    { mmI2c, 0,  mmInGainCoef, mmInOffsetCoef, mmInChannelEnable,
      mmInTriggerControl, mmInDecimation, mmInPacketHeader, mmRxPacketControl
    };

//---------------------------------------------------------------------------
//  constructor for class X5_G12DeviceMap
//---------------------------------------------------------------------------

#ifdef _MSC_VER
    #pragma warning(push)
    #pragma warning(disable : 4355)
#endif

X5_G12DeviceMap::X5_G12DeviceMap(X5_G12 * owner)
    : IUsesStreamDeviceEventCallbacks(owner),
      FInput(new X5_G12InputDevice(this, X5G12IoDeviceRegisters)),
      Owner(owner), Debug(true),
      Clock(this),
      Pll(this, LogicMemory, X5_G12_PllRegisters),
      DESParamRomSection(Rom(), PmcIdrom::waModule,
                           romX5_G12_PhaseCoeffOffset, romX5_G12_PhaseCoeffSize),
      VcxoI2CReg(LogicMemory, mmVcxoI2C),
      VcxoIntf(VcxoI2CReg),
      Vcxo_I2C(VcxoIntf, I2CDevice::SubAddress, VcxoI2C_SubAddr),
      Vcxo(Vcxo_I2C, Si570::tcA, kVcxoPowerOnRate),
      RefVcxoIntf(VcxoI2CReg),
      RefVcxo_I2C(RefVcxoIntf, I2CDevice::SubAddress, VcxoI2C_SubAddr),
      RefVcxo(RefVcxo_I2C, Si570::tcC, kRefVcxoPowerOnRate)
{
    //
    //  Hook stream events from base class
    PDMPtr()->OnStreamConfigure.SetEvent(this,   &X5_G12DeviceMap::OnStreamConfigureHandler );
    PDMPtr()->OnBeforeStreamStart.SetEvent(this, &X5_G12DeviceMap::OnBeforeStreamStartHandler );
    PDMPtr()->OnAfterStreamStart.SetEvent(this,   &X5_G12DeviceMap::OnAfterStreamStartHandler );
    PDMPtr()->OnBeforeStreamStop.SetEvent(this,   &X5_G12DeviceMap::OnBeforeStreamStopHandler );
    PDMPtr()->OnAfterStreamStop.SetEvent(this,   &X5_G12DeviceMap::OnAfterStreamStopHandler );

//    DDMPtr()->OnStreamConfigure.SetEvent(this,   &X5_G12DeviceMap::OnStreamConfigureHandler );
//    DDMPtr()->OnBeforeStreamStart.SetEvent(this, &X5_G12DeviceMap::OnBeforeStreamStartHandler );
//    DDMPtr()->OnAfterStreamStart.SetEvent(this,   &X5_G12DeviceMap::OnAfterStreamStartHandler );
//    DDMPtr()->OnBeforeStreamStop.SetEvent(this,   &X5_G12DeviceMap::OnBeforeStreamStopHandler );
//    DDMPtr()->OnAfterStreamStop.SetEvent(this,   &X5_G12DeviceMap::OnAfterStreamStopHandler );
}

#ifdef _MSC_VER
    #pragma warning(pop)
#endif

//---------------------------------------------------------------------------
//  destructor for class X5_G12DeviceMap
//---------------------------------------------------------------------------

X5_G12DeviceMap::~X5_G12DeviceMap()
{
    delete FInput;
    FInput = 0;
}

//---------------------------------------------------------------------------
//  X5_G12DeviceMap::OpenHardware()
//---------------------------------------------------------------------------

void  X5_G12DeviceMap::OpenHardware()
{
    X5DeviceMap::OpenHardware();

    RefVcxo.Init();
    Vcxo.Init();
    Pll.Powered(true);
    Pll.Init();

    // Configure A/Ds
    FInput->PacketSize(0x4000);
    FInput->PacketId(0x0);
    Input().Unframed();
    // Gains and offsets for HS channels only
    for (int ch = 0; ch < MaxHsDevices; ++ch)
        {
        FInput->Gain(ch, 1.0);
        FInput->Offset(ch, 0.0);
        }

    //  Initial Defaults for Low Speed Data
    Input().LowSpeedPacketId(0x08);
    Input().LowSpeedPacketSize(0x800);
    Input().TestFrequency(1.0e6);

    //  Initial Defaults for FFT
    //Input().FFTPacketId(0x10);
    //Input().FFTSize(X5_G12InputDevice::fft4k);
    //Input().FFTScaling(0, 0, 0, 0, 0);

    //  Determine DES Capability
    Input().DESCapabilityCheck();

}

//---------------------------------------------------------------------------
//  X5_G12DeviceMap::OnStreamConfigureHandler()
//---------------------------------------------------------------------------

void  X5_G12DeviceMap::OnStreamConfigureHandler(OpenWire::NotifyEvent & Event)
{
    // Family-common init
    X5DeviceMap::ConfigureHardware();

    // Reapply PLL settings
    EnableRefVco(true);

    RefVcxo.Reapply();
    Vcxo.Reapply();

    Pll.Reapply();
    Clock.Reapply();
    Innovative::Sleep(100);
    if (!Clock.Locked())
        Debug.Log("ERROR: Pll not locked!");

    //
    //  Apply input settings
    //  NOTE: Channel enables MUST PRECEDE input configuration for
    //        proper G12 initialization!
    for (int ch = 0; ch < FInput->Info().Channels().Channels(); ++ch)
        FInput->ChannelEnabled(ch, FInput->Info().Channels().Enabled(ch));

    FInput->PacketId(FInput->PacketId());
    FInput->PacketSize(FInput->PacketSize());

    // Reapply Hardware Settings for Input Device
    FInput->ConfigureHardware();
    //
    //  Apply Gain/Offset Coefficients
    if (FInput->Calibrated())
        FInput->ApplyRomCoefficients();
    else
        FInput->ApplyDefaultCoefficients();

    //  Reapply DES Phase 
    if (FInput->DESModeCapable())
        {
        FInput->DESPhase(0, FInput->DESPhase(0));
        FInput->DESPhase(1, FInput->DESPhase(1));
        }

    //  Forward notify to baseboard handler
    OwnerCB()->CallOnStreamConfigure(Event);
}

//---------------------------------------------------------------------------
//  X5_G12DeviceMap::OnBeforeStreamStartHandler()
//---------------------------------------------------------------------------

void  X5_G12DeviceMap::OnBeforeStreamStartHandler(OpenWire::NotifyEvent & Event)
{
    //  Forward notify to baseboard handler
    OwnerCB()->CallOnBeforeStreamStart(Event);
    //
    //  Start execution
    UserFpgaControlReg.InGo.Set();
    UserFpgaControlReg.Apply();
}

//---------------------------------------------------------------------------
//  X5_G12DeviceMap::OnAfterStreamStartHandler()
//---------------------------------------------------------------------------

void  X5_G12DeviceMap::OnAfterStreamStartHandler(OpenWire::NotifyEvent & Event)
{
    //  Forward notify to baseboard handler
    OwnerCB()->CallOnAfterStreamStart(Event);
}

//---------------------------------------------------------------------------
//  X5_G12DeviceMap::OnBeforeStreamStopHandler()
//---------------------------------------------------------------------------

void  X5_G12DeviceMap::OnBeforeStreamStopHandler(OpenWire::NotifyEvent & Event)
{
    //  Forward notify to baseboard handler
    OwnerCB()->CallOnBeforeStreamStop(Event);
}

//---------------------------------------------------------------------------
//  X5_G12DeviceMap::OnAfterStreamStopHandler()
//---------------------------------------------------------------------------

void  X5_G12DeviceMap::OnAfterStreamStopHandler(OpenWire::NotifyEvent & Event)
{
    //  Disable User Logic Streaming
    UserFpgaControlReg.InGo.Clear();
    UserFpgaControlReg.Apply();

    Input().SoftwareTrigger(false);
    Input().HsRun(false);
    Input().LsRun(false);

    //  Forward notify to baseboard handler
    OwnerCB()->CallOnAfterStreamStop(Event);
}

//---------------------------------------------------------------------------
//  X5_G12DeviceMap::Input() -- Reference to input object
//---------------------------------------------------------------------------

X5_G12InputDevice & X5_G12DeviceMap::Input()
{
    return *FInput;
}

//---------------------------------------------------------------------------
//  X5_G12DeviceMap::EnableRefVco() -- enable/disable Ref Vco
//---------------------------------------------------------------------------

void  X5_G12DeviceMap::EnableRefVco(bool enabled)
{
    VcxoI2CReg.RefVco_En.Value(enabled);
}

//---------------------------------------------------------------------------
//  X5_G12DeviceMap::SetDESPhase() -- Write DES Phase to ROM Map
//---------------------------------------------------------------------------

void  X5_G12DeviceMap::SetDESPhase(int ch, unsigned int phase)
{
    if (ch<0 || ch>=2)
        return;

    DESParamRomSection.AsInt(ch, phase);
}

//---------------------------------------------------------------------------
//  X5_G12DeviceMap::GetDESPhase() -- Read DES Phase from ROM Map
//---------------------------------------------------------------------------

unsigned int  X5_G12DeviceMap::GetDESPhase(int ch) const
{
    if (ch<0 || ch>=2)
        return 0;

    if (GetDESPhaseCalibrated())
        {
        return static_cast<unsigned int>(DESParamRomSection.AsInt(ch));
        }
    else
        return 0;  // default value
}

//---------------------------------------------------------------------------
//  X5_G12DeviceMap::SetDESPhaseCalibrated() -- Write DES Cal Bit
//---------------------------------------------------------------------------

void  X5_G12DeviceMap::SetDESPhaseCalibrated(bool state)
{
    if (state)
        DESParamRomSection.AsInt(romX5_G12_Phase_CalibratedIdx, 'Y');
    else
        DESParamRomSection.AsInt(romX5_G12_Phase_CalibratedIdx, 0);
}

//---------------------------------------------------------------------------
//  X5_G12DeviceMap::GetDESPhaseCalibrated() -- Read Phase cal bit
//---------------------------------------------------------------------------

bool  X5_G12DeviceMap::GetDESPhaseCalibrated() const
{
    if (DESParamRomSection.AsInt(romX5_G12_Phase_CalibratedIdx)=='Y')
        return true;
    else
        return false;
}

} // namespace Innovative


