// X5_GSPSDeviceMap_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2005 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <X5_GSPS_Mb.h>
#include <X5_GSPSDeviceMap_Mb.h>
#include <X5MemoryMap_Mb.h>
#include <X5_GSPSMemoryMap_Mb.h>
#include <SystemSupport_Mb.h>
#include <Magic_Mb.h>
#include <sstream>

//---------------------------------------------------------------------------
namespace Innovative
{

using namespace X5_MemoryMap;
using namespace X5_GSPSMemoryMap;

//---------------------------------------------------------
//  Module Constants Definition, Clock is either external or internal fixed crystal.

const int    MaxInDevices      = 2;
const int    InputWidth_Bits   = 8;
const int    InputMuxCount     = 0;
const bool   InputIsUnipolar   = false;
const int    InputClockFactor  = 1;
const float  InputMaxRate      = 3000.0e6f;
const float  DesMinRate        = InputMaxRate / 2.0f;
const float  OedMinRate        = 680.0e6f;
const float  InputMinRate      = 200.0e6f;
const int    InputGangSize     = 2;
const float  Input_ReferenceClk = 100.0e6;
const float  Input_VcoMinClk    = 2550.0e6f;
const float  Input_VcoMaxClk    = 2995.0e6f;

//---------------------------------------------------------
//  IDROM Section Map
//  ...First section: Input and Output Coefficient section
//       2channels x 2 values + 1 calibration flag + 1 spare
const int romX5_GSPS_InputCoeffOffset   = 0;
const int romX5_GSPS_InChannels         = MaxInDevices;
const int romX5_GSPS_InputCoeffSize     = romX5_GSPS_InChannels * 2 + 1 + 1;
const int romX5_GSPS_CoeffSize          = romX5_GSPS_InputCoeffSize;


//===========================================================================
// class X5_GSPSInputInfoIntf
//===========================================================================

class X5_GSPSInputInfoIntf : public IoInfoIntf
{
public:
    X5_GSPSInputInfoIntf()
        : //    devices,           bits,    muxChannelCount,   unipolar
          FChannels(MaxInDevices, InputWidth_Bits, InputMuxCount, InputIsUnipolar)
        {  FSpan.push_back(SpanInfo(1000, -1000));  }
    virtual ~X5_GSPSInputInfoIntf()  {  }

    virtual UnpairedChannelInfo &       Channels()              {  return FChannels;}
    virtual float                       MaxRate() const         {  return InputMaxRate;  }
    virtual float                       MinRate() const         {  return InputMinRate;  }
    virtual const SpanArray &           Span() const            {  return FSpan;  }
    virtual int                         GangSize() const        {  return InputGangSize;  }
    virtual bool                        IsStackable() const     {  return true;  }
	virtual bool                        IsInverted() const      {  return true;  }
    virtual bool                        IsMuxed() const         {  return false;  }
    virtual bool                        MuxDepth() const        {  return InputMuxCount;  }
    virtual int                         BufferSize() const      {  return 65536;  }
    virtual int                         ClockFactor() const     {  return InputClockFactor;  }
    virtual float                       ReferenceClock() const  {  return Input_ReferenceClk;  }
	virtual int               	        SamplesPerFrame() const {  return 32;  }

private:
    UnpairedChannelInfo         FChannels;
    SpanArray                   FSpan;
};

class ADC08D1500;
//=============================================================================
//  CLASS X5_AdcShadowRegister  --
//=============================================================================

class X5_AdcShadowRegister : public MemoryBitManipulator
{
public:
    X5_AdcShadowRegister(ADC08D1500 * owner, int mux, int value)
        : MemoryBitManipulator(value), Owner(owner), Mux(mux)
    {
    }

protected:
    // Data
    ADC08D1500 *    Owner;
    int             Mux;

    unsigned int ReadV() const
        {  return FValue;  }
    void WriteV(unsigned int value);
};

//===========================================================================
//  CLASS ADC08D1500  -- Register & low level support for GSPS A/D
//===========================================================================
class ADC08D1500
{
    friend class X5_AdcShadowRegister;

public:
    ADC08D1500(X5_GSPSInputDevice * owner, AddressingSpace &space, int spi)
        : Owner(owner), 
          SpiControlReg(space, spi),
          SpiStatusReg(space, spi),
          ConfigReg(this, 0x01, 0xb2ff),
          IOffsetReg(this, 0x02, 0x7f),
          IGainReg(this, 0x03, 0x807f),
          ExtendedConfigReg(this, 0x09, 0x03ff),
          QOffsetReg(this, 0x0a, 0x7f),
          QGainReg(this, 0x0b, 0x807f),
          DESEnableReg(this, 0x0d, 0x3fff),
          DESCourseAdjustReg(this, 0x0e, 0x7ff),
          DESFineAdjustReg(this, 0x0f, 0x7f)
        {  }

    // Properties

    // Methods
	// Gain/Offset
    void		SetCalibrationOffset(int ch, int value)
		{
        // Each count in the offset reg is .176 mV at 700 mV full scale
        // But, GSPS input range is ~2.127 Vp-p instead of 700 mVp-p
        const float correction = (2.127f/.7000f)/.176f;
        value = static_cast<int>(value * correction);

        // Clip to usable range
        value = std::max(std::min(value, 255), -255);
        int magnitude = abs(value) << 8;
        int sign = (value < 0) ? 0x80 : 0x00;
        int code = (sign | magnitude) | 0x7f;

        if (ch)
            QOffsetReg.Value(code);
        else
            IOffsetReg.Value(code);
        }
    void 		SetCalibrationGain(int ch, int value)
        {
        // Gain reg can adjust only 20% (not 100%) of full-scale.
        const float correction = -1.0f/0.2f;
        value = static_cast<int>(value * correction);

        int code = value | 0x7f;

        if (ch)
            QGainReg.Value(code);
        else
            IGainReg.Value(code);
        }

    // Dual edge sampling
    void        DESEnable(bool state)
        {
        if (!Owner->IsADC08D1500())
            ExtendedConfigReg.Bit(13, state);
        else
            DESEnableReg.Field(14, 2, state ? 3 : 0);
        }
    bool        DESEnable() const
        {
        if (!Owner->IsADC08D1500())
            return ExtendedConfigReg.Bit(13) ? true : false;
        else
            return DESEnableReg.Field(14, 2) == 3;
		}

    // Double data rate
    void        DdrEnable(bool state)
        {  ConfigReg.Bit(10, state ? 0 : 1);  }
    bool        DdrEnable() const
        {  return ConfigReg.Bit(10) ? false : true;  }

    // Demultiplexor control
    void        DemuxDisable(bool state)
        {  ConfigReg.Bit(8, state ? 1 : 0);  }
    bool        DemuxDisable() const
        {  return ConfigReg.Bit(8) ? true : false;  }

    // Test pattern
    void        TestEnable(bool state)
        {
        if (!Owner->IsADC08D1500())
            ExtendedConfigReg.Bit(15, state ? 1 : 0);
        }
    bool        TestEnable() const
        {
        if (!Owner->IsADC08D1500())
            return ExtendedConfigReg.Bit(15) ? true : false;
        else
            return false;
        }

    // Improved AC performance below 900 MHz
    void        LowFrequencyEnable(bool state)
        {
        if (!Owner->IsADC08D1500())
            ExtendedConfigReg.Bit(10, state ? 1 : 0);
        }
    bool        LowFrequencyEnable() const
        {
        if (!Owner->IsADC08D1500())
            return ExtendedConfigReg.Bit(10) ? true : false;
        else
            return false;
        }

    // Input channel selection
    void        QInputSelect(bool state)
        {
		if (!Owner->IsADC08D1500())
            ExtendedConfigReg.Bit(12, state ? 1 : 0);
        }
    bool        QInputSelect() const
        {
        if (!Owner->IsADC08D1500())
            return ExtendedConfigReg.Bit(12) ? true : false;
        else
            return false;
        }

protected:
    // Register definitions
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//  CLASS SpiControlRegister  -- A/D SPI interface
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	class SpiControlRegister : public CachedShadowRegister
	{
		typedef CachedShadowRegister inherited;

	public:
		RegisterBitGroup      Data;
		RegisterBitGroup      Address;

		SpiControlRegister(AddressingSpace &space, int offset)
			:  inherited("ADC08D1500::AdcSpiRegister", space, offset),
			   Data(*this, 0, 16), Address(*this, 16, 4)
			{
			//  Default setup ????
			}
	};

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//  CLASS SpiStatusRegister  -- PLL SPI interface
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	class SpiStatusRegister : public Register
	{
		typedef Register inherited;

	public:
		RegisterBit           Ready;

		SpiStatusRegister(AddressingSpace &space, int offset)
			: inherited("ADC08D1500::SpiStatusRegister", space, offset),
              Ready(*this, 31)
			{
			//  Default setup ????
			}
    };

    // Data
    X5_GSPSInputDevice *    Owner;
    Trace                   Debug;
    bool                    FDESEnable;

    // Registers
    SpiControlRegister      SpiControlReg;
    SpiStatusRegister       SpiStatusReg;

    // Shadows
    X5_AdcShadowRegister    ConfigReg;
    X5_AdcShadowRegister    IOffsetReg;
    X5_AdcShadowRegister    IGainReg;
    X5_AdcShadowRegister    ExtendedConfigReg;
    X5_AdcShadowRegister    QOffsetReg;
    X5_AdcShadowRegister    QGainReg;
    X5_AdcShadowRegister    DESEnableReg;
    X5_AdcShadowRegister    DESCourseAdjustReg;
    X5_AdcShadowRegister    DESFineAdjustReg;

public:
    //  Raw configuration accesses
	void 			Port(unsigned short reg, unsigned short value);
};

//-----------------------------------------------------------------------------
//  ADC08D1500::Port()  --  Write to PLL data port
//-----------------------------------------------------------------------------

void ADC08D1500::Port(unsigned short reg, unsigned short value)
{
    // Wait for previous operation to complete
    for (unsigned int i = 0; (i < 10) && !SpiStatusReg.Ready.Value(); ++i)
        Sleep(1);
    if (!SpiStatusReg.Ready.Value())
        Debug.Log("Adc SPI ready timeout error!");

    SpiControlReg.Address.Value(reg);
    SpiControlReg.Data.Value(value);
    SpiControlReg.Apply();
}

//-----------------------------------------------------------------------------
//  X5_AdcShadowRegisterWriteV::WriteV()  --  Write to data port
//-----------------------------------------------------------------------------

void X5_AdcShadowRegister::WriteV(unsigned int value)
    {
    FValue = value;
    Owner->Port(Mux, value);
    }

//===========================================================================
//  CLASS X5_GSPSInputDevice  -- Register & low level support for X5
//===========================================================================

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice Ctor
//---------------------------------------------------------------------------

X5_GSPSInputDevice::X5_GSPSInputDevice(X5_GSPSDeviceMap * owner,
									   X5IoDeviceRegisters & reg)
	: inherited(owner,
				reg,
				new X5_GSPSInputInfoIntf,
				PmcIdromSection(owner->Rom(),
								PmcIdrom::waModule,
								romX5_GSPS_InputCoeffOffset,
								romX5_GSPS_InputCoeffSize),
				0x8000),	// 'one' value for gain registers
      Owner(owner), Debug(true), 
      AdcControlReg(owner->LogicMemory, mmInControl),
      AdcStatusReg(owner->LogicMemory, mmInControl),
      TriggerReg(owner->LogicMemory, reg.TriggerControlRegister_Addr),
	  AdcClockPhaseReg(owner->LogicMemory, mmAdcClockPhase),
      AdcConfigReg(new ADC08D1500(this, owner->LogicMemory, mmInSpi)),
      FDecimation(1), FFrameSize(0x1000), FFrameMode(false),
      FAdcClockPhase(0x3d), FPatternEnable(false)
{
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice Ctor
//---------------------------------------------------------------------------

X5_GSPSInputDevice::~X5_GSPSInputDevice()
{
    delete AdcConfigReg;
    AdcConfigReg = 0;
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::Framed() --
//---------------------------------------------------------------------------

void X5_GSPSInputDevice::Framed(int samples)
{
    FFrameSize = samples;
    FFrameMode = true;
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::Unframed() --
//---------------------------------------------------------------------------

void X5_GSPSInputDevice::Unframed()
{
    FFrameMode = false;
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::ExternalTrigger() --  Set external trigger enable
//---------------------------------------------------------------------------

void X5_GSPSInputDevice::ExternalTrigger(bool state)
{
    TriggerReg.External.Value(state);
    AdcControlReg.SyncPipelineReset.Value(0);
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::ExternalTrigger() --  Query ext trigger enable
//---------------------------------------------------------------------------

bool X5_GSPSInputDevice::ExternalTrigger() const
{
    return TriggerReg.External.Value() ? true : false;
}


//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::DualEdgeSampling() --
//---------------------------------------------------------------------------

bool X5_GSPSInputDevice::DualEdgeSampling() const
{
	return AdcConfigReg->DESEnable();
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::DualEdgeSampling() --
//---------------------------------------------------------------------------

void X5_GSPSInputDevice::DualEdgeSampling(bool value)
{
    AdcConfigReg->DESEnable(value);
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::Decimation() --
//---------------------------------------------------------------------------

void X5_GSPSInputDevice::Decimation(unsigned int factor)
{
    // 1 >= Factor <= 8.  Even values only   
    factor &= ~1;

    factor = std::min<unsigned int>(factor, 8u);
    factor = std::max<unsigned int>(factor, 1u);

	FDecimation = factor;
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::Decimation() --
//---------------------------------------------------------------------------

unsigned int X5_GSPSInputDevice::Decimation() const
{
    return FDecimation;
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::IsADC08D1500() -- True of original IsADC08D1500 present
//---------------------------------------------------------------------------

bool X5_GSPSInputDevice::IsADC08D1500() const
{
    return ((Owner->FpgaLogicVariant() & 1) == 0) && (Owner->FpgaLogicRevision() == 0);
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::ConfigureHardware() -- Apply hardware settings after reset
//---------------------------------------------------------------------------

void X5_GSPSInputDevice::ConfigureHardware()
{
//    AdcInterfaceEnable

    // dclk phase offset
    AdcClockPhaseReg.Value(FAdcClockPhase);
    Sleep(1);

    // Re-apply config.  DDR must be enabled.
    AdcConfigReg->DdrEnable(true);

	// Disable demux if sample rate < 680 MHz
	bool LowSpeedMode = Owner->Clock.Frequency() < OedMinRate;
	if (!IsADC08D1500())
        AdcConfigReg->DemuxDisable(LowSpeedMode);

    // Improve Re-apply config.  DDR must be enabled.
    bool lfe = Owner->Clock.Frequency() < 900.e6;
	AdcConfigReg->LowFrequencyEnable(lfe);

    AdcControlReg.Reset.Set();
    AdcControlReg.Reset.Clear();
    Sleep(1);
    if (!AdcStatusReg.PhyDcmLocked.Value())
        Debug.Log("Error: DCM lock failure!");
	if (!AdcStatusReg.ClockCalibrationDone.Value())
        Debug.Log("Error: Clock calibration incomplete!");
    if (AdcStatusReg.ClockCalibrationFailed.Value())
        Debug.Log("Error: Clock calibration failure!");

    // DES must be disabled during calibration
    DualEdgeSampling(false);

    // Configure pattern enable
    AdcConfigReg->TestEnable(FPatternEnable);

    // Calibrate the A/D
    AdcControlReg.Calibrate.Value(1);
    Sleep(5);
    if (!AdcStatusReg.ClockCalibrationDone.Value())
        Debug.Log("Error: Clock calibration incomplete!");
    AdcControlReg.Calibrate.Value(0);

    // Reset external sync
    AdcControlReg.SyncPipelineReset.Value(1);
    AdcControlReg.DclkReset.Value(1);
    AdcControlReg.DclkReset.Value(0);

    AdcControlReg.Reset.Set();
    AdcControlReg.Reset.Clear();
    Sleep(1);
    if (!AdcStatusReg.PhyDcmLocked.Value())
        Debug.Log("Error: DCM lock failure!");
    if (!AdcStatusReg.ClockCalibrationDone.Value())
        Debug.Log("Error: Clock calibration incomplete!");
    if (AdcStatusReg.ClockCalibrationFailed.Value())
        Debug.Log("Error: Clock calibration failure!");

    // Re-apply DES mode settings
    bool des = Owner->Clock.Frequency() > DesMinRate;
    DualEdgeSampling(des);

    AdcConfigReg->QInputSelect(des && ChannelEnabled(1));

    // Apply decimation setting
    AdcControlReg.Decimation.Value(FloorLog2(FDecimation));

    // Apply frame settings
	int SamplesPerFrame = DualEdgeSampling() ? 16 : 8;
    int SpeedFactor = LowSpeedMode ? 2 : 1;
    TriggerReg.Count.Value((FFrameSize*FDecimation*SpeedFactor)/SamplesPerFrame);
    TriggerReg.FramedMode.Value(FFrameMode);

    // Reapply trigger settings
    TriggerReg.Apply();
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::SetCalibrationOffset()
//---------------------------------------------------------------------------

void X5_GSPSInputDevice::SetCalibrationOffset(int ch, int value)
{
    AdcConfigReg->SetCalibrationOffset(ch, value);
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::SetCalibrationGain()
//---------------------------------------------------------------------------

void X5_GSPSInputDevice::SetCalibrationGain(int ch, int value)
{
    AdcConfigReg->SetCalibrationGain(ch, value);
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::ChannelEnabled() -- Enable specified I/O channel
//---------------------------------------------------------------------------

void X5_GSPSInputDevice::ChannelEnabled(int ch, bool state)
{
    AdcControlReg.ChannelEnable.Bit(ch, state);
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::ChannelEnabled() -- Query state of specified I/O channel
//---------------------------------------------------------------------------

bool X5_GSPSInputDevice::ChannelEnabled(int ch) const
{
    return AdcControlReg.ChannelEnable.Bit(ch);
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::TestEnable()
//---------------------------------------------------------------------------

void X5_GSPSInputDevice::TestEnable(bool enable)
{
    AdcControlReg.TestMode.Value(enable);
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::TestEnable()
//---------------------------------------------------------------------------

bool X5_GSPSInputDevice::TestEnable() const
{
	return AdcControlReg.TestMode.Value() ? true : false;
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::PatternEnable()
//---------------------------------------------------------------------------

void X5_GSPSInputDevice::PatternEnable(bool state)
{
    FPatternEnable = state;
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::PatternEnable()
//---------------------------------------------------------------------------

bool X5_GSPSInputDevice::PatternEnable() const
{
    return FPatternEnable;
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::AdcClockPhase()
//---------------------------------------------------------------------------

void X5_GSPSInputDevice::AdcClockPhase(unsigned int value)
{
    FAdcClockPhase = value;
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::AdcClockPhase()
//---------------------------------------------------------------------------

unsigned int X5_GSPSInputDevice::AdcClockPhase() const
{
    return FAdcClockPhase;
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::AcknowledgeAlert()
//---------------------------------------------------------------------------

void X5_GSPSInputDevice::AcknowledgeAlert()
{
    AdcControlReg.AcknowledgeAlert.Toggle();
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::Spi()
//---------------------------------------------------------------------------

unsigned short X5_GSPSInputDevice::Spi(int /*addr*/)
{
	return 0;        // not readable
}

//---------------------------------------------------------------------------
//  X5_GSPSInputDevice::Spi()
//---------------------------------------------------------------------------

void X5_GSPSInputDevice::Spi(int addr, unsigned short value)
{
	AdcConfigReg->Port(addr, value);
}

//==============================================================================
//  CLASS X5_GSPS_PRI_Intf -- Interface X5 Pulse repetition interval functionality
//==============================================================================

class X5_GSPS_PRI_Intf : public IPulseRepetitionIntervalIntf
{
public:
	// Ctor
	X5_GSPS_PRI_Intf(X5_GSPSDeviceMap * owner, AddressingSpace & space,
		const int period, const int width, const int ctl, const int rst)
		: Owner(owner), FEnabled(false), PeriodReg(space, period),
		  WidthReg(space, width), ControlReg(space, ctl), ResetReg(space, rst)
		  {  }

	// Properties
	void    		    Enabled(bool state)
		{  FEnabled = state;  }
	bool	    		Enabled() const
		{  return FEnabled;  }

	void    		    Run(bool state)
		{  ControlReg.Run.Value(state ? 0 : 1);  }
	bool	    		Run() const
		{  return ControlReg.Run.Value() ? false : true;  }

   	virtual void    		    FiniteFrames(bool /*state*/)  {  }
	virtual bool	    		FiniteFrames() const {  return false;  }

  	virtual void    		    FiniteFrameRearm(bool /*state*/)  {  }
	virtual bool	    		FiniteFrameRearm() const {  return false;  }

   	virtual void    		    FiniteFrameCount(unsigned short /*count*/)  {  }
	virtual unsigned short 		FiniteFrameCount() const {  return 0;  }

	void    		    AddEvent(unsigned int period,
								 unsigned int delay,
								 unsigned int width)
		{  CaptureEvents.push_back(CaptureEvent(period, delay, width));  }

	// Methods
	void 				Reset()
		{
		Run(false);
		Enable(false);

		CaptureEvents.clear();
		}
	void 				Apply()
		{
		Enable(FEnabled);
		if (!FEnabled)
			return;

		// Register divisor is a function of sample rate, since data format
		// from A/D changes with data rate
		int divisor = 4;
		if (Owner->Clock.Frequency() >= OedMinRate)
			divisor *= 2;
		if (Owner->Clock.Frequency() > DesMinRate)
			divisor *= 2;

		ResetReg.Access.Value(1);
		for (CaptureEventArray::iterator i = CaptureEvents.begin(); i != CaptureEvents.end(); ++i)
			{
			// Period register must be loaded with N-1
			PeriodReg.Value((i->Period/divisor) - 1);
			ControlReg.Delay.Value(i->Delay/divisor);
			WidthReg.Value(i->Width/divisor);
			}
		}

protected:
	class CaptureEvent
	{
	public:
		CaptureEvent(unsigned int period,
					 unsigned int delay,
					 unsigned int width)
			: Period(period), Delay(delay), Width(width)
			{  }
		// Data
		unsigned int Period;
		unsigned int Delay;
		unsigned int Width;
	};
	// Data
	X5_GSPSDeviceMap * Owner;

	typedef std::vector<CaptureEvent> CaptureEventArray;
	CaptureEventArray	CaptureEvents;

	// Fields
	bool		FEnabled;

	// Register definitions
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//  CLASS PeriodRegister  -- Pulse period control
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	class PeriodRegister : public ShadowRegister
	{
		typedef ShadowRegister inherited;

	public:
		RegisterBitGroup      Count;

		PeriodRegister(AddressingSpace &space, int offset)
			:  inherited("X5_GSPSDeviceMap::PeriodRegister", space, offset),
			   Count(*this, 0, 32)
			{  }
	};

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//  CLASS WidthRegister  -- Pulse capture width control
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	class WidthRegister : public ShadowRegister
	{
		typedef ShadowRegister inherited;

	public:
		RegisterBitGroup      Count;

		WidthRegister(AddressingSpace &space, int offset)
			:  inherited("X5_GSPSDeviceMap::WidthRegister", space, offset),
			   Count(*this, 0, 24)
			{  }
	};

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//  CLASS ControlRegister  -- Pulse repetition duration control
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	class ControlRegister : public ShadowRegister
	{
		typedef ShadowRegister inherited;

	public:
		RegisterBitGroup      	Delay;
		RegisterBit				Enable;
		RegisterBit				Run;

		ControlRegister(AddressingSpace &space, int offset)
			:  inherited("X5_GSPSDeviceMap::ControlRegister", space, offset),
			   Delay(*this, 0, 24), Enable(*this, 24), Run(*this, 28)
			{  }
	};

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//  CLASS ResetRegister  -- Pulse repetition duration control
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	class ResetRegister : public ShadowRegister
	{
		typedef ShadowRegister inherited;

	public:
		RegisterBit				Access;

		ResetRegister(AddressingSpace &space, int offset)
			:  inherited("X5_GSPSDeviceMap::ResetRegister", space, offset),
			   Access(*this, 0)
			{  }
	};

	// Registers
	PeriodRegister		PeriodReg;
	WidthRegister 		WidthReg;
	ControlRegister		ControlReg;
	ResetRegister		ResetReg;

	// Support methods
	void    		    Enable(bool state)
		{  ControlReg.Enable.Value(state ? 1 : 0);  }
	bool	    		Enable() const
		{  return ControlReg.Enable.Value() ? true : false;  }
};

//==============================================================================
//  CLASS X5_GSPSClock -- GSPS clock control
//==============================================================================

//---------------------------------------------------------------------------
//  Constructor
//---------------------------------------------------------------------------

X5_GSPSClock::X5_GSPSClock(X5_GSPSDeviceMap * owner, AddressingSpace &space,
	int ctl, int spi)
	: Owner(owner), Clock(0, space, ctl, spi)
{
	Clock.Reference(Input_ReferenceClk);
	Clock.Min(Input_VcoMinClk/Clock.MaxOutputDivider());
	Clock.Max(Input_VcoMaxClk);
}

//---------------------------------------------------------------------------
//  X5_GSPSClock::Frequency() -- Change output frequency
//---------------------------------------------------------------------------

void X5_GSPSClock::Frequency(double value)
{
	Clock.ClockFactor((value > DesMinRate) ? 0.5 : 1.0);
	Clock.Frequency(value);
}

//---------------------------------------------------------------------------
//  X5_GSPSClock::Frequency() -- Query output frequency
//---------------------------------------------------------------------------

double X5_GSPSClock::Frequency() const
{
	return Clock.Frequency();
}

//---------------------------------------------------------------------------
//  X5_GSPSClock::FrequencyActual() -- Query actual output frequency
//---------------------------------------------------------------------------

double X5_GSPSClock::FrequencyActual() const
{
    return Clock.FrequencyActual();
}

//---------------------------------------------------------------------------
//  X5_GSPSClock::Source() -- Retrieve clock source
//---------------------------------------------------------------------------

X5_GSPSClock::IIClockSource X5_GSPSClock::Source() const
{
    if (Clock.OutputClock() == Ad9516::oVco)
        return csInternal;
    else
		return csExternal;
}

//---------------------------------------------------------------------------
//  X5_GSPSClock::Source() -- Select clock source
//---------------------------------------------------------------------------

void X5_GSPSClock::Source(IIClockSource src)
{
    if (src == csInternal)
        Clock.OutputClock(Ad9516::oVco);
    else
        Clock.OutputClock(Ad9516::oExternal);
}

//---------------------------------------------------------------------------
//  X5_GSPSClock::Reference() -- Retrieve reference source
//---------------------------------------------------------------------------

X5_GSPSClock::IIReferenceSource X5_GSPSClock::Reference() const
{
    if (Clock.ReferenceClock() == Ad9516::rInternal)
        return rsInternal;
    else
		return rsExternal;
}

//---------------------------------------------------------------------------
//  X5_GSPSClock::Reference() -- Select reference source
//---------------------------------------------------------------------------

void X5_GSPSClock::Reference(IIReferenceSource src)
{
    if (src == rsInternal)
		Clock.ReferenceClock(Ad9516::rInternal);
    else
        Clock.ReferenceClock(Ad9516::rExternal);
}

//---------------------------------------------------------------------------
//  X5_GSPSClock::ReferenceFrequency() -- Change reference frequency
//---------------------------------------------------------------------------

void X5_GSPSClock::ReferenceFrequency(double value)
{
    Clock.Reference(value);
}

//---------------------------------------------------------------------------
//  X5_GSPSClock::ReferenceFrequency() -- Query reference frequency
//---------------------------------------------------------------------------

double X5_GSPSClock::ReferenceFrequency() const
{
    return Clock.Reference();
}

//---------------------------------------------------------------------------
//  X5_GSPSClock::Reapply() -- Reapply all clock settings
//---------------------------------------------------------------------------

void X5_GSPSClock::Reapply()
{
	// Route clock to active analog devices
    Source(Source());

	// Select reference clock
    Reference(Reference());

	// Set reference based on clock source to obtain correct FrequencyActual
	if (Source() == csExternal)
    	Clock.Reference(ReferenceFrequency());
	else
		Clock.Reference(Owner->Input().Info().ReferenceClock());

    Clock.Reapply();
}

//---------------------------------------------------------------------------
//  X5_GSPSClock::Locked() -- Return PLL lock status
//---------------------------------------------------------------------------

bool X5_GSPSClock::Locked() const
{
    return Clock.Locked();
}

//===========================================================================

static X5IoDeviceRegisters X5GspsIoDeviceRegisters =
	{ mmI2c, 0,  mmInGainCoef, mmInOffsetCoef, mmInChannelEnable,
	  mmInTriggerControl, mmInDecimation, mmInPacketHeader, mmRxPacketControl
	};

//===========================================================================
//  CLASS X5_GSPSDeviceMap  -- Register & low level support for SD module
//===========================================================================

//---------------------------------------------------------------------------
//  constructor
//---------------------------------------------------------------------------

#ifdef _MSC_VER
    #pragma warning(push)
	#pragma warning(disable : 4355)
#endif

X5_GSPSDeviceMap::X5_GSPSDeviceMap(X5_GSPS * owner)
	: IUsesStreamDeviceEventCallbacks(owner),
	  FInput (new X5_GSPSInputDevice (this, X5GspsIoDeviceRegisters)),
	  FPulse(new X5_GSPS_PRI_Intf(this, LogicMemory,
		mmPulsePeriod, mmPulseWidth, mmPulseControl, mmPulseReset)),
	  Owner(owner), Debug(true),
	  Clock(this, LogicMemory, mmPllCtl, mmPllSpi)
{
	//
	//  Hook stream events from base class
	PDMPtr()->OnStreamConfigure.SetEvent(this,   &X5_GSPSDeviceMap::OnStreamConfigureHandler );
	PDMPtr()->OnBeforeStreamStart.SetEvent(this, &X5_GSPSDeviceMap::OnBeforeStreamStartHandler );
	PDMPtr()->OnAfterStreamStart.SetEvent(this,   &X5_GSPSDeviceMap::OnAfterStreamStartHandler );
	PDMPtr()->OnBeforeStreamStop.SetEvent(this,   &X5_GSPSDeviceMap::OnBeforeStreamStopHandler );
	PDMPtr()->OnAfterStreamStop.SetEvent(this,   &X5_GSPSDeviceMap::OnAfterStreamStopHandler );
}

#ifdef _MSC_VER
	#pragma warning(pop)
#endif

//---------------------------------------------------------------------------
//  destructor
//---------------------------------------------------------------------------

X5_GSPSDeviceMap::~X5_GSPSDeviceMap()
{
	delete FInput;
	FInput = 0;

	delete FPulse;
	FPulse = 0;
}

//---------------------------------------------------------------------------
//  X5_GSPSDeviceMap::OpenHardware()
//---------------------------------------------------------------------------

void  X5_GSPSDeviceMap::OpenHardware()
{
    X5DeviceMap::OpenHardware();

    // Configure A/Ds
    for (int ch = 0; ch < Input().Info().Channels().Channels(); ++ch)
		{
        FInput->PacketSize(0x4000);
		FInput->PacketId(0x0);
		FInput->Gain(ch, 1.0);
        FInput->Offset(ch, 0.0);
        Input().Unframed();
		}

	Pulse().Reset();
}

//---------------------------------------------------------------------------
//  X5_GSPSDeviceMap::OnStreamConfigureHandler()
//---------------------------------------------------------------------------

void  X5_GSPSDeviceMap::OnStreamConfigureHandler(OpenWire::NotifyEvent & Event)
{
	// Family-common init
	X5DeviceMap::ConfigureHardware();

	// Reapply PLL settings
    Clock.Reapply();
    Sleep(100);
	if (!Clock.Locked())
        Debug.Log("ERROR: Pll not locked!");

	//
	//  Apply input settings
	//  NOTE: Channel enables MUST PRECEED input configuration for 
	//        proper GSPS initialization!
	for (int ch = 0; ch < FInput->Info().Channels().Channels(); ++ch)
		{
		FInput->ChannelEnabled(ch, FInput->Info().Channels().Enabled(ch));
        FInput->PacketId(FInput->PacketId());
		FInput->PacketSize(FInput->PacketSize());
		}

	// Reapply Hardware Settings for  X5 IO Devices
	FInput->ConfigureHardware();

	//
	//  Apply Gain/Offset Coefficients
	if (FInput->Calibrated())
		FInput->ApplyRomCoefficients();
	else
		FInput->ApplyDefaultCoefficients();


	// Configure triggers
	FInput->TriggerReg.Apply();

	Pulse().Apply();

	//  Forward notify to baseboard handler
	OwnerCB()->CallOnStreamConfigure(Event);
}

//---------------------------------------------------------------------------
//  X5_GSPSDeviceMap::OnBeforeStreamStartHandler()
//---------------------------------------------------------------------------

void  X5_GSPSDeviceMap::OnBeforeStreamStartHandler(OpenWire::NotifyEvent & Event)
{
    //  Forward notify to baseboard handler
    OwnerCB()->CallOnBeforeStreamStart(Event);
	//
    //  Start execution
    UserFpgaControlReg.InGo.Set();
	UserFpgaControlReg.Apply();
}

//---------------------------------------------------------------------------
//  X5_GSPSDeviceMap::OnAfterStreamStartHandler()
//---------------------------------------------------------------------------

void  X5_GSPSDeviceMap::OnAfterStreamStartHandler(OpenWire::NotifyEvent & Event)
{
	//  Forward notify to baseboard handler
	OwnerCB()->CallOnAfterStreamStart(Event);

	Pulse().Run(true);
}

//---------------------------------------------------------------------------
//  X5_GSPSDeviceMap::OnBeforeStreamStopHandler()
//---------------------------------------------------------------------------

void  X5_GSPSDeviceMap::OnBeforeStreamStopHandler(OpenWire::NotifyEvent & Event)
{
	//  Forward notify to baseboard handler
	OwnerCB()->CallOnBeforeStreamStop(Event);
}

//---------------------------------------------------------------------------
//  X5_GSPSDeviceMap::OnAfterStreamStopHandler()
//---------------------------------------------------------------------------

void  X5_GSPSDeviceMap::OnAfterStreamStopHandler(OpenWire::NotifyEvent & Event)
{
	//  Disable User Logic Streaming
	UserFpgaControlReg.InGo.Clear();
	UserFpgaControlReg.Apply();

	Trigger(tInput, false);
	Pulse().Run(false);
//    Input().TriggerEnable(false);

	//  Forward notify to baseboard handler
	OwnerCB()->CallOnAfterStreamStop(Event);
}

//---------------------------------------------------------------------------
//  X5_GSPSDeviceMap::Input() -- Reference to input object
//---------------------------------------------------------------------------

X5_GSPSInputDevice & X5_GSPSDeviceMap::Input()
{
	return *FInput;
}

//---------------------------------------------------------------------------
//  X5_GSPSDeviceMap::Pulse() -- Reference to pulse object
//---------------------------------------------------------------------------

IPulseRepetitionIntervalIntf & X5_GSPSDeviceMap::Pulse()
{
	return *FPulse;
}

//---------------------------------------------------------------------------
//  X5_GSPSDeviceMap::SetCalibrationOffset()
//---------------------------------------------------------------------------

void X5_GSPSDeviceMap::SetCalibrationOffset(int /*ch*/, int /*value*/)
{
//    Owner->SetCalibrationOffset(ch, value);
}

//---------------------------------------------------------------------------
//  X5_GSPSDeviceMap::SetCalibrationGain()
//---------------------------------------------------------------------------

void X5_GSPSDeviceMap::SetCalibrationGain(int /*ch*/, int /*value*/)
{
//    Owner->SetCalibrationGain(ch, value);
}


} // namespace Innovative


