//---------------------------------------------------------------------------
// X6Idrom_Mb.cpp  --  X6 Idrom Access
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2011 Innovative Integration Corporation.
//  All Rights Reserved.
//
#include <X6Idrom_Mb.h>

namespace Innovative
{
#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif
//=============================================================================
//  CLASS X6IdromPage  --  Access to a segment of the X6 ROM as an IDROM
//=============================================================================
//-----------------------------------------------------------------------------
//  housekeeping for class X6IdromPage
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
//  constructor for class X6IdromPage
//-----------------------------------------------------------------------------

X6IdromPage::X6IdromPage(IIdRomDevice & rom, unsigned int page)
    :  IdromIoDG(this->IView()), ROM(rom), Page(page)
{
    Clear();
}

#ifdef _MSC_VER
	#pragma warning(pop)
#endif
//-----------------------------------------------------------------------------
//  destructor for class X6IdromPage
//-----------------------------------------------------------------------------

X6IdromPage::~X6IdromPage()
{
}

//-----------------------------------------------------------------------------
//  X6IdromPage::Clear() -- Blank out in-memory image
//-----------------------------------------------------------------------------

void  X6IdromPage:: Clear()
{
    for (int i=0; i<IdromPageSize; i++)
        Packet[i] = 0;
}

//-----------------------------------------------------------------------------
//  X6IdromPage::CalculateChecksum() -- Calculate checksum of packet
//-----------------------------------------------------------------------------

void  X6IdromPage::CalculateChecksum()
{
    unsigned int checksum(0);

    for ( unsigned int i = 0; i < idrom_data_size(); i++)
        checksum += Packet[i];

    Packet[IdromPageSize-ReservedSize] = checksum;
}

//-----------------------------------------------------------------------------
//  X6IdromPage::ErasePage() -- Use ROM to erase the IDROM page
//-----------------------------------------------------------------------------

void  X6IdromPage::ErasePage()
{
    ROM.ErasePage(Page);
}

//-----------------------------------------------------------------------------
//  X6IdromPage::LoadFromRom() -- Use ROM to read the IDROM page into cache
//-----------------------------------------------------------------------------

void  X6IdromPage::LoadFromRom()
{
    ROM.ReadPage(Page, Packet, IdromPageSize);
}

//-----------------------------------------------------------------------------
//  X6IdromPage::StoreToRom() -- Use ROM to write the IDROM page from cache
//-----------------------------------------------------------------------------

void  X6IdromPage::StoreToRom()
{
    ErasePage();
    ROM.WritePage(Page, Packet, IdromPageSize);
}

//=============================================================================
//  CLASS X6SystemIdromPage  --  System Reserved IDROM page
//=============================================================================
//------------------------------------------------------------------------
//   X6SystemIdromPage::Name()  --  Read ROM name code from IDROM
//------------------------------------------------------------------------

std::string  X6SystemIdromPage::Name() const
{
    std::string name;
    for (int i=0; i<16; i++)
        {
        char c = (char)AsInt(i);
        if (c=='\0')
            break;
        name += c;
        }
    return name;
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::Name()  --  Write ROM name code to Image
//------------------------------------------------------------------------

void  X6SystemIdromPage::Name(const std::string & name)
{
    for (int i=0; i<16; i++)
        {
        AsInt( i, static_cast<unsigned int>(name[i]) );
        if (name[i]=='\0')
            break;
        }
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::Revision()  --  Read ROM rev code from IDROM
//------------------------------------------------------------------------

std::string  X6SystemIdromPage::Revision() const
{
    std::string name;

    name += static_cast<char>(AsInt(16));
    return name;
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::Revision()  --  Write ROM rev code to IDROM
//------------------------------------------------------------------------

void  X6SystemIdromPage::Revision(const std::string & revision)
{
    AsInt(16, revision[0]);
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::PartNumber()  --  Read PartNumber from IDROM
//------------------------------------------------------------------------

unsigned int  X6SystemIdromPage::PartNumber() const
{
    unsigned int part;
    part = AsInt(17);
    return part;
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::PartNumber()  --  Write PartNumber to Image
//------------------------------------------------------------------------

void  X6SystemIdromPage::PartNumber(unsigned int part)
{
    AsInt(17, part);
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::DashNumber()  --  Read Board Dash Code from IDROM
//------------------------------------------------------------------------

unsigned int  X6SystemIdromPage::DashNumber() const
{
    unsigned int dash;

    dash = AsInt(18);
    return dash;
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::DashNumber()  --  Write Board Dash Code to Image
//------------------------------------------------------------------------

void  X6SystemIdromPage::DashNumber(unsigned int dash)
{
    AsInt(18, dash);
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::DashVariant()  --  Read Board Dash Variant Code from IDROM
//------------------------------------------------------------------------

std::string  X6SystemIdromPage::DashVariant() const
{
    std::string dashV;

    dashV.push_back(AsInt(19));
    return dashV;
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::DashVariant()  --  Write Board Dash Variant to Image
//------------------------------------------------------------------------

void  X6SystemIdromPage::DashVariant(const std::string & dashV)
{
    AsInt(19, dashV[0]);
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::RuggedizationNumber()  --  Read Board Dash Code from IDROM
//------------------------------------------------------------------------

unsigned int  X6SystemIdromPage::RuggedizationNumber() const
{
    unsigned int rugged;

    rugged = AsInt(20);
    return rugged;
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::RuggedizationNumber()  --  Write Board Dash Code to Image
//------------------------------------------------------------------------

void  X6SystemIdromPage::RuggedizationNumber(unsigned int rugged)
{
    AsInt(20, rugged);
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::EcoSubRevNumber()  --  Read ECO Code from IDROM
//------------------------------------------------------------------------

unsigned int  X6SystemIdromPage::EcoSubRevNumber() const
{
    unsigned int subrev;

    subrev = AsInt(21);
    return subrev;
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::EcoSubRevNumber()  --  Write ECO Code to Image
//------------------------------------------------------------------------

void  X6SystemIdromPage::EcoSubRevNumber(unsigned int subrev)
{
    AsInt(21, subrev);
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::Updated()  -- read 'written to' flag for a section
//------------------------------------------------------------------------

bool  X6SystemIdromPage::Updated() const
{
    return (AsInt(22)==1) ? true : false; // make sure junk is 'false'
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::Updated()  -- Mark data area as 'written to'
//------------------------------------------------------------------------

void  X6SystemIdromPage::Updated(bool state)
{
    AsInt(22, (state) ? 1 : 0);
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::SerialNumber()  --  Read Serial from IDROM
//------------------------------------------------------------------------

std::string  X6SystemIdromPage::SerialNumber() const
{
    std::string sn;

    sn = AsString(23, 5);   // 20 characters max
    return sn;
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::SerialNumber()  --  Write Serial to Image
//------------------------------------------------------------------------

void  X6SystemIdromPage::SerialNumber(const std::string & sn)
{
    AsString(23, 5, sn);
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::FpgaType()  --  Read Fpga Type Code from IDROM
//------------------------------------------------------------------------

unsigned int  X6SystemIdromPage::FpgaType() const
{
    unsigned int type;

    type = AsInt(28);
    return type;
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::FpgaType()  --  Write Fpga Type to Image
//------------------------------------------------------------------------

void  X6SystemIdromPage::FpgaType(unsigned int type)
{
    AsInt(28, type);
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::FpgaSpeed()  --  Read Fpga Type Code from IDROM
//------------------------------------------------------------------------

unsigned int  X6SystemIdromPage::FpgaSpeed() const
{
    unsigned int speed;

    speed = AsInt(29);
    return speed;
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::FpgaSpeed()  --  Write Fpga Type to Image
//------------------------------------------------------------------------

void  X6SystemIdromPage::FpgaSpeed(unsigned int speed)
{
    AsInt(29, speed);
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::FpgaTempGrade()  --  Read Fpga Type Code from IDROM
//------------------------------------------------------------------------

unsigned int  X6SystemIdromPage::FpgaTempGrade() const
{
    unsigned int tempgrade;

    tempgrade = AsInt(30);
    return tempgrade;
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::FpgaTempGrade()  --  Write Fpga Type to Image
//------------------------------------------------------------------------

void  X6SystemIdromPage::FpgaTempGrade(unsigned int tempgrade)
{
    AsInt(30, tempgrade);
}

//------------------------------------------------------------------------
//   X6SystemIdromPage::FirstUnusedWord()  -- First safe int for addons
//------------------------------------------------------------------------

unsigned int  X6SystemIdromPage::FirstUnusedWord() const
{
	return 31;
}

#ifdef _MSC_VER
	#pragma warning(push)
	#pragma warning(disable : 4355)
#endif
//=============================================================================
//  CLASS X6IdromPageSection  --  Object for access to a section of IDROM Page
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class X6IdromPageSection
//-----------------------------------------------------------------------------

X6IdromPageSection::X6IdromPageSection(X6IdromPage & rom, int offset,
                           int section_data_size, int section_reserve_size)
    :  IdromIoDG(this->IView()), Rom(rom), Offset(offset),
       SectionSize(section_data_size + section_reserve_size),
       SectionReserveSize(section_reserve_size)

{
}

#ifdef _MSC_VER
	#pragma warning(pop)
#endif
//-----------------------------------------------------------------------------
//  X6IdromPageSection::Clear() -- Erase the section
//-----------------------------------------------------------------------------

void  X6IdromPageSection::Clear()
{
    for (unsigned int i=0; i<SizeInInts(); i++)
        AsInt(i, 0);
}

//-----------------------------------------------------------------------------
//  X6IdromPageSection::LoadFromRom() -- Read data from ROM
//-----------------------------------------------------------------------------

void  X6IdromPageSection::LoadFromRom()
{
    Rom.LoadFromRom();
}

//-----------------------------------------------------------------------------
//  X6IdromPageSection::StoreToRom() -- Write data to ROM
//-----------------------------------------------------------------------------

void  X6IdromPageSection::StoreToRom()
{
    Rom.StoreToRom();
}

//------------------------------------------------------------------------
//  X6IdromPageSection::Updated()  -- read 'written to' flag for a section
//------------------------------------------------------------------------

bool  X6IdromPageSection::Updated() const
{
    if (SectionReserveSize == 0)
        return true;

    //  Use first word of reserved region as update flag
    const int * update_flag = const_idrom_base() + idrom_data_size();

    return ((*update_flag)==1) ? true : false ;
}

//------------------------------------------------------------------------
//   X6IdromPageSection::Updated()  -- Mark data area as 'written to'
//------------------------------------------------------------------------

void  X6IdromPageSection::Updated(bool state)
{
    if (SectionReserveSize == 0)
        return;

    //  Use first word of reserved region as update flag
    int * update_flag = idrom_base() + idrom_data_size();
    (*update_flag) = (state) ? 1 : 0 ;
}

//=============================================================================
//  CLASS X6CalibrationSection  --  IDROM section for analog calibration
//=============================================================================
//-----------------------------------------------------------------------------
//  constructor for class X6CalibrationSection
//-----------------------------------------------------------------------------

X6CalibrationSection::X6CalibrationSection(X6IdromPage & rom, int offset, int channels, int one,
                                           IGainCalibrationDevice & gain_cal)
	:  X6IdromPageSection(rom, offset, channels*2), Channels(channels), FOne(one),
       GainCal(gain_cal)
{
}

//-----------------------------------------------------------------------------
//  X6CalibrationSection::ApplyDefaultCoefficients() -- load default set of coeffs into memory
//-----------------------------------------------------------------------------

void  X6CalibrationSection::ApplyDefaultCoefficients()
{
	for (int ch = 0; ch < Channels; ++ch)
		{
		Offset(ch, 0);
		Gain(ch, 1.0);
		}
}

//-----------------------------------------------------------------------------
//  X6CalibrationSection::ApplyRomCoefficients() --
//-----------------------------------------------------------------------------

void  X6CalibrationSection::ApplyRomCoefficients()
{
    if (!Calibrated())
        {
        ApplyDefaultCoefficients();
        return;
        }

	for (int ch = 0; ch < Channels; ++ch)
		{
		Offset(ch, Offset(ch));
		Gain(ch, Gain(ch));
		}
}

//-----------------------------------------------------------------------------
//  X6CalibrationSection::ToFixed() - Converts a float gain coeff to internal fixed-point format
//-----------------------------------------------------------------------------

unsigned int X6CalibrationSection::ToFixed(float value)
{
    return static_cast<int>(value*One());
}

//------------------------------------------------------------------------------
//  X6CalibrationSection::FromFixed() - Converts an internal fixed-point format to float
//------------------------------------------------------------------------------

float X6CalibrationSection::FromFixed(unsigned int value)
{
    return static_cast<float>(value)/One();
}

//-----------------------------------------------------------------------------
//  X6CalibrationSection::Gain()  --  Set gain coefficient
//-----------------------------------------------------------------------------

void X6CalibrationSection::Gain(int ch, float gain)
{
	int idx = (Channels + ch);

	AsFloat(idx, gain);
	GainCal.SetCalibrationGain(ch, ToFixed(gain));
}

//-----------------------------------------------------------------------------
//  X6CalibrationSection::Gain()  --  Get gain coefficient
//-----------------------------------------------------------------------------

float X6CalibrationSection::Gain(int ch) const
{
	int idx = (Channels + ch);
	return AsFloat(idx);
}

//-----------------------------------------------------------------------------
//  X6CalibrationSection::Offset()  --  Set offset coefficient
//-----------------------------------------------------------------------------

void X6CalibrationSection::Offset(int ch, float offset)
{
	int idx = ch;

	AsFloat(idx, offset);
	GainCal.SetCalibrationOffset(ch, (int)offset);
}

//-----------------------------------------------------------------------------
//  X6CalibrationSection::Offset()  --  Get offset coefficient
//-----------------------------------------------------------------------------

float X6CalibrationSection::Offset(int ch) const
{
	int idx = ch;
	return AsFloat(idx);
}

//-----------------------------------------------------------------------------
//  X6CalibrationSection::One()  --  Get Q-notation unity-value
//-----------------------------------------------------------------------------

int X6CalibrationSection::One() const
{
	return FOne;
}

}  // end namespace Innovative
