//---------------------------------------------------------------------------
// X6Idrom_Mb.h  --  X6 Idrom Access
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2011 Innovative Integration Corporation.
//  All Rights Reserved.
//

#ifndef X6Idrom_MbH
#define X6Idrom_MbH

#include <IdromDatagram_Mb.h>
#include <string>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//=============================================================================
//  CLASS IIdRomDevice  --  Interface to devices supporting IDROM pages
//=============================================================================

class IIdRomDevice
{
public:
    virtual void   ErasePage(unsigned int page) = 0;
    virtual void   ReadPage(unsigned int page, int * pkt, unsigned int size) = 0;
    virtual void   WritePage(unsigned int page, int * pkt, unsigned int size) = 0;

    virtual unsigned int  JdecId() = 0;
    virtual unsigned int  Rdid() {  return 0; }

    virtual ~IIdRomDevice()
        {}

};

//=============================================================================
//  CLASS IGainCalibrationDevice  --  Interface to gain calibration registers
//=============================================================================

class IGainCalibrationDevice
{
public:
    virtual void   SetCalibrationOffset(unsigned int ch, unsigned int value) = 0;
    virtual void   SetCalibrationGain(unsigned int ch, unsigned int value) = 0;

    virtual ~IGainCalibrationDevice()
        {}

};

//=============================================================================
//  CLASS X6IdromPage  --  Access to a segment of the X6 ROM as an IDROM
//=============================================================================
//  The X6 IDROM can be logically broken up into 1 kW pages that can be erased
//    independently of the others.  This class provides an interface to one
//    of those pages
//
//  The last word of the IDROM page is a checksum of the rest of the page
//

class X6IdromPage :  public IIdromViewable, public IdromIoDG
{
    static const int IdromPageSize = 1024;   // IDROM size in words
    static const int ReservedSize  = 1;      //  Checksum area

public:

    friend class X6IdromPageSection;
    //
    //  Member Functions
    X6IdromPage(IIdRomDevice & rom, unsigned int page);
    virtual ~X6IdromPage();

    //  Data Operations
    void   Clear();
    int    Checksum() const
        {  return Packet[IdromPageSize-1];  }

    //  ROM Operations
    void   ErasePage();
    void   LoadFromRom();
    void   StoreToRom();

protected:
    //
    //  IIdromViewable Interface Implementation
    virtual const int *  const_idrom_base() const
                            {  return Packet;  }
    virtual int *        idrom_base()
                            {  return Packet;  }

    virtual size_t       idrom_size() const
                            {  return IdromPageSize;  }
    virtual size_t       idrom_reserve_size() const
                            {  return ReservedSize;  }
    size_t               idrom_data_size() const
                            {  return idrom_size() - idrom_reserve_size();  }

    virtual void         after_data_write()
                            {  CalculateChecksum();  }

    IIdromViewable & IView()
        {  return *this;  }
        
private:
    //
    //  Member Data
    IIdRomDevice &       ROM;
    int                  Packet[IdromPageSize];
    unsigned int         Page;
    //
    //  Support methods
    void        CalculateChecksum();
};

//=============================================================================
//  CLASS X6SystemIdromPage  --  System Reserved IDROM page
//=============================================================================
//  Page dedicated to the standard Board ID information
//

class X6SystemIdromPage :  public X6IdromPage
{
public:
    //
    //  Member Functions
    X6SystemIdromPage(IIdRomDevice & rom, unsigned int page) : X6IdromPage(rom, page)
        {}
    ~X6SystemIdromPage()
        {}
    //
    //  Data Access Properties
    std::string     Name() const;
    void            Name(const std::string & name);
    std::string     Revision() const;
    void            Revision(const std::string & revision);
    unsigned int    PartNumber() const;
    void            PartNumber(unsigned int part);
    unsigned int    DashNumber() const;
    void            DashNumber(unsigned int dash);
    std::string     DashVariant() const;
    void            DashVariant(const std::string & dashV);
    unsigned int    RuggedizationNumber() const;
    void            RuggedizationNumber(unsigned int rugged);
    unsigned int    EcoSubRevNumber() const;
    void            EcoSubRevNumber(unsigned int eco_subrev);
    std::string     SerialNumber() const;
    void            SerialNumber(const std::string & sn);
    unsigned int    FpgaType() const;
    void            FpgaType(unsigned int type);
    unsigned int    FpgaSpeed() const;
    void            FpgaSpeed(unsigned int speed);
    unsigned int    FpgaTempGrade() const;
    void            FpgaTempGrade(unsigned int temp_grade);

    bool            Updated() const;          // 'Section written' flag
    void            Updated(bool state);

    unsigned int    FirstUnusedWord() const;

};

//=============================================================================
//  CLASS X6IdromPageSection  --  Object for access to a section of IDROM Page
//=============================================================================
//  When a single IDROM Work area is to be used by multiple clients, you can use a
//     predefined section object to manage the location of the data in the ROM

class X6IdromPageSection  :  public IIdromViewable, public IdromIoDG
{
public:
    //
    //  Member Functions
    X6IdromPageSection(X6IdromPage & rom, int offset,
                           int section_data_size, int section_reserve_size=1);
    virtual ~X6IdromPageSection()
        {  }

    void   Clear();
    void   LoadFromRom();
    void   StoreToRom();

    bool   Updated() const;          // 'Section written' flag
    void   Updated(bool state);

protected:
    //
    //  IIdromViewable Interface Implementation
    virtual const int *  const_idrom_base() const
                            {  return Rom.const_idrom_base() + Offset;  }
    virtual int *        idrom_base()
                            {  return Rom.idrom_base() + Offset;;  }

    virtual size_t       idrom_size() const
                            {  return SectionSize;  }
    virtual size_t       idrom_reserve_size() const
                            {  return SectionReserveSize;  }
    size_t               idrom_data_size() const
                            {  return idrom_size() - idrom_reserve_size();  }

    virtual void         after_data_write()
                            {  Rom.after_data_write();  }

    IIdromViewable & IView()
        {  return *this;  }

private:
    X6IdromPage &       Rom;
    int                 Offset;
    int                 SectionSize;
    int                 SectionReserveSize;
};

//=============================================================================
//  CLASS X6CalibrationSection  --  IDROM section for analog calibration
//=============================================================================

class X6CalibrationSection :  public X6IdromPageSection
{
public:

    X6CalibrationSection(X6IdromPage & rom, int offset, int channels, int one,
                         IGainCalibrationDevice & gain_cal);
    virtual ~X6CalibrationSection()
        {  }

    // Properties
	bool	Calibrated() const
                {  return Updated();  }
	void	Calibrated(bool state)
                {  Updated(state);  }

    float 	Offset(int ch) const;
    void	Offset(int ch, float value);

	float 	Gain(int ch) const;
    void	Gain(int ch, float value);

    //  Loading functions
	void	ApplyRomCoefficients();
    void    ApplyDefaultCoefficients();

protected:
	// Data
	int					Channels;
    int                 FOne;

	// Methods
	unsigned int 		ToFixed(float value);
	float 				FromFixed(unsigned int value);

	int					One() const;

    IGainCalibrationDevice & GainCal;

};

#ifdef __CLR_VER
#pragma managed(pop)
#endif

}  // end namespace Innovative


//---------------------------------------------------------------------------
#endif
