// X6_RXDeviceMap_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2010 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <X6_RX_Mb.h>
#include <X6_RXDeviceMap_Mb.h>
#include <X6MemoryMap_Mb.h>
#include <X6_RXMemoryMap_Mb.h>
#include <SystemSupport_Mb.h>

//---------------------------------------------------------------------------
namespace Innovative
{

using namespace X6_MemoryMap;
using namespace X6_RXMemoryMap;

//---------------------------------------------------------
//  Module Constants Definition, Clock is either external or internal fixed crystal.

const int    MaxInDevices      = 2;
const int    ChannelsPerDevice = 2;
const int    MaxInChannels     = MaxInDevices*ChannelsPerDevice;

const float         InputMaxRate            = 160.0e6f;
const float         InputMinRate            = 0.0e6f;
const unsigned int  InputWidth_Bits         = 16;
const          int  InputMax_mV             = 1200;
const          int  InputMin_mV             = -1200;
const unsigned int  InputSamplesPerWord     = 2;
const unsigned int  InputTriggerGranularity = 256;

const unsigned int X6_RX_AdcSpiDevices      = MaxInDevices;
const unsigned int X6RX_VeloInStreamCount   = 1;   // Number of Velocia Incoming PIDs
const unsigned int X6RX_VitaInStreamCount   = MaxInChannels;   // Number of VITA Incoming SIDs

const unsigned int X6RX_AdcCount            = MaxInChannels;  // Number of Adc Channels
const unsigned int X6RX_CalibrationAdcCount = MaxInChannels;  // Number of Adc Calibrations

const unsigned int X6_RX_DacSpiDevices      = 0;   // No DACs with SPIs

const unsigned int X6RX_VeloOutStreamCount  = 1;   // Number of Velo PIDs
const unsigned int X6RX_VitaOutStreamCount  = 1;   // Number of VITA Outgoing SIDs

const unsigned int X6_RX_Adc_One = 0x10000;    // representation of floating point 1.0

//---------------------------------------------------------
//  IDROM Section Information
//  ...System Section: Standard Data only
const int romX6_RX_SysInfoOffset        = 0;
const int romX6_RX_SysInfoOffset_avail  = 18;  // first 'free' word

//  ...Adc Cal Idrom: ADC coefficients
const int romX6_RX_InputCoeffOffset       = 0;
const int romX6_RX_InputCoeffOffset_avail = 9;   // 4 * 2 coefficients plus 'calibrated' word

//  ...Dac Cal Idrom: DAC coefficients
const int romX6_RX_OutputCoeffOffset       = 0;
const int romX6_RX_OutputCoeffOffset_avail = 0;  // none used

//===========================================================================
// Class X6_RX_AnalogInInfo
//===========================================================================

class X6_RX_AnalogInInfo : public IX6AnalogIoInfo
{
public:
    X6_RX_AnalogInInfo() : FSpan(InputMax_mV, InputMin_mV)
        {   }
    virtual ~X6_RX_AnalogInInfo()  {  }

    virtual float     MaxRate() const         {  return InputMaxRate;  }
    virtual float     MinRate() const         {  return InputMinRate;  }
    virtual SpanInfo  Span() const            {  return FSpan;  }
    virtual SpanInfo  Span(size_t /*gain_range*/) const
                {  return Span();  }
    virtual unsigned int  Bits() const        {  return InputWidth_Bits;  }

    virtual unsigned int  SamplesPerWord() const {  return InputSamplesPerWord;  }
    virtual unsigned int  TriggerFrameGranularity() const   {  return InputTriggerGranularity;  }

private:
    SpanInfo   FSpan;
};

//===========================================================================
//  CLASS X6_RXInputDevice  -- Register & low level support for X6 RX
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class X6_RXInputDevice
//---------------------------------------------------------------------------

X6_RXInputDevice::X6_RXInputDevice(X6_RXDeviceMap * owner, IX6AnalogIoInfo * info)
    : inherited(owner), FInfo(info),
      CalRomSection(AdcCalIdrom(), romX6_RX_InputCoeffOffset, MaxInChannels,
                       X6_RX_Adc_One, AdcCal() )
{
}

//---------------------------------------------------------------------------
//  destructor for class X6_RXInputDevice
//---------------------------------------------------------------------------

X6_RXInputDevice::~X6_RXInputDevice()
{
	delete FInfo;
	FInfo = 0;
}

//---------------------------------------------------------------------------
//  X6_RXInputDevice::Info() -- Analog Information
//---------------------------------------------------------------------------

IX6AnalogIoInfo &  X6_RXInputDevice::Info()
{
	return *FInfo;
}

//---------------------------------------------------------------------------
//  X6_RXInputDevice::Channels() -- Number of enable-able ADC channels
//---------------------------------------------------------------------------

unsigned int  X6_RXInputDevice::Channels()
{
	return MaxInChannels;
}

//---------------------------------------------------------------------------
//  X6_RXInputDevice::ConfigureHardware()
//---------------------------------------------------------------------------
//
//void X6_RXInputDevice::ConfigureHardware()
//{
//    inherited::ConfigureHardware();
//}
//
//=============================================================================
//  CLASS X6_RX_Pll  --  PLL timebase control
//=============================================================================
//---------------------------------------------------------------------------
//  housekeeping for class X6_RX_Pll
//---------------------------------------------------------------------------

const float  Input_ReferenceClk = 10.0e6;

//---------------------------------------------------------------------------
//  constructor for class X6_RX_Pll
//---------------------------------------------------------------------------

X6_RX_Pll::X6_RX_Pll(X6_RXDeviceMap * owner, IAddressingSpace &space, Cdce72010Registers & reg)
    : Cdce72010(space, reg), Owner(owner), FPDF(100.0e3)
{
    SetMasterOutputSelection(1);  // Use Output 1 as guideline output

    ReferenceFrequency(0, Input_ReferenceClk);  // All Refs are 10
    ReferenceFrequency(1, Input_ReferenceClk);
}

//---------------------------------------------------------------------------
//  X6_RX_Pll::OnInit() -- inital configuration
//---------------------------------------------------------------------------

void  X6_RX_Pll::OnInit()
{
    //  Device Powered/Enabled Setup
    SetPllEnable(true);
    SetFBDividerEnable(true);
    //
    //  Set Input Configuration
    SetInputConfig(icLvPecl);
    SetInputBufferConfig(false,  // AC Terminated
                         false,  // Input Term NOT disabled
                         false,  // Primary Bias OFF
                         true,   // Secondary Bias ON
                         false   // NO failsafe
                         );
    //
    //  Set Vco/Aux Configuration
    SetVcoConfig(vcLvPecl);
    SetVcoBufferConfig(true,   // DC Terminated
                       false,  // Hysteresis
                       false,  // Input Term NOT disabled
                       false   // Vco Bias OFF
                       );
    //
    //  Set Output Configuration
    OutputInUse(0, true);   SetOutputConfig(0, ocLvds);     // FPGA
    OutputInUse(1, true);   SetOutputConfig(1, ocLvPecl);   // fudge to get 0's divider turned on...
    //OutputInUse(1, false);  SetOutputConfig(1, ocDisabled);
    OutputInUse(2, false);  SetOutputConfig(2, ocDisabled);
    OutputInUse(3, false);  SetOutputConfig(3, ocDisabled);
    OutputInUse(4, false);  SetOutputConfig(4, ocDisabled);
    OutputInUse(5, true);   SetOutputConfig(5, ocLvCmos_P_Active | ocLvCmos_N_TriStated);   // ADC Dev 0 (Ch0-1)
    OutputInUse(6, false);  SetOutputConfig(6, ocDisabled);
    OutputInUse(7, true);   SetOutputConfig(7, ocLvCmos_P_Active | ocLvCmos_N_TriStated);   // ADC Dev 1 (Ch2-3)
    OutputInUse(8, false);  SetOutputConfig(8, ocDisabled);
    OutputInUse(9, false);  SetOutputConfig(9, ocDisabled);

    //  Phase = 0
    SetPhaseAdjust(0, 1, 0);   // Set all outputs to 0 phase
    SetPhaseAdjust(5, 1, 0);
    SetPhaseAdjust(7, 1, 0);

    //  Configure Aux
    SetAuxInputConfig(aeEnabled, aiNormal);
    SetLockConfig(false);   // NOT analog lock
    //
    //  Set Frequency Limits for Outputs (Preset to 1500.0 MHz)
    //    (Outputs 0 and 1 match preset value, so leave alone)
    //OutputFrequencyLimit[6] = 100.0e6;
    //OutputFrequencyLimit[8] = 250.0e6;
    //
    //  Reference
    SetExternalReferenceControl(true);  // Use external pin
}

//---------------------------------------------------------------------------
//  X6_RX_Pll::Vcxo() -- return Vcxo
//---------------------------------------------------------------------------

IVcxoControl &  X6_RX_Pll::Vcxo()
{
    return Owner->AFEClk_SDev.Vcxo();
}

//---------------------------------------------------------------------------
//  X6_RX_Pll::OnReapply()
//---------------------------------------------------------------------------

void  X6_RX_Pll::OnReapply()
{
}

//=============================================================================
//  CLASS X6_RX_SioPll  --  SIO PLL timebase control
//=============================================================================
//---------------------------------------------------------------------------
//  housekeeping for class X6_RX_SioPll
//---------------------------------------------------------------------------

const float  SioPll_ReferenceClk = 10.0e6;

//---------------------------------------------------------------------------
//  constructor for class X6_RX_SioPll
//---------------------------------------------------------------------------

X6_RX_SioPll::X6_RX_SioPll(X6_RXDeviceMap * owner, I2CDevice & device)
    : Si5338(device), Owner(owner)
{
    SetMasterOutputSelection(0);  // Use Output 0 as guideline output

    ReferenceFrequency(0, SioPll_ReferenceClk);  // All Refs are 10
    ReferenceFrequency(1, SioPll_ReferenceClk);
    ReferenceFrequency(2, SioPll_ReferenceClk);
    ReferenceFrequency(3, SioPll_ReferenceClk);

    ReferenceSource(ref3SingleEnded);  // Use IN3 as input clock

    Debug.Enabled(false);  // turn off verbose object output
}

//---------------------------------------------------------------------------
//  X6_RX_SioPll::OnInit() -- inital configuration
//---------------------------------------------------------------------------

void  X6_RX_SioPll::OnInit()
{
    DenomForN(1);
    DenomForM(0, 1);     // match settings for utility
    DenomForM(1, 1);
    DenomForM(2, 1);
    DenomForM(3, 1);

    //  PLL enabled by default in SettingsInit()
    //  Device Powered/Enabled Setup
    SetClkPowerDown(0, false);    SetMultiSynthPowerDown(0, false);
    SetClkPowerDown(1, false);    SetMultiSynthPowerDown(1, false);
    SetClkPowerDown(2, true);     SetMultiSynthPowerDown(2, true);
    SetClkPowerDown(3, false);    SetMultiSynthPowerDown(3, false);
    //
    //  Set Input Configuration
    SetInputMuxConfig(ra3SingleEnded, SioPll_ReferenceClk);  // rate needed if ra12Xtal used
    SetFeedbackMuxConfig(fbNoClk);    
    SetPllReferenceInput(prRefClk);
    SetFeedbackSource(false, efbNoClk);     // use internal feedback
    //
    //  Set Output Configuration
    //  SetOutDividerSource(0, N) (Use standard setting of from SettingsInit()
    SetOutDividerSource(2, diNoClk);   // override since not using output

    OutputInUse(0, true);   SetOutputConfig(0, ocLvds, oiNormal);  OutputEnabled(0, true);
    OutputInUse(1, true);   SetOutputConfig(1, ocLvds, oiNormal);  OutputEnabled(1, true);
    OutputInUse(2, false);  SetOutputConfig(2, ocLvds, oiNormal);  OutputEnabled(2, false);
    OutputInUse(3, true);   SetOutputConfig(3, ocLvds, oiNormal);  OutputEnabled(3, true);

    SetOutputTrim(0, ot_2p5lvds);
    SetOutputTrim(1, ot_2p5lvds);
    SetOutputTrim(2, ot_2p5lvds);
    SetOutputTrim(3, ot_2p5lvds);

    SetOutputVoltage(0, vo_2p5volts);
    SetOutputVoltage(1, vo_2p5volts);
    SetOutputVoltage(2, vo_2p5volts);   // actually off
    SetOutputVoltage(3, vo_2p5volts);

    SetOutputDisableState(0, odHighZ);
    SetOutputDisableState(1, odHighZ);
    SetOutputDisableState(2, odHighZ);
    SetOutputDisableState(3, odHighZ);

    OutputFrequency(0, 100.0e6);
    OutputFrequency(1, 100.0e6);
    OutputFrequency(2, 100.0e6);  // actually off
    OutputFrequency(3, 100.0e6);

}

//---------------------------------------------------------------------------
//  X6_RX_SioPll::OnReapply()
//---------------------------------------------------------------------------

void  X6_RX_SioPll::OnReapply()
{
}

//===========================================================================
//  CLASS X6_RXDeviceMap  -- Register & low level support for SD module
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class X6_RXDeviceMap
//---------------------------------------------------------------------------

const double  kVcxoPowerOnRate = 10.0;    // Si570 power on rate (MHz)

static Cdce72010Registers X6_RX_PllRegisters =
    { mmAFE_PllControl,  mmAFE_PllSpi  };

static X6_Std_I2CLinkage::MapRegisters X6_RX_VcoI2CLinkRegAddrs =
    {  mmAFE_VcxoControl,      //  AfeVcxoControl_Addr
    };
    
//---------------------------------------------------------------------------
//  constructor for class X6_RXDeviceMap
//---------------------------------------------------------------------------

#ifdef _MSC_VER
    #pragma warning(push)
    #pragma warning(disable : 4355)
#endif

X6_RXDeviceMap::X6_RXDeviceMap(X6_RX * owner)
    :  X6DeviceMap(owner),
       FInput(new X6_RXInputDevice(this, new X6_RX_AnalogInInfo)),
       Owner(owner)
{
    //
    //  Since this device has no DACs, disable all DAC subdevices
    DisableDacDevices();
    //
    //  SubDevice Customization
    X6DeviceMap::ConstructCustomParts();
    //
    //  Hook stream events from base class
    PDMPtr()->OnStreamPreconfig.SetEvent(this,   &X6_RXDeviceMap::OnStreamPreconfigHandler );
    PDMPtr()->OnStreamSingleTimeAtStartConfig.SetEvent(this, &X6_RXDeviceMap::OnStreamSingleTimeAtStartConfigHandler );
    PDMPtr()->OnStreamConfigure.SetEvent(this,   &X6_RXDeviceMap::OnStreamConfigureHandler );
    PDMPtr()->OnBeforeStreamStart.SetEvent(this, &X6_RXDeviceMap::OnBeforeStreamStartHandler );
    PDMPtr()->OnAfterStreamStart.SetEvent(this,   &X6_RXDeviceMap::OnAfterStreamStartHandler );
    PDMPtr()->OnBeforeStreamStop.SetEvent(this,   &X6_RXDeviceMap::OnBeforeStreamStopHandler );
    PDMPtr()->OnAfterStreamStop.SetEvent(this,   &X6_RXDeviceMap::OnAfterStreamStopHandler );


}

#ifdef _MSC_VER
    #pragma warning(pop)
#endif

//---------------------------------------------------------------------------
//  destructor for class X6_RXDeviceMap
//---------------------------------------------------------------------------

X6_RXDeviceMap::~X6_RXDeviceMap()
{
	delete FInput;
	FInput = 0;
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::OpenHardware()
//---------------------------------------------------------------------------

void  X6_RXDeviceMap::OpenHardware()
{
    X6DeviceMap::OpenHardware();
    FInput->Cal().ApplyRomCoefficients();
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::OnStreamPreconfigHandler() - preconfigure operations
//---------------------------------------------------------------------------

void  X6_RXDeviceMap::OnStreamPreconfigHandler(OpenWire::NotifyEvent & Event)
{
    FInput->Cal().ApplyRomCoefficients();

	// Family-common init
	X6DeviceMap::PreconfigureHardware();

    //  Forward notify to baseboard handler
    OwnerCB()->CallOnStreamPreconfig(Event);
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::OnStreamSingleTimeAtStartConfigHandler()
//---------------------------------------------------------------------------

void  X6_RXDeviceMap::OnStreamSingleTimeAtStartConfigHandler(OpenWire::NotifyEvent & Event)
{
    //  Forward notify to baseboard handler
    OwnerCB()->CallOnStreamSingleTimeAtStartConfig(Event);
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::OnStreamConfigureHandler()
//---------------------------------------------------------------------------

void  X6_RXDeviceMap::OnStreamConfigureHandler(OpenWire::NotifyEvent & Event)
{
    // Family-common init
    X6DeviceMap::ConfigureHardware();
    //
    //  Special Config for A/D Devices
    Input().Spi(0, 0x1, 0);    // set A/D clock phase for 160 MHz
    Sleep(10);
    Input().Spi(1, 0x1, 0);    //   bits 4:2 to 000 to set 31 degree clock phase
    Sleep(10);

    //  Forward notify to baseboard handler
    OwnerCB()->CallOnStreamConfigure(Event);
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::OnBeforeStreamStartHandler()
//---------------------------------------------------------------------------

void  X6_RXDeviceMap::OnBeforeStreamStartHandler(OpenWire::NotifyEvent & Event)
{
    //  Forward notify to baseboard handler
    OwnerCB()->CallOnBeforeStreamStart(Event);
    //
    //  Start execution
    AFECommon_SDev.Engine(X6AFE_Common::tInput, true);
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::OnAfterStreamStartHandler()
//---------------------------------------------------------------------------

void  X6_RXDeviceMap::OnAfterStreamStartHandler(OpenWire::NotifyEvent & Event)
{
    //  Forward notify to baseboard handler
    OwnerCB()->CallOnAfterStreamStart(Event);
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::OnBeforeStreamStopHandler()
//---------------------------------------------------------------------------

void  X6_RXDeviceMap::OnBeforeStreamStopHandler(OpenWire::NotifyEvent & Event)
{
    //  Forward notify to baseboard handler
    OwnerCB()->CallOnBeforeStreamStop(Event);
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::OnAfterStreamStopHandler()
//---------------------------------------------------------------------------

void  X6_RXDeviceMap::OnAfterStreamStopHandler(OpenWire::NotifyEvent & Event)
{
    //  Base class "on stream stop" behavior
    X6DeviceMap::StreamStopHardware();

	//  Forward notify to baseboard handler
	OwnerCB()->CallOnAfterStreamStop(Event);
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::MakeAfePllObject() -- Reference to input object
//---------------------------------------------------------------------------

Cdce72010 *  X6_RXDeviceMap::MakeAfePllObject()
{
    //  Make our local PLL object (subdev will delete it)
    return new X6_RX_Pll(this, WB_AFE, X6_RX_PllRegisters );
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::MakeAfeVcxoObject() --  Build custom VCO object
//---------------------------------------------------------------------------

Si570 *  X6_RXDeviceMap::MakeAfeVcxoObject(IRequires_I2C_RegIo & device)
{
    //  Make our local Vcxo object (subdev will delete it)
    return new Si570(device, Si570::tcC, kVcxoPowerOnRate );
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::MakeAfeVcxoI2CLink() --  Build custom Vco I2C Link
//---------------------------------------------------------------------------

IX6_I2C_uDevice *  X6_RXDeviceMap::MakeAfeVcxoI2CLink()
{
    //  Make our local Vcxo I2C Link (subdev will delete it)
    return new X6_Std_I2CLinkage(WB_AFE, X6_RX_VcoI2CLinkRegAddrs);
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::MakeAfeAdcSpiDevices() -- Build ADC Spi Devices
//---------------------------------------------------------------------------

void  X6_RXDeviceMap::MakeAfeAdcSpiDevices()
{
    //  Make one SPI device for each ADC device
    unsigned int Write_Addr = mmAFE_AdcSpiCtrlBase;
    unsigned int Read_Addr  = mmAFE_AdcSpiStatBase;

    for (unsigned int i=0; i<X6_RX_AdcSpiDevices; ++i)
        {
        AFEAdcSpi_SDev.MakeSpiDevice("ADC", Write_Addr, Read_Addr);
        Write_Addr += mmAFE_AdcSpiCtrlDelta;
        Read_Addr  += mmAFE_AdcSpiStatDelta;
        }
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::MakeAfeDacSpiDevices() -- Build DAC Spi Devices
//---------------------------------------------------------------------------

void  X6_RXDeviceMap::MakeAfeDacSpiDevices()
{
/*  ...No DAC SPI for X6_RX
    //  Make one SPI device for each DAC device
    unsigned int Write_Addr = mmAFE_DacSpiCtrlBase;
    unsigned int Read_Addr  = mmAFE_DacSpiStatBase;

    for (unsigned int i=0; i<X6_RX_DacSpiDevices; ++i)
        {
        AFEDacSpi_SDev.MakeSpiDevice("DAC", Write_Addr, Read_Addr);
        Write_Addr += mmAFE_DacSpiCtrlDelta;
        Read_Addr  += mmAFE_DacSpiStatDelta;
        }
*/
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::MakeSioPllObject() -- Build custom PLL object
//---------------------------------------------------------------------------

Si5338 *  X6_RXDeviceMap::MakeSioPllObject(I2CDevice & device)
{
    //  Make our local SioPll object (subdev will delete it)
    return new X6_RX_SioPll(this, device);
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::GetVeloInStreamCount() -- Return Velo In PIDs
//---------------------------------------------------------------------------

unsigned int  X6_RXDeviceMap::GetVeloInStreamCount()
{
    return X6RX_VeloInStreamCount;
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::GetVeloOutStreamCount() -- Return Velo Out PIDs
//---------------------------------------------------------------------------

unsigned int  X6_RXDeviceMap::GetVeloOutStreamCount()
{
    return X6RX_VeloOutStreamCount;
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::GetVitaInStreamCount() -- Return Vita In SIDs
//---------------------------------------------------------------------------

unsigned int  X6_RXDeviceMap::GetVitaInStreamCount()
{
    return X6RX_VitaInStreamCount;
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::GetVitaOutStreamCount() -- Return Vita Out SIDs
//---------------------------------------------------------------------------

unsigned int  X6_RXDeviceMap::GetVitaOutStreamCount()
{
    return X6RX_VitaOutStreamCount;
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::GetCalibrationAdcCount() -- Return ADCs needing Cal
//---------------------------------------------------------------------------

unsigned int  X6_RXDeviceMap::GetCalibrationAdcCount()
{
    return X6RX_VitaInStreamCount;
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::GetCalibrationDacCount() -- Return DACs needing Cal
//---------------------------------------------------------------------------

unsigned int  X6_RXDeviceMap::GetCalibrationDacCount()
{
    return X6RX_VitaOutStreamCount;
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::GetAdcCalibrationOverride() -- Return override object or 0
//---------------------------------------------------------------------------

ICalibrationValueOverride *  X6_RXDeviceMap::GetAdcCalibrationOverride()
{
    return 0;   // no override
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::GetDacCalibrationOverride() -- Return override object or 0
//---------------------------------------------------------------------------

ICalibrationValueOverride *  X6_RXDeviceMap::GetDacCalibrationOverride()
{
    return 0;   // no override
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::GetDacLatencyCalibrator() -- Get Latency Cal object
//---------------------------------------------------------------------------

IDacCalibrator *  X6_RXDeviceMap::GetDacLatencyCalibrator()
{
    // No dacs, no calibrator needed
    return 0;
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::MakePriClockDevice() -- Get Pri Device object
//---------------------------------------------------------------------------

IPriClockDevice *  X6_RXDeviceMap::MakePriClockDevice(PriDeviceCode /*which*/)
{
    // return same object for both directions
    return new X6StandardPriClockDevice(this, 2);
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::MakeFrameCalcDevice() -- Get Frame Calc Device object
//---------------------------------------------------------------------------

IFrameCalcDevice *  X6_RXDeviceMap::MakeFrameCalcDevice(FrameDeviceCode /*which*/)
{
    // return same object for both directions
    return new X6StandardFrameCalcDevice(this, 1);
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::GetBaseboardPtr() -- Get Baseboard Device object ptr
//---------------------------------------------------------------------------

IIsAVitaPacketStreamBoard *   X6_RXDeviceMap::GetBaseboardPtr()
{
    return Owner;
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::GetPatternModeAlignment() --
//---------------------------------------------------------------------------
//  Patterns must start on an even multiple of N (32-bit words)

unsigned int   X6_RXDeviceMap::GetPatternModeAlignment()
{
    return 8;   // standard X6 alignment of 8 words
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::GetPatternModeScaleFactor() --
//---------------------------------------------------------------------------
//  Divide 32 bit address and size by this number when writing to command packet

unsigned int   X6_RXDeviceMap::GetPatternModeScaleFactor()
{
    return 8;   // standard X6 scale factor of 8 words
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::GetPatternModeSingleMemoryFlag() --
//---------------------------------------------------------------------------
//  return true if we use a single memory for pattern mode

bool   X6_RXDeviceMap::GetPatternModeSingleMemoryFlag()
{
    return false;  // dual memory
}

//---------------------------------------------------------------------------
//  X6_RXDeviceMap::Input() -- Reference to input object
//---------------------------------------------------------------------------

X6_RXInputDevice & X6_RXDeviceMap::Input()
{
    return *FInput;
}


} // namespace Innovative


