// XU_RTDeviceMap_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2017 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <XU_RT_Mb.h>
#include <XU_RTDeviceMap_Mb.h>
#include <XUFamilyMemoryMap_Mb.h>
#include <XU_RTMemoryMap_Mb.h>
#include <FloatingPtSupport_Mb.h>
#include <SystemSupport_Mb.h>

//---------------------------------------------------------------------------
namespace Innovative
{

using namespace XU_RTMemoryMap;

//---------------------------------------------------------
//  Subdevice parameters Definition
const double  kVcxoPowerOnRate = 125.0;    // Si570 power on rate (MHz)

//  Parameters for XU Stream Configuration
const unsigned int  XU_RT_OutgoingStreams = 2;
const unsigned int  XU_RT_IncomingStreams = 2;  // actually 0, but support 2 for loopback tests

//  Parameters for SubDevices
const int  XU_RT_AdcLogicalDevices = 2;
const int  XU_RT_AdcLogicalChannelsPerDevice = 1;
const int  XU_RT_AdcChannels = XU_RT_AdcLogicalDevices*XU_RT_AdcLogicalChannelsPerDevice;

const int     XU_RT_AdcPhysicalDevices = 2;
const int     XU_RT_AdcPhysicalChannelsPerDevice = 1;

unsigned int  XU_RT_AdcSpiDeviceCount = XU_RT_AdcPhysicalDevices;
unsigned int  XU_RT_AdcCalibrationDeviceCount = XU_RT_AdcChannels;

const int  XU_RT_DacLogicalDevices = 2;
const int  XU_RT_DacLogicalChannelsPerDevice = 1;
const int  XU_RT_DacChannels = XU_RT_DacLogicalDevices*XU_RT_DacLogicalChannelsPerDevice;

const int     XU_RT_DacPhysicalDevices = 2;
const int     XU_RT_DacPhysicalChannelsPerDevice = 1;

unsigned int  XU_RT_DacSpiDeviceCount = XU_RT_DacPhysicalDevices;
unsigned int  XU_RT_DacCalibrationDeviceCount = XU_RT_DacChannels;

//  Input Info Data
const float         InputMaxRate            = 250.0e6f;
const float         InputMinRate            = 0.0e6f;
const unsigned int  InputWidth_Bits         = 14;
const          int  InputMax_mV             = 750;
const          int  InputMin_mV             = -750;
const unsigned int  InputSamplesPerWord     = 2;
const unsigned int  InputTriggerGranularity = 16;

//  Output Info Data
const float         OutputMaxRate            = 250.0e6f;
const float         OutputMinRate            = 0.0e6f;
const unsigned int  OutputWidth_Bits         = 14;
const          int  OutputMax_mV             = 750;
const          int  OutputMin_mV             = -750;
const unsigned int  OutputSamplesPerWord     = 2;
const unsigned int  OutputTriggerGranularity = 16;


const unsigned int XU_RT_Adc_One = 0x10000;    // representation of floating point 1.0
const unsigned int XU_RT_Dac_One = 0x10000;    // representation of floating point 1.0

const int XU_RT_Si570_SubAddr = 0xAC;

//---------------------------------------------------------
//  IDROM Section Information
//  ...System Section: Standard Data only
const int romXU_RT_SysInfoOffset        = 0;
const int romXU_RT_SysInfoOffset_avail  = 18;  // first 'free' word

//  ...Adc Cal Idrom: ADC coefficients
const int romXU_RT_InputCoeffOffset       = 0;  // Not used
const int romXU_RT_InputCoeffOffset_avail = 0;

//  ...Dac Cal Idrom: DAC coefficients
const int romXU_RT_OutputCoeffOffset       = 0;
const int romXU_RT_OutputCoeffOffset_avail = XU_RT_DacChannels * 2 + 1;   // XU_RT_DacChannels * 2 coefficients plus 'calibrated' word

//===========================================================================
// Class XU_RT_AnalogInInfo
//===========================================================================

class XU_RT_AnalogInInfo : public IX6AnalogIoInfo
{
public:
    XU_RT_AnalogInInfo(XU_RTDeviceMap * owner)
         :  Owner(owner), FSpan(InputMax_mV, InputMax_mV)
        {   }
    virtual ~XU_RT_AnalogInInfo()  {  }

    virtual float     MaxRate() const
        {
            return InputMaxRate;
        }
    virtual float         MinRate() const     {  return InputMinRate;  }
    virtual SpanInfo      Span() const        {  return FSpan;  }
    virtual unsigned int  Bits() const        {  return InputWidth_Bits;  }

    virtual unsigned int  SamplesPerWord() const {  return InputSamplesPerWord;  }
    virtual unsigned int  TriggerFrameGranularity() const   {  return InputTriggerGranularity;  }

private:
    XU_RTDeviceMap * Owner;
    SpanInfo   FSpan;
};


//===========================================================================
// Class XU_RT_AnalogOutInfo
//===========================================================================

class XU_RT_AnalogOutInfo : public IX6AnalogIoInfo
{
public:
    XU_RT_AnalogOutInfo(XU_RTDeviceMap * owner)
         :  Owner(owner), FSpan(OutputMax_mV, OutputMax_mV)
        {   }
    virtual ~XU_RT_AnalogOutInfo()  {  }

    virtual float     MaxRate() const
        {
            return OutputMaxRate;
        }
    virtual float         MinRate() const     {  return OutputMinRate;  }
    virtual SpanInfo      Span() const        {  return FSpan;  }
    virtual unsigned int  Bits() const        {  return OutputWidth_Bits;  }

    virtual unsigned int  SamplesPerWord() const {  return OutputSamplesPerWord;  }
    virtual unsigned int  TriggerFrameGranularity() const   {  return OutputTriggerGranularity;  }

private:
    XU_RTDeviceMap * Owner;
    SpanInfo   FSpan;
};

//=============================================================================
//  CLASS XU_RT_Pll  --  PLL timebase control
//=============================================================================
//---------------------------------------------------------------------------
//  housekeeping for class XU_RT_Pll
//---------------------------------------------------------------------------

const float  Input_ReferenceClk = 156.25e6;

//---------------------------------------------------------------------------
//  constructor for class XU_RT_Pll
//---------------------------------------------------------------------------

XU_RT_Pll::XU_RT_Pll(XU_RTDeviceMap * owner, IAddressingSpace &space, Lmk04820Registers & reg)
    : Lmk04820(space, reg, omNestedZeroDelay, tsmLogic), Owner(owner), FMasterOutIdx(0)
{
    OutputSyncMode(osmSysRefSync);
    Allow_Pll1_FeedbackAdjustment();
    SetMasterOutputSelection(DacMasterOutput());  // Use Output 0 as guideline output
    ReferenceFrequency(0, Input_ReferenceClk);    // All Refs are 156.25MHz
    ReferenceFrequency(1, Input_ReferenceClk);
    ReferenceFrequency(2, Input_ReferenceClk);
}

//---------------------------------------------------------------------------
//  destructor for class XU_RT_Pll
//---------------------------------------------------------------------------

XU_RT_Pll::~XU_RT_Pll()
{
}

//---------------------------------------------------------------------------
//  XU_RT_Pll::OnInit() -- inital configuration
//---------------------------------------------------------------------------

void  XU_RT_Pll::OnInit()
{
    ScopedTraceRef ccc(Debug, "XU_RT_Pll::OnInit()");
    //
    //  Input Configuration
    SetInputConfig(0, icBipolar);
    SetInputConfig(1, icBipolar);
    SetInputConfig(2, icBipolar);

    //
    //  Set Source Configuration
    SourceInUse(0, true);   SetSourceMapping(0, IXUClockIo::rsCarrier );    // clkin0 = carrier
    SourceInUse(1, true);   SetSourceMapping(1, IXUClockIo::rsFrontPanel);  // clkin1 = front panel
    SourceInUse(2, true);   SetSourceMapping(2, IXUClockIo::rsInternal);    // clkin2 = internal

    //  Device Powered/Enabled Setup
    //PowerDownOutputGroup(2/2, true);
    //
    //  Set Output Configuration
    OutputInUse(0,  true);    SetOutputConfig(0,  ocLvds);     // DCLK_REF - MGT Reference Clock
    OutputInUse(1,  true);    SetOutputConfig(1,  ocLvds);     // FP_SDCLK - System Reference
    OutputInUse(2,  true);    SetOutputConfig(2,  ocLvds);     // B2B_DCLK - (PClk = LineRate/40)
    OutputInUse(3,  true);    SetOutputConfig(3,  ocLvds);     // TRIG_REG_CLK
                              SetOutput_AsCopy(3);             //  - is a copy of output 2
    OutputInUse(4,  true);    SetOutputConfig(4,  ocLvds);     // RT_DEV_CLK_1B - (PClk = LineRate/40)
    OutputInUse(5,  true);    SetOutputConfig(5,  ocLvds);     // RT_SYSREF_1B  - FPGA RT1 System Reference
    OutputInUse(6,  false);   SetOutputConfig(6,  ocDisabled); // (unused)
    OutputInUse(7,  true);    SetOutputConfig(7,  ocLvds);     // RT_SYSREF_1A  - DAC1 System Reference
    OutputInUse(8,  true);    SetOutputConfig(8,  ocLvds);     // RT_DEV_CLK_0A - DAC0 uw PLL Reference
    OutputInUse(9,  true);    SetOutputConfig(9,  ocLvds);     // RT_SYSREF_0A  - DAC0 System Reference
    OutputInUse(10, false);   SetOutputConfig(10, ocDisabled); // RT_DEV_CLK_0B - (unused) (reverse polarity)
    OutputInUse(11, true);    SetOutputConfig(11, ocLvds);     // RT_SYSREF_0B  - System reference (reverse polarity) (used to register the ext sync)
    OutputInUse(12, true);    SetOutputConfig(12, ocLvds);     // RT_DEV_CLK_1A - DAC1 uw PLL Reference
    OutputInUse(13, true);    SetOutputConfig(13, ocLvds);     // RT_SDCLK_B    - FPGA RT0 System reference

    FeedbackOutput(fbm_Sysref);  // set output to use as PLL feedback in 0-delay modes

    //  Set Reduced Frequency Limits for Outputs (Preset to 3200.0 MHz)
    //  to set a limit, use a line like "OutputFrequencyLimit[6] = 100.0e6;
    //      if the physical rate limit is 100 MHz
    OutputFrequencyLimit[8]  = 600.0e6;
    OutputFrequencyLimit[12] = 600.0e6;

    //
    //  Reference
//    SetExternalReferenceControl(true);  // Use Internal Select
    ReferenceSource(refPrimary);    // Set to use primary reference
}

//---------------------------------------------------------------------------
//  XU_RT_Pll::OnReapply()
//---------------------------------------------------------------------------

void  XU_RT_Pll::OnReapply()
{
}

//---------------------------------------------------------------------------
//  XU_RT_Pll::AdcMasterOutput() - choose master if ADC is Master
//---------------------------------------------------------------------------

unsigned int  XU_RT_Pll::AdcMasterOutput()
{
    return FMasterOutIdx;
}

//---------------------------------------------------------------------------
//  XU_RT_Pll::DacMasterOutput() - choose master if Dac is Master
//---------------------------------------------------------------------------

unsigned int  XU_RT_Pll::DacMasterOutput()
{
    return FMasterOutIdx;
}

//---------------------------------------------------------------------------
//  XU_RT_Pll::IsFeedbackOutputAdc() - which group is Feedback Out in
//---------------------------------------------------------------------------

bool XU_RT_Pll::IsFeedbackOutputAdc()
{
    return false;
}

//---------------------------------------------------------------------------
//  XU_RT_Pll::SetAdcOutputRates() -- Set ADC group frequencies
//---------------------------------------------------------------------------
//  ADC/DAC clock outputs must be set in unison, at fixed ratios

void  XU_RT_Pll::SetAdcOutputRates(double /*adc_rate*/)
{
    // none
}

//---------------------------------------------------------------------------
//  XU_RT_Pll::SetDacOutputRates() -- Set DAC group frequencies
//---------------------------------------------------------------------------

void  XU_RT_Pll::SetDacOutputRates(double /*dac_rate*/)
{
    XU_Chrono_Type2::LmkClkRatesStruct & rates = Owner->Clock_SDev.LmkClkRates;

    FMasterOutIdx = rates.MasterClkIdx;

    OutputFrequency(0,   rates.MgtRefClk);   // DCLK_REF
    OutputFrequency(1,   rates.SysRefClk);   // FP_SDCLK
    OutputFrequency(2,   rates.FpgaDacClk);  // B2B_DCLK
//    OutputFrequency(3,   rates.FpgaDacClk);  // TRIG_REG_CLK copy of 2
    OutputFrequency(4,   rates.FpgaDacClk);  // RT_DEV_CLK_1B
    OutputFrequency(5,   rates.SysRefClk);   // RT_SYSREF_1B
    OutputFrequency(7,   rates.SysRefClk);   // RT_SYSREF_1A
    OutputFrequency(8,   rates.TxPllRefClk); // RT_DEV_CLK_0A
    OutputFrequency(9,   rates.SysRefClk);   // RT_SYSREF_0A
    OutputFrequency(11,  rates.SysRefClk);   // RT_SYSREF_0B
    OutputFrequency(12,  rates.TxPllRefClk); // RT_DEV_CLK_1A
    OutputFrequency(13,  rates.SysRefClk);   // RT_SDCLK_B
}

//===========================================================================
//  CLASS XU_RT_Xo  --  Programmable XO/VCXO Chip Support
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class XU_RT_Xo
//---------------------------------------------------------------------------

XU_RT_Xo::XU_RT_Xo(IRequires_I2C_RegIo & device)
    :  Lmk61E2(device)
{
    //  ADD mode controls the I2C Address used in Init()
    // SetADD_Mode(addFloating);   // uncomment to change from default floating
}

//-----------------------------------------------------------------------------
//  XU_RT_Xo::OnInit() - customizaton on open
//-----------------------------------------------------------------------------

void XU_RT_Xo::OnInit()
{
    // nothing yet
}

//-----------------------------------------------------------------------------
//  XU_RT_Xo::OnReapply() - customizaton on stream start or config change
//-----------------------------------------------------------------------------

void XU_RT_Xo::OnReapply()
{
    // nothing yet
}

//=============================================================================
//  CLASS XU_RT_TxPll  --  RT Support PLL for XU_RT
//=============================================================================
//---------------------------------------------------------------------------
//  housekeeping for class XU_RT_TxPll
//---------------------------------------------------------------------------

const float  RT_TxPll_Input_ReferenceClk = 100.0e6;

//---------------------------------------------------------------------------
//  constructor for class XU_RT_TxPll
//---------------------------------------------------------------------------

XU_RT_TxPll::XU_RT_TxPll(XU_RTDeviceMap * owner, IAddressingSpace &space, ADF4355Registers & reg)
    : ADF4355(space, reg), Owner(owner)
{
    SetMasterOutputSelection(DacMasterOutput());
    SetType(adf4356);

    ReferenceFrequency(0, RT_TxPll_Input_ReferenceClk);
}

//---------------------------------------------------------------------------
//  destructor for class XU_RT_TxPll
//---------------------------------------------------------------------------

XU_RT_TxPll::~XU_RT_TxPll()
{
}

//---------------------------------------------------------------------------
//  XU_RT_TxPll::OnInit() -- inital configuration
//---------------------------------------------------------------------------

void  XU_RT_TxPll::OnInit()
{
    ScopedTraceRef ccc(Debug, "XU_RT_TxPll::OnInit()");

    //  Set Output Configuration
    OutputInUse(0, true);   OutputEnabled(0, true);   // RT0/1_CLKA - DAC0/1 Sampling Clock
    OutputInUse(1, false);  OutputEnabled(1, false);  // unused

    //  Reference
    ReferenceSource(refPrimary); // Set to use primary reference
}

//---------------------------------------------------------------------------
//  XU_RT_TxPll::OnReapply()
//---------------------------------------------------------------------------

void  XU_RT_TxPll::OnReapply()
{
}

//---------------------------------------------------------------------------
//  XU_RT_TxPll::AdcMasterOutput() - choose master if ADC is Master
//---------------------------------------------------------------------------

unsigned int  XU_RT_TxPll::AdcMasterOutput()
{
    return 0;
}

//---------------------------------------------------------------------------
//  XU_RT_TxPll::DacMasterOutput() - choose master if Dac is Master
//---------------------------------------------------------------------------

unsigned int  XU_RT_TxPll::DacMasterOutput()
{
    return 0;
}

//---------------------------------------------------------------------------
//  XU_RT_TxPll::IsFeedbackOutputAdc() - which group is Feedback Out in
//---------------------------------------------------------------------------

bool XU_RT_TxPll::IsFeedbackOutputAdc()
{
    return false;
}

//---------------------------------------------------------------------------
//  XU_RT_TxPll::SetAdcOutputRates() -- Set ADC group frequencies
//---------------------------------------------------------------------------
//  ADC/DAC clock outputs must be set in unison, at fixed ratios

void  XU_RT_TxPll::SetAdcOutputRates(double /*adc_rate*/)
{
}

//---------------------------------------------------------------------------
//  XU_RT_TxPll::SetDacOutputRates() -- Set DAC group frequencies
//---------------------------------------------------------------------------

void  XU_RT_TxPll::SetDacOutputRates(double dac_rate)
{
    OutputFrequency(0,    dac_rate);
    //OutputFrequency(1,   dac_rate);   // copy of 0
}

//=============================================================================
//  CLASS XU_RT_RxPll  --  RT Support PLL for XU_RT
//=============================================================================
//---------------------------------------------------------------------------
//  housekeeping for class XU_RT_RxPll
//---------------------------------------------------------------------------

const float  RT_RxPll_Input_ReferenceClk = 100.0e6;

//---------------------------------------------------------------------------
//  constructor for class XU_RT_RxPll
//---------------------------------------------------------------------------

XU_RT_RxPll::XU_RT_RxPll(XU_RTDeviceMap * owner, IAddressingSpace &space, ADF4355Registers & reg)
    : ADF4355(space, reg), Owner(owner)
{
    SetMasterOutputSelection(AdcMasterOutput());
    SetType(adf4356);

    ReferenceFrequency(0, RT_RxPll_Input_ReferenceClk);
}

//---------------------------------------------------------------------------
//  destructor for class XU_RT_RxPll
//---------------------------------------------------------------------------

XU_RT_RxPll::~XU_RT_RxPll()
{
}

//---------------------------------------------------------------------------
//  XU_RT_RxPll::OnInit() -- inital configuration
//---------------------------------------------------------------------------

void  XU_RT_RxPll::OnInit()
{
    ScopedTraceRef ccc(Debug, "XU_RT_RxPll::OnInit()");

    //  Set Output Configuration
    OutputInUse(0, true);   OutputEnabled(0, true);   // RT0/1_CLKA - DAC0/1 Sampling Clock
    OutputInUse(1, false);  OutputEnabled(1, false);  // unused

    //  Reference
    ReferenceSource(refPrimary); // Set to use primary reference
}

//---------------------------------------------------------------------------
//  XU_RT_RxPll::OnReapply()
//---------------------------------------------------------------------------

void  XU_RT_RxPll::OnReapply()
{
}

//---------------------------------------------------------------------------
//  XU_RT_RxPll::AdcMasterOutput() - choose master if ADC is Master
//---------------------------------------------------------------------------

unsigned int  XU_RT_RxPll::AdcMasterOutput()
{
    return 0;
}

//---------------------------------------------------------------------------
//  XU_RT_RxPll::DacMasterOutput() - choose master if Dac is Master
//---------------------------------------------------------------------------

unsigned int  XU_RT_RxPll::DacMasterOutput()
{
    return 0;
}

//---------------------------------------------------------------------------
//  XU_RT_RxPll::IsFeedbackOutputAdc() - which group is Feedback Out in
//---------------------------------------------------------------------------

bool XU_RT_RxPll::IsFeedbackOutputAdc()
{
    return false;
}

//---------------------------------------------------------------------------
//  XU_RT_RxPll::SetAdcOutputRates() -- Set ADC group frequencies
//---------------------------------------------------------------------------
//  ADC/DAC clock outputs must be set in unison, at fixed ratios

void  XU_RT_RxPll::SetAdcOutputRates(double /*adc_rate*/)
{
}

//---------------------------------------------------------------------------
//  XU_RT_RxPll::SetDacOutputRates() -- Set DAC group frequencies
//---------------------------------------------------------------------------

void  XU_RT_RxPll::SetDacOutputRates(double dac_rate)
{
    OutputFrequency(0,    dac_rate);
    //OutputFrequency(1,   dac_rate);   // copy of 0
}


//===========================================================================
//  CLASS XU_RT_SysMonInterface -- Device Specific System Monitor Methods
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class XU_RT_SysMonInterface
//---------------------------------------------------------------------------

// linear function Vout = (slope)*T + (offset)

const double Temp_Slope  = .00625;
const double Temp_Offset = 0.424;

const unsigned int CommonFactorGroups = 2;  // Sets of three DACs used to set 2 values in 2 pairs

//---------------------------------------------------------------------------
//  constructor for class XU_RT_SysMonInterface
//---------------------------------------------------------------------------

XU_RT_SysMonInterface::XU_RT_SysMonInterface(XU_RTDeviceMap * owner)
    :  SysMonInterface(&owner->SysMon_SDev, TotalPins, CommonFactorGroups)
{
    Pin_Init();
}

//---------------------------------------------------------------------------
//  XU_RT_SysMonInterface::Pin_Init()
//---------------------------------------------------------------------------

void  XU_RT_SysMonInterface::Pin_Init()
{
    Location[pin_Io0]         = dp(0, 0) ;  ConversionMultiplier[pin_Io0]         =  1.0;
    Location[pin_Io1]         = dp(0, 1) ;  ConversionMultiplier[pin_Io1]         =  1.0;
    Location[pin_Io2]         = dp(0, 2) ;  ConversionMultiplier[pin_Io2]         =  1.0;
    Location[pin_Io3]         = dp(0, 3) ;  ConversionMultiplier[pin_Io3]         =  1.0;
    Location[pin_Io4]         = dp(0, 4) ;  ConversionMultiplier[pin_Io4]         =  1.0;
    Location[pin_Io5]         = dp(0, 5) ;  ConversionMultiplier[pin_Io5]         =  1.0;
    Location[pin_V3p3]        = dp(0, 6) ;  ConversionMultiplier[pin_V3p3]        =  2.0;

    Location[pin_Vp1p5_cnfg]  = dp(1, 0) ;  ConversionMultiplier[pin_Vp1p5_cnfg]  =  1.0;
    Location[pin_Vp1p8_cnfg]  = dp(1, 1) ;  ConversionMultiplier[pin_Vp1p8_cnfg]  =  1.0;
    Location[pin_Vp5p0]       = dp(1, 2) ;  ConversionMultiplier[pin_Vp5p0]       =  2.5;
    Location[pin_Vp1p3_dac]   = dp(1, 3) ;  ConversionMultiplier[pin_Vp1p3_dac]   =  1.0;
    Location[pin_Vp1p2]       = dp(1, 4) ;  ConversionMultiplier[pin_Vp1p2]       =  1.0;
    Location[pin_Vp1p0]       = dp(1, 5) ;  ConversionMultiplier[pin_Vp1p0]       =  1.0;
    Location[pin_VPwr]        = dp(1, 6) ;  ConversionMultiplier[pin_VPwr]        = 11.0;

    Location[pin_T_U78]       = dp(2, 0) ;  ConversionMultiplier[pin_T_U78]       =  1.0;
    Location[pin_T_U77]       = dp(2, 1) ;  ConversionMultiplier[pin_T_U77]       =  1.0;
    Location[pin_T_U76]       = dp(2, 2) ;  ConversionMultiplier[pin_T_U76]       =  1.0;
    Location[pin_T_U75]       = dp(2, 3) ;  ConversionMultiplier[pin_T_U75]       =  1.0;
    Location[pin_T_U74]       = dp(2, 4) ;  ConversionMultiplier[pin_T_U74]       =  1.0;
    Location[pin_T_U73]       = dp(2, 5) ;  ConversionMultiplier[pin_T_U73]       =  1.0;
    Location[pin_Vp1p8]       = dp(2, 6) ;  ConversionMultiplier[pin_Vp1p8]       =  1.0;

}

//---------------------------------------------------------------------------
//  XU_RT_SysMonInterface::CustomInit()
//---------------------------------------------------------------------------

void  XU_RT_SysMonInterface::CustomInit()
{    
    CommonFactorLevel[0] = 0xccc;   // Trig+, Trig- range from 2.0 to -0.5 v
    CommonFactorLevel[1] = 0x800;   // RT0, RT1 range from -1.25 to 1.25 v.

    CommonFactorScale[0] = 0.5;     // Trig+, Trig- gain from Dac to output
    CommonFactorScale[1] = 0.25;    // RT0, RT1 gain from Dac to output

    // Initialize driver Device 0
    Device->MultiIo(0).SetOutputConfig(0, XUSystemMonitor::AD5592R::ocDac_AdcCheck);  //  I/O 0
    Device->MultiIo(0).SetOutputConfig(1, XUSystemMonitor::AD5592R::ocDac_AdcCheck);  //  I/O 1
    Device->MultiIo(0).SetOutputConfig(2, XUSystemMonitor::AD5592R::ocDac_AdcCheck);  //  I/O 2
    Device->MultiIo(0).SetOutputConfig(3, XUSystemMonitor::AD5592R::ocDac_AdcCheck);  //  I/O 3
    Device->MultiIo(0).SetOutputConfig(4, XUSystemMonitor::AD5592R::ocDac_AdcCheck);  //  I/O 4
    Device->MultiIo(0).SetOutputConfig(5, XUSystemMonitor::AD5592R::ocDac_AdcCheck);  //  I/O 5
    Device->MultiIo(0).SetOutputConfig(6, XUSystemMonitor::AD5592R::ocAdc);           //  3.3V
    //  Dev 0, Output 7 auto_configured            //  BUSY_1

    // Initialize driver Device 1
    Device->MultiIo(1).SetOutputConfig(0, XUSystemMonitor::AD5592R::ocAdc);  //  VP_1.5V_CNFG
    Device->MultiIo(1).SetOutputConfig(1, XUSystemMonitor::AD5592R::ocAdc);  //  VP_1.8V_CNFG
    Device->MultiIo(1).SetOutputConfig(2, XUSystemMonitor::AD5592R::ocAdc);  //  VP_5.0V_LN
    Device->MultiIo(1).SetOutputConfig(3, XUSystemMonitor::AD5592R::ocAdc);  //  VP_1.3V_DAC
    Device->MultiIo(1).SetOutputConfig(4, XUSystemMonitor::AD5592R::ocAdc);  //  VP_1.2V
    Device->MultiIo(1).SetOutputConfig(5, XUSystemMonitor::AD5592R::ocAdc);  //  VP_1.0V
    Device->MultiIo(1).SetOutputConfig(6, XUSystemMonitor::AD5592R::ocAdc);  //  VPWR
    //  Dev 1, Output 7 auto_configured            //  BUSY_1


    // Initialize driver Device 2
    Device->MultiIo(2).SetOutputConfig(0, XUSystemMonitor::AD5592R::ocAdc);  //  U78 - High load near U1, U20 Top Temp
    Device->MultiIo(2).SetOutputConfig(1, XUSystemMonitor::AD5592R::ocAdc);  //  U77 - Low load near  J1 BB Temp
    Device->MultiIo(2).SetOutputConfig(2, XUSystemMonitor::AD5592R::ocAdc);  //  U76 - Ref PLL U58  BB Temp
    Device->MultiIo(2).SetOutputConfig(3, XUSystemMonitor::AD5592R::ocAdc);  //  U75 - RT0 HS Temp
    Device->MultiIo(2).SetOutputConfig(4, XUSystemMonitor::AD5592R::ocAdc);  //  U74 - RT1 HS Temp
    Device->MultiIo(2).SetOutputConfig(5, XUSystemMonitor::AD5592R::ocAdc);  //  U73 - FPGA U1 BB Temp
    Device->MultiIo(2).SetOutputConfig(6, XUSystemMonitor::AD5592R::ocAdc);  //  VP_1.8V
    //  Dev 2, Output 7 auto_configured            //  BUSY_1

}

//---------------------------------------------------------------------------
//  XU_RT_SysMonInterface::OpenHardware()
//---------------------------------------------------------------------------
//  Called on open, after above configuration written to chip...

void XU_RT_SysMonInterface::OpenHardware()
{
    Trace Debug;
    Debug.Log("---------------------- XU-RT System Monitor Init ----------------------");
    Init_Trig(0.0, 1.5);
    Init_RT(0.0, 0.0);
    Debug.Log("---------------------- XU-RT End of System Monitor Init ---------------");
}

//---------------------------------------------------------------------------
//  XU_RT_SysMonInterface::ReadTemperature()
//---------------------------------------------------------------------------

double  XU_RT_SysMonInterface::ReadTemperature(XU_RT_pins pin_code)
{
    double raw_voltage = ReadAdc(pin_code);

    // linear function Vout = (slope)*T + (offset)
    double T = (raw_voltage-Temp_Offset)/Temp_Slope;

    return T;

}

//===========================================================================
//  CLASS XU_RT_AdcJesdIo -- Interface to ADC JESD
//===========================================================================

class XU_RT_AdcJesdIo : public IJESD_IoDevice
{
public:
    XU_RT_AdcJesdIo( XU_RTDeviceMap * owner )
        :  Owner(owner)
            {
            }
    virtual ~XU_RT_AdcJesdIo()
            {}
    virtual bool  IsDeviceActive()
                    {  return false;  }   // we use input - NOT
    virtual double CalcJesdLineRate(double SamplingRate)
                    {  return Owner->Dac(0).CalcDacJesdLineRate(SamplingRate);  }

private:
    XU_RTDeviceMap * Owner;

};

//===========================================================================
//  CLASS XU_RT_DacJesdIo -- Interface to DAC JESD
//===========================================================================

class XU_RT_DacJesdIo : public IJESD_IoDevice
{
public:
    XU_RT_DacJesdIo( XU_RTDeviceMap * owner, unsigned int device )
        :  Owner(owner), Device(device)
            {
            }
    virtual ~XU_RT_DacJesdIo()
        {}
    virtual bool  IsDeviceActive()
                    {  return true;  }   // we use output
    virtual double CalcJesdLineRate(double SamplingRate)
                    {  return Owner->Dac(Device).CalcDacJesdLineRate(SamplingRate);  }

private:
    XU_RTDeviceMap * Owner;
    const unsigned int Device;

};

//===========================================================================
//  CLASS XU_RT_JesdPllRates -- Interface to JESD Clock
//===========================================================================

class XU_RT_JesdPllRates : public IJESD_PllRates
{
public:
    XU_RT_JesdPllRates( XU_RTDeviceMap * owner, unsigned int device )
        :  Owner(owner), Device(device)
            {
            }
    virtual ~XU_RT_JesdPllRates()
        {}

    virtual double AdcFrequencyActual()
        {  return Owner->Clock().Adc().FrequencyActual();  }
    virtual double DacFrequencyActual()
        {  return Owner->Clock().Dac().FrequencyActual();  }
    virtual double MgtRefFrequencyActual()
        {  return Owner->MgtRefActual();  }

private:
    XU_RTDeviceMap * Owner;
    const unsigned int Device;

};

//===========================================================================
//  CLASS IXU_Chrono_Type1_Callbacks -- Callback Customization Interface
//===========================================================================
//---------------------------------------------------------------------------
//  declaration of class XU_RTDeviceMap
//---------------------------------------------------------------------------

class XU_RT_ChronoCallback : public IXU_Chrono_Type2_Callbacks
{
public:
    XU_RT_ChronoCallback(XU_RTDeviceMap * device)  :  Device(device)
        {}
    virtual ~XU_RT_ChronoCallback()
        {}

    virtual void  OnChrono_SetFrequency(class XU_Chrono_Type2 * chrono, double freq);

private:

    XU_RTDeviceMap * Device;
};

//---------------------------------------------------------------------------
//  XU_RT_ChronoCallback::OnChrono_SetFrequency
//---------------------------------------------------------------------------

void  XU_RT_ChronoCallback::OnChrono_SetFrequency(XU_Chrono_Type2* chrono, double freq)
{
    // XU-RT specific code for setting system frequency
    Device->SetLmkClkRates(freq, chrono->LmkClkRates);
}

//=============================================================================
//  CLASS XU_RT_DacCalOverride  --  Cal Override for XU_RT Dacs
//=============================================================================

class XU_RT_DacCalOverride : public ICalibrationValueOverride
{
public:
    virtual unsigned int   OverrideOffset(unsigned int value) ;
    virtual unsigned int   OverrideGain(unsigned int value) ;

    virtual bool OverrideOffsetUpdate() {  return true;  }
    virtual bool OverrideGainUpdate() {  return true;  }

    virtual void WriteOffset(unsigned int ch, unsigned int value);
    virtual void WriteGain(unsigned int ch, unsigned int value);

    XU_RT_DacCalOverride(XU_RTDeviceMap *  owner) : Owner(owner)
        {  RT_Ofs[0] = 0.0;   RT_Ofs[1]= 0.0;   }
    virtual ~XU_RT_DacCalOverride()
        {}

private:
    XU_RTDeviceMap *  Owner;
    double  RT_Ofs[2];
};

//---------------------------------------------------------------------------
//  XU_RT_DacCalOverride::OverrideOffset() -- adapt for nonstandard cal
//---------------------------------------------------------------------------
//  Adapt the 'offset' value for the

unsigned int  XU_RT_DacCalOverride::OverrideOffset(unsigned int value)
{
    // trim to 16 bit value
    return value&0xFFFF;
}

//---------------------------------------------------------------------------
//  XU_RT_DacCalOverride::OverrideGain() -- adapt for nonstandard cal
//---------------------------------------------------------------------------

unsigned int  XU_RT_DacCalOverride::OverrideGain(unsigned int value)
{
    return value;
}

//---------------------------------------------------------------------------
//  XU_RT_DacCalOverride::WriteOffset() -- adapt for nonstandard cal
//---------------------------------------------------------------------------

void  XU_RT_DacCalOverride::WriteOffset(unsigned int ch, unsigned int value)
{
    short signed_value = static_cast<short>(value);

    RT_Ofs[ch] = static_cast<double>(signed_value)/1000.0;
    Owner->SysMonInterface.Init_RT(RT_Ofs[0], RT_Ofs[1]);
}

//---------------------------------------------------------------------------
//  XU_RT_DacCalOverride::WriteGain() -- adapt for nonstandard cal
//---------------------------------------------------------------------------

void  XU_RT_DacCalOverride::WriteGain(unsigned int ch, unsigned int value)
{
    double gain = static_cast<double>(value)/XU_RT_Dac_One;

    unsigned int UnityGainFSCSetting = 0x358;     // This number might change

    // IoutFs = 32mA x (ANA_FSC/1023) + 8mA
    double UnityGainFSC_mA = 32.0 * (static_cast<double>(UnityGainFSCSetting)/1023.0) + 8.0;

    double FullScaleCurrent_mA = gain * UnityGainFSC_mA;

    Owner->Dac(ch).SetFullScaleCurrent_mA(FullScaleCurrent_mA);
}

//===========================================================================
//  CLASS XU_RT_SioPll  --  Programmable SioPll/VCSioPll Chip Support
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class XU_RT_SioPll
//---------------------------------------------------------------------------

XU_RT_SioPll::XU_RT_SioPll(IRequires_I2C_RegIo & device)
    :  Lmk61E2(device)
{
    //  ADD mode controls the I2C Address used in Init()
    // SetADD_Mode(addFloating);   // uncomment to change from default floating
}

//-----------------------------------------------------------------------------
//  XU_RT_SioPll::OnInit() - customizaton on open
//-----------------------------------------------------------------------------

void XU_RT_SioPll::OnInit()
{
    Frequency(125e6);
}

//-----------------------------------------------------------------------------
//  XU_RT_SioPll::OnReapply() - customizaton on stream start or config change
//-----------------------------------------------------------------------------

void XU_RT_SioPll::OnReapply()
{
    // nothing yet
}

//===========================================================================
//  CLASS XU_RTDeviceMap  -- Register & low level support for SD module
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class XU_RTDeviceMap
//---------------------------------------------------------------------------

using namespace XUFamily_MemoryMap;
using namespace XU_RTMemoryMap;

namespace
{

XU_Chrono_Type2::MapRegisters  ChronoMapAddrs =
    {
    mmAFE_Pll_Control,    //  AfePllControl_Addr
    mmAFE_Pll_Spi,        //  AfePllSpi_Addr
    };

Lmk04820Registers  XU_RT_Lmk04820Registers =
    {
    mmAFE_Pll_Control,    //  Control_Addr;
    mmAFE_Pll_Spi,        //  SpiIntf_Addr
    };

X6_Std_I2CLinkage::MapRegisters XU_RT_VcoI2CLinkRegAddrs =
    {
    mmAFE_Vcxo_I2C,       //  AfeVcxoControl_Addr
    };

X6_Std_I2CLinkage::MapRegisters XU_RT_XoI2CLinkRegAddrs =
    {
    mmAFE_XO_I2C,         //  AfeVcxoControl_Addr
    };

ADF4355Registers  XU_RT_TX_ADF4355Registers =
    {
    mmAFE_TXPll_Control,    //  Control_Addr
    mmAFE_TXPll_Spi,        //  SpiIntf_Addr
    };

ADF4355Registers  XU_RT_RX_ADF4355Registers =
    {
    mmAFE_RXPll_Control,    //  Control_Addr
    mmAFE_RXPll_Spi,        //  SpiIntf_Addr
    };

X6_Std_I2CLinkage::MapRegisters XU_RT_SioI2CLinkRegAddrs =
    {
    mmAuroraI2C,         //  AfeVcxoControl_Addr
    };

}

//---------------------------------------------------------------------------
//  constructor for class XU_RTDeviceMap
//---------------------------------------------------------------------------

#ifdef _MSC_VER
    #pragma warning(push)
    #pragma warning(disable : 4355)
#endif

XU_RTDeviceMap::XU_RTDeviceMap(XU_RT * owner)
    :  XUBidirDeviceMap(owner),
       Owner(owner),
       Clock_SDev(this, AXI_AFE, ChronoMapAddrs),
       JESD_PHY0_SDev(this, 0),
       JESD_PHY1_SDev(this, 1),
       JESD_ADC_SDev(this),
       JESD_DAC0_SDev(this, 0),
       //JESD_DAC1_SDev(this, 1),
       ADC_SDev(this, JESD_ADC_SDev.Device()),
       DAC0_SDev(this, 0),
       //DAC1_SDev(this, 1),
       FInput(new XU_RTInputDevice(this, new XU_RT_AnalogInInfo(this))),
       FOutput(new XU_RTOutputDevice(this, new XU_RT_AnalogOutInfo(this))),
       SysMonInterface(this)
{
    //
    //  Customize Input and Output stream In Use count
    BmInInit().SetMcChannels(XU_RT_IncomingStreams);
    BmOutInit().SetMcChannels(XU_RT_OutgoingStreams);
    //
    //  SubDevice Customization
    XUBidirDeviceMap::AtConstructCustomPartTime();

}

#ifdef _MSC_VER
    #pragma warning(pop)
#endif

//---------------------------------------------------------------------------
//  destructor for class XU_RTDeviceMap
//---------------------------------------------------------------------------

XU_RTDeviceMap::~XU_RTDeviceMap()
{
    delete FInput;
    FInput = 0;
    delete FOutput;
    FOutput = 0;
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::ConstructCustomParts()
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::ConstructCustomParts()
{
    XUBidirDeviceMap::ConstructCustomParts();
    Clock_SDev.ConstructCustomParts();
    JESD_PHY0_SDev.ConstructCustomParts();
    JESD_PHY1_SDev.ConstructCustomParts();
    JESD_DAC0_SDev.ConstructCustomParts();
    //JESD_DAC1_SDev.ConstructCustomParts();
    ADC_SDev.ConstructCustomParts();
    DAC0_SDev.ConstructCustomParts();
    //DAC1_SDev.ConstructCustomParts();
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::PreOpenHardware()
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::PreOpenHardware()
{

    XUBidirDeviceMap::PreOpenHardware();
    //Clock_SDev.PreOpenHardware();
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::OpenHardware()
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::OpenHardware()
{
    XUBidirDeviceMap::OpenHardware();
    Clock_SDev.OpenHardware();
    JESD_PHY0_SDev.OpenHardware();
    JESD_PHY1_SDev.OpenHardware();
    JESD_DAC0_SDev.OpenHardware();
    //JESD_DAC1_SDev.OpenHardware();
    ADC_SDev.OpenHardware();
    DAC0_SDev.OpenHardware();
    //DAC1_SDev.OpenHardware();
    Input().OpenHardware();
    Output().OpenHardware();

    SysMonInterface.OpenHardware();
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::PreconfigureHardware() - preconfigure operations
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::PreconfigureHardware()
{
    XUBidirDeviceMap::PreconfigureHardware();
    Clock_SDev.PreconfigureHardware();
    // Initialize the FPGA MGTs
    JESD_PHY0_SDev.PreconfigureHardware();
    JESD_PHY1_SDev.PreconfigureHardware();
    JESD_DAC0_SDev.PreconfigureHardware();
    //JESD_DAC1_SDev.PreconfigureHardware();
    ADC_SDev.PreconfigureHardware();
    DAC0_SDev.PreconfigureHardware();

    Input().PreconfigureHardware();
    //
    //  Special Config for D/A Devices
    Output().PreconfigureHardware();
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::OnSingleStreamAtStartConfig()
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::OnSingleStreamAtStartConfig()
{
    XUBidirDeviceMap::OnSingleStreamAtStartConfig();
    //Clock_SDev.OnSingleStreamAtStartConfig();
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::ConfigureHardware()
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::ConfigureHardware()
{
    XUBidirDeviceMap::ConfigureHardware();
    Clock_SDev.ConfigureHardware();

    JESD_PHY0_SDev.ConfigureHardware();
    JESD_PHY1_SDev.ConfigureHardware();
    JESD_DAC0_SDev.ConfigureHardware();
//    JESD_DAC1_SDev.ConfigureHardware();
    ADC_SDev.PreconfigureHardware();
    DAC0_SDev.PreconfigureHardware();

    //ConfigureAdcs();

    ConfigureDacs();
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::ConfigureDacs()
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::ConfigureDacs()
{
    //
    //  Special Config for D/A Devices
    if (Output().ChannelPowered(0)==false && Output().ChannelPowered(1)==false)
        {
        Debug.Log("DAC Power off, config skipped.");
        return;   // no need to go on, power is off.
        }

    Output().ConfigureHardware();  // reset DACs

    Dac(0).ConfigureHardware();
    Dac(1).ConfigureHardware();
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::OnBeforeStreamStart()
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::OnBeforeStreamStart()
{
    XUBidirDeviceMap::OnBeforeStreamStart();
    //Clock_SDev.OnBeforeStreamStart();
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::OnAfterStreamStart()
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::OnAfterStreamStart()
{
    XUBidirDeviceMap::OnAfterStreamStart();
    //Clock_SDev.OnAfterStreamStart();
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::OnBeforeStreamStop()
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::OnBeforeStreamStop()
{
    XUBidirDeviceMap::OnBeforeStreamStop();
    //Clock_SDev.OnBeforeStreamStop();
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::StreamStopHardware()
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::StreamStopHardware()
{
    XUBidirDeviceMap::StreamStopHardware();
    Clock_SDev.StreamStopHardware();
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::SetLmkClkRates()
//---------------------------------------------------------------------------

void XU_RTDeviceMap::SetLmkClkRates(double sampling_rate, XU_Chrono_Type2::LmkClkRatesStruct & rates)
{
    double Dac0LineRate = Owner->Dac(0).CalcDacJesdLineRate(sampling_rate);
    double Dac1LineRate = Owner->Dac(1).CalcDacJesdLineRate(sampling_rate);

    double PclkRate;
    double MgtRefClk;
    double LmfcRate;

    if (IsAboutEqual(Dac0LineRate, 0.0, 10.0) == false)      // means channel0 is active and not in NCO only mode
        {
        PclkRate  = Dac0LineRate/40;
        MgtRefClk = Owner->JesdPhy(0).CalcMgtRefClk(Dac0LineRate);
        LmfcRate  = Owner->Dac(0).CalcLmfcRate();
        }
    else if (IsAboutEqual(Dac1LineRate, 0.0, 10.0) == false) // means channel1 is active and not in NCO only mode
        {
        PclkRate  = Dac1LineRate/40;
        MgtRefClk = Owner->JesdPhy(1).CalcMgtRefClk(Dac1LineRate);
        LmfcRate  = Owner->Dac(1).CalcLmfcRate();
        }
    else                                                   // means the enabled channels are in NCO only mode
        {
        PclkRate  = sampling_rate/16;
        MgtRefClk = sampling_rate/8;
        LmfcRate  = sampling_rate/128;
        }

    const double TxPllRefInMax = 600.0e6f;
    double TxPllRefIn = PclkRate;

    if (sampling_rate > 0.0)
        {
        while (TxPllRefIn*2.0 < TxPllRefInMax)
            TxPllRefIn = TxPllRefIn * 2.0;
        }

    rates.FpgaDacClk  = PclkRate;
    rates.MgtRefClk   = MgtRefClk;
    rates.SysRefClk   = LmfcRate;    // SysRefClk = LMFC rate
    rates.TxPllRefClk = TxPllRefIn;

    if (TxPllRefIn >= MgtRefClk)
        rates.MasterClkIdx = 8;
    else
        rates.MasterClkIdx = 0;
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::Input() -- Reference to Input object
//---------------------------------------------------------------------------

XU_RTInputDevice & XU_RTDeviceMap::Input()
{
    return *FInput;
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::Output() -- Reference to output object
//---------------------------------------------------------------------------

XU_RTOutputDevice & XU_RTDeviceMap::Output()
{
    return *FOutput;
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::Make_AfePllObject()
//---------------------------------------------------------------------------

Lmk04820 *  XU_RTDeviceMap::Make_AfePllObject()
{
    return new XU_RT_Pll(this, AXI_AFE, XU_RT_Lmk04820Registers );
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::Make_AfeVcxoI2CLink()
//---------------------------------------------------------------------------

IX6_I2C_uDevice*XU_RTDeviceMap::Make_AfeVcxoI2CLink()
{
    //  Make our local Vcxo I2C Link (subdev will delete it)
    return new X6_Std_I2CLinkage(AXI_AFE, XU_RT_VcoI2CLinkRegAddrs);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::Make_AfeVcxoObject()
//---------------------------------------------------------------------------

Si570 * XU_RTDeviceMap::Make_AfeVcxoObject(IRequires_I2C_RegIo& device)
{
    //  Make our local Vcxo object (subdev will delete it)
    Si570 * vco = new Si570(device, Si570::tcB, kVcxoPowerOnRate );
    //  Override I2C SubAddr
    vco->SetI2C_SubAddr(XU_RT_Si570_SubAddr);
    return vco;
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::Make_AfeXoI2CLink()
//---------------------------------------------------------------------------

IX6_I2C_uDevice * XU_RTDeviceMap::Make_AfeXoI2CLink()
{
    //  Make our local Vcxo I2C Link (subdev will delete it)
    return new X6_Std_I2CLinkage(AXI_AFE, XU_RT_XoI2CLinkRegAddrs);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::Make_AfeXoObject()
//---------------------------------------------------------------------------

Lmk61E2 * XU_RTDeviceMap::Make_AfeXoObject(IRequires_I2C_RegIo& device)
{
    return new XU_RT_Xo(device);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::Make_AfeTxPllObject()
//---------------------------------------------------------------------------

ADF4355 * XU_RTDeviceMap::Make_AfeTxPllObject()
{
    return new XU_RT_TxPll(this, AXI_AFE, XU_RT_TX_ADF4355Registers );
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::Make_AfeTxPllObject()
//---------------------------------------------------------------------------

ADF4355 * XU_RTDeviceMap::Make_AfeRxPllObject()
{
    return new XU_RT_RxPll(this, AXI_AFE, XU_RT_RX_ADF4355Registers );
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::Make_Xu_Chrono_CB()
//---------------------------------------------------------------------------

IXU_Chrono_Type2_Callbacks* XU_RTDeviceMap::Make_Xu_Chrono_CB()
{
    return new XU_RT_ChronoCallback(this);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::XU_Type2_ChannelEnabled()
//---------------------------------------------------------------------------

bool XU_RTDeviceMap::XU_Type2_ChannelEnabled(int ch)
{
    return Output().ChannelEnabled(ch);    // IS THIS RIGHT?
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::MakePriClockDevice() -- Get Pri Device object
//---------------------------------------------------------------------------

IPriClockDevice *  XU_RTDeviceMap::MakePriClockDevice(PriDeviceCode /*which*/)
{
    // return same object for both directions
    return new XUStandardPriClockDevice(this, 2);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::MakeFrameCalcDevice() -- Get Frame Calc Device object
//---------------------------------------------------------------------------

IFrameCalcDevice *  XU_RTDeviceMap::MakeFrameCalcDevice(FrameDeviceCode /*which*/)
{
    // return same object for both directions  (we don't need an object)
    return new XUStandardFrameCalcDevice(this, 1);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::GetAnalogDeviceInfo() -- Return devices, channels on device
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::GetAnalogDeviceInfo(FrameDeviceCode /*which*/,
                            unsigned int & devices, unsigned int & ch_per_dev)
{
    devices = XU_RT_DacLogicalDevices;   // SID devices and channels INFO
    ch_per_dev = XU_RT_DacLogicalChannelsPerDevice;
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::GetAnalogPhysicalDeviceInfo() -- Return devices, channels on PHYSICAL device
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::GetAnalogPhysicalDeviceInfo(FrameDeviceCode /*which*/,
                            unsigned int & devices, unsigned int & ch_per_dev)
{
    devices = XU_RT_DacPhysicalDevices;
    ch_per_dev = XU_RT_DacPhysicalChannelsPerDevice;
}


//---------------------------------------------------------------------------
//  XU_RTDeviceMap::SetXuStreamChannelActive() -- set XU stream ch to active
//---------------------------------------------------------------------------

void XU_RTDeviceMap::SetXuStreamChannelActive(FrameDeviceCode /*which*/, int ch, bool isActive)
{
    BmOutInit().SetMcChannelActive(ch, isActive);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::GetDacSpiDeviceCount() -- Return Dac SPI devices
//---------------------------------------------------------------------------

unsigned int  XU_RTDeviceMap::GetDacSpiDeviceCount()
{
    return XU_RT_DacSpiDeviceCount;
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::GetDacCalibrationOverride() -- Return override object or 0
//---------------------------------------------------------------------------

ICalibrationValueOverride *  XU_RTDeviceMap::GetDacCalibrationOverride()
{
    return new XU_RT_DacCalOverride(this);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::GetAdcSpiDeviceCount() --
//---------------------------------------------------------------------------

unsigned int  XU_RTDeviceMap::GetAdcSpiDeviceCount()
{
    return XU_RT_AdcSpiDeviceCount;
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::GetCalibrationAdcCount() --
//---------------------------------------------------------------------------

unsigned int  XU_RTDeviceMap::GetCalibrationAdcCount()
{
    return XU_RT_AdcCalibrationDeviceCount;
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::GetAdcCalibrationOverride() --
//---------------------------------------------------------------------------

ICalibrationValueOverride *  XU_RTDeviceMap::GetAdcCalibrationOverride()
{
    return 0;  // no override?
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::GetCalibrationDacCount() -- Return ADCs needing Cal
//---------------------------------------------------------------------------

unsigned int  XU_RTDeviceMap::GetCalibrationDacCount()
{
    return XU_RT_DacCalibrationDeviceCount;
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::PokeJESDPHY() -- write to the JESD PHY
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::PokeJESDPHY(unsigned int device, unsigned int address, unsigned int data)
{
    unsigned int phy_id = (device==0) ? 1 : 3 ;
    PokeJesdLogic(phy_id, address, data);

//    std::stringstream msg;
//    msg << std::hex << "JESD204B PHY IP(0x" << address << ", 0x" << data << ")";
//    Debug.Log(msg.str());
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::PeekJESDPHY() -- read from JESD PHY
//---------------------------------------------------------------------------

unsigned int  XU_RTDeviceMap::PeekJESDPHY(unsigned int device, unsigned int address)
{
    unsigned int phy_id = (device==0) ? 1 : 3 ;
    return PeekJesdLogic(phy_id, address);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::GetAdcJesd_IoDevice() -- Create interface object to JESD ADC
//---------------------------------------------------------------------------

IJESD_IoDevice *  XU_RTDeviceMap::GetAdcJesd_IoDevice(unsigned int /*device*/)
{
    return new XU_RT_AdcJesdIo(this);   // This is dummy anyway, don't need 2 definitions
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::GetDacJesd_IoDevice() -- Create interface object to JESD DAC
//---------------------------------------------------------------------------

IJESD_IoDevice *  XU_RTDeviceMap::GetDacJesd_IoDevice(unsigned int device)
{
    return new XU_RT_DacJesdIo(this, device);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::GetJesd_PllRatesDevice() -- Create interface to clocks for JESD
//---------------------------------------------------------------------------

IJESD_PllRates *  XU_RTDeviceMap::GetJesd_PllRatesDevice(unsigned int device)
{
    return new XU_RT_JesdPllRates(this, device);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::PokeJESDADC() -- write to the JESD ADC
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::PokeJESDADC(unsigned int /*device*/, unsigned int address, unsigned int data)
{
    PokeJesdLogic(0, address, data);
    std::stringstream msg;
    msg << std::hex << "Adc JESD204B IP(0x" << address << ", 0x" << data << ")";
    Debug.Log(msg.str());
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::PeekJESDADC() -- read from JESD ADC
//---------------------------------------------------------------------------

unsigned int  XU_RTDeviceMap::PeekJESDADC(unsigned int /*device*/, unsigned int address)
{
    return PeekJesdLogic(0, address);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::PokeJESDDAC() -- write to the JESD DAC
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::PokeJESDDAC(unsigned int device_id, unsigned int address, unsigned int data)
{
    unsigned int logic_id = (device_id==0) ? 0 : 2 ;
    PokeJesdLogic(logic_id, address, data);
    std::stringstream msg;
    msg << std::hex << "Dac JESD204B IP (Dev" << device_id << " 0x" << address << ", 0x" << data << ")";
    Debug.Log(msg.str());
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::PeekJESDDAC() -- read from JESD DAC
//---------------------------------------------------------------------------

unsigned int  XU_RTDeviceMap::PeekJESDDAC(unsigned int device_id, unsigned int address)
{
    unsigned int logic_id = (device_id==0) ? 0 : 2 ;
    return PeekJesdLogic(logic_id, address);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::SetAdcLaneConfiguration() - pre-set up ADC options
//---------------------------------------------------------------------------
//  This matches FMC-1000 setup

void XU_RTDeviceMap::SetAdcLaneConfiguration(unsigned int /*device*/)
{
    // Load LanePN_Inverted "bits"
    Adc().LanePN_Inverted[0] = 1;    Adc().LanePN_Inverted[1] = 1;
    Adc().LanePN_Inverted[2] = 1;    Adc().LanePN_Inverted[3] = 1;

    // ADC physical to logical lanes mapping
    Adc().PhysToLogical_LaneMap[0] = 0;    Adc().PhysToLogical_LaneMap[1] = 1;
    Adc().PhysToLogical_LaneMap[2] = 2;    Adc().PhysToLogical_LaneMap[3] = 3;
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::AdcSpi() -- read from ADC SPI
//---------------------------------------------------------------------------

unsigned int XU_RTDeviceMap::AdcSpi(unsigned int device, unsigned int address)
{
    return AFEAdcSpi_SDev.Spi(device, address);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::AdcSpi() -- write to ADC SPI
//---------------------------------------------------------------------------

void XU_RTDeviceMap::AdcSpi(unsigned int device, unsigned int address, unsigned int data)
{
    AFEAdcSpi_SDev.Spi(device, address, data);
}

////---------------------------------------------------------------------------
////  XU_RTDeviceMap::PulsorModeEnter() -- ADC Clock Interface
////---------------------------------------------------------------------------

//void XU_RTDeviceMap::PulsorModeEnter()
//{
//    Clock_SDev.Pll().PulsorModeEnter();
//}

////---------------------------------------------------------------------------
////  XU_RTDeviceMap::PulsorModeFire() -- ADC Clock Interface
////---------------------------------------------------------------------------

//void XU_RTDeviceMap::PulsorModeFire(unsigned int Log2PulseCnt)
//{
//    Clock_SDev.Pll().PulsorModeFire(Log2PulseCnt);
//}

////---------------------------------------------------------------------------
////  XU_RTDeviceMap::PulsorModeExit() -- ADC Clock Interface
////---------------------------------------------------------------------------

//void XU_RTDeviceMap::PulsorModeExit()
//{
//    Clock_SDev.Pll().PulsorModeExit();
//}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::AdcChannelEnabled() -- read channel enabled interface
//---------------------------------------------------------------------------

bool  XU_RTDeviceMap::AdcChannelEnabled(int ch)
{
    return Input().ChannelEnabled(ch);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::AdcChannelEnabled() -- set channel enabled interface
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::AdcChannelEnabled(int ch, bool state)
{
    Input().ChannelEnabled(ch, state);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::AdcActiveChannels() -- count enabled ADC channels
//---------------------------------------------------------------------------

unsigned int  XU_RTDeviceMap::AdcActiveChannels()
{
    return Input().ActiveChannels();
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::SetDacLaneConfiguration() - pre-set up DAC options
//---------------------------------------------------------------------------

void XU_RTDeviceMap::SetDacLaneConfiguration(unsigned int device)
{
    // Load LanePN_Inverted "bits"
    Dac(device).LanePN_Inverted[0] = 0;    Dac(device).LanePN_Inverted[1] = 0;
    Dac(device).LanePN_Inverted[2] = 0;    Dac(device).LanePN_Inverted[3] = 0;
    Dac(device).LanePN_Inverted[4] = 0;    Dac(device).LanePN_Inverted[5] = 0;
    Dac(device).LanePN_Inverted[6] = 0;    Dac(device).LanePN_Inverted[7] = 0;

    // DAC physical to logical lanes mapping
    if (device == 0)
        {
        Dac(0).PhysToLogical_LaneMap[0] = 0;    Dac(0).PhysToLogical_LaneMap[1] = 1;
        Dac(0).PhysToLogical_LaneMap[2] = 2;    Dac(0).PhysToLogical_LaneMap[3] = 3;
        Dac(0).PhysToLogical_LaneMap[4] = 4;    Dac(0).PhysToLogical_LaneMap[5] = 5;
        Dac(0).PhysToLogical_LaneMap[6] = 6;    Dac(0).PhysToLogical_LaneMap[7] = 7;
        }
    else        // device == 1
        {
        Dac(1).PhysToLogical_LaneMap[0] = 7;    Dac(1).PhysToLogical_LaneMap[1] = 6;
        Dac(1).PhysToLogical_LaneMap[2] = 5;    Dac(1).PhysToLogical_LaneMap[3] = 4;
        Dac(1).PhysToLogical_LaneMap[4] = 3;    Dac(1).PhysToLogical_LaneMap[5] = 2;
        Dac(1).PhysToLogical_LaneMap[6] = 1;    Dac(1).PhysToLogical_LaneMap[7] = 0;
        }
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::DacSpi() -- read from DAC SPI
//---------------------------------------------------------------------------

unsigned int XU_RTDeviceMap::DacSpi(unsigned int device, unsigned int address)
{
    return AFEDacSpi_SDev.Spi(device, address);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::DacSpi() -- write to DAC SPI
//---------------------------------------------------------------------------

void XU_RTDeviceMap::DacSpi(unsigned int device, unsigned int address, unsigned int data)
{
    AFEDacSpi_SDev.Spi(device, address, data);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::DacChannelEnabled() -- read channel enabled interface
//---------------------------------------------------------------------------

bool  XU_RTDeviceMap::DacChannelEnabled(int ch)
{
    return Output().ChannelEnabled(ch);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::DacChannelEnabled() -- set channel enabled interface
//---------------------------------------------------------------------------

void  XU_RTDeviceMap::DacChannelEnabled(int ch, bool state)
{
    Output().ChannelEnabled(ch, state);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::GetPatternModeAlignment() --
//---------------------------------------------------------------------------
//  Patterns must start on an even multiple of N (32-bit words)

unsigned int   XU_RTDeviceMap::GetPatternModeAlignment()
{
    return 16;   // standard XU alignment of 16 words
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::GetPatternModeScaleFactor() --
//---------------------------------------------------------------------------
//  Divide 32 bit address and size by this number when writing to command packet

unsigned int   XU_RTDeviceMap::GetPatternModeScaleFactor()
{
    return 1;   // standard XU scale factor of 1 words
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::SetSysmonDeviceConfiguration() --
//---------------------------------------------------------------------------
//  Board-specific INIT of system monitor outputs

void   XU_RTDeviceMap::SetSysmonDeviceConfiguration()
{
    SysMonInterface.CustomInit();
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::MakeXuSioI2CLink()
//---------------------------------------------------------------------------

IX6_I2C_uDevice * XU_RTDeviceMap::MakeXuSioI2CLink()
{
    //  Make our local Vcxo I2C Link (subdev will delete it)
    return new X6_Std_I2CLinkage(AXI_Periph, XU_RT_SioI2CLinkRegAddrs);
}

//---------------------------------------------------------------------------
//  XU_RTDeviceMap::MakeXuSioPllObject()
//---------------------------------------------------------------------------

Lmk61E2 * XU_RTDeviceMap::MakeXuSioPllObject(IRequires_I2C_RegIo& device)
{
    return new XU_RT_SioPll(device);
}

//===========================================================================
//  CLASS XU_RTInputDevice  -- XU_RT Input Devices
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class XU_RTInputDevice
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//  constructor for class XU_RTInputDevice
//---------------------------------------------------------------------------

XU_RTInputDevice::XU_RTInputDevice(XU_RTDeviceMap * owner, IX6AnalogIoInfo * info)
    : inherited(owner), OwnerRT(owner), FInfo(info),
      CalRomSection(AdcCalIdrom(), romXU_RT_InputCoeffOffset, XU_RT_AdcChannels,
                       XU_RT_Adc_One, AdcCal() )
{
}

//---------------------------------------------------------------------------
//  destructor for class XU_RTInputDevice
//---------------------------------------------------------------------------

XU_RTInputDevice::~XU_RTInputDevice()
{
    delete FInfo;
    FInfo = 0;
}

//---------------------------------------------------------------------------
//  XU_RTInputDevice::Info() -- Analog Information
//---------------------------------------------------------------------------

IX6AnalogIoInfo &  XU_RTInputDevice::Info()
{
    return *FInfo;
}

//---------------------------------------------------------------------------
//  XU_RTInputDevice::OpenHardware() -- on open
//---------------------------------------------------------------------------

void  XU_RTInputDevice::OpenHardware()
{
    CalRomSection.ApplyRomCoefficients();
}

//---------------------------------------------------------------------------
//  XU_RTInputDevice::PreconfigureHardware() --
//---------------------------------------------------------------------------

void  XU_RTInputDevice::PreconfigureHardware()
{
    CalRomSection.ApplyRomCoefficients();
}

//---------------------------------------------------------------------------
//  XU_RTInputDevice::ConfigureHardware() --
//---------------------------------------------------------------------------

void  XU_RTInputDevice::ConfigureHardware()
{
    OwnerRT->Debug.Log("ADC ConfigureHardware()");

    // Toggle the DacReset
//    DacCommon().DacReset(false);
//    DacCommon().DacReset(true);
//    Sleep(1);
//    DacCommon().DacReset(false);
}

//===========================================================================
//  CLASS XU_RTOutputDevice  -- XU_RT Output Devices
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class XU_RTOutputDevice
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//  constructor for class XU_RTOutputDevice
//---------------------------------------------------------------------------

XU_RTOutputDevice::XU_RTOutputDevice(XU_RTDeviceMap * owner, IX6AnalogIoInfo * info)
    : inherited(owner), OwnerRT(owner), FInfo(info),
      CalRomSection(DacCalIdrom(), romXU_RT_OutputCoeffOffset, XU_RT_DacChannels,
                       XU_RT_Dac_One, DacCal() )
{
}

//---------------------------------------------------------------------------
//  destructor for class XU_RTOutputDevice
//---------------------------------------------------------------------------

XU_RTOutputDevice::~XU_RTOutputDevice()
{
    delete FInfo;
    FInfo = 0;
}

//---------------------------------------------------------------------------
//  XU_RTOutputDevice::Info() -- Analog Information
//---------------------------------------------------------------------------

IX6AnalogIoInfo &  XU_RTOutputDevice::Info()
{
    return *FInfo;
}

//---------------------------------------------------------------------------
//  XU_RTOutputDevice::OpenHardware() -- on open
//---------------------------------------------------------------------------

void  XU_RTOutputDevice::OpenHardware()
{
    CalRomSection.ApplyRomCoefficients();
}

//---------------------------------------------------------------------------
//  XU_RTOutputDevice::PreconfigureHardware() --
//---------------------------------------------------------------------------

void  XU_RTOutputDevice::PreconfigureHardware()
{
    CalRomSection.ApplyRomCoefficients();
}

//---------------------------------------------------------------------------
//  XU_RTOutputDevice::ConfigureHardware() --
//---------------------------------------------------------------------------

void  XU_RTOutputDevice::ConfigureHardware()
{
    OwnerRT->Debug.Log("DAC ConfigureHardware()");

    // Toggle the DacReset
    DacCommon().DacReset(false);
    DacCommon().DacReset(true);
    Sleep(1);
    DacCommon().DacReset(false);
}


} // namespace Innovative


