// XU_TXDeviceMap_Mb.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2016 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <XU_TX_Mb.h>
#include <XU_TXDeviceMap_Mb.h>
#include <XUFamilyMemoryMap_Mb.h>
#include <XU_TXMemoryMap_Mb.h>
#include <FloatingPtSupport_Mb.h>
#include <SystemSupport_Mb.h>

//---------------------------------------------------------------------------
namespace Innovative
{

using namespace XU_TXMemoryMap;

//---------------------------------------------------------
//  Subdevice parameters Definition
const double  kVcxoPowerOnRate = 125.0;    // Si570 power on rate (MHz)

//  Parameters for XU Stream Configuration
const unsigned int  XU_TX_OutgoingStreams = 2;
const unsigned int  XU_TX_IncomingStreams = 2;  // actually 0, but support 2 for loopback tests

//  Parameters for SubDevices
const int  XU_TX_DacLogicalDevices = 2;
const int  XU_TX_DacLogicalChannelsPerDevice = 1;
const int  XU_TX_DacChannels = XU_TX_DacLogicalDevices*XU_TX_DacLogicalChannelsPerDevice;

const int     XU_TX_DacPhysicalDevices = 2;
const int     XU_TX_DacPhysicalChannelsPerDevice = 1;

unsigned int  XU_TX_DacSpiDeviceCount = XU_TX_DacPhysicalDevices;
unsigned int  XU_TX_DacCalibrationDeviceCount = XU_TX_DacChannels;

//  Output Info Data
const float         OutputMaxRate            = 5.0e9f;  // max "raw data" rate
const float         OutputMinRate            = 0.0e6f;
const unsigned int  OutputWidth_Bits         = 16;
const          int  OutputMax_mV             = 750;
const          int  OutputMin_mV             = -750;
const unsigned int  OutputSamplesPerWord     = 2;
const unsigned int  OutputTriggerGranularity = 16;


const unsigned int XU_TX_Dac_One = 0x10000;    // representation of floating point 1.0

const int XU_TX_Si570_SubAddr = 0xAC;

//---------------------------------------------------------
//  IDROM Section Information
//  ...System Section: Standard Data only
const int romXU_TX_SysInfoOffset        = 0;
const int romXU_TX_SysInfoOffset_avail  = 18;  // first 'free' word

//  ...Adc Cal Idrom: ADC coefficients
const int romXU_TX_InputCoeffOffset       = 0;  // Not used
const int romXU_TX_InputCoeffOffset_avail = 0;

//  ...Dac Cal Idrom: DAC coefficients
const int romXU_TX_OutputCoeffOffset       = 0;
const int romXU_TX_OutputCoeffOffset_avail = XU_TX_DacChannels * 2 + 1;   // XU_TX_DacChannels * 2 coefficients plus 'calibrated' word

//===========================================================================
// Class XU_TX_AnalogOutInfo
//===========================================================================

class XU_TX_AnalogOutInfo : public IX6AnalogIoInfo
{
public:
    XU_TX_AnalogOutInfo(XU_TXDeviceMap * owner)
         :  Owner(owner), FSpan(OutputMax_mV, OutputMax_mV)
        {   }
    virtual ~XU_TX_AnalogOutInfo()  {  }

    virtual float     MaxRate() const
        {
            return OutputMaxRate;
        }
    virtual float         MinRate() const     {  return OutputMinRate;  }
    virtual SpanInfo      Span() const        {  return FSpan;  }
    virtual unsigned int  Bits() const        {  return OutputWidth_Bits;  }

    virtual unsigned int  SamplesPerWord() const {  return OutputSamplesPerWord;  }
    virtual unsigned int  TriggerFrameGranularity() const   {  return OutputTriggerGranularity;  }

private:
    XU_TXDeviceMap * Owner;
    SpanInfo   FSpan;
};

//=============================================================================
//  CLASS XU_TX_Pll  --  PLL timebase control
//=============================================================================
//---------------------------------------------------------------------------
//  housekeeping for class XU_TX_Pll
//---------------------------------------------------------------------------

const float  Input_ReferenceClk = 156.25e6;

//---------------------------------------------------------------------------
//  constructor for class XU_TX_Pll
//---------------------------------------------------------------------------

XU_TX_Pll::XU_TX_Pll(XU_TXDeviceMap * owner, IAddressingSpace &space, Lmk04820Registers & reg)
    : Lmk04820(space, reg, omNestedZeroDelay, tsmLogic), Owner(owner), FMasterOutIdx(0)
{
    OutputSyncMode(osmSysRefSync);
    Allow_Pll1_FeedbackAdjustment();
    SetMasterOutputSelection(DacMasterOutput());  // Use Output 0 as guideline output
    ReferenceFrequency(0, Input_ReferenceClk);    // All Refs are 156.25MHz
    ReferenceFrequency(1, Input_ReferenceClk);
    ReferenceFrequency(2, Input_ReferenceClk);
}

//---------------------------------------------------------------------------
//  destructor for class XU_TX_Pll
//---------------------------------------------------------------------------

XU_TX_Pll::~XU_TX_Pll()
{
}

//---------------------------------------------------------------------------
//  XU_TX_Pll::OnInit() -- inital configuration
//---------------------------------------------------------------------------

void  XU_TX_Pll::OnInit()
{
    ScopedTraceRef ccc(Debug, "XU_TX_Pll::OnInit()");
    //
    //  Input Configuration
    SetInputConfig(0, icBipolar);
    SetInputConfig(1, icBipolar);
    SetInputConfig(2, icBipolar);

    //
    //  Set Source Configuration
    SourceInUse(0, true);   SetSourceMapping(0, IXUClockIo::rsCarrier );    // clkin0 = carrier
    SourceInUse(1, true);   SetSourceMapping(1, IXUClockIo::rsFrontPanel);  // clkin1 = front panel
    SourceInUse(2, true);   SetSourceMapping(2, IXUClockIo::rsInternal);    // clkin2 = internal

    //  Device Powered/Enabled Setup
    //PowerDownOutputGroup(2/2, true);
    //
    //  Set Output Configuration
    OutputInUse(0,  true);    SetOutputConfig(0,  ocLvds);     // DCLK_REF - MGT Reference Clock
    OutputInUse(1,  true);    SetOutputConfig(1,  ocLvds);     // FP_SDCLK - System Reference
    OutputInUse(2,  true);    SetOutputConfig(2,  ocLvds);     // B2B_DCLK - (PClk = LineRate/40)
    OutputInUse(3,  true);    SetOutputConfig(3,  ocLvds);     // TRIG_REG_CLK
                              SetOutput_AsCopy(3);             //  - is a copy of output 2
    OutputInUse(4,  true);    SetOutputConfig(4,  ocLvds);     // TX_DEV_CLK_1B - (PClk = LineRate/40)
    OutputInUse(5,  true);    SetOutputConfig(5,  ocLvds);     // TX_SYSREF_1B  - FPGA TX1 System Reference
    OutputInUse(6,  false);   SetOutputConfig(6,  ocDisabled); // (unused)
    OutputInUse(7,  true);    SetOutputConfig(7,  ocLvds);     // TX_SYSREF_1A  - DAC1 System Reference
    OutputInUse(8,  true);    SetOutputConfig(8,  ocLvds);     // TX_DEV_CLK_0A - DAC0 uw PLL Reference
    OutputInUse(9,  true);    SetOutputConfig(9,  ocLvds);     // TX_SYSREF_0A  - DAC0 System Reference
    OutputInUse(10, false);   SetOutputConfig(10, ocDisabled); // TX_DEV_CLK_0B - (unused) (reverse polarity)
    OutputInUse(11, true);    SetOutputConfig(11, ocLvds);     // TX_SYSREF_0B  - System reference (reverse polarity) (used to register the ext sync)
    OutputInUse(12, true);    SetOutputConfig(12, ocLvds);     // TX_DEV_CLK_1A - DAC1 uw PLL Reference
    OutputInUse(13, true);    SetOutputConfig(13, ocLvds);     // TX_SDCLK_B    - FPGA TX0 System reference

    FeedbackOutput(fbm_Sysref);  // set output to use as PLL feedback in 0-delay modes

    //  Set Reduced Frequency Limits for Outputs (Preset to 3200.0 MHz)
    //  to set a limit, use a line like "OutputFrequencyLimit[6] = 100.0e6;
    //      if the physical rate limit is 100 MHz
    OutputFrequencyLimit[8]  = 600.0e6;
    OutputFrequencyLimit[12] = 600.0e6;

    //
    //  Reference
//    SetExternalReferenceControl(true);  // Use Internal Select
    ReferenceSource(refPrimary);    // Set to use primary reference
}

//---------------------------------------------------------------------------
//  XU_TX_Pll::OnReapply()
//---------------------------------------------------------------------------

void  XU_TX_Pll::OnReapply()
{
}

//---------------------------------------------------------------------------
//  XU_TX_Pll::AdcMasterOutput() - choose master if ADC is Master
//---------------------------------------------------------------------------

unsigned int  XU_TX_Pll::AdcMasterOutput()
{
    return FMasterOutIdx;
}

//---------------------------------------------------------------------------
//  XU_TX_Pll::DacMasterOutput() - choose master if Dac is Master
//---------------------------------------------------------------------------

unsigned int  XU_TX_Pll::DacMasterOutput()
{
    return FMasterOutIdx;
}

//---------------------------------------------------------------------------
//  XU_TX_Pll::IsFeedbackOutputAdc() - which group is Feedback Out in
//---------------------------------------------------------------------------

bool XU_TX_Pll::IsFeedbackOutputAdc()
{
    return false;
}

//---------------------------------------------------------------------------
//  XU_TX_Pll::SetAdcOutputRates() -- Set ADC group frequencies
//---------------------------------------------------------------------------
//  ADC/DAC clock outputs must be set in unison, at fixed ratios

void  XU_TX_Pll::SetAdcOutputRates(double /*adc_rate*/)
{
    // none
}

//---------------------------------------------------------------------------
//  XU_TX_Pll::SetDacOutputRates() -- Set DAC group frequencies
//---------------------------------------------------------------------------

void  XU_TX_Pll::SetDacOutputRates(double /*dac_rate*/)
{
    XU_Chrono_Type1::LmkClkRatesStruct & rates = Owner->Clock_SDev.LmkClkRates;

    FMasterOutIdx = rates.MasterClkIdx;

    OutputFrequency(0,   rates.MgtRefClk);   // DCLK_REF
    OutputFrequency(1,   rates.SysRefClk);   // FP_SDCLK
    OutputFrequency(2,   rates.FpgaDacClk);  // B2B_DCLK
//    OutputFrequency(3,   rates.FpgaDacClk);  // TRIG_REG_CLK copy of 2
    OutputFrequency(4,   rates.FpgaDacClk);  // TX_DEV_CLK_1B
    OutputFrequency(5,   rates.SysRefClk);   // TX_SYSREF_1B
    OutputFrequency(7,   rates.SysRefClk);   // TX_SYSREF_1A
    OutputFrequency(8,   rates.TxPllRefClk); // TX_DEV_CLK_0A
    OutputFrequency(9,   rates.SysRefClk);   // TX_SYSREF_0A
    OutputFrequency(11,  rates.SysRefClk);   // TX_SYSREF_0B
    OutputFrequency(12,  rates.TxPllRefClk); // TX_DEV_CLK_1A
    OutputFrequency(13,  rates.SysRefClk);   // TX_SDCLK_B
}

//===========================================================================
//  CLASS XU_TX_Xo  --  Programmable XO/VCXO Chip Support
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class XU_TX_Xo
//---------------------------------------------------------------------------

XU_TX_Xo::XU_TX_Xo(IRequires_I2C_RegIo & device)
    :  Lmk61E2(device)
{
    //  ADD mode controls the I2C Address used in Init()
    // SetADD_Mode(addFloating);   // uncomment to change from default floating
}

//-----------------------------------------------------------------------------
//  XU_TX_Xo::OnInit() - customizaton on open
//-----------------------------------------------------------------------------

void XU_TX_Xo::OnInit()
{
    // nothing yet
}

//-----------------------------------------------------------------------------
//  XU_TX_Xo::OnReapply() - customizaton on stream start or config change
//-----------------------------------------------------------------------------

void XU_TX_Xo::OnReapply()
{
    // nothing yet
}

//=============================================================================
//  CLASS XU_TX_TxPll  --  Tx Support PLL for XU_TX
//=============================================================================
//---------------------------------------------------------------------------
//  housekeeping for class XU_TX_TxPll
//---------------------------------------------------------------------------

const float  TxPll_Input_ReferenceClk = 100.0e6;

//---------------------------------------------------------------------------
//  constructor for class XU_TX_TxPll
//---------------------------------------------------------------------------

XU_TX_TxPll::XU_TX_TxPll(XU_TXDeviceMap * owner, IAddressingSpace &space, ADF4355Registers & reg)
    : ADF4355(space, reg), Owner(owner)
{
    SetMasterOutputSelection(DacMasterOutput());
    SetType(adf4355);

    ReferenceFrequency(0, TxPll_Input_ReferenceClk);
}

//---------------------------------------------------------------------------
//  destructor for class XU_TX_TxPll
//---------------------------------------------------------------------------

XU_TX_TxPll::~XU_TX_TxPll()
{
}

//---------------------------------------------------------------------------
//  XU_TX_TxPll::OnInit() -- inital configuration
//---------------------------------------------------------------------------

void  XU_TX_TxPll::OnInit()
{
    ScopedTraceRef ccc(Debug, "XU_TX_TxPll::OnInit()");

    //  Set Output Configuration
    OutputInUse(0, true);   OutputEnabled(0, true);   // TX0/1_CLKA - DAC0/1 Sampling Clock
    OutputInUse(1, false);  OutputEnabled(1, false);  // unused

    //  Reference
    ReferenceSource(refPrimary); // Set to use primary reference
}

//---------------------------------------------------------------------------
//  XU_TX_TxPll::OnReapply()
//---------------------------------------------------------------------------

void  XU_TX_TxPll::OnReapply()
{
}

//---------------------------------------------------------------------------
//  XU_TX_TxPll::AdcMasterOutput() - choose master if ADC is Master
//---------------------------------------------------------------------------

unsigned int  XU_TX_TxPll::AdcMasterOutput()
{
    return 0;
}

//---------------------------------------------------------------------------
//  XU_TX_TxPll::DacMasterOutput() - choose master if Dac is Master
//---------------------------------------------------------------------------

unsigned int  XU_TX_TxPll::DacMasterOutput()
{
    return 0;
}

//---------------------------------------------------------------------------
//  XU_TX_TxPll::IsFeedbackOutputAdc() - which group is Feedback Out in
//---------------------------------------------------------------------------

bool XU_TX_TxPll::IsFeedbackOutputAdc()
{
    return false;
}

//---------------------------------------------------------------------------
//  XU_TX_TxPll::SetAdcOutputRates() -- Set ADC group frequencies
//---------------------------------------------------------------------------
//  ADC/DAC clock outputs must be set in unison, at fixed ratios

void  XU_TX_TxPll::SetAdcOutputRates(double /*adc_rate*/)
{
}

//---------------------------------------------------------------------------
//  XU_TX_TxPll::SetDacOutputRates() -- Set DAC group frequencies
//---------------------------------------------------------------------------

void  XU_TX_TxPll::SetDacOutputRates(double dac_rate)
{
    OutputFrequency(0,    dac_rate);
    //OutputFrequency(1,   dac_rate);   // copy of 0
}


//===========================================================================
//  CLASS XU_TX_SysMonInterface -- Device Specific System Monitor Methods
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class XU_TX_SysMonInterface
//---------------------------------------------------------------------------

// linear function Vout = (slope)*T + (offset)

const double Temp_Slope  = .00625;
const double Temp_Offset = 0.424;

const unsigned int CommonFactorGroups = 2;  // Sets of three DACs used to set 2 values in 2 pairs

//---------------------------------------------------------------------------
//  constructor for class XU_TX_SysMonInterface
//---------------------------------------------------------------------------

XU_TX_SysMonInterface::XU_TX_SysMonInterface(XU_TXDeviceMap * owner)
    :  SysMonInterface(&owner->SysMon_SDev, TotalPins, CommonFactorGroups)
{
    Pin_Init();
}

//---------------------------------------------------------------------------
//  XU_TX_SysMonInterface::Pin_Init()
//---------------------------------------------------------------------------

void  XU_TX_SysMonInterface::Pin_Init()
{
    Location[pin_Io0]         = dp(0, 0) ;  ConversionMultiplier[pin_Io0]         =  1.0;
    Location[pin_Io1]         = dp(0, 1) ;  ConversionMultiplier[pin_Io1]         =  1.0;
    Location[pin_Io2]         = dp(0, 2) ;  ConversionMultiplier[pin_Io2]         =  1.0;
    Location[pin_Io3]         = dp(0, 3) ;  ConversionMultiplier[pin_Io3]         =  1.0;
    Location[pin_Io4]         = dp(0, 4) ;  ConversionMultiplier[pin_Io4]         =  1.0;
    Location[pin_Io5]         = dp(0, 5) ;  ConversionMultiplier[pin_Io5]         =  1.0;
    Location[pin_V3p3]        = dp(0, 6) ;  ConversionMultiplier[pin_V3p3]        =  2.0;

    Location[pin_Vp1p5_cnfg]  = dp(1, 0) ;  ConversionMultiplier[pin_Vp1p5_cnfg]  =  1.0;
    Location[pin_Vp1p8_cnfg]  = dp(1, 1) ;  ConversionMultiplier[pin_Vp1p8_cnfg]  =  1.0;
    Location[pin_Vp5p0]       = dp(1, 2) ;  ConversionMultiplier[pin_Vp5p0]       =  2.5;
    Location[pin_Vp1p3_dac]   = dp(1, 3) ;  ConversionMultiplier[pin_Vp1p3_dac]   =  1.0;
    Location[pin_Vp1p2]       = dp(1, 4) ;  ConversionMultiplier[pin_Vp1p2]       =  1.0;
    Location[pin_Vp1p0]       = dp(1, 5) ;  ConversionMultiplier[pin_Vp1p0]       =  1.0;
    Location[pin_VPwr]        = dp(1, 6) ;  ConversionMultiplier[pin_VPwr]        = 11.0;

    Location[pin_T_U78]       = dp(2, 0) ;  ConversionMultiplier[pin_T_U78]       =  1.0;
    Location[pin_T_U77]       = dp(2, 1) ;  ConversionMultiplier[pin_T_U77]       =  1.0;
    Location[pin_T_U76]       = dp(2, 2) ;  ConversionMultiplier[pin_T_U76]       =  1.0;
    Location[pin_T_U75]       = dp(2, 3) ;  ConversionMultiplier[pin_T_U75]       =  1.0;
    Location[pin_T_U74]       = dp(2, 4) ;  ConversionMultiplier[pin_T_U74]       =  1.0;
    Location[pin_T_U73]       = dp(2, 5) ;  ConversionMultiplier[pin_T_U73]       =  1.0;
    Location[pin_Vp1p8]       = dp(2, 6) ;  ConversionMultiplier[pin_Vp1p8]       =  1.0;

}

//---------------------------------------------------------------------------
//  XU_TX_SysMonInterface::CustomInit()
//---------------------------------------------------------------------------

void  XU_TX_SysMonInterface::CustomInit()
{    
    CommonFactorLevel[0] = 0xccc;   // Trig+, Trig- range from 2.0 to -0.5 v
    CommonFactorLevel[1] = 0x800;   // Tx0, Tx1 range from -1.25 to 1.25 v.

    CommonFactorScale[0] = 0.5;     // Trig+, Trig- gain from Dac to output
    CommonFactorScale[1] = 0.25;    // Tx0, Tx1 gain from Dac to output

    // Initialize driver Device 0
    Device->MultiIo(0).SetOutputConfig(0, XUSystemMonitor::AD5592R::ocDac_AdcCheck);  //  I/O 0
    Device->MultiIo(0).SetOutputConfig(1, XUSystemMonitor::AD5592R::ocDac_AdcCheck);  //  I/O 1
    Device->MultiIo(0).SetOutputConfig(2, XUSystemMonitor::AD5592R::ocDac_AdcCheck);  //  I/O 2
    Device->MultiIo(0).SetOutputConfig(3, XUSystemMonitor::AD5592R::ocDac_AdcCheck);  //  I/O 3
    Device->MultiIo(0).SetOutputConfig(4, XUSystemMonitor::AD5592R::ocDac_AdcCheck);  //  I/O 4
    Device->MultiIo(0).SetOutputConfig(5, XUSystemMonitor::AD5592R::ocDac_AdcCheck);  //  I/O 5
    Device->MultiIo(0).SetOutputConfig(6, XUSystemMonitor::AD5592R::ocAdc);           //  3.3V
    //  Dev 0, Output 7 auto_configured            //  BUSY_1

    // Initialize driver Device 1
    Device->MultiIo(1).SetOutputConfig(0, XUSystemMonitor::AD5592R::ocAdc);  //  VP_1.5V_CNFG
    Device->MultiIo(1).SetOutputConfig(1, XUSystemMonitor::AD5592R::ocAdc);  //  VP_1.8V_CNFG
    Device->MultiIo(1).SetOutputConfig(2, XUSystemMonitor::AD5592R::ocAdc);  //  VP_5.0V_LN
    Device->MultiIo(1).SetOutputConfig(3, XUSystemMonitor::AD5592R::ocAdc);  //  VP_1.3V_DAC
    Device->MultiIo(1).SetOutputConfig(4, XUSystemMonitor::AD5592R::ocAdc);  //  VP_1.2V
    Device->MultiIo(1).SetOutputConfig(5, XUSystemMonitor::AD5592R::ocAdc);  //  VP_1.0V
    Device->MultiIo(1).SetOutputConfig(6, XUSystemMonitor::AD5592R::ocAdc);  //  VPWR
    //  Dev 1, Output 7 auto_configured            //  BUSY_1


    // Initialize driver Device 2
    Device->MultiIo(2).SetOutputConfig(0, XUSystemMonitor::AD5592R::ocAdc);  //  U78 - High load near U1, U20 Top Temp
    Device->MultiIo(2).SetOutputConfig(1, XUSystemMonitor::AD5592R::ocAdc);  //  U77 - Low load near  J1 BB Temp
    Device->MultiIo(2).SetOutputConfig(2, XUSystemMonitor::AD5592R::ocAdc);  //  U76 - Ref PLL U58  BB Temp
    Device->MultiIo(2).SetOutputConfig(3, XUSystemMonitor::AD5592R::ocAdc);  //  U75 - TX0 HS Temp
    Device->MultiIo(2).SetOutputConfig(4, XUSystemMonitor::AD5592R::ocAdc);  //  U74 - TX1 HS Temp
    Device->MultiIo(2).SetOutputConfig(5, XUSystemMonitor::AD5592R::ocAdc);  //  U73 - FPGA U1 BB Temp
    Device->MultiIo(2).SetOutputConfig(6, XUSystemMonitor::AD5592R::ocAdc);  //  VP_1.8V
    //  Dev 2, Output 7 auto_configured            //  BUSY_1

}

//---------------------------------------------------------------------------
//  XU_TX_SysMonInterface::OpenHardware()
//---------------------------------------------------------------------------
//  Called on open, after above configuration written to chip...

void XU_TX_SysMonInterface::OpenHardware()
{
    Trace Debug;
    Debug.Log("---------------------- XU-TX System Monitor Init ----------------------");
    Init_Trig(0.0, 1.5);
    Init_Tx(0.0, 0.0);
    Debug.Log("---------------------- XU-TX End of System Monitor Init ---------------");
}

//---------------------------------------------------------------------------
//  XU_TX_SysMonInterface::ReadTemperature()
//---------------------------------------------------------------------------

double  XU_TX_SysMonInterface::ReadTemperature(XU_TX_pins pin_code)
{
    double raw_voltage = ReadAdc(pin_code);

    // linear function Vout = (slope)*T + (offset)
    double T = (raw_voltage-Temp_Offset)/Temp_Slope;

    return T;

}

//===========================================================================
//  CLASS XU_TX_AdcJesdIo -- Interface to ADC JESD
//===========================================================================

class XU_TX_AdcJesdIo : public IJESD_IoDevice
{
public:
    XU_TX_AdcJesdIo( XU_TXDeviceMap * owner )
        :  Owner(owner)
            {
            }
    virtual ~XU_TX_AdcJesdIo()
            {}
    virtual bool  IsDeviceActive()
                    {  return false;  }   // we use input - NOT
    virtual double CalcJesdLineRate(double SamplingRate)
                    {  return Owner->Dac(0).CalcDacJesdLineRate(SamplingRate);  }

private:
    XU_TXDeviceMap * Owner;

};

//===========================================================================
//  CLASS XU_TX_DacJesdIo -- Interface to DAC JESD
//===========================================================================

class XU_TX_DacJesdIo : public IJESD_IoDevice
{
public:
    XU_TX_DacJesdIo( XU_TXDeviceMap * owner, unsigned int device )
        :  Owner(owner), Device(device)
            {
            }
    virtual ~XU_TX_DacJesdIo()
        {}
    virtual bool  IsDeviceActive()
                    {  return true;  }   // we use output
    virtual double CalcJesdLineRate(double SamplingRate)
                    {  return Owner->Dac(Device).CalcDacJesdLineRate(SamplingRate);  }

private:
    XU_TXDeviceMap * Owner;
    const unsigned int Device;

};

//===========================================================================
//  CLASS XU_TX_JesdPllRates -- Interface to JESD Clock
//===========================================================================

class XU_TX_JesdPllRates : public IJESD_PllRates
{
public:
    XU_TX_JesdPllRates( XU_TXDeviceMap * owner, unsigned int device )
        :  Owner(owner), Device(device)
            {
            }
    virtual ~XU_TX_JesdPllRates()
        {}

    virtual double AdcFrequencyActual()
        {  return Owner->Clock().Adc().FrequencyActual();  }
    virtual double DacFrequencyActual()
        {  return Owner->Clock().Dac().FrequencyActual();  }
    virtual double MgtRefFrequencyActual()
        {  return Owner->MgtRefActual();  }

private:
    XU_TXDeviceMap * Owner;
    const unsigned int Device;

};

//===========================================================================
//  CLASS IXU_Chrono_Type1_Callbacks -- Callback Customization Interface
//===========================================================================
//---------------------------------------------------------------------------
//  declaration of class XU_TXDeviceMap
//---------------------------------------------------------------------------

class XU_TX_ChronoCallback : public IXU_Chrono_Type1_Callbacks
{
public:
    XU_TX_ChronoCallback(XU_TXDeviceMap * device)  :  Device(device)
        {}
    virtual ~XU_TX_ChronoCallback()
        {}

    virtual void  OnChrono_SetFrequency(class XU_Chrono_Type1 * chrono, double freq);

private:

    XU_TXDeviceMap * Device;
};

//---------------------------------------------------------------------------
//  XU_TX_ChronoCallback::OnChrono_SetFrequency
//---------------------------------------------------------------------------

void  XU_TX_ChronoCallback::OnChrono_SetFrequency(class XU_Chrono_Type1 * chrono, double freq)
{
    // XU-TX specific code for setting system frequency
    Device->SetLmkClkRates(freq, chrono->LmkClkRates);
}

//=============================================================================
//  CLASS XU_TX_DacCalOverride  --  Cal Override for XU_TX Dacs
//=============================================================================

class XU_TX_DacCalOverride : public ICalibrationValueOverride
{
public:
    virtual unsigned int   OverrideOffset(unsigned int value) ;
    virtual unsigned int   OverrideGain(unsigned int value) ;

    virtual bool OverrideOffsetUpdate() {  return true;  }
    virtual bool OverrideGainUpdate() {  return true;  }

    virtual void WriteOffset(unsigned int ch, unsigned int value);
    virtual void WriteGain(unsigned int ch, unsigned int value);

    XU_TX_DacCalOverride(XU_TXDeviceMap *  owner) : Owner(owner)
        {  Tx_Ofs[0] = 0.0;   Tx_Ofs[1]= 0.0;   }
    virtual ~XU_TX_DacCalOverride()
        {}

private:
    XU_TXDeviceMap *  Owner;
    double  Tx_Ofs[2];
};

//---------------------------------------------------------------------------
//  XU_TX_DacCalOverride::OverrideOffset() -- adapt for nonstandard cal
//---------------------------------------------------------------------------
//  Adapt the 'offset' value for the

unsigned int  XU_TX_DacCalOverride::OverrideOffset(unsigned int value)
{
    // trim to 16 bit value
    return value&0xFFFF;
}

//---------------------------------------------------------------------------
//  XU_TX_DacCalOverride::OverrideGain() -- adapt for nonstandard cal
//---------------------------------------------------------------------------

unsigned int  XU_TX_DacCalOverride::OverrideGain(unsigned int value)
{
    return value;
}

//---------------------------------------------------------------------------
//  XU_TX_DacCalOverride::WriteOffset() -- adapt for nonstandard cal
//---------------------------------------------------------------------------

void  XU_TX_DacCalOverride::WriteOffset(unsigned int ch, unsigned int value)
{
    short signed_value = static_cast<short>(value);

    Tx_Ofs[ch] = static_cast<double>(signed_value)/1000.0;
    Owner->SysMonInterface.Init_Tx(Tx_Ofs[0], Tx_Ofs[1]);
}

//---------------------------------------------------------------------------
//  XU_TX_DacCalOverride::WriteGain() -- adapt for nonstandard cal
//---------------------------------------------------------------------------

void  XU_TX_DacCalOverride::WriteGain(unsigned int ch, unsigned int value)
{
    double gain = static_cast<double>(value)/XU_TX_Dac_One;

    unsigned int UnityGainFSCSetting = 0x358;     // This number might change

    // IoutFs = 32mA x (ANA_FSC/1023) + 8mA
    double UnityGainFSC_mA = 32.0 * (static_cast<double>(UnityGainFSCSetting)/1023.0) + 8.0;

    double FullScaleCurrent_mA = gain * UnityGainFSC_mA;

    Owner->Dac(ch).SetFullScaleCurrent_mA(FullScaleCurrent_mA);
}

//===========================================================================
//  CLASS XU_TX_SioPll  --  Programmable SioPll/VCSioPll Chip Support
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class XU_TX_SioPll
//---------------------------------------------------------------------------

XU_TX_SioPll::XU_TX_SioPll(IRequires_I2C_RegIo & device)
    :  Lmk61E2(device)
{
    //  ADD mode controls the I2C Address used in Init()
    // SetADD_Mode(addFloating);   // uncomment to change from default floating
}

//-----------------------------------------------------------------------------
//  XU_TX_SioPll::OnInit() - customizaton on open
//-----------------------------------------------------------------------------

void XU_TX_SioPll::OnInit()
{
    Frequency(125e6);
}

//-----------------------------------------------------------------------------
//  XU_TX_SioPll::OnReapply() - customizaton on stream start or config change
//-----------------------------------------------------------------------------

void XU_TX_SioPll::OnReapply()
{
    // nothing yet
}

//===========================================================================
//  CLASS XU_TXDeviceMap  -- Register & low level support for SD module
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class XU_TXDeviceMap
//---------------------------------------------------------------------------

using namespace XUFamily_MemoryMap;
using namespace XU_TXMemoryMap;

namespace
{

XU_Chrono_Type1::MapRegisters  ChronoMapAddrs =
    {
    mmAFE_Pll_Control,    //  AfePllControl_Addr
    mmAFE_Pll_Spi,        //  AfePllSpi_Addr
    };

Lmk04820Registers  XU_TX_Lmk04820Registers =
    {
    mmAFE_Pll_Control,    //  Control_Addr;
    mmAFE_Pll_Spi,        //  SpiIntf_Addr
    };

X6_Std_I2CLinkage::MapRegisters XU_TX_VcoI2CLinkRegAddrs =
    {
    mmAFE_Vcxo_I2C,       //  AfeVcxoControl_Addr
    };

X6_Std_I2CLinkage::MapRegisters XU_TX_XoI2CLinkRegAddrs =
    {
    mmAFE_XO_I2C,         //  AfeVcxoControl_Addr
    };

ADF4355Registers  XU_TX_ADF4355Registers =
    {
    mmAFE_TXPll_Control,    //  Control_Addr
    mmAFE_TXPll_Spi,        //  SpiIntf_Addr
    };

X6_Std_I2CLinkage::MapRegisters XU_TX_SioI2CLinkRegAddrs =
    {
    mmAuroraI2C,         //  AfeVcxoControl_Addr
    };

}

//---------------------------------------------------------------------------
//  constructor for class XU_TXDeviceMap
//---------------------------------------------------------------------------

#ifdef _MSC_VER
    #pragma warning(push)
    #pragma warning(disable : 4355)
#endif

XU_TXDeviceMap::XU_TXDeviceMap(XU_TX * owner)
    :  XUDacDeviceMap(owner),
       Owner(owner),
       Clock_SDev(this, AXI_AFE, ChronoMapAddrs),
       JESD_PHY0_SDev(this, 0),
       JESD_PHY1_SDev(this, 1),
       JESD_DAC0_SDev(this, 0),
       JESD_DAC1_SDev(this, 1),
       DAC0_SDev(this, 0),
       DAC1_SDev(this, 1),
       FOutput(new XU_TXOutputDevice(this, new XU_TX_AnalogOutInfo(this))),
       SysMonInterface(this)
{
    //
    //  Customize Input and Output stream In Use count
    BmInInit().SetMcChannels(XU_TX_IncomingStreams);
    BmOutInit().SetMcChannels(XU_TX_OutgoingStreams);
    //
    //  SubDevice Customization
    XUDeviceMap::AtConstructCustomPartTime();

}

#ifdef _MSC_VER
    #pragma warning(pop)
#endif

//---------------------------------------------------------------------------
//  destructor for class XU_TXDeviceMap
//---------------------------------------------------------------------------

XU_TXDeviceMap::~XU_TXDeviceMap()
{
    delete FOutput;
    FOutput = 0;
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::ConstructCustomParts()
//---------------------------------------------------------------------------

void  XU_TXDeviceMap::ConstructCustomParts()
{
    XUDacDeviceMap::ConstructCustomParts();
    Clock_SDev.ConstructCustomParts();
    JESD_PHY0_SDev.ConstructCustomParts();
    JESD_PHY1_SDev.ConstructCustomParts();
    JESD_DAC0_SDev.ConstructCustomParts();
    JESD_DAC1_SDev.ConstructCustomParts();
    DAC0_SDev.ConstructCustomParts();
    DAC1_SDev.ConstructCustomParts();
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::PreOpenHardware()
//---------------------------------------------------------------------------

void  XU_TXDeviceMap::PreOpenHardware()
{

    XUDacDeviceMap::PreOpenHardware();
    //Clock_SDev.PreOpenHardware();
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::OpenHardware()
//---------------------------------------------------------------------------

void  XU_TXDeviceMap::OpenHardware()
{
    XUDacDeviceMap::OpenHardware();
    Clock_SDev.OpenHardware();
    JESD_PHY0_SDev.OpenHardware();
    JESD_PHY1_SDev.OpenHardware();
    JESD_DAC0_SDev.OpenHardware();
    JESD_DAC1_SDev.OpenHardware();
    DAC0_SDev.OpenHardware();
    DAC1_SDev.OpenHardware();
    Output().OpenHardware();

    SysMonInterface.OpenHardware();
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::PreconfigureHardware() - preconfigure operations
//---------------------------------------------------------------------------

void  XU_TXDeviceMap::PreconfigureHardware()
{
    XUDacDeviceMap::PreconfigureHardware();
    Clock_SDev.PreconfigureHardware();
    // Initialize the FPGA MGTs
    JESD_PHY0_SDev.PreconfigureHardware();
    JESD_PHY1_SDev.PreconfigureHardware();
    JESD_DAC0_SDev.PreconfigureHardware();
    JESD_DAC1_SDev.PreconfigureHardware();
    //
    //  Special Config for D/A Devices
    Output().PreconfigureHardware();
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::OnSingleStreamAtStartConfig()
//---------------------------------------------------------------------------

void  XU_TXDeviceMap::OnSingleStreamAtStartConfig()
{
    XUDacDeviceMap::OnSingleStreamAtStartConfig();
    //Clock_SDev.OnSingleStreamAtStartConfig();
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::ConfigureHardware()
//---------------------------------------------------------------------------

void  XU_TXDeviceMap::ConfigureHardware()
{
    XUDacDeviceMap::ConfigureHardware();
    Clock_SDev.ConfigureHardware();

    JESD_PHY0_SDev.ConfigureHardware();
    JESD_PHY1_SDev.ConfigureHardware();
    JESD_DAC0_SDev.ConfigureHardware();
    JESD_DAC1_SDev.ConfigureHardware();

    ConfigureDacs();
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::ConfigureDacs()
//---------------------------------------------------------------------------

void  XU_TXDeviceMap::ConfigureDacs()
{
    //
    //  Special Config for D/A Devices
    if (Output().ChannelPowered(0)==false && Output().ChannelPowered(1)==false)
        {
        Debug.Log("DAC Power off, config skipped.");
        return;   // no need to go on, power is off.
        }

    Output().ConfigureHardware();  // reset DACs

    Dac(0).ConfigureHardware();
    Dac(1).ConfigureHardware();
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::OnBeforeStreamStart()
//---------------------------------------------------------------------------

void  XU_TXDeviceMap::OnBeforeStreamStart()
{
    XUDacDeviceMap::OnBeforeStreamStart();
    //Clock_SDev.OnBeforeStreamStart();
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::OnAfterStreamStart()
//---------------------------------------------------------------------------

void  XU_TXDeviceMap::OnAfterStreamStart()
{
    XUDacDeviceMap::OnAfterStreamStart();
    //Clock_SDev.OnAfterStreamStart();
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::OnBeforeStreamStop()
//---------------------------------------------------------------------------

void  XU_TXDeviceMap::OnBeforeStreamStop()
{
    XUDacDeviceMap::OnBeforeStreamStop();
    //Clock_SDev.OnBeforeStreamStop();
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::StreamStopHardware()
//---------------------------------------------------------------------------

void  XU_TXDeviceMap::StreamStopHardware()
{
    XUDacDeviceMap::StreamStopHardware();
    Clock_SDev.StreamStopHardware();
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::SetLmkClkRates()
//---------------------------------------------------------------------------

void XU_TXDeviceMap::SetLmkClkRates(double sampling_rate, XU_Chrono_Type1::LmkClkRatesStruct & rates)
{
    double Dac0LineRate = Owner->Dac(0).CalcDacJesdLineRate(sampling_rate);
    double Dac1LineRate = Owner->Dac(1).CalcDacJesdLineRate(sampling_rate);

    double PclkRate;
    double MgtRefClk;
    double LmfcRate;

    if (IsAboutEqual(Dac0LineRate, 0.0, 10.0) == false)      // means channel0 is active and not in NCO only mode
        {
        PclkRate  = Dac0LineRate/40;
        MgtRefClk = Owner->JesdPhy(0).CalcMgtRefClk(Dac0LineRate);
        LmfcRate  = Owner->Dac(0).CalcLmfcRate();
        }
    else if (IsAboutEqual(Dac1LineRate, 0.0, 10.0) == false) // means channel1 is active and not in NCO only mode
        {
        PclkRate  = Dac1LineRate/40;
        MgtRefClk = Owner->JesdPhy(1).CalcMgtRefClk(Dac1LineRate);
        LmfcRate  = Owner->Dac(1).CalcLmfcRate();
        }
    else                                                   // means the enabled channels are in NCO only mode
        {
        PclkRate  = sampling_rate/16;
        MgtRefClk = sampling_rate/8;
        LmfcRate  = sampling_rate/128;
        }

    const double TxPllRefInMax = 600.0e6f;
    double TxPllRefIn = PclkRate;

    if (sampling_rate > 0.0)
        {
        while (TxPllRefIn*2.0 < TxPllRefInMax)
            TxPllRefIn = TxPllRefIn * 2.0;
        }

    rates.FpgaDacClk  = PclkRate;
    rates.MgtRefClk   = MgtRefClk;
    rates.SysRefClk   = LmfcRate;    // SysRefClk = LMFC rate
    rates.TxPllRefClk = TxPllRefIn;

    if (TxPllRefIn >= MgtRefClk)
        rates.MasterClkIdx = 8;
    else
        rates.MasterClkIdx = 0;
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::Output() -- Reference to output object
//---------------------------------------------------------------------------

XU_TXOutputDevice & XU_TXDeviceMap::Output()
{
    return *FOutput;
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::Make_AfePllObject()
//---------------------------------------------------------------------------

Lmk04820 *  XU_TXDeviceMap::Make_AfePllObject()
{
    return new XU_TX_Pll(this, AXI_AFE, XU_TX_Lmk04820Registers );
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::Make_AfeVcxoI2CLink()
//---------------------------------------------------------------------------

IX6_I2C_uDevice*XU_TXDeviceMap::Make_AfeVcxoI2CLink()
{
    //  Make our local Vcxo I2C Link (subdev will delete it)
    return new X6_Std_I2CLinkage(AXI_AFE, XU_TX_VcoI2CLinkRegAddrs);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::Make_AfeVcxoObject()
//---------------------------------------------------------------------------

Si570 * XU_TXDeviceMap::Make_AfeVcxoObject(IRequires_I2C_RegIo& device)
{
    //  Make our local Vcxo object (subdev will delete it)
    Si570 * vco = new Si570(device, Si570::tcB, kVcxoPowerOnRate );
    //  Override I2C SubAddr
    vco->SetI2C_SubAddr(XU_TX_Si570_SubAddr);
    return vco;
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::Make_AfeXoI2CLink()
//---------------------------------------------------------------------------

IX6_I2C_uDevice * XU_TXDeviceMap::Make_AfeXoI2CLink()
{
    //  Make our local Vcxo I2C Link (subdev will delete it)
    return new X6_Std_I2CLinkage(AXI_AFE, XU_TX_XoI2CLinkRegAddrs);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::Make_AfeXoObject()
//---------------------------------------------------------------------------

Lmk61E2 * XU_TXDeviceMap::Make_AfeXoObject(IRequires_I2C_RegIo& device)
{
    return new XU_TX_Xo(device);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::Make_AfeTxPllObject()
//---------------------------------------------------------------------------

ADF4355 * XU_TXDeviceMap::Make_AfeTxPllObject()
{
    return new XU_TX_TxPll(this, AXI_AFE, XU_TX_ADF4355Registers );
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::Make_Xu_Chrono_CB()
//---------------------------------------------------------------------------

IXU_Chrono_Type1_Callbacks * XU_TXDeviceMap::Make_Xu_Chrono_CB()
{
    return new XU_TX_ChronoCallback(this);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::XU_Type1_ChannelEnabled()
//---------------------------------------------------------------------------

bool XU_TXDeviceMap::XU_Type1_ChannelEnabled(int ch)
{
    return Output().ChannelEnabled(ch);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::MakePriClockDevice() -- Get Pri Device object
//---------------------------------------------------------------------------

IPriClockDevice *  XU_TXDeviceMap::MakePriClockDevice(PriDeviceCode /*which*/)
{
    // return same object for both directions
    return new XUStandardPriClockDevice(this, 2);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::MakeFrameCalcDevice() -- Get Frame Calc Device object
//---------------------------------------------------------------------------

IFrameCalcDevice *  XU_TXDeviceMap::MakeFrameCalcDevice(FrameDeviceCode /*which*/)
{
    // return same object for both directions  (we don't need an object)
    return new XUStandardFrameCalcDevice(this, 1);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::GetAnalogDeviceInfo() -- Return devices, channels on device
//---------------------------------------------------------------------------

void  XU_TXDeviceMap::GetAnalogDeviceInfo(FrameDeviceCode /*which*/,
                            unsigned int & devices, unsigned int & ch_per_dev)
{
    devices = XU_TX_DacLogicalDevices;   // SID devices and channels INFO
    ch_per_dev = XU_TX_DacLogicalChannelsPerDevice;
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::GetAnalogPhysicalDeviceInfo() -- Return devices, channels on PHYSICAL device
//---------------------------------------------------------------------------

void  XU_TXDeviceMap::GetAnalogPhysicalDeviceInfo(FrameDeviceCode /*which*/,
                            unsigned int & devices, unsigned int & ch_per_dev)
{
    devices = XU_TX_DacPhysicalDevices;
    ch_per_dev = XU_TX_DacPhysicalChannelsPerDevice;
}


//---------------------------------------------------------------------------
//  XU_TXDeviceMap::SetXuStreamChannelActive() -- set XU stream ch to active
//---------------------------------------------------------------------------

void XU_TXDeviceMap::SetXuStreamChannelActive(FrameDeviceCode /*which*/, int ch, bool isActive)
{
    BmOutInit().SetMcChannelActive(ch, isActive);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::GetDacSpiDeviceCount() -- Return Dac SPI devices
//---------------------------------------------------------------------------

unsigned int  XU_TXDeviceMap::GetDacSpiDeviceCount()
{
    return XU_TX_DacSpiDeviceCount;
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::GetDacCalibrationOverride() -- Return override object or 0
//---------------------------------------------------------------------------

ICalibrationValueOverride *  XU_TXDeviceMap::GetDacCalibrationOverride()
{
    return new XU_TX_DacCalOverride(this);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::GetCalibrationDacCount() -- Return ADCs needing Cal
//---------------------------------------------------------------------------

unsigned int  XU_TXDeviceMap::GetCalibrationDacCount()
{
    return XU_TX_DacCalibrationDeviceCount;
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::PokeJESDPHY() -- write to the JESD PHY
//---------------------------------------------------------------------------

void  XU_TXDeviceMap::PokeJESDPHY(unsigned int device, unsigned int address, unsigned int data)
{
    unsigned int phy_id = (device==0) ? 1 : 3 ;
    PokeJesdLogic(phy_id, address, data);

//    std::stringstream msg;
//    msg << std::hex << "JESD204B PHY IP(0x" << address << ", 0x" << data << ")";
//    Debug.Log(msg.str());
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::PeekJESDPHY() -- read from JESD PHY
//---------------------------------------------------------------------------

unsigned int  XU_TXDeviceMap::PeekJESDPHY(unsigned int device, unsigned int address)
{
    unsigned int phy_id = (device==0) ? 1 : 3 ;
    return PeekJesdLogic(phy_id, address);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::GetAdcJesd_IoDevice() -- Create interface object to JESD ADC
//---------------------------------------------------------------------------

IJESD_IoDevice *  XU_TXDeviceMap::GetAdcJesd_IoDevice(unsigned int /*device*/)
{
    return new XU_TX_AdcJesdIo(this);   // This is dummy anyway, don't need 2 definitions
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::GetDacJesd_IoDevice() -- Create interface object to JESD DAC
//---------------------------------------------------------------------------

IJESD_IoDevice *  XU_TXDeviceMap::GetDacJesd_IoDevice(unsigned int device)
{
    return new XU_TX_DacJesdIo(this, device);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::GetJesd_PllRatesDevice() -- Create interface to clocks for JESD
//---------------------------------------------------------------------------

IJESD_PllRates *  XU_TXDeviceMap::GetJesd_PllRatesDevice(unsigned int device)
{
    return new XU_TX_JesdPllRates(this, device);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::PokeJESDDAC() -- write to the JESD DAC
//---------------------------------------------------------------------------

void  XU_TXDeviceMap::PokeJESDDAC(unsigned int device_id, unsigned int address, unsigned int data)
{
    unsigned int logic_id = (device_id==0) ? 0 : 2 ;
    PokeJesdLogic(logic_id, address, data);
    std::stringstream msg;
    msg << std::hex << "Dac JESD204B IP (Dev" << device_id << " 0x" << address << ", 0x" << data << ")";
    Debug.Log(msg.str());
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::PeekJESDDAC() -- read from JESD DAC
//---------------------------------------------------------------------------

unsigned int  XU_TXDeviceMap::PeekJESDDAC(unsigned int device_id, unsigned int address)
{
    unsigned int logic_id = (device_id==0) ? 0 : 2 ;
    return PeekJesdLogic(logic_id, address);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::SetDacLaneConfiguration() - pre-set up DAC options
//---------------------------------------------------------------------------

void XU_TXDeviceMap::SetDacLaneConfiguration(unsigned int device)
{
    // Load LanePN_Inverted "bits"
    Dac(device).LanePN_Inverted[0] = 0;    Dac(device).LanePN_Inverted[1] = 0;
    Dac(device).LanePN_Inverted[2] = 0;    Dac(device).LanePN_Inverted[3] = 0;
    Dac(device).LanePN_Inverted[4] = 0;    Dac(device).LanePN_Inverted[5] = 0;
    Dac(device).LanePN_Inverted[6] = 0;    Dac(device).LanePN_Inverted[7] = 0;

    // DAC physical to logical lanes mapping
    if (device == 0)
        {
        Dac(0).PhysToLogical_LaneMap[0] = 0;    Dac(0).PhysToLogical_LaneMap[1] = 1;
        Dac(0).PhysToLogical_LaneMap[2] = 2;    Dac(0).PhysToLogical_LaneMap[3] = 3;
        Dac(0).PhysToLogical_LaneMap[4] = 4;    Dac(0).PhysToLogical_LaneMap[5] = 5;
        Dac(0).PhysToLogical_LaneMap[6] = 6;    Dac(0).PhysToLogical_LaneMap[7] = 7;
        }
    else        // device == 1
        {
        Dac(1).PhysToLogical_LaneMap[0] = 7;    Dac(1).PhysToLogical_LaneMap[1] = 6;
        Dac(1).PhysToLogical_LaneMap[2] = 5;    Dac(1).PhysToLogical_LaneMap[3] = 4;
        Dac(1).PhysToLogical_LaneMap[4] = 3;    Dac(1).PhysToLogical_LaneMap[5] = 2;
        Dac(1).PhysToLogical_LaneMap[6] = 1;    Dac(1).PhysToLogical_LaneMap[7] = 0;
        }
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::DacSpi() -- read from DAC SPI
//---------------------------------------------------------------------------

unsigned int XU_TXDeviceMap::DacSpi(unsigned int device, unsigned int address)
{
    return AFEDacSpi_SDev.Spi(device, address);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::DacSpi() -- write to DAC SPI
//---------------------------------------------------------------------------

void XU_TXDeviceMap::DacSpi(unsigned int device, unsigned int address, unsigned int data)
{
    AFEDacSpi_SDev.Spi(device, address, data);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::DacChannelEnabled() -- read channel enabled interface
//---------------------------------------------------------------------------

bool  XU_TXDeviceMap::DacChannelEnabled(int ch)
{
    return Output().ChannelEnabled(ch);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::DacChannelEnabled() -- set channel enabled interface
//---------------------------------------------------------------------------

void  XU_TXDeviceMap::DacChannelEnabled(int ch, bool state)
{
    Output().ChannelEnabled(ch, state);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::GetPatternModeAlignment() --
//---------------------------------------------------------------------------
//  Patterns must start on an even multiple of N (32-bit words)

unsigned int   XU_TXDeviceMap::GetPatternModeAlignment()
{
    return 16;   // standard XU alignment of 16 words
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::GetPatternModeScaleFactor() --
//---------------------------------------------------------------------------
//  Divide 32 bit address and size by this number when writing to command packet

unsigned int   XU_TXDeviceMap::GetPatternModeScaleFactor()
{
    return 1;   // standard XU scale factor of 1 words
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::SetSysmonDeviceConfiguration() --
//---------------------------------------------------------------------------
//  Board-specific INIT of system monitor outputs

void   XU_TXDeviceMap::SetSysmonDeviceConfiguration()
{
    SysMonInterface.CustomInit();
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::MakeXuSioI2CLink()
//---------------------------------------------------------------------------

IX6_I2C_uDevice * XU_TXDeviceMap::MakeXuSioI2CLink()
{
    //  Make our local Vcxo I2C Link (subdev will delete it)
    return new X6_Std_I2CLinkage(AXI_Periph, XU_TX_SioI2CLinkRegAddrs);
}

//---------------------------------------------------------------------------
//  XU_TXDeviceMap::MakeXuSioPllObject()
//---------------------------------------------------------------------------

Lmk61E2 * XU_TXDeviceMap::MakeXuSioPllObject(IRequires_I2C_RegIo& device)
{
    return new XU_TX_SioPll(device);
}

//===========================================================================
//  CLASS XU_TXOutputDevice  -- XU_TX Output Devices
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class XU_TXOutputDevice
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//  constructor for class XU_TXOutputDevice
//---------------------------------------------------------------------------

XU_TXOutputDevice::XU_TXOutputDevice(XU_TXDeviceMap * owner, IX6AnalogIoInfo * info)
    : inherited(owner), OwnerTX(owner), FInfo(info),
      CalRomSection(DacCalIdrom(), romXU_TX_OutputCoeffOffset, XU_TX_DacChannels,
                       XU_TX_Dac_One, DacCal() )
{
}

//---------------------------------------------------------------------------
//  destructor for class XU_TXOutputDevice
//---------------------------------------------------------------------------

XU_TXOutputDevice::~XU_TXOutputDevice()
{
    delete FInfo;
    FInfo = 0;
}

//---------------------------------------------------------------------------
//  XU_TXOutputDevice::Info() -- Analog Information
//---------------------------------------------------------------------------

IX6AnalogIoInfo &  XU_TXOutputDevice::Info()
{
    return *FInfo;
}

//---------------------------------------------------------------------------
//  XU_TXOutputDevice::OpenHardware() -- on open
//---------------------------------------------------------------------------

void  XU_TXOutputDevice::OpenHardware()
{
    CalRomSection.ApplyRomCoefficients();
}

//---------------------------------------------------------------------------
//  XU_TXOutputDevice::PreconfigureHardware() --
//---------------------------------------------------------------------------

void  XU_TXOutputDevice::PreconfigureHardware()
{
    CalRomSection.ApplyRomCoefficients();
}

//---------------------------------------------------------------------------
//  XU_TXOutputDevice::ConfigureHardware() --
//---------------------------------------------------------------------------

void  XU_TXOutputDevice::ConfigureHardware()
{
    OwnerTX->Debug.Log("DAC ConfigureHardware()");

    // Toggle the DacReset
    DacCommon().DacReset(false);
    DacCommon().DacReset(true);
    Sleep(1);
    DacCommon().DacReset(false);
}


} // namespace Innovative


