//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2013 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include "GpDsp_DrcBoard_Gsps_Impl.h"
#include <ctime>
#include <iomanip>

namespace   Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//===========================================================================
//  CLASS X6_GSPS_Drc  -- Implementation of DRC module for X6-250M
//===========================================================================
//---------------------------------------------------------------------------
//  FACTORY FUNCTION MakeGspsModule(Drc_Provider * provider)
//---------------------------------------------------------------------------

IDrc_Module *  MakeModule(class Drc_Provider * provider)
{
    return new X6_GSPS_Drc(provider);
}


IDrc_Module *  MakeModule(class Duc_Provider * /*provider*/)
{
    return 0;
}

//===========================================================================
//  CLASS X6_GSPS_Drc  -- Implementation of DRC module for X6-250M
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class X6_GSPS_Drc
//---------------------------------------------------------------------------

X6_GSPS_Drc::X6_GSPS_Drc(Drc_Provider * provider)
    :  Provider(provider), Opened(false),
       GpDsp_Ex(Module), GpRegs(GpDsp_Ex)
{

}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::AttachSettings() -- Change INI setting defaults
//---------------------------------------------------------------------------

void X6_GSPS_Drc::AttachSettings(Innovative::IniSaver& /*saver*/)
{
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::PassDefaults() -- Change INI setting defaults
//---------------------------------------------------------------------------
//  Override library defaults by filling in your own here.
//    NOTE: you need to have a board for these to be applied

void X6_GSPS_Drc::PassDefaults(Drc_ProviderSettings& defs)
{
    
    defs.SampleRate = 1800.0;     // MHz
    defs.DacSampleRate = 1800.0;  // MHz

    defs.EnableDateTimeLogging = false;
    defs.PlotEnable = true;
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::OpenHardware() -- Open Hardware & set up callbacks
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::OpenHardware()
{
    Module.RawClockDevice().Pll().SetPhaseDetectorFrequency(Settings().PhaseDetectorFrequency);
    //
    //  Configure Module Event Handlers
    Module.OnAfterStreamStop.SetEvent(this, &X6_GSPS_Drc::HandleAfterStreamStop);
    Module.OnAfterStreamStop.Synchronize();
    //
    //  Alerts
    HookAlerts();

    // Insure BM size is a multiple of four MB
    const int Meg = 1024 * 1024;
    const int BmSize = std::max(Settings().BusmasterSize/4, 1) * 4;
    Module.IncomingBusMasterSize(BmSize * Meg);
    Module.OutgoingBusMasterSize(1 * Meg);
    Module.Target(Personality().Target);
    //
    //  Open Device
    try
        {
        Module.Open();

        std::stringstream msg;
        msg << "Bus master size: " << BmSize << " MB";
        Log(msg.str());
        }
    catch(Innovative::MalibuException & e)
        {
        Log("Module Device Open Failure:");
        Log(e.what());
        return;
        }
    catch(...)
        {
        Log("Module Device Open Failure: Unknown Exception");
        Opened = false;
        return;
        }

    Module.Reset();
    Opened = true;

}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::CloseHardware() -- Close board
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::CloseHardware()
{
    if (!Opened)
        return;

    Module.Close();
    Opened = false;
}

//---------------------------------------------------------------------------
//  ValidateStreamParameters() -- check stream parameters
//---------------------------------------------------------------------------

bool  X6_GSPS_Drc::ValidateStreamParameters()
{
    if (Settings().SampleRate*1.e6 > Module.Input().Info().MaxRate())
        {
        Log("Sample rate too high.");
        return false;
        }

    return true;
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::ConnectToStream() -- associate stream and board
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::ConnectToStream(VitaPacketStream & stream)
{
    stream.ConnectTo(&Module);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::BeforePreconfigure() -- Setup for Preconfigure stage
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::BeforePreconfigure()
{
    //
    //  Channel Enables
    //  NOTE:  in the end, we will enable all channels...
    Module.Input().ChannelDisableAll();
    for (unsigned int i = 0; i < Module.Input().Channels(); ++i)
        {
            Module.Input().ChannelEnabled(i, true);
        }

    // Route ext clock source
    IX6ClockIo::IIClockSelect cks[] = { IX6ClockIo::cslFrontPanel, IX6ClockIo::cslP16 };
    Module.Clock().ExternalClkSelect(cks[Settings().ExtClockSrcSelection]);

    // Route reference.
    IX6ClockIo::IIReferenceSource ref[] = { IX6ClockIo::rsExternal, IX6ClockIo::rsInternal };
    Module.Clock().Reference(ref[Settings().ReferenceClockSource]);
    Module.Clock().ReferenceFrequency(Settings().ReferenceRate * 1e6);

    // Route clock
    IX6ClockIo::IIClockSource src[] = { IX6ClockIo::csExternal, IX6ClockIo::csInternal };
    Module.Clock().Source(src[Settings().SampleClockSource]);
    Module.Clock().Frequency(Settings().SampleRate * 1e6);

    // Readback Frequency
    double freq_actual = Module.Clock().FrequencyActual();
    {
    std::stringstream msg;
    msg << std::setprecision(9) << "Actual PLL Frequency: " << freq_actual ;
    Log(msg.str());
    }

    //  This isn't needed because we will turn on all channels
//    int ActiveChannels = Module().Input().ActiveChannels();
//    if (!ActiveChannels)
//        {
//        Log("Error: Must enable at least one channel");
//        UI->AfterStreamAutoStop();
//        return;
//        }
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::BeforeStreaming() -- Setup for Streaming stage
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::BeforeStreaming()
{
    //
    //  Velocia Packet Size
    SetInputPacketDataSize(Settings().PacketSize);
   
    //
    //  Input Test Generator Setup
    SetTestConfiguration(false, 0);

    // Set Decimation Factor
    Module.Input().Decimation(0);

    // Frame Triggering and other Trigger Config
    Module.Input().Trigger().FramedMode((Settings().Framed)? true : false);
    Module.Input().Trigger().Edge((Settings().EdgeTrigger)? true : false);
    Module.Input().Trigger().FrameSize(Settings().FrameSize);

    // Route External Trigger source
    IX6IoDevice::AfeExtSyncOptions syncsel[] = { IX6IoDevice::essFrontPanel, IX6IoDevice::essP16 };
    Module.Input().Trigger().ExternalSyncSource( syncsel[ Settings().ExtTriggerSrcSelection ] );

    std::vector<float> Delays;
    std::vector<float> Widths;

    //  ...add to module configuration
    SetPulseTriggerConfiguration(false, 0, Delays, Widths);

    //  Alert Config
    ConfigureAlerts(Settings().AlertEnable);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::isDdcAvailable(GpProject_DDC &GpRegs) --
//---------------------------------------------------------------------------

bool  X6_GSPS_Drc::isDdcAvailable()
{
    if (GpRegs.ProjectInfo.Enabled(GspDspProjectInfo::DDC))
        return true;
    else
        return false;
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::isDucAvailable(GpProject_DDC &GpRegs) --
//---------------------------------------------------------------------------

bool  X6_GSPS_Drc::isDucAvailable()
{
    if (GpRegs.ProjectInfo.Enabled(GspDspProjectInfo::DUC))
        return true;
    else
        return false;
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::isFFTAvailable(GpProject_DDC &GpRegs) --
//---------------------------------------------------------------------------

bool  X6_GSPS_Drc::isFFTAvailable()
{
    if (GpRegs.ProjectInfo.Enabled(GspDspProjectInfo::Spectrum))
        return true;
    else
        return false;
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::SetSoftwareTrigger() -- Set software trigger
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::SetSoftwareTrigger(bool state)
{
   SetInputSoftwareTrigger(state);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::SetExternalTrigger() -- Set software trigger
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::SetExternalTrigger(bool state)
{
    Module.Input().Trigger().External(state);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::BoardCount() -- get count of boards in system
//---------------------------------------------------------------------------

unsigned int  X6_GSPS_Drc::BoardCount()
{
   return Module.BoardCount();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::WordsFromSamples() -- convert samples to words
//---------------------------------------------------------------------------

ii64  X6_GSPS_Drc::WordsFromSamples(ii64 samples)
{
   ii64 SamplesPerWord = Module.Input().Info().SamplesPerWord();
   return samples / SamplesPerWord;
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::Temperature() --  Query current module temperature
//---------------------------------------------------------------------------

int X6_GSPS_Drc::Temperature()
{
    return Module.Thermal().LogicTemperature();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::PllLocked() --  Query module PLL Lock Status
//---------------------------------------------------------------------------

bool  X6_GSPS_Drc::PllLocked()
{
    return Module.Clock().Locked();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::Input_Bits() --  Query module sample size
//---------------------------------------------------------------------------

int  X6_GSPS_Drc::Input_Bits()
{
    return Module.Input().Info().Bits();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::Input_Unipolar() --  Query sample polarity
//---------------------------------------------------------------------------

bool  X6_GSPS_Drc::Input_Unipolar()
{
    return Module.Input().Info().Span().IsUnipolar();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::Input_FullSpan() --  Query analog span in mV
//---------------------------------------------------------------------------

int  X6_GSPS_Drc::Input_FullSpan()
{
    return Module.Input().Info().Span().Delta();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::Input_ActiveChannels() --  Query active channel count
//---------------------------------------------------------------------------

unsigned int  X6_GSPS_Drc::Input_ActiveChannels()
{
    return Module.Input().ActiveChannels();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::Input_ActiveDevices() --  Query active devices count
//---------------------------------------------------------------------------

unsigned int  X6_GSPS_Drc::Input_ActiveDevices()
{
    return Module.Input().ActiveChannels();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::VeloDataSize() --  Query data size of velo packet
//---------------------------------------------------------------------------

unsigned int  X6_GSPS_Drc::VeloDataSize()
{
    return Module.Velo().VeloDataSize(0);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::AdcClockFrequencyActual() --  Query clock frequency
//---------------------------------------------------------------------------

double  X6_GSPS_Drc::AdcClockFrequencyActual()
{
    return Module.Clock().FrequencyActual();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::TimeStamp() --
//---------------------------------------------------------------------------
void X6_GSPS_Drc::TimeStamp(bool utc, bool extPps, int tsf_mode, time_t now)
{
    IVitaTimestamp::TimestampFsModes tsfMode = static_cast<IVitaTimestamp::TimestampFsModes>(tsf_mode);

    if (utc)
        {
        IVitaTimestamp::TimestampPpsModes pps = extPps ? IVitaTimestamp::tsPpsMode : IVitaTimestamp::tsInternalTimer;
        Module.VitaIn_TS().TimestampCtrl(pps, IVitaTimestamp::tsmUtc, tsfMode);
        std::stringstream txt;
        txt << "Current time : " << static_cast<unsigned int>(time(0)) << " \n";
        Module.VitaIn_TS().LoadTimestamp(static_cast<unsigned int>(now));       

        txt << "Loaded timestamp : " <<static_cast<unsigned int>(now) << " \n";
        txt << "Loaded timestamp : " << "0x" << std::hex <<static_cast<unsigned int>(now) << "\n";

        Log(txt.str());
        }
    else
        {
        Module.VitaIn_TS().TimestampIsMode(IVitaTimestamp::tsmElapsed);
        Module.VitaIn_TS().TimestampFsMode(tsfMode);
        Module.VitaIn_TS().LoadTimestamp(0);
        }
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::Tag() --
//---------------------------------------------------------------------------

string X6_GSPS_Drc::Tag()
{
    std::stringstream msg;
    msg << "Module[" << Personality().Ordinal << "]";

    return msg.str();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::Make_Ficls() -- Dump clock info
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::Make_Ficls(FiclGroup & group, IFicl_UserInterface * mgr)
{
    std::string help_command("cr ?system cr ?X6-GSPS");
    Make_Ficls_Impl(Module, group, mgr, help_command);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::Make_FftChannels() --
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::Make_FftChannels(FftChannelsProxy & group)
{
    Make_FftChannels_Impl(Module, GpRegs, group);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::Make_DdcChannels() --
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::Make_DdcChannels(DdcChannelsProxy & group)
{
    string str;

 #ifdef LINUX
    str = "../../Config/ddc_2gsps_Fs1800MHz_BW10MHz.ini";
 #else
    str = "..\\..\\Config\\ddc_2gsps_Fs1800MHz_BW10MHz.ini";
 #endif

    vector<DdcChannelInfo> defs;
    // 8
    defs.push_back( DdcChannelInfo(0, 70.0f, 5.0f, false, false, str) );
    defs.push_back( DdcChannelInfo(0, 70.0f, 5.0f, false, false, str) );
    defs.push_back( DdcChannelInfo(0, 70.0f, 5.0f, false, false, str) );
    defs.push_back( DdcChannelInfo(0, 70.0f, 5.0f, false, false, str) );
    defs.push_back( DdcChannelInfo(0, 70.0f, 5.0f, false, false, str) );
    defs.push_back( DdcChannelInfo(0, 70.0f, 5.0f, false, false, str) );
    defs.push_back( DdcChannelInfo(0, 70.0f, 5.0f, false, false, str) );
    defs.push_back( DdcChannelInfo(0, 70.0f, 5.0f, false, false, str) );

    Make_DdcChannels_Impl(GpRegs, group, defs);

//   Make_DdcChannels_Impl(GpRegs, group);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::Make_PowerMeter() --
//---------------------------------------------------------------------------

void X6_GSPS_Drc::Make_PowerMeter(PowerMeterProxy &group)
{
    Make_PowerMeter_Impl(GpRegs, group);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::Make_OverflowDetector() --
//---------------------------------------------------------------------------

void X6_GSPS_Drc::Make_OverflowDetector(OverflowProxy &group)
{
    Make_OverflowDetector_Impl(GpRegs, group);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::Make_ToneGenerator() --
//---------------------------------------------------------------------------
void X6_GSPS_Drc::Make_ToneGenerator(ToneGeneratorProxy &group)
{
    vector<ToneGeneratorDataType> defs;
    defs.push_back(ToneGeneratorDataType(71.0f, 0.0f, 1.0f));

    Make_ToneGenerator_Impl(GpRegs, group, defs);
//    Make_ToneGenerator_Impl(GpRegs, group);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::NumberOfDDC() --
//---------------------------------------------------------------------------

unsigned int X6_GSPS_Drc::DdcCoresCount()
{
    if(IsOpen() && isDdcAvailable())
    {
        int group_size = GpRegs.DdcM.ChannelCountPerCore() == 0 ? 1 : GpRegs.DdcM.ChannelCountPerCore();
        size_t total_cores = GpRegs.DdcM.CoreCount() * group_size;
        return total_cores;
    }
    else
        return 0;
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::NumberOfDUC() --
//---------------------------------------------------------------------------

unsigned int X6_GSPS_Drc::DucCoresCount()
{
    if(IsOpen() && isDucAvailable())
        return GpRegs.DucM.CoreCount();
    else
        return 0;
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::NumberOfFFT() --
//---------------------------------------------------------------------------

unsigned int X6_GSPS_Drc::FftCoresCount()
{
    if(IsOpen() && isFFTAvailable())
         return GpRegs.SpectralFftM.CoreCount();
    else
        return 0;
}


//---------------------------------------------------------------------------
//  X6_GSPS_Drc::SampleWidth() --
//---------------------------------------------------------------------------

unsigned int X6_GSPS_Drc::SampleWidth() const
{
	return GpRegs.DdcM.SampleWidth();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::IsIQEnabled() --
//---------------------------------------------------------------------------

bool X6_GSPS_Drc::IsIQEnabled() const
{
	return GpRegs.DdcM.IsIQEnabled();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::IsUnsigned() --
//---------------------------------------------------------------------------

bool X6_GSPS_Drc::IsUnsigned() const
{
        return GpRegs.DdcM.IsUnsigned();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::SoftwareAlert() --  Send software alert
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::SoftwareAlert(unsigned int code)
{
    Module.Alerts().SoftwareAlert(code);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::LogicVersion() --
//---------------------------------------------------------------------------

string X6_GSPS_Drc::LogicVersion()
{
    stringstream logic;

    const char *hw_type[] = {"No Board",
                       "X6-RX",
                       "X6-400M",
                       "X6-GSPS",
                       "X6-250M"};

    const char *fpga_type[] = {"LX550T",
                         "LX240T",
                         "SX315T",
                         "SX475T"};

	char RevNumber = 'A' + static_cast<char>(Module.Info().FpgaChipType());
	const char *hw_rev = "Rev";

    logic  << std::hex << string(hw_type[Module.Info().PciLogicType()]) << "."
            << hw_rev << RevNumber << "."
            << string(fpga_type[Module.Info().FpgaChipType()]);

    return logic.str();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::CoreVersion() --
//---------------------------------------------------------------------------

pair<string, string> X6_GSPS_Drc::CoreVersion()
{
    pair<string, string> result;
    stringstream logic;

    MemoryBitManipulator ver(GpRegs.ProjectInfo.Version());
    MemoryBitManipulator sver(GpRegs.ProjectInfo.SubVersion());

    logic     << hex << std::setfill('0') << setw(2)
              << ver.Field(24, 8) << "_" << setw(3) << ver.Field(12, 12)
              << "_" << setw(3) << ver.Field(0, 12);

    result.first = logic.str();

    logic.str("");

    logic   << "v"
            << sver.Field(8, 8) << "." << sver.Field(0, 8)
            << "." << sver.Field(24, 8) <<"." << sver.Field(16, 8);

    result.second = logic.str();

    return result;
}


//---------------------------------------------------------------------------
//  X6_GSPS_Drc::LogLogicVersion() --  Dump logic version info
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::LogLogicVersion()
{
    std::stringstream msg;
    msg << std::hex << "Logic Version: " << Module.Info().FpgaLogicVersion()
        << ", Hdw Variant: " << Module.Info().FpgaHardwareVariant()
        << ", Revision: " << Module.Info().PciLogicRevision()
        << ", Subrevision: " << Module.Info().FpgaLogicSubrevision();
    Log(msg.str());

    std::stringstream msg2;
    msg2 << std::hex << "Board Family: " << Module.Info().PciLogicFamily()
        << ", Type: " << Module.Info().PciLogicType()
        << ", Board Revision: " << Module.Info().PciLogicPcb()
        << ", Chip: " << Module.Info().FpgaChipType();
    Log(msg2.str());

    msg.str("");
    msg << "PCI Express Lanes: " << Module.Debug()->LaneCount();
    if (Module.Debug()->IsGen2Capable())
        msg << " Gen 2";
    else
        msg << " Gen 1";
    Log(msg.str());
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::LogClockInfo() --  Dump clock info
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::LogClockInfo()
{
    Log("Clock output data");
    for (unsigned int i=0; i<Module.RawClockDevice().Pll().Outputs(); i++)
        {
        std::stringstream msg;
        msg << std::hex << "Out # " << i << ": "
            << " Divider: "   << Module.RawClockDevice().Pll().OutputDivider(i)
            << ", Out Freq: " << Module.RawClockDevice().Pll().OutputFrequency(i)
            << ", Actual: "   << Module.RawClockDevice().Pll().OutputFrequencyActual(i) ;
        Log(msg.str());
        }
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::PokeLogic() --
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::PokeLogic(int offset, unsigned int value)
{
    Module.PokeLogic(offset,value);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::PeekLogic() --
//---------------------------------------------------------------------------

unsigned int  X6_GSPS_Drc::PeekLogic(int offset)
{
    return Module.PeekLogic(offset);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::EnableCalibration() --
//---------------------------------------------------------------------------

void X6_GSPS_Drc::EnableCalibration()
{
    bool calibrated = Module.Input().Cal().Calibrated();

    if(!calibrated)
    {
         Module.Input().Cal().Calibrated(true);
         Module.IdRom().AdcCal().StoreToRom();
    }
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::HandleAfterStreamStop() --  Post stream termination event
//---------------------------------------------------------------------------
//  Hardware only commands here please!

void X6_GSPS_Drc::HandleAfterStreamStop(OpenWire::NotifyEvent & /*Event*/)
{
    // Disable external triggering initially
    SetSoftwareTrigger(false);
    SetExternalTrigger(false);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::HookAlerts() --  Hook Alerts
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::HookAlerts()
{
    Module.Alerts().OnTimeStampRolloverAlert.SetEvent(this, &X6_GSPS_Drc::HandleTimestampRolloverAlert);
    Module.Alerts().OnSoftwareAlert.SetEvent(this, &X6_GSPS_Drc::HandleSoftwareAlert);
    Module.Alerts().OnWarningTemperature.SetEvent(this, &X6_GSPS_Drc::HandleWarningTempAlert);
    Module.Alerts().OnInputOverflow.SetEvent(this, &X6_GSPS_Drc::HandleInputFifoOverrunAlert);
    Module.Alerts().OnTrigger.SetEvent(this, &X6_GSPS_Drc::HandleTriggerAlert);
    Module.Alerts().OnInputOverrange.SetEvent(this, &X6_GSPS_Drc::HandleInputOverrangeAlert);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::ConfigureAlerts() --  Configure Alerts
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::ConfigureAlerts(std::vector<char> & AlertEnable)
{
    enum IUsesX6Alerts::AlertType Alert[] = {
        IUsesX6Alerts::alertTimeStampRollover, IUsesX6Alerts::alertSoftware,
        IUsesX6Alerts::alertWarningTemperature,
        IUsesX6Alerts::alertInputOverflow,
        IUsesX6Alerts::alertTrigger, IUsesX6Alerts::alertInputOverrange };


    for (unsigned int i = 0; i < AlertEnable.size(); ++i)
        Module.Alerts().AlertEnable(Alert[i], AlertEnable[i] ? true : false);

}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::SetInputSoftwareTrigger() --  Set or Clear Software Trigger
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::SetInputSoftwareTrigger(bool state)
{
    Module.Input().SoftwareTrigger(state);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::SetInputPacketDataSize() --  Set Velocia packet data size
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::SetInputPacketDataSize( unsigned int data_size )
{
    // Load all velo stream sizes to the our data size
    Module.Velo().VeloDataSize(0, data_size);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::SetTestConfiguration() --  Configure Test Setup
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::SetTestConfiguration( bool enable, int mode_idx )
{
    // Optionally enable ramp generator
    Module.Input().TestModeEnabled(enable, mode_idx);

    //  Do special test config too
    //Module.Input().AdcDeviceTestMode(AdcTestEnable,
    //                    static_cast<X6_250MInputDevice::AdcDevTestModes>(AdcTestMode) );
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::SetPulseTriggerConfiguration() --  Configure Pulse Trigger
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::SetPulseTriggerConfiguration(bool enable, float period,
                                const PulseSettingArray & Delays, const PulseSettingArray & Widths )
{
    // Optionally enable ramp generator
    Module.Input().Pulse().Reset();
    Module.Input().Pulse().Enabled(enable);

    for (unsigned int i=0; i<Delays.size(); i++)
        {
        Module.Input().Pulse().AddEvent(static_cast<unsigned int>(period),
                                        static_cast<unsigned int>(Delays[i]),
                                        static_cast<unsigned int>(Widths[i]));
        }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  Alert Handlers
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//------------------------------------------------------------------------------
//  X6_GSPS_Drc::HandleTimestampRolloverAlert() --
//------------------------------------------------------------------------------

void  X6_GSPS_Drc::HandleTimestampRolloverAlert(Innovative::AlertSignalEvent & event)
{
    std::stringstream msg;
    msg << "Time stamp rollover 0x" << std::hex << event.Argument
        << " after " << std::dec << Elapsed(event.TimeStamp);
    Log(msg.str());
}

//------------------------------------------------------------------------------
//  X6_GSPS_Drc::HandleSoftwareAlert() --
//------------------------------------------------------------------------------

void  X6_GSPS_Drc::HandleSoftwareAlert(Innovative::AlertSignalEvent & event)
{
    std::stringstream msg;
    msg << "Software alert 0x" << std::hex << event.Argument
        << " after " << std::dec << Elapsed(event.TimeStamp);
    Log(msg.str());
}

//------------------------------------------------------------------------------
//  X6_GSPS_Drc::HandleWarningTempAlert() --
//------------------------------------------------------------------------------

void  X6_GSPS_Drc::HandleWarningTempAlert(Innovative::AlertSignalEvent & event)
{
    // Clear warning condition (reading clears)
    //Module.Thermal().LogicWarningTemperature();

    std::stringstream msg;
    msg << "Temp warning alert 0x" << std::hex << event.Argument
        << " after " << std::dec << Elapsed(event.TimeStamp);

    Log(msg.str());
}

//------------------------------------------------------------------------------
//  X6_GSPS_Drc::HandleInputFifoOverrunAlert() --
//------------------------------------------------------------------------------

void  X6_GSPS_Drc::HandleInputFifoOverrunAlert(Innovative::AlertSignalEvent & event)
{
    std::stringstream msg;
    msg << "Vita Framer Overflow 0x" << std::hex << event.Argument
        << " after " << std::dec << Elapsed(event.TimeStamp);
    Log(msg.str());


    //
    //  Alert Acknowledgement - some alerts can recur so frequently as to potentially
    //    overload the system with warnings.  Therefore, these alerts require a request
    //    to clear the lock that prevents a second alert be made at the application level.
    //
    //  The line below is how the event is acked. Even clearing it automatically here risks
    //    a cascade of alerts - ideally this would be cleared by the app elsewhere...
    //
    //  Module.Alerts().AlertClear(IUsesX6Alerts::alertInputOverflow);
}

//------------------------------------------------------------------------------
//  X6_GSPS_Drc::HandleTriggerAlert() --
//------------------------------------------------------------------------------

void  X6_GSPS_Drc::HandleTriggerAlert(Innovative::AlertSignalEvent & event)
{
    std::string triggerType;
    switch (event.Argument & 0x3)
        {
        case 0:  triggerType = "? ";  break;
        case 1:  triggerType = "Input ";  break;
        case 2:  triggerType = "Output ";  break;
        case 3:  triggerType = "Input and Output ";  break;
        }
    std::stringstream msg;
    msg << "Trigger 0x" << std::hex << event.Argument
        << " Type: " <<  triggerType
        << " after " << std::dec << Elapsed(event.TimeStamp);
    Log(msg.str());
}

//------------------------------------------------------------------------------
//  X6_GSPS_Drc::HandleInputOverrangeAlert() --
//------------------------------------------------------------------------------

void  X6_GSPS_Drc::HandleInputOverrangeAlert(Innovative::AlertSignalEvent & event)
{
    std::stringstream msg;
    msg << "Input overrange 0x " << std::hex << event.Argument
        << " after " << std::dec << Elapsed(event.TimeStamp);
    Log(msg.str());
    //
    //  Alert Acknowledgement - some alerts can recur so frequently as to potentially
    //    overload the system with warnings.  Therefore, these alerts require a request
    //    to clear the lock that prevents a second alert be made at the application level.
    //
    //  The line below is how the event is acked. Even clearing it automatically here risks
    //    a cascade of alerts - ideally this would be cleared by the app elsewhere...
    //
    //  Module.Alerts().AlertClear(IUsesX6Alerts::alertInputOverrange);
}

//------------------------------------------------------------------------------
//  X6_GSPS_Drc::Elapsed() --  Display timestamp as elapsed MCLKs
//------------------------------------------------------------------------------

std::string X6_GSPS_Drc::Elapsed(size_t timestamp)
{
    stringstream msg;
    msg << timestamp << " master clocks";
    return msg.str();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::Log() --  Log message thunked to owner
//---------------------------------------------------------------------------

void  X6_GSPS_Drc::Log(const std::string & msg)
{
    ProcessStatusEvent e(msg);
    OnLog.Execute(e);
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::FpgaLogicVersion() --
//---------------------------------------------------------------------------

std::string X6_GSPS_Drc::FpgaLogicVersion()
{
    stringstream ss;
    ss << std::hex << Module.Info().FpgaLogicVersion();
    return  ss.str();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::FpgaHardwareVariant() --
//---------------------------------------------------------------------------

short X6_GSPS_Drc::FpgaHardwareVariant()
{
    return  Module.Info().FpgaHardwareVariant();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::PciLogicRevision() --
//---------------------------------------------------------------------------

short X6_GSPS_Drc::PciLogicRevision()
{
    return  Module.Info().PciLogicRevision();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::FpgaLogicSubrevision() --
//---------------------------------------------------------------------------

short X6_GSPS_Drc::FpgaLogicSubrevision()
{
    return  Module.Info().FpgaLogicSubrevision();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::PciLogicFamily() --
//---------------------------------------------------------------------------

std::string X6_GSPS_Drc::PciLogicFamily()
{
    return  Module.FamilyName(Module.Info().PciLogicFamily());
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::PciLogicType() --
//---------------------------------------------------------------------------

std::string X6_GSPS_Drc::PciLogicType()
{
    return  Module.BoardName(Module.Info().PciLogicType());
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::PciLogicPcb() --
//---------------------------------------------------------------------------

std::string X6_GSPS_Drc::PciLogicPcb()
{
    char RevNumber = 'A' + static_cast<char>(Module.Info().PciLogicPcb());
    stringstream ss;
    ss << "Rev" << RevNumber;
    return ss.str();
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::FpgaName() --
//---------------------------------------------------------------------------

std::string X6_GSPS_Drc::FpgaName()
{
    return Module.FpgaName(Module.Info().FpgaChipType());
}

//---------------------------------------------------------------------------
//  X6_GSPS_Drc::PCIExpressLanes() --
//---------------------------------------------------------------------------

std::string X6_GSPS_Drc::PCIExpressLanes()
{
    std::stringstream msg;
        msg << Module.Debug()->LaneCount();
        if (Module.Debug()->IsGen2Capable())
            msg << " Gen 2";
        else
            msg << " Gen 1";
    return msg.str();
}


#ifdef __CLR_VER
#pragma managed(pop)
#endif

}  // namespace Innovative


