#ifndef GPDSP_DRCBOARD_GSPS_IMPL_H
#define GPDSP_DRCBOARD_GSPS_IMPL_H


#include <DrcProvider_Gp.h>

#include <VitaPacketStream_Mb.h>
#include <Ficl/FiclSystem_Mb.h>
// ...Board Specific
#include <X6_GSPS_Mb.h>

#include <GpProject_DDC_Gp.h>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//===========================================================================
//  CLASS X6_GSPS_Drc  -- Implementation of DRC module for X6-250M
//===========================================================================

class X6_GSPS_Drc : public IDrc_MakerModule
{
public:
    X6_GSPS_Drc(Drc_Provider * provider);
    virtual ~X6_GSPS_Drc()  {  }
    //
    //  Operations
    virtual void  OpenHardware();
    virtual void  CloseHardware();
    //  ...Open Time
    virtual void  ConnectToStream(VitaPacketStream & stream);
    //  ...Stream Config Time
    virtual bool  ValidateStreamParameters();
    virtual void  BeforePreconfigure();
    virtual void  BeforeStreaming();
    //
    //  Trigger Manipulation
    virtual void  SetSoftwareTrigger(bool state);
    virtual void  SetExternalTrigger(bool state);
    //
    //  Information (Pre Open)
    virtual unsigned int BoardCount();
    virtual bool         IsOpen() {  return Opened;  }
    virtual string       Tag();
    //
    //  Information (Post Open)
    virtual ii64        WordsFromSamples(ii64 samples);
    virtual int         Temperature();
    virtual bool        PllLocked();
    virtual int         Input_Bits();
    virtual bool        Input_Unipolar();
    virtual int         Input_FullSpan();
    virtual unsigned int  SampleWidth() const;
    virtual bool		IsIQEnabled() const;
    virtual bool		IsUnsigned() const;

    virtual unsigned int   Input_ActiveChannels();
    virtual unsigned int   Input_ActiveDevices();

    //
    virtual bool           isDdcAvailable();
    virtual bool           isDucAvailable();
    virtual bool           isFFTAvailable();


    virtual unsigned int   DdcCoresCount();
    virtual unsigned int   DucCoresCount();
    virtual unsigned int   FftCoresCount();

    virtual unsigned int   VeloDataSize();
    virtual double         AdcClockFrequencyActual();
    virtual void           TimeStamp(bool utc, bool extPps, int tsf_mode, time_t t);
    virtual void           AttachSettings(Innovative::IniSaver& saver);
    virtual void           PassDefaults(Drc_ProviderSettings& defs);

    virtual void          EnableCalibration();

    string LogicVersion();
    pair<string,string> CoreVersion();

    //
    //  Debug
    virtual void  Make_Ficls(FiclGroup & group, IFicl_UserInterface * mgr);
    virtual void  Make_DdcChannels(DdcChannelsProxy & group);
	
    virtual void  Make_DucChannels(DucChannelsProxy & /*group*/)
    {}

    virtual void  Make_PowerMeter(PowerMeterProxy & group);
    virtual void  Make_OverflowDetector(OverflowProxy &group);
    virtual void  Make_ToneGenerator(ToneGeneratorProxy &group);
    virtual void  Make_FftChannels(FftChannelsProxy & group);
    virtual void  SoftwareAlert(unsigned int code);
    virtual void  LogLogicVersion();
    virtual void  LogClockInfo();

    virtual unsigned int  PeekLogic(int offset);
    virtual void          PokeLogic(int offset, unsigned int value);

    virtual GpProject_DDC* GetProjectRegs()
    {
        return &GpRegs;
    }


    ///////////////////
    //Board Information
    ///////////////////
    //Logic Version
    virtual std::string FpgaLogicVersion();
    //Hardware Variant
    virtual short FpgaHardwareVariant();
    //Revision
    virtual short PciLogicRevision();
    //Subrevision
    virtual short FpgaLogicSubrevision();
    //Board Family
    virtual std::string PciLogicFamily();
    //Type
    virtual std::string PciLogicType();
    //Board Revision
    virtual std::string PciLogicPcb();
    //Chip
    virtual std::string FpgaName();
    //PCI EXPRESS
    virtual std::string PCIExpressLanes();

    std::string PrintDevices()
       { return  Module.PrintAllDevices(); }

protected:
    void  Log(const std::string & msg);

private:
    //
    //  Data Members
    Drc_Provider *       Provider;
    Innovative::X6_GSPS  Module;
    bool                 Opened;

    GpDspMemoryExtension_BindTo<Innovative::X6_GSPS>  GpDsp_Ex;   // needs to be after Module
    GpProject_DDC        GpRegs;
    //
    //  Support Methods
    Drc_ProviderSettings &     Settings()     {  return Provider->Settings;  }
    Drc_ProviderPersonality &  Personality()  {  return Provider->Personality;  }
    //
    //  Module Methods
    void  HookAlerts();
    void  ConfigureAlerts(std::vector<char> & AlertEnable);
    void  SetInputSoftwareTrigger(bool state);
    void  SetInputPacketDataSize(unsigned int size);
    void  SetTestConfiguration( bool enable, int mode_idx );
    typedef std::vector<float> PulseSettingArray;
    void  SetPulseTriggerConfiguration(bool enable, float period,
                                 const PulseSettingArray & Delays,
                                 const PulseSettingArray & Widths );
    //
    //  Event Handlers
    //  ...Streaming
    void  HandleBeforeStreamStart(OpenWire::NotifyEvent & Event);
    void  HandleAfterStreamStop(OpenWire::NotifyEvent & Event);
    //  ...Alerts
    void  HandleTimestampRolloverAlert(Innovative::AlertSignalEvent & event);
    void  HandleSoftwareAlert(Innovative::AlertSignalEvent & event);
    void  HandleWarningTempAlert(Innovative::AlertSignalEvent & event);
    void  HandleInputFifoOverrunAlert(Innovative::AlertSignalEvent & event);
    void  HandleTriggerAlert(Innovative::AlertSignalEvent & event);
    void  HandleInputOverrangeAlert(Innovative::AlertSignalEvent & event);

    std::string Elapsed(size_t timestamp);
};


#ifdef __CLR_VER
#pragma managed(pop)
#endif

}  // end namespace Innovative

#endif // GPDSP_DRCBOARD_GSPS_IMPL_H
