% Copyright 2014 by Innovative Integration Inc., All Rights Reserved.
% Design Name: Digital Down Converter Design
% Tool versions: Matlab R2013b
%
% Description: the Polyphase Digital Down Converter, consisting of a 
% programmable tuner, a polyphase /8 filter (Polyfir), Cascaded
% Integrator-Comb (CIC) filter or the FIR1 decimation filter, a 
% Compensation filter (CFIR), and a Programmable filter (PFIR), decimates 
% the RF to the baseband signal, providing signal compensation and shaping. 
% CIC filter is useful to realize large sample rate changes in digital 
% systems. CIC filters are multiplierless structures, consisting of only 
% adders, subtractors and registers, which is a benefit for hardware 
% implementation. The compensation filter is used to flatten the passband 
% frequency response. The low-pass programmable filter is used to create
% the quick roll-off transition band.
%
% Author: Billy Kao
% Date Created: 12/12/2013
% Date Modified: 07/24/2015

function ddc_2gsps_filter(Fs, ddc_rate, BW)

close all, clc;

num_adc_bits = 16;

% test input
% Fs = 1800e6;
% ddc_rate = 0;
% BW = 100e3;
% BW = 5e6;
% BW = 10e6;
% BW = 15e6;
% BW = 20e6;
% BW = 30e6;
% BW = 35e6;
% BW = 40e6;
% BW = 45e6;   % filt_sel1=0, filt_sel2=1
% BW = 50e6;
% BW = 60e6;   % filt_sel1=2, filt_sel2=0 
% BW = 70e6;
% BW = 80e6;
% BW = 90e6;   % filt_sel1=1, filt_sel2=0
% BW = 100e6;  % filt_sel1=0, filt_sel2=0

% ===================================
% system parameter
% ===================================
Ts = 1/250e6;
Tsample = 1/Fs;

% ===================================
% filter design parameter
% ===================================
Fpass = BW/2;                  % bandwidth/2
polyfir_deci = 8;              % polyphase filter decimation rate
cfir_deci = 2;                 % compensation filter decimation rate
pfir_deci = 2;                 % programmable filter decimation rate

if mod(ddc_rate, 8) ~= 0
    error('The DDC decimation rate shall be multiple integer of 8!');
elseif ddc_rate == 0
    R = floor(Fs/(2*(Fpass/0.8)*polyfir_deci));
else
    R = ddc_rate/polyfir_deci;
end;

if R == 0
    disp(['Decimation Rate R: ' num2str(R)]);
    error('The bandwidth is not supported!');
elseif R == 1
    filt_sel1 = 0;
    filt_sel2 = 0;
    fir1_deci = 1;         % fir1 decimation rate
    cic_deci = 4;          % CIC filter decimation rate
elseif R == 2
    filt_sel1 = 1;
    filt_sel2 = 0;
    fir1_deci = 2;
    cic_deci = 4;
elseif R == 3
    filt_sel1 = 2;
    filt_sel2 = 0;
    fir1_deci = 3;
    cic_deci = 4;
elseif R >= 4 && R < 8
    filt_sel1 = 0;
    filt_sel2 = 1;
    fir1_deci = 1;
    cic_deci = 4;
elseif R >= 8 && R < 12
    filt_sel1 = 1;
    filt_sel2 = 1;
    fir1_deci = 2;
    cic_deci = 4;
elseif R >= 12 && R < 16
    filt_sel1 = 2;
    filt_sel2 = 1;
    fir1_deci = 3;
    cic_deci = 4;
elseif R >= 16 && R <= 32768
    filt_sel1 = 3;
    filt_sel2 = 1;
    fir1_deci = 1;
    cic_deci = floor(R/4);
else
    disp(['Decimation Rate R: ' num2str(R)]);
    error('The bandwidth is not supported!');
end;

if filt_sel1 <= 2 && filt_sel2 == 0
    ddc_rate = polyfir_deci*fir1_deci;
elseif filt_sel1 <= 2 && filt_sel2 == 1
    ddc_rate = polyfir_deci*fir1_deci*cfir_deci*pfir_deci;
elseif filt_sel1 == 3
    ddc_rate = polyfir_deci*cic_deci*cfir_deci*pfir_deci;
end;

Fddc = Fs/ddc_rate;
Tfft = 1/Fddc;

Apass = 0.02;          % passband ripple
Astop = 10^(-80/20);   % stopband attenuation (dB)
Aslope = 10;           % stopband slope
CoeffWordLength = 18;

% --------------------------------------
% filter design starts
% 1. polyphase /8 filter
% --------------------------------------
polyfir_order = 39;            % polyfir_taps = polyfir_order+1 

if Fpass > 300e3
    F = [0 Fpass/(Fs/2) (Fs/polyfir_deci-Fpass*1.5)/(Fs/2) 1];
else
    F = [0 Fpass/(Fs/2) (Fs/polyfir_deci-Fpass*3)/(Fs/2) 1];
end;

A = [1 1 0 0];
W = [1 20];
polyfir = firgr(polyfir_order,F,A,W);

hpolyfir = mfilt.firdecim(polyfir_deci, polyfir);
polyfir_coef = round(polyfir*2^(CoeffWordLength-1)*0.94);

% create default polyphase filter coefficients
polyfir_reshape = reshape(polyfir_coef,8,[]);
polyfir0 = polyfir_reshape(1,:);
polyfir1 = polyfir_reshape(2,:);
polyfir2 = polyfir_reshape(3,:);
polyfir3 = polyfir_reshape(4,:);
polyfir4 = polyfir_reshape(5,:);
polyfir5 = polyfir_reshape(6,:);
polyfir6 = polyfir_reshape(7,:);
polyfir7 = polyfir_reshape(8,:);

polyfir_flip = fliplr(polyfir_reshape);

% --------------------------------------
% 2. cascaded integrator-comb filter (CIC)
% --------------------------------------
Fcic = Fs/polyfir_deci;
M = 2;                         % differential delay
N = 5;                         % number of either comb or integrator sections
IWL = 24;                      % input word length
OWL = 94;                      % output word length
hcic = mfilt.cicdecim(cic_deci, M, N, IWL, OWL);
hgain = dfilt.scalar(1/gain(hcic));      % create a scalar filter
hcic_norm = mfilt.cascade(hgain, hcic);  % cic after normalization

% --------------------------------------
% 3. fir1 decimator (FIR1)
% --------------------------------------
fir1_order = 23;               % fir1_taps = fir1_order+1 
Ffir1 = Fcic;

if Fpass/(Ffir1/2) >= 0.4
    F = [0 Fpass/(Ffir1/2) (Fpass*1.5)/(Ffir1/2) 1];
    A = [1 1 0 0];
    W = [1 10];
    fir1 = firgr(fir1_order,F,A,W);
else
    if filt_sel1 == 0
        fir1 = firceqrip(fir1_order, Fpass/(Ffir1/2), [Apass, 10^(-55/20)], 'slope', 10, 'passedge');
    elseif filt_sel1 == 1 || filt_sel1 == 2
        fir1 = firceqrip(fir1_order, Fpass/(Ffir1/2), [Apass, 10^(-80/20)], 'slope', 10, 'passedge');
    elseif filt_sel1 == 3
        fir1 = zeros(1, fir1_order+1);
    end;
end;

hfir1 = mfilt.firdecim(fir1_deci, fir1);    
fir1_coef = round(fir1*2^(CoeffWordLength-1));

% --------------------------------------
% 4. compensation fir decimator (CFIR)
% --------------------------------------
cfir_order = 19;               % cfir_taps = cfir_order+1 

if filt_sel2 == 1
    if filt_sel1 == 3
        Fcfir = Fs/polyfir_deci/cic_deci;
        cfir_Fstop = Fs/polyfir_deci/cic_deci/cfir_deci/pfir_deci;  
        d = fdesign.decimator(cfir_deci, 'ciccomp', hcic.DifferentialDelay, hcic.NumberOfSections, 'N,Fp,Fst', cfir_order, Fpass, cfir_Fstop, Fcfir);
        hcfir = design(d, 'equiripple', 'StopbandShape', 'linear', 'StopbandDecay', 10);
        cfir = hcfir.Numerator;
    else
        Fcfir = Fs/polyfir_deci/fir1_deci;
        cfir = firceqrip(cfir_order, Fpass/(Fcfir/2), [Apass, 10^(-70/20)], 'slope', 20, 'passedge');
    end;
else
    Fcfir = Fs/polyfir_deci/fir1_deci;
    cfir = zeros(1, cfir_order+1);
end;

hcfir = mfilt.firdecim(cfir_deci, cfir);
cfir_coef = round(cfir*2^(CoeffWordLength-1)); 
    
% --------------------------------------
% 5. programmable filter (PFIR)
% --------------------------------------
pfir_order = 39;        % pfir_taps = pfir_order+1 
Fpfir = Fcfir/cfir_deci;

if filt_sel2 == 1
    pfir = firceqrip(pfir_order, Fpass/(Fpfir/2), [Apass, Astop], 'slope', 0, 'passedge');
else
    pfir = zeros(1, pfir_order+1);
end;

hpfir = mfilt.firdecim(pfir_deci, pfir);
pfir_coef = round(pfir*2^(CoeffWordLength-1));

% --------------------------------------
% 6. plot
% --------------------------------------
fvtool(hpolyfir, 'Fs', Fs);
title('Polyfir /8 filter');
ylim([-120 5]) 

hcfir_pfir = mfilt.cascade(hcfir, hpfir);
if filt_sel1 <= 2 && filt_sel2 == 0
    hddc = mfilt.cascade(hpolyfir, hfir1);
    fvtool(hfir1);%, 'Fs', Ffir1);
    title('FIR1 filter');
    h=fvtool(hpolyfir, hfir1, 'Fs', [Fs, Ffir1]);   
    legend(h, 'polyfir', 'fir1');
    xlim([0 Fs/8/1e6])
    fvtool(hddc, 'Fs', Fs);
    title('DDC Frequency Response');
    ylim([-100 5])      
elseif filt_sel1 <= 2 && filt_sel2 == 1
    hfir1_cfir_pfir = mfilt.cascade(hfir1, hcfir_pfir);
    hddc = mfilt.cascade(hpolyfir, hfir1_cfir_pfir);
    fvtool(hfir1, 'Fs', Ffir1);
    title('FIR1 filter');
    fvtool(hcfir, 'Fs', Fcfir);
    title('CFIR');
    fvtool(hpfir, 'Fs', Fpfir);
    title('PFIR');
    h=fvtool(hpolyfir, hfir1, hcfir, hpfir, 'Fs', [Fs, Ffir1, Fcfir, Fpfir]);   
    legend(h, 'polyfir', 'fir1', 'cfir', 'pfir');
    xlim([0 Fs/8/1e6])
    fvtool(hddc, 'Fs', Fs);
    title('DDC Frequency Response');
    xlim([0 Fs/8/1e6])
    ylim([-120 5])      
elseif filt_sel1 == 3
    hcic_cfir_pfir = mfilt.cascade(hcic_norm, hcfir_pfir);
    hddc = mfilt.cascade(hpolyfir, hcic_cfir_pfir);
    fvtool(hcic_norm, 'Fs', Fcic);
    title('CIC filter');
    fvtool(hcfir, 'Fs', Fcfir);
    title('CFIR');
    fvtool(hpfir, 'Fs', Fpfir);
    title('PFIR');
    h=fvtool(hpolyfir, hcic_norm, hcfir, hpfir, 'Fs', [Fs, Fcic, Fcfir, Fpfir]);   
    legend(h, 'polyfir', 'cic', 'cfir', 'pfir');
    xlim([0 8*Fs/ddc_rate/1e6])
    fvtool(hddc, 'Fs', Fs);
    title('DDC Frequency Response');
    xlim([0 Fs/8/1e6])
end;

disp(['The DDC Overall Decimation Rate: ' num2str(ddc_rate)]);
disp(['The DDC I/Q Output Data Rate: ' num2str(Fddc/1e6) ' MHz']);
disp(['Filt_sel1 = ' num2str(filt_sel1) '; Filt_sel2 = ' num2str(filt_sel2)]);

% --------------------------------------
% 7. create filter configuration file .coe
% --------------------------------------
coef_sets = 1;
coe_creator(polyfir_coef,'polyfir',coef_sets); 
coe_creator(fir1_coef,'fir1',coef_sets); 
coe_creator(cfir_coef,'cfir',coef_sets);  
coe_creator(pfir_coef,'pfir',coef_sets); 

% --------------------------------------
% 8. create filter configuration file .ini
% --------------------------------------
polyfir_load_idx = [0 1 2 3 4];        % 0 to polyfir_taps/8-1
fir1_load_idx = [11 10 9 8 7 6 5 4 3 2 1 0];
cfir_load_idx = [9 8 7 6 5 4 3 2 1 0];
pfir_load_idx = [17 19 16 18 13 15 12 14 9 11 8 10 5 7 4 6 1 3 0 2];  
polyfir_taps = polyfir_order+1;
fir1_taps = fir1_order+1;
cfir_taps = cfir_order+1;
pfir_taps = pfir_order+1;

polyfir_scale = 8;
if filt_sel1 <= 2 && filt_sel2 == 0
    fir1_scale = 17;
    cic_scale = 62;
    cfir_scale = 30;
    pfir_scale = 30;
elseif filt_sel1 <= 2 && filt_sel2 == 1
    fir1_scale = 17;
    cic_scale = 62;
    cfir_scale = 17;
    pfir_scale = 17;    
elseif filt_sel1 == 3
    fir1_scale = 40;
    cic_scale = ceil(5*log2(cic_deci*2));
    cfir_scale = 17;
    pfir_scale = 17;    
end;

ddc_scale = [polyfir_scale fir1_scale cic_scale cfir_scale pfir_scale];

ddc_ini_creator('./',Fs,BW,ddc_rate,filt_sel1,filt_sel2,polyfir_flip,polyfir_taps,polyfir_load_idx,cic_deci,fir1_coef,fir1_taps,fir1_load_idx,cfir_coef,cfir_taps,cfir_load_idx,pfir_coef,pfir_taps,pfir_load_idx,ddc_scale);

% --------------------------------------
% 9. create filter coefficient dataset for DDC simulation
% --------------------------------------
polyfir0_coef_data = [zeros(1,10),upsample([polyfir_taps/8,0,polyfir_flip(1,:)],2)];
polyfir1_coef_data = [zeros(1,10),upsample([polyfir_taps/8,1,polyfir_flip(2,:)],2)];
polyfir2_coef_data = [zeros(1,10),upsample([polyfir_taps/8,2,polyfir_flip(3,:)],2)];
polyfir3_coef_data = [zeros(1,10),upsample([polyfir_taps/8,3,polyfir_flip(4,:)],2)];
polyfir4_coef_data = [zeros(1,10),upsample([polyfir_taps/8,4,polyfir_flip(5,:)],2)];
polyfir5_coef_data = [zeros(1,10),upsample([polyfir_taps/8,5,polyfir_flip(6,:)],2)];
polyfir6_coef_data = [zeros(1,10),upsample([polyfir_taps/8,6,polyfir_flip(7,:)],2)];
polyfir7_coef_data = [zeros(1,10),upsample([polyfir_taps/8,7,polyfir_flip(8,:)],2)];
fir_coef_data_wt   = [zeros(1,10),upsample(ones(1,polyfir_taps/8+2),2)];
polyfir_coef_data  = [polyfir0_coef_data polyfir1_coef_data polyfir2_coef_data polyfir3_coef_data polyfir4_coef_data polyfir5_coef_data polyfir6_coef_data polyfir7_coef_data];
polyfir_coef_data_wt = [fir_coef_data_wt fir_coef_data_wt fir_coef_data_wt fir_coef_data_wt fir_coef_data_wt fir_coef_data_wt fir_coef_data_wt fir_coef_data_wt];

fir1_coef_data    = [zeros(1,10),upsample([fir1_taps/2,0,fir1_coef(12),fir1_coef(11),fir1_coef(10),fir1_coef(9),fir1_coef(8),fir1_coef(7),fir1_coef(6),fir1_coef(5),fir1_coef(4),fir1_coef(3),fir1_coef(2),fir1_coef(1)],2)];
fir1_coef_data_wt = [zeros(1,10),upsample(ones(1,fir1_taps/2+2),2)];
cfir_coef_data    = [zeros(1,10),upsample([cfir_taps/2,0,cfir_coef(10),cfir_coef(9),cfir_coef(8),cfir_coef(7),cfir_coef(6),cfir_coef(5),cfir_coef(4),cfir_coef(3),cfir_coef(2),cfir_coef(1)],2)];
cfir_coef_data_wt = [zeros(1,10),upsample(ones(1,cfir_taps/2+2),2)];
pfir_coef_data    = [zeros(1,10),upsample([pfir_taps/2,0,pfir_coef(18),pfir_coef(20),pfir_coef(17),pfir_coef(19),pfir_coef(14),pfir_coef(16),pfir_coef(13),pfir_coef(15),pfir_coef(10),pfir_coef(12),pfir_coef(9),pfir_coef(11),pfir_coef(6),pfir_coef(8),pfir_coef(5),pfir_coef(7),pfir_coef(2),pfir_coef(4),pfir_coef(1),pfir_coef(3)],2)];
pfir_coef_data_wt = [zeros(1,10),upsample(ones(1,pfir_taps/2+2),2)];

save ddc_2gsps_filter.mat Tsample Ts Tfft num_adc_bits filt_sel1 filt_sel2 polyfir0 polyfir1 polyfir2 polyfir3 polyfir4 polyfir5 polyfir6 polyfir7 fir1_coef cic_deci cfir_coef pfir_coef polyfir_coef_data polyfir_coef_data_wt fir1_coef_data fir1_coef_data_wt cfir_coef_data cfir_coef_data_wt pfir_coef_data pfir_coef_data_wt polyfir_scale fir1_scale cic_scale cfir_scale pfir_scale;

return

function coe_creator(coef,filename,ChannelNum)

fid = fopen(['./',filename,'.coe'],'wt');
fprintf(fid,'radix=10;\n');
fprintf(fid,'coefdata=\n');

for j=1:ChannelNum,
  for i = 1:length(coef)-1
      if j==1
          fprintf(fid,[num2str(coef(i)),',\n']);
      else
          fprintf(fid,['0',',\n']);
      end
  end
  if j==ChannelNum 
      if ChannelNum==1
          fprintf(fid,[num2str(coef(end)),';\n']);
      else
          fprintf(fid,[num2str(0),';\n']);
      end
  else
      if j==1
          fprintf(fid,[num2str(coef(end)),',\n']);
      else
          fprintf(fid,[num2str(0),',\n']);
      end
  end
end

fclose(fid);

return

function ddc_ini_creator(filepath,Fs,BW,ddc_rate,filt_sel1,filt_sel2,polyfir_coef,polyfir_taps,polyfir_load_idx,cic_deci,fir1_coef,fir1_taps,fir1_load_idx,cfir_coef,cfir_taps,cfir_load_idx,pfir_coef,pfir_taps,pfir_load_idx,ddc_scale)

% taps load sequence
m_polyfir_seq = length(polyfir_load_idx);
m_fir1_seq = length(fir1_load_idx);
m_cfir_seq = length(cfir_load_idx);
m_pfir_seq = length(pfir_load_idx);

% open file    
fid = fopen([filepath,'ddc_2gsps_Fs',num2str(Fs/1e6),'MHz_BW',num2str(BW/1e6),'MHz.ini'],'wt');

% scale factor
polyfir_scale = ddc_scale(1);
fir1_scale = ddc_scale(2);
cic_scale = ddc_scale(3);
cfir_scale = ddc_scale(4);
pfir_scale = ddc_scale(5);

% DDC overall decimation rate
fprintf(fid, 'DecimationFactor=%d\n', ddc_rate);

% Filter selection
fprintf(fid, '\n[Filt_Sel]\n');
fprintf(fid, 'filt_sel1=%d\n', filt_sel1);
fprintf(fid, 'filt_sel2=%d\n', filt_sel2);

% CIC  
fprintf(fid, '\n[CIC]\n');
fprintf(fid, 'ShiftDir=%d\n', 0);
fprintf(fid, 'ShiftNum=%d\n', cic_scale);  
% Rate  
fprintf(fid, 'CicRate=%d\n', cic_deci);
    
% FIR1
fprintf(fid, '\n[Fir1]\n');
fprintf(fid, 'ShiftDir=%d\n', 0);
fprintf(fid, 'ShiftNum=%d\n', fir1_scale); 
% Coefficient    
for i = 1:m_fir1_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, fir1_coef(i));
end
% Index
for i = 1:m_fir1_seq,
    kth_tap = fir1_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end  
fprintf(fid, 'TapsToLoad=%d\n', m_fir1_seq);
fprintf(fid, 'Taps=%d\n', fir1_taps);

% CFIR
fprintf(fid, '\n[CFir]\n');
fprintf(fid, 'ShiftDir=%d\n', 0);
fprintf(fid, 'ShiftNum=%d\n', cfir_scale);
% Coefficient    
for i = 1:m_cfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, cfir_coef(i));
end        
% Index
for i = 1:m_cfir_seq,
    kth_tap = cfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end    
fprintf(fid, 'TapsToLoad=%d\n', m_cfir_seq);
fprintf(fid, 'Taps=%d\n', cfir_taps);

% PFIR      
fprintf(fid, '\n[PFir]\n');
fprintf(fid, 'ShiftDir=%d\n', 0);
fprintf(fid, 'ShiftNum=%d\n', pfir_scale);   
% Coefficient
for i = 1:m_pfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, pfir_coef(i));
end     
% Index
for i = 1:m_pfir_seq,
    kth_tap = pfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end     
fprintf(fid, 'TapsToLoad=%d\n', m_pfir_seq);
fprintf(fid, 'Taps=%d\n', pfir_taps);    

% Polyphase filter overview
fprintf(fid, '\n[Poly_Overview]\n');
fprintf(fid, 'NPoly=%d\n', 8);
fprintf(fid, 'Taps=%d\n', polyfir_taps);
fprintf(fid, 'ShiftDir=%d\n', 0);
fprintf(fid, 'ShiftNum=%d\n', polyfir_scale);

% Polyfir0
fprintf(fid, '\n[Poly0]\n');
% Coefficient
for i = 1:m_polyfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, polyfir_coef(1,i));
end
% Index
for i = 1:m_polyfir_seq,
    kth_tap = polyfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end
fprintf(fid, 'TapsToLoad=%d\n', m_polyfir_seq);

% Polyfir1
fprintf(fid, '\n[Poly1]\n');
% Coefficient    
for i = 1:m_polyfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, polyfir_coef(2,i));
end
% Index
for i = 1:m_polyfir_seq,
    kth_tap = polyfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end    
fprintf(fid, 'TapsToLoad=%d\n', m_polyfir_seq);

% Polyfir2
fprintf(fid, '\n[Poly2]\n');
% Coefficient
for i = 1:m_polyfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, polyfir_coef(3,i));
end
% Index
for i = 1:m_polyfir_seq,
    kth_tap = polyfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end
fprintf(fid, 'TapsToLoad=%d\n', m_polyfir_seq);

% Polyfir3
fprintf(fid, '\n[Poly3]\n');
% Coefficient    
for i = 1:m_polyfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, polyfir_coef(4,i));
end
% Index
for i = 1:m_polyfir_seq,
    kth_tap = polyfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end    
fprintf(fid, 'TapsToLoad=%d\n', m_polyfir_seq);

% Polyfir4
fprintf(fid, '\n[Poly4]\n');    
% Coefficient    
for i = 1:m_polyfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, polyfir_coef(5,i));
end
% Index
for i = 1:m_polyfir_seq,
    kth_tap = polyfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end
fprintf(fid, 'TapsToLoad=%d\n', m_polyfir_seq);

% Polyfir5
fprintf(fid, '\n[Poly5]\n');   
% Coefficient    
for i = 1:m_polyfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, polyfir_coef(6,i));
end 
% Index
for i = 1:m_polyfir_seq,
    kth_tap = polyfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end 
fprintf(fid, 'TapsToLoad=%d\n', m_polyfir_seq);

% Polyfir6
fprintf(fid, '\n[Poly6]\n');   
% Coefficient    
for i = 1:m_polyfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, polyfir_coef(7,i));
end    
% Index
for i = 1:m_polyfir_seq,
    kth_tap = polyfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end 
fprintf(fid, 'TapsToLoad=%d\n', m_polyfir_seq);

% Polyfir7
fprintf(fid, '\n[Poly7]\n');  
% Coefficient    
for i = 1:m_polyfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, polyfir_coef(8,i));
end   
% Index
for i = 1:m_polyfir_seq,
    kth_tap = polyfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end  
fprintf(fid, 'TapsToLoad=%d', m_polyfir_seq);

fclose(fid);

return
