% Copyright 2014 by Innovative Integration Inc., All Rights Reserved.
% Design Name: Digital Down Converter Design
% Tool versions: Matlab R2013b
%
% Description: the Polyphase Digital Down Converter, consisting of a 
% programmable tuner and a polyphase filter (Polyfir), decimates the RF to 
% the baseband signal, providing channel rejection and shaping.
%
% Author: Billy Kao
% Date Created: 12/12/2013
% Date Modified: 07/24/2015

function uwbddc_filter(Fs, ddc_rate, BW)

close all, clc;

num_adc_bits = 16;

% test input
% Fs = 1800e6;
% ddc_rate = 0;
% BW = 500e6;

% ===================================
% system parameter
% ===================================
Ts = 1/251e6;
Tsample = 1/Fs;

% ===================================
% filter design parameter
% ===================================
Fpass = BW/2;                  % bandwidth/2
polyfir_deci = 2;              % polyphase filter decimation rate

if ddc_rate == 1 || ddc_rate == 3 || ddc_rate ==5 || ddc_rate == 6 || ddc_rate == 7 || ddc_rate > 8
    error('The DDC decimation rate is not supported!');
elseif ddc_rate == 0
    R = floor(Fs/(2*(Fpass/0.85)*polyfir_deci));
    if R >= 4
        filt_sel1 = 2;
        R = 4;
    elseif R >= 2 && R < 4
        filt_sel1 = 1;
        R = 2;
    elseif R < 2
        filt_sel1 = 0;
        R = 1;
    end;
elseif ddc_rate == 2
    filt_sel1 = 0;
    R = 1;
elseif ddc_rate == 4
    filt_sel1 = 1;
    R = 2;
elseif ddc_rate == 8
    filt_sel1 = 2;
    R = 4;
end;

ddc_rate = polyfir_deci*R;
Fddc = Fs/ddc_rate;
Tfft = 1/Fddc;
filt_sel2 = 0;

Apass = 0.04;        % passband ripple
CoeffWordLength = 18;

% --------------------------------------
% filter design starts
% 1. polyphase /2 filter
% --------------------------------------
polyfir_order = 79;            % polyfir_taps = polyfir_order+1 

if filt_sel1 == 0
    Astop = 10^(-90/20);
    Aslope = 10; 
elseif filt_sel1 == 1
    Astop = 10^(-80/20);
    Aslope = 10; 
elseif filt_sel1 == 2
    Astop = 10^(-60/20);
    Aslope = 10;   
end;

polyfir = firceqrip(polyfir_order, Fpass/(Fs/2), [Apass, Astop], 'slope', Aslope, 'passedge');
hpolyfir = mfilt.firdecim(polyfir_deci, polyfir);
polyfir_coef = round(polyfir*2^(CoeffWordLength-1)*0.94);

% create default polyphase filter coefficients
polyfir_reshape = reshape(polyfir_coef,8,[]);
polyfir0 = polyfir_reshape(1,:);
polyfir1 = polyfir_reshape(2,:);
polyfir2 = polyfir_reshape(3,:);
polyfir3 = polyfir_reshape(4,:);
polyfir4 = polyfir_reshape(5,:);
polyfir5 = polyfir_reshape(6,:);
polyfir6 = polyfir_reshape(7,:);
polyfir7 = polyfir_reshape(8,:);

%polyfir_flip = fliplr(polyfir_reshape);

% --------------------------------------
% 2. plot
% --------------------------------------
fvtool(hpolyfir, 'Fs', Fs);
title('UWBDDC Frequency Response');
xlim([0 Fddc/1e6])
ylim([-120 5]) 

disp(['The DDC Overall Decimation Rate: ' num2str(ddc_rate)]);
disp(['The DDC I/Q Output Data Rate: ' num2str(Fddc/1e6) ' MHz']);
disp(['Filt_sel1 = ' num2str(filt_sel1) '; Filt_sel2 = ' num2str(filt_sel2)]);

% --------------------------------------
% 3. create filter configuration file .coe
% --------------------------------------
coef_sets = 1;
coe_creator(polyfir_coef,'polyfir',coef_sets); 

% --------------------------------------
% 4. create filter configuration file .ini
% --------------------------------------
polyfir_load_idx = [0 1 2 3 4 5 6 7 8 9];
polyfir_taps = polyfir_order+1;

polyfir_scale = 8;
ddc_scale = [polyfir_scale];

ddc_ini_creator('./',Fs,BW,ddc_rate,filt_sel1,filt_sel2,polyfir_reshape,polyfir_taps,polyfir_load_idx,ddc_scale);

% --------------------------------------
% 5. create polyfir_coef_data for multi-channel DDC simulation
% --------------------------------------
polyfir0_coef_data = [zeros(1,10),upsample([polyfir_taps/8,0,polyfir_reshape(1,:)],2)];
polyfir1_coef_data = [zeros(1,10),upsample([polyfir_taps/8,1,polyfir_reshape(2,:)],2)];
polyfir2_coef_data = [zeros(1,10),upsample([polyfir_taps/8,2,polyfir_reshape(3,:)],2)];
polyfir3_coef_data = [zeros(1,10),upsample([polyfir_taps/8,3,polyfir_reshape(4,:)],2)];
polyfir4_coef_data = [zeros(1,10),upsample([polyfir_taps/8,4,polyfir_reshape(5,:)],2)];
polyfir5_coef_data = [zeros(1,10),upsample([polyfir_taps/8,5,polyfir_reshape(6,:)],2)];
polyfir6_coef_data = [zeros(1,10),upsample([polyfir_taps/8,6,polyfir_reshape(7,:)],2)];
polyfir7_coef_data = [zeros(1,10),upsample([polyfir_taps/8,7,polyfir_reshape(8,:)],2)];
fir_coef_data_wt   = [zeros(1,10),upsample(ones(1,polyfir_taps/8+2),2)];
polyfir_coef_data  = [polyfir0_coef_data polyfir1_coef_data polyfir2_coef_data polyfir3_coef_data polyfir4_coef_data polyfir5_coef_data polyfir6_coef_data polyfir7_coef_data];
polyfir_coef_data_wt = [fir_coef_data_wt fir_coef_data_wt fir_coef_data_wt fir_coef_data_wt fir_coef_data_wt fir_coef_data_wt fir_coef_data_wt fir_coef_data_wt];

save uwbddc_filter.mat Tsample Ts Tfft num_adc_bits filt_sel1 filt_sel2 polyfir0 polyfir1 polyfir2 polyfir3 polyfir4 polyfir5 polyfir6 polyfir7 polyfir_scale polyfir_coef_data polyfir_coef_data_wt;

return

function coe_creator(coef,filename,ChannelNum)

fid = fopen(['./',filename,'.coe'],'wt');
fprintf(fid,'radix=10;\n');
fprintf(fid,'coefdata=\n');

for j=1:ChannelNum,
  for i = 1:length(coef)-1
      if j==1
          fprintf(fid,[num2str(coef(i)),',\n']);
      else
          fprintf(fid,['0',',\n']);
      end
  end
  if j==ChannelNum 
      if ChannelNum==1
          fprintf(fid,[num2str(coef(end)),';\n']);
      else
          fprintf(fid,[num2str(0),';\n']);
      end
  else
      if j==1
          fprintf(fid,[num2str(coef(end)),',\n']);
      else
          fprintf(fid,[num2str(0),',\n']);
      end
  end
end

fclose(fid);

return

function ddc_ini_creator(filepath,Fs,BW,ddc_rate,filt_sel1,filt_sel2,polyfir_coef,polyfir_taps,polyfir_load_idx,ddc_scale)

% taps load sequence
m_polyfir_seq = length(polyfir_load_idx);
    
% open file    
fid = fopen([filepath,'uwbddc_Fs',num2str(Fs/1e6),'MHz_BW',num2str(BW/1e6),'MHz.ini'],'wt');

% scale factor
polyfir_scale = ddc_scale(1);

% DDC overall decimation rate
fprintf(fid, 'DecimationFactor=%d\n', ddc_rate);

% Filter selection
fprintf(fid, '\n[Filt_Sel]\n');
fprintf(fid, 'filt_sel1=%d\n', filt_sel1);
fprintf(fid, 'filt_sel2=%d\n', filt_sel2);

% Polyphase filter overview
fprintf(fid, '\n[Poly_Overview]\n');
fprintf(fid, 'NPoly=%d\n', 8);
fprintf(fid, 'Taps=%d\n', polyfir_taps);
fprintf(fid, 'ShiftDir=%d\n', 0);
fprintf(fid, 'ShiftNum=%d\n', polyfir_scale);

% Polyfir0
fprintf(fid, '\n[Poly0]\n');
% Coefficient
for i = 1:m_polyfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, polyfir_coef(1,i));
end
% Index
for i = 1:m_polyfir_seq,
    kth_tap = polyfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end
fprintf(fid, 'TapsToLoad=%d\n', m_polyfir_seq);

% Polyfir1
fprintf(fid, '\n[Poly1]\n');
% Coefficient    
for i = 1:m_polyfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, polyfir_coef(2,i));
end
% Index
for i = 1:m_polyfir_seq,
    kth_tap = polyfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end    
fprintf(fid, 'TapsToLoad=%d\n', m_polyfir_seq);

% Polyfir2
fprintf(fid, '\n[Poly2]\n');
% Coefficient
for i = 1:m_polyfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, polyfir_coef(3,i));
end
% Index
for i = 1:m_polyfir_seq,
    kth_tap = polyfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end
fprintf(fid, 'TapsToLoad=%d\n', m_polyfir_seq);

% Polyfir3
fprintf(fid, '\n[Poly3]\n');
% Coefficient    
for i = 1:m_polyfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, polyfir_coef(4,i));
end
% Index
for i = 1:m_polyfir_seq,
    kth_tap = polyfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end    
fprintf(fid, 'TapsToLoad=%d\n', m_polyfir_seq);

% Polyfir4
fprintf(fid, '\n[Poly4]\n');    
% Coefficient    
for i = 1:m_polyfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, polyfir_coef(5,i));
end
% Index
for i = 1:m_polyfir_seq,
    kth_tap = polyfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end
fprintf(fid, 'TapsToLoad=%d\n', m_polyfir_seq);

% Polyfir5
fprintf(fid, '\n[Poly5]\n');   
% Coefficient    
for i = 1:m_polyfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, polyfir_coef(6,i));
end 
% Index
for i = 1:m_polyfir_seq,
    kth_tap = polyfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end 
fprintf(fid, 'TapsToLoad=%d\n', m_polyfir_seq);

% Polyfir6
fprintf(fid, '\n[Poly6]\n');   
% Coefficient    
for i = 1:m_polyfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, polyfir_coef(7,i));
end    
% Index
for i = 1:m_polyfir_seq,
    kth_tap = polyfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end 
fprintf(fid, 'TapsToLoad=%d\n', m_polyfir_seq);

% Polyfir7
fprintf(fid, '\n[Poly7]\n');  
% Coefficient    
for i = 1:m_polyfir_seq,
    fprintf(fid, 'Coeff%d=%d\n', i-1, polyfir_coef(8,i));
end   
% Index
for i = 1:m_polyfir_seq,
    kth_tap = polyfir_load_idx(i);
    fprintf(fid, 'Index%d=%d\n', i-1, kth_tap);
end  
fprintf(fid, 'TapsToLoad=%d', m_polyfir_seq);

fclose(fid);

return
