#include "mainwindow.h"
#include "ui_mainwindow.h"
#include <Malibu_Mb.h>
#include <StringSupport_Mb.h>
#include <SystemSupport_Mb.h>
#include <GpDsp_Config_Gp.h>
#include <DataGrid.h>
#include <GpDsp_Config_Impl.h>
#include <DataPlayer_Mb.h>
#include <iostream>
#include <iomanip>
#include <limits>
#include <algorithm>
#include <QMessageBox>
#include <QStylePainter>
#include <Fft_WindowingCoeffs.h>
#include <MalibuVersion_Mb.h>
#include "Poco/Path.h"


using namespace Innovative;
using namespace std;
using namespace Innovative::GpDsp_Gui;


//===========================================================================
//  CLASS MainWindow  -- GUI Form
//===========================================================================
//---------------------------------------------------------------------------
//  constructor of class MainWindow
//---------------------------------------------------------------------------

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent), ui(new Ui::MainWindow)
{

    ui->setupUi(this);

    Inited = false;
    psignalMapper = 0;
    SelectedFftRowIndex = 0;

    //
    //  Set up Manager
    MUI    = new MGR_GUIInterface(this);
    MFICL  = new MGR_FiclInterface(this);
    Mgr    = new GpDspManager(*MUI, *MFICL);

    //  Create Physical Configuraton (and FICL configuraton)
    Mgr->ConfigureSystem();

    ddcTable.reset(new QTableWidgetEx(ui->DdcTable));
    fftTable.reset(new QTableWidgetEx(ui->FftSetupTable));

    QObject::connect(ddcTable.data(), SIGNAL(updateRequired(const std::vector<int> &)),
                     this, SLOT(DdcTableChanged(const std::vector<int> &)));

    QObject::connect(ddcTable.data(), SIGNAL(currentCellChanged(int, int, int, int)),
                     this, SLOT(DdcTableCurrentCellChanged(int, int, int, int)));

    QObject::connect(fftTable.data(), SIGNAL(currentCellChanged(int, int, int, int)),
                     this, SLOT(FftSetupTableCurrentCellChanged(int, int, int, int)));


    ////////////
    //GUI setup
    ////////////
    InstallSettings(Mgr->DrcSettings);
    InstallSettings(Mgr->ClockModules._cache);

    //  Nice stuff
    //  Form Title and Build date Labels
    FormTitle(this, MainFormTitleString());
    //  show config page at start
    ShowTab(ui->ConfigTab, ui->Tabs, true);

    ToControl(ui->ConfigSwVersionLineEdit, SoftwareVersionString());
    ToControl(ui->ConfigSwBuildDateLineEdit, std::string(__DATE__));
    ToControl(ui->ConfigSwBuildTimeLineEdit, std::string(__TIME__));
    ToControl(ui->ConfigMalibuVersionLineEdit, MalibuVersion().Malibu());

    // Setup Power Meter table
    vector<std::string> PowerMetertableVerticalHeader(DdcSourceList());
    PowerMetertableVerticalHeader.pop_back();
    PowerMeterTableInit(ui->PowerMeterTable, PowerMetertableVerticalHeader);

    // Setup Test Generator table
    vector<std::string> TestGeneratortableVerticalHeader;
    TestGeneratortableVerticalHeader.push_back("Tone1");
    TestGeneratorTableInit(ui->DDCTestSignalTable, TestGeneratortableVerticalHeader);

    //  Populate Listbox for Module Alerts in board independent way
    FillControl(ui->AlertEnableCheckListBox, AlertStrings());

    // Enable buttons
    ControlEnabled(ui->CloseBtn, false);
    ControlEnabled(ui->OpenBtn, true);
    ControlEnabled(ui->StreamStartBtn, false);
    ControlEnabled(ui->StreamStopBtn, false);
    ControlEnabled(ui->BusMasterSizeEditBox, true);
    //
    vector<string> SrcList = DdcSourceList();
    FillControl(ui->DDCSourceSelectComboBox_2, SrcList);
    // Clock Module
    ClockModuleGuiConfig();

    //set controls according to settings
    SetSettings();
    //vector<std::string> StatusVerticalHeader;
    BoardStatusTableInit(ui->tableStatus, vector<std::string>());
    SetupBoardsTable();

    NameUpperLeftCornerOfTheTableWidget();

    // Context menu
    ui->GraphPanel->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(ui->GraphPanel, SIGNAL(customContextMenuRequested(const QPoint&)),
       this, SLOT(ShowContextMenu(const QPoint&)));

    // Create actions
    pEnableAutoScale = new QAction(tr("Log Plot Data"), this);
    connect(pEnableAutoScale, SIGNAL(triggered()), this, SLOT(EnableAutoScale()));
    pEnableAutoScale->setCheckable(true);
    pEnableAutoScale->setChecked(Mgr->DrcSettings.EnableFFTLogging);

    pEnablePlot = new QAction(tr("Enable Plotting"), this);
    connect(pEnablePlot, SIGNAL(triggered()), this, SLOT(EnablePlotting()));
    pEnablePlot->setCheckable(true);
    pEnablePlot->setChecked(Mgr->DrcSettings.EnablePlotting);

    pSavePlot = new QAction(tr("Save Plot"), this);
    connect(pSavePlot, SIGNAL(triggered()), this, SLOT(SavePlot()));

    pRawxAxis = new QAction(tr("Show bin numbers"), this);
    connect(pRawxAxis, SIGNAL(triggered()), this, SLOT(TogglexScale()));
    pRawxAxis->setCheckable(true);

    ContextMenu.addAction(pEnableAutoScale);
    ContextMenu.addAction(pEnablePlot);
    ContextMenu.addAction(pSavePlot);
    ContextMenu.addAction(pRawxAxis);

#ifdef DRC
    RemoveTab(ui->Tabs, ui->Tabs->indexOf(ui->DucTab));
    ControlVisible(ui->DacSampleRateEdit,false);
    ControlVisible(ui->DacFreqLabel,false);
#endif


#if defined(LINUX)
    std::string expPath(Poco::Path::expand("/usr/Innovative/Lib/Dll/"));
#else

#if WIN64
	std::string expPath(Poco::Path::expand("%InnovativeCommon%\\Lib\\Dll_x64\\"));
#else
    std::string expPath(Poco::Path::expand("%InnovativeCommon%\\Lib\\Dll\\"));
#endif
#endif
    FFTGraph.Path(expPath);
   // Attach Graph to Panel
    FFTGraph.Attach(ui->GraphPanel);

    //Print out the graph version
    std::stringstream GraphVersion;
    GraphVersion << "Graph Version = " << FFTGraph->version();
    Log(GraphVersion.str());

    TableSelectionBehavior(ui->DucTable);
    if(Mgr->BoardCount()>0)
      Log(Mgr->Boards(0).PrintDevices());

    (void)swin_coeff_info;
}

//---------------------------------------------------------------------------
//  MainWindow::ClockModuleGuiConfig()
//---------------------------------------------------------------------------

void MainWindow::ClockModuleGuiConfig()
{
    // Make sure we have some clock modules
    if (Mgr->ClockModules.Size() == 0)
        return;

    vector<string> rowlabel;
    rowlabel.push_back("Module");
    TimingModulesTableInit(ui->TT_TimingTable, rowlabel);

    bool en = Mgr->ClockModules.IsGpsEnabled();
    ControlEnabled(ui->TT_GPSEnableB, !en);
    ControlEnabled(ui->TT_GPSDisableB, en);
    en = Mgr->ClockModules.IsSoftwareTriggerActivated();
    ControlEnabled(ui->TT_SoftwareTriggerEnableB,  en);
    ControlEnabled(ui->TT_SoftwareTriggerDisableB, !en);

    if(Mgr->ClockModules.isAtropos())
    {
        ControlEnabled(ui->TT_PPSSourceGB, false);
        ControlEnabled(ui->TT_GPSTriggerGB, false);
        ControlEnabled(ui->TT_ExtTriggerSourceLB,  true);
    }

    if(Mgr->ClockModules.isX3Timing())
    {
        ControlEnabled(ui->TT_TypeGB, false);
        ControlEnabled(ui->TT_PPSSourceGB, false);
        ControlEnabled(ui->TT_ExtTriggerSourceLB,  true);
    }

    if(Mgr->ClockModules.isDuo())
    {
        ControlEnabled(ui->TT_GPSTriggerGB, false);
        ControlEnabled(ui->TT_ExtTriggerSourceLB,  false);
    }


    // Reference Source
    {
    ClearControl(ui->TT_ReferenceClockSourceLB);
    vector<string> list = Mgr->ClockModules.Proxy(0).ReferenceSources();
    for (size_t i = 0; i < list.size(); i++)
        AddToControl(ui->TT_ReferenceClockSourceLB, list[i]);
    }
    // Clock Source
    {
    ClearControl(ui->TT_SampleClockSourceLB);
    vector<string> list = Mgr->ClockModules.Proxy(0).ClockSources();
    for (size_t i = 0; i < list.size(); i++)
        AddToControl(ui->TT_SampleClockSourceLB, list[i]);
    }
    // External Clock Source
    {
    ClearControl(ui->TT_ExtClockSourceLB);
    vector<string> list = Mgr->ClockModules.Proxy(0).ExternalClockSources();
    for (size_t i = 0; i < list.size(); i++)
        AddToControl(ui->TT_ExtClockSourceLB, list[i]);
    }
    // External Trig Source
    {
    ClearControl(ui->TT_ExtTriggerSourceLB);
    vector<string> list = Mgr->ClockModules.Proxy(0).ExternalTriggerSources();
    for (size_t i = 0; i < list.size(); i++)
        AddToControl(ui->TT_ExtTriggerSourceLB, list[i]);
    }

    if (Mgr->ClockModules.Size())
        ControlEnabled(ui->TT_GPSGB, Mgr->ClockModules.Proxy(0).GpsSupported());
}

//---------------------------------------------------------------------------
//  MainWindow::DdcTableChanged()
//---------------------------------------------------------------------------

void MainWindow::DdcTableChanged(const std::vector<int> &rows)
{
    GL.Invalidate();
    GetSettings();
    if (WarnChannelSelect(rows))
       {
       Mgr->DdcChannels.ApplyCached(rows);
       }
    SetSettings();
}

//---------------------------------------------------------------------------
//  MainWindow::resizeEvent
//---------------------------------------------------------------------------

void MainWindow::resizeEvent(QResizeEvent* event)
{
   QMainWindow::resizeEvent(event);

   FFTGraph->resize();
}

//---------------------------------------------------------------------------
//  destructor of class MainWindow
//---------------------------------------------------------------------------

MainWindow::~MainWindow()
{
    GetSettings();

    Mgr->Close();

    delete Mgr;

    delete ui;
    delete MFICL;
    delete MUI;

    psignalMapper = 0;
}

//---------------------------------------------------------------------------
//  MainWindow::EnablePlotting()
//---------------------------------------------------------------------------

void MainWindow::EnablePlotting()
{
    Mgr->DrcSettings.EnablePlotting = pEnablePlot->isChecked();
}

//---------------------------------------------------------------------------
//  MainWindow::EnableAutoScale()
//---------------------------------------------------------------------------

void MainWindow::EnableAutoScale()
{
    Mgr->DrcSettings.EnableFFTLogging = pEnableAutoScale->isChecked();
}

//---------------------------------------------------------------------------
//   MainWindow::SavePlot()
//---------------------------------------------------------------------------

void MainWindow::SavePlot()
{
    FFTGraph->takeSnapshot("Snapshot.pdf");
}

//---------------------------------------------------------------------------
//  MainWindow::TogglexScale()
//---------------------------------------------------------------------------

void MainWindow::TogglexScale()
{
    //FFTGraph->togglexScale(); FIXME
    // pRawxAxis->setChecked(FFTGraph->isBinModeEnabled());
}

//---------------------------------------------------------------------------
//  MainWindow::ShowContextMenu()
//---------------------------------------------------------------------------

void MainWindow::ShowContextMenu(const QPoint& pos) // this is a slot
{
    // for most widgets
   QPoint globalPos = ui->GraphPanel->mapToGlobal(pos);
    // for QAbstractScrollArea and derived classes you would use:
    // QPoint globalPos = myWidget->viewport()->mapToGlobal(pos);
   ContextMenu.exec(globalPos);
}



//---------------------------------------------------------------------------
//  MainWindow::NameUpperLeftCornerOfTheTableWidget() --
//---------------------------------------------------------------------------

void MainWindow::NameUpperLeftCornerOfTheTableWidget()
{
    QAbstractButton* cornerButton = ui->DdcTable->findChild<QAbstractButton*>();

    if (cornerButton)
        {
        cornerButton->setText(" Select All");
        // Install this object as the event filter
        cornerButton->installEventFilter(this);
        }
}

//---------------------------------------------------------------------------
//  MainWindow::AddComboWidgetToTableWidget() --
//---------------------------------------------------------------------------

void MainWindow::AddComboWidgetToTableWidget(int row, int col, const vector<string> &items, int idx)
{
    if (psignalMapper == 0)
        psignalMapper = new QSignalMapper(this);

    QComboBox *combo = new QComboBox();
    FillControl(combo, items);
    ToControl(combo, idx);

    // Insert combo
    ui->FftPlotTableWidget->setCellWidget(row, col, combo);
    // Connect combo to mapper
    connect(combo, SIGNAL(currentIndexChanged(int)), psignalMapper, SLOT(map()));
    psignalMapper->setMapping(combo, QString("%1-%2").arg(row).arg(col));
    // Connect mapper to mainwindow
    connect(psignalMapper, SIGNAL(mapped(const QString &)), this, SLOT(EmbeddedCombo_currentIndexChanged(const QString &)));
}

//---------------------------------------------------------------------------
//  MainWindow::EmbeddedCombo_currentIndexChanged() --
//---------------------------------------------------------------------------

void MainWindow::EmbeddedCombo_currentIndexChanged(const QString &position)
{
    EmbeddedCombo_currentIndexChangedEmitCall(position,true);
}

//---------------------------------------------------------------------------
//  MainWindow::eventFilter() -- Intercept corner button PAINT events
//---------------------------------------------------------------------------

bool MainWindow::eventFilter(QObject* o, QEvent* e)
{
    if (e->type() == QEvent::Paint)
        {
        QAbstractButton* btn = qobject_cast<QAbstractButton*>(o);
        if (btn)
            {
            // paint by hand (borrowed from QTableCornerButton)
            QStyleOptionHeader opt;
            opt.init(btn);
            QStyle::State styleState = QStyle::State_None;
            if (btn->isEnabled())      styleState |= QStyle::State_Enabled;
            if (btn->isActiveWindow()) styleState |= QStyle::State_Active;
            if (btn->isDown())         styleState |= QStyle::State_Sunken;
            opt.state = styleState;
            opt.rect = btn->rect();
            opt.text = btn->text();
            opt.icon = btn->icon();
            opt.position = QStyleOptionHeader::OnlyOneSection;
            QStylePainter painter(btn);
            painter.drawControl(QStyle::CE_Header, opt);
            return true; // eat event
            }
        }
    return false;
}


//---------------------------------------------------------------------------
//  MainWindow::ComponentInit() --
//---------------------------------------------------------------------------

void MainWindow::ComponentInit()
{
    if(!Inited)
    {
    // Make components
    Mgr->Make_Components();

    // Connect components GUI
    InstallComponentSettings();

    // Setup Stream Status
    vector<std::string> StreamStatusTableVerticalHeader;
    Mgr->BoardList(StreamStatusTableVerticalHeader, true);
    StreamStatusTableInitDdc(ui->StreamStatusTable, StreamStatusTableVerticalHeader);
    // Setup Overflow table
    vector<std::string> OverflowTableVerticalHeader;
    Mgr->OverflowDetector.OverflowDetectorList(OverflowTableVerticalHeader, 0);
    //  WAS PRE_MERGE...
    //Mgr->DdcChannels.ChannelTags(OverflowTableVerticalHeader, 0);
    OverflowTableInit(ui->OverflowTable, OverflowTableVerticalHeader);
    // Setup Ddc table
    vector<std::string> VerticalHeader;
    Mgr->DdcChannelList(VerticalHeader);
    DrcTableInit(ddcTable.data(), VerticalHeader);
    // Setup Power Meter table
    LoadPowerMeterTargetSelector();

    if (Mgr->FftChannels.Size() > 0)
        {
        // Setup Fft Setup table
        vector<std::string> FftCoreTags;
        Mgr->FftChannels.ChannelTags(FftCoreTags);
        FftSetupTableInit(fftTable.data(), FftCoreTags);
        FftPlotTableInit(ui->FftPlotTableWidget, FftCoreTags);
        // Fft setup combo init
        PopulateFftSourceCombo(SelectedFftBoard());
        CreateFftSourceMap();
        // Fft setup window combo init
        vector<string> WindowingList;
        for (size_t i = 0; i < sizeof(win_coeff_info) / sizeof(win_coeff_info[0]); i++)
            WindowingList.push_back(win_coeff_info[i].name);

        // disconnect signal
        ui->FftSetupWindowingCombo->disconnect(this);
        FillControl(ui->FftSetupWindowingCombo, WindowingList);
        // reconnect signal
        connect(ui->FftSetupWindowingCombo, SIGNAL(currentIndexChanged(int)),
                this, SLOT(on_FftSetupWindowingCombo_currentIndexChanged(int)));

        AddComboWidgetToTableWidget(0, 0, Mgr->FftSources(Mgr->FftChannels.Wrapper(0).WhichBoard()),
                                         ReadControl<int>(ui->FftSetupInputSourceCombo));
        AddComboWidgetToTableWidget(0, 1, WindowingList, ReadControl<int>(ui->FftSetupWindowingCombo));

        // SetSettings will be called by this
        FftSetupInputSourceComboEmitCall(Mgr->FftChannels.Cache.Source,true); //fixme

        on_FftSetupOutputSpectrumComboBox_currentIndexChanged(Mgr->FftChannels.Cache.OutputMode);


      }

    Inited = true;
    }

    if (Mgr->FftChannels.Size() > 0)
        {
        // Enable tab pages for FFT
        ui->FftTab->setEnabled(true);
        ui->FftSetupTab->setEnabled(true);
        }
    else
        {
        // Disable tab pages for FFT
        ui->FftTab->setEnabled(false);
        ui->FftSetupTab->setEnabled(false);
        SetSettings();
        }
}

//---------------------------------------------------------------------------
//  MainWindow::CreateFftSourceMap() --
//---------------------------------------------------------------------------

void MainWindow::CreateFftSourceMap()
{
    // DdcSourceList includes Test; so minus 1
    int adc_count = (DdcSourceList().size() - 1) < 8 ? (DdcSourceList().size() - 1) : 8;

    Mgr->SetSourceMaps(adc_count);
}

//---------------------------------------------------------------------------
//  MainWindow::InitMaps() --
//---------------------------------------------------------------------------

 map<string, vector<string> > MainWindow::InitMaps(unsigned int ordinal)
 {
     map<string, vector<string> > temp;

     vector<string> WindowingList;

     for (size_t i = 0; i < sizeof(win_coeff_info) / sizeof(win_coeff_info[0]); i++)
         WindowingList.push_back(win_coeff_info[i].name);

     vector<string> SpectrumList;
     SpectrumList.push_back("Full Spectrum");
     SpectrumList.push_back("TL Spectrum");

     vector<string> ModesList;
     ModesList.push_back("Disable");
     ModesList.push_back("Single");
     ModesList.push_back("Continuous");

     vector<string> SourceList = Mgr->FftSources(ordinal);

     vector<string> OnOffList;
     OnOffList.push_back("Off");
     OnOffList.push_back("On");

     temp["OutputMode"]   = SpectrumList;
     temp["Source"]       = SourceList;
     temp["Window"]       = WindowingList;
     temp["TriggerMode"]  = ModesList;
     temp["OnOff"]        = OnOffList;

     return temp;
}

//---------------------------------------------------------------------------
//  MainWindow::HandleStatusTimer() -- 1 sec
//---------------------------------------------------------------------------

void  MainWindow::HandleStatusTimer(OpenWire::NotifyEvent & /*Event*/)
{
}

//---------------------------------------------------------------------------
//  MainWindow::~changeEvent() --
//---------------------------------------------------------------------------

void MainWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

//===========================================================================
//  Event Handlers
//===========================================================================

//---------------------------------------------------------------------------
//  MainWindow::on_OpenBtn_clicked() --  Open driver
//---------------------------------------------------------------------------

void MainWindow::on_OpenBtn_clicked()
{
    GetSettings();

    //Configure Ficl targer list
    //before open because user may change
    //target or change number of modules
    LoadFiclTargetSelector();

    //open each installed board
    Mgr->Open();

    ControlEnabled(ui->OpenBtn, false);
    ControlEnabled(ui->CloseBtn, true);
    ControlEnabled(ui->StreamStartBtn, true);
    ControlEnabled(ui->StreamStopBtn, false);
    ControlEnabled(ui->BusMasterSizeEditBox, false);
    ControlVisible(ui->SampleClockSourceRadioGroup, HasClockMux());
    ControlVisible(ui->ReferenceGroupBox, HasProgrammableReference());

    ComponentInit();

    UpdateBoardsTable();
    on_tableStatus_currentCellChanged(0,0,0,0);
    SetSelectedRow(fftTable.data(), 0);

    if (Mgr->ClockModules.Size())
       {
       TimerStatusInfo stat = Mgr->ClockModules.GetStatus()[0];
       ToControl(ui->TT_TimingTable, stat, 0);
       }
}

//---------------------------------------------------------------------------
//  MainWindow::on_CloseBtn_clicked() --  Close driver
//---------------------------------------------------------------------------

void MainWindow::on_CloseBtn_clicked()
{
    Mgr->Close();
    ControlEnabled(ui->OpenBtn, true);
    ControlEnabled(ui->CloseBtn, false);
    ControlEnabled(ui->StreamStartBtn, false);
    ControlEnabled(ui->StreamStopBtn, false);

    ControlEnabled(ui->BusMasterSizeEditBox, true);

    UpdateBoardsTable();
}

//---------------------------------------------------------------------------
//  MainWindow::on_StreamStartBtn_clicked() --  Start stream
//---------------------------------------------------------------------------

void MainWindow::on_StreamStartBtn_clicked()
{

    //  Activate stop button, deactivate start btn
    ControlEnabled(ui->OpenBtn, false);
    ControlEnabled(ui->CloseBtn, false);
    ControlEnabled(ui->StreamStartBtn, false);
    ControlEnabled(ui->StreamStopBtn, true);

    GetSettings();

    for (size_t i = 0; i < Mgr->FftChannels.ChannelGroup.size(); i++)  {
        Mgr->FftChannels.Wrapper(i).SetNumberOfDDC(Mgr->DdcChannels.Size());
        Mgr->FftChannels.Wrapper(i).SetFrameModeInfo(Mgr->DdcChannels.Cache.FrameMode);
    }

    // In case we are in Frame Mode calculate packet size
    if (Mgr->DdcChannels.Cache.FrameMode.mode == 1)  {
        Mgr->DrcSettings.PacketSize = Mgr->DdcChannels.CalculatePacketSize();
        ToControl(ui->PacketSizeEdit, Mgr->DrcSettings.PacketSize, 16);
        ControlEnabled(ui->PacketSizeEdit, false);
    } else {
        ControlEnabled(ui->PacketSizeEdit, true);
    }

    // Update the Sample to Log
    Mgr->DrcSettings.SamplesToLog = SamplingDurationToSamples(Mgr->DrcSettings.SamplingDuration);

    Mgr->StartStreaming();
}

//---------------------------------------------------------------------------
//  MainWindow::on_StreamStopBtn_clicked() --  Stop stream
//---------------------------------------------------------------------------

void MainWindow::on_StreamStopBtn_clicked()
{
    //
    //  Deectivate stop button, activate start btn
    ControlEnabled(ui->OpenBtn, false);
    ControlEnabled(ui->CloseBtn, true);
    ControlEnabled(ui->StreamStartBtn, true);
    ControlEnabled(ui->StreamStopBtn, false);
    ControlEnabled(ui->FftSetupUpdateButton, true);
    ControlEnabled(ui->PacketSizeEdit, true);

    Mgr->StopStreaming();
}


//===========================================================================
//  Support Methods
//===========================================================================

//---------------------------------------------------------------------------
//  MainWindow::Log() --
//---------------------------------------------------------------------------

void  MainWindow::Log(const std::string & a_string)
{
    AddToControl(ui->MessageLog, a_string);
}

//---------------------------------------------------------------------------
//  MainWindow::Status() --
//---------------------------------------------------------------------------

void  MainWindow::Status(const std::string & /*a_string*/)
{
}

//---------------------------------------------------------------------------
//  MainWindow::PeriodicStatus() --
//---------------------------------------------------------------------------

void MainWindow::PeriodicStatus(unsigned int board)
{
    unsigned int index = ReadControl<unsigned int>(ui->StatusTargetList);

    if (ui->PowerMeterGroup->isChecked())
        {
        PowerMeterProxy::PowerMeterGroupType ps = Mgr->PowerMeter.PowerMeters(index);

        for (size_t i = 0; i < Drc::AnalogInChannels(); i++)
            {
            float d1 = -std::numeric_limits<float>::infinity();
            if (ps.size())
                d1 = ps[0]->PowerData_dBm(i) + Mgr->DrcSettings.PowerMeterCalibrationOffset;
            ToControl(ui->PowerMeterTable, 0, i, d1);
            }
        }

    if (ui->ArithmeticOverflowGroup->isChecked())
        {
        // Overflow proxies for a given board
        OverflowProxy::OverflowGroupType os = Mgr->OverflowDetector.OverflowDetectors(index);

        for (size_t i = 0; i < os.size(); i++)
            {
            Innovative::MemoryBitManipulator BitManipulate(os[i]->OverflowStatus());
            bool b1 = BitManipulate.Bit(OverflowWrapper::cic_ovflo);
            bool b2 = BitManipulate.Bit(OverflowWrapper::cfir_ovflo);
            bool b3 = BitManipulate.Bit(OverflowWrapper::pfir_ovflo);
            bool b4 = BitManipulate.Bit(OverflowWrapper::channelizer_ovflo);
            bool b5 = BitManipulate.Bit(OverflowWrapper::fft_ovflo);
            bool b6 = BitManipulate.Bit(OverflowWrapper::fir1_ovflo);
            bool b0 = BitManipulate.Bit(OverflowWrapper::tuner_ovflo);
            bool b8 = BitManipulate.Bit(OverflowWrapper::ppf_ovflo);

            /// //////////////////
            ///OverFlow Table Data
            /// //////////////////
            CheckOverflowTableCell(b0,0,i);
            CheckOverflowTableCell(b4,1,i);
            CheckOverflowTableCell(b8,2,i);
            CheckOverflowTableCell(b1,3,i);
            CheckOverflowTableCell(b2,4,i);
            CheckOverflowTableCell(b3,5,i);
            CheckOverflowTableCell(b6,6,i);
            CheckOverflowTableCell(b5,7,i);

        }
        }

    int row = 0;
    for (size_t i = 0; i < Mgr->Settings.DRC_Count; i++)
        {
        if (Mgr->Settings.DRC_Targets_Enabled[i])
            {
            ToControl(ui->StreamStatusTable, 0, row, Mgr->Boards(i).SampleCount() / 1e6);
            ToControl(ui->StreamStatusTable, 1, row, static_cast<float>( Mgr->Boards(i).BlockRate()));
            ToControl(ui->StreamStatusTable, 2, row,  Mgr->Boards(i).FftSampleCount() / 1e6 );
            ToControl(ui->StreamStatusTable, 3, row, static_cast<float>( Mgr->Boards(i).FftBlockRate()));
            ToControl(ui->StreamStatusTable, 4, row,  Mgr->Boards(i).Temperature());

            //colorful pll lock cell
            if(Mgr->Boards(i).PllLocked())
            {
                ToControl(ui->StreamStatusTable, 5, row, string("+"));
                SetCell_BackgroundColor(ui->StreamStatusTable, 5, row,IIGreen);
            }
            else
            {
                ToControl(ui->StreamStatusTable, 5, row, string("-"));
                SetCell_BackgroundColor(ui->StreamStatusTable, 5, row,IIRed);
            }

            row++;
            }
        }

    if (Mgr->FftChannels.Size() && (board == SelectedFftBoard()) )
        {
        FFTGraph->setRate(Mgr->Boards(SelectedFftBoard()).FramePerSec());
        int dec = Mgr->FftChannels.Wrapper(SelectedFftIndex()).DecimationRate();
        FFTGraph->setBandwidth((Mgr->DrcSettings.SampleRate / dec) / 32768.0 * 1000);
        }


    if (Mgr->ClockModules.Size())
       {
       TimerStatusInfo stat = Mgr->ClockModules.GetStatus()[0];
       ToControl(ui->TT_TimingTable, stat, 0);
       }
}

//---------------------------------------------------------------------------
//  MainWindow::CheckOverflowTableCell() --
//---------------------------------------------------------------------------

void MainWindow::CheckOverflowTableCell(bool val, int col, int row)
{
    if(val)
    {
        ToControl(ui->OverflowTable, col, row, string("+"));
        SetCell_BackgroundColor(ui->OverflowTable, col, row,IIRed);
    }
    else
    {
        ToControl(ui->OverflowTable, col, row, string("-"));
        SetCell_BackgroundColor(ui->OverflowTable, col, row,IIGreen);
    }
}

//---------------------------------------------------------------------------
//  MainWindow::InstallSettings() --  Initialize Transfer Object
//---------------------------------------------------------------------------

void MainWindow::InstallSettings(Drc_ProviderSettings &Settings)
{
    //  Board Settings
    GL.Install( ToGui(ui->BusMasterSizeEditBox,           Settings.BusmasterSize) );
    GL.Install( ToGui(ui->DebugScriptEdit,                Settings.DebugScript) );
    //  Config Data
    GL.Install( ToGui(ui->PacketSizeEdit,                 Settings.PacketSize, 16) );
    GL.Install( ToGui(ui->SampleClockSourceRadioGroup,    Settings.SampleClockSource) );
    GL.Install( ToGui(ui->AdcSampleRateEdit,              Settings.SampleRate, 3) );
    GL.Install( ToGui(ui->ReferenceClockSourceRadioGroup, Settings.ReferenceClockSource) );
    GL.Install( ToGui(ui->ReferenceRateEdit,              Settings.ReferenceRate, 3) );
    GL.Install( ToGui(ui->AlertEnableCheckListBox,        Settings.AlertEnable) );
    GL.Install( ToGui(ui->ExtClockSrcSelectLB,            Settings.ExtClockSrcSelection) );


    //  Trigger
    GL.Install( ToGui(ui->TriggerSourceRadioGroup,       Settings.ExternalTrigger) );
    GL.Install( ToGui(ui->TriggerSignalTypeRadioGroup,   Settings.EdgeTrigger) );
    GL.Install( ToGui(ui->ExtTriggerSrcSelectLB,         Settings.ExtTriggerSrcSelection) );
  //  GL.Install( ToGui(ui->TxTriggerModeListBox,          Settings.Framed) );
  //  GL.Install( ToGui(ui->TxFrameSizeEdit,               Settings.FrameSize, 16) );


    //  Streaming and log
    GL.Install( ToGui(ui->LogEnableChk,                   Settings.LoggerEnable) );
    GL.Install( ToGui(ui->StreamLogForeverCheckBox,       Settings.ContinuousLogging));
    GL.Install( ToGui(ui->PlotEnableChk,                  Settings.PlotEnable) );
    GL.Install( ToGui(ui->SamplingDurationEdit,           Settings.SamplingDuration, 3) );
    GL.Install( ToGui(ui->AutoStopCheckBox,               Settings.AutoStop) );
    GL.Install( ToGui(ui->OverwriteBddCheckBox,           Settings.OverwriteBdd) );
    GL.Install( ToGui(ui->UtcCheckBox,                    Settings.Utc) );
    GL.Install( ToGui(ui->ExtPpsCheckBox,                 Settings.ExternalPps) );
    GL.Install( ToGui(ui->TimeStampledFilenameCheckBox,   Settings.EnableDateTimeLogging) );

    GL.Install( ToGui(ui->PowerMeterCalibrationOffsetEdit,Settings.PowerMeterCalibrationOffset) );
}

//---------------------------------------------------------------------------
//  MainWindow::InstallClockSettings() --
//---------------------------------------------------------------------------

void MainWindow::InstallSettings(ClockGUISettings &Settings)
{
    //
    GL.Install( ToGui(ui->TT_ReferenceRateED,       Settings.ReferenceRate) );
    GL.Install( ToGui(ui->TT_ReferenceClockSourceLB,Settings.RefSource) );
    GL.Install( ToGui(ui->TT_ExtClockSourceLB,      Settings.ExtClockSrc) );

    GL.Install( ToGui(ui->TT_ExtTriggerSourceLB,    Settings.ExternalTrigger) );
    GL.Install( ToGui(ui->TT_TypeLB,                Settings.TriggerType) );
    GL.Install( ToGui(ui->TT_PPSSourceLB,           Settings.PpsSrc) );
    //
    GL.Install( ToGui(ui->TT_SampleClockRateED,     Settings.Frequency, 3) );
    GL.Install( ToGui(ui->TT_SampleClockSourceLB,   Settings.SampleSource) );
    //  Trigger Page
    GL.Install( ToGui(ui->TT_TriggerSourceLB,       Settings.ExternalTrigger) );
    GL.Install( ToGui(ui->TT_FrameModeLB,           Settings.Framed) );
    GL.Install( ToGui(ui->TT_TriggerInEdit,         Settings.TriggerInSec) );
    GL.Install( ToGui(ui->TT_FrameSizeED,           Settings.FrameSize, 16) );
}

//---------------------------------------------------------------------------
//  MainWindow::InstallComponentSettings() --  Initialize Transfer Object
//---------------------------------------------------------------------------

void MainWindow::InstallComponentSettings()
{
    // Ddc component
    GL.Install( ToGui(ui->DDCOutputDataFmtListBx,         Mgr->DdcChannels.BypassDdc) );
    GL.Install( ToGui(ui->FrameModeFreqsListFileEdit,         Mgr->DdcChannels.ScheduleFreqsFile) );

    map<string, int> m;
    vector<string>   s;
    // Add on-off map
    m.insert(make_pair(string("On"),  1));
    m.insert(make_pair(string("Off"), 0));
    // Add DDC sources
    s = DdcSourceList();
    for (size_t i = 0 ; i < s.size(); i++)
        m.insert(make_pair(s[i], i));

    for (size_t i = 0; i < Mgr->DdcChannels.Size(); i++)
        GL.Install( ToGui(ddcTable.data(), Mgr->DdcChannels[i]->info, make_pair(s, i), make_pair(m, i)) );
    // Cache
    GL.Install( ToGui(ui->DDCSourceSelectComboBox_2,       Mgr->DdcChannels.Cache.Source) );
    GL.Install( ToGui(ui->DDCChanConfigFileEditBx_2,       Mgr->DdcChannels.Cache.File) );
    GL.Install( ToGui(ui->DDCChanTuningFreqEditBx_2,       Mgr->DdcChannels.Cache.Frequency, 3) );
    GL.Install( ToGui(ui->DDCSpecInvCheckBox_2,            Mgr->DdcChannels.Cache.SpectrumInvert) );

    GL.Install( ToGui(ui->FrameModeListBox,                Mgr->DdcChannels.Cache.FrameMode.mode) );
    GL.Install( ToGui(ui->FrameModeSizeEdit,               Mgr->DdcChannels.Cache.FrameMode.sz) );
    GL.Install( ToGui(ui->FrameModeDelayEdit,              Mgr->DdcChannels.Cache.FrameMode.delay) );
    GL.Install( ToGui(ui->FrameModeSchedulerCombo,         Mgr->DdcChannels.Cache.FrameMode.sched) );
    GL.Install( ToGui(ui->FrameModeEnableTuningScheduleCheck, Mgr->DdcChannels.Cache.FrameMode.en) );

    // Tone Generator
    for (size_t i = 0; i < Mgr->ToneGenerator.Size(); i++)
        GL.Install( ToGui(ui->DDCTestSignalTable, Mgr->ToneGenerator.Proxy(i).tones[0], 0, 0) ); // FIXME
    // FFT setup table
    for (size_t i = 0; i < Mgr->FftChannels.Size(); i++)
        GL.Install( ToGui(fftTable.data(),      Mgr->FftChannels.Wrapper(i).Info, i,
                          InitMaps(Mgr->FftChannels.Wrapper(i).WhichBoard()), i) );

    // FFT plot table
    //fixme
//    for (size_t i = 0; i < Mgr->FftChannels.Size(); i++)
//        {
//        vector<string> SrcList = Mgr->FftSources(Mgr->FftChannels.Wrapper(i).WhichBoard());
//        GL.Install( ToGui(ui->FftPlotTableWidget,      Mgr->FftChannels.Wrapper(i).Info, i, SrcList, i) );
//        }

    // Cache
    GL.Install( ToGui(ui->FftSetupInputSourceCombo,          Mgr->FftChannels.Cache.Source) );
    GL.Install( ToGui(ui->FftSetupWindowingCombo,            Mgr->FftChannels.Cache.Window) );
    GL.Install( ToGui(ui->FftSetupFastAttackCheckBox,        Mgr->FftChannels.Cache.FasdEnable) );
    GL.Install( ToGui(ui->FftSetupFastAttackEdit,            Mgr->FftChannels.Cache.Fasd, 4) );
    GL.Install( ToGui(ui->FftSetupOutputSpectrumComboBox,    Mgr->FftChannels.Cache.OutputMode) );
    GL.Install( ToGui(ui->FftSetupPeakMonitorFileEdit,       Mgr->FftChannels.Cache.ThresholdFile) );
    GL.Install( ToGui(ui->FftSetupFramePerSecondEdit,        Mgr->FftChannels.Cache.Fps, 3) );
    GL.Install( ToGui(ui->FftSetupTriggerModeComboBox,       Mgr->FftChannels.Cache.TriggerMode) );
}

//---------------------------------------------------------------------------
//  MainWindow::GetSettings() --  Form controls to options object
//---------------------------------------------------------------------------

void MainWindow::GetSettings()
{
    Mgr->DrcSettings.EnablePowerMeterReads      = ui->PowerMeterGroup->isChecked();
    Mgr->DrcSettings.EnableArithmeticOflowReads = ui->ArithmeticOverflowGroup->isChecked();

    GL.GetSettings();    // alternate name
}

//---------------------------------------------------------------------------
//  MainWindow::SetSettings() --  Settings object to form controls
//---------------------------------------------------------------------------

void MainWindow::SetSettings()
{
    ui->PowerMeterGroup->setChecked(Mgr->DrcSettings.EnablePowerMeterReads);
    ui->ArithmeticOverflowGroup->setChecked(Mgr->DrcSettings.EnableArithmeticOflowReads);

    GL.SetSettings();
}

//---------------------------------------------------------------------------
//  MainWindow::SetupBoardsTable() --
//---------------------------------------------------------------------------

void MainWindow::SetupBoardsTable()
{
    GetSettings();

    Mgr->AdjustTargets();

    // Row headers
    vector<std::string> StatusVerticalHeader;
    Mgr->BoardList(StatusVerticalHeader, false);
    // Coloumn headers
    BoardStatusTableInit(ui->tableStatus, StatusVerticalHeader);
    // Add rows content for each board
    UpdateBoardsTable();
}

//---------------------------------------------------------------------------
//  MainWindow::SamplingDurationToSamples() --
//---------------------------------------------------------------------------

ii64 MainWindow::SamplingDurationToSamples(float duration)
{
    ii64 samples = Mgr->SamplingDurationToSamples(duration);

    stringstream str;
    str << "IsIQEnabled: " << Mgr->Boards(0).IsIQEnabled() << " SampleWidth: " << Mgr->Boards(0).SampleWidth();
    Log(str.str());

    return samples;
}

//---------------------------------------------------------------------------
//  MainWindow::BeforeStreamStart() --
//---------------------------------------------------------------------------

void MainWindow::BeforeStreamStart(unsigned int /*ordinal*/)
{
    // SetSettings will be called by this
    if (Mgr->FftChannels.Size())
        on_FftSetupInputSourceCombo_currentIndexChanged(Mgr->FftChannels.Cache.Source);
}

//---------------------------------------------------------------------------
//  MainWindow::AfterStreamStart() --
//---------------------------------------------------------------------------

void MainWindow::AfterStreamStart(unsigned int )
{
}

//---------------------------------------------------------------------------
//  MainWindow::AfterStreamAutoStop() --
//---------------------------------------------------------------------------

void MainWindow::AfterStreamAutoStop(unsigned int )
{
    //
    //  Deectivate stop button, activate start btn
    ControlEnabled(ui->OpenBtn, false);
    ControlEnabled(ui->CloseBtn, true);
    ControlEnabled(ui->StreamStartBtn, true);
    ControlEnabled(ui->StreamStopBtn, false);
    ControlEnabled(ui->FftSetupUpdateButton, true);
}


//---------------------------------------------------------------------------
//  MainWindow::UpdateBoardsTable() --
//---------------------------------------------------------------------------

void MainWindow::UpdateBoardsTable()
{
    for (unsigned int i = 0; i < Mgr->Settings.DRC_Count; i++)
        {
        UpdateBoardEntry(ui->tableStatus,
          BoardStatusInfo(
             Mgr->Settings.DRC_Targets[i],       Mgr->Settings.DRC_Targets_Enabled[i],
             Mgr->IsOpen(i) ? "Open" : "Closed", Mgr->CoreVersion(i).first,
             Mgr->CoreVersion(i).second,         Mgr->LogicVersion(i), Mgr->Boards(i).DdcCoresCount(),
             Mgr->Boards(i).DucCoresCount(),     Mgr->Boards(i).FftCoresCount()),
          i);
        }
}

//---------------------------------------------------------------------------
//  Other debug tab methods
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//  MainWindow::LoadPowerMeterTargetSelector()
//---------------------------------------------------------------------------

void  MainWindow::LoadPowerMeterTargetSelector()
{
    ClearControl(ui->StatusTargetList);

    std::vector<std::string> target_names;
    Mgr->GetFiclTargets(target_names);
    for (unsigned int i=0; i<target_names.size(); i++)
        {
        AddToControl(ui->StatusTargetList, target_names[i]);
        }
    if (target_names.size())
        ToControl(ui->StatusTargetList, 0);

}

//---------------------------------------------------------------------------
//  FICL debug tab interface and other methods
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//  MainWindow::LoadFiclTargetSelector()
//---------------------------------------------------------------------------

void  MainWindow::LoadFiclTargetSelector()
{
    ClearControl(ui->FiclTargetList);

    std::vector<std::string> target_names;
    Mgr->GetFiclTargets(target_names);

    for (unsigned int i=0; i < Mgr->Settings.DRC_Count; i++)
        {
        if(Mgr->Settings.DRC_Targets_Enabled[i])
            AddToControl(ui->FiclTargetList, target_names[i]);
        }

    if (target_names.size())
        ToControl(ui->FiclTargetList, 0);
}

//---------------------------------------------------------------------------
//  MainWindow::on_FiclCommandEdit_returnPressed() --
//---------------------------------------------------------------------------

void MainWindow::on_FiclCommandEdit_returnPressed()
{
    char Key = 0x0d;

    Mgr->Ficl_OnCommandKeyPress(ReadControl<unsigned int>(ui->FiclTargetList), Key);
}

//-----------------------------------------------------------------------------
//  MainWindow::FiclReadCommandString()  --
//-----------------------------------------------------------------------------

std::string MainWindow::FiclReadCommandString()
{
    return ReadControl<std::string>(ui->FiclCommandEdit);
}

//-----------------------------------------------------------------------------
//  MainWindow::FiclCommandComplete()  --
//-----------------------------------------------------------------------------

void MainWindow::FiclCommandComplete()
{
    ClearControl(ui->FiclCommandEdit);
}

//-----------------------------------------------------------------------------
//  MainWindow::on_FiclCommandEdit_textEdited()  --
//-----------------------------------------------------------------------------

void MainWindow::on_FiclCommandEdit_textEdited(QString )
{
    //Mgr->FiclIo_OnCommandKeyPress(Key);
}

//---------------------------------------------------------------------------
//  MainWindow::on_DebugScriptBrowseButton_clicked() --  Select debug script
//---------------------------------------------------------------------------

void MainWindow::on_DebugScriptBrowseButton_clicked()
{
    SelectFileDialog(ui->DebugScriptEdit, "Script Files (*.fr *.*");
}

//---------------------------------------------------------------------------
//  MainWindow::on_DebugScriptExecuteButton_clicked() --  Execute debug script
//---------------------------------------------------------------------------

void MainWindow::on_DebugScriptExecuteButton_clicked()
{
    Mgr->Ficl_ExecuteDebugScript(ReadControl<unsigned int>(ui->FiclTargetList),
                                ReadControl<std::string>(ui->DebugScriptEdit));
}

//---------------------------------------------------------------------------
//  MainWindow::on_LogicReadButton_clicked() --  Read logic port
//---------------------------------------------------------------------------

void MainWindow::on_LogicReadButton_clicked()
{
    Mgr->Ficl_Logic(ReadControl<unsigned int>(ui->FiclTargetList),
                    ReadControl<std::string>(ui->AddressEdit));
}

//---------------------------------------------------------------------------
//  MainWindow::on_LogicWriteButton_clicked() --  Write logic port
//---------------------------------------------------------------------------

void MainWindow::on_LogicWriteButton_clicked()
{
    Mgr->Ficl_Logic(ReadControl<unsigned int>(ui->FiclTargetList),
                    ReadControl<std::string>(ui->AddressEdit),
                    ReadControl<std::string>(ui->DataEdit));
}

//---------------------------------------------------------------------------
//  MainWindow::on_PortReadButton_clicked() --  Read PCI port
//---------------------------------------------------------------------------

void MainWindow::on_PortReadButton_clicked()
{
    Mgr->Ficl_Port(ReadControl<unsigned int>(ui->FiclTargetList),
                    ReadControl<std::string>(ui->AddressEdit));
}

//---------------------------------------------------------------------------
//  MainWindow::on_PortWriteButton_clicked() --  Write PCI port
//---------------------------------------------------------------------------

void MainWindow::on_PortWriteButton_clicked()
{
    Mgr->Ficl_Port(ReadControl<unsigned int>(ui->FiclTargetList),
                    ReadControl<std::string>(ui->AddressEdit),
                    ReadControl<std::string>(ui->DataEdit));
}

//---------------------------------------------------------------------------
//  MainWindow::on_DeviceReadButton_clicked() --  Read device port
//---------------------------------------------------------------------------

void MainWindow::on_DeviceReadButton_clicked()
{
    Mgr->Ficl_Device(ReadControl<unsigned int>(ui->FiclTargetList),
                     ReadControl<std::string>(ui->AddressEdit),
                     ReadControl<std::string>(ui->DeviceEdit));
}

//---------------------------------------------------------------------------
//  MainWindow::on_DeviceWriteButton_clicked() --  Write device port
//---------------------------------------------------------------------------

void MainWindow::on_DeviceWriteButton_clicked()
{
    Mgr->Ficl_Device(ReadControl<unsigned int>(ui->FiclTargetList),
                     ReadControl<std::string>(ui->AddressEdit),
                     ReadControl<std::string>(ui->DeviceEdit),
                     ReadControl<std::string>(ui->DataEdit));
}

//---------------------------------------------------------------------------
//  MainWindow::on_tableStatus_cellChanged() --
//---------------------------------------------------------------------------

void MainWindow::on_tableStatus_cellChanged(int row, int )
{

    if (row >= 0 && row < Mgr->BoardCount())
        {
        ReadBoardEntry(ui->tableStatus, Mgr->Settings.DRC_Targets[row], Mgr->Settings.DRC_Targets_Enabled[row], row);

        Mgr->Boards(row).Personality.Target  = Mgr->Settings.DRC_Targets[row];
        Mgr->Boards(row).Personality.Enabled = Mgr->Settings.DRC_Targets_Enabled[row];
        }
}

//---------------------------------------------------------------------------
//  MainWindow::on_DDCTestSignalUpdateButton_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_DDCTestSignalUpdateButton_clicked()
{
    GetSettings();

    // Adc and Dac sample rates are the same
    Mgr->ToneGenerator.SamplingRate(Mgr->DrcSettings.SampleRate, Mgr->DrcSettings.SampleRate);
    Mgr->ToneGenerator.ApplyCached();

    Mgr->CommandTrigger(GpDspManager::eCommon);

}

//---------------------------------------------------------------------------
//  MainWindow::on_DDCChanConfigFileSelBtn_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_DDCChanConfigFileSelBtn_clicked()
{
    SelectFileDialog(ui->DDCChanConfigFileEditBx_2, "Filter INI Files (*.ini *.INI)");
}

//---------------------------------------------------------------------------
//  MainWindow::on_DDCUpdateChanButton_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_DDCUpdateChanButton_clicked()
{
    vector<int> rows;
    SelectedRows(ddcTable.data(), rows);
    GL.Invalidate();
    GetSettings();
    if (WarnChannelSelect(rows))
       {
       Mgr->DdcChannels.Apply(rows);
       }
    else
        return;

    SetSettings();

    // DDC setting have changed, rescale the plot axis for the new values
    if ( Mgr->FftChannels.Size() && ((size_t)Mgr->FftChannels.Cache.Source < Mgr->DdcChannels.Size()) )
        {
        // Load FFT with new decimation
        Mgr->DdcChannels.Wrapper(Mgr->FftChannels.Cache.Source).ReloadFile();
        int decrate = Mgr->DdcChannels.DecimationCoeffs(Mgr->FftChannels.Cache.Source);
        Mgr->FftChannels.Wrapper(SelectedFftIndex()).DecimationRate(decrate);

        EmbeddedCombo_currentIndexChangedEmitCall(QString("0-0"),false);
        }
}

//---------------------------------------------------------------------------
//  MainWindow::WarnChannelSelect() --
//---------------------------------------------------------------------------

bool MainWindow::WarnChannelSelect(const vector<int> &rows)
{
    if (rows.size())
        {
        return true;
        }
    else
        {
        Log("Please select channel(s) to apply");
        return false;
        }
}

//---------------------------------------------------------------------------
//  MainWindow::DdcTableWidgetCurrentCellChanged() --
//---------------------------------------------------------------------------

void MainWindow::DdcTableCurrentCellChanged(int currentRow, int /*currentColumn*/, int /*previousRow*/, int /*previousColumn*/)
{
    GL.Invalidate();
    Mgr->DdcChannels.Cache = Mgr->DdcChannels[currentRow]->info ;

    ToControl(ui->DDCSourceSelectComboBox_2,       Mgr->DdcChannels.Cache.Source);
    ToControl(ui->DDCChanConfigFileEditBx_2,       Mgr->DdcChannels.Cache.File);
    ToControl(ui->DDCChanTuningFreqEditBx_2,       Mgr->DdcChannels.Cache.Frequency, 3);
    ToControl(ui->DDCSourceSelectComboBox_2,       Mgr->DdcChannels.Cache.Source);
    ToControl(ui->DDCSpecInvCheckBox_2,            Mgr->DdcChannels.Cache.SpectrumInvert);

    ControlEnabled(ui->DdcChannelEnableButton, !Mgr->DdcChannels.Cache.Enable);
    ControlEnabled(ui->DdcChannelDisableButton, Mgr->DdcChannels.Cache.Enable);
}

//---------------------------------------------------------------------------
//  MainWindow::on_StreamLogForeverCheckBox_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_StreamLogForeverCheckBox_clicked(bool checked)
{
    if (checked)
    {
        ToControl(ui->AutoStopCheckBox, false);
        ControlEnabled(ui->SamplingDurationEdit, false);
        ControlEnabled(ui->AutoStopCheckBox, false);
    }
    else
    {
        ControlEnabled(ui->SamplingDurationEdit, true);
        ControlEnabled(ui->AutoStopCheckBox, true);
    }
}

//---------------------------------------------------------------------------
//  MainWindow::ChannelEnable() --
//---------------------------------------------------------------------------

void MainWindow::ChannelEnable(bool value)
{
    GL.Invalidate();
    Mgr->DdcChannels.Cache.Enable = value;

    vector<int> rows;
    SelectedRows(ddcTable.data(), rows);
    GetSettings();
    if (WarnChannelSelect(rows))
       {
       Mgr->DdcChannels.ApplyEnable(rows);
       }
    SetSettings();

    ControlEnabled(ui->DdcChannelEnableButton, !Mgr->DdcChannels.Cache.Enable);
    ControlEnabled(ui->DdcChannelDisableButton, Mgr->DdcChannels.Cache.Enable);
}

//---------------------------------------------------------------------------
//  MainWindow::on_DdcChannelEnableButton_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_DdcChannelEnableButton_clicked()
{
    ChannelEnable(true);
}

//---------------------------------------------------------------------------
//  MainWindow::on_DdcChannelDisableButton_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_DdcChannelDisableButton_clicked()
{
    ChannelEnable(false);
}

//---------------------------------------------------------------------------
//  MainWindow::on_FftSetupPeakMonitorFileSelectButton_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_FftSetupPeakMonitorFileSelectButton_clicked()
{
    SelectFileDialog(ui->FftSetupPeakMonitorFileEdit, "Filter INI Files (*.ini *.INI)");
}

//---------------------------------------------------------------------------
//  MainWindow::on_FftSetupUpdateButton_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_FftSetupUpdateButton_clicked()
{
    vector<int> rows;
    SelectedRows(fftTable.data(), rows);
    GetSettings();
    if (WarnChannelSelect(rows))
       {
       if (Mgr->FftChannels.Cache.Source < static_cast<int>(Mgr->DdcChannels.Size()))
           {
            Mgr->DdcChannels.Wrapper(Mgr->FftChannels.Cache.Source).ReloadFile();
            vector<float> cfs = Mgr->DdcChannels.DecimationCoeffs();
            int decrate = cfs.size() ? cfs[Mgr->FftChannels.Cache.Source] : 1;
            Mgr->FftChannels.Wrapper(SelectedFftIndex()).DecimationRate(decrate);
           }
       else
           {
           Mgr->FftChannels.Wrapper(SelectedFftIndex()).DecimationRate(1);
           }

       Mgr->FftChannels.SamplingRate(Mgr->DrcSettings.SampleRate);
       Mgr->FftChannels.Wrapper(SelectedFftIndex()).SetNumberOfDDC(Mgr->DdcChannels.Size());
       Mgr->FftChannels.Apply(rows);
       }

    SetSettings();
}

//---------------------------------------------------------------------------
//  MainWindow::Plot() --
//---------------------------------------------------------------------------

void MainWindow::Plot(unsigned int /*board*/, VeloBuffer &buffer, int type)
{
    if (type == 0)
        FFTGraph->setFftSamples(0,  buffer);
    else
        FFTGraph->setTlsSamples(1,  buffer);
}

//---------------------------------------------------------------------------
//  MainWindow::DataPacketAvailable() --
//---------------------------------------------------------------------------

void MainWindow::DataPacketAvailable(unsigned int /*ordinal*/,VeloBuffer &buffer)
{
   VeloHeaderDatagram Hdr(buffer);
   if (Hdr.PeripheralId() == 0) {
        if  ((Mgr->DdcChannels.Cache.FrameMode.sched == 0) &&
              (Mgr->DdcChannels.Cache.FrameMode.mode == 1))
            Mgr->DdcChannels.Strobe();
    }
}

//---------------------------------------------------------------------------
//  MainWindow::PopulateFftSourceCombo() --
//---------------------------------------------------------------------------

void MainWindow::PopulateFftSourceCombo(unsigned int ordinal)
{
    // disconnect signal
    ui->FftSetupInputSourceCombo->disconnect(this);

    vector<string> SrcList = Mgr->FftSources(ordinal);
    ClearControl(ui->FftSetupInputSourceCombo);
    FillControl(ui->FftSetupInputSourceCombo, SrcList);

    // reconnect signal
    connect(ui->FftSetupInputSourceCombo, SIGNAL(currentIndexChanged(int)),
            this, SLOT(on_FftSetupInputSourceCombo_currentIndexChanged(int)));
}

//---------------------------------------------------------------------------
//  MainWindow::DrawSelectedFftRow() --
//---------------------------------------------------------------------------

void MainWindow::DrawSelectedFftRow(int crow)
{

    vector<string> rows;

    rows.push_back(Mgr->FftChannels.Wrapper(crow).Tag());
    FftPlotTableInit(ui->FftPlotTableWidget, rows);

    int srcidx = ReadControl<int>(ui->FftSetupInputSourceCombo);
    int wndidx = ReadControl<int>(ui->FftSetupWindowingCombo);

    // Windowing names
    vector<string> WindowingList;
    for (size_t i = 0; i < sizeof(win_coeff_info) / sizeof(win_coeff_info[0]); i++)
        WindowingList.push_back(win_coeff_info[i].name);

    vector<string> FftSourceList = Mgr->FftSources(Mgr->FftChannels.Wrapper(crow).WhichBoard());

    // Add columns to table
    AddComboWidgetToTableWidget(0, 0, FftSourceList, srcidx);
    AddComboWidgetToTableWidget(0, 1, WindowingList, wndidx);
    // Trigger a source change event so FFTGraph will adjust its x-axis
    FftSetupInputSourceComboEmitCall(srcidx,false);

}

//---------------------------------------------------------------------------
//  MainWindow::SelectedFftIndex() --
//---------------------------------------------------------------------------

unsigned int MainWindow::SelectedFftIndex()
{
    return SelectedFftRowIndex;
}

//---------------------------------------------------------------------------
//  MainWindow::SelectedFftBoard() --
//---------------------------------------------------------------------------

unsigned int MainWindow::SelectedFftBoard()
{
    return Mgr->FftChannels.Wrapper(SelectedFftIndex()).WhichBoard();
}


//---------------------------------------------------------------------------
//  MainWindow::on_FftSetupManualTriggerButton_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_FftSetupManualTriggerButton_clicked()
{
    GetSettings();
    Mgr->FftChannels.SingleTrigger();
    SetSettings();
}

//---------------------------------------------------------------------------
//  MainWindow::on_FftSetupInputSourceCombo_currentIndexChanged() --
//---------------------------------------------------------------------------

void MainWindow::on_FftSetupInputSourceCombo_currentIndexChanged(int idx)
{
    FftSetupInputSourceComboEmitCall(idx,true);
}


//---------------------------------------------------------------------------
//  MainWindow::on_FftSetupWindowingCombo_currentIndexChanged() --
//---------------------------------------------------------------------------

void MainWindow::on_FftSetupWindowingCombo_currentIndexChanged(int idx)
{
    FftSetupWindowingComboEmitCall(idx,true);
}

//---------------------------------------------------------------------------
//  MainWindow::on_FftSetupTableWidget_currentCellChanged() --
//---------------------------------------------------------------------------

void MainWindow::FftSetupTableCurrentCellChanged(int crow, int , int , int )
{
    SelectedFftRowIndex = crow;

    PopulateFftSourceCombo(SelectedFftRowIndex);

    for (size_t i = 0; i < Mgr->FftChannels.Size(); i++)
        {
        // Put the selected fft to contionuous mode while disabling others
        Mgr->FftChannels.Cache.TriggerMode = (crow == static_cast<int>(i)) ? 2 : 0;
        Mgr->FftChannels.TriggerModeApply(i);
        }
    Mgr->FftChannels.Cache = Mgr->FftChannels.Wrapper(crow).Info;

    SetSettings();

    DrawSelectedFftRow(crow);
}

//---------------------------------------------------------------------------
//  MainWindow::on_FftSetupOutputSpectrumComboBox_currentIndexChanged() --
//---------------------------------------------------------------------------

void MainWindow::on_FftSetupOutputSpectrumComboBox_currentIndexChanged(int index)
{
   if (index < 0)
        return;

   FFTGraph->clearAll();
   FFTGraph->showSamples(index, true);
}

//---------------------------------------------------------------------------
//  MainWindow::on_SnapshotButton_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_SnapshotButton_clicked()
{
    FFTGraph->takeSnapshot("snap.pdf");
}

//---------------------------------------------------------------------------
//  MainWindow::on_ArithmeticOverflowGroup_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_ArithmeticOverflowGroup_clicked(bool checked)
{
    if (checked && Mgr != 0)
        {
        // Setup Overflow table
        vector<std::string> OverflowTableVerticalHeader;
        Mgr->OverflowDetector.OverflowDetectorList(OverflowTableVerticalHeader, 0);
        OverflowTableInit(ui->OverflowTable, OverflowTableVerticalHeader);
        }
}

//---------------------------------------------------------------------------
//  MainWindow::on_PowerMeterGroup_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_PowerMeterGroup_clicked()
{
    GetSettings();
}

//---------------------------------------------------------------------------
//  MainWindow::on_ArithmeticOverflowGroup_clicked() --
//---------------------------------------------------------------------------

void MainWindow::on_ArithmeticOverflowGroup_clicked()
{
    GetSettings();
}

//---------------------------------------------------------------------------
//  MainWindow::on_Tabs_currentChanged() --
//---------------------------------------------------------------------------

void MainWindow::on_Tabs_currentChanged(int index)
{
    if (index == ui->Tabs->indexOf(ui->FftTab))
       FFTGraph->resize();
}

//---------------------------------------------------------------------------
//  MainWindow::BeforeDrcOpen() --
//---------------------------------------------------------------------------

void  MainWindow::BeforeDrcOpen()
{

}

//---------------------------------------------------------------------------
//  MainWindow::AfterDrcOpen() --
//---------------------------------------------------------------------------

void  MainWindow::AfterDrcOpen()
{

}

//---------------------------------------------------------------------------
//  MainWindow::BeforeDrcClose() --
//---------------------------------------------------------------------------

void  MainWindow::BeforeDrcClose()
{

}

//---------------------------------------------------------------------------
//  MainWindow::AfterDrcClose() --
//---------------------------------------------------------------------------

void  MainWindow::AfterDrcClose()
{

}

//---------------------------------------------------------------------------
//  MainWindow::AfterDrcStreamStart() --
//---------------------------------------------------------------------------

void  MainWindow::AfterDrcStreamStart()
{

}

//---------------------------------------------------------------------------
//  MainWindow::BeforeDrcStreamStart() --
//---------------------------------------------------------------------------

void  MainWindow::BeforeDrcStreamStart()
{

}

//---------------------------------------------------------------------------
//  MainWindow::AfterDrcStreamStop() --
//---------------------------------------------------------------------------

void  MainWindow::AfterDrcStreamStop()
{
    //disable software trigger
    Mgr->ClockModules.SoftwareTrigger(false);
    ControlEnabled(ui->TT_SoftwareTriggerEnableB, true);
    ControlEnabled(ui->TT_SoftwareTriggerDisableB, false);
}

//---------------------------------------------------------------------------
//  MainWindow::BeforeDrcStreamStop() --
//---------------------------------------------------------------------------

void  MainWindow::BeforeDrcStreamStop()
{

}

//---------------------------------------------------------------------------
//  MainWindow::on_tableStatus_currentCellChanged() --
//---------------------------------------------------------------------------

void MainWindow::on_tableStatus_currentCellChanged(int currentRow, int /*currentColumn*/,
                                                   int /*previousRow*/, int /*previousColumn*/)
{
    if(Mgr->Boards(currentRow).IsOpen())
    {
        ToControl(ui->LogicVersionEdit,Mgr->Boards(currentRow).FpgaLogicVersion());
        ToControl(ui->PciRevisionEdit,Mgr->Boards(currentRow).PciLogicRevision(),16);
        ToControl(ui->PCIExpressLinesEdit, Mgr->Boards(currentRow).PCIExpressLanes());
        ToControl(ui->BoardFamilyEdit,Mgr->Boards(currentRow).PciLogicFamily());
        ToControl(ui->BoardTypeEdit,Mgr->Boards(currentRow).PciLogicType());
        ToControl(ui->BoardRevisionEdit,Mgr->Boards(currentRow).PciLogicPcb());
        ToControl(ui->BoardChipEdit, Mgr->Boards(currentRow).FpgaName());
        ToControl(ui->HdwVariantEdit,Mgr->Boards(currentRow).FpgaHardwareVariant(),16);
    }

}

//---------------------------------------------------------------------------
//  MainWindow::FftSetupInputSourceComboEmitCall() --
//---------------------------------------------------------------------------

void MainWindow::FftSetupInputSourceComboEmitCall(int idx, bool update)
{
    Mgr->FftChannels.Cache.Source = idx;

    FFTGraph->setSamplingRate(Mgr->DrcSettings.SampleRate);


    bool TestSelected = Mgr->FftSources(SelectedFftBoard()).size()-1 == (size_t)idx;

    if (Mgr->FftChannels.Cache.Source >= static_cast<int>(Mgr->DdcChannels.ChannelCount(SelectedFftBoard())) || TestSelected)
       { // ADC
       // Update Fft Channel
       Mgr->UpdateAdcInfo(SelectedFftIndex());
       // Update FFT Graph
       FFTGraph->setDecimationRate(1);
       FFTGraph->setRealAxis();
       }
    else
       { // DDC
        // Update Fft Channel
       Mgr->UpdateDdcInfo(SelectedFftIndex());
       // Update FFT Graph
       int decrate = Mgr->DdcChannels.DecimationCoeffs()[Mgr->FftChannels.Cache.Source];
       FFTGraph->setDecimationRate(decrate);
       FFTGraph->setComplexAxis();
       }

    if (psignalMapper)
        {
        // Change the table combo index
        QComboBox* combo = (QComboBox*)ui->FftPlotTableWidget->cellWidget(SelectedFftIndex(), 0);
        psignalMapper->disconnect(this);
        if (combo) combo->setCurrentIndex(idx);
        // reconnect
        connect(psignalMapper, SIGNAL(mapped(const QString &)), this,
                SLOT(EmbeddedCombo_currentIndexChanged(const QString &)));
        }

     if(update)
        SetSettings();
}

//---------------------------------------------------------------------------
//  MainWindow::FftSetupInputSourceComboEmitCall() --
//---------------------------------------------------------------------------

void MainWindow::FftSetupWindowingComboEmitCall(int idx, bool update)
{
    Mgr->FftChannels.Cache.Window = idx;

    Mgr->FftChannels.WindowChangeApply(SelectedFftIndex());

    if (psignalMapper)
        {
        // Change the table combo index
        QComboBox* combo = (QComboBox*)ui->FftPlotTableWidget->cellWidget(SelectedFftIndex(), 1);
        psignalMapper->disconnect(this);
        if (combo) combo->setCurrentIndex(idx);
        // reconnect
        connect(psignalMapper, SIGNAL(mapped(const QString &)), this,
                SLOT(EmbeddedCombo_currentIndexChanged(const QString &)));
        }

    if(update) SetSettings();
}

//---------------------------------------------------------------------------
//  MainWindow::EmbeddedCombo_currentIndexChangedEmitCall() --
//---------------------------------------------------------------------------

void MainWindow::EmbeddedCombo_currentIndexChangedEmitCall(QString position, bool /*update*/)
{
    QStringList coordinates = position.split("-");
    int row = coordinates[0].toInt();
    int col = coordinates[1].toInt();
    QComboBox* combo = (QComboBox*)ui->FftPlotTableWidget->cellWidget(row, col);

    // Change the combo index on the setup tab
    if (combo && (col == 0))
        FftSetupInputSourceComboEmitCall(combo->currentIndex(),false);
    else if (combo && (col == 1))
        FftSetupWindowingComboEmitCall(combo->currentIndex(), false);


     Mgr->FftChannels.Wrapper(SelectedFftIndex()).SetNumberOfDDC(Mgr->DdcChannels.Size());
     Mgr->FftChannels.Wrapper(SelectedFftIndex()).ApplyFrameDecimation();
}

void MainWindow::on_TT_SoftwareTriggerEnableB_clicked()
{
    Mgr->ClockModules.SoftwareTrigger(true);
    ControlEnabled(ui->TT_SoftwareTriggerEnableB, false);
    ControlEnabled(ui->TT_SoftwareTriggerDisableB, true);
}

void MainWindow::on_TT_SoftwareTriggerDisableB_clicked()
{
    Mgr->ClockModules.SoftwareTrigger(false);
    ControlEnabled(ui->TT_SoftwareTriggerEnableB, true);
    ControlEnabled(ui->TT_SoftwareTriggerDisableB, false);
}

void MainWindow::on_TT_GPSEnableB_clicked()
{
    Mgr->ClockModules.GpsEnable(true);
    ControlEnabled(ui->TT_GPSEnableB, false);
    ControlEnabled(ui->TT_GPSDisableB, true);
}

void MainWindow::on_TT_GPSDisableB_clicked()
{
    Mgr->ClockModules.GpsEnable(false);
    ControlEnabled(ui->TT_GPSEnableB, true);
    ControlEnabled(ui->TT_GPSDisableB, false);
}

void MainWindow::on_TT_PpsArmButton_clicked()
{
    Mgr->ClockModules.PpsArm();
}

void MainWindow::on_FrameModeFreqsListFileButton_clicked()
{
    SelectFileDialog(ui->FrameModeFreqsListFileEdit, "Filter INI Files (*.ini *.INI)");
}

void MainWindow::on_TT_TriggerInBtn_clicked()
{
    Mgr->ClockModules.GpsTriggerSoon(ReadControl<int>(ui->TT_TriggerInEdit));
}

