
#define NOMINMAX

#include "AtroposClockModule_Gp.h"
#include <Analysis_Mb.h>
#include <IppMemoryUtils_Mb.h>
#include <Exception_Mb.h>


namespace Innovative     {
namespace Atropos_Clock  {

using namespace Innovative;
using namespace std;

//---------------------------------------------------------------------------
//  ExtClockSrcSelectStrings() -- return strings for Ext Src Select Listbox
//---------------------------------------------------------------------------

std::vector<std::string>  ExtClockSrcSelectStrings()
{
    std::vector<std::string>  List;

    List.push_back( std::string("Front Panel") );
    List.push_back( std::string("P16") );

    return List;
}

//---------------------------------------------------------------------------
//  ExtTriggerSrcSelectStrings() -- return strings for Trigger Src Select LB
//---------------------------------------------------------------------------

std::vector<std::string>  ExtTriggerSrcSelectStrings()
{
    std::vector<std::string>  List;

    List.push_back( std::string("Front Panel") );
    List.push_back( std::string("P16") );

    return List;
}

//===========================================================================
//  CLASS ModuleIo  -- Hardware Access and Application Io Class
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class ModuleIo
//---------------------------------------------------------------------------

ModuleIo::ModuleIo()
{
}

//---------------------------------------------------------------------------
//  destructor for class ModuleIo
//---------------------------------------------------------------------------

ModuleIo::~ModuleIo()
{
}

//---------------------------------------------------------------------------
//  ModuleIo::PreOpen
//---------------------------------------------------------------------------

void ModuleIo::PreOpen()
{
}

//---------------------------------------------------------------------------
//  ModuleIo::FiclConnectTo() --  Connect to FICL system
//---------------------------------------------------------------------------

void  ModuleIo::FiclConnectTo(FiclSystem & ficl)
{
    // do nothing if FICL is not supported by the board

    //  Attach board specific commands to FICL
    ficl.ConnectTarget(FiclTarget());
}

//---------------------------------------------------------------------------
//  ModuleIo::Log() --  Log message thunked to owner
//---------------------------------------------------------------------------

void  ModuleIo::Log(const std::string & msg)
{
    ProcessStatusEvent e(msg);
    OnLog.Execute(e);
}

//---------------------------------------------------------------------------
//  ModuleIo::Help() --  module-specific help command
//---------------------------------------------------------------------------

std::string ModuleIo::Help() const
{
    return "cr ?system cr ?Atropos";
}

//---------------------------------------------------------------------------
//  ModuleIo::SetSoftwareTrigger() --  Set or Clear Software Trigger
//---------------------------------------------------------------------------

void  ModuleIo::SetSoftwareTrigger(bool state)
{
    Module.SoftwareTrigger(state);
}

//---------------------------------------------------------------------------
//  ModuleIo::SetPulseTriggerConfiguration() --  Configure Pulse Trigger
//---------------------------------------------------------------------------

void  ModuleIo::SetPulseTriggerConfiguration(bool enable, bool finite_frames, bool finite_frame_rearm,
                                             unsigned int period, unsigned int finite_frame_count,
                                             std::vector<unsigned int> & delays,
                                             std::vector<unsigned int> & widths )
{
    // Optionally enable ramp generator
    Module.Pulse().Reset();
    Module.Pulse().Enabled(enable);
    Module.Pulse().FiniteFrames(finite_frames);
    Module.Pulse().FiniteFrameRearm(finite_frame_rearm);
    Module.Pulse().FiniteFrameCount(finite_frame_count);

    size_t pulses = std::min(delays.size(), widths.size());
    pulses = (pulses<32) ? pulses : 32;   // cap at 32 pulses

    for (size_t i=0; i<pulses; i++)
        {
        Module.Pulse().AddEvent(static_cast<unsigned int>(period),
                                static_cast<unsigned int>(delays[i]),
                                static_cast<unsigned int>(widths[i]));
        }
}

//===========================================================================
//  CLASS ApplicationIo  -- Hardware Access and Application I/O Class
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class ApplicationIo
//---------------------------------------------------------------------------

ApplicationIo::ApplicationIo(IUserInterface * ui)
    :  FiclIo(ui), Settings("AtroposClockSettings0.ini"), UI(ui),
       FOpened(false), Stopped(true)
{
    TraceVerbosity(Settings.DebugVerbosity);

    OnLog.SetEvent(this, &ApplicationIo::HandleOnLog);
    Module.OnLog.SetEvent(this, &ApplicationIo::HandleOnLog);
    Module.Atropos().RawClockDevice().Pll().OnPllTuning.SetEvent(this, &ApplicationIo::HandlePllTuning);

    //
    // Use IPP performance memory functions.
    Init::UsePerformanceMemoryFunctions();

    Timer.Interval(1000);

    MasterTriggerAsserted = true;

}

//---------------------------------------------------------------------------
//  destructor for class ApplicationIo
//---------------------------------------------------------------------------

ApplicationIo::~ApplicationIo()
{
    Close();
}

//---------------------------------------------------------------------------
//  ApplicationIo::BoardCount() -- Query number of installed boards
//---------------------------------------------------------------------------

unsigned int ApplicationIo::BoardCount()
{
    return static_cast<unsigned int>(Module.Atropos().BoardCount());
}

//---------------------------------------------------------------------------
// ApplicationIo::Open()
//---------------------------------------------------------------------------

void ApplicationIo::Open()
{
    //UI->GetSettings();

    Timer.OnElapsed.SetEvent(this, &ApplicationIo::HandleTimer);
    Timer.OnElapsed.Thunk();

    Module.Atropos().Target(Settings.Target);
    //
    //  Open Devices
    try
        {
        Module.PreOpen();
        Module.Atropos().Open();
        }
    catch (Innovative::MalibuException & exception)
        {
        Log("Open Failure:");
        Log(exception.what());
        return;
        }
    catch(...)
        {
        Log("Module Device Open Failure!");
        return;
        }

    Module.Atropos().Reset();
    Log("Atropos Opened Successfully...");
    FOpened = true;

    DisplayLogicVersion();
}

//---------------------------------------------------------------------------
// ApplicationIo::Close()
//---------------------------------------------------------------------------

void ApplicationIo::Close()
{
    Module.Atropos().Close();
    FOpened = false;
    Log("Atropos Closed...");
}

//---------------------------------------------------------------------------
// ApplicationIo::ValidateSampleRate()
//---------------------------------------------------------------------------

bool ApplicationIo::ValidateSampleRate(double should_be, double was, double allowed_range)
{
    double error = (should_be-was)/should_be;
    if (error < (-allowed_range) || error > allowed_range)
        return false;
    else
        return true;
}

//---------------------------------------------------------------------------
// ApplicationIo::StartTimer()
//---------------------------------------------------------------------------

bool ApplicationIo::StartTimer()
{
    if (Settings.Trig.Framed)
        {
        // Granularity is firmware limitation
        int framesize = Module.Atropos().Trigger().TriggerFrameGranularity();

        if (Settings.Trig.FrameSize % framesize)
            {
            std::stringstream msg;
            msg << "Error: Output frame count must be a multiple of " << framesize;
            Log(msg.str());
            //UI->AfterStop();
            return false;
            }
        }

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  Clock Configuration
    //   Route ext clock source
    AtroposLMK20_Clock::AtroposClockIo::IIClockSelect cks[] = { AtroposLMK20_Clock::AtroposClockIo::cslFrontPanel, AtroposLMK20_Clock::AtroposClockIo::cslP16 };
    Module.Atropos().Clock().ExternalClkSelect(cks[Settings.ExtClockSrcSelection]);
    //   Route reference.
    AtroposLMK20_Clock::AtroposClockIo::IIReferenceSource ref[] = { AtroposLMK20_Clock::AtroposClockIo::rsExternal, AtroposLMK20_Clock::AtroposClockIo::rsInternal };
    Module.Atropos().Clock().Reference(ref[Settings.ReferenceClockSource]);
    Module.Atropos().Clock().ReferenceFrequency(Settings.ReferenceRate * 1e6);
    //   Route clock
    AtroposLMK20_Clock::AtroposClockIo::IIClockSource src[] = { AtroposLMK20_Clock::AtroposClockIo::csExternal, AtroposLMK20_Clock::AtroposClockIo::csInternal };
    Module.Atropos().Clock().Source(src[Settings.SampleClockSource]);
    Module.Atropos().Clock().Frequency(Settings.Clk.SampleRate * 1e6);


    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  Trigger Configuration
    //
    //  Frame Triggering
    Module.Atropos().Trigger().FramedMode((Settings.Trig.Framed)? true : false);
    Module.Atropos().Trigger().Edge((Settings.Trig.EdgeTrigger)? true : false);
    Module.Atropos().Trigger().FrameSize(Settings.Trig.FrameSize);
    Module.Atropos().Trigger().External((Settings.Trig.ExternalTrigger)? true : false);

    //  Route External Trigger source
    IAtroposTriggerCfg::ExtSyncOptions syncsel[] = { IAtroposTriggerCfg::essFrontPanel,
                                                     IAtroposTriggerCfg::essP16 };
    Module.Atropos().Trigger().ExternalSyncSource( syncsel[ Settings.ExtTriggerSrcSelection ] );

    //  Pulse Trigger Config
    //  ...read pattern from file -
    //
    //  File is INI format with these fields. Delay is offset from Period start,
    //  width is width of trigger
    //
    //  --------------------------
    //  [Delay]
    //  ArraySize=2
    //  P0=100
    //  P1=500
    //
    //  [Width]
    //  ArraySize=2
    //  P0=50
    //  P1=150
    //
    ReadPriPatternFile(Settings.Trig.Pulse.PulsePatternINIFile);

     //  ...add to module configuration
    Module.SetPulseTriggerConfiguration(Settings.Trig.Pulse.Enable, Settings.Trig.Pulse.FiniteFrames,
                                        Settings.Trig.Pulse.FiniteFrameRearm,
                                        Settings.Trig.Pulse.Period, Settings.Trig.Pulse.FiniteFrameCount,
                                        Settings.Trig.Pulse.Delays,
                                        Settings.Trig.Pulse.Widths);

    // Set Decimation Factor
    int factor = Settings.Trig.DecimationEnable ? Settings.Trig.DecimationFactor : 0;
    Module.Atropos().Trigger().Decimation(factor);

    Stopped = false;
    Log("Clocks Started");
    Timer.Enabled(true);

    Module.Atropos().Start();

    MasterTrigger(true);

    // Readback Frequency
    double freq_actual = Module.Atropos().Clock().System().FrequencyActual();
    double freq        = Module.Atropos().Clock().System().Frequency();
    {
    std::stringstream msg;
    msg << "Actual PLL Frequency: " << freq_actual   ;
    Log(msg.str());
    }
    if (ValidateSampleRate(freq, freq_actual, .05)==false)
        {
        std::stringstream msg;
        msg << "WARNING: Sample Rate Mismatch - should be " << freq << " but was " << freq_actual  ;
        Log(msg.str());
        }

    return true;
}

//---------------------------------------------------------------------------
// ApplicationIo::StopTimer()
//---------------------------------------------------------------------------

void ApplicationIo::StopTimer()
{
    Stopped = true;
    Timer.Enabled(false);

    Module.Atropos().SoftwareTrigger(false);
    MasterTrigger(false);

    Module.Atropos().Stop();
    Log("Clocks Stopped");

    //UI->AfterStop();
}

void ApplicationIo::Clock10AnalogDelay(unsigned int delay)
{
    Atropos_Pll & A_Pll = dynamic_cast<Atropos_Pll &>(Module.Atropos().RawClockDevice().Pll());
    A_Pll.AnalogDelay_10(delay);
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandlePllTuning() --
//---------------------------------------------------------------------------

void ApplicationIo::HandlePllTuning(Innovative::Lmk04820PllTuningEvent & Event)
{
    if (Settings.PllTuningOverrideEnable)
        {
        Event.CLKin0_R = Settings.PllTuningOverride.CLKin0_R;
        Event.CLKin1_R = Settings.PllTuningOverride.CLKin1_R;
        Event.CLKin2_R = Settings.PllTuningOverride.CLKin2_R;
        Event.N1 = Settings.PllTuningOverride.N1;
        Event.R2 = Settings.PllTuningOverride.R2;
        Event.N2 = Settings.PllTuningOverride.N2;
        Event.N2_Cal = Settings.PllTuningOverride.N2_Cal;
        Event.N2_Pre = Settings.PllTuningOverride.N2_Pre;
        Event.OD = Settings.PllTuningOverride.OD;
        Event.Feedback_OD = Settings.PllTuningOverride.Feedback_OD;

        stringstream msg;
        msg << "CLKin0_R: " << Event.CLKin0_R
            << ", CLKin1_R: " << Event.CLKin1_R
            << ", CLKin2_R: " << Event.CLKin2_R
            << ", N1: " << Event.N1
            << ", R2: " << Event.R2
            << ", N2: " << Event.N2
            << ", N2_Cal: " << Event.N2_Cal
            << ", N2_Pre: " << Event.N2_Pre
            << ", OD: " << Event.OD
            << ", Feedback_OD: " << Event.Feedback_OD;

        Log(msg.str());
        }

}

//------------------------------------------------------------------------------
//  ApplicationIo::ReadPriPatternFile() --  Read PRI pattern from file
//------------------------------------------------------------------------------

void  ApplicationIo::ReadPriPatternFile(const std::string & file)
{
    Settings.Trig.Pulse.Delays.clear();
    Settings.Trig.Pulse.Widths.clear();

    IniReader patternFile(file);

    //  Define Data in Pattern File
    patternFile.Install( ToIni("Delay", "P",       Settings.Trig.Pulse.Delays,  (unsigned int)0x100 ));
    patternFile.Install( ToIni("Width", "P",       Settings.Trig.Pulse.Widths,  (unsigned int)0x100 ));

    patternFile.Load();

}

//---------------------------------------------------------------------------
//  ApplicationIo::SoftwareTrigger() --
//---------------------------------------------------------------------------

void ApplicationIo::SoftwareTrigger(bool state)
{
    Module.Atropos().SoftwareTrigger(state);
    FTriggered = state;
}

//---------------------------------------------------------------------------
//  ApplicationIo::MasterTrigger() --
//---------------------------------------------------------------------------

void ApplicationIo::MasterTrigger(bool state)
{
    Module.Atropos().MasterTrigger(state);
    MasterTriggerAsserted = state;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  Support Functions
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//---------------------------------------------------------------------------
//  ApplicationIo::Temperature() --  Query current module temperature
//---------------------------------------------------------------------------

float ApplicationIo::Temperature()
{
    return static_cast<float>(Module.Atropos().Thermal().LogicTemperature());
}

//---------------------------------------------------------------------------
//  ApplicationIo::PllLocked() --  Query pll lock status
//---------------------------------------------------------------------------

bool ApplicationIo::PllLocked()
{
    return Module.Atropos().Clock().Locked();
}

//---------------------------------------------------------------------------
//  ApplicationIo::SerDesSyncStatus() --  Query pll trigger sync
//---------------------------------------------------------------------------

bool ApplicationIo::SerDesSyncStatus()
{
    return Module.Atropos().Clock().TriggerSync();
}

//------------------------------------------------------------------------------
//  ApplicationIo::DisplayLogicVersion() --  Log version info
//------------------------------------------------------------------------------

void  ApplicationIo::DisplayLogicVersion()
{
    std::stringstream msg;
    msg << std::hex << "Logic Version: " << Module.Atropos().Info().FpgaLogicVersion()
        << ", Hdw Variant: " << Module.Atropos().Info().FpgaHardwareVariant()
        << ", Revision: " << Module.Atropos().Info().PciLogicRevision()
        << ", Subrevision: " << Module.Atropos().Info().FpgaLogicSubrevision();
    Log(msg.str());

    std::stringstream msg2;
    msg2 << std::hex << "Board Family: " << Module.Atropos().Info().PciLogicFamily()
        << ", Type: " << Module.Atropos().Info().PciLogicType()
        << "  (" << Module.Atropos().Name() << ")"
        << ", Board Revision: " << Module.Atropos().Info().PciLogicPcb()
        << ", Chip: " << Module.Atropos().FpgaName(Module.Atropos().Info().FpgaChipType());
    Log(msg2.str());

}


//---------------------------------------------------------------------------
// ApplicationIo::WriteRom()
//---------------------------------------------------------------------------

void ApplicationIo::WriteRom()
{
    //  System Page Operations
    Module.Atropos().IdRom().Name(Settings.ModuleName);
    Module.Atropos().IdRom().Revision(Settings.ModuleRevision);

    Module.Atropos().IdRom().StoreToRom();

}

//---------------------------------------------------------------------------
// ApplicationIo::ReadRom()
//---------------------------------------------------------------------------

void ApplicationIo::ReadRom()
{
    //  System Page Operations
    Module.Atropos().IdRom().LoadFromRom();

    Settings.ModuleName = Module.Atropos().IdRom().Name();
    Settings.ModuleRevision = Module.Atropos().IdRom().Revision();

}


//---------------------------------------------------------------------------
//  ApplicationIo::Log() --  Log message thunked to main thread
//---------------------------------------------------------------------------

void  ApplicationIo::Log(const std::string & msg)
{
    ProcessStatusEvent e(msg);
    OnLog.Execute(e);
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleOnLog() --
//---------------------------------------------------------------------------

void ApplicationIo::HandleOnLog(Innovative::ProcessStatusEvent & /*Event*/)
{
    //UI->Log(Event.Message);
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleTimer() --  Per-second status timer event
//---------------------------------------------------------------------------

void ApplicationIo::HandleTimer(OpenWire::NotifyEvent & /*Event*/)
{
    // Display status
    //UI->PeriodicStatus();
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleDisableTrigger() --  Trigger Manager Trig OFF
//------------------------------------------------------------------------------

void  ApplicationIo::HandleDisableTrigger(OpenWire::NotifyEvent & /*Event*/)
{
    Module.Atropos().Trigger().External(false);
    Module.SetSoftwareTrigger(false);
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleSoftwareTrigger() --  Trigger Manager Trig OFF
//------------------------------------------------------------------------------

void  ApplicationIo::HandleSoftwareTrigger(OpenWire::NotifyEvent & /*Event*/)
{
    if (Settings.Trig.ExternalTrigger == 0)
        Module.SetSoftwareTrigger(true);
}


//---------------------------------------------------------------------------
//  ApplicationIo::HandleExternalTrigger() --  Enable External trigger
//------------------------------------------------------------------------------

void  ApplicationIo::HandleExternalTrigger(OpenWire::NotifyEvent & /*Event*/)
{
    if (Settings.Trig.ExternalTrigger == 1)
        Module.Atropos().Trigger().External(true);
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleManualTrigger() --  Trigger Manager Trig OFF
//------------------------------------------------------------------------------

void  ApplicationIo::HandleManualTrigger(ManualTriggerEvent & Event)
{
    Module.SetSoftwareTrigger(Event.State);
}

//---------------------------------------------------------------------------
//  ApplicationIo::ClockInfo() -- Dump clock info
//---------------------------------------------------------------------------

void  ApplicationIo::ClockInfo()
{

    Log("Clock output data");
    for (unsigned int i=0; i<Module.Atropos().RawClockDevice().Pll().Outputs(); i++)
        {
        std::stringstream msg;
        msg << /*std::hex <<*/ "Out # " << i << ": "
            << " Divider: "   << Module.Atropos().RawClockDevice().Pll().OutputDivider(i)
            << ", Out Freq: " << Module.Atropos().RawClockDevice().Pll().OutputFrequency(i)
            << ", Actual: "   << Module.Atropos().RawClockDevice().Pll().OutputFrequencyActual(i) ;
        Log(msg.str());
        }
}


//---------------------------------------------------------------------------
//  ApplicationIo::PllLockedStr() -- Display Pll Lock Status
//---------------------------------------------------------------------------

std::string  ApplicationIo::PllLockedStr()
{
    if (PllLocked())
        return std::string(" + ");
    else
        return std::string(" - ");
}

//------------------------------------------------------------------------------
//  ApplicationIo::ModuleName() --  Return name of module as string
//------------------------------------------------------------------------------

std::string ApplicationIo::ModuleName()
{
    return Module.Atropos().Name();
}

//------------------------------------------------------------------------------
//  ApplicationIo::ModuleRevision() --  Return rev of module as string
//------------------------------------------------------------------------------

std::string ApplicationIo::ModuleRevision()
{
    std::stringstream msg;
    char c = 'A' + static_cast<char>(Module.Atropos().Info().PciLogicPcb());
    msg << c;
    return msg.str();
}

//---------------------------------------------------------------------------
//  ApplicationIo::TriggerStatus() -- Display Trigger Status
//---------------------------------------------------------------------------

std::string ApplicationIo::TriggerStatus()
{
    bool master_state = MasterTriggerAsserted;
    bool sw_state = FTriggered;
    std::string master_str = (master_state) ? "+" : "-" ;
    std::string sw_str = (sw_state) ? "+" : "-" ;

    std::stringstream msg;
    msg << std::hex << "M: " << master_str
        << "  SW: " << sw_str;
    return msg.str();
}

//==============================================================================
//  CLASS ApplicationSettings
//==============================================================================
//------------------------------------------------------------------------------
//  ApplicationSettings::ApplicationSettings() --  Ctor, load from INI file
//------------------------------------------------------------------------------

ApplicationSettings::ApplicationSettings( const std::string &file )
    :  IniSaver(file)
{
    //  Board Settings
    Install( ToIni("Target",                    Target,          0)  );

    //  Config Data
    //  ..Clock
    Install( ToIni("ExtClockSrcSelection", ExtClockSrcSelection, 0)  );
    Install( ToIni("ReferenceClockSource", ReferenceClockSource, 1)  );
    Install( ToIni("ReferenceRate",        ReferenceRate,       10.0f)  );
    Install( ToIni("SampleClockSource",    SampleClockSource, 1)  );
    //Install( ToIni("PLL PFD Rate",         PllPfdRate, 100.0e3)  );

    Install( ToIni("Clk.SampleRate",        Clk.SampleRate,  static_cast<float>(MaxInRateMHz()))  );
    //  ..Trigger
    Install( ToIni("ExternalTrigger",        Trig.ExternalTrigger,     0)  );
    Install( ToIni("EdgeTrigger",            Trig.EdgeTrigger,         0)  );
    Install( ToIni("Framed",                 Trig.Framed,              0)  );
    Install( ToIni("FrameSize",              Trig.FrameSize,           0x4000)  );
    Install( ToIni("ExtTriggerSrcSelection", ExtTriggerSrcSelection, 0)  );
    Install( ToIni("DecimationEnable",       Trig.DecimationEnable,    false)  );
    Install( ToIni("DecimationFactor",       Trig.DecimationFactor,    1)  );
    Install( ToIni("TriggerType",            Trig.TriggerType,    1)  );
    Install( ToIni("PpsSource",              Trig.PpsSource,    1)  );
    Install( ToIni("TriggerInSec",          Trig.TriggerInSec         , 1)  );

    Install( ToIni("Pulse.Enable",              Trig.Pulse.Enable,              false)  );
    Install( ToIni("Pulse.Period",              Trig.Pulse.Period,            10000000u)  );
    Install( ToIni("Pulse.FiniteFrames",        Trig.Pulse.FiniteFrames,        false)  );
    Install( ToIni("Pulse.FiniteFrameRearm",    Trig.Pulse.FiniteFrameRearm,    false)  );
    Install( ToIni("Pulse.FiniteFrameCount",    Trig.Pulse.FiniteFrameCount,    100u)  );
    Install( ToIni("Pulse.PulsePatternINIFile", Trig.Pulse.PulsePatternINIFile,   std::string("PulsePattern.ini"))  );
    //  ..Other
    Install( ToIni("Debug Script",         DebugScript,  std::string("")) );
    Install( ToIni("DebugVerbosity",       DebugVerbosity,  5) );   // vNormal

    PllTuningOverrideEnable = false;
    PllTuningOverride.CLKin0_R = 0; PllTuningOverride.CLKin1_R = 0;
    PllTuningOverride.CLKin2_R = 0; PllTuningOverride.N1 = 0;
    PllTuningOverride.R2 = 0;       PllTuningOverride.N2 = 0;
    PllTuningOverride.N2_Cal = 0;   PllTuningOverride.N2_Pre = 0;
    PllTuningOverride.OD = 0;       PllTuningOverride.Feedback_OD = 0;

    Load();
    //
    //  Sanity Check on Target
    if (Target < 0)
        Target = 0;

}

//------------------------------------------------------------------------
//  ApplicationSettings::~ApplicationSettings() --  Dtor, save to INI file
//------------------------------------------------------------------------

ApplicationSettings::~ApplicationSettings()
{
    //
    //  Sanity Check on Target
    if (Target < 0)
        Target = 0;

    Save();
}









} } // Innovative::Clock
