#ifndef ATROPOSCLOCKMODULE_GP_H
#define ATROPOSCLOCKMODULE_GP_H

#include <ProcessEvents_Mb.h>
#include <VitaPacketStream_Mb.h>
#include <BinView_Mb.h>
#include <Benchmark_Mb.h>
#include <SoftwareTimer_Mb.h>
#include <DataPlayer_Mb.h>
#include <Exception_Mb.h>

#include <Application/FiclIo_App.h>
#include <Application/IniSaver_App.h>
#include <Application/TriggerManager_App.h>
#include <Application/StaticWaveform_App.h>
#include <Application/BinviewPlotter_App.h>

#include <Atropos_Mb.h>

namespace Innovative {
namespace Atropos_Clock      {


//===========================================================================
//  Information Functions
//===========================================================================
//  NOTE - these are often used in constructors so cannot depend on objects
//         without due care taken
//
//   General Info
inline std::string ModuleNameStr()
            {  return  "Atropos Board";  }

inline float  MaxInRateMHz()
          {  return 400.0;  }     // In MHz
//inline float  MaxOutRateMHz()
//          {  return 1000.0;  }     // In MHz

//
//   Application Feature Presence  -- used to hide GUI Portions

inline bool   HasClockMux()
        {  return true; }
inline bool   HasProgrammableReference()
        {  return true; }
inline bool   HasFiclSupport()
        {  return true; }
inline bool   HasExtClockSrcSelectMux()
        {  return true; }
inline bool   HasExtTriggerSrcSelectMux()
        {  return true; }


typedef std::vector<std::string> StringArray;
StringArray  ExtClockSrcSelectStrings();
StringArray  ExtTriggerSrcSelectStrings();

//===========================================================================
//  CLASS ModuleIo  -- Hardware Access and Application Io Class
//===========================================================================

class ModuleIo
{
public:
    ModuleIo();
    ~ModuleIo();

    //  Module Aliases
    Innovative::Atropos & Atropos()
        {  return Module;  }
    Innovative::IFiclTarget *  FiclTarget()
            {   return &Module;  }

    OpenWire::ThunkedEventHandler<Innovative::ProcessStatusEvent>  OnLog;

    //
    //  App System Methods
    void  FiclConnectTo(Innovative::FiclSystem & ficl);
    std::string Help() const;

    //
    //  Module Methods

    void  SetSoftwareTrigger(bool state);
    void  SetPulseTriggerConfiguration(bool enable, bool finite_frames, bool finite_frame_rearm,
                                       unsigned int period, unsigned int finite_frame_count,
                                       std::vector<unsigned int> & delays,
                                       std::vector<unsigned int> & widths );

    void PreOpen();


protected:
    void  Log(const std::string & msg);

private:
    Innovative::Atropos  Module;


};

class ApplicationIo;


//==============================================================================
//  CLASS ApplicationSettings
//==============================================================================

class ApplicationSettings : public Innovative::IniSaver
{
public:

    typedef std::vector<int>    IntArray;

    // Ctor
    ApplicationSettings( const std::string &file );
    ~ApplicationSettings();

    //
    //  Board Settings
    int             Target;


    //  Config Data
    //  Clock
    int             ExtClockSrcSelection;
    int             ReferenceClockSource;
    float           ReferenceRate;
    int             SampleClockSource;
    double          PllPfdRate;
    int             ExtTriggerSrcSelection;
    //  Analog

    // Debug
    int             DebugVerbosity;
    std::string     DebugScript;


    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS ClockSettings
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    struct ClockSettings
    {
        ClockSettings()
            {}

        float           SampleRate;
    };

    ClockSettings   Clk;

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS TuningEvent
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    struct TuningEvent
    {
        unsigned int CLKin0_R;
        unsigned int CLKin1_R;
        unsigned int CLKin2_R;
        unsigned int N1;
        unsigned int R2;
        unsigned int N2;
        unsigned int N2_Cal;
        unsigned int N2_Pre;
        unsigned int OD;
        unsigned int Feedback_OD;
    };

    TuningEvent PllTuningOverride;
    bool        PllTuningOverrideEnable;

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS PulseD - PRI pulse trigger settings
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    struct PulseD
    {
        bool          Enable;
        bool          FiniteFrames;
        bool          FiniteFrameRearm;
        unsigned int  Period;
        unsigned int  FiniteFrameCount;
        std::string   PulsePatternINIFile;

        std::vector<unsigned int>   Delays;
        std::vector<unsigned int>   Widths;

    };

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS TriggerSettings
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    struct TriggerSettings
    {
        TriggerSettings()
            {}

        //  Trigger
        int             ExternalTrigger;
        int             EdgeTrigger;
        int             Framed;
        int             FrameSize;
        PulseD          Pulse;
        bool            DecimationEnable;
        int             DecimationFactor;
        int             TriggerType;
        int             PpsSource;
        int             TriggerInSec;
    };

    TriggerSettings   Trig;

    //
    //  Not saved in INI file
    //  ..Eeprom
    std::string     ModuleName;
    std::string     ModuleRevision;



};


//===========================================================================
//  CLASS IUserInteface  -- Interface for callbacks to UI from App Logic
//===========================================================================

class IUserInterface : public IFicl_UserInterface
{
public:
    virtual ~IUserInterface()  {  }
    virtual void  Log(const std::string & a_string) = 0;
    virtual void  GetSettings() = 0;
    virtual void  SetSettings() = 0;
    virtual void  AfterStop() = 0;
    virtual void  PeriodicStatus() = 0;
};

//===========================================================================
//  CLASS ApplicationIo  -- Hardware Access and Application Io Class
//===========================================================================

class ApplicationIo : public FiclIo
{
public:
    // Ctor
    ApplicationIo(IUserInterface * gui);
    ~ApplicationIo();

    ModuleIo &  ModIo()
        {  return Module;  }

    // Methods
    unsigned int    BoardCount();

    void            Open();
    bool            IsOpen()
        {  return FOpened;  }
    void            Close();

//	void            StreamPreconfigure();

    bool            StartTimer();
    void            StopTimer();
    void            Clock10AnalogDelay(unsigned int delay);

    void            SoftwareTrigger(bool state);
    void            MasterTrigger(bool state);

    bool            IsRunning()
        {  return Timer.Enabled();  }

    void            WriteRom();
    void            ReadRom();

    float   Temperature();
    bool    PllLocked();
    bool    Triggered()  {  return FTriggered;  }
    bool    SerDesSyncStatus();

    void    ExecuteDebugScript(const std::string & cmd)
        {  Execute(std::string("load ") + cmd);  }
    void    Help()
        {  Execute(Module.Help());  }

    void    ClockInfo();

    // Module status
    std::string  ModuleName();
    std::string  ModuleRevision();
    std::string  TriggerStatus();
    std::string  PllLockedStr();


public:  // Pseudo-protected
    // Data
    ApplicationSettings             Settings;

private:
    //
    //  Member Data
    ModuleIo                        Module;
    IUserInterface *                UI;

    Innovative::SoftwareTimer       Timer;

    // App State Variables
    bool                            FOpened;
    bool                            FTriggered;
    bool                            Stopped;
    bool                            MasterTriggerAsserted;

protected:

    void  HandleOnLog(Innovative::ProcessStatusEvent & Event);
    void  HandleTimer(OpenWire::NotifyEvent & Event);
    void  HandlePllTuning(Innovative::Lmk04820PllTuningEvent & Event);

    void  HandleDisableTrigger(OpenWire::NotifyEvent & Event);
    void  HandleExternalTrigger(OpenWire::NotifyEvent & Event);
    void  HandleSoftwareTrigger(OpenWire::NotifyEvent & Event);
    void  HandleManualTrigger(Innovative::ManualTriggerEvent & Event);

    void  Log(const std::string & msg);

    OpenWire::ThunkedEventHandler<Innovative::ProcessStatusEvent>  OnLog;

    void  ReadPriPatternFile(const std::string & file);

    bool  ValidateSampleRate(double should_be, double was, double allowed_range);
    void  DisplayLogicVersion();

};








} } // Innovative::Atropos_Clock

#endif // ATROPOSCLOCKMODULE_GP_H
