#ifndef RATECONTROLIO_H
#define RATECONTROLIO_H

#include "Filter.h"
#include "CommonRegisters_Gp.h"
#include "DdcModule_Gp.h"
#include "DucModule_Gp.h"
#include <Application/IniSaver_App.h>

namespace GP {


struct DdcChannelInfo
{
    DdcChannelInfo() :
        Source(0), Frequency(0.0), ModFrequency(0.0), Enable(false), SpectrumInvert(false), File("")
    {
    }

    DdcChannelInfo(int _src, double _freq, double _mfreq, bool _en, bool _sp, string _file) :
        Source(_src), Frequency(_freq), ModFrequency(_mfreq), Enable(_en), SpectrumInvert(_sp), File(_file)
    {
    }

    int    Source;
    double Frequency;
    double ModFrequency;
    bool   Enable;
    bool   SpectrumInvert;
    string File;
    FrameModeInfo FrameMode;

    bool operator==(const DdcChannelInfo &other)
    {
        return
           other.Source          == Source           &&
           other.Frequency       == Frequency        &&
           other.ModFrequency    == ModFrequency     &&
           other.Enable          == Enable           &&
           other.SpectrumInvert  == SpectrumInvert   &&
           other.File            == File;
    }

};

struct DucChannelInfo
{
    DucChannelInfo() :
        Source(0), StreamId(0), Frequency(0.0), Enable(false), SpectrumInvert(false), File("")
    {
    }

    DucChannelInfo(int _src, int _sid, double _freq, bool _en, bool _sp, string _file) :
        Source(_src),  StreamId(_sid), Frequency(_freq), Enable(_en), SpectrumInvert(_sp), File(_file)
    {
    }

    int    Source;
	int    StreamId;
    double Frequency;
    bool   Enable;
    bool   SpectrumInvert;
    string File;


    bool operator==(const DucChannelInfo &other)
    {
        return
           other.Source          == Source           &&
           other.StreamId        == StreamId         &&
           other.Frequency       == Frequency        &&
           other.Enable          == Enable           &&
           other.SpectrumInvert  == SpectrumInvert   &&
           other.File            == File;
    }
};

class IChannel
{
public:
    virtual void Reset(bool state) = 0;
    virtual void Apply() = 0;
    virtual void ApplyEnable() = 0;
    virtual string Tag() = 0;
    virtual void AttachSettings(Innovative::IniSaver& saver) = 0;
    virtual unsigned int WhichBoard() = 0;

    IChannel(int aid)
        : ch_id(aid)
    {}
    IChannel(int aid, const DdcChannelInfo &ainfo)
        : info(ainfo), ch_id(aid)
    {}
    
	IChannel(int aid, const DucChannelInfo &ainfo)
        : ducinfo(ainfo), ch_id(aid)
    {}

    DdcChannelInfo   info;
	DucChannelInfo   ducinfo;
    Filter           filter;
    unsigned int     ch_id;
    float            dec;
    string           tag;
};

//===========================================================================
//  CLASS DdcTunerWrapper --
//===========================================================================
class DdcTunerWrapper
{
public:
    DdcTunerWrapper(DdcTuner *atune, string atag, unsigned int acore);
    virtual ~DdcTunerWrapper();

    void Frequency(unsigned int parent, double freq, double srate);
    void ModFrequency(unsigned int parent, double freq, double srate);
    void Phase(unsigned int parent, double value);
    void TuningFrequencies(unsigned int parent, const vector<double> &freqs, double srate);
    void TuningEnable(unsigned int parent, bool en);
    unsigned int TuningDataOut(unsigned int parent);

    void Reset(unsigned int parent, bool value);
    void CommandTrigger(unsigned int parent, bool value);

    unsigned int CoreCount()
    {
        return tune->CoreCount();
    }

private:
    DdcTuner         * const tune;
    string                   tag;
    const unsigned int       core;


private:
    DdcTunerWrapper(const DdcTunerWrapper &other);
    DdcTunerWrapper &operator=(const DdcTunerWrapper &other);
};

//===========================================================================
//  CLASS DdcLowPassFilterWrapper --
//===========================================================================
class DdcLowPassFilterWrapper
{
public:
    DdcLowPassFilterWrapper(DdcLowPassFilter *alp, string atag, unsigned int acore);
    virtual ~DdcLowPassFilterWrapper();

    void CicRate(unsigned int parent, unsigned int value);
    void CicBStart(unsigned int parent, unsigned int num, unsigned int dir);
    void CFirCoeffs(unsigned int parent, std::vector<int> cfs);
    void CFirBStart(unsigned int parent, unsigned int num, unsigned int dir);
    void PFirCoeffs(unsigned int parent, std::vector<int> cfs);
    void PFirBStart(unsigned int parent, unsigned int num, unsigned int dir);
    void FirlCoeffs(unsigned int parent, std::vector<int> cfs);
    void FirlBStart(unsigned int parent, unsigned int num, unsigned int dir);
    void Fir2Coeffs(unsigned int parent, std::vector<int> cfs);
    void Fir2BStart(unsigned int parent, unsigned int num, unsigned int dir);

    void Reset(unsigned int parent, bool value);
    void CommandTrigger(unsigned int parent, bool value);

private:
    DdcLowPassFilter * const lp;
    string                  tag;
    const unsigned int      core;
private:
    DdcLowPassFilterWrapper(const DdcLowPassFilterWrapper &other);
    DdcLowPassFilterWrapper &operator=(const DdcLowPassFilterWrapper &other);
};

//===========================================================================
//  CLASS DdcPolyphaseFilterWrapper --
//===========================================================================
class DdcPolyphaseFilterWrapper
{
public:
    DdcPolyphaseFilterWrapper(DdcPolyPhaseFilter *alp, string atag, unsigned int acore);
    virtual ~DdcPolyphaseFilterWrapper();

    void PpfBStart(unsigned int parent, unsigned int num, unsigned int dir);
    void PpfCoeffs(unsigned int parent, unsigned int id, std::vector<int> cfs);
    void Reset(unsigned int parent, bool value);
    void CommandTrigger(unsigned int parent, bool value);

private:
    DdcPolyPhaseFilter * const lp;
    string                  tag;
    const unsigned int      core;
private:
    DdcPolyphaseFilterWrapper(const DdcPolyphaseFilterWrapper &other);
    DdcPolyphaseFilterWrapper &operator=(const DdcPolyphaseFilterWrapper &other);
};


//===========================================================================
//  CLASS DdcVitaFramerWrapper --
//===========================================================================
class DdcVitaFramerWrapper
{
public:
    DdcVitaFramerWrapper(DdcVitaFramer *avf, string atag, unsigned int acore);
    virtual ~DdcVitaFramerWrapper();

    void Enable(unsigned int parent, bool value);
    void Reset(unsigned int parent, bool value);
    void CommandTrigger(unsigned int parent, bool value);
    void Size(unsigned int parent, unsigned int value);
    unsigned int Size(unsigned int parent);
    void Identity(unsigned int parent, unsigned int did, unsigned int sid);
    void FrameMode(unsigned int parent, unsigned int mode, bool loc, unsigned int src);
    void FrameSize(unsigned int parent, unsigned int value);
    unsigned int FrameCount(unsigned int parent);
    void FrameDelay(unsigned int parent, unsigned int value);
    void ReEnable(unsigned int parent, unsigned int value);

private:
    DdcVitaFramer    * const vf;
    string                   tag;
    const unsigned int       core;

private:
    DdcVitaFramerWrapper(const DdcVitaFramerWrapper &other);
    DdcVitaFramerWrapper &operator=(const DdcVitaFramerWrapper &other);
};

//===========================================================================
//  CLASS ChannelizerWrapper --
//===========================================================================
class ChannelizerWrapper
{
public:
    ChannelizerWrapper(DdcChannelizer *achizer, string atag, unsigned int acore);
    virtual ~ChannelizerWrapper();

    void Reset(unsigned int parent, bool value);
    void CommandTrigger(unsigned int parent, bool value);
    void ScalingSchedule(unsigned int parent, unsigned int value);

private:
    DdcChannelizer      * const chizer;
    string                     tag;
    const unsigned int         core;
private:
    ChannelizerWrapper(const ChannelizerWrapper &other);
    ChannelizerWrapper &operator=(const ChannelizerWrapper &other);
};

//===========================================================================
//  CLASS DdcPeripheralWrapper --
//===========================================================================
class DdcPeripheralWrapper
{
public:
    DdcPeripheralWrapper(DdcPeripheralSM *ape, string atag, unsigned int acore);
    virtual ~DdcPeripheralWrapper();

    void BypassCore(unsigned int parent, bool value);
    void Source(unsigned int parent, unsigned int in);
    void EnableSpectrumInversion(unsigned int parent, bool value);
    void Filter1_Select(unsigned int parent, unsigned int value);
    void Filter2_Select(unsigned int parent, unsigned int value);
    void Decimation(unsigned int parent, unsigned int value);
    void Reset(unsigned int parent, bool value);
    void CommandTrigger(unsigned int parent, bool value);

private:
    DdcPeripheralSM   * const pe;
    string                   tag;
    const unsigned int       core;

private:
    DdcPeripheralWrapper(const DdcPeripheralWrapper &other);
    DdcPeripheralWrapper &operator=(const DdcPeripheralWrapper &other);
};

class DdcChannelsProxy;
//===========================================================================
//  CLASS ChannelWrapper --
//===========================================================================

class ChannelWrapper : public IChannel
{
public:
    ChannelWrapper(Ddc_Module *amod, string atag, int aid, const DdcChannelInfo &ainfo, DdcChannelsProxy *aparent, int abid);
    virtual ~ChannelWrapper();

    void Clear();
    string Tag();
    void AttachSettings(Innovative::IniSaver& saver);
    void ApplyEnable();
    void Apply();
    void Reset(bool state);
    void CommandTrigger();
    void AddTunerWrapper(DdcTunerWrapper *atune);
    void AddLowPassFilterWrapper(DdcLowPassFilterWrapper *alp);
    void AddVitaFramerWrapper(DdcVitaFramerWrapper *avf);
    void AddPeripheralWrapper(DdcPeripheralWrapper *ape);
    void AddChannelizerWrapper(ChannelizerWrapper *ach);
    void AddPolyPassFilterWrapper(DdcPolyphaseFilterWrapper *app);

    unsigned int WhichBoard();
    void ReloadFile();

    void TestExecute();
    void Flush(bool value);
    void Strobe(bool en);
    void Frequency(double freq);

    void Group(int value);
    int Group();

    size_t ChannelizerChannelCount()
    {
        return ChannelizerGroup.size();
    }

    void UpdateFrequencies(const vector<double> &v)
    {  Frequencies = v; }

    unsigned int VitaSize();

private:
    Ddc_Module                         *DdcM;
    vector<DdcPeripheralWrapper *>      PeripGroup;
    vector<DdcTunerWrapper *>           TunerGroup;
    vector<DdcLowPassFilterWrapper *>   LPGroup;
    vector<DdcPolyphaseFilterWrapper *> PGroup;
    vector<DdcVitaFramerWrapper *>      VFramerGroup;
    vector<ChannelizerWrapper *>        ChannelizerGroup;

    std::string                         LoadedFilter;
    vector<double>                      Frequencies;
    unsigned int                        BoardOrdinal;
    int                                 GroupId;
    DdcChannelsProxy * const parent;

private:
    ChannelWrapper(const ChannelWrapper &other);
    ChannelWrapper &operator=(const ChannelWrapper &other);

};

//===========================================================================
//  CLASS DdcChannelsProxy --
//===========================================================================

class DdcChannelsProxy
{
public:
    DdcChannelsProxy();
    ~DdcChannelsProxy();

    void AttachSettings(Innovative::IniSaver &saver);
    void ApplyCached();
    void ApplyCached(const vector<int> &values);
    void Apply();
    void Apply(unsigned int ch);
    void Apply(const vector<int> &values);
    void ApplyEnable(unsigned int ch);
    void ApplyEnable(const vector<int> &values);
    void ChannelTags(vector<string> &tags);
    void ChannelTags(vector<string> &tags, unsigned int board);
    void Reset(bool state);
    void ChannelizerReset(bool state);
    void CommandTrigger();
    void SamplingRate(double value);
    double SamplingRate();
    void Update(unsigned int ordinal, int group);
    void UpdateEnable(unsigned int ordinal, int group);
    int  ChannelCount(unsigned int board);

    vector<float> DecimationCoeffs();
    float DecimationCoeffs(unsigned int ch);

    void Clear();
    size_t Size();
    IChannel *operator[](unsigned int n);
    void Append(ChannelWrapper *val);
    ChannelWrapper &Wrapper(unsigned int channel);

    pair<size_t, int> EnabledChannelCount();
    size_t CalculatePacketSize();

    void            TestExecute();
    void            Strobe();
    void            NextScheduledFrequency(bool init = false);

    double             SRate;
	double             IpolRate;
    int                BypassDdc;
    string             ScheduleFreqsFile;
    vector<size_t>     NextSchedIndexes;
    DdcChannelInfo     Cache;
    vector<ChannelWrapper *> Chans;

private:


    vector<vector<double> > ReadFrequencyList(const std::string &file);
    vector<vector<double> > FrequencyList;

    DdcChannelsProxy(const DdcChannelsProxy &other);
    DdcChannelsProxy &operator=(const DdcChannelsProxy &other);
};


/////////////////////////////////////////////////
// DUC
/////////////////////////////////////////////////

class DucChannelsProxy;
class DucPeripheralWrapper;
class DucTunerWrapper;
class DucVitaFramerWrapper;
class DucLowPassFilterWrapper;
class DucPolyphaseFilterWrapper;
//===========================================================================
//  CLASS DucChannelWrapper --
//===========================================================================
class DucChannelWrapper : public IChannel
{
public:
    DucChannelWrapper(Duc_Module *amod, string atag, int aid, const DucChannelInfo &ainfo, DucChannelsProxy *aparent, int abid);
    virtual ~DucChannelWrapper();

    void Clear();

    string Tag();

    void ApplyEnable();
    void Apply();
    void Reset(bool state);
    void CommandTrigger();

    void AttachSettings(Innovative::IniSaver& saver);
	void AddTunerWrapper(DucTunerWrapper *atune);
    void AddVitaFramerWrapper(DucVitaFramerWrapper *avf);
    void AddPeripheralWrapper(DucPeripheralWrapper *ape);
	void AddLowPassFilterWrapper(DucLowPassFilterWrapper *alp);
    void AddPolyPassFilterWrapper(DucPolyphaseFilterWrapper *app);

    unsigned int WhichBoard();
    void         ReloadFile();

private:
    Duc_Module                         *DucM;
    vector<DucPeripheralWrapper *>      PeripGroup;
    vector<DucTunerWrapper *>           TunerGroup;
    vector<DucVitaFramerWrapper *>      VFramerGroup;
    vector<DucLowPassFilterWrapper *>   LPGroup;
	vector<DucPolyphaseFilterWrapper *> PGroup;

    unsigned int                        BoardOrdinal;
	    
	DucChannelsProxy				  * const parent;
private:
    DucChannelWrapper(const DucChannelWrapper &other);
    DucChannelWrapper &operator=(const DucChannelWrapper &other);
};

//===========================================================================
//  CLASS DucChannelsProxy --
//===========================================================================

class DucChannelsProxy
{
public:
    DucChannelsProxy();
    ~DucChannelsProxy();
	    
    void AttachSettings(Innovative::IniSaver &saver);
    void ApplyCached();
    void ApplyCached(const vector<int> &values);
    void Apply();
    void Apply(unsigned int ch);
    void Apply(const vector<int> &values);
    void ApplyEnable(unsigned int ch);
    void ApplyEnable(const vector<int> &values);
    void ChannelTags(vector<string> &tags);
    void ChannelTags(vector<string> &tags, unsigned int board);
    void Reset(bool state);
    void CommandTrigger();
    void SamplingRate(double value);
    double SamplingRate();
    int  ChannelCount(unsigned int board);
	vector<float> DecimationCoeffs();
    float DecimationCoeffs(unsigned int ch);

	void Clear();
    size_t Size();
    IChannel *operator[](unsigned int n);
    void Append(DucChannelWrapper *val);
    DucChannelWrapper &Wrapper(unsigned int channel);

    double             SRate;
    int                BypassDdc;
	//  SamplingRateProxy           SamplingRate;
    DucChannelInfo                 Cache;
    vector<DucChannelWrapper *>    Chans;
private:
    DucChannelsProxy(const DucChannelsProxy &other);
    DucChannelsProxy &operator=(const DucChannelsProxy &other);
};

//===========================================================================
//  CLASS DucPeripheralWrapper --
//===========================================================================
class DucPeripheralWrapper
{
public:
    DucPeripheralWrapper(DucPeripheralSM *ape, string atag, unsigned int acore);
    virtual ~DucPeripheralWrapper();

    void BypassCore(unsigned int parent, bool value);
    void Source(unsigned int parent, unsigned int in);
	void Sid(unsigned int parent, unsigned int sid);
    void EnableSpectrumInversion(unsigned int parent, bool value);
    void Filter1_Select(unsigned int parent, unsigned int value);
    void Filter2_Select(unsigned int parent, unsigned int value);
    void Decimation(unsigned int parent, unsigned int value);
    void Reset(unsigned int parent, bool value);
    void CommandTrigger(unsigned int parent, bool value);

private:
    DucPeripheralSM   * const pe;
    string                   tag;
    const unsigned int       core;

private:
    DucPeripheralWrapper(const DucPeripheralWrapper &other);
    DucPeripheralWrapper &operator=(const DucPeripheralWrapper &other);
};

//===========================================================================
//  CLASS DucTunerWrapper --
//===========================================================================
class DucTunerWrapper
{
public:
    DucTunerWrapper(DucTuner *atune, string atag, unsigned int acore);
    virtual ~DucTunerWrapper();

    void Frequency(unsigned int parent, double freq, double srate);
    void Phase(unsigned int parent, double value);
    void Reset(unsigned int parent, bool value);
    void CommandTrigger(unsigned int parent, bool value);

private:
    DucTuner         * const tune;
    string                   tag;
    const unsigned int       core;


private:
    DucTunerWrapper(const DucTunerWrapper &other);
    DucTunerWrapper &operator=(const DucTunerWrapper &other);
};

//===========================================================================
//  CLASS DucVitaFramerWrapper --
//===========================================================================
class DucVitaFramerWrapper
{
public:
    DucVitaFramerWrapper(DucVitaFramer *avf, string atag, unsigned int acore);
    virtual ~DucVitaFramerWrapper();

    void Enable(unsigned int parent, bool value);
    void Reset(unsigned int parent, bool value);
    void CommandTrigger(unsigned int parent, bool value);
    void Size(unsigned int parent, unsigned int value);
    void Identity(unsigned int parent, unsigned int did, unsigned int sid);

private:
    DucVitaFramer    * const vf;
    string                   tag;
    const unsigned int       core;

private:
    DucVitaFramerWrapper(const DucVitaFramerWrapper &other);
    DucVitaFramerWrapper &operator=(const DucVitaFramerWrapper &other);
};

//===========================================================================
//  CLASS DucLowPassFilterWrapper --
//===========================================================================
class DucLowPassFilterWrapper
{
public:
    DucLowPassFilterWrapper(DucLowPassFilter *alp, string atag, unsigned int acore);
    virtual ~DucLowPassFilterWrapper();

    void CicRate(unsigned int parent, unsigned int value);
    void CicBStart(unsigned int parent, unsigned int num, unsigned int dir);
    void CFirCoeffs(unsigned int parent, std::vector<int> cfs);
    void CFirBStart(unsigned int parent, unsigned int num, unsigned int dir);
    void PFirCoeffs(unsigned int parent, std::vector<int> cfs);
    void PFirBStart(unsigned int parent, unsigned int num, unsigned int dir);
    void FirlCoeffs(unsigned int parent, std::vector<int> cfs);
    void FirlBStart(unsigned int parent, unsigned int num, unsigned int dir);
    void Fir2Coeffs(unsigned int parent, std::vector<int> cfs);
    void Fir2BStart(unsigned int parent, unsigned int num, unsigned int dir);

    void Reset(unsigned int parent, bool value);
    void CommandTrigger(unsigned int parent, bool value);

private:
    DucLowPassFilter * const lp;
    string                  tag;
    const unsigned int      core;
private:
    DucLowPassFilterWrapper(const DucLowPassFilterWrapper &other);
    DucLowPassFilterWrapper &operator=(const DucLowPassFilterWrapper &other);
};

//===========================================================================
//  CLASS DucPolyphaseFilterWrapper --
//===========================================================================
class DucPolyphaseFilterWrapper
{
public:
    DucPolyphaseFilterWrapper(DucPolyPhaseFilter *alp, string atag, unsigned int acore);
    virtual ~DucPolyphaseFilterWrapper();

    void PpfBStart(unsigned int parent, unsigned int num, unsigned int dir);
    void PpfCoeffs(unsigned int parent, unsigned int id, std::vector<int> cfs);
    void Reset(unsigned int parent, bool value);
    void CommandTrigger(unsigned int parent, bool value);

private:
    DucPolyPhaseFilter * const lp;
    string                  tag;
    const unsigned int      core;
private:
    DucPolyphaseFilterWrapper(const DucPolyphaseFilterWrapper &other);
    DucPolyphaseFilterWrapper &operator=(const DucPolyphaseFilterWrapper &other);
};

} // GS

#endif // RATECONTROLIO_H
