#include "ClockModule_Gp.h"



namespace Innovative {
namespace Clock      {




//===========================================================================
//  CLASS ApplicationIo  -- Hardware Access and Application I/O Class
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class ApplicationIo
//---------------------------------------------------------------------------

ApplicationIo::ApplicationIo(unsigned int target, ClockSettings &settings)
    : Settings(settings), Opened(false), Target(target)
{
    OnLog.SetEvent(this, &ApplicationIo::HandleOnLog);

    std::stringstream txt;
    txt << "ClockSettings" << target << ".ini";
    SettingsFileName = txt.str();

    // As we know the target number, get settings for the given target
    LoadSettingsFromFile(SettingsFileName);

    TraceVerbosity(Trace::vNormal);

    UpdateSummaryOutputDisplayPrefixes();
}

//---------------------------------------------------------------------------
//  destructor for class ApplicationIo
//---------------------------------------------------------------------------

ApplicationIo::~ApplicationIo()
{
    Close();

    SaveSettingsToFile(SettingsFileName);
}

//---------------------------------------------------------------------------
//  ApplicationIo::BoardCount() -- Query number of installed boards
//---------------------------------------------------------------------------

int ApplicationIo::BoardCount()
{
    return static_cast<int>(Module.BoardCount());
}

//---------------------------------------------------------------------------
//  ApplicationIo::Open() -- Open Hardware & set up callbacks
//---------------------------------------------------------------------------

void  ApplicationIo::Open()
{
    Module.Gps().OnEpoch.SetEvent(this, &ApplicationIo::HandleEpoch);
    Module.Gps().OnGpsArm.SetEvent(this, &ApplicationIo::HandleGpsArm);

    //
    //  Open Devices
    Module.Target(Target);
    try
        {
        Module.Open();
        }
    catch(...)
        {
        Opened = false;
        return;
        }
    Module.Reset();
    Opened = true;
}

//---------------------------------------------------------------------------
//  ApplicationIo::Close() -- Close Hardware & set up callbacks
//---------------------------------------------------------------------------

void  ApplicationIo::Close()
{
    Module.Close();
    Opened = false;
}

//---------------------------------------------------------------------------
//  ApplicationIo::AfterGetSettings() -- Validate settings before use
//---------------------------------------------------------------------------

void  ApplicationIo::AfterGetSettings()
{
    if (Settings.Board.Target < 0)
        Settings.Board.Target = 0;
}


//---------------------------------------------------------------------------
//  ApplicationIo::LoadSettingsFromFile() -- Load settings file
//---------------------------------------------------------------------------

void  ApplicationIo::LoadSettingsFromFile(const std::string & file_name)
{
    IniSaver  Saver(file_name, false);   // not local file path

    InstallFileIoSettings(Saver);

    Saver.Load();   // To Settings

}

//---------------------------------------------------------------------------
//  ApplicationIo::SaveSettingsToFile() -- Save settings file
//---------------------------------------------------------------------------

void  ApplicationIo::SaveSettingsToFile(const std::string & file_name)
{
    IniSaver  Saver(file_name, false);   // not local file path
    InstallFileIoSettings(Saver);
    Saver.Save();   // To File
}


//------------------------------------------------------------------------------
//  ApplicationIo::InstallFileIoSettings() -- Install data for Settings Load
//------------------------------------------------------------------------------
//  These should follow those in the Settings object closely
//

void  ApplicationIo::InstallFileIoSettings(IniSaver & Saver)
{
    // Install Settings
    //  Config Page
    //  InMux Page
    Saver.Install( ToIni("InMux0Source",     Settings.InMux.Channel[0].Source,      0) );
    Saver.Install( ToIni("InMux0DivIdx",     Settings.InMux.Channel[0].DivideByIdx, 0) );
    Saver.Install( ToIni("InMux0Phase",      Settings.InMux.Channel[0].Phase,      0u) );
    Saver.Install( ToIni("InMux1Source",     Settings.InMux.Channel[1].Source,      0) );
    Saver.Install( ToIni("InMux1DivIdx",     Settings.InMux.Channel[1].DivideByIdx, 0) );
    Saver.Install( ToIni("InMux1Phase",      Settings.InMux.Channel[1].Phase,      0u) );
    Saver.Install( ToIni("InMux2Source",     Settings.InMux.Channel[2].Source,      0) );
    Saver.Install( ToIni("InMux2DivIdx",     Settings.InMux.Channel[2].DivideByIdx, 0) );
    Saver.Install( ToIni("InMux2Phase",      Settings.InMux.Channel[2].Phase,      0u) );
    Saver.Install( ToIni("InMux3Source",     Settings.InMux.Channel[3].Source,      0) );
    Saver.Install( ToIni("InMux3DivIdx",     Settings.InMux.Channel[3].DivideByIdx, 0) );
    Saver.Install( ToIni("InMux3Phase",      Settings.InMux.Channel[3].Phase,      0u) );
    Saver.Install( ToIni("InMux4Source",     Settings.InMux.Channel[4].Source,      0) );
    Saver.Install( ToIni("InMux4DivIdx",     Settings.InMux.Channel[4].DivideByIdx, 0) );
    Saver.Install( ToIni("InMux4Phase",      Settings.InMux.Channel[4].Phase,      0u) );
    //  Vcxo Page
    Saver.Install( ToIni("FrontPanelInputRate",   Settings.Vcxo.FrontPanelInRateMhz, 20.0) );
    //  Pll Page
    Saver.Install( ToIni("PllFrequency",      Settings.Pll.Frequency,   250.0) );
    Saver.Install( ToIni("PllOutPhase",       Settings.Pll.PllOut.Phase,   0u) );
    Saver.Install( ToIni("DebugOutPhase",     Settings.Pll.DebugOut.Phase,   0u) );
    Saver.Install( ToIni("PhaseDetectorFrequency",     Settings.Pll.PhaseDetectorFrequency, 10e-3) );
    Saver.Install( ToIni("PllBoost",                   Settings.Pll.Boost,   0) );
    //  Gps Page
    Saver.Install( ToIni("GpsEnabled",      Settings.Gps.Enabled,          true) );

    //  OutMux Page
    Saver.Install( ToIni("OutMux0Source",     Settings.OutMux.Channel[0].Source,      0) );
    Saver.Install( ToIni("OutMux0DivIdx",     Settings.OutMux.Channel[0].DivideByIdx, 0) );
    Saver.Install( ToIni("OutMux0Phase",      Settings.OutMux.Channel[0].Phase,      0u) );
    Saver.Install( ToIni("OutMux1Source",     Settings.OutMux.Channel[1].Source,      0) );
    Saver.Install( ToIni("OutMux1DivIdx",     Settings.OutMux.Channel[1].DivideByIdx, 0) );
    Saver.Install( ToIni("OutMux1Phase",      Settings.OutMux.Channel[1].Phase,      0u) );
    Saver.Install( ToIni("OutMux2Source",     Settings.OutMux.Channel[2].Source,      0) );
    Saver.Install( ToIni("OutMux2DivIdx",     Settings.OutMux.Channel[2].DivideByIdx, 0) );
    Saver.Install( ToIni("OutMux2Phase",      Settings.OutMux.Channel[2].Phase,      0u) );
    Saver.Install( ToIni("OutMux3Source",     Settings.OutMux.Channel[3].Source,      0) );
    Saver.Install( ToIni("OutMux3DivIdx",     Settings.OutMux.Channel[3].DivideByIdx, 0) );
    Saver.Install( ToIni("OutMux3Phase",      Settings.OutMux.Channel[3].Phase,      0u) );

    //  PxiMux Page
    Saver.Install( ToIni("PxiMux0Source",     Settings.PxiMux.Channel[0].Source,      0) );
    Saver.Install( ToIni("PxiMux0DivIdx",     Settings.PxiMux.Channel[0].DivideByIdx, 0) );
    Saver.Install( ToIni("PxiMux0Phase",      Settings.PxiMux.Channel[0].Phase,      0u) );
    Saver.Install( ToIni("PxiMux1Source",     Settings.PxiMux.Channel[1].Source,      0) );
    Saver.Install( ToIni("PxiMux1DivIdx",     Settings.PxiMux.Channel[1].DivideByIdx, 0) );
    Saver.Install( ToIni("PxiMux1Phase",      Settings.PxiMux.Channel[1].Phase,      0u) );
    Saver.Install( ToIni("PxiMux2Source",     Settings.PxiMux.Channel[2].Source,      0) );
    Saver.Install( ToIni("PxiMux2DivIdx",     Settings.PxiMux.Channel[2].DivideByIdx, 0) );
    Saver.Install( ToIni("PxiMux2Phase",      Settings.PxiMux.Channel[2].Phase,      0u) );
    Saver.Install( ToIni("PxiMux3Source",     Settings.PxiMux.Channel[3].Source,      0) );
    Saver.Install( ToIni("PxiMux3DivIdx",     Settings.PxiMux.Channel[3].DivideByIdx, 0) );
    Saver.Install( ToIni("PxiMux3Phase",      Settings.PxiMux.Channel[3].Phase,      0u) );

    //  Trigger Page
    Saver.Install( ToIni("ExternalTrigger",       Settings.Trigger.ExternalTrigger, 0) );
    Saver.Install( ToIni("Framed",                Settings.Trigger.Framed, 0) );
    Saver.Install( ToIni("FrameSize",             Settings.Trigger.FrameSize,   0x4000) );
    Saver.Install( ToIni("TriggerMasterEnable",   Settings.Trigger.MasterTriggerEnabled,  true) );
    Saver.Install( ToIni("TriggerActivate",       Settings.Trigger.SoftwareTriggerActivated,  true) );
    Saver.Install( ToIni("TriggerSampleClockSrc",  Settings.Trigger.SampleClockSource,  0) );

    //  Debug Page

}


//---------------------------------------------------------------------------
//  ApplicationIo::ApplyInMuxSettings() -- Load settings for InMux
//---------------------------------------------------------------------------

void  ApplicationIo::ApplyInMuxSettings()
{
    //  Apply Front Panel Rate
    Module.FrontPanelRateMhz(Settings.Vcxo.FrontPanelInRateMhz);

    Module.InMux().Powered(true);

    for (unsigned int i=0; i<MaxInMuxChannels; ++i)
        {
        Module.InMux().Source(i, Settings.InMux.Channel[i].Source);
        Module.InMux().Divider(i, Module.InMux().ToDivider(Settings.InMux.Channel[i].DivideByIdx));
        Module.InMux().Phase(i, Module.InMux().ToDivider(Settings.InMux.Channel[i].DivideByIdx),
                                Settings.InMux.Channel[i].Phase);
        Module.InMux().Boost(i, (Settings.InMux.Boost)? true : false);
        }
    Module.InMux().Reapply();

    for (unsigned int i=0; i<3; i++)
        Settings.InMux.SourceRate[i] = Module.InMux().SourceFrequency(i)/1e6;
    for (unsigned int i=0; i<5; i++)
        Settings.InMux.OutputRate[i] = Module.InMux().OutputFrequencyActual(i)/1e6;
}

//---------------------------------------------------------------------------
//  ApplicationIo::ChangeVcxoFrequency() --  Manually set Vcxo
//---------------------------------------------------------------------------

void  ApplicationIo::ChangeVcxoFrequency(float mhz)
{
    Module.Pll().SetActiveVco(0);
    Module.Vcxo().Frequency(mhz);
}

//---------------------------------------------------------------------------
//  ApplicationIo::LoadVco() --  Manually set Vcxo
//---------------------------------------------------------------------------

void  ApplicationIo::LoadVco(int hsdiv, int n1, ii64 rfreq)
{
    Module.Pll().SetActiveVco(0);
    Module.Vcxo().LoadFreq(hsdiv, n1, rfreq);
}

//---------------------------------------------------------------------------
//  ApplicationIo::ApplyRefSettings() --  Change Reference Settings
//---------------------------------------------------------------------------

void  ApplicationIo::ApplyRefSettings()
{
    Module.Ref().TempCompEnabled(Settings.Vcxo.RefTempCompEnable);
    Module.Ref().TuningPPB(Settings.Vcxo.RefFrequencyAdjust);
    Module.Ref().Reapply();
}

//---------------------------------------------------------------------------
//  ApplicationIo::TimerStart() --  Turn on timers
//---------------------------------------------------------------------------

void  ApplicationIo::TimerStart()
{
    //  Apply useful settings before start
    ApplyInMuxSettings();
    ApplyPllSettings();
    ApplyOutMuxSettings();
    ApplyPxiMuxSettings();
    ApplyTriggerSettings();

    Module.Start();
}

//---------------------------------------------------------------------------
//  ApplicationIo::TimerStop() --  Turn off timers
//---------------------------------------------------------------------------

void  ApplicationIo::TimerStop()
{
    Module.Stop();
}

//---------------------------------------------------------------------------
//  ApplicationIo::DisplayPllData() -- Display PLL Info
//---------------------------------------------------------------------------

void  ApplicationIo::DisplayPllData()
{
    //  Read the data
    Cdce72010::InfoBlock Info;
    Module.Pll().ReadInfo(Info);

    std::stringstream txt;

    txt << "------ PLL Info -------" << "\r\n"
        << " M = " << Info.M << "\r\n"
        << " R = " << Info.R << "\r\n"
        << " N = " << Info.N << "\r\n"
        << " FB = " << Info.Fb << "\r\n"
        << " X = " << Info.X << "\r\n"
        << std::scientific
        << " Ref Rate = " << Info.RefFrequency << "\r\n"
        << " Vcxo Rate = " << Info.VcoFrequency << "\r\n"
        << " Out Rate = " << Info.FrequencyActual << "\r\n";

   // Log(txt.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::ChangePllFrequency() --  Manually set Pll Clock
//---------------------------------------------------------------------------

void  ApplicationIo::ChangePllFrequency()
{
    Module.Pll().ReferenceSource(Cdce72010::refPrimary);
    Module.Pll().Frequency(Settings.Pll.Frequency*1e6);
    Module.Pll().Reapply();
    //
    //  After setting the frequency, read back the mux setting
    Settings.Pll.PllOut.DivideByIdx =
                Module.Pll().ToDividerIdx(Module.Pll().PllOutputDivider());
    Settings.Pll.DebugOut.DivideByIdx =
                Module.Pll().ToDividerIdx(Module.Pll().DebugDivider());

    for (unsigned int i=0; i<2; i++)
        Settings.Pll.ReferenceRate[i] = Module.Pll().SourceFrequency(i)/1e6;

    Settings.Pll.OutRateActual = Module.Pll().FrequencyActual()/1e6;
    Settings.Pll.DebugRateActual = Module.Pll().OutputFrequencyActual(8)/1e6;
}

//---------------------------------------------------------------------------
//  ApplicationIo::LoadPll() --  Manually set Pll
//---------------------------------------------------------------------------

void  ApplicationIo::LoadPll(int m, int n, int fb, int divider_idx)
{
    int divider = Module.Pll().ToDivider(divider_idx);
    Module.Pll().SetActiveVco(0);
    Module.Pll().LoadFreq(m, n, fb, divider);
}

//---------------------------------------------------------------------------
//  ApplicationIo::ApplyPllSettings() -- Load settings for Pll
//---------------------------------------------------------------------------

void  ApplicationIo::ApplyPllSettings()
{
    Module.Pll().Powered(true);

    Module.Pll().SetPDF(Settings.Pll.PhaseDetectorFrequency*1e6);
    Module.Pll().PllOutputBoost((Settings.Pll.Boost)? true : false);
    Module.Pll().DebugBoost((Settings.Pll.Boost)? true : false);

    Module.Pll().Reapply();

    ChangePllFrequency();

}

//---------------------------------------------------------------------------
//  ApplicationIo::GpsCommand() --  Turn off gps
//---------------------------------------------------------------------------

void  ApplicationIo::GpsCommand(const std::string & text)
{
    Module.Gps().Command(text);
}

//---------------------------------------------------------------------------
//  ApplicationIo::GpsEnable() --
//---------------------------------------------------------------------------

void  ApplicationIo::GpsEnable()
{
    Module.Gps().Enabled(Settings.Gps.Enabled);
}

//---------------------------------------------------------------------------
//  ApplicationIo::GpsTriggerSoon() --
//---------------------------------------------------------------------------
#ifdef _MSC_VER
    #pragma warning(push)
    #pragma warning(disable : 4996)
#endif

void  ApplicationIo::GpsTriggerSoon(int seconds_from_now)
{
    std::time_t  trigger_time = Module.Gps().SecondsFromNow(seconds_from_now);
    Module.Gps().GpsTriggerTime(trigger_time);

    std::stringstream msg;
    msg << "Trigger to fire at: " << std::ctime(&trigger_time);
    //Log(msg.str());
}
#ifdef _MSC_VER
    #pragma warning(pop)
#endif

//---------------------------------------------------------------------------
//  ApplicationIo::HandleEpoch() --  Epoch interrupt
//---------------------------------------------------------------------------

void  ApplicationIo::HandleEpoch(Innovative::GpsEpochEvent & Event)
{
    if (Settings.Gps.DisplayRawInfo)
        Log(Event.List.AsText());   // dump raw list
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleGpsArm() --  Trigger Arming Interrupt
//---------------------------------------------------------------------------

void  ApplicationIo::HandleGpsArm(OpenWire::NotifyEvent & /*Event*/)
{
    //UI->PeriodicStatus();
    Log("Gps Trigger Armed: trigger in 1 second...");
}


//---------------------------------------------------------------------------
//  ApplicationIo::GpsDateTime() --  Display current Gps Date/Time
//---------------------------------------------------------------------------
#ifdef _MSC_VER
    #pragma warning(push)
    #pragma warning(disable : 4996)
#endif

std::string  ApplicationIo::GpsDateTime()
{
    std::time_t  Time = Module.Gps().Time();
    char * text =  std::ctime(&Time);
    std::string  TimeText( text );
    return Trim(TimeText);
}
#ifdef _MSC_VER
    #pragma warning(pop)
#endif

//---------------------------------------------------------------------------
//  ApplicationIo::GpsSatellites() --  Display current Satellites
//---------------------------------------------------------------------------

unsigned int  ApplicationIo::GpsSatellites()
{
    return Module.Gps().Satellites();
}

//---------------------------------------------------------------------------
//  ApplicationIo::GpsFix() --  Display current Fix Status
//---------------------------------------------------------------------------

std::string  ApplicationIo::GpsFix()
{
    IGps::FixStatus fix = Module.Gps().Fix();
    switch (fix)
        {
            default:
            case IGps::fixNoFix:     return "NF";
            case IGps::fixGps:       return "GPS";
            case IGps::fixDiffGps:   return "DGPS";
        }
}

//---------------------------------------------------------------------------
//  ApplicationIo::GpsLongitude() --  Display current Fix Status
//---------------------------------------------------------------------------

std::string  ApplicationIo::GpsLongitude()
{
    GpsCoordinate Long = Module.Gps().Longitude();

    std::stringstream txt;

    txt << Long.Degrees << "\u00b0" << Long.Minutes << "  " << Long.Units;

    return Trim(txt.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::GpsLatitude() --  Display current Fix Status
//---------------------------------------------------------------------------

std::string  ApplicationIo::GpsLatitude()
{
    GpsCoordinate Lat = Module.Gps().Latitude();

    std::stringstream txt;


    txt << Lat.Degrees << "\u00b0" << Lat.Minutes << "  " << Lat.Units;

    return Trim(txt.str());
}


//---------------------------------------------------------------------------
//  ApplicationIo::ApplyOutMuxSettings() -- Load settings for OutMux
//---------------------------------------------------------------------------

void  ApplicationIo::ApplyOutMuxSettings()
{
    Module.OutMux().Powered(true);

    for (unsigned int i=0; i<MaxOutMuxChannels; ++i)
        {
        Module.OutMux().Source(i, Settings.OutMux.Channel[i].Source);
        Module.OutMux().Divider(i, Module.OutMux().ToDivider(Settings.OutMux.Channel[i].DivideByIdx));
        Module.OutMux().Phase(i, Module.OutMux().ToDivider(Settings.OutMux.Channel[i].DivideByIdx),
                                Settings.OutMux.Channel[i].Phase);
        Module.OutMux().Boost(i, (Settings.OutMux.Boost) ? true : false);
        }
    Module.OutMux().Reapply();

    for (unsigned int i=0; i<3; ++i)
        Settings.OutMux.SourceRate[i] = Module.OutMux().SourceFrequency(i)/1e6;
    for (unsigned int i = 0; i<MaxOutMuxChannels; ++i)
        Settings.OutMux.OutputRate[i] = Module.OutMux().OutputFrequencyActual(i)/1e6;
}

//---------------------------------------------------------------------------
//  ApplicationIo::ApplyPxiMuxSettings() -- Load settings for PxiMux
//---------------------------------------------------------------------------

void  ApplicationIo::ApplyPxiMuxSettings()
{
    Module.PxiMux().Powered(true);

    Module.PxiMux().DriveDSTAROutputs(Settings.PxiMux.DriveDSTAROut);
    Module.PxiMux().Drive10MHzOutput(Settings.PxiMux.Drive10MHzOut);
    Module.PxiMux().TriggerOutputEnable(Settings.PxiMux.EnableTriggerOut);

    for (unsigned int i=0; i<MaxPxiMuxChannels; ++i)
        {
        Module.PxiMux().Source(i, Settings.PxiMux.Channel[i].Source);
        Module.PxiMux().Divider(i, Module.PxiMux().ToDivider(Settings.PxiMux.Channel[i].DivideByIdx));
        Module.PxiMux().Phase(i, Module.PxiMux().ToDivider(Settings.PxiMux.Channel[i].DivideByIdx),
                                Settings.PxiMux.Channel[i].Phase);
        Module.PxiMux().Boost(i, (Settings.PxiMux.Boost) ? true : false );
        }
    Module.PxiMux().Reapply();

    for (unsigned int i=0; i<3; i++)
        Settings.PxiMux.SourceRate[i] = Module.PxiMux().SourceFrequency(i)/1e6;
    for (unsigned int i=0; i<MaxPxiMuxChannels; i++)
        Settings.PxiMux.OutputRate[i] = Module.PxiMux().OutputFrequencyActual(i)/1e6;
}

//---------------------------------------------------------------------------
//  ApplicationIo::ApplyTriggerSettings() -- Set up trigger stuff
//---------------------------------------------------------------------------

void  ApplicationIo::ApplyTriggerSettings()
{
    if (Settings.Trigger.Framed)
        Module.Framed(Settings.Trigger.FrameSize);
    else
        Module.Unframed();

    Module.ExternalTrigger((Settings.Trigger.ExternalTrigger == 1));

    TriggerMasterEnable(Settings.Trigger.MasterTriggerEnabled);
    TriggerActivate(Settings.Trigger.SoftwareTriggerActivated);
    Module.TriggerSampleClock(Settings.Trigger.SampleClockSource);

    Module.PxiMux().Reapply();
    Module.OutMux().Reapply();

    int source = Settings.Trigger.SampleClockSource;
    Settings.Trigger.SourceRate    = Module.OutMux().SourceFrequency(source)/1e6;
    Settings.Trigger.OutputRate[0] = Module.OutMux().OutputFrequencyActual(4)/1e6;
    Settings.Trigger.OutputRate[1] = Module.PxiMux().OutputFrequencyActual(4)/1e6;
    Settings.Trigger.DivideByIdx[0] = Module.OutMux().Divider(4);
    Settings.Trigger.DivideByIdx[1] = Module.PxiMux().Divider(4);
}

//---------------------------------------------------------------------------
//  ApplicationIo::TriggerMasterEnable() -- Set/Reset Master Trigger
//---------------------------------------------------------------------------

void  ApplicationIo::TriggerMasterEnable(bool state)
{
    Settings.Trigger.MasterTriggerEnabled = state;
    //  Hit hardware
    Module.ArmTrigger(state);
}

//---------------------------------------------------------------------------
//  ApplicationIo::TriggerActivate() -- SW Trigger On/Off
//---------------------------------------------------------------------------

void  ApplicationIo::TriggerActivate(bool state)
{
    Settings.Trigger.SoftwareTriggerActivated = state;
    //  Hit hardware
    Module.SoftwareTrigger(state);
}

//---------------------------------------------------------------------------
//  ApplicationIo::Log() --  Log message thunked to main thread
//---------------------------------------------------------------------------

void  ApplicationIo::Log(const std::string & msg)
{
    ProcessStatusEvent e(msg);
    OnLog.Execute(e);
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleOnLog() --
//---------------------------------------------------------------------------

void ApplicationIo::HandleOnLog(Innovative::ProcessStatusEvent & Event)
{
   // UI->Log(Event.Message);
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  Support Functions
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//---------------------------------------------------------------------------
//  ApplicationIo::Temperature() --  Query current module temperature
//---------------------------------------------------------------------------

float ApplicationIo::Temperature()
{
    return Module.Thermal().LogicTemperature();
}

//---------------------------------------------------------------------------
//  ApplicationIo::ReferenceTemperature() --  Query Reference Temperature
//---------------------------------------------------------------------------

float ApplicationIo::ReferenceTemperature()
{
    return Module.Ref().Temperature();
}

//---------------------------------------------------------------------------
//  ApplicationIo::TriggerStatus() -- Display Trigger Status
//---------------------------------------------------------------------------

std::string  ApplicationIo::TriggerStatus()
{
    bool master_state = Module.ArmTrigger();
    bool sw_state = Module.SoftwareTrigger();
    std::string master_str = (master_state) ? "+" : "-" ;
    std::string sw_str = (sw_state) ? "+" : "-" ;

    std::stringstream msg;
    msg << std::hex << "M: " << master_str
        << "  SW: " << sw_str;
    return msg.str();
}

//---------------------------------------------------------------------------
//  ApplicationIo::PllLockedStr() -- Display Pll Lock Status
//---------------------------------------------------------------------------

std::string  ApplicationIo::PllLockedStr()
{
    bool locked_state = Module.Pll().Locked();

    if (locked_state)
        return std::string(" + ");
    else
        return std::string(" - ");
}

//------------------------------------------------------------------------------
//  ApplicationIo::ModuleName() --  Return name of module as string
//------------------------------------------------------------------------------

std::string ApplicationIo::ModuleName() const
{
    return Module.Name();
}

//------------------------------------------------------------------------------
//  ApplicationIo::ModuleRevision() --  Return rev of module as string
//------------------------------------------------------------------------------

std::string ApplicationIo::ModuleRevision()
{
    std::stringstream msg;
    char c = 'A' + static_cast<char>(Module.PciLogicVersion().PciLogicPcb());
    msg << c;
    return msg.str();
}

//---------------------------------------------------------------------------
//  ApplicationIo::Peek() --
//---------------------------------------------------------------------------

unsigned int ApplicationIo::Peek(const string &type, int offset)
{
    (void) type;
    (void) offset;

    return 0;
}

//---------------------------------------------------------------------------
//  ApplicationIo::Poke() --
//---------------------------------------------------------------------------

void ApplicationIo::Poke(const string &type, int offset, unsigned int value)
{
    (void) type;
    (void) value;
    (void) offset;
}

//---------------------------------------------------------------------------
//  ApplicationIo::UpdateSummaryOutputDisplayPrefixes() --  Change display
//---------------------------------------------------------------------------

void ApplicationIo::UpdateSummaryOutputDisplayPrefixes()
{
    if (Settings.Board.SummaryDisplayIdx==0)
        {
        Settings.Board.Prefix[0] = "Out 0: ";  Settings.Board.Prefix[1] = "Out 1: ";
        Settings.Board.Prefix[2] = "Out 2: ";  Settings.Board.Prefix[3] = "Out 3: ";
        }
    else
        {
        Settings.Board.Prefix[0] = "DSTAR A: "; Settings.Board.Prefix[1] = "DSTAR B: ";
        Settings.Board.Prefix[2] = "DSTAR C: "; Settings.Board.Prefix[3] = "Pxi In: ";
        }

}


// Ctor
ClockSettings::ClockSettings()
{

}


ClockSettings::~ClockSettings()
{

}






} // namespace Clock
} // namespace Innovative
