#ifndef CLOCKMODULE_GP_H
#define CLOCKMODULE_GP_H

#include <X3_Timing_Mb.h>
#include <StopWatch_Mb.h>
#include <IniFile_Mb.h>
#include <Benchmark_Mb.h>
#include <SoftwareTimer_Mb.h>
#include <fstream>
#include <Application/ScriptPlayer_App.h>
#include <Application/IniSaver_App.h>
#include <Exception_Mb.h>


namespace Innovative {
namespace Clock      {

//===========================================================================
//  CLASS Settings --
//===========================================================================

const unsigned int MaxInMuxChannels = 5;
const unsigned int MaxOutMuxChannels = 4;
const unsigned int MaxPxiMuxChannels = 4;

struct ClockSettings
{
    // Ctor
    ClockSettings();
    ~ClockSettings();

    struct MuxSetting
    {
        int            Source;
        int            DivideByIdx;
        unsigned int   Phase;
    };

    //
    //  Debug Settings
    struct BoardCfg
    {
        int             Target;
        int             Digits;
        int             SummaryDisplayIdx;
        std::string     Prefix[4];
        bool            DisplayDebugControls;
    };

    struct InMuxCfg
    {
        int             Boost;
        MuxSetting      Channel[MaxInMuxChannels];
        double          SourceRate[3];
        double          OutputRate[MaxInMuxChannels];
    };

    struct VcxoCfg
    {
        bool            RefTempCompEnable;
        int             RefFrequencyAdjust;
        double          FrontPanelInRateMhz;
    };

    struct PllCfg
    {
        int             Boost;
        double          Frequency;
        double          PhaseDetectorFrequency;
        MuxSetting      PllOut;
        MuxSetting      DebugOut;
        double          ReferenceRate[2];
        double          OutRateActual;
        double          DebugRateActual;
    };

    struct GpsCfg
    {
        bool            DisplayRawInfo;
        bool            Enabled;
    };

    struct OutMuxCfg
    {
        int             Boost;
        MuxSetting      Channel[MaxOutMuxChannels];
        double          SourceRate[3];
        double          OutputRate[MaxOutMuxChannels];
    };

    struct PxiMuxCfg
    {
        bool            DriveDSTAROut;
        bool            Drive10MHzOut;
        bool            EnableTriggerOut;
        int             Boost;
        MuxSetting      Channel[MaxPxiMuxChannels];
        double          SourceRate[3];
        double          OutputRate[MaxPxiMuxChannels];
    };

    struct TriggerCfg
    {
        bool            MasterTriggerEnabled;
        bool            SoftwareTriggerActivated;
        int             Framed;
        int             FrameSize;
        int             ExternalTrigger;
        int             SampleClockSource;
        double          SourceRate;
        double          OutputRate[2];
        int             DivideByIdx[2];
    };

    struct DebugCfg
    {
        std::string     Script;
        int             DioConfig;
    };

    struct BoardPartInfo
    {
        unsigned int   PartNo;
        unsigned int   DashNo;
        std::string    DashVar;
        unsigned int   RuggedNo;
    };

    BoardCfg        Board;
    InMuxCfg        InMux;
    VcxoCfg         Vcxo;
    PllCfg          Pll;
    GpsCfg          Gps;
    OutMuxCfg       OutMux;
    PxiMuxCfg       PxiMux;
    TriggerCfg      Trigger;
    DebugCfg        Debug;

    BoardPartInfo   PartInfo;

    std::string     ModuleName;
    std::string     ModuleRevision;
    double          VcxoPwrOnRateMhz;
};

//===========================================================================
//  CLASS ApplicationIo  -- Hardware Access and Application Io Class
//===========================================================================

class ApplicationIo
{
public:
    //
    //  Member Functions
    ApplicationIo(unsigned int target, ClockSettings &settings);
    ~ApplicationIo();
    //
    //  Config Page
    int    BoardCount();
    void   Open();
    bool   IsOpen()
        {  return Opened;  }
    void   Close();
    void  AfterGetSettings();

    void  LoadSettingsFromFile(const std::string & file_name);
    void  SaveSettingsToFile(const std::string & file_name);
    void  InstallFileIoSettings(Innovative::IniSaver & saver);

    unsigned int Peek(const string &type, int offset);
    void         Poke(const string &type, int offset, unsigned int value);
    //
    //   InMux Page
    void  ApplyInMuxSettings();
    //
    //   Vco Page
    void   ChangeVcxoFrequency(float mhz);
    void   LoadVco(int hsdiv, int n1, ii64 rfreq);
    void   ApplyRefSettings();
    //
    //  ToolButton Bar
    void   TimerStart();
    void   TimerStop();
    void   DisplayPllData();
    //
    //   Pll Page
    void   ChangePllFrequency();
    void   LoadPll(int m, int n, int fb, int divider_idx);
    void   ApplyPllSettings();
    //
    //   Gps Page
    void   GpsCommand(const std::string & text);
    void   GpsEnable();
    void   GpsTriggerSoon(int seconds_from_now);

    std::string   GpsDateTime();
    unsigned int  GpsSatellites();
    std::string   GpsFix();
    std::string   GpsLongitude();
    std::string   GpsLatitude();
    //
    //   OutMux Page
    void  ApplyOutMuxSettings();
    //
    //   PxiMux Page
    void  ApplyPxiMuxSettings();
    //
    //   Trigger Page
    void  TriggerMasterEnable(bool state);
    void  TriggerActivate(bool state);
    void  ApplyTriggerSettings();

    //
    //   Status Bar
    float  Temperature();
    float  ReferenceTemperature();
    std::string  TriggerStatus();
    std::string  PllLockedStr();
    //
    //  Where are these?
    bool   IsTriggered()
        {  return FTriggered >= 0;  }

    void UpdateSummaryOutputDisplayPrefixes();

    std::string     ModuleName() const;
    std::string     ModuleRevision();

    // Data
    ClockSettings     &Settings;

private:
    //
    //  Member Data
    Innovative::X3_Timing               Module;
    bool                                Opened;
    unsigned int                        Target;
    bool                                Stopped;
    float                               ActualSampleRate;
    int                                 FTicks;
    int                                 FTriggered;

    std::string                         SettingsFileName;

    //
    //  Member Functions
    void  HandleEpoch(Innovative::GpsEpochEvent & Event);
    void  HandleGpsArm(OpenWire::NotifyEvent & Event);

    void  HandleOnLog(Innovative::ProcessStatusEvent & Event);
    void  Log(const std::string & msg);
    OpenWire::ThunkedEventHandler<Innovative::ProcessStatusEvent>  OnLog;


};



} // namespace Clock
} // namespace Innovative

#endif // CLOCKMODULE_GP_H
