#include <Components_Gp.h>
#include <iomanip>

namespace   Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

using namespace GpDsp_MemoryMap;

//===========================================================================
//  CLASS GspDspProjectInfo  -- Project wide Information Block
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class GspDspProjectInfo
//---------------------------------------------------------------------------

static GspDspProjectInfo::MapRegisters GpDspProject_RegAddrs =
    {  mmPrjVersion,        //  ProjectVersion_Addr
       mmPrj_MajorCores,    //  ModuleCoreEnable_Addr
       mmPrjSubVersion
    };

//---------------------------------------------------------------------------
//  constructor for class GspDspProjectInfo
//---------------------------------------------------------------------------

GspDspProjectInfo::GspDspProjectInfo(IAddressingSpace & space)
    :  ProjectVersionReg(space, GpDspProject_RegAddrs.ProjectVersion_Addr),
       ModuleCoreEnableReg(space, GpDspProject_RegAddrs.ModuleCoreEnable_Addr),
       ProjectSubVersionReg(space, GpDspProject_RegAddrs.ProjectSubVersion_Addr)
{
}

//---------------------------------------------------------------------------
//  GspDspProjectInfo::Enable() -
//---------------------------------------------------------------------------

bool GspDspProjectInfo::Enabled(Cores c)
{
    struct CoreMapType
    {
        Cores        ct;
        unsigned int val;
    } CoreMap[] = {
       { CustorProj ,0x01 },
       { DspCommon  ,0x02 },
       { DDC        ,0x04 },
       { DUC        ,0x08 },
       { Spectrum   ,0x10 },
       { Beamforming,0x20 },
       { PSK        ,0x40 },
       { FSK        ,0x80 }
    };

    bool result = false;
    unsigned int bits = ModuleCoreEnableReg.CoreEnabled.Value();
    for (size_t i = 0; i < sizeof(CoreMap) / sizeof(CoreMapType); i++)
        if (c == CoreMap[i].ct)
            result = bits & CoreMap[i].val;

    //avoid the warning
    return result;
}

//---------------------------------------------------------------------------
//  GspDspProjectInfo::Version() -
//---------------------------------------------------------------------------

unsigned int GspDspProjectInfo::Version()
{
    return ProjectVersionReg.Value();
}

//---------------------------------------------------------------------------
//  GspDspProjectInfo::VersionString() -
//---------------------------------------------------------------------------

std::string GspDspProjectInfo::VersionString()
{
    std::stringstream ss;
    ss << std::hex <<
          std::setfill('0') << std::setw(2) <<
          ProjectVersionReg.Additions.Value() << "_" <<
          std::setfill('0') << std::setw(2) <<
          ProjectVersionReg.TxCoreType.Value() << std::setw(0) << ProjectVersionReg.TxCores.Value() << "_" <<
          std::setfill('0') << std::setw(2) <<
          ProjectVersionReg.RxCoreType.Value() << std::setw(0) << ProjectVersionReg.RxCores.Value();

    return ss.str();
}

//---------------------------------------------------------------------------
//  GspDspProjectInfo::SubVersion() -
//---------------------------------------------------------------------------

unsigned int GspDspProjectInfo::SubVersion()
{
    return ProjectSubVersionReg.Value();
}

//---------------------------------------------------------------------------
//  GspDspProjectInfo::SubVersionString() -
//---------------------------------------------------------------------------

std::string GspDspProjectInfo::SubVersionString()
{
    std::stringstream ss;
    ss << std::hex << "v" <<
                 ProjectSubVersionReg.Revision_Hi.Value() << "." <<
                 ProjectSubVersionReg.Revision_Lo.Value() << "." <<
                 ProjectSubVersionReg.EngineerID.Value() << "." <<
                 ProjectSubVersionReg.SubRevInternal.Value();
    return ss.str();
}

//===========================================================================
//  CLASS GspDspModuleBase  -- Component shared features
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class GspDspModuleBase
//---------------------------------------------------------------------------

static GspDspModuleBase::MapRegisters DspBase_RegAddrs =
    {  mmHeader_CoresPresent,     //  ModuleEn_Addr
       mmHeader_CmdTrigger,       //  CmdTrigger_Addr
       mmHeader_CoreIndex,        //  CoreIndex_Addr;
       mmHeader_ParentIndex,      //  ParentIndex_Addr;
       mmSyncReset                //  SyncReset_Addr;
    };

//---------------------------------------------------------------------------
//  constructor for class GspDspModuleBase
//---------------------------------------------------------------------------

GspDspModuleBase::GspDspModuleBase(IAddressingSpace & space, WishboneType awtype)
    :  ModuleEnReg(space, DspBase_RegAddrs.ModuleEn_Addr),
       CmdTriggerReg(space, DspBase_RegAddrs.CmdTrigger_Addr, awtype),
       CoreIndexReg(space, DspBase_RegAddrs.CoreIndex_Addr, awtype),
       ParentIndexReg(space, DspBase_RegAddrs.ParentIndex_Addr, awtype),
       SyncResetReg(space, DspBase_RegAddrs.SyncReset_Addr, awtype)
{
}

//---------------------------------------------------------------------------
//  GspDspModuleBase::GetEnableBits() - read module enable field
//---------------------------------------------------------------------------

unsigned int  GspDspModuleBase::GetEnableBits()
{
    return ModuleEnReg.ModuleEnabled.Value();
}

//---------------------------------------------------------------------------
//  GspDspModuleBase::SetCommandTrigger() - set command trigger
//---------------------------------------------------------------------------

void  GspDspModuleBase::SetCommandTrigger(bool state)
{
    CmdTriggerReg.Trigger.Value(state);
}

//---------------------------------------------------------------------------
//  GspDspModuleBase::IsBitEnabled() -
//---------------------------------------------------------------------------

bool GspDspModuleBase::IsBitEnabled(GspDspModuleBase::EnableBitsEnum value)
{
    return (GetEnableBits() & value) != 0;
}

//---------------------------------------------------------------------------
//  GspDspModuleBase::CommandTrigger() - set command trigger on-off
//---------------------------------------------------------------------------

void  GspDspModuleBase::CommandTrigger()
{
    SetCommandTrigger(true);
    SetCommandTrigger(false);
}

//---------------------------------------------------------------------------
//  GspDspModuleBase::SetIndex() - set index registers
//---------------------------------------------------------------------------

void  GspDspModuleBase::SetIndex(unsigned int parent, unsigned int core)
{
    ParentIndexReg.ParentIndex.Value(parent);
    CoreIndexReg.CoreIndex.Value(core);
}

//---------------------------------------------------------------------------
//  GspDspModuleBase::Reset() - set command trigger on-off
//---------------------------------------------------------------------------

void  GspDspModuleBase::Reset(bool state)
{
    SyncResetReg.Value(state);
}


//===========================================================================
//  CLASS GpDspSubmodule  -- Subcomponent shared registers
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class GpDspSubmodule
//---------------------------------------------------------------------------

GpDspSubmodule::GpDspSubmodule(GspDspModuleBase & cbase,
                               IAddressingSpace & space, MapRegisters & regs, WishboneType awtype)
    :  VersionReg(space, regs.Version_Addr),
       ConfigReg(space, regs.Config_Addr),
       ResetReg(space, regs.Reset_Addr, awtype),
       Module(cbase)
{
}

//---------------------------------------------------------------------------
//  GpDspSubmodule::SetIndex() -
//---------------------------------------------------------------------------

void  GpDspSubmodule::SetIndex(unsigned int core)
{
    Module.SetIndex(core);
}

//---------------------------------------------------------------------------
//  GpDspSubmodule::SetIndex() -
//---------------------------------------------------------------------------

void  GpDspSubmodule::SetIndex(unsigned int parent, unsigned int core)
{
    Module.SetIndex(parent, core);
}

//---------------------------------------------------------------------------
//  GpDspSubmodule::CoreCount() -
//---------------------------------------------------------------------------

unsigned int GpDspSubmodule::CoreCount() const
{
    return VersionReg.CoreCount.Value();
}

//---------------------------------------------------------------------------
//  GspDspModuleBase::WhichArchitecture() -
//---------------------------------------------------------------------------

GpDspSubmodule::ArhitectureType GpDspSubmodule::WhichArchitecture()
{
    ArhitectureType arc = aParallel;
    switch (ConfigReg.Architecture.Value())
    {
    case 0: arc = aParallel;
        break;
    case 1: arc = aSerial;
        break;
    }
    return arc;
}



//===========================================================================
//  CLASS DdcModuleBase  -- Component shared features
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class DdcModuleBase
//---------------------------------------------------------------------------

static DdcModuleBase::MapRegisters DdcModuleBase_RegAddrs =
    {  mmDdcInfo,            //  mmDdcInfo_Addr;
       mmDdcConfig,          //  mmDdcConfig_Addr;
       mmDdcDecimation       //  mmDdcDecimation_Addr;
    };

//===========================================================================
//  CLASS DdcModuleBase  --
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class DdcModuleBase
//---------------------------------------------------------------------------


DdcModuleBase::DdcModuleBase(IAddressingSpace & space, WishboneType awtype)
    :  GspDspModuleBase(space, awtype),
       VersionReg(space, DdcModuleBase_RegAddrs.Info_Addr),
       ConfigReg(space, DdcModuleBase_RegAddrs.Config_Addr),
       DecimationReg(space, DdcModuleBase_RegAddrs.Decimation_Addr)
{
}

//---------------------------------------------------------------------------
//  DdcModuleBase::IsIQEnabled() -
//---------------------------------------------------------------------------

bool DdcModuleBase::IsIQEnabled() const
{
    return ConfigReg.NonInterleaved.Value() == 0;
}

//---------------------------------------------------------------------------
//  DdcModuleBase::IsUnsigned() -
//---------------------------------------------------------------------------

bool DdcModuleBase::IsUnsigned() const
{
    return ConfigReg.Unsigned.Value();
}

//---------------------------------------------------------------------------
//  DdcCommon::IsComplexInput() -
//---------------------------------------------------------------------------

bool DdcModuleBase::IsComplexInput() const
{
    return ConfigReg.ComplexInput.Value() == 1;
}

//---------------------------------------------------------------------------
//  DdcCommon::IsVitaPacket() -
//---------------------------------------------------------------------------

bool DdcModuleBase::IsVitaPacket() const
{
    return ConfigReg.OutputFormat.Value() == 0;
}

//---------------------------------------------------------------------------
//  DdcCommon::GetEnableBits() - read core count
//---------------------------------------------------------------------------

unsigned int DdcModuleBase::CoreCount() const
{
    return VersionReg.CoreCount.Value();
}

//---------------------------------------------------------------------------
//  DdcModuleBase::ChannelCountPerCore() - read channel count per core
//---------------------------------------------------------------------------

unsigned int DdcModuleBase::ChannelCountPerCore() const
{
    return VersionReg.ChannelPerCore.Value();
}

//---------------------------------------------------------------------------
//  DdcModuleBase::SampleWidth() -
//---------------------------------------------------------------------------

unsigned int DdcModuleBase::SampleWidth() const
{
    unsigned int bits = 8;

    switch(ConfigReg.SampleWidth.Value())
    {
    case SubModuleConfigRegister::sw08:
        bits =  8; break;
    case SubModuleConfigRegister::sw16:
        bits = 16; break;
    case SubModuleConfigRegister::sw24:
        bits = 24; break;
    case SubModuleConfigRegister::sw32:
        bits = 32; break;
    }

    return bits;
}

//---------------------------------------------------------------------------
//  DdcModuleBase::Decimation() -
//---------------------------------------------------------------------------

unsigned int DdcModuleBase::Decimation()
{
    return DecimationReg.Value();
}

//---------------------------------------------------------------------------
//  DdcModuleBase::Decimation() -
//---------------------------------------------------------------------------

void DdcModuleBase::Decimation(unsigned int value)
{
    DecimationReg.Value(value);
}


//===========================================================================
//  CLASS ComponentModuleBase  -- Component shared features
//===========================================================================
//---------------------------------------------------------------------------
//  housekeeping for class DdcModuleBase
//---------------------------------------------------------------------------

static ComponentModuleBase::MapRegisters ComponentModuleBase_RegAddrs =
    {  mmCommonCompSampleRate,            //  InputSampleRate_Addr
       mmCommonCompDataLogTrig           //   DataLogTrigger_Addr
    };

//===========================================================================
//  CLASS ComponentModuleBase  --
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class ComponentModuleBase
//---------------------------------------------------------------------------

ComponentModuleBase::ComponentModuleBase(IAddressingSpace & space, WishboneType awtype)
    :  GspDspModuleBase(space, awtype),
       InputSampleRateReg(space, ComponentModuleBase_RegAddrs.InputSampleRate_Addr),
       DataLogTriggerReg(space, ComponentModuleBase_RegAddrs.DataLogTrigger_Addr)
{
}

//---------------------------------------------------------------------------
//  ComponentModuleBase::SampleRate() -
//---------------------------------------------------------------------------

void ComponentModuleBase::SampleRate(unsigned int value)
{
    InputSampleRateReg.Value(value);
}

//---------------------------------------------------------------------------
//  ComponentModuleBase::SampleRate() -
//---------------------------------------------------------------------------

unsigned int ComponentModuleBase::SampleRate()
{
    return InputSampleRateReg.Value();
}

//---------------------------------------------------------------------------
//  ComponentModuleBase::Decimation() -
//---------------------------------------------------------------------------

void ComponentModuleBase::TriggerDataLog(bool value)
{
    DataLogTriggerReg.Value(value);
}

//===========================================================================
//  CLASS GpDsp_1D_Submodule  -- Shared Submodule w/one_D indexing
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class GpDsp_1D_Submodule
//---------------------------------------------------------------------------

GpDsp_1D_Submodule::GpDsp_1D_Submodule(GspDspModuleBase & cbase,
                               IAddressingSpace & space, MapRegisters & regs, WishboneType awtype)
    :  GpDspSubmodule(cbase, space, regs, awtype)
{
}

//---------------------------------------------------------------------------
//  GpDsp_1D_Submodule::Reset() - reset power meter
//---------------------------------------------------------------------------

void  GpDsp_1D_Submodule::Reset(unsigned int core, bool state)
{
    SetIndex(core);
    ResetReg.Reset.Value(state);
}

//---------------------------------------------------------------------------
//  GpDsp_1D_Submodule::CommandTrigger() -
//---------------------------------------------------------------------------

void  GpDsp_1D_Submodule::CommandTrigger()
{
    Module.CommandTrigger();
}

//===========================================================================
//  CLASS GpDsp_2D_Submodule  -- Shared Submodule w/one_D indexing
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class GpDsp_2D_Submodule
//---------------------------------------------------------------------------

GpDsp_2D_Submodule::GpDsp_2D_Submodule(GspDspModuleBase & cbase,
                               IAddressingSpace & space, MapRegisters & regs, WishboneType awtype)
    :  GpDspSubmodule(cbase, space, regs, awtype)
{
}

//---------------------------------------------------------------------------
//  GpDsp_2D_Submodule::Reset() - reset power meter
//---------------------------------------------------------------------------

void  GpDsp_2D_Submodule::Reset(unsigned int parent, unsigned int core, bool state)
{
    SetIndex(parent, core);
    ResetReg.Reset.Value(state);
}

//---------------------------------------------------------------------------
//  GpDsp_2D_Submodule::CommandTrigger() -
//---------------------------------------------------------------------------

void  GpDsp_2D_Submodule::CommandTrigger()
{
    Module.CommandTrigger();
}

#ifdef __CLR_VER
#pragma managed(pop)
#endif

}  // namespace Innovative



