#ifndef Components_GpH
#define Components_GpH

#include <CommonRegisters_Gp.h>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//===========================================================================
//  CLASS GspDspProjectInfo  -- Project wide Information Block
//===========================================================================

class GspDspProjectInfo
{
public:
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS GspDspProjectInfo::MapRegisters  -- Device Register Offsets
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    class MapRegisters
    {
    public:
        unsigned int   ProjectVersion_Addr;
        unsigned int   ModuleCoreEnable_Addr;
        unsigned int   ProjectSubVersion_Addr;
    };

    GspDspProjectInfo(IAddressingSpace & space);
    virtual ~GspDspProjectInfo() {}

    enum Cores { CustorProj, DspCommon, DDC, DUC, Spectrum, Beamforming, PSK, FSK, LastCore };

    bool Enabled(Cores c);

    unsigned int Version();
    unsigned int SubVersion();
    std::string  VersionString();
    std::string  SubVersionString();

private:
    ProjectVersionRegister             ProjectVersionReg;
    ProjectCoreEnableRegister          ModuleCoreEnableReg;
    ProjectSubVersionRegister          ProjectSubVersionReg;
};

//===========================================================================
//  CLASS IGspDspModule  -- Interface to common Header registers of a module
//===========================================================================

class IGspDspModule
{
public:
    virtual ~IGspDspModule() {};

    virtual unsigned int  GetEnableBits() = 0;
    virtual void  SetCommandTrigger(bool state) = 0;

    virtual void  SetIndex(unsigned int parent, unsigned int core) = 0; // two dimensional
    void  SetIndex(unsigned int core)
        {  SetIndex(0, core);  }   // one dimensional - set parent to 0

};

//===========================================================================
//  CLASS GspDspModuleBase  -- Component shared features
//===========================================================================

class GspDspModuleBase : public IGspDspModule
{
public:


    enum EnableBitsEnum
    {
        ePeripheral    = 0x0002,
        eTuner         = 0x0004,
        ePolyFilter    = 0x0008,
        eLowPassFilter = 0x0010,
        eOffset        = 0x0040,
        eChannelizer   = 0x0080,
        eVitaFramer    = 0x4000,
        eDebug         = 0x8000
    };


    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS GspDspModuleBase::MapRegisters  -- Device Register Offsets
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    class MapRegisters
    {
    public:
        unsigned int   ModuleEn_Addr;
        unsigned int   CmdTrigger_Addr;
        unsigned int   CoreIndex_Addr;
        unsigned int   ParentIndex_Addr;
        unsigned int   SyncReset_Addr;
    };
    GspDspModuleBase(IAddressingSpace & space, WishboneType awtype);
    virtual ~GspDspModuleBase() {}

    virtual unsigned int  GetEnableBits();
    virtual void  SetCommandTrigger(bool state);

    bool IsBitEnabled(EnableBitsEnum value);

    void CommandTrigger();
    virtual void  SetIndex(unsigned int parent, unsigned int core);
    void  SetIndex(unsigned int core)
        {  IGspDspModule::SetIndex(core);  }

    void Reset(bool state);

private:
    ComponentModuleEnableRegister      ModuleEnReg;
    ComponentCommandTriggerRegister    CmdTriggerReg;
    ComponentCoreIndexRegister         CoreIndexReg;
    ComponentParentIndexRegister       ParentIndexReg;
    SubModuleResetRegister             SyncResetReg;

};

//===========================================================================
//  CLASS DdcModuleBase  --
//===========================================================================

class DdcModuleBase : public GspDspModuleBase
{
public:
    class MapRegisters
    {
    public:
        unsigned int   Info_Addr;
        unsigned int   Config_Addr;
        unsigned int   Decimation_Addr;
    };

    DdcModuleBase(IAddressingSpace & space, WishboneType awtype);
    virtual ~DdcModuleBase() {}

    virtual unsigned int CoreCount() const;
    virtual unsigned int ChannelCountPerCore() const;
    virtual bool IsIQEnabled() const;
    virtual bool IsUnsigned() const;
    virtual unsigned int  SampleWidth() const;

    virtual bool IsComplexInput() const;
    virtual bool IsVitaPacket() const;

    void    Decimation(unsigned int value);
    unsigned int Decimation();

private:
    SubModuleVersionRegister           VersionReg;
    SubModuleConfigRegister            ConfigReg;
    DdcShadowRegister                  DecimationReg;

};


//===========================================================================
//  CLASS ComponentModuleBase  --
//===========================================================================

class ComponentModuleBase : public GspDspModuleBase
{
public:
    class MapRegisters
    {
    public:
        unsigned int   InputSampleRate_Addr;
        unsigned int   DataLogTrigger_Addr;
    };

    ComponentModuleBase(IAddressingSpace & space, WishboneType awtype);
    virtual ~ComponentModuleBase() {}

    unsigned int SampleRate();
    void SampleRate(unsigned int value);

    void TriggerDataLog(bool value);

private:
    Register                           InputSampleRateReg;
    Register                           DataLogTriggerReg;
};

//===========================================================================
//  CLASS GpDspSubmodule  -- Submodule shared registers
//===========================================================================

class GpDspSubmodule
{
public:
    enum ArhitectureType { aParallel, aSerial };

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS GpDspSubmodule::MapRegisters  -- Device Register Offsets
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    class MapRegisters
    {
    public:
        unsigned int   Version_Addr;
        unsigned int   Config_Addr;
        unsigned int   Reset_Addr;
    };

    GpDspSubmodule(GspDspModuleBase & cbase, IAddressingSpace & space, MapRegisters & regs, WishboneType awtype);
    virtual ~GpDspSubmodule() {}

    // version?
    // config?
    unsigned int CoreCount() const;
    ArhitectureType WhichArchitecture();


protected:

    SubModuleVersionRegister  VersionReg;
    SubModuleConfigRegister   ConfigReg;
    SubModuleResetRegister    ResetReg;

    GspDspModuleBase &  Module;

    void  SetIndex(unsigned int core);
    void  SetIndex(unsigned int parent, unsigned int core);

};

//===========================================================================
//  CLASS GpDsp_1D_Submodule  -- Shared Submodule w/one_D indexing
//===========================================================================

class GpDsp_1D_Submodule : public GpDspSubmodule
{
public:

    GpDsp_1D_Submodule(GspDspModuleBase & cbase, IAddressingSpace & space, MapRegisters & regs, WishboneType awtype);
    virtual ~GpDsp_1D_Submodule() {}

	void  SetIndex(unsigned int core) 
	{ GpDspSubmodule::SetIndex(core);}
    
	// version?
    // config?
    void  Reset(unsigned int core, bool state);
    void  CommandTrigger();
};

//===========================================================================
//  CLASS GpDsp_2D_Submodule  -- Shared Submodule w/two_D indexing
//===========================================================================

class GpDsp_2D_Submodule : public GpDspSubmodule
{
public:

    GpDsp_2D_Submodule(GspDspModuleBase & cbase, IAddressingSpace & space, MapRegisters & regs, WishboneType awtype);
    virtual ~GpDsp_2D_Submodule() {}

	void  SetIndex(unsigned int parent, unsigned int core) 
	{ GpDspSubmodule::SetIndex(parent, core);}

    // version?
    // config?
    void  Reset(unsigned int parent, unsigned int core, bool state);
    void  CommandTrigger();
};

#ifdef __CLR_VER
#pragma managed(pop)
#endif

}  // namespace Innovative

#endif // Components_GpH
