// Drc_Provider_Gp.cpp
//
// Data flow and hardware I/O for DRC units
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2013 Innovative Integration Corporation.
//  All Rights Reserved.
//
//---------------------------------------------------------------------------

#include "DrcProvider_Gp.h"
#include <SystemSupport_Mb.h>
#include <StringSupport_Mb.h>
#include <limits>
#include <cmath>
#include <numeric>
#include <Exception_Mb.h>
#include <FileSupport_Mb.h>
#include <DeviceEnum_Mb.h>
#include <Buffer_Mb.h>
#include <BufferDatagrams_Mb.h>
#include <BufferHeader_Mb.h>
#include <sstream>
#include <cstdio>
#include <Magic_Mb.h>

#include <GpDsp_Config_Gp.h>

#include <ctime>
#include <cstring>
#include <chrono>
#include <thread>

using namespace std;


namespace   Innovative
{

using namespace Drc;

//===========================================================================
//  CLASS Drc_Provider  -- Hardware Access and Application I/O Class
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class Drc_Provider
//---------------------------------------------------------------------------

Drc_Provider::Drc_Provider(Drc_ProviderSettings & settings, IDrcManagerInterface * mgr)
    : Settings(settings), MGR(mgr),
      StreamConnected(false), Stopped(true),
      FBlockRate(0.0f), FWordCount(0), SamplesPerWord(2),
      WordsToLog(0), Time(6), FftTime(6), BytesPerBlock(6),
      FftBytesPerBlock(6)
{
    //
    //  Create DRC Module Object using Factory
    Module = MakeModule(this);

    OnLog.SetEvent(this, &Drc_Provider::HandleOnLog);
    Module->OnLog.SetEvent(this, &Drc_Provider::HandleOnLog);

    Module->OnSend.SetEvent(this, &Drc_Provider::HandleOnSend);

    Timer.Interval(1000);
}

//---------------------------------------------------------------------------
//  destructor for class Drc_Provider
//---------------------------------------------------------------------------

Drc_Provider::~Drc_Provider()
{
    Close();
}

//---------------------------------------------------------------------------
//  Drc_Provider::BoardCount() -- Query number of installed boards
//---------------------------------------------------------------------------

unsigned int  Drc_Provider::BoardCount()
{
    return Module->BoardCount();
}

//---------------------------------------------------------------------------
//  Drc_Provider::Open() -- Open Hardware & set up callbacks
//---------------------------------------------------------------------------

void  Drc_Provider::Open()
{
    //
    //  Configure Trigger Manager Event Handlers
    Trig.OnDisableTrigger.SetEvent(this, &Drc_Provider::HandleDisableTrigger);
    Trig.OnExternalTrigger.SetEvent(this, &Drc_Provider::HandleExternalTrigger);
    Trig.OnSoftwareTrigger.SetEvent(this, &Drc_Provider::HandleSoftwareTrigger);

    Trig.DelayedTrigger(true);   // trigger delayed after start
    //
    //  Configure Stream Event Handlers
    Stream.OnVeloDataAvailable.SetEvent(this, &Drc_Provider::HandleDataAvailable);
    Stream.OnVeloDataAvailable.Synchronize();
    Stream.DirectDataMode(false);

    Stream.OnAfterStop.SetEvent(this, &Drc_Provider::HandleAfterStop);
    Stream.OnAfterStop.Synchronize();
    Stream.OnBeforeStart.SetEvent(this, &Drc_Provider::HandleBeforeStreamStart);
    Stream.OnBeforeStart.Synchronize();
    Stream.OnAfterStart.SetEvent(this, &Drc_Provider::HandleAfterStreamStart);
    Stream.OnAfterStart.Synchronize();


    Stream.RxLoadBalancing(false);
    Stream.TxLoadBalancing(false);

    Timer.OnElapsed.SetEvent(this, &Drc_Provider::HandleTimer);
    Timer.OnElapsed.Thunk();

    Module->AttachSettings(Settings);
    Module->OpenHardware();

    if (Module->IsOpen()==false)
        return;

    Log("Module Device opened successfully...");

    Module->EnableCalibration();

    //
    //  Connect Stream
    Module->ConnectToStream(Stream);
    StreamConnected = true;
    Log("Stream Connected...");
	
    std::stringstream sv;
    sv << "Stream" << Ordinal();
    Graph.System().ServerSlotName(sv.str());

    DisplayLogicVersion();
    {
    std::stringstream ss;
    ss << Settings.Root() << "FFTData" << Ordinal() << ".bin";
    FFTLogger.FileName(ss.str());
    }

}

//---------------------------------------------------------------------------
//  Drc_Provider::Close() -- Close Hardware & set up callbacks
//---------------------------------------------------------------------------

void  Drc_Provider::Close()
{
    if (!IsOpen())
        return;

    Stream.Disconnect();
    StreamConnected = false;

    Module->CloseHardware();

    Log("Stream Disconnected...");
}

// 
//---------------------------------------------------------------------------
//  Drc_Provider::CurrentDateTime()  -- Get current date/time, format is YYYY-MM-DD.HH:mm:ss
//---------------------------------------------------------------------------
const std::string Drc_Provider::CurrentDateTime() 
{
    char       buf[80];

	time_t     now = time(0);
    struct tm  tstruct = *std::localtime(&now);
    
    strftime(buf, sizeof(buf), "%Y-%m-%d@%X", &tstruct);
	// ':' is illegal in filenames
	replace(buf, buf + strlen(buf), ':', '.');
    return buf;
}

//---------------------------------------------------------------------------
//  Drc_Provider::StartStreaming()  -- Initiate Data Flow
//---------------------------------------------------------------------------

void Drc_Provider::StartStreaming()
{
	//
    //  Set up Loggers and graphs
    std::stringstream ss;

    if (!Settings.DestinationDirectory.empty())
        ss << Settings.DestinationDirectory << GetSlash();

    if (Settings.EnableDateTimeLogging)
        ss << "Data" << Ordinal() << "_" << CurrentDateTime() << ".bin";
    else
        ss << "Data" << Ordinal() << ".bin";

    Logger.FileName(ss.str());
    Graph.BinFile(Logger.FileName());

    StartStreaming_Orig();

 }

//---------------------------------------------------------------------------
//  Drc_Provider::StartStreaming_Orig() --  Standard ftn to start data flow
//---------------------------------------------------------------------------

void  Drc_Provider::StartStreaming_Orig()
{
    if (!StreamConnected)
        {
        Log("Stream not connected! -- Open the boards");
        return;
        }

    //
    //  Set up Parameters for Data Streaming
    //  ...First have UI get settings into our settings store
    if (Module->ValidateStreamParameters()==false)
        {
        StopStreaming();
        return;
        }

    // Wait for the millisecond edge of the clock
    std::stringstream txt;
    txt << "Current time : " << static_cast<unsigned int>(time(0)) << " \n";
    Log(txt.str());
    using std::chrono::milliseconds;
    using std::chrono::duration_cast;
    milliseconds ms = duration_cast<milliseconds>(std::chrono::system_clock::now().time_since_epoch());
    int t = 1000 - (ms.count() % 1000);
    std::this_thread::sleep_for(milliseconds(t));

    time_t now = time(0);
    Module->TimeStamp(Settings.Utc, Settings.ExternalPps, Settings.TsfMode, now);

    FWordCount = 0;
    FFftWordCount = 0;
    FSentPacketCount = 0;
    FDroppedPacketCount = 0;
    WordsToLog = Module->WordsFromSamples(Settings.SamplesToLog);
    FftBlockCount = 0;
    FBlockRate = 0.0f;

    // Disable triggers initially  (done by library)
    Trig.DelayedTriggerPeriod(Settings.DelayedTriggerPeriod);
    Trig.ExternalTrigger(Settings.ExternalTrigger == 1);
    Trig.AtConfigure();

    Module->BeforePreconfigure();
    //
    //  Always preconfigure
    Stream.Preconfigure();

    // Let the module know the actual frequency
    Module->GetProjectRegs()->CommonM.SampleRate(Module->AdcClockFrequencyActual());

    Module->BeforeStreaming();

    //
    //  Start Loggers on active channels
    if (Settings.PlotEnable)
        {
        Graph.Quit();
        }

    if (Settings.LoggerEnable || Settings.PlotEnable)
        {
        try {
        Logger.WriteThrough(Settings.LoggerWriteThrough);
        Logger.CoalesceWrites(Settings.LoggerWriteCoalesce);

        Logger.Start();
        FFTLogger.Start();
        } catch (...) {
        Log("LOGGING FAILED !!!");
        }
        }


    Trig.AtStreamStart();

    //
    //  Start Streaming
    Stopped = false;
    Stream.Start();
    RunTimeSW.Start();
    Log("Stream Mode started");

    Timer.Enabled(true);
}

//---------------------------------------------------------------------------
//  Drc_Provider::StopStreaming() --  Terminate data flow
//---------------------------------------------------------------------------

void  Drc_Provider::StopStreaming()
{
    if (!StreamConnected)
        {
        Log("Stream not connected! -- Open the boards");
        return;
        }

    //
    //  Stop Streaming
    Stream.Stop();
    Stopped = true;
    Timer.Enabled(false);

    Trig.AtStreamStop();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  Data Flow Event Handlers
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//---------------------------------------------------------------------------
//  Drc_Provider::HandleDataAvailable() --  Handle received packet
//---------------------------------------------------------------------------

void  Drc_Provider::HandleDataAvailable(VitaPacketStreamDataEvent & Event)
{
    if (Stopped)
        return;

    VeloBuffer Packet;
    //
    //  Extract the packet from the Incoming Queue...
    Event.Sender->Recv(Packet);
    VeloHeaderDatagram Hdr(Packet);
    MGR->DataPacketAvailable(Personality.Ordinal, Packet);

    if (Hdr.PeripheralId() == 0x41)
        {

        FftBlockCount++;
        if (Settings.EnablePlotting)
        {
            MGR->Plot(Personality.Ordinal, Packet, 1);
        }

        if (Settings.EnableFFTLogging)
        {
            FFTLogger.LogWithHeader(Packet);
        }

        return;
        }

    if (Hdr.PeripheralId() == 0x40)
        {
        double AvgBytes = FftBytesPerBlock.Process(IntegerDG(Packet).size() * sizeof(int));
        FFftBlockRate   = AvgBytes / (FftTime.Differential()*1.0e6);
        FFftWordCount  += Packet.SizeInInts();

        FftBlockCount++;
        if (Settings.EnablePlotting)
        {
            MGR->Plot(Personality.Ordinal, Packet, 0);
        }

        if (Settings.EnableFFTLogging)
        {
            FFTLogger.LogWithHeader(Packet);
        }

        return;
        }

    if (Settings.LoggerEnable)
        if (Settings.ContinuousLogging || (WordsToLog==0 || (FWordCount < WordsToLog)))
            {
            Logger.LogWithHeader(Packet);
            }

    FWordCount += Packet.SizeInInts();

    IntegerDG Packet_DG(Packet);

    TallyBlock(Packet_DG.size()*sizeof(int));

    // Per block triggering actions
    Trig.AtBlockProcess(Packet_DG.size()*sizeof(int));
}

//---------------------------------------------------------------------------
//  Drc_Provider::TallyBlock() --  Finish processing of received packet
//---------------------------------------------------------------------------

void  Drc_Provider::TallyBlock(size_t bytes)
{
    double Period = Time.Differential();
    double AvgBytes = BytesPerBlock.Process(static_cast<double>(bytes));

    if (Period)
        {
        FBlockRate   = AvgBytes / (Period*1.0e6);
        }
	else
		FBlockRate = 0.0f;
    //
    //  Stop streaming when both Channels have passed their limit
    if (Settings.AutoStop && IsDataLoggingCompleted() && !Stopped)
        {
        // Stop counter and display it
        double elapsed = RunTimeSW.Stop();

        StopStreaming();
        Log("Stream Mode Stopped automatically");
        Log(std::string("Elasped (S): ") + FloatToString(elapsed));
        }
}


//---------------------------------------------------------------------------
//  Drc_Provider::HandleAfterStop() --  Post stream termination event
//---------------------------------------------------------------------------

void Drc_Provider::HandleAfterStop(OpenWire::NotifyEvent & /*Event*/)
{
    //
    //  Stop Loggers on active Channels
    if (Settings.LoggerEnable || Settings.PlotEnable)
        {
        Logger.Stop();
        InitBddFile(Graph);

        FFTLogger.Stop();

        if (Logger.Logged() && Settings.PlotEnable)
            Graph.Plot();
        }
    Log(std::string("Analog I/O Stopped"));
    MGR->AfterStreamAutoStop(Ordinal());
}

//---------------------------------------------------------------------------
//  Drc_Provider::HandleBeforeStreamStart() --  Pre-streaming init event
//---------------------------------------------------------------------------

void Drc_Provider::HandleBeforeStreamStart(OpenWire::NotifyEvent & /*Event*/)
{
    MGR->BeforeStreamStart(Ordinal());
}

//---------------------------------------------------------------------------
//  Drc_Provider::HandleAfterStreamStart() --  Post streaming init event
//---------------------------------------------------------------------------

void Drc_Provider::HandleAfterStreamStart(OpenWire::NotifyEvent & /*Event*/)
{
    Log(std::string("Analog I/O started"));
}

//---------------------------------------------------------------------------
//  Drc_Provider::HandleTimer() --  Per-second status timer event
//---------------------------------------------------------------------------

void Drc_Provider::HandleTimer(OpenWire::NotifyEvent & /*Event*/)
{
    // Collect Data for Periodic Status call...
    // Display status
    MGR->PeriodicStatusUpdate(Ordinal());

    Trig.AtTimerTick();
}

//---------------------------------------------------------------------------
//  Drc_Provider::Log() --  Log message thunked to main thread
//---------------------------------------------------------------------------

void  Drc_Provider::Log(const std::string & msg)
{
    ProcessStatusEvent e(msg);
    OnLog.Execute(e);
}

//---------------------------------------------------------------------------
//  Drc_Provider::HandleOnLog() --
//---------------------------------------------------------------------------

void Drc_Provider::HandleOnLog(Innovative::ProcessStatusEvent & Event)
{
	MGR->Log(Ordinal(), Event.Message);
}

//---------------------------------------------------------------------------
//  Drc_Provider::HandleOnSend() --
//---------------------------------------------------------------------------

void Drc_Provider::HandleOnSend(Innovative::SenderEvent<VeloBuffer> & Event)
{
	if (Stream.Internals()->XmitQDepth() > 20)
		{
			Log("Error : " + IntToString(Stream.Internals()->XmitQDepth()));
			FDroppedPacketCount++;
			return;
		}
	FSentPacketCount++;	
	Stream.Send(0x0, Event.Sender);
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  Support Functions
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//---------------------------------------------------------------------------
//  Drc_Provider::IsDataLoggingCompleted() --check if done logging
//---------------------------------------------------------------------------

bool  Drc_Provider::IsDataLoggingCompleted()
{
    if (WordsToLog==0)
        return false;
    else
        return FWordCount >= WordsToLog;
}

//---------------------------------------------------------------------------
//  Drc_Provider::SoftwareAlert() --  Issue specified software alert
//---------------------------------------------------------------------------

void Drc_Provider::SoftwareAlert(unsigned int value)
{
    Log(std::string("Posting SW alert..."));
    Module->SoftwareAlert(value);
}

//---------------------------------------------------------------------------
//  Drc_Provider::Temperature() --  Query current module temperature
//---------------------------------------------------------------------------

int Drc_Provider::Temperature()
{
    return Module->Temperature();
}

//---------------------------------------------------------------------------
//  Drc_Provider::PllLocked() --  Query current PLL lock status
//---------------------------------------------------------------------------

bool  Drc_Provider::PllLocked()
{
    return Module->PllLocked();
}

//---------------------------------------------------------------------------
//  Drc_Provider::HandleDisableTrigger() --  Trigger Manager Trig OFF
//------------------------------------------------------------------------------

void  Drc_Provider::HandleDisableTrigger(OpenWire::NotifyEvent & /*Event*/)
{
    Module->SetExternalTrigger(false);
}

//---------------------------------------------------------------------------
//  Drc_Provider::HandleSoftwareTrigger() --  Trigger Manager Trig OFF
//------------------------------------------------------------------------------

void  Drc_Provider::HandleSoftwareTrigger(OpenWire::NotifyEvent & /*Event*/)
{
    Module->SetSoftwareTrigger(true);
}

//---------------------------------------------------------------------------
//  Drc_Provider::HandleExternalTrigger() --  Enable External trigger
//------------------------------------------------------------------------------

void  Drc_Provider::HandleExternalTrigger(OpenWire::NotifyEvent & /*Event*/)
{
    Module->SetExternalTrigger(true);
}

//------------------------------------------------------------------------------
//  Drc_Provider::DisplayLogicVersion() --  Log version info
//------------------------------------------------------------------------------

void  Drc_Provider::DisplayLogicVersion()
{
    Module->LogLogicVersion();
}

//------------------------------------------------------------------------------
//  Drc_Provider::LogicVersion() --  Version info
//------------------------------------------------------------------------------

string Drc_Provider::LogicVersion()
{
    return Module->LogicVersion();
}

//------------------------------------------------------------------------------
//  Drc_Provider::PeekLogic() --
//------------------------------------------------------------------------------

unsigned int Drc_Provider::PeekLogic(int offset)
{
    return Module->PeekLogic(offset);
}

//------------------------------------------------------------------------------
//  Drc_Provider::PokeLogic() --
//------------------------------------------------------------------------------

void Drc_Provider::PokeLogic(int offset, unsigned int value)
{
    return Module->PokeLogic(offset, value);
}

//------------------------------------------------------------------------------
//  Drc_Provider::LogicVersion() --  Version info
//------------------------------------------------------------------------------

pair<string,string> Drc_Provider::CoreVersion()
{
    return Module->CoreVersion();
}

//---------------------------------------------------------------------------
//  Drc_Provider::InitBddFile() --  Set up BDD File for an output file
//---------------------------------------------------------------------------

void Drc_Provider::InitBddFile(BinView & graph)
{
    PathSpec spec(graph.BinFile());
    spec.Ext(".bdd");
    if (FileExists(spec.Full()) && !Settings.OverwriteBdd)
        return;

    remove(spec.Full().c_str());

    // Make a new BDD
    const int bits = Module->Input_Bits();

    const int limit = 1 << (bits-1);

    graph.Time().LowerLimit(-limit);
    graph.Time().UpperLimit(limit);
    graph.Time().Break(1);
    graph.Time().SeamSize(Module->VeloDataSize()*sizeof(int));
    graph.Time().AnnotateSeams(true);

    graph.Fft().LowerLimit(-150);
    graph.Fft().ScaleYType(BinViewOptions::FftOptions::sLog);
    graph.Fft().ScaleXType(BinViewOptions::FftOptions::sLog);
    graph.Fft().CriteriaThreshold(75);
    int pktsize = std::min(Settings.PacketSize, 0x10000);
    BinViewOptions::FftOptions::IIPoints pts(static_cast<BinViewOptions::FftOptions::IIPoints>(
        TrailingZeroCount(pktsize)-7));
    graph.Fft().Points(pts);
    graph.Fft().Window(BinViewOptions::FftOptions::wBlackmann);
    graph.Fft().Average(1);

    graph.Text().DataFormat("%11.3f");
    graph.System().UpdateRate(BinViewOptions::SystemOptions::ms1000);
//    graph.System().ServerSlotName("BinView");
    graph.Leap(Settings.PacketSize);

    graph.Polarity(IsUnsigned() ? BinView::pUnsigned : BinView::pSigned);
    if (graph.Polarity() == BinView::pUnsigned)
        graph.DcOffset(-750.0f);
    else
        graph.DcOffset(0);

    graph.DataSpan(300);
    graph.DataType(BinView::tInt);
    graph.Units("mV");
    graph.Samples(1000);

    graph.SampleRate(static_cast<float>(Module->AdcClockFrequencyActual() / 1.0e3f) / MGR->DecimationFactor(0));
    graph.Channels(MGR->EnabledChannels(Ordinal()));
    graph.Devices(MGR->EnabledDevices(Ordinal()));

    int percore    = Module->GetProjectRegs()->DdcM.ChannelCountPerCore();
    int group_size = (percore == 0) ? 1 : percore;
    if (MGR->EnabledDevices(Ordinal()) == MGR->EnabledChannels(Ordinal()))
        {
        graph.SignificantBits(bits);
        }
    else
        {
        graph.SignificantBits(Module->SampleWidth()); // DDC
        graph.Devices(MGR->EnabledDevices(Ordinal()));
        graph.Channels(MGR->EnabledChannels(Ordinal()) * group_size);
        }

    if (!Module->GetProjectRegs()->DdcM.IsVitaPacket())
        graph.Devices(1);

    graph.System().Source(Module->BddOutputFileType());
    graph.NullDataValue(.12345e-19f);
    graph.InputSpan((float)Module->Input_FullSpan());
    graph.ScalingEnabled(true);
	
    graph.Comments().Clear();
    graph.Comments().Add(MGR->ChannelMap(Ordinal()));

    //  Board Specific Modifications
    //Module.ModifyBddFile(graph);

    graph.Save();
}

//---------------------------------------------------------------------------
//  Drc_Provider::ClockInfo() -- Dump clock info
//---------------------------------------------------------------------------

void  Drc_Provider::ClockInfo()
{
    Module->LogClockInfo();
}

//---------------------------------------------------------------------------
//  Drc_Provider::Make() -- Dump clock info
//---------------------------------------------------------------------------

void  Drc_Provider::Make(FiclGroup & group, IFicl_UserInterface * mgr)
{
    //  Fill in FICL object(s) for the current Module
    Module->Make_Ficls(group, mgr);
}

//---------------------------------------------------------------------------
//  Drc_Provider::Make() --
//---------------------------------------------------------------------------

void  Drc_Provider::Make(DdcChannelsProxy & group)
{
    Module->Make_DdcChannels(group);
}

//---------------------------------------------------------------------------
//  Drc_Provider::Make_FftChannels() --
//---------------------------------------------------------------------------

void  Drc_Provider::Make(FftChannelsProxy & group)
{
    Module->Make_FftChannels(group);
}

//---------------------------------------------------------------------------
//  Drc_Provider::Make_DdcChannels() --
//---------------------------------------------------------------------------

void Drc_Provider::Make(PowerMeterProxy & group)
{
    Module->Make_PowerMeter(group);
}

//---------------------------------------------------------------------------
//  Drc_Provider::Make() --
//---------------------------------------------------------------------------

void Drc_Provider::Make(OverflowProxy & group)
{
    Module->Make_OverflowDetector(group);
}

//---------------------------------------------------------------------------
//  Drc_Provider::Make() --
//---------------------------------------------------------------------------

void Drc_Provider::Make(ToneGeneratorProxy &group)
{
    Module->Make_ToneGenerator(group);
}

//---------------------------------------------------------------------------
//  Drc_Provider::SampleWidth() --
//---------------------------------------------------------------------------

unsigned int Drc_Provider::SampleWidth() const
{
    return Module->SampleWidth();
}

//---------------------------------------------------------------------------
//  Drc_Provider::IsIQEnabled() --
//---------------------------------------------------------------------------

bool Drc_Provider::IsIQEnabled() const
{
    return Module->IsIQEnabled();
}

//---------------------------------------------------------------------------
//  Drc_Provider::IsIQEnabled() --
//---------------------------------------------------------------------------

bool Drc_Provider::IsUnsigned() const
{
    return Module->IsUnsigned();
}

//==============================================================================
//  CLASS Drc_ProviderSettings  - shared Settings for all DRC Providers
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class Drc_ProviderSettings
//------------------------------------------------------------------------------

Drc_ProviderSettings::Drc_ProviderSettings()
    :   AlertEnable(AnalogInAlerts(), false),
       Gain(AnalogInChannels()), Offset(AnalogInChannels())
{
    //  Set Default Defaults (can be overridden later)
    //  Board Settings
#ifdef LINUX
    int kDefaultBMSize = 4;
#else
    int kDefaultBMSize = 32;
#endif
    BusmasterSize       = kDefaultBMSize;

    //  Config Data
    PacketSize           = SystemDefaultPacketSize();
    SampleClockSource    = 1;
    SampleRate           = MaxInRateMHz();
    DacSampleRate        = MaxInRateMHz();
    ExtClockSrcSelection = 0;
    ReferenceClockSource = 1;
    ReferenceRate        = 10.0f;
    AlertEnable          = DefaultAlertList();
    PhaseDetectorFrequency = PhaseDetectorDefault();

    //  Trigger
    ExternalTrigger          = 0;
    EdgeTrigger              = 1;
    DelayedTriggerPeriod     = 2;
    ExtTriggerSrcSelection   = 0;
    Framed                   = 0;
    FrameSize                = 0x4000;

    Calibrated = true;
    //  Streaming
    LoggerEnable = true;
    PlotEnable = true;
    OverwriteBdd = true;
    ContinuousLogging = false;
    SamplingDuration = 1.0f;

    AutoStop = true;
    Utc = true;
    ExternalPps = false;

    EnablePowerMeterReads = false;
    EnableArithmeticOflowReads = false;

    EnablePlotting = true;
    EnableFFTLogging = false;
    EnableDateTimeLogging = false;

    Help = true;
    DebugScript = std::string("");

    TsfMode = 1;
    TimerInterval = 1000;
    CalibOffset = 0;

#ifdef LINUX
        DestinationDirectory        = ".//";
#else
        DestinationDirectory        = ".\\";
#endif

    //     
    LoggerWriteCoalesce      = true;
    LoggerWriteThrough       = false;

    PowerMeterCalibrationOffset = 0.0f;
}

//------------------------------------------------------------------------------
//  Load INI Settings
//------------------------------------------------------------------------------

void Drc_ProviderSettings::LoadINISettings()
{
    //  Board Settings
    Install( ToIni("BMSize",               BusmasterSize,   BusmasterSize)  );

    //  Config Data
    Install( ToIni("PacketSize",           PacketSize,           PacketSize)  );   //
    Install( ToIni("SampleClockSource",    SampleClockSource,    SampleClockSource)  );
    Install( ToIni("SampleRate",           SampleRate,           SampleRate)  );
    Install( ToIni("DacSampleRate",        DacSampleRate,        DacSampleRate)  );
    Install( ToIni("ExtClockSrcSelection", ExtClockSrcSelection, ExtClockSrcSelection)  );
    Install( ToIni("ReferenceClockSource", ReferenceClockSource, ReferenceClockSource)  );
    Install( ToIni("ReferenceRate",        ReferenceRate,        ReferenceRate)  );

    Install( ToIni("AlertEnable", "Alert", AlertEnable,          DefaultAlertList() ));

    //  Trigger
    Install( ToIni("ExternalTrigger",        ExternalTrigger,        ExternalTrigger)  );
    Install( ToIni("ExtTriggerSrcSelection", ExtTriggerSrcSelection, ExtTriggerSrcSelection)  );
    Install( ToIni("EdgeTrigger",            EdgeTrigger,            EdgeTrigger)  );
    Install( ToIni("TriggerDelay",           DelayedTriggerPeriod,   DelayedTriggerPeriod)  );
    Install( ToIni("Framed",                 Framed,                 Framed)  );
    Install( ToIni("FrameSize",              FrameSize,              FrameSize)  );


    //  Streaming
    Install( ToIni("LoggerEnable",           LoggerEnable,                LoggerEnable)  );
    Install( ToIni("PlotEnable",             PlotEnable,                  PlotEnable)  );
    Install( ToIni("OverwriteBdd",           OverwriteBdd,                OverwriteBdd)  );
    Install( ToIni("ContinuousLogging",      ContinuousLogging,           ContinuousLogging)  );
    Install( ToIni("SamplingDuration",       SamplingDuration,            SamplingDuration)  );

    Install( ToIni("AutoStop",               AutoStop,                    AutoStop)  );
    Install( ToIni("Utc",                    Utc,                         Utc)  );
    Install( ToIni("ExternalPps",            ExternalPps,                 ExternalPps)  );

    Install( ToIni("EnablePowerMeterReads",       EnablePowerMeterReads,       EnablePowerMeterReads)  );
    Install( ToIni("EnableArithmeticOflowReads",  EnableArithmeticOflowReads,  EnableArithmeticOflowReads)  );

    Install( ToIni("EnablePlotting",              EnablePlotting,              EnablePlotting)  );
    Install( ToIni("EnableFFTLogging",            EnableFFTLogging,            EnableFFTLogging)  );
    Install( ToIni("EnableDateTimeLogging",       EnableDateTimeLogging,       EnableDateTimeLogging)  );

    Install( ToIni("LoggerWriteCoalesce",         LoggerWriteCoalesce,         LoggerWriteCoalesce)  );
    Install( ToIni("LoggerWriteThrough",          LoggerWriteThrough,          LoggerWriteThrough)  );

    Install( ToIni("Help",             Help,  Help) );

    Install( ToIni("AdcCalibrated",    Calibrated,  Calibrated)  );

    Install( ToIni("Debug Script",     DebugScript,  DebugScript) );

    Install( ToIni("PhaseDetectorFrequency", PhaseDetectorFrequency, PhaseDetectorFrequency) );

    Install( ToIni("TsfMode",        TsfMode,     TsfMode)  );

    Install( ToIni("TimerInterval",  TimerInterval,   TimerInterval)  );
    Install( ToIni("CalibOffset",    CalibOffset,  CalibOffset)  );

    Install( ToIni("DestinationDirectory",     DestinationDirectory,  DestinationDirectory) );

    Install( ToIni("PowerMeterCalibrationOffset",     PowerMeterCalibrationOffset,  PowerMeterCalibrationOffset) );


    Load();
}

//------------------------------------------------------------------------------
//  destructor for class Drc_ProviderSettings
//------------------------------------------------------------------------------

Drc_ProviderSettings::~Drc_ProviderSettings()
{
}




}  // namespace Innovative

