// DrcProvider_Gp.h
//
// Data flow and hardware I/O for DRC units
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2013 Innovative Integration Corporation.
//  All Rights Reserved.
//
//---------------------------------------------------------------------------

#ifndef DrcProvider_GpH
#define DrcProvider_GpH

#include <GpProviderBase_Gp.h>
#include <ProcessEvents_Mb.h>
#include <VitaPacketStream_Mb.h>
#include <StopWatch_Mb.h>

#include <Benchmark_Mb.h>
#include <SoftwareTimer_Mb.h>
#include <DataLogger_Mb.h>

#include <Application/TriggerManager_App.h>
#include <Application/BinviewPlotter_App.h>
#include <Application/IniSaver_App.h>
#include <Application/FiclIo_App.h>
#include <Exception_Mb.h>


#include <ChannelsProxy_Gp.h>
#include <FftProxy_Gp.h>

#include <SupportProxy_Gp.h>
#include <PacketStreamSupport_Mb.h>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//==============================================================================
//  CLASS Drc_ProviderSettings  - shared Settings for all DRC Providers
//==============================================================================

class Drc_ProviderSettings  :  public Innovative::IniSaver
{
public:

    typedef std::vector<char>   BoolArray;
    typedef std::vector<int>    IntArray;
    typedef std::vector<float>  FloatArray;

    // Ctor
    Drc_ProviderSettings();
    ~Drc_ProviderSettings();
    void LoadINISettings();
    //
    //  Board Settings
    int             BusmasterSize;

    //  Config Data
    int             PacketSize;
    int             SampleClockSource;
    double          SampleRate;
    double          DacSampleRate;
    int             ExtClockSrcSelection;
    int             ReferenceClockSource;
    float           ReferenceRate;
    BoolArray       AlertEnable;


    //  Trigger
    int             ExternalTrigger;
    int             EdgeTrigger;
    int             ExtTriggerSrcSelection;
    int             DelayedTriggerPeriod;
    int             Framed;
    int             FrameSize;

    //  Logging
    bool            LoggerEnable;
    bool            PlotEnable;
    ii64            SamplesToLog;
    float	        SamplingDuration;
    bool            OverwriteBdd;
    bool            AutoStop;
    bool            ForcePacketSize;
    bool            ContinuousLogging;
    bool            Utc;
    bool            ExternalPps;
    bool            EnablePowerMeterReads;
    bool            EnableArithmeticOflowReads;

    bool            EnablePlotting;
    bool            EnableFFTLogging;
    bool            EnableDateTimeLogging;

    bool            LoggerWriteThrough;
    bool            LoggerWriteCoalesce;

    float           PowerMeterCalibrationOffset;

    //  Log Page Data
//    struct LogD
//    {
//        std::vector<int> LogData;
//        std::vector<int> IsrStatusData;
//        std::vector<int> VppData;
//    };

//    LogD    Log;
    //  Eeprom
    FloatArray      Gain;
    FloatArray      Offset;
    float           PllCorrection;
    bool            Calibrated;

    bool            Help;
    std::string     ModuleName;
    std::string     ModuleRevision;
    bool            InitDdcs;

    int             Divider;

    int             TsfMode;

    std::string     DebugScript;
	std::string     DestinationDirectory;
    double          PhaseDetectorFrequency;

    int             TimerInterval;
    float           CalibOffset;

private:
};

//  Factory Function Declaration to be implemented elsewhere
extern IDrc_Module *  MakeModule(class Drc_Provider * provider);

//===========================================================================
//  CLASS IDrcManagerInterface  -- Callback from DRC provider to manager
//===========================================================================

class IDrcManagerInterface
{
public:
    virtual ~IDrcManagerInterface()  {  }

    virtual void  Log(unsigned int ordinal, const std::string & a_string) = 0;
    virtual void  AfterStreamAutoStop(unsigned int ordinal) = 0;
    virtual void  PeriodicStatusUpdate(unsigned int ordinal) = 0;
    virtual void  BeforeStreamStart(unsigned int ordinal) = 0;
    virtual void  AfterStreamStart(unsigned int ordinal) = 0;
    virtual unsigned int EnabledChannels(unsigned int ordinal) = 0;
    virtual unsigned int EnabledDevices(unsigned int ordinal) = 0;
    virtual float DecimationFactor(unsigned int ordinal) = 0;
    virtual unsigned int EnabledChannelizerChannels(unsigned int ordinal) = 0;
    virtual void Plot(unsigned int ordinal, VeloBuffer &buffer, int type) = 0;
    virtual void DataPacketAvailable(unsigned int ordinal, VeloBuffer &buffer) = 0;
    virtual string ChannelMap(unsigned int ordinal) = 0;
};

//===========================================================================
//  CLASS Drc_Provider  -- Data flow and hardware I/O for DRC units
//===========================================================================

class Drc_Provider : public GpProviderBase
{
//    friend class ApplicationSettings;
//    typedef std::vector<__int64>    IntArray;

public:
    //
    //  Member Functions
    Drc_Provider(Drc_ProviderSettings & settings, IDrcManagerInterface * mgr);
    virtual ~Drc_Provider();

    void DumpInterruptInfo()
        { Module->DumpInterruptInfo();  }

    //
    //  Public Interface to Manager
    void Open();
    bool IsOpen()
        {  return Module->IsOpen();  }
    void Close();
    void StartStreaming();
    void StopStreaming();
    bool IsStreaming()
        {  return Timer.Enabled();  }

    unsigned int BoardCount();

    void  Make(FiclGroup & group, IFicl_UserInterface * mgr);

    void  Make(DdcChannelsProxy & group);
    void  Make(FftChannelsProxy & group);
    void  Make(PowerMeterProxy & group);
    void  Make(OverflowProxy & group);
    void  Make(ToneGeneratorProxy &group);

    string LogicVersion();
    pair<string,string> CoreVersion();
	unsigned int SampleWidth() const;
	bool IsIQEnabled() const;
        bool IsUnsigned() const;
    double AdcClockFrequencyActual()
    {   return Module->AdcClockFrequencyActual(); }
    double DacClockFrequencyActual()
    {   return Module->DacClockFrequencyActual(); }


    unsigned int  PeekLogic(int offset);
    void          PokeLogic(int offset, unsigned int value);

    //
    //  Private Support Ftns for Manager
private:
    void StartStreaming_Orig();
	const std::string CurrentDateTime();
public:
    void SoftwareAlert(unsigned int value);
    int  Temperature();
    bool PllLocked();

    double BlockRate() const
        {  return FBlockRate;  }

    double FftBlockRate() const
        {  return FFftBlockRate;  }


    ii64 FramePerSec()
        {
            ii64 temp = FftBlockCount;
            FftBlockCount = 0;

            return temp;
        }


    ii64    SampleCount() const
        {  return FWordCount*SamplesPerWord;  }

    ii64    FftSampleCount() const
        {  return FFftWordCount*SamplesPerWord;  }

    ii64    SentPacketCount() const
        {  return FSentPacketCount;  }

	ii64    DroppedPacketCount() const
        {  return FDroppedPacketCount;  }
	
	void PassDefaults()
		{ Module->PassDefaults(Settings); }

    void ClockInfo();

    virtual unsigned int   DdcCoresCount()
        { return Module->DdcCoresCount(); }
    virtual unsigned int   DucCoresCount()
        { return Module->DucCoresCount(); }
    virtual unsigned int   FftCoresCount()
        { return Module->FftCoresCount(); }
    virtual void  SetSoftwareTrigger(bool state)
        { Module->SetSoftwareTrigger(state);}
    virtual void  SetExternalTrigger(bool state)
        {Module->SetExternalTrigger(state);}

    ///////////////////
    //Board Information
    ///////////////////
    //Logic Version
    std::string FpgaLogicVersion()
    { return  Module->FpgaLogicVersion(); }
    //Hardware Variant
    short FpgaHardwareVariant()
    { return  Module->FpgaHardwareVariant(); }
    //Revision
    short PciLogicRevision()
    { return  Module->PciLogicRevision(); }
    //Subrevision
    short FpgaLogicSubrevision()
    { return Module->FpgaLogicSubrevision(); }
    //Board Family
    std::string PciLogicFamily()
    { return  Module->PciLogicFamily(); }
    //Type
    std::string PciLogicType()
    { return  Module->PciLogicType(); }
    //Board Revision
    std::string PciLogicPcb()
    { return  Module->PciLogicPcb(); }
    //Chip
    std::string FpgaName()
    { return Module->FpgaName(); }
    //PCI EXPRESS
    std::string PCIExpressLanes()
    { return Module->PCIExpressLanes(); }


    std::string PrintDevices()
    { return  Module->PrintDevices(); }


public: // Pseudo-protected
    // Data
    Drc_ProviderSettings &        Settings;
    Drc_ProviderPersonality       Personality;

    unsigned int Ordinal()  {  return Personality.Ordinal;  }
    ii64                                FftBlockCount;


	size_t QDepth()
	{ return Stream.Internals()->XmitQDepth(); }

private:
    //
    //  Member Data
    IDrc_Module *                       Module;
    IDrcManagerInterface *              MGR;
    Innovative::VitaPacketStream        Stream;
    Innovative::TriggerManager          Trig;
    Innovative::SoftwareTimer           Timer;
    Innovative::StopWatch               RunTimeSW;
    Innovative::DataLogger              Logger;
    Innovative::BinviewPlotter          RtPlot;
    Innovative::BinView                 Graph;

    Innovative::DataLogger              FFTLogger;

    //  App State variables
    bool                                StreamConnected;
    bool                                Stopped;
    //  App Status variables
    double                              FBlockRate;
    double                              FFftBlockRate;
    ii64                                FWordCount;
    ii64                                FFftWordCount;
    ii64                                FSentPacketCount;
	ii64                                FDroppedPacketCount;
    int                                 SamplesPerWord;
    ii64                                WordsToLog;

    Innovative::AveragedRate            Time;
    Innovative::AveragedRate            FftTime;
    Innovative::AveragedRate            BytesPerBlock;
    Innovative::AveragedRate            FftBytesPerBlock;

    //
    //  Member Functions
    void  HandleDataAvailable(Innovative::VitaPacketStreamDataEvent & Event);
    void  Handle_VPP_ImageAvailable(Innovative::VitaPacketParserImageAvailable & Event);

    void  HandleAfterStop(OpenWire::NotifyEvent & Event);
    void  HandleBeforeStreamStart(OpenWire::NotifyEvent & /*Event*/);
    void  HandleAfterStreamStart(OpenWire::NotifyEvent & Event);
    void  HandleScriptCommand(Innovative::ProcessStatusEvent & Event);
    void  HandleScriptMessage(Innovative::ProcessStatusEvent & Event);
    void  HandleTimer(OpenWire::NotifyEvent & Event);
    void  HandleOnLog(Innovative::ProcessStatusEvent & Event);
    void  HandleOnSend(Innovative::SenderEvent<VeloBuffer> & Event);

    void  HandleDisableTrigger(OpenWire::NotifyEvent & Event);
    void  HandleExternalTrigger(OpenWire::NotifyEvent & Event);
    void  HandleSoftwareTrigger(OpenWire::NotifyEvent & Event);

    void  Log(const std::string & msg);

    OpenWire::ThunkedEventHandler<Innovative::ProcessStatusEvent>  OnLog;
    OpenWire::ThunkedEventHandler<Innovative::SendBufferEvent>     OnSend;

    bool  IsDataLoggingCompleted();
    void  InitBddFile(Innovative::BinView & graph);

    void  DisplayLogicVersion();
    void  TallyBlock(size_t bytes);
};


#ifdef __CLR_VER
#pragma managed(pop)
#endif
}; // namespace Innovative

#endif

