// Duc_Provider_Gp.cpp
//
// Data flow and hardware I/O for Duc units
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2013 Innovative Integration Corporation.
//  All Rights Reserved.
//
//---------------------------------------------------------------------------

#include "DucProvider_Gp.h"
#include <SystemSupport_Mb.h>
#include <StringSupport_Mb.h>
#include <limits>
#include <cmath>
#include <numeric>
#include <Exception_Mb.h>
#include <FileSupport_Mb.h>
#include <DeviceEnum_Mb.h>
#include <Buffer_Mb.h>
#include <BufferDatagrams_Mb.h>
#include <BufferHeader_Mb.h>
#include <sstream>
#include <cstdio>
#include <Magic_Mb.h>

#include <GpDsp_Config_Gp.h>
#include <ctime>
#include <cstring>

using namespace std;


namespace   Innovative
{

using namespace Drc;
using namespace Duc;

//===========================================================================
//  CLASS Duc_Provider  -- Hardware Access and Application I/O Class
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class Duc_Provider
//---------------------------------------------------------------------------

Duc_Provider::Duc_Provider(Duc_ProviderSettings & settings, IDucManagerInterface * mgr)
: Settings(settings), MGR(mgr),
StreamConnected(false), Stopped(true),
RxTime(6), RxBytesPerBlock(6),
TxTime(6), TxBytesPerBlock(6),
FftTime(6),   FftBytesPerBlock(6),
FWordCount(0), FFftWordCount(0),
SamplesPerWord(2), WordsToLog(0)
{
        //
        //  Create Duc Module Object using Factory
        Module = MakeModule(this);

        OnLog.SetEvent(this, &Duc_Provider::HandleOnLog);
        Module->OnLog.SetEvent(this, &Duc_Provider::HandleOnLog);

        Timer.Interval(1000);

}

//---------------------------------------------------------------------------
//  destructor for class Duc_Provider
//---------------------------------------------------------------------------

Duc_Provider::~Duc_Provider()
{
        Close();
}

//---------------------------------------------------------------------------
//  Duc_Provider::BoardCount() -- Query number of installed boards
//---------------------------------------------------------------------------

unsigned int  Duc_Provider::BoardCount()
{
        return Module->BoardCount();
}

//---------------------------------------------------------------------------
//  Duc_Provider::Open() -- Open Hardware & set up callbacks
//---------------------------------------------------------------------------

void  Duc_Provider::Open()
{
        //
        //  Configure Trigger Manager Event Handlers
        Trig.OnDisableTrigger.SetEvent(this, &Duc_Provider::HandleDisableTrigger);
        Trig.OnExternalTrigger.SetEvent(this, &Duc_Provider::HandleExternalTrigger);
        Trig.OnSoftwareTrigger.SetEvent(this, &Duc_Provider::HandleSoftwareTrigger);

        Trig.DelayedTrigger(true);   // trigger delayed after start
        //
        //  Configure Stream Event Handlers
        Stream.OnVeloDataAvailable.SetEvent(this, &Duc_Provider::HandleDataAvailable);
        Stream.OnVeloDataAvailable.Synchronize();
        Stream.OnVeloDataRequired.SetEvent(this, &Duc_Provider::HandleDataRequired);
        Stream.OnVeloDataRequired.Synchronize();
        Stream.DirectDataMode(false);

        Stream.OnAfterStop.SetEvent(this, &Duc_Provider::HandleAfterStop);
        Stream.OnAfterStop.Synchronize();
        Stream.OnBeforeStart.SetEvent(this, &Duc_Provider::HandleBeforeStreamStart);
        Stream.OnBeforeStart.Synchronize();
        Stream.OnAfterStart.SetEvent(this, &Duc_Provider::HandleAfterStreamStart);
        Stream.OnAfterStart.Synchronize();


        Stream.RxLoadBalancing(false);
        Stream.TxLoadBalancing(false);

        Timer.OnElapsed.SetEvent(this, &Duc_Provider::HandleTimer);
        Timer.OnElapsed.Thunk();

        Module->AttachSettings(Settings);
        Module->OpenHardware();


        if (Module->IsOpen()==false)
        return;

        Log("Module Device opened successfully...");

        Module->EnableCalibration();

        //
        //  Connect Stream
        Module->ConnectToStream(Stream);
        StreamConnected = true;
        Log("Stream Connected...");

        PrefillPacketCount = Stream.PrefillPacketCount();
        //
        //  Set up Loggers and graphs
        std::stringstream sv;
        sv << "Stream" << Ordinal();
        Graph.System().ServerSlotName(sv.str());

        DisplayLogicVersion();
        {
                std::stringstream ss;
                ss << Settings.Root() << "FFTData" << Ordinal() << ".bin";
                FFTLogger.FileName(ss.str());
        }


}

//---------------------------------------------------------------------------
//  Duc_Provider::Close() -- Close Hardware & set up callbacks
//---------------------------------------------------------------------------

void  Duc_Provider::Close()
{
        if (!IsOpen())
        return;

        Stream.Disconnect();
        StreamConnected = false;

        Module->CloseHardware();

        Log("Stream Disconnected...");
}

//---------------------------------------------------------------------------
//  Duc_Provider::CurrentDateTime()  -- Get current date/time, format is YYYY-MM-DD.HH:mm:ss
//---------------------------------------------------------------------------
const std::string Duc_Provider::CurrentDateTime()
{
        char       buf[80];

        time_t     now = time(0);
        struct tm  tstruct = *std::localtime(&now);

        strftime(buf, sizeof(buf), "%Y-%m-%d@%X", &tstruct);
        // ':' is illegal in filenames
        replace(buf, buf + std::strlen(buf), ':', '.');
        return buf;
}

//---------------------------------------------------------------------------
//  Duc_Provider::StartStreaming()  -- Initiate Data Flow
//---------------------------------------------------------------------------

void Duc_Provider::StartStreaming()
{
        //
        //  Set up Loggers and graphs
        std::stringstream ss;

        if (!Settings.DestinationDirectory.empty())
        ss << Settings.DestinationDirectory << GetSlash();

        if (Settings.Rx.EnableDateTimeLogging)
        ss << "Data" << Ordinal() << "_" << CurrentDateTime() << ".bin";
        else
        ss << "Data" << Ordinal() << ".bin";

        Logger.FileName(ss.str());
        Graph.BinFile(Logger.FileName());

        StartStreaming_Orig();
}

//---------------------------------------------------------------------------
//  Duc_Provider::StartStreaming_Orig() --  Standard ftn to start data flow
//---------------------------------------------------------------------------

void  Duc_Provider::StartStreaming_Orig()
{
        if (!StreamConnected)
        {
                Log("Stream not connected! -- Open the boards");
                return;
        }

        //
        //  Set up Parameters for Data Streaming
        //  ...First have UI get settings into our settings store
        if (Module->ValidateStreamParameters()==false)
        {
                StopStreaming();
                return;
        }

        time_t now = time(0);
        Module->TimeStamp(Settings.Rx.Utc, Settings.Rx.ExternalPps, Settings.Rx.TsfMode, now);

        FWordCount = 0;
        FFftWordCount = 0;
        WordsToLog = Module->WordsFromSamples(Settings.Rx.SamplesToLog);
        FftBlockCount = FFftBlockRate = 0;
        FTxBlockCount = FRxBlockCount = 0;
        FTxBlockRate  = FRxBlockRate  = 0;

        // Disable triggers initially  (done by library)
        Trig.DelayedTriggerPeriod(Settings.TriggerDelayPeriod);
        //Trig.ExternalTrigger(Settings.Rx.ExternalTrigger == 1);
        Trig.ExternalTrigger((Settings.Rx.ExternalTrigger ? true : false) || (Settings.Tx.ExternalTrigger ? true : false));
        Trig.AtConfigure();

        Module->BeforePreconfigure();
        //
        //  Always preconfigure
        Stream.Preconfigure();
        // Let the module know the actual frequency
        Module->GetProjectRegs()->CommonM.SampleRate(Module->AdcClockFrequencyActual());

        Module->BeforeStreaming();

        // Output Prefill
        if (!Settings.Tx.PlayFromFile.Enable)
        Stream.PrefillPacketCount(0);  // no prefill or data signalling
        else
        Stream.PrefillPacketCount(PrefillPacketCount);

        //
        //  Start Player if in use
        if (Settings.Tx.PlayFromFile.Enable)
        {
                Player.FileName(Settings.Tx.PlayFromFile.Filename);
                Player.Mode(Settings.Tx.PlayFromFile.Repeat ? DataPlayer::pmRepeat : DataPlayer::pmStop);
                try {
                        if (!Player.Start())
                        Log("PlayFromFile Open Error.");

                } catch (const MalibuException &e) {
                        Log(e.what());
                        return;
                }

        }

        //
        //  Start Loggers on active channels
        if (Settings.Rx.PlotEnable)
        {
                Graph.Quit();
        }

        if (Settings.Rx.LoggerEnable || Settings.Rx.PlotEnable)
        {
                try
                {
                Logger.WriteThrough(Settings.LoggerWriteThrough);
                Logger.CoalesceWrites(Settings.LoggerWriteCoalesce);
                        Logger.Start();
                        FFTLogger.Start();
                }
                catch (...)
                {
                        Log("LOGGING FAILED !!!");
                }
        }

        Trig.AtStreamStart();

        //
        //  Start Streaming
        Stopped = false;
        Stream.Start();
        RunTimeSW.Start();
        Log("Stream Mode started");

        Timer.Enabled(true);
}

//---------------------------------------------------------------------------
//  Duc_Provider::StopStreaming() --  Terminate data flow
//---------------------------------------------------------------------------

void  Duc_Provider::StopStreaming()
{
        if (!StreamConnected)
        {
                Log("Stream not connected! -- Open the boards");
                return;
        }

        if (Settings.Tx.PlayFromFile.Enable)
        {
                Player.Stop();
        }
        //
        //  Stop Streaming
        Stream.Stop();
        Stopped = true;
        Timer.Enabled(false);
        Trig.AtStreamStop();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  Data Flow Event Handlers
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//---------------------------------------------------------------------------
//  Duc_Provider::HandleDataAvailable() --  Handle received packet
//---------------------------------------------------------------------------

void  Duc_Provider::HandleDataAvailable(VitaPacketStreamDataEvent & Event)
{
        if (Stopped) return;

        VeloBuffer Packet;

        Event.Sender->Recv(Packet);
        VeloHeaderDatagram Hdr(Packet);
        MGR->DataPacketAvailable(Personality.Ordinal, Packet);

        if (Hdr.PeripheralId() == 0x41)
        {

                FftBlockCount++;
                if (Settings.Rx.EnablePlotting)
                {
                        MGR->Plot(Personality.Ordinal, Packet, 1);
                }

                if (Settings.Rx.EnableFFTLogging)
                {
                        FFTLogger.LogWithHeader(Packet);
                }

                return;
        }

        if (Hdr.PeripheralId() == 0x40)
        {
                double AvgBytes = FftBytesPerBlock.Process(IntegerDG(Packet).size() * sizeof(int));
                FFftBlockRate   = AvgBytes / (FftTime.Differential()*1.0e6);
                FFftWordCount  += Packet.SizeInInts();

                FftBlockCount++;
                if (Settings.Rx.EnablePlotting)
                {
                        MGR->Plot(Personality.Ordinal, Packet, 0);
                }

                if (Settings.Rx.EnableFFTLogging)
                {
                        FFTLogger.LogWithHeader(Packet);
                }

                return;
        }

        if (Settings.Rx.LoggerEnable)
        if (Settings.Rx.ContinuousLogging || (WordsToLog==0 || (FWordCount < WordsToLog)))
        {
                Logger.LogWithHeader(Packet);
        }

        FWordCount += Packet.SizeInInts();

        IntegerDG Packet_DG(Packet);

        TallyBlock(Packet_DG.size()*sizeof(int));

        // Per block triggering actions
        Trig.AtBlockProcess(Packet_DG.size()*sizeof(int));
}

//---------------------------------------------------------------------------
//  Duc_Provider::TallyBlock() --  Finish processing of received packet
//---------------------------------------------------------------------------

void  Duc_Provider::TallyBlock(size_t bytes)
{
        double Period = RxTime.Differential();
        double AvgBytes = RxBytesPerBlock.Process(static_cast<double>(bytes));
        if (Period)
        {
                FRxBlockRate = AvgBytes / (Period*1.0e6);
        }
        //
        //  Tally the blocks received
        ++FRxBlockCount;
        //
        //  Stop streaming when both Channels have passed their limit
        if (Settings.Rx.AutoStop && IsDataLoggingCompleted() && !Stopped)
        {
                // Stop counter and display it
                double elapsed = RunTimeSW.Stop();

                StopStreaming();
                Log("Stream Mode Stopped automatically");
                Log(std::string("Elasped (S): ") + FloatToString(elapsed));
        }
}

//---------------------------------------------------------------------------
// Duc_Provider::HandleDataRequired()
//---------------------------------------------------------------------------

void Duc_Provider::HandleDataRequired(VitaPacketStreamDataEvent & Event)
{
        SendOneBlock(Event.Sender);

        // Per block triggering actions (not used on X6)
        //Trig.AtBlockProcess( WaveformPacket.SizeInInts() );
}

//------------------------------------------------------------------------------
// Duc_Provider::SendOneBlock()
//------------------------------------------------------------------------------

void Duc_Provider::SendOneBlock(VitaPacketStream * PS)
{
        //ShortDG Packet_DG(WaveformPacket);
        size_t packet_size_bytes(0);
        //
        //  No matter what channels are enabled, we have one packet type
        //    to send here
        if (Settings.Tx.PlayFromFile.Enable)
        {
                VeloBuffer VB;
                InitHeader(VB);

                Player.PlayWithHeader(VB);
                PS->Send(0, VB);
                ++FTxBlockCount;
                packet_size_bytes = VB.SizeInBytes();


                std::string test;
                AlignedMemoryManager::Info(test);
        }
        else
        {
                //  Send the single WF packet
                //PS->Send(0, WaveformPacket);
                //++FTxBlockCount;
                //packet_size_bytes = WaveformPacket.SizeInBytes();
        }

        // Calculate transfer rate in kB/s
        double Period = TxTime.Differential();

        double AvgBytes = TxBytesPerBlock.Process(static_cast<double>(packet_size_bytes));
        if (Period)
        FTxBlockRate = AvgBytes / (Period*1.0e6);

}

//---------------------------------------------------------------------------
//  Duc_Provider::HandleAfterStop() --  Post stream termination event
//---------------------------------------------------------------------------

void Duc_Provider::HandleAfterStop(OpenWire::NotifyEvent & /*Event*/)
{
        //
        //  Stop Loggers on active Channels
        if (Settings.Rx.LoggerEnable || Settings.Rx.PlotEnable)
        {
                Logger.Stop();
                InitBddFile(Graph);

                FFTLogger.Stop();

                if (Logger.Logged() && Settings.Rx.PlotEnable)
                Graph.Plot();
        }
        Log(std::string("Analog I/O Stopped"));
        MGR->AfterStreamAutoStop(Ordinal());
}

//---------------------------------------------------------------------------
//  Duc_Provider::HandleBeforeStreamStart() --  Pre-streaming init event
//---------------------------------------------------------------------------

void Duc_Provider::HandleBeforeStreamStart(OpenWire::NotifyEvent & /*Event*/)
{
        MGR->BeforeStreamStart(Ordinal());
}

//---------------------------------------------------------------------------
//  Duc_Provider::HandleAfterStreamStart() --  Post streaming init event
//---------------------------------------------------------------------------

void Duc_Provider::HandleAfterStreamStart(OpenWire::NotifyEvent & /*Event*/)
{
        Log(std::string("Analog I/O started"));
}

//---------------------------------------------------------------------------
//  Duc_Provider::HandleTimer() --  Per-second status timer event
//---------------------------------------------------------------------------

void Duc_Provider::HandleTimer(OpenWire::NotifyEvent & /*Event*/)
{
        // Collect Data for Periodic Status call...
        // Display status
        MGR->PeriodicStatusUpdate(Ordinal());

        Trig.AtTimerTick();
}

//---------------------------------------------------------------------------
//  Duc_Provider::Log() --  Log message thunked to main thread
//---------------------------------------------------------------------------

void  Duc_Provider::Log(const std::string & msg)
{
        ProcessStatusEvent e(msg);
        OnLog.Execute(e);
}

//---------------------------------------------------------------------------
//  Duc_Provider::HandleOnLog() --
//---------------------------------------------------------------------------

void Duc_Provider::HandleOnLog(Innovative::ProcessStatusEvent & Event)
{
        MGR->Log(Ordinal(), Event.Message);
}



//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  Support Functions
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//---------------------------------------------------------------------------
//  Duc_Provider::IsDataLoggingCompleted() --check if done logging
//---------------------------------------------------------------------------

bool  Duc_Provider::IsDataLoggingCompleted()
{
        if (WordsToLog==0)
        return false;
        else
        return FWordCount >= WordsToLog;
}

//---------------------------------------------------------------------------
//  Duc_Provider::SoftwareAlert() --  Issue specified software alert
//---------------------------------------------------------------------------

void Duc_Provider::SoftwareAlert(unsigned int value)
{
        Log(std::string("Posting SW alert..."));
        Module->SoftwareAlert(value);
}

//---------------------------------------------------------------------------
//  Duc_Provider::Temperature() --  Query current module temperature
//---------------------------------------------------------------------------

int Duc_Provider::Temperature()
{
        return Module->Temperature();
}

//---------------------------------------------------------------------------
//  Duc_Provider::PllLocked() --  Query current PLL lock status
//---------------------------------------------------------------------------

bool  Duc_Provider::PllLocked()
{
        return Module->PllLocked();
}

//---------------------------------------------------------------------------
//  Duc_Provider::HandleDisableTrigger() --  Trigger Manager Trig OFF
//------------------------------------------------------------------------------

void  Duc_Provider::HandleDisableTrigger(OpenWire::NotifyEvent & /*Event*/)
{
        Module->SetExternalTrigger(false);
}

//---------------------------------------------------------------------------
//  Duc_Provider::HandleSoftwareTrigger() --  Trigger Manager Trig OFF
//------------------------------------------------------------------------------

void  Duc_Provider::HandleSoftwareTrigger(OpenWire::NotifyEvent & /*Event*/)
{
        Module->SetSoftwareTrigger(true);
}

//---------------------------------------------------------------------------
//  Duc_Provider::HandleExternalTrigger() --  Enable External trigger
//------------------------------------------------------------------------------

void  Duc_Provider::HandleExternalTrigger(OpenWire::NotifyEvent & /*Event*/)
{
        Module->SetExternalTrigger(true);
}

//------------------------------------------------------------------------------
//  Duc_Provider::DisplayLogicVersion() --  Log version info
//------------------------------------------------------------------------------

void  Duc_Provider::DisplayLogicVersion()
{
        Module->LogLogicVersion();
}

//------------------------------------------------------------------------------
//  Duc_Provider::LogicVersion() --  Version info
//------------------------------------------------------------------------------

string Duc_Provider::LogicVersion()
{
        return Module->LogicVersion();
}

//------------------------------------------------------------------------------
//  Duc_Provider::PeekLogic() --
//------------------------------------------------------------------------------

unsigned int Duc_Provider::PeekLogic(int offset)
{
    return Module->PeekLogic(offset);
}

//------------------------------------------------------------------------------
//  Duc_Provider::PokeLogic() --
//------------------------------------------------------------------------------

void Duc_Provider::PokeLogic(int offset, unsigned int value)
{
    return Module->PokeLogic(offset, value);
}

//------------------------------------------------------------------------------
//  Duc_Provider::CoreVersion() --  Version info
//------------------------------------------------------------------------------

pair<string,string> Duc_Provider::CoreVersion()
{
        return Module->CoreVersion();
}

//---------------------------------------------------------------------------
//  Duc_Provider::InitBddFile() --  Set up BDD File for an output file
//---------------------------------------------------------------------------

void Duc_Provider::InitBddFile(BinView & graph)
{
        PathSpec spec(graph.BinFile());
        spec.Ext(".bdd");
        if (FileExists(spec.Full()) && !Settings.Rx.OverwriteBdd)
        return;

        remove(spec.Full().c_str());

        // Make a new BDD
        const int bits = Module->Input_Bits();

        const int limit = 1 << (bits-1);

        graph.Time().LowerLimit(-limit);
        graph.Time().UpperLimit(limit);
        graph.Time().Break(1);
        graph.Time().SeamSize(Module->VeloDataSize()*sizeof(int));
        graph.Time().AnnotateSeams(true);

        graph.Fft().LowerLimit(-150);
        graph.Fft().ScaleYType(BinViewOptions::FftOptions::sLog);
        graph.Fft().ScaleXType(BinViewOptions::FftOptions::sLog);
        graph.Fft().CriteriaThreshold(75);
        int pktsize = std::min(Settings.Rx.PacketSize, 0x10000);
        BinViewOptions::FftOptions::IIPoints pts(static_cast<BinViewOptions::FftOptions::IIPoints>(
        TrailingZeroCount(pktsize)-7));
        graph.Fft().Points(pts);
        graph.Fft().Window(BinViewOptions::FftOptions::wBlackmann);
        graph.Fft().Average(1);

        graph.Text().DataFormat("%11.3f");
        graph.System().UpdateRate(BinViewOptions::SystemOptions::ms1000);
        //    graph.System().ServerSlotName("BinView");
        graph.Leap(Settings.Rx.PacketSize);

        graph.Polarity(Module->Input_Unipolar() ? BinView::pUnsigned : BinView::pSigned);
        graph.DataSpan(300);
        graph.DataType(BinView::tInt);
        graph.Units("mV");
        graph.Samples(1000);

        graph.SampleRate(static_cast<float>(Module->AdcClockFrequencyActual() / 1.0e3f) / MGR->DecimationFactor(0));
        graph.Channels(MGR->EnabledChannels(Ordinal()));
        graph.Devices(MGR->EnabledDevices(Ordinal()));

        int percore    = Module->GetProjectRegs()->DdcM.ChannelCountPerCore();
        int group_size = (percore == 0) ? 1 : percore;
        if (MGR->EnabledDevices(Ordinal()) == MGR->EnabledChannels(Ordinal()))
            {
            graph.SignificantBits(bits);
            }
        else
            {
            graph.SignificantBits(Module->SampleWidth()); // DDC
            graph.Devices(MGR->EnabledDevices(Ordinal()));
            graph.Channels(MGR->EnabledChannels(Ordinal()) * group_size);
            }

        if (!Module->GetProjectRegs()->DdcM.IsVitaPacket())
            graph.Devices(1);

        graph.System().Source(Module->BddOutputFileType());
        graph.NullDataValue(.12345e-19f);
        graph.InputSpan((float)Module->Input_FullSpan());
        graph.ScalingEnabled(true);

        //  Board Specific Modifications
        //Module.ModifyBddFile(graph);

        graph.Save();
}

//---------------------------------------------------------------------------
//  Duc_Provider::ClockInfo() -- Dump clock info
//---------------------------------------------------------------------------

void  Duc_Provider::ClockInfo()
{
        Module->LogClockInfo();
}

//---------------------------------------------------------------------------
//  Duc_Provider::Make() -- Dump clock info
//---------------------------------------------------------------------------

void  Duc_Provider::Make(FiclGroup & group, IFicl_UserInterface * mgr)
{
        //  Fill in FICL object(s) for the current Module
        Module->Make_Ficls(group, mgr);
}

//---------------------------------------------------------------------------
//  Duc_Provider::Make() --
//---------------------------------------------------------------------------

void  Duc_Provider::Make(DdcChannelsProxy & group)
{
        Module->Make_DdcChannels(group);
}

//---------------------------------------------------------------------------
//  Duc_Provider::Make() --
//---------------------------------------------------------------------------

void  Duc_Provider::Make(DucChannelsProxy & group)
{
        Module->Make_DucChannels(group);
}

//---------------------------------------------------------------------------
//  Duc_Provider::Make_FftChannels() --
//---------------------------------------------------------------------------

void  Duc_Provider::Make(FftChannelsProxy & group)
{
        Module->Make_FftChannels(group);
}

//---------------------------------------------------------------------------
//  Duc_Provider::Make() --
//---------------------------------------------------------------------------

void Duc_Provider::Make(PowerMeterProxy & group)
{
        Module->Make_PowerMeter(group);
}

//---------------------------------------------------------------------------
//  Duc_Provider::Make() --
//---------------------------------------------------------------------------

void Duc_Provider::Make(OverflowProxy & group)
{
        Module->Make_OverflowDetector(group);
}

//---------------------------------------------------------------------------
//  Duc_Provider::Make() --
//---------------------------------------------------------------------------

void Duc_Provider::Make(ToneGeneratorProxy &group)
{
        Module->Make_ToneGenerator(group);
}

//---------------------------------------------------------------------------
//  Duc_Provider::SampleWidth() --
//---------------------------------------------------------------------------

unsigned int Duc_Provider::SampleWidth() const
{
        return Module->SampleWidth();
}

//---------------------------------------------------------------------------
//  Duc_Provider::IsIQEnabled() --
//---------------------------------------------------------------------------

bool Duc_Provider::IsIQEnabled() const
{
        return Module->IsIQEnabled();
}

//---------------------------------------------------------------------------
//  Duc_Provider::IsIQEnabled() --
//---------------------------------------------------------------------------

bool Duc_Provider::IsUnsigned() const
{
        return Module->IsUnsigned();
}

//==============================================================================
//  CLASS Duc_ProviderPersonality  - unique Settings for a Duc Provider
//==============================================================================
//==============================================================================
//  CLASS Duc_TxSettings
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class Duc_TxSettings
//------------------------------------------------------------------------------

Duc_TxSettings::Duc_TxSettings()
:  //ActiveChannels(AnalogOutChannels()),
Gain(AnalogOutChannels()), Offset(AnalogOutChannels())
{
}

//==============================================================================
//  CLASS Duc_RxSettings
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class Duc_RxSettings
//------------------------------------------------------------------------------

Duc_RxSettings::Duc_RxSettings()
:  //ActiveChannels(AnalogInChannels()),
Gain(AnalogInChannels()), Offset(AnalogInChannels())
{

}

//==============================================================================
//  CLASS Duc_ProviderSettings  - shared Settings for all Duc Providers
//==============================================================================
//------------------------------------------------------------------------------
//  constructor for class Duc_ProviderSettings
//------------------------------------------------------------------------------

Duc_ProviderSettings::Duc_ProviderSettings()
:  AlertEnable(AnalogInAlerts(), false)
{
        // Init defaults
#ifdef LINUX
        int kDefaultBMSize = 4;
#else
        int kDefaultBMSize = 32;
#endif
        Rx.BusmasterSize			= kDefaultBMSize;
        Tx.BusmasterSize			= kDefaultBMSize;
        PhaseDetectorFrequency		= PhaseDetectorDefault();

        SampleClockSource			= 1;
        ExtClockSrcSelection		= 0;
        ReferenceClockSource		= 1;
        ReferenceRate			= 10.0f;
        TriggerDelayPeriod			= 2;

        Tx.Calibrated = true;

        Tx.ExtTriggerSrcSelection   = 0;

        AlertEnable[2]			= char(1);
        AlertEnable[3]			= char(1);
        AlertEnable[5]			= char(1);
        AlertEnable[7]			= char(1);

        Rx.SampleRate			= static_cast<float>(MaxInRateMHz());
        Tx.SampleRate			= static_cast<float>(MaxOutRateMHz());
        Tx.ExternalTrigger			= 0;
        Tx.EdgeTrigger			= 1;

        Tx.PlayFromFile.Enable		= true;
        Tx.PlayFromFile.Repeat		= true;
        Tx.PlayFromFile.Filename    = std::string("../../Playback/Data0.bin");

#ifdef LINUX
        DestinationDirectory        = ".//";
#else
        DestinationDirectory        = ".\\";
#endif

        PowerMeterCalibrationOffset         = 0;

        Tx.PacketSize			= 0x10000;
        Rx.PacketSize			= 0x10000;
        //  Streaming
        Rx.LoggerEnable             = true;
        Rx.PlotEnable               = true;
        Rx.OverwriteBdd             = true;
        Rx.ContinuousLogging        = false;
        Rx.AutoStop                 = false;
        Rx.EnableDateTimeLogging    = false;
        Rx.ExternalTrigger          = 0;
        Rx.EdgeTrigger              = 1;
        Rx.CalibOffset              = 0;
        //  Timestamp Config
        Rx.Utc                      = true;
        Rx.ExternalPps              = false;
        Rx.TsfMode                  = 1;
        //  FFT Graphing
        Rx.EnablePlotting           = true;
        Rx.EnableFFTLogging         = false;

        Rx.ExtTriggerSrcSelection   = 0;
        Rx.Framed                   = 0;
        Rx.FrameSize                = 0x4000;

        Rx.SamplingDuration         = 1.0f;

        Rx.Calibrated = true;

        Rx.EnablePowerMeterReads     = false;
        Rx.EnableArithmeticOflowReads= false;
}

//------------------------------------------------------------------------------
//  Load INI Settings
//------------------------------------------------------------------------------

void Duc_ProviderSettings::LoadINISettings()
{

        //  Board Settings
        Install( ToIni("Rx.BMSize",                 Rx.BusmasterSize,			Rx.BusmasterSize)  );
        Install( ToIni("Tx.BMSize",                 Tx.BusmasterSize,			Tx.BusmasterSize)  );

        //  Config Data
        Install( ToIni("PhaseDetectorFrequency",	PhaseDetectorFrequency,		PhaseDetectorFrequency) );
        Install( ToIni("SampleClockSource",			SampleClockSource,			SampleClockSource)  );
        Install( ToIni("ExtClockSrcSelection",		ExtClockSrcSelection,		ExtClockSrcSelection)  );
        Install( ToIni("ReferenceClockSource",		ReferenceClockSource,		ReferenceClockSource)  );
        Install( ToIni("ReferenceRate",				ReferenceRate,				ReferenceRate)  );
        //Install( ToIni("AlertEnable", "Alert",		AlertEnable,				AlertEnable ));
        Install( ToIni("TriggerDelay",				TriggerDelayPeriod,			TriggerDelayPeriod)  );

        //
        Install( ToIni("AlertBit0",					AlertEnable[0],				AlertEnable[0])  );
        Install( ToIni("AlertBit1",					AlertEnable[1],				AlertEnable[1])  );
        Install( ToIni("AlertBit2",					AlertEnable[2],				AlertEnable[2])  );
        Install( ToIni("AlertBit3",					AlertEnable[3],				AlertEnable[3])  );
        Install( ToIni("AlertBit4",					AlertEnable[4],				AlertEnable[4])  );
        Install( ToIni("AlertBit5",					AlertEnable[5],				AlertEnable[5])  );
        Install( ToIni("AlertBit6",					AlertEnable[6],				AlertEnable[6])  );
        Install( ToIni("AlertBit7",					AlertEnable[7],				AlertEnable[7])  );

        Install( ToIni("Rx.SampleRate",				Rx.SampleRate,				Rx.SampleRate )  );
        Install( ToIni("Tx.SampleRate",				Tx.SampleRate,				Tx.SampleRate )  );

        //  Trigger
        Install( ToIni("Tx.ExternalTrigger",		Tx.ExternalTrigger,			Tx.ExternalTrigger)  );
        Install( ToIni("Tx.ExtTriggerSrcSelection", Tx.ExtTriggerSrcSelection,  Tx.ExtTriggerSrcSelection)  );
        Install( ToIni("Tx.EdgeTrigger",			Tx.EdgeTrigger,				Tx.EdgeTrigger)  );

        //  ..PlayFromFile
        Install( ToIni("Tx.PlayFromFile.Enable",    Tx.PlayFromFile.Enable,     Tx.PlayFromFile.Enable)  );
        Install( ToIni("Tx.PlayFromFile.Filename",  Tx.PlayFromFile.Filename,   Tx.PlayFromFile.Filename)  );
        Install( ToIni("Tx.PlayFromFile.Repeat",    Tx.PlayFromFile.Repeat,     Tx.PlayFromFile.Repeat)  );
        //  ..Streaming
        Install( ToIni("Tx.PacketSize",				Tx.PacketSize,				Tx.PacketSize)  );
        Install( ToIni("Rx.SamplingDuration",       Rx.SamplingDuration,     Rx.SamplingDuration)  );

        Install( ToIni("Rx.EnablePowerMeterReads",     Rx.EnablePowerMeterReads,      Rx.EnablePowerMeterReads)  );
        Install( ToIni("Rx.EnableArithmeticOflowReads",Rx.EnableArithmeticOflowReads, Rx.EnableArithmeticOflowReads)  );

        // Rx
        Install( ToIni("Rx.ExternalTrigger",        Rx.ExternalTrigger,             Rx.ExternalTrigger)  );
        Install( ToIni("Rx.EdgeTrigger",            Rx.EdgeTrigger,                 Rx.EdgeTrigger)  );

        Install( ToIni("Rx.CalibOffset",            Rx.CalibOffset,                 Rx.CalibOffset)  );
        Install( ToIni("Rx.PacketSize",             Rx.PacketSize,               Rx.PacketSize)  );
        Install( ToIni("Framed",                    Rx.Framed,                   Rx.Framed)  );
        Install( ToIni("FrameSize",                 Rx.FrameSize,                Rx.FrameSize)  );
        Install( ToIni("ExtTriggerSrcSelection",    Rx.ExtTriggerSrcSelection,   Rx.ExtTriggerSrcSelection)  );

        //  Streaming
        Install( ToIni("Rx.LoggerEnable",           Rx.LoggerEnable,             Rx.LoggerEnable)  );
        Install( ToIni("Rx.PlotEnable",             Rx.PlotEnable,               Rx.PlotEnable)  );
        Install( ToIni("Rx.OverwriteBdd",           Rx.OverwriteBdd,             Rx.OverwriteBdd)  );
        Install( ToIni("Rx.ContinuousLogging",      Rx.ContinuousLogging,        Rx.ContinuousLogging)  );
        Install( ToIni("Rx.AutoStop",               Rx.AutoStop,                 Rx.AutoStop)  );
        Install( ToIni("Rx.EnableDateTimeLogging",  Rx.EnableDateTimeLogging,    Rx.EnableDateTimeLogging)  );

        Install( ToIni("LoggerWriteCoalesce",         LoggerWriteCoalesce,         LoggerWriteCoalesce)  );
        Install( ToIni("LoggerWriteThrough",          LoggerWriteThrough,          LoggerWriteThrough)  );

        //  Timestamp Config
        Install( ToIni("Rx.Utc",                    Rx.Utc,                      Rx.Utc)  );
        Install( ToIni("Rx.ExternalPps",            Rx.ExternalPps,              Rx.ExternalPps)  );
        Install( ToIni("Rx.TsfMode",                Rx.TsfMode,                  Rx.TsfMode)  );
        //  FFT Graphing
        Install( ToIni("EnablePlotting",            Rx.EnablePlotting,           Rx.EnablePlotting)  );
        Install( ToIni("EnableFFTLogging",          Rx.EnableFFTLogging,         Rx.EnableFFTLogging)  );


        Install( ToIni("DestinationDirectory",     DestinationDirectory,  DestinationDirectory) );


        Install( ToIni("AdcCalibrated",    Rx.Calibrated,  Rx.Calibrated)  );
        Install( ToIni("DacCalibrated",    Tx.Calibrated,  Tx.Calibrated)  );

        Install( ToIni("PowerMeterCalibrationOffset",  PowerMeterCalibrationOffset,  PowerMeterCalibrationOffset)  );


        Load();
}

//------------------------------------------------------------------------------
//  destructor for class Duc_ProviderSettings
//------------------------------------------------------------------------------

Duc_ProviderSettings::~Duc_ProviderSettings()
{
}


}  // namespace Innovative



