// DucProvider_Gp.h
//
// Data flow and hardware I/O for Duc units
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2013 Innovative Integration Corporation.
//  All Rights Reserved.
//
//---------------------------------------------------------------------------

#ifndef DucProvider_GpH
#define DucProvider_GpH

#include <GpProviderBase_Gp.h>
#include <StopWatch_Mb.h>

#include <Benchmark_Mb.h>
#include <SoftwareTimer_Mb.h>
#include <DataLogger_Mb.h>
#include <DataPlayer_Mb.h>

#include <Application/TriggerManager_App.h>
#include <Application/BinviewPlotter_App.h>
#include <Application/IniSaver_App.h>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

	//==============================================================================
	//  CLASS Duc_RxSettings
	//==============================================================================
	struct Duc_RxSettings
	{
		typedef std::vector<char>   BoolArray;
		typedef std::vector<int>    IntArray;
		typedef std::vector<float>  FloatArray;

		Duc_RxSettings();

		int             BusmasterSize;
		float           SampleRate;

		//  Trigger
		int             ExternalTrigger;
		int             EdgeTrigger;
        int             Framed;
        int             FrameSize;
        int             ExtTriggerSrcSelection;

		//  Streaming
		int             PacketSize;
		bool            ForcePacketSize;

		bool            Utc;               // Timestamp
		bool            ExternalPps;
		int             TsfMode;

		//  Logging
		bool            LoggerEnable;
		bool            PlotEnable;
		ii64            SamplesToLog;
		float	        SamplingDuration;
		bool            OverwriteBdd;
		bool            AutoStop;
		bool            ContinuousLogging;
		bool            EnablePowerMeterReads;
		bool            EnableArithmeticOflowReads;

		bool            EnablePlotting;
		bool            EnableFFTLogging;
		bool            EnableDateTimeLogging;

		//  Eeprom
		FloatArray      Gain;
		FloatArray      Offset;
		float           PllCorrection;
		bool            Calibrated;

		bool            Help;
		std::string     ModuleName;
		std::string     ModuleRevision;
		bool            InitDdcs;

		int             Divider;

		std::string     DebugScript;
		std::string     DestinationDirectory;
		double          PhaseDetectorFrequency;


		int             TimerInterval;
		float           CalibOffset;
	};

	//==============================================================================
	//  CLASS Duc_TxSettings
	//==============================================================================

	struct Duc_TxSettings
	{
		Duc_TxSettings();

		////////////
        //  Struct PlayFromFileD
		////////////
		struct PlayFromFileD
		{
			bool         Enable;
			bool         Repeat;
			std::string  Filename;
		};

		typedef std::vector<char>   BoolArray;
		typedef std::vector<int>    IntArray;
		typedef std::vector<float>  FloatArray;

		int             BusmasterSize;

		//  Trigger
		int             ExternalTrigger;
		int             ExtTriggerSrcSelection;
		int             EdgeTrigger;
		//  Streaming
		int             PacketSize;
		PlayFromFileD   PlayFromFile;
		float           SampleRate;
		//
		//  Not saved in INI file
		//  ..Eeprom
		FloatArray      Gain;
		FloatArray      Offset;
		bool            Calibrated;
		BoolArray       ActiveChannels;
	};

	//==============================================================================
	//  CLASS Duc_ProviderSettings  - shared Settings for all Duc Providers
	//==============================================================================
	class Duc_Provider;
	class Duc_ProviderSettings  :  public Innovative::IniSaver
	{
	public:
		typedef std::vector<char>   BoolArray;
		typedef std::vector<int>    IntArray;
		typedef std::vector<float>  FloatArray;

		// Ctor
		Duc_ProviderSettings();
		~Duc_ProviderSettings();
		void LoadINISettings();

		Duc_RxSettings      Rx;
		Duc_TxSettings      Tx;
		//
		//  Board Settings

		//  Common Config Data
		double          PhaseDetectorFrequency;
		int             SampleClockSource;
		int             ExtClockSrcSelection;
		int             ReferenceClockSource;
		float           ReferenceRate;
		BoolArray       AlertEnable;
		int             TriggerDelayPeriod;

		std::string     DestinationDirectory;

        float            PowerMeterCalibrationOffset;
        bool            LoggerWriteThrough;
        bool            LoggerWriteCoalesce;

	private:
		Duc_Provider *Module;
	};

	//  Factory Function Declaration to be implemented elsewhere
	extern IDrc_Module *  MakeModule(class Duc_Provider * provider);

	//===========================================================================
	//  CLASS IDucManagerInterface  -- Callback from Duc provider to manager
	//===========================================================================

	class IDucManagerInterface
	{
	public:
		virtual ~IDucManagerInterface()  {  }

		virtual void  Log(unsigned int ordinal, const std::string & a_string) = 0;
		virtual void  AfterStreamAutoStop(unsigned int ordinal) = 0;
		virtual void  PeriodicStatusUpdate(unsigned int ordinal) = 0;
		virtual void  BeforeStreamStart(unsigned int ordinal) = 0;
		virtual void  AfterStreamStart(unsigned int ordinal) = 0;
		virtual unsigned int EnabledChannels(unsigned int ordinal) = 0;
		virtual unsigned int EnabledDevices(unsigned int ordinal) = 0;
		virtual float DecimationFactor(unsigned int ordinal) = 0;
		virtual unsigned int EnabledChannelizerChannels(unsigned int ordinal) = 0;
		virtual void Plot(unsigned int ordinal, VeloBuffer &buffer, int type) = 0;
        virtual void DataPacketAvailable(unsigned int ordinal, VeloBuffer &buffer) = 0;
	};

	//===========================================================================
	//  CLASS Duc_Provider  -- Data flow and hardware I/O for Duc units
	//===========================================================================

	class Duc_Provider
	{

	public:
		//
		//  Member Functions
		Duc_Provider(Duc_ProviderSettings & settings, IDucManagerInterface * mgr);
		virtual ~Duc_Provider();
		//
		//  Public Interface to Manager
		void Open();
		bool IsOpen()
		{  return Module->IsOpen();  }
		void Close();
		void StartStreaming();
		void StopStreaming();
		bool IsStreaming()
		{  return Timer.Enabled();  }

		unsigned int BoardCount();

		const std::string CurrentDateTime();

		void  Make(FiclGroup & group, IFicl_UserInterface * mgr);
		void  Make(DdcChannelsProxy & group);
		void  Make(DucChannelsProxy & group);
		void  Make(FftChannelsProxy & group);
		void  Make(PowerMeterProxy & group);
		void  Make(OverflowProxy & group);
		void  Make(ToneGeneratorProxy &group);

		virtual void PassDefaults()
		{ Module->PassDefaults(Settings); }

		string LogicVersion();
		pair<string,string> CoreVersion();
		unsigned int SampleWidth() const;
		bool IsIQEnabled() const;
		bool IsUnsigned() const;

		double AdcClockFrequencyActual()
		{   return Module->AdcClockFrequencyActual(); }
		double DacClockFrequencyActual()
		{   return Module->DacClockFrequencyActual(); }

		//
		//  Private Support Ftns for Manager
	private:
		void StartStreaming_Orig();

	public:
		void SoftwareAlert(unsigned int value);
		int  Temperature();
		bool PllLocked();

		double FftBlockRate() const
		{  return FFftBlockRate;  }
		
		double RxBlockRate() const
		{  return FRxBlockRate;  }
		double TxBlockRate() const
		{  return FTxBlockRate;  }

		ii64 FramePerSec()
		{
			ii64 temp = FftBlockCount;
			FftBlockCount = 0;

			return temp;
		}


		ii64    SampleCount() const
		{  return FWordCount*SamplesPerWord;  }

		ii64    FftSampleCount() const
		{  return FFftWordCount*SamplesPerWord;  }

		ii64    SentPacketCount() const
		{  return FTxBlockCount;  }

		ii64    DroppedPacketCount() const
		{  return 0;  }

        virtual unsigned int   DdcCoresCount()
            { return Module->DdcCoresCount(); }
        virtual unsigned int   DucCoresCount()
            { return Module->DucCoresCount(); }
        virtual unsigned int   FftCoresCount()
            { return Module->FftCoresCount(); }

		void ClockInfo();
        unsigned int  PeekLogic(int offset);
        void          PokeLogic(int offset, unsigned int value);

	public: // Pseudo-protected
		// Data
		Duc_ProviderSettings &        Settings;
		Drc_ProviderPersonality       Personality;

		unsigned int Ordinal()  {  return Personality.Ordinal;  }
		ii64                                FftBlockCount;


        ///////////////////
        //Board Information
        ///////////////////
        //Logic Version
        std::string FpgaLogicVersion()
        { return  Module->FpgaLogicVersion(); }
        //Hardware Variant
        short FpgaHardwareVariant()
        { return  Module->FpgaHardwareVariant(); }
        //Revision
        short PciLogicRevision()
        { return  Module->PciLogicRevision(); }
        //Subrevision
        short FpgaLogicSubrevision()
        { return Module->FpgaLogicSubrevision(); }
        //Board Family
        std::string PciLogicFamily()
        { return  Module->PciLogicFamily(); }
        //Type
        std::string PciLogicType()
        { return  Module->PciLogicType(); }
        //Board Revision
        std::string PciLogicPcb()
        { return  Module->PciLogicPcb(); }
        //Chip
        std::string FpgaName()
        { return Module->FpgaName(); }
        //PCI EXPRESS
        std::string PCIExpressLanes()
        { return Module->PCIExpressLanes(); }



        std::string PrintDevices()
        { return  Module->PrintDevices(); }

	private:
		//
		//  Member Data
		IDrc_Module *                       Module;
		IDucManagerInterface *              MGR;
		Innovative::VitaPacketStream        Stream;
		Innovative::TriggerManager          Trig;
		Innovative::SoftwareTimer           Timer;
		Innovative::StopWatch               RunTimeSW;
		Innovative::DataLogger              Logger;
		Innovative::BinviewPlotter          RtPlot;
		Innovative::BinView                 Graph;

		Innovative::DataLogger              FFTLogger;

		Innovative::DataPlayer              Player;   // for "Play From File" mode

		//  App State variables
		bool                                StreamConnected;
		bool                                Stopped;
		int                                 PrefillPacketCount;

		//  App Status variables
		Innovative::AveragedRate            RxTime;
		Innovative::AveragedRate            RxBytesPerBlock;
		Innovative::AveragedRate            TxTime;
		Innovative::AveragedRate            TxBytesPerBlock;
		Innovative::AveragedRate            FftTime;
		Innovative::AveragedRate            FftBytesPerBlock;

		double                              FTxBlockRate;
		double                              FRxBlockRate;
		double                              FFftBlockRate;
		int                                 FTxBlockCount;
		int                                 FRxBlockCount;

		ii64                                FWordCount;
		ii64                                FFftWordCount;
		int                                 SamplesPerWord;
		ii64                                WordsToLog;
		//
		//  Member Functions
		void  HandleDataRequired(Innovative::VitaPacketStreamDataEvent & Event);
		void  HandleDataAvailable(Innovative::VitaPacketStreamDataEvent & Event);
		void  Handle_VPP_ImageAvailable(Innovative::VitaPacketParserImageAvailable & Event);

		void  HandleAfterStop(OpenWire::NotifyEvent & Event);
		void  HandleBeforeStreamStart(OpenWire::NotifyEvent & /*Event*/);
		void  HandleAfterStreamStart(OpenWire::NotifyEvent & Event);
		void  HandleScriptCommand(Innovative::ProcessStatusEvent & Event);
		void  HandleScriptMessage(Innovative::ProcessStatusEvent & Event);
		void  HandleTimer(OpenWire::NotifyEvent & Event);
		void  HandleOnLog(Innovative::ProcessStatusEvent & Event);

		void  HandleDisableTrigger(OpenWire::NotifyEvent & Event);
		void  HandleExternalTrigger(OpenWire::NotifyEvent & Event);
		void  HandleSoftwareTrigger(OpenWire::NotifyEvent & Event);

		void  Log(const std::string & msg);

		OpenWire::ThunkedEventHandler<Innovative::ProcessStatusEvent>  OnLog;

		bool  IsDataLoggingCompleted();
		void  InitBddFile(Innovative::BinView & graph);

		void  DisplayLogicVersion();
		void  TallyBlock(size_t bytes);
		void  SendOneBlock(VitaPacketStream * PS);

	};


#ifdef __CLR_VER
#pragma managed(pop)
#endif
}; // namespace Innovative

#endif


















#if 0
//  GpDsp_DucBoard_1000M_Impl.h
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2014 Innovative Integration Corporation.
//  All Rights Reserved.
//




#ifndef GpDsp_DucBoard_1000M_ImplH
#define GpDsp_DucBoard_1000M_ImplH

#include <DucProvider_Gp.h>

#include <VitaPacketStream_Mb.h>
#include <Ficl/FiclSystem_Mb.h>
// ...Board Specific
#include <X6_1000M_Mb.h>

#include <GpProject_DDC_Gp.h>

namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

	//===========================================================================
	//  CLASS X6_1000M_Duc  -- Implementation of DRC module for X6-1000M
	//===========================================================================

	class X6_1000M_Duc : public IDrc_MakerModule
	{
	public:
		X6_1000M_Duc(Duc_Provider * provider);
		virtual ~X6_1000M_Duc()  {  }
		//
		//  Operations
		virtual void  OpenHardware();
		virtual void  CloseHardware();
		//  ...Open Time
		virtual void  ConnectToStream(VitaPacketStream & stream);
		//  ...Stream Config Time
		virtual bool  ValidateStreamParameters();
		virtual void  BeforePreconfigure();
		virtual void  BeforeStreaming();
		//
		//  Trigger Manipulation
		virtual void  SetSoftwareTrigger(bool state);
		virtual void  SetExternalTrigger(bool state);
		//
		//  Information (Pre Open)
		virtual unsigned int BoardCount();
		virtual bool         IsOpen() {  return Opened;  }
		//virtual string       Tag();
		//
		//  Information (Post Open)
		virtual ii64        WordsFromSamples(ii64 samples);
		virtual int         Temperature();
		virtual bool        PllLocked();
		virtual int         Input_Bits();
		virtual bool        Input_Unipolar();
		virtual int         Input_FullSpan();

		virtual unsigned int   Input_ActiveChannels();
		virtual unsigned int   Input_ActiveDevices();

		virtual unsigned int   VeloDataSize();
		virtual double         ClockFrequencyActual();
        virtual void           TimeStamp(bool utc, bool extPps, int tsf_mode, time_t t);

		string LogicVersion();
		string CoreVersion();

		//
		//  Debug
		virtual void  Make_Ficls(FiclGroup & group, IFicl_UserInterface * mgr);
		virtual void  Make_DdcChannels(DdcChannelsProxy & group);
		virtual void  Make_DucChannels(DucChannelsProxy & group);
		virtual void  Make_PowerMeter(PowerMeterProxy & group);
		virtual void  Make_OverflowDetector(OverflowProxy &group);
		virtual void  Make_ToneGenerator(ToneGeneratorProxy &group);
		virtual void  Make_FftChannels(FftChannelsProxy & group);

		virtual void  SoftwareAlert(unsigned int code);
		virtual void  LogLogicVersion();
		virtual void  LogClockInfo();

		//    Innovative::X6_1000M  &Ref()  // I don't want to expose the module if I can avoid it.
		//            { return Module; }

	protected:
		//
		//  Support Methods
		virtual Drc_ProviderPersonality &  Personality()  {  return Provider->Personality;  }

	private:
		//
		//  Data Members
		Duc_Provider *        Provider;
		Innovative::X6_1000M  Module;
		bool                  Opened;

		GpDspMemoryExtension_BindTo<Innovative::X6_1000M>  GpDsp_Ex;   // needs to be after Module
		GpProject_DDC        GpRegs;
		//
		//  Support Methods
		Duc_ProviderSettings &     Settings()     {  return Provider->Settings;  }
		//
		//  Module Methods
		void  HookAlerts();
		void  ConfigureAlerts(std::vector<char> & AlertEnable);
		void  SetInputSoftwareTrigger(bool state);
		void  SetOutputSoftwareTrigger(bool state);
		void  SetInputPacketDataSize(unsigned int size);
		void  SetOutputPacketDataSize(unsigned int size);
		void  SetInputTestConfiguration( bool enable, int mode_idx );
		void  SetOutputTestConfiguration( bool enable, int mode_idx );

		typedef std::vector<float> PulseSettingArray;
		void  SetInputPulseTriggerConfiguration(bool enable, float period,
		const PulseSettingArray & Delays,
		const PulseSettingArray & Widths );
		void  SetOutputPulseTriggerConfiguration(bool enable, float period,
		const PulseSettingArray & Delays,
		const PulseSettingArray & Widths );
		//
		//  Event Handlers
		//  ...Streaming
		void  HandleBeforeStreamStart(OpenWire::NotifyEvent & Event);
		void  HandleAfterStreamStop(OpenWire::NotifyEvent & Event);
		//  ...Alerts
		void  HandleTimestampRolloverAlert(Innovative::AlertSignalEvent & event);
		void  HandleSoftwareAlert(Innovative::AlertSignalEvent & event);
		void  HandleWarningTempAlert(Innovative::AlertSignalEvent & event);
		void  HandleInputFifoOverrunAlert(Innovative::AlertSignalEvent & event);
		void  HandleTriggerAlert(Innovative::AlertSignalEvent & event);
		void  HandleInputOverrangeAlert(Innovative::AlertSignalEvent & event);

		std::string Elapsed(size_t timestamp);

		bool  ValidateSampleRate(double should_be, double was, double allowed_range);

	};







#ifdef __CLR_VER
#pragma managed(pop)
#endif

}  // end namespace Innovative


#endif // GpDsp_DrcBoard_ImplH


#endif
