// GpManagerBase_Gp.cpp
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2013 Innovative Integration Corporation.
//  All Rights Reserved.
//

#include <GpManagerBase_Gp.h>
#include <Malibu_Mb.h>
#include <IppMemoryUtils_Mb.h>
#include <GpDsp_Config_Gp.h>

namespace   Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif


//===========================================================================
//  CLASS GpManagerBase  -- Hardware Access and Application I/O Base Class
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class GpManagerBase
//---------------------------------------------------------------------------

GpManagerBase::GpManagerBase(IManager_GUIInterface & ui, IManager_FiclInterface & ficl)
    :  UI(ui), FICL(ficl)//, MGR(new GpDucManagerInterface(this))
{
    // Use IPP performance memory functions.
    Init::UsePerformanceMemoryFunctions();

    SetTraceVerbosity(Settings.DebugVerbosity);
}

//---------------------------------------------------------------------------
//  destructor for class GpManagerBase
//---------------------------------------------------------------------------

GpManagerBase::~GpManagerBase()
{
    //DrcSettings.Save();
    Settings.Save();

   // Close();
}

//---------------------------------------------------------------------------
//  GpManagerBase::BoardTag() --
//---------------------------------------------------------------------------

string GpManagerBase::BoardTag(unsigned int /*b*/)
{
    return "Module";
}

//---------------------------------------------------------------------------
//  GpManagerBase::UpdateAdcInfo() --
//---------------------------------------------------------------------------

void GpManagerBase::UpdateAdcInfo(unsigned int ch)
{
    // Update Fft Channel decimation
    FftChannels.Wrapper(ch).DecimationRate(1);
    //
    FftChannels.Wrapper(ch).ReCalculateWindow();
    FftChannels.SourceChangeApply(ch);
}

//---------------------------------------------------------------------------
//  GpManagerBase::UpdateDdcInfo() --
//---------------------------------------------------------------------------

void GpManagerBase::UpdateDdcInfo(unsigned int ch)
{
    // Read decimation factor from the newly loaded file
    DdcChannels.Wrapper(FftChannels.Cache.Source).ReloadFile();

    vector<float> cfs = DdcChannels.DecimationCoeffs();
    float decrate = cfs.size() ? cfs[FftChannels.Cache.Source] : 1.0f;
    // Update Fft Channel decimation
    FftChannels.Wrapper(ch).DecimationRate(decrate);
    //
    FftChannels.Wrapper(ch).ReCalculateWindow();
    FftChannels.SourceChangeApply(ch);
}

//---------------------------------------------------------------------------
//  GpManagerBase::SetSourceMaps() --
//---------------------------------------------------------------------------

void GpManagerBase::SetSourceMaps(unsigned int adc_count)
{
    vector<int> values;
    static int adc_map[] = { 0x010000, 0x020000, 0x040000, 0x080000,
                             0x100000, 0x200000, 0x400000, 0x800000,
                           };
    // ddc
    for(size_t i = 0; i < DdcChannels.Size(); i++)
        values.push_back(i);
    // adc
    copy(&adc_map[0], &adc_map[0] + adc_count, back_inserter(values));
    // test
    values.push_back(0x01000000);

    for (size_t i = 0; i < FftChannels.Size(); i++)
        {
        FftChannels.Wrapper(i).SetSourceMap(values);
        }
}

//---------------------------------------------------------------------------
//  GpManagerBase::FftSources() --
//---------------------------------------------------------------------------

vector<std::string> GpManagerBase::FftSources(unsigned int ordinal)
{
    vector<std::string> ddcs;

    DdcChannels.ChannelTags(ddcs, ordinal);
    vector<string> adcs = Innovative::GpDsp_Gui::DdcSourceList();
    // Fft input src = ddc channels + adc channels
    copy(adcs.begin(), adcs.end(), back_inserter(ddcs));

    return ddcs;
}

//---------------------------------------------------------------------------
//  GpManagerBase::AdjustTargets() --
//---------------------------------------------------------------------------

void GpManagerBase::AdjustTargets()
{
    // DRC_Count size and DRC_Targets vector size mismatch
    if (Settings.DRC_Targets.size() < Settings.DRC_Count)
        {
        // Make the vectors the same size as the DRC_Count
        Settings.DRC_Targets.resize(Settings.DRC_Count);
        Settings.DRC_Targets_Enabled.resize(Settings.DRC_Count);
        // Initialize newly allocated slots with default values.
        for (unsigned int i = 0; i < Settings.DRC_Count; i++)
            {
            Settings.DRC_Targets[i] = i;
            Settings.DRC_Targets_Enabled[i] = true;
            }
        }
}

//---------------------------------------------------------------------------
//  GpManagerBase::BoardList() --
//---------------------------------------------------------------------------

void GpManagerBase::BoardList(vector<string> &List, bool only_enabled)
{
    List.clear();
    for (unsigned int i = 0; i < Settings.DRC_Count; i++)
        {
        if (only_enabled)
            {
            if (Settings.DRC_Targets_Enabled[i])
                List.push_back(BoardTag(i) + string(":") + IntToString(i));
            }
        else
            {
            List.push_back(BoardTag(i) + string(":") + IntToString(i));
            }
        }
}

//---------------------------------------------------------------------------
//  GpManagerBase::ConfigureSystem() -
//---------------------------------------------------------------------------

void GpManagerBase::ConfigureSystem()
{
    Make_Drc();
    Make_Ficls();

    Make_Clocks();
}

//---------------------------------------------------------------------------
//  GpManagerBase::Make_Components() - These objects require opened board
//---------------------------------------------------------------------------

void GpManagerBase::Make_Components()
{
    Make_DdcChannels();
    Make_DucChannels();
    Make_FftChannels();
    Make_ToneGenerator();
    Make_PowerMeter();
    Make_OverflowDetector();

    LoadDrcSettings();
}

//---------------------------------------------------------------------------
//  GpManagerBase::Reset() -
//---------------------------------------------------------------------------

void GpManagerBase::Reset(bool state)
{
    DdcChannels.Reset(state);
    DucChannels.Reset(state);
    PowerMeter.Reset(state);
    OverflowDetector.Reset(state);
    ToneGenerator.Reset(state);
    FftChannels.Reset(state);
}

//---------------------------------------------------------------------------
//  GpManagerBase::CommandTrigger() -
//---------------------------------------------------------------------------

void GpManagerBase::CommandTrigger(CommandTriggerType which)
{
    switch(which)
    {
    case eDDC:
        DdcChannels.CommandTrigger();
        break;
    case eDUC:
        DucChannels.CommandTrigger();
        break;
    case eCommon:
        PowerMeter.CommandTrigger();
        OverflowDetector.CommandTrigger();
        ToneGenerator.CommandTrigger();
        break;
    case eBoth:
        CommandTrigger(eDDC);
        CommandTrigger(eDUC);
        CommandTrigger(eCommon);
    }
}

//---------------------------------------------------------------------------
//  GpManagerBase::CoreReset() -
//---------------------------------------------------------------------------

void GpManagerBase::CoreReset()
{
    Reset(true);
}

//---------------------------------------------------------------------------
//  GpManagerBase::Open() - open all hardware
//---------------------------------------------------------------------------

void GpManagerBase::Open()
{
    UI.BeforeDrcOpen();
    Open_Drc();
    ClockModules.Open();
    UI.AfterDrcOpen();
}

//---------------------------------------------------------------------------
//  GpManagerBase::BeforeStreamStart()
//---------------------------------------------------------------------------

void GpManagerBase::BeforeStreamStart(unsigned int ordinal)
{
    CoreReset();
    // Ddc stuff
    DucChannels.SamplingRate(DucClockFrequencyActual(ordinal) / 1.0e6f);
    DucChannels.ApplyCached();
    DdcChannels.SamplingRate(DdcClockFrequencyActual(ordinal) / 1.0e6f);
    DdcChannels.ApplyCached();

    if (DdcChannels.Cache.FrameMode.en)
        DdcChannels.NextScheduledFrequency(true);

    // Tone generator
	
    ToneGenerator.SamplingRate(DdcClockFrequencyActual(ordinal)/ 1.0e6f,
                               DucClockFrequencyActual(ordinal)/ 1.0e6f
                               );
    if (DucChannels.Size())
        ToneGenerator.InterpolationRate(DucChannels[0]->filter.InterpolationRate);
    ToneGenerator.ApplyCached();
    // Power Meter
    PowerMeter.Init();
    // Go
    CommandTrigger();
    // Fft
    //FftChannels.SamplingRate(DrcSettings.Rx.SampleRate);
    FftChannels.SamplingRate(DdcSampleFrequency(ordinal));
    FftChannels.ApplyCached();
}

//---------------------------------------------------------------------------
//  GpManagerBase::Close() - close all hardware
//---------------------------------------------------------------------------

void GpManagerBase::Close()
{
    UI.BeforeDrcClose();
    Close_Drc();
    ClockModules.Close();
    UI.AfterDrcClose();
}

//---------------------------------------------------------------------------
//  GpManagerBase::StartStreaming() - start streaming on boards
//---------------------------------------------------------------------------

void GpManagerBase::StartStreaming()
{
    UI.BeforeDrcStreamStart();
    ClockModules.Start();
    StartStreaming_Drc();
    UI.AfterDrcStreamStart();
}

//---------------------------------------------------------------------------
//  GpManagerBase::StopStreaming()  - stop streaming on boards
//---------------------------------------------------------------------------

void GpManagerBase::StopStreaming()
{
    UI.BeforeDrcStreamStop();
    ClockModules.Stop();
    StopStreaming_Drc();
    UI.AfterDrcStreamStop();
}

//---------------------------------------------------------------------------
//  GpManagerBase::Plot()
//---------------------------------------------------------------------------

void GpManagerBase::Plot(unsigned int ordinal, VeloBuffer &buffer, int type)
{
    UI.Plot(ordinal, buffer, type);
}

//---------------------------------------------------------------------------
//  GpManagerBase::DataPacketAvailable()
//---------------------------------------------------------------------------
void GpManagerBase::Drc_DataPacketAvailable(unsigned int ordinal, VeloBuffer &buffer)
{
    UI.DataPacketAvailable(ordinal, buffer);
}

//---------------------------------------------------------------------------
//  GpManagerBase::Drc_Log()
//---------------------------------------------------------------------------

void GpManagerBase::Drc_Log(unsigned int ordinal, const std::string & a_string)
{
    UI.Log("[" + IntToString(ordinal) + "]-> " + a_string);
}

//---------------------------------------------------------------------------
//  GpManagerBase::Drc_AfterStreamAutoStop()
//---------------------------------------------------------------------------

void GpManagerBase::Drc_AfterStreamAutoStop(unsigned int ordinal)
{
    UI.AfterStreamAutoStop(ordinal);
}

//---------------------------------------------------------------------------
//  GpManagerBase::Drc_BeforeStreamStart()
//---------------------------------------------------------------------------

void GpManagerBase::Drc_BeforeStreamStart(unsigned int ordinal)
{
    BeforeStreamStart(ordinal);

    UI.BeforeStreamStart(ordinal);
}

//---------------------------------------------------------------------------
//  GpManagerBase::Drc_AfterStreamStart()
//---------------------------------------------------------------------------

void GpManagerBase::Drc_AfterStreamStart(unsigned int ordinal)
{
    UI.AfterStreamStart(ordinal);
}

//---------------------------------------------------------------------------
//  GpManagerBase::Drc_PeriodicStatusUpdate()
//---------------------------------------------------------------------------

void GpManagerBase::Drc_PeriodicStatusUpdate(unsigned int ordinal)
{
    UI.PeriodicStatus(ordinal);
}

//---------------------------------------------------------------------------
//  GpManagerBase::DecimationFactor()
//---------------------------------------------------------------------------

float GpManagerBase::DecimationFactor(unsigned int /*ordinal*/)
{
    float factor = 0;
    // First enabled channel's filter is used
    for (size_t i = 0; i < DdcChannels.Size(); i++)
        if (DdcChannels[i]->info.Enable)
            {
            factor = DdcChannels[i]->filter.DecimationFactor;
            break;
            }
    // DdcChannels.BypassDdc == 1 is Raw
    return DdcChannels.BypassDdc ? 1.0f : factor;
}

//---------------------------------------------------------------------------
//  GpManagerBase::EnabledChannelizerChannels()
//---------------------------------------------------------------------------

unsigned int GpManagerBase::EnabledChannelizerChannels(unsigned int /*ordinal*/)
{
    unsigned int count = 0;

    for (size_t i = 0; i < DdcChannels.Size(); i++)
        if (DdcChannels[i]->info.Enable)
            {
            count += DdcChannels.Chans[i]->ChannelizerChannelCount();
            }
    return count;
}

//---------------------------------------------------------------------------
//  GpManagerBase::EnabledChannels()
//---------------------------------------------------------------------------

unsigned int GpManagerBase::EnabledChannels(unsigned int ordinal)
{
    // DdcChannels.BypassDdc == 1 is Raw
    return DdcChannels.BypassDdc ? EnabledDevices(ordinal) : EnabledDevices(ordinal) * 2;
}

//---------------------------------------------------------------------------
//  GpManagerBase::EnabledDevices()
//---------------------------------------------------------------------------

unsigned int GpManagerBase::EnabledDevices(unsigned int ordinal)
{
    unsigned int devs = 0;
    int cg = -1;
    for (size_t i = 0; i < DdcChannels.Size(); i++)
        if (DdcChannels[i]->info.Enable && DdcChannels[i]->WhichBoard() == ordinal )
            {
            if (cg != DdcChannels.Wrapper(i).Group())
                {
                devs++;
                cg = DdcChannels.Wrapper(i).Group();
                }
            }
    return devs;
}


//---------------------------------------------------------------------------
//  GpManagerBase::GetFiclTargets()
//---------------------------------------------------------------------------

void GpManagerBase::GetFiclTargets(std::vector<std::string>& ficl_targets)
{
    //  ask FiclIo units for thier descriptive names
    for (size_t i = 0; i < FiclUnits.Size(); i++)
        {
        ficl_targets.push_back(FiclUnits[i].Tag());
        }
}

//---------------------------------------------------------------------------
//  GpManagerBase::Ficl_OnCommandKeyPress()
//---------------------------------------------------------------------------

void GpManagerBase::Ficl_OnCommandKeyPress(unsigned int ficlT, char &Key)
{
    if (ficlT<FiclUnits.Size())
        FiclUnits[ficlT].FiclIo_OnCommandKeyPress(Key);
}

//---------------------------------------------------------------------------
//  GpManagerBase::Ficl_ExecuteDebugScript()
//---------------------------------------------------------------------------

void GpManagerBase::Ficl_ExecuteDebugScript(unsigned int ficlT, const std::string & script_file)
{
    if (ficlT<FiclUnits.Size())
        FiclUnits[ficlT].ExecuteDebugScript(script_file);
}

//---------------------------------------------------------------------------
//  GpManagerBase::Ficl_Logic()
//---------------------------------------------------------------------------

void GpManagerBase::Ficl_Logic(unsigned int ficlT, const std::string & address, const std::string & data)
{
    if (ficlT<FiclUnits.Size())
        FiclUnits[ficlT].Logic(address, data);
}

//---------------------------------------------------------------------------
//  GpManagerBase::Ficl_Logic()
//---------------------------------------------------------------------------

void GpManagerBase::Ficl_Logic(unsigned int ficlT, const std::string & address)
{
    if (ficlT<FiclUnits.Size())
        FiclUnits[ficlT].Logic(address);
}

//---------------------------------------------------------------------------
//  GpManagerBase::Ficl_Port()
//---------------------------------------------------------------------------

void GpManagerBase::Ficl_Port(unsigned int ficlT, const std::string & address, const std::string & data)
{
    if (ficlT<FiclUnits.Size())
        FiclUnits[ficlT].Port(address, data);
}

//---------------------------------------------------------------------------
//  GpManagerBase::Ficl_Port()
//---------------------------------------------------------------------------

void GpManagerBase::Ficl_Port(unsigned int ficlT, const std::string & address)
{
    if (ficlT<FiclUnits.Size())
        FiclUnits[ficlT].Port(address);
}

//---------------------------------------------------------------------------
//  GpManagerBase::Ficl_Device()
//---------------------------------------------------------------------------

void GpManagerBase::Ficl_Device(unsigned int ficlT, const std::string & address, const std::string & device,
                               const std::string & data)
{
    if (ficlT<FiclUnits.Size())
       FiclUnits[ficlT].Device(address, device, data);
}

//---------------------------------------------------------------------------
//  GpManagerBase::Ficl_Device()
//---------------------------------------------------------------------------

void GpManagerBase::Ficl_Device(unsigned int ficlT, const std::string & address, const std::string & device)
{
    if (ficlT<FiclUnits.Size())
       FiclUnits[ficlT].Device(address, device);
}

//---------------------------------------------------------------------------
//  GpManagerBase::DdcChannelList()
//---------------------------------------------------------------------------

void GpManagerBase::DdcChannelList(std::vector<std::string>& channel_list)
{
    DdcChannels.ChannelTags(channel_list);
}

//---------------------------------------------------------------------------
//  GpManagerBase::DucChannelList()
//---------------------------------------------------------------------------

void GpManagerBase::DucChannelList(std::vector<std::string>& channel_list)
{
    DucChannels.ChannelTags(channel_list);
}

//---------------------------------------------------------------------------
//  GpManagerBase::SetTraceVerbosity() - change debug trace level
//---------------------------------------------------------------------------

void GpManagerBase::SetTraceVerbosity(unsigned int level)
{
    TraceVerbosity(level);
}

//---------------------------------------------------------------------------
//  GpManagerBase::ChannelMap()
//---------------------------------------------------------------------------
std::string  GpManagerBase::ChannelMap(unsigned int ordinal)
{
    string map;
    for (size_t i = 0; i < DdcChannels.Size(); i++)
        if (DdcChannels[i]->info.Enable && DdcChannels[i]->WhichBoard() == ordinal )
            map += " DDC" + IntToString(i) + "<= ADC" + IntToString(DdcChannels[i]->info.Source) + ", ";
    return map;
}

//===========================================================================
//  CLASS GpDrcManagerSettings  -- Hardware Access and Application I/O Class
//===========================================================================
//---------------------------------------------------------------------------
//  constructor for class GpDrcManagerSettings
//---------------------------------------------------------------------------

GpDrcManagerSettings::GpDrcManagerSettings()
    :  IniSaver("Settings_Manager.ini")//,
{
    //
    //  DRC Settings
    Install( ToIni("DRC_Count", DRC_Count, 1u) );
    Install( ToIni("DRC_Targets", "Target", DRC_Targets, 0) );
    Install( ToIni("DRC_Targets_Enabled", "Target_Enabled", DRC_Targets_Enabled, 1) );

    Install( ToIni("Clock_Module_Enabled", Clock_Module_Enabled, 0u) );
    Install( ToIni("Clock_Module_Type", Clock_Module_Type,   std::string("sbc-duo"))  );

    //
    //  Misc Settings
    Install( ToIni("DebugVerbosity", DebugVerbosity, int(Trace::vNormal)));

    static int DRC_Targets_Enabled0;
    Install( ToIni("DRC_Targets_Enabled0", DRC_Targets_Enabled0, 1) );

    Load();

    if (DRC_Targets_Enabled.empty() && DRC_Count > 0)
        {
        DRC_Targets.resize(DRC_Count);
        DRC_Targets_Enabled.resize(DRC_Count);
        // Enable the first board
        DRC_Targets_Enabled[0] = DRC_Targets_Enabled0;
        }

}

//---------------------------------------------------------------------------
//  destructor for class GpDrcManagerSettings
//---------------------------------------------------------------------------

GpDrcManagerSettings::~GpDrcManagerSettings()
{

}

#ifdef __CLR_VER
#pragma managed(pop)
#endif

}  // namespace Innovative

