// GpProviderBase_Gp.h
//
// Data flow and hardware I/O for Duc units
//
//    INNOVATIVE INTEGRATION CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or nondisclosure
//  agreement with Innovative Integration Corporation and may not be copied or
//  disclosed except in accordance with the terms of that agreement.
//  Copyright (c) 2000..2014 Innovative Integration Corporation.
//  All Rights Reserved.
//
//---------------------------------------------------------------------------

#ifndef GpProviderBase_GpH
#define GpProviderBase_GpH

#include <VitaPacketStream_Mb.h>
#include <Application/FiclIo_App.h>
#include <BinView_Mb.h>

#include <GpProject_DDC_Gp.h>
#include <ChannelsProxy_Gp.h>
#include <FftProxy_Gp.h>
#include <SupportProxy_Gp.h>

using GP::DdcChannelsProxy;
using GP::DucChannelsProxy;
using GP::DdcChannelInfo;
using GP::DucChannelInfo;


namespace Innovative
{
#ifdef __CLR_VER
#pragma managed(push, off)
#endif

//==============================================================================
//  CLASS Drc_ProviderPersonality  - unique Settings for a  Provider
//==============================================================================

class Drc_ProviderPersonality
{
public:
    int             Ordinal;   // unique identifer for Drc board
    int             Target;
    bool            Enabled;
};

//===========================================================================
//  CLASS IDrc_Module  -- Interface to a Duc Module Implementation
//===========================================================================

class IDrc_Module
{
public:
    virtual ~IDrc_Module()  {  }

    //
    //  Operations
    virtual void  OpenHardware() = 0;
    virtual void  CloseHardware() = 0;
    //  ...Open Time
    virtual void  ConnectToStream(VitaPacketStream & stream) = 0;
    //  ...Stream Config Time
    virtual bool  ValidateStreamParameters() = 0;
    virtual void  BeforePreconfigure() = 0;
    virtual void  BeforeStreaming() = 0;
    //
    //  Trigger Manipulation
    virtual void  SetSoftwareTrigger(bool state) = 0;
    virtual void  SetExternalTrigger(bool state) = 0;
    //
    //  Information (Pre Open)
    virtual unsigned int BoardCount() = 0;
    virtual bool         IsOpen() = 0;
    //
    //  Information (Post Open)
    virtual ii64        WordsFromSamples(ii64 samples) = 0;
    virtual int         Temperature() = 0;
    virtual bool        PllLocked() = 0;
    virtual int         Input_Bits() = 0;
    virtual bool        Input_Unipolar() = 0;
    virtual int         Input_FullSpan() = 0;
    virtual unsigned int SampleWidth() const = 0;
    virtual bool	   IsIQEnabled() const = 0;
    virtual bool           IsUnsigned() const = 0;
    virtual unsigned int   Input_ActiveChannels() = 0;
    virtual unsigned int   Input_ActiveDevices() = 0;

    virtual unsigned int   VeloDataSize() = 0;
    virtual double         AdcClockFrequencyActual() = 0;
    virtual double         DacClockFrequencyActual()
	{ return 1.0; }
    virtual void           TimeStamp(bool utc, bool extPps, int tsf_mode, time_t t) = 0;
    virtual BinViewOptions::SystemOptions::IISource BddOutputFileType()
                                    {  return BinViewOptions::SystemOptions::sVita;  }

    //
    //
    virtual void  Make_Ficls(FiclGroup & group, IFicl_UserInterface * mgr) = 0;
    virtual void  Make_DdcChannels(DdcChannelsProxy & group) = 0;
    virtual void  Make_DucChannels(DucChannelsProxy & group) = 0;
    virtual void  Make_FftChannels(FftChannelsProxy & group) = 0;
    virtual void  Make_PowerMeter(PowerMeterProxy & group) = 0;
    virtual void  Make_OverflowDetector(OverflowProxy & group) = 0;
    virtual void  Make_ToneGenerator(ToneGeneratorProxy &group) = 0;

    virtual void  SoftwareAlert(unsigned int code) = 0;
    virtual void  LogLogicVersion() = 0;
    virtual void  LogClockInfo() = 0;
    virtual unsigned int  PeekLogic(int offset) = 0;
    virtual void          PokeLogic(int offset, unsigned int value) = 0;

    virtual void          EnableCalibration() = 0;

    virtual unsigned int   DdcCoresCount() = 0;
    virtual unsigned int   DucCoresCount() = 0;
    virtual unsigned int   FftCoresCount() = 0;


    virtual string LogicVersion() = 0;
    virtual pair<string,string> CoreVersion() = 0;
    virtual void AttachSettings(Innovative::IniSaver& saver) = 0;

    virtual void PassDefaults(class Drc_ProviderSettings& /*defs*/)
    { }
    virtual void PassDefaults(class Duc_ProviderSettings& /*defs*/)
    { }
    virtual void DumpInterruptInfo()
    { }
    virtual GpProject_DDC* GetProjectRegs() = 0;


    OpenWire::ThunkedEventHandler<Innovative::ProcessStatusEvent>  OnLog;
    void  Log(const std::string & msg);

    OpenWire::ThunkedEventHandler<Innovative::SenderEvent<VeloBuffer> > OnSend;



    ///////////////////
    //Board Information
    ///////////////////
    //Logic Version
    virtual std::string FpgaLogicVersion() = 0;
    //Hardware Variant
    virtual short FpgaHardwareVariant() = 0;
    //Revision
    virtual short PciLogicRevision() = 0;
    //Subrevision
    virtual short FpgaLogicSubrevision() = 0;
    //Board Family
    virtual std::string PciLogicFamily() = 0;
    //Type
    virtual std::string PciLogicType() = 0;
    //Board Revision
    virtual std::string PciLogicPcb() = 0;
    //Chip
    virtual std::string FpgaName() = 0;
    //PCI EXPRESS
    virtual std::string PCIExpressLanes() = 0;


    virtual std::string PrintDevices() = 0;

protected:
    //
    //  Support Methods
    virtual Drc_ProviderPersonality &  Personality() = 0;
    virtual std::string                Tag();

};

//===========================================================================
//  CLASS IDrc_MakerModule  -- IDrc_Module with implementation for Make_ ers
//===========================================================================

class IDrc_MakerModule : public IDrc_Module
{
public:
    virtual ~IDrc_MakerModule()  {  }
    //
    //
    //---------------------------------------------------------------------------
    template <class ModuleT>
    void  Make_Ficls_Impl(ModuleT & Module, FiclGroup & group, IFicl_UserInterface * mgr, const std::string & help_command)
    {
        //  make FICL Io
        FiclIo * x6_ficl = new FiclIo(Tag(), mgr, help_command);
        x6_ficl->FiclSys().ConnectTarget(&Module);  //  Attach board specific commands to FICL
        x6_ficl->FiclIo_Init();  //  Start Ficl System

        group.Append(x6_ficl);

    }
    //---------------------------------------------------------------------------
    template <class ModuleT>
    void  Make_Ficls_Impl_Tag(ModuleT & Module, FiclGroup & group, IFicl_UserInterface * mgr, const std::string & help_command,
                              const std::string & tag_end)
    {
        //  make FICL Io
        std::string ficl_tag = Tag() + tag_end;
        FiclIo * x6_ficl = new FiclIo(ficl_tag, mgr, help_command);
        x6_ficl->FiclSys().ConnectTarget(&Module);  //  Attach board specific commands to FICL
        x6_ficl->FiclIo_Init();  //  Start Ficl System

        group.Append(x6_ficl);

    }
    //---------------------------------------------------------------------------
    template <class ModuleT>
    void  Make_FftChannels_Impl(ModuleT & Module, GpProject_DDC &GpRegs, FftChannelsProxy & group)
    {
        if (!GpRegs.ProjectInfo.Enabled(GspDspProjectInfo::Spectrum))
        {
            //double check
            if(GpRegs.SpectralM.GetEnableBits() == 0xda55) //No FFT
            {
                Log("No FFT Cores.");
                return;
            }
        }

        std::stringstream msg;
        msg << "Number of FFT Cores : " << GpRegs.SpectralFftM.CoreCount();
        Log(msg.str());

        msg.str("");

        //if (GpRegs.SpectralFftM.WhichArchitecture() == GpDspSubmodule::aParallel)
        //    {
        //    msg << "Parallel architecture ";
        //    }
        //else
        //    {
        //    msg << "Serial architecture ";
        //    }
        //Log(msg.str());

        // Enable channels
        size_t count = GpRegs.SpectralFftM.CoreCount();

    #if 0
        // Set PIDs
        Module.VeloIn().Custom_Fft_VeloPacketConfig(count);
        Module.VeloIn().Custom_Tls_VeloPacketConfig(count);
    #else
        Module.VeloIn().SetStreamCount(2 + 2 * count);
        // Disable loopback
        Module.VeloIn().Enable(1, 0);

       unsigned int ver = (GpRegs.ProjectInfo.SubVersion() & 0xff00) >> 8;

       if (ver < 5)
          {
          // Force packet size for Data & Loopback
          Module.VeloIn().ForceVeloPacketSize(0, true);
          Module.VeloIn().ForceVeloPacketSize(1, true);
          }


       // Sets Velo parameters for FFT & TLS only if the logic version is less than 5
       for (size_t i = 0; ver < 5 && i < count; i++)
            {
            // Set pids
            Module.VeloIn().VeloPacketId(2*i + 2, 0x40 + 2*i + 0);
            Module.VeloIn().VeloPacketId(2*i + 3, 0x40 + 2*i + 1);
            // Set packet size
            Module.VeloIn().VeloDataSize(2*i + 2, 16448);
            Module.VeloIn().VeloDataSize(2*i + 3, 520);
            // Force packet size
            Module.VeloIn().ForceVeloPacketSize(2*i + 2, true);
            Module.VeloIn().ForceVeloPacketSize(2*i + 3, true);
            }

    #endif

        for (size_t i = 0; i < count; i++)
            {
            FftChannelWrapper *ch = new FftChannelWrapper(&GpRegs.SpectralM, i, FftInfo(), Personality().Ordinal);
            ch->ConfigureWindowCoeffs(GpRegs.SpectralFftM.WhichArchitecture());
            ch->AddWrapper(new SpectralPeripheralWrapper(&GpRegs.SpectralPeriphM, 0));
            ch->AddWrapper(new SpectralMultiBufferWrapper(&GpRegs.SpectralMultiBufferM, 0));
            ch->AddWrapper(new SpectralWindowingWrapper(&GpRegs.SpectralWindowingM, 0));
            ch->AddWrapper(new SpectralFftWrapper(&GpRegs.SpectralFftM, 0));
            ch->AddWrapper(new SpectralPostProcWrapper(&GpRegs.SpectralPostProcM, 0));
            ch->AddWrapper(new SpectralVitaFramerWrapper(&GpRegs.SpectralVitaFramerM, 0));

            group.Append(ch);
            }
    }

    void  Make_DdcChannels_Impl(GpProject_DDC &GpRegs, DdcChannelsProxy & group, const vector<DdcChannelInfo> &defs);
	void  Make_DucChannels_Impl(GpProject_DDC &GpRegs, DucChannelsProxy & group, const vector<DucChannelInfo> &defs);
    void  Make_FftChannels_Impl(GpProject_DDC &GpRegs, FftChannelsProxy & group);
    void  Make_PowerMeter_Impl(GpProject_DDC &GpRegs, PowerMeterProxy & group);
    void  Make_OverflowDetector_Impl(GpProject_DDC &GpRegs, OverflowProxy & group);
    void  Make_ToneGenerator_Impl(GpProject_DDC &GpRegs, ToneGeneratorProxy &group, const vector<ToneGeneratorDataType> &tonedef);
protected:

};

//==============================================================================
//  CLASS  GpProviderBase  --  Base for all Providers
//==============================================================================

class GpProviderBase
{
public:
    GpProviderBase();
    virtual ~GpProviderBase();

};

#ifdef __CLR_VER
#pragma managed(pop)
#endif
}; // namespace Innovative

#endif
