#include "SbcDuoClockModule_Gp.h"
#include <sstream>
#include <iomanip>
#include <StringSupport_Mb.h>
#include <SystemSupport_Mb.h>
#include <DeviceEnum_Mb.h>
#include <Exception_Mb.h>
#include <Gps_Mb.h>
#include <string.h>

namespace Innovative   {
namespace SbcDuo_Clock {



//==============================================================================
//  CLASS ApplicationIo
//==============================================================================

//---------------------------------------------------------------------------
//  ApplicationIo::ApplicationIo() --  Ctor
//---------------------------------------------------------------------------

ApplicationIo::ApplicationIo(IUserInterface * ui)
    : FiclIo(ui), Settings("SbcClockSettings0.ini"), FOpened(false), UI(ui)
{
    OnLog.SetEvent(this, &ApplicationIo::HandleOnLog);
    Timer.OnElapsed.SetEvent(this, &ApplicationIo::HandleTimer);
    GpsInitTimer.OnElapsed.SetEvent(this, &ApplicationIo::HandleGpsInitTimer);
    Board.RawClockDevice().Pll().OnPllTuning.SetEvent(this, &ApplicationIo::HandlePllTuning);

    Status(sStatus, "Status: Idle");

    Timer.Enabled(true);

    //WARNING: vAll verbosity will cause bad clock output
    TraceVerbosity(Trace::vNormal);

    MasterTriggerAsserted = true;
    SoftwareTriggerAsserted = false;

}

//---------------------------------------------------------------------------
//  ApplicationIo::~ApplicationIo() --  Dtor
//---------------------------------------------------------------------------

ApplicationIo::~ApplicationIo()
{
    Timer.Enabled(false);
    Close();
}

//---------------------------------------------------------------------------
//  ApplicationIo::Open() --  Open board driver
//---------------------------------------------------------------------------

bool ApplicationIo::Open()
{
    if (FOpened)
        return true;

    //UI->GetSettings();

    // Gps
    Board.SelectGps(Settings.WhichGps);   // Hook GPS handlers after you select which
                                          //   will be used...

    Board.GpsCtrl().Gps().OnEpoch.SetEvent(this, &ApplicationIo::HandleOnEpoch);
    //
    //  Open Devices
    Board.Target(Settings.Target);
    try
        {
        Board.Open();
        }
    catch (Innovative::MalibuException & ex)
        {
        Log("Device open failure!! Cause:");
        Log(ex.what());
        return false;
        }
    catch (...)
        {
        Log("Device open failure---.");
        return false;
        }


    //
    // If we are using the IEEE1588 "GPS" we need to route PPS and clock reference from that source.
    ISbcDuoSourceSelector::ExtClockRefSignalSources whichRef =
        (Settings.WhichGps) ? ISbcDuoSourceSelector::clkExternal : ISbcDuoSourceSelector::clkGps;

    ISbcDuoSourceSelector::PulsePerSecSources whichPps =
            (Settings.WhichGps) ? ISbcDuoSourceSelector::ppsInternalFpga : ISbcDuoSourceSelector::ppsGps;

    Board.Selector().MasterReferenceSource(whichRef);
    Board.Selector().MasterPulsePerSecSource(whichPps);


    FOpened = true;
    Log("SbcDuo opened...");
    Status(sStatus, "Status: Open");
    DisplayLogicVersion();

    return true;
}

//---------------------------------------------------------------------------
//  ApplicationIo::Close() --  Close board driver
//---------------------------------------------------------------------------

void ApplicationIo::Close()
{
    if (!FOpened)
        return;

    GpsEnable(false);

    Board.Close();
    FOpened = false;


    Log("SbcDuo closed...");
}


//---------------------------------------------------------------------------
// ApplicationIo::Start()
//---------------------------------------------------------------------------

bool ApplicationIo::Start()
{
    if (!FOpened)
        return false;

    SetClock();
    SetTrig();
    SetDio();
    SetMezz();

    Board.Start();

    MasterTrigger(true);

    Log("Clocks Started");
    Status(sStatus, "Status: Running");

    return true;

}


//---------------------------------------------------------------------------
// ApplicationIo::StopTimer()
//---------------------------------------------------------------------------

void ApplicationIo::Stop()
{
    SoftwareTrigger(false);
    MasterTrigger(false);

    Board.Stop();

    Log("Clocks Stopped");

}

//---------------------------------------------------------------------------
//  ApplicationIo::BoardCount() -- Query number of installed boards
//---------------------------------------------------------------------------

int ApplicationIo::BoardCount()
{
    return Board.BoardCount();
}


//---------------------------------------------------------------------------
// ApplicationIo::ValidateSampleRate()
//---------------------------------------------------------------------------

bool ApplicationIo::ValidateSampleRate(double should_be, double was, double allowed_range)
{
    double error = (should_be-was)/should_be;
    if (error < (-allowed_range) || error > allowed_range)
        return false;
    else
        return true;
}

//---------------------------------------------------------------------------
//  ApplicationIo::SetClock() --
//---------------------------------------------------------------------------

void ApplicationIo::SetClock()
{
    //UI->GetSettings();

    //   Route reference.
    ISbcDuoClockIo::IIReferenceSource ref[] = { ISbcDuoClockIo::rsExternal, ISbcDuoClockIo::rsInternal };
    Board.Clock().Reference(ref[Settings.ReferenceClockSource]);

    Board.Clock().ReferenceFrequency(Settings.ReferenceRate * 1e6);
    //   Route clock
    ISbcDuoClockIo::IIClockSource src[] = { ISbcDuoClockIo::csExternal, ISbcDuoClockIo::csInternal };
    Board.Clock().Source(src[Settings.SampleClockSource]);
    //   Route External Ref Clock.
    ISbcDuoSourceSelector::ExtClockRefSignalSources xref[] = { ISbcDuoSourceSelector::clkExternal, ISbcDuoSourceSelector::clkGps };
    Board.Selector().MasterReferenceSource(xref[Settings.ExternalClockSource]);

    Board.Clock().System().Frequency(Settings.OutputFrequency * 1e6);
    {
        std::stringstream msg;
        msg << "Requested PLL Frequency: " << Settings.OutputFrequency;
        Log(msg.str());
    }

    Board.Clock().LowLatencyTriggerMode(Settings.LowLatencyTrigger);
    if (Settings.LowLatencyTrigger)
        {
        Log("Low Latency Trigger Selected. Warning: Framed Trigger and PRI Trigger now unavailable");
        }

    // Readback Frequency
    double sys_freq_actual = Board.Clock().System().FrequencyActual();
    double sys_freq = Board.Clock().System().Frequency();
    {
        std::stringstream msg;
        msg << "Actual PLL Frequency: " << sys_freq_actual  ;
        Log(msg.str());
    }
    if (ValidateSampleRate(sys_freq, sys_freq_actual, .05)==false)
    {
        std::stringstream msg;
        msg << "WARNING: Duo Sample Rate Mismatch - should be " << sys_freq << " but was " << sys_freq_actual  ;
        Log(msg.str());
    }

    //
    //  This code shows how to use the Gps on te Duo to initialize the timestamp for streaming
    //     packets on an XMC card
    //  Load "Vita Timestamp" Registers
//    if (Board.GpsCtrl().GpsEnabled())
//        {
//        //  GPS is on, use it
//        Xmc().VitaIn_TS().TimestampPpsMode(IVitaTimestamp::tsPpsMode);
//        Xmc().VitaIn_TS().TimestampIsMode(IVitaTimestamp::tsmUtc);
//        Xmc().VitaIn_TS().TimestampFsMode(IVitaTimestamp::tsfSamples);

//        unsigned int gps_now = static_cast<unsigned int>(Board.GpsCtrl().Gps().Time());
//        Xmc().VitaIn_TS().LoadTimestamp(gps_now);
//        }
//    else
//        {
//        //  GPS is off, don't use it
//        Xmc().VitaIn_TS().TimestampPpsMode(IVitaTimestamp::tsInternalTimer);
//        Xmc().VitaIn_TS().TimestampIsMode(IVitaTimestamp::tsmUtc);
//        Xmc().VitaIn_TS().TimestampFsMode(IVitaTimestamp::tsfSamples);

//        unsigned int now = static_cast<unsigned int>(time(NULL));
//        Xmc().VitaIn_TS().LoadTimestamp(now);
//        }


}

//---------------------------------------------------------------------------
//  ApplicationIo::SetTrig() --
//---------------------------------------------------------------------------

void ApplicationIo::SetTrig()
{
    //UI->GetSettings();

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  Trigger Configuration
    //
    //  Frame Triggering
    Board.Trigger().FramedMode((Settings.Trig.Framed)? true : false);
    Board.Trigger().Edge((Settings.Trig.EdgeTrigger)? true : false);
    Board.Trigger().FrameSize(Settings.Trig.FrameSize);
    Board.Trigger().External(Settings.Trig.TriggerSource);

    //Pps Config
    ISbcDuoSourceSelector::PulsePerSecSources Pps[] = { ISbcDuoSourceSelector::ppsInternalFpga, ISbcDuoSourceSelector::ppsGps };
    Board.Selector().MasterPulsePerSecSource(Pps[Settings.Trig.PpsSource]);

    //  Pulse Trigger Config
    //  ...read pattern from file -
    //
    //  File is INI format with these fields. Delay is offset from Period start,
    //  width is width of trigger
    //
    //  --------------------------
    //  [Delay]
    //  ArraySize=2
    //  P0=100
    //  P1=500
    //
    //  [Width]
    //  ArraySize=2
    //  P0=50
    //  P1=150
    //
    ReadPriPatternFile(Settings.Trig.Pulse.PulsePatternINIFile);

     //  ...add to module configuration
    SetPulseTriggerConfiguration(Settings.Trig.Pulse.Enable, Settings.Trig.Pulse.FiniteFrames,
                                 Settings.Trig.Pulse.FiniteFrameRearm,
                                 Settings.Trig.Pulse.Period, Settings.Trig.Pulse.FiniteFrameCount,
                                 Settings.Trig.Pulse.Delays,
                                 Settings.Trig.Pulse.Widths);

    // Set Decimation Factor
    int factor = Settings.Trig.DecimationEnable ? Settings.Trig.DecimationFactor : 0;
    Board.Trigger().Decimation(factor);

}

//---------------------------------------------------------------------------
//  ApplicationIo::SetPulseTriggerConfiguration() --  Configure Pulse Trigger
//---------------------------------------------------------------------------

void  ApplicationIo::SetPulseTriggerConfiguration(bool enable, bool finite_frames, bool finite_frame_rearm,
                                                  unsigned int period, unsigned int finite_frame_count,
                                                  std::vector<unsigned int> & delays,
                                                  std::vector<unsigned int> & widths )
{
    // Optionally enable ramp generator
    Board.Pulse().Reset();
    Board.Pulse().Enabled(enable);
    Board.Pulse().FiniteFrames(finite_frames);
    Board.Pulse().FiniteFrameRearm(finite_frame_rearm);
    Board.Pulse().FiniteFrameCount(finite_frame_count);

    size_t pulses = std::min(delays.size(), widths.size());
    pulses = (pulses<32) ? pulses : 32;   // cap at 32 pulses

    for (size_t i=0; i<pulses; i++)
        {
        Board.Pulse().AddEvent(static_cast<unsigned int>(period),
                                static_cast<unsigned int>(delays[i]),
                                static_cast<unsigned int>(widths[i]));
        }
}

//------------------------------------------------------------------------------
//  ApplicationIo::ReadPriPatternFile() --  Read PRI pattern from file
//------------------------------------------------------------------------------

void  ApplicationIo::ReadPriPatternFile(const std::string & file)
{
    Settings.Trig.Pulse.Delays.clear();
    Settings.Trig.Pulse.Widths.clear();

    IniReader patternFile(file);

    //  Define Data in Pattern File
    patternFile.Install( ToIni("Delay", "P",       Settings.Trig.Pulse.Delays,  (unsigned int)0x100 ));
    patternFile.Install( ToIni("Width", "P",       Settings.Trig.Pulse.Widths,  (unsigned int)0x100 ));

    patternFile.Load();

}

//---------------------------------------------------------------------------
//  ApplicationIo::SetDio() --
//---------------------------------------------------------------------------

void ApplicationIo::SetDio()
{
    //UI->GetSettings();

    Board.Dio().DioPortConfig(Settings.DioConfig);
    Board.Dio().DioPortData().Value(Settings.DioDataLow);
    Board.Dio().DioPortDataHigh().Value(Settings.DioDataHigh);
}

//---------------------------------------------------------------------------
//  ApplicationIo::SetMezz() --
//---------------------------------------------------------------------------

void ApplicationIo::SetMezz()
{
    //UI->GetSettings();

    Board.Mezz().OutputEnable(Settings.MezzOutputEnabled);

    SbcDuo_Mezzanine::XmcVSel_options  XmcOpts[]= { SbcDuo_Mezzanine::xvs_3p3v,SbcDuo_Mezzanine::xvs_2p5v
        ,SbcDuo_Mezzanine::xvs_1p8v,SbcDuo_Mezzanine::xvs_1p5v };
    //Xmc0
    Board.Mezz().XmcEnable(0,Settings.Xmc0Enabled);
    if (Settings.Xmc0Enabled)
    {
        Board.Mezz().Xmc_VSel(0,XmcOpts[Settings.Xmc0Voltage]);
        for(int i=0;i < 4;i++)
            Board.Mezz().Xmc_Direction(0,i,Settings.Xmc0Direction[i]);
    }

    //Xmc1
    Board.Mezz().XmcEnable(1,Settings.Xmc1Enabled);
    if(Settings.Xmc1Enabled)
    {
        Board.Mezz().Xmc_VSel(1,XmcOpts[Settings.Xmc1Voltage]);
        for(int i=0;i < 4;i++)
            Board.Mezz().Xmc_Direction(1,i,Settings.Xmc1Direction[i]);
    }

    //io
    SbcDuo_Mezzanine::IoVSel_options IoOpts[] = { SbcDuo_Mezzanine::ivs_5v,SbcDuo_Mezzanine::ivs_3p3v
        ,SbcDuo_Mezzanine::ivs_2p5v,SbcDuo_Mezzanine::ivs_1p8v };
    Board.Mezz().Io_VSel(IoOpts[Settings.MezzIoVoltage]);

}

//------------------------------------------------------------------------------
//  ApplicationIo::DisplayLogicVersion() --  Log version info
//------------------------------------------------------------------------------

void  ApplicationIo::DisplayLogicVersion()
{
    std::stringstream msg;
    msg << std::hex << "Logic Version: " << Board.Info().FpgaLogicVersion()
    << ", Hdw Variant: " << Board.Info().FpgaHardwareVariant()
    << ", Revision: " << Board.Info().PciLogicRevision()
    << ", Subrevision: " << Board.Info().FpgaLogicSubrevision();
    Log(msg.str());

    std::stringstream msg2;
    msg2 << std::hex << "Board Family: " << Board.Info().PciLogicFamily()
    << ", Type: " << Board.Info().PciLogicType()
    << "(" << Board.Name() << ")"
    << ", Board Revision: " << Board.Info().PciLogicPcb()
    << ", Chip: " << Board.FpgaName(Board.Info().FpgaChipType());
    Log(msg2.str());

}

//---------------------------------------------------------------------------
//  ApplicationIo::Log() --  Log message thunked to main thread
//---------------------------------------------------------------------------

void  ApplicationIo::Log(const string & msg)
{
    ProcessStatusEvent e(msg);
    OnLog.Execute(e);
}

//---------------------------------------------------------------------------
//  ApplicationIo::Status() --  Log message thunked to main thread
//---------------------------------------------------------------------------

void  ApplicationIo::Status(IIStatusType type, const string & msg)
{
    StatusMessageEvent e(type, msg);
    OnStatus.Execute(e);
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleTimer() --  Periodic status check
//---------------------------------------------------------------------------

void ApplicationIo::HandleTimer(OpenWire::NotifyEvent & /*event*/)
{
    //UI->PeriodicStatus();

}

//---------------------------------------------------------------------------
//  ApplicationIo::Temperature() --  return Logic Temperature
//---------------------------------------------------------------------------

int  ApplicationIo::Temperature()
{
    if (!FOpened)
        return 0;
    return Board.Thermal().LogicTemperature();
}

//---------------------------------------------------------------------------
//  ApplicationIo::PllLocked() --  return Logic Temperature
//---------------------------------------------------------------------------

bool  ApplicationIo::PllLocked()
{
    if (!FOpened)
        return false;
    return  Board.Clock().Locked();
}

//---------------------------------------------------------------------------
//  ApplicationIo::GpsEnable() --
//---------------------------------------------------------------------------

void ApplicationIo::GpsEnable(bool state)
{
    //UI->GetSettings();

    Settings.GpsStat.EpochTally = 0;
    GpsInitTally = 0;

    Board.GpsCtrl().GpsEnabled(state);
    if (state)
        {
         Log("Initializing GPS, this might take a while...");
        }

    GpsInitTimer.Enabled(state);

}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleGpsInitTimer() --  Periodic status check
//---------------------------------------------------------------------------

void ApplicationIo::HandleGpsInitTimer(OpenWire::NotifyEvent & /*event*/)
{
//  if (Settings.EpochTally && !Board.GpsCtrl().IsAntenna())
//        Log("Warning: Antenna is disconnected");

    if (!Settings.GpsStat.EpochTally && (GpsInitTally > 2))
        Log("Warning: GPS is not generating epoch interrupts");

//  if (++GpsInitTally > 2)
//        GpsInitTimer.Enabled(false);

    //UI->DisplayGpsStatus();
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleOnEpoch() --  Periodic status check
//---------------------------------------------------------------------------

void ApplicationIo::HandleOnEpoch(Innovative::GpsEpochEvent & Event)
{
    {
    stringstream msg;
    msg.precision(4);
    msg << "Epochs: " << ++Settings.GpsStat.EpochTally;
    }

    {
    stringstream msg;
    time_t currentTime = Board.GpsCtrl().Gps().Time();
//  msg << "UTC: " << ctime(&currentTime);
    msg << "UTC: " << Innovative::Trim(asctime(localtime(&currentTime)));
    Settings.GpsStat.UtcTime = msg.str();
    }

    {
    stringstream msg;
    GpsCoordinate Lat = Board.GpsCtrl().Gps().Latitude();
    msg << "Latitude: " << Lat.Print();
    Settings.GpsStat.Latitude = msg.str();
    }

    {
    stringstream msg;
    GpsCoordinate Long = Board.GpsCtrl().Gps().Longitude();
    msg << "Longitude: " << Long.Print();
    Settings.GpsStat.Longitude = msg.str();
    }

    {
    stringstream msg;
    msg << "Quality: " << Board.GpsCtrl().Gps().Fix();
    Settings.GpsStat.Quality = msg.str();
    }

    {
    stringstream msg;
    msg << "Satellites: " << Board.GpsCtrl().Gps().Satellites();
    Settings.GpsStat.Satellites = msg.str();
    }

    {
    stringstream msg;
    msg << "Lock: " <<Board.GpsCtrl().Gps().IsLocked();
    Settings.GpsStat.Lock = msg.str();
    }

    if (!Settings.DebugGps)
        return;

    for (GpsMessageList::iterator i = Event.List.begin(); i != Event.List.end(); ++i)
        DumpGpsMessage(*i);
}

//---------------------------------------------------------------------------
//  ApplicationIo::DumpGpsMessage() --
//---------------------------------------------------------------------------

void ApplicationIo::DumpGpsMessage(GpsMessage & msg)
{
    stringstream txt;
    txt << "Gps: " << msg.AsText();

    Log(txt.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::GpsCommand() --
//---------------------------------------------------------------------------

void ApplicationIo::GpsCommand(const std::string & message)
{
    Board.GpsCtrl().Gps().Command(message);
}

ii64 ApplicationIo::IeeeClockId()
{
    return Board.Babushka().ClockID();
}

bool ApplicationIo::IeeePllLocked()
{
    return Board.Babushka().PllLocked();
}

bool ApplicationIo::IeeeInSync()
{
    return Board.Babushka().InSync();
}

bool ApplicationIo::IeeeLinked()
{
    return Board.Babushka().Linked();
}

bool ApplicationIo::IeeeGigabit()
{
    return Board.Babushka().Gigabit();
}

void ApplicationIo::IeeeInit()
{
    Board.Babushka().InitIeee1588();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  EEProm support
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//---------------------------------------------------------------------------
//  ApplicationIo::WriteRom() --  Write rom using relevant settings
//---------------------------------------------------------------------------

void ApplicationIo::WriteRom()
{
    Board.IdRom().Name(Settings.ModuleName);
    Board.IdRom().Revision(Settings.ModuleRevision);

    Board.IdRom().StoreToRom();
}

//---------------------------------------------------------------------------
//  ApplicationIo::ReadRom() --  Read rom and update relevant settings
//---------------------------------------------------------------------------

void ApplicationIo::ReadRom()
{
    Board.IdRom().LoadFromRom();

    Settings.ModuleName = Board.IdRom().Name();
    Settings.ModuleRevision = Board.IdRom().Revision();
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandleOnLog() --
//---------------------------------------------------------------------------

void ApplicationIo::HandleOnLog(Innovative::ProcessStatusEvent & /*Event*/)
{
    //UI->Log(Event.Message);
}

//---------------------------------------------------------------------------
//  ApplicationIo::HandlePllTuning() --
//---------------------------------------------------------------------------

void ApplicationIo::HandlePllTuning(Innovative::Lmk04820PllTuningEvent & Event)
{
    if (Settings.PllTuningOverrideEnable)
        {
        Event.CLKin0_R = Settings.PllTuningOverride.CLKin0_R;
        Event.CLKin1_R = Settings.PllTuningOverride.CLKin1_R;
        Event.CLKin2_R = Settings.PllTuningOverride.CLKin2_R;
        Event.N1 = Settings.PllTuningOverride.N1;
        Event.R2 = Settings.PllTuningOverride.R2;
        Event.N2 = Settings.PllTuningOverride.N2;
        Event.N2_Cal = Settings.PllTuningOverride.N2_Cal;
        Event.N2_Pre = Settings.PllTuningOverride.N2_Pre;
        Event.OD = Settings.PllTuningOverride.OD;
        Event.Feedback_OD = Settings.PllTuningOverride.Feedback_OD;

        stringstream msg;
        msg << "CLKin0_R: " << Event.CLKin0_R
            << ", CLKin1_R: " << Event.CLKin1_R
            << ", CLKin2_R: " << Event.CLKin2_R
            << ", N1: " << Event.N1
            << ", R2: " << Event.R2
            << ", N2: " << Event.N2
            << ", N2_Cal: " << Event.N2_Cal
            << ", N2_Pre: " << Event.N2_Pre
            << ", OD: " << Event.OD
            << ", Feedback_OD: " << Event.Feedback_OD;

        Log(msg.str());
        }

}

//---------------------------------------------------------------------------
//  ApplicationIo::SoftwareTrigger() --
//---------------------------------------------------------------------------

void ApplicationIo::SoftwareTrigger(bool state)
{
    SoftwareTriggerAsserted = state;
    Board.SoftwareTrigger(state);
}

//---------------------------------------------------------------------------
//  ApplicationIo::MasterTrigger() --
//---------------------------------------------------------------------------

void ApplicationIo::MasterTrigger(bool state)
{
    MasterTriggerAsserted = state;
    Board.MasterTrigger(state);
}

//---------------------------------------------------------------------------
//  ApplicationIo::PpsTriggerArm() --
//---------------------------------------------------------------------------

void ApplicationIo::PpsTriggerArm()
{
    Board.ArmPpsTrigger();
}

//---------------------------------------------------------------------------
//  ApplicationIo::UartForIdx() -- return UART for index given
//---------------------------------------------------------------------------

Innovative::IMulticharUart & ApplicationIo::UartForIdx(unsigned int port_idx)
{
   switch (port_idx)
       {
       default:
       case 0:  // Gps
           return Board.Debug()->GpsUart(0);
       case 1:
           return Board.Debug()->GpsUart(1);
       case 2:
           return Board.Uart(port_idx-1);
       case 3:
           return Board.Uart(port_idx-1);
       }
}

//---------------------------------------------------------------------------
//  ApplicationIo::ClockInfo() -- Dump clock info
//---------------------------------------------------------------------------

void  ApplicationIo::ClockInfo()
{

    Log("Clock output data");
    for (unsigned int i=0; i<Duo().RawClockDevice().Pll().Outputs(); i++)
        {
        std::stringstream msg;
        msg <<  "Out # " << i << ": "
            << " Divider: "   << Duo().RawClockDevice().Pll().OutputDivider(i)
            << ", Out Freq: " << Duo().RawClockDevice().Pll().OutputFrequency(i)
            << ", Actual: "   << Duo().RawClockDevice().Pll().OutputFrequencyActual(i) ;
        Log(msg.str());
        }
}

//---------------------------------------------------------------------------
//  ApplicationIo::PllLockedStr() -- Display Pll Lock Status
//---------------------------------------------------------------------------

std::string  ApplicationIo::PllLockedStr()
{
    if (PllLocked())
        return std::string(" + ");
    else
        return std::string(" - ");
}

//------------------------------------------------------------------------------
//  ApplicationIo::ModuleName() --  Return name of module as string
//------------------------------------------------------------------------------

std::string ApplicationIo::ModuleName() const
{
    return Board.Name();
}

//------------------------------------------------------------------------------
//  ApplicationIo::ModuleRevision() --  Return rev of module as string
//------------------------------------------------------------------------------

std::string ApplicationIo::ModuleRevision()
{
    std::stringstream msg;
    char c = 'A' + static_cast<char>(Board.Info().PciLogicPcb());
    msg << c;
    return msg.str();
}

//---------------------------------------------------------------------------
//  ApplicationIo::TriggerStatus() -- Display Trigger Status
//---------------------------------------------------------------------------

std::string ApplicationIo::TriggerStatus()
{
    bool master_state = MasterTriggerAsserted;
    bool sw_state = SoftwareTriggerAsserted;
    std::string master_str = (master_state) ? "+" : "-" ;
    std::string sw_str = (sw_state) ? "+" : "-" ;

    std::stringstream msg;
    msg << std::hex << "M: " << master_str
        << "  SW: " << sw_str;
    return msg.str();
}

//---------------------------------------------------------------------------
//  ApplicationIo::GpsDateTime() --  Display current Gps Date/Time
//---------------------------------------------------------------------------

std::string ApplicationIo::GpsDateTime()
{
    std::time_t  Time = Board.GpsCtrl().Gps().Time();
    char * text =  std::ctime(&Time);
    std::string  TimeText( text );
    return Trim(TimeText);
}

//---------------------------------------------------------------------------
//  ApplicationIo::GpsSatellites() --  Display current Satellites
//---------------------------------------------------------------------------

unsigned int ApplicationIo::GpsSatellites()
{
    return Board.GpsCtrl().Gps().Satellites();
}

//---------------------------------------------------------------------------
//  ApplicationIo::GpsFix() --  Display current Fix Status
//---------------------------------------------------------------------------

std::string ApplicationIo::GpsFix()
{
    IGps::FixStatus fix = Board.GpsCtrl().Gps().Fix();
    switch (fix)
        {
            default:
            case IGps::fixNoFix:     return "NF";
            case IGps::fixGps:       return "GPS";
            case IGps::fixDiffGps:   return "DGPS";
        }
}

//---------------------------------------------------------------------------
//  ApplicationIo::GpsLongitude() --  Display current Fix Status
//---------------------------------------------------------------------------

std::string ApplicationIo::GpsLongitude()
{
    GpsCoordinate Long = Board.GpsCtrl().Gps().Longitude();

    std::stringstream txt;

    txt << Long.Degrees << "\u00b0" << Long.Minutes << "  " << Long.Units;

    return Trim(txt.str());
}

//---------------------------------------------------------------------------
//  ApplicationIo::GpsLatitude() --  Display current Fix Status
//---------------------------------------------------------------------------

std::string ApplicationIo::GpsLatitude()
{
    GpsCoordinate Lat = Board.GpsCtrl().Gps().Latitude();

    std::stringstream txt;


    txt << Lat.Degrees << "\u00b0" << Lat.Minutes << "  " << Lat.Units;

    return Trim(txt.str());
}



//==============================================================================
//  CLASS ApplicationSettings
//==============================================================================

//------------------------------------------------------------------------
//  ApplicationSettings::ApplicationSettings() --  Ctor, load from INI file
//------------------------------------------------------------------------

ApplicationSettings::ApplicationSettings(const std::string &file)
    :  IniSaver(file), Xmc0Direction(4, false),    Xmc1Direction(4, false)
{

    //  Target
    Install( ToIni("Target",                  Target,     0)  );

    //  Clock
    Install( ToIni("SampleClockSource",       SampleClockSource,     1u)  );
    Install( ToIni("ReferenceClockSource",    ReferenceClockSource,  1u)  );
    Install( ToIni("OutputFrequency",         OutputFrequency,       MaxOutRateMHz())  );
    Install( ToIni("ReferenceRate",           ReferenceRate,         10.0)  );
    Install( ToIni("ExtClkSrc",                 ExternalClockSource,   0)  );
    Install( ToIni("LowLatencyTrigger",       LowLatencyTrigger,     false)  );



    Install( ToIni("WhichGps",                  WhichGps,   0)  );


    //  ..Trigger
    Install( ToIni("PpsSource",              Trig.PpsSource, 0)  );


    Install( ToIni("EdgeTrigger",            Trig.EdgeTrigger,         0)  );
    Install( ToIni("Framed",                 Trig.Framed,              0)  );
    Install( ToIni("FrameSize",              Trig.FrameSize,           0x4000)  );
    Install( ToIni("TriggerSource",          Trig.TriggerSource         , 0)  );
    Install( ToIni("DecimationEnable",       Trig.DecimationEnable,    false)  );
    Install( ToIni("DecimationFactor",       Trig.DecimationFactor,    1)  );
    Install( ToIni("TriggerType",            Trig.TriggerType         , 0)  );
    Install( ToIni("TriggerInSec",           Trig.TriggerInSec         , 1)  );

    Install( ToIni("Pulse.Enable",              Trig.Pulse.Enable,              false)  );
    Install( ToIni("Pulse.Period",              Trig.Pulse.Period,            10000000u)  );
    Install( ToIni("Pulse.FiniteFrames",        Trig.Pulse.FiniteFrames,        false)  );
    Install( ToIni("Pulse.FiniteFrameRearm",    Trig.Pulse.FiniteFrameRearm,    false)  );
    Install( ToIni("Pulse.FiniteFrameCount",    Trig.Pulse.FiniteFrameCount,    100u)  );
    Install( ToIni("Pulse.PulsePatternINIFile", Trig.Pulse.PulsePatternINIFile,   std::string("PulsePattern.ini"))  );

    // Mezzanine
    Install( ToIni("Xmc0Enable",                    Xmc0Enabled,          false)  );
    Install( ToIni("Xmc0Voltage",                   Xmc0Voltage,          0)  );
    Install( ToIni("Xmc0Direction", "Direction",    Xmc0Direction,       char(0) ));
    Install( ToIni("Xmc1Enable",                    Xmc1Enabled,         false)  );
    Install( ToIni("Xmc1Voltage",                   Xmc1Voltage,          0)  );
    Install( ToIni("Xmc1Direction", "Direction",    Xmc1Direction,        char(0) ));
    Install( ToIni("OutputEnable",                  MezzOutputEnabled,     false)  );
    Install( ToIni("IoVoltage",                     MezzIoVoltage,          0)  );

    GpsStat.EpochTally = 0;

    PllTuningOverrideEnable = false;
    PllTuningOverride.CLKin0_R = 0; PllTuningOverride.CLKin1_R = 0;
    PllTuningOverride.CLKin2_R = 0; PllTuningOverride.N1 = 0;
    PllTuningOverride.R2 = 0;       PllTuningOverride.N2 = 0;
    PllTuningOverride.N2_Cal = 0;   PllTuningOverride.N2_Pre = 0;
    PllTuningOverride.OD = 0;       PllTuningOverride.Feedback_OD = 0;

    Load();
}

//------------------------------------------------------------------------
//  ApplicationSettings::~ApplicationSettings() --  Dtor, save to INI file
//------------------------------------------------------------------------

ApplicationSettings::~ApplicationSettings()
{
    Save();
}




} } // Innovative::Clock
