#ifndef SBCDUOCLOCKMODULE_GP_H
#define SBCDUOCLOCKMODULE_GP_H

#include <SbcDuo_Mb.h>
#include <Application/FiclIo_App.h>
#include <Application/IniSaver_App.h>


#include <SoftwareTimer_Mb.h>
#include <Gps_Mb.h>

namespace Innovative   {
namespace SbcDuo_Clock {



enum IIStatusType { sStatus, sEpoch, sPll, sDio,                                // Misc
    sUtc, sLatitude, sLongitude, sQuality, sSatellites, sLock,  // Gps
    sLast };

//===========================================================================
// CLASS StatusMessageEvent  --  In process status reporting
//===========================================================================

class StatusMessageEvent : public OpenWire::Event
{
public:
    IIStatusType Type;
    std::string Message;

    StatusMessageEvent(IIStatusType type, const std::string & msg)
    : Type(type), Message(msg)
    {
    }
};

inline double  MaxOutRateMHz()
{  return 3100.0;  }     // In MHz
inline bool   HasFiclSupport()
{  return true; }

//==============================================================================
//  CLASS ApplicationSettings
//==============================================================================

class ApplicationSettings : public Innovative::IniSaver
{
public:
    // Ctor
    ApplicationSettings(const std::string &file);
    ~ApplicationSettings();

    typedef std::vector<float> FloatArray;
    typedef std::vector<int>   IntArray;
    typedef std::vector<char>  BoolArray;

    // Data
    std::string     ModuleName;
    std::string     ModuleRevision;

    double          ReferenceRate;
    double			OutputFrequency;
    unsigned int    SampleClockSource;
    unsigned int    ReferenceClockSource;
    int               ExternalClockSource;
    bool            LowLatencyTrigger;

    int             Target;



    std::string     DebugScript;

    //  Digital I/O
    int				DioConfig;
    int				DioDataLow;
    int				DioDataHigh;

    //  Mezzanine
    bool           Xmc0Enabled;
    bool           Xmc1Enabled;
    BoolArray      Xmc0Direction;
    BoolArray      Xmc1Direction;
    int            Xmc0Voltage;
    int            Xmc1Voltage;
    int            MezzIoVoltage;
    bool           MezzOutputEnabled;

    //  Gps Status Data
    struct GpsD
    {
        unsigned int  EpochTally;
        std::string   UtcTime;
        std::string   Latitude;
        std::string   Longitude;
        std::string   Quality;
        std::string   Satellites;
        std::string   Lock;
    };

    GpsD    GpsStat;
    bool    DebugGps;
    int     WhichGps;

    // Pll Tuning Event
    struct TuningEvent
    {
        unsigned int CLKin0_R;
        unsigned int CLKin1_R;
        unsigned int CLKin2_R;
        unsigned int N1;
        unsigned int R2;
        unsigned int N2;
        unsigned int N2_Cal;
        unsigned int N2_Pre;
        unsigned int OD;
        unsigned int Feedback_OD;
    };

    TuningEvent PllTuningOverride;
    bool        PllTuningOverrideEnable;


    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  CLASS PulseD - PRI pulse trigger settings
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    struct PulseD
    {
        bool          Enable;
        bool          FiniteFrames;
        bool          FiniteFrameRearm;
        unsigned int  Period;
        unsigned int  FiniteFrameCount;
        std::string   PulsePatternINIFile;

        std::vector<unsigned int>   Delays;
        std::vector<unsigned int>   Widths;

    };

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //  STRUCT TriggerSettings
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    struct TriggerSettings
    {
        TriggerSettings()
        {}

        //  Trigger
        //  int             ExternalTrigger;
        int             TriggerSource;
        int             EdgeTrigger;
        int             Framed;
        int             FrameSize;
        PulseD          Pulse;
        bool            DecimationEnable;
        int             DecimationFactor;
        int             TriggerType;
        int             PpsSource;
        int             TriggerInSec;
    };

    TriggerSettings   Trig;



private:

};

//==============================================================================
//  CLASS IUserInterface
//==============================================================================

class IUserInterface : public IFicl_UserInterface
{
public:
    virtual void PeriodicStatus() = 0;
    virtual void Log(const std::string & msg) = 0;
    virtual void GetSettings() = 0;
};

//==============================================================================
//  CLASS ApplicationIo
//==============================================================================

class ApplicationIo  : public FiclIo
{
public:
    // Ctor
    ApplicationIo(IUserInterface * gui);
    ~ApplicationIo();

    // Methods
    int             BoardCount();
    bool            Open();
    void            Close();
    bool            Opened() const
                        {  return FOpened;  }

    int             Peek(int addr)
                        {  return Board.PeekPort(addr);  }
    void            Poke(int addr, int data)
                        {  Board.PokePort(addr, data);  }
    unsigned int    Peek(const string &/*type*/, int offset)
                    { return Peek(offset); }
    void            Poke(const string &/*type*/, int offset, unsigned int value)
                    { Poke(offset, value); }
    bool            Start();
    void            Stop();

    void            SetDio();
    void            SetClock();
    void            SetTrig();
    void            SetMezz();

    void            SoftwareTrigger(bool state);
    void            MasterTrigger(bool state);

    void            PpsTriggerArm();

    void  SetPulseTriggerConfiguration(bool enable, bool finite_frames, bool finite_frame_rearm,
                                       unsigned int period, unsigned int finite_frame_count,
                                       std::vector<unsigned int> & delays,
                                       std::vector<unsigned int> & widths );
    void  ReadPriPatternFile(const std::string & file);


    // Gps Stuff
    //void  GpsTriggerArm();
    void  GpsEnable(bool state);
    void  GpsCommand(const std::string & message);

    //  IEEE Uart
    void  UartEnable(unsigned int baud, bool state);
    void  UartCommand(const std::string & cmd);

    ii64 IeeeClockId();
    bool IeeePllLocked();
    bool IeeeInSync();
    bool IeeeLinked();
    bool IeeeGigabit();

    void IeeeInit();



    // Status Readback
    int   Temperature();
    bool  PllLocked();

    // Idrom Programming
    void            WriteRom();
    void            ReadRom();

    // Debug
    void  ClockInfo();

    // Exposed data
    ApplicationSettings     Settings;

    Innovative::SbcDuo &    Duo()
                                {  return Board;  }

    // Module status
    std::string  ModuleName() const;
    std::string  ModuleRevision();
    std::string  TriggerStatus();
    std::string  PllLockedStr();
    std::string  GpsDateTime();
    unsigned int GpsSatellites();
    std::string  GpsFix();
    std::string  GpsLongitude();
    std::string  GpsLatitude();

protected:
    // Fields
    bool                            FOpened;

    // Data
    Innovative::SbcDuo              Board;
    IUserInterface *                UI;

    Innovative::SoftwareTimer		Timer;
    Innovative::SoftwareTimer		GpsInitTimer;
    unsigned int                    GpsInitTally;
    unsigned int 					Lost;
    Innovative::SoftwareTimer       UartTimer;

    Innovative::StopWatch           SW;
    bool                            MasterTriggerAsserted;
    bool                            SoftwareTriggerAsserted;

    // Methods
    void  Log(const std::string & msg);
    void  Status(IIStatusType type, const std::string & msg);
    OpenWire::ThunkedEventHandler<Innovative::ProcessStatusEvent>  OnLog;
    OpenWire::ThunkedEventHandler<StatusMessageEvent>  OnStatus;

    void HandleTimer(OpenWire::NotifyEvent & event);
    void HandleGpsInitTimer(OpenWire::NotifyEvent & event);
    void DisplayLogicVersion();
    void DumpGpsMessage(Innovative::GpsMessage & msg);

    void  HandleOnLog(Innovative::ProcessStatusEvent & Event);
    void  HandleOnEpoch(Innovative::GpsEpochEvent & Event);
    void  HandlePllTuning(Innovative::Lmk04820PllTuningEvent & Event);
    void  HandleUartTimer(OpenWire::NotifyEvent & Event);

    bool  ValidateSampleRate(double should_be, double was, double allowed_range);

    Innovative::IMulticharUart & UartForIdx(unsigned int port_idx);
};






} } // Innovative::Clock

#endif // SBCDUOCLOCKMODULE_GP_H
