#ifndef SUPPORTPROXY_GP_H
#define SUPPORTPROXY_GP_H

#include <CommonRegisters_Gp.h>
#include <CommonModule_Gp.h>
#include <string>
#include <Application/IniSaver_App.h>

using namespace Innovative;
using namespace std;

//===========================================================================
//  CLASS IPowerMeter --
//===========================================================================
class IPowerMeter
{
public:
    virtual void  AccumulatePoints(unsigned int power_of_2) = 0;
    virtual void  Source(unsigned int source) = 0;
    virtual void Reset(bool state) = 0;
    virtual void CommandTrigger() = 0;
    virtual void Init() = 0;

    virtual double  PowerData_dBm(unsigned int source) = 0;
};

//===========================================================================
//  CLASS PowerMeterWrapper --
//===========================================================================
class PowerMeterWrapper : public IPowerMeter
{
public:
    PowerMeterWrapper(DspPowerMeterSM *apm, string atag, unsigned int acore);
    virtual ~PowerMeterWrapper();


    virtual void  AccumulatePoints(unsigned int power_of_2);
    virtual void  Source(unsigned int source);
    virtual void Reset(bool state);
    virtual void CommandTrigger();
    virtual void Init();

    virtual double  PowerData_dBm(unsigned int source);
    string  Tag() const;
    void WhichBoard(unsigned int ordinal);
    unsigned int WhichBoard();
private:
    DspPowerMeterSM * const pm;
    string                  tag;
    const unsigned int      core;
    unsigned int            BoardOrdinal;
private:
    PowerMeterWrapper(const PowerMeterWrapper &other);
    PowerMeterWrapper &operator=(const PowerMeterWrapper &other);
};

//===========================================================================
//  CLASS PowerMeterProxy --
//===========================================================================
class PowerMeterProxy
{
public:
    typedef vector<PowerMeterWrapper *> PowerMeterGroupType;

    PowerMeterProxy();
    virtual ~PowerMeterProxy();

    void Clear();

    PowerMeterWrapper &Proxy(unsigned int index);

    void Append(PowerMeterWrapper *value);
    void PowerMeterList(vector<string> &list);
    PowerMeterGroupType PowerMeters(unsigned int board);

    void Reset(bool state);
    void CommandTrigger();
    void Init();

private:
    PowerMeterGroupType group;
private:
    PowerMeterProxy(const PowerMeterProxy &other);
    PowerMeterProxy &operator=(const PowerMeterProxy &other);
};

//===========================================================================
//  CLASS IOverflow --
//===========================================================================
class IOverflow
{
public:
    virtual unsigned int OverflowStatus() = 0;
    virtual void Reset(bool state) = 0;
    virtual void CommandTrigger() = 0;
};

//===========================================================================
//  CLASS OverflowWrapper --
//===========================================================================
class OverflowWrapper : public IOverflow
{
public:
    OverflowWrapper(DspOverflowSM *aov, string atag, unsigned int acore);
    virtual ~OverflowWrapper();

    enum OverflowBits {
                       tuner_ovflo   = 0,
                       channelizer_ovflo,
                       ppf_ovflo        ,
                       cic_ovflo        ,
                       cfir_ovflo       ,
                       pfir_ovflo       ,
                       fir1_ovflo       ,
                       fft_ovflo
                      };

    virtual unsigned int OverflowStatus();
    virtual void Reset(bool state);
    virtual void CommandTrigger();
    string  Tag() const;

    void WhichBoard(unsigned int ordinal);
    unsigned int WhichBoard();
private:
    DspOverflowSM   * const ov;
    string                  tag;
    const unsigned int      core;

    unsigned int    BoardOrdinal;
private:
    OverflowWrapper(const OverflowWrapper &other);
    OverflowWrapper &operator=(const OverflowWrapper &other);
};

//===========================================================================
//  CLASS OverflowProxy --
//===========================================================================
class OverflowProxy
{
public:
    typedef vector<OverflowWrapper *> OverflowGroupType;

    OverflowProxy();
    virtual ~OverflowProxy();

    void Clear();

    OverflowWrapper &Proxy(unsigned int index);

    void Append(OverflowWrapper *value);
    void OverflowDetectorList(vector<string> &list);
    void OverflowDetectorList(vector<string> &list, unsigned int board);
    OverflowGroupType OverflowDetectors(unsigned int board);

    size_t Count() const;
    size_t Count(unsigned int board) const;
    void Reset(bool state);
    void CommandTrigger();

private:
    OverflowGroupType group;
private:
    OverflowProxy(const OverflowProxy &other);
    OverflowProxy &operator=(const OverflowProxy &other);
};


struct ToneGeneratorDataType
{
	ToneGeneratorDataType() 
	{}

	ToneGeneratorDataType(float f, float p, float g) :
	Freq(f), Phase(p), Gain(g) 
	{}

    float Freq;
    float Phase;
    float Gain;

    bool operator==(const ToneGeneratorDataType &other)
    {
        return
           other.Freq       == Freq   &&
           other.Phase      == Phase  &&
           other.Gain       == Gain;
    }

};


//===========================================================================
//  CLASS IToneGenerator --
//===========================================================================
class IToneGenerator
{
public:
    virtual void Frequency1(double freq, double srate[2], double /*fdiv*/) = 0;
    virtual void Phase1(double value) = 0;
    virtual void Gain1(double value) = 0;
    virtual void Frequency2(double freq, double srate) = 0;
    virtual void Phase2(double value) = 0;
    virtual void Gain2(double value) = 0;
    virtual void Reset(bool state) = 0;
    virtual void CommandTrigger() = 0;

    ToneGeneratorDataType tones[2];
};


//===========================================================================
//  CLASS ToneGeneratorWrapper --
//===========================================================================
class ToneGeneratorWrapper : public IToneGenerator
{
public:
    ToneGeneratorWrapper(DspToneGeneratorSM *atone, string atag, unsigned int acore, const ToneGeneratorDataType &tonedef);
    virtual ~ToneGeneratorWrapper();

    virtual void Frequency1(double freq, double srate[2], double fdiv);
    virtual void Phase1(double value);
    virtual void Gain1(double value);
    virtual void Frequency2(double value, double srate);
    virtual void Phase2(double value);
    virtual void Gain2(double value);
    virtual void Reset(bool state);
    virtual void CommandTrigger();
    void AttachSettings(Innovative::IniSaver& saver);

private:
    DspToneGeneratorSM   * const tone;
    string                       tag;
    const unsigned int           core;
private:
    ToneGeneratorWrapper(const ToneGeneratorWrapper &other);
    ToneGeneratorWrapper &operator=(const ToneGeneratorWrapper &other);

protected:
    void Set_Frequency1(double freq, double srate);  // internal setter
};

//===========================================================================
//  CLASS DucToneGeneratorWrapper --
//===========================================================================

class DucToneGeneratorWrapper : public ToneGeneratorWrapper
{
public:

	DucToneGeneratorWrapper(DspToneGeneratorSM *atone, string atag, unsigned int acore, const ToneGeneratorDataType &tonedef):
		ToneGeneratorWrapper(atone, atag, acore, tonedef)
		{}

    virtual void Frequency1(double freq, double srate[2], double fdiv)
	{
        // Use the DAC frequency here
        Set_Frequency1(freq, srate[1] / fdiv);
	}
};


//===========================================================================
//  CLASS ToneGeneratorProxy --
//===========================================================================
class ToneGeneratorProxy
{
public:
    ToneGeneratorProxy();
    virtual ~ToneGeneratorProxy();

    void Clear();

    size_t Size();
    void Append(ToneGeneratorWrapper *value);
    void Reset(bool state);
    void CommandTrigger();
	void CommandTrigger(unsigned int n);
    void AttachSettings(Innovative::IniSaver& saver);
    void ApplyCached();
	void ApplyCached(unsigned int n);
    void SamplingRate(double adc_rate, double dac_rate);
    void InterpolationRate(double value);

    ToneGeneratorWrapper &Proxy(unsigned int index);

private:
    vector<ToneGeneratorWrapper *> group;
    double                         SRate[2]; // 0=Adc Frequency, 1=Dac Frequency
	double                         IpolRate;

private:
    ToneGeneratorProxy(const ToneGeneratorProxy &other);
    ToneGeneratorProxy &operator=(const ToneGeneratorProxy &other);
};


#endif // SUPPORTPROXY_GP_H
